// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2007, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.eaacs;

//TODO: GetPathAndName
//TODO: numberOfQuarters
//TODO: Begin- EndQuarter
//TODO: in_out_indicator
//TODO: application
//TODO: database
//TODO: addConfigLine();

import java.sql.*;
import java.util.*;
import java.text.*;
import java.io.*;

/* ***************************** Class ****************************** */

/**
 * Export utility providing the methods to be called as
 * Hyperion/essbase user-defined functions.
 * This way, cube data can be extracted from Hyperion.
 *
 * @author
 * <a href="mailto:ana.padrao@ecb.europa.eu?subject=GESMES/TS-Suite">Ana Padrao</a>
 *
 * @version @(#)$Id: EaacsExport.java,v 1.1 2017/01/19 16:14:05 trainot Exp $
 */
public class EaacsExport {

	/* ===================== Static Fields ====================== */

	/** Essbase uses this value to indicate empty cells. */
	public static final double HYPERION_MISSING = -2.0E-15;

	static Locale currentLocale = new Locale("en","US");
	static NumberFormat nf = NumberFormat.getNumberInstance(currentLocale);
	static DecimalFormat df = (DecimalFormat)nf;

	static Map fileMap = new HashMap();
	/* 
	 * this counter should be linked to a file.  But as the method memberIndex
	 * does not provide any way to know the file, Let's IMPLICITELY link this
	 * to the last file we opened.  YUCK!!! 
	 */
	static int callsToMemberIndex;
	/* 
	 * Same remark (YUCK!!!) plus:
	 * We expect each call to writeToFile to add a line.
	 * So numberOfLine ~ callsTo writeToFile().  
	 */
	static int numberOfLines;
	/* 
	 * Same remark (YUCKY!!!)
	 */
	static String fullFilePath;
	
	static String inOutFlag;

	/* ==================Public Static Methods ================== */

	/** Report data via the named EaacsExport.
	 * This method is typically called in the center of a FIX loop
	 * of a CALC-script to export a subcube.
	 * It received the cell coordinates and the associated value.
	 *
	 * @param channelName uniquely identifies the channel.
	 * @param filename identifies the file where to add data to.
	 * @param key of the time series (ordered according to the key family).
	 * @param year of the observation (prefixed with "Y", that is "Ynnnn").
	 * @param period of the observation inside the year.
	 * @param value of the observation.
	 */
	public static void main ( 
		String args[]
	) {
		System.out.println("EaacsExport.main===>");
	}

	public static int memberIndex (
		String member, 
		String[] listMembers
	) {
		int mIndex = 0;
		
		for (int i = 0; i < listMembers.length; i++) {
			if ( member.equals(listMembers[i]) ) {
				mIndex = i + 1;
			}
		}
		++callsToMemberIndex;
		return mIndex;
	}

    
	/* 
	 *    Export to a text file
	 *   arg 2:  file name.  This file name must be used to close the file 
	 *           after the calculation completes.
	 *   arg 6:  an array of member names (Use "" if you are not exporting 
	 *           members names).
	 *   arg 7:  an array of data (Use 0 if you are not exporting Data values)
	 *
	 */
	public static void exportToFile (
		String fileName, 
		String inOut,
		String[] strArgs, 
		String strDelimiter,
		int[] matrixArgs,
		double[] dblArgs,
		/* no alternative file name, we write EITHER in OR out, exclusively.*/
		//String outFileName, 
		String outSuffix
	) {
		System.out.println(
			"EaacsExport.exportToFile===> filename: "
			+ fileName + " (" + inOut + ")"
		);
		openFile( fileName );
		
		inOutFlag = inOut.toUpperCase(); 
		
		System.out.println( 
			"EaacsExport.exportToFile===> inout : " + "IN".equals( inOutFlag )
		);
	
		insertToFile(
			fileName,
			strArgs,
			strDelimiter,
			matrixArgs,
			dblArgs,
			"IN".equals( inOutFlag ),
			outSuffix
		);
	}

	public static void closeFiles( 
		String fileName,
		String confg,    // configuration file to be amended
		String range,    // range of periods
		String appli,    // hyperion application
		String dbase     // hyperion database
	) {
		System.out.println( "EaacsExport.closeFiles---> " + fileName );
		try {
			if ( fileMap.containsKey(fileName) ) {
				System.out.println(
					"EaacsExport.closeFiles---> fileMap contains file: "
					+ fileName
				);
				FileWriter thisOut = (FileWriter) fileMap.get(fileName);
				thisOut.close() ;
				thisOut = null;
				fileMap.remove(fileName);
				System.out.println( fileName + ": file Closed");
			} else {
				System.out.println(
					"EaacsExport.closeFiles---> "
					+ "fileMap DOES NOT contain file: " + fileName
				);
			}
		} catch (Exception e) {
			System.out.println( "Can not close file: " + e.getMessage() );
		}
		updateConfigFile( confg, range, appli, dbase );
	}
	
	public static String getFileName() {
		return (new File(fullFilePath)).getName();
	}
	public static String getFilePath() {
		// stupid java.io.file.getPath() yields path with file basename!!!
		return fullFilePath.substring( 0, fullFilePath.lastIndexOf("/") ) + "/";
	}
	public static int getNumberOfLines() {
		return numberOfLines;
	}
	public static int getNumberOfIndices() {
		return callsToMemberIndex / getNumberOfLines() ;
	}
	public static int getNumberOfPeriods( String range ) {
		// these two without leading 'Y'.
		String beginPeriod = getBeginPeriod( range );
		String endPeriod = getEndPeriod( range );
		StringTokenizer beginTok = new StringTokenizer( beginPeriod, "Q" );
		StringTokenizer endTok = new StringTokenizer( endPeriod, "Q" );
		
		int beginY = Integer.valueOf( beginTok.nextToken() ).intValue();
		int endY = Integer.valueOf( endTok.nextToken() ).intValue();
		
		if ( beginTok.hasMoreTokens() ) {
			int beginQ = Integer.valueOf( beginTok.nextToken() ).intValue();
			int endQ = Integer.valueOf( endTok.nextToken() ).intValue();
			return ( endY * 4 + endQ ) - ( beginY * 4 + beginQ ) + 1 ;
		} else {
			return endY - beginY + 1 ;
		}
	}
	public static String getBeginPeriod( String range ) {
		StringTokenizer periodTok = new StringTokenizer( range, ":" );
		return periodTok.nextToken().substring(1);
	}	
	public static String getEndPeriod( String range ) {
		StringTokenizer periodTok = new StringTokenizer( range, ":" );
		periodTok.nextToken(); // ignore the first period
		return periodTok.nextToken().substring(1);
	}	
	public static String getInOutFlag() {
		return inOutFlag;
	}	
	public static void updateConfigFile(
		String configFile,
		String periodRange,
		String application,
		String database
	) {
		System.out.println( "EaacsExport.updateConfigFile--->starting" );
		StringBuffer insertString = new StringBuffer();
		insertString.append( getFileName() + "," );
		insertString.append( getFilePath() + "," );
		insertString.append( getNumberOfLines() + "," );
		insertString.append( getNumberOfIndices() + "," );
		insertString.append( getNumberOfPeriods(periodRange) + "," );
		insertString.append( getBeginPeriod(periodRange) + "," );
		insertString.append( getEndPeriod(periodRange) + "," );
		insertString.append( getInOutFlag() + "," );
		insertString.append( application + "," );
		insertString.append( database + "\n" );
		System.out.println( "EaacsExport.updateConfigFile--->"
			+ insertString.toString()
		);
		// Try to write in a tmp file.
		// On success, we replaced the original file.
		File orgConfigFile = new File( configFile );
		File tmpConfigFile = new File( configFile + ".tmp" );
		try {	
			System.out.println( "EaacsExport.updateConfigFile--->reading "
				+ configFile
			);
			FileWriter fw = new FileWriter( tmpConfigFile );
			// if an original file already exist, copy the contents into
			// a tmp file, but remove the lines that are like our insertString.
			if ( orgConfigFile.exists() ) {
				FileReader fr = new FileReader( orgConfigFile );
				BufferedReader br = new BufferedReader( fr ); 
				String s;
				while((s = br.readLine()) != null) { 
					StringTokenizer strtok = new StringTokenizer( s, "," );
					String f = strtok.nextToken();
					String p = strtok.nextToken();
					if ( getFileName().equals(f) && getFilePath().equals(p) )
						; // ignore this line, we will replace it
					else 
						fw.write( s + "\n" );
				} 
				fr.close();
			}
			//
			fw.write( insertString.toString() );
			fw.close();
			System.out.println( "EaacsExport.updateConfigFile--->new written" );
		} catch (Throwable t) {
			System.out.println( "EaacsExport.updateConfigFile--->problem on writing config file " + t.getMessage()
			);
		}
		System.out.println( "EaacsExport.updateConfigFile--->"
			+ "Renaming");
		tmpConfigFile.renameTo( orgConfigFile );
		// tmpConfigFile.delete();
	}
   
	/* ==================Private Static Methods ================= */
   
	private static void openFile (
		String fileName
	) {
		System.out.println("EaacsExport.openFile===> " + fileName);
		if ( !fileMap.containsKey( fileName ) ) {
			System.out.println(
				"EaacsExport.openFile---> fileMap does not contain Key "
				+ fileName
			);
			try {
				fileMap.put( fileName, new FileWriter( new File( fileName ) ) );
				callsToMemberIndex = 0; 
				numberOfLines = 0; 
				fullFilePath = fileName;
				System.out.println( fileName + ": file opened." );
			} catch (Throwable t) {
				System.out.println(
					fileName + ": Can't open file " + t.getMessage()
				);
			}
		} else {
			System.out.println(
				"EaacsExport.openFile---> fileMap contains Key " + fileName
			);
		}
	}

	private static void insertToFile (
		String fileName, 
  		String[] strArgs, 
		String strDelimiter,
		int[] matrixArgs,
		double[] dblArgs,
		boolean writingAnInFile,
		String outSuffix 
	) {
		System.out.println("EaacsExport.insertToFile===> " + fileName);

		StringBuffer insertString = new StringBuffer();
		String numToFormat = null;
		boolean hasString = true;
		boolean hasMatrix = true;
		boolean hasDouble = true;
		if ( strArgs.length == 1 && strArgs[0].trim().length() == 0 ) {
			hasString = false;
		}
		if ( matrixArgs.length == 1 && matrixArgs[0] == 0 ) {
			hasMatrix = false;
		}
		if ( dblArgs.length == 1 && dblArgs[0] == 0 ) {
			hasDouble = false;
		}
		try {
			if ( hasString ) { // For serieskey the delimiter is "."
				for (int i = 0; i < strArgs.length-1; i++) { 
					// Removes 3 first characters, prefix of each member
					insertString.append( strArgs[i] + strDelimiter );
				}
				if ( writingAnInFile )
					insertString.append( strArgs[strArgs.length -1] );
				else
					insertString.append( outSuffix );
         }
			
		if ( hasMatrix ) { // For matrix the delimiter is ","
            insertString.append( "," );
            for (int i = 0; i < matrixArgs.length-1; i++) {
  	            insertString.append(
  	            	String.valueOf( matrixArgs[i] ) + ","
  	            );
			}
            insertString.append(
            	String.valueOf( matrixArgs[matrixArgs.length-1] )
            );
		}

		if ( hasDouble ) { // For values the delimiter is ","
            insertString.append( "," );
            for (int i = 0; i < dblArgs.length; i++) {
            	if (! writingAnInFile) 
            		numToFormat = "";
            	else if (
            		com.hyperion.essbase.calculator.Calculator.isMissing(
            			dblArgs[i]
            		)
            	)
					numToFormat = "ND";	// replace missing values by NA
            	else
            		numToFormat = String.valueOf( dblArgs[i] );
            	insertString.append( numToFormat );
            	if ( i != dblArgs.length-1 ) 
            		insertString.append( "," );
            }
			System.out.println(
				"EaacsExport.insertToFile---> hasDouble - insertString: "
				+ insertString.toString()
			);	
		} 
			
		insertString.append( "\n" );				
						
        if ( hasString || hasMatrix || hasDouble)
        	writeToFile( fileName, insertString.toString() );
		} catch(Exception e) {
			System.out.println(e.getMessage());
		} finally {
			insertString = null;
			numToFormat = null;
		}
	}

	private static void writeToFile (
		String fileName, 
		String output
	) {
		System.out.println(
			"EaacsExport.writeToFile---> " + fileName + " output " + output
		);
		FileWriter thisOut = null;
		try {
			thisOut = (FileWriter) fileMap.get( fileName );
			thisOut.write( output ) ;
			++numberOfLines;
		} catch( Exception e ) {
			System.out.println( "Can not write to the file: " + e.getMessage() );
		} finally {
			thisOut = null;
		}
	}

}

