// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2002, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb;

import java.text.ParseException;
import java.text.FieldPosition;

import Int.ecb.logging.Level;
import Int.ecb.logging.Logger;

import Int.ecb.prefs.Preferences;

import Int.ecb.gesmescb.time.RegularInterval;
import Int.ecb.gesmescb.util.PackageProperties;
import Int.ecb.gesmescb.edifact.RegularIntervalFormat;

/* ***************************** Class ****************************** */

/**
 * Implements common behaviour of Gesmes/TS data objects.
 *
 * @author
 * <A HREF="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</A>
 *
 * @version @(#)$Id: AbstractCube.java,v 1.1 2017/01/19 16:14:04 trainot Exp $
 */

public abstract class AbstractCube extends GESMESCBObject {

	/* ======================= Constants ======================== */

	/**
	 * The separator character used in the key. This should always be a dot.
	 */
	public static final char KEY_SEPARATOR;

	/* =================== Static Initializer =================== */

	static {
		final String packageName = PackageProperties.getName(AbstractCube.class.getName());
		final Logger logger = Logger.getLogger( packageName );
		final Preferences
			userPrefs = Preferences.userRoot().node( packageName ),
			systemPrefs = Preferences.systemRoot().node( packageName );
		final String keySeparator
		= systemPrefs.get(
			"keySeparator",
			userPrefs.get( "keySeparator", "." )
		);
		if ( keySeparator.length() == 1 ) {
			logger.log(
				Level.FINE,
				"AbstractCube",
				"static initialiser",
				"Using '" + keySeparator + "' as series key separator."
			);
			KEY_SEPARATOR = keySeparator.charAt( 0 );
		} else {
			logger.log(
				Level.SEVERE,
				"AbstractCube",
				"static initialiser",
				"Configuration problem: Key separator \""
				+ keySeparator
				+ "\" should be exactly one character."
			);
			KEY_SEPARATOR = '.';
		}
	}

	/* ========================= Fields ========================= */

	/**
	 * Key the AbstractCube relates to
	 * in {@link #KEY_SEPARATOR}-separated notation.
	 *
	 * @serial
	 */
	private final String key;

	/**
	 * Time period value.
	 * The format is according to the field periodFormat.
	 *
	 * @serial
	 */
	private final String period;

	/**
	 * Format of period value.
	 * The meaning according to the GESMES/TS User Guide,
	 * see description of main ARR and 2nd DTM segments.
	 *
	 * The GESMES/TS period format specification:
	 *
	 * <P><CENTER><TABLE BORDER=3 CELLPADDING=3>
	 *
	 * <CAPTION>GESMES/TS period format codes</CAPTION>
	 *
	 * <TR><TD></TD><TH>Dates</TH>
	 * <TD></TD><TD></TD><TH>Date Ranges</TH></TR>
	 *
	 * <TR><TH>203</TH><TD>CCYYMMDDhhmm</TD>
	 * <TD></TD><TH></TH><TD></TD></TR>
	 *
	 * <TR><TH>102</TH><TD>CCYYMMDD</TD>
	 * <TD></TD><TH>711</TH><TD>CCYYMMDDCCYYMMDD</TD></TR>
	 *
	 * <TR><TH>616</TH><TD>CCYYWW</TD>
	 * <TD></TD><TH>716</TH><TD>CCYYWWCCYYWW</TD></TR>
	 *
	 * <TR><TH>610</TH><TD>CCYYMM</TD>
	 * <TD></TD><TH>710</TH><TD>CCYYMMCCYYMM</TD></TR>
	 *
	 * <TR><TH>608</TH><TD>CCYYQ</TD>
	 * <TD></TD><TH>708</TH><TD>CCYYQCCYYQ</TD></TR>
	 *
	 * <TR><TH>604</TH><TD>CCYYS</TD>
	 * <TD></TD><TH>704</TH><TD>CCYYHCCYYH</TD></TR>
	 *
	 * <TR><TH>602</TH><TD>CCYY</TD>
	 * <TD></TD><TH>702</TH><TD>CCYYCCYY</TD></TR>
	 *
	 * </TABLE></CENTER></P>
	 *
	 * @serial
	 */
	private final String periodFormat;

	/* -------------------- Transient Fields -------------------- */

	/**
	 * The key length.
	 */
	private transient int keyLength;

	/**
	 * The time interval.
	 */
	private transient RegularInterval interval;

	/**
	 * Precomputed hash code.
	 */
	private transient int hashCode;

	/* ====================== Constructors ====================== */

	/**
	 * Construct a new AbstractCube.
	 *
	 * @param key which the AbstractCube relates to.
	 * The components must be separated by {@link #KEY_SEPARATOR}.
	 * @param period time period.
	 * @param periodFormat time period format.
	 *
	 * @throws NullPointerException
	 * if any if the arguments is <CODE>null</CODE>.
	 */
	public AbstractCube (
		final String key,
		final String period,
		final String periodFormat
	) {
		if ( ( this.key = key ) == null ) {
			throw new NullPointerException(
				"Trying to create " + getClass().getName()
				+ " with null key."
			);
		}

		if ( ( this.period = period ) == null ) {
			throw new NullPointerException(
				"Trying to create " + getClass().getName()
				+ " with null period."
			);
		}

		if ( ( this.periodFormat = periodFormat ) == null ) {
			throw new NullPointerException(
				"Trying to create " + getClass().getName()
				+ " with null periodFormat."
			);
		}
	}

	/**
	 * Construct a new AbstractCube.
	 *
	 * @param key which the AbstractCube relates to.
	 * The components must be separated by {@link #KEY_SEPARATOR}.
	 * @param interval time period.
	 *
	 * @throws NullPointerException
	 * if any if the arguments is <CODE>null</CODE>.
	 */
	public AbstractCube (
		final String key,
		final RegularInterval interval
	) {
		if ( ( this.key = key ) == null ) {
			throw new NullPointerException(
				"Trying to create " + getClass().getName()
				+ " with null key."
			);
		}

		if ( ( this.interval = interval ) == null ) {
			period = "";
			periodFormat = "";
		} else {
			final RegularIntervalFormat format = new RegularIntervalFormat();
			final StringBuffer buffer = new StringBuffer();
			final FieldPosition position = new FieldPosition(
				RegularIntervalFormat.COLON_FIELD
			);
			format.format( interval, buffer, position );
			period = buffer.substring( 0, position.getBeginIndex() );
			periodFormat = buffer.substring( position.getEndIndex() );
		}
	}

	/* ===================== Public Methods ===================== */

	/**
	 * Returns the key which the AbstractCube relates to.
	 *
	 * @return the key.
	 */
	public String getKey () {
		return key;
	}

	/**
	 * Computes the key length.
	 * This is the number of {@link #KEY_SEPARATOR} characters
	 * in the key plus 1.
	 *
	 * @return the key length.
	 */
	public int getKeyLength () {
		// If not yet initialized.
		if ( keyLength == 0 ) {
			keyLength = computeKeyLength( key );
		}
		return keyLength;
	}

	/**
	 * Returns time period.
	 *
	 * @return period.
	 */
	public String getPeriod () {
		return period;
	}

	/**
	 * Returns period format.
	 *
	 * @return period format.
	 */
	public String getPeriodFormat () {
		return periodFormat;
	}

	/**
	 * Returns the spanned period.
	 *
	 * @return time interval or null if invalid or not available.
	 */
	public synchronized RegularInterval getInterval () {
		// synchronised for Java VMM compliance.
		if ( interval == null && !"".equals( period ) ) {
			try {
				interval = new RegularIntervalFormat().parse(
					period, periodFormat
				);
			}
			catch ( final ParseException pe ) {}
		}
		return interval;
	}

	/**
	 * Indicates whether some other Object is "equal to" this AbstractCube.
	 *
	 * @param object the reference object with which to compare
	 *
	 * @return true if this AbstractCube is equal to the argument;
	 * false otherwise.
	 */
	public boolean equals ( final Object object ) {
		if ( this == object ) return true;
		if ( ! ( object instanceof AbstractCube ) ) return false;
		final AbstractCube other = (AbstractCube) object;
		return
			key.equals( other.key )
			&& period.equals( other.period )
			&& periodFormat.equals( other.periodFormat );
	}

	/**
	 * Returns the hash code for this AbstractCube.
	 * If a1.equals(a2) then a1.hashCode()==a2.hashCode()
	 * for any two AbstractCube a1 and a2,
	 * as required by the general contract of {@link Object#hashCode}.
	 *
	 * @return hash code for this object.
	 */
	public int hashCode () {
		// if not yet computed (or accidentally zero)
		if ( hashCode == 0 ) {
			hashCode = key.hashCode()
				^ period.hashCode()
				^ periodFormat.hashCode();
		}
		return hashCode;
	}

	/**
	 * Returns a string representation of the AbstractCube
	 * in a human readable format.
	 *
	 * @return a string representation of the AbstractCube.
	 */
	public String toString () {
		return
			getClass().getName()
			+ "( key = " + key
			+ ", period = " + period
			+ ", periodFormat = " + periodFormat
			+ " )";
	}

	/* ==================== Package Methods ===================== */

	/**
	 * Computes the key length.
	 * This is the number of {@link #KEY_SEPARATOR} characters
	 * in the <CODE>keyString</CODE> plus 1.
	 *
	 * @param keyString the key to be considered.
	 *
	 * @return the key length of <CODE>keyString</CODE>.
	 */
	private static int computeKeyLength ( final String keyString ) {
		final int length = keyString.length();
		int keySections = 1;
		for ( int i = 0; i < length; ++i ) {
			if ( keyString.charAt( i ) == KEY_SEPARATOR ) ++keySections;
		}
		return keySections;
	}
}
