// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2002, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb;

import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;

import Int.ecb.gesmescb.time.RegularInterval;
import Int.ecb.gesmescb.time.GregorianRegularInterval;

/* ***************************** Class ****************************** */

/**
 * <P>Tests the implementation of the class {@link AbstractCube}.</P>
 *
 * @author
 * <A HREF="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</A>
 *
 * @version @(#)$Id: AbstractCubeTest.java,v 1.1 2017/01/19 16:14:04 trainot Exp $
 */

public class AbstractCubeTest extends TestCase {

	/* ====================== Constructors ====================== */

	public AbstractCubeTest ( final String name ) {
		super( name );
	}

	/* ===================== Public Methods ===================== */

	/**
	 * Assembles and returns a test suite for all
	 * the test methods of this test case.
	 *
	 * @return a non-null test suite
	 */
	public static Test suite () {
		return new TestSuite( AbstractCubeTest.class );
	}

	/**
	 * Runs the test case.
	 */
	public static void main ( String args[] ) {
		final String[] testCaseName = { AbstractCubeTest.class.getName() };
		junit.textui.TestRunner.main( testCaseName );
	}

	/* ---------------------- Test Methods ---------------------- */

	/** Test the default {@link AbstractCube#KEY_SEPARATOR}. */
	public void testDefaultKeySeparator () {
		assertEquals( '.', AbstractCube.KEY_SEPARATOR );
	}

	/**
	 * Test the two constructors.
	 */
	public void testConstruct () {
		final String
			key = "key",
			period = "200001200012",
			periodFormat = "710";

		final RegularInterval interval = new GregorianRegularInterval(
			RegularInterval.MONTHLY, 2000, 0, 12
		);

		final AbstractCube c1 = new TestAbstractCube(
			key, period, periodFormat
		);
		final AbstractCube c2 = new TestAbstractCube( key, interval );

		assertEquals( "equal objects", c1, c2 );
		assertEquals( "equal hashcodes", c1.hashCode(), c2.hashCode() );
		assertEquals( "equal intervals", c1.getInterval(), c2.getInterval() );
	}

	/**
	 * Test the equals, hashCode and toString methods.
	 */
	public void testEquals () {
		final String[] original = {
			"M.AT.CPI.CH.Z",
			"199504200106",
			"708"
		};
		final String[] modified = {
			"M.AT.CPI.CH",
			"199503200101",
			"716"
		};

		final AbstractCube c0 = new TestAbstractCube(
			original[0],
			original[1],
			original[2]
		);
		final AbstractCube c1 = new TestAbstractCube(
			new String( original[0] ),
			new String( original[1] ),
			new String( original[2] )
		);
		assertEquals( "equal objects", c0, c1 );
		assertEquals( "equal hashcodes", c0.hashCode(), c1.hashCode() );
		assertEquals( "equal strings", c0.toString(), c1.toString() );

		for ( int i = 0; i < original.length; ++i ) {
			final AbstractCube ci = new TestAbstractCube(
				i == 0 ? modified[0] : original[0],
				i == 1 ? modified[1] : original[1],
				i == 2 ? modified[2] : original[2]
			);

			assertTrue( "unequal objects["+i+"]", !ci.equals( c0 ) );
			// Inequality of the hashCode is not really contract.
			// But the probability of equal hashCode is
			// far less than that of a buggy hashCode-implementation.
			assertTrue(
				"unequal hashcodes[" + i + "]",
				ci.hashCode() != c0.hashCode()
			);
			assertTrue(
				"unequal strings[" + i + "]",
				!ci.toString().equals( c0.toString() )
			);
		}
	}

	/**
	 * Test the get-type methods.
	 */
	public void testProperties () {
		final AbstractCube c0 = new TestAbstractCube(
			"BO.GU.S", "200109200110", "716"
		);

		assertEquals( "BO.GU.S", c0.getKey() );
		assertEquals( "200109200110", c0.getPeriod() );
		assertEquals( "716", c0.getPeriodFormat() );

		assertEquals( "key length", 3, c0.getKeyLength() );
		assertEquals(
			new GregorianRegularInterval(
				RegularInterval.WEEKLY, 2001, 8, 2
			),
			c0.getInterval()
		);
	}

	/** Test null checks. */
	public void testNullChecks () {
		final AbstractCube c0 = new TestAbstractCube(
			"B.DI.D", null
		);
		assertEquals( "period", "", c0.getPeriod() );
		assertEquals( "periodFormat", "", c0.getPeriodFormat() );
		assertEquals( "interval", null, c0.getInterval() );
		try {
			//final AbstractCube c1 =
			new TestAbstractCube(
				null, "200208", "616"
			);
		} catch ( final NullPointerException e ) {}
		try {
			//final AbstractCube c2 =
			new TestAbstractCube(
				null,
				new GregorianRegularInterval(
					RegularInterval.WEEKLY, 2001, 8, 2
				)
			);
		} catch ( final NullPointerException e ) {}
	}

	/**
	 * Tests serialization.
	 */
	public void testSerialization () {
		final AbstractCube c0 = new TestAbstractCube(
			"B.DI.D", "200208", "616"
		);
		final RegularInterval interval = new GregorianRegularInterval(
			RegularInterval.WEEKLY, 2002, 7, 1
		);
		final AbstractCube c1
		= (AbstractCube) AttributeTest.transserialize( c0 );
		assertEquals( c0, c1 );
		assertEquals( "key length", 3, c1.getKeyLength() );
		assertEquals( "interval", interval, c1.getInterval() );
	}
}

/** Concrete test class. */
class TestAbstractCube extends AbstractCube {

	/**
	 *   serialVersionUID generated by eclipse.
	 */
	private static final long serialVersionUID = -8298839820471647512L;

	public TestAbstractCube (
		final String key, final String period, final String periodFormat
	) {
		super( key, period, periodFormat );
	}

	public TestAbstractCube (
		final String key, final RegularInterval interval
	) {
		super( key, interval );
	}
}
