// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2002, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb;

import java.util.Date;

/* ***************************** Class ****************************** */

/**
 * Implements common behaviour of Gesmes/TS structure objects.
 *
 * @author
 * <A HREF="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</A>
 *
 * @version @(#)$Id: AbstractStructuralDefinition.java,v 1.1 2017/01/19 16:14:04 trainot Exp $
 */

abstract class AbstractStructuralDefinition extends GESMESCBObject {

	/* ========================= Fields ========================= */

	/**
	 * Identifier, in older parlance also called mnemonic.
	 *
	 * @serial
	 */
	private final String identifier;

	/**
	 * Name.
	 *
	 * @serial
	 */
	private final String name;

	/**
	 * Description.
	 *
	 * @serial
	 */
	private final String description;

	/**
	 * Version timestamp.
	 *
	 * @serial
	 */
	private final Date versionDate;

	/* -------------------- Transient Fields -------------------- */

	/**
	 * Precomputed hash code.
	 */
	private transient int hashCode;

	/* ====================== Constructors ====================== */

	/**
	 * Construct a new AbstractStructuralDefinition object.
	 *
	 * @param description human readable description.
	 * Note that descriptions are not
	 * supported by Gesmes/TS-EDIFACT.
	 * @param versionDate to be used as version timestamp.
	 * It should be computed from the {@link MessageBegin}'s
	 * extractionDateTime.
	 */
	public AbstractStructuralDefinition (
		String identifier,
		String name,
		String description,
		Date versionDate
	) {
		if ( ( this.identifier = identifier ) == null ) {
			throw new NullPointerException(
				"Trying to create " + getClass().getName()
				+ " with null identifier."
			);
		}

		if ( ( this.name = name ) == null ) {
			throw new NullPointerException(
				"Trying to create " + getClass().getName()
				+ " with null name."
			);
		}

		if ( ( this.description = description ) == null ) {
			throw new NullPointerException(
				"Trying to create " + getClass().getName()
				+ " with null description."
			);
		}

		if ( ( this.versionDate = versionDate ) == null ) {
			throw new NullPointerException(
				"Trying to create " + getClass().getName()
				+ " with null versionDate."
			);
		}
	}

	// TODO: Complete the transition and remoce this constructor.
	/**
	 * Construct a new AbstractStructuralDefinition object
	 * without versionDate.
	 *
	 * @deprecated You should always construct with version date.
	 * This constructor is only included to allow a smoother transition
	 * to versioned Structural Definitions.
	 * It will be removed once the transition is done.
	 */
	public AbstractStructuralDefinition (
		String identifier,
		String name,
		String description
	) {
		if ( ( this.identifier = identifier ) == null ) {
			throw new NullPointerException(
				"Trying to create " + getClass().getName()
				+ " with null identifier."
			);
		}

		if ( ( this.name = name ) == null ) {
			throw new NullPointerException(
				"Trying to create " + getClass().getName()
				+ " with null name."
			);
		}

		if ( ( this.description = description ) == null ) {
			throw new NullPointerException(
				"Trying to create " + getClass().getName()
				+ " with null description."
			);
		}

		versionDate = null;
	}

	/* ===================== Public Methods ===================== */

	/** Returns the identifier. */
	public String getId () { return identifier; }

	/** Returns the name. */
	public String getName () { return name; }

	/** Returns the description. */
	public String getDescription () { return description; }

	/** Returns the version timestamp. */
	public Date getVersionDate () { return versionDate; }

	/**
	 * Returns the hash code value for this object.
	 * For any two AbstractStructuralDefinitions sd1 and sd2,
	 * sd1.equals(sd2) implies sd1.hashCode()==sd2.hashCode(),
	 * as required by the general contract of {@link Object#hashCode}.
	 *
	 * @return a hash code value for this object.
	 */
	public int hashCode () {
		// if not yet computed (or accidentally zero)
		if ( hashCode == 0 ) {
			hashCode = identifier.hashCode()
				^ name.hashCode()
				^ description.hashCode()
				^ ( versionDate != null ? versionDate.hashCode() : 0 );
		}
		return hashCode;
	}

	/**
	 * Indicates whether some other AbstractStructuralDefinition object
	 * is "equal to" this one.
	 *
	 * @param obj the other reference object with which to compare
	 *
	 * @return true if this AbstractStructuralDefinition object
	 * is equal to the argument; false otherwise.
	 */
	public boolean equals ( Object object ) {
		if ( this == object ) return true;
		if ( ! ( object instanceof AbstractStructuralDefinition ) ) {
			return false;
		}
		final AbstractStructuralDefinition other
		= (AbstractStructuralDefinition) object;

		// TODO: Enforce non-null versionDate.
		// replaced != (false when both date are non null????) with ||.
		if ( ( versionDate == null ) || ( other.versionDate == null ) ) {
			return false;
		}

		return
			identifier.equals( other.identifier )
			&& name.equals( other.name )
			&& description.equals( other.description )
			&& (
				versionDate == null
				||
				versionDate.equals( other.versionDate )
			);
	}

	/**
	 * Return a string representation of
	 * this <CODE>AbstractStructuralDefinition</CODE>.
	 * This method is intended to be used only for debugging purposes,
	 * and the format of the returned string may vary between
	 * implementations.
	 * The returned string may be empty but may not be null.
	 *
	 * @return a string representation
	 * of this <CODE>AbstractStructuralDefinition</CODE>.
	 */
	public String toString () {
		return getClass().getName() + "("
			+ "identifier=" + identifier + ", "
			+ "name=" + name + ", "
			+ "description=" + description + ", "
			+ "versionDate=" + versionDate
			+ ")";
	}
}
