// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2002, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb;

import java.util.Date;
import java.util.Calendar;

import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;

/* ***************************** Class ****************************** */

/**
 * <P>Tests the implementation of the class {@link AbstractStructuralDefinition}.</P>
 *
 * @author
 * <A HREF="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</A>
 *
 * @version @(#)$Id: AbstractStructuralDefinitionTest.java,v 1.1 2017/01/19 16:14:04 trainot Exp $
 */

public class AbstractStructuralDefinitionTest extends TestCase {

	/* ====================== Constructors ====================== */

	public AbstractStructuralDefinitionTest ( final String name ) {
		super( name );
	}

	/* ===================== Public Methods ===================== */

	/**
	 * Assembles and returns a test suite for all
	 * the test methods of this test case.
	 *
	 * @return a non-null test suite
	 */
	public static Test suite () {
		return new TestSuite( AbstractStructuralDefinitionTest.class );
	}

	/**
	 * Runs the test case.
	 */
	public static void main ( String args[] ) {
		final String[] testCaseName = {
			AbstractStructuralDefinitionTest.class.getName()
		};
		junit.textui.TestRunner.main( testCaseName );
	}

	/* ---------------------- Test Methods ---------------------- */

	/**
	 * Test the two constructors.
	 */
	public void testConstruct () {
		final String
			identifier = "DEF1",
			name = "Definition 1",
			description = "This is just the first one";
		final Date versionDate = new Date();

		final AbstractStructuralDefinition s1
		= new TestAbstractStructuralDefinition(
			identifier, name, description, versionDate
		);
		final AbstractStructuralDefinition s2
		= new TestAbstractStructuralDefinition(
			identifier, name, description
		);

		assertSame( identifier, s1.getId() );
		assertSame( name, s1.getName() );
		assertSame( description, s1.getDescription() );
		assertSame( versionDate, s1.getVersionDate() );
		assertSame( identifier, s2.getId() );
		assertSame( name, s2.getName() );
		assertSame( description, s2.getDescription() );
		assertTrue( "s2 shoud have a version date",
			s2.getVersionDate() != null
		);

		assertTrue( "not1equals2", !s1.equals( s2 ) );
		assertTrue( "not2equals1", !s2.equals( s1 ) );
		assertTrue( "hashcodes", s1.hashCode() != s2.hashCode() );
	}

	/**
	 * Test the equals, hashCode and toString methods.
	 */
	public void testEquals () {
		final Calendar calendar = Calendar.getInstance();
		calendar.clear();
		final Date versionDate1 = calendar.getTime();
		calendar.add( Calendar.DAY_OF_YEAR, 1 );
		final Date versionDate2 = calendar.getTime();

		final Object[] original = {
			"DEF1",
			"Definition 1",
			"Some definition",
			versionDate1
		};
		final Object[] modified = {
			"DEF2",
			"Definition 2",
			"Some other definition",
			versionDate2
		};

		final AbstractStructuralDefinition s0 = new TestAbstractStructuralDefinition(
			(String) original[0],
			(String) original[1],
			(String) original[2],
			(Date) original[3]
		);
		final AbstractStructuralDefinition s1 = new TestAbstractStructuralDefinition(
			new String( (String) original[0] ),
			new String( (String) original[1] ),
			new String( (String) original[2] ),
			(Date) ( (Date) original[3] ).clone()
		);
		assertEquals( "equal objects", s0, s1 );
		assertEquals( "equal hashcodes", s0.hashCode(), s1.hashCode() );
		assertEquals( "equal strings", s0.toString(), s1.toString() );

		for ( int i = 0; i < original.length; ++i ) {
			final AbstractStructuralDefinition si = new TestAbstractStructuralDefinition(
				i == 0 ? (String) modified[0] : (String) original[0],
				i == 1 ? (String) modified[1] : (String) original[1],
				i == 2 ? (String) modified[2] : (String) original[2],
				i == 3 ? (Date) modified[3] : (Date) original[3]
			);

			assertTrue( "unequal objects["+i+"]", !si.equals( s0 ) );
			// Inequality of the hashCode is not really contract.
			// But the probability of equal hashCode is
			// far less than that of a buggy hashCode-implementation.
			assertTrue(
				"unequal hashcodes[" + i + "]",
				si.hashCode() != s0.hashCode()
			);
			assertTrue(
				"unequal strings[" + i + "]",
				!si.toString().equals( s0.toString() )
			);
		}
	}

	/**
	 * Tests serialization.
	 */
	public void testSerialization () {
		final Date versionDate = new Date();
		final AbstractStructuralDefinition s0
		= new TestAbstractStructuralDefinition(
			"DEF1", "Definition 1", "", versionDate
		);
		final AbstractStructuralDefinition s1
		= (AbstractStructuralDefinition) AttributeTest.transserialize( s0 );
		assertEquals( s0, s1 );
	}
}

/** Concrete test class. */
class TestAbstractStructuralDefinition extends AbstractStructuralDefinition {

	/**
	 *   serialVersionUID generated by eclipse.
	 */
	private static final long serialVersionUID = 6778289103622420879L;

	public TestAbstractStructuralDefinition (
		final String identifier,
		final String name,
		final String description,
		final Date versionDate
	) {
		super( identifier, name, description, versionDate );
	}

	/** Also make the deprecated constructor testable. */
	public TestAbstractStructuralDefinition (
		final String identifier,
		final String name,
		final String description
	) {
		// date is roughly 2005 in number of ms since 1st jan 1970.
		super( identifier, name, description, new Date( 1104516000 ) );
	}
}
