// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, 2003 European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb;

import Int.ecb.gesmescb.time.RegularInterval;

/* ***************************** Class ****************************** */

/**
 * The Attribute class holds the attribute definitions for the attribute event.
 *
 * @author
 * <A HREF="mailto:klo@FAME.com">Kam Lo</A>
 * @author
 * <A HREF="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</A>
 *
 * @version @(#)$Id: Attribute.java,v 1.1 2017/01/19 16:14:04 trainot Exp $
 */

public class Attribute extends AbstractCube {

	/* ======================= Constants ======================== */

	/**
	 *   serialVersionUID generated by eclipse.
	 */
	private static final long serialVersionUID = -7891667993240764280L;

	/**
	 * This value for scope means that the attribute relates to a data set.
	 *
	 * @deprecated The constructors using the parameter <code>scope</code>
	 * are obsolete. Therefore, this constant is no longer useful.
	 */
	public static final String DATA_SET_SCOPE = "1";

	/**
	 * This value for scope means that
	 * the attribute relates to a series or sibling group.
	 *
	 * @deprecated The constructors using the parameter <code>scope</code>
	 * are obsolete. Therefore, this constant is no longer useful.
	 */
	public static final String SERIES_OR_SIBLING_SCOPE = "4";

	/**
	 * This value for scope means that
	 * the attribute relates to a single observation.
	 *
	 * @deprecated The constructors using the parameter <code>scope</code>
	 * are obsolete. Therefore, this constant is no longer useful.
	 */
	public static final String OBSERVATION_SCOPE = "5";

	/**
	 * This value for coded means that
	 * the attribute takes values from a code list.
	 *
	 * @deprecated The constructors using the parameter <code>coded</code>
	 * are obsolete. Therefore, this constant is no longer useful.
	 */
	public static final String CODED = "Z10";

	/**
	 * This value for coded means that
	 * the attribute takes free text values.
	 *
	 * @deprecated The constructors using the parameter <code>coded</code>
	 * are obsolete. Therefore, this constant is no longer useful.
	 */
	public static final String UNCODED = "Z11";

	/* ========================= Fields ========================= */

	/**
	 * Name of the attribute.
	 *
	 * @serial
	 */
	private String name;

	/**
	 * Value of the attribute.
	 *
	 * @serial
	 */
	private final String value;

	/* -------------------- Transient Fields -------------------- */

	/**
	 * Precomputed hash code.
	 */
	private transient int hashCode;

	/* ====================== Constructors ====================== */

	/**
	 * Construct a new <code>Attribute</code>.
	 *
	 * @param name name of attribute.
	 * @param key which the attribute relates to.
	 * The components should be separated by {@link #KEY_SEPARATOR}.
	 * @param period observation period.
	 * @param periodFormat observation period format.
	 * @param scope of the attribute.
	 * @param coded whether attributed value is from a code list.
	 * This value is supposed to be one of
	 * {@link #CODED} or {@link #UNCODED}
	 * @param value value of attribute.
	 *
	 * @throws NullPointerException
	 * if any if the arguments is <CODE>null</CODE>.
	 *
	 * @deprecated The parameters <code>scope</code> and <code>coded</code>
	 * are obsolete.
	 * Use {@link #Attribute(String,String,String,String,String)} instead.
	 * But note the different order of parameters used there!
	 */
	public Attribute (
		final String name,
		final String key,
		final String period,
		final String periodFormat,
		final String scope,
		final String coded,
		final String value
	) {
		super( key, period, periodFormat );

		if ( ( this.name = name ) == null ) {
			throw new NullPointerException(
				"Trying to create Attribute with null name."
			);
		}

		if ( scope == null ) {
			throw new NullPointerException(
				"Trying to create Attribute with null scope."
			);
		}

		if ( coded == null ) {
			throw new NullPointerException(
				"Trying to create Attribute with null coded."
			);
		}

		if ( ( this.value = value ) == null ) {
			throw new NullPointerException(
				"Trying to create Attribute with null value."
			);
		}
	}

	/**
	 * Construct a new <code>Attribute</code>.
	 *
	 * @param name name of attribute.
	 * @param key which the attribute relates to.
	 * The components should be separated by {@link #KEY_SEPARATOR}.
	 * @param interval for which the observation is valid.
	 * @param scope of the attribute.
	 * @param coded whether attributed value is from a code list.
	 * This value is supposed to be one of
	 * {@link #CODED} or {@link #UNCODED}
	 * @param value value of attribute.
	 *
	 * @throws NullPointerException
	 * if any if the arguments is <CODE>null</CODE>.
	 *
	 * @deprecated The parameters <code>scope</code> and <code>coded</code>
	 * are obsolete.
	 * Use {@link #Attribute(String,RegularInterval,String,String)} instead.
	 * But note the different order of parameters used there!
	 */
	public Attribute (
		final String name,
		final String key,
		final RegularInterval interval,
		final String scope,
		final String coded,
		final String value
	) {
		super( key, interval );

		if ( ( this.name = name ) == null ) {
			throw new NullPointerException(
				"Trying to create Attribute with null name."
			);
		}

		if ( scope == null ) {
			throw new NullPointerException(
				"Trying to create Attribute with null scope."
			);
		}

		if ( coded == null ) {
			throw new NullPointerException(
				"Trying to create Attribute with null coded."
			);
		}

		if ( ( this.value = value ) == null ) {
			throw new NullPointerException(
				"Trying to create Attribute with null value."
			);
		}
	}

	/**
	 * Construct a new <code>Attribute</code>.
	 *
	 * @param name name of attribute.
	 * @param key which the attribute relates to.
	 * The components should be separated by {@link #KEY_SEPARATOR}.
	 * @param period observation period.
	 * @param periodFormat observation period format.
	 *
	 * @throws NullPointerException
	 * if any if the arguments is <CODE>null</CODE>.
	 */
	public Attribute (
		final String key,
		final String period,
		final String periodFormat,
		final String name,
		final String value
	) {
		super( key, period, periodFormat );

		if ( ( this.name = name ) == null ) {
			throw new NullPointerException(
				"Trying to create Attribute with null name."
			);
		}

		if ( ( this.value = value ) == null ) {
			throw new NullPointerException(
				"Trying to create Attribute with null value."
			);
		}
	}

	/**
	 * Construct a new <code>Attribute</code>.
	 *
	 * @param name name of attribute.
	 * @param key which the attribute relates to.
	 * The components should be separated by {@link #KEY_SEPARATOR}.
	 * @param interval for which the observation is valid.
	 * This value is supposed to be one of
	 * {@link #CODED} or {@link #UNCODED}
	 * @param value value of attribute.
	 *
	 * @throws NullPointerException
	 * if any if the arguments is <CODE>null</CODE>.
	 */
	public Attribute (
		final String key,
		final RegularInterval interval,
		final String name,
		final String value
	) {
		super( key, interval );

		if ( ( this.name = name ) == null ) {
			throw new NullPointerException(
				"Trying to create Attribute with null name."
			);
		}

		if ( ( this.value = value ) == null ) {
			throw new NullPointerException(
				"Trying to create Attribute with null value."
			);
		}
	}

	/* ===================== Public Methods ===================== */

	/**
	 * Returns attribute name.
	 *
	 * @return the identifier of the respective Statistical Concept.
	 */
	public String getName () {
		return name;
	}

	/**
	 * Sets the name of the attribute.
	 * Needed for attributes at dataset level, which get the dataset name attached 
	 * @param name
	 */
	public void setName (final String name) {
		this.name = name;
	}

	
	/**
	 * Returns attribute value.
	 *
	 * @return attribute value.
	 */
	public String getValue () {
		return value;
	}

	/**
	 * Indicates whether some other Object is "equal to" this Attribute.
	 *
	 * @param obj the reference object with which to compare
	 *
	 * @return true if this Attribute is equal to the argument;
	 * false otherwise.
	 */
	public boolean equals ( final Object obj ) {
		if ( this == obj ) {
			return true;
		} else if ( obj instanceof Attribute ) {
			final Attribute other = (Attribute) obj;
			return
				super.equals( other )
				&& name.equals( other.name )
				&& value.equals( other.value );
		} else {
			return false;
		}
	}

	/**
	 * Returns the hash code value for this Attribute.
	 * If a1.equals(a2) then a1.hashCode()==a2.hashCode()
	 * for any two Attributes a1 and a2,
	 * as required by the general contract of {@link Object#hashCode}.
	 *
	 * @return hash code value for this object.
	 */
	public int hashCode () {
		// if not yet computed (or accidentally zero)
		if ( hashCode == 0 ) {
			hashCode = super.hashCode()
				^ name.hashCode()
				^ value.hashCode();
		}
		return hashCode;
	}

	/**
	 * Returns a string representation of the Attribute
	 * in a human readable format.
	 *
	 * @return a string representation of the Attribute.
	 */
	public String toString () {
		return
			getClass().getName()
			+ "( key = " + getKey()
			+ ", period = " + getPeriod()
			+ ", periodFormat = " + getPeriodFormat()
			+ ", name = " + name
			+ ", value = " + value + " )";
	}
}
