// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb;

import java.io.*;

import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;

import Int.ecb.gesmescb.time.RegularInterval;
import Int.ecb.gesmescb.time.GregorianRegularInterval;

/* ***************************** Class ****************************** */

/**
 * <P>Tests the implementation of the class {@link Attribute}.</P>
 *
 * @author
 * <A HREF="mailto:Sami.Airo@ECB.int">Sami Airo</A>
 * @author
 * <A HREF="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</A>
 *
 * @version @(#)$Id: AttributeTest.java,v 1.1 2017/01/19 16:14:04 trainot Exp $
 */

public class AttributeTest extends TestCase {

	/* ====================== Constructors ====================== */

	/**
	 * Constructs an AttributeTest with the specified name.
	 *
	 * @param name Test case name
	 */
	public AttributeTest ( final String name ) {
		super( name );
	}

	/* ===================== Public Methods ===================== */

	/**
	 * Assembles and returns a test suite for all
	 * the test methods of this test case.
	 *
	 * @return a non-null test suite
	 */
	public static Test suite () {
		return new TestSuite( AttributeTest.class );
	}

	/**
	 * Runs the test case.
	 */
	public static void main ( final String args[] ) {
		final String[] testCaseName = { AttributeTest.class.getName() };
		junit.textui.TestRunner.main( testCaseName );
	}

	/* ---------------------- Test Methods ---------------------- */

	/** Test the two constructors. */
	public void testConstructors () {
		final Attribute a1 = new Attribute(
			"BLACK.SOX", "2003", "602", "TITLE", "The duff stuff."
		);
		final Attribute a2 = new Attribute(
			"BLACK.SOX",
			new GregorianRegularInterval(
				RegularInterval.ANNUAL, 2003, 0, 1
			),
			"TITLE", "The duff stuff."
		);
		assertEquals( a1, a2 );
	}

	/**
	 * Test the equals, hashCode and toString methods.
	 */
	public void testEquals () {
		// Create Strings for constructor args.
		String[] original = {
			"name",
			"1995040220010402",
			"711",
			"key",
			"Rsel"
		};
		// Create modifications.
		String[] modified = {
			"Key",
			"1995040220000402",
			"911",	// (is invalid, but do not care here)
			"Name",
			"Brsel"
		};

		// Create Attribute.
		Attribute a0 = new Attribute(
			original[0],
			original[1],
			original[2],
			original[3],
			original[4]
		);
		// Create "equal" Attribute from equal, non-identical Strings.
		Attribute a1 = new Attribute(
			new String( original[0] ),
			new String( original[1] ),
			new String( original[2] ),
			new String( original[3] ),
			new String( original[4] )
		);
		assertTrue( "a0 equals a1", a0.equals( a1 ) );
		assertTrue(
			"a0.hashCode == a1.hashCode",
			a0.hashCode() == a1.hashCode()
		);
		assertTrue(
			"a0.toString equals a1.toString",
			a0.toString().equals( a1.toString() )
		);

		// Next create unequal Attributes,
		// which are unequal at varying positions.
		Attribute[] a = new Attribute[original.length];
		for ( int i = 0; i < a.length; ++i ) {
			a[i] = new Attribute(
				i == 0 ? modified[0] : original[0],
				i == 1 ? modified[1] : original[1],
				i == 2 ? modified[2] : original[2],
				i == 3 ? modified[3] : original[3],
				i == 4 ? modified[4] : original[4]
			);

			assertTrue( "a["+i+"] not equals a0", !a[i].equals( a0 ) );
			// Inequality of the hashCode is not really contract.
			// But the probability of eventual equal hashCode is
			// far less than that of a buggy
			// hashCode-implementation.
			// Therefore check.
			assertTrue(
				"a[" + i + "].hashCode != a0.hashCode",
				a[i].hashCode() != a0.hashCode()
			);
			assertTrue(
				"a[" + i + "].toString != a0.toString",
				!a[i].toString().equals( a0.toString() )
			);
		}
	}

	/**
	 * Test the get-type methods.
	 */
	public void testSimpleProperties () {
		Attribute a0 = new Attribute(
			"AR.RA", "200109200110", "710", "Sane", "Paranoia"
		);
		Attribute a1 = new Attribute(
			".", "200109200110", "710", "Insane", "Sox"
		);
		Attribute a2 = new Attribute(
			"", "200109200110", "710", "Infaust", "Pfllorch"
		);

		assertEquals( "Sane", a0.getName() );
		assertEquals( "AR.RA", a0.getKey() );
		assertEquals( "200109200110", a0.getPeriod() );
		assertEquals( "710", a0.getPeriodFormat() );
		assertEquals( "Paranoia", a0.getValue() );

		assertEquals( "a0.keyLength", 2, a0.getKeyLength() );
		assertEquals( "a1.keyLength", 2, a1.getKeyLength() );
		assertEquals( "a2.keyLength", 1, a2.getKeyLength() );
	}

	/**
	 * Tests serialization.
	 */
	public void testSerialization () {
		Attribute a0 = new Attribute(
			"INS", "200109200110", "710", "Insane", "Paranoia"
		);
		Attribute a1 = (Attribute) transserialize( a0 );
		assertEquals( a0, a1 );
		assertEquals( "key length", 1, a1.getKeyLength() );
	}

	/** Temporarily serialize and then deserialize again.
	 * @param object to serialize.
	 * @return deserialized <CODE>object</CODE>.
	 */
	public static Serializable transserialize ( final Serializable object ) {
		Serializable transserialized = null;

		try {
			final ByteArrayOutputStream outputStream
			= new ByteArrayOutputStream();
			final ObjectOutput objectOut
			= new ObjectOutputStream( outputStream );

			objectOut.writeObject( object );
			objectOut.close();

			final byte[] buffer = outputStream.toByteArray();

			final InputStream inputStream = new ByteArrayInputStream( buffer );
			final ObjectInput objectIn = new ObjectInputStream( inputStream );

			transserialized = (Serializable) objectIn.readObject();
			objectIn.close();
		} catch ( final Throwable e ) {
			fail( e.getClass().getName() + " thrown" );
		}
		return transserialized;
	}
}
