// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb;

import java.util.*;

/* ***************************** Class ****************************** */

/**
 * <P>Holds one GESMES/TS code list.
 * CodeList resembles a Java Collection class in a way
 * that it stores code values and their associated descriptions.</P>
 *
 * <P>A CodeList object is meant to be immutable
 * so that it does not change after construction.</P>
 *
 * @author
 * <A HREF="mailto:klo@FAME.com">Kam Lo</A>
 * @author
 * <A HREF="mailto:Sami.Airo@ECB.int">Sami Airo</A>
 * @author
 * <A HREF="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</A>
 *
 * @version @(#)$Id: CodeList.java,v 1.1 2017/01/19 16:14:04 trainot Exp $
 */

public class CodeList extends AbstractStructuralDefinition {

	/* ======================= Constants ======================== */

	/**
	 *   serialVersionUID generated by eclipse.
	 */
	private static final long serialVersionUID = 4283642818994444223L;

	/**
	 * A constant empty array to support type checking in the call
	 * to the method List.toArray( String[] ).
	 * Only the type String[] is relevant.
	 *
	 * The size must be 0, otherwise instance data could be written
	 * to the static array by List.toArray!
	 *
	 * Note that in case of empty keys
	 * a reference to the static array itself
	 * becomes part of the instance.
	 * But empty is anyway empty.
	 */
	private static final String[] emptyStringArray = new String[0];

	/* ========================= Fields ========================= */

	/**
	 * List of code values.
	 * See the field codeMap.
	 *
	 * @serial
	 */
	private final String[] codeValues;

	/**
	 * List of code descriptions.
	 * See the field codeMap.
	 *
	 * @serial
	 */
	private final String[] codeDescriptions;

	/* -------------------- Transient Fields -------------------- */

	/**
	 * Map of the code values to their descriptions.
	 * The codeMap contains the same information as
	 * the fields codeValues plus codeDescriptions
	 * unless there are code values with more than one
	 * description.
	 * This redundance is the price for consistent overall design
	 * and high performance.
	 */
	private transient Map codeMap;

	/**
	 * Precomputed hash code.
	 */
	private transient int hashCode;

	/* ====================== Constructors ====================== */

	/**
	 * Constructs a new code list.
	 * The Lists used as parameters to the constructor
	 * are guaranteed to be copied.
	 * The Lists must contain only Strings, otherwise an ArrayStoreException
	 * will be thrown.
	 *
	 * @param identifier			code list identifier.
	 * @param name			code list name.
	 * @param description		code list description.
	 * @param versionDate	concept version timestamp.
	 * @param codeValues		List of code values.
	 * @param codeDescriptions	List of code descriptions.
	 *
	 * @throws ArrayStoreException if codeValues or codeDescriptions
	 * contains objects of a different class than {@link String}.
	 */
	public CodeList (
		final String identifier,
		final String name,
		final String description,
		final Date versionDate,
		final List codeValues,
		final List codeDescriptions
	) {
		super( identifier, name, description, versionDate );

		this.codeValues
		= (String[])codeValues.toArray( emptyStringArray );
		this.codeDescriptions
		= (String[])codeDescriptions.toArray( emptyStringArray );

		initCodeMap();	// init transient.
	}

	/**
	 * Constructs a new code list.
	 * The Lists used as parameters to the constructor
	 * are guaranteed to be copied.
	 * The Lists must contain only Strings, otherwise an ArrayStoreException
	 * will be thrown.
	 *
	 * @param identifier			code list identifier.
	 * @param name			code list name.
	 * @param description		code list description.
	 * @param codeValues		List of code values.
	 * @param codeDescriptions	List of code descriptions.
	 *
	 * @throws ArrayStoreException if codeValues or codeDescriptions
	 * contains objects of a different class than {@link String}.
	 *
	 * @deprecated Do not construct a CodeList without a versionDate.
	 * Use {@link #CodeList(String,String,String,Date,List,List)}.
	 */
	public CodeList (
		final String identifier,
		final String name,
		final String description,
		final List codeValues,
		final List codeDescriptions
	) {
		super( identifier, name, description );

		this.codeValues
		= (String[])codeValues.toArray( emptyStringArray );
		this.codeDescriptions
		= (String[])codeDescriptions.toArray( emptyStringArray );

		initCodeMap();	// init transient.
	}

	/**
	 * Constructs a new code list.
	 * The arrays used as parameters to the constructor
	 * are guaranteed to be copied.
	 *
	 * @param identifier			code list identifier.
	 * @param name			code list name.
	 * @param description		code list description.
	 * @param versionDate	concept version timestamp.
	 * @param codeValues		array of code values.
	 * @param codeDescriptions	array of code descriptions.
	 */
	public CodeList (
		final String identifier,
		final String name,
		final String description,
		final Date versionDate,
		final String[] codeValues,
		final String[] codeDescriptions
	) {
		super( identifier, name, description, versionDate );

		this.codeValues = (String[]) codeValues.clone();
		this.codeDescriptions = (String[]) codeDescriptions.clone();

		initCodeMap();	// init transient.
	}

	/**
	 * Constructs a new code list.
	 * The arrays used as parameters to the constructor
	 * are guaranteed to be copied.
	 *
	 * @param identifier			code list identifier.
	 * @param name			code list name.
	 * @param description		code list description.
	 * @param codeValues		array of code values.
	 * @param codeDescriptions	array of code descriptions.
	 *
	 * @deprecated Do not construct a CodeList without a versionDate.
	 * Use {@link #CodeList(String,String,String,Date,String[],String[])}.
	 */
	public CodeList (
		final String identifier,
		final String name,
		final String description,
		final String[] codeValues,
		final String[] codeDescriptions
	) {
		super( identifier, name, description );

		this.codeValues = (String[]) codeValues.clone();
		this.codeDescriptions = (String[]) codeDescriptions.clone();

		initCodeMap();	// init transient.
	}

	/* ===================== Public Methods ===================== */

	/**
	 * Returns the number of code values in this CodeList.
	 *
	 * @return the number of code values in this CodeList.
	 */
	public int getCodeValueCount () {
		return codeValues.length;
	}

	/**
	 * Returns the description for a code value.
	 * If the CodeList contains no description for this code value,
	 * then null is returned.
	 * If there are duplicate values in the code list with
	 * different descriptions then the return value is
	 * one of them.
	 *
	 * @param codeValue	code value.
	 *
	 * @return the description for a code value.
	 */
	public synchronized String getCodeDescription ( final String codeValue ) {
		// synchronized is necessary because of Java Virtual Memory model.
		if ( codeMap == null ) initCodeMap();
		return (String) codeMap.get( codeValue );
	}

	/**
	 * Tests if the code list contains the specified code value.
	 *
	 * @param codeValue code value whose presence in this code list
	 * should be tested.
	 *
	 * @return true if this code list contains the specified code value
	 * and false otherwise.
	 */
	public synchronized boolean containsCodeValue ( final String codeValue ) {
		// synchronized is necessary because of Java Virtual Memory model.
		if ( codeMap == null ) initCodeMap();
		return codeMap.containsKey( codeValue );
	}

	/**
	 * Returns an iterator over the code values in this code list.
	 * The values are returned in an appropriate order to match
	 * the descriptions.
	 * The iterator does not support modifying access.
	 *
	 * @return an iterator over the code values in this code list.
	 */
	public Iterator codeValueIterator () {
		return Collections.unmodifiableList(
			Arrays.asList( codeValues )
		).iterator();
	}

	/**
	 * Returns an iterator over the code descriptions in this code list.
	 * The descriptions are returned in an appropriate order to match
	 * the values.
	 * The iterator does not support modifying access.
	 *
	 * @return an iterator over the code descriptions in this code list.
	 */
	public Iterator codeDescriptionIterator () {
		return Collections.unmodifiableList(
			Arrays.asList( codeDescriptions )
		).iterator();
	}

	/**
	 * Returns the hash code value for this CodeList.
	 *
	 * If cl1.equals(cl2) then cl1.hashCode()==cl2.hashCode()
	 * for any two CodeLists cl1 and cl2,
	 * as required by the general contract of {@link Object#hashCode}.
	 *
	 * @return hash code value for this object.
	 */
	public int hashCode () {
		// if not yet computed (or accidentally zero)
		if ( hashCode == 0 ) {
			hashCode = super.hashCode()
				^ Arrays.asList( codeValues ).hashCode()
				^ Arrays.asList( codeDescriptions ).hashCode();
		}
		return hashCode;
	}

	/**
	 * Indicates whether some other CodeList is "equal to" this one.
	 *
	 * @param object with which to compare
	 *
	 * @return true if this CodeList is equal to the argument;
	 * false otherwise.
	 */
	public boolean equals ( final Object object ) {
		if ( this == object ) return true;
		if ( object instanceof CodeList ) {
			final CodeList other = (CodeList) object;
			return
				super.equals( other )
				&& Arrays.equals( codeValues, other.codeValues )
				&& Arrays.equals(
					codeDescriptions,
					other.codeDescriptions
				);
		}
		return false;
	}

	/**
	 * Returns a String representation of the CodeList.
	 */
	public String toString() {
		// Build a string for the actual code mapping.
		StringBuffer stringBuffer = new StringBuffer();
		for (
			int i = 0;
			i < codeValues.length && i < codeDescriptions.length;
			++i
		) {
			if ( i > 0 ) stringBuffer.append( ", " );
			stringBuffer.append( codeValues[i] );
			stringBuffer.append( '=' );
			stringBuffer.append( codeDescriptions[i] );
		}

		return getClass().getName()
			+ " ( identifier=" + getId() + ", "
			+ "name=" + getName() + ", "
			+ "description=" + getDescription() + ", "
			+ "versionDate=" + getVersionDate() + ", "
			+ ", { " + stringBuffer.toString() + " } )";
	}

	/* ==================== Private Methods ===================== */

	/**
	 * Derive the codeMap field from the object's state.
	 * Since the serialized state does not contain transient fields,
	 * they must be computed upon request.
	 */
	private synchronized void initCodeMap () {
		codeMap = new HashMap();
		for (
			int i = 0;
			i < codeValues.length && i < codeDescriptions.length;
			++i
		) {
			codeMap.put(
				codeValues[i],
				codeDescriptions[i]
			);
		}
	}
}
