package Int.ecb.gesmescb;

import java.util.Collection;
import java.util.Iterator;
import java.util.Collections;
import java.util.Arrays;

import java.io.Serializable;

import Int.ecb.gesmescb.ContactChannel;

/* ***************************** Class ****************************** */

/**
 * Holds GESMES/TS message contact information.
 *
 * @author <A HREF="mailto:Sami.Airo@ECB.int">Sami Airo</A>
 *
 * @version @(#)$Id: Contact.java,v 1.1 2017/01/19 16:14:04 trainot Exp $
 */

public final class Contact implements Serializable {

	/* ========================= Constants ====================== */

	/* -------------------- Public constants -------------------- */

	/**
	 *   serialVersionUID generated by eclipse.
	 */
	private static final long serialVersionUID = -4009662444257171544L;

	/**
	 * Value of the function field indicating responsible for information production.
	 */
	public static final String INFORMATION_PRODUCTION = "CC";

	/**
	 * Value of the function field indicating responsible for data processing.
	 */
	public static final String DATA_PROCESSING = "CP";

	/**
	 * Value of the function field indicating head of information production.
	 */
	public static final String HEAD_OF_INFORMATION_PRODUCTION = "CF";

	/**
	 * Value of the function field indicating head of data processing.
	 */
	public static final String HEAD_OF_DATA_PROCESSING = "CE";

	/* ------------------- Private constants -------------------- */

	/**
	 * A constant empty array to support type checking in the call
	 * to the method Collection.toArray( ContactChannel[] ).
	 * Only the type ContactChannel[] is relevant.
	 *
	 * The size must be 0, otherwise instance data would be written
	 * to the static array by Collection.toArray!
	 *
	 * Note that in case of empty keys
	 * a reference to the static array itself
	 * becomes part of the instance.
	 * But empty is anyway empty.
	 */
	private static final ContactChannel[] emptyContactChannelArray
	= new ContactChannel[0];

	/* ========================= Fields ========================= */

	/**
	 * Function of the contact.
	 *
	 * @serial
	 */
	private final String function;

	/**
	 * Identity of the contact.
	 *
	 * @serial
	 */
	private final String identity;

	/**
	 * Name of the contact.
	 *
	 * @serial
	 */
	private final String name;

	/**
	 * Contact channels.
	 *
	 * @serial
	 */
	private ContactChannel[] contactChannels;

	/* -------------------- Transient Fields -------------------- */

	/**
	 * Precomputed hash code.
	 * This will be lazily computed on demand if it is zero.
	 */
	private transient int hashCode;

	/* ====================== Constructors ====================== */

	/**
	 * Construct a new <CODE>Contact</CODE> from message header information.
	 *
	 * @param function Function of the contact.
	 * @param identity Identity of the contact.
	 * @param name Name of the contact.
	 * @param contactChannels Contact channels.
	 *
	 * @throws NullPointerException if any of the parameters is null.
	 */
	public Contact (
		final String function,
		final String identity,
		final String name,
		final Collection contactChannels
	)
		throws NullPointerException
	{
		if ( ( this.function = function ) == null )
			throw new NullPointerException(
				"Initializing Contact with null function"
			);

		if ( ( this.identity = identity ) == null )
			throw new NullPointerException(
				"Initializing Contact with null identity"
			);

		if ( ( this.name = name ) == null )
			throw new NullPointerException(
				"Initializing Contact with null name"
			);

		if ( contactChannels == null ) {
			throw new NullPointerException(
				"Initializing Contact with null contactChannels"
			);
		} else {
			this.contactChannels = (ContactChannel[])contactChannels.toArray(
				emptyContactChannelArray
			);
		}
	}

	/**
	 * Construct a new <CODE>Contact</CODE> from message header information.
	 *
	 * @param function of the contact.
	 * @param identity of the contact. E.g. the department name.
	 * @param name of the contact. E.g. a hotline or person's name.
	 * @param contactChannels available.
	 *
	 * @throws NullPointerException if any of the parameters is null.
	 */
	public Contact (
		final String function,
		final String identity,
		final String name,
		final ContactChannel[] contactChannels
	)
		throws NullPointerException
	{
		if ( ( this.function = function ) == null )
			throw new NullPointerException(
				"Initializing Contact with null function"
			);

		if ( ( this.identity = identity ) == null )
			throw new NullPointerException(
				"Initializing Contact with null identity"
			);

		if ( ( this.name = name ) == null )
			throw new NullPointerException(
				"Initializing Contact with null name"
			);

		if ( contactChannels == null ) {
			throw new NullPointerException(
				"Initializing Contact with null contactChannels"
			);
		} else {
			// do a real copy for immutability.
			this.contactChannels = new ContactChannel[contactChannels.length];
			System.arraycopy(
				contactChannels, 0,
				this.contactChannels, 0,
				contactChannels.length
			);
		}
	}

	/* ===================== Public Methods ===================== */

	/**
	 * Returns the function of the contact.
	 *
	 * @return function of the contact.
	 */
	public String getFunction () {
		return function;
	}

	/**
	 * Returns the identity of the contact.
	 *
	 * @return identity of the contact.
	 */
	public String getIdentity () {
		return identity;
	}

	/**
	 * Returns the name of the contact.
	 *
	 * @return name of the contact.
	 */
	public String getName () {
		return name;
	}

	/**
	 * Returns an iterator to the contact channels.
	 *
	 * @return iterator to the contact channels.
	 */
	public Iterator channelIterator () {
		return (
			Collections.unmodifiableList(
				Arrays.asList( contactChannels )
			)
		).iterator();
	}

	/**
	 * Tests if the contact is responsible for information production.
	 *
	 * @return true if the contact is responsible for information production; false otherwise.
	 */
	public boolean isResponsibleForInformationProduction () {
		return function.equals(INFORMATION_PRODUCTION);
	}

	/**
	 * Tests if the contact is responsible for data processing.
	 *
	 * @return true if the contact is responsible for data processing; false otherwise.
	 */
	public boolean isResponsibleForDataProcessing () {
		return function.equals(DATA_PROCESSING);
	}

	/**
	 * Tests if the contact is the head of information production.
	 *
	 * @return true if the contact is the head of information production; false otherwise.
	 */
	public boolean isHeadOfInformationProduction () {
		return function.equals(HEAD_OF_INFORMATION_PRODUCTION);
	}

	/**
	 * Tests if the contact is the head of data processing.
	 *
	 * @return true if the contact is the head of data processing; false otherwise.
	 */
	public boolean isHeadOfDataProcessing () {
		return function.equals(HEAD_OF_DATA_PROCESSING);
	}

	/* ------------------ Other Public Methods ------------------ */

	/**
	 * Returns the hash code value for this Contact object.
	 * Equality according to {@link #equals} implies that
	 * the hashcodes are the same
	 * as required by the general contract of {@link Object#hashCode}.
	 *
	 * @return hash code value for this object.
	 */
	public int hashCode () {
		// if not yet computed (or accidentally zero)
		if ( hashCode == 0 ) {
			hashCode =
				function.hashCode()
				^ identity.hashCode()
				^ name.hashCode()
				^ Arrays.asList( contactChannels ).hashCode();
		}
		return hashCode;
	}

	/**
	 * Indicates whether some other Contact object
	 * is "equal to" this one.
	 *
	 * @param obj the other reference object with which to compare
	 *
	 * @return true if this Contact object
	 * is equal to the argument; false otherwise.
	 */
	public boolean equals ( Object obj ) {
		if ( this == obj ) {
			return true;
		} else if ( obj instanceof Contact ) {
			Contact other = (Contact) obj;

			return
				function.equals( other.function )
				&& identity.equals( other.identity )
				&& name.equals( other.name )
				&& Arrays.equals( contactChannels, other.contactChannels );
		} else {
			return false;
		}
	}

	/**
	 * Returns a string representation of the Contact
	 * in a human readable format.
	 *
	 * @return a string representation of the Contact.
	 */
	public String toString () {
		// Build a string for the contactChannels.
		StringBuffer stringBuffer = new StringBuffer();
		for ( int i = 0; i < contactChannels.length; ++i ) {
			if ( i > 0 ) stringBuffer.append( ", " );
			stringBuffer.append( contactChannels[i].toString() );
		}

		return getClass().getName() + "("
			+ "function=" + function + ", "
			+ "identity=" + identity + ", "
			+ "name=" + name + ", "
			+ "contactChannels={ " + stringBuffer.toString() + " } )";
	}
}
