package Int.ecb.gesmescb;

import java.io.Serializable;

/* ***************************** Class ****************************** */

/**
 * Holds the GESMES/TS communications channel information for a message contact.
 *
 * @author <A HREF="mailto:Sami.Airo@ECB.int">Sami Airo</A>
 *
 * @version @(#)$Id: ContactChannel.java,v 1.1 2017/01/19 16:14:04 trainot Exp $
 */

public final class ContactChannel implements Serializable {

	/* ========================= Constants ====================== */

	/**
	 *   serialVersionUID generated by eclipse.
	 */
	private static final long serialVersionUID = 4271204325096131869L;

	/**
	 * Value of the type field indicating e-mail.
	 */
	public static final String EMAIL = "EM";

	/**
	 * Value of the type field indicating telephone.
	 */
	public static final String TELEPHONE = "TE";

	/**
	 * Value of the type field indicating fax.
	 */
	public static final String FAX = "FX";

	/**
	 * Value of the type field indicating X.400 mail.
	 */
	public static final String X400 = "XF";

	/* ========================= Fields ========================= */

	/**
	 * Number for the communication channel.
	 *
	 * @serial
	 */
	private final String number;

	/**
	 * Type of the communication channel.
	 *
	 * @serial
	 */
	private final String type;

	/* -------------------- Transient Fields -------------------- */

	/**
	 * Precomputed hash code.
	 * This will be lazily computed on demand if it is zero.
	 */
	private transient int hashCode;

	/* ====================== Constructors ====================== */

	/**
	 * Construct a new <CODE>ContactChannel</CODE> from message header information.
	 *
	 * @param type type of the contact channel.
	 * @param number Number for the contact channel.
	 *
	 * @throws NullPointerException if any of the parameters is null.
	 */
	public ContactChannel (
		final String type,
		final String number
	)
		throws NullPointerException
	{
		if ( ( this.number = number ) == null )
			throw new NullPointerException(
				"Initializing ContactChannel with null number"
			);

		if ( ( this.type = type ) == null )
			throw new NullPointerException(
				"Initializing ContactChannel with null type"
			);
	}

	/* ===================== Public Methods ===================== */

	/**
	 * Returns the number.
	 *
	 * @return number.
	 */
	public String getNumber () {
		return number;
	}

	/**
	 * Returns the channel type.
	 *
	 * @return channel type.
	 */
	public String getType () {
		return type;
	}

	/**
	 * Tests if the number is an email address.
	 *
	 * @return true if the number is an email address; false otherwise.
	 */
	public boolean isEmailAddress () {
		return type.equals(EMAIL);
	}

	/**
	 * Tests if the number is a telephone number.
	 *
	 * @return true if the number is a telephone number; false otherwise.
	 */
	public boolean isTelephoneNumber () {
		return type.equals(TELEPHONE);
	}

	/**
	 * Tests if the number is a fax number.
	 *
	 * @return true if the number is a fax number; false otherwise.
	 */
	public boolean isFaxNumber () {
		return type.equals(FAX);
	}

	/**
	 * Tests if the number is a X.400 address.
	 *
	 * @return true if the number is a X.400 address; false otherwise.
	 */
	public boolean isX400Address () {
		return type.equals(X400);
	}

	/* ------------------ Other Public Methods ------------------ */

	/**
	 * Returns the hash code value for this ContactChannel object.
	 * Equality according to {@link #equals} implies that
	 * the hashcodes are the same
	 * as required by the general contract of {@link Object#hashCode}.
	 *
	 * @return hash code value for this object.
	 */
	public int hashCode () {
		// if not yet computed (or accidentally zero)
		if ( hashCode == 0 ) {
			hashCode =
				number.hashCode()
				^ type.hashCode();
		}
		return hashCode;
	}

	/**
	 * Indicates whether some other ContactChannel object
	 * is "equal to" this one.
	 *
	 * @param obj the other reference object with which to compare
	 *
	 * @return true if this ContactChannel object
	 * is equal to the argument; false otherwise.
	 */
	public boolean equals ( final Object obj ) {
		if ( this == obj ) {
			return true;
		} else if ( obj instanceof ContactChannel ) {
			ContactChannel other = (ContactChannel) obj;

			return
				number.equals( other.number )
				&& type.equals( other.type );
		} else {
			return false;
		}
	}

	/**
	 * Returns a string representation of the ContactChannel
	 * in a human readable format.
	 *
	 * @return a string representation of the ContactChannel.
	 */
	public String toString () {
		return getClass().getName() + "("
			+ "number=" + number + ", "
			+ "type=" + type + ")";
	}
}
