// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb;

import java.text.ParsePosition;
import java.text.SimpleDateFormat;
import java.util.Date;

/* ***************************** Class ****************************** */

/**
 * Holds the GESMES/TS interchange header information.
 *
 * @author
 * <A HREF="mailto:klo@FAME.com">Kam Lo</A>
 * @author
 * <A HREF="mailto:Sami.Airo@ECB.int">Sami Airo</A>
 * @author
 * <A HREF="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</A>
 *
 * @version @(#)$Id: InterchangeBegin.java,v 1.1 2017/01/19 16:14:04 trainot Exp $
 */

public final class InterchangeBegin extends GESMESCBObject {

	/* ======================= Constants ======================== */

	/**
	 *   serialVersionUID generated by eclipse.
	 */
	private static final long serialVersionUID = 6165036547660056036L;

	/**
	 * Value returned by {@link #getIrefNo} if the
	 * interchange reference number is in an invalid format.
	 */
	public static final int INVALID_IREF = -1;

	/* ========================= Fields ========================= */

	/**
	 * Interchange reference identifyer.
	 *
	 * @serial
	 */
	private final String iref;

	/**
	 * Alledged sender of the interchange.
	 *
	 * @serial
	 */
	private final String sender;

	/**
	 * Intended receiver of the interchange.
	 *
	 * @serial
	 */
	private final String receiver;

	/**
	 * Timestamp of preparation (as Date).
	 *
	 * @serial
	 */
	private final Date preparationDate;

	/**
	 * True if this interchange is for testing only; false otherwise.
	 *
	 * @serial
	 */
	private final boolean isTest;

	/* ------------------- Transient Fields -------------------- */

	/**
	 * Interchange reference number.
	 */
	private transient int irefNo;

	/**
	 * Precomputed hashcode.
	 */
	private transient int hashCode;

	/* ====================== Constructors ====================== */

	/**
	 * Construct a new interchange.
	 *
	 * @param iref interchange reference number.
	 * @param sender alledged sender of interchange.
	 * @param receiver intended receiver of interchange.
	 * @param preparationDateTime preparation time stamp in format
	 * <code><i>YYMMDD</i>:<i>hhmm</i></code>.
	 * @param isTest indicates whether this interchange is for testing only.
	 *
	 * @throws NullPointerException if any of the String arguments is null.
	 */
	public InterchangeBegin (
		final String iref,
		final String sender,
		final String receiver,
		final String preparationTimestamp,
		final boolean isTest
	) {
		this(
			iref,
			sender,
			receiver,
			preparationTimestampToDate( preparationTimestamp ),
			isTest
		);
	}

	/**
	 * Construct a new interchange.
	 *
	 * @param iref interchange reference number.
	 * @param sender alledged sender of interchange.
	 * @param receiver intended receiver of interchange.
	 * @param preparationDate preparation time stamp.
	 * @param isTest indicates whether this interchange is for testing only.
	 *
	 * @throws NullPointerException if any of the String arguments is null.
	 */
	public InterchangeBegin (
		final String iref,
		final String sender,
		final String receiver,
		final Date preparationDate,
		final boolean isTest
	) {
		if ( ( this.iref = iref ) == null )
			throw new NullPointerException(
				"Initializing InterchangeBegin with null iref"
			);

		if ( ( this.sender = sender ) == null )
			throw new NullPointerException(
				"Initializing InterchangeBegin with null sender"
			);

		if ( ( this.receiver = receiver ) == null )
			throw new NullPointerException(
				"Initializing InterchangeBegin with null receiver"
			);
		
		this.preparationDate = preparationDate;
		this.isTest = isTest;
	}

	/* ===================== Public Methods ===================== */

	/**
	 * Returns interchange reference.
	 *
	 * @return interchange reference.
	 */
	public String getIref () {
		return iref;
	}

	/**
	 * Returns sender of interchange.
	 *
	 * @return sender of interchange.
	 */
	public String getSender () {
		return sender;
	}

	/**
	 * Returns receiver of interchange.
	 *
	 * @return receiver of interchange.
	 */
	public String getReceiver () {
		return receiver;
	}

	/**
	 * Returns preparation date and time
	 * in the format YYMMDD:hhmm.
	 *
	 * @return preparation date and time.
	 */
	public String getPreparationDateTime () {
		return preparationTimestampToString (
			preparationDate
		);
	}
	
	/**
	 * Tests if this interchange is for testing purposes.
	 *
	 * @return true if interchange is for testing purposes.
	 */
	public boolean isTest () {
		return isTest;
	}

	/**
	 * Returns the interchange reference number.
	 *
	 * @return the IREF number or {@link #INVALID_IREF}
	 * if the IREF is invalid.
	 */
	public int getIrefNo () {
		// If not yet computed (or accidentally zero)
		if ( irefNo == 0 ) {
			try {
				irefNo = Integer.parseInt(
					iref.substring( 4 )
				);
			} catch ( IndexOutOfBoundsException e ) {
				irefNo = INVALID_IREF;
			} catch ( NumberFormatException e ) {
				irefNo = INVALID_IREF;
			}
		}
		return irefNo;
	}

	/**
	 * Returns the preparation date.
	 *
	 * @return the preparation date
	 *
	 */
	public Date getPreparationDateTimeAsDate () {
		return preparationDate;
	}



	/* ------------------ Other Public Methods ------------------ */

	/**
	 * Returns the hash code value for this InterchangeBegin object.
	 * Equality according to {@link #equals} implies that
	 * the hashcodes are the same
	 * as required by the general contract of {@link Object#hashCode}.
	 *
	 * @return hash code value for this object.
	 */
	public int hashCode () {
		// if not yet computed (or accidentally zero)
		if ( hashCode == 0 ) {
			hashCode = iref.hashCode()
				^ sender.hashCode()
				^ receiver.hashCode()
				^ ((null == preparationDate) ? 0 : preparationDate.hashCode())
				//TODO: Bernhard is this good coding?
				^ ( isTest ? 1231 : 1237 );
				// These are the values used by Boolean.hashCode().
		}
		return hashCode;
	}

	/**
	 * Indicates whether some other InterchangeBegin object
	 * is "equal to" this one.
	 *
	 * @param obj the other reference object with which to compare
	 *
	 * @return true if this InterchangeBegin object
	 * is equal to the argument; false otherwise.
	 */
	public boolean equals ( Object obj ) {

		if ( this == obj ) {
			return true;
		} else if ( obj instanceof InterchangeBegin ) {
			InterchangeBegin other = (InterchangeBegin) obj;

			return
				iref.equals( other.iref )
				&& sender.equals( other.sender )
				&& receiver.equals( other.receiver )
				&& (
				   (null != preparationDate
				       && preparationDate.equals( other.preparationDate )
				   )
				   ||
				   (null == preparationDate && null == other.preparationDate)
				)
				&& isTest == other.isTest;
		} else {
			return false;
		}
	}

	/**
	 * Returns a string representation of this InterchangeBegin.
	 * This method is intended to be used only for debugging purposes,
	 * and the format of the returned string may vary between
	 * implementations. The returned string may be empty but may not be null.
	 *
	 * @return a string representation of this InterchangeBegin.
	 */
	public String toString () {
		return getClass().getName() + "("
			+ "iref=" + iref + ", "
			+ "sender=" + sender + ", "
			+ "receiver=" + receiver + ", "
			+ "preparationDate=" + preparationDate + ", "
			+ "isTest=" + String.valueOf(isTest) + ")";
	}

	/* ==================== Private Methods ===================== */

	private static String preparationTimestampToString (
		final Date prepared
	) {
		if ( null == prepared ) return "";
		return new SimpleDateFormat( "yyMMdd:HHmm" ).format( prepared );
	}
	
	private static Date preparationTimestampToDate (
		final String preparationTimestamp
	) {
		final String dateFormat = "yyMMdd:HHmm";
		final SimpleDateFormat formatter
		= new SimpleDateFormat( dateFormat );
		formatter.setLenient( false );
		final ParsePosition pos = new ParsePosition( 0 );
		Date date = formatter.parse( preparationTimestamp, pos );
		if ( pos.getIndex() < dateFormat.length() ) {
			date = null; // null if parsing failed.
		}
		return date;
	}

}
