// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb;

import java.util.Date;

import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.text.ParsePosition;
import java.text.ParseException;

import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;

/* ***************************** Class ****************************** */

/**
 * <P>Tests the implementation of {@link InterchangeBegin}.</P>
 *
 * @author Sami Airo
 * @author
 * <a href="mailto:sis.external@ecb.int">Bernhard Bodenstorfer</a>
 *
 * @version @(#)$Id: InterchangeBeginTest.java,v 1.1 2017/01/19 16:14:04 trainot Exp $
 */

public class InterchangeBeginTest extends TestCase {

	/* ====================== Constructors ====================== */

	/**
	 * Constructs an <CODE>InterchangeBeginTest</CODE> with the specified name.
	 *
	 * @param name Test case name
	 */
	public InterchangeBeginTest ( String name ) {
		super(name);
	}

	/* ===================== Public Methods ===================== */

	/**
	 * Assembles and returns a test suite for all
	 * the test methods of this test case.
	 *
	 * @return a non-null test suite
	 */
	public static Test suite () {
		TestSuite suite = new TestSuite( InterchangeBeginTest.class );
		return suite;
	}

	/**
	 * Runs the test case
	 */
	public static void main ( String args[] ) {
		String[] testCaseName = { InterchangeBeginTest.class.getName() };
		junit.textui.TestRunner.main( testCaseName );
	}

	/* ---------------------- Test Methods ---------------------- */

	/** Test the date-based utility constructor. */
	public void testDatebasedConstructor () throws ParseException {
		final DateFormat format
		= new SimpleDateFormat( "yyyy-MM-dd'T'HH:mm:ss" );
		assertEquals(
			"Preparation timestamp set",
			new InterchangeBegin(
				"IREF000001", "DE2", "4F0", "000315:1245", true
			),
			new InterchangeBegin(
				"IREF000001", "DE2", "4F0",
				format.parse( "2000-03-15T12:45:00" ),
				true
			)
		);
	}

	/**
	 * Tests the equals and hashcode methods method
	 */
	public void testEquals () {
		InterchangeBegin o1 = new InterchangeBegin(
			"IREF0001", "DE2", "4F0", "000315:1245", false
		);
		InterchangeBegin o2 = new InterchangeBegin(
			"IREF0001", "DE2", "4F0", "000315:1245", false
		);
		InterchangeBegin o3 = new InterchangeBegin(
			"IREF0001", "FI2", "4F0", "000315:1245", false
		);

		assertTrue("o1 equals o2", o1.equals(o2));
		assertTrue("o1 not equals o3", !o1.equals(o3));
		assertTrue("o1 hashcode == o2.hashcode", o1.hashCode() == o2.hashCode());
		assertTrue("o1 hashcode != o3.hashcode", o1.hashCode() != o3.hashCode());
	}

	/**
	 * Tests the normal fields
	 */
	public void testNormalFields () {
		InterchangeBegin o1 = new InterchangeBegin(
			"IREF0001", "DE2", "4F0", "000315:1245", false
		);

		assertEquals("DE2", o1.getSender());
		assertEquals("4F0", o1.getReceiver());
		assertEquals("IREF0001", o1.getIref());
		assertEquals("000315:1245", o1.getPreparationDateTime());
		assertTrue(!o1.isTest());
	}

	/**
	 * Tests the computed fields
	 */
	public void testComputedFields () {
		SimpleDateFormat formatter = new SimpleDateFormat("yyMMddHHmm");
		Date date = formatter.parse("0003151245", new ParsePosition(0));
		InterchangeBegin o1 = new InterchangeBegin(
			"IREF0001", "DE2", "4F0", "000315:1245", false
		);

		assertNotNull(o1.getPreparationDateTimeAsDate());
		assertEquals("Date", date, o1.getPreparationDateTimeAsDate());
		assertEquals("IREF", 1, o1.getIrefNo());

		try {
			o1 = new InterchangeBegin(
				"IREF0001", "DE2", "4F0", "000315:1299", false
			);
		} catch (Throwable t) {
			fail("exception thrown from constructor with invalid date: " + t);
		}

		assertNull("000315:1299 -> null", o1.getPreparationDateTimeAsDate());

		try {
			o1 = new InterchangeBegin(
				"IREFXXXX", "DE2", "4F0", "000315:1245", false
			);
		} catch (Throwable t) {
			fail("exception thrown from constructor with invalid IREF: " + t);
		}

		assertEquals("INVALID_IREF", o1.getIrefNo(), InterchangeBegin.INVALID_IREF);
	}

	/**
	 * Tests serialization
	 */
	public void testSerialization () {
		final InterchangeBegin ib1 = new InterchangeBegin(
			"IREF0001", "DE2", "4F0", "000315:1245", false
		);
		final InterchangeBegin transserialized
		= (InterchangeBegin) AttributeTest.transserialize( ib1 );
		assertEquals( ib1, transserialized );
		assertEquals( ib1.hashCode(), transserialized.hashCode() );
		assertEquals( ib1.getIrefNo(), transserialized.getIrefNo() );
	}
}
