// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb;

import java.lang.String;
import java.lang.Integer;

import Int.ecb.gesmescb.InterchangeBegin;

/* ***************************** Class ****************************** */

/**
 * Holds the GESMES/TS interchange trailer information.
 *
 * @author
 * <A HREF="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</A>
 *
 * @version @(#)$Id: InterchangeEnd.java,v 1.1 2017/01/19 16:14:04 trainot Exp $
 */

public final class InterchangeEnd extends GESMESCBObject {

	/* ======================= Constants ======================== */

	/**
	 *   serialVersionUID generated by eclipse.
	 */
	private static final long serialVersionUID = 7455115554672874918L;

	/**
	 * Value returned by {@link #getIrefNo} if the
	 * interchange reference number is in an invalid format.
	 */
	public static final int INVALID_IREF = InterchangeBegin.INVALID_IREF;

	/* ========================= Fields ========================= */

	/**
	 * Interchange reference identifyer.
	 *
	 * @serial
	 */
	private final String iref;

	/* ------------------- Transient Fields -------------------- */

	/**
	 * Interchange reference number.
	 */
	private transient int irefNo;

	/**
	 * Precomputed hashcode.
	 */
	private transient int hashCode;

	/* ====================== Constructors ====================== */

	/**
	 * Construct a new interchange.
	 *
	 * @param iref interchange reference number.
	 *
	 * @throws NullPointerException upon a <CODE>null</CODE> parameter.
	 */
	public InterchangeEnd (
		final String iref
	) {
		if ( ( this.iref = iref ) == null )
			throw new NullPointerException(
				"Initializing InterchangeEnd with null iref"
			);
	}

	/* ===================== Public Methods ===================== */

	/**
	 * Returns interchange reference.
	 *
	 * @return interchange reference.
	 */
	public String getIref () {
		return iref;
	}

	/**
	 * Returns the interchange reference number.
	 *
	 * @return the IREF number or {@link #INVALID_IREF}
	 * if the IREF is invalid.
	 */
	public int getIrefNo () {
		// If not yet computed (or accidentally zero)
		if ( irefNo == 0 ) {
			try {
				irefNo = Integer.parseInt(
					iref.substring( 4 )
				);
			} catch ( IndexOutOfBoundsException e ) {
				irefNo = INVALID_IREF;
			} catch ( NumberFormatException e ) {
				irefNo = INVALID_IREF;
			}
		}
		return irefNo;
	}

	/**
	 * Returns whether the <CODE>InterchangeEnd</CODE> object corresponds
	 * to the given <CODE>InterchangeBegin</CODE> object.
	 *
	 * @return whether this <CODE>InterchangeEnd</CODE>
	 * can be considered to end the given
	 * <CODE>InterchangeBegin</CODE>.
	 */
	public boolean isEndFor(
		final InterchangeBegin interchangeBegin
	) {
		return iref.equals( interchangeBegin.getIref() );
	}

	/* ------------------ Other Public Methods ------------------ */

	/**
	 * Returns the hash code value for this InterchangeEnd object.
	 * Equality according to {@link #equals} implies that
	 * the hashcodes are the same
	 * as required by the general contract of {@link Object#hashCode}.
	 *
	 * @return hash code value for this object.
	 */
	public int hashCode () {
		// if not yet computed (or accidentally zero)
		if ( hashCode == 0 ) {
			hashCode = iref.hashCode();
		}
		return hashCode;
	}

	/**
	 * Indicates whether some other InterchangeEnd object
	 * is "equal to" this one.
	 *
	 * @param obj the other reference object with which to compare
	 *
	 * @return true if this InterchangeEnd object
	 * is equal to the argument; false otherwise.
	 */
	public boolean equals ( Object obj ) {

		if ( this == obj ) {
			return true;
		} else if ( obj instanceof InterchangeEnd ) {
			InterchangeEnd other = (InterchangeEnd) obj;

			return
				iref.equals( other.iref );
		} else {
			return false;
		}
	}

	/**
	 * Returns a string representation of this InterchangeEnd.
	 * This method is intended to be used only for debugging purposes,
	 * and the format of the returned string may vary between
	 * implementations. The returned string may be empty but may not be null.
	 *
	 * @return a string representation of this InterchangeEnd.
	 */
	public String toString () {
		return getClass().getName() + "("
			+ "iref=" + iref + ")";
	}
}
