// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, 2004, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import Int.ecb.gesmescb.util.KeyTokenizer;
import Int.ecb.logging.Level;
import Int.ecb.logging.Logger;
import Int.ecb.prefs.Preferences;

/* ***************************** Class ****************************** */

/**
 * <P>The class KeyFamily holds the components for one key family.
 * The objects of this class are immutable and they cannot be changed after
 * construction.</P>
 *
 * <P>Each key family holds two collections: components and key structure. The
 * key structure is a subset of the components collection than contains all
 * the dimensions in the key structure.</P>
 *
 * @author
 * <A HREF="mailto:klo@FAME.com">Kam Lo</A>
 * @author
 * <A HREF="mailto:Sami.Airo@ECB.int">Sami Airo</A>
 * @author
 * <A HREF="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</A>
 *
 * @version @(#)$Id: KeyFamily.java,v 1.1 2017/01/19 16:14:04 trainot Exp $
 */
public class KeyFamily extends AbstractStructuralDefinition {

	/* ======================= Constants ======================== */

	/**
	 *   serialVersionUID generated by eclipse.
	 */
	private static final long serialVersionUID = -4419012355459597708L;

	/**
	 * A constant empty array to support type checking in the call
	 * to the method List.toArray( KeyFamilyComponent[] ).
	 * Only the type KeyFamilyComponent[] is relevant.
	 *
	 * The size must be 0, otherwise instance data would be written
	 * to the static array by List.toArray!
	 *
	 * Note that in case of empty keys
	 * a reference to the static array itself
	 * becomes part of the instance.
	 * But empty is anyway empty.
	 */
	private static final KeyFamilyComponent[] emptyKeyFamilyComponentArray
	= new KeyFamilyComponent[0];

	/** Whether to sort key family components.
	 * Note that this setting also influences the equals() method.
	 * If the setting is true, the order of components passed to
	 * the constructor is regarded as being significant.
	 * Use the "orderedComponents" option only if you know what you are doing.
	 */
	private static final boolean ORDERED_COMPONENTS;

	/* =================== Static Initializer =================== */

	static {
		final String className = KeyFamily.class.getName();
		final Logger logger = Logger.getLogger( className );
		final Preferences
			userPrefs = Preferences.userRoot().node( className ),
			systemPrefs = Preferences.systemRoot().node( className );
		final String orderedComponents
		= systemPrefs.get(
			"orderedComponents",
			userPrefs.get( "orderedComponents", "false" )
		);
		ORDERED_COMPONENTS = "true".equalsIgnoreCase( orderedComponents );
		logger.log(
			Level.FINE,
			"KeyFamily",
			"static initialiser",
			"Component order significant: " + ORDERED_COMPONENTS
		);
	}

	/* ========================= Fields ========================= */

	/**
	 * Key family components.
	 *
	 * @serial
	 */
	private final KeyFamilyComponent[] components;

	/* -------------------- Transient Fields -------------------- */

	/**
	 * Flag whether the transient fields have been initialized.
	 * This helps serialization.
	 * Note that initializedTransients directly after having read from
	 * a {@link java.io.ObjectInput} object is false.
	 */
	private transient boolean initializedTransients;

	/**
	 * Mapping of the conceptIds to components.
	 * This is computed from components.
	 */
	private transient Map componentMap;

	/**
	 * An array holding only the key structure for the key family.
	 * It is sorted by the natural order.
	 */
	private transient KeyFamilyComponent[] keyStructure;

	/**
	 * An array holding only the time dimensions for the key family.
	 * It is sorted by the natural order.
	 */
	private transient KeyFamilyComponent[] timeDimensionStructure;

	/**
	 * An array holding only the array cell components for the key family.
	 * It is sorted by the natural order.
	 */
	private transient KeyFamilyComponent[] arrayCellStructure;

	/** An array holding the attributes for the key family. */
	private transient KeyFamilyComponent[] attributes;

	/** An array holding the data set attributes for the key family. */
	private transient KeyFamilyComponent[] dataSetAttributes;

	/** An array holding the sibling group attributes for the key family. */
	private transient KeyFamilyComponent[] groupAttributes;

	/** An array holding the time series attributes for the key family. */
	private transient KeyFamilyComponent[] seriesAttributes;

	/** An array holding the observation attributes for the key family. */
	private transient KeyFamilyComponent[] observationAttributes;

	/**
	 * Position of the frequency dimension.
	 * This is computed from components.
	 */
	private transient int frequencyDimensionPosition;

	/** The frequency or null if none. */
	private transient KeyFamilyComponent frequency;

	/**
	 * Precomputed hash code.
	 */
	private transient int hashCode;

	/* ====================== Constructors ====================== */

	/**
	 * Constructs a new key family.
	 * The component list is guaranteed to be copied.
	 *
	 * @param identifier		Key family identifier.
	 * @param name		Key family name.
	 * @param description	Key family description.
	 * @param versionDate	concept version timestamp.
	 * @param components	key family component list.
	 *
	 * @throws ArrayStoreException if components
	 * contains objects of a different class than
	 * {@link KeyFamilyComponent}.
	 */
	public KeyFamily (
		final String identifier,
		final String name,
		final String description,
		final Date versionDate,
		final List components
	) {
		super( identifier, name, description, versionDate );

		this.components = (KeyFamilyComponent[])components.toArray(
			emptyKeyFamilyComponentArray
		);
		if ( ! ORDERED_COMPONENTS ) {
			Arrays.sort( this.components );
		}

		// Compute the derived transient fields.
		initTransients();
	}

	/**
	 * Constructs a new key family.
	 * The component array is guaranteed to be copied.
	 *
	 * @param identifier		Key family identifier.
	 * @param name		Key family name.
	 * @param description	Key family description.
	 * @param versionDate	concept version timestamp.
	 * @param components	key family component array.
	 */
	public KeyFamily (
		final String identifier,
		final String name,
		final String description,
		final Date versionDate,
		final KeyFamilyComponent[] components
	) {
		super( identifier, name, description, versionDate );

		this.components = (KeyFamilyComponent[]) components.clone();
		if ( ! ORDERED_COMPONENTS ) {
			Arrays.sort( this.components );
		}

		// Compute the derived transient fields.
		initTransients();
	}

	/* ===================== Public Methods ===================== */

	/**
	 * Returns the component for a concept identifier.
	 * Returns null if the key family contains no component
	 * for this concept identifier.
	 * If there are duplicate components in the code list,
	 * then the return value is one of them.
	 *
	 * @param conceptId concept identifier.
	 *
	 * @return component with the concept identifier.
	 */
	public synchronized KeyFamilyComponent getComponent ( String conceptId ) {
		// synchronized for Java VMM.
		if ( !initializedTransients ) initTransients();
		return (KeyFamilyComponent) componentMap.get( conceptId );
	}

	/**
	 * Returns the component count for this key family.
	 *
	 * @return component count.
	 */
	public int getComponentCount () {
		return components.length;
	}

	/**
	 * Returns an iterator over the {@link KeyFamilyComponent}s
	 * in this <CODE>KeyFamily</CODE>.
	 *
	 * @return an iterator over the {@link KeyFamilyComponent}s.
	 */
	public Iterator componentIterator () {
		return Collections.unmodifiableList(
			Arrays.asList( components )
		).iterator();
	}

	/**
	 * Returns the key structure dimension count for this key family.
	 *
	 * @return the number of key structure dimensions.
	 */
	public synchronized int getKeyDimensionCount () {
		// synchronized for Java VMM.
		if ( !initializedTransients ) initTransients();
		return keyStructure.length;
	}

	/**
	 * Returns an iterator over the key structure
	 * of this <CODE>KeyFamily</CODE>.
	 * The iterator reports the key structure dimensions
	 * sorted by increasing position.
	 *
	 * @return an iterator over the key structure.
	 */
	public synchronized Iterator keyStructureIterator () {
		// synchronized for Java VMM.
		if ( !initializedTransients ) initTransients();
		return Collections.unmodifiableList(
			Arrays.asList( keyStructure )
		).iterator();
	}

	/** Returns the frequency dimension. */
	public synchronized KeyFamilyComponent getFrequency () {
		if ( !initializedTransients ) initTransients();
		return frequency;
	}

	/**
	 * Returns the number of time dimensions.
	 * This should always be 2.
	 */
	public synchronized int getTimeDimensionCount () {
		// synchronized for Java VMM.
		if ( !initializedTransients ) initTransients();
		return timeDimensionStructure.length;
	}

	/** Returns the time dimension. */
	public synchronized KeyFamilyComponent getTimePeriod () {
		if ( !initializedTransients ) initTransients();
		return timeDimensionStructure.length >= 1
			? timeDimensionStructure[0] : null;
	}

	/** Returns the time format concept. */
	public synchronized KeyFamilyComponent getTimeFormat () {
		if ( !initializedTransients ) initTransients();
		return timeDimensionStructure.length >= 2
			? timeDimensionStructure[1] : null;
	}

	/**
	 * Returns an iterator over the time dimensions.
	 * The iterator sorts by increasing position.
	 */
	public synchronized Iterator timeDimensionIterator () {
		// synchronized for Java VMM.
		if ( !initializedTransients ) initTransients();
		return Collections.unmodifiableList(
			Arrays.asList( timeDimensionStructure )
		).iterator();
	}

	/**
	 * Returns the number of array cells for this key family.
	 * This should always be 4.
	 *
	 * @return the number of number of array cells dimensions.
	 */
	public synchronized int getArrayCellCount () {
		// synchronized for Java VMM.
		if ( !initializedTransients ) initTransients();
		return arrayCellStructure.length;
	}

	/**
	 * Returns an iterator over the array cells
	 * of this <CODE>KeyFamily</CODE>.
	 * The iterator reports the array cells
	 * sorted by increasing position.
	 *
	 * @return an iterator over the array cell structure.
	 */
	public synchronized Iterator arrayCellIterator () {
		// synchronized for Java VMM.
		if ( !initializedTransients ) initTransients();
		return Collections.unmodifiableList(
			Arrays.asList( arrayCellStructure )
		).iterator();
	}

	/** Returns the first array cell (observation value). */
	public synchronized KeyFamilyComponent getObsValue () {
		if ( !initializedTransients ) initTransients();
		return arrayCellStructure.length >= 1
			? arrayCellStructure[0] : null;
	}

	/** Returns the second array cell (observation status). */
	public synchronized KeyFamilyComponent getObsStatus () {
		if ( !initializedTransients ) initTransients();
		return arrayCellStructure.length >= 2
			? arrayCellStructure[1] : null;
	}

	/** Returns the third array cell (observation confidentiality). */
	public synchronized KeyFamilyComponent getObsConf () {
		if ( !initializedTransients ) initTransients();
		return arrayCellStructure.length >= 3
			? arrayCellStructure[2] : null;
	}

	/** Returns the fourth array cell (observation pre-break value). */
	public synchronized KeyFamilyComponent getObsPreBreak () {
		if ( !initializedTransients ) initTransients();
		return arrayCellStructure.length >= 4
			? arrayCellStructure[3] : null;
	}

	/**
	 * Returns an iterator over the attributes
	 * of this <code>KeyFamily</code>.
	 *
	 * @return an iterator over the attributes.
	 */
	public synchronized Iterator attributeIterator () {
		// synchronized for Java VMM.
		if ( !initializedTransients ) initTransients();
		return Collections.unmodifiableList(
			Arrays.asList( attributes )
		).iterator();
	}

	/**
	 * Returns an iterator over the attributes
	 * for a given attachment level of this <code>KeyFamily</code>.
	 *
	 * @param attachmentLevel must be one of
	 * {@link KeyFamilyComponent#ATTACHED_TO_DATA_SET},
	 * {@link KeyFamilyComponent#ATTACHED_TO_SIBLING_GROUP},
	 * {@link KeyFamilyComponent#ATTACHED_TO_TIME_SERIES}, or
	 * {@link KeyFamilyComponent#ATTACHED_TO_OBSERVATION}.
	 * @return an iterator over attributes.
	 */
	public synchronized Iterator attributeIterator (
		final String attachmentLevel
	) {
		// synchronized for Java VMM.
		if ( !initializedTransients ) initTransients();
		final KeyFamilyComponent[] attributeArray;
		if (
			KeyFamilyComponent.ATTACHED_TO_DATA_SET.equals( attachmentLevel )
		) {
			attributeArray = dataSetAttributes;
		} else if (
			KeyFamilyComponent.ATTACHED_TO_SIBLING_GROUP.equals( attachmentLevel )
		) {
			attributeArray = groupAttributes;
		} else if (
			KeyFamilyComponent.ATTACHED_TO_TIME_SERIES.equals( attachmentLevel )
		) {
			attributeArray = seriesAttributes;
		} else if (
			KeyFamilyComponent.ATTACHED_TO_OBSERVATION.equals( attachmentLevel )
		) {
			attributeArray = observationAttributes;
		} else {
			throw new IllegalArgumentException(
				"Invalid attachment level code \"" + attachmentLevel + "\"."
			);
		}
		return Collections.unmodifiableList(
			Arrays.asList( attributeArray )
		).iterator();
	}

	/**
	 * Returns the position of the frequency dimension in the key structure.
	 * First position in GESMES/TS is 1.
	 *
	 * @return the position of the frequency dimension in the key structure.
	 */
	public synchronized int getFrequencyDimensionPosition () {
		// synchronized for Java VMM.
		if ( !initializedTransients ) initTransients();
		return frequencyDimensionPosition;
	}

	/** Compute the sibling group key to a given <code>key</code>.
	 * If the key is empty (i.e. a data set key)
	 * or a sibling group key, itself is returned.
	 *
	 * @return the sibling group key for the given key.
	 */
	public String computeSiblingGroupKey (
		final String key
	) {
		return computeSiblingGroupKeyOrFrequencyValue( key, true );
	}

	/** Compute the frequency value to a given <code>key</code>.
	 * If the key is empty (i.e. a data set key)
	 * or a sibling group key, the empty string is returned.
	 *
	 * @return the frequency dimension value for the given key.
	 */
	public String computeFrequencyValue (
		final String key
	) {
		return computeSiblingGroupKeyOrFrequencyValue( key, false );
	}

	/**
	 * Returns the hash code value for this key family.
	 * If kf1.equals(kf2) then kf1.hashCode()==kf2.hashCode()
	 * for any two KeyFamilies kf1 and kf2,
	 * as required by the general contract of {@link Object#hashCode}.
	 *
	 * @return hash code value for this object.
	 */
	public int hashCode () {
		// if not yet computed (or accidentally zero)
		if ( hashCode == 0 ) {
			hashCode = super.hashCode()
				^ Arrays.asList( components ).hashCode();
		}
		return hashCode;
	}

	/**
	 * Indicates whether some other key family is "equal to" this one.
	 *
	 * @param object the reference object with which to compare
	 *
	 * @return true if this KeyFamily is the same as the other argument;
	 * false otherwise.
	 */
	public boolean equals ( final Object object ) {
		if ( this == object ) return true;
		if ( object instanceof KeyFamily ) {
			final KeyFamily other = (KeyFamily) object;
			return
				super.equals( other )
				&& Arrays.equals( components, other.components);
		}
		return false;
	}

	/**
	 * Returns a String representation of the <CODE>KeyFamily</CODE>.
	 */
	public String toString() {
		// Build a string for the components.
		StringBuffer stringBuffer = new StringBuffer();
		for (
			int i = 0;
			i < components.length;
			++i
		) {
			if ( i > 0 ) stringBuffer.append( ", " );
			stringBuffer.append( components[i].toString() );
		}

		return getClass().getName()
			+ " ( identifier=" + getId() + ", "
			+ "name=" + getName() + ", "
			+ "description=" + getDescription() + ", "
			+ "versionDate=" + getVersionDate()
			+ ", { " + stringBuffer.toString() + " } )";
	}

	/* ==================== Private Methods ===================== */

	/**
	 * Derive the transient fields from the object's state.
	 * Since the serialized state does not contain transient fields,
	 * they must be computed upon request.
	 * Finally, the flag initializedTransients is set true
	 * to avoid unnecessarily repeated computation of the fields.
	 */
	private synchronized void initTransients () {
		componentMap = new HashMap();
		frequencyDimensionPosition = 0;
		final List keyDimensions = new ArrayList();
		final List timeDimensions = new ArrayList();
		final List arrayCellDimensions = new ArrayList();
		final List attributeList = new ArrayList();
		final List dataSetAttributeList = new ArrayList();
		final List groupAttributeList = new ArrayList();
		final List seriesAttributeList = new ArrayList();
		final List observationAttributeList = new ArrayList();
		frequency = null;

		// Iterate through all components and put them into the
		// map for fast lookup.
		// If they are dimensions,
		// then also put them into the key structure.
		for ( int n = 0; n < components.length; n++ ) {
			final KeyFamilyComponent component = components[n];

			// Everything into the lookup map.
			componentMap.put(
				component.getConceptId(), component
			);

			if ( component.isKeyStructureDimension() ) {
				// dimensions into the key structure.
				keyDimensions.add( component );

				// Compute the frequencyDimensionPosition
				// on the fly.
				if ( component.isFrequencyDimension() ) {
					frequency = component;
					try {
						frequencyDimensionPosition
						= Integer.parseInt( component.getPosition() );
					}
					catch ( NumberFormatException nfe ) {
						// Leave zero (or previous value)
					}
				}
			} else if ( component.isTimeDimension() ) {
				timeDimensions.add( component );
			} else if ( component.isArrayCell() ) {
				arrayCellDimensions.add( component );
			} else if ( component.isAttribute() ) {
				attributeList.add( component );
				if ( component.isAttachedToDataSet() ) {
					dataSetAttributeList.add( component );
				} else if ( component.isAttachedToSiblingGroup() ) {
					groupAttributeList.add( component );
				} else if ( component.isAttachedToTimeSeries() ) {
					seriesAttributeList.add( component );
				} else if ( component.isAttachedToObservation() ) {
					observationAttributeList.add( component );
				}
			}
		}
		keyStructure = (KeyFamilyComponent[])keyDimensions.toArray(
			emptyKeyFamilyComponentArray
		);
		Arrays.sort( keyStructure );
		timeDimensionStructure = (KeyFamilyComponent[])timeDimensions.toArray(
			emptyKeyFamilyComponentArray
		);
		Arrays.sort( timeDimensionStructure );
		arrayCellStructure = (KeyFamilyComponent[])arrayCellDimensions.toArray(
			emptyKeyFamilyComponentArray
		);
		Arrays.sort( arrayCellStructure );
		attributes = (KeyFamilyComponent[]) attributeList.toArray(
			emptyKeyFamilyComponentArray
		);
		Arrays.sort( attributes );
		dataSetAttributes = (KeyFamilyComponent[]) dataSetAttributeList.toArray(
			emptyKeyFamilyComponentArray
		);
		Arrays.sort( dataSetAttributes );
		groupAttributes = (KeyFamilyComponent[]) groupAttributeList.toArray(
			emptyKeyFamilyComponentArray
		);
		Arrays.sort( groupAttributes );
		seriesAttributes = (KeyFamilyComponent[]) seriesAttributeList.toArray(
			emptyKeyFamilyComponentArray
		);
		Arrays.sort( seriesAttributes );
		observationAttributes = (KeyFamilyComponent[]) observationAttributeList.toArray(
			emptyKeyFamilyComponentArray
		);
		Arrays.sort( observationAttributes );

		// Finally set the flag that transients have been computed.
		initializedTransients = true;
	}

	/**
	 * @param chooseSiblingGroupKey compute sibling group key if true,
	 * otherwise frequency value.
	 * @return the sibling group key or the frequency value for the given key.
	 */
	private String computeSiblingGroupKeyOrFrequencyValue (
		final String key, final boolean chooseSiblingGroupKey
	) {
		final StringBuffer choosenKey = new StringBuffer( key.length() );
		final Iterator
			dimensions = keyStructureIterator(),
			counter = keyStructureIterator(),
			dimensionValues = new KeyTokenizer( key );
		// There is a problem with attributes with sibling attachement level.
		// Their key contains no dimension.
		// Not knowing the attachment level, we have to resort to the trick 
		// of checking whether the length of the key is in line with the number
		// of dims of the key family.
		final String[] list = key.split( "[" + TimeSeries.KEY_SEPARATOR + "]" );
		int dimensionSize = 0;
		while ( counter.hasNext() ) {
			counter.next();
			dimensionSize++;
		}
		
		while ( dimensions.hasNext() && dimensionValues.hasNext() ) {
			KeyFamilyComponent dimension
			= (KeyFamilyComponent) dimensions.next();

			if ( chooseSiblingGroupKey && dimension.isFrequencyDimension()
					&& list.length < dimensionSize
			) {
				dimension = (KeyFamilyComponent) dimensions.next();
			}
			final String value = (String) dimensionValues.next();
			if ( chooseSiblingGroupKey ^ dimension.isFrequencyDimension() ) {
				choosenKey.append( value );
				choosenKey.append( TimeSeries.KEY_SEPARATOR );
			}
		}
		if ( choosenKey.length() > 0 ) {
			choosenKey.setLength( choosenKey.length() - 1 );
		}
		return choosenKey.toString();
	}
}
