// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb;

import java.text.ParseException;

import java.io.Serializable;

/* ***************************** Class ****************************** */

/**
 * <P>Holds one component of a GESMES/TS key family.
 * It is used together with {@link KeyFamily} to defined one GESMES/TS
 * key family. This class is used for all component types: structure dimensions,
 * time dimensions, array cells, and attributes. Some of the fields like
 * usageStatus only apply to one type of component and not others.</P>
 *
 * <P>The objects of this class are immutable so they cannot be changed after
 * construction.</P>
 *
 * @see KeyFamily
 *
 * @author
 * <A HREF="mailto:klo@FAME.com">Kam Lo</A>
 * @author
 * <A HREF="mailto:Sami.Airo@ECB.int">Sami Airo</A>
 * @author
 * <A HREF="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</A>
 *
 * @version @(#)$Id: KeyFamilyComponent.java,v 1.2 2018/03/09 12:27:10 kortest Exp $
 */

public final class KeyFamilyComponent implements Serializable, Comparable
{
	/* ======================= Constants ======================== */

	/**
	 *   serialVersionUID generated by eclipse.
	 */
	private static final long serialVersionUID = 2805037901259028145L;
	/** Concept type code. */
	public static final String TIME_DIMENSION = "1";
	public static final String ARRAY_CELL = "3";
	/** @deprecated use {@link #ARRAY_CELL} instead. */
	public static final String ARRAY_CELL_DIMENSION = ARRAY_CELL;
	public static final String KEY_STRUCTURE_DIMENSION = "4";
	public static final String FREQUENCY_DIMENSION = "13";
	public static final String ATTRIBUTE = "Z09";

	/** Attachment level code. */
	public static final String ATTACHED_TO_DATA_SET = "1";
	public static final String ATTACHED_TO_TIME_SERIES = "4";
	public static final String ATTACHED_TO_OBSERVATION = "5";
	public static final String ATTACHED_TO_SIBLING_GROUP = "9";
	public static final String ATTACHED_TO_DIMENSIONS = "8";
	/** @deprecated use {@link #ATTACHED_TO_DATA_SET} instead. */
	public static final String ATTACHED_TO_KEY_FAMILY = ATTACHED_TO_DATA_SET;

	/** Usage status code. */
	public static final String CONDITIONAL = "1";
	public static final String MANDATORY = "2";

	/**
	 * Value returned by {@link #getPositionNo} if the
	 * position is not available or in an invalid format.
	 */
	public static final int INVALID_POSITION = -1;

	/* ========================= Fields ========================= */

	/**
	 * Statistical concept identifier.
	 *
	 * @serial
	 */
	private final String conceptId;

	/**
	 * Statistical concept type.
	 *
	 * @see #getConceptType
	 *
	 * @serial
	 */
	private final String conceptType;

	/**
	 * Position of the concept in the key family structure.
	 *
	 * @serial
	 */
	private final String position;

	/**
	 * Format specification for the code values.
	 * This tells length and type (numeric/alphanumeric)
	 * of the code values.
	 *
	 * @serial
	 */
	private final String format;

	/**
	 * Usage status of an attribute.
	 *
	 * @see #getUsageStatus
	 *
	 * @serial
	 */
	private final String usageStatus;

	/**
	 * Attachment level of an attribute.
	 * This determines where an attribute applies to.
	 *
	 * @see #getAttachmentLevel
	 *
	 * @serial
	 */
	private final String attachmentLevel;

	/**
	 * Code list identifier of the applicable code list.
	 *
	 * @serial
	 */
	private final String codeListId;

	/**
	 * Support Gesmes/TS frequency dimension type.
	 *
	 * @serial
	 */
	private final boolean isGesmesTS;

	/* ------------------- Transient Fields -------------------- */

	/** Precomputed text format. */
	private transient TextFormat textFormat;

	/**
	 * Precomputed hash code.
	 */
	private transient int hashCode;

	/* ====================== Constructors ====================== */

	/**
	 * <p>Constructs a Gesmes/TS key family component with specified fields.</p>
	 *
	 * <p>The parameters cannot be null references.
	 * If the component is dimension and
	 * thus the attachment level does not apply
	 * an empty string should be used.</p>
	 *
	 * <p>For the old Gesmes/TS format use the constructor
	 * {@link #KeyFamilyComponent(String,String,String,String,String,String,String)}
	 * which replaces the improved flagging of the frequency dimension
	 * by a heuristics.</p>
	 *
	 * @param conceptId statistical concept identifier.
	 * @param conceptType statistical concept type.
	 * @param position of the concept in the key structure.
	 * @param format restricting text.
	 * @param usageStatus of an attribute:
	 * {@link #MANDATORY} or {@link #CONDITIONAL}.
	 * @param attachmentLevel of an attribute.
	 * @param codeListId code list identifier.
	 * @param isGesmesTS whether to rely on type {@link #FREQUENCY_DIMENSION}
	 * for identifying the frequency dimension.
	 */
	public KeyFamilyComponent (
		final String conceptId,
		final String conceptType,
		final String position,
		final String format,
		final String usageStatus,
		final String attachmentLevel,
		final String codeListId,
		final boolean isGesmesTS
	) {
		// Set the internal status fields.
		this.conceptId = conceptId;
		this.conceptType = conceptType;
		this.position = position;
		this.format = format;
		this.usageStatus = usageStatus;
		this.codeListId = codeListId;
		this.attachmentLevel = attachmentLevel;
		this.isGesmesTS = isGesmesTS;
	}

	/**
	 * <p>Constructs a Gesmes/TS key family component with specified fields.</p>
	 *
	 * <p>The parameters cannot be null references.
	 * If the component is dimension and
	 * thus the attachment level does not apply
	 * an empty string should be used.</p>
	 *
	 * <p>For Gesmes/TS use the constructor
	 * {@link #KeyFamilyComponent(String,String,String,String,String,String,String,boolean)}
	 * in order to take advantage of the better flagging of
	 * the frequency dimension.</p>
	 *
	 * @param conceptId		statistical concept identifier.
	 * @param conceptType		statistical concept type.
	 * @param position		position of the concept
	 *				in the key family structure.
	 * @param format		format specification for
	 *				the code values.
	 * @param usageStatus		usage status of an attribute.
	 * @param attachmentLevel	attachment level of an attribute.
	 * @param codeListId		code list identifier.
	 */
	public KeyFamilyComponent (
		final String conceptId,
		final String conceptType,
		final String position,
		final String format,
		final String usageStatus,
		final String attachmentLevel,
		final String codeListId
	) {
		// Set the internal status fields.
		this.conceptId = conceptId;
		this.conceptType = conceptType;
		this.position = position;
		this.format = format;
		this.usageStatus = usageStatus;
		this.codeListId = codeListId;
		this.attachmentLevel = attachmentLevel;
		this.isGesmesTS = false;
	}

	/* ===================== Public Methods ===================== */

	/**
	 * Returns statistical concept identifier.
	 *
	 * @return the statistical concept identifier.
	 */
	public String getConceptId () {
		return conceptId;
	}

	/**
	 * Returns statistical concept type.
	 *
	 * <P><CENTER><TABLE BORDER=3 CELLPADDING=3>
	 * <caption>GESMES/TS concept type codes</caption>
	 * <tr><td>{@link #FREQUENCY_DIMENSION}</td>
	 * <td>frequency (in Gesmes/TS messages)</td></tr>
	 * <TR><TD>{@link #ARRAY_CELL_DIMENSION}</TD>
	 * <TD>observation value</TD></TR>
	 * <tr><td>{@link #TIME_DIMENSION}</td>
	 * <td>time</td></tr>
	 * <TR><TD>{@link #KEY_STRUCTURE_DIMENSION}</TD>
	 * <TD>key</TD></TR>
	 * <TR><TD>{@link #ATTRIBUTE}</TD>
	 * <TD>attribute value</TD></TR>
	 * </TABLE></CENTER></P>
	 *
	 * @return the statistical concept type.
	 */
	public String getConceptType () {
		return conceptType;
	}

	/**
	 * Returns position of the concept in the key family structure.
	 *
	 * @return the position of the concept in the key family structure.
	 */
	public String getPosition () {
		return position;
	}

	/**
	 * Returns the format specification for the code values.
	 * For example, this could be <CODE>AN..35</CODE>.
	 *
	 * @return format specification for the code values.
	 *
	 * @see #getTextFormat()
	 */
	public String getFormat () {
		return format;
	}

	/**
	 * Returns the usage status of an attribute.
	 *
	 * <P><CENTER><TABLE BORDER=3 CELLPADDING=3>
	 * <caption>GESMES/TS usage status codes</caption>
	 * <TR><TD>{@link #MANDATORY}</TD><TD>conditional</TD></TR>
	 * <TR><TD>{@link #CONDITIONAL}</TD><TD>mandatory</TD></TR>
	 * </TABLE></CENTER></P>
	 *
	 * @return usage status of an attribute.
	 */
	public String getUsageStatus () {
		return usageStatus;
	}

	/**
	 * Returns the attachment level of an attribute.
	 *
	 * <P><CENTER><TABLE BORDER=3 CELLPADDING=3>
	 * <caption>GESMES/TS attachment level codes</caption>
	 * <TR><TH>Code</TH><TH>means attachment to</TH></TR>
	 * <TR><TD>{@link #ATTACHED_TO_DATA_SET}</TD>
	 * <TD>data set</TD></TR>
	 * <TR><TD>{@link #ATTACHED_TO_TIME_SERIES}</TD>
	 * <TD>time series</TD></TR>
	 * <TR><TD>{@link #ATTACHED_TO_OBSERVATION}</TD>
	 * <TD>observation</TD></TR>
	 * <TR><TD>{@link #ATTACHED_TO_SIBLING_GROUP}</TD>
	 * <TD>sibling group</TD></TR>
	 * </TABLE></CENTER></P>
	 *
	 * @return attachment level of an attribute.
	 */
	public String getAttachmentLevel () {
		return attachmentLevel;
	}

	/**
	 * Returns the code list identifier.
	 *
	 * @return code list identifier.
	 */
	public String getCodeListId () {
		return codeListId;
	}

	/**
	 * Tests if this key family component is a dimension
	 * in the key structure.
	 *
	 * @return true if this key family component is a dimension
	 * in the enclosing key structure; false otherwise.
	 */
	public boolean isKeyStructureDimension () {
		return
			KEY_STRUCTURE_DIMENSION.equals( conceptType )
			||
			FREQUENCY_DIMENSION.equals( conceptType );
	}

	/**
	 * Tests if this key family component is a time dimension.
	 *
	 * @return true if this key family component is a time dimension;
	 * false otherwise.
	 */
	public boolean isTimeDimension () {
		return conceptType.equals(TIME_DIMENSION);
	}

	/** @deprecated Use {@link #isArrayCell()} instead. */
	public boolean isArrayCellDimension () {
		return isArrayCell();
	}

	/**
	 * Tests if this key family component is an array cell.
	 *
	 * @return true if this key family component is a array cell;
	 * false otherwise.
	 */
	public boolean isArrayCell () {
		return conceptType.equals( ARRAY_CELL );
	}

	/**
	 * Tests if this key family component is the frequency dimension.
	 *
	 * @return true if this key family component is the frequency dimension;
	 * false otherwise.
	 */
	public boolean isFrequencyDimension () {
		return
			FREQUENCY_DIMENSION.equals( conceptType )
			||
			!isGesmesTS
			&&
			KEY_STRUCTURE_DIMENSION.equals( conceptType )
			&&
			"1".equals( position );
	}

	/**
	 * Tests if this key family component is an attribute.
	 *
	 * @return true if this key family component is an attribute;
	 * false otherwise.
	 */
	public boolean isAttribute () {
		return conceptType.equals(ATTRIBUTE);
	}

	/**
	 * Tests if this key family component is an coded.
	 *
	 * @return true if this key family component is coded;
	 * false if it is uncoded.
	 */
	public boolean isCoded () {
		return !codeListId.equals("");
	}

	/**
	 * Tests if this key family component is mandatory.
	 *
	 * @return true if this key family component is mandatory;
	 * false otherwise.
	 */
	public boolean isMandatory () {
		return usageStatus.equals(MANDATORY);
	}

	/**
	 * Tests if the specified format of this this key family component
	 * admits only numeric codes.
	 *
	 * @return true if the format specifies only numeric codes;
	 * false otherwise.
	 */
	public boolean isNumeric () {
		return format.startsWith("N");
	}

	/**
	 * Tests if the format of this this key family component
	 * specifies fixed length codes.
	 *
	 * @return true if the format specifies fixed length codes;
	 * false otherwise.
	 */
	public boolean isFixedLength () {
		return format.indexOf(".") == -1;
	}

	/** @deprecated Use {@link #isAttachedToDataSet()} instead. */
	public boolean isAttachedToKeyFamily () {
		return isAttachedToDataSet();
	}

	/**
	 * Tests if this key family component is attached to the data set.
	 *
	 * @return true if this key family component is attached to the data set;
	 * false otherwise.
	 */
	public boolean isAttachedToDataSet () {
		return attachmentLevel.equals( ATTACHED_TO_DATA_SET );
	}

	/**
	 * Tests if this key family component is attached to sibling group.
	 *
	 * @return true if this key family component is attached
	 * to sibling group; false otherwise.
	 */
	public boolean isAttachedToSiblingGroup () {
		return attachmentLevel.equals( ATTACHED_TO_SIBLING_GROUP );
	}

	/**
	 * Tests if this key family component is attached to time series.
	 *
	 * @return true if this key family component is attached to time series;
	 * false otherwise.
	 */
	public boolean isAttachedToTimeSeries () {
		return attachmentLevel.equals( ATTACHED_TO_TIME_SERIES );
	}

	/**
	 * Tests if this key family component is attached to observation.
	 *
	 * @return true if this key family component
	 * is attached to observation; false otherwise.
	 */
	public boolean isAttachedToObservation () {
		return attachmentLevel.equals( ATTACHED_TO_OBSERVATION );
	}

	/**
	 * Return {@link #getPosition} as <code>int</code>.
	 */
	public int getPositionNo () {
		final String posi = getPosition();
		int pos = INVALID_POSITION;
		if ( !"".equals( posi ) ) {
			try {
				pos = Integer.parseInt( posi );
			}
			catch ( final NumberFormatException nfe ) {}
		}
		return pos;
	}

	/**
	 * Return {@link #getFormat} expression as <code>TextFormat</code>.
	 *
	 * @return <code>null</code> if the format cannot be parsed.
	 */
	public synchronized TextFormat getTextFormat () {
		if ( textFormat == null ) {
			try {
				textFormat = TextFormat.parse( format );
			}
			catch ( final ParseException pe ) {}
		}
		return textFormat;
	}

	/**
	 * Compares this <CODE>KeyFamilyComponent</CODE>
	 * with the specified KeyFamilyComponent for order.
	 * The ordering is primarily based on the position field.
	 * This orders the dimensions according to their position.
	 * Next, ordering is done by ascending attachment level,
	 * starting from Observation level.
	 * Then ordering is by the concept identifier.
	 *
	 * @param obj the KeyFamilyComponent to be compared.
	 *
	 * @return a negative integer, zero, or a positive integer as this
	 * object is less than, equal to, or greater than the specified object.
	 *
	 * @throws java.lang.ClassCastException if obj is no KeyFamilyComponent
	 * in agreement with the contract of {@link Comparable#compareTo}.
	 */
	public int compareTo ( Object obj ) {

		// This may throw a ClassCastException.
		KeyFamilyComponent other = (KeyFamilyComponent) obj;

		int c;

		// Order the components by their position
		// - attributes have an empty position
		//   but they should be ordered last
		// - positions are strings
		//   but they should be orderer numerically ("2" < "11").
		if (
			position.length() != 0 && other.position.length() == 0
		) {
			return -1;
		} else if (
			position.length() == 0 && other.position.length() != 0
		) {
			return 1;
		} else {
			int thisPosition = 0, thatPosition = 0;
			boolean thisIllegal, thatIllegal;
			try {
				thisPosition = Integer.parseInt( position );
				thisIllegal = false;
			} catch ( NumberFormatException nfe ) {
				thisIllegal = true;
			}
			try {
				thatPosition = Integer.parseInt( other.position );
				thatIllegal = false;
			} catch ( NumberFormatException nfe ) {
				thatIllegal = true;
			}
			// Legal ones are smaller.
			if ( thisIllegal ) {
				if ( thatIllegal ) {
					c = position.compareTo( other.position );
				} else {
					return 1;
				}
			} else {
				if ( thatIllegal ) {
					return -1;
				} else {
					c = thisPosition - thatPosition;
				}
			}
		}
		if ( c != 0 ) return c;

		if ( !attachmentLevel.equals( other.attachmentLevel ) ) {
			if ( isAttachedToObservation() ) return -1;
			if ( isAttachedToDataSet() ) return 1;
			if ( other.isAttachedToObservation() ) return 1;
			if ( other.isAttachedToDataSet() ) return -1;
			if ( isAttachedToTimeSeries() ) return -1;
			if ( isAttachedToSiblingGroup() ) return 1;
		}

		c = conceptId.compareTo( other.conceptId );
		if ( c != 0 ) return c;

		c = conceptType.compareTo( other.conceptType );
		if ( c != 0 ) return c;

		c = format.compareTo( other.format );
		if ( c != 0 ) return c;

		c = usageStatus.compareTo( other.usageStatus );
		if ( c != 0 ) return c;

		c = codeListId.compareTo( other.codeListId );
		if ( c != 0 ) return c;

		return ( isGesmesTS ? 1 : 0 ) - ( other.isGesmesTS ? 1 : 0 );
	}

	/**
	 * Indicates whether some other KeyFamilyComponent
	 * is "equal to" this one.
	 *
	 * @param obj the object with which to compare.
	 *
	 * @return true if this KeyFamilyComponent is the same
	 * as the other argument; false otherwise.
	 */
	public boolean equals ( final Object obj ) {
		return
			this == obj
			||
				obj instanceof KeyFamilyComponent
				&&
				compareTo( obj ) == 0;
	}

	/**
	 * Returns the hash code value for this KeyFamilyComponent.
	 * If kfc1.equals(kfc2) then kfc1.hashCode()==kfc2.hashCode()
	 * for any two KeyFamilyComponents
	 * kfc1 and kfc2, as required by the general contract of
	 * {@link Object#hashCode}.
	 *
	 * @return a hash code value for this object.
	 */
	public int hashCode () {
		// if not yet computed (or accidentally zero)
		if ( hashCode == 0 ) {
			hashCode = conceptId.hashCode()
				^ conceptType.hashCode()
				^ position.hashCode()
				^ format.hashCode()
				^ usageStatus.hashCode()
				^ attachmentLevel.hashCode()
				^ codeListId.hashCode()
				^ (
					isGesmesTS
					? Boolean.TRUE.hashCode()
					: Boolean.FALSE.hashCode()
				);
		}
		return hashCode;
	}

	/**
	 * Returns a String representation of the KeyFamilyComponent.
	 */
	public String toString() {
		return getClass().getName()
			+ "(conceptId=" + conceptId
			+ ", conceptType=" + conceptType
			+ ", position=" + position
			+ ", format=" + format
			+ ", usageStatus=" + usageStatus
			+ ", attachmentLevel=" + attachmentLevel
			+ ", codeListId=" + codeListId
			+ ", isGesmes/TS=" + isGesmesTS
			+ ")";
	}
}
