// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb;

import java.text.ParseException;
import java.text.ParsePosition;
import java.text.SimpleDateFormat;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.Date;
import java.util.Iterator;

import Int.ecb.gesmescb.edifact.RegularIntervalFormat;
import Int.ecb.gesmescb.time.Interval;
import Int.ecb.gesmescb.time.RegularInterval;

/* ***************************** Class ****************************** */

/**
 * Holds the GESMES/TS message header and optionally data set header
 * information.
 *
 * @author
 * <A HREF="mailto:klo@FAME.com">Kam Lo</A>
 * @author
 * <A HREF="mailto:Sami.Airo@ECB.int">Sami Airo</A>
 * @author
 * <A HREF="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</A>
 *
 * @version @(#)$Id: MessageBegin.java,v 1.1 2017/01/19 16:14:04 trainot Exp $
 */

public final class MessageBegin extends GESMESCBObject {

	/* ========================= Constants ====================== */

	/* ------------------- Message Constants -------------------- */

	/**
	 *   serialVersionUID generated by eclipse.
	 */
	private static final long serialVersionUID = -6834200097190083301L;

	/**
	 * Value of the function field indicating structural definitions.
	 */
	public static final String STATISTICAL_DEFINITIONS = "73";

	/**
	 * Value of the function field indicating statistical data.
	 */
	public static final String STATISTICAL_DATA = "74";

	/**
	 * Value of the function field indicating data request.
	 */
	public static final String DATA_REQUEST = "75";

	/**
	 * Value returned by {@link #getMrefNo} if the
	 * message reference number is in an invalid format.
	 */
	public static final int INVALID_MREF = -1;

	/* ---------------- Message Sender Constants ---------------- */

	/**
	 * A constant empty array to support type checking in the call
	 * to the method Collection.toArray( Contact[] ).
	 * Only the type Contact[] is relevant.
	 *
	 * The size must be 0, otherwise instance data would be written
	 * to the static array by Collection.toArray!
	 *
	 * Note that in case of empty contact list,
	 * a reference to the static array itself
	 * becomes part of the instance.
	 * But empty is anyway empty.
	 */
	private static final Contact[] EMPTY_CONTACT_ARRAY = new Contact[0];

	/* ------------------- Data Set Constants ------------------- */

	/**
	 * Value of the action field indicating update or replace.
	 */
	public static final String UPDATE_OR_REPLACE_ACTION = "7";

	/**
	 * Value of the action field indicating delete.
	 */
	public static final String DELETE_ACTION = "6";

	/**
	 * Value of the action field indicating testing.
	 */
	public static final String TESTING_ACTION = "0";
	
	/**
	 * Value of the action field indicating object level revision.
	 */
	public static final String OBJECT_LEVEL_ACTION = "8";

	/* ========================= Fields ========================= */

	/* --------------------- Message Header --------------------- */

	/**
	 * Message reference identifier.
	 *
	 * @serial
	 */
	private final String mref;

	/**
	 * Alledged sender of the message.
	 *
	 * @serial
	 */
	private final String sender;

	/**
	 * Intended receiver of the message.
	 *
	 * @serial
	 */
	private final String receiver;

	/**
	 * Maintenance agency.
	 *
	 * @serial
	 */
	private final String maintenanceAgency;

	/**
	 * Message function.
	 *
	 * @serial
	 */
	private final String function;

	/* --------------------- Message Sender --------------------- */

	/**
	 * Identity of the message.
	 * This is a kind of free text subject as stated by the sender.
	 *
	 * @serial
	 */
	private final String identity;

	/**
	 * Sender contacts.
	 *
	 * @serial
	 */
	private final Contact[] contacts;

	/* -------------------- Data Set Header --------------------- */

	/**
	 * Data set identifier.
	 * This makes explicit to which data set the subsequent
	 * manipulations will apply.
	 *
	 * @serial
	 */
	private final String dataSetId;

	/**
	 * Data set identifier.
	 * Keeps track of the original dataSetId after conversions.
	 * 
	 * @serial
	 */
	private final String originalDataSetId;

	/**
	 * Data for update, delete or testing.
	 * <P><CENTER><TABLE BORDER=3 CELLPADDING=3>
	 * <caption>GESMES/TS data contents codes</caption>
	 * <TR><TD>6</TD><TD>delete</TD></TR>
	 * <TR><TD>7</TD><TD>update or replace</TD></TR>
	 * <TR><TD>8</TD><TD>update or replace on object level</TD></TR>
	 * <TR><TD>0</TD><TD>provisional test value</TD></TR>
	 * </TABLE></CENTER></P>
	 *
	 * @serial
	 */
	private final String action;

	/**
	 * Extraction date and time.
	 *
	 * @serial
	 */
	private final Date extractionDate;

	/**
	 * Reporting period.
	 *
	 * @serial
	 */
	private final String reportingPeriod;

	/**
	 * Reporting period format.
	 *
	 * @serial
	 */
	private final String reportingPeriodFormat;

	/**
	 * key family identifier.
	 *
	 * @serial
	 */
	private final String keyFamilyId;

	/* ------------------- Transient Fields -------------------- */

	/**
	 * Message reference number.
	 */
	private transient int mrefNo;

	/**
	 * Precomputed hashcode.
	 */
	private transient int hashCode;

	/* ====================== Constructors ====================== */

	/**
	 * Construct a new <CODE>MessageBegin</CODE> from message header
	 * and data set header information with a <CODE>Collection</CODE> of
	 * <CODE>Contact</CODE>s.
	 *
	 * @param mref message reference.
	 * @param sender asserted sender of message.
	 * @param receiver intended receiver of message.
	 * @param maintenanceAgency maintenance agency.
	 * @param function message function.
	 * @param identity of the message. This is a free text subject line.
	 * @param contacts to contact the sender.
	 * @param dataSetId data set identifier.
	 * @param action for update, delete or provisional testing.
	 * @param extractionDateTime extraction date and time.
	 * @param reportingPeriod reporting period.
	 * @param reportingPeriodFormat reporting period format.
	 * @param keyFamilyId specifies the keyFamily embracing the dataset.
	 *
	 * @throws NullPointerException if any of the parameters is null.
	 */
	public MessageBegin (
		final String mref,
		final String sender,
		final String receiver,
		final String maintenanceAgency,
		final String function,
		final String identity,
		final Collection contacts,
		final String dataSetId,
		final String action,
		final String extractionDateTime,
		final String reportingPeriod,
		final String reportingPeriodFormat,
		final String keyFamilyId
	)
		throws NullPointerException
	{
		this(
			mref,
			sender,
			receiver,
			maintenanceAgency,
			function,
			identity,
			contacts,
			dataSetId,
			action,
			extractedToDate(extractionDateTime),
			reportingPeriod,
			reportingPeriodFormat,
			keyFamilyId,
			dataSetId
		);
	}
	/**
	 * Construct a new <CODE>MessageBegin</CODE> from message header
	 * and data set header information with a <CODE>Collection</CODE> of
	 * <CODE>Contact</CODE>s.
	 *
	 * @param mref message reference.
	 * @param sender asserted sender of message.
	 * @param receiver intended receiver of message.
	 * @param maintenanceAgency maintenance agency.
	 * @param function message function.
	 * @param identity of the message. This is a free text subject line.
	 * @param contacts to contact the sender.
	 * @param dataSetId data set identifier.
	 * @param action for update, delete or provisional testing.
	 * @param extractionDateTime extraction date and time.
	 * @param reportingPeriod reporting period.
	 * @param reportingPeriodFormat reporting period format.
	 * @param keyFamilyId specifies the keyFamily embracing the dataset.
	 * @param originalDataSetId the dataSetId before any conversion.
	 * @throws NullPointerException if any of the parameters is null.
	 */
	public MessageBegin (
		final String mref,
		final String sender,
		final String receiver,
		final String maintenanceAgency,
		final String function,
		final String identity,
		final Collection contacts,
		final String dataSetId,
		final String action,
		final String extractionDateTime,
		final String reportingPeriod,
		final String reportingPeriodFormat,
		final String keyFamilyId,
		final String originalDataSetId
	)
		throws NullPointerException
	{
		this(
			mref,
			sender,
			receiver,
			maintenanceAgency,
			function,
			identity,
			contacts,
			dataSetId,
			action,
			extractedToDate(extractionDateTime),
			reportingPeriod,
			reportingPeriodFormat,
			keyFamilyId,
			originalDataSetId
		);
	}
	
	/**
	 * Construct a new <CODE>MessageBegin</CODE> from message header
	 * and data set header information with array of
	 * <CODE>Contact</CODE>s.
	 *
	 * @param mref message reference.
	 * @param sender asserted sender of message.
	 * @param receiver intended receiver of message.
	 * @param maintenanceAgency maintenance agency.
	 * @param function message function.
	 * @param identity of the message. This is a free text subject line.
	 * @param contacts to contact the sender.
	 * @param dataSetId data set identifier.
	 * @param action for update, delete or provisional testing.
	 * @param extractionDateTime extraction date and time.
	 * @param reportingPeriod reporting period.
	 * @param reportingPeriodFormat reporting period format.
	 * @param keyFamilyId specifies the keyFamily embracing the dataset.
	 *
	 * @throws NullPointerException if any of the parameters is null.
	 */
	public MessageBegin (
		final String mref,
		final String sender,
		final String receiver,
		final String maintenanceAgency,
		final String function,
		final String identity,
		final Contact[] contacts,
		final String dataSetId,
		final String action,
		final String extractionDateTime,
		final String reportingPeriod,
		final String reportingPeriodFormat,
		final String keyFamilyId
	)
		throws NullPointerException
	{
		this(
			mref,
			sender,
			receiver,
			maintenanceAgency,
			function,
			identity,
			contacts,
			dataSetId,
			action,
			extractedToDate(extractionDateTime),
			reportingPeriod,
			reportingPeriodFormat,
			keyFamilyId,
			dataSetId
		);
	}
	/**
	 * Construct a new <CODE>MessageBegin</CODE> from message header
	 * and data set header information with array of
	 * <CODE>Contact</CODE>s.
	 *
	 * @param mref message reference.
	 * @param sender asserted sender of message.
	 * @param receiver intended receiver of message.
	 * @param maintenanceAgency maintenance agency.
	 * @param function message function.
	 * @param identity of the message. This is a free text subject line.
	 * @param contacts to contact the sender.
	 * @param dataSetId data set identifier.
	 * @param action for update, delete or provisional testing.
	 * @param extractionDateTime extraction date and time.
	 * @param reportingPeriod reporting period.
	 * @param reportingPeriodFormat reporting period format.
	 * @param keyFamilyId specifies the keyFamily embracing the dataset.
	 * @param originalDataSetId the dataSetId before any conversion took place
	 *
	 * @throws NullPointerException if any of the parameters is null.
	 */
	public MessageBegin (
		final String mref,
		final String sender,
		final String receiver,
		final String maintenanceAgency,
		final String function,
		final String identity,
		final Contact[] contacts,
		final String dataSetId,
		final String action,
		final String extractionDateTime,
		final String reportingPeriod,
		final String reportingPeriodFormat,
		final String keyFamilyId,
		final String originalDataSetId
	)
		throws NullPointerException
	{
		this(
			mref,
			sender,
			receiver,
			maintenanceAgency,
			function,
			identity,
			contacts,
			dataSetId,
			action,
			extractedToDate(extractionDateTime),
			reportingPeriod,
			reportingPeriodFormat,
			keyFamilyId,
			originalDataSetId
		);
	}

	/**
	 * Construct a new <code>MessageBegin</code> from message header
	 * and data set header information with a <code>Collection</code> of
	 * <code>Contact</code>s and reporting period as interval.
	 *
	 * @param mref message reference.
	 * @param sender asserted sender of message.
	 * @param receiver intended receiver of message.
	 * @param maintenanceAgency maintenance agency.
	 * @param function message function.
	 * @param identity of the message. This is a free text subject line.
	 * @param contacts to contact the sender.
	 * @param dataSetId data set identifier.
	 * @param action for update, delete or provisional testing.
	 * @param extracted extraction time stamp.
	 * @param reportingInterval reporting interval.
	 * @param keyFamilyId specifies the keyFamily embracing the dataset.
	 *
	 * @throws NullPointerException if any of the parameters is null.
	 */
	public MessageBegin (
		final String mref,
		final String sender,
		final String receiver,
		final String maintenanceAgency,
		final String function,
		final String identity,
		final Collection contacts,
		final String dataSetId,
		final String action,
		final Date extracted,
		final RegularInterval reportingInterval,
		final String keyFamilyId
	)
		throws NullPointerException
	{
		this(
			mref,
			sender,
			receiver,
			maintenanceAgency,
			function,
			identity,
			contacts,
			dataSetId,
			action,
			extracted,
			intervalToPeriod( reportingInterval ),
			intervalToPeriodFormat( reportingInterval ),
			keyFamilyId,
			dataSetId
		);
	}

	/**
	 * Construct a new <code>MessageBegin</code> from message header
	 * and data set header information with a <code>Collection</code> of
	 * <code>Contact</code>s and reporting period as interval.
	 *
	 * @param mref message reference.
	 * @param sender asserted sender of message.
	 * @param receiver intended receiver of message.
	 * @param maintenanceAgency maintenance agency.
	 * @param function message function.
	 * @param identity of the message. This is a free text subject line.
	 * @param contacts to contact the sender.
	 * @param dataSetId data set identifier.
	 * @param action for update, delete or provisional testing.
	 * @param extracted extraction time stamp.
	 * @param reportingInterval reporting interval.
	 * @param keyFamilyId specifies the keyFamily embracing the dataset.
	 * @param originalDataSetId data set id before any conversion took place.
	 *
	 * @throws NullPointerException if any of the parameters is null.
	 */
	public MessageBegin (
		final String mref,
		final String sender,
		final String receiver,
		final String maintenanceAgency,
		final String function,
		final String identity,
		final Collection contacts,
		final String dataSetId,
		final String action,
		final Date extracted,
		final RegularInterval reportingInterval,
		final String keyFamilyId,
		final String originalDataSetId
	)
		throws NullPointerException
	{
		this(
			mref,
			sender,
			receiver,
			maintenanceAgency,
			function,
			identity,
			contacts,
			dataSetId,
			action,
			extracted,
			intervalToPeriod( reportingInterval ),
			intervalToPeriodFormat( reportingInterval ),
			keyFamilyId,
			originalDataSetId
		);
	}
	
	/**
	 * Construct a new <code>MessageBegin</code> from message header
	 * and data set header information with array of
	 * <code>Contact</code>s and reporting period as interval.
	 *
	 * @param mref message reference.
	 * @param sender asserted sender of message.
	 * @param receiver intended receiver of message.
	 * @param maintenanceAgency maintenance agency.
	 * @param function message function.
	 * @param identity of the message. This is a free text subject line.
	 * @param contacts to contact the sender.
	 * @param dataSetId data set identifier.
	 * @param action for update, delete or provisional testing.
	 * @param extractionDateTime extraction date and time.
	 * @param extracted extraction time stamp.
	 * @param reportingInterval reporting interval.
	 * @param keyFamilyId specifies the keyFamily embracing the dataset.
	 *
	 * @throws NullPointerException if any of the parameters is null.
	 */
	public MessageBegin (
		final String mref,
		final String sender,
		final String receiver,
		final String maintenanceAgency,
		final String function,
		final String identity,
		final Contact[] contacts,
		final String dataSetId,
		final String action,
		final Date extracted,
		final RegularInterval reportingInterval,
		final String keyFamilyId
	)
		throws NullPointerException
	{
		this(
			mref,
			sender,
			receiver,
			maintenanceAgency,
			function,
			identity,
			contacts,
			dataSetId,
			action,
			extracted,
			intervalToPeriod( reportingInterval ),
			intervalToPeriodFormat( reportingInterval ),
			keyFamilyId,
			dataSetId
		);
	}

	/**
	 * Construct a new <CODE>MessageBegin</CODE> from message header
	 * information with a <CODE>Collection</CODE> of
	 * <CODE>Contact</CODE>s.
	 *
	 * @param mref message reference.
	 * @param sender asserted sender of message.
	 * @param receiver intended receiver of message.
	 * @param maintenanceAgency maintenance agency.
	 * @param function message function.
	 * @param identity of the message. This is a free text subject line.
	 * @param contacts to contact the sender.
	 *
	 * @throws NullPointerException if any of the parameters is null.
	 */
	public MessageBegin (
		final String mref,
		final String sender,
		final String receiver,
		final String maintenanceAgency,
		final String function,
		final String identity,
		final Collection contacts
	)
		throws NullPointerException
	{
		this(
			mref,
			sender,
			receiver,
			maintenanceAgency,
			function,
			identity,
			contacts,
			"", "", "", "", "", "",	// No data set info.
			""                      // Thus no original data set info
		);
	}

	/**
	 * Construct a new <CODE>MessageBegin</CODE> from message header
	 * information with array of <CODE>Contact</CODE>s.
	 *
	 * @param mref message reference.
	 * @param sender asserted sender of message.
	 * @param receiver intended receiver of message.
	 * @param maintenanceAgency maintenance agency.
	 * @param function message function.
	 * @param identity of the message. This is a free text subject line.
	 * @param contacts to contact the sender.
	 *
	 * @throws NullPointerException if any of the parameters is null.
	 */
	public MessageBegin (
		final String mref,
		final String sender,
		final String receiver,
		final String maintenanceAgency,
		final String function,
		final String identity,
		final Contact[] contacts
	)
		throws NullPointerException
	{
		this(
			mref,
			sender,
			receiver,
			maintenanceAgency,
			function,
			identity,
			contacts,
			"", "", "", "", "", "",	// No data set info.
			""                      // Thus no original data set info
		);
	}

	/**
	 * Construct a new <CODE>MessageBegin</CODE> from message header
	 * and data set header information.
	 * In contrast to the public constructors, this one takes an Object type
	 * reference to the contacts.
	 *
	 * @param mref message reference.
	 * @param sender asserted sender of message.
	 * @param receiver intended receiver of message.
	 * @param maintenanceAgency maintenance agency.
	 * @param function message function.
	 * @param identity of the message. This is a free text subject line.
	 * @param contacts to contact the sender.
	 * @param dataSetId data set identifier.
	 * @param action for update, delete or provisional testing.
	 * @param extractionDateTime extraction date and time.
	 * @param reportingPeriod reporting period.
	 * @param reportingPeriodFormat reporting period format.
	 * @param keyFamilyId specifies the keyFamily embracing the dataset.
	 *
	 * @throws NullPointerException if any of the parameters is null.
	 */
	public MessageBegin (
		final String mref,
		final String sender,
		final String receiver,
		final String maintenanceAgency,
		final String function,
		final String identity,
		final Object contacts,
		final String dataSetId,
		final String action,
		final Date extractionDate,
		final String reportingPeriod,
		final String reportingPeriodFormat,
		final String keyFamilyId,
		final String originalDataSetId
	)
		throws NullPointerException
	{
		if ( ( this.mref = mref ) == null )
			throw new NullPointerException(
				"Initializing MessageBegin with null mref"
			);

		if ( ( this.sender = sender ) == null )
			throw new NullPointerException(
				"Initializing MessageBegin with null sender"
			);

		if ( ( this.receiver = receiver ) == null )
			throw new NullPointerException(
				"Initializing MessageBegin with null receiver"
			);

		if ( ( this.maintenanceAgency = maintenanceAgency ) == null )
			throw new NullPointerException(
				"Initializing MessageBegin with null maintenanceAgency"
			);

		if ( ( this.function = function ) == null )
			throw new NullPointerException(
				"Initializing MessageBegin with null function"
			);

		if ( ( this.identity = identity ) == null )
			throw new NullPointerException(
				"Initializing MessageBegin with null identity"
			);

		if ( contacts == null ) {
			throw new NullPointerException(
				"Initializing MessageBegin with null contacts"
			);
		}

		if ( contacts instanceof Contact[] ) {
			this.contacts = (Contact[])( (Contact[])contacts ).clone();
		} else if ( contacts instanceof Collection ) {
			this.contacts = (Contact[])(
				( (Collection)contacts ).toArray( EMPTY_CONTACT_ARRAY )
			);
		} else {
			throw new IllegalArgumentException(
				"Initializing MessageBegin with wrong type contacts"
			);
		}

		if ( ( this.dataSetId = dataSetId ) == null )
			throw new NullPointerException(
				"Initializing MessageBegin with null dataSetId"
			);

		if ( ( this.action = action ) == null )
			throw new NullPointerException(
				"Initializing MessageBegin with null action"
			);

		this.extractionDate = extractionDate;
		
		if ( ( this.reportingPeriod = reportingPeriod ) == null )
			throw new NullPointerException(
				"Initializing MessageBegin with null reportingPeriod"
			);

		if ( ( this.reportingPeriodFormat = reportingPeriodFormat ) == null )
			throw new NullPointerException(
				"Initializing MessageBegin with null reportingPeriodFormat"
			);
		if ( ( this.keyFamilyId = keyFamilyId ) == null )
			throw new NullPointerException(
				"Initializing MessageBegin with null keyFamilyId"
			);

		if ( ( this.originalDataSetId = originalDataSetId ) == null )
			throw new NullPointerException(
				"Initializing MessageBegin with null originalDataSetId"
			);
	
}

	/* ===================== Public Methods ===================== */

	/* -------------------- Message Methods --------------------- */

	/**
	 * Returns message reference.
	 *
	 * @return message reference.
	 */
	public String getMref () {
		return mref;
	}

	/**
	 * Returns sender of message.
	 *
	 * @return sender of message.
	 */
	public String getSender () {
		return sender;
	}

	/**
	 * Returns receiver of message.
	 *
	 * @return receiver of message.
	 */
	public String getReceiver () {
		return receiver;
	}

	/**
	 * Returns maintenance agency identifier.
	 *
	 * @return maintenance agency identifier.
	 */
	public String getMaintenanceAgencyId () {
		return maintenanceAgency;
	}

	/**
	 * Returns message function.
	 *
	 * @return message function.
	 */
	public String getFunction () {
		return function;
	}

	/**
	 * Returns the MREF number.
	 *
	 * @return the MREF number or {@link #INVALID_MREF}
	 * if the MREF is invalid.
	 */
	public int getMrefNo () {
		// if mrefNo not yet initialized (or accidentally zero)
		if ( mrefNo == 0 ) {
			try {
				mrefNo = Integer.parseInt(
					mref.substring( 4 )
				);
			} catch ( IndexOutOfBoundsException e ) {
				mrefNo = INVALID_MREF;
			} catch ( NumberFormatException e ) {
				mrefNo = INVALID_MREF;
			}
		}
		return mrefNo;
	}

	/**
	 * Tests if the function is statistical definitions.
	 *
	 * @return true if the function is statistical definitions;
	 * false otherwise.
	 */
	public boolean isStatisticalDefinitions () {
		return function.equals( STATISTICAL_DEFINITIONS );
	}

	/**
	 * Tests if the function is statistical data.
	 *
	 * @return true if the function is statistical data;
	 * false otherwise.
	 */
	public boolean isStatisticalData () {
		return function.equals( STATISTICAL_DATA );
	}

	/**
	 * Tests if the function is data request.
	 *
	 * @return true if the function is data request;
	 * false otherwise.
	 */
	public boolean isDataRequest () {
		return function.equals( DATA_REQUEST );
	}

	/* --------------------- Sender Methods --------------------- */

	/**
	 * Returns identity of the message.
	 *
	 * @return identity of the message.
	 */
	public String getIdentity () {
		return identity;
	}

	/**
	 * Returns an iterator to the message contacts.
	 *
	 * @return iterator to the message contacts.
	 */
	public Iterator contactIterator () {
		return (
			Collections.unmodifiableList(
				Arrays.asList( contacts )
			)
		).iterator();
	}

	/**
	 * Returns an collection of contacts.
	 *
	 * @return collection of the message contacts.
	 */
	public Collection contactAsCollection () {
		return Arrays.asList( contacts );
	}	
	
	/* -------------------- Data Set Methods -------------------- */

	/**
	 * Returns the data set identifier.
	 *
	 * @return data set id.
	 */
	public String getDataSetId () {
		return dataSetId;
	}

	/**
	 * Returns the data set identifier.
	 *
	 * @return data set id.
	 */
	public String getOriginalDataSetId () {
		return originalDataSetId;
	}
	
	/**
	 * Returns the action to be performed: update/delete/provisional test.
	 *
	 * @return action to be performed.
	 */
	public String getAction () {
		return action;
	}

	/**
	 * Returns extraction date and time in the format "YYYYMMDDhhmm".
	 *
	 * @return extraction date and time.
	 */
	public String getExtractionDateTime () {
		return extractedToPeriod(extractionDate);
	}

	/**
	 * Returns the reporting period.
	 *
	 * @return reporting period.
	 */
	public String getReportingPeriod () {
		return reportingPeriod;
	}

	/**
	 * Returns the reporting format.
	 *
	 * @return reporting period format.
	 */
	public String getReportingPeriodFormat () {
		return reportingPeriodFormat;
	}

	/**
	 * Returns the key family identifier.
	 *
	 * @return key family id.
	 */
	public String getKeyFamilyId () {
		return keyFamilyId;
	}

	/**
	 * Tests if the action is update or replace
	 *
	 * @return true if the action is update or replace;
	 * false otherwise.
	 */
	public boolean isUpdateOrReplaceAction () {
		return
			action.equals(UPDATE_OR_REPLACE_ACTION)
			||
			action.equals(OBJECT_LEVEL_ACTION)
		;
	}

	/**
	 * Tests if the action is delete
	 *
	 * @return true if the action is delete;
	 * false otherwise.
	 */
	public boolean isDeleteAction () {
		return action.equals(DELETE_ACTION);
	}

	/**
	 * Tests if the action is testing
	 *
	 * @return true if the action is testing;
	 * false otherwise.
	 */
	public boolean isTestingAction () {
		return action.equals(TESTING_ACTION);
	}

	/**
	 * Returns the extraction date as Date.
	 *
	 * @return the extraction date
	 * or null if the extraction date string is invalid.
	 */
	public Date getExtractionDateTimeAsDate () {
		return extractionDate;
	}

	public Interval getReportingPeriodAsInterval () {
		final RegularIntervalFormat formatter = new RegularIntervalFormat();

		try {
			final Interval interval = formatter.parse( reportingPeriod, reportingPeriodFormat );
			return interval;
		}
		catch ( final ParseException pe ) {
			return null;
		}
	}

	/* ------------------ Other Public Methods ------------------ */

	/**
	 * Returns the hash code value for this MessageBegin object.
	 * Equality according to {@link #equals} implies that
	 * the hashcodes are the same
	 * as required by the general contract of {@link Object#hashCode}.
	 *
	 * @return hash code value for this object.
	 */
	public int hashCode () {
		// if not yet computed (or accidentally zero)
		if ( hashCode == 0 ) {
			hashCode =
			/*	hash collision occurred. Many little string hashcodes
			    xored together makes this a rather probable event.
				Let's hashcode one big string instead. 
				^ sender.hashCode()
				^ receiver.hashCode()
				^ maintenanceAgency.hashCode()
				^ function.hashCode()
				^ identity.hashCode()
				^ Arrays.asList( contacts ).hashCode()
				^ dataSetId.hashCode()
				^ action.hashCode()
				^ ((null == extractionDate) ? 0 : extractionDate.hashCode())
				^ reportingPeriod.hashCode()
				^ reportingPeriodFormat.hashCode()
				^ keyFamilyId.hashCode()
				^ ((null == originalDataSetId) ? 0 : originalDataSetId.hashCode());
			*/
				(
					mref + sender + receiver + maintenanceAgency + function
					+ identity + dataSetId + action.hashCode()
					+ ((null == extractionDate) ? 0 : extractionDate.hashCode())
					+ reportingPeriod + reportingPeriodFormat.hashCode() + keyFamilyId.hashCode()
					+ ((null == originalDataSetId) ? 0 : originalDataSetId.hashCode())
				).hashCode()
				^ Arrays.asList( contacts ).hashCode();
		}
		return hashCode;
	}

	/**
	 * Indicates whether some other MessageBegin object
	 * is "equal to" this one.
		}
		return hashCode;
	}

	/**
	 * Indicates whether some other MessageBegin object
	 * is "equal to" this one.
	 *
	 * @param obj the other reference object with which to compare
	 *
	 * @return true if this MessageBegin object
	 * is equal to the argument; false otherwise.
	 */
	public boolean equals ( Object obj ) {
		if ( this == obj ) {
			return true;
		} else if ( obj instanceof MessageBegin ) {
			MessageBegin other = (MessageBegin) obj;
			return
				mref.equals( other.mref )
				&& sender.equals( other.sender )
				&& receiver.equals( other.receiver )
				&& maintenanceAgency.equals(
					other.maintenanceAgency
				)
				&& function.equals( other.function )
				&& identity.equals( other.identity )
				&& Arrays.equals( contacts, other.contacts )
				&& dataSetId.equals( other.dataSetId )
				&& action.equals( other.action )
				&& (
				    (null != extractionDate
				    &&
				    extractionDate.equals( other.extractionDate )
				   )
				   || (null == extractionDate && null == other.extractionDate)
				   )
				&& reportingPeriod.equals(
					other.reportingPeriod
				)
				&& reportingPeriodFormat.equals(
					other.reportingPeriodFormat
				)
				&& keyFamilyId.equals( other.keyFamilyId )
				&& originalDataSetId.equals( other.originalDataSetId );
		} else {
			return false;
		}
	}

	/**
	 * Return a string representation of this MessageBegin.
	 * This method is intended to be used only for debugging purposes,
	 * and the format of the returned string may vary between
	 * implementations.
	 * The returned string may be empty but may not be null.
	 *
	 * @return a string representation of this MessageBegin.
	 */
	public String toString () {
		StringBuffer stringBuffer = new StringBuffer();
		for ( int i = 0; i < contacts.length; ++i ) {
			if ( i > 0 ) stringBuffer.append( ", " );
			stringBuffer.append( contacts[i].toString() );
		}
		return getClass().getName() + "("
			+ "mref=" + mref + ", "
			+ "sender=" + sender + ", "
			+ "receiver=" + receiver + ", "
			+ "maintenanceAgency=" + maintenanceAgency + ", "
			+ "function=" + function + ", "
			+ "identity=" + identity + ", "
			+ "contacts={" + stringBuffer.toString() + " }"
			+ (
				// If the MessageBegin has been constructed
				// without data set information,
				// the fields are really == "",
				// not only equal.
				// Anyway, it does not harm.
				dataSetId != ""
				|| action != ""
				|| extractionDate != null
				|| reportingPeriod != ""
				|| reportingPeriodFormat != ""
				|| keyFamilyId != ""
				|| originalDataSetId != ""
				? ", dataSetId=" + dataSetId + ", "
					+ "action=" + action + ", "
					+ "extractionDateTime=" + extractionDate + ", "
					+ "reportingPeriod=" + reportingPeriod + ", "
					+ "reportingPeriodFormat=" + reportingPeriodFormat + ", "
					+ "keyFamilyId=" + keyFamilyId + ", "
					+ "originalDataSetId=" + originalDataSetId
				: ""
			) + ")";
	}

	/* ==================== Private Methods ===================== */

	private static String extractedToPeriod ( final Date extracted ) {
		if ( null == extracted ) return "";
		return new SimpleDateFormat( "yyyyMMddHHmm" ).format( extracted );
	}

	private static String intervalToPeriod (
		final RegularInterval interval
	) {
		if ( null == interval ) return "";
		final RegularIntervalFormat formatter = new RegularIntervalFormat();
		final String periodAndFormat = formatter.format( interval );
		return periodAndFormat.substring( 0, periodAndFormat.indexOf( ':' ) );
	}

	private static String intervalToPeriodFormat (
		final RegularInterval interval
	) {
		if ( null == interval ) return "";
		final RegularIntervalFormat formatter = new RegularIntervalFormat();
		final String periodAndFormat = formatter.format( interval );
		return periodAndFormat.substring( periodAndFormat.indexOf( ':' ) + 1 );
	}
	
	private static Date extractedToDate(final String extracted) {
	    final String dateFormat = "yyyyMMddHHmm";
        final SimpleDateFormat formatter
        = new SimpleDateFormat( dateFormat );
        formatter.setLenient( false );
        final ParsePosition pos = new ParsePosition( 0 );

        Date date = formatter.parse( extracted, pos );
        if ( pos.getIndex() < dateFormat.length() ) {
            date = null;
        }
        return date;
	}
}
