// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, 2007 European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb;

import java.util.Collection;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.Calendar;

import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.text.ParseException;

import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;

import Int.ecb.gesmescb.time.Interval;
import Int.ecb.gesmescb.time.RegularInterval;
import Int.ecb.gesmescb.time.GregorianRegularInterval;

/* ***************************** Class ****************************** */

/**
 * <P>Tests the implementation of the class {@link MessageBegin}.</P>
 *
 * @author Sami Airo
 * @author <a href="mailto:sis.external@ecb.int">Bernhard Bodenstorfer</a>
 *
 * @version @(#)$Id: MessageBeginTest.java,v 1.1 2017/01/19 16:14:04 trainot Exp $
 */

public class MessageBeginTest extends TestCase {

	/* ======================= Constants ======================== */

	// Strings for constructor args.
	private static final Object[] original = {
		"MREF0001",
		"DE2",
		"4F0",
		"ECB",
		MessageBegin.STATISTICAL_DATA,
		"The subject",
		new Contact[] {
			new Contact(
				Contact.DATA_PROCESSING, "ECB-SIS", "Quasimodo",
				new ContactChannel[] {
					new ContactChannel(
						ContactChannel.EMAIL,
						"s-xtermal@ecb.int"
					)
				}
			)
		},
		"DATASET1",
		MessageBegin.DELETE_ACTION,
		"200103201423",
		"1999",
		"602",
		"ECB_BOP1"
	};

	private static final Object[] originalWithOrgDSI = {
		"MREF0001",
		"DE2",
		"4F0",
		"ECB",
		MessageBegin.STATISTICAL_DATA,
		"The subject",
		new Contact[] {
			new Contact(
				Contact.DATA_PROCESSING, "ECB-SIS", "Quasimodo",
				new ContactChannel[] {
					new ContactChannel(
						ContactChannel.EMAIL,
						"s-xtermal@ecb.int"
					)
				}
			)
		},
		"DATASET1",
		MessageBegin.DELETE_ACTION,
		"200103201423",
		"1999",
		"602",
		"ECB_BOP1",
		"DATASET1" // the original external dsi
	};
	
	// Modifications.
	private static final Object[] modified = {
		"MREFPULP",	// Intentionally invalid
		"4F0",
		"DE2",
		"BIS",
		MessageBegin.DATA_REQUEST,
		"No subject",
		new Contact[] {},
		"DATASET2",
		MessageBegin.TESTING_ACTION,
		"200103201723",
		"1998",
		"610",
		"ECB_BOPP"
	};

	/* ========================= Fields ========================= */

	/**
	 * The "standard" messages, once with data set info, once short.
	 */
	private MessageBegin m0, ms0, m00;

	/**
	 * An "equal" message, once with data set info, once short.
	 */
	private MessageBegin m1, ms1;

	/**
	 * An array of "unequal" messages, once with data set info, once short.
	 */
	private MessageBegin[] m, ms;

	/* ====================== Constructors ====================== */

	/**
	 * Constructs a MessageBeginTest with the specified name.
	 *
	 * @param name Test case name
	 */
	public MessageBeginTest ( String name ) {
		super( name );
	}

	/* ================== Protected Methods ===================== */

	/**
	 * Sets up the test fixture
	 */
	protected void setUp () {
		// Create full MessageBegin.
		m0 = new MessageBegin(
			(String) original[0],
			(String) original[1],
			(String) original[2],
			(String) original[3],
			(String) original[4],
			(String) original[5],
			(Contact[]) original[6],
			(String) original[7],
			(String) original[8],
			(String) original[9],
			(String) original[10],
			(String) original[11],
			(String) original[12]
		);
		// Create full MessageBegin.
		m00 = new MessageBegin(
			(String) originalWithOrgDSI[0],
			(String) originalWithOrgDSI[1],
			(String) originalWithOrgDSI[2],
			(String) originalWithOrgDSI[3],
			(String) originalWithOrgDSI[4],
			(String) originalWithOrgDSI[5],
			(Contact[]) originalWithOrgDSI[6],
			(String) originalWithOrgDSI[7],
			(String) originalWithOrgDSI[8],
			(String) originalWithOrgDSI[9],
			(String) originalWithOrgDSI[10],
			(String) originalWithOrgDSI[11],
			(String) originalWithOrgDSI[12],
			(String) originalWithOrgDSI[13]
		);
		
		// Without dataset info.
		ms0 = new MessageBegin(
			(String) original[0],
			(String) original[1],
			(String) original[2],
			(String) original[3],
			(String) original[4],
			(String) original[5],
			(Contact[]) original[6]
		);

		// Create "equal" MessageBegin from equal, non-identical Strings.
		m1 = new MessageBegin(
			new String( (String) original[0] ),
			new String( (String) original[1] ),
			new String( (String) original[2] ),
			new String( (String) original[3] ),
			new String( (String) original[4] ),
			new String( (String) original[5] ),
			(Contact[]) ( (Contact[]) original[6] ).clone(),
			new String( (String) original[7] ),
			new String( (String) original[8] ),
			new String( (String) original[9] ),
			new String( (String) original[10] ),
			new String( (String) original[11] ),
			new String( (String) original[12] )
		);
		// Without dataset info.
		ms1 = new MessageBegin(
			new String( (String) original[0] ),
			new String( (String) original[1] ),
			new String( (String) original[2] ),
			new String( (String) original[3] ),
			new String( (String) original[4] ),
			new String( (String) original[5] ),
			(Contact[]) ( (Contact[]) original[6] ).clone()
		);
		// Create "equal" MessageBegin from equal, non-identical Strings.
		m = new MessageBegin[original.length];
		for ( int i = 0; i < m.length; ++i ) {
			m[i] = new MessageBegin(
				(String) ( i == 0 ? modified[0] : original[0] ),
				(String) ( i == 1 ? modified[1] : original[1] ),
				(String) ( i == 2 ? modified[2] : original[2] ),
				(String) ( i == 3 ? modified[3] : original[3] ),
				(String) ( i == 4 ? modified[4] : original[4] ),
				(String) ( i == 5 ? modified[5] : original[5] ),
				(Contact[]) ( i == 6 ? modified[6] : original[6] ),
				(String) ( i == 7 ? modified[7] : original[7] ),
				(String) ( i == 8 ? modified[8] : original[8] ),
				(String) ( i == 9 ? modified[9] : original[9] ),
				(String) ( i == 10 ? modified[10] : original[10] ),
				(String) ( i == 11 ? modified[11] : original[11] ),
				(String) ( i == 12 ? modified[12] : original[12] )
			);
		}
		// Without dataset info.
		ms = new MessageBegin[5];	// Only five Constructor args.
		for ( int i = 0; i < ms.length; ++i ) {
			ms[i] = new MessageBegin(
				(String) ( i == 0 ? modified[0] : original[0] ),
				(String) ( i == 1 ? modified[1] : original[1] ),
				(String) ( i == 2 ? modified[2] : original[2] ),
				(String) ( i == 3 ? modified[3] : original[3] ),
				(String) ( i == 4 ? modified[4] : original[4] ),
				(String) ( i == 5 ? modified[5] : original[5] ),
				(Contact[]) ( i == 6 ? modified[6] : original[6] )
			);
		}
	}

	/**
	 * Tears down the test fixture
	 */
	protected void tearDown () {
	}

	/* ===================== Public Methods ===================== */

	/**
	 * Assembles and returns a test suite for all
	 * the test methods of this test case.
	 *
	 * @return a non-null test suite
	 */
	public static Test suite () {
		TestSuite suite = new TestSuite( MessageBeginTest.class );
		return suite;
	}

	/**
	 * Runs the test case.
	 */
	public static void main ( String args[] ) {
		String[] testCaseName = { MessageBeginTest.class.getName() };
		junit.textui.TestRunner.main( testCaseName );
	}

	/* --------------------- Test Methods --------------------- */

	/** Test constructor equivalences. */
	public void testConstructors () {
		final Collection contacts = new ArrayList();
		contacts.add( ( (Contact[]) original[6] )[0] );
		final MessageBegin other = new MessageBegin(
			(String) original[0],
			(String) original[1],
			(String) original[2],
			(String) original[3],
			(String) original[4],
			(String) original[5],
			contacts
		);
		assertEquals( "contacts collection", ms0, other );

		// A message simulating "no dataset info".
		MessageBegin mnodsi = new MessageBegin(
			new String( (String) original[0] ),
			new String( (String) original[1] ),
			new String( (String) original[2] ),
			new String( (String) original[3] ),
			new String( (String) original[4] ),
			new String( (String) original[5] ),
			(Contact[]) ( (Contact[]) original[6] ).clone(),
			new String( "" ),
			new String( "" ),
			new String( "" ),
			new String( "" ),
			new String( "" ),
			new String( "" )
		);
		assertEquals( "mnodsi.equals", mnodsi, ms1 );
		assertEquals( "mnodsi.hashCode", mnodsi, ms1 );
		assertEquals( "mnodsi.toString", mnodsi, ms1 );

		for ( int i = 0; i < ms.length; ++i ) {
			assertTrue( "mnodsi unequal", !mnodsi.equals( ms[i] ) );
			// Inequality of the hashCode is not really contract.
			// But the probability of eventual equal hashCode is
			// far less than that of a buggy
			// hashCode-implementation.
			// Therefore check.
			assertTrue(
				"mnodsi.hashCode unequal",
				mnodsi.hashCode() != ms[i].hashCode()
			);
			assertTrue(
				"mnodsi.toString unequal",
				!mnodsi.toString().equals( ms[i].toString() )
			);
		}
	}

	/** Test the date-based utility constructor. */
	public void testDatebasedConstructor () throws ParseException {
		final DateFormat format
		= new SimpleDateFormat( "yyyy-MM-dd'T'HH:mm:ss" );
		assertEquals(
			"Extracted and reporting period set",
			new MessageBegin(
				"MREF000", "4F0", "DE2", "BIS",
				MessageBegin.DATA_REQUEST,
				"No subject",
				new Contact[] {},
				"DATASET2",
				MessageBegin.TESTING_ACTION,
				"199506131629",
				"205603205802", "710",
				"ECB_BLUP"
			),
			new MessageBegin(
				"MREF000", "4F0", "DE2", "BIS",
				MessageBegin.DATA_REQUEST,
				"No subject",
				new Contact[] {},
				"DATASET2",
				MessageBegin.TESTING_ACTION,
				format.parse( "1995-06-13T16:29:00" ),
				new GregorianRegularInterval(
					RegularInterval.MONTHLY,
					format.parse( "2056-03-01T00:00:00" ),
					24
				),
				"ECB_BLUP"
			)
		);
		assertEquals(
			"Reporting period not set",
			new MessageBegin(
				"MREF000", "4F0", "DE2", "BIS",
				MessageBegin.DATA_REQUEST,
				"No subject",
				new Contact[] {},
				"DATASET2",
				MessageBegin.TESTING_ACTION,
				"183412310059",
				"", "",
				"ECB_WLUP"
			),
			new MessageBegin(
				"MREF000", "4F0", "DE2", "BIS",
				MessageBegin.DATA_REQUEST,
				"No subject",
				new Contact[] {},
				"DATASET2",
				MessageBegin.TESTING_ACTION,
				format.parse( "1834-12-31T00:59:00" ),
				null,
				"ECB_WLUP"
			)
		);
	}

	/**
	 * Tests the equals and hashcode methods method
	 */
	public void testEquals () {
		assertEquals( "m0 equals m1", m0, m1 );
		
		assertEquals( "m0 equals m00", m0, m00 );
		
		assertEquals(
			"m0.hashCode == m1.hashCode",
			m0.hashCode(),
			m1.hashCode()
		);
		assertEquals(
			"m0.toString equals m1.toString",
			m0.toString(),
			m1.toString()
		);

		for ( int i = 0; i < m.length; ++i ) {
			assertTrue( "m0 unequal", !m0.equals( m[i] ) );
			// Inequality of the hashCode is not really contract.
			// But the probability of eventual equal hashCode is
			// far less than that of a buggy
			// hashCode-implementation.
			// Therefore check.
			assertTrue(
				"m0.hashCode("+m0.hashCode()
				+"), with String\n"+m0.toString()
				+"\nshould be diffrent from m["+i+"].hashcode (" + m[i].hashCode()
				+"), with string:\n"+m[i].toString()
				+".\nAnd actually their originalDatasetId hashcodes should be different too:\n"
				+m0.getOriginalDataSetId().hashCode()
				+" as opposed to :\n"+m[i].getOriginalDataSetId().hashCode()+".",
				m0.hashCode() != m[i].hashCode()
			);	
			assertTrue(
				"m0.toString unequal",
				!m0.toString().equals( m[i].toString() )
			);
		}
	}

	/**
	 * Tests the normal fields.
	 */
	public void testNormalFields () {
		assertEquals( "MREF0001", m1.getMref() );
		assertEquals( "DE2", m1.getSender() );
		assertEquals( "4F0", m1.getReceiver() );
		assertEquals( "ECB", m1.getMaintenanceAgencyId() );
		assertEquals( MessageBegin.STATISTICAL_DATA, m1.getFunction() );
		assertEquals( "The subject", m1.getIdentity() );
		final Iterator iterator = m1.contactIterator();
		assertEquals(
			"contact 0", ((Contact[]) original[6])[0], iterator.next()
		);
		assertTrue( "no contact 1", !iterator.hasNext() );
		assertEquals( "DATASET1", m1.getDataSetId() );
		assertEquals( MessageBegin.DELETE_ACTION, m1.getAction() );
		assertEquals( "200103201423", m1.getExtractionDateTime() );
		assertEquals( "1999", m1.getReportingPeriod() );
		assertEquals( "602", m1.getReportingPeriodFormat() );
		assertEquals( "ECB_BOP1", m1.getKeyFamilyId() );
	}

	/**
	 * Tests the computed fields.
	 */
	public void testComputedFields () {

		assertTrue( m1.isStatisticalData() );
		assertTrue( !m1.isStatisticalDefinitions() );
		assertTrue( !m1.isDataRequest() );
		assertEquals( "MREF", 1, m1.getMrefNo() );

		MessageBegin m2 = new MessageBegin(
			(String) modified[0],
			(String) modified[1],
			(String) modified[2],
			(String) modified[3],
			(String) modified[4],
			(String) modified[5],
			(Contact[]) modified[6],
			(String) modified[7],
			(String) modified[8],
			(String) modified[9],
			(String) modified[10],
			(String) modified[11],
			(String) modified[12]
		);

		assertTrue( !m2.isStatisticalData() );
		assertTrue( !m2.isStatisticalDefinitions() );
		assertTrue( m2.isDataRequest() );
		assertEquals(
			"INVALID_MREF",
			m2.getMrefNo(),
			MessageBegin.INVALID_MREF
		);
	}

	/** Test the time converting methods. */
	public void testTimeConversions () {
		final Calendar calendar = Calendar.getInstance();

		{
			final MessageBegin message = new MessageBegin(
				"MREF234768", "BBB", "4F0", "BODI",
				MessageBegin.STATISTICAL_DEFINITIONS,
				"Sample Definiitons",
				new Contact[0]
			);
			assertNull( "date 1", message.getExtractionDateTimeAsDate() );
			assertNull( "interval 1", message.getReportingPeriodAsInterval() );
		}

		{
			final MessageBegin message = new MessageBegin(
				"MREF234768", "BBB", "4F0", "BODI",
				MessageBegin.STATISTICAL_DATA,
				"Sample Data",
				new Contact[0],
				"BBB_BOP1",
				MessageBegin.UPDATE_OR_REPLACE_ACTION,
				"200202031856",
				"200204", "610",
				"BBB_BOP1"
			);

			calendar.clear();
			calendar.set( 2002, Calendar.FEBRUARY, 3, 18, 56 );
			assertEquals(
				"date 2",
				calendar.getTime(),
				message.getExtractionDateTimeAsDate()
			);

			final Interval reportingInterval
			= message.getReportingPeriodAsInterval();
			calendar.clear();
			calendar.set( 2002, Calendar.APRIL, 1 );
			assertEquals(
				"begin 2",
				calendar.getTime(),
				reportingInterval.getBegin()
			);
			calendar.set( 2002, Calendar.MAY, 1 );
			assertEquals(
				"end 2",
				calendar.getTime(),
				reportingInterval.getEnd()
			);
		}

		{
			final MessageBegin message = new MessageBegin(
				"MREF234768", "BBB", "4F0", "BODI",
				MessageBegin.STATISTICAL_DATA,
				"Sample Data",
				new Contact[0],
				"BBB_BOP1",
				MessageBegin.UPDATE_OR_REPLACE_ACTION,
				"200200031856",
				"20025", "608",
				"BBB_BOP1"
			);
			assertNull( "date 3", message.getExtractionDateTimeAsDate() );
			assertNull( "interval 3", message.getReportingPeriodAsInterval() );
		}
	}

	/**
	 * Tests serialization.
	 */
	public void testSerialization () {
		final MessageBegin transserialized
		= (MessageBegin) AttributeTest.transserialize( m0 );
		assertEquals( m0, transserialized );
		assertEquals( m0.hashCode(), transserialized.hashCode() );
		assertEquals( m0.getMrefNo(), transserialized.getMrefNo() );
	}
}
