// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb;

import java.lang.String;
import java.lang.Integer;

import Int.ecb.gesmescb.MessageBegin;

/* ***************************** Class ****************************** */

/**
 * Holds the GESMES/TS message trailer
 * information.
 *
 * @author
 * <A HREF="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</A>
 *
 * @version @(#)$Id: MessageEnd.java,v 1.1 2017/01/19 16:14:04 trainot Exp $
 */

public final class MessageEnd extends GESMESCBObject {

	/* ======================= Constants ======================== */

	/**
	 *   serialVersionUID generated by eclipse.
	 */
	private static final long serialVersionUID = -1386698407850686107L;

	/**
	 * Value returned by {@link #getMrefNo} if the
	 * message reference number is in an invalid format.
	 */
	public static final int INVALID_MREF = MessageBegin.INVALID_MREF;

	/* ========================= Fields ========================= */

	/**
	 * Message reference identifyer.
	 *
	 * @serial
	 */
	private final String mref;

	/* ------------------- Transient Fields -------------------- */

	/**
	 * message reference number
	 */
	private transient int mrefNo;

	/**
	 * Precomputed hashcode.
	 */
	private transient int hashCode;

	/* ====================== Constructors ====================== */

	/**
	 * Construct a new <CODE>MessageEnd</CODE> from message header
	 * and data set header information.
	 *
	 * @param mref message reference.
	 *
	 * @throws NullPointerException upon a <CODE>null</CODE> parameter.
	 */
	public MessageEnd (
		final String mref
	)
		throws NullPointerException
	{
		if ( ( this.mref = mref ) == null )
			throw new NullPointerException(
				"Initializing MessageEnd with null mref"
			);
	}

	/* ===================== Public Methods ===================== */

	/* -------------------- Message Methods --------------------- */

	/**
	 * Returns message reference.
	 *
	 * @return message reference.
	 */
	public String getMref () {
		return mref;
	}

	/**
	 * Returns the MREF number.
	 *
	 * @return the MREF number or {@link #INVALID_MREF}
	 * if the MREF is invalid.
	 */
	public int getMrefNo () {
		// if mrefNo not yet initialized (or accidentally zero)
		if ( mrefNo == 0 ) {
			try {
				mrefNo = Integer.parseInt(
					mref.substring( 4 )
				);
			} catch ( IndexOutOfBoundsException e ) {
				mrefNo = INVALID_MREF;
			} catch ( NumberFormatException e ) {
				mrefNo = INVALID_MREF;
			}
		}
		return mrefNo;
	}

	/**
	 * Returns whether the <CODE>MessageEnd</CODE> object corresponds
	 * to the given <CODE>MessageBegin</CODE> object.
	 *
	 * @return whether this <CODE>MessageEnd</CODE>
	 * can be considered to end the given
	 * <CODE>MessageBegin</CODE>.
	 */
	public boolean isEndFor(
		final MessageBegin messageBegin
	) {
		return mref.equals( messageBegin.getMref() );
	}

	/* ------------------ Other Public Methods ------------------ */

	/**
	 * Returns the hash code value for this MessageEnd object.
	 * Equality according to {@link #equals} implies that
	 * the hashcodes are the same
	 * as required by the general contract of {@link Object#hashCode}.
	 *
	 * @return hash code value for this object.
	 */
	public int hashCode () {
		// if not yet computed (or accidentally zero)
		if ( hashCode == 0 ) {
			hashCode =
				mref.hashCode();
		}
		return hashCode;
	}

	/**
	 * Indicates whether some other MessageEnd object
	 * is "equal to" this one.
	 *
	 * @param obj the other reference object with which to compare
	 *
	 * @return true if this MessageEnd object
	 * is equal to the argument; false otherwise.
	 */
	public boolean equals ( Object obj ) {

		if ( this == obj ) {
			return true;
		} else if ( obj instanceof MessageEnd ) {
			MessageEnd other = (MessageEnd) obj;

			return
				mref.equals( other.mref );
		} else {
			return false;
		}
	}

	/**
	 * Return a string representation of this MessageEnd.
	 * This method is intended to be used only for debugging purposes,
	 * and the format of the returned string may vary between
	 * implementations.
	 * The returned string may be empty but may not be null.
	 *
	 * @return a string representation of this MessageEnd.
	 */
	public String toString () {
		return getClass().getName() + "("
			+ "mref=" + mref + ")";
	}
}
