// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb;

import java.lang.String;

/* ***************************** Class ****************************** */

/**
 * The class Observation holds one observation of the time series object.
 *
 * @author
 * <A HREF="mailto:klo@FAME.com">Kam Lo</A>
 * @author
 * <A HREF="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</A>
 *
 * @version @(#)$Id: Observation.java,v 1.1 2017/01/19 16:14:04 trainot Exp $
 */

public final class Observation implements java.io.Serializable {

	/* ======================= Constants ======================== */

	/**
	 *   serialVersionUID generated by eclipse.
	 */
	private static final long serialVersionUID = -4612757792300979515L;

	/**
	 * An empty observation.
	 * This can be used by {@link Int.ecb.gesmescb.event.GESMESCBReader}
	 * objects to quickly feed a {@link TimeSeries}
	 * with empty observations. This is particularly
	 * handy for delete messages.
	 */
	public static final Observation EMPTY_OBSERVATION
	= new Observation( "", "", "", "" );

	/**
	 * As an observation value it indicates a missing value.
	 */
	public static final String MISSING_VALUE = "-";

	/* ========================= Fields ========================= */

	/**
	 * Observation value.
	 *
	 * @serial
	 */
	private final String value;

	/**
	 * Observation status.
	 *
	 * @serial
	 */
	private final String status;

	/**
	 * Confidentiality value.
	 *
	 * @serial
	 */
	private final String confidentiality;

	/**
	 * Pre-break observation value.
	 *
	 * @serial
	 */
	private final String preBreakValue;

	/**
	 * Precomputed hash code.
	 * This will be lazily computed on demand if it is zero.
	 */
	private transient int hashCode;

	/* ====================== Constructors ====================== */

	/**
	 * Construct a new observation.
	 *
	 * @param value observation value.
	 * @param status observation status.
	 * @param confidentiality observation confidentiality.
	 * @param preBreakValue pre-break observation value.
	 */
	public Observation (
		String value,
		String status,
		String confidentiality,
		String preBreakValue
	) {
		this.value = value;
		this.status = status;
		this.confidentiality = confidentiality;
		this.preBreakValue = preBreakValue;
	}

	/* ===================== Public Methods ===================== */

	/**
	 * Returns the observation value.
	 *
	 * @return observation value.
	 */
	public String getValue () {
		return value;
	}

	/**
	 * Returns the observation status.
	 *
	 * @return observation status.
	 */
	public String getStatus () {
		return status;
	}

	/**
	 * Returns the observation confidentiality value.
	 *
	 * @return observation confidentiality value.
	 */
	public String getConfidentiality () {
		return confidentiality;
	}

	/**
	 * Returns the pre-break observation value.
	 *
	 * @return pre-break observation value.
	 */
	public String getPreBreakValue () {
		return preBreakValue;
	}

	/**
	 * Return whether the observation is empty.
	 * This is the case if it has been constructed from empty strings only.
	 *
	 * @return whether the value is marked as missing.
	 */
	public boolean isEmpty () {
		return equals( EMPTY_OBSERVATION );
	}

	/**
	 * Return whether a missing value indicator is carried.
	 *
	 * @return whether the value is marked as missing.
	 */
	public boolean hasMissingValue () {
		return value.equals( MISSING_VALUE );
	}

	/**
	 * Return whether a missing value indicator is carried as pre break value.
	 *
	 * @return whether the pre break value is marked as missing.
	 */
	public boolean hasMissingPreBreakValue () {
		return preBreakValue.equals( MISSING_VALUE );
	}

	/**
	 * Indicates whether some other Object is "equal to" this Observation.
	 *
	 * @param obj the reference object with which to compare
	 *
	 * @return true if this Observation is equal to the argument;
	 * false otherwise.
	 */
	public boolean equals ( final Object obj ) {
		if ( this == obj ) return true;
		if ( obj instanceof Observation ) {
			final Observation other = (Observation) obj;
			return
				value.equals( other.value )
				&& status.equals( other.status )
				&& confidentiality.equals( other.confidentiality )
				&& preBreakValue.equals( other.preBreakValue );
		}
		return false;
	}

	/**
	 * Returns the hash code value for this Observation.
	 * If o1.equals(o2) then o1.hashCode()==o2.hashCode()
	 * for any two Observations o1 and o2,
	 * as required by the general contract of {@link Object#hashCode}.
	 *
	 * @return hash code value for this object.
	 */
	public int hashCode () {
		// hashCode itself is used as flag to recompute
		// for thread safety.
		if ( hashCode == 0 ) {
			hashCode = value.hashCode()
				^ status.hashCode()
				^ confidentiality.hashCode()
				^ preBreakValue.hashCode();
		}
		return hashCode;
	}

	/**
	 * Returns a string representation of the Observation
	 * in a human readable format.
	 *
	 * @return a string representation of the Observation.
	 */
	public String toString () {
		return
			value + ":"
			+ status + ":"
			+ confidentiality + ":"
			+ preBreakValue;
	}
}
