// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb;

import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;

/* ***************************** Class ****************************** */

/**
 * <P>Tests the implementation of the class {@link Observation}.</P>
 *
 * @author
 * <A HREF="mailto:Sami.Airo@ECB.int">Sami Airo</A>
 * @author
 * <A HREF="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</A>
 *
 * @version @(#)$Id: ObservationTest.java,v 1.1 2017/01/19 16:14:04 trainot Exp $
 */

public class ObservationTest extends TestCase {

	/* ====================== Constructors ====================== */

	/**
	 * Constructs an ObservationTest with the specified name.
	 *
	 * @param name Test case name
	 */
	public ObservationTest ( final String name ) {
		super( name );
	}

	/* ===================== Public Methods ===================== */

	/**
	 * Assembles and returns a test suite for all
	 * the test methods of this test case.
	 *
	 * @return a non-null test suite
	 */
	public static Test suite () {
		return new TestSuite( ObservationTest.class );
	}

	/**
	 * Runs the test case.
	 */
	public static void main ( final String args[] ) {
		final String[] testCaseName = { ObservationTest.class.getName() };
		junit.textui.TestRunner.main( testCaseName );
	}

	/* --------------------- Test Methods --------------------- */

	/**
	 * Test the equals, hashCode and toString methods.
	 */
	public void testEquals () {
		// Create Strings for constructor args.
		final String[] original = {
			"value",
			"status",
			"confidentiality",
			"preBreakValue"
		};
		// Create modifications.
		final String[] modified = {
			"Value",
			"Status",
			"Confidentiality",
			"PreBreakValue"
		};

		// Create Observation.
		final Observation obs0 = new Observation(
			original[0],
			original[1],
			original[2],
			original[3]
		);
		// Create "equal" Observation from equal, non-identical Strings.
		final Observation obs1 = new Observation(
			new String( original[0] ),
			new String( original[1] ),
			new String( original[2] ),
			new String( original[3] )
		);
		assertTrue( "obs0 equals obs1", obs0.equals( obs1 ) );
		assertTrue(
			"obs0.hashCode == obs1.hashCode",
			obs0.hashCode() == obs1.hashCode()
		);
		assertTrue(
			"obs0.toString equals obs1.toString",
			obs0.toString().equals( obs1.toString() )
		);

		// Next create unequal Observations,
		// which are unequal at varying positions.
		final Observation[] obs = new Observation[original.length];
		for ( int i = 0; i < obs.length; ++i ) {
			obs[i] = new Observation(
				i == 0 ? modified[0] : original[0],
				i == 1 ? modified[1] : original[1],
				i == 2 ? modified[2] : original[2],
				i == 3 ? modified[3] : original[3]
			);

			assertTrue(
				"obs["+i+"] not equals obs0",
				!obs[i].equals( obs0 )
			);
			// Inequality of the hashCode is not really contract.
			// But the probability of eventual equal hashCode is
			// far less than that of a buggy
			// hashCode-implementation.
			// Therefore check.
			assertTrue(
				"obs[" + i + "].hashCode != obs0.hashCode",
				obs[i].hashCode() != obs0.hashCode()
			);
			assertTrue(
				"obs[" + i + "].toString != obs0.toString",
				!obs[i].toString().equals( obs0.toString() )
			);
		}
	}

	/**
	 * Test the get-type methods.
	 */
	public void testProperties () {
		final Observation obs0 = new Observation( "1", "2", "3", "4" );

		assertEquals( "1", obs0.getValue() );
		assertEquals( "2", obs0.getStatus() );
		assertEquals( "3", obs0.getConfidentiality() );
		assertEquals( "4", obs0.getPreBreakValue() );
	}

	/**
	 * Test the boolean shortcut methods.
	 */
	public void testShortcuts () {
		final Observation obs0 = new Observation( "1", "2", "3", "4" );
		final Observation obs1 = new Observation( "-", "X", "Y", "Z" );
		final Observation obs2 = new Observation( "", "", "", "" );
		final Observation obs3 = new Observation( "E", "X", "Y", "-" );

		assertTrue( "obs0 nonempty", !obs0.isEmpty() );
		assertTrue( "obs0 nonmissing", !obs0.hasMissingValue() );
		assertTrue( "obs1 nonempty", !obs1.isEmpty() );
		assertTrue( "obs1 missing", obs1.hasMissingValue() );
		assertTrue( "obs1 present preBreak", !obs1.hasMissingPreBreakValue() );
		assertTrue( "obs2 empty", obs2.isEmpty() );
		assertTrue( "obs2 nonmissing", !obs2.hasMissingValue() );
		assertTrue( "obs2 present preBreak", !obs2.hasMissingPreBreakValue() );
		assertTrue( "obs3 missing", obs3.hasMissingPreBreakValue() );

		final Observation[] obs = new Observation[16];
		for ( int i = 0; i < obs.length; ++i ) {
			obs[i] = new Observation(
				( i & 1 ) > 0 ? new String( "A" ) : new String ( "" ),
				( i & 2 ) > 0 ? new String( "B" ) : new String ( "" ),
				( i & 4 ) > 0 ? new String( "Ch" ) : new String ( "" ),
				( i & 8 ) > 0 ? new String( "Dh" ) : new String ( "" )
			);
		}
		assertTrue( "obs[0] empty", obs[0].isEmpty() );
		for ( int i = 1; i < obs.length; ++i ) {
			assertTrue( "obs[" + i + "] nonempty", !obs[i].isEmpty() );
		}
	}

	/**
	 * Tests serialization.
	 */
	public void testSerialization () {
		final Observation obs0 = new Observation(
			"BLDI", "Fatal", "", "Jomk"
		);
		final Observation obs1 = (Observation) AttributeTest.transserialize( obs0 );
		assertEquals( obs0, obs1 );
		assertEquals( obs0.hashCode(), obs1.hashCode() );
	}
}
