// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb;

import java.util.Date;

/* ***************************** Class ****************************** */

/**
 * GESMES/TS statistical concept.
 *
 * StatisticalConcept object is meant to be immutable - so that it does not
 * change after construction.
 *
 * @author
 * <A HREF="mailto:Sami.Airo@ECB.int">Sami Airo</A>
 * @author
 * <A HREF="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</A>
 *
 * @version @(#)$Id: StatisticalConcept.java,v 1.1 2017/01/19 16:14:04 trainot Exp $
 */

public class StatisticalConcept extends AbstractStructuralDefinition {

	/* ========================= Fields ========================= */

	/**
	 *   serialVersionUID generated by eclipse.
	 */
	private static final long serialVersionUID = 8544535861042183597L;

	/**
	 * Concept representation.
	 * Currently part of the GESMES/TS data model but not
	 * supported by the EDIFACT syntax.
	 *
	 * @serial
	 */
	private final String representation;

	/* -------------------- Transient Fields -------------------- */

	/**
	 * Precomputed hash code.
	 */
	private transient int hashCode;

	/* ====================== Constructors ====================== */

	/**
	 * Construct a new StatisticalConcept object.
	 *
	 * @param identifier		concept identifier.
	 * @param name			concept name.
	 * @param description		concept description.
	 * @param versionDate	concept version timestamp.
	 */
	public StatisticalConcept (
		final String identifier,
		final String name,
		final String description,
		final Date versionDate
	) {
		super( identifier, name, description, versionDate );
		representation = "";
	}

	/**
	 * Construct a new StatisticalConcept object.
	 *
	 * @param identifier		concept identifier.
	 * @param name			concept name.
	 * @param description		concept description.
	 * @param representation	concept representation.
	 *
	 * @deprecated Do not construct StructuralDefinitions without versionDate
	 * but with superfluous representation.
	 * Use {@link #StatisticalConcept(String,String,String,Date)}.
	 */
	public StatisticalConcept (
		final String identifier,
		final String name,
		final String description,
		final String representation
	) {
		super( identifier, name, description );

		if ( ( this.representation = representation ) == null ) {
			throw new NullPointerException(
				"Trying to create StatisticalConcept"
				+ " with null representation."
			);
		}
	}

	/* ===================== Public Methods ===================== */

	/**
	 * Returns concept representation.
	 * The representation is currently part
	 * of the GESMES/TS data model but not
	 * supported by the EDIFACT syntax.
	 *
	 * @return concept representation.
	 *
	 * @deprecated Representation is never used and will be dropped.
	 * Refer to {@link KeyFamilyComponent} for representation information.
	 */
	public String getRepresentation () {
		return representation;
	}

	/**
	 * Returns the hash code value for this object.
	 * If co1.equals(co2) then co1.hashCode()==co2.hashCode()
	 * for any two Concepts objects co1 and co2,
	 * as required by the general contract of {@link Object#hashCode}.
	 *
	 * @return a hash code value for this object.
	 */
	public int hashCode () {
		// if not yet computed (or accidentally zero)
		if ( hashCode == 0 ) {
			hashCode = super.hashCode()
				^ representation.hashCode();
		}
		return hashCode;
	}

	/**
	 * Indicates whether some other StatisticalConcept object
	 * is "equal to" this one.
	 *
	 * @param object with which to compare
	 *
	 * @return true if this StatisticalConcept object
	 * is equal to the argument; false otherwise.
	 */
	public boolean equals ( final Object object ) {
		if ( this == object ) return true;
		if ( object instanceof StatisticalConcept ) {
			final StatisticalConcept other = (StatisticalConcept) object;
			return
				super.equals( other )
				&& representation.equals( other.representation );
		}
		return false;
	}

	/**
	 * Return a string representation of
	 * this <CODE>StatisticalConcept</CODE>.
	 * This method is intended to be used only for debugging purposes,
	 * and the format of the returned string may vary between
	 * implementations.
	 * The returned string may be empty but may not be null.
	 *
	 * @return a string representation
	 * of this <CODE>StatisticalConcept</CODE>.
	 */
	public String toString () {
		return getClass().getName() + "("
			+ "identifier=" + getId() + ", "
			+ "name=" + getName() + ", "
			+ "description=" + getDescription() + ", "
			+ "versionDate=" + getVersionDate() + ", "
			+ "representation=" + representation
			+ ")";
	}
}
