// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb;

import java.text.ParseException;

/* ***************************** Class ****************************** */

/**
 * <p>Checks {@link String} length and character type.
 * The formats correspond to EDIFACT specification formats,
 * e.g. <CODE>AN..35</CODE> or <CODE>N8</CODE>.
 * This restricts the character set and the fixed or maximum length
 * of a string.</P>
 *
 * <P><CODE>TextFormat</CODE> objects are immutable.</P>
 *
 * @author
 * <A HREF="mailto:klo@FAME.com">Kam Lo</A>
 * @author
 * <A HREF="mailto:Sami.Airo@ECB.int">Sami Airo</A>
 * @author
 * <A HREF="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</A>
 *
 * @version @(#)$Id: TextFormat.java,v 1.1 2017/01/19 16:14:04 trainot Exp $
 */
public class TextFormat {

	/* ======================= Constants ======================== */

	/** Any character type except control characters. */
	public static final int ANY = 0;

	/** Any character type.
	 * @deprecated use {@link #ANY}.
	 */
	private static final int ALPHANUMERIC = 1;

	/** Numeric character type. */
	public static final int NUMERIC = 2;

	/** Alphabetic character type. */
	public static final int ALPHABETIC = 3;

	/** Identified character type. */
	public static final int IDENTIFIER = 4;

	/** Coordinate value character type. */
	public static final int COORDINATE = 5;

	/** Code value character type of Gesmes/CB version 2.1.
	 * @deprecated Use {@link #ANY} for Gesmes/TS 3.0 code values.
	 */
	public static final int CODE = 6;

	private static final String[] TYPE_CODE = new String[CODE+1];

	static {
		TYPE_CODE[ANY]
		= TYPE_CODE[ALPHANUMERIC] = "AN";
		TYPE_CODE[ALPHABETIC] = "A";
		TYPE_CODE[NUMERIC] = "N";
		TYPE_CODE[IDENTIFIER] = "ID";
		TYPE_CODE[COORDINATE] = "CO";
		TYPE_CODE[CODE] = "CV";
	}

	/* ========================= Fields ========================= */

	/**
	 * Maximum value length or exact required length.
	 * @see #isVariableLength
	 */
	private final int referenceLength;

	/**
	 * Whether the {@link #limitLength} specifies a maximum
	 * or otherwise must be met.
	 */
	private final boolean isVariableLength;

	/* ====================== Constructors ====================== */

	/**
	 * Construct a new data value format object with specific type and length.
	 *
	 * @param length specified (maximum or exact) length
	 * @param variableLength whether the length specifies a maximum only
	 * or must be matched exactly.
	 *
	 * @throws IllegalArgumentException if length is invalid.
	 */
	TextFormat (
		final int length,
		final boolean variableLength
	) {
		this.isVariableLength = variableLength;

		if ( ( this.referenceLength = length ) < 0 ) {
			throw new IllegalArgumentException(
				"Invalid length passed to TextFormat."
			);
		}
	}

	/* ===================== Public Methods ===================== */

	/** Works as factory method for <code>TextFormat</code>s.
	 *
	 * @param characterType one of:
	 * {@link #ANY},
	 * {@link #ALPHABETIC},
	 * {@link #NUMERIC},
	 * {@link #IDENTIFIER},
	 * {@link #COORDINATE},
	 * {@link #CODE}.
	 * @param length specified (maximum or exact) length
	 * @param variableLength whether the length specifies a maximum only
	 * or must be matched exactly.
	 *
	 * @throws IllegalArgumentException if characterType or length is invalid.
	 */
	public static TextFormat create (
		final int characterType,
		final int length,
		final boolean variableLength
	) {
		switch ( characterType ) {
		case ANY:
		case ALPHANUMERIC:
			return new AnyFormat( length, variableLength );
		case NUMERIC: return new NumericFormat( length, variableLength );
		case ALPHABETIC: return new AlphabeticFormat( length, variableLength );
		case IDENTIFIER: return new IdentifierFormat( length, variableLength );
		case COORDINATE: return new CoordinateFormat( length, variableLength );
		case CODE: return new CodeFormat( length, variableLength );
		default:
			throw new IllegalArgumentException(
				"Invalid character type "
				+ characterType
				+ " passed to TextFormat."
			);
		}
	}

	/**
	 * Parse a new <code>TextFormat</code> object from a string.
	 * Format should be NX, N..X, NX..X, ANX, AN..X where AN is
	 * alphanumeric type, N is numeric type and X is a numeber specifying
	 * minimum and maximum length of data.
	 *
	 * @param format string containing format of data.
	 *
	 * @throws ParseException on error parsing format string.
	 */
	public static TextFormat parse (
		final String format
	) throws ParseException {
		int characterType = -1, parseRest = 0;

		for ( int i = 0; i < TYPE_CODE.length; ++i ) {
			if ( format.startsWith( TYPE_CODE[i] ) ) {
				characterType = i;
				parseRest = TYPE_CODE[i].length();
				break;
			}
		}

		if ( characterType < 0 ) {
			throw new ParseException(
				"Invalid format \"" + format + "\"", 0
			);
		}

		// Parse length-mode.
		final boolean variableLength
		= format.regionMatches( parseRest, "..", 0, 2 );
		if ( variableLength ) parseRest += 2;

		final String rest = format.substring( parseRest );
		try {
			final int length = Integer.parseInt( rest );
			return create( characterType, length, variableLength );
		}
		catch ( final NumberFormatException nfe ) {
			throw new ParseException(
				"Invalid format \"" + format + "\". \""
				+ rest + "\" is not a number.", parseRest
			);
		}
		catch ( final IllegalArgumentException iae ) {
			throw new ParseException(
				"Invalid format \"" + format + "\": " + iae.getMessage(), 0
			);
		}
	}

	/** Return the character type code.
	 * @return one of
	 * {@link #ANY},
	 * {@link #ALPHABETIC},
	 * {@link #NUMERIC},
	 * {@link #IDENTIFIER},
	 * {@link #COORDINATE},
	 * {@link #CODE}.
	 */
	public int getCharacterType () {
		return ANY;
	}

	/** Return the reference length.
	 * @see #isVariableLength()
	 */
	public int getReferenceLength () {
		return referenceLength;
	}

	/** Return whether the reference {#length()} is a maximum only.
	 * Otherwise it is exact.
	 */
	public boolean isVariableLength () {
		return isVariableLength;
	}

	/**
	 * Tests whether a value is valid.
	 *
	 * @param value to be tested.
	 *
	 * @return true if value is valid; false otherwise.
	 */
	public boolean isValid ( final String value ) {
		final int length = value.length();

		if ( !isVariableLength && length != referenceLength ) return false;
		return length <= referenceLength;
	}

	public int hashCode () {
		return
			getClass().hashCode()
			^
			getReferenceLength()
			^
			( isVariableLength() ? 1231 : 1237 );
			// These are the values used by Boolean.hashCode()
	}

	public boolean equals ( final Object object ) {
		if ( ! ( object instanceof TextFormat ) ) return false;
		final TextFormat that = (TextFormat) object;
		return
			getClass() == that.getClass()
			&&
			getReferenceLength() == that.getReferenceLength()
			&&
			isVariableLength() == that.isVariableLength();
	}

	public String toString () {
		final StringBuffer stringBuffer = new StringBuffer( 10 );
		stringBuffer.append( TYPE_CODE[getCharacterType()] );
		if ( isVariableLength ) stringBuffer.append( ".." );
		stringBuffer.append( referenceLength );
		return stringBuffer.toString();
	}
}

/* ***************************** Class ****************************** */

/** Accepts all strings without control characters. */
class AnyFormat extends TextFormat {

	public AnyFormat (
		final int length,
		final boolean variableLength
	) {
		super( length, variableLength );
	}

	public int getCharacterType () {
		return ANY;
	}

	public boolean isValid ( final String value ) {
		final int length = value.length();
		for ( int i = 0; i < length; ++i ) {
			final char c = value.charAt( i );
			if ( Character.isISOControl( c ) ) {
				return false;
			}
		}
		return super.isValid( value );
	}
}

/** Accepts only alphabetic strings. */
class AlphabeticFormat extends TextFormat {

	public AlphabeticFormat (
		final int length,
		final boolean variableLength
	) {
		super( length, variableLength );
	}

	public int getCharacterType () {
		return ALPHABETIC;
	}

	public boolean isValid ( final String value ) {
		final int length = value.length();
		for ( int i = 0; i < length; ++i ) {
			final char c = value.charAt( i );
			// Do not use Character.isLetter unless you want
			// Arabic, Bopomofo, Gurmukhi, etc letters.
			if (
				( c < 'A' || 'Z' < c )
				&&
				( c < 'a' || 'z' < c )
			) return false;
		}
		return super.isValid( value );
	}
}

/** Accepts only strings composed from digits. */
class NumericFormat extends TextFormat {

	public NumericFormat (
		final int length,
		final boolean variableLength
	) {
		super( length, variableLength );
	}

	public int getCharacterType () {
		return NUMERIC;
	}

	public boolean isValid ( final String value ) {
		final int length = value.length();
		for ( int i = 0; i < length; ++i ) {
			final char c = value.charAt( i );
			// Do not use Character.isDigit unless you want
			// Arabic, Gurmukhi, etc digits.
			if ( c < '0' || '9' < c ) return false;
		}
		return super.isValid( value );
	}
}

/** Accepts only strings of characters allowed for identifiers. */
class IdentifierFormat extends TextFormat {

	public IdentifierFormat (
		final int length,
		final boolean variableLength
	) {
		super( length, variableLength );
	}

	public int getCharacterType () {
		return IDENTIFIER;
	}

	public boolean isValid ( final String value ) {
		final int length = value.length();

		// The first character must be alphabetic
		if ( length > 0 ) {
			final char c = value.charAt( 0 );
			if ( c < 'A' || 'Z' < c ) return false;
		} else {
			return false;
		}

		for ( int i = 1; i < length; ++i ) {
			final char c = value.charAt( i );
			if (
				( c < 'A' || 'Z' < c )
				&&
				( c < '0' || '9' < c )
				&&
				c != '_'
			) return false;
		}
		return super.isValid( value );
	}
}

/** Accepts only strings of characters allowed for coordinates. */
class CoordinateFormat extends TextFormat {

	public CoordinateFormat (
		final int length,
		final boolean variableLength
	) {
		super( length, variableLength );
	}

	public int getCharacterType () {
		return COORDINATE;
	}

	public boolean isValid ( final String value ) {
		final int length = value.length();
		for ( int i = 0; i < length; ++i ) {
			final char c = value.charAt( i );
			if (
				( c < 'A' || 'Z' < c )
				&&
				( c < '0' || '9' < c )
				&&
				c != '_'
			) return false;
		}
		return super.isValid( value );
	}
}

/** Accepts only strings of characters allowed for code values. */
class CodeFormat extends TextFormat {

	public CodeFormat (
		final int length,
		final boolean variableLength
	) {
		super( length, variableLength );
	}

	public int getCharacterType () {
		return CODE;
	}

	public boolean isValid ( final String value ) {
		final int length = value.length();
		for ( int i = 0; i < length; ++i ) {
			final char c = value.charAt( i );
			if (
				( c < 'A' || 'Z' < c )
				&&
				( c < '0' || '9' < c )
				&&
				"_-/".indexOf( c ) < 0
			) return false;
		}
		return super.isValid( value );
	}
}
