// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb;

import java.text.ParseException;

import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;

/* ***************************** Class ****************************** */

/**
 * Tests the implementation of {@link TextFormat}.
 *
 * @author
 * <a href="mailto:Sami.Airo@ECB.int">Sami Airo</a>
 * @author
 * <a href="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</a>
 *
 * @version @(#)$Id: TextFormatTest.java,v 1.1 2017/01/19 16:14:04 trainot Exp $
 */
public class TextFormatTest extends TestCase {

	/* ====================== Constructors ====================== */

	/**
	 * Constructs a InterchangeTest with the specified name.
	 *
	 * @param name Test case name
	 */
	public TextFormatTest ( final String name ) {
		super( name );
	}

	/* ===================== Public Methods ===================== */

	/**
	 * Assembles and returns a test suite for all
	 * the test methods of this test case.
	 *
	 * @return a non-null test suite
	 */
	public static Test suite () {
		return new TestSuite(TextFormatTest.class);
	}

	/**
	 * Runs the test case
	 */
	public static void main ( final String args[] ) {
		final String[] testCaseName = {TextFormatTest.class.getName()};
		junit.textui.TestRunner.main( testCaseName );
	}

	/* --------------------- Test Methods --------------------- */

	public void testCreation () throws ParseException {
		final TextFormat an0a = TextFormat.create(
			TextFormat.ANY, 0, true
		);
		final TextFormat an0b = TextFormat.parse( "AN..0" );
		assertEqualities( "AN..0", an0a, an0b );

		final TextFormat a5a = TextFormat.create(
			TextFormat.ALPHABETIC, 5, true
		);
		final TextFormat a5b = TextFormat.parse( "A..5" );
		assertEqualities( "A..5", a5a, a5b );

		final TextFormat n1a = TextFormat.create(
			TextFormat.NUMERIC, 1, true
		);
		final TextFormat n1b = TextFormat.parse( "N..1" );
		assertEqualities( "N..1", n1a, n1b );

		final TextFormat id8a = TextFormat.create(
			TextFormat.IDENTIFIER, 8, false
		);
		final TextFormat id8b = TextFormat.parse( "ID8" );
		assertEqualities( "ID8", id8a, id8b );

		final TextFormat co3a = TextFormat.create(
			TextFormat.COORDINATE, 3, false
		);
		final TextFormat co3b = TextFormat.parse( "CO3" );
		assertEqualities( "CO3", co3a, co3b );

		final TextFormat cv4a = TextFormat.create(
			TextFormat.CODE, 4, true
		);
		final TextFormat cv4b = TextFormat.parse( "CV..4" );
		assertEqualities( "CV..4", cv4a, cv4b );
	}

	public void testParseFail () {
		final String[] failing = {
			"X5", "X..6", "AB", "A..F",
			"", "A4F", "ID8.", "CV-4",
			"N 3", " A4", "2", "-4",
			"..5", "CO..", "CO..-5", "AN..ID",
			" ", "AN", "A", "N",
			"ID..", "CV..", "CO3..7"
		};

		for ( int i = 0; i < failing.length; ++i ) {
			try {
				TextFormat.parse( failing[i] );
				fail( "Format \"" + failing[i] + "\" should not pass." );
			} catch ( final ParseException x ) {}
		}
	}

	public void testProperties () throws ParseException {
		{
			final TextFormat an5 = TextFormat.parse( "AN5" );
			assertEquals(
				"AN5 ctype", TextFormat.ANY, an5.getCharacterType()
			);
			assertEquals(
				"AN5 length", 5, an5.getReferenceLength()
			);
			assertEquals(
				"AN5 variable", false, an5.isVariableLength()
			);
		}

		{
			final TextFormat a50 = TextFormat.parse( "A..50" );
			assertEquals(
				"A..50 ctype", TextFormat.ALPHABETIC, a50.getCharacterType()
			);
			assertEquals(
				"A..50 length", 50, a50.getReferenceLength()
			);
			assertEquals(
				"A..50 variable", true, a50.isVariableLength()
			);
		}

		{
			final TextFormat n5 = TextFormat.parse( "N5");
			assertEquals(
				"N5 ctype", TextFormat.NUMERIC, n5.getCharacterType()
			);
		}

		{
			final TextFormat id8 = TextFormat.parse( "ID..8");
			assertEquals(
				"ID8 ctype", TextFormat.IDENTIFIER, id8.getCharacterType()
			);
		}

		{
			final TextFormat co17 = TextFormat.parse( "CO17");
			assertEquals(
				"CO17 ctype", TextFormat.COORDINATE, co17.getCharacterType()
			);
		}

		{
			final TextFormat cv8 = TextFormat.parse( "CV..8");
			assertEquals(
				"CV8 ctype", TextFormat.CODE, cv8.getCharacterType()
			);
		}
	}

	public void testIsValid () {
		{
			final TextFormat an5 = TextFormat.create(
				TextFormat.ANY, 5, true
			);
			assertTrue( an5.isValid( "129" ) );
			assertTrue( !an5.isValid( "ABC123" ) );
		}

		{
			final TextFormat an5 = TextFormat.create(
				TextFormat.ANY, 5, false
			);
			assertTrue( an5.isValid( "abcde" ) );
			assertTrue( !an5.isValid( "ABC1" ) );
		}
	}

	public void testIsValidAny () {
		final TextFormat an5 = TextFormat.create(
			TextFormat.ANY, 5, true
		);
		assertTrue( an5.isValid( "ABZ" ) );
		assertTrue( an5.isValid( "ABZ09" ) );
		assertTrue( an5.isValid( "129'" ) );
		assertTrue( an5.isValid( "_/? \"" ) );
		assertTrue( !an5.isValid( "ABC123" ) );
		assertTrue( !an5.isValid( "A\nA" ) );
		assertTrue( !an5.isValid( "0\r0" ) );
		assertTrue( !an5.isValid( "a\ta" ) );
	}

	public void testIsValidAlphabetic () {
		final TextFormat a5 = TextFormat.create(
			TextFormat.ALPHABETIC, 5, true
		);
		assertTrue( a5.isValid( "AZ" ) );
		assertTrue( !a5.isValid( "AZ029" ) );
	}

	public void testIsValidNumeric () {
		final TextFormat n5 = TextFormat.create(
			TextFormat.NUMERIC, 5, true
		);
		assertTrue( n5.isValid( "12" ) );
		assertTrue( !n5.isValid( "AB123" ) );

		final TextFormat n5b = TextFormat.create(
			TextFormat.NUMERIC, 5, false
		);
		assertTrue( n5b.isValid( "02349" ) );
		assertTrue( !n5b.isValid( "0239" ) );
	}

	/**
	 * Tests {@link TextFormat#isValid} for mnemonic formats.
	 */
	public void testIsValidIdentifier () {
		final TextFormat id5 = TextFormat.create(
			TextFormat.IDENTIFIER,
			5,
			true
		);
		assertTrue( id5.isValid( "ABZ" ) );
		assertTrue( !id5.isValid( "A1_2_B_" ) );
		assertTrue( !id5.isValid( "1_09" ) );
		assertTrue( !id5.isValid( "_A09" ) );
		assertTrue( id5.isValid( "A_C_2" ) );
		assertTrue( !id5.isValid( "A-/_9" ) );
		assertTrue( id5.isValid( "A1_09" ) );
		assertTrue( id5.isValid( "A" ) );
		assertTrue( !id5.isValid( "" ) );
	}

	/**
	 * Tests {@link TextFormat#isValid} for code value formats.
	 */
	public void testIsValidCoordinate () {
		final TextFormat co5 = TextFormat.create(
			TextFormat.COORDINATE,
			5,
			true
		);
		assertTrue( co5.isValid( "A09" ) );
		assertTrue( !co5.isValid( "A1_2_B_" ) );
		assertTrue( !co5.isValid( "E14b" ) );
		assertTrue( co5.isValid( "1_2Z" ) );
		assertTrue( !co5.isValid( "A-/_9" ) );

		final TextFormat co35 = TextFormat.create(
			TextFormat.COORDINATE, 35, true
		);
		assertTrue( !co35.isValid( "DEM/C.DEM" ) );
	}

	/**
	 * Tests {@link TextFormat#isValid} for code value formats.
	 */
	public void testIsValidCode () {
		final TextFormat cv5 = TextFormat.create(
			TextFormat.CODE,
			5,
			true
		);
		assertTrue( !cv5.isValid( "A=09" ) );
		assertTrue( !cv5.isValid( "A1_-_B_" ) );
		assertTrue( !cv5.isValid( "E14b" ) );
		assertTrue( cv5.isValid( "1_2Z" ) );
		assertTrue( cv5.isValid( "A-/_9" ) );

		final TextFormat cv35 = TextFormat.create(
			TextFormat.CODE, 35, true
		);

		assertTrue( !cv35.isValid( "DEM/C.DEM" ) );
		assertTrue( !cv35.isValid( "DE\\/C.DEM" ) );
	}

	/* ==================== Private Methods ===================== */

	private void assertEqualities (
		final String label,
		final TextFormat a,
		final TextFormat b
	) {
		assertEquals( label, a, b );
		assertEquals( label + " hashcode", a.hashCode(), b.hashCode() );
		assertEquals( label + "a toString", label, a.toString() );
		assertEquals( label + "b toString", label, b.toString() );
	}
}
