// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Date;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import Int.ecb.gesmescb.time.RegularInterval;

/* ***************************** Class ****************************** */

/**
 * Contain the identification
 * and data information which represent a time series in Gesmes/TS.
 *
 * @author
 * <A HREF="mailto:klo@FAME.com">Kam Lo</A>
 * @author
 * <A HREF="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</A>
 *
 * @version @(#)$Id: TimeSeries.java,v 1.1 2017/01/19 16:14:04 trainot Exp $
 */

public class TimeSeries extends AbstractCube {

	/* ======================= Constants ======================== */

	/**
	 *   serialVersionUID generated by eclipse.
	 */
	private static final long serialVersionUID = 8336784118558517273L;

	/**
	 * A constant empty array to support type checking in the call
	 * to the method List.toArray( String[] ).
	 * Only the type String[] is relevant.
	 *
	 * The size must be 0, otherwise instance data would be written
	 * to the static array by List.toArray!
	 *
	 * Note that in case of empty keys
	 * a reference to the static array itself
	 * becomes part of the instance.
	 * But empty is anyway empty.
	 */
	private static final Observation[] emptyObservationArray
	= new Observation[0];

	/* ========================= Fields ========================= */

	/**
	 * The array containing all observations.
	 *
	 * @serial
	 */
	private Observation[] observations;

	/* -------------------- Transient Fields -------------------- */

	/**
	 * Precomputed hash code.
	 */
	private transient int hashCode;

	/* ====================== Constructors ====================== */

	/**
	 * Construct a new time series object.
	 * Observations are passed in an array.
	 *
	 * @param key time series key.
	 * @param period time period of observations.
	 * @param periodFormat time period format.
	 * @param observations observation list.
	 *
	 * @throws NullPointerException
	 * if any of the arguments is <code>null</code>.
	 */
	public TimeSeries (
		final String key,
		final String period,
		final String periodFormat,
		final Observation[] observations
	) {
		super ( key, period, periodFormat );
		checkObservations(
			this.observations = convertArrayToArray( observations )
		);
	}

	/**
	 * Construct a new time series object.
	 * Observations are passed in a List.
	 *
	 * @param key time series key.
	 * @param period time period of observations.
	 * @param periodFormat time period format.
	 * @param observations observation list.
	 *
	 * @throws ArrayStoreException if observations contains
	 * objects of a different class than {@link Observation}.
	 * @throws NullPointerException
	 * if any of the arguments is <code>null</code>.
	 */
	public TimeSeries (
		final String key,
		final String period,
		final String periodFormat,
		final List observations
	) {
		super ( key, period, periodFormat );
		checkObservations(
			this.observations = convertListToArray( observations )
		);
	}

	/**
	 * Construct a new time series object.
	 * Observations are passed by iteration.
	 *
	 * @param key time series key.
	 * @param period time period of observations.
	 * @param periodFormat time period format.
	 * @param observations observation iterator.
	 *
	 * @throws ClassCastException if iterator produces
	 * objects of a different class than {@link Observation}.
	 * @throws NullPointerException
	 * if any of the arguments is <code>null</code>.
	 */
	public TimeSeries (
		final String key,
		final String period,
		final String periodFormat,
		final Iterator observations
	) {
		super ( key, period, periodFormat );
		checkObservations(
			this.observations = convertIteratorToArray( observations )
		);
	}

	/**
	 * Construct a new time series object.
	 * Observations are passed in an array.
	 *
	 * @param key time series key.
	 * @param interval time range of observations.
	 * @param observations observation array.
	 *
	 * @throws ClassCastException if iterator produces
	 * objects of a different class than {@link Observation}.
	 * @throws NullPointerException
	 * if <code>key</code> or <code>observations</code>
	 * is <code>null</code>.
	 * @throws IllegalArgumentException
	 * if <code>interval</code> is <code>null</code>,
	 * but <code>observations</code> is not empty.
	 */
	public TimeSeries (
		final String key,
		final RegularInterval interval,
		final Observation[] observations
	) {
		super ( key, interval );
		if (
			interval == null
			&&
			observations.length > 0
		) {
			throw new IllegalArgumentException(
				"Trying to construct a nonempty TimeSeries"
				+ " with null time interval."
			);
		}
		checkObservations(
			this.observations = convertArrayToArray( observations )
		);
	}

	/**
	 * Construct a new time series object.
	 * Observations are passed in a List.
	 *
	 * @param key time series key.
	 * @param interval time range of observations.
	 * @param observations observation list.
	 *
	 * @throws ClassCastException if iterator produces
	 * objects of a different class than {@link Observation}.
	 * @throws NullPointerException
	 * if <code>key</code> or <code>observations</code>
	 * is <code>null</code>.
	 * @throws IllegalArgumentException
	 * if <code>interval</code> is <code>null</code>,
	 * but <code>observations</code> is not empty.
	 */
	public TimeSeries (
		final String key,
		final RegularInterval interval,
		final List observations
	) {
		super ( key, interval );
		if (
			interval == null
			&&
			observations.size() > 0
		) {
			throw new IllegalArgumentException(
				"Trying to construct a nonempty TimeSeries"
				+ " with null time interval."
			);
		}
		checkObservations(
			this.observations = convertListToArray( observations )
		);
	}

	/**
	 * Construct a new time series object.
	 * Observations are passed by iteration.
	 *
	 * @param key time series key.
	 * @param interval time range of observations.
	 * @param observations observation iterator.
	 *
	 * @throws ClassCastException if iterator produces
	 * objects of a different class than {@link Observation}.
	 * @throws NullPointerException
	 * if any of the arguments is <code>null</code>.
	 * @throws IllegalArgumentException
	 * if <code>interval</code> is <code>null</code>,
	 * but <code>observations</code> has some items.
	 */
	public TimeSeries (
		final String key,
		final RegularInterval interval,
		final Iterator observations
	) {
		super ( key, interval );
		if (
			interval == null
			&&
			observations.hasNext()
		) {
			throw new IllegalArgumentException(
				"Trying to construct a nonempty TimeSeries"
				+ " with null time interval."
			);
		}
		checkObservations(
			this.observations = convertIteratorToArray( observations )
		);
	}

    /**
	 * Construct a new time series object.
	 *
	 * Observations are passed as a Map.
	 *
	 * @param key time series key.
	 * @param interval time range of observations.
	 * @param tree Map of observations indexed by date.
	 * @throws IllegalArgumentException
	 * if <code>interval</code> is <code>null</code>,
	 * but <code>observations</code> has some items.
	 */
	public TimeSeries (
	    final String key,
        final RegularInterval interval,
	    final Map tree
	) {
	    super( key, interval );
	    // ensures that the list contains EMPTY_OBSERVATIONs for the dates
	    // which are not keys of the TreeMap.
	    // Trailing EMPTY_OBSERVATIONs are not considered.
	    final List observations = new ArrayList(); // we want to return
	    final Iterator intervalIterator = interval.iterator();  
	    while ( intervalIterator.hasNext() ) {
	        final Date period = 
	            ( (RegularInterval) intervalIterator.next() ).getBegin();
	        if ( tree.containsKey(period)) {
	            observations.add( tree.get( period ) );
	        } else {
	            observations.add( Observation.EMPTY_OBSERVATION );
	        }
	    }
        if (
            interval == null
            &&
            observations.size() > 0
        ) {
            throw new IllegalArgumentException(
                "Trying to construct a nonempty TimeSeries"
                + " with null time interval."
            );
        }
        checkObservations(
            this.observations = convertListToArray( observations )
        );
	}

	/* ===================== Public Methods ===================== */

	/**
	 * Returns an iterator over the observations.
	 *
	 * @return iterator over observations.
	 */
	public Iterator observationIterator () {
		return (
			Collections.unmodifiableList(
				Arrays.asList( observations )
			)
		).iterator();
	}

	/**
	 * Returns an iterator over the observations.
	 *
	 * @return iterator over observations.
	 *
	 * @deprecated Renamed to {@link #observationIterator}.
	 */
	public Iterator getObservationIterator () {
		return observationIterator();
	}

	/**
	 * Returns the number of {@link Observation}s stored in this
	 * <CODE>TimeSeries</CODE>.
	 */
	public int length () {
		return observations.length;
	}

	/**
	 * Indicates whether some other Object is "equal to" this TimeSeries.
	 *
	 * @param obj the reference object with which to compare
	 *
	 * @return true if this TimeSeries is equal to the argument;
	 * false otherwise.
	 */
	public boolean equals ( final Object obj ) {
		if ( this == obj ) return true;
		if ( obj instanceof TimeSeries ) {
			final TimeSeries other = (TimeSeries) obj;
			return
				super.equals( other )
				&& Arrays.equals(
					observations, other.observations
				);
		}
		return false;
	}

	/**
	 * Returns the hash code value for this TimeSeries.
	 * If ts1.equals(ts2) then ts1.hashCode()==ts2.hashCode()
	 * for any two TimeSeries ts1 and ts2,
	 * as required by the general contract of {@link Object#hashCode}.
	 *
	 * @return hash code value for this object.
	 */
	public int hashCode () {
		// hashCode itself is used as flag to recompute
		// for thread safety.
		if ( hashCode == 0 ) {
			hashCode = super.hashCode()
				^ Arrays.asList( observations ).hashCode();
		}
		return hashCode;
	}

	/**
	 * Returns a string representation of the TimeSeries
	 * in a human readable format.
	 *
	 * @return a string representation of the TimeSeries.
	 */
	public String toString () {
		// Build a string for the observations.
		final StringBuffer stringBuffer = new StringBuffer();
		for ( int i = 0; i < observations.length; ++i ) {
			if ( i > 0 ) stringBuffer.append( ", " );
			stringBuffer.append( observations[i].toString() );
		}

		return getClass().getName()
			+ " ( key = " + getKey()
			+ ", period = " + getPeriod()
			+ ", periodFormat = " + getPeriodFormat()
			+ ", observations = ( "
			+ stringBuffer.toString() + " ) )";
	}

	/* ==================== Private Methods ===================== */

	private static Observation[] convertArrayToArray (
		final Observation[] array
	) {
		return (Observation[]) array.clone();
	}

	private static Observation[] convertListToArray (
		final List list
	) {
		return (Observation[]) list.toArray(
			emptyObservationArray
		);
	}

	private static Observation[] convertIteratorToArray (
		final Iterator iterator
	) {
		final List list = new ArrayList();
		while ( iterator.hasNext() ) {
			list.add( iterator.next() );
		}
		return convertListToArray( list );
	}

	private static void checkObservations (
		final Observation[] observations
	) {
		for ( int i = 0; i < observations.length; ++i ) {
			if ( observations[i] == null ) {
				throw new NullPointerException(
					"Trying to construct a TimeSeries with a null observation."
				);
			}
		}
	}
}
