// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb;

import java.util.Iterator;
import java.util.Arrays;

import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;

import Int.ecb.gesmescb.time.RegularInterval;
import Int.ecb.gesmescb.time.GregorianRegularInterval;

import Int.ecb.gesmescb.Observation;
import Int.ecb.gesmescb.TimeSeries;

/* ***************************** Class ****************************** */

/**
 * <P>Tests the implementation of the class {@link TimeSeries}.</P>
 *
 * @author
 * <A HREF="mailto:Sami.Airo@ECB.int">Sami Airo</A>
 * @author
 * <A HREF="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</A>
 *
 * @version @(#)$Id: TimeSeriesTest.java,v 1.1 2017/01/19 16:14:04 trainot Exp $
 */

public class TimeSeriesTest extends TestCase {

	/* ========================= Fields ========================= */

	/* ====================== Constructors ====================== */

	/**
	 * Constructs a TimeSeriesTest with the specified name.
	 *
	 * @param name Test case name
	 */
	public TimeSeriesTest ( final String name ) {
		super( name );
	}

	/* ===================== Public Methods ===================== */

	/**
	 * Assembles and returns a test suite for all
	 * the test methods of this test case.
	 *
	 * @return a non-null test suite
	 */
	public static Test suite () {
		return new TestSuite( TimeSeriesTest.class );
	}

	/**
	 * Runs the test case.
	 */
	public static void main ( final String args[] ) {
		final String[] testCaseName = { TimeSeriesTest.class.getName() };
		junit.textui.TestRunner.main( testCaseName );
	}

	/* ---------------------- Test Methods ---------------------- */

	/**
	 * Test the constructors.
	 */
	public void testConstruct () {
		final String
			key = "key",
			period = "20010202",
			periodFormat = "102";
		final Observation[] observations = {
			new Observation( "", "", "", "" ),
			new Observation( "1", "", "", "" ),
			new Observation( "1", "2", "", "" ),
			new Observation( "1", "2", "3", "" ),
			new Observation( "1", "2", "3", "4" )
		};
		final RegularInterval interval = new GregorianRegularInterval(
			RegularInterval.DAILY, 2001, 32, 1
		);

		final TimeSeries ts1 = new TimeSeries(
			key, period, periodFormat, observations
		);
		final TimeSeries ts2 = new TimeSeries(
			key, period, periodFormat, Arrays.asList( observations )
		);
		final TimeSeries ts3 = new TimeSeries(
			key, period, periodFormat, Arrays.asList( observations ).iterator()
		);
		final TimeSeries ts4 = new TimeSeries(
			key, interval, observations
		);
		final TimeSeries ts5 = new TimeSeries(
			key, interval, Arrays.asList( observations )
		);
		final TimeSeries ts6 = new TimeSeries(
			key, interval, Arrays.asList( observations ).iterator()
		);

		assertEquals( "1 equal 2", ts1, ts2 );
		assertEquals( "2 equal 3", ts2, ts3 );
		assertEquals( "3 equal 4", ts3, ts4 );
		assertEquals( "4 equal 5", ts4, ts5 );
		assertEquals( "5 equal 6", ts5, ts6 );
		assertEquals( "6 equal 1", ts6, ts1 );

		assertEquals( "1 hash 2", ts1.hashCode(), ts2.hashCode() );
		assertEquals( "2 hash 3", ts2.hashCode(), ts3.hashCode() );
		assertEquals( "3 hash 4", ts3.hashCode(), ts4.hashCode() );
		assertEquals( "4 hash 5", ts4.hashCode(), ts5.hashCode() );
		assertEquals( "5 hash 6", ts5.hashCode(), ts6.hashCode() );
	}

	/**
	 * Test the equals, hashCode and toString methods.
	 */
	public void testEquals () {
		// Create necessary Observations.
		Observation[] original_observations = new Observation[10];
		Observation[] equal_observations
		= new Observation[original_observations.length];
		Observation[] modified_observations
		= new Observation[original_observations.length+1];
		for ( int i = 0; i < original_observations.length; ++i ) {
			// Create observations.
			original_observations[i] = new Observation(
				"value" + i,
				"status" + i,
				"confidentiality" + i,
				"prebreakvalue" + i
			);
			// And equal observations.
			equal_observations[i] = new Observation(
				"value" + i,
				"status" + i,
				"confidentiality" + i,
				"prebreakvalue" + i
			);
			// And modified observations
			// (which are equal but there is one more).
			modified_observations[i] = new Observation(
				"value" + i,
				"status" + i,
				"confidentiality" + i,
				"prebreakvalue" + i
			);
		}
		modified_observations[original_observations.length]
		= new Observation( "one", "more", "observation", "here" );

		// Create constructor arguments.
		Object[] original = {
			"key",
			"1995040220010402",
			"711",
			original_observations
		};
		// Create modifications.
		Object[] modified = {
			"Key",
			"1995040220000402",
			"911",	// invalid GESMES/TS, but does not matter here.
			modified_observations
		};

		// Create TimeSeries.
		TimeSeries ts0 = new TimeSeries(
			(String) original[0],
			(String) original[1],
			(String) original[2],
			(Observation[]) original[3]
		);
		// Create "equal" TimeSeries from equal, non-identical Strings.
		TimeSeries ts1 = new TimeSeries(
			new String( (String) original[0] ),
			new String( (String) original[1] ),
			new String( (String) original[2] ),
			equal_observations
		);
		assertTrue( "ts0 equals ts1", ts0.equals( ts1 ) );
		assertTrue(
			"ts0.hashCode == ts1.hashCode",
			ts0.hashCode() == ts1.hashCode()
		);
		assertTrue(
			"ts0.toString equals ts1.toString",
			ts0.toString().equals( ts1.toString() )
		);

		// Next create unequal TimeSeries,
		// which are unequal at varying positions.
		TimeSeries[] ts = new TimeSeries[original.length];
		for ( int i = 0; i < ts.length; ++i ) {
			ts[i] = new TimeSeries(
				(String)( i == 0 ? modified[0] : original[0] ),
				(String)( i == 1 ? modified[1] : original[1] ),
				(String)( i == 2 ? modified[2] : original[2] ),
				(Observation[])(
					i == 3 ? modified[3] : original[3]
				)
			);

			assertTrue(
				"ts["+i+"] not equals ts0",
				!ts[i].equals( ts0 )
			);
			// Inequality of the hashCode is not really contract.
			// But the probability of eventual equal hashCode is
			// far less than that of a buggy
			// hashCode-implementation.
			// Therefore check.
			assertTrue(
				"ts[" + i + "].hashCode != ts0.hashCode",
				ts[i].hashCode() != ts0.hashCode()
			);
			assertTrue(
				"ts[" + i + "].toString != ts0.toString",
				!ts[i].toString().equals( ts0.toString() )
			);
		}
	}

	/**
	 * Test the get-type methods.
	 */
	public void testProperties () {
		Observation o0, o1, o2;
		TimeSeries ts0 = new TimeSeries(
			"CO.SI.NUS", "200109200110", "710",
			new Observation[] {
				o0 = new Observation( "1", "2", "3", "4" ),
				o1 = new Observation( "5", "6", "7", "8" ),
				o2 = new Observation( "9", "10", "11", "12" )
			}
		);

		Iterator iterator = ts0.observationIterator();
		assertTrue( "0th observation == o0", o0 == iterator.next() );
		assertTrue( "1st observation == o1", o1 == iterator.next() );
		assertTrue( "2nd observation == o2", o2 == iterator.next() );
		assertTrue( "Only 3 observations", !iterator.hasNext() );

		assertTrue( "length check", ts0.length() == 3 );
	}

	/**
	 * Tests serialization.
	 */
	public void testSerialization () {
		TimeSeries ts0 = new TimeSeries(
			"B.DI.D", "0815", "1827",
			new Observation[] {
				new Observation( "Bh", "Wh", "Gh", "Prx" ),
				new Observation( "Jh", "Bh", "Gh", "F" )
			}
		);
		final TimeSeries ts1
		= (TimeSeries) AttributeTest.transserialize( ts0 );
		assertEquals( ts0, ts1 );
		assertEquals( "key length", 3, ts1.getKeyLength() );
	}
}
