// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2003, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.application;

import java.util.List;
import java.util.Properties;
import java.util.Iterator;

import java.io.File;
import java.io.InputStream;
import java.io.BufferedInputStream;
import java.io.FileInputStream;
import java.io.OutputStream;
import java.io.BufferedOutputStream;
import java.io.FileOutputStream;
import java.io.Reader;
import java.io.InputStreamReader;
import java.io.IOException;
import java.io.FileNotFoundException;

import java.awt.Frame;
import java.awt.MenuBar;
import java.awt.Menu;
import java.awt.MenuItem;
import java.awt.MenuShortcut;
import java.awt.CheckboxMenuItem;
import java.awt.TextArea;
import java.awt.FileDialog;
import java.awt.Color;

import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import java.awt.event.ItemListener;
import java.awt.event.ItemEvent;

import java.awt.dnd.DropTarget;
import java.awt.dnd.DropTargetDragEvent;
import java.awt.dnd.DropTargetDropEvent;
import java.awt.dnd.DnDConstants;
import java.awt.datatransfer.DataFlavor;
import java.awt.datatransfer.Transferable;
import java.awt.datatransfer.UnsupportedFlavorException;

import Int.ecb.logging.LogManager;
import Int.ecb.logging.Handler;
import Int.ecb.logging.Level;

import Int.ecb.gesmescb.util.ComponentHelper;
import Int.ecb.gesmescb.util.NormalLogFormatter;
import Int.ecb.gesmescb.util.TextAreaHandler;
import Int.ecb.gesmescb.util.InfoDialog;

import Int.ecb.gesmescb.sieve.PreferenceConfiguredFilter;

import Int.ecb.gesmescb.edifact.EDIFACTReader;
import Int.ecb.gesmescb.fame.FAMEReader;
import Int.ecb.gesmescb.xml.XMLReader;
import Int.ecb.gesmescb.csv.SPFReader;

/* ***************************** Class ****************************** */

/**
 * This is the top application class for the graphical user interface.
 * For the actual job it uses the {@link BatchTextUI} and its preferences.
 *
 * @author
 * <a href="mailto:Paula.Rocha@ECB.int">Paula Rocha</a>
 * @author
 * <a href="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</a>
 *
 * @version @(#)$Id: AwUI.java,v 1.2 2020/07/20 07:03:02 delical Exp $
 */
public class AwUI {

	/* ========================= Fields ========================= */

	private final Frame frame;
	protected final ComponentHelper componentHelper;
	private final String configFile;
	private final Properties properties;
	private final CheckGroupListener checkGroupListener;

	/* ====================== Constructors ====================== */

	private AwUI () throws IOException {
		final TextArea textArea = new TextArea( 25, 80 );
		textArea.setEditable( false );
		final LogManager logManager = LogManager.getLogManager();
		final Handler handler = new TextAreaHandler(
			textArea, new NormalLogFormatter()
		);
		handler.setLevel( Level.INFO );
		logManager.addGlobalHandler( handler );

		componentHelper = new ComponentHelper( "", this );
		configFile = componentHelper.getPreference( "config", "awui.cnf" );
		properties = new Properties();
		try {
			final InputStream inputStream = new BufferedInputStream(
				new FileInputStream( configFile )
			);
			properties.load( inputStream );
			inputStream.close();
		}
		catch ( final FileNotFoundException e ) {}
		catch ( final IOException e ) {
			componentHelper.logger.log(
				Level.WARNING, componentHelper.logName, "AwUI",
				"Could not read application defaults: " + e.getMessage()
			);
		}

		frame = new Frame( "ECB's Checker" );
		final ExitListener exitListener = new ExitListener( frame );
		frame.addWindowListener( exitListener );
		final MenuBar menuBar = new MenuBar();
		frame.setMenuBar( menuBar );
		final Menu fileMenu = new Menu( "File" );
		menuBar.add( fileMenu );
		final MenuItem openFileItem = new MenuItem(
			"Open...", new MenuShortcut( 'O' )
		);
		fileMenu.add( openFileItem );
		final FileDialog openFileDialog = new FileDialog(
			frame, "File Chooser"
		);
		openFileItem.addActionListener( new OpenFileDialogListener(
			openFileDialog, this
		) );
		final MenuItem clearItem = new MenuItem(
			"Clear Window", new MenuShortcut( 'N' )
		);
		fileMenu.add( clearItem );
		clearItem.addActionListener( new ClearListener( textArea ) );
		final MenuItem exitItem = new MenuItem(
			"Exit...", new MenuShortcut( 'Q' )
		);
		fileMenu.add( exitItem );
		exitItem.addActionListener( exitListener );
		final Menu structureMenu = new Menu( "Settings" );
		menuBar.add( structureMenu );
		final MenuItem structureFileItem = new MenuItem(
			"Set Default Structural Definitions...", new MenuShortcut( 'D' )
		);
		structureMenu.add( structureFileItem );
		final FileDialog structureFileDialog = new FileDialog(
			frame, "Structural Definitions Chooser"
		);
		structureFileItem.addActionListener( new StructureFileDialogListener(
			structureFileDialog, this
		) );
		final CheckboxMenuItem
			trust = new CheckboxMenuItem( "Trust Default Structural Definitions Without Checking", false ),
			reject = new CheckboxMenuItem( "Reject Structural Definitions", false ),
			check = new CheckboxMenuItem( "Check Structural Definitions", false ),
			memorise = new CheckboxMenuItem( "Memorise Structural Definitions", false );
		trust.setName( "Do Not Check Default" );
		reject.setName( "Reject" );
		check.setName( "Check" );
		memorise.setName( "Memorise" );
		trust.setShortcut( new MenuShortcut( 'T' ) );
		reject.setShortcut( new MenuShortcut( 'R' ) );
		check.setShortcut( new MenuShortcut( 'K' ) );
		memorise.setShortcut( new MenuShortcut( 'M' ) );
		structureMenu.add( trust );
		structureMenu.add( reject );
		structureMenu.add( check );
		structureMenu.add( memorise );
		final String structureHandling = properties.getProperty(
			"structureHandling", "reject"
		);
		checkGroupListener = new CheckGroupListener(
			structureHandling.equalsIgnoreCase( "memorise" )
			? memorise : (
				structureHandling.equalsIgnoreCase( "check" )
				? check : reject
			),
			this
		);
		trust.addItemListener( new TrustDefaultsListener( this ) );
		reject.addItemListener( checkGroupListener );
		check.addItemListener( checkGroupListener );
		memorise.addItemListener( checkGroupListener );
		final Menu helpMenu = new Menu( "Help" );
		menuBar.add( helpMenu );
		final MenuItem helpItem = new MenuItem(
			"Help", new MenuShortcut( 'H' )
		);
		helpMenu.add( helpItem );
		helpItem.addActionListener( new HelpDialogListener(
			frame
		) );
		final MenuItem licenseItem = new MenuItem(
			"License", new MenuShortcut( 'L' )
		);
		helpMenu.add( licenseItem );
		licenseItem.addActionListener( new LicenseDialogListener(
			frame
		) );
		frame.add( textArea );
		textArea.setDropTarget(
			new CheckingTarget( this )
		);
		frame.pack();
		frame.setVisible(true);

		final String structureFile = properties.getProperty( "structureFile" );
		if ( structureFile != null ) {
			// TODO synchronized ( this ) { // Force write back Thread's cache
				checkFile( structureFile, true );
			// }
		}
	}

	/* ===================== Public Methods ===================== */

	public static void main ( final String[] arg ) throws IOException {
		if ( arg.length > 0 ) {
			BatchTextUI.main( arg );
		} else {
			final AwUI gui = new AwUI();
		}
	}

	/* ==================== Package Methods ===================== */

	String getProperty ( final String name, final String defaultValue ) {
		return properties.getProperty( name, defaultValue );
	}

	void setProperty ( final String name, final String value ) {
		properties.setProperty( name, value );
	}

	void saveProperties () {
		try {
			final OutputStream outputStream = new BufferedOutputStream(
				new FileOutputStream( configFile )
			);
			properties.store(
				outputStream, "AwUI properties - overwritten automatically"
			);
			outputStream.close();
		}
		catch ( final IOException e ) {
			componentHelper.logger.log(
				Level.WARNING, componentHelper.logName, "AwUI",
				"Could not save application defaults: " + e.getMessage()
			);
		}
	}

	/**
	 * @param isDefaults whether the file is treated
	 * as the default structural definitions.
	 */
	void checkFile ( final String filename, final boolean isDefaults ) {
		final int dotIndex = filename.lastIndexOf( '.' );
		final String extension = dotIndex >= 0
			? filename.substring( dotIndex + 1 ) : "";
		final String readerClassName;
		if ( "csv".equalsIgnoreCase( extension ) ) {
			readerClassName = SPFReader.class.getName();
		} else if ( "xml".equalsIgnoreCase( extension ) ) {
			readerClassName = XMLReader.class.getName();
		} else if ( "db".equalsIgnoreCase( extension ) ) {
			readerClassName = FAMEReader.class.getName();
		} else {
			readerClassName = EDIFACTReader.class.getName();
		}
		System.setProperty(
			BatchTextUI.class.getName() + ".dataReaderClass",
			readerClassName
		);

		if (
			isDefaults
			&&
			"true".equalsIgnoreCase( getProperty( "trustDefaults", "false" ) )
		) {
			System.setProperty(
				BatchTextUI.class.getName()
				+ ".useSemanticChecker",
				"false"
			);
		} else {
			System.setProperty(
				BatchTextUI.class.getName()
				+ ".useSemanticChecker",
				"true"
			);
		}

		final String checkedName = checkGroupListener.getCheckedName();
		if ( isDefaults || checkedName.equalsIgnoreCase( "memorise" ) ) {
			System.setProperty(
				BatchTextUI.class.getName()
				+ ".acceptStructuralDefinitions",
				"true"
			);
			System.setProperty(
				BatchTextUI.class.getName() + ".useSieve",
				"false"
			);
		} else if ( checkedName.equalsIgnoreCase( "check" ) ) {
			System.setProperty(
				BatchTextUI.class.getName()
				+ ".acceptStructuralDefinitions",
				"false"
			);
			System.setProperty(
				BatchTextUI.class.getName() + ".useSieve",
				"false"
			);
		} else if ( checkedName.equalsIgnoreCase( "reject" ) ) {
			System.setProperty(
				BatchTextUI.class.getName()
				+ ".acceptStructuralDefinitions",
				"false"
			);
			System.setProperty(
				BatchTextUI.class.getName() + ".useSieve",
				isDefaults ? "false" : "true"
			);
			System.setProperty(
				PreferenceConfiguredFilter.class.getName()
				+ ".Data.*.*.*.*.*",
				"*"
			);
		} else {
			// This should not be possible
			componentHelper.logger.log(
				Level.SEVERE, componentHelper.logName, "AwUI",
				"AwUI implementation error 526: keyword \""
				 + checkedName + "\"."
			);
		}

		final BatchTextUI btui = new BatchTextUI();
		btui.run( new String[]{ filename } );

		final Properties properties = System.getProperties();
		properties.remove(
			BatchTextUI.class.getName() + ".dataReaderClass"
		);
		properties.remove(
			BatchTextUI.class.getName() + ".useSemanticChecker"
		);
		properties.remove(
			BatchTextUI.class.getName() + ".acceptStructuralDefinitions"
		);
		properties.remove(
			BatchTextUI.class.getName() + ".useSieve"
		);
		properties.remove(
			PreferenceConfiguredFilter.class.getName() + ".Data.*.*.*.*.*"
		);
	}

	/* ==================== Private Methods ===================== */
}

/* ***************************** Class ****************************** */

class ExitListener extends WindowAdapter implements ActionListener {
	private final InfoDialog infoDialog;

	public ExitListener ( final Frame owner ) {
		infoDialog = new InfoDialog(
			owner,
			"Exit",
			"Do you truly want to EXIT?",
			new String[] { "Noo, bro'", "Damn straight" },
			true
		);
		infoDialog.pack();
	}

	public void windowClosing ( final WindowEvent event ) {
		exit();
	}

	public void actionPerformed ( final ActionEvent event ) {
		exit();
	}

	private void exit () {
		infoDialog.setVisible(true);
		switch ( infoDialog.getChosen() ) {
			case 0: break;
			case 1:
			case 2: System.exit( 0 );
		}
	}
}

/* ***************************** Class ****************************** */

class ClearListener implements ActionListener {
	private final TextArea textArea;
	private final Color color;

	public ClearListener ( final TextArea textArea ) {
		this.textArea = textArea;
		color = textArea.getBackground();
	}

	public void actionPerformed ( final ActionEvent event ) {
		textArea.setText( "" );
		textArea.setBackground( color );
	}
}

/* ***************************** Class ****************************** */

class HelpDialogListener implements ActionListener {
	private final InfoDialog infoDialog;
	private static final String helpText
	= "Dear User!\n\n"
		+ "We congratulate you on using the ECB's Gesmes/TS checker!"
		+ " This is free software. You can find the license information"
		+ " in the \"Help\" menu under the item \"License\".\n\n"
		+ "Purpose\n"
		+ "--------------------------------------------------------\n"
		+ "The checker validates files in the Gesmes/TS and earlier Gesmes/CB"
		+ " EDIFACT formats. To learn more about these formats,"
		+ " visit http://www.ecb.int/stats/services/gesmes/html/index.en.html"
		+ " Gesmes/TS and Gesmes/CB foresee two kinds of files:\n"
		+ "1) Structural definitions files determine"
		+ " the structure of data (breakdown of multi-dimensional"
		+ " time series keys and allowed code values)\n"
		+ "2) Data files transport the actual time series data"
		+ " (observation values and attributes)\n\n"
		+ "Checking data\n"
		+ "--------------------------------------------------------\n"
		+ "To check a file, drag and drop it into the checker window."
		+ " After the checks are finished, the window background colour"
		+ " indicates the highest reported severity of problem encountered"
		+ " while checking. To reset the background, use in the"
		+ " \"File\" menu the item \"Clear\"."
		+ " Remember that you must provide suitable structural definitions"
		+ " before you can check a data file.\n\n"
		+ "Checking structural definitions\n"
		+ "--------------------------------------------------------\n"
		+ "In the \"Structural Definitions\" menu"
		+ " activate either \"check\" or \"memorise\"."
		+ " In the latter case, the checked structural definitions will"
		+ " be used to check subsequent data files."
		+ " Memorised structural definitions will be forgotten only upon exit.\n\n"
		+ "To memorise structural definitions without checking, activate"
		+ " \"trust\". This can be used if they are not 100% living up to the"
		+ " standard, but still meaningful. However, it is not recommended.\n\n"
		+ "Setting default structural definitions\n"
		+ "--------------------------------------------------------\n"
		+ "If you usually use the same structural definitions file"
		+ " over and over again,"
		+ " you can set it as the default. Its content will then be read"
		+ " and memorised upon starting up the graphical user interface.\n";

	public HelpDialogListener ( final Frame owner ) {
		infoDialog = new InfoDialog( owner, "Help", helpText );
		infoDialog.pack();
	}

	public void actionPerformed ( final ActionEvent event ) {
		infoDialog.setVisible(true);
	}
}

/* ***************************** Class ****************************** */

class LicenseDialogListener implements ActionListener {
	private final InfoDialog infoDialog;

	public LicenseDialogListener ( final Frame owner ) throws IOException {
		final ClassLoader loader = getClass().getClassLoader();
		final InputStream in = loader.getResourceAsStream( "license.txt" );
		if ( in == null ) {
			throw new FileNotFoundException(
				"License file not found."
			);
		}
		final Reader reader = new InputStreamReader(
			new BufferedInputStream( in )
		);
		infoDialog = new InfoDialog( owner, "License", reader );
		reader.close();
		infoDialog.pack();
	}

	public void actionPerformed ( final ActionEvent event ) {
		infoDialog.setVisible(true);
	}
}

/* ***************************** Class ****************************** */

class OpenFileDialogListener implements ActionListener {
	private final FileDialog fileDialog;
	private final AwUI awui;

	public OpenFileDialogListener (
		final FileDialog fileDialog,
		final AwUI awui
	) {
		this.fileDialog = fileDialog;
		this.awui = awui;
	}

	public void actionPerformed ( final ActionEvent event ) {
		fileDialog.setVisible(true);
		final String
			directory = fileDialog.getDirectory(),
			file = fileDialog.getFile();
		if ( file != null ) {
			final File newFile = new File( directory, file );
			final String filename = newFile.getPath();
			awui.checkFile( filename, false );
		}
	}
}

/* ***************************** Class ****************************** */

class StructureFileDialogListener implements ActionListener {
	private final FileDialog fileDialog;
	private final AwUI awui;

	public StructureFileDialogListener (
		final FileDialog fileDialog,
		final AwUI awui
	) {
		this.fileDialog = fileDialog;
		this.awui = awui;
	}

	public void actionPerformed ( final ActionEvent event ) {
		final String filename = awui.getProperty( "structureFile", null );
		final File oldFile;
		if ( filename != null ) {
			oldFile = new File( filename );
			fileDialog.setDirectory( oldFile.getParent() );
			fileDialog.setFile( oldFile.getName() );
		} else {
			oldFile = null;
		}
		fileDialog.setVisible(true);
		final String
			directory = fileDialog.getDirectory(),
			file = fileDialog.getFile();
		if ( file != null ) {
			final File newFile = new File( directory, file );
			if ( oldFile == null || !oldFile.equals( newFile ) ) {
				final String newFilename = newFile.getPath();
				awui.checkFile( newFilename, true );
				awui.setProperty( "structureFile", newFilename );
				awui.saveProperties();
			}
		}
	}
}

/* ***************************** Class ****************************** */

class TrustDefaultsListener implements ItemListener {
	private final AwUI awui;

	/** @param checked sets the originally true item */
	public TrustDefaultsListener (
		final AwUI awui
	) {
		this.awui = awui;
	}

	public void itemStateChanged ( final ItemEvent e ) {
		switch ( e.getStateChange() ) {
			case ItemEvent.SELECTED:
				awui.setProperty( "trustDefaults", "true" );
				break;
			case ItemEvent.DESELECTED:
				awui.setProperty( "trustDefaults", "false" );
				break;
		}
		awui.saveProperties();
	}
}

/* ***************************** Class ****************************** */

class CheckGroupListener implements ItemListener {
	private final AwUI awui;
	private CheckboxMenuItem checked;

	/** @param checked sets the originally true item */
	public CheckGroupListener (
		final CheckboxMenuItem checked,
		final AwUI awui
	) {
		this.awui = awui;
		this.checked = checked;
		if ( !checked.getState() ) {
			checked.setState( true );
		}
	}

	public void itemStateChanged ( final ItemEvent e ) {
		final CheckboxMenuItem item = (CheckboxMenuItem) e.getItemSelectable();
		switch ( e.getStateChange() ) {
			case ItemEvent.SELECTED:
				if ( checked != null && checked != item ) {
					checked.setState( false );
				}
				checked = item;
				awui.setProperty( "structureHandling", item.getName() );
				awui.saveProperties();
				break;
			case ItemEvent.DESELECTED:
				item.setState( true );
				break;
		}
	}

	public String getCheckedName () {
		return checked.getName();
	}
}

/* ***************************** Class ****************************** */

class CheckingTarget extends DropTarget {
	private final AwUI awui;

	public CheckingTarget ( final AwUI awui ) {
		this.awui = awui;
	}

	public void dragEnter ( final DropTargetDragEvent e ) {
		judgeDrag( e );
	}

	public void dragOver ( final DropTargetDragEvent e ) {
		judgeDrag( e );
	}

	public void dragActionChanged ( final DropTargetDragEvent e ) {
		judgeDrag( e );
	}

	public void drop ( final DropTargetDropEvent e ) {
		if ( e.isDataFlavorSupported( DataFlavor.javaFileListFlavor ) ) {
			try {
				e.acceptDrop( DnDConstants.ACTION_COPY_OR_MOVE );
				final Transferable transferable = e.getTransferable();
				final List files = (List) transferable.getTransferData(
					DataFlavor.javaFileListFlavor
				);

				for (
					final Iterator iterator = files.iterator();
					iterator.hasNext();
				) {
					final File file = (File) iterator.next();
					final String filename = file.getPath();
					awui.checkFile( filename, false );
				}

				e.dropComplete( true );
			} catch ( final UnsupportedFlavorException ex ) {
				awui.componentHelper.logger.log(
					Level.SEVERE, awui.componentHelper.logName,
					"CheckingTarget",
					"Drag and drop failed: " + ex.getMessage()
				);
			} catch ( final IOException ioe ) {
				awui.componentHelper.logger.log(
					Level.SEVERE, awui.componentHelper.logName,
					"CheckingTarget",
					"Opening file failed: " + ioe.getMessage()
				);
			}
		}
	}

	private void judgeDrag ( final DropTargetDragEvent e ) {
		if ( e.isDataFlavorSupported( DataFlavor.javaFileListFlavor ) ) {
			e.acceptDrag( DnDConstants.ACTION_COPY_OR_MOVE );
		} else {
			e.rejectDrag();
		}
	}
}
