// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.application;

import java.lang.System;
import java.lang.String;
import java.lang.Package;
import java.lang.reflect.Constructor;

import java.io.IOException;

import java.util.Collection;
import java.util.Iterator;
import java.util.ArrayList;

import Int.ecb.gesmescb.metabase.Metabase;
import Int.ecb.gesmescb.edifact.EDIFACTReader;
import Int.ecb.gesmescb.util.ComponentHelper;
import Int.ecb.gesmescb.util.TimeSeriesJoiner;
import Int.ecb.gesmescb.util.RuleBasedMapper;
import Int.ecb.gesmescb.util.DecimalsTrimmer;
import Int.ecb.gesmescb.util.InterchangeBeginOverwriter;
import Int.ecb.gesmescb.util.MessageBeginOverwriter;
import Int.ecb.gesmescb.event.GESMESCBReader;
import Int.ecb.gesmescb.event.GESMESCBWriter;
import Int.ecb.gesmescb.event.ThreadPump;
import Int.ecb.gesmescb.event.GESMESCBEventSource;
import Int.ecb.gesmescb.event.DataCollector;
import Int.ecb.gesmescb.error.AcknowledgementManager;
import Int.ecb.gesmescb.semanticchecker.CheckerBase;
import Int.ecb.gesmescb.semanticchecker.PreBreakRestrictionsChecker;
import Int.ecb.gesmescb.semanticchecker.SemanticChecker;
import Int.ecb.gesmescb.semanticchecker.ECBRestrictionsChecker;
import Int.ecb.gesmescb.sieve.GESMESCBEventSieve;
import Int.ecb.gesmescb.sieve.EmptyMessageBlocker;
import Int.ecb.gesmescb.sieve.ObservationSieve;
import Int.ecb.gesmescb.sieve.ObservationSieveWithReject;
import Int.ecb.gesmescb.sieve.ObsLevelAttributeFilter;
import Int.ecb.gesmescb.conversion.SimpleDSIConverter;
import Int.ecb.gesmescb.conversion.FAMEIREFConverter;
import Int.ecb.gesmescb.conversion.PreferenceConfiguredKeyConverter;
import Int.ecb.gesmescb.conversion.PreferenceConfiguredReceiverConverter;

import Int.ecb.logging.LogManager;
import Int.ecb.logging.Level;



/* ***************************** Class ****************************** */

/**
 * This acts as a top application class.
 *
 * <center><table border="3" cellpadding="3">
 * <caption>
 * <a name="prefs">Available {@link Int.ecb.prefs preference} settings</a>
 * </caption>
 * <tr>
 * <th>preference name</th>
 * <th>type</th>
 * <th>meaning</th>
 * <th>default</th>
 * </tr>
 * <tr><td><code>acceptStructuralDefinitions</code></td><td>boolean</td>
 *	<td>whether to use Structural Definitions from the data file</td>
 *	<td><code>true</code></td></tr>
 * <tr><td><code>dataReaderClass</code></td><td>string</td>
 *	<td>Name of {@link GESMESCBReader} class to read data</td>
 *	<td>{@link EDIFACTReader Int.ecb.gesmescb.edifact.EDIFACTReader}</td></tr>
 * <tr><td><code>dataWriterClass.<i>n</i></code></td><td>string</td>
 *	<td>Name of the <i>n</i>-th {@link GESMESCBWriter} class to write data</td>
 * </tr>
 * <tr><td><code>structuralDefinitionsReaderClass</code></td><td>string</td>
 *	<td>Name of {@link GESMESCBReader} class to read structural definitions</td>
 *	<td>{@link EDIFACTReader Int.ecb.gesmescb.edifact.EDIFACTReader}</td></tr>
 * <tr><td><code>useSemanticChecker</code></td><td>boolean</td>
 *	<td>whether to check input (see {@link SemanticChecker})</td>
 *	<td><code>true</code></td></tr>
 * <tr><td><code>useECBRestrictionsChecker</code></td><td>boolean</td>
 *	<td>whether to check additional restrictions of the
 *	ECB restriction environment (see {@link ECBRestrictionsChecker})</td>
 *	<td><code>false</code></td></tr>
 * * <tr><td><code>usePreBreakRestrictionsChecker</code></td><td>boolean</td>
 *	<td>whether to check consistency between the presence of a B pre-break indicator
 *	and presence of a pre-break value (see {@link PreBreakRestrictionsChecker})</td>
 *	<td><code>false</code></td></tr>
 * <tr><td><code>checkStructuralDefinitions</code></td><td>boolean</td>
 *	<td>whether to check Structural Definitions if checking is enabled</td>
 *	<td><code>true</code></td></tr>
 * <tr><td><code>useSieve</code></td><td>boolean</td>
 *	<td>whether to check for unexpected objects
 *	(see {@link GESMESCBEventSieve})</td>
 *	<td><code>false</code></td></tr>
 * <tr><td><code>useTimeSeriesJoiner</code></td><td>boolean</td>
 *	<td>whether to try to join {@link Int.ecb.gesmescb.TimeSeries}
 *	for better writing performance
 *	(see {@link TimeSeriesJoiner})</td>
 *	<td><code>false</code></td></tr>
 * <tr><td><code>useDecimalsTrimmer</code></td><td>boolean</td>
 *	<td>whether the number of decimals should be changed in the obs value.
 *	(see {@link DecimalsTrimmer})</td>
 *	<td><code>false</code></td></tr>
 * <tr><td><code>useThreadPump</code></td><td>boolean</td>
 *	<td>whether to use multiple threads for better writing performance
 *	(see {@link ThreadPump})</td>
 *	<td><code>false</code></td></tr>
 * <tr><td><code>useObsLevelAttributeFilter</code></td><td>boolean</td>
 *	<td>whether to filter out the observation level attributes
 * (see {@link ObsLevelAttributeFilter})</td>
 *	<td><code>false</code></td></tr> * <tr><td><code>useFAMEIREFConverter</code></td><td>boolean</td>
 *	<td>whether to convert IREF number using a {@link FAMEIREFConverter}</td>
 *	<td><code>false</code></td></tr>
 * <tr><td><code>useDSIConverter</code></td><td>boolean</td>
 *	<td>whether to convert DSI using a {@link SimpleDSIConverter}</td>
 *	<td><code>true</code></td></tr>
 * <tr><td><code>useKeyConverter</code></td><td>boolean</td>
 *	<td>whether to convert keys and data flows using an
 * {@link PreferenceConfiguredKeyConverter}</td>
 *	<td><code>false</code></td></tr>
 * <tr><td><code>useEmptyMessageBlocker</code></td><td>boolean</td>
 *	<td>whether to block empty messages with
 * {@link EmptyMessageBlocker}</td>
 *	<td><code>false</code></td></tr>
 * <tr><td><code>useReceiverConverter</code></td><td>boolean</td>
 *	<td>whether to convert Receiver with
 * {@link PreferenceConfiguredReceiverConverter}</td>
 *	<td><code>false</code></td></tr>
 * <tr><td><code>useObservationSieve</code></td><td>boolean</td>
 *	<td>whether to sieve observations with the
 * {@link ObservationSieve}</td>
 *	<td><code>false</code></td></tr>
 * <tr><td><code>useObservationSieveWithReject</code></td><td>boolean</td>
 *	<td>whether to sieve observations with the
 * {@link ObservationSieveWithReject} so as to be able to sieve later
 * the attributes corresponding to non accepted observations.</td>
 *	<td><code>false</code></td></tr>
 * </table></center>
 *
 * @author
 * <A HREF="mailto:Sami.Airo@ECB.int">Sami Airo</A>
 * @author
 * <A HREF="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</A>
 *
 * @version @(#)$Id: BatchTextUI.java,v 1.4 2018/04/19 14:46:28 kortest Exp $
 */
public class BatchTextUI {

	/* ========================= Fields ========================= */

	/** Takes care of logging and preferences. */
	private final ComponentHelper componentHelper;

	private boolean
		useSemanticChecker,
		useECBRestrictionsChecker,
		usePreBreakRestrictionsChecker,
		checkStructuralDefinitions,
		acceptStructuralDefinitions,
		useSieve,
		useFAMEIREFConverter,
		useJoiner,
		useDecimalsTrimmer,
		useDSIConverter,
		useKeyConverter,
		useReceiverConverter,
		useInterchangeBeginOverwriter,
		useMessageBeginOverwriter,
		useEmptyMessageBlocker,
		useObservationSieve,
		useObservationSieveWithReject,
		useObsLevelAttributeFilter,
		useThreadPump,
		useRuleBasedMapper;

	/* ====================== Constructors ====================== */

	BatchTextUI () {
		componentHelper = new ComponentHelper( "", this );
		logImplementationDetails();
		getPreferences();
	}

	/* ===================== Public Methods ===================== */

	/**
	 * Run through all files on the command line.
	 */
	public static void main ( final String[] arg ) {
		try {
			final BatchTextUI app = new BatchTextUI();
			System.exit( app.run( arg ) );
		}
		catch ( final Throwable t ) {
			System.err.println( "UNEXPECTED FAILURE" );
			t.printStackTrace();
			System.exit( 96 );	// Should never occur.
		}
	}

	/* ==================== Package Methods ===================== */

	/** Execute the whole process.
	 * @return the intended program exit code.
	 */
	int run ( final String[] arg ) {
		final AcknowledgementManager acknowledgementManager = new AcknowledgementManager();
		LogManager.getLogManager().addGlobalHandler(
			acknowledgementManager.getErrorHandler()
		);

		if ( arg.length <= 0 ) {
			componentHelper.logger.log(
				Level.SEVERE, 
				componentHelper.logName, 
				"run",
				"Empty argument list."
			);
			return 80;
		}

		int argNum = 0;
		try {
			for ( argNum = 0; argNum < arg.length - 1; ++argNum ) {
				loadStructuralDefinitionFile(acknowledgementManager, arg[argNum]);

				if ( acknowledgementManager.computeExitCode() != 0 ) {
					componentHelper.logger.log(
						Level.SEVERE, 
						componentHelper.logName, 
						"run",
						"Errors in loading structural definitions file: " + arg[argNum]
					);
					acknowledgementManager.close();
					return acknowledgementManager.computeExitCode();
				}
			}

			processDataFile(acknowledgementManager, arg[argNum]);
		} catch ( final Throwable e ) {
			componentHelper.logger.log(
				Level.SEVERE, 
				componentHelper.logName, 
				"run",
				"Exception in processing file: " + arg[argNum], 
				e
			);
		}

		acknowledgementManager.close();
		return acknowledgementManager.computeExitCode();
	}

	/* ==================== Private Methods ===================== */

	private void logImplementationDetails() {
		final Package application = Package.getPackage(
			"Int.ecb.gesmescb.application"
		);
		if ( application != null ) {
			componentHelper.logger.log(
				Level.INFO,
				componentHelper.logName,
				"logImplementationDetails",
				application.getImplementationTitle()
			);
			componentHelper.logger.log(
				Level.INFO,
				componentHelper.logName,
				"logImplementationDetails",
				"Implementation Version: " + application.getImplementationVersion()
			);
			componentHelper.logger.log(
				Level.INFO,
				componentHelper.logName,
				"logImplementationDetails",
				"Running on: "
				+ System.getProperty( "os.name" )
				+ "; "
				+ System.getProperty( "java.runtime.name" )
				+ ", "
				+ System.getProperty( "java.runtime.version" )
				+ " ("
				+ System.getProperty( "java.vendor.url" )
				+ ")"
			);
		}
	}

	private GESMESCBReader getReader(
		final String objectName,
		final String dataFileName
	) {
		final String readerClass = componentHelper.getPreference(
			objectName + "Class", EDIFACTReader.class.getName()
		);

		componentHelper.logger.log(
			Level.CONFIG,
			componentHelper.logName,
			"getReader",
			"Using " + readerClass + " to read from " + dataFileName + "."
		);

		GESMESCBReader configuredReader = null;
		try {
			final Class valueClass = Class.forName( readerClass );
			try {
				final Constructor constructor
				= valueClass.getConstructor(
					new Class[]{ String.class, String.class }
				);
				try { 
					configuredReader
					= (GESMESCBReader)constructor.newInstance(
						new Object[]{ objectName, dataFileName }
					);
				} catch ( final Exception e ) {
					componentHelper.logger.log(
						Level.SEVERE,
						componentHelper.logName,
						"getReader",
						"Cannot create instance for class \""
						+ readerClass + "\".",
						e.getCause().toString()
					);
				}
			} catch ( final Exception e ) {
				componentHelper.logger.log(
					Level.SEVERE,
					componentHelper.logName,
					"getReader",
					"Cannot get constructor for class \""
					+ readerClass + "\".",
					e
				);
			}
		} catch ( final Exception e ) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"getReader",
				"Cannot get class for name \""
				+ readerClass + "\".",
				e
			);
		}
		return configuredReader;
	}

	private void getPreferences() {
		useSemanticChecker = componentHelper.getPreference(
			"useSemanticChecker", "true"
		).equalsIgnoreCase( "true" );

		useECBRestrictionsChecker = componentHelper.getPreference(
			"useECBRestrictionsChecker", "false"
		).equalsIgnoreCase( "true" );

		usePreBreakRestrictionsChecker = componentHelper.getPreference(
			"usePreBreakRestrictionsChecker", "false"
		).equalsIgnoreCase( "true" );
		
		checkStructuralDefinitions = componentHelper.getPreference(
			"checkStructuralDefinitions", "true"
		).equalsIgnoreCase( "true" );

		acceptStructuralDefinitions = componentHelper.getPreference(
			"acceptStructuralDefinitions", "true"
		).equalsIgnoreCase( "true" );

		useSieve = componentHelper.getPreference(
			"useSieve", "false"
		).equalsIgnoreCase( "true" );

		useJoiner = componentHelper.getPreference(
			"useTimeSeriesJoiner", "false"
		).equalsIgnoreCase( "true" );

		useDecimalsTrimmer = componentHelper.getPreference(
			"useDecimalsTrimmer", "false"
		).equalsIgnoreCase( "true" );

		useObsLevelAttributeFilter = componentHelper.getPreference(
			"useObsLevelAttributeFilter", "false"
		).equalsIgnoreCase( "true" );

		useThreadPump = componentHelper.getPreference(
			"useThreadPump", "false"
		).equalsIgnoreCase( "true" );

		useDSIConverter = componentHelper.getPreference(
			"useDSIConverter", "true"
		).equalsIgnoreCase( "true" );

		useKeyConverter = componentHelper.getPreference(
			"useKeyConverter", "false"
		).equalsIgnoreCase( "true" );

		useEmptyMessageBlocker = componentHelper.getPreference(
			"useEmptyMessageBlocker", "false"
		).equalsIgnoreCase( "true" );

		useObservationSieve = componentHelper.getPreference(
			"useObservationSieve", "false"
		).equalsIgnoreCase( "true" );

		useObservationSieveWithReject = componentHelper.getPreference(
			"useObservationSieveWithReject", "false"
		).equalsIgnoreCase( "true" );
		
		useReceiverConverter = componentHelper.getPreference(
			"useReceiverConverter", "false"
		).equalsIgnoreCase( "true" );

		useInterchangeBeginOverwriter = componentHelper.getPreference(
				"useInterchangeBeginOverwriter", "false"
		).equalsIgnoreCase( "true" );

		useMessageBeginOverwriter = componentHelper.getPreference(
				"useMessageBeginOverwriter", "false"
		).equalsIgnoreCase( "true" );
		
		useFAMEIREFConverter = componentHelper.getPreference(
			"useFAMEIREFConverter", "false"
		).equalsIgnoreCase( "true" );
		
		useRuleBasedMapper = componentHelper.getPreference(
			"useRuleBasedMapper", "false"
		).equalsIgnoreCase( "true" );		
	}

	private void loadStructuralDefinitionFile(
		final AcknowledgementManager acknowledgementManager,
		final String fileName
	) 
		throws IOException
	{
		final GESMESCBReader reader = getReader("structuralDefinitionsReader", fileName);
		if (reader == null) return;

		GESMESCBEventSource source = reader;

		CheckerBase checker = null, ecbChecker = null;
		if ( checkStructuralDefinitions ) {
			if ( useSemanticChecker ) {
				componentHelper.logger.log(
					Level.CONFIG, 
					componentHelper.logName, 
					"loadStructuralDefinitionFile",
					"Using SemanticChecker on structural definitions"
				);
				checker = new SemanticChecker();
				source.addListener( checker, true );
				source = checker;
			}
			if ( useECBRestrictionsChecker ) {
				componentHelper.logger.log(
					Level.CONFIG, 
					componentHelper.logName, 
					"loadStructuralDefinitionFile",
					"Using ECBRestrictionsChecker on structural definitions"
				);
				ecbChecker = new ECBRestrictionsChecker();
				source.addListener( ecbChecker, true );
				source = ecbChecker;
			}
		}

		source.addListener(
			acknowledgementManager.getErrorListener(), false
		);

		source.addListener( Metabase.getMetabase(), true );
		reader.run();
		reader.close();

		if ( checker != null ) {
			checker.close();
		}

		if ( ecbChecker != null ) {
			ecbChecker.close();
		}
	}

	private void processDataFile(
		final AcknowledgementManager acknowledgementManager,
		final String dataFileName
	) {
		final GESMESCBReader dataReader = getReader("dataReader", dataFileName);
		if (dataReader == null) return;

		GESMESCBEventSource source = dataReader;

		if ( useReceiverConverter ) {
			componentHelper.logger.log(
				Level.CONFIG, 
				componentHelper.logName, 
				"processDataFile",
				"Using PreferenceConfiguredReceiverConverter on data"
			);
			final PreferenceConfiguredReceiverConverter converter = new PreferenceConfiguredReceiverConverter();
			source.addListener( converter, true );
			source = converter;
		}

		CheckerBase checker = null, ecbChecker = null, pbChecker = null;
		if ( useSemanticChecker ) {
			componentHelper.logger.log(
				Level.CONFIG, 
				componentHelper.logName, 
				"processDataFile",
				"Using SemanticChecker on data"
			);
			checker = new SemanticChecker();
			source.addListener( checker, true );
			source = checker;
		}

		if ( useECBRestrictionsChecker ) {
			componentHelper.logger.log(
				Level.CONFIG, 
				componentHelper.logName, 
				"processDataFile",
				"Using ECBRestrictionsChecker on data"
			);
			ecbChecker = new ECBRestrictionsChecker();
			source.addListener( ecbChecker, true );
			source = ecbChecker;
		}

		if ( usePreBreakRestrictionsChecker ) {
			componentHelper.logger.log(
				Level.CONFIG, 
				componentHelper.logName, 
				"processDataFile",
				"Using PreBreakRestrictionsChecker on data"
			);
			pbChecker = new PreBreakRestrictionsChecker();
			source.addListener( pbChecker, true );
			source = pbChecker;
		}
		
		source.addListener(
			acknowledgementManager.getEnvelopeListener(), false
		);

		source.addListener(
			acknowledgementManager.getErrorListener(), false
		);

		if ( acceptStructuralDefinitions ) {
			componentHelper.logger.log(
				Level.CONFIG,
				componentHelper.logName,
				"processDataFile",
				"Allowing to use structural definitions from the input."
			);
			source.addListener( Metabase.getMetabase(), true );
		}

		if ( useSieve ) {
			componentHelper.logger.log(
				Level.CONFIG, 
				componentHelper.logName, 
				"processDataFile",
				"Using GESMESCBEventSieve"
			);
			final GESMESCBEventSieve sieve = new GESMESCBEventSieve();
			source.addListener( sieve, true );
			source = sieve.getAcceptedSource();
			sieve.getRejectedSource().addListener(
				acknowledgementManager.getUnexpectedListener(), true
			);
		}

		if ( useKeyConverter ) {
			componentHelper.logger.log(
				Level.CONFIG, 
				componentHelper.logName, 
				"processDataFile",
				"Using PreferenceConfiguredKeyConverter on data"
			);
			final PreferenceConfiguredKeyConverter converter = new PreferenceConfiguredKeyConverter();
			source.addListener( converter, true );
			source = converter;
		}

		if ( useDSIConverter ) {
			componentHelper.logger.log(
				Level.CONFIG, 
				componentHelper.logName, 
				"processDataFile",
				"Using SimpleDSIConverter"
			);
			final SimpleDSIConverter simpleDSIConverter = new SimpleDSIConverter();
			source.addListener( simpleDSIConverter, true );
			source = simpleDSIConverter;
		}
		
		if ( useRuleBasedMapper ) {
			componentHelper.logger.log(
					Level.CONFIG, 
					componentHelper.logName, 
					"processDataFile",
					"Using RuleBasedMapper"
				);
				final RuleBasedMapper ruleBasedMapper = new RuleBasedMapper();
				source.addListener( ruleBasedMapper, true );
				source = ruleBasedMapper;	
				final GESMESCBEventSieve sieve = new GESMESCBEventSieve();
				source.addListener( sieve, true );
				source = sieve.getAcceptedSource();
				sieve.getRejectedSource().addListener(
					acknowledgementManager.getUnexpectedListener(), true
				);
				
		}

		if ( useJoiner ) {
			componentHelper.logger.log(
				Level.CONFIG, 
				componentHelper.logName, 
				"processDataFile",
				"Using TimeSeriesJoiner"
			);
			TimeSeriesJoiner joiner = new TimeSeriesJoiner();
			source.addListener( joiner, true );
			source = joiner;
		}
		
		if ( useDecimalsTrimmer ) {
			componentHelper.logger.log(
				Level.CONFIG, 
				componentHelper.logName, 
				"processDataFile",
				"Using DecimalsTrimmer"
			);
			DecimalsTrimmer trimmer = new DecimalsTrimmer();
			source.addListener( trimmer, true );
			source = trimmer;
		}

		ThreadPump pump = null;
		if ( useThreadPump ) {
			componentHelper.logger.log(
				Level.CONFIG, 
				componentHelper.logName, 
				"processDataFile",
				"Using ThreadPump"
			);
			pump = new ThreadPump( 10 );
			source.addListener( pump, true );
			source = pump;
		}

		if ( useEmptyMessageBlocker ) {
			componentHelper.logger.log(
				Level.CONFIG, 
				componentHelper.logName, 
				"processDataFile",
				"Using EmptyMessageBlocker"
			);
			EmptyMessageBlocker blocker = new EmptyMessageBlocker();
			source.addListener( blocker, true );
			source = blocker;
		}

		if ( useFAMEIREFConverter ) {
			componentHelper.logger.log(
				Level.CONFIG, 
				componentHelper.logName, 
				"processDataFile",
				"Using FAMEIREFConverter"
			);
			FAMEIREFConverter irefConverter = new FAMEIREFConverter();
			source.addListener( irefConverter, true );
			source = irefConverter;
		}

		if ( useObservationSieve ) {
			componentHelper.logger.log(
				Level.CONFIG, 
				componentHelper.logName, 
				"processDataFile",
				"Using ObservationSieve"
			);
			ObservationSieve observationSieve = new ObservationSieve();
			source.addListener( observationSieve, true );
			source = observationSieve;
		}

		if ( useObservationSieveWithReject ) {
			componentHelper.logger.log(
				Level.CONFIG, 
				componentHelper.logName, 
				"processDataFile",
				"Using ObservationSieveWithReject"
			);
			ObservationSieveWithReject observationSieveWithReject
			= new ObservationSieveWithReject();
			source.addListener( observationSieveWithReject, true );
			source = observationSieveWithReject;
			DataCollector dataCollector = DataCollector.getDataCollector();
			dataCollector.clear();
			observationSieveWithReject.addRejectionListener(
				dataCollector,
				true
			);
		}

		if ( useObsLevelAttributeFilter ) {
			componentHelper.logger.log(
				Level.CONFIG, 
				componentHelper.logName, 
				"processDataFile",
				"Using ObsLevelAttributeFilter"
			);
			final ObsLevelAttributeFilter obsLevelAttributeFilter
			= new ObsLevelAttributeFilter ();
			final GESMESCBEventSieve obsLevelAttributeSieve
			= new GESMESCBEventSieve ( obsLevelAttributeFilter );

			source.addListener( obsLevelAttributeSieve, true );
			source = obsLevelAttributeSieve.getAcceptedSource();
			obsLevelAttributeSieve.getRejectedSource().addListener(
				acknowledgementManager.getUnexpectedListener(), true
			);
		}
		
		if ( useInterchangeBeginOverwriter ) {
			componentHelper.logger.log(
				Level.CONFIG, 
				componentHelper.logName, 
				"processDataFile",
				"Using InterchangeBeginOverwriter"
			);
			InterchangeBeginOverwriter interchangeBeginOverwriter
			= new InterchangeBeginOverwriter();
			source.addListener( interchangeBeginOverwriter, true );
			source = interchangeBeginOverwriter;
		}

		if ( useMessageBeginOverwriter ) {
			componentHelper.logger.log(
				Level.CONFIG, 
				componentHelper.logName, 
				"processDataFile",
				"Using MessageBeginOverwriter"
			);
			MessageBeginOverwriter messageBeginOverwriter
			= new MessageBeginOverwriter();
			source.addListener( messageBeginOverwriter, true );
			source = messageBeginOverwriter;
		}
		
		final Collection writers = new ArrayList();

		int count = 1;
		while (1==1) {
			final String dataWriterClass = componentHelper.getPreference(
				"dataWriterClass." + count, null
			);

			if (dataWriterClass == null) {
				break;
			}

			final GESMESCBWriter writer = getWriter(dataWriterClass);
			if (writer == null) return;

			source.addListener( writer, true );
			writers.add( writer );

			count++;
		}

		source.addListener(
			acknowledgementManager.getErrorListener(), false
		);

		dataReader.run();

		dataReader.close();

		if (checker != null) {
			checker.close();
		}

		if ( ecbChecker != null) {
			ecbChecker.close();
		}

		if ( pump != null ) {
			pump.close();
		}

		for (
			final Iterator iterator = writers.iterator();
			iterator.hasNext();
		) {
			final GESMESCBWriter writer = (GESMESCBWriter) iterator.next();
			writer.close();
		}

		componentHelper.logger.log(
			Level.INFO, 
			componentHelper.logName, 
			"processDataFile",
			"Finished processing " + dataFileName
		);
	}

	private GESMESCBWriter getWriter(
		final String writerClass
	) {
		componentHelper.logger.log(
			Level.CONFIG,
			componentHelper.logName,
			"getWriter",
			"Using " + writerClass + " for writing."
		);

		GESMESCBWriter configuredWriter = null;
		try {
			final Class valueClass = Class.forName( writerClass );
			final Constructor constructor = valueClass.getConstructor(
				new Class[]{ String.class }
			);
			configuredWriter = (GESMESCBWriter)constructor.newInstance(
				new Object[]{ "" }
			);
		}
		catch ( final Exception e ) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"getWriter",
				"Cannot construct writer from class \"" + writerClass + "\"."
				+" Cause:\"" + e.getCause() + "\"",
				e
			);
		}
		return configuredWriter;
	}
}
