// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2007, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.application;

import java.lang.reflect.Constructor;
import java.lang.reflect.Method;
import java.util.Collection;
import java.util.ArrayList;
import java.util.Map;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Collections;
import java.io.InputStream;
import java.io.FileInputStream;
import java.io.BufferedInputStream;
import java.io.IOException;

import com.hyperion.essbase.calculator.Context;
import com.hyperion.essbase.calculator.InvalidContextException;

import Int.ecb.gesmescb.util.ComponentHelper;
import Int.ecb.gesmescb.metabase.Metabase;
import Int.ecb.gesmescb.event.GESMESCBEventSource;
import Int.ecb.gesmescb.event.GESMESCBReader;
import Int.ecb.gesmescb.event.GESMESCBWriter;
import Int.ecb.gesmescb.error.ErrorWatcher;
import Int.ecb.gesmescb.error.ErrorEventListener;
import Int.ecb.gesmescb.error.ErrorHandler;
import Int.ecb.gesmescb.edifact.EDIFACTReader;
import Int.ecb.gesmescb.essbase.FunctionInvokedSource;
import Int.ecb.gesmescb.semanticchecker.SemanticChecker;
import Int.ecb.gesmescb.semanticchecker.ECBRestrictionsChecker;
import Int.ecb.gesmescb.conversion.CodeBasedConverter;
import Int.ecb.gesmescb.conversion.SimpleDSIConverter;

import Int.ecb.logging.LogManager;
import Int.ecb.logging.Logger;
import Int.ecb.logging.Level;

import Int.ecb.prefs.Preferences;

/* ***************************** Class ****************************** */

/**
 * Faade utility providing the methods to be called as
 * Hyperion/essbase user-defined functions.
 * This way, cube data can be extracted from Hyperion.
 *
 * <center><table border="3" cellpadding="3">
 * <caption>
 * <a name="prefs">Available {@link Int.ecb.prefs preference} settings</a>
 * </caption>
 * <tr>
 * <th>preference name</th>
 * <th>type</th>
 * <th>meaning</th>
 * <th>default</th>
 * </tr>
 * <tr><td><code>useCodeConverter</code></td><td>boolean</td>
 *	<td>whether to convert series key codes using a
 *	{@link CodeBasedConverter} on coordinates received</td>
 *	<td><code>false</code></td></tr>
 * <tr><td><code>useSemanticChecker</code></td><td>boolean</td>
 *	<td>whether to check input (see {@link SemanticChecker})</td>
 *	<td><code>true</code></td></tr>
 * <tr><td><code>useECBRestrictionsChecker</code></td><td>boolean</td>
 *	<td>whether to check additional restrictions of the
 *	ECB restriction environment (see {@link ECBRestrictionsChecker})</td>
 *	<td><code>true</code></td></tr>
 * <tr><td><code>useDSIConverter</code></td><td>boolean</td>
 *	<td>whether to convert DSI using a {@link SimpleDSIConverter}</td>
 *	<td><code>false</code></td></tr>
 * <tr><td><code>dataWriterClass.<i>n</i></code></td><td>string</td>
 *	<td>Name of the <i>n</i>-th {@link GESMESCBWriter} class to write data</td>
 * </tr>
 * </table></center>
 *
 * @author
 * <a href="mailto:sis.external@ecb.int">Bernhard Bodenstorfer</A>
 *
 * @version @(#)$Id: EaacsFacade.java,v 1.1 2017/01/19 16:14:05 trainot Exp $
 */
public class EaacsFacade {

	/* ===================== Static Fields ====================== */

	/** Essbase uses this value to indicate empty cells. */
	private static final double HACKY_MISSING = -2.0E-15;

	/** Used for List conversion. */
	private static final Class[] EMPTY_CLASS_ARRAY = new Class[0];

	/** Class name for logging. */
	private static final String CLASS_NAME = EaacsFacade.class.getName();

	/** Holds the currently active EaacsFacades.
	 * They are indexed by the name passed to {@link #init}.
	 */
	private static final Map facades = Collections.synchronizedMap(
		new HashMap()
	);

	/* ===================== Static Methods ===================== */

	/** Send version info to the logs. */
	public static void logImplementationDetails () {
		final Package application = Package.getPackage(
			"Int.ecb.gesmescb.application"
		);
		logStatic(
			Level.INFO,
			"logImplementationDetails",
			CLASS_NAME + " implementation."
		);
		if ( application != null ) {
			logStatic(
				Level.INFO,
				"logImplementationDetails",
				application.getImplementationTitle()
			);
			logStatic(
				Level.INFO,
				"logImplementationDetails",
				"Implementation Version: "
				+ application.getImplementationVersion()
			);
			logStatic(
				Level.INFO,
				"logImplementationDetails",
				"Running on: "
				+ System.getProperty( "os.name" )
				+ "; "
				+ System.getProperty( "java.runtime.name" )
				+ ", "
				+ System.getProperty( "java.runtime.version" )
				+ " ("
				+ System.getProperty( "java.vendor.url" )
				+ ")"
			);
		} else {
			logStatic(
				Level.INFO,
				"logImplementationDetails",
				"Implementation details are not available."
			);
		}
	}

	/** Load the application and logging preferences. */
	public static void loadPreferences ( final String filename ) {
		try {
			final InputStream in = new BufferedInputStream(
				new FileInputStream( filename )
			);
			LogManager.getLogManager().readConfiguration();
			in.close();
			Preferences.load( filename );
		} catch ( final IOException ioe ) {
			final String message
				= "Cannot open preferences file \"" + filename + "\": " + ioe;
			logStatic(
				Level.SEVERE,
				"loadPreferences",
				message
			);
			System.err.println( message );
		}
	}

	/** Unload the application preferences. */
	public static void unloadPreferences () {
		Preferences.clear();
	}

	/** Change a preference setting.
	 *
	 * @param preference names the setting.
	 */
	public static void setPreference (
		final String preference,
		final String value
	) {
		System.setProperty( preference, value );
	}

	/** Change a preference setting.
	 *
	 * @param preference names the setting.
	 */
	public static void unsetPreference (
		final String preference
	) {
		System.getProperties().remove( preference );
	}

	/** Load the structural definitions. */
	public static void loadStructure ( final String filename ) {
		try {
			final GESMESCBReader reader = new EDIFACTReader(
				"structuralDefinitionsReader", filename
			);
			reader.addListener( Metabase.getMetabase(), true );
			reader.run();
			reader.close();
		} catch ( final IOException ioe ) {
			final String message
				= "Cannot open structure file \"" + filename + "\": " + ioe;
			logStatic(
				Level.SEVERE,
				"loadPreferences",
				message
			);
		}
	}

	/** Create and initialise a new EaacsFacade data channel.
	 * This method must be called before any further static method
	 * which takes a name as an argument.
	 *
	 * @param name uniquely identifies the channel. This is useful
	 * for simultaneous invocation of several EaacsFacade channels
	 * with independent data transfer.
	 */
	public static void open ( final String name ) {
		if ( facades.containsKey( name ) ) {
			logStatic(
				Level.SEVERE,
				"open",
				"EaacsFacade channel name \"" + name + "\" already exists."
			);
		} else {
			logStatic(
				Level.FINE,
				"open",
				"Open new channel name \"" + name + "\"."
			);
			facades.put( name, new EaacsFacade( name ) );
		}
	}

	/** Report data via the named EaacsFacade.
	 * This method is typically called in the center of a FIX loop
	 * of a CALC-script.
	 * It received the cell coordinates and the associated value.
	 *
	 * @param name uniquely identifies the channel.
	 * @param key of the time series (ordered according to the key family).
	 * @param year of the observation (prefixed with "Y", that is "Ynnnn").
	 * @param period of the observation inside the year.
	 * @param value of the observation.
	 */
	public static void handle (
		final String name,
		final String[] key,
		final String year,
		final String period,
		double value
	) {
		if ( HACKY_MISSING == value ) {
			logStatic(
				Level.FINEST,
				"handle",
				"Replacing proprietary hacky missing on channel \""
				+ name + "\" with NaN."
			);
			value = Double.NaN;
		}
		final EaacsFacade facade = (EaacsFacade) facades.get( name );
		if ( null == facade ) {
			logStatic(
				Level.SEVERE,
				"handle",
				"EaacsFacade channel name \"" + name + "\" does not exist."
			);
		} else {
			logStatic(
				Level.FINEST,
				"handle",
				"Handle observation on channel \"" + name + "\"."
			);
			facade.functionInvokedSource.handle( key, year, period, value );
		}
	}

	/** Close a EaacsFacade data channel and recollect all its resources.
	 * This method should be called when the channel has done its job.
	 * It causes any cached data to be written and
	 * then frees most recources allocated to the channel.
	 *
	 * @param name uniquely identifies the channel.
	 */
	public static void close ( final String name ) {
		final EaacsFacade facade = (EaacsFacade) facades.get( name );
		if ( null == facade ) {
			logStatic(
				Level.SEVERE,
				"close",
				"EaacsFacade channel name \"" + name + "\" does not exist."
			);
		} else {
			logStatic(
				Level.FINE,
				"close",
				"Close channel name \"" + name + "\"."
			);
			facade.close();
		}
	}

	/** Get the error status of a EaacsFacade data channel.
	 * This is usually inspected after closing the channel,
	 * but before removing it.
	 *
	 * @param name uniquely identifies the channel.
	 */
	public static double result ( final String name ) {
		final EaacsFacade facade = (EaacsFacade) facades.get( name );
		if ( null == facade ) {
			logStatic(
				Level.SEVERE,
				"result",
				"EaacsFacade channel name \"" + name + "\" does not exist."
			);
			return 97;
		} else {
			logStatic(
				Level.FINE,
				"result",
				"Result of channel name \"" + name + "\"."
			);
			return facade.result();
		}
	}

	/** Remove a EaacsFacade data channel.
	 * This method should be called when the channel has done its job.
	 * It frees the remaining recources allocated to the channel.
	 * After this, no further reference to the channel
	 * via the name is possible.
	 *
	 * @param name uniquely identifies the channel.
	 */
	public static void remove ( final String name ) {
		final EaacsFacade facade = (EaacsFacade) facades.remove( name );
		if ( null == facade ) {
			logStatic(
				Level.SEVERE,
				"remove",
				"EaacsFacade channel name \"" + name + "\" does not exist."
			);
		} else {
			logStatic(
				Level.FINE,
				"remove",
				"Close channel name \"" + name + "\", just in case."
			);
			facade.close();
		}
	}

	/** Remove all open EaacsFacade data channels and recollect their resources.
	 * This method should be called when there is a suspicion that
	 * unused channels are dangling around.
	 */
	public static void removeAll () {
		for (
			final Iterator names = facades.keySet().iterator();
			names.hasNext();
		) {
			final String name = (String) names.next();
			remove( name );
		}
	}

	/** Discard the parameter.
	 * This is useful when a returned value should not be picked up
	 * by the calcscript to modify values.
	 */
	public static void toVoid( final String text ) {}

	/** Discard the parameter.
	 * This is useful when a returned value should not be picked up
	 * by the calcscript to modify values.
	 */
	public static void toVoid( final double value ) {}

	/** Informs about the usage. */
	public static void main ( final String[] args ) {
		System.err.println(
			CLASS_NAME
			+ " is an entry point for user-defined-functions"
			+ " and cannot be used directly as an application."
			+ " It is meant to be linked into the Java VM as JAR file"
			+ " and then invoked using the public static functions."
		);
		for ( int i = 0; i < args.length; ++i ) {
			System.err.println( "arg[" + i + "]=\"" + args[i] + "\"" );
		}
	}

	/** Provides the entry point for the <code>RUNJAVA</code> statement. */
	public static void main ( final Context context, final String[] args ) {
		if ( 0 == args.length ) {
			try {
				context.sendMessage(
					"Usage: RUNJAVA "
					+ CLASS_NAME
					+ " method_name [arguments]"
				);
			} catch ( final InvalidContextException e ) {
				System.err.println( "SendMessage failed: " + e.toString() );
			}
		} else {
			try {
				final int argN = args.length - 1;
				final Method method = EaacsFacade.class.getMethod(
					args[0],
					(Class[]) Collections.nCopies( argN, String.class )
					.toArray( EMPTY_CLASS_ARRAY )
				);
				final String[] argV = new String[ argN ];
				System.arraycopy( args, 1, argV, 0, argN );
				method.invoke( null, argV );
			} catch ( final Exception e ) {
				System.err.println(
					"Call \"" + args[0] + "\" failed with parameters:"
				);
				for ( int i = 1; i < args.length; ++i ) {
					System.err.println( "arg[" + i + "]=\"" + args[i] + "\"" );
				}
				System.err.println( "Exception was: " + e );
			}
		}
	}

	/** Log a message. */
	private static void logStatic (
		final Level level,
		final String method,
		final String message
	) {
		final Logger logger = Logger.getLogger( CLASS_NAME );
		logger.log( level, "{static}", method, message );
	}

	/* ==================== Instance Fields ===================== */

	/** Takes care of logging and preferences. */
	private final ComponentHelper componentHelper;

	/** Receives data from method invocations. */
	private final FunctionInvokedSource functionInvokedSource;

	/** Listens to the log stream. */
	private final ErrorHandler errorHandler;

	/** Listens to the event stream. */
	private final ErrorEventListener errorListener;

	/** Key coordinate conversion. */
	private final CodeBasedConverter codeBasedConverter;

	/** GESMES/TS standard semantic checking. */
	private final SemanticChecker semanticChecker;

	/** Extended semantic checking. */
	private final ECBRestrictionsChecker ecbRestrictionsChecker;

	/** DSI conversion. */
	private final SimpleDSIConverter dsiConverter;

	/** Writes data to the destination. */
	private final Collection writers;

	/* ====================== Constructors ====================== */

	/** Create an instance.
	 *
	 * @param name used for configuration and logging
	 */
	private EaacsFacade ( final String name ) {
		componentHelper = new ComponentHelper( name, this );

		errorHandler = new ErrorHandler();
		LogManager.getLogManager().addGlobalHandler(
			errorHandler
		);

		functionInvokedSource = new FunctionInvokedSource( name );

		GESMESCBEventSource source = functionInvokedSource;

		if (
			componentHelper.getPreference(
				"useCodeConverter", "false"
			).equalsIgnoreCase( "true" )
		) {
			componentHelper.logger.log(
				Level.CONFIG,
				componentHelper.logName,
				"EaacsFacade()",
				"Using CodeBasedConverter"
			);
			codeBasedConverter = new CodeBasedConverter();
			source.addListener( codeBasedConverter, true );
			source = codeBasedConverter;
		} else {
			codeBasedConverter = null;
		}

		if (
			! componentHelper.getPreference(
				"useSemanticChecker", "true"
			).equalsIgnoreCase( "false" )
		) {
			componentHelper.logger.log(
				Level.CONFIG,
				componentHelper.logName,
				"EaacsFacade()",
				"Using SemanticChecker on data"
			);
			semanticChecker = new SemanticChecker();
			source.addListener( semanticChecker, true );
			source = semanticChecker;
		} else {
			semanticChecker = null;
		}

		if (
			! componentHelper.getPreference(
				"useECBRestrictionsChecker", "true"
			).equalsIgnoreCase( "false" )
		) {
			componentHelper.logger.log(
				Level.CONFIG,
				componentHelper.logName,
				"EaacsFacade()",
				"Using ECBRestrictionsChecker on data"
			);
			ecbRestrictionsChecker = new ECBRestrictionsChecker();
			source.addListener( ecbRestrictionsChecker, true );
			source = ecbRestrictionsChecker;
		} else {
			ecbRestrictionsChecker = null;
		}

		if (
			componentHelper.getPreference(
				"useDSIConverter", "false"
			).equalsIgnoreCase( "true" )
		) {
			componentHelper.logger.log(
				Level.CONFIG,
				componentHelper.logName,
				"EaacsFacade()",
				"Using SimpleDSIConverter"
			);
			dsiConverter = new SimpleDSIConverter();
			source.addListener( dsiConverter, true );
			source = dsiConverter;
		} else {
			dsiConverter = null;
		}

		writers = new ArrayList();

		for ( int i = 1;; ++i ) {
			final String dataWriterClass = componentHelper.getPreference(
				"dataWriterClass." + i, null
			);

			if ( dataWriterClass == null ) {
				if ( 1 == i ) {
					componentHelper.logger.log(
						Level.SEVERE,
						componentHelper.logName,
						"EaacsFacade()",
						"No dataWriterClass configured."
					);
				}
				break;
			}

			componentHelper.logger.log(
				Level.CONFIG,
				componentHelper.logName,
				"EaacsFacade()",
				"Using " + dataWriterClass + " for writing."
			);

			try {
				final Class valueClass = Class.forName( dataWriterClass );
				final Constructor constructor = valueClass.getConstructor(
					new Class[]{ String.class }
				);
				final GESMESCBWriter writer
					= (GESMESCBWriter) constructor.newInstance(
						new Object[]{ name }
					);
				source.addListener( writer, true );
				writers.add( writer );
			} catch ( final Exception e ) {
				componentHelper.logger.log(
					Level.SEVERE,
					componentHelper.logName,
					"EaacsFacade()",
					"Cannot construct writer from class \""
					+ dataWriterClass + "\": " + e,
					e
				);
			}
		}

		errorListener = new ErrorEventListener();
		source.addListener( errorListener, false );
	}

	/** Flush buffers and release all held resources.
	 */
	private void close () {
		functionInvokedSource.close();
		if ( null != semanticChecker ) {
			semanticChecker.close();
		}
		if ( null != ecbRestrictionsChecker ) {
			ecbRestrictionsChecker.close();
		}
		for (
			final Iterator iterator = writers.iterator();
			iterator.hasNext();
		) {
			final GESMESCBWriter writer = (GESMESCBWriter) iterator.next();
			writer.close();
		}
		LogManager.getLogManager().removeGlobalHandler( errorHandler );
	}

	/** Retrieve the error status.
	 *
	 * @return a nonzero exit status in case of problems
	 * registered during the activity of the data channel.
	 */
	private double result () {
		if ( errorListener.getSeverity() >= ErrorWatcher.FAILURE ) return 99;
		if ( errorListener.getSeverity() == ErrorWatcher.ERROR ) return 2;
		if ( errorHandler.getSeverity() >= ErrorWatcher.FAILURE ) return 98;
		return 0;
	}
}
