// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2007, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.application;

import java.io.File;
import java.io.StringReader;
import java.io.FileReader;
import java.io.BufferedReader;
import java.io.Writer;
import java.io.FileWriter;
import java.io.IOException;

import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;

import Int.ecb.gesmescb.essbase.FunctionInvokedSource;
import Int.ecb.gesmescb.csv.EaacsLabConfig;
import Int.ecb.gesmescb.conversion.CodeBasedConverter;
import Int.ecb.gesmescb.conversion.SimpleDSIConverter;
import Int.ecb.gesmescb.edifact.EDIFACTWriter;

/* ***************************** Class ****************************** */

/**
 * Implements a few tests for the {@link EaacsFacade}.
 *
 * @author
 * <a href="mailto:sis.external@ecb.int">Bernhard Bodenstorfer</a>
 *
 * @version @(#)$Id: EaacsFacadeTest.java,v 1.1 2017/01/19 16:14:05 trainot Exp $
 */
public class EaacsFacadeTest extends TestCase {

	/* ======================= Constants ======================== */

	private static final String
		eaacsFacadeName = EaacsFacade.class.getName() + ".Bobby",
		functionInvokedSourceName = FunctionInvokedSource.class.getName()
			+ ".Bobby",
		codeConverterName = CodeBasedConverter.class.getName(),
		eaacsLabConfigName = EaacsLabConfig.class.getName() + ".Bobby",
		dsiConverterName = SimpleDSIConverter.class.getName(),
		edifactWriterName = EDIFACTWriter.class.getName() + ".Bobby";

	private static final String[][]
		eaacsFacadeProperties = {
			{ "useCodeConverter", "true" },
			{ "useSemanticChecker", "false" },
			{ "useECBRestrictionsChecker", "false" },
			{ "useDSIConverter", "true" },
			{ "dataWriterClass.1", EDIFACTWriter.class.getName() }
		},
		functionInvokedSourceProperties = {
			{ "maintenanceAgency", "ECB" },
			{ "keyFamily", "ECB_IEA1" },
			{ "dataSet", "ECB_IEAQ" },
			{ "prepared", "2007-04-04T00:45:00" },
			{ "extracted", "2007-04-04T00:44:00" },
			{ "maxCache", "100" }
		},
		codeConverterProperties = {
			{ "1.dataSet", "ECB_IEAQ" },
			{ "1.maintenanceAgency", "ECB" },
			{ "1.keyFamily", "ECB_IEA1" },
			{ "1.inverse", "true" },
			{ "1.component.FREQ.prefix", "FQ_" },
			{ "1.component.REF_AREA.prefix", "RA_" }
		},
		eaacsLabConfigProperties = {
			{ "begin", "2000-01-01T00:00:00" },
			{ "end", "2002-01-01T00:00:00" },
			{ "1.dataSet", "ECB_IEAQ" },
			{ "1.maintenanceAgency", "ECB" },
			{ "1.keyFamily", "ECB_IEA1" },
			{ "1.dimension.FREQ.accept", "A" },
			{ "1.dimension.REF_AREA.translate", "EE,U2" }
		},
		dsiConverterProperties = {
			{ "internalName.ECB_IEAQ", "IEAQ" }
		};

	/** Structure file content for the {@link EaacsFacade} under test. */
	private static final String testStructure
	= "UNA:+.? '\n"
	+ "UNB+UNOC:3+4F0+ZZZ+070330:1357+IREF800528++GESMES/TS'\n"
	+ "UNH+MREF000001+GESMES:2:1:E6'\n"
	+ "BGM+73'\n"
	+ "NAD+Z02+ECB'\n"
	+ "NAD+MR+ZZZ'\n"
	+ "NAD+MS+4F0'\n"
	+ "IDE+10+Test KF structure'\n"
	+ "ASI+ECB_IEA1'\n"
	+ "FTX+ACM+++Euro area Accounts'\n"
	+ "SCD+13+FREQ++++:1'ATT+3+5+:::AN1'IDE+1+CL_FREQ'\n"
	+ "SCD+4+REF_AREA++++:2'ATT+3+5+:::AN2'IDE+1+CL_AREA_EE'\n"
	+ "SCD+1+TIME_PERIOD++++:3'ATT+3+5+:::AN..35'\n"
	+ "SCD+1+TIME_FORMAT++++:4'ATT+3+5+:::AN3'\n"
	+ "SCD+3+OBS_VALUE++++:5'ATT+3+5+:::AN..15'\n"
	+ "SCD+3+OBS_STATUS++++:6'ATT+3+5+:::AN1'ATT+3+35+2:USS'ATT+3+32+5:ALV'\n"
	+ "SCD+3+OBS_CONF++++:7'ATT+3+5+:::AN1'ATT+3+35+2:USS'ATT+3+32+5:ALV'\n"
	+ "SCD+3+OBS_PRE_BREAK++++:8'ATT+3+5+:::AN..15'\n"
	+ "UNT+31+MREF000001'\n"
	+ "UNZ+1+IREF800528'\n";

	/** Expected output file content from the {@link EaacsFacade} under test. */
	private static final String testOutput
	= "UNA:+.? '\n"
	+ "UNB+UNOC:3+ZZZ+ZZZ+070404:0045+IREF000000++GESMES/TS'\n"
	+ "UNH+MREF000001+GESMES:2:1:E6'\n"
	+ "BGM+74'\n"
	+ "NAD+Z02+ECB'\n"
	+ "NAD+MR+ZZZ'\n"
	+ "NAD+MS+ZZZ'\n"
	+ "DSI+IEAQ'\n"
	+ "STS+3+7'\n"
	+ "DTM+242:200704040044:203'\n"
	+ "DTM+Z02:20002001:702'\n"
	+ "IDE+5+ECB_IEA1'\n"
	+ "GIS+AR3'\n"
	+ "GIS+1:::-'\n"
	+ "ARR++A:DE:20002001:702:3.5:A:F+8.125:A:F'\n"
	+ "UNT+14+MREF000001'\n"
	+ "UNZ+1+IREF000000'\n";

	/* ========================= Fields ========================= */

	/** The files used for the location of test data. */
	private File testPropertiesFile, testStructureFile, testOutputFile;

	/* ====================== Constructors ====================== */

	public EaacsFacadeTest ( final String name ) {
		super( name );
	}

	/* ================== Protected Methods ===================== */

	/** Sets up the test fixture. */
	protected void setUp () throws IOException {
		testPropertiesFile = File.createTempFile( "test_properties.", ".txt" );
		final Writer propertiesWriter = new FileWriter( testPropertiesFile );

		final Object[][] propertySets = new Object[][]{
			{ eaacsFacadeName, eaacsFacadeProperties },
			{ functionInvokedSourceName, functionInvokedSourceProperties },
			{ codeConverterName, codeConverterProperties },
			{ eaacsLabConfigName, eaacsLabConfigProperties },
			{ dsiConverterName, dsiConverterProperties }
		};
		for ( int i = 0; i < propertySets.length; ++i ) {
			final String objectName = (String) propertySets[i][0];
			final String[][] objectProperties = (String[][]) propertySets[i][1];
			for ( int j = 0; j < objectProperties.length; ++j ) {
				propertiesWriter.write(
					objectName + "." + objectProperties[j][0]
					+ " = " + objectProperties[j][1] + "\n"
				);
			}
		}
		propertiesWriter.close();

		testStructureFile = File.createTempFile( "test_structure.", ".gsm" );
		final Writer structureWriter = new FileWriter( testStructureFile );
		structureWriter.write( testStructure );
		structureWriter.close();

		testOutputFile = File.createTempFile( "test_output.", ".ges" );
		System.setProperty(
			edifactWriterName + ".destination",
			testOutputFile.getAbsolutePath()
		);
	}

	/** Tears down the test fixture. */
	protected void tearDown () {
		System.getProperties().remove( edifactWriterName + ".destination" );
		testOutputFile.delete();
		testStructureFile.delete();
		testPropertiesFile.delete();
	}

	/* ===================== Public Methods ===================== */

	/**
	 * Assembles and returns a test suite for all
	 * the test methods of this test case.
	 *
	 * @return a non-null test suite
	 */
	public static Test suite () {
		return new TestSuite( EaacsFacadeTest.class );
	}

	/**
	 * @param arg the array of argument {@link String}s
	 */
	public static void main ( final String[] arg ) {
		final String[] testCaseName = {
			EaacsFacadeTest.class.getName()
		};
		junit.textui.TestRunner.main( testCaseName );
	}

	/* ---------------------- Test Methods ---------------------- */

	/** Test {@link EaacsFacade} usage cycle. */
	public void testCycle () throws IOException {
		EaacsFacade.loadPreferences( testPropertiesFile.getAbsolutePath() );
		EaacsFacade.loadStructure( testStructureFile.getAbsolutePath() );

		EaacsFacade.open( new String( "Bobby" ) );

		EaacsFacade.handle(
			"Bobby",
			new String[]{ "FQ_A", "RA_DE" },
			"Y2000", "A",
			3.5
		);
		EaacsFacade.handle(
			"Bobby",
			new String[]{ "FQ_A", "RA_DE" },
			"Y2001", "A",
			8.125
		);

		EaacsFacade.close( "Bobby" );

		final double result = EaacsFacade.result( "Bobby" );

		assertEquals( "All o.k.", 0.0, result, 0 );

		EaacsFacade.remove( "Bobby" );

		final BufferedReader
			expectedReader = new BufferedReader(
				new StringReader( testOutput )
			),
			actualReader = new BufferedReader(
				new FileReader( testOutputFile )
			);

		for ( int i = 1;; ++i ) {
			final String
				expected = expectedReader.readLine(),
				actual = actualReader.readLine();
			if ( null == expected ) {
				assertNull( "Last line " + i, actual );
				break;
			}
			assertEquals( "Line " + i, expected, actual );
		}

		actualReader.close();
		expectedReader.close();
	}
}
