// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.application;

import java.lang.System;
import java.lang.String;
import java.lang.Package;
import java.lang.reflect.Constructor;

import java.util.Collection;
import java.util.Iterator;
import java.util.ArrayList;

import Int.ecb.gesmescb.metabase.Metabase;
import Int.ecb.gesmescb.edifact.EDIFACTReader;
import Int.ecb.gesmescb.edifact.EDIFACTWriter;
import Int.ecb.gesmescb.util.ComponentHelper;
import Int.ecb.gesmescb.util.TimeSeriesJoiner;
import Int.ecb.gesmescb.event.GESMESCBReader;
import Int.ecb.gesmescb.event.GESMESCBWriter;
import Int.ecb.gesmescb.event.ThreadPump;
import Int.ecb.gesmescb.event.GESMESCBEventSource;
import Int.ecb.gesmescb.error.AcknowledgementManager;
import Int.ecb.gesmescb.semanticchecker.CheckerBase;
import Int.ecb.gesmescb.semanticchecker.PreBreakRestrictionsChecker;
import Int.ecb.gesmescb.semanticchecker.SemanticChecker;
import Int.ecb.gesmescb.semanticchecker.ECBRestrictionsChecker;
import Int.ecb.gesmescb.sieve.GESMESCBEventSieve;
import Int.ecb.gesmescb.conversion.SimpleDSIConverter;
import Int.ecb.gesmescb.conversion.FAMEIREFConverter;
import Int.ecb.gesmescb.fame.SimpleESDBWriter;
import Int.ecb.gesmescb.fame.ESDBWriter;
import Int.ecb.gesmescb.fame.SDPEWriter;
import Int.ecb.gesmescb.fame.SDPEReceiverWriter;
import Int.ecb.gesmescb.fame.FAME4GLWriter;
import Int.ecb.gesmescb.csv.EurostatWriter;
import Int.ecb.gesmescb.xml.GesmesTSXMLWriter;
import Int.ecb.gesmescb.xml.SDMXMLWriter;
import Int.ecb.gesmescb.sql.SQLDataDefinitionWriter;
import Int.ecb.gesmescb.sql.SQLDataManipulationWriter;
import Int.ecb.gesmescb.sql.SDWWriter;

import Int.ecb.logging.LogManager;

import Int.ecb.logging.Level;


/* ***************************** Class ****************************** */

/**
 * Loads GESMES/TS and GESMES/CB to FAME.
 * This acts as a top application class.
 *
 * <center><table border="3" cellpadding="3">
 * <caption>
 * <a name="prefs">Available {@link Int.ecb.prefs preference} settings</a>
 * </caption>
 * <tr>
 * <th>preference name</th>
 * <th>type</th>
 * <th>meaning</th>
 * <th>default</th>
 * </tr>
 * <tr><td><code>dataReaderClass</code></td><td>string</td>
 *	<td>Name of {@link GESMESCBReader} class to read data</td>
 *	<td>{@link EDIFACTReader Int.ecb.gesmescb.edifact.EDIFACTReader}</td></tr>
 * <tr><td><code>useSemanticChecker</code></td><td>boolean</td>
 *	<td>whether to check input (see {@link SemanticChecker})</td>
 *	<td><code>true</code></td></tr>
 * <tr><td><code>useECBRestrictionsChecker</code></td><td>boolean</td>
 *	<td>whether to check additional restrictions of the
 *	ECB restriction environment (see {@link ECBRestrictionsChecker})</td>
 *	<td><code>false</code></td></tr>
 * <tr><td><code>usePreBreakRestrictionsChecker</code></td><td>boolean</td>
 *	<td>whether to check consistency between the presence of a B pre-break indicator
 *	and presence of a pre-break value (see {@link PreBreakRestrictionsChecker})</td>
 *	<td><code>false</code></td></tr>
 * <tr><td><code>checkStructuralDefinitions</code></td><td>boolean</td>
 *	<td>whether to check Structural Definitions if checking is enabled</td>
 *	<td><code>true</code></td></tr>
 * <tr><td><code>useSieve</code></td><td>boolean</td>
 *	<td>whether to check for unexpected objects
 *	(see {@link GESMESCBEventSieve})</td>
 *	<td><code>false</code></td></tr>
 * <tr><td><code>useTimeSeriesJoiner</code></td><td>boolean</td>
 *	<td>whether to try to join {@link Int.ecb.gesmescb.TimeSeries}
 *	for better writing performance
 *	(see {@link TimeSeriesJoiner})</td>
 *	<td><code>false</code></td></tr>
 * <tr><td><code>useThreadPump</code></td><td>boolean</td>
 *	<td>whether to use multiple threads for better writing performance
 *	(see {@link ThreadPump})</td>
 *	<td><code>false</code></td></tr>
 * <tr><td><code>useEDIFACTWriter</code></td><td>boolean</td>
 *	<td>whether to write to EDIFACT (see {@link EDIFACTWriter})</td>
 *	<td><code>false</code></td></tr>
 * <tr><td><code>useSimpleESDBWriter</code></td><td>boolean</td>
 *	<td>whether to write to FAME using a {@link SimpleESDBWriter}</td>
 *	<td><code>false</code></td></tr>
 * <tr><td><code>useESDBWriter</code></td><td>boolean</td>
 *	<td>whether to write to FAME using an {@link ESDBWriter}</td>
 *	<td><code>false</code></td></tr>
 * <tr><td><code>useSDPEWriter</code></td><td>boolean</td>
 *	<td>whether to write to FAME using an {@link SDPEWriter}</td>
 *	<td><code>false</code></td></tr>
 * <tr><td><code>useSDPEReceiverWriter</code></td><td>boolean</td>
 *	<td>whether to write to FAME using an {@link SDPEReceiverWriter}</td>
 *	<td><code>false</code></td></tr>
 * <tr><td><code>useFAMEIREFConverter</code></td><td>boolean</td>
 *	<td>whether to convert IREF number using a {@link FAMEIREFConverter}</td>
 *	<td><code>false</code></td></tr>
 * <tr><td><code>useGesmesTSXMLWriter</code></td><td>boolean</td>
 *	<td>whether to write GESMES/TS-XML using an {@link GesmesTSXMLWriter}</td>
 *	<td><code>false</code></td></tr>
 * <tr><td><code>useSDMXMLWriter</code></td><td>boolean</td>
 *	<td>whether to write SDMX-ML using an {@link SDMXMLWriter}</td>
 *	<td><code>false</code></td></tr>
 * <tr><td><code>useSDWWriter</code></td><td>boolean</td>
 *	<td>whether to write to the SDW using an {@link SDWWriter}</td>
 *	<td><code>false</code></td></tr>
 * </table></center>
 *
 * @author
 * <A HREF="mailto:Sami.Airo@ECB.int">Sami Airo</A>
 * @author
 * <A HREF="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</A>
 *
 * @version @(#)$Id: Loader.java,v 1.1 2017/01/19 16:14:05 trainot Exp $
 */
public class Loader {

	/* ========================= Fields ========================= */

	/** Takes care of logging and preferences. */
	protected final ComponentHelper componentHelper;

	private final boolean
		useSemanticChecker,
		useECBRestrictionsChecker,
		usePreBreakRestrictionsChecker,
		checkStructuralDefinitions,
		acceptStructuralDefinitions,
		useSieve,
		useJoiner,
		useThreadPump,
		useEDIFACTWriter,
		useSimpleESDBWriter,
		useESDBWriter,
		useSDPEWriter,
		useSDPEReceiverWriter,
		useEurostatWriter,
		useFAMEIREFConverter,
		useSQLDataDefinitionWriter,
		useSQLDataManipulationWriter,
		useFAME4GLWriter,
		useGesmesTSXMLWriter,
		useSDMXMLWriter,
		useSDWWriter;

	private final GESMESCBReader dataReader;

	/* ====================== Constructors ====================== */

	private Loader ( final String dataFileName ) {
		componentHelper = new ComponentHelper( "", this );

		final Package application = Package.getPackage(
			"Int.ecb.gesmescb.application"
		);
		if ( application != null ) {
			componentHelper.logger.log(
				Level.INFO,
				componentHelper.logName,
				"Loader",
				application.getImplementationTitle()
			);
			componentHelper.logger.log(
				Level.INFO,
				componentHelper.logName,
				"Loader",
				"Implementation Version: "
				+ application.getImplementationVersion()
			);
			componentHelper.logger.log(
				Level.INFO,
				componentHelper.logName,
				"logImplementationDetails",
				"Running on: "
				+ System.getProperty( "os.name" )
				+ "; "
				+ System.getProperty( "java.runtime.name" )
				+ ", "
				+ System.getProperty( "java.runtime.version" )
				+ " ("
				+ System.getProperty( "java.vendor.url" )
				+ ")"
			);
		}

		final String dataReaderClass = componentHelper.getPreference(
			"dataReaderClass", EDIFACTReader.class.getName()
		);

		componentHelper.logger.log(
			Level.CONFIG,
			componentHelper.logName,
			"Loader",
			"Using " + dataReaderClass + " to read data."
		);

		GESMESCBReader configuredDataReader = null;
		try {
			final Class valueClass = Class.forName( dataReaderClass );
			try {
				final Constructor constructor = valueClass.getConstructor(
					new Class[]{ String.class, String.class }
				);
				try {
					configuredDataReader
					= (GESMESCBReader)constructor.newInstance(
						new Object[]{ "DataReader", dataFileName }
					);
				} catch ( final Exception e ) {
					componentHelper.logger.log(
						Level.SEVERE,
						componentHelper.logName,
						"Loader",
						"Cannot create instance for class \""
						+ dataReaderClass + "\".",
						e.getCause().toString()
					);
				}			
			} catch ( final Exception e ) {
				componentHelper.logger.log(
					Level.SEVERE,
					componentHelper.logName,
					"Loader",
					"Cannot get constructor for class \""
					+ dataReaderClass + "\".",
					e
				);
			}
		} catch ( final Exception e ) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"Loader",
				"Cannot get class for name \""
				+ dataReaderClass + "\".",
				e
			);
		}
		dataReader = configuredDataReader;

		useSemanticChecker = componentHelper.getPreference(
			"useSemanticChecker", "true"
		).equals( "true" );

		useECBRestrictionsChecker = componentHelper.getPreference(
			"useECBRestrictionsChecker", "false"
		).equals( "true" );

		usePreBreakRestrictionsChecker = componentHelper.getPreference(
			"usePreBreakRestrictionsChecker", "false"
		).equals( "true" );

		checkStructuralDefinitions = componentHelper.getPreference(
			"checkStructuralDefinitions", "true"
		).equals( "true" );

		acceptStructuralDefinitions = componentHelper.getPreference(
			"acceptStructuralDefinitions", "true"
		).equals( "true" );

		useSieve = componentHelper.getPreference(
			"useSieve", "false"
		).equalsIgnoreCase( "true" );

		useJoiner = componentHelper.getPreference(
			"useTimeSeriesJoiner", "false"
		).equals( "true" );

		useThreadPump = componentHelper.getPreference(
			"useThreadPump", "false"
		).equals( "true" );

		useEDIFACTWriter = componentHelper.getPreference(
			"useEDIFACTWriter", "false"
		).equals( "true" );

		useSimpleESDBWriter = componentHelper.getPreference(
			"useSimpleESDBWriter", "false"
		).equals( "true" );

		useESDBWriter = componentHelper.getPreference(
			"useESDBWriter", "false"
		).equals( "true" );

		useSDPEWriter = componentHelper.getPreference(
			"useSDPEWriter", "false"
		).equals( "true" );

		useSDPEReceiverWriter = componentHelper.getPreference(
			"useSDPEReceiverWriter", "false"
		).equals( "true" );

		useEurostatWriter = componentHelper.getPreference(
			"useEurostatWriter", "false"
		).equals( "true" );

		useFAMEIREFConverter = componentHelper.getPreference(
			"useFAMEIREFConverter", "false"
		).equals( "true" );

		useGesmesTSXMLWriter = componentHelper.getPreference(
			"useGesmesTSXMLWriter", "false"
		).equalsIgnoreCase( "true" );

		useSDMXMLWriter = componentHelper.getPreference(
			"useSDMXMLWriter", "false"
		).equalsIgnoreCase( "true" );

		useFAME4GLWriter = componentHelper.getPreference(
			"useFAME4GLWriter", "false"
		).equalsIgnoreCase( "true" );

		useSQLDataDefinitionWriter = componentHelper.getPreference(
			"useSQLDataDefinitionWriter", "false"
		).equalsIgnoreCase( "true" );

		useSQLDataManipulationWriter = componentHelper.getPreference(
			"useSQLDataManipulationWriter", "false"
		).equalsIgnoreCase( "true" );

		useSDWWriter = componentHelper.getPreference(
			"useSDWWriter", "false"
		).equalsIgnoreCase( "true" );

		{
			final Collection activities = new ArrayList();

			if ( useSemanticChecker ) activities.add( "check" );
			if ( useSieve ) activities.add( "sieve" );
			if ( useSimpleESDBWriter ) activities.add( "load to FAME with SimpleESDBWriter" );
			if ( useESDBWriter ) activities.add( "load to FAME with ESDBWriter" );
			if ( useSDPEWriter ) activities.add( "load to FAME with SDPEWriter" );
			if ( useEDIFACTWriter ) activities.add( "write EDIFACT" );
			if ( useEurostatWriter ) activities.add( "write Eurostat CSV" );
			if ( useFAMEIREFConverter ) activities.add( "Convert IREF numbers" );
			if ( useGesmesTSXMLWriter ) activities.add( "write GESMES/TS-XML" );
			if ( useSDMXMLWriter ) activities.add( "write SDMX-ML" );
			if ( useSQLDataDefinitionWriter ) activities.add( "write SQL for data definition" );
			if ( useSQLDataManipulationWriter ) activities.add( "write SQL for data manipulation" );
			if ( useFAME4GLWriter ) activities.add( "write FAME 4GL" );
			if ( useSDWWriter ) activities.add( "write to SDW" );

			final StringBuffer buffer = new StringBuffer( "Configured to {" );
			for (
				final Iterator iterator = activities.iterator();
				iterator.hasNext();
			) {
				final String act = (String) iterator.next();
				buffer.append( act );
				if ( iterator.hasNext() ) buffer.append( ", " );
			}
			buffer.append( "}" );

			componentHelper.logger.log(
				Level.INFO,
				componentHelper.logName,
				"Loader",
				buffer.toString()
			);
		}
	}

	/* ===================== Public Methods ===================== */

	/**
	 * Run through all files on the command line.
	 */
	public static void main ( final String[] arg ) {
		try {
			final Loader loader = new Loader( arg[ arg.length - 1 ] );
			System.exit( loader.run( arg ) );
		}
		catch ( final Throwable t ) {
			System.err.println( "UNEXPECTED ECB LOADER FAILURE" );
			t.printStackTrace();
			System.exit( 96 );	// Should never occur.
		}
	}

	/* ==================== Private Methods ===================== */

	/** Execute the whole process.
	 * @return the intended program exit code.
	 */
	private int run ( final String[] arg ) {

		// Set up the handlers to catch errors
		final AcknowledgementManager acknowledgementManager
		= new AcknowledgementManager();
		LogManager.getLogManager().addGlobalHandler(
			acknowledgementManager.getErrorHandler()
		);

		if ( arg.length <= 0 ) {
			componentHelper.logger.log(
				Level.SEVERE, componentHelper.logName, "run",
				"Loader called with empty argument list."
			);
			return 80;
		}

		int argNum = 0;

		try {
			GESMESCBEventSource source;

			// First load all the structural definitions files

			for ( argNum = 0; argNum < arg.length - 1; ++argNum ) {
				final GESMESCBReader reader
				= new EDIFACTReader(
					"StructuralDefinitionsReader",
					arg[argNum]
				);
				source = reader;

				CheckerBase checker = null, ecbChecker = null;
				if ( checkStructuralDefinitions ) {
					if ( useSemanticChecker ) {
						componentHelper.logger.log(
							Level.FINE, componentHelper.logName, "run",
							"Using SemanticChecker on structural definitions"
						);
						checker = new SemanticChecker();
						source.addListener( checker, true );
						source = checker;
					}
					if ( useECBRestrictionsChecker ) {
						ecbChecker = new ECBRestrictionsChecker();
						source.addListener( ecbChecker, true );
						source = ecbChecker;
					}
				}

				source.addListener(
					acknowledgementManager.getErrorListener(), false
				);

				source.addListener( Metabase.getMetabase(), true );

				reader.run();

				reader.close();

				if ( checker != null ) {
					checker.close();
				}

				if ( ecbChecker != null ) {
					ecbChecker.close();
				}

				if ( acknowledgementManager.computeExitCode() != 0 ) {
					componentHelper.logger.log(
						Level.SEVERE, componentHelper.logName, "run",
						"Errors in loading structural definitions file: "
						+ arg[argNum]
					);
					acknowledgementManager.close();
					return acknowledgementManager.computeExitCode();
				}
			}

			source = dataReader;

			CheckerBase checker = null, ecbChecker = null, pbChecker = null;
			if ( useSemanticChecker ) {
				componentHelper.logger.log(
					Level.FINE, componentHelper.logName, "run",
					"Using SemanticChecker"
				);
				checker = new SemanticChecker();
				source.addListener( checker, true );
				source = checker;
			}
			if ( useECBRestrictionsChecker ) {
				ecbChecker = new ECBRestrictionsChecker();
				source.addListener( ecbChecker, true );
				source = ecbChecker;
			} 
			if ( usePreBreakRestrictionsChecker ) {
				pbChecker = new PreBreakRestrictionsChecker();
				source.addListener( pbChecker, true );
				source = pbChecker;
			}

			source.addListener(
				acknowledgementManager.getEnvelopeListener(), false
			);

			source.addListener(
				acknowledgementManager.getErrorListener(), false
			);

			if ( acceptStructuralDefinitions ) {
				componentHelper.logger.log(
					Level.FINE,
					componentHelper.logName,
					"run",
					"Allowing to use structural definitions from the input."
				);
				source.addListener( Metabase.getMetabase(), false );
			}

			if ( useSieve ) {
				componentHelper.logger.log(
					Level.FINE, componentHelper.logName, "run",
					"Using GESMESCBEventSieve"
				);
				final GESMESCBEventSieve sieve = new GESMESCBEventSieve();
				source.addListener( sieve, true );
				source = sieve.getAcceptedSource();
				sieve.getRejectedSource().addListener(
					acknowledgementManager.getUnexpectedListener(), true
				);
			}

			componentHelper.logger.log(
				Level.FINE, componentHelper.logName, "main",
				"Using SimpleDSIConverter"
			);
			final SimpleDSIConverter simpleDSIConverter = new SimpleDSIConverter();
			source.addListener( simpleDSIConverter, true );
			source = simpleDSIConverter;

			TimeSeriesJoiner joiner = null;
			if ( useJoiner ) {
				componentHelper.logger.log(
					Level.FINE, componentHelper.logName, "run",
					"Using TimeSeriesJoiner"
				);
				joiner = new TimeSeriesJoiner();
				source.addListener( joiner, true );
				source = joiner;
			}

			ThreadPump pump = null;
			if ( useThreadPump ) {
				componentHelper.logger.log(
					Level.FINE, componentHelper.logName, "run",
					"Using ThreadPump"
				);
				pump = new ThreadPump( 10 );
				source.addListener( pump, true );
				source = pump;
			}

			FAMEIREFConverter irefConverter = null;
			if ( useFAMEIREFConverter ) {
				componentHelper.logger.log(
					Level.FINE, componentHelper.logName, "run",
					"Using FAMEIREFConverter"
				);
				irefConverter = new FAMEIREFConverter();
				source.addListener( irefConverter, true );
				source = irefConverter;
			}

			// Remember writers to close.
			final Collection writers = new ArrayList();

			if ( useEDIFACTWriter ) {
				componentHelper.logger.log(
					Level.INFO, componentHelper.logName, "run",
					"Going to write in EDIFACT."
				);
				final GESMESCBWriter writer = new EDIFACTWriter( "" );
				source.addListener( writer, true );
				writers.add( writer );
			}

			if ( useSimpleESDBWriter ) {
				componentHelper.logger.log(
					Level.INFO, componentHelper.logName, "main",
					"Going to load to FAME using SimpleESDBWriter."
				);
				final SimpleESDBWriter writer = new SimpleESDBWriter( "" );
				source.addListener( writer, true );
				writers.add( writer );
			}

			if ( useESDBWriter ) {
				componentHelper.logger.log(
					Level.INFO, componentHelper.logName, "main",
					"Going to load to FAME using ESDBWriter."
				);
				final ESDBWriter writer = new ESDBWriter( "" );
				source.addListener( writer, true );
				writers.add( writer );
			}

			if ( useSDPEWriter ) {
				componentHelper.logger.log(
					Level.INFO, componentHelper.logName, "main",
					"Going to load to FAME using SDPEWriter."
				);
				final SDPEWriter writer = new SDPEWriter( "" );
				source.addListener( writer, true );
				writers.add( writer );
			}

			if ( useSDPEReceiverWriter ) {
				componentHelper.logger.log(
					Level.INFO, componentHelper.logName, "main",
					"Going to load to FAME using SDPEReceiverWriter."
				);
				final SDPEReceiverWriter writer = new SDPEReceiverWriter( "" );
				source.addListener( writer, true );
				writers.add( writer );
			}

			if ( useEurostatWriter ) {
				componentHelper.logger.log(
					Level.CONFIG, componentHelper.logName, "run",
					"Going to write CSV in Eurostat format."
				);
				final GESMESCBWriter writer = new EurostatWriter( "" );
				source.addListener( writer, true );
				writers.add( writer );
			}

			if ( useGesmesTSXMLWriter ) {
				componentHelper.logger.log(
					Level.CONFIG, componentHelper.logName, "run",
					"Going to write GESMES/TS-XML."
				);
				final GESMESCBWriter writer = new GesmesTSXMLWriter( "" );
				source.addListener( writer, true );
				writers.add( writer );
			}

			if ( useSDMXMLWriter ) {
				componentHelper.logger.log(
					Level.CONFIG, componentHelper.logName, "run",
					"Going to write SDMX-ML."
				);
				final GESMESCBWriter writer = new SDMXMLWriter( "" );
				source.addListener( writer, true );
				writers.add( writer );
			}

			if ( useSQLDataDefinitionWriter ) {
				componentHelper.logger.log(
					Level.CONFIG, componentHelper.logName, "run",
					"Going to write SQL for data definition."
				);
				final SQLDataDefinitionWriter writer = new SQLDataDefinitionWriter( "" );
				source.addListener( writer, true );
				writers.add( writer );
			}

			if ( useSQLDataManipulationWriter ) {
				componentHelper.logger.log(
					Level.CONFIG, componentHelper.logName, "run",
					"Going to write SQL for data manipulation."
				);
				final SQLDataManipulationWriter writer = new SQLDataManipulationWriter( "" );
				source.addListener( writer, true );
				writers.add( writer );
			}

			if ( useFAME4GLWriter ) {
				componentHelper.logger.log(
					Level.CONFIG, componentHelper.logName, "run",
					"Going to write FAME 4GL."
				);
				final FAME4GLWriter writer = new FAME4GLWriter( "" );
				source.addListener( writer, true );
				writers.add( writer );
			}

			if ( useSDWWriter ) {
				componentHelper.logger.log(
					Level.CONFIG, componentHelper.logName, "run",
					"Going to write to SDW."
				);
				final SDWWriter writer = new SDWWriter( "" );
				source.addListener( writer, true );
				writers.add( writer );
			}

			source.addListener(
				acknowledgementManager.getErrorListener(), false
			);

			dataReader.run();

			dataReader.close();

			if (checker != null) {
				checker.close();
			}

			if ( ecbChecker != null) {
				 ecbChecker.close();
			}
			
			if ( pbChecker != null) {
				 pbChecker.close();
			}

			if ( pump != null ) {
				pump.close();
			}

			for (
				final Iterator iterator = writers.iterator();
				iterator.hasNext();
			) {
				final GESMESCBWriter writer = (GESMESCBWriter) iterator.next();
				writer.close();
			}

			componentHelper.logger.log(
				Level.INFO, componentHelper.logName, "run",
				"Finished processing " + arg[argNum]
			);
		} catch ( final Throwable e ) {
			componentHelper.logger.log(
				Level.SEVERE, componentHelper.logName, "run",
				"Exception in processing file: " + arg[argNum], e
			);
		}

		acknowledgementManager.close();
		return acknowledgementManager.computeExitCode();
	}
}
