// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2007, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.conversion;

import java.util.List;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.Map;
import java.util.HashMap;

import Int.ecb.logging.Level;

import Int.ecb.gesmescb.*;

import Int.ecb.gesmescb.metabase.Metabase;
import Int.ecb.gesmescb.metabase.MaintenanceAgency;

import Int.ecb.gesmescb.util.KeyTokenizer;
import Int.ecb.gesmescb.util.ComponentHelper;

import Int.ecb.gesmescb.event.GESMESCBEventBroker;

import Int.ecb.gesmescb.error.ErrorWatcher;

/* **************************** Class ***************************** */

/**
 * Converts dimension values and attribute values
 * according to preference settings.
 * Thus, keys of time series and attributes,
 * but also values of array cells and attributes are converted.
 *
 * @author
 * <a href="mailto:sis.external@ecb.int">Bernhard Bodenstorfer</a>
 *
 * @version @(#)$Id: CodeBasedConverter.java,v 1.1 2017/01/19 16:14:04 trainot Exp $
 */

public class CodeBasedConverter extends GESMESCBEventBroker {

	/* ========================= Fields ========================= */

	/** Holds conversion info objects per data set. */
	private final Map conversionMap;

	/** Currently active conversion strategy. */
	private DecomposeToCodeConverter decomposeToCodeConverter;

	/* ====================== Constructors ====================== */

	public CodeBasedConverter() {
		this( "" );
	}

	public CodeBasedConverter( final String name ) {
		super( name, true );

		conversionMap = new HashMap();

		String dataSetId;
		for (
			int i = 1;
			null != ( dataSetId = componentHelper.getPreference(
				i + ".dataSet", null
			) );
			++i
		) {
			final String maintenanceAgencyId = componentHelper.getPreference(
				i + ".maintenanceAgency", null
			);
			if ( null == maintenanceAgencyId ) {
				complain(
					"No maintenance agency configured for dataset "
					+ dataSetId + ".",
					"CodeBasedConverter()",
					dataSetId,
					ErrorWatcher.ERROR
				);
				break;
			}
			final String keyFamilyId = componentHelper.getPreference(
				i + ".keyFamily", null
			);
			if ( null == keyFamilyId ) {
				complain(
					"No key family configured for dataset "
					+ dataSetId + ".",
					"CodeBasedConverter()",
					dataSetId,
					ErrorWatcher.ERROR
				);
				break;
			}
			final boolean inverse = "true".equalsIgnoreCase(
				componentHelper.getPreference(
					i + ".inverse", "false"
				)
			);
			if ( conversionMap.containsKey( dataSetId ) ) {
				final DecomposeToCodeConverter converter
				= (DecomposeToCodeConverter) conversionMap.get( dataSetId );
				final boolean isSame
				= maintenanceAgencyId.equals(
					converter.getMaintenanceAgencyId()
				) && keyFamilyId.equals(
					converter.getKeyFamilyId()
				) && inverse == converter.isInverse();
				complain(
					"Configured data set " + dataSetId + " twice "
					+ ( isSame ? "the same" : "differently" )
					+ " with indices "
					+ converter.getIndex()
					+ " and " + i + ".",
					"CodeBasedConverter()",
					dataSetId,
					isSame ? ErrorWatcher.WARNING : ErrorWatcher.ERROR
				);
			} else {
				conversionMap.put(
					dataSetId,
					new DecomposeToCodeConverter(
						componentHelper,
						i,
						maintenanceAgencyId,
						keyFamilyId,
						inverse
					)
				);
			}
		}
	}

	/* ===================== Public Methods ===================== */

	public void handle ( final MessageBegin messageBegin ) {
		final String dataSetId = messageBegin.getDataSetId();
		final boolean isConvertibleMessage = messageBegin.isStatisticalData()
			||
			! conversionMap.containsKey( dataSetId );

		if ( isConvertibleMessage ) {
			componentHelper.logger.log(
				Level.FINE,
				componentHelper.logName,
				"handle(MessageBegin)",
				"Starting to convert from data set " + dataSetId + "."
			);
			decomposeToCodeConverter
			= (DecomposeToCodeConverter) conversionMap.get(
				messageBegin.getDataSetId()
			);
		} else {
			componentHelper.logger.log(
				Level.FINE,
				componentHelper.logName,
				"handle(MessageBegin)",
				"No conversion defined for data set " + dataSetId + "."
			);
			decomposeToCodeConverter = null;
		}

		dispatch( messageBegin );
	}

	public void handle ( final TimeSeries timeSeries ) {
		componentHelper.logger.log(
			Level.FINER, componentHelper.logName, "handle(TimeSeries)",
			"handling TimeSeries", timeSeries
		);

		if ( null == decomposeToCodeConverter ) {
			super.handle( timeSeries );
			return;
		}

		final List convertedObservations = new ArrayList();
		for (
			final Iterator observations = timeSeries.observationIterator();
			observations.hasNext();
		) {
			final Observation observation = (Observation) observations.next();
			convertedObservations.add(
				decomposeToCodeConverter.convertObservation( observation )
			);
		}

		dispatch( new TimeSeries(
			decomposeToCodeConverter.convertKey( timeSeries.getKey() ),
			timeSeries.getPeriod(),
			timeSeries.getPeriodFormat(),
			convertedObservations
		) );
	}


	public void handle ( final Attribute attribute ) {
		componentHelper.logger.log(
			Level.FINER, componentHelper.logName, "handle(Attribute)",
			"handling Attribute", attribute
		);

		if ( null == decomposeToCodeConverter ) {
			super.handle( attribute );
			return;
		}

		dispatch( new Attribute(
			decomposeToCodeConverter.convertKey( attribute.getKey() ),
			attribute.getPeriod(),
			attribute.getPeriodFormat(),
			attribute.getName(),
			decomposeToCodeConverter.convertValue(
				attribute.getName(),
				attribute.getValue()
			)
		) );
	}
}

/* **************************** Class ***************************** */

class DecomposeToCodeConverter {

	private final int index;
	private final ComponentHelper componentHelper;
	private final String maintenanceAgencyId, keyFamilyId;
	private final boolean isInverse;

	private transient KeyFamily keyFamily;
	private transient CodeConverter[] dimensionMappings;
	private transient Map componentMappings;

	public DecomposeToCodeConverter (
		final ComponentHelper componentHelper,
		final int index,
		final String maintenanceAgencyId,
		final String keyFamilyId,
		final boolean isInverse
	) {
		this.componentHelper = componentHelper;
		this.index = index;
		this.maintenanceAgencyId = maintenanceAgencyId;
		this.keyFamilyId = keyFamilyId;
		this.isInverse = isInverse;
	}

	public int getIndex () { return index; }

	public String getMaintenanceAgencyId () { return maintenanceAgencyId; }

	public String getKeyFamilyId () { return keyFamilyId; }

	public boolean isInverse () { return isInverse; }

	public String convertCoordinate ( final int index, final String code ) {
		initKeyFamily();
		final CodeConverter converter
		= (CodeConverter) dimensionMappings[ index ];
		return null == converter
			? code
			: converter.convert( code );
	}

	public String convertValue ( final String conceptId, final String code ) {
		initKeyFamily();
		final CodeConverter converter
		= (CodeConverter) componentMappings.get( conceptId );
		return null == converter
			? code
			: converter.convert( code );
	}

	public String convertKey ( final String key ) {
		final KeyTokenizer tokenizer = new KeyTokenizer( key );
		final StringBuffer buffer = new StringBuffer();

		for ( int i = 0; tokenizer.hasNext(); ++i ) {
			final String coordinate = tokenizer.nextToken();
			buffer.append( convertCoordinate( i, coordinate ) );
			buffer.append( TimeSeries.KEY_SEPARATOR );
		}

		if ( 0 < buffer.length() ) {
			buffer.setLength( buffer.length() - 1 );
		}

		return buffer.toString();
	}

	public Observation convertObservation ( final Observation observation ) {
		initKeyFamily();
		return new Observation(
			convertValue(
				keyFamily.getObsValue().getConceptId(),
				observation.getValue()
			),
			convertValue(
				keyFamily.getObsStatus().getConceptId(),
				observation.getStatus()
			),
			convertValue(
				keyFamily.getObsConf().getConceptId(),
				observation.getConfidentiality()
			),
			convertValue(
				keyFamily.getObsPreBreak().getConceptId(),
				observation.getPreBreakValue()
			)
		);
	}

	private void initKeyFamily () {
		if ( null != keyFamily ) return;

		final MaintenanceAgency maintenanceAgency
		= Metabase.getMetabase().getMaintenanceAgency( maintenanceAgencyId );
		if ( null == maintenanceAgency ) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"CodeConverter.initKeyFamily()",
				"Unknown maintenance agency " + maintenanceAgencyId + "."
			);
			return;
		}

		keyFamily = maintenanceAgency.getKeyFamily( keyFamilyId );
		if ( null == keyFamily ) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"CodeConverter.initKeyFamily()",
				"Unknown key family " + keyFamilyId + "."
			);
			return;
		}

		dimensionMappings = new CodeConverter[keyFamily.getKeyDimensionCount()];
		componentMappings = new HashMap();

		for (
			final Iterator components = keyFamily.componentIterator();
			components.hasNext();
		) {
			final KeyFamilyComponent component
			= (KeyFamilyComponent) components.next();

			final String prefix = componentHelper.getPreference(
				index + ".component."
				+ component.getConceptId() + ".prefix",
				null
			);

			final CodeConverter codeConverter;

			if ( null == prefix ) {
				componentHelper.logger.log(
					Level.CONFIG,
					componentHelper.logName,
					"CodeConverter.initKeyFamily()",
					"No conversion for key family " + keyFamilyId
					+ " component " + component.getConceptId() + "."
				);
				codeConverter = null;
			} else {
				componentHelper.logger.log(
					Level.CONFIG,
					componentHelper.logName,
					"CodeConverter.initKeyFamily()",
					"Configure prefix " + prefix
					+ " for key family " + keyFamilyId
					+ " component " + component.getConceptId() + "."
				);
				codeConverter = isInverse
					? (CodeConverter) new PrefixRemoveCodeConverter(
						componentHelper, prefix
					)
					: (CodeConverter) new PrefixCodeConverter( prefix );
			}

			if ( component.isKeyStructureDimension() ) {
				dimensionMappings[ component.getPositionNo() - 1 ]
				= codeConverter;
			}
			componentMappings.put(
				component.getConceptId(),
				codeConverter
			);
		}
	}
}

/* *************************** Interface **************************** */

interface CodeConverter {
	String convert ( String code );
}

/* **************************** Class ***************************** */

class PrefixCodeConverter implements CodeConverter {
	private final String prefix;

	public PrefixCodeConverter (
		final String prefix
	) {
		this.prefix = prefix;
	}

	public String convert ( final String code ) {
		if ( "".equals( code ) ) {
			return code;
		} else {
			return prefix + code;
		}
	}

	public String toString () {
		return getClass().getName() + "( " + prefix + " )";
	}
}

/* **************************** Class ***************************** */

class PrefixRemoveCodeConverter implements CodeConverter {
	private final ComponentHelper componentHelper;
	private final String prefix;

	public PrefixRemoveCodeConverter (
		final ComponentHelper componentHelper,
		final String prefix
	) {
		this.componentHelper = componentHelper;
		this.prefix = prefix;
	}

	public String convert ( final String code ) {
		if ( "".equals( code ) ) {
			return code;
		} else if ( code.startsWith( prefix ) ) {
			return code.substring( prefix.length() );
		} else {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"PrefixRemoveConverter.convert()",
				"Code \"" + code
				+ "\" does not start with prefix \"" + prefix + "\"."
			);
			return code;
		}
	}

	public String toString () {
		return getClass().getName() + "( " + prefix + " )";
	}
}
