// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2007, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.conversion;

import java.util.Collections;
import java.util.Date;

import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;

import Int.ecb.gesmescb.*;
import Int.ecb.gesmescb.event.*;
import Int.ecb.gesmescb.metabase.Metabase;

/* ***************************** Class ****************************** */

/**
 * Tests the implementation of the class {@link CodeBasedConverter}.
 *
 * @author
 * <a href="mailto:sis.external@ecb.int">Bernhard Bodenstorfer</a>
 *
 * @version @(#)$Id: CodeBasedConverterTest.java,v 1.1 2017/01/19 16:14:04 trainot Exp $
 */

public class CodeBasedConverterTest extends TestCase {

	/* ======================= Constants ======================== */

	private static final String objectName
	= CodeBasedConverter.class.getName() + ".Paula";

	private static final String[][] testProperties = {
		{ "1.maintenanceAgency", "ECB" },
		{ "1.keyFamily", "ECB_IEA1" },
		{ "1.dataSet", "ECB_IEAQ" },
		{ "1.component.FREQ.prefix", "FQ_" },
		{ "1.component.REF_AREA.prefix", "RA_" },
		{ "1.component.OBS_CONF.prefix", "OC_" },
		{ "1.component.UNIT.prefix", "UT_" }
	};

	private static final MessageBegin
		STRUCTURE_MESSAGE_BEGIN = new MessageBegin(
			"MREF000001", "LT2", "4F0", "ECB",
			MessageBegin.STATISTICAL_DEFINITIONS,
			"Test Update", Collections.EMPTY_SET
		),
		MESSAGE_BEGIN = new MessageBegin(
			"MREF000001", "LT2", "4F0", "ECB",
			MessageBegin.STATISTICAL_DATA,
			"Test Update", Collections.EMPTY_SET,
			"ECB_IEAQ",
			MessageBegin.UPDATE_OR_REPLACE_ACTION,
			"20040804:1112", "20002003", "702",
			"ECB_IEA1"
		),
		DELETE_MESSAGE_BEGIN = new MessageBegin(
			"MREF000001", "LT2", "4F0", "ECB",
			MessageBegin.STATISTICAL_DATA,
			"Test Delete", Collections.EMPTY_SET,
			"ECB_IEAQ",
			MessageBegin.DELETE_ACTION,
			"20040804:1112", "20002003", "702",
			"ECB_IEA1"
		);

	private static final MessageEnd MESSAGE_END = new MessageEnd(
		"MREF000001"
	);

	private static final KeyFamily KEY_FAMILY = new KeyFamily(
		"ECB_IEA1", "Test KF", "", new Date( 0 ),
		new KeyFamilyComponent[]{
			new KeyFamilyComponent(
				"FREQ", KeyFamilyComponent.KEY_STRUCTURE_DIMENSION,
				"1", "AN..1", "", "", "CL_FREQ"
			),
			new KeyFamilyComponent(
				"REF_AREA", KeyFamilyComponent.KEY_STRUCTURE_DIMENSION,
				"2", "AN1", "", "", "CL_AREA_EE"
			),
			new KeyFamilyComponent(
				"ITEM", KeyFamilyComponent.KEY_STRUCTURE_DIMENSION,
				"3", "AN2", "", "", "CL_ITEM"
			),
			new KeyFamilyComponent(
				"TIME_PERIOD", KeyFamilyComponent.TIME_DIMENSION,
				"4", "AN..35", "", "", ""
			),
			new KeyFamilyComponent(
				"TIME_FORMAT", KeyFamilyComponent.TIME_DIMENSION,
				"5", "AN3", "", "", ""
			),
			new KeyFamilyComponent(
				"OBS_VALUE", KeyFamilyComponent.ARRAY_CELL,
				"6", "AN..15", KeyFamilyComponent.MANDATORY,
				KeyFamilyComponent.ATTACHED_TO_OBSERVATION, ""
			),
			new KeyFamilyComponent(
				"OBS_STATUS", KeyFamilyComponent.ARRAY_CELL,
				"7", "AN1", KeyFamilyComponent.MANDATORY,
				KeyFamilyComponent.ATTACHED_TO_OBSERVATION, "CL_OBS_STATUS"
			),
			new KeyFamilyComponent(
				"OBS_CONF", KeyFamilyComponent.ARRAY_CELL,
				"8", "AN1", KeyFamilyComponent.CONDITIONAL,
				KeyFamilyComponent.ATTACHED_TO_OBSERVATION, "CL_OBS_CONF"
			),
			new KeyFamilyComponent(
				"OBS_PRE_BREAK", KeyFamilyComponent.ARRAY_CELL,
				"9", "AN..15", KeyFamilyComponent.CONDITIONAL,
				KeyFamilyComponent.ATTACHED_TO_OBSERVATION, ""
			),
			new KeyFamilyComponent(
				"UNIT", KeyFamilyComponent.ATTRIBUTE,
				"", "AN..20", KeyFamilyComponent.CONDITIONAL,
				KeyFamilyComponent.ATTACHED_TO_SIBLING_GROUP, "CL_UNIT"
			),
			new KeyFamilyComponent(
				"TITLE", KeyFamilyComponent.ATTRIBUTE,
				"", "A2", KeyFamilyComponent.MANDATORY,
				KeyFamilyComponent.ATTACHED_TO_SIBLING_GROUP, ""
			)
		}
	);

	private static final TimeSeries
		TIME_SERIES = new TimeSeries(
			"A.B.C", "2005", "602", new Observation[]{
				new Observation( "13.6", "A", "F", "" )
			}
		),
		TIME_SERIES_CONVERTED = new TimeSeries(
			"FQ_A.RA_B.C", "2005", "602", new Observation[]{
				new Observation( "13.6", "A", "OC_F", "" )
			}
		);

	private static final Attribute
		ATTRIBUTE = new Attribute(
			".B.C", "", "", "UNIT", "KILO"
		),
		ATTRIBUTE_CONVERTED = new Attribute(
			".RA_B.C", "", "", "UNIT", "UT_KILO"
		);

	final Object[] testInput = {
		STRUCTURE_MESSAGE_BEGIN,
		KEY_FAMILY,
		MESSAGE_END,
		MESSAGE_BEGIN,
		TIME_SERIES,
		ATTRIBUTE,
		MESSAGE_END,
		DELETE_MESSAGE_BEGIN,
		TIME_SERIES,
		ATTRIBUTE,
		MESSAGE_END
	},
	testOutput = {
		STRUCTURE_MESSAGE_BEGIN,
		KEY_FAMILY,
		MESSAGE_END,
		MESSAGE_BEGIN,
		TIME_SERIES_CONVERTED,
		ATTRIBUTE_CONVERTED,
		MESSAGE_END,
		DELETE_MESSAGE_BEGIN,
		TIME_SERIES_CONVERTED,
		ATTRIBUTE_CONVERTED,
		MESSAGE_END
	};

	/* ====================== Constructors ====================== */

	/** Constructs a CodeBasedConverterTest with the specified name. */
	public CodeBasedConverterTest ( final String name ) {
		super( name );
	}

	/* ================== Protected Methods ===================== */

	/** Sets up the test fixture. */
	protected void setUp () {
		for ( int i = 0; i < testProperties.length; ++i ) {
			System.setProperty(
				objectName + "." + testProperties[i][0],
				testProperties[i][1]
			);
		}
	}

	/** Tears down the test fixture. */
	protected void tearDown () {
		for ( int i = 0; i < testProperties.length; ++i ) {
			System.getProperties().remove(
				objectName + "." + testProperties[i][0]
			);
		}
	}

	/* ===================== Public Methods ===================== */

	/**
	 * Assembles and returns a test suite for all
	 * the test methods of this test case.
	 *
	 * @return a non-null test suite.
	 */
	public static Test suite () {
		return new TestSuite( CodeBasedConverterTest.class );
	}

	/** Runs the test case. */
	public static void main ( final String args[] ) {
		final String[] testCaseName = {
			CodeBasedConverterTest.class.getName()
		};
		junit.textui.TestRunner.main( testCaseName );
	}

	/* --------------------- Test Methods --------------------- */

	public void testConvert () {
		final TestEventSource source = new TestEventSource();
		final CodeBasedConverter converter = new CodeBasedConverter( "Paula" );
		final TestEventListener listener = new TestEventListener( testOutput );

		source.addListener( Metabase.getMetabase(), false );
		source.addListener( converter, true );
		converter.addListener( listener, true );
		source.dispatch( testInput );
		listener.close();
	}

	public void testInverseConvert () {
		System.setProperty( objectName + ".1.inverse", "true" );
		final TestEventSource source = new TestEventSource();
		final CodeBasedConverter converter = new CodeBasedConverter( "Paula" );
		final TestEventListener listener = new TestEventListener( testInput );
		System.getProperties().remove( objectName + ".1.inverse" );

		source.addListener( Metabase.getMetabase(), false );
		source.addListener( converter, true );
		converter.addListener( listener, true );
		source.dispatch( testOutput );
		listener.close();
	}

	public void testNotConvert () {
		System.getProperties().remove( objectName + ".1.dataSet" );
		final TestEventSource source = new TestEventSource();
		final CodeBasedConverter converter = new CodeBasedConverter( "Paula" );
		final TestEventListener listener = new TestEventListener( testInput );

		source.addListener( Metabase.getMetabase(), false );
		source.addListener( converter, true );
		converter.addListener( listener, true );
		source.dispatch( testInput );
		listener.close();
	}
}
