// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.conversion;

import Int.ecb.logging.Level;

import com.fame.timeiq.TiqObject;

import Int.ecb.gesmescb.*;
import Int.ecb.gesmescb.event.ErrorEvent;
import Int.ecb.gesmescb.event.GESMESCBEventBroker;

import Int.ecb.gesmescb.fame.FAMEDb;

/* **************************** Class ***************************** */

/**
 * Converts IREF by requesting new IREF from the MCADBS server.
 *
 * <center><table border="3" cellpadding="3">
 * <caption>
 * <a name="prefs">Available {@link Int.ecb.prefs preference} settings</a>
 * </caption>
 * <tr>
 * <th>preference name</th>
 * <th>type</th>
 * <th>meaning</th>
 * <th>default</th>
 * </tr>
 * <tr><td><code>serverName</code></td><td>string</td>
 *	<td>MCADBS server name (like dissserv@stat-prod)</td></tr>
 * <tr><td><code>fameExpression</code></td><td>string</td>
 *	<td>FAME 4GL expression to evaluate on the MCADBS server</td>
 *	<td><code>$iref()</code></td></tr>
 * </table></center>
 *
 * @author
 * <A HREF="mailto:Peter.Lindlacher@ECB.int">Peter Lindlacher</A>
 * @author
 * <A HREF="mailto:Sami.Airo@ECB.int">Sami Airo</A>
 *
 * @version @(#)$Id: FAMEIREFConverter.java,v 1.1 2017/01/19 16:14:04 trainot Exp $
 */

public class FAMEIREFConverter extends GESMESCBEventBroker {

	/* ======================= Constants ======================== */

	/* ========================= Fields ========================= */

   /* Stores the current iref*/
   private String iref;

	/* ====================== Constructors ====================== */

	public FAMEIREFConverter () {
		this( "" );
	}

	public FAMEIREFConverter( final String name ) {
		super( name, true );
	}

	/* ===================== Public Methods ===================== */

	public void handle ( final InterchangeBegin interchangeBegin ) {
      final String serverName = componentHelper.getPreference("serverName", null);

		if (interchangeBegin.isTest()) {
			componentHelper.logger.log(
				Level.INFO,
				componentHelper.logName,
				"handle",
				"Interchange is a test: IREF not converted."
			);
			iref = interchangeBegin.getIref();
         dispatch( interchangeBegin );
         return;
		}
      
      if (serverName == null) {
         final String desc = "Cannot read property \"serverName\".";
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"handle",
				desc
			);
         setActive( false );
         dispatch( new ErrorEvent( desc, ErrorEvent.FAILURE ) );
         return;
      }

		componentHelper.logger.log(
			Level.CONFIG,
			componentHelper.logName,
			"handle",
			"Using MCADBS server " + serverName
		);
   
      final String expression = componentHelper.getPreference("fameExpression", "$iref()");

		componentHelper.logger.log(
			Level.CONFIG,
			componentHelper.logName,
			"handle",
			"Using FAME expression " + expression
		);
   
  		FAMEDb db = new FAMEDb("");
		db.open(serverName, "", null);
      TiqObject irefScalar = db.remEval(expression);
      db.close(true);

      if (irefScalar == null) {
         final String desc = "Cannot retrieve new IREF number.";
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"handle",
				desc
			);
         setActive( false );
         dispatch( new ErrorEvent( desc, ErrorEvent.FAILURE ) );
         return;
      }

      iref = "IREF" + irefScalar.getTiqValue().getStringValue();
   
		if (! iref.equals( interchangeBegin.getIref() ) ) {
			componentHelper.logger.log(
				Level.INFO,
				componentHelper.logName,
				"handle",
				interchangeBegin.getIref() + " converted to " + iref + "."
			);
		}

      dispatch( 
         new InterchangeBegin( 
            iref,
            interchangeBegin.getSender(),
            interchangeBegin.getReceiver(),
            interchangeBegin.getPreparationDateTime(),
            interchangeBegin.isTest()
         )
      );
	}

   public void handle ( final InterchangeEnd interchangeEnd ) {
      dispatch( new InterchangeEnd( iref ) );
   }

}
