// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.conversion;

import java.util.Iterator;
import java.util.List;
import java.util.ArrayList;
import java.util.Map;
import java.util.HashMap;
import java.util.StringTokenizer;
import java.util.NoSuchElementException;

import java.io.File;
import java.io.FileReader;
import java.io.BufferedReader;
import java.io.IOException;

import Int.ecb.logging.Level;

import Int.ecb.gesmescb.*;
import Int.ecb.gesmescb.event.ErrorEvent;
import Int.ecb.gesmescb.event.GESMESCBEventBroker;

/* **************************** Class ***************************** */

/**
 * Converts attribute and time series keys according to preference settings.
 *
 * @author
 * <A HREF="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</A>
 * @author
 * <A HREF="mailto:Sami.Airo@ECB.int">Sami Airo</A>
 *
 * @version @(#)$Id: PreferenceConfiguredKeyConverter.java,v 1.1 2017/01/19 16:14:04 trainot Exp $
 */

public class PreferenceConfiguredKeyConverter extends GESMESCBEventBroker {

	/* ======================= Constants ======================== */

	/* ========================= Fields ========================= */

	private Map keyMap;

	private Map attributeMap;

	private boolean dispatchedMessageBegin;

	private boolean filterMessages;
	
	private boolean reportConversionInMessageIdentity;

	/* ====================== Constructors ====================== */

	public PreferenceConfiguredKeyConverter () {
		this( "" );
	}

	public PreferenceConfiguredKeyConverter( final String name ) {
		super( name, true );
		filterMessages = componentHelper.getPreference(
				"filterMessages", "false"
			).equalsIgnoreCase("true");
		reportConversionInMessageIdentity 
			= componentHelper.getPreference(
				"reportConversionInMessageIdentity",
				"false"
			).equalsIgnoreCase("true");
	}

	/* ===================== Public Methods ===================== */

	public void handle ( final MessageBegin messageBegin ) {
		dispatchedMessageBegin = false;
		keyMap = null;
		attributeMap = null;

		if (messageBegin.isStatisticalDefinitions()) {
			dispatch(messageBegin);
			dispatchedMessageBegin= true;
			return;
		}

		final String dataFlow = messageBegin.getSender()
										+ "." + messageBegin.getReceiver()
										+ "." + messageBegin.getMaintenanceAgencyId()
										+ "." + messageBegin.getKeyFamilyId()
										+ "." + messageBegin.getDataSetId();

		final String convertedDataFlow
		= componentHelper.getPreference("convert." + dataFlow, null);

		if (convertedDataFlow == null) {
			componentHelper.logger.log(
				Level.CONFIG,
				componentHelper.logName,
				"handle",
				"No preference found for converting data flow: " + dataFlow
			);
			if (filterMessages) return;
			dispatch(messageBegin);
			dispatchedMessageBegin = true;
			return;
		}

		keyMap = getMap("keyFile." + dataFlow, dataFlow, "keys");
		attributeMap = getMap("attributeFile." + dataFlow, dataFlow, "attributes");

		componentHelper.logger.log(
			Level.CONFIG,
			componentHelper.logName,
			"handle",
			"Data flow " + dataFlow + " converted to " + convertedDataFlow
		);

		StringTokenizer tokenizer = new StringTokenizer(convertedDataFlow, ".");

		String convertedSender = null;
		String convertedReceiver = null;
		String convertedMaintenanceAgency = null;
		String convertedKeyFamily = null;
		String convertedDataSet = null;
		try {
			convertedSender = tokenizer.nextToken();
			convertedReceiver = tokenizer.nextToken();
			convertedMaintenanceAgency = tokenizer.nextToken();
			convertedKeyFamily = tokenizer.nextToken();
			convertedDataSet = tokenizer.nextToken();
		} catch (final NoSuchElementException e) {
			logAndDispatchSevereError(
				"Error in parsing data flow conversion",
				"handle",
				e
			);
			return;
		}

		final List contacts = new ArrayList();
		for (
			final Iterator iterator = messageBegin.contactIterator();
			iterator.hasNext();
		) {
			contacts.add( iterator.next() );
		}

		String identity = messageBegin.getIdentity();
		
		if (reportConversionInMessageIdentity) {
			identity += "\nOriginal dataflow " + dataFlow
			+ " converted to " + convertedDataFlow;
		}

		dispatch(
			new MessageBegin(
				messageBegin.getMref(),
				convertedSender,
				convertedReceiver,
				convertedMaintenanceAgency,
				messageBegin.getFunction(),
				identity,
				contacts,
				convertedDataSet,
				messageBegin.getAction(),
				messageBegin.getExtractionDateTime(),
				messageBegin.getReportingPeriod(),
				messageBegin.getReportingPeriodFormat(),
				convertedKeyFamily,
				messageBegin.getOriginalDataSetId()
			)
		);

		dispatchedMessageBegin = true;
	}


	public void handle ( final MessageEnd messageEnd ) {
		if (!dispatchedMessageBegin) return;
		dispatch(messageEnd);
	}


	public void handle ( final TimeSeries timeSeries ) {
		if (!dispatchedMessageBegin) return;

		if (keyMap == null) {
			dispatch(timeSeries);
			return;
		}

		componentHelper.logger.log(
			Level.FINE, componentHelper.logName, "handle",
			"handling TimeSeries", timeSeries
		);

		String convertedKey = getConvertedKey(timeSeries.getKey());
		if (convertedKey == null) return;

		dispatch( new TimeSeries(
								convertedKey,
								timeSeries.getPeriod(),
								timeSeries.getPeriodFormat(),
								timeSeries.observationIterator()
							)
		);
	}


	public void handle ( final Attribute attribute ) {
		if (!dispatchedMessageBegin) return;

		if (keyMap == null && attributeMap == null) {
			dispatch(attribute);
			return;
		}

		componentHelper.logger.log(
			Level.FINE, componentHelper.logName, "handle",
			"handling Attribute", attribute
		);

		String convertedKey = getConvertedKey(attribute.getKey());
		if (convertedKey == null) return;

		String convertedName = getConvertedName(attribute.getName());
		if (convertedName == null)	return;

		dispatch( new Attribute(
			convertedKey,
			attribute.getPeriod(),
			attribute.getPeriodFormat(),
			convertedName,
			attribute.getValue()
		) );
	}


	public Map getMap(
		final String preference,
		final String dataFlow,
		final String elementType
	) {
		final String fileName = componentHelper.getPreference(preference, null);

		if (fileName == null) {
			logAndDispatchSevereError(
				"No value for preference variable \"" + preference + "\" found",
				"getMap"
			);
			return new HashMap();
		}

		if ("*".equals(fileName)) {
			componentHelper.logger.log(
				Level.CONFIG,
				componentHelper.logName,
				"getMap",
				"No conversion of " + elementType + " for data flow " + dataFlow
			);
			return null;
		}

		componentHelper.logger.log(
			Level.CONFIG,
			componentHelper.logName,
			"getMap",
			"Reading conversions for " + elementType + " from file " + fileName
		);

		final Map keyMap = new HashMap();
		final File sourceFile = new File( fileName );

		try {
			final BufferedReader br = new BufferedReader(
				new FileReader( sourceFile )
			);

			String line;
			while ( ( line = br.readLine() ) != null ) {
				final StringTokenizer tokenizer = new StringTokenizer(line);
				final String key = tokenizer.nextToken();
				final String convertedKey = tokenizer.nextToken();
				keyMap.put(key, convertedKey);
			}

			br.close();
		} catch ( final IOException e ) {
			logAndDispatchSevereError(
				"Error in reading from file " + sourceFile.getAbsolutePath(),
				"readMap",
				e
			);
			return new HashMap();
		} catch (final NoSuchElementException e) {
			logAndDispatchSevereError(
				"Error in parsing from file " + sourceFile.getAbsolutePath(),
				"readMap",
				e
			);
			return new HashMap();
		}

		return keyMap;
	}

	protected void logAndDispatchSevereError (
		final String msg,
		final String procedureName
	) {
		componentHelper.logger.log(
			Level.SEVERE,
			componentHelper.logName,
			procedureName,
			msg
		);

		dispatch(
			new ErrorEvent( msg, ErrorEvent.ERROR )
		);
	}

	protected void logAndDispatchSevereError (
		final String msg,
		final String procedureName,
		final Exception e
	) {
		componentHelper.logger.log(
			Level.SEVERE,
			componentHelper.logName,
			procedureName,
			msg,
			e
		);

		dispatch(
			new ErrorEvent( msg, ErrorEvent.ERROR )
		);
	}

	public String getConvertedName (
		final String name
	) {
		if (attributeMap == null) {
			return name;
		}

		String convertedName = (String)attributeMap.get(name);

		if (convertedName == null) {
			return null;
		}

		componentHelper.logger.log(
			Level.FINER,
			componentHelper.logName,
			"getConvertedName",
			"Converted attribute name from " + name + " to " + convertedName
		);

		return convertedName;
	}


	public String getConvertedKey (
		final String key
	) {
		if (keyMap == null) {
			return key;
		}

		String convertedKey = (String)keyMap.get(key);

		if (convertedKey == null) {
			logAndDispatchSevereError(
				"No conversion for key " + key,
				"getConvertedKey"
			);
			return null;
		}

		componentHelper.logger.log(
			Level.FINER,
			componentHelper.logName,
			"getConvertedString",
			"Converted key from " + key + " to " + convertedKey
		);

		return convertedKey;
	}

}
