// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001-2005, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.conversion;

import java.util.Properties;
import java.util.Collections;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;

import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;

import Int.ecb.gesmescb.*;
import Int.ecb.gesmescb.event.*;

/* ***************************** Class ****************************** */

/**
 * Tests the implementation of the class {@link PreferenceConfiguredKeyConverter}.
 *
 * @author
 * <A HREF="mailto:Sami.Airo@ECB.int">Sami Airo</A>
 * <a href="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</A>
 *
 * @version @(#)$Id: PreferenceConfiguredKeyConverterTest.java,v 1.1 2017/01/19 16:14:04 trainot Exp $
 */

public class PreferenceConfiguredKeyConverterTest extends TestCase {

	/* ======================= Constants ======================== */

	protected static final InterchangeBegin INTERCHANGE_BEGIN
	= new InterchangeBegin(
		"IREF000001", "LT2", "4F0", "040804:1143", true
	);

	protected static final MessageBegin DATA_MESSAGE_BEGIN_MATCHING = new MessageBegin(
		"MREF000001", "LT2", "4F0", "ECB",
		MessageBegin.STATISTICAL_DATA,
		"Identity", Collections.EMPTY_SET,
		"ECB_QFG",
		MessageBegin.UPDATE_OR_REPLACE_ACTION,
		"20040804:1112", "20002003", "702",
		"ECB_MUFA"
	);

	protected static final MessageEnd DATA_MESSAGE_END = new MessageEnd(
		"MREF000001"
	);

	protected static final MessageBegin DATA_MESSAGE_BEGIN_MISMATCHING = new MessageBegin(
		"MREF000001", "LU2", "4F0", "ECB",
		MessageBegin.STATISTICAL_DATA,
		"Identity", Collections.EMPTY_SET,
		"ECB_QFG",
		MessageBegin.UPDATE_OR_REPLACE_ACTION,
		"20040804:1112", "20002003", "702",
		"ECB_MUFA"
	);

	protected static final MessageBegin DELETE_MESSAGE_BEGIN
	= new MessageBegin(
		"MREF000001", "LT2", "4F0", "ECB",
		MessageBegin.STATISTICAL_DATA,
		"Identity", Collections.EMPTY_SET,
		"ECB_MUFA",
		MessageBegin.DELETE_ACTION,
		"20040804:1112", "20002003", "702",
		"ECB_QFG"
	);

	protected static final TimeSeries TIME_SERIES_MATCHING = new TimeSeries(
		"ABC", "", "", Collections.EMPTY_LIST
	);

	protected static final TimeSeries TIME_SERIES_MISMATCHING = new TimeSeries(
		"BCD", "", "", Collections.EMPTY_LIST
	);

	protected static final Attribute ATTRIBUTE_MATCHING = new Attribute(
		"ABC", "", "", "AVAILABILITY", "A"
	);

	protected static final Attribute ATTRIBUTE_KEY_MISMATCHING = new Attribute(
		"BCD", "", "", "AVAILABILITY", "A"
	);

	protected static final Attribute ATTRIBUTE_NAME_MISMATCHING = new Attribute(
		"ABC", "", "", "TITLE", "A"
	);

	protected static final Attribute ATTRIBUTE_ALL_MISMATCHING = new Attribute(
		"BCD", "", "", "TITLE", "A"
	);

	/* ========================= Fields ========================= */

	private File keyFile, attributeFile;

	/* ====================== Constructors ====================== */

	/**
	 * Constructs a PreferenceConfiguredKeyConverterTest with the specified name.
	 *
	 * @param name Test case name
	 */
	public PreferenceConfiguredKeyConverterTest ( final String name ) {
		super( name );
	}

	/* ================== Protected Methods ===================== */

	/**
	 * Sets up the test fixture.
	 */
	protected void setUp () throws IOException {
		System.setProperty( ".level", "ALL" );
		System.setProperty( "handlers", "Int.ecb.logging.ConsoleHandler" );
		System.setProperty(
			"Int.ecb.logging.ConsoleHandler.formatter",
			"Int.ecb.gesmescb.util.NormalLogFormatter"
		);

		keyFile = File.createTempFile( "keyfile", null );

		final FileWriter fileWriter = new FileWriter( keyFile );
		fileWriter.write( "ABC\tDEF" );
		fileWriter.close();

		keyFile.deleteOnExit();

		attributeFile = File.createTempFile( "attributeFile", null );

		final FileWriter fileWriter2 = new FileWriter( attributeFile );
		fileWriter2.write( "AVAILABILITY\tDISTRIBUTIBILITY" );
		fileWriter2.close();

		attributeFile.deleteOnExit();

		final String className = PreferenceConfiguredKeyConverter.class.getName();
		System.setProperty(
			className + ".convert.LT2.4F0.ECB.ECB_MUFA.ECB_QFG", "LT1.4F1.BIS.BIS_GST1.BIS_QFG"
		);
		System.setProperty(
			className + ".keyFile.LT2.4F0.ECB.ECB_MUFA.ECB_QFG",
			keyFile.getAbsolutePath()
		);
		System.setProperty(
			className + ".attributeFile.LT2.4F0.ECB.ECB_MUFA.ECB_QFG",
			attributeFile.getAbsolutePath()
		);
		System.setProperty(
			className + ".filterMessages",
			"true"
		);
		System.setProperty(
			className + ".reportConversionInMessageIdentity",
			"false"
		);
	}

	/**
	 * Tears down the test fixture.
	 */
	protected void tearDown () {
		final String className = PreferenceConfiguredKeyConverter.class.getName();
		final Properties properties = System.getProperties();
		properties.remove(
			className + "convert.LT2.4F0.ECB.ECB_MUFA.ECB_QFG"
		);
		properties.remove(
			className + ".keyFile.LT2.4F0.ECB.ECB_MUFA.ECB_QFG"
		);
		properties.remove(
			className + ".attributeFile.LT2.4F0.ECB.ECB_MUFA.ECB_QFG"
		);
		properties.remove(
			className + ".filterMessages"
		);
		properties.remove(
			className + ".reportConversionInMessageIdentity"
		);

		keyFile.delete();
		keyFile = null;
		attributeFile.delete();
		attributeFile = null;
	}

	/* ===================== Public Methods ===================== */

	/**
	 * Assembles and returns a test suite for all
	 * the test methods of this test case.
	 *
	 * @return a non-null test suite.
	 */
	public static Test suite () {
		return new TestSuite( PreferenceConfiguredKeyConverterTest.class );
	}

	/**
	 * Runs the test case.
	 */
	public static void main ( final String args[] ) {
		final String[] testCaseName = {
			PreferenceConfiguredKeyConverterTest.class.getName()
		};
		junit.textui.TestRunner.main( testCaseName );
	}

	/* --------------------- Test Methods --------------------- */

	public void testFilterConvertMessage () {
		final String className = PreferenceConfiguredKeyConverter.class.getName();
		System.setProperty(
			className + ".filterMessages",
			"true"
		);
		final PreferenceConfiguredKeyConverter converter = new PreferenceConfiguredKeyConverter();
		final TestEventListener listener = new TestEventListener();
		converter.addListener( listener, true );

		converter.handle( DATA_MESSAGE_BEGIN_MATCHING );
		assertEquals( "MessageBegin size 1", 1, listener.size() );
		assertEquals(
			"MessageBegin converted OK",
			new MessageBegin(
				"MREF000001", "LT1", "4F1", "BIS",
				MessageBegin.STATISTICAL_DATA,
				"Identity", Collections.EMPTY_SET,
				"BIS_QFG",
				MessageBegin.UPDATE_OR_REPLACE_ACTION,
				"20040804:1112", "20002003", "702",
				"BIS_GST1",
				DATA_MESSAGE_BEGIN_MATCHING.getDataSetId()
			),
			listener.getLastObject()
		);

		converter.handle( TIME_SERIES_MATCHING );
		assertEquals( "Timeseries size 2", 2, listener.size() );
		assertEquals(
			"Timeseries converted OK",
			new TimeSeries(
				"DEF", "", "", Collections.EMPTY_LIST
			),
			listener.getLastObject()
		);

		converter.handle( TIME_SERIES_MISMATCHING );
		assertEquals( "Error event 3", 3, listener.size() );
		assertTrue( "ErrorEvent received", listener.getLastObject().getClass() == ErrorEvent.class );

		converter.handle( ATTRIBUTE_MATCHING );
		assertEquals( "ATTRIBUTE_MATCHING size 4", 4, listener.size() );
		assertEquals(
			"ATTRIBUTE_MATCHING converted OK",
			new Attribute(
				"DEF", "", "", "DISTRIBUTIBILITY", "A"
			),
			listener.getLastObject()
		);

		converter.handle( ATTRIBUTE_KEY_MISMATCHING );
		assertEquals( "Error event 5", 5, listener.size() );
		assertTrue( "ErrorEvent received", listener.getLastObject().getClass() == ErrorEvent.class );

		converter.handle( ATTRIBUTE_NAME_MISMATCHING );
		assertEquals( "ATTRIBUTE_MATCHING unknown 5", 5, listener.size() );

		converter.handle( ATTRIBUTE_ALL_MISMATCHING );
		assertEquals( "Error event 6", 6, listener.size() );
		assertTrue( "ErrorEvent received", listener.getLastObject().getClass() == ErrorEvent.class );

		converter.handle( DATA_MESSAGE_END );
		assertEquals( "MessageEnd size 7", 7, listener.size() );
		assertEquals(
			"MessageEnd converted OK",
			DATA_MESSAGE_END,
			listener.getLastObject()
		);
	}

	public void testFilterNotConvertMessage () {
		final String className = PreferenceConfiguredKeyConverter.class.getName();
		System.setProperty(
			className + ".filterMessages",
			"true"
		);
		final PreferenceConfiguredKeyConverter converter = new PreferenceConfiguredKeyConverter();
		final TestEventListener listener = new TestEventListener();
		converter.addListener( listener, true );

		converter.handle( DATA_MESSAGE_BEGIN_MISMATCHING );
		assertEquals( "MessageBegin size 0", 0, listener.size() );

		converter.handle( TIME_SERIES_MATCHING );
		assertEquals( "Timeseries size 0", 0, listener.size() );

		converter.handle( ATTRIBUTE_MATCHING );
		assertEquals( "Attribute size 0", 0, listener.size() );

		converter.handle( DATA_MESSAGE_END );
		assertEquals( "MessageEnd size 0", 0, listener.size() );
	}

	public void testNotFilterNotConvertMessage () {
		final String className = PreferenceConfiguredKeyConverter.class.getName();
		System.setProperty(
			className + ".filterMessages",
			"false"
		);
		final PreferenceConfiguredKeyConverter converter = new PreferenceConfiguredKeyConverter();
		final TestEventListener listener = new TestEventListener();
		converter.addListener( listener, true );

		converter.handle( DATA_MESSAGE_BEGIN_MISMATCHING );
		assertEquals( "MessageBegin size 1", 1, listener.size() );
		assertEquals(
			"MessageBegin converted OK",
			DATA_MESSAGE_BEGIN_MISMATCHING,
			listener.getLastObject()
		);

		converter.handle( TIME_SERIES_MATCHING );
		assertEquals( "Timeseries size 2", 2, listener.size() );
		assertEquals(
			"Timeseries converted OK",
			TIME_SERIES_MATCHING,
			listener.getLastObject()
		);

		converter.handle( ATTRIBUTE_MATCHING );
		assertEquals( "ATTRIBUTE_MATCHING size 3", 3, listener.size() );
		assertEquals(
			"ATTRIBUTE_MATCHING converted OK",
			ATTRIBUTE_MATCHING,
			listener.getLastObject()
		);

		converter.handle( DATA_MESSAGE_END );
		assertEquals( "MessageEnd size 4", 4, listener.size() );
		assertEquals(
			"MessageEnd converted OK",
			DATA_MESSAGE_END,
			listener.getLastObject()
		);
	}

	public void testConvertReportMessage () {
		final String className = PreferenceConfiguredKeyConverter.class.getName();

		System.setProperty(
			className + ".reportConversionInMessageIdentity",
			"true"
		);

		final PreferenceConfiguredKeyConverter converter = new PreferenceConfiguredKeyConverter();
		final TestEventListener listener = new TestEventListener();
		converter.addListener( listener, true );

		converter.handle( DATA_MESSAGE_BEGIN_MATCHING );
		assertEquals( "MessageBegin size 1", 1, listener.size() );
		assertEquals(
			"MessageBegin converted OK",
			new MessageBegin(
				"MREF000001", "LT1", "4F1", "BIS",
				MessageBegin.STATISTICAL_DATA,
				DATA_MESSAGE_BEGIN_MATCHING.getIdentity() + 
				"\nOriginal dataflow LT2.4F0.ECB.ECB_MUFA.ECB_QFG converted to LT1.4F1.BIS.BIS_GST1.BIS_QFG", Collections.EMPTY_SET,
				"BIS_QFG",
				MessageBegin.UPDATE_OR_REPLACE_ACTION,
				"20040804:1112", "20002003", "702",
				"BIS_GST1",
				DATA_MESSAGE_BEGIN_MATCHING.getDataSetId()
			),
			listener.getLastObject()
		);

		converter.handle( TIME_SERIES_MATCHING );
		assertEquals( "Timeseries size 2", 2, listener.size() );
		assertEquals(
			"Timeseries converted OK",
			new TimeSeries(
				"DEF", "", "", Collections.EMPTY_LIST
			),
			listener.getLastObject()
		);

		converter.handle( TIME_SERIES_MISMATCHING );
		assertEquals( "Error event 3", 3, listener.size() );
		assertTrue( "ErrorEvent received", listener.getLastObject().getClass() == ErrorEvent.class );

		converter.handle( ATTRIBUTE_MATCHING );
		assertEquals( "ATTRIBUTE_MATCHING size 4", 4, listener.size() );
		assertEquals(
			"ATTRIBUTE_MATCHING converted OK",
			new Attribute(
				"DEF", "", "", "DISTRIBUTIBILITY", "A"
			),
			listener.getLastObject()
		);

		converter.handle( ATTRIBUTE_KEY_MISMATCHING );
		assertEquals( "Error event 5", 5, listener.size() );
		assertTrue( "ErrorEvent received", listener.getLastObject().getClass() == ErrorEvent.class );

		converter.handle( ATTRIBUTE_NAME_MISMATCHING );
		assertEquals( "ATTRIBUTE_MATCHING unknown 5", 5, listener.size() );

		converter.handle( ATTRIBUTE_ALL_MISMATCHING );
		assertEquals( "Error event 6", 6, listener.size() );
		assertTrue( "ErrorEvent received", listener.getLastObject().getClass() == ErrorEvent.class );

		converter.handle( DATA_MESSAGE_END );
		assertEquals( "MessageEnd size 7", 7, listener.size() );
		assertEquals(
			"MessageEnd converted OK",
			DATA_MESSAGE_END,
			listener.getLastObject()
		);
	}
}
