// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2003, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.csv;

import java.util.NoSuchElementException;

import java.io.Reader;
import java.io.BufferedReader;
import java.io.IOException;

/* ***************************** Class ****************************** */

/**
 * <p>Parses a CSV character stream into cells.</p>
 *
 * @author
 * <A HREF="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</A>
 * @author
 * <A HREF="mailto:Ana.Padrao@ECB.int">Ana Garcia Padrao</A>
 *
 * @version @(#)$Id: CSVTokenizer.java,v 1.1 2017/01/19 16:14:04 trainot Exp $
 */
public class CSVTokenizer {

	/* =================== Nested Interfaces ==================== */

	public interface Line {
		boolean hasMoreCells () throws IOException;
		String nextCell () throws IOException;
		int getLineNo ();
		int getCellNo ();
	}

	/* ========================= Fields ========================= */

	/** Source stream. */
	private final BufferedReader reader;
	private int lineNo;

	/* ====================== Constructors ====================== */

	/**
	 * Construct a <code>CSVTokenizer</code> to read from a file.
	 *
	 * @throws IOException on <code>reader</code> errors.
	 */
	public CSVTokenizer ( final Reader reader )
		throws IOException
	{
		this.reader = reader instanceof BufferedReader
			? (BufferedReader) reader
			: new BufferedReader( reader );
		lineNo = 0;
	}

	/* ===================== Public Methods ===================== */

	/** Tells if more lines are available after the current one. */
	public boolean hasMoreLines () throws IOException {
		reader.mark( 1 );
		final int c = reader.read();
		if ( c != -1 ) {
			reader.reset();
			return true;
		}
		return false;
	}

	/** Gets a new line. */
	public Line nextLine () throws IOException {
		final String line = reader.readLine();
		if ( line == null ) {
			throw new NoSuchElementException( "No more CSV lines." );
		}
		++lineNo;
		return new CSVTokenizerLine ( line, lineNo );
	}
}

/* ***************************** Class ****************************** */

class CSVTokenizerLine implements CSVTokenizer.Line {

	/* ========================= Fields ========================= */

	private final String line;
	private final int lineNo, length;
	private int commas, column, cursor;

	/* ====================== Constructors ====================== */

	CSVTokenizerLine ( final String line, final int lineNo ) {
		this.line = line;
		this.lineNo = lineNo;
		length = line.length();
		commas = column = cursor = 0;
	}

	/* ===================== Public Methods ===================== */

	/** Tells if more cells are available. */
	public boolean hasMoreCells () {
		return cursor < length || commas > 0;
	}

	/** Gets next cell. */
	public String nextCell () {
		if ( hasMoreCells() ) {
			if ( commas > 0 ) {
				--commas;
				++column;
				return "";
			}
			final int nextComma = line.indexOf( ',', cursor );
			final String cell;
			if ( nextComma >= 0 ) {
				cell = line.substring( cursor, nextComma );
				cursor = nextComma + 1;
				while ( cursor < length && line.charAt( cursor ) == ',' ) {
					++commas;
					++cursor;
				}
			} else {
				cell = line.substring( cursor );
				cursor = length;
			}
			if ( cursor == length ) {
				commas = 0;
			}
			++column;
			return cell;
		} else {
			throw new NoSuchElementException( "No more CSV cells." );
		}
	}

	public int getLineNo () {
		return lineNo;
	}

	public int getCellNo () {
		return column;
	}
}
