// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2007, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.csv;

import java.io.File;
import java.io.StringReader;
import java.io.FileReader;
import java.io.BufferedReader;
import java.io.IOException;

import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;

import Int.ecb.gesmescb.*;

import Int.ecb.gesmescb.event.GESMESCBWriter;
import Int.ecb.gesmescb.event.TestEventSource;

/* ***************************** Class ****************************** */

/**
 * Implements a few tests for the {@link EaacsDumpWriter}.
 *
 * @author
 * <a href="mailto:sis.external@ecb.int">Bernhard Bodenstorfer</a>
 *
 * @version @(#)$Id: EaacsDumpWriterTest.java,v 1.1 2017/01/19 16:14:04 trainot Exp $
 */
public class EaacsDumpWriterTest extends TestCase {

	/* ======================= Constants ======================== */

	private static final String
		objectName = EaacsDumpWriter.class.getName() + ".Xavier";

	private static final MessageBegin UPDATE_MESSAGE
	= new MessageBegin(
		"MREF000002", "4F0", "ZZZ", "ECB",
		MessageBegin.STATISTICAL_DATA,
		"Test data", new Contact[]{},
		"ECB_IEAQ", MessageBegin.UPDATE_OR_REPLACE_ACTION,
		"199901011432", "", "", "ECB_IEA1"
	);

	private static final MessageBegin DELETE_MESSAGE
	= new MessageBegin(
		"MREF000002", "4F0", "ZZZ", "ECB",
		MessageBegin.STATISTICAL_DATA,
		"Test delete", new Contact[]{},
		"ECB_IEAQ", MessageBegin.DELETE_ACTION,
		"199901011432", "", "", "ECB_IEA1"
	);

	private static final MessageEnd MESSAGE_END = new MessageEnd(
		"MREF000001"
	);

	private static final TimeSeries
		TS1 = new TimeSeries(
			"A.U2.C", "19951998", "702", new Observation[]{
				new Observation( "3.6", "A", "F", "" ),
				Observation.EMPTY_OBSERVATION,
				new Observation( "23.6", "A", "F", "" ),
				new Observation( "33.6", "A", "F", "" )
			}
		),
		TS2 = new TimeSeries(
			"Q.EE.C", "1995119954", "708", new Observation[]{
				new Observation( "4.6", "A", "C", "-" ),
				new Observation( "14.6", "A", "C", "-" ),
				Observation.EMPTY_OBSERVATION,
				new Observation( "-3.6", "B", "F", "" )
			}
		),
		TS3 = new TimeSeries(
			"Q.U2.C", "1995319981", "708", new Observation[]{
				new Observation( "7.6", "A", "C", "-" ),
				Observation.EMPTY_OBSERVATION,

				Observation.EMPTY_OBSERVATION,
				Observation.EMPTY_OBSERVATION,
				Observation.EMPTY_OBSERVATION,
				Observation.EMPTY_OBSERVATION,

				Observation.EMPTY_OBSERVATION,
				new Observation( "74.6", "A", "C", "-" ),
				new Observation( "17.6", "A", "C", "-" ),
				Observation.EMPTY_OBSERVATION,

				new Observation( "14.7", "A", "C", "-" )
			}
		);

	private static final String testOutput
	= "\"A\" \"Q1\" \"Q2\" \"Q3\" \"Q4\"\n"
	+ "\"A\" \"U2\" \"C\"\n"
	+ "\"1995\" 3.6\n"
	+ "\"1997\" 23.6\n"
	+ "\"1998\" 33.6\n"
	+ "\"Q\" \"EE\" \"C\"\n"
	+ "\"1995\" #Mi 4.6 14.6 #Mi -3.6\n"
	+ "\"Q\" \"U2\" \"C\"\n"
	+ "\"1995\" #Mi #Mi #Mi 7.6\n"
	+ "\"1997\" #Mi #Mi 74.6 17.6\n"
	+ "\"1998\" #Mi 14.7\n";

	/* ========================= Fields ========================= */

	/** The temporary destination for test output. */
	private File file;

	/* ====================== Constructors ====================== */

	public EaacsDumpWriterTest ( final String name ) {
		super( name );
	}

	/* ================== Protected Methods ===================== */

	/** Sets up the test fixture. */
	protected void setUp () throws IOException {
		file = File.createTempFile( "EaacsDumpWriterTest", ".dump" );
		file.deleteOnExit();
		System.setProperty(
			objectName + ".destination",
			file.getAbsolutePath()
		);
	}

	/** Tears down the test fixture. */
	protected void tearDown () {
		System.getProperties().remove( objectName + ".destination" );
		file.delete();
		file = null;
	}

	/* ===================== Public Methods ===================== */

	/**
	 * Assembles and returns a test suite for all
	 * the test methods of this test case.
	 *
	 * @return a non-null test suite
	 */
	public static Test suite () {
		return new TestSuite( EaacsDumpWriterTest.class );
	}

	/**
	 * @param arg the array of argument {@link String}s
	 */
	public static void main ( final String[] arg ) {
		final String[] testCaseName = {
			EaacsDumpWriterTest.class.getName()
		};
		junit.textui.TestRunner.main( testCaseName );
	}

	/* ---------------------- Test Methods ---------------------- */

	/**
	 * Test {@link EaacsDumpWriter}.
	 *
	 * @throws IOException if opening the EaacsDumpWriter yields a problem.
	 */
	public void testAnnual () throws IOException {
		final TestEventSource source = new TestEventSource();
		final GESMESCBWriter writer = new EaacsDumpWriter( "Xavier" );

		source.addListener( writer, true );

		source.dispatch( new Object[]{
			UPDATE_MESSAGE,
			TS1, TS2, TS3,
			MESSAGE_END
		} );

		writer.close();

		assertFileContent( "Test update", testOutput, file );
	}

	/**
	 * Test {@link EaacsDumpWriter}.
	 *
	 * @throws IOException if opening the EaacsDumpWriter yields a problem.
	 */
	public void testIgnoreDelete () throws IOException {
		final TestEventSource source = new TestEventSource();
		final GESMESCBWriter writer = new EaacsDumpWriter( "Xavier" );

		source.addListener( writer, true );

		source.dispatch( new Object[]{
			DELETE_MESSAGE,
			TS1, TS2, TS3,
			MESSAGE_END
		} );

		writer.close();

		assertFileContent( "Test delete", "", file );
	}

	private void assertFileContent (
		final String message,
		final String expectedContent,
		final File file
	) throws IOException {
		final BufferedReader
			expectedReader = new BufferedReader( new StringReader( expectedContent ) ),
			actualReader = new BufferedReader( new FileReader( file ) );

		String expected = "", actual = "";
		for ( int i = 1; null != expected; ++i ) {
			expected = expectedReader.readLine();
			actual = actualReader.readLine();
			assertEquals( message + " line " + i, expected, actual );
		}
		actualReader.close();
		expectedReader.close();
	}
}
