// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2007, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.csv;

import java.util.Date;
import java.util.StringTokenizer;
import java.util.Iterator;
import java.util.Map;
import java.util.HashMap;

import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.text.ParseException;

import Int.ecb.logging.Level;

import Int.ecb.gesmescb.KeyFamily;
import Int.ecb.gesmescb.KeyFamilyComponent;

import Int.ecb.gesmescb.time.RegularisableInterval;

import Int.ecb.gesmescb.metabase.Metabase;
import Int.ecb.gesmescb.metabase.MaintenanceAgency;

import Int.ecb.gesmescb.util.ComponentHelper;

/* ***************************** Class ****************************** */

/**
 * <p>Administers the configuration for the EAACS Matlab transformation.</p>
 *
 * <center><table border="3" cellpadding="3">
 * <caption>
 * <a name="prefs">Available {@link Int.ecb.prefs preference} settings</a>
 * </caption>
 * <tr>
 * <th>preference name</th>
 * <th>type</th>
 * <th>meaning</th>
 * <th>default</th>
 * </tr>
 * <tr>
 *	<td><code>begin</code></td>
 *	<td>string</td>
 *	<td>Begin time point of transmitted data in ISO format
 *		<i>yyyy-mm-ddThh:mm:ss</i></td>
 * </tr>
 * <tr>
 *	<td><code>end</code></td>
 *	<td>string</td>
 *	<td>End time point of transmitted data in ISO format
 *		<i>yyyy-mm-ddThh:mm:ss</i></td>
 * </tr>
 * <tr>
 *	<td><code><i>i</i>.dataSet</code></td>
 *	<td>string</td>
 *	<td><i>i</i>-th configuration record, data set</td>
 * </tr>
 * <tr>
 *	<td><code><i>i</i>.maintenanceAgency</code></td>
 *	<td>string</td>
 *	<td><i>i</i>-th configuration record, maintenanceAgency agency</td>
 * </tr>
 * <tr>
 *	<td><code><i>i</i>.keyFamily</code></td>
 *	<td>string</td>
 *	<td><i>i</i>-th configuration record, key family</td>
 * </tr>
 * <tr>
 *	<td><code><i>i</i>.dimension.<i>concept</i>.accept</code></td>
 *	<td>string</td>
 *	<td><i>i</i>-th configured key family,
 *		the accepted codes for the dimension specified as concept;
 *		use this or the respective <code>translate</code> option</td>
 * </tr>
 * <tr>
 *	<td><code><i>i</i>.dimension.<i>concept</i>.translate</code></td>
 *	<td>string</td>
 *	<td><i>i</i>-th configured key family,
 *		the translateable codes for the dimension specified as concept;
 *		use this or the respective <code>accept</code> option</td>
 * </tr>
 * </table></center>
 *
 * @author
 * <a href="mailto:sis.external@ecb.int">Bernhard Bodenstorfer</a>
 *
 * @version @(#)$Id: EaacsLabConfig.java,v 1.1 2017/01/19 16:14:04 trainot Exp $
 */
public class EaacsLabConfig {

	/* ========================= Fields ========================= */

	/** Takes care of logging and preferences. */
	protected final ComponentHelper componentHelper;

	/** The configured begin date. */
	private final Date begin;

	/** The configured end date. */
	private final Date end;

	/** The covering intervals administration. */
	private final RegularisableInterval possibleIntervals;

	/** For configured data set. */
	private String dataSetId, maintenanceAgencyId, keyFamilyId;

	/** For configured data set. */
	private KeyFamily keyFamily;

	/** Holds the objects to evaluate code values for all dimensions. */
	private CodeEvaluator[] dimensionEvaluators;

	/* ====================== Constructors ====================== */

	/**
	 * Construct a <code>EaacsLabConfig</code>.
	 *
	 * @param name used for logging and preferences.
	 * @param filename where to write to.
	 *
	 * @throws NullPointerException if <code>filename==null</code>
	 * and the preference variable
	 * <code>destination</code> is undefined.
	 * @throws IOException if the file cannot be opened for writing.
	 */
	public EaacsLabConfig ( final String name ) {
		componentHelper = new ComponentHelper( name, this );

		final DateFormat format = new SimpleDateFormat(
			"yyyy-MM-dd'T'HH:mm:ss"
		);

		{
			final String beginString = componentHelper.getPreference(
				"begin", null
			);
			Date begin;
			if ( null == beginString ) {
				componentHelper.logger.log(
					Level.SEVERE,
					componentHelper.logName,
					"EaacsLabConfig()",
					"No begin time configuration found."
				);
				begin = null;
			} else try {
				begin = format.parse( beginString );
			} catch ( final ParseException e ) {
				componentHelper.logger.log(
					Level.SEVERE,
					componentHelper.logName,
					"EaacsLabConfig()",
					"Begin time configuration \"" + beginString
					+ "\" is not in ISO format yyyy-mm-ddThh:mm:ss.",
					e
				);
				begin = null;
			}
			this.begin = begin;
		}

		{
			final String endString = componentHelper.getPreference(
				"end", null
			);
			Date end;
			if ( null == endString ) {
				componentHelper.logger.log(
					Level.SEVERE,
					componentHelper.logName,
					"EaacsLabConfig()",
					"No end time configuration found."
				);
				end = null;
			} else try {
				end = format.parse( endString );
			} catch ( final ParseException e ) {
				componentHelper.logger.log(
					Level.SEVERE,
					componentHelper.logName,
					"EaacsLabConfig()",
					"End time configuration \"" + endString
					+ "\" is not in ISO format yyyy-mm-ddThh:mm:ss.",
					e
				);
				end = null;
			}
			this.end = end;
		}

		possibleIntervals = null != begin && null != end
			? new RegularisableInterval( begin, end )
			: null;
	}

	/* ===================== Public Methods ===================== */

	/** Retrieve the configured time interval begin. */
	public Date getBegin () {
		return begin;
	}

	/** Retrieve the configured time interval end. */
	public Date getEnd () {
		return end;
	}

	/** Get the possible intervals from begin to end.
	 *
	 * @return <code>null</code> if <code>begin</code> or <code>end</code>
	 * not configured.
	 */
	public RegularisableInterval getPossibleIntervals () {
		return possibleIntervals;
	}

	/** Obtain and store the configuration for the given dataset.
	 *
	 * This method must be called before any other data set based method.
	 */
	public void configure ( final String dataSetId ) {
		componentHelper.logger.log(
			Level.FINE,
			componentHelper.logName,
			"configure(dataSetId)",
			"Retrieve configuration for data set " + dataSetId,
			dataSetId
		);

		maintenanceAgencyId = null;
		keyFamilyId = null;
		keyFamily = null;
		this.dataSetId = null;
		int dataSetIndex = 0;

		for ( int configIndex = 1;; ++configIndex ) {
			final String dataSetConfig = componentHelper.getPreference(
				configIndex + ".dataSet", null
			);
			if ( null == dataSetConfig ) {
				if ( null == this.dataSetId ) {
					componentHelper.logger.log(
						Level.SEVERE,
						componentHelper.logName,
						"configure( dataSetId )",
						"No configuration found for data set \""
						+ dataSetId + "\"."
					);
					return;
				} else {
					break;
				}
			} else if ( dataSetId.equals( dataSetConfig ) ) {
				if ( null != this.dataSetId ) {
					componentHelper.logger.log(
						Level.WARNING,
						componentHelper.logName,
						"configure( dataSetId )",
						"Duplicate configuration for data set \""
						+ dataSetId + "\"."
					);
				}
				dataSetIndex = configIndex;
				this.dataSetId = dataSetConfig;

				final String maintenanceAgencyConfig
				= componentHelper.getPreference(
					configIndex + ".maintenanceAgency", null
				);
				if ( null == maintenanceAgencyConfig ) {
					componentHelper.logger.log(
						Level.SEVERE,
						componentHelper.logName,
						"configure( dataSetId )",
						"No maintenance agency configuration"
						+ " found for data set \"" + dataSetId + "\"."
					);
				} else if (
					null != maintenanceAgencyId
					&&
					! maintenanceAgencyId.equals( maintenanceAgencyConfig )
				) {
					componentHelper.logger.log(
						Level.SEVERE,
						componentHelper.logName,
						"configure( dataSetId )",
						"Contradicting maintenance agency configurations"
						+ " found for data set \"" + dataSetId + "\": "
						+ maintenanceAgencyId + " and "
						+ maintenanceAgencyConfig
					);
				} else {
					maintenanceAgencyId = maintenanceAgencyConfig;
					componentHelper.logger.log(
						Level.CONFIG,
						componentHelper.logName,
						"configure( dataSetId )",
						"Configured maintenance agency \""
						+ maintenanceAgencyConfig
						+ "\" for data set \"" + dataSetId + "\"."
					);
				}

				final String keyFamilyConfig = componentHelper.getPreference(
					configIndex + ".keyFamily", null
				);
				if ( null == keyFamilyConfig ) {
					componentHelper.logger.log(
						Level.SEVERE,
						componentHelper.logName,
						"configure( dataSetId )",
						"No key family configuration"
						+ " found for data set \"" + dataSetId + "\"."
					);
				} else if (
					null != keyFamilyId
					&&
					! keyFamilyId.equals( keyFamilyConfig )
				) {
					componentHelper.logger.log(
						Level.SEVERE,
						componentHelper.logName,
						"configure( dataSetId )",
						"Contradicting key family configurations"
						+ " found for data set \"" + dataSetId + "\": "
						+ keyFamilyId + " and "
						+ keyFamilyConfig
					);
				} else {
					keyFamilyId = keyFamilyConfig;
					componentHelper.logger.log(
						Level.CONFIG,
						componentHelper.logName,
						"configure( dataSetId )",
						"Configured key family \""
						+ keyFamilyConfig
						+ "\" for data set \"" + dataSetId + "\"."
					);
				}
			}
		}

		if ( null != maintenanceAgencyId && null != keyFamilyId ) {

			final Metabase metabase = Metabase.getMetabase();
			final MaintenanceAgency maintenanceAgency
			= metabase.getMaintenanceAgency( maintenanceAgencyId );
			if ( maintenanceAgency == null ) {
				componentHelper.logger.log(
					Level.SEVERE,
					componentHelper.logName,
					"configure( dataSetId )",
					"Unknown maintenance agengy \"" + maintenanceAgencyId
					+ "\"."
				);
				return;
			}

			keyFamily = maintenanceAgency.getKeyFamily( keyFamilyId );
			if ( keyFamily == null ) {
				componentHelper.logger.log(
					Level.SEVERE,
					componentHelper.logName,
					"configure( dataSetId )",
					"Maintenance agengy " + maintenanceAgencyId
					+ " does not know key family \""
					+ keyFamilyId + "\"."
				);
				return;
			}

			final int dimensionCount =  keyFamily.getKeyDimensionCount();
			dimensionEvaluators = new CodeEvaluator[ dimensionCount ];
			for (
				final Iterator dimensions
				= keyFamily.keyStructureIterator();
				dimensions.hasNext();
			) {
				final KeyFamilyComponent dimension
				= ( KeyFamilyComponent) dimensions.next();
				final String conceptId = dimension.getConceptId();
				final int index = dimension.getPositionNo() - 1;

				String acceptedCodes
				= componentHelper.getPreference(
					dataSetIndex + ".dimension." + conceptId + ".accept",
					null
				);
				final String translatedCodes
				= componentHelper.getPreference(
					dataSetIndex + ".dimension." + conceptId + ".translate",
					null
				);

				if ( null == acceptedCodes && null == translatedCodes ) {
					componentHelper.logger.log(
						Level.CONFIG,
						componentHelper.logName,
						"configure( dataSetId )",
						"Missing configuration for data set \""
						+ dataSetId
						+ "\" maintenance agency \""
						+ maintenanceAgencyId
						+ "\" key family \""
						+ keyFamilyId
						+ "\" concept \""
						+ conceptId
						+ "\" of accepted or translated codes."
						+ " Using default: accept = *."
					);
				}

				if ( null != acceptedCodes && null != translatedCodes ) {
					componentHelper.logger.log(
						Level.SEVERE,
						componentHelper.logName,
						"configure( dataSetId )",
						"Duplicate configuration for data set \""
						+ dataSetId
						+ "\" maintenance agency \""
						+ maintenanceAgencyId
						+ "\" key family \""
						+ keyFamilyId
						+ "\" concept \""
						+ conceptId
						+ "\": Accepted codes \""
						+ acceptedCodes
						+ "\" plus translated codes \""
						+ translatedCodes
						+ "\"."
					);
				}

				if ( null != acceptedCodes ) {
					dimensionEvaluators[ index ]
						= "*".equals( acceptedCodes )
							? new CodeAcceptor()
							: new CodeAcceptor( acceptedCodes );
				}

				if ( null != translatedCodes ) {
					dimensionEvaluators[ index ]
						= "*".equals( translatedCodes )
							? new CodeToIndexTranslator()
							: new CodeToIndexTranslator( translatedCodes );
				}
			}
		}
	}

	/** Return the data set identifier
	 * which has been configured most recently.
	 */
	public String getDataSetId () {
		return dataSetId;
	}

	/** Return the maintenance agency identifier for the data set
	 * which has been configured most recently.
	 */
	public String getMaintenanceAgencyId () {
		return maintenanceAgencyId;
	}

	/** Return the key family identifier for the data set
	 * which has been configured most recently.
	 */
	public String getKeyFamilyId () {
		return keyFamilyId;
	}

	/** Return the key family for the data set
	 * which has been configured most recently.
	 */
	public KeyFamily getKeyFamily () {
		return keyFamily;
	}

	/** Return the number of key structure dimensions
	 * of the key family for the configured data set.
	 */
	public int getDimensionCount () {
		return dimensionEvaluators.length;
	}

	/** Return whether the given dimension is to be translated into indexes. */
	public boolean isIndexed ( final int dimensionIndex ) {
		return dimensionEvaluators[ dimensionIndex ]
			instanceof CodeToIndexTranslator;
	}

	/** Return for the given dimension how many possible codes are allowed. */
	public int codeCount ( final int dimensionIndex ) {
		final CodeEvaluator evaluator = dimensionEvaluators[ dimensionIndex ];
		return evaluator.size();
	}

	/** Return whether the given code is fine.
	 * This method will raise an exception if called on indexed dimensions.
	 */
	public boolean codeAccept ( final int dimensionIndex, final String code ) {
		final CodeAcceptor acceptor
			= ( CodeAcceptor ) dimensionEvaluators[ dimensionIndex ];
		return acceptor.accept( code );
	}

	/** Return for the given dimension whether codes are automatically indexed.
	 * This method will raise an exception if called on non-indexed dimensions.
	 */
	public boolean isAutomatic ( final int dimensionIndex ) {
		final CodeToIndexTranslator translator
			= ( CodeToIndexTranslator ) dimensionEvaluators[ dimensionIndex ];
		return translator.isAutomatic();
	}

	/** Return for the given dimension the index of the given code.
	 * This method will raise an exception if called on non-indexed dimensions.
	 *
	 * @return the index or <code>null</code> if the code is not registered.
	 */
	public Integer codeIndex ( final int dimensionIndex, final String code ) {
		final CodeToIndexTranslator translator
			= ( CodeToIndexTranslator ) dimensionEvaluators[ dimensionIndex ];
		return translator.translate( code );
	}
}

/* *************************** Interface **************************** */

/** Commonly needed for assessing or translating codes. */
interface CodeEvaluator {
	/** Return either a Boolean or an Integer.
	 * Boolean would return whether the code is accepted.
	Object evaluate ( final String code );
	/** How many codes are known. */
	int size ();
}

/* ***************************** Class ****************************** */

/** Translates codes to Indexes. */
class CodeAcceptor implements CodeEvaluator {

	private String code;

	/** Construct one which automatically inserts the first encountered code. */
	public CodeAcceptor () {
		code = null;
	}

	/** Construct one with pre-set code accepted. */
	public CodeAcceptor ( final String code ) {
		this.code = code;
	}

	public boolean accept ( final String code ) {
		if ( null == this.code ) {
			this.code = code;
		}
		return this.code.equals( code );
	}

	public int size () {
		return null == this.code ? 0 : 1;
	}
}

/* ***************************** Class ****************************** */

/** Translates codes to Indexes. */
class CodeToIndexTranslator implements CodeEvaluator {

	private final Map indexes;
	private final boolean autoUpdate;

	/** Construct one which automatically adds new codes. */
	public CodeToIndexTranslator () {
		indexes = new HashMap();
		autoUpdate = true;
	}

	/** Construct one with a constant set of codes. */
	public CodeToIndexTranslator ( final String codes ) {
		indexes = new HashMap();
		for (
			final StringTokenizer translate
				= new StringTokenizer( codes, ", " );
			translate.hasMoreTokens();
		) {
			final String code = translate.nextToken();
			update( code );
		}
		autoUpdate = false;
	}

	/** Translate the code to the associated index. */
	public Integer translate ( final String code ) {
		if ( autoUpdate ) {
			update( code );
		}
		return (Integer) indexes.get( code );
	}

	/** Whether new codes will be indexed as they turn up. */
	public boolean isAutomatic () {
		return autoUpdate;
	}

	public int size () {
		return indexes.size();
	}

	/** Add a code, if necessary. */
	private void update ( final String code ) {
		if ( ! indexes.containsKey( code ) ) {
			indexes.put( code, new Integer( 1 + indexes.size() ) );
		}
	}
}
