// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2007, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.csv;

import java.util.Properties;
import java.util.Date;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.text.ParseException;

import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;

import Int.ecb.gesmescb.*;
import Int.ecb.gesmescb.time.RegularisableInterval;
import Int.ecb.gesmescb.metabase.Metabase;

import Int.ecb.gesmescb.event.TestEventSource;

/* ***************************** Class ****************************** */

/**
 * Implements a few tests for the {@link EaacsLabConfig}.
 *
 * @author
 * <a href="mailto:sis.external@ecb.int">Bernhard Bodenstorfer</a>
 *
 * @version @(#)$Id: EaacsLabConfigTest.java,v 1.1 2017/01/19 16:14:04 trainot Exp $
 */
public class EaacsLabConfigTest extends TestCase {

	/* ======================= Constants ======================== */

	private static final DateFormat ISO_FORMAT
		= new SimpleDateFormat( "yyyy-MM-dd'T'HH:mm:ss" );

	private static final String objectName
	= EaacsLabConfig.class.getName() + ".Nuno";

	private static final String[][] testProperties = {
		{ "begin", "2005-03-11T19:14:12" },
		{ "end", "2007-02-28T05:32:44" },
		{ "1.dataSet", "ECB_IEAQ" },
		{ "1.dataSet", "ECB_IEAQ" },
		{ "1.maintenanceAgency", "ECB" },
		{ "1.keyFamily", "ECB_IEA1" },
		{ "1.dimension.FREQ.accept", "A" },
		{ "1.dimension.REF_AREA.translate", "EE,U2" },
		{ "2.dataSet", "ECB_IEAA" },
		{ "2.maintenanceAgency", "ECB" },
		{ "2.keyFamily", "ECB_IEA1" },
		{ "2.dimension.FREQ.translate", "D" },
		{ "2.dimension.REF_AREA.translate", "EE" },
		{ "3.dataSet", "ECB_IEAS" },
		{ "3.maintenanceAgency", "ECB" },
		{ "3.keyFamily", "ECB_IEA1" },
		{ "3.dimension.FREQ.accept", "*" },
		{ "3.dimension.REF_AREA.translate", "*" },
		{ "4.dataSet", "ECB_MUFI" },
		{ "4.maintenanceAgency", "ECB" },
		{ "4.keyFamily", "ECB_MUFA" },
		{ "4.dimension.ITEM.accept", "800" }
	};

	public static final MessageBegin STRUCTURE_MESSAGE
	= new MessageBegin(
		"MREF000001", "4F0", "ZZZ", "ECB",
		MessageBegin.STATISTICAL_DEFINITIONS,
		"Set up", new Contact[]{}
	);

	public static final KeyFamily KF_IEA1 = new KeyFamily(
		"ECB_IEA1", "Test KF", "Dimensions only", new Date( 0 ),
		new KeyFamilyComponent[]{
			new KeyFamilyComponent(
				"FREQ",
				KeyFamilyComponent.KEY_STRUCTURE_DIMENSION,
				"1", "AN..1", "", "", "CL_FREQ"
			),
			new KeyFamilyComponent(
				"REF_AREA",
				KeyFamilyComponent.KEY_STRUCTURE_DIMENSION,
				"2", "AN1", "", "", "CL_AREA_EE"
			),
			new KeyFamilyComponent(
				"ITEM",
				KeyFamilyComponent.KEY_STRUCTURE_DIMENSION,
				"3", "AN2", "", "", "CL_ITEM"
			)
		}
	),
	KF_MUFA = new KeyFamily(
		"ECB_MUFA", "Test KF2", "Dimensions only", new Date( 0 ),
		new KeyFamilyComponent[]{
			new KeyFamilyComponent(
				"FREQ",
				KeyFamilyComponent.KEY_STRUCTURE_DIMENSION,
				"1", "AN..1", "", "", "CL_FREQ"
			),
			new KeyFamilyComponent(
				"ITEM",
				KeyFamilyComponent.KEY_STRUCTURE_DIMENSION,
				"2", "AN1", "", "", "CL_AREA_EE"
			)
		}
	);

	public static final MessageEnd MESSAGE_END = new MessageEnd(
		"MREF000001"
	);

	/* ====================== Constructors ====================== */

	public EaacsLabConfigTest ( final String name ) {
		super( name );
	}

	/* ================== Protected Methods ===================== */

	/** Sets up the test fixture. */
	protected void setUp () {
		for ( int i = 0; i < testProperties.length; ++i ) {
			System.setProperty(
				objectName + "." + testProperties[i][0],
				testProperties[i][1]
			);
		}

		final TestEventSource source = new TestEventSource();
		final Metabase metabase = Metabase.getMetabase();
		source.addListener( metabase, true );
		source.dispatch( STRUCTURE_MESSAGE );
		source.dispatch( KF_IEA1 );
		source.dispatch( KF_MUFA );
		source.dispatch( MESSAGE_END );
	}

	/** Tears down the test fixture. */
	protected void tearDown () {
		for ( int i = 0; i < testProperties.length; ++i ) {
			System.getProperties().remove(
				objectName + "." + testProperties[i][0]
			);
		}
	}

	/* ===================== Public Methods ===================== */

	/**
	 * Assembles and returns a test suite for all
	 * the test methods of this test case.
	 *
	 * @return a non-null test suite
	 */
	public static Test suite () {
		return new TestSuite( EaacsLabConfigTest.class );
	}

	/**
	 * @param arg the array of argument {@link String}s
	 */
	public static void main ( final String[] arg ) {
		final String[] testCaseName = {
			EaacsLabConfigTest.class.getName()
		};
		junit.textui.TestRunner.main( testCaseName );
	}

	/* ---------------------- Test Methods ---------------------- */

	/**
	 * Test {@link EaacsLabConfig}.
	 */
	public void testProperties () throws ParseException {
		final EaacsLabConfig config = new EaacsLabConfig( "Nuno" );

		assertEquals(
			"begin",
			ISO_FORMAT.parse( "2005-03-11T19:14:12" ),
			config.getBegin()
		);
		assertEquals(
			"end",
			ISO_FORMAT.parse( "2007-02-28T05:32:44" ),
			config.getEnd()
		);

		config.configure( "ECB_IEAQ" );
		assertEquals( "DS1", "ECB_IEAQ", config.getDataSetId() );
		assertEquals( "MA1", "ECB", config.getMaintenanceAgencyId() );
		assertEquals( "KF1", "ECB_IEA1", config.getKeyFamilyId() );
		assertEquals( "KF1 obj", KF_IEA1, config.getKeyFamily() );
		assertEquals( "1 dim", 3, config.getDimensionCount() );
		assertTrue( "1 dim 1 unlisted", ! config.isIndexed( 0 ) );
		assertEquals( "1 map 1 size", 1, config.codeCount( 0 ) );
		assertTrue(
			"1 map 1 content 1",
			config.codeAccept( 0, "A" )
		);
		assertTrue(
			"1 map 1 not content",
			! config.codeAccept( 0, "Q" )
		);
		assertTrue( "1 dim 2 listed", config.isIndexed( 1 ) );
		assertTrue( "1 dim 2 non-auto", ! config.isAutomatic( 1 ) );
		assertEquals( "1 map 2 size", 2, config.codeCount( 1 ) );
		assertEquals(
			"1 map 2 content EE",
			new Integer( 1 ),
			config.codeIndex( 1, "EE" )
		);
		assertEquals(
			"1 map 2 content U2",
			new Integer( 2 ),
			config.codeIndex( 1, "U2" )
		);
		assertTrue( "1 dim 3 unlisted", ! config.isIndexed( 2 ) );
		try {
			config.codeCount( 2 );
			fail( "1 map 3 uninitialised must be recognised" );
		} catch ( final Exception e ) {}
		try {
			config.codeCount( 3 );
			fail( "No more than three dimension maps 1." );
		} catch ( final Exception e ) {}

		config.configure( "ECB_IEAA" );
		assertEquals( "DS2", "ECB_IEAA", config.getDataSetId() );
		assertEquals( "MA2", "ECB", config.getMaintenanceAgencyId() );
		assertEquals( "KF2", "ECB_IEA1", config.getKeyFamilyId() );
		assertEquals( "KF2 obj", KF_IEA1, config.getKeyFamily() );
		assertEquals( "2 dim", 3, config.getDimensionCount() );
		assertTrue( "2 dim 1 listed", config.isIndexed( 0 ) );
		assertTrue( "2 dim 1 non-automatic", ! config.isAutomatic( 0 ) );
		assertEquals( "2 map 1 size", 1, config.codeCount( 0 ) );
		assertEquals(
			"2 map 1 content",
			new Integer( 1 ),
			config.codeIndex( 0, "D" )
		);
		assertTrue( "2 dim 2 listed", config.isIndexed( 1 ) );
		assertEquals( "2 map 2 size", 1, config.codeCount( 1 ) );
		assertEquals(
			"2 map 2 content EE",
			new Integer( 1 ),
			config.codeIndex( 1, "EE" )
		);
		assertTrue( "2 dim 3 unlisted", ! config.isIndexed( 2 ) );
		try {
			config.codeCount( 2 );
			fail( "2 map 3 uninitialised must be recognised" );
		} catch ( final Exception e ) {}
		try {
			config.codeCount( 3 );
			fail( "No more than three dimension maps." );
		} catch ( final Exception e ) {}

		config.configure( "ECB_IEAS" );
		assertEquals( "DS3", "ECB_IEAS", config.getDataSetId() );
		assertEquals( "MA3", "ECB", config.getMaintenanceAgencyId() );
		assertEquals( "KF3", "ECB_IEA1", config.getKeyFamilyId() );
		assertEquals( "KF3 obj", KF_IEA1, config.getKeyFamily() );
		assertEquals( "3 dim", 3, config.getDimensionCount() );
		assertTrue( "3 dim 1 unlisted", ! config.isIndexed( 0 ) );
		assertEquals( "3 map 1 size", 0, config.codeCount( 0 ) );
		assertTrue(
			"3 map 1 content 1",
			config.codeAccept( 0, "A" )
		);
		assertEquals( "3 map 1 size again", 1, config.codeCount( 0 ) );
		assertTrue(
			"3 map 1 not content",
			! config.codeAccept( 0, "Q" )
		);
		assertEquals( "3 map 1 size yet again", 1, config.codeCount( 0 ) );
		assertTrue( "3 dim 2 listed", config.isIndexed( 1 ) );
		assertTrue( "3 dim 2 automatic", config.isAutomatic( 1 ) );
		assertEquals( "3 map 2 size", 0, config.codeCount( 1 ) );
		assertEquals(
			"3 map 2 content EE",
			new Integer( 1 ),
			config.codeIndex( 1, "EE" )
		);
		assertEquals( "3 map 2 size again", 1, config.codeCount( 1 ) );
		assertEquals(
			"3 map 2 content U2",
			new Integer( 2 ),
			config.codeIndex( 1, "U2" )
		);
		assertEquals( "3 map 2 size yet again", 2, config.codeCount( 1 ) );
		assertEquals(
			"3 map 2 content EE again",
			new Integer( 1 ),
			config.codeIndex( 1, "EE" )
		);
		assertEquals( "3 map 2 size yet yet again", 2, config.codeCount( 1 ) );
		assertTrue( "3 dim 3 unlisted", ! config.isIndexed( 2 ) );
		try {
			config.codeCount( 2 );
			fail( "3 map 3 uninitialised must be recognised" );
		} catch ( final Exception e ) {}
		try {
			config.codeCount( 3 );
			fail( "No more than three dimension maps 1." );
		} catch ( final Exception e ) {}

		config.configure( "ECB_MUFI" );
		assertEquals( "DS4", "ECB_MUFI", config.getDataSetId() );
		assertEquals( "MA4", "ECB", config.getMaintenanceAgencyId() );
		assertEquals( "KF4", "ECB_MUFA", config.getKeyFamilyId() );
		assertEquals( "KF4 obj", KF_MUFA, config.getKeyFamily() );
		assertEquals( "4 dim", 2, config.getDimensionCount() );
		assertTrue( "4 dim 1 unlisted", ! config.isIndexed( 0 ) );
		try {
			config.codeCount( 0 );
			fail( "4 map 1 uninitialised must be recognised" );
		} catch ( final Exception e ) {}
		assertTrue( "4 dim 2 unlisted", ! config.isIndexed( 1 ) );
		assertEquals( "4 map 2 size", 1, config.codeCount( 1 ) );
		try {
			config.codeCount( 2 );
			fail( "4 no more than two dimension maps." );
		} catch ( final Exception e ) {}
	}

	/** Test
	 * {@link EaacsLabConfig#getIntervalByFrequency( int )}
	 * and {@link EaacsLabConfig#getIntervalBySize( int )}.
	 */
	public void testIntervalMaps () throws ParseException {
		final Properties properties = System.getProperties();

		{
			final EaacsLabConfig config = new EaacsLabConfig( "Nuno" );
			assertNotNull(
				"Some maps with begin and end",
				config.getPossibleIntervals()
			);
		}

		properties.remove( objectName + ".begin" );
		{
			final EaacsLabConfig config = new EaacsLabConfig( "Nuno" );
			assertNull(
				"No maps without begin",
				config.getPossibleIntervals()
			);
		}
		properties.put( objectName + ".begin", "2001-01-01T00:00:00" );

		properties.remove( objectName + ".end" );
		{
			final EaacsLabConfig config = new EaacsLabConfig( "Nuno" );
			assertNull(
				"No maps without end",
				config.getPossibleIntervals()
			);
		}
		properties.put( objectName + ".end", "2007-01-01T00:00:00" );

		{
			final Date
				begin = ISO_FORMAT.parse( "2001-01-01T00:00:00" ),
				end = ISO_FORMAT.parse( "2007-01-01T00:00:00" );
			final EaacsLabConfig config = new EaacsLabConfig( "Nuno" );
			assertEquals(
				"The maps",
				new RegularisableInterval( begin, end ),
				config.getPossibleIntervals()
			);
		}
	}
}
