// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2007, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.csv;

import java.io.File;
import java.io.Reader;
import java.io.FileReader;
import java.io.IOException;

import java.util.Date;
import java.util.List;
import java.util.ArrayList;
import java.util.Iterator;

import java.text.SimpleDateFormat;
import java.text.NumberFormat;
import java.text.DecimalFormat;
import java.text.ParseException;

import Int.ecb.logging.Level;

import Int.ecb.gesmescb.InterchangeBegin;
import Int.ecb.gesmescb.InterchangeEnd;
import Int.ecb.gesmescb.MessageBegin;
import Int.ecb.gesmescb.MessageEnd;
import Int.ecb.gesmescb.Contact;
import Int.ecb.gesmescb.KeyFamilyComponent;
import Int.ecb.gesmescb.TimeSeries;
import Int.ecb.gesmescb.Observation;

import Int.ecb.gesmescb.event.ErrorEvent;
import Int.ecb.gesmescb.event.GESMESCBEventSourceAdapter;
import Int.ecb.gesmescb.event.GESMESCBReader;

import Int.ecb.gesmescb.util.Named;
import Int.ecb.gesmescb.util.ComponentHelper;
import Int.ecb.gesmescb.util.KeyTokenizer;

import Int.ecb.gesmescb.time.RegularInterval;
import Int.ecb.gesmescb.time.GregorianRegularInterval;

import Int.ecb.gesmescb.xml.XMLDateFormat;
import Int.ecb.gesmescb.edifact.RegularIntervalFormat;

/* ***************************** Class ****************************** */

/**
 * <p>Reads {@link TimeSeries} from a EAACS project style flat file
 * from Matlab.</p>
 *
 * <center><table border="3" cellpadding="3">
 * <caption>
 * <a name="prefs">Available {@link Int.ecb.prefs preference} settings</a>
 * </caption>
 * <tr>
 * <th>preference name</th>
 * <th>type</th>
 * <th>meaning</th>
 * <th>default</th>
 * </tr>
 * <tr><td><code>source</td><td>string</td><td>input file path</td></tr>
 * </table></center>
 *
 * @see EaacsLabConfig for shared configuration
 *
 * @author
 * <a href="mailto:sis.external@ecb.int">Bernhard Bodenstorfer</a>
 *
 * @version @(#)$Id: EaacsLabReader.java,v 1.1 2017/01/19 16:14:04 trainot Exp $
 */
public class EaacsLabReader extends GESMESCBEventSourceAdapter
	implements Named, GESMESCBReader
{
	/* ======================= Constants ======================== */

	private final Observation
		MISSING_OBSERVATION_M = new Observation(
			Observation.MISSING_VALUE, "M", "N", ""
		),
		MISSING_OBSERVATION_L = new Observation(
			Observation.MISSING_VALUE, "L", "N", ""
		);

	/* ========================= Fields ========================= */

	/** Takes care of logging and preferences. */
	protected final ComponentHelper componentHelper;

	/** Takes care of shared preferences. */
	private final EaacsLabConfig eaacsLabConfig;

	/** EDIFACT interval format for reporting period. */
	private final RegularIntervalFormat edifactIntervalFormat;

	/**
	 * The filename of the file to read from
	 * if the <code>EaacsLabReader</code> has been
	 * constructed from a filename.
	 * Otherwise it is null.
	 */
	private final String filename;

	/** The <code>Reader</code> for reading from the CSV-input file. */
	private final Reader reader;

	/* ====================== Constructors ====================== */

	/**
	 * Construct a <code>EaacsLabReader</code> to read from a file.
	 *
	 * @param name used for logging and preferences.
	 * @param filename where to read from.
	 *
	 * @throws NullPointerException if <code>filename==null</code>
	 * and the preference variable
	 * <code>source</code> is undefined.
	 * @throws IOException if the file cannot be opened for reading.
	 */
	public EaacsLabReader ( final String name, final String filename )
		throws IOException
	{
		componentHelper = new ComponentHelper( name, this );
		eaacsLabConfig = new EaacsLabConfig( name );
		edifactIntervalFormat = new RegularIntervalFormat();
		this.filename = filename != null
			? filename
			: componentHelper.getPreference( "source", null );
		if ( this.filename == null ) {
			final String desc
			= "No preference variable \"source\" found"
				+ " to determine the input file.";
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"EaacsLabReader()",
				desc
			);
			throw new NullPointerException( desc );
		}
		reader = new FileReader( this.filename );
	}

	/**
	 * Construct a <code>EaacsLabReader</code>
	 * reading from a file according to preferences.
	 *
	 * @param name used for logging and preferences.
	 *
	 * @throws NullPointerException if the preference variable
	 * <code>source</code> is undefined.
	 * @throws IOException if the file cannot be opened for reading.
	 */
	public EaacsLabReader ( final String name )
		throws IOException
	{
		this( name, null );
	}

	/* ===================== Public Methods ===================== */

	public String getName () {
		return componentHelper.name;
	}

	/**
	 * Parse a GESMES/TS interchange in CSV syntax used by the EAACS project
	 * into {@link Int.ecb.gesmescb.GESMESCBObject}s.
	 * These are then dispatched to all listeners.
	 */
	public void run () {
		final NumberFormat mrefFormat = new DecimalFormat( "000000" );

		try {
			componentHelper.logger.log(
				Level.INFO,
				componentHelper.logName,
				"run()",
				"Starting to parse EAACS/Matlab input from \""
				+ filename
				+ "\"."
			);

			final String iref
			= componentHelper.getPreference( "iref", "IREF000000" );
			final String sender
			= componentHelper.getPreference( "sender", "ZZZ" );
			final String receiver
			= componentHelper.getPreference( "receiver", "ZZZ" );
			Date prepared;
			{
				final String preparedString = componentHelper.getPreference(
					"prepared", null
				);
				if ( null == preparedString ) {
					prepared = new Date();
				} else try {
					prepared = new XMLDateFormat().parse( preparedString );
				} catch ( final ParseException pe ) {
					componentHelper.logger.log(
						Level.SEVERE,
						componentHelper.logName,
						"run()",
						"Configured preparation timestamp \""
						+ preparedString
						+ "\" is not in required ISO dateTime format.",
						pe
					);
					prepared = null;
				}
			}
			final boolean test
			= ! "false".equalsIgnoreCase(
				componentHelper.getPreference( "test", "false" )
			);
			final InterchangeBegin interchangeBegin = new InterchangeBegin(
				iref, sender, receiver,
				null != prepared
					? new SimpleDateFormat( "yyMMdd:HHmm" ).format( prepared )
					: "",
				test
			);
			componentHelper.logger.log(
				Level.FINE,
				componentHelper.logName,
				"run()",
				"Dispatching InterchangeBegin",
				interchangeBegin
			);
			dispatch( interchangeBegin );

			final File sourceFile = new File( filename );
			final String subject;
			{
				final String subjectString = componentHelper.getPreference(
					"subject", sourceFile.getName()
				);
				if ( 35 < subjectString.length() ) {
					subject = subjectString.substring( 0, 35 );
					componentHelper.logger.log(
						Level.WARNING,
						componentHelper.logName,
						"run()",
						"Subject string \""
						+ subjectString
						+ "\" has been truncated to 35 characters.",
						subject
					);
				} else {
					subject = subjectString;
				}
			}

			Date extracted;
			{
				final String extractedString = componentHelper.getPreference(
					"extracted", null
				);
				if ( null == extractedString ) {
					extracted = new Date( sourceFile.lastModified() );
				} else try {
					extracted = new XMLDateFormat().parse( extractedString );
				} catch ( final ParseException pe ) {
					componentHelper.logger.log(
						Level.SEVERE,
						componentHelper.logName,
						"run()",
						"Configured extraction timestamp \""
						+ extractedString
						+ "\" is not in required ISO dateTime format.",
						pe
					);
					extracted = null;
				}
			}

			final Date reportingBegin = eaacsLabConfig.getBegin();
			final Date reportingEnd = eaacsLabConfig.getEnd();
			final RegularInterval reportingInterval;
			if ( reportingBegin != null && reportingEnd != null ) {
				reportingInterval = GregorianRegularInterval.searchCover(
					reportingBegin, reportingEnd, true
				);
			} else {
				reportingInterval = null;
			}
			final String reportingPeriod, reportingPeriodFormat;
			if ( reportingInterval != null ) {
				final String reportingPaF = edifactIntervalFormat.format(
					reportingInterval
				);
				final int separator = reportingPaF.indexOf( ':' );
				reportingPeriod = reportingPaF.substring( 0, separator );
				reportingPeriodFormat = reportingPaF.substring( separator + 1 );
			} else {
				reportingPeriod = reportingPeriodFormat = "";
			}

			int messageCounter = 0;
			for (
				final CSVTokenizer tokenizer = new CSVTokenizer( reader );
				tokenizer.hasMoreLines();
			) {
				final CSVTokenizer.Line line = tokenizer.nextLine();
				final String dsiPrefixedKey = line.nextCell();
				final int dotPosition = dsiPrefixedKey.indexOf(
					TimeSeries.KEY_SEPARATOR
				);
				final String dataSetId = dsiPrefixedKey.substring(
					0, dotPosition
				);

				if (
					null == eaacsLabConfig.getDataSetId()
					||
					! dataSetId.equals( eaacsLabConfig.getDataSetId() )
				) {
					if ( null != eaacsLabConfig.getDataSetId() ) {
						final MessageEnd messageEnd = new MessageEnd(
							"MREF" + mrefFormat.format( messageCounter )
						);
						componentHelper.logger.log(
							Level.FINE,
							componentHelper.logName,
							"run()",
							"Dispatching MessageEnd",
							messageEnd
						);
						dispatch( messageEnd );
					}

					eaacsLabConfig.configure( dataSetId );

					final MessageBegin messageBegin = new MessageBegin(
						"MREF" + mrefFormat.format( ++messageCounter ),
						sender, receiver,
						eaacsLabConfig.getMaintenanceAgencyId(),
						MessageBegin.STATISTICAL_DATA,
						subject,
						new Contact[]{},
						dataSetId,
						MessageBegin.UPDATE_OR_REPLACE_ACTION,
						null != extracted
						? new SimpleDateFormat( "yyyyMMddHHmm" ).format(
							extracted
						) : "",
						reportingPeriod, reportingPeriodFormat,
						eaacsLabConfig.getKeyFamilyId()
					);
					componentHelper.logger.log(
						Level.FINE,
						componentHelper.logName,
						"run()",
						"Dispatching MessageBegin",
						messageBegin
					);
					dispatch( messageBegin );
				}

				final String key = dsiPrefixedKey.substring(
					dotPosition + 1
				);

				final Iterator
					dimensions = eaacsLabConfig.getKeyFamily()
						.keyStructureIterator(),
					dimensionValues = new KeyTokenizer( key );
				for (
					int i = 0;
					dimensions.hasNext()
					&&
					dimensionValues.hasNext();
					++i
				) {
					final KeyFamilyComponent dimension
						= (KeyFamilyComponent) dimensions.next();
					final String coordinate = (String) dimensionValues.next();
					if ( eaacsLabConfig.isIndexed( i ) ) {
						final String read = line.nextCell();
						final Integer translated
						= eaacsLabConfig.codeIndex( i, coordinate );
						if ( null == translated ) {
							componentHelper.logger.log(
								Level.SEVERE,
								componentHelper.logName,
								"run()",
								"Dimension "
								+ dimension.getConceptId()
								+ " code \""
								+ coordinate
								+ "\" is not in the configured list."
							);
						} else if (
							! eaacsLabConfig.isAutomatic( i )
							&&
							! translated.equals( read )
						) {
							componentHelper.logger.log(
								Level.SEVERE,
								componentHelper.logName,
								"run()",
								"Dimension "
								+ dimension.getConceptId()
								+ " code \""
								+ coordinate
								+ "\" is configured as index "
								+ translated
								+ ", but read \""
								+ read + "\"."
							);
						}
					} else if (
						! eaacsLabConfig.codeAccept( i, coordinate )
					) {
						componentHelper.logger.log(
							Level.WARNING,
							componentHelper.logName,
							"run()",
							"Dimension code \""
							+ coordinate
							+ "\" not in configured list."
						);
					}
				}

				final List observations = new ArrayList();
				while ( line.hasMoreCells() ) {
					final String value = line.nextCell();
					final Observation observation
					= "".equals( value ) || "ND".equals( value )
						? Observation.EMPTY_OBSERVATION
						: "NA".equals( value )
						? MISSING_OBSERVATION_M
						: "NC".equals( value )
						? MISSING_OBSERVATION_L
						: new Observation( value, "A", "N", "" );
					observations.add( observation );
				}
				final RegularInterval interval
				= eaacsLabConfig.getPossibleIntervals().getIntervalBySize(
					observations.size()
				);
				if ( null == interval ) {
					componentHelper.logger.log(
						Level.SEVERE,
						componentHelper.logName,
						"run()",
						"Time series " + key + " is given with "
						+ observations.size()
						+ " observations, but this number does not fit"
						+ " the interval from " + reportingBegin
						+ " to " + reportingEnd
						+ " with any known frequency.",
						observations
					);
				} else {
					final TimeSeries timeSeries = new TimeSeries(
						key, interval, observations
					);
					componentHelper.logger.log(
						Level.FINE,
						componentHelper.logName,
						"run()",
						"Dispatching TimeSeries " + key,
						timeSeries
					);
					dispatch( timeSeries );
				}
			}

			if ( null != eaacsLabConfig.getDataSetId() ) {
				final MessageEnd messageEnd = new MessageEnd(
					"MREF" + mrefFormat.format( messageCounter )
				);
				componentHelper.logger.log(
					Level.FINE,
					componentHelper.logName,
					"run()",
					"Dispatching MessageEnd",
					messageEnd
				);
				dispatch( messageEnd );
			}

			final InterchangeEnd interchangeEnd = new InterchangeEnd(
				iref
			);
			componentHelper.logger.log(
				Level.FINE,
				componentHelper.logName,
				"run()",
				"Dispatching InterchangeEnd",
				interchangeEnd
			);
			dispatch( interchangeEnd );

			componentHelper.logger.log(
				Level.INFO,
				componentHelper.logName,
				"run()",
				"Ending to parse EAACS/Matlab input from \""
				+ filename
				+ "\"."
			);
		}
		catch ( final Exception e ) {
			final String desc = "Problem: " + e;

			componentHelper.logger.log(
				Level.SEVERE,
				getName(),
				"run()",
				desc,
				e
			);
			dispatch(
				new ErrorEvent( desc, ErrorEvent.ERROR )
			);
		}
	}

	/**
	 * Close the <code>EaacsLabReader</code>.
	 */
	public void close () {
		try {
			componentHelper.logger.log(
				Level.FINE,
				componentHelper.logName,
				"close()",
				"Closing EAACS/Matlab input file \""
				+ filename
				+ "\"."
			);
			reader.close();
		}
		catch ( final IOException ioe ) {
			final String desc = "Close failed on: " + ioe.getMessage() + ".";
			componentHelper.logger.log(
				Level.SEVERE,
				getName(),
				"close()",
				desc,
				ioe
			);
			dispatch(
				new ErrorEvent( desc, ErrorEvent.ERROR )
			);
		}
	}
}
