// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2007, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.csv;

import java.io.File;
import java.io.Writer;
import java.io.FileWriter;
import java.io.IOException;

import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;

import Int.ecb.gesmescb.*;
import Int.ecb.gesmescb.metabase.Metabase;

import Int.ecb.gesmescb.event.TestEventSource;
import Int.ecb.gesmescb.event.TestEventListener;

/* ***************************** Class ****************************** */

/**
 * Implements a few tests for the {@link EaacsLabReader}.
 *
 * @author
 * <a href="mailto:sis.external@ecb.int">Bernhard Bodenstorfer</a>
 *
 * @version @(#)$Id: EaacsLabReaderTest.java,v 1.1 2017/01/19 16:14:04 trainot Exp $
 */
public class EaacsLabReaderTest extends TestCase {

	/* ======================= Constants ======================== */

	private static final String
		objectName = EaacsLabReader.class.getName() + ".Clara",
		sharedName = EaacsLabConfig.class.getName() + ".Clara";

	private static final String[][]
		objectProperties = {
			{ "iref", "IREF000035" },
			{ "prepared", "2000-12-01T14:10:00" },
			{ "test", "true" },
			{ "subject", "eaacs-test" },
			{ "extracted", "2002-03-14T00:15:00" }
		},
		sharedProperties = {
			{ "begin", "2006-01-01T00:00:00" },
			{ "end", "2010-01-01T00:00:00" },
			{ "1.dataSet", "ECB_IEAQ" },
			{ "1.maintenanceAgency", "ECB" },
			{ "1.keyFamily", "ECB_IEA1" },
			{ "1.dimension.FREQ.accept", "A" },
			{ "1.dimension.REF_AREA.translate", "EE,U2" },
			{ "1.dimension.ITEM.accept", "*" },
			{ "2.dataSet", "ECB_MUFI" },
			{ "2.maintenanceAgency", "ECB" },
			{ "2.keyFamily", "ECB_MUFA" },
			{ "2.dimension.FREQ.translate", "A,Q" },
			{ "2.dimension.ITEM.translate", "800,900" }
		};

	/** The input to the {@link EaacsLabReader} under Test. */
	private static final String testInput
	= "ECB_IEAQ.A.U2.C,2,3.6,13.6,23.6,33.6\n"
	+ "ECB_IEAQ.A.EE.C,1,4.6,14.6,,-3.6\n"
	+ "ECB_MUFI.A.900,1,2,2.2,,,2.6\n"
	+ "ECB_MUFI.Q.800,2,1,6.2,,NC,6.6,NA,ND,,8,1,,,,,,7,8\n";

	/**
	 * The objects into which
	 * the {@link EaacsLabReader} should translate
	 * {@link #testInput} when called in the correct way.
	 */
	private static final Object[] testOutput = {
		new InterchangeBegin(
			"IREF000035", "ZZZ", "ZZZ", "001201:1410", true
		),
		new MessageBegin(
			"MREF000001", "ZZZ", "ZZZ", "ECB",
			MessageBegin.STATISTICAL_DATA, "eaacs-test",
			new Contact[]{},
			"ECB_IEAQ", MessageBegin.UPDATE_OR_REPLACE_ACTION,
			"200203140015", "20062009", "702", "ECB_IEA1"
		),
		new TimeSeries(
			"A.U2.C", "20062009", "702",
			new Observation[] {
				new Observation( "3.6", "A", "N", "" ),
				new Observation( "13.6", "A", "N", "" ),
				new Observation( "23.6", "A", "N", "" ),
				new Observation( "33.6", "A", "N", "" )
			}
		),
		new TimeSeries(
			"A.EE.C", "20062009", "702",
			new Observation[] {
				new Observation( "4.6", "A", "N", "" ),
				new Observation( "14.6", "A", "N", "" ),
				Observation.EMPTY_OBSERVATION,
				new Observation( "-3.6", "A", "N", "" )
			}
		),
		new MessageEnd( "MREF000001" ),
		new MessageBegin(
			"MREF000002", "ZZZ", "ZZZ", "ECB",
			MessageBegin.STATISTICAL_DATA, "eaacs-test",
			new Contact[]{},
			"ECB_MUFI", MessageBegin.UPDATE_OR_REPLACE_ACTION,
			"200203140015", "20062009", "702", "ECB_MUFA"
		),
		new TimeSeries(
			"A.900", "20062009", "702",
			new Observation[] {
				new Observation( "2.2", "A", "N", "" ),
				Observation.EMPTY_OBSERVATION,
				Observation.EMPTY_OBSERVATION,
				new Observation( "2.6", "A", "N", "" )
			}
		),
		new TimeSeries(
			"Q.800", "2006120094", "708",
			new Observation[] {
				new Observation( "6.2", "A", "N", "" ),
				Observation.EMPTY_OBSERVATION,
				new Observation( Observation.MISSING_VALUE, "L", "N", "" ),
				new Observation( "6.6", "A", "N", "" ),
				new Observation( Observation.MISSING_VALUE, "M", "N", "" ),
				Observation.EMPTY_OBSERVATION,
				Observation.EMPTY_OBSERVATION,
				new Observation( "8", "A", "N", "" ),
				new Observation( "1", "A", "N", "" ),
				Observation.EMPTY_OBSERVATION,
				Observation.EMPTY_OBSERVATION,
				Observation.EMPTY_OBSERVATION,
				Observation.EMPTY_OBSERVATION,
				Observation.EMPTY_OBSERVATION,
				new Observation( "7", "A", "N", "" ),
				new Observation( "8", "A", "N", "" )
			}
		),
		new MessageEnd( "MREF000002" ),
		new InterchangeEnd( "IREF000035" )
	};

	/* ========================= Fields ========================= */

	/** The file used for the location of test data. */
	private File testFile;

	/* ====================== Constructors ====================== */

	public EaacsLabReaderTest ( final String name ) {
		super( name );
	}

	/* ================== Protected Methods ===================== */

	/** Sets up the test fixture. */
	protected void setUp () throws IOException {
		testFile = File.createTempFile( "eaacs_testinput.", null );
		testFile.deleteOnExit();
		final Writer writer = new FileWriter( testFile );
		writer.write( testInput );
		writer.close();
		System.setProperty(
			objectName + ".source",
			testFile.getAbsolutePath()
		);

		for ( int i = 0; i < sharedProperties.length; ++i ) {
			System.setProperty(
				sharedName + "." + sharedProperties[i][0],
				sharedProperties[i][1]
			);
		}
		for ( int i = 0; i < objectProperties.length; ++i ) {
			System.setProperty(
				objectName + "." + objectProperties[i][0],
				objectProperties[i][1]
			);
		}

		final TestEventSource source = new TestEventSource();
		final Metabase metabase = Metabase.getMetabase();
		source.addListener( metabase, true );
		source.dispatch( EaacsLabConfigTest.STRUCTURE_MESSAGE );
		source.dispatch( EaacsLabConfigTest.KF_IEA1 );
		source.dispatch( EaacsLabConfigTest.KF_MUFA );
		source.dispatch( EaacsLabConfigTest.MESSAGE_END );
	}

	/** Tears down the test fixture. */
	protected void tearDown () {
		System.getProperties().remove( objectName + ".source" );
		testFile.delete();
		for ( int i = 0; i < sharedProperties.length; ++i ) {
			System.getProperties().remove(
				sharedName + "." + sharedProperties[i][0]
			);
		}
		for ( int i = 0; i < objectProperties.length; ++i ) {
			System.getProperties().remove(
				objectName + "." + objectProperties[i][0]
			);
		}
	}

	/* ===================== Public Methods ===================== */

	/**
	 * Assembles and returns a test suite for all
	 * the test methods of this test case.
	 *
	 * @return a non-null test suite
	 */
	public static Test suite () {
		return new TestSuite( EaacsLabReaderTest.class );
	}

	/**
	 * @param arg the array of argument {@link String}s
	 */
	public static void main ( final String[] arg ) {
		final String[] testCaseName = {
			EaacsLabReaderTest.class.getName()
		};
		junit.textui.TestRunner.main( testCaseName );
	}

	/* ---------------------- Test Methods ---------------------- */

	/** Test {@link EaacsLabReader#run}. */
	public void testRunFromFile () throws IOException {
		final EaacsLabReader reader = new EaacsLabReader( "Clara" );
		final TestEventListener listener = new TestEventListener( testOutput );
		reader.addListener( listener, true );
		reader.run();
		reader.close();
		listener.close();
	}
}
