// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2007, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.csv;

import java.io.Writer;
import java.io.FileWriter;
import java.io.BufferedWriter;
import java.io.IOException;

import java.util.Date;
import java.util.Iterator;

import Int.ecb.logging.Level;

import Int.ecb.gesmescb.InterchangeBegin;
import Int.ecb.gesmescb.InterchangeEnd;
import Int.ecb.gesmescb.MessageBegin;
import Int.ecb.gesmescb.MessageEnd;
import Int.ecb.gesmescb.KeyFamilyComponent;
import Int.ecb.gesmescb.TimeSeries;
import Int.ecb.gesmescb.Observation;
import Int.ecb.gesmescb.Attribute;

import Int.ecb.gesmescb.time.Interval;
import Int.ecb.gesmescb.time.RegularInterval;
import Int.ecb.gesmescb.time.RegularisableInterval;

import Int.ecb.gesmescb.event.DataListener;
import Int.ecb.gesmescb.event.GESMESCBEventListenerAdapter;
import Int.ecb.gesmescb.event.GESMESCBWriter;

import Int.ecb.gesmescb.util.Named;
import Int.ecb.gesmescb.util.ComponentHelper;
import Int.ecb.gesmescb.util.KeyTokenizer;
import Int.ecb.gesmescb.util.ObservationMap;

/* ***************************** Class ****************************** */

/**
 * <p>Writes {@link TimeSeries} to a EAACS project style flat file
 * for Matlab.</p>
 *
 * <center><table border="3" cellpadding="3">
 * <caption>
 * <a name="prefs">Available {@link Int.ecb.prefs preference} settings</a>
 * </caption>
 * <tr>
 * <th>preference name</th>
 * <th>type</th>
 * <th>meaning</th>
 * <th>default</th>
 * </tr>
 * <tr><td><code>destination</td><td>string</td><td>output file path</td></tr>
 * </table></center>
 *
 * @see EaacsLabConfig for shared configuration
 *
 * @author
 * <a href="mailto:sis.external@ecb.int">Bernhard Bodenstorfer</a>
 *
 * @version @(#)$Id: EaacsLabWriter.java,v 1.1 2017/01/19 16:14:04 trainot Exp $
 */
public class EaacsLabWriter extends GESMESCBEventListenerAdapter
	implements Named, DataListener, GESMESCBWriter
{
	/* ======================= Constants ======================== */

	protected static final ValueRepresentationStrategy
		FAME_STRATEGY = new FameStyleStrategy(),
		EMPTY_STRATEGY = new EmptyTemplateStrategy();

	/* ========================= Fields ========================= */

	/** Takes care of logging and preferences. */
	protected final ComponentHelper componentHelper;

	/** Takes care of shared preferences. */
	private final EaacsLabConfig eaacsLabConfig;

	/** How to represent empty observations and missing values. */
	private final ValueRepresentationStrategy valueRepresentationStrategy;

	/**
	 * The filename of the file to write to
	 * if the <code>EaacsLabWriter</code> has been
	 * constructed from a filename.
	 * Otherwise it is null.
	 */
	private final String filename;

	/** The <code>Writer</code> for writing to the CSV-output file. */
	private final Writer writer;

	/** Whether current is a deletion message. */
	private boolean isDeleteMessage;

	/* ====================== Constructors ====================== */

	/**
	 * Construct a <code>EaacsLabWriter</code> to write to a file.
	 *
	 * @param name used for logging and preferences.
	 * @param filename where to write to.
	 *
	 * @throws NullPointerException if <code>filename==null</code>
	 * and the preference variable
	 * <code>destination</code> is undefined.
	 * @throws IOException if the file cannot be opened for writing.
	 */
	protected EaacsLabWriter (
		final String name,
		final String filename,
		final ValueRepresentationStrategy valueRepresentationStrategy
	)
		throws IOException
	{
		componentHelper = new ComponentHelper( name, this );
		eaacsLabConfig = new EaacsLabConfig( name );
		this.valueRepresentationStrategy = valueRepresentationStrategy;

		this.filename = filename != null
			? filename
			: componentHelper.getPreference( "destination", null );
		if ( this.filename == null ) {
			final String desc
			= "No preference variable \"destination\" found"
				+ " to determine the output file.";
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"EaacsLabWriter()",
				desc
			);
			throw new NullPointerException( desc );
		}
		writer = new BufferedWriter( new FileWriter( this.filename ) );
		isDeleteMessage = false;
	}

	/**
	 * Construct a <code>EaacsLabWriter</code> to write to a file.
	 *
	 * @param name used for logging and preferences.
	 * @param filename where to write to.
	 *
	 * @throws NullPointerException if <code>filename==null</code>
	 * and the preference variable
	 * <code>destination</code> is undefined.
	 * @throws IOException if the file cannot be opened for writing.
	 */
	public EaacsLabWriter ( final String name, final String filename )
		throws IOException
	{
		this( name, filename, FAME_STRATEGY );
	}

	/**
	 * Construct a <code>EaacsLabWriter</code>
	 * writing to a file according to preferences.
	 *
	 * @param name used for logging and preferences.
	 *
	 * @throws NullPointerException if the preference variable
	 * <code>destination</code> is undefined.
	 * @throws IOException if the file cannot be opened for writing.
	 */
	public EaacsLabWriter ( final String name )
		throws IOException
	{
		this( name, null );
	}

	/* ===================== Public Methods ===================== */

	public String getName () {
		return componentHelper.name;
	}

	/**
	 * Close the <code>EaacsLabWriter</code>.
	 */
	public void close () {
		try {
			componentHelper.logger.log(
				Level.FINE,
				componentHelper.logName,
				"close()",
				"Closing"
			);
			setActive( false );
			writer.close();
		}
		catch ( final IOException ioe ) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"close()",
				"Closing failed.",
				ioe
			);
		}
	}

	/* ---------------------- Interchange ----------------------- */

	public void handle ( final InterchangeBegin interchangeBegin ) {
		componentHelper.logger.log(
			Level.FINE,
			componentHelper.logName,
			"handle( InterchangeBegin )",
			"Received InterchangeBegin",
			interchangeBegin
		);
	}

	public void handle ( final InterchangeEnd interchangeEnd ) {
		componentHelper.logger.log(
			Level.FINE,
			componentHelper.logName,
			"handle( InterchangeEnd )",
			"Received InterchangeEnd",
			interchangeEnd
		);
	}

	/* ------------------------ Message ------------------------- */

	public void handle ( final MessageBegin messageBegin ) {
		componentHelper.logger.log(
			Level.FINE,
			componentHelper.logName,
			"handle( MessageBegin )",
			"Received MessageBegin",
			messageBegin
		);

		final String dataSetConfig = messageBegin.getDataSetId();
		eaacsLabConfig.configure( dataSetConfig );

		final String maintenanceAgencyId
		= messageBegin.getMaintenanceAgencyId();
		final String maintenanceAgencyConfig
		= eaacsLabConfig.getMaintenanceAgencyId();
		if ( ! maintenanceAgencyId.equals( maintenanceAgencyConfig ) ) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"handle( MessageBegin )",
				"Configuration for data set \""
				+ dataSetConfig
				+ "\" stipulates maintenance agency \""
				+ maintenanceAgencyConfig
				+ "\" but received \""
				+ maintenanceAgencyId
				+ "\".",
				messageBegin
			);
		}

		final String keyFamilyId
		= messageBegin.getKeyFamilyId();
		final String keyFamilyConfig
		= eaacsLabConfig.getKeyFamilyId();
		if ( ! keyFamilyId.equals( keyFamilyConfig ) ) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"handle( MessageBegin )",
				"Configuration for data set \""
				+ dataSetConfig
				+ "\" stipulates key family \""
				+ keyFamilyConfig
				+ "\" but received \""
				+ keyFamilyId
				+ "\".",
				messageBegin
			);
		}

		final Interval reportingPeriod
			= messageBegin.getReportingPeriodAsInterval();
		if ( null != reportingPeriod ) {
			final Date configuredBegin = eaacsLabConfig.getBegin();
			final Date configuredEnd = eaacsLabConfig.getEnd();
			final Date reportingBegin = reportingPeriod.getBegin();
			final Date reportingEnd = reportingPeriod.getEnd();
			if ( ! configuredBegin.equals( reportingBegin ) ) {
				componentHelper.logger.log(
					Level.WARNING,
					componentHelper.logName,
					"handle( MessageBegin )",
					"Reporting period (" + reportingPeriod
					+ ") does not begin at configured begin point "
					+ configuredBegin + ".",
					messageBegin
				);
			}
			if ( ! configuredEnd.equals( reportingEnd ) ) {
				componentHelper.logger.log(
					Level.WARNING,
					componentHelper.logName,
					"handle( MessageBegin )",
					"Reporting period (" + reportingPeriod
					+ ") does not end at configured end point "
					+ configuredBegin + ".",
					messageBegin
				);
			}
		}

		if ( ( isDeleteMessage = messageBegin.isDeleteAction() ) ) {
			componentHelper.logger.log(
				Level.WARNING,
				componentHelper.logName,
				"handle( MessageBegin )",
				"Message \"" + messageBegin.getMref()
				+ "\" is a delete message and will be ignored.",
				messageBegin
			);
		}
	}

	public void handle ( final MessageEnd messageEnd ) {
		componentHelper.logger.log(
			Level.FINE,
			componentHelper.logName,
			"handle( MessageEnd )",
			"Received MessageEnd",
			messageEnd
		);
	}

	/* ---------------------- Time Series ----------------------- */

	public void handle ( final TimeSeries timeSeries ) {
		if ( isDeleteMessage ) {
			componentHelper.logger.log(
				Level.FINER,
				componentHelper.logName,
				"handle( TimeSeries )",
				"Ignoring TimeSeries deletion",
				timeSeries
			);
			return;
		}
		componentHelper.logger.log(
			Level.FINER,
			componentHelper.logName,
			"handle( TimeSeries )",
			"Writing TimeSeries",
			timeSeries
		);

		try {
			final StringBuffer line = new StringBuffer(
				eaacsLabConfig.getDataSetId()
			);
			line.append( TimeSeries.KEY_SEPARATOR );
			line.append( timeSeries.getKey() );
			line.append( ',' );

			final Iterator
				dimensions
					= eaacsLabConfig.getKeyFamily().keyStructureIterator(),
				dimensionValues = new KeyTokenizer( timeSeries.getKey() );
			for (
				int i = 0;
				i < eaacsLabConfig.getDimensionCount()
				&&
				dimensionValues.hasNext();
				++i
			) {
				final KeyFamilyComponent dimension
					= (KeyFamilyComponent) dimensions.next();
				final String value = (String) dimensionValues.next();
				if ( eaacsLabConfig.isIndexed( i ) ) {
					final Integer translated
					= eaacsLabConfig.codeIndex( i, value );
					if ( null != translated ) {
						line.append( translated );
					} else {
						componentHelper.logger.log(
							Level.SEVERE,
							componentHelper.logName,
							"handle( TimeSeries )",
							"Dimension \"" + dimension.getConceptId()
							+ "\" code \"" + value
							+ "\" not in configured list."
						);
					}
					line.append( ',' );
				} else if (
					! eaacsLabConfig.codeAccept( i, value )
				) {
					componentHelper.logger.log(
						Level.SEVERE,
						componentHelper.logName,
						"handle( TimeSeries )",
						"Dimension \"" + dimension.getConceptId()
						+ "\" code \"" + value
						+ "\" not allowed by configuration."
					);
				}
			}

			final RegularisableInterval possibleIntervals
				= eaacsLabConfig.getPossibleIntervals();
			final RegularInterval
				coveredInterval = timeSeries.getInterval(),
				configuredInterval = possibleIntervals.getIntervalByFrequency(
					coveredInterval.getFrequency()
				);
			if ( null == configuredInterval ) {
				componentHelper.logger.log(
					Level.SEVERE,
					componentHelper.logName,
					"handle( TimeSeries )",
					"Time series \"" + timeSeries.getKey()
					+ "\" frequency cannot be aligned with configured interval"
					+ " from " + eaacsLabConfig.getBegin()
					+ " to " + eaacsLabConfig.getEnd() + "."
				);
				return;
			}

			final int
				beginConfiguredVsCovered = coveredInterval.getBegin().compareTo(
					configuredInterval.getBegin()
				),
				endConfiguredVsCovered = coveredInterval.getEnd().compareTo(
					configuredInterval.getEnd()
				);
			if (
				0 < beginConfiguredVsCovered
				||
				0 > endConfiguredVsCovered
			) {
				componentHelper.logger.log(
					Level.FINER,
					componentHelper.logName,
					"handle( TimeSeries )",
					"Time series \"" + timeSeries.getKey()
					+ "\" does not cover configured interval"
					+ " from " + eaacsLabConfig.getBegin()
					+ " to " + eaacsLabConfig.getEnd() + "."
					+ " Filling up with empty observations."
				);
			}
			if (
				0 > beginConfiguredVsCovered
				||
				0 < endConfiguredVsCovered
			) {
				componentHelper.logger.log(
					Level.FINER,
					componentHelper.logName,
					"handle( TimeSeries )",
					"Time series \"" + timeSeries.getKey()
					+ "\" extends outside configured interval"
					+ " from " + eaacsLabConfig.getBegin()
					+ " to " + eaacsLabConfig.getEnd() + "."
					+ " Ignoring all observations outside."
				);
			}

			final ObservationMap observationMap = new ObservationMap();
			observationMap.put( timeSeries );

			for (
				final Iterator periods = configuredInterval.iterator();
				periods.hasNext();
			) {
				final RegularInterval period = (RegularInterval) periods.next();
				final Observation observation = observationMap.get( period );
				line.append(
					null == observation || observation.isEmpty()
					? valueRepresentationStrategy.getEmptyCode()
					: observation.hasMissingValue()
					? valueRepresentationStrategy.getMissingCode()
					: valueRepresentationStrategy.represent(
						observation.getValue()
					)
				);
				line.append( periods.hasNext() ? ',' : '\n' );
			}

			writer.write( line.toString() );
		}
		catch ( final IOException ioe ) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"handle( TimeSeries )",
				"Output problem.",
				ioe
			);
			close();
		}
	}

	/* ----------------------- Attribute ------------------------ */

	public void handle ( final Attribute attribute ) {
		componentHelper.logger.log(
			Level.FINER,
			componentHelper.logName,
			"handle( Attribute )",
			"Ignoring Attribute",
			attribute
		);
	}
}

/* *************************** Interface **************************** */

/* ======================= Interfaces ======================= */

/** Determines the representation of empty observations and missing values. */
interface ValueRepresentationStrategy {

	/** Return the string to use for normal observations. */
	String represent ( String value );

	/** Return the string to use for empty observations. */
	String getEmptyCode ();

	/** Return the string to use for observations with missing value. */
	String getMissingCode ();
}


/* ***************************** Class ****************************** */

class FameStyleStrategy implements ValueRepresentationStrategy {

	public String represent ( final String value ) {
		return value;
	}

	public String getEmptyCode () {
		return "ND";
	}

	public String getMissingCode () {
		return "NC";
	}
}

/* ***************************** Class ****************************** */

class EmptyTemplateStrategy implements ValueRepresentationStrategy {

	public String represent ( final String value ) {
		return "";
	}

	public String getEmptyCode () {
		return "";
	}

	public String getMissingCode () {
		return "";
	}
}
