// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2007, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.csv;

import java.io.File;
import java.io.FileReader;
import java.io.BufferedReader;
import java.io.IOException;

import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;

import Int.ecb.gesmescb.*;
import Int.ecb.gesmescb.metabase.Metabase;

import Int.ecb.gesmescb.event.GESMESCBWriter;
import Int.ecb.gesmescb.event.TestEventSource;

/* ***************************** Class ****************************** */

/**
 * Implements a few tests for the {@link EaacsLabWriter}.
 *
 * @author
 * <a href="mailto:sis.external@ecb.int">Bernhard Bodenstorfer</a>
 *
 * @version @(#)$Id: EaacsLabWriterTest.java,v 1.1 2017/01/19 16:14:04 trainot Exp $
 */
public class EaacsLabWriterTest extends TestCase {

	/* ======================= Constants ======================== */

	private static final String
		objectName = EaacsLabWriter.class.getName() + ".Tjeerd",
		sharedName = EaacsLabConfig.class.getName() + ".Tjeerd";

	private static final String[][] sharedProperties = {
		{ "1.dataSet", "ECB_IEAQ" },
		{ "1.maintenanceAgency", "ECB" },
		{ "1.keyFamily", "ECB_IEA1" },
		{ "1.dimension.FREQ.accept", "A" },
		{ "1.dimension.REF_AREA.translate", "EE,U2" },
		{ "1.dimension.ITEM.accept", "*" },
		{ "begin", "1995-01-01T00:00:00" },
		{ "end", "1999-01-01T00:00:00" }
	};

	private static final MessageBegin UPDATE_MESSAGE
	= new MessageBegin(
		"MREF000002", "4F0", "ZZZ", "ECB",
		MessageBegin.STATISTICAL_DATA,
		"Test data", new Contact[]{},
		"ECB_IEAQ", MessageBegin.UPDATE_OR_REPLACE_ACTION,
		"199901011432", "", "", "ECB_IEA1"
	);

	private static final MessageBegin DELETE_MESSAGE
	= new MessageBegin(
		"MREF000002", "4F0", "ZZZ", "ECB",
		MessageBegin.STATISTICAL_DATA,
		"Test delete", new Contact[]{},
		"ECB_IEAQ", MessageBegin.DELETE_ACTION,
		"199901011432", "", "", "ECB_IEA1"
	);

	private static final MessageEnd MESSAGE_END = new MessageEnd(
		"MREF000001"
	);

	/* ========================= Fields ========================= */

	/** The temporary destination for test output. */
	private File file;

	/* ====================== Constructors ====================== */

	public EaacsLabWriterTest ( final String name ) {
		super( name );
	}

	/* ================== Protected Methods ===================== */

	/** Sets up the test fixture. */
	protected void setUp () throws IOException {
		for ( int i = 0; i < sharedProperties.length; ++i ) {
			System.setProperty(
				sharedName + "." + sharedProperties[i][0],
				sharedProperties[i][1]
			);
		}

		file = File.createTempFile( "EaacsLabWriterTest", ".csv" );
		file.deleteOnExit();
		System.setProperty(
			objectName + ".destination",
			file.getAbsolutePath()
		);

		final TestEventSource source = new TestEventSource();
		final Metabase metabase = Metabase.getMetabase();
		source.addListener( metabase, true );
		source.dispatch( EaacsLabConfigTest.STRUCTURE_MESSAGE );
		source.dispatch( EaacsLabConfigTest.KF_IEA1 );
		source.dispatch( EaacsLabConfigTest.MESSAGE_END );
	}

	/** Tears down the test fixture. */
	protected void tearDown () {
		for ( int i = 0; i < sharedProperties.length; ++i ) {
			System.getProperties().remove(
				sharedName + "." + sharedProperties[i][0]
			);
		}
		System.getProperties().remove( objectName + ".destination" );
		file.delete();
		file = null;
	}

	/* ===================== Public Methods ===================== */

	/**
	 * Assembles and returns a test suite for all
	 * the test methods of this test case.
	 *
	 * @return a non-null test suite
	 */
	public static Test suite () {
		return new TestSuite( EaacsLabWriterTest.class );
	}

	/**
	 * @param arg the array of argument {@link String}s
	 */
	public static void main ( final String[] arg ) {
		final String[] testCaseName = {
			EaacsLabWriterTest.class.getName()
		};
		junit.textui.TestRunner.main( testCaseName );
	}

	/* ---------------------- Test Methods ---------------------- */

	/**
	 * Test {@link EaacsLabWriter}.
	 *
	 * @throws IOException if opening the EaacsLabWriter yields a problem.
	 */
	public void testAnnual () throws IOException {
		final TestEventSource source = new TestEventSource();
		final GESMESCBWriter writer = new EaacsLabWriter( "Tjeerd" );

		source.addListener( writer, true );

		source.dispatch( UPDATE_MESSAGE );

		source.dispatch( new TimeSeries(
			"A.U2.C", "19951998", "702", new Observation[]{
				new Observation( "3.6", "A", "F", "" ),
				new Observation( "13.6", "A", "F", "" ),
				new Observation( "23.6", "A", "F", "" ),
				new Observation( "33.6", "A", "F", "" )
			}
		) );

		source.dispatch( new TimeSeries(
			"A.EE.C", "19951998", "702", new Observation[]{
				new Observation( "4.6", "A", "C", "-" ),
				new Observation( Observation.MISSING_VALUE, "M", "C", "-" ),
				Observation.EMPTY_OBSERVATION,
				new Observation( "-3.6", "B", "F", "" )
			}
		) );

		source.dispatch( MESSAGE_END );

		writer.close();

		final BufferedReader reader
		= new BufferedReader( new FileReader( file ) );
		assertEquals(
			"1", "ECB_IEAQ.A.U2.C,2,3.6,13.6,23.6,33.6", reader.readLine()
		);
		assertEquals(
			"2", "ECB_IEAQ.A.EE.C,1,4.6,NC,ND,-3.6", reader.readLine()
		);
		assertNull( "3", reader.readLine() );
		reader.close();
	}

	/**
	 * Test {@link EaacsLabWriter} respect of begin and end times.
	 *
	 * @throws IOException if opening the EaacsLabWriter yields a problem.
	 */
	public void testBeginEnd () throws IOException {
		final TestEventSource source = new TestEventSource();
		final GESMESCBWriter writer = new EaacsLabWriter( "Tjeerd" );

		source.addListener( writer, true );

		source.dispatch( UPDATE_MESSAGE );

		source.dispatch( new TimeSeries(
			"A.U2.C", "19961997", "702", new Observation[]{
				new Observation( "0", "A", "F", "18" ),
				new Observation( Observation.MISSING_VALUE, "M", "C", "-" )
			}
		) );

		source.dispatch( new TimeSeries(
			"A.EE.C", "19941999", "702", new Observation[]{
				new Observation( "4.6", "A", "C", "-" ),
				new Observation( Observation.MISSING_VALUE, "M", "C", "-" ),
				new Observation( Observation.MISSING_VALUE, "M", "C", "-" ),
				Observation.EMPTY_OBSERVATION,
				new Observation( Observation.MISSING_VALUE, "M", "C", "-" ),
				new Observation( "-3.6", "B", "F", "" )
			}
		) );

		source.dispatch( MESSAGE_END );

		writer.close();

		final BufferedReader reader
		= new BufferedReader( new FileReader( file ) );
		assertEquals(
			"1", "ECB_IEAQ.A.U2.C,2,ND,0,NC,ND", reader.readLine()
		);
		assertEquals(
			"2", "ECB_IEAQ.A.EE.C,1,NC,NC,ND,NC", reader.readLine()
		);
		assertNull( "3", reader.readLine() );
		reader.close();
	}

	/**
	 * Test {@link EaacsLabWriter} with empty value strategy.
	 *
	 * @throws IOException if opening the EaacsLabWriter yields a problem.
	 */
	public void testEmptyStrategy () throws IOException {
		final TestEventSource source = new TestEventSource();
		final GESMESCBWriter writer = new EaacsLabWriter(
			"Tjeerd", null, EaacsLabWriter.EMPTY_STRATEGY
		);

		source.addListener( writer, true );

		source.dispatch( UPDATE_MESSAGE );

		source.dispatch( new TimeSeries(
			"A.U2.C", "19951998", "702", new Observation[]{
				new Observation( "6.6", "E", "F", "" ),
				new Observation( Observation.MISSING_VALUE, "L", "F", "" ),
				Observation.EMPTY_OBSERVATION,
				new Observation( "6.6", "F", "F", "" )
			}
		) );

		source.dispatch( MESSAGE_END );

		writer.close();

		final BufferedReader reader
		= new BufferedReader( new FileReader( file ) );
		assertEquals(
			"1", "ECB_IEAQ.A.U2.C,2,,,,", reader.readLine()
		);
		assertNull( "2", reader.readLine() );
		reader.close();
	}

	/**
	 * Test {@link EaacsLabWriter}.
	 *
	 * @throws IOException if opening the EaacsLabWriter yields a problem.
	 */
	public void testIgnoreDelete () throws IOException {
		final TestEventSource source = new TestEventSource();
		final GESMESCBWriter writer = new EaacsLabWriter( "Tjeerd" );

		source.addListener( writer, true );

		source.dispatch( DELETE_MESSAGE );

		source.dispatch( new TimeSeries(
			"A.U2.C", "19951998", "702", new Observation[]{
				new Observation( "3.6", "A", "F", "" )
			}
		) );

		writer.close();

		final BufferedReader reader
		= new BufferedReader( new FileReader( file ) );
		assertNull( "1", reader.readLine() );
		reader.close();
	}
}
