// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2007, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.csv;

import java.io.Writer;
import java.io.FileWriter;
import java.io.BufferedWriter;
import java.io.IOException;

import java.util.Iterator;

import Int.ecb.logging.Level;

import Int.ecb.gesmescb.InterchangeBegin;
import Int.ecb.gesmescb.InterchangeEnd;
import Int.ecb.gesmescb.MessageBegin;
import Int.ecb.gesmescb.MessageEnd;
import Int.ecb.gesmescb.TimeSeries;
import Int.ecb.gesmescb.Observation;
import Int.ecb.gesmescb.Attribute;

import Int.ecb.gesmescb.event.DataListener;
import Int.ecb.gesmescb.event.GESMESCBEventListenerAdapter;
import Int.ecb.gesmescb.event.GESMESCBWriter;

import Int.ecb.gesmescb.util.Named;
import Int.ecb.gesmescb.util.ComponentHelper;

import Int.ecb.gesmescb.time.RegularInterval;

/* ***************************** Class ****************************** */

/**
 * <p>Writes {@link TimeSeries} to Hyperion/essbase style flat file
 * for import.</p>
 *
 * <center><table border="3" cellpadding="3">
 * <caption>
 * <a name="prefs">Available {@link Int.ecb.prefs preference} settings</a>
 * </caption>
 * <tr>
 * <th>preference name</th>
 * <th>type</th>
 * <th>meaning</th>
 * <th>default</th>
 * </tr>
 * <tr><td><code>destination</td><td>string</td><td>output file path</td></tr>
 * </table></center>
 *
 * @author
 * <a href="mailto:sis.external@ecb.int">Xavier Sosnowska</a>
 *
 * @author
 * <a href="mailto:sis.external@ecb.int">Bernhard Bodenstorfer</a>
 *
 * @version @(#)$Id: EssbaseLoadWriter.java,v 1.1 2017/01/19 16:14:04 trainot Exp $
 */
public class EssbaseLoadWriter extends GESMESCBEventListenerAdapter
	implements Named, DataListener, GESMESCBWriter
{
	/* ========================= Fields ========================= */

	/** Takes care of logging and preferences. */
	protected final ComponentHelper componentHelper;

	/**
	 * The filename of the file to write to
	 * if the <code>EssbaseLoadWriter</code> has been
	 * constructed from a filename.
	 * Otherwise it is null.
	 */
	private final String filename;

	/** The <code>Writer</code> for writing to the CSV-output file. */
	private final Writer writer;

	/** Whether current is a deletion message. */
	private boolean isDeleteMessage;

	/* ====================== Constructors ====================== */

	/**
	 * Construct a <code>EssbaseLoadWriter</code> to write to a file.
	 *
	 * @param name used for logging and preferences.
	 * @param filename where to write to.
	 *
	 * @throws NullPointerException if <code>filename==null</code>
	 * and the preference variable
	 * <code>destination</code> is undefined.
	 * @throws IOException if the file cannot be opened for writing.
	 */
	public EssbaseLoadWriter ( final String name, final String filename )
		throws IOException
	{
		componentHelper = new ComponentHelper( name, this );
		this.filename = filename != null
			? filename
			: componentHelper.getPreference( "destination", null );
		if ( this.filename == null ) {
			final String desc
			= "No preference variable \"destination\" found"
				+ " to determine the output file.";
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"EssbaseLoadWriter()",
				desc
			);
			throw new NullPointerException( desc );
		}
		writer = new BufferedWriter( new FileWriter( this.filename ) );
		isDeleteMessage = false;
	}

	/**
	 * Construct a <code>EssbaseLoadWriter</code>
	 * writing to a file according to preferences.
	 *
	 * @param name used for logging and preferences.
	 *
	 * @throws NullPointerException if the preference variable
	 * <code>destination</code> is undefined.
	 * @throws IOException if the file cannot be opened for writing.
	 */
	public EssbaseLoadWriter ( final String name )
		throws IOException
	{
		this( name, null );
	}

	/* ===================== Public Methods ===================== */

	public String getName () {
		return componentHelper.name;
	}

	/* ---------------------- Interchange ----------------------- */

	public void handle ( final InterchangeBegin interchangeBegin ) {
		componentHelper.logger.log(
			Level.FINE,
			componentHelper.logName,
			"handle( InterchangeBegin )",
			"Received InterchangeBegin",
			interchangeBegin
		);
	}

	public void handle ( final InterchangeEnd interchangeEnd ) {
		componentHelper.logger.log(
			Level.FINE,
			componentHelper.logName,
			"handle( InterchangeEnd )",
			"Received InterchangeEnd",
			interchangeEnd
		);
	}

	/* ------------------------ Message ------------------------- */

	public void handle ( final MessageBegin messageBegin ) {
		componentHelper.logger.log(
			Level.FINE,
			componentHelper.logName,
			"handle( MessageBegin )",
			"Received MessageBegin",
			messageBegin
		);

		if ( ( isDeleteMessage = messageBegin.isDeleteAction() ) ) {
			componentHelper.logger.log(
				Level.WARNING,
				componentHelper.logName,
				"handle( MessageBegin )",
				"Message \"" + messageBegin.getMref()
				+ "\" is a delete message and will be ignored.",
				messageBegin
			);
		}
	}

	public void handle ( final MessageEnd messageEnd ) {
		componentHelper.logger.log(
			Level.FINE,
			componentHelper.logName,
			"handle( MessageEnd )",
			"Received MessageEnd",
			messageEnd
		);
	}

	/* ---------------------- Time Series ----------------------- */

	public void handle ( final TimeSeries timeSeries ) {
		if ( isDeleteMessage ) {
			componentHelper.logger.log(
				Level.FINER,
				componentHelper.logName,
				"handle( TimeSeries )",
				"Ignoring TimeSeries deletion",
				timeSeries
			);
			return;
		}

		final String key = timeSeries.getKey();

		componentHelper.logger.log(
			Level.FINER,
			componentHelper.logName,
			"handle( TimeSeries )",
			"Writing TimeSeries " + key,
			timeSeries
		);

		try {
			final String commaKey
				= key.replace( TimeSeries.KEY_SEPARATOR, ',' );

			final RegularInterval interval = timeSeries.getInterval();
			for (
				final Iterator
					periods = interval.iterator(),
					observations = timeSeries.observationIterator();
				periods.hasNext() && observations.hasNext();
			) {
				final RegularInterval period
					= (RegularInterval) periods.next();
				final Observation observation
					= (Observation) observations.next();
				if ( ! observation.isEmpty() ) {
					writer.write( commaKey );
					writer.write( ',' );
					final int currentYear
						= period.getFirstPeriodYear();
					writer.write( Integer.toString( currentYear ) );
					writer.write( ',' );
					if ( RegularInterval.ANNUAL == period.getFrequency() ) {
						writer.write( 'A' );
					} else {
						final int currentIndex = period.getFirstPeriodIndex();
						if (
							RegularInterval.QUARTERLY == period.getFrequency()
						) {
							writer.write( 'Q' );
						} else if (
							RegularInterval.MONTHLY == period.getFrequency()
						) {
							writer.write( 'M' );
						} else if (
							RegularInterval.DAILY == period.getFrequency()
						) {
							writer.write( 'D' );
						} else {
							componentHelper.logger.log(
								Level.SEVERE,
								componentHelper.logName,
								"handle( TimeSeries )",
								"Time series " + key
								+ "[" + interval + "] has a frequency"
								+ " which is not implemented for EAACS",
								timeSeries
							);
							writer.write( '?' );
						}
						writer.write( Integer.toString( currentIndex + 1 ) );
					}
					writer.write( ',' );
					final String value = observation.getValue();
					writer.write(
						Observation.MISSING_VALUE.equals( value )
						? "#Mi" : value
					);
					writer.write( '\n' );
				}
			}
		}
		catch ( final IOException ioe ) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"handle( TimeSeries )",
				"Output problem.",
				ioe
			);
			close();
		}
	}

	/* ----------------------- Attribute ------------------------ */

	public void handle ( final Attribute attribute ) {
		componentHelper.logger.log(
			Level.FINER,
			componentHelper.logName,
			"handle( Attribute )",
			"Ignoring Attribute",
			attribute
		);
	}

	/* ------------------------- close -------------------------- */

	/**
	 * Close the <code>EssbaseLoadWriter</code>.
	 */
	public void close () {
		try {
			componentHelper.logger.log(
				Level.FINE,
				componentHelper.logName,
				"close()",
				"Closing"
			);
			setActive( false );
			writer.close();
		}
		catch ( final IOException ioe ) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"close()",
				"Closing failed.",
				ioe
			);
		}
	}
}
