// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2002, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.csv;

import java.io.Writer;
import java.io.FileWriter;
import java.io.BufferedWriter;
import java.io.IOException;

import java.util.Iterator;
import java.text.FieldPosition;
import java.text.ParseException;

import Int.ecb.logging.Level;

import Int.ecb.gesmescb.*;
import Int.ecb.gesmescb.metabase.*;

import Int.ecb.gesmescb.event.DataListener;
import Int.ecb.gesmescb.event.GESMESCBEventListenerAdapter;
import Int.ecb.gesmescb.event.GESMESCBWriter;

import Int.ecb.gesmescb.util.Named;
import Int.ecb.gesmescb.util.ComponentHelper;
import Int.ecb.gesmescb.util.KeyTokenizer;

import Int.ecb.gesmescb.time.RegularInterval;
import Int.ecb.gesmescb.edifact.RegularIntervalFormat;

/* ***************************** Class ****************************** */

/**
 * <P>Writes {@link TimeSeries} to a Eurostat style CSV file.</P>
 *
 * <center><table border="3" cellpadding="3">
 * <caption>
 * <a name="prefs">Available {@link Int.ecb.prefs preference} settings</a>
 * </caption>
 * <tr>
 * <th>preference name</th>
 * <th>type</th>
 * <th>meaning</th>
 * <th>default</th>
 * </tr>
 * <tr><td><code>destination</td><td>string</td><td>output file path</td></tr>
 * <tr>
 * <td><code>decimal.separator</td><td>string</td>
 * <td>decimal separator for observation value and pre break value</td>
 * <td>dot (".")</td></tr>
 * </table></center>
 *
 * @author
 * <a href="mailto:sis.external@ecb.int">Bernhard Bodenstorfer</a>
 *
 * @version @(#)$Id: EurostatWriter.java,v 1.1 2017/01/19 16:14:04 trainot Exp $
 */
public class EurostatWriter extends GESMESCBEventListenerAdapter
	implements Named, DataListener, GESMESCBWriter
{
	/* ========================= Fields ========================= */

	/**
	 * Takes care of logging and preferences.
	 */
	protected final ComponentHelper componentHelper;

	/**
	 * The filename of the file to write to
	 * if the <CODE>EDIFACTWriter</CODE> has been
	 * constructed from a filename.
	 * Otherwise it is null.
	 */
	private final String filename;

	/**
	 * The <CODE>Writer</CODE> for writing to the CSV-output file.
	 */
	private final Writer writer;

	/** Character to use as decimal separator.
	 *
	 * This is to support legacy applications which still need
	 * something else than a dot.
	 *
	 * @see #useSeparator
	 */
	private final String decimalSeparator;

	/** Whether current is a deletion message. */
	private boolean isDeleteMessage;

	/**
	 * Number of array cells in the current <code>KeyFamily</code>.
	 * Used in data messages.
	 */
	private int arrayCellCount;

	/* ====================== Constructors ====================== */

	/**
	 * Construct a <CODE>EurostatWriter</CODE> to write to a file.
	 *
	 * @param name used for logging and preferences.
	 * @param filename where to write to.
	 *
	 * @throws NullPointerException if <CODE>filename==null</CODE>
	 * and the preference variable
	 * <CODE>destination</CODE> is undefined.
	 * @throws IOException if the file cannot be opened for writing.
	 */
	public EurostatWriter ( final String name, final String filename )
		throws IOException
	{
		componentHelper = new ComponentHelper( name, this );
		this.filename = filename != null
			? filename
			: componentHelper.getPreference( "destination", null );
		if ( this.filename == null ) {
			final String desc
			= "No preference variable \"destination\" found"
				+ " to determine the output file.";
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"EurostatWriter()",
				desc
			);
			throw new NullPointerException( desc );
		}
		writer = new BufferedWriter( new FileWriter( this.filename ) );
		decimalSeparator = componentHelper.getPreference(
			"decimal.separator", "."
		);
		if ( decimalSeparator.length() != 1 ) {
			final String desc
			= "Preference \"decimal.separator\" is \""
				+ decimalSeparator
				+ "\". This is other than a single character.";
			componentHelper.logger.log(
				Level.WARNING,
				componentHelper.logName,
				"EurostatWriter()",
				desc
			);
		}
		arrayCellCount = 4;	// default value
		isDeleteMessage = false;
	}

	/**
	 * Construct a <CODE>EurostatWriter</CODE>
	 * writing to a file according to preferences.
	 *
	 * @param name used for logging and preferences.
	 *
	 * @throws NullPointerException if the preference variable
	 * <CODE>destination</CODE> is undefined.
	 * @throws IOException if the file cannot be opened for writing.
	 */
	public EurostatWriter ( final String name )
		throws IOException
	{
		this( name, null );
	}

	/* ===================== Public Methods ===================== */

	public String getName () {
		return componentHelper.name;
	}

	/**
	 * <P>Close the <CODE>EurostatWriter</CODE>.</P>
	 */
	public void close () {
		try {
			componentHelper.logger.log(
				Level.FINE,
				componentHelper.logName,
				"close()",
				"Closing"
			);
			setActive( false );
			writer.close();
		}
		catch ( final IOException ioe ) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"close()",
				"Closing failed.",
				ioe
			);
		}
	}

	/* ---------------------- Interchange ----------------------- */

	public void handle ( final InterchangeBegin interchangeBegin ) {
		componentHelper.logger.log(
			Level.FINE,
			componentHelper.logName,
			"handle( InterchangeBegin )",
			"Received InterchangeBegin",
			interchangeBegin
		);
	}

	public void handle ( final InterchangeEnd interchangeEnd ) {
		componentHelper.logger.log(
			Level.FINE,
			componentHelper.logName,
			"handle( InterchangeEnd )",
			"Received InterchangeEnd",
			interchangeEnd
		);
	}

	/* ------------------------ Message ------------------------- */

	public void handle ( final MessageBegin messageBegin ) {
		componentHelper.logger.log(
			Level.FINE,
			componentHelper.logName,
			"handle( MessageBegin )",
			"Received MessageBegin",
			messageBegin
		);

		if (
			( isDeleteMessage = messageBegin.isDeleteAction() )
		) {
			componentHelper.logger.log(
				Level.WARNING,
				componentHelper.logName,
				"handle( MessageBegin )",
				"Message \"" + messageBegin.getMref()
				+ "\" is a delete message and will be ignored.",
				messageBegin
			);
		}
		arrayCellCount = 4;	// default value
		if ( ! "".equals( messageBegin.getKeyFamilyId() ) ) {
			final Metabase metabase = Metabase.getMetabase();
			final String maintenanceAgencyId
			= messageBegin.getMaintenanceAgencyId();
			final MaintenanceAgency maintenanceAgency
			= metabase.getMaintenanceAgency( maintenanceAgencyId );
			if ( maintenanceAgency != null ) {
				final String keyFamilyId = messageBegin.getKeyFamilyId();
				final KeyFamily keyFamily
				= maintenanceAgency.getKeyFamily( keyFamilyId );
				if ( keyFamily == null ) {
					componentHelper.logger.log(
						Level.SEVERE,
						componentHelper.logName,
						"handle( MessageBegin )",
						"Maintenance agengy " + maintenanceAgencyId
						+ " does not know key family \""
						+ keyFamilyId + "\"."
					);
					close();
				} else {
					arrayCellCount = keyFamily.getArrayCellCount();
				}
			} else {
				componentHelper.logger.log(
					Level.SEVERE,
					componentHelper.logName,
					"handle( MessageBegin )",
					"Unknown maintenance agengy \"" + maintenanceAgencyId
					+ "\"."
				);
				close();
			}
		}
	}

	public void handle ( final MessageEnd messageEnd ) {
		componentHelper.logger.log(
			Level.FINE,
			componentHelper.logName,
			"handle( MessageEnd )",
			"Received MessageEnd",
			messageEnd
		);
	}

	/* ---------------------- Time Series ----------------------- */

	public void handle ( final TimeSeries timeSeries ) {
		if ( isDeleteMessage ) {
			componentHelper.logger.log(
				Level.FINER,
				componentHelper.logName,
				"handle( TimeSeries )",
				"Ignoring TimeSeries deletion",
				timeSeries
			);
		} else try {
			componentHelper.logger.log(
				Level.FINER,
				componentHelper.logName,
				"handle( TimeSeries )",
				"Writing TimeSeries",
				timeSeries
			);

			final StringBuffer line = new StringBuffer();

			for (
				final Iterator dimensionValues
				= new KeyTokenizer( timeSeries.getKey() );
				dimensionValues.hasNext();
			) {
				final String value = (String) dimensionValues.next();
				line.append( escapeComma( value ) );
				line.append( ',' );
			}
			final String keyPattern = line.toString();

			final RegularIntervalFormat format = new RegularIntervalFormat();
			final RegularInterval interval = format.parse(
				timeSeries.getPeriod(), timeSeries.getPeriodFormat()
			);
			final FieldPosition periodPosition = new FieldPosition(
				RegularIntervalFormat.PERIOD_RANGE_FIELD
			);
			for (
				final Iterator
					observations = timeSeries.observationIterator(),
					periods = interval.iterator();
					observations.hasNext() && periods.hasNext();
			) {
				final Observation observation
				= (Observation) observations.next();
				final RegularInterval period
				= (RegularInterval) periods.next();

				line.setLength( 0 );
				line.append( keyPattern );
				format.format( period, line, periodPosition );
				line.setLength( periodPosition.getEndIndex() );
				if ( arrayCellCount >= 1 ) {
					line.append( ',' );
					line.append(
						escapeComma( useSeparator( observation.getValue() ) )
					);
				}
				if ( arrayCellCount >= 2 ) {
					line.append( ',' );
					line.append(
						escapeComma( observation.getStatus() )
					);
				}
				if ( arrayCellCount >= 3 ) {
					line.append( ',' );
					line.append(
						escapeComma( observation.getConfidentiality() )
					);
				}
				if ( arrayCellCount >= 4 ) {
					line.append( ',' );
					line.append(
						escapeComma(
							useSeparator( observation.getPreBreakValue() )
						)
					);
				}
				line.append( '\n' );
				writer.write( line.toString() );
			}
		}
		catch ( final ParseException pe ) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"handle( TimeSeries )",
				"Failed to parse TimeSeries period information.",
				pe
			);
			close();
		}
		catch ( final IOException ioe ) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"handle( TimeSeries )",
				"Output problem.",
				ioe
			);
			close();
		}
	}

	/* ----------------------- Attribute ------------------------ */

	public void handle ( final Attribute attribute ) {
		componentHelper.logger.log(
			Level.FINER,
			componentHelper.logName,
			"handle( Attribute )",
			"Ignoring Attribute",
			attribute
		);
	}

	/* ==================== Private Methods ===================== */

	/**
	 * Use {@link #decimalSeparator} in place of dot in numbers.
	 * This is to support legacy applications which still need
	 * something else than a dot.
	 */
	private String useSeparator ( final String value ) {
		if (
			decimalSeparator.length() != 1
			||
			TimeSeries.KEY_SEPARATOR != decimalSeparator.charAt( 0 )
		) {
			final int dotPosition = value.indexOf( TimeSeries.KEY_SEPARATOR );
			if ( dotPosition >= 0 ) {
				final StringBuffer buffer = new StringBuffer( value );
				buffer.replace(
					dotPosition, dotPosition + 1, decimalSeparator
				);
				return buffer.toString();
			}
		}
		return value;
	}

	/**
	 * Put in Quotes if <code>value</code> contains a comma.
	 * This is necessary not ot mess up the CSV format.
	 */
	private String escapeComma ( final String value ) {
		if ( value.indexOf( "," ) >= 0 ) {
			return "\"" + value + "\"";
		} else {
			return value;
		}
	}
}
