// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.csv;

import java.util.Date;
import java.util.Calendar;

import java.io.File;
import java.io.FileReader;
import java.io.BufferedReader;
import java.io.IOException;

import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;

import Int.ecb.gesmescb.*;
import Int.ecb.gesmescb.metabase.Metabase;

import Int.ecb.gesmescb.event.GESMESCBWriter;
import Int.ecb.gesmescb.event.TestEventSource;

import Int.ecb.gesmescb.csv.EurostatWriter;

/* ***************************** Class ****************************** */

/**
 * Implements a few tests for the {@link EurostatWriter}.
 *
 * @author
 * <A HREF="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</A>
 *
 * @version @(#)$Id: EurostatWriterTest.java,v 1.1 2017/01/19 16:14:04 trainot Exp $
 */
public class EurostatWriterTest extends TestCase {

	/* ======================= Constants ======================== */

	private static final MessageBegin structureMessage
	= new MessageBegin(
		"MREF000001", "4F0", "ZZZ", "ECB",
		MessageBegin.STATISTICAL_DEFINITIONS,
		"Set up", new Contact[]{}
	);

	private static final MessageBegin updateMessage
	= new MessageBegin(
		"MREF000002", "4F0", "ZZZ", "ECB",
		MessageBegin.STATISTICAL_DATA,
		"Set up KF reference", new Contact[]{},
		"DS0", MessageBegin.UPDATE_OR_REPLACE_ACTION,
		"199901011432", "", "", "CSV_TEST_KF"
	);

	/**
	 * Version date of the Key Family.
	 */
	private static final Date VERSION_DATE;
	static {
		final Calendar calendar = Calendar.getInstance();
		calendar.clear();
		calendar.set( 2000, Calendar.OCTOBER, 5, 4, 22 );
		VERSION_DATE = calendar.getTime();
	}

	private static final KeyFamily CSV_TEST_KF
	= new KeyFamily(
		"CSV_TEST_KF", "", "only three array cells",
		VERSION_DATE,
		new KeyFamilyComponent[]{
			new KeyFamilyComponent(
				"OBS_VALUE",
				KeyFamilyComponent.ARRAY_CELL,
				"6", "AN..15", "", "", ""
			),
			new KeyFamilyComponent(
				"OBS_STATUS",
				KeyFamilyComponent.ARRAY_CELL,
				"7", "A1", "", "", ""
			),
			new KeyFamilyComponent(
				"OBS_CONF",
				KeyFamilyComponent.ARRAY_CELL,
				"8", "A1", "", "", ""
			)
		}
	);

	/* ========================= Fields ========================= */

	/**
	 * The temporary destination for test output.
	 */
	private File file;

	/* ====================== Constructors ====================== */

	public EurostatWriterTest ( final String name ) {
		super( name );
	}

	/* ================== Protected Methods ===================== */

	/**
	 * Sets up the test fixture.
	 */
	protected void setUp () throws IOException {
		file = File.createTempFile( "EurostatWriterTest", ".csv" );
		file.deleteOnExit();
		System.setProperty(
			EurostatWriter.class.getName() + ".Bjorn.destination",
			file.getAbsolutePath()
		);
	}

	/**
	 * Tears down the test fixture.
	 */
	protected void tearDown () {
		System.getProperties().remove(
			EurostatWriter.class.getName() + ".Bjorn.destination"
		);
		file.delete();
		file = null;
	}

	/* ===================== Public Methods ===================== */

	/**
	 * Assembles and returns a test suite for all
	 * the test methods of this test case.
	 *
	 * @return a non-null test suite
	 */
	public static Test suite () {
		TestSuite suite = new TestSuite( EurostatWriterTest.class );
		return suite;
	}

	/**
	 * @param arg the array of argument {@link String}s
	 */
	public static void main ( final String[] arg ) {
		final String[] testCaseName = {
			EurostatWriterTest.class.getName()
		};
		junit.textui.TestRunner.main( testCaseName );
	}

	/* ---------------------- Test Methods ---------------------- */

	/**
	 * Test {@link EurostatWriter}.
	 *
	 * @throws IOException if opening the EurostatWriter yields a problem.
	 */
	public void testDefault () throws IOException {
		final TestEventSource source = new TestEventSource();
		final GESMESCBWriter writer = new EurostatWriter( "Bjorn" );

		source.addListener( writer, true );

		source.dispatch( new TimeSeries(
			"A.B.C", "199504", "610", new Observation[]{
				new Observation( "13.6", "A", "F", "" )
			}
		) );

		source.dispatch( new TimeSeries(
			"A.B.E", "1999051319990514", "711", new Observation[]{
				new Observation( "14.6", "A", "C", "-" ),
				new Observation( "-13.6", "B", "F", "" )
			}
		) );

		writer.close();

		final BufferedReader reader
		= new BufferedReader( new FileReader( file ) );

		assertEquals( "1", "A,B,C,199504,13.6,A,F,", reader.readLine() );
		assertEquals( "2", "A,B,E,19990513,14.6,A,C,-", reader.readLine() );
		assertEquals( "3", "A,B,E,19990514,-13.6,B,F,", reader.readLine() );
		assertNull( "4", reader.readLine() );
		reader.close();
	}

	/**
	 * Test special wishes from Eurostat.
	 * Comma instead of decimal dot, no pre-break, quoting.
	 */
	public void testSpecial () throws IOException {
		System.setProperty(
			EurostatWriter.class.getName() + ".Bjorn.decimal.separator",
			","
		);

		final TestEventSource source = new TestEventSource();
		final Metabase metabase = Metabase.getMetabase();
		final GESMESCBWriter writer = new EurostatWriter( "Bjorn" );

		source.addListener( metabase, false );
		source.addListener( writer, true );

		source.dispatch( structureMessage );

		source.dispatch( CSV_TEST_KF );

		source.dispatch( new MessageEnd( "MREF000001" ) );

		source.dispatch( updateMessage );

		source.dispatch( new TimeSeries(
			"A,C.B.C", "199504", "610", new Observation[]{
				new Observation( "13.6", "A", "F,r", "" )
			}
		) );

		writer.close();

		final BufferedReader reader
		= new BufferedReader( new FileReader( file ) );

		assertEquals(
			"1", "\"A,C\",B,C,199504,\"13,6\",A,\"F,r\"",
			reader.readLine()
		);
		assertNull( "2", reader.readLine() );

		reader.close();

		System.getProperties().remove(
			EurostatWriter.class.getName() + ".Bjorn.decimal.separator"
		);
	}

	/**
	 * Test {@link EurostatWriter}.
	 *
	 * @throws IOException if opening the EurostatWriter yields a problem.
	 */
	public void testIgnoreDelete () throws IOException {
		final TestEventSource source = new TestEventSource();
		final Metabase metabase = Metabase.getMetabase();
		final GESMESCBWriter writer = new EurostatWriter( "Bjorn" );

		source.addListener( metabase, false );
		source.addListener( writer, true );

		source.dispatch( structureMessage );

		source.dispatch( CSV_TEST_KF );

		source.dispatch( new TimeSeries(
			"A.B.C", "199504", "610", new Observation[]{
				new Observation( "13.6", "A", "", "" )
			}
		) );

		source.dispatch( new MessageBegin(
			"MREF000002", "EE2", "4F0", "ECB",
			MessageBegin.STATISTICAL_DATA, "Delete test message",
			new Contact[]{},
			"CSV_TEST_KF", MessageBegin.DELETE_ACTION,
			"200008091923", "", "", "CSV_TEST_KF"
		) );

		source.dispatch( new TimeSeries(
			"A.B.E", "1999051319990514", "711", new Observation[]{
				new Observation( "14.6", "A", "", "-" ),
				new Observation( "-13.6", "A", "", "" )
			}
		) );

		writer.close();

		final BufferedReader reader
		= new BufferedReader( new FileReader( file ) );

		assertEquals( "1", "A,B,C,199504,13.6,A,,", reader.readLine() );
		assertNull( "2", reader.readLine() );
		reader.close();
	}

}
