// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.csv;

import java.io.FileReader;
import java.io.BufferedReader;
import java.io.IOException;

import Int.ecb.logging.Level;

import Int.ecb.gesmescb.*;
import Int.ecb.gesmescb.event.*;

import Int.ecb.gesmescb.util.ComponentHelper;

/* **************************** Class ***************************** */

/**
 * Reader for timeseries key files.
 *
 * <P>Reads time series keys from text file specified</P>
 *
 * <CENTER><TABLE BORDER=3 CELLPADDING=3>
 * <CAPTION>
 * <A NAME="prefs">Available {@link Int.ecb.prefs preference} settings</A>
 * </CAPTION>
 * <TR>
 * <TH>preference name</TH>
 * <TH>preference type</TH>
 * <TH>preference value</TH>
 * </TR>
 * <TR>
 * <TD>source</TD><TD>string</TD>
 * <TD>Input file name for time series keys</TD>
 * </TR>
 * <TR>
 * <TD>keyFamily</TD><TD>string</TD>
 * <TD>Key family identifier</TD>
 * </TR>
 * <TR>
 * <TD>maintenanceAgency</TD><TD>string</TD>
 * <TD>name of the key family maintenance agency</TD>
 * </TR>
 * </TABLE></CENTER>
 *
 * @author
 * <A HREF="mailto:Sami.Airo@ECB.int">Sami Airo</A>
 *
 * @version @(#)$Id: KeyFileReader.java,v 1.1 2017/01/19 16:14:04 trainot Exp $
 */
public class KeyFileReader
	extends GESMESCBEventSourceAdapter
	implements GESMESCBReader
{
	/* ======================= Constants ======================== */

	private static final String
		IREF = "IREF000001",
		MREF = "MREF000001",
		SENDER = "4F0",
		RECEIVER = "4F0",
		PREPARATION_DATETIME = "010101:0101",
		EXTRACTION_DATETIME = "200101010101";

	/* ========================= Fields ========================= */

	/** Gets preferences. */
	protected final ComponentHelper componentHelper;

	/** The source filename. */
	private final String filename;

	/* ====================== Constructors ====================== */

	public KeyFileReader () {
		this( "" );
	}

	public KeyFileReader( final String name ) {
		this( "", null );
	}

	public KeyFileReader( final String name, final String filename ) {
		componentHelper = new ComponentHelper( name, this );
		if ( filename != null ) {
			this.filename = filename;
		} else {
			this.filename = componentHelper.getPreference( "source", null );
			if ( this.filename == null ) {
				final String desc
				= "Cannot read source name from preference \"source\".";
				componentHelper.logger.log(
					Level.SEVERE,
					componentHelper.logName,
					"handle",
					desc
				);
				dispatch(
					new ErrorEvent( desc, ErrorEvent.FAILURE )
				);
				return;
			}
		}
	}

	/* ===================== Public methods ====================== */

	public void run () {
		final String maintenanceAgency = componentHelper.getPreference(
			"maintenanceAgency", null
		);

		if ( maintenanceAgency == null ) {
			final String desc
			= "Cannot read maintenance agency from preference"
				+ " \"maintenanceAgency\".";
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"handle",
				desc
			);
			dispatch(
				new ErrorEvent( desc, ErrorEvent.FAILURE )
			);
			return;
		}

		final String keyFamilyID = componentHelper.getPreference(
			"keyFamily", null
		);

		if ( keyFamilyID == null ) {
			final String desc
			= "Cannot read key family from preference \"keyFamily\".";
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"handle",
				desc
			);
			dispatch(
				new ErrorEvent( desc, ErrorEvent.FAILURE )
			);
			return;
		}

		dispatch(
			new InterchangeBegin(
				IREF,
				SENDER,
				RECEIVER,
				PREPARATION_DATETIME,
				false
			)
		);
		dispatch(
			new MessageBegin(
				MREF,
				SENDER,
				RECEIVER,
				maintenanceAgency,
				MessageBegin.STATISTICAL_DATA,
				"",
				new Contact[]{},
				keyFamilyID,
				MessageBegin.DELETE_ACTION,
				EXTRACTION_DATETIME,
				"", "",
				keyFamilyID
			)
		);

		try {
			final BufferedReader br = new BufferedReader(
				new FileReader( filename )
			);

			String line;
			while ( ( line = br.readLine() ) != null ) {
				componentHelper.logger.log(
					Level.FINEST,
					componentHelper.logName,
					"run",
					"Dispatching key " + line
				);

				final TimeSeries timeSeries = new TimeSeries(
					line,
					"",
					"",
					new Observation[]{}
				);

				if (!dispatch(timeSeries)) break;
			}

			br.close();
		} catch ( final IOException e ) {
			final String desc
			= "Error in reading from a file " + filename;
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"getExpectedKeys",
				desc,
				e
			);

			dispatch(
				new ErrorEvent( desc, ErrorEvent.FAILURE )
			);
		}

		dispatch( new MessageEnd( MREF ) );
		dispatch( new InterchangeEnd( IREF ) );
	}

	public void close () {
		componentHelper.logger.log(
			Level.FINE,
			componentHelper.logName,
			"close",
			"Closing."
		);
	}
}
