// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2003, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.csv;

import java.lang.String;
import java.lang.Number;
import java.lang.IllegalArgumentException;

import java.util.Calendar;
import java.util.Date;

import java.text.Format;
import java.text.FieldPosition;
import java.text.ParsePosition;
import java.text.ParseException;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.text.DecimalFormat;

import Int.ecb.gesmescb.time.RegularInterval;
import Int.ecb.gesmescb.time.GregorianRegularInterval;

/* ***************************** Class ****************************** */

/**
 * Formats {@link RegularInterval}s.
 *
 * @author
 * <A HREF="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</A>
 *
 * @version @(#)$Id: RegularIntervalFormat.java,v 1.1 2017/01/19 16:14:04 trainot Exp $
 */
public class RegularIntervalFormat extends Format {

	/* ======================= Constants ======================== */

	/**
	 * generated by eclipse.
	 */
	private static final long serialVersionUID = 8458479653761413860L;

	/** Field number for the year specification.
	 * @see FieldPosition */
	public static final int YEAR_FIELD = 0x1;

	/** Field number for the quarter field.
	 * @see FieldPosition */
	public static final int QUARTER_FIELD = 0x2;

	/** Field number for the month field.
	 * @see FieldPosition */
	public static final int MONTH_FIELD = 0x4;

	/** Field number for the quarter or month field, whichever present.
	 * @see FieldPosition */
	public static final int QUARTER_OR_MONTH_FIELD
	= QUARTER_FIELD | MONTH_FIELD;

	// TODO: Implement the field only formatting.

	/* ========================= Fields ========================= */

	private transient DateFormat monthlyFormat, quarterlyFormat, yearlyFormat;

	/* ====================== Constructors ====================== */

	/**
	 * Construct a RegularIntervalFormat.
	 */
	public RegularIntervalFormat () {
		monthlyFormat = quarterlyFormat = yearlyFormat = null;
	}

	/* ===================== Public Methods ===================== */

	/* --------------------- Format Methods --------------------- */

	public final StringBuffer format (
		final Object object,
		final StringBuffer toAppendTo,
		final FieldPosition position
	) {
		if ( object instanceof RegularInterval ) {
			return format( (RegularInterval) object, toAppendTo, position );
		} else {
			throw new IllegalArgumentException(
				"The object is not a RegularInterval:" + object
			);
		}
	}

	/**
	 * Compute the full SPF representation
	 * of a <code>RegularInterval</code>.
	 *
	 * @return <code>toAppendTo</code>
	 *
	 * @throws IllegalArgumentException if the <code>interval</code>
	 * is of a frequency or size not supported.
	 */
	public StringBuffer format (
		final RegularInterval interval,
		final StringBuffer toAppendTo,
		final FieldPosition position
	) throws IllegalArgumentException {
		if ( interval.getSize() != 1 ) {
			throw new IllegalArgumentException(
				getClass().getName()
				+ " cannot translate " + interval
				+ " because it is not a single base period."
			);
		}
		final Date begin = interval.getBegin();
		switch ( interval.getFrequency() ) {
			case RegularInterval.ANNUAL:
				return getYearlyFormat().format(
					begin, toAppendTo, position
				);
			case RegularInterval.QUARTERLY:
				return getQuarterlyFormat().format(
					begin, toAppendTo, position
				);
			case RegularInterval.MONTHLY:
				return getMonthlyFormat().format(
					begin, toAppendTo, position
				);
			default:
				throw new IllegalArgumentException(
					getClass().getName()
					+ " cannot translate " + interval
					+ " due to unsupported frequency."
				);
		}
	}

	/* --------------------- Parse Methods ---------------------- */

	/**
	 * <p>Parse a <code>RegularInterval</code> from SPF CSV format.</p>
	 *
	 * @see #parse(String,ParsePosition)
	 */
	public final Object parseObject(
		final String text,
		final ParsePosition position
	) {
		return parse( text, position );
	}

	/**
	 * <p>Parse a <code>RegularInterval</code> from SPF CSV format.</p>
	 *
	 * @param text to parse.
	 * @param position tracker for parsing process.
	 *
	 * @return the new <code>RegularInterval</code>
	 * or <code>null</code> if parsing fails.
	 */
	public RegularInterval parse (
		final String text,
		final ParsePosition position
	) {
		final int hyphen = text.indexOf( '-', position.getIndex() );
		final int quarterQ = text.indexOf( 'Q', position.getIndex() );
		final Date begin;
		final int frequency;
		if ( hyphen >= 0 ) {
			begin = getMonthlyFormat().parse( text, position );
			frequency = RegularInterval.MONTHLY;
		} else if ( quarterQ >= 0 ) {
			begin = getQuarterlyFormat().parse( text, position );
			frequency = RegularInterval.QUARTERLY;
		} else {
			begin = getYearlyFormat().parse( text, position );
			frequency = RegularInterval.ANNUAL;
		}
		if ( position.getErrorIndex() < 0 ) {
			return new GregorianRegularInterval( frequency, begin, 1 );
		} else {
			return null;
		}
	}

	/**
	 * Parse a <code>RegularInterval</code> from SPF format.
	 * Parsing starts at position 0.
	 * @see #parse(String,ParsePosition)
	 */
	public final RegularInterval parse(
		final String text
	) throws ParseException {
		final ParsePosition position = new ParsePosition( 0 );
		final RegularInterval interval = parse( text, position );
		if ( position.getErrorIndex() >= 0 ) {
			throw new ParseException(
				"Cannot parse \"" + text + "\" to a time interval.",
				position.getErrorIndex()
			);
		}
		if ( position.getIndex() < text.length() ) {
			throw new ParseException(
				"Unparsed text after time period string\""
				+ text + "\".",
				position.getIndex()
			);
		}
		return interval;
	}

	/* ==================== Private Methods ===================== */

	private DateFormat getMonthlyFormat () {
		if ( monthlyFormat == null ) {
			monthlyFormat = new SimpleDateFormat( "MMM-yyyy" );
		}
		return monthlyFormat;
	}

	private DateFormat getQuarterlyFormat () {
		if ( quarterlyFormat == null ) {
			quarterlyFormat = new QuarterlyFormat();
		}
		return quarterlyFormat;
	}

	private DateFormat getYearlyFormat () {
		if ( yearlyFormat == null ) {
			yearlyFormat = new SimpleDateFormat( "yyyy" );
		}
		return yearlyFormat;
	}
}

/* ***************************** Class ****************************** */

/** Formats quarterly time points. */
class QuarterlyFormat extends DateFormat {
	/**
	 * generated by eclipse.
	 */
	private static final long serialVersionUID = -3641045677303752417L;

	public QuarterlyFormat () {
		setNumberFormat( new DecimalFormat( "####" ) );
	}

	public StringBuffer format (
		final Date date,
		final StringBuffer appendTo,
		final FieldPosition position
	) {
		initCalendar();
		calendar.setTime( date );
		final int month = calendar.get( Calendar.MONTH );
		if ( month % 3 != 0 ) {
			throw new IllegalArgumentException(
				date + " is not at the start of a quarter."
			);
		}
		// TODO: Check beginning of month.
		final int year = calendar.get( Calendar.YEAR );
		final int quarter = month / 3 + 1;
		appendTo.append( year );
		appendTo.append( 'Q' );
		appendTo.append( quarter );
		return appendTo;
	}

	public Date parse (
		final String text,
		final ParsePosition position
	) {
		initCalendar();
		final ParsePosition tentativePosition = new ParsePosition(
			position.getIndex()
		);
		final Number yearNumber = numberFormat.parse( text, tentativePosition );
		if ( yearNumber != null ) {
			calendar.clear();
			calendar.set( Calendar.YEAR, yearNumber.intValue() );
			final String quarterCode
				= tentativePosition.getIndex() + 2 < text.length()
				? text.substring(
					tentativePosition.getIndex(),
					tentativePosition.getIndex() + 2
				) : text.substring( tentativePosition.getIndex() );
			if ( "Q1".equals( quarterCode ) ) {
				calendar.set( Calendar.MONTH, Calendar.JANUARY );
				position.setIndex( tentativePosition.getIndex() + 2 );
				return calendar.getTime();
			} else if ( "Q2".equals( quarterCode ) ) {
				calendar.set( Calendar.MONTH, Calendar.APRIL );
				position.setIndex( tentativePosition.getIndex() + 2 );
				return calendar.getTime();
			} else if ( "Q3".equals( quarterCode ) ) {
				calendar.set( Calendar.MONTH, Calendar.JULY );
				position.setIndex( tentativePosition.getIndex() + 2 );
				return calendar.getTime();
			} else if ( "Q4".equals( quarterCode ) ) {
				calendar.set( Calendar.MONTH, Calendar.OCTOBER );
				position.setIndex( tentativePosition.getIndex() + 2 );
				return calendar.getTime();
			} else {
				position.setErrorIndex( tentativePosition.getIndex() );
			}
		} else {
			position.setErrorIndex( tentativePosition.getErrorIndex() );
		}
		return null;
	}

	/* ==================== Private Methods ===================== */

	private void initCalendar () {
		if ( getCalendar() == null ) {
			setCalendar( Calendar.getInstance() );
		}
	}
}
