// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2003, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.csv;

import java.text.ParsePosition;
import java.text.ParseException;

import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;

import Int.ecb.gesmescb.time.RegularInterval;
import Int.ecb.gesmescb.time.GregorianRegularInterval;

/* ***************************** Class ****************************** */

/**
 * <p>Tests the implementation of {@link RegularIntervalFormat}.</p>
 *
 * @author
 * <A HREF="mailto:Sami.Airo@ECB.int">Sami Airo</A>
 * @author
 * <A HREF="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</A>
 *
 * @version @(#)$Id: RegularIntervalFormatTest.java,v 1.1 2017/01/19 16:14:04 trainot Exp $
 */
public class RegularIntervalFormatTest extends TestCase {

	/* ====================== Constructors ====================== */

	/**
	 * Constructs a RegularIntervalFormatTest with the specified name.
	 *
	 * @param name Test case name
	 */
	public RegularIntervalFormatTest ( final String name ) {
		super( name );
	}

	/* ===================== Public Methods ===================== */

	/**
	 * Assembles and returns a test suite for all
	 * the test methods of this test case.
	 *
	 * @return a non-null test suite
	 */
	public static Test suite () {
		return new TestSuite( RegularIntervalFormatTest.class );
	}

	/**
	 * Runs the test case.
	 */
	public static void main ( final String args[] ) {
		final String[] testCaseName = {
			RegularIntervalFormatTest.class.getName()
		};
		junit.textui.TestRunner.main( testCaseName );
	}

	/* --------------------- Test Methods --------------------- */

	public void testMonthly () throws ParseException {
		final RegularIntervalFormat format = new RegularIntervalFormat();

		final String text = "Sep-2001";
		final RegularInterval interval = new GregorianRegularInterval(
			RegularInterval.MONTHLY, 2001, 8, 1
		);

		assertEquals( text, format.format( interval ) );
		assertEquals( interval, format.parse( text ) );
	}

	public void testQuarterly () throws ParseException {
		final RegularIntervalFormat format = new RegularIntervalFormat();

		final String text = "2001Q2";
		final RegularInterval interval = new GregorianRegularInterval(
			RegularInterval.QUARTERLY, 2001, 1, 1
		);

		assertEquals( text, format.format( interval ) );
		assertEquals( interval, format.parse( text ) );
	}

	public void testYearly () throws ParseException {
		final RegularIntervalFormat format = new RegularIntervalFormat();

		final String text = "2001";
		final RegularInterval interval = new GregorianRegularInterval(
			RegularInterval.ANNUAL, 2001, 0, 1
		);

		assertEquals( text, format.format( interval ) );
		assertEquals( interval, format.parse( text ) );
	}

	public void testValidation () {
		validFormatTest( "Jan-1999" );
		validFormatTest( "Feb-1949" );
		validFormatTest( "Mar-1766" );
		validFormatTest( "1766Q3" );
		validFormatTest( "2000Q1" );
		validFormatTest( "1998Q4" );
		validFormatTest( "1998" );
		validFormatTest( "2000" );
		// Be warned: Those are valid.
		validFormatTest( "37266" );
		validFormatTest( "19802" );
		// validFormatTest( "Mar-02" ); backconversion to Mar-0002!

		invalidFormatTest( "2000X" );
		invalidFormatTest( "20Q0" );
		invalidFormatTest( "2002Q0" );
		invalidFormatTest( "2002Q01" );
		invalidFormatTest( "2002Q10" );
		invalidFormatTest( "2002Q1X" );
		invalidFormatTest( "2002Q" );
		invalidFormatTest( "Jan 2000" );
		invalidFormatTest( "Sept. 1999" );
		invalidFormatTest( "Sep. 1999" );
		invalidFormatTest( "2003q1" );
		invalidFormatTest( "Mar 02" );
		invalidFormatTest( "Mar. 02" );
		invalidFormatTest( "Mar2002" );
		invalidFormatTest( "1980-02" );
	}

	/* ==================== Private Methods ===================== */

	private void validFormatTest ( final String text ) {
		final RegularIntervalFormat format = new RegularIntervalFormat();
		final ParsePosition position = new ParsePosition( 0 );
		final RegularInterval interval = format.parse( text, position );
		assertEquals( "index on end", text.length(), position.getIndex() );
		assertEquals( "no errorIndex", -1, position.getErrorIndex() );
		assertEquals( text, format.format( interval ) );
	}

	private void invalidFormatTest ( final String text ) {
		final RegularIntervalFormat format = new RegularIntervalFormat();
		final ParsePosition position = new ParsePosition( 0 );
		final RegularInterval interval = format.parse( text, position );
		if ( position.getIndex() >= text.length() ) {
			assertNull(
				"Invalid format \"" + text
				+ "\" must yield null but gives " + interval + ".",
				interval
			);
			assertTrue(
				"Failed to set errorIndex for invalid format \"" + text + "\".",
				position.getErrorIndex() >= 0
			);
			assertEquals(
				"Consumed positions in invalid format \"" + text + "\".",
				0,
				position.getIndex()
			);
		}
		try {
			format.parse( text );
			fail(
				"Parse of complete text did not fail with \""
				+ text + "\"."
			);
		}
		catch ( final ParseException pe ) {}
	}
}
