// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2003, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.csv;

import java.util.List;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.Date;
import java.util.Calendar;

import java.text.SimpleDateFormat;
import java.text.ParseException;

import java.io.File;
import java.io.Reader;
import java.io.FileReader;
import java.io.IOException;

import Int.ecb.logging.Level;

import Int.ecb.gesmescb.*;

import Int.ecb.gesmescb.time.RegularInterval;
import Int.ecb.gesmescb.time.GregorianRegularInterval;

import Int.ecb.gesmescb.event.ErrorEvent;
import Int.ecb.gesmescb.event.GESMESCBReader;
import Int.ecb.gesmescb.event.GESMESCBEventSourceAdapter;

import Int.ecb.gesmescb.util.Named;
import Int.ecb.gesmescb.util.ComponentHelper;

/* ***************************** Class ****************************** */

/**
 * <p>Parses CSV in SPF syntax to
 * {@link Int.ecb.gesmescb.GESMESCBObject}s.</p>
 *
 * <center><table border="3" cellpadding="3">
 * <caption>
 * <a name="prefs">Available {@link Int.ecb.prefs preference} settings</a>
 * </caption>
 * <tr>
 * <th>preference name</th>
 * <th>type</th>
 * <th>meaning</th>
 * <th>default</th>
 * </tr>
 * <tr><td><code>source</td><td>string</td><td>input file path</td></tr>
 * <tr>
 *	<td><code>iref</td><td>string</td>
 *	<td>interchange reference number</td>
 *	<td><code>IREF000000</code></td>
 * </tr>
 * <tr>
 *	<td><code>sender</td><td>string</td>
 *	<td>sender organisation code</td>
 *	<td><code>ZZZ</code></td>
 * </tr>
 * <tr>
 *	<td><code>receiver</td><td>string</td>
 *	<td>receiver organisation code</td>
 *	<td><code>ZZZ</code></td>
 * </tr>
 * <tr>
 *	<td><code>test</td><td>boolean</td>
 *	<td>whether to flag data as test data</td>
 *	<td><code>false</code></td>
 * </tr>
 * <tr>
 *	<td><code>maintenanceAgency</td><td>string</td>
 *	<td>the key family maintenance agency</td>
 *	<td><code>ECB</code></td>
 * </tr>
 * <tr>
 *	<td><code>keyFamily</td><td>string</td>
 *	<td>the identifier of the key family structuring the data</td>
 *	<td><code>ECB_FCT1</code></td>
 * </tr>
 * <tr>
 *	<td><code>dataSet</td><td>string</td>
 *	<td>the identifier of the data set</td>
 *	<td><code>ECB_SPF</code></td>
 * </tr>
 * <tr>
 *	<td><code>generateAttributes</td><td>boolean</td>
 *	<td>whether to produce series/sibling group level attributes</td>
 *	<td><code>false</code></td>
 * </tr>
 * </table></center>
 *
 * @author
 * <a href="mailto:sis.external@ecb.int">Bernhard Bodenstorfer</a>
 * @author
 * <a href="mailto:sis.external@ecb.int">Ana Garcia Padrao</a>
 *
 * @version @(#)$Id: SPFReader.java,v 1.2 2018/12/17 09:35:17 kortest Exp $
 */
public class SPFReader
	extends GESMESCBEventSourceAdapter
	implements GESMESCBReader, Named
{
	/* ========================= Fields ========================= */

	/**
	 * Takes care of logging and preferences.
	 */
	protected final ComponentHelper componentHelper;

	/**
	 * The name of the file read.
	 * This can be used for log messages.
	 */
	private final String filename;

	/** Reads lines from the input file. */
	private final Reader reader;
	private final CSVTokenizer tokenizer;
	private CSVTokenizer.Line line;

	private final List blockDescriptors;
	private final List columnDescriptors;
	private final SPFObservationFactory observationFactory;

	private final ColumnDescriptor emptyColumnDescriptor;

	/* ====================== Constructors ====================== */

	/**
	 * Construct a <code>SPFReader</code> to read from a file.
	 *
	 * @param name used for logging and preferences.
	 * @param filename where to read from.
	 *
	 * @throws NullPointerException if <code>filename==null</code>
	 * and the preference variable
	 * <code>source</code> is undefined.
	 * @throws IOException if the file cannot be opened for reading.
	 */
	public SPFReader ( final String name, final String filename )
		throws IOException
	{
		componentHelper = new ComponentHelper( name, this );
		this.filename = filename != null
			? filename
			: componentHelper.getPreference( "source", null );
		if ( this.filename == null ) {
			final String desc
			= "No preference variable \"source\" found"
				+ " to determine the input file.";
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"SPFReader()",
				desc
			);
			throw new NullPointerException( desc );
		}

		reader = new FileReader( this.filename );
		tokenizer = new CSVTokenizer( reader );
		blockDescriptors = new ArrayList();
		columnDescriptors = new ArrayList();
		observationFactory = new SPFObservationFactory( this );
		emptyColumnDescriptor = new EmptyColumnDescriptor( observationFactory );
	}

	/**
	 * Construct a <code>SPFReader</code>
	 * reading from a file according to preferences.
	 *
	 * @param name used for logging and preferences.
	 *
	 * @throws NullPointerException if the preference variable
	 * <code>source</code> is undefined.
	 * @throws IOException if the file cannot be opened for reading.
	 */
	public SPFReader ( final String name )
		throws IOException
	{
		this( name, null );
	}

	/* ===================== Public Methods ===================== */

	public String getName () {
		return componentHelper.name;
	}

	/**
	 * Parse a GESMES/TS interchange in CSV syntax used by the SPF project into
	 * {@link Int.ecb.gesmescb.GESMESCBObject}s.
	 * These then are dispatched to all listeners.
	 */
	public void run () {
		try {
			componentHelper.logger.log(
				Level.INFO,
				componentHelper.logName,
				"run()",
				"Starting to parse SPF input from \""
				+ filename
				+ "\"."
			);

			final String iref
			= componentHelper.getPreference( "iref", "IREF000000" );
			final String sender
			= componentHelper.getPreference( "sender", "4F0" );
			final String receiver
			= componentHelper.getPreference( "receiver", "4F0" );
			final String prepared
			= new SimpleDateFormat( "yyMMdd:HHmm" ).format( new Date() );
			final boolean test
			= ! "false".equalsIgnoreCase(
				componentHelper.getPreference( "test", "false" )
			);
			final InterchangeBegin interchangeBegin = new InterchangeBegin(
				iref, sender, receiver, prepared, test
			);
			componentHelper.logger.log(
				Level.FINE,
				componentHelper.logName,
				"run()",
				"Dispatching InterchangeBegin",
				interchangeBegin
			);
			dispatch( interchangeBegin );

			final String maintenanceAgencyId
			= componentHelper.getPreference( "maintenanceAgency", "ECB" );
			final String keyFamilyId
			= componentHelper.getPreference( "keyFamily", "ECB_FCT1" );
			final File sourceFile = new File( filename );
			final String subject = sourceFile.getName();
			final String dataSetId
			= componentHelper.getPreference( "dataSet", "ECB_SPF" );
			final String extracted
			= new SimpleDateFormat( "yyyyMMddHHmm" ).format(
				new Date( sourceFile.lastModified() )
			);
			final MessageBegin messageBegin = new MessageBegin(
				"MREF000001", sender, receiver, maintenanceAgencyId,
				MessageBegin.STATISTICAL_DATA,
				subject,
				new Contact[]{},
				dataSetId,
				MessageBegin.UPDATE_OR_REPLACE_ACTION,
				extracted,
				"", "",
				keyFamilyId
			);
			componentHelper.logger.log(
				Level.FINE,
				componentHelper.logName,
				"run()",
				"Dispatching MessageBegin",
				messageBegin
			);
			dispatch( messageBegin );

			// Process first line
			if ( tokenizer.hasMoreLines() ) {
				line = tokenizer.nextLine();
				{
					final String cell = line.nextCell();
					if ( "SPF".equals( cell ) ) {
						// o.k.
					} else if ( "Survey".equals( cell )	) {
						warn(
							"First line starts with \"" + cell
							+ "\", which is deprecated, but still allowed"
							+ " for backward compatibility for some time."
							+ " First cell content in new files"
							+ " should be \"SPF\"."
						);
					} else {
						complain(
							"Parsing problem: First line starts with \"" + cell
							+ "\", but first cell content should be \"SPF\"."
						);
					}
				}

				{
					final String cell = line.nextCell();
					observationFactory.setSurveyDate( cell );
				}

				while ( line.hasMoreCells() ) {
					final String cell = line.nextCell();
					if ( "".equals( cell ) ) {
					} else if ( "HICP".equals( cell ) ) {
						blockDescriptors.add (
							new BlockDescriptor( "HICP", line.getCellNo() )
						);
					} else if ( "GDP".equals( cell ) ) {
						blockDescriptors.add (
							new BlockDescriptor( "RGDP", line.getCellNo() )
						);
					} else if ( "Unemployment".equals( cell ) ) {
						blockDescriptors.add (
							new BlockDescriptor( "UNEM", line.getCellNo() )
						);
					} else if ( "Assumptions".equals( cell ) ) {
						blockDescriptors.add (
							new BlockDescriptor( "ASSU", line.getCellNo() )
						);
					} else if ( "HICP Core".equals( cell ) ) {
						blockDescriptors.add(
							new BlockDescriptor( "CORE", line.getCellNo() ) 
						);
					}
						else {
						complain(
							"Parsing problem: First line contains \"" + cell
							+ "\" but should only contain"
							+ " HICP, GDP, Unemployment,"
							+ " Assumptions, HICP Core or empty cells."
						);
					}
				}
				// TODO: Check that all four appear exactly once?
			}

			// Process second line
			if ( tokenizer.hasMoreLines() ) {
				line = tokenizer.nextLine();
				if ( line.hasMoreCells() ) {
					final String cell = line.nextCell();
					if ( !"Period".equals( cell ) ) {
						complain(
							"Parsing problem: Second line starts with \"" + cell
							+ "\" but should contain \"Period\" there."
						);
					} else {
						columnDescriptors.add(
							new PeriodCodeColumnDescriptor(
								observationFactory
							)
						);
					}
				}

				if ( line.hasMoreCells() ) {
					final String cell = line.nextCell();
					if ( !"Forecaster".equals( cell ) ) {
						complain(
							"Parsing problem: Second line starts with \"" + cell
							+ "\" but should contain \"Forecaster\" there."
						);
					} else {
						columnDescriptors.add(
							new ForecasterColumnDescriptor(
								observationFactory
							)
						);
					}
				}

				final Iterator blocks = blockDescriptors.iterator();
				BlockDescriptor
					currentBlock = null,
					nextBlock = blocks.hasNext()
						? (BlockDescriptor) blocks.next()
						: null;
				boolean firstColumnOfBlock = false;
				while ( line.hasMoreCells() ) {
					final String cell = line.nextCell();
					boolean isDateColumn = false;
					if (
						nextBlock != null
						&&
						line.getCellNo() >= nextBlock.startColumn
					) {
						currentBlock = nextBlock;
						nextBlock = blocks.hasNext()
							? (BlockDescriptor) blocks.next()
							: null;
						firstColumnOfBlock = true;
					}
					if ( "".equals( cell ) ) {
						if ( firstColumnOfBlock ) {
							columnDescriptors.add( new DateColumnDescriptor(
								observationFactory
							) );
							isDateColumn = true;
						} else {
							columnDescriptors.add( new EmptyColumnDescriptor(
								observationFactory
							) );
						}
					} else if ( "Sum".equals( cell ) ) {
						columnDescriptors.add(
							new ProbabilityColumnDescriptor(
								observationFactory,
								currentBlock,
								"SUM",
								"PC"
							)
						);
					} else if (
						"Point".equals( cell )
						||
						"ir".equals( cell )
						||
						"lab".equals( cell )
					) {
						columnDescriptors.add(
							new PercentageColumnDescriptor(
								observationFactory,
								currentBlock,
								cell.toUpperCase(),
								"UNEM".equals( currentBlock.topic )
								? "PC" : "PCPA"
							)
						);
					} else if (
						"oil".equals( cell )
						||
						"usd".equals( cell )
					) {
						columnDescriptors.add(
							new PlainColumnDescriptor(
								observationFactory,
								currentBlock,
								cell.toUpperCase(),
								"USD"
							)
						);
					} else {
						columnDescriptors.add(
							new ProbabilityColumnDescriptor(
								observationFactory,
								currentBlock,
								intervalBreakDownCode( cell ),
								"PC"
							)
						);
					}
					if (
						currentBlock != null
						&&
						currentBlock.startColumn == line.getCellNo()
					) {
						if ( ! isDateColumn ) {
							warn(
								"Only the date column should be below"
								+ " a topic heading in the first line but there"
								+ " was found \"" + cell + "\"."
							);
						}
					} else {
						if ( isDateColumn ) {
							warn(
								"The date column should be below"
								+ " the topic heading in the first line."
							);
						}
					}
					firstColumnOfBlock = false;
				}
			}

			// Process lines after second
			while ( tokenizer.hasMoreLines() ) {
				line = tokenizer.nextLine();

				final Iterator descriptors = columnDescriptors.iterator();

				while (
					descriptors.hasNext()
					||
					line.hasMoreCells()
				) {
					final ColumnDescriptor descriptor = descriptors.hasNext()
						? (ColumnDescriptor) descriptors.next()
						: emptyColumnDescriptor;
					final String cell = line.hasMoreCells()
						? line.nextCell()
						: "";
					descriptor.processValue( cell );
				}
			}

			final MessageEnd messageEnd = new MessageEnd( "MREF000001" );
			componentHelper.logger.log(
				Level.FINE,
				componentHelper.logName,
				"run()",
				"Dispatching MessageEnd",
				messageEnd
			);
			dispatch( messageEnd );

			final InterchangeEnd interchangeEnd = new InterchangeEnd(
				iref
			);
			componentHelper.logger.log(
				Level.FINE,
				componentHelper.logName,
				"run()",
				"Dispatching InterchangeEnd",
				interchangeEnd
			);
			dispatch( interchangeEnd );

			componentHelper.logger.log(
				Level.INFO,
				componentHelper.logName,
				"run()",
				"Ending to parse SPF input from \""
				+ filename
				+ "\"."
			);
		}
		catch ( final Exception e ) {
			final String desc = "Problem: " + e.getMessage() + ".";

			componentHelper.logger.log(
				Level.SEVERE,
				getName(),
				"run()",
				desc,
				e
			);
			dispatch(
				new ErrorEvent( desc, ErrorEvent.ERROR )
			);
		}
	}

	/**
	 * Close the <code>SPFReader</code>.
	 */
	public void close () {
		try {
			componentHelper.logger.log(
				Level.FINE,
				componentHelper.logName,
				"close()",
				"Closing SPF input file \""
				+ filename
				+ "\"."
			);
			reader.close();
		}
		catch ( final Exception e ) {
			final String desc = "Problem: " + e.getMessage() + ".";

			componentHelper.logger.log(
				Level.SEVERE,
				getName(),
				"close()",
				desc,
				e
			);
			dispatch(
				new ErrorEvent( desc, ErrorEvent.ERROR )
			);
		}
	}

	/* ==================== Package Methods ===================== */

	/** Raise severe error from {@link #run()}. */
	void complain ( final String description ) {
		final String desc = "File \"" + filename
			+ ( line != null
				? "\" line " + line.getLineNo() + " cell " + line.getCellNo()
				: ""
			) + ": "
			+ description;
		componentHelper.logger.log(
			Level.SEVERE,
			componentHelper.logName,
			"run()",
			desc
		);
		dispatch(
			new ErrorEvent( desc, ErrorEvent.ERROR )
		);
	}

	void warn ( final String description ) {
		final String desc = "File \"" + filename
			+ ( line != null
				? "\" line " + line.getLineNo() + " cell " + line.getCellNo()
				: ""
			) + ": "
			+ description;
		componentHelper.logger.log(
			Level.WARNING,
			componentHelper.logName,
			"run()",
			desc
		);
		dispatch(
			new ErrorEvent( desc, ErrorEvent.WARNING )
		);
	}

	/** Lets package members dispatch. */
	void dispatchSeries( final TimeSeries timeSeries ) {
		componentHelper.logger.log(
			Level.FINER,
			componentHelper.logName,
			"dispatchSeries()",
			"Dispatching TimeSeries",
			timeSeries
		);
		dispatch( timeSeries );
	}

	/** Lets package members dispatch. */
	void dispatchAttribute( final Attribute attribute ) {
		componentHelper.logger.log(
			Level.FINER,
			componentHelper.logName,
			"dispatchAttribute()",
			"Dispatching Attribute",
			attribute
		);
		dispatch( attribute );
	}

	/* ==================== Private Methods ===================== */

	private String intervalBreakDownCode ( final String cell ) {
		final StringBuffer buffer = new StringBuffer();
		final int length = cell.length();
		boolean replacedHyphen = false;
		for ( int i = 0; i < length; ++i ) {
			final char c = cell.charAt( i );
			switch ( c ) {
				case '<': buffer.append( 'T' ); break;
				case '>': buffer.append( 'F' ); break;
				case '.': buffer.append( '_' ); break;
				case '-':
					if ( i > 0 && ! replacedHyphen ) {
						buffer.append( 'T' );
						replacedHyphen = true;
					} else {
						buffer.append( 'M' );
					}
				break;
				// case N is a hack to circumvent our non-understanding 
				// of this method. Good that the original data already use N
				// instead of hyphen for negatives.  
				case 'N': 
					if ( i == 0 ) {
						buffer.append( 'F' );
					}
					buffer.append( 'N' );
				break;
				case ' ':
				case '%':
					warn(
						"Interval breakdown \"" + cell
						+ "\" contains deprecated character (blank or %)."
					);
				break;
				case '0':
				case '1':
				case '2':
				case '3':
				case '4':
				case '5':
				case '6':
				case '7':
				case '8':
				case '9':
					if ( i == 0 ) {
						buffer.append( 'F' );
					}
					buffer.append( c );
					break;
				default:
					complain(
						"Character '" + c
						+ "' not allowed in breakdown code \"" + cell
						+ "\"."
					);
			}
		}
		return buffer.toString();
	}
}

/* ***************************** Class ****************************** */

class SPFObservationFactory {

	/* ========================= Fields ========================= */

	private final SPFReader reader;
	private final RegularIntervalFormat periodFormat;

	private final Calendar calendar;
	private RegularInterval surveyMonth, surveyQuarter;
	private int surveyQuarterIndex;

	/** Used to successively construct the observations. */
	private String
		surveyMonthText,
		periodText,
		frequency,
		refArea,
		topic,
		breakDown,
		horizon,
		surveyFrequency,
		forecaster,
		obsStatus,
		obsConf;

	private RegularInterval period;

	private final boolean generateAttributes;

	/* ====================== Constructors ====================== */

	public SPFObservationFactory ( final SPFReader reader ) {
		this.reader = reader;
		periodFormat = new RegularIntervalFormat();
		surveyMonthText = periodText = "";
		refArea = "U2";
		surveyFrequency = "Q";
		generateAttributes = "true".equalsIgnoreCase(
			reader.componentHelper.getPreference(
				"generateAttributes",
				"false"
			)
		);
		reader.componentHelper.logger.log(
			Level.CONFIG,
			reader.componentHelper.logName,
			"SPFObservationFactory()",
			"Generate attributes: " + generateAttributes
		);
		calendar = Calendar.getInstance();
	}

	/* ===================== Public Methods ===================== */

	public void setSurveyDate ( final String surveyDate ) {
		surveyMonthText = surveyDate;
		try {
			reader.componentHelper.logger.log(
				Level.FINE,
				reader.componentHelper.logName,
				"setSurveyDate()",
				"Survey date: " + surveyDate
			);
			surveyMonth = periodFormat.parse( surveyDate );
			surveyQuarter = new GregorianRegularInterval(
				RegularInterval.QUARTERLY,
				surveyMonth.getFirstPeriodYear(),
				surveyMonth.getFirstPeriodIndex() / 3,
				1
			);
			reader.componentHelper.logger.log(
				Level.FINE,
				reader.componentHelper.logName,
				"setSurveyDate()",
				"Survey quarter begins at: " + surveyQuarter.getBegin()
			);
			calendar.setTime( surveyQuarter.getBegin() );
			surveyQuarterIndex
			= 4 * calendar.get( Calendar.YEAR )
			+ calendar.get( Calendar.MONTH ) / 3;
		}
		catch ( final ParseException pe ) {
			complain(
				"Problem with survey cycle date: " + pe
			);
		}
	}

	public void setTopic ( final String topic ) {
		this.topic = topic;
	}

	public void setBreakDown ( final String breakDown ) {
		this.breakDown = breakDown;
	}

	/** Set the forecast interval in SPF style.
	 * 2002 ... year; 2002Q3 ... quarter; Sept-2002 ... month.
	 */
	public void setInterval ( final String interval ) {
		periodText = interval;
		if (
			"".equals( interval )
			||
			"X".equals( interval )
		) {
			period = null;
		} else {
			try {
				period = periodFormat.parse( interval );
				calendar.setTime( period.getEnd() );
				switch ( period.getFrequency() ) {
					case RegularInterval.ANNUAL:
						frequency = "A";
						break;
					case RegularInterval.QUARTERLY:
						frequency = "Q";
						break;
					case RegularInterval.MONTHLY:
						frequency = "M";
						calendar.add(
							Calendar.MONTH,
							2 - ( calendar.get( Calendar.MONTH ) + 2 ) % 3
						);
						break;
					default:
						complain(
							"In cell \"" + interval + "\":"
							+ " Frequencies other than annual, quarterly, monthly"
							+ " are not supported."
						);
				}
				int index
				= 4 * calendar.get( Calendar.YEAR )
				+ calendar.get( Calendar.MONTH ) / 3
				- surveyQuarterIndex;
				final StringBuffer buffer = new StringBuffer();
				if ( index <= 0 ) {
					index = -index;
					buffer.append( "M" );
					obsStatus = "A";
				} else {
					buffer.append( "P" );
					obsStatus = "F";
				}
				buffer.append( 3 * index );
				buffer.append( 'M' );
				horizon = buffer.toString();
			}
			catch ( final ParseException pe ) {
				complain(
					"\"" + interval + "\" is not a valid time period: "
					+ pe
				);
			}
		}
	}

	public void setForecaster ( final String forecaster ) {
		final int length = forecaster.length();
		switch ( length ) {
			case 0: this.forecaster = null; break;
			case 1: this.forecaster = "00" + forecaster; break;
			case 2: this.forecaster = "0" + forecaster; break;
			case 3: this.forecaster = forecaster; break;
			default:
				complain(
					"Bad forecaster id \"" + forecaster + "\"."
				);
		}
		if ( "ECB".equalsIgnoreCase( forecaster ) ) {
			obsConf = "F";
		} else {
			obsConf = "C";
		}
	}

	/** Dispatch a {@link TimeSeries} with given observation value. */
	public void dispatchValue ( final String value ) {
		if (
			"ECB".equalsIgnoreCase( forecaster )
			&& (
				"HICP".equals( topic ) && ! "M0M".equals( horizon )
				||
				"RGDP".equals( topic ) && ! "M3M".equals( horizon )
				||
				"UNEM".equals( topic ) && ! "M0M".equals( horizon )
				||
				"CORE".equals( topic )  && ! "M0M".equals( horizon )
			)
		) {
			complain(
				"Survey month " + surveyMonthText
				+ " does not fit " + topic
				+ " reference period " + periodText + "."
			);
		}
		final TimeSeries timeSeries = new TimeSeries(
			frequency + TimeSeries.KEY_SEPARATOR
			+ refArea + TimeSeries.KEY_SEPARATOR
			+ topic + TimeSeries.KEY_SEPARATOR
			+ breakDown + TimeSeries.KEY_SEPARATOR
			+ horizon + TimeSeries.KEY_SEPARATOR
			+ surveyFrequency + TimeSeries.KEY_SEPARATOR
			+ forecaster,
			period,
			new Observation[]{
				new Observation(
					value,
					"-".equals( value ) ? "M" : obsStatus,
					obsConf,
					""
				)
			}
		);
		reader.dispatchSeries( timeSeries );
	}

	/** Dispatch a UNIT {@link Attribute} with value. */
	public void dispatchUnit ( final String value ) {
		if ( generateAttributes ) {
			final Attribute attribute = new Attribute(
				TimeSeries.KEY_SEPARATOR
				+ refArea + TimeSeries.KEY_SEPARATOR
				+ topic + TimeSeries.KEY_SEPARATOR
				+ breakDown + TimeSeries.KEY_SEPARATOR
				+ horizon + TimeSeries.KEY_SEPARATOR
				+ surveyFrequency + TimeSeries.KEY_SEPARATOR
				+ forecaster,
				"", "",
				"UNIT",
				value
			);
			reader.dispatchAttribute( attribute );
		}
	}

	public boolean isCubeDefined () {
		return forecaster != null && period != null;
	}

	/* =================== Protected Methods ==================== */

	protected void complain ( final String description ) {
		reader.complain( description );
	}

	protected void warn ( final String description ) {
		reader.warn( description );
	}
}

/* ***************************** Class ****************************** */

/** Describes the setup of a conceptual block. */
class BlockDescriptor {

	/* ========================= Fields ========================= */

	public final String topic;
	public final int startColumn;

	/* ====================== Constructors ====================== */

	public BlockDescriptor (
		final String topic,
		final int startColumn
	) {
		this.topic = topic;
		this.startColumn = startColumn;
	}

	/* ===================== Public Methods ===================== */

	/** Turns into <code>String</code> for debugging only. */
	public String toString () {
		return getClass().getName() + "(" + topic + "," + startColumn + ")";
	}
}

/* ***************************** Class ****************************** */

/** Describes the setup of a column of a conceptual block. */
abstract class ColumnDescriptor {

	/* ========================= Fields ========================= */

	protected final SPFObservationFactory observationFactory;

	/* ====================== Constructors ====================== */

	public ColumnDescriptor (
		final SPFObservationFactory observationFactory
	) {
		this.observationFactory = observationFactory;
	}

	/* ===================== Public Methods ===================== */

	/** Do what must be done for a cell with given value. */
	public abstract void processValue ( final String value );

	/* =================== Protected Methods ==================== */

	protected void complain ( final String description ) {
		observationFactory.complain( description );
	}

	protected void warn ( final String description ) {
		observationFactory.warn( description );
	}
}


/* ***************************** Class ****************************** */

/** Processes period id column entries. */
class PeriodCodeColumnDescriptor extends ColumnDescriptor {

	/* ====================== Constructors ====================== */

	public PeriodCodeColumnDescriptor (
		final SPFObservationFactory observationFactory
	) {
		super( observationFactory );
	}

	/* ===================== Public Methods ===================== */

	public void processValue ( final String value ) {
		// ignore
	}
}

/* ***************************** Class ****************************** */

/** Processes forecaster id column entries. */
class ForecasterColumnDescriptor extends ColumnDescriptor {

	public ForecasterColumnDescriptor (
		final SPFObservationFactory observationFactory
	) {
		super( observationFactory );
	}

	/* ===================== Public Methods ===================== */

	public void processValue ( final String value ) {
		observationFactory.setForecaster( value );
	}
}

/* ***************************** Class ****************************** */

/** Processes date column entries. */
class DateColumnDescriptor extends ColumnDescriptor {

	/* ====================== Constructors ====================== */

	public DateColumnDescriptor (
		final SPFObservationFactory observationFactory
	) {
		super( observationFactory );
	}

	/* ===================== Public Methods ===================== */

	public void processValue ( final String value ) {
		observationFactory.setInterval( value );
	}
}

/* ***************************** Class ****************************** */

/** Processes empty column entries. */
class EmptyColumnDescriptor extends ColumnDescriptor {

	/* ====================== Constructors ====================== */

	public EmptyColumnDescriptor (
		final SPFObservationFactory observationFactory
	) {
		super( observationFactory );
	}

	/* ===================== Public Methods ===================== */

	public void processValue ( final String value ) {
		if (
			! "".equals( value )
			&&
			! "X".equals( value )
		) {
			complain(
				"The cell contains \"" + value + "\" but should be empty."
			);
		}
	}
}

/* ***************************** Class ****************************** */

/** Processes column entries without value translations. */
class PlainColumnDescriptor extends ColumnDescriptor {

	/* ========================= Fields ========================= */

	private final BlockDescriptor block;
	private final String breakDown;
	private final String unit;

	/* ====================== Constructors ====================== */

	public PlainColumnDescriptor (
		final SPFObservationFactory observationFactory,
		final BlockDescriptor block,
		final String breakDown,
		final String unit
	) {
		super( observationFactory );
		this.block = block;
		this.breakDown = breakDown;
		this.unit = unit;
	}

	/* ===================== Public Methods ===================== */

	public void processValue ( final String value ) {
		observationFactory.setTopic( block.topic );
		observationFactory.setBreakDown( breakDown );
		if ( observationFactory.isCubeDefined() ) {
			detailProcessValue( value );
			observationFactory.dispatchUnit( unit );
		} else {
			if (
				! "".equals( value )
				&&
				! "X".equals( value )
				&&
				! "0.00%".equals( value )
			) {
				complain(
					"Expecting empty cell if no forecaster and date is set,"
					+ " but got \"" + value + "\"."
				);
			}
		}
	}

	/** Turns into <code>String</code> for debugging only. */
	public String toString () {
		return getClass().getName() + "(" + block.topic + "," + breakDown + "," + unit + ")";
	}

	/* =================== Protected Methods ==================== */

	/** Make nonempty value ready for dispatching and dispatch. */
	protected void detailProcessValue ( final String value ) {
		if ( ! "".equals( value ) ) {
			observationFactory.dispatchValue( value );
		} else {
			observationFactory.dispatchValue( "-" );
		}
	}
}

/* ***************************** Class ****************************** */

/** Processes point estimate and IRS column entries. */
class PercentageColumnDescriptor extends PlainColumnDescriptor {

	/* ====================== Constructors ====================== */

	public PercentageColumnDescriptor (
		final SPFObservationFactory observationFactory,
		final BlockDescriptor block,
		final String breakDown,
		final String unit
	) {
		super( observationFactory, block, breakDown, unit );
	}

	/* =================== Protected Methods ==================== */

	/** Process nonempty values as percentage. */
	protected void detailProcessValue ( final String value ) {
		super.detailProcessValue( percentage( value ) );
	}

	/* ==================== Private Methods ===================== */

	/** Strip percentage sign if present. */
	private String percentage ( final String value ) {
		return value.endsWith( "%" )
			? value.substring( 0, value.length() - 1 )
			: value;
	}
}

/* ***************************** Class ****************************** */

/** Processes probability column entries. */
class ProbabilityColumnDescriptor extends PercentageColumnDescriptor {

	/* ====================== Constructors ====================== */

	public ProbabilityColumnDescriptor (
		final SPFObservationFactory observationFactory,
		final BlockDescriptor block,
		final String breakDown,
		final String unit
	) {
		super( observationFactory, block, breakDown, unit );
	}

	/* =================== Protected Methods ==================== */

	/** Process nonempty values as probability percentage (divide by 100). */
	protected void detailProcessValue ( final String value ) {
		if ( "".equals( value ) ) {
			observationFactory.dispatchValue( "0" );
		} else {
			super.detailProcessValue( value );
		}
	}
}
