// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (C) 2004, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.database;

import java.util.SortedMap;
import java.util.TreeMap;
import java.util.Iterator;

import Int.ecb.gesmescb.KeyFamily;
import Int.ecb.gesmescb.TimeSeries;
import Int.ecb.gesmescb.Observation;
import Int.ecb.gesmescb.Attribute;

import Int.ecb.gesmescb.time.RegularInterval;
import Int.ecb.gesmescb.time.GregorianRegularInterval;

import Int.ecb.gesmescb.util.ComponentHelper;
import Int.ecb.gesmescb.util.KeyExtractor;
import Int.ecb.gesmescb.util.SiblingGroupKeyExtractor;
import Int.ecb.gesmescb.util.ObsIntervalExtractor;
import Int.ecb.gesmescb.util.FilterIterator;
import Int.ecb.gesmescb.util.FilterIterator.Condition;
import Int.ecb.gesmescb.util.CombinedCondition;
import Int.ecb.gesmescb.util.InstanceOfCondition;
import Int.ecb.gesmescb.util.EqualsCondition;
import Int.ecb.gesmescb.util.KeyEqualsCondition;
import Int.ecb.gesmescb.util.IntervalEqualsCondition;
import Int.ecb.gesmescb.util.CubeClassCondition;
import Int.ecb.gesmescb.util.SubcubeCondition;

/* ***************************** Class ****************************** */

/**
 * Implements common basic behaviour of {@link DataSet}s.
 *
 * @author
 * <a href="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</a>
 *
 * @version @(#)$Id: AbstractDataSet.java,v 1.1 2017/01/19 16:14:03 trainot Exp $
 */

public abstract class AbstractDataSet implements DataSet {

	/* ========================= Fields ========================= */

	/** The ComponentHelper used for preferences and logging. */
	protected final ComponentHelper componentHelper;

	/** Structure of the data set. */
	protected final KeyFamily keyFamily;

	/* ====================== Constructors ====================== */

	/**
	 * Construct a new <code>DataSet</code>.
	 *
	 * @param name to be used for logging and preferences.
	 * @param keyFamily determining the structure of the data set.
	 */
	public AbstractDataSet (
		final String name,
		final KeyFamily keyFamily
	) {
		this.componentHelper = new ComponentHelper( name, this );

		if ( ( this.keyFamily = keyFamily ) == null ) {
			throw new NullPointerException(
				"Trying to create an AbstractDataSet with null keyFamily."
			);
		}
	}

	/* ===================== Public Methods ===================== */

	/* ------------------------ General ------------------------- */

	/** Retrieva all contents.
	 * In principle, this is the only method
	 * which needs to be implemented by concrete subclasses.
	 * Of course, implementation of other methods can give better performance
	 * than the default implementation in terms of <code>iterateData</code>.
	 */
	public abstract Iterator iterateData ();

	public boolean isEmpty () {
		return ! iterateData().hasNext();
	}

	public Iterator iterateTimeSeries () {
		return new FilterIterator(
			new InstanceOfCondition( TimeSeries.class ),
			iterateData()
		);
	}

	public Iterator iterateAttributes () {
		return new FilterIterator(
			new InstanceOfCondition( Attribute.class ),
			iterateData()
		);
	}

	/* ------------------------ Data Set ------------------------ */

	public Iterator iterateDataSetAttributes () {
		return new FilterIterator(
			new KeyEqualsCondition( "" ),
			iterateAttributes()
		);
	}

	public String getDataSetAttributeValue ( final String conceptId ) {
		String value = null;
		for (
			final Iterator iterator = iterateDataSetAttributes();
			iterator.hasNext();
		) {
			final Attribute attribute = (Attribute) iterator.next();
			if ( conceptId.equals( attribute.getName() ) ) {
				value = attribute.getValue();
			}
		}
		return value;
	}

	/* --------------------- Sibling Group ---------------------- */

	public Iterator iterateSiblingGroupKeys () {
		return new SiblingGroupKeyExtractor(
			keyFamily,
			new FilterIterator(
				new CubeClassCondition(
					keyFamily,
					CubeClassCondition.SIBLING_GROUP
					|
					CubeClassCondition.TIME_SERIES
					|
					CubeClassCondition.OBSERVATION
				),
				iterateData()
			)
		);
	}

	public boolean knowsSiblingGroup ( final String siblingGroupKey ) {
		return new FilterIterator(
			new EqualsCondition( siblingGroupKey ),
			iterateSiblingGroupKeys()
		).hasNext();
	}

	public Iterator iterateSiblingGroupAttributes (
		final String siblingGroupKey
	) {
		return new FilterIterator(
			new KeyEqualsCondition( siblingGroupKey ),
			iterateAttributes()
		);
	}

	public String getSiblingGroupAttributeValue (
		final String siblingGroupKey,
		final String conceptId
	) {
		String value = null;
		for (
			final Iterator iterator = iterateSiblingGroupAttributes(
				siblingGroupKey
			);
			iterator.hasNext();
		) {
			final Attribute attribute = (Attribute) iterator.next();
			if ( conceptId.equals( attribute.getName() ) ) {
				value = attribute.getValue();
			}
		}
		return value;
	}

	/* ---------------------- Time Series ----------------------- */

	public Iterator iterateTimeSeriesKeys ( final String siblingGroupKey ) {
		return new KeyExtractor(
			new FilterIterator(
				new CombinedCondition(
					CombinedCondition.AND,
					new Condition[]{
						new CubeClassCondition(
							keyFamily,
							CubeClassCondition.TIME_SERIES
							|
							CubeClassCondition.OBSERVATION
						),
						new SubcubeCondition(
							keyFamily,
							siblingGroupKey
						)
					}
				),
				iterateData()
			)
		);
	}

	public String getSiblingGroupKey ( final String timeSeriesKey ) {
		return keyFamily.computeSiblingGroupKey( timeSeriesKey );
	}

	public String getFrequencyValue ( final String timeSeriesKey ) {
		return keyFamily.computeFrequencyValue( timeSeriesKey );
	}

	public boolean knowsTimeSeries ( final String timeSeriesKey ) {
		final String siblingGroupKey = getSiblingGroupKey( timeSeriesKey );
		return new FilterIterator(
			new EqualsCondition( timeSeriesKey ),
			iterateTimeSeriesKeys( siblingGroupKey )
		).hasNext();
	}

	public TimeSeries getTimeSeries ( final String timeSeriesKey ) {
		final SortedMap observationsMap = new TreeMap();
		for (
			final Iterator iterator = new FilterIterator(
				new KeyEqualsCondition( timeSeriesKey ),
				iterateTimeSeries()
			);
			iterator.hasNext();
		) {
			final TimeSeries timeSeries = (TimeSeries) iterator.next();
			final RegularInterval range = timeSeries.getInterval();
			if ( range != null ) {
				for (
					final Iterator
						intervals = range.iterator(),
						observations = timeSeries.observationIterator();
					intervals.hasNext() && observations.hasNext();
				) {
					final RegularInterval interval
					= (RegularInterval) intervals.next();
					final Observation observation
					= (Observation) observations.next();
					if ( ! observation.isEmpty() ) {
						observationsMap.put( interval, observation );
					}
				}
			}
		}
		if ( observationsMap.isEmpty() ) {
			return null;
		} else {
			final RegularInterval first
			= (RegularInterval) observationsMap.firstKey();
			final RegularInterval last
			= (RegularInterval) observationsMap.lastKey();
			final int size = first.computeIndex( last ) + 1;
			final RegularInterval range = new GregorianRegularInterval(
				first.getFrequency(),
				first.getBegin(),
				size
			);
			final Observation[] observations = new Observation[size];
			int i = 0;
			for (
				final Iterator intervals = range.iterator();
				intervals.hasNext();
			) {
				final RegularInterval interval
				= (RegularInterval) intervals.next();
				final Observation observation
				= (Observation) observationsMap.get( interval );
				observations[i++] = observation != null
					? observation : Observation.EMPTY_OBSERVATION;
			}
			return new TimeSeries(
				timeSeriesKey,
				range,
				observations
			);
		}
	}

	public Iterator iterateTimeSeriesAttributes ( final String timeSeriesKey ) {
		return new FilterIterator(
			new CombinedCondition(
				CombinedCondition.AND,
				new Condition[]{
					new KeyEqualsCondition( timeSeriesKey ),
					new CubeClassCondition(
						keyFamily,
						CubeClassCondition.TIME_SERIES
					)
				}
			),
			iterateAttributes()
		);
	}

	public String getTimeSeriesAttributeValue (
		final String timeSeriesKey,
		final String conceptId
	) {
		String value = null;
		for (
			final Iterator iterator = iterateTimeSeriesAttributes(
				timeSeriesKey
			);
			iterator.hasNext();
		) {
			final Attribute attribute = (Attribute) iterator.next();
			if ( conceptId.equals( attribute.getName() ) ) {
				value = attribute.getValue();
			}
		}
		return value;
	}

	/* ---------------------- Observation ----------------------- */

	public Iterator iterateObservationIntervals ( final String timeSeriesKey ) {
		return new ObsIntervalExtractor(
			new FilterIterator(
				new KeyEqualsCondition( timeSeriesKey ),
				iterateData()
			)
		);
	}

	public boolean knowsObservation (
		final String timeSeriesKey,
		final RegularInterval interval
	) {
		return new FilterIterator(
			new EqualsCondition( interval ),
			iterateObservationIntervals( timeSeriesKey )
		).hasNext();
	}

	public Observation getObservation (
		final String timeSeriesKey,
		final RegularInterval interval
	) {
		final TimeSeries timeSeries = getTimeSeries( timeSeriesKey );
		if ( timeSeries == null ) return null;
		final RegularInterval period = timeSeries.getInterval();
		if ( period == null ) return null;
		final Iterator
			intervals = period.iterator(),
			observations = timeSeries.observationIterator();
		while (
			intervals.hasNext()
			&&
			observations.hasNext()
		) {
			final RegularInterval obsInterval
			= (RegularInterval) intervals.next();
			final Observation observation
			= (Observation) observations.next();
			if ( interval.equals( obsInterval ) ) {
				return observation;
			}
		}
		return null;
	}

	public Iterator iterateObservationAttributes (
		final String timeSeriesKey,
		final RegularInterval interval
	) {
		return new FilterIterator(
			new CombinedCondition(
				CombinedCondition.AND,
				new Condition[]{
					new KeyEqualsCondition( timeSeriesKey ),
					new IntervalEqualsCondition( interval )
				}
			),
			iterateAttributes()
		);
	}

	public String getObservationAttributeValue (
		final String timeSeriesKey,
		final RegularInterval interval,
		final String conceptId
	) {
		String value = null;
		for (
			final Iterator iterator = iterateObservationAttributes(
				timeSeriesKey, interval
			);
			iterator.hasNext();
		) {
			final Attribute attribute = (Attribute) iterator.next();
			if ( conceptId.equals( attribute.getName() ) ) {
				value = attribute.getValue();
			}
		}
		return value;
	}

	/* -------------------- Basic Behaviour --------------------- */

	public String toString () {
		return componentHelper.logName;
	}
}
