// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (C) 2004, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.database;

import java.util.Iterator;

import Int.ecb.gesmescb.KeyFamily;
import Int.ecb.gesmescb.TimeSeries;
import Int.ecb.gesmescb.Attribute;
import Int.ecb.gesmescb.Observation;

import Int.ecb.gesmescb.time.RegularInterval;

/**
 * <p>Specifies read access to data according to the Gesmes/TS data model.</p>
 *
 * <p>On a very abstract level, a data set maps keys to cubes
 * and cubes carry values.</p>
 *
 * <p>However,
 * this interface is very specific to the needs of the Gesmes/TS-Suite.
 * It is not the most general or elegant implementation of the
 * framework of nested multi-dimensional cubes.</p>
 *
 * <p>The nesting hierarchy is linear:
 * data set, sibling group, time series, observation.
 * Each of these types of cubes can carry attribute values for
 * all attributes foreseen at this level by the <code>DataSet</code>'s
 * {@link KeyFamily}.</p>
 *
 * <h3>Important notes</h3>
 *
 * <p>Note 1: {@link Iterator}s returned by various operations may
 * report objects repeatedly unless a stricter contract is specified
 * for an implementing class. To make returned objects unique,
 * you can use the {@link Int.ecb.gesmescb.util.UniqueIterator}.</p>
 *
 * <p>Note 2: {@link Iterator}s returned by various operations may
 * report objects in parts. For example a {@link TimeSeries} may
 * come as a sequence of time series, one for each observation.
 * Even objects reported later can overwrite or partially overwrite
 * earlier ones in which case the last should be considered the intended.</p>
 *
 * <p>However, other operations than iterations, e.g. the various
 * <code>get</code>s, must report only the single intended result.</p>
 *
 * <p>Also, many classes implementing
 * <code>DataSet</code> obey a stricter contract
 * and do not report duplicates or different versions of objects.
 * E.g. {@link MemoryDataStore} or {@link FileSystemDataSet}.</p>
 *
 * @see Int.ecb.gesmescb.util.TimeSeriesJoiner to build
 * one <code>TimeSeries</code> from a sequence of parts of it.
 *
 * @author
 * <a href="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</a>
 *
 * @version @(#)$Id: DataSet.java,v 1.1 2017/01/19 16:14:03 trainot Exp $
 */
public interface DataSet {

	/* ------------------------ General ------------------------- */

	/** Tells whether there are no stored objects. */
	boolean isEmpty ();

	/** Retrieve all objects stored in the data set.
	 * Objects are returned in unspecified order, but as fast as possible.
	 * @return an iterator over {@link TimeSeries} and {@link Attribute}s.
	 * @see #iterateTimeSeries
	 * @see #iterateAttributes
	 */
	Iterator iterateData ();

	/** Retrieve all time series stored in the data set.
	 * Time series are returned in unspecified order, but fast.
	 * @return an iterator over {@link TimeSeries}.
	 * @see #iterateData
	 */
	Iterator iterateTimeSeries ();

	/** Retrieve all attributes stored in the data set.
	 * Attributes are returned in unspecified order, but fast.
	 * @return an iterator over {@link Attribute}s.
	 * @see #iterateData
	 */
	Iterator iterateAttributes ();

	/* ------------------------ Data Set ------------------------ */

	/** Get all attributes of the data set.
	 * @return an iterator over {@link Attribute}s.
	 */
	Iterator iterateDataSetAttributes ();

	/** Get the attribute value for a given concept. */
	String getDataSetAttributeValue ( String conceptId );

	/* --------------------- Sibling Group ---------------------- */

	/** Get all sibling group keys in the data set.
	 * @return an iterator over {@link String}s.
	 */
	Iterator iterateSiblingGroupKeys ();

	/** Tells whether the sibling group is known,
	 * i.e. attributes are stored in the database
	 * or contained timeseries are known to the database.
	 */
	boolean knowsSiblingGroup ( String siblingGroupKey );

	/** Get all attributes of a sibling group.
	 * @return an iterator over {@link Attribute}s.
	 */
	Iterator iterateSiblingGroupAttributes ( String siblingGroupKey );

	/** Get the attribute value for conceptId for a given sibling group. */
	String getSiblingGroupAttributeValue (
		String siblingGroupKey, String conceptId
	);

	/* ---------------------- Time Series ----------------------- */

	/** Get all series keys in a sibling group.
	 * @return an iterator over {@link String}s.
	 */
	Iterator iterateTimeSeriesKeys ( String siblingGroupKey );

	/** Determine the sibling group a series is member of. */
	String getSiblingGroupKey ( String timeSeriesKey );

	/** Determine the frequency dimension value from a series key. */
	String getFrequencyValue ( String timeSeriesKey );

	/** Tells whether the time series is known,
	 * i.e. attributes are stored in the database
	 * or contained observations are known to the database.
	 */
	boolean knowsTimeSeries ( String timeSeriesKey );

	/** Get a series by its key. */
	TimeSeries getTimeSeries ( String timeSeriesKey );

	/** Get all attributes of a time series.
	 * @return an iterator over {@link Attribute}s.
	 */
	Iterator iterateTimeSeriesAttributes ( String timeSeriesKey );

	/** Get the attribute value for conceptId for a given time series. */
	String getTimeSeriesAttributeValue (
		String timeSeriesKey, String conceptId
	);

	/* ---------------------- Observation ----------------------- */

	/** Get all periods in a time series,
	 * where a value or attribute data is available in the data set.
	 * The iteration follows the ascending order in time,
	 * but there may be gaps when no data is available.
	 * @return an <code>Iterator</code> over
	 * {@link Int.ecb.gesmescb.time.RegularInterval}s.
	 */
	Iterator iterateObservationIntervals ( String timeSeriesKey );

	/** Tells whether the observation is known,
	 * i.e. attributes are stored in the database.
	 */
	boolean knowsObservation (
		String timeSeriesKey, RegularInterval interval
	);

	/** Get an observation from a series. */
	Observation getObservation (
		String timeSeriesKey, RegularInterval interval
	);

	/** Get all attributes of an observation.
	 * @return an iterator over {@link Attribute}s.
	 */
	Iterator iterateObservationAttributes (
		String timeSeriesKey, RegularInterval interval
	);

	/** Get the attribute value for conceptId for a given observation. */
	String getObservationAttributeValue (
		String timeSeriesKey, RegularInterval interval, String conceptId
	);

	/* -------------------- Basic Behaviour --------------------- */

	/** Free any occupied resources.
	 * Call this before disposing of the <code>DataSet</code>.
	 */
	void close ();
}
