// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (C) 2004, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.database;

import java.util.*;

import junit.framework.TestCase;

import Int.ecb.gesmescb.*;
import Int.ecb.gesmescb.time.*;

/* ***************************** Class ****************************** */

/**
 * <p>Basic testing of {@link DataStore} implementations.</P>
 *
 * <p>Extend this unit test class to build unit tests for classes
 * which should implement the <code>DataStore</code>.</p>
 *
 * @author
 * <a href="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</a>
 *
 * @version @(#)$Id: DataStoreTest.java,v 1.1 2017/01/19 16:14:03 trainot Exp $
 */
public abstract class DataStoreTest extends TestCase {

	/* ======================= Constants ======================== */

	protected static final Date versionDate;
	static {
		final Calendar calendar = Calendar.getInstance();
		calendar.clear();
		versionDate = calendar.getTime();
	}

	protected static final KeyFamilyComponent
		FREQ = new KeyFamilyComponent(
			"FREQ", KeyFamilyComponent.FREQUENCY_DIMENSION,
			"1", "AN..1", "", "", "CL_FREQ", true
		),
		DIM1 = new KeyFamilyComponent(
			"DIM1", KeyFamilyComponent.KEY_STRUCTURE_DIMENSION,
			"2", "AN1", "", "", "CL_DIM1", true
		),
		DIM2 = new KeyFamilyComponent(
			"DIM2", KeyFamilyComponent.KEY_STRUCTURE_DIMENSION,
			"3", "AN2", "", "", "CL_DIM2", true
		),
		TIME_PERIOD = new KeyFamilyComponent(
			"TIME_PERIOD", KeyFamilyComponent.TIME_DIMENSION,
			"4", "AN..35", "", "", "", true
		),
		TIME_FORMAT = new KeyFamilyComponent(
			"TIME_FORMAT", KeyFamilyComponent.TIME_DIMENSION,
			"5", "AN3", "", "", "", true
		),
		OBS_VALUE = new KeyFamilyComponent(
			"OBS_VALUE", KeyFamilyComponent.ARRAY_CELL,
			"6", "AN..15", "", "", ""
		),
		OBS_STATUS = new KeyFamilyComponent(
			"OBS_STATUS", KeyFamilyComponent.ARRAY_CELL,
			"7", "AN1", KeyFamilyComponent.MANDATORY,
			KeyFamilyComponent.ATTACHED_TO_OBSERVATION, "CL_OBS_STATUS"
		),
		OBS_CONF = new KeyFamilyComponent(
			"OBS_CONF", KeyFamilyComponent.ARRAY_CELL,
			"8", "AN1", KeyFamilyComponent.CONDITIONAL,
			KeyFamilyComponent.ATTACHED_TO_OBSERVATION, "CL_OBS_CONF"
		),
		OBS_PRE_BREAK = new KeyFamilyComponent(
			"OBS_PRE_BREAK", KeyFamilyComponent.ARRAY_CELL,
			"9", "AN..15", KeyFamilyComponent.CONDITIONAL,
			KeyFamilyComponent.ATTACHED_TO_OBSERVATION, ""
		),
		OBS_ATT = new KeyFamilyComponent(
			"OBS_ATT", KeyFamilyComponent.ATTRIBUTE,
			"", "AN..20", KeyFamilyComponent.CONDITIONAL,
			KeyFamilyComponent.ATTACHED_TO_OBSERVATION, "CL_OBS_ATT"
		),
		TS_ATT = new KeyFamilyComponent(
			"TS_ATT", KeyFamilyComponent.ATTRIBUTE,
			"", "A2", KeyFamilyComponent.MANDATORY,
			KeyFamilyComponent.ATTACHED_TO_TIME_SERIES, "CL_TS_ATT"
		),
		SG_ATT = new KeyFamilyComponent(
			"SG_ATT", KeyFamilyComponent.ATTRIBUTE,
			"", "N..10", KeyFamilyComponent.CONDITIONAL,
			KeyFamilyComponent.ATTACHED_TO_SIBLING_GROUP, ""
		),
		DS_ATT = new KeyFamilyComponent(
			"DS_ATT", KeyFamilyComponent.ATTRIBUTE,
			"", "A..10", KeyFamilyComponent.MANDATORY,
			KeyFamilyComponent.ATTACHED_TO_DATA_SET, ""
		);

	protected static final KeyFamilyComponent[]
		KFC_ARR = new KeyFamilyComponent[] {
			FREQ,
			DIM1,
			DIM2,
			TIME_PERIOD,
			TIME_FORMAT,
			OBS_VALUE,
			OBS_STATUS,
			OBS_CONF,
			OBS_PRE_BREAK,
			OBS_ATT,
			TS_ATT,
			SG_ATT,
			DS_ATT
		};

	protected static final KeyFamily
		KF = new KeyFamily(
			"KF", "Name", "Description", versionDate, KFC_ARR
		);

	protected static final Observation[]
		OBS_ARR = new Observation[] {
			new Observation( "13.6", "A", "F", "" ),
			new Observation( "13.7", "A", "F", "" ),
			new Observation( "13.8", "A", "F", "" )
		};

	protected static final TimeSeries
		TS1 = new TimeSeries( "M.B.LO", "199010199012", "710", OBS_ARR ),
		TS2 = new TimeSeries( "Q.B.LO", "1990119903", "708", OBS_ARR ),
		TS3 = new TimeSeries( "M.J.LO", "199010199012", "710", OBS_ARR );

	private static final TimeSeries[] TS = new TimeSeries[]{ TS1, TS2, TS3 };

	protected static final Attribute
		ATT_OBS = new Attribute(
			"M.B.LO", "199001", "610", "OBS_ATT", "Z"
		),
		ATT1_OBS = new Attribute(
			"M.B.LO", "199001", "610", "OBS_ATT", "Y"
		),
		ATT_TS = new Attribute(
			"M.B.LO", "", "", "TS_ATT", "JU"
		),
		ATT1_TS = new Attribute(
			"M.B.LO", "", "", "TS_ATT", "LI"
		),
		ATT2_TS = new Attribute(
			"M.W.HO", "", "", "TS_ATT", "JU"
		),
		ATT_SG = new Attribute(
			"B.LO", "", "", "SG_ATT", "123"
		),
		ATT1_SG = new Attribute(
			"B.LO", "", "", "SG_ATT", "456"
		),
		ATT2_SG = new Attribute(
			"A.HI", "", "", "SG_ATT", "123"
		),
		ATT_DS = new Attribute(
			"", "", "", "DS_ATT", "Irrdzdz"
		),
		ATT1_DS = new Attribute(
			"", "", "", "DS_ATT", "nasowas"
		);

	/* ====================== Constructors ====================== */

	public DataStoreTest ( final String name ) {
		super( name );
	}

	/* =================== Protected Methods ==================== */

	/** Obtain the guinea pig <code>DataStore</code> to test. */
	protected abstract DataStore createDataStore (
		final KeyFamily keyFamily
	);

	/**
	 * Add all retrieved objects from <code>Iterator</code>
	 * into a <code>Collection</code>.
	 */
	protected static Collection slurp (
		final Iterator iterator,
		final Collection collection
	) {
		while ( iterator.hasNext() ) {
			collection.add( iterator.next() );
		}
		return collection;
	}

	/* ===================== Public Methods ===================== */

	/* --------------------- Test Methods --------------------- */

	public void testIsEmpty () {
		final DataStore store = createDataStore( KF );
		assertTrue( "Empty", store.isEmpty() );
		store.store( ATT_DS );
		assertTrue( "Nonempty", ! store.isEmpty() );
		store.close();
	}

	public void testIterators () {
		final DataStore store = createDataStore( KF );
		store.store( TS3 );
		store.store( TS2 );
		store.store( TS1 );
		store.store( ATT_OBS );
		store.store( ATT_TS );
		store.store( ATT_SG );
		store.store( ATT_DS );

		{
			final Iterator iterator = store.iterateData();
			final Collection all = slurp( iterator, new HashSet() );
			assertEquals( "Seven objects", 7, all.size() );
			assertTrue( "TS1 in all", all.contains( TS1 ) );
			assertTrue( "TS2 in all", all.contains( TS2 ) );
			assertTrue( "TS3 in all", all.contains( TS3 ) );
			assertTrue( "ATT_OBS in all", all.contains( ATT_OBS ) );
			assertTrue( "ATT_TS in all", all.contains( ATT_TS ) );
			assertTrue( "ATT_SG in all", all.contains( ATT_SG ) );
			assertTrue( "ATT_DS in all", all.contains( ATT_DS ) );
		}

		{
			final Iterator iterator = store.iterateTimeSeries();
			final Collection series = slurp( iterator, new HashSet() );
			assertEquals( "Three series", 3, series.size() );
			assertTrue( "TS1 in series", series.contains( TS1 ) );
			assertTrue( "TS2 in series", series.contains( TS2 ) );
			assertTrue( "TS3 in series", series.contains( TS3 ) );
		}

		{
			final Iterator iterator = store.iterateAttributes();
			final Collection atts = slurp( iterator, new HashSet() );
			assertEquals( "Four atts", 4, atts.size() );
			assertTrue( "ATT_OBS in atts", atts.contains( ATT_OBS ) );
			assertTrue( "ATT_TS in atts", atts.contains( ATT_TS ) );
			assertTrue( "ATT_SG in atts", atts.contains( ATT_SG ) );
			assertTrue( "ATT_DS in atts", atts.contains( ATT_DS ) );
		}

		{
			final Iterator dsai = store.iterateDataSetAttributes();
			final Collection dsas = slurp( dsai, new HashSet() );
			assertEquals( "One DS_ATT", 1, dsas.size() );
			assertTrue( "The DS_ATT", dsas.contains( ATT_DS ) );
		}

		{
			final Iterator sgi = store.iterateSiblingGroupKeys();
			final Collection sgs = slurp( sgi, new HashSet() );
			assertEquals( "Two SGs", 2, sgs.size() );
			assertTrue( "Exists SG1", sgs.contains( "B.LO" ) );
			assertTrue( "Exists SG2", sgs.contains( "J.LO" ) );
		}

		{
			final Iterator sg1ai = store.iterateSiblingGroupAttributes( "B.LO" );
			final Collection sg1as = slurp( sg1ai, new HashSet() );
			assertEquals( "One SG1_ATT", 1, sg1as.size() );
			assertTrue( "The SG_ATT", sg1as.contains( ATT_SG ) );
		}

		{
			final Iterator sg2ai = store.iterateSiblingGroupAttributes( "J.LO" );
			final Collection sg2as = slurp( sg2ai, new HashSet() );
			assertEquals( "No SG2_ATT", 0, sg2as.size() );
		}

		{
			final Iterator sg1tsi = store.iterateTimeSeriesKeys( "B.LO" );
			final Collection sg1ts = slurp( sg1tsi, new HashSet() );
			assertEquals( "Two SG1_TS", 2, sg1ts.size() );
			assertTrue( "The SG1_TS1", sg1ts.contains( TS1.getKey() ) );
			assertTrue( "The SG1_TS2", sg1ts.contains( TS2.getKey() ) );
		}

		{
			final Iterator sg2tsi = store.iterateTimeSeriesKeys( "J.LO" );
			final Collection sg2ts = slurp( sg2tsi, new HashSet() );
			assertEquals( "One SG2_TS", 1, sg2ts.size() );
			assertTrue( "The SG2_TS3", sg2ts.contains( TS3.getKey() ) );
		}

		{
			final Iterator ts1ai = store.iterateTimeSeriesAttributes(
				TS1.getKey()
			);
			final Collection ts1as = slurp( ts1ai, new HashSet() );
			assertEquals( "One TS1_ATT", 1, ts1as.size() );
			assertTrue( "The TS1_ATT", ts1as.contains( ATT_TS ) );
		}

		{
			final Iterator ts2ai = store.iterateTimeSeriesAttributes(
				TS2.getKey()
			);
			final Collection ts2as = slurp( ts2ai, new HashSet() );
			assertEquals( "No TS2_ATT", 0, ts2as.size() );
		}

		{
			final Iterator ts3ai = store.iterateTimeSeriesAttributes(
				TS3.getKey()
			);
			final Collection ts3as = slurp( ts3ai, new HashSet() );
			assertEquals( "No TS3_ATT", 0, ts3as.size() );
		}

		for ( int i = 0; i < TS.length; ++i ) {
			final Iterator tsoi = store.iterateObservationIntervals(
				TS[i].getKey()
			);
			final Collection tsos = slurp( tsoi, new HashSet() );
			assertEquals(
				"Number of observations [i=" + i + "]",
				i == 0 ? 4 : 3,
				tsos.size()
			);
			final RegularInterval range = TS[i].getInterval();
			for (
				final Iterator periods = range.iterator();
				periods.hasNext();
			) {
				final Object period = periods.next();
				assertTrue(
					"TS" + ( i + 1 ) + "[" + period.toString() + "]",
					tsos.contains( period )
				);
			}
		}

		for ( int i = 0; i < TS.length; ++i ) {
			final RegularInterval range = TS[i].getInterval();
			for (
				final Iterator periods = range.iterator();
				periods.hasNext();
			) {
				final RegularInterval period = (RegularInterval) periods.next();
				final Iterator obsAtti = store.iterateObservationAttributes(
					TS[i].getKey(), period
				);
				final Collection obsAtts = slurp( obsAtti, new HashSet() );
				if ( i == 0 && period.equals( ATT_OBS.getInterval() ) ) {
					assertEquals( "One OBS_ATT", 1, obsAtts.size() );
					assertTrue( "The OBS_ATT", obsAtts.contains( ATT_OBS ) );
				} else {
					assertEquals(
						"No OBS_ATT[" + i + "," + period + "]",
						0, obsAtts.size()
					);
				}
			}
		}

		store.close();
	}

	public void testKnows () {
		final DataStore store = createDataStore( KF );
		assertTrue(
			"Initially B.LO unknown",
			! store.knowsSiblingGroup( "B.LO" )
		);
		assertTrue(
			"Initially Q.B.LO unknown",
			! store.knowsTimeSeries( "Q.B.LO" )
		);
		assertTrue(
			"Initially M.B.LO[Jan1990] unknown",
			! store.knowsObservation(
				"M.B.LO",
				new GregorianRegularInterval(
					RegularInterval.MONTHLY, 1990, 0, 1
				)
			)
		);

		store.store( ATT_OBS );
		assertTrue(
			"Now B.LO is known",
			store.knowsSiblingGroup( "B.LO" )
		);
		assertTrue(
			"Also M.B.LO known",
			store.knowsTimeSeries( "M.B.LO" )
		);
		assertTrue(
			"And also M.B.LO[Jan1990]",
			store.knowsObservation(
				"M.B.LO",
				new GregorianRegularInterval(
					RegularInterval.MONTHLY, 1990, 0, 1
				)
			)
		);
		assertTrue(
			"Still M.B.LO[Feb1990] unknown",
			! store.knowsObservation(
				"M.B.LO",
				new GregorianRegularInterval(
					RegularInterval.MONTHLY, 1990, 1, 1
				)
			)
		);
		assertTrue(
			"Still M.B.LO[Dec1989] unknown",
			! store.knowsObservation(
				"M.B.LO",
				new GregorianRegularInterval(
					RegularInterval.MONTHLY, 1989, 11, 1
				)
			)
		);
		assertTrue(
			"Still Q.B.LO unknown",
			! store.knowsTimeSeries( "Q.B.LO" )
		);
		assertTrue(
			"Still M.J.LO unknown",
			! store.knowsTimeSeries( "M.J.LO" )
		);
		assertTrue(
			"Still J.LO unknown",
			! store.knowsSiblingGroup( "J.LO" )
		);

		store.store( TS3 );
		assertTrue(
			"Still B.LO is known",
			store.knowsSiblingGroup( "B.LO" )
		);
		assertTrue(
			"Also M.B.LO known",
			store.knowsTimeSeries( "M.B.LO" )
		);
		assertTrue(
			"And also M.J.LO[Oct1990]",
			store.knowsObservation(
				"M.J.LO",
				new GregorianRegularInterval(
					RegularInterval.MONTHLY, 1990, 9, 1
				)
			)
		);
		assertTrue(
			"And also M.J.LO[Nov1990]",
			store.knowsObservation(
				"M.J.LO",
				new GregorianRegularInterval(
					RegularInterval.MONTHLY, 1990, 10, 1
				)
			)
		);
		assertTrue(
			"And also M.J.LO[Dec1990]",
			store.knowsObservation(
				"M.J.LO",
				new GregorianRegularInterval(
					RegularInterval.MONTHLY, 1990, 11, 1
				)
			)
		);
		assertTrue(
			"Still M.J.LO[Jan1991] unknown",
			! store.knowsObservation(
				"M.J.LO",
				new GregorianRegularInterval(
					RegularInterval.MONTHLY, 1991, 0, 1
				)
			)
		);
		assertTrue(
			"Still M.J.LO[Sep1990] unknown",
			! store.knowsObservation(
				"M.J.LO",
				new GregorianRegularInterval(
					RegularInterval.MONTHLY, 1990, 8, 1
				)
			)
		);
		assertTrue(
			"Still Q.B.LO unknown",
			! store.knowsTimeSeries( "Q.B.LO" )
		);
		assertTrue(
			"Newly M.J.LO known",
			store.knowsTimeSeries( "M.J.LO" )
		);
		assertTrue(
			"Newly J.LO known",
			store.knowsSiblingGroup( "J.LO" )
		);

		assertTrue(
			"Still M.W.HO unknown",
			! store.knowsTimeSeries( "M.W.HO" )
		);
		assertTrue(
			"Still W.HO unknown",
			! store.knowsSiblingGroup( "W.HO" )
		);
		store.store( ATT2_TS );
		assertTrue(
			"Now M.W.HO known",
			store.knowsTimeSeries( "M.W.HO" )
		);
		assertTrue(
			"Now W.HO known",
			store.knowsSiblingGroup( "W.HO" )
		);

		assertTrue(
			"Still A.HI unknown",
			! store.knowsSiblingGroup( "A.HI" )
		);
		assertTrue(
			"Still M.A.HI unknown",
			! store.knowsTimeSeries( "M.A.HI" )
		);
		store.store( ATT2_SG );
		assertTrue(
			"Now A.HI known",
			store.knowsSiblingGroup( "A.HI" )
		);
		assertTrue(
			"Still M.A.HI unknown",
			! store.knowsTimeSeries( "M.A.HI" )
		);

		store.store( TS1 );
		store.store( TS2 );
		store.store( ATT_TS );
		store.store( ATT_SG );
		store.store( ATT_DS );

		assertTrue(
			"Finally B.LO known",
			store.knowsSiblingGroup( "B.LO" )
		);
		assertTrue(
			"Finally M.B.LO known",
			store.knowsTimeSeries( "M.B.LO" )
		);
		assertTrue(
			"Finally Q.B.LO known",
			store.knowsTimeSeries( "Q.B.LO" )
		);

		store.close();
	}

	public void testGetSiblingGroupKey () {
		final DataStore store = createDataStore( KF );
		assertEquals( "B.LO", store.getSiblingGroupKey( "M.B.LO" ) );
		assertEquals( "B.LO", store.getSiblingGroupKey( "Q.B.LO" ) );
	}

	public void testGetFrequencyValue () {
		final DataStore store = createDataStore( KF );
		assertEquals( "M", store.getFrequencyValue( "M.B.LO" ) );
		assertEquals( "Q", store.getFrequencyValue( "Q.B.LO" ) );
	}

	public void testGetSeries () {
		final DataStore store = createDataStore( KF );
		store.store( TS1 );
		assertNull( "NoTS2", store.getTimeSeries( TS2.getKey() ) );
		store.store( TS2 );
		assertEquals( "TS1", TS1, store.getTimeSeries( TS1.getKey() ) );
		assertEquals( "TS2", TS2, store.getTimeSeries( TS2.getKey() ) );
		assertNull( "NoTS3", store.getTimeSeries( TS3.getKey() ) );

		for ( int i = 0; i < OBS_ARR.length; ++i ) {
			assertEquals(
				"OBS"+i, OBS_ARR[i],
				store.getObservation(
					TS1.getKey(), TS1.getInterval().computeBasePeriod( i )
				)
			);
		}

		store.close();
	}

	public void testGetAttribute () {
		final DataStore store = createDataStore( KF );

		assertNull(
			"NoATT_DS", store.getDataSetAttributeValue( ATT_DS.getName() )
		);
		store.store( ATT_DS );
		assertEquals(
			"ATT_DS",
			ATT_DS.getValue(),
			store.getDataSetAttributeValue( ATT_DS.getName() )
		);

		assertNull(
			"NoATT_SG", store.getDataSetAttributeValue( ATT_SG.getName() )
		);
		store.store( ATT_SG );
		assertEquals(
			"ATT_SG",
			ATT_SG.getValue(),
			store.getSiblingGroupAttributeValue(
				ATT_SG.getKey(), ATT_SG.getName()
			)
		);

		assertNull(
			"NoATT_TS", store.getDataSetAttributeValue( ATT_TS.getName() )
		);
		store.store( ATT_TS );
		assertEquals(
			"ATT_TS",
			ATT_TS.getValue(),
			store.getTimeSeriesAttributeValue(
				ATT_TS.getKey(), ATT_TS.getName()
			)
		);

		assertNull(
			"NoATT_OBS", store.getDataSetAttributeValue( ATT_OBS.getName() )
		);
		store.store( ATT_OBS );
		assertEquals(
			"ATT_OBS",
			ATT_OBS.getValue(),
			store.getObservationAttributeValue(
				ATT_OBS.getKey(),
				ATT_OBS.getInterval(),
				ATT_OBS.getName()
			)
		);
		store.close();
	}

	public void testEmptySeries () {
		final DataStore store = createDataStore( KF );

		store.store(
			new TimeSeries(
				"A.K.LO", "20032002", "702", new Observation[]{}
			)
		);
		assertTrue(
			"Don't have any observations",
			! store.iterateObservationIntervals( "A.K.LO" ).hasNext()
		);
		assertTrue(
			"Don't know empty series",
			! store.knowsTimeSeries( "A.K.LO" )
		);
		assertTrue(
			"Don't know empty sibling group",
			! store.knowsSiblingGroup( "K.LO" )
		);
		assertTrue(
			"Means it's all empty",
			store.isEmpty()
		);

		store.store(
			new TimeSeries(
				"H.K.LO", "20031", "604",
				new Observation[]{
					Observation.EMPTY_OBSERVATION
				}
			)
		);
		assertTrue(
			"Don't have empty observations",
			! store.iterateObservationIntervals( "H.K.LO" ).hasNext()
		);
		assertTrue(
			"Don't know empty obs series",
			! store.knowsTimeSeries( "H.K.LO" )
		);
		assertTrue(
			"Don't know empty obs sibling group",
			! store.knowsSiblingGroup( "K.LO" )
		);
		assertTrue(
			"Means it's still all empty",
			store.isEmpty()
		);

		store.close();
	}

	public void testOverwriteAttribute () {
		final Attribute[][] attributes = new Attribute[][]{
			{ ATT_OBS, ATT_TS, ATT_SG, ATT_DS },
			{ ATT1_OBS, ATT1_TS, ATT1_SG, ATT1_DS }
		};

		final DataStore store = createDataStore( KF );
		for ( int i = 0; i < 4; ++i ) {
			final int k = i % 2;
			for ( int j = 0; j < 4; ++j ) {
				store.store( attributes[k][j] );
			}
			assertEquals(
				"obs[" + k + "]",
				attributes[k][0].getValue(),
				store.getObservationAttributeValue(
					ATT_OBS.getKey(),
					ATT_OBS.getInterval(),
					ATT_OBS.getName()
				)
			);
			assertEquals(
				"ts[" + k + "]",
				attributes[k][1].getValue(),
				store.getTimeSeriesAttributeValue(
					ATT_TS.getKey(),
					ATT_TS.getName()
				)
			);
			assertEquals(
				"sg[" + k + "]",
				attributes[k][2].getValue(),
				store.getSiblingGroupAttributeValue(
					ATT_SG.getKey(),
					ATT_SG.getName()
				)
			);
			assertEquals(
				"ds[" + k + "]",
				attributes[k][3].getValue(),
				store.getDataSetAttributeValue(
					ATT_DS.getName()
				)
			);
		}

		store.close();
	}

	public void testJoinSeries () {
		final Observation obs1 = new Observation( "44.7", "E", "", "" );
		final DataStore store = createDataStore( KF );

		store.store( TS1 );
		store.store(
			new TimeSeries(
				TS1.getKey(), "199011", "610",
				new Observation[]{ obs1 }
			)
		);
		assertEquals(
			"Overwritten obs",
			new TimeSeries(
				TS1.getKey(), TS1.getInterval(),
				new Observation[]{
					OBS_ARR[0], obs1, OBS_ARR[2]
				}
			),
			store.getTimeSeries( TS1.getKey() )
		);

		store.store( TS1 );
		assertEquals(
			"Overwritten back",
			TS1,
			store.getTimeSeries( TS1.getKey() )
		);

		store.store(
			new TimeSeries(
				TS1.getKey(), "199009199101", "710",
				new Observation[]{
					obs1,
					Observation.EMPTY_OBSERVATION,
					Observation.EMPTY_OBSERVATION,
					Observation.EMPTY_OBSERVATION,
					obs1
				}
			)
		);
		assertEquals(
			"Extended time series",
			new TimeSeries(
				TS1.getKey(), "199009199101", "710",
				new Observation[]{
					obs1,
					OBS_ARR[0], OBS_ARR[1], OBS_ARR[2],
					obs1
				}
			),
			store.getTimeSeries( TS1.getKey() )
		);

		store.store(
			new TimeSeries(
				TS1.getKey(), "199007", "610",
				new Observation[]{
					obs1
				}
			)
		);
		assertEquals(
			"Hole in time series",
			new TimeSeries(
				TS1.getKey(), "199007199101", "710",
				new Observation[]{
					obs1,
					Observation.EMPTY_OBSERVATION,
					obs1, OBS_ARR[0], OBS_ARR[1], OBS_ARR[2], obs1
				}
			),
			store.getTimeSeries( TS1.getKey() )
		);

		store.store(
			new TimeSeries(
				TS1.getKey(), "199011199103", "710",
				new Observation[]{
					obs1,
					Observation.EMPTY_OBSERVATION,
					Observation.EMPTY_OBSERVATION,
					Observation.EMPTY_OBSERVATION,
					obs1
				}
			)
		);
		assertEquals(
			"Combined attack",
			new TimeSeries(
				TS1.getKey(), "199007199103", "710",
				new Observation[]{
					obs1,
					Observation.EMPTY_OBSERVATION,
					obs1, OBS_ARR[0], obs1, OBS_ARR[2], obs1,
					Observation.EMPTY_OBSERVATION,
					obs1
				}
			),
			store.getTimeSeries( TS1.getKey() )
		);

		store.close();
	}
}
