// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (C) 2005, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.database;

import java.util.Iterator;

import Int.ecb.gesmescb.database.memory.MemoryCubeImplementation;

/* ***************************** Class ****************************** */

/**
 * <p>Basic testing of {@link HierarchicalCube} implementations.</P>
 *
 * @author
 * <a href="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</a>
 * @author
 * <a href="mailto:Xavier.Sosnowska@ECB.int">Xavier Sosnowska</a>
 *
 * @version @(#)$Id: HierarchicalCubeTest.java,v 1.1 2017/01/19 16:14:03 trainot Exp $
 */
public abstract class HierarchicalCubeTest extends CuboidTest {

	/* ====================== Constructors ====================== */

	public HierarchicalCubeTest ( final String name ) {
		super( name );
	}

	/* =================== Protected Methods ==================== */

	protected abstract HierarchicalCube createCube ();

	protected Cuboid createCuboid () {
		return createCube();
	}

	/* ===================== Public Methods ===================== */

	public void testChild () {
		final HierarchicalCube root = createCube();
		assertTrue( "Initially root empty", root.isEmpty() );
		assertTrue( "No Child", !root.iterateChildKeys().hasNext() );

		final HierarchicalCube child = root.getChild( "child" );
		assertTrue( "Root still empty", root.isEmpty() );
		assertTrue( "Child is unknown", !root.knowsChild( "child" ) );
		assertTrue( "No Child, because no value yet", !root.iterateChildKeys().hasNext() );
		child.setAttribute( "COMMENT", "Hi, there!" );
		assertTrue( "Root not empty anymore", !root.isEmpty() );
		assertEquals( "One child 1", child, root.getChild( "child" ) );
		assertTrue( "Child is known", root.knowsChild( "child" ) );
		final Iterator iterator = root.iterateChildKeys();
		assertTrue( "Some children", iterator.hasNext() );
		assertEquals( "The child's key", "child", iterator.next() );
		assertTrue( "No more children", !iterator.hasNext() );
		child.removeAttribute( "COMMENT" );
		assertTrue( "Root again empty", root.isEmpty() );
		assertTrue( "Child is unknown", !root.knowsChild( "child" ) );
		assertTrue( "No Child, because no values anymore", !root.iterateChildKeys().hasNext() );

		final HierarchicalCube grandchild = child.getChild( "grandchild" );
		assertTrue( "Root still empty in spite of grandchild", root.isEmpty() );
		assertTrue( "Child still empty in spite of grandchild", child.isEmpty() );
		assertTrue( "Child is unknown", !root.knowsChild( "child" ) );
		assertTrue( "No Child 3", !root.iterateChildKeys().hasNext() );
		assertTrue( "SubChild is unknown", !child.knowsChild( "grandchild" ) );
		assertTrue( "No subChild", !child.iterateChildKeys().hasNext() );

		grandchild.setAttribute( "REMARK", "Hello, hello" );
		assertTrue( "Root no longer empty due to grandchild", !root.isEmpty() );
		assertTrue( "Child no longer empty due to grandchild", !child.isEmpty() );
		assertTrue( "Child is known because of grandchild", root.knowsChild( "child" ) );
		assertTrue( "Grandchild is not known by root", !root.knowsChild( "grandchild" ) );
		final Iterator iterator2 = root.iterateChildKeys();
		assertTrue( "Some children 2", iterator2.hasNext() );
		assertEquals( "The child's key 2", "child", iterator2.next() );
		assertTrue( "No more children 2", !iterator2.hasNext() );
		final Iterator iterator3 = child.iterateChildKeys();
		assertTrue( "Some grandchildren", iterator3.hasNext() );
		assertEquals( "The grandchild's key", "grandchild", iterator3.next() );
		assertTrue( "No more grandchildren", !iterator3.hasNext() );
		assertTrue( "Grandchild is known", child.knowsChild( "grandchild" ) );

		grandchild.removeAttribute( "REMARK" );
		assertTrue( "Grandchild is now unknown", !child.knowsChild( "grandchild" ) );
		assertTrue( "No subChild 2", !child.iterateChildKeys().hasNext() );
		assertTrue( "Root finally empty", root.isEmpty() );
		assertTrue( "Child finally empty", child.isEmpty() );
		assertTrue( "Child is now unknown", !root.knowsChild( "child" ) );
		assertTrue( "No Child 2", !root.iterateChildKeys().hasNext() );
	}

	public void testTwoWayAccess () {
		final HierarchicalCube root = createCube();
		final HierarchicalCube childRoute1 = root.getChild( "child" );
		final HierarchicalCube childRoute2 = root.getChild( "child" );
		childRoute1.setAttribute( "COMMENT", "432" );
		assertEquals( "Second knows attribute", "432", childRoute2.getAttribute( "COMMENT" ) );

		final HierarchicalCube grandchildRoute1 = childRoute1.getChild( "grandchild" );
		assertTrue( "Route 2 does not yet know Child", !childRoute2.knowsChild( "grandchild" ) );
		assertTrue( "Route 2 does not yet iterate Child", !childRoute2.iterateChildKeys().hasNext() );
		grandchildRoute1.setAttribute( "REMARK", "321" );
		assertTrue( "Route 2 knows Child", childRoute2.knowsChild( "grandchild" ) );
		assertTrue( "Route 2 does iterate Child", childRoute2.iterateChildKeys().hasNext() );

		childRoute2.setAttribute( "COMMENT", "234" );
		assertEquals( "Route 1 learns from route 2", "234", childRoute1.getAttribute( "COMMENT" ) );
	}

	public void testEqualsHashCode () {
		final HierarchicalCube cube1 = createCube();
		final HierarchicalCube cube2 = createCube();
		equalAndHashCodeTest( cube1, cube2 );
		equalAndHashCodeTest( cube2, cube1 );
	}

	public void testEqualsHashCodeConsistentWithReferenceImplementation () {
		final HierarchicalCube cube1 = createCube();
		final HierarchicalCube cube2 = HierarchicalCubifier.newCubifier( new MemoryCubeImplementation() );
		equalAndHashCodeTest( cube1, cube2 );
		equalAndHashCodeTest( cube2, cube1 );
	}

	private void equalAndHashCodeTest (
		final HierarchicalCube cube1,
		final HierarchicalCube cube2
	) {
		assertEquals( "All empty cubes are equal", cube1, cube2 );
		assertEquals(
			"All empty cubes have equal hash value",
			cube1.hashCode(), cube2.hashCode()
		);
		cube1.setAttribute( "COMMENT", "foo" );
		assertTrue( "Empty cube not equal nonempty", ! cube1.equals( cube2 ) );
		assertTrue(
			"Non empty cubes most likely have non equal hash value",
			cube1.hashCode() != cube2.hashCode()
		);
		cube2.setAttribute( "COMMENT", "foo" );
		assertEquals( "Equally attributed cubes are equal", cube1, cube2 );
		assertEquals(
			"All equally attributed cubes have equal hash value",
			cube1.hashCode(), cube2.hashCode()
		);
		final HierarchicalCube child1 = cube1.getChild( "child" );
		assertEquals( "Virtual child must not break equality", cube1, cube2 );
		assertEquals(
			"Virtual child must not break equality of hash value",
			cube1.hashCode(), cube2.hashCode()
		);
		child1.setAttribute( "COMMENT", "foo" );
		assertTrue( "Non empty child makes not equal", ! cube1.equals( cube2 ) );
		assertTrue(
			"Non empty child most likely implies non equal hash value",
			cube1.hashCode() != cube2.hashCode()
		);
		final HierarchicalCube child2 = cube2.getChild( "child" );
		assertTrue(
			"Virtual child will not re-establish equality",
			! cube1.equals( cube2 )
		);
		assertTrue(
			"Virtual child will most likely not re-establish equality of hash value",
			cube1.hashCode() != cube2.hashCode()
		);
		child2.setAttribute( "COMMENT", "foo" );
		assertEquals( "Equal children re-establish equality", cube1, cube2 );
		assertEquals(
			"Equal children re-establish equality of hash value",
			cube1.hashCode(), cube2.hashCode()
		);
		cube1.removeAttribute( "COMMENT" );
		cube2.removeAttribute( "COMMENT" );
		assertEquals( "Simultaneous attribute removal keeps equality", cube1, cube2 );
		assertEquals(
			"Simultaneous attribute removal keeps equality of hash value",
			cube1.hashCode(), cube2.hashCode()
		);
		child1.removeAttribute( "COMMENT" );
		child2.removeAttribute( "COMMENT" );
		assertEquals( "Simultaneous child attribute removal keeps equality", cube1, cube2 );
		assertEquals(
			"Simultaneous child attribute removal keeps equality of hash value",
			cube1.hashCode(), cube2.hashCode()
		);
	}
}
