// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (C) 2005, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.database;

import java.util.Iterator;
import java.util.Collections;

/**
 * Makes a {@link HierarchicalCube} out of a {@link HierarchicalCubeImplementation}.
 * I.e.: it takes care of allocating/deallocating children.
 * This relies on the assumption that it has full control over
 * the <code>HierarchicalCubeImplementation</code>, i.e. that this
 * is not messed up by direct access from elsewhere.
 *
 * @author
 * <a href="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</a>
 * @author
 * <a href="mailto:Xavier.Sosnowska@ECB.int">Xavier Sosnowska</a>
 *
 * @version @(#)$Id: HierarchicalCubifier.java,v 1.1 2017/01/19 16:14:03 trainot Exp $
 */
public abstract class HierarchicalCubifier extends AbstractHierarchicalCube {

	/* ====================== Constructors ====================== */

	HierarchicalCubifier () {}

	/* ===================== Public Methods ===================== */

	public boolean isEmpty() {
		final HierarchicalCubeImplementation cubeImplementation = getCubeImplementation();
		if ( cubeImplementation == null ) {
			return true;
		} else {
			return cubeImplementation.isEmpty();
		}
	}

	public static HierarchicalCubifier newCubifier( final HierarchicalCubeImplementation cubeImplementation ) {
		return new HierarchicalRootCubifier( cubeImplementation );
	}

	public Iterator iterateAttributes () {
		final HierarchicalCubeImplementation cubeImplementation = getCubeImplementation();
		if ( cubeImplementation == null ) {
			return Collections.EMPTY_SET.iterator();
		}
		return new CubifyingIterator( this, cubeImplementation.iterateAttributes() );
	}

	public String getAttribute ( final String conceptId ) {
		final HierarchicalCubeImplementation cubeImplementation = getCubeImplementation();
		return cubeImplementation == null
			? null
			: cubeImplementation.getAttribute( conceptId );
	}

	public void setAttribute ( final String conceptId, final String value ) {
		createOrGetCubeImplementation().setAttribute( conceptId, value );
	}

	public void removeAttribute ( final String conceptId ) {
		final HierarchicalCubeImplementation cubeImplementation = getCubeImplementation();
		if ( cubeImplementation == null ) return;
		cubeImplementation.removeAttribute( conceptId );
		possiblyRemoveCubeImplementation();
	}

	public Iterator iterateChildKeys () {
		final HierarchicalCubeImplementation cubeImplementation = getCubeImplementation();
		if ( cubeImplementation == null ) {
			return Collections.EMPTY_SET.iterator();
		}
		return new CubifyingIterator( this, cubeImplementation.iterateChildKeys() );
	}

	public boolean knowsChild ( final String key ) {
		final HierarchicalCubeImplementation cubeImplementation = getCubeImplementation();
		return
			cubeImplementation != null
			&&
			cubeImplementation.existsChild( key );
	}

	public HierarchicalCube getChild ( final String key ) {
		return new HierarchicalChildCubifier( this, key );
	}

	/* =================== Protected Methods ==================== */

	/** Get the <code>HierarchicalCubeImplementation</code> which corresponds to
	 * this <code>HierarchicalCubifier</code> in the parallel hierarchy of
	 * <code>HierarchicalCubeImplementations</code>.
	 * Return <code>null</code> if none yet exists. */
	protected abstract HierarchicalCubeImplementation getCubeImplementation ();

	/** Get the <code>HierarchicalCubeImplementation</code> which corresponds to
	 * this <code>HierarchicalCubifier</code> in the parallel hierarchy of
	 * <code>HierarchicalCubeImplementations</code>.
	 * Create a new one and return it if none yet exists. */
	protected abstract HierarchicalCubeImplementation createOrGetCubeImplementation ();

	/** Remove the <code>HierarchicalCubeImplementation</code> which corresponds to
	 * this <code>HierarchicalCubifier</code> in the parallel hierarchy of
	 * <code>HierarchicalCubeImplementations</code> if it is empty. */
	protected abstract void possiblyRemoveCubeImplementation ();
}

/* ***************************** Class ****************************** */

/** Iterator which takes care of deallocation in case of {@link #remove()}. */
class CubifyingIterator implements Iterator {

	/* ========================= Fields ========================= */

	private final HierarchicalCubifier cubifier;
	private final Iterator iterator;

	/* ====================== Constructors ====================== */

	public CubifyingIterator (
		final HierarchicalCubifier cubifier, final Iterator iterator
	) {
		this.cubifier = cubifier;
		this.iterator = iterator;
	}

	/* ===================== Public Methods ===================== */

	public boolean hasNext () {
		return iterator.hasNext();
	}

	public Object next () {
		return iterator.next();
	}

	public void remove () {
		iterator.remove();
		cubifier.possiblyRemoveCubeImplementation();
	}
}

/* ***************************** Class ****************************** */

/** HierarchicalCubifier connecting to the root <code>HierarchicalCubeImplementation</code>. */
class HierarchicalRootCubifier extends HierarchicalCubifier {

	/* ========================= Fields ========================= */

	private HierarchicalCubeImplementation cubeImplementation;

	/* ====================== Constructors ====================== */

	public HierarchicalRootCubifier ( final HierarchicalCubeImplementation cubeImplementation ) {
		this.cubeImplementation = cubeImplementation;
	}

	/* =================== Protected Methods ==================== */

	protected HierarchicalCubeImplementation getCubeImplementation () {
		return cubeImplementation;
	}

	protected HierarchicalCubeImplementation createOrGetCubeImplementation () {
		return cubeImplementation;
	}

	protected void possiblyRemoveCubeImplementation () {}
}


/* ***************************** Class ****************************** */

/** HierarchicalCubifier connecting via its ancestors to a
 * <code>HierarchicalCubeImplementation</code>. */
class HierarchicalChildCubifier extends HierarchicalCubifier {

	/* ========================= Fields ========================= */

	private HierarchicalCubifier parent;
	private String key;

	/* ====================== Constructors ====================== */

	public HierarchicalChildCubifier ( final HierarchicalCubifier parent, final String key ) {
		this.parent = parent;
		this.key = key;
	}

	/* =================== Protected Methods ==================== */

	protected HierarchicalCubeImplementation getCubeImplementation () {
		return parent.getCubeImplementation().getChild( key );
	}

	protected HierarchicalCubeImplementation createOrGetCubeImplementation () {
		final HierarchicalCubeImplementation parentImplementation = parent.createOrGetCubeImplementation();
		return parentImplementation.existsChild( key )
			? parentImplementation.getChild( key )
			: parentImplementation.createChild( key );
	}

	protected void possiblyRemoveCubeImplementation () {
		final HierarchicalCubeImplementation cubeImplementation = getCubeImplementation();
		if (
			cubeImplementation != null
			&&
			cubeImplementation.isEmpty()
		) {
			parent.getCubeImplementation().removeChild( key );
			parent.possiblyRemoveCubeImplementation();
		}
	}
}
