// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.edifact;

import java.lang.Character;
import java.lang.String;
import java.io.Reader;
import java.io.IOException;
import java.lang.NullPointerException;
import java.lang.IllegalArgumentException;

import Int.ecb.logging.Level;

import Int.ecb.gesmescb.util.ComponentHelper;
import Int.ecb.gesmescb.util.FastStringBuffer;

/* ***************************** Class ****************************** */

/**
 * <P>The class <CODE>EDIFACTReadBuffer</CODE>
 * acts as a fast buffer to treat incoming GESMES/CB data.
 * It is used by the {@link EDIFACTReader}.</P>
 *
 * <P>Its main task is to ignore control characters and
 * help split the input data into tokens,
 * which are either data elements or component elements.
 * The <CODE>EDIFACTReadBuffer</CODE> implements its own buffering scheme.
 * Thus no external {@link java.io.BufferedReader} is needed.</P>
 *
 * <P>Choosing a reasonably large buffer reduces the effect
 * of i/o overhead if reading is done directly from
 * a {@link java.io.FileReader} and not a {@link java.io.BufferedReader}.
 * But note that a too large buffer might impact cache memory usage.</P>
 *
 * <P>The <CODE>EDIFACTReadBuffer</CODE> is not internally synchronized.
 * External synchronisation is necessary if multiple threads want to
 * use the same <CODE>EDIFACTReadBuffer</CODE>.</P>
 *
 * @author
 * <A HREF="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</A>
 *
 * @version @(#)$Id: EDIFACTReadBuffer.java,v 1.1 2017/01/19 16:14:05 trainot Exp $
 */
class EDIFACTReadBuffer {

	/* ======================= Constants ======================== */

	/**
	 * A sensible default buffer capacity.
	 */
	private static final int DEFAULT_CAPACITY = 1024;

	/* ========================= Fields ========================= */

	/** To support logging. */
	private final ComponentHelper componentHelper;

	/**
	 * The Reader oject from which data should be read.
	 */
	private final Reader reader;

	/**
	 * The data input buffer.
	 * During normal operation, it contains no control characters.
	 * These are deleted just after having read from the Reader.
	 */
	private char[] buffer;

	/**
	 * The position of the characters in buffer[] as characters
	 * in the input data.
	 * Necessary to give information in case of errors.
	 * The array position[] is always as long as buffer[].
	 */
	private long[] position;

	/**
	 * The line numbers of the characters in buffer[] as
	 * character in the input data.
	 * Necessary to give information in case of errors.
	 * The array line[] is always as long as buffer[].
	 */
	private long[] line;

	/**
	 * The positions of the characters in buffer[] as
	 * characters in their respective lines.
	 * Note that control characters also occupy a "column"
	 * but are not contqained in buffer[].
	 * Necessary to give information in case of errors.
	 * The array column[] is always as long as buffer[].
	 */
	private long[] column;

	/**
	 * The number of valid characters in buffer[].
	 */
	private int top;

	/**
	 * The index of the next character to be read from buffer[].
	 * This is called the character at the <I>current position</I>.
	 */
	private int index;

	/**
	 * The current position in the input data of the character
	 * after the last character read from the reader.
	 * This field is used to process the raw input data
	 * containing control characters.
	 */
	private long topPosition;

	/**
	 * The line number in the input data of the character
	 * after the last character read from the reader.
	 * This field is used to process the raw input data
	 * containing control characters.
	 */
	private long topLine;

	/**
	 * The column number within the respective line
	 * in the input data of the character
	 * after the last character read from the reader.
	 * This field is used to process the raw input data
	 * containing control characters.
	 */
	private long topColumn;

	/**
	 * Signals that there was a '\r' without '\n' at the end
	 * of the last data read from the reader.
	 * This is needed to calculate correct line numbers on
	 * platforms which use only '\r'.
	 */
	private boolean pendingCR;

	/**
	 * Holds a FastStringBuffer which is used when Strings are compiled
	 * to be output.
	 * Keeping the FastStringBuffer only saves all that unneccessary
	 * construction and finalization.
	 * The FastStringBuffer contents are by no means essential to the
	 * status of the <CODE>EDIFACTReadBuffer</CODE>
	 * between calls to its methods.
	 */
	private final FastStringBuffer outputBuffer;

	/* ====================== Constructors ====================== */

	/**
	 * Construct an <CODE>EDIFACTReadBuffer</CODE>
	 * with specified minimum initial capacity
	 * from a {@link java.io.Reader}.
	 *
	 * @param reader where to read from.
	 * @param capacity the minimal initial buffer capacity to be used.
	 * If necessary, the <CODE>EDIFACTReadBuffer</CODE>
	 * may transparently increase
	 * its used internal buffer capacity during operation.
	 * Nevertheless, a reasonably large initial buffer capacity should be
	 * chosen in order to avoid loss of performance.
	 * @param componentHelper to facilitate logging.
	 *
	 * @throws NullPointerException
	 * if <CODE>reader</CODE> is <CODE>null</CODE>.
	 */
	public EDIFACTReadBuffer ( final Reader reader, int capacity, final ComponentHelper componentHelper )
		throws NullPointerException
	{
		this.componentHelper = componentHelper;

		if ( ( this.reader = reader ) == null ) {
			throw new NullPointerException(
				"Construction of a "
				+ getClass().getName()
				+ " from a null pointer is inhibited."
			);
		}

		// Since the EDIFACTReadBuffer does not work with
		// buffer sizes less than one, increase such a capacity.
		if ( capacity < 1 ) {
			capacity = DEFAULT_CAPACITY;
		}
		buffer = new char[ capacity ];

		// The error information arrays.
		position = new long[ capacity ];
		line = new long[ capacity ];
		column = new long[ capacity ];

		// Some further initialisation in order to be explicit.
		index = top = 0;	// Init pointers within the buffer[].

		// Though it is automatically done by Java.
		// This concerns the position etc. numbering.
		topPosition = 0;	// Set the position read so far.
		topLine = 0;		// Start with line number 0.
		topColumn = 0;		// And position 0 in the current line.

		pendingCR = false;	// No '\r' read yet.

		// Initialize the FastStringBuffer used for output buffer
		outputBuffer = new FastStringBuffer();
	}

	/**
	 * Construct an <CODE>EDIFACTReadBuffer</CODE>
	 * with default initial capacity
	 * from a {@link java.io.Reader}.
	 *
	 * @param reader where to read from.
	 * @param componentHelper to facilitate logging.
	 *
	 * @throws NullPointerException
	 * if <CODE>reader</CODE> is <CODE>null</CODE>.
	 */
	public EDIFACTReadBuffer ( final Reader reader, final ComponentHelper componentHelper )
		throws NullPointerException
	{
		this( reader, DEFAULT_CAPACITY, componentHelper );
	}

	/* ===================== Public Methods ===================== */

	/**
	 * Read the next character from the current position.
	 * Escaped characters are not translated.
	 *
	 * @return the next character in the input or -1 if the end of input
	 * has been reached.
	 *
	 * @throws IOException if the underlying Reader does so.
	 */
	public int readRawChar ()
		throws IOException
	{
		if ( request( 1 ) ) {
			return buffer[index++];
		} else {
			return -1;
		}
	}

	/**
	 * Read the next component from the current position.
	 * Escaped characters are not translated.
	 *
	 * @return the next component or null if end of input has been reached.
	 *
	 * @throws IOException if the underlying Reader does so.
	 */
	public String readRawComponent ()
		throws IOException
	{
		boolean escaped = false;	// escape char '?' encountered

		outputBuffer.clear();	// clear the output buffer

		// Test end of input.
		if ( !request( 1 ) ) return null;

		while (
			request( 1 )
			&&
			(
				buffer[index] != ':'
				&&
				buffer[index] != '+'
				&&
				buffer[index] != '\''
				||
				escaped
			)
		) {
			int nextBoundary;

			// To reduce calls to FastStringBuffer.append,
			// append not characters but array sections.
			// Seek component end or buffer end.
			for (
				nextBoundary = index;
				nextBoundary < top
				&&
				(
					buffer[nextBoundary] != ':'
					&&
					buffer[nextBoundary] != '+'
					&&
					buffer[nextBoundary] != '\''
					||
					escaped
				);
				++nextBoundary
			) {
				if ( escaped ) {
					escaped = false;
				} else if ( buffer[nextBoundary] == '?' ) {
					escaped = true;
				}
			}
			outputBuffer.append(
				buffer, index, nextBoundary - index
			);
			index = nextBoundary;
		}

		return outputBuffer.toString();
	}

	/**
	 * Read the next component from the current position
	 * translating escaped characters.
	 *
	 * @return the next component with escaped characters translated
	 * or null if the end of input has been reached.
	 *
	 * @throws IOException if the underlying Reader does so.
	 */
	public String readEscapedComponent ()
		throws IOException
	{
		// It is assumed that escaped characters are rare.
		// So read the string raw first and then check for
		// the escape character.
		String string = readRawComponent();

		// Also simply forward null references.
		if ( string != null && string.indexOf( '?' ) >= 0 ) {
			final int stringLen = string.length();

			outputBuffer.clear();	// clear output buffer

			int i = 0;	// Position in string

			while ( i < stringLen ) {
				int next = string.indexOf( '?', i );
				// If there is a '?'.
				if ( next >= 0 ) {
					// Add the string before '?'.
					outputBuffer.append(
						string, i, next - i
					);
					i = next + 1;
					// Add the character after '?'
					// if it exists.
					if ( i < stringLen ) {
						outputBuffer.append(
							string.charAt( i )
						);
					}
					// And jump over written substring
					// plus escape sequence.
					++i;
				} else {
					// Add all the rest.
					outputBuffer.append(
						string, i
					);
					// Ready to leave loop
					i = stringLen;
				}
			}

			string = outputBuffer.toString();
		}

		return string;
	}

	/**
	 * Try to read the given character from the the current position.
	 * If it is not there, then no character is read and false is
	 * returned.
	 * Escape sequences are not translated.
	 *
	 * @param c Character to compare with the character on the current
	 * position.
	 *
	 * @return true if and only if the character on the
	 * current position exists and matches c.
	 * Note that false is returned at end of input.
	 *
	 * @throws IOException if the underlying Reader does so.
	 */
	public boolean verifyRawChar ( final char c )
		throws IOException
	{
		if ( request( 1 ) && buffer[index] == c ) {
			++index;
			return true;
		} else {
			return false;
		}
	}

	/**
	 * Try to read the given string from the the current position.
	 * If it is not there, then no characters are read and false is
	 * returned.
	 * Escape sequences are not translated.
	 *
	 * @param string String to compare with the characters
	 * starting at the current position.
	 *
	 * @return true if and only if the characters starting at the
	 * current position match string.
	 * Note that false is returned if the end of input is encountered
	 * before string has been completely matched.
	 *
	 * @throws IOException if the underlying Reader does so.
	 */
	public boolean verifyRawString ( final String string )
		throws IOException
	{
		int stringLen = string.length();

		if ( request( stringLen ) ) {
			int i;
			// Compare string to buffer[] contents.
			for (
				i = 0;
				i < stringLen
				&&
				buffer[index+i] == string.charAt( i );
				++i
			);
			// If all match, mark as read and return true.
			if ( i >= stringLen ) {
				index += stringLen;
				return true;
			}
		}

		// If end of input or mismatch.
		return false;
	}

	/**
	 * Query if there is still data to be read from the EDIFACTBuffer.
	 *
	 * @return true if and only if there is some data left.
	 * This means that the end of input has not yet been reached.
	 *
	 * @throws IOException if the underlying Reader does so.
	 */
	public boolean dataAvailable ()
		throws IOException
	{
		// A faster implementation would remember
		// the end of input in a member variable.
		return request(1);
	}

	/**
	 * Query the current position in the input data.
	 * Position counting starts at 0.
	 *
	 * @return the current position.
	 */
	public long getPosition () {
		if ( index < top ) {
			return position[index];
		} else {
			return topPosition;
		}
	}

	/**
	 * Query the current line number in the input data.
	 * Line number counting starts at 0.
	 * This is not what usual text editors do.
	 *
	 * @return the current line number.
	 */
	public long getLineNumber () {
		if ( index < top ) {
			return line[index];
		} else {
			return topLine;
		}
	}

	/**
	 * Query the current position in the current input line.
	 * Column counting starts at 0 in each line and
	 * also control characters count for one column.
	 * This is not what usual text editors do.
	 *
	 * @return the current position in the current line.
	 */
	public long getColumn () {
		if ( index < top ) {
			return column[index];
		} else {
			return topColumn;
		}
	}

	/**
	 * View a few characters from the current position.
	 *
	 * @return the present buffer[] contents.
	 */
	public String getTextSample () {
		// Truncate longer buffer contents in the dump String.
		int len = 64;
		String truncateDots = "...";

		// If the buffer[] contents are not too long,
		// they are not truncated.
		if ( top - index <= len ) {
			len = top - index;
			truncateDots = "";
		}
		return new String(
			buffer, index, len
		) + truncateDots;
	}

	/**
	 * Transform the present valid buffer[] contents into a String.
	 *
	 * @return a human readable dump
	 * of the <CODE>EDIFACTReadBuffer</CODE> and
	 * some of its contents.
	 */
	public String toString () {
		return super.toString()
			+ ": buffer["
			+ getPosition()
			+ ','
			+ getLineNumber()
			+ ','
			+ getColumn()
			+ "]=\""
			+ getTextSample()
			+ '"';
	}

	/* ==================== Private Methods ===================== */

	/**
	 * Request that buffer[] contain at least count noncontrol characters
	 * starting from the current position unless the end of file is hit.
	 * This important method handles all reading from reader
	 * into the buffer.
	 *
	 * @param count indicates the number of characters requested.
	 *
	 * @return true if the request could be successfully served
	 * and false if the underlying Reader reached the end of data.
	 *
	 * @throws IOException if the underlying Reader does so.
	 */
	private boolean request ( final int count )
		throws IOException
	{
		// If loading new data is necessary
		if ( index + count > top ) {

			boolean endOfInput;

			// Shift left remaining buffer[] contents.
			forget();

			// If the requested amount of data is too large,
			// increase the buffer capacity.
			if ( count > buffer.length ) {
				// Resize to a sufficiently large power of 2.
				// Resizing to count itself could lead
				// to too much resizing.
				// The factor 2 helps avoid
				// excessive reading activity
				// in the pathological case of
				// extremely many control
				// characters in the input stream.
				resize(
					FastStringBuffer.powerBound( 2*count )
				);
			}

			// Then read as much as possible into buffer[]
			do {
				int rawTop = top + reader.read(
					buffer,
					top,
					buffer.length - top
				);

				// End of input encountered?
				endOfInput = rawTop < buffer.length;

				// Compress away the control characters read
				// and incorporate data to the end of buffer[].
				incorporate( rawTop );

			} while ( index + count > top && !endOfInput );
		}

		return index + count <= top;
	}

	/**
	 * Resize the internal buffer[].
	 * This is used to allow runtime buffer capacity variation,
	 * which is necessary to support the test and check methods.
	 *
	 * Presently, only increasing the buffer capacity is allowed.
	 *
	 * @param capacity the desired new buffer capacity.
	 *
	 * @throws IllegalArgumentException if an attempt is made to
	 * decrease the buffer capacity.
	 */
	private void resize ( final int capacity )
		throws IllegalArgumentException
	{
		// Test if buffer[] would become smaller
		if ( capacity < buffer.length ) {
			throw new IllegalArgumentException(
				"Decreasing the buffer capacity is forbidden."
			);
		}
		if ( capacity > buffer.length ) {
			char[] newBuffer = new char[ capacity ];
			long[] newPosition = new long[ capacity ];
			long[] newLine = new long[ capacity ];
			long[] newColumn = new long[ capacity ];

			System.arraycopy( buffer, 0, newBuffer, 0, top );
			System.arraycopy( position, 0, newPosition, 0, top );
			System.arraycopy( line, 0, newLine, 0, top );
			System.arraycopy( column, 0, newColumn, 0, top );

			buffer = newBuffer;
			position = newPosition;
			line = newLine;
			column = newColumn;
		}
	}

	/**
	 * Forget the used-up buffer[] contents from the beginning.
	 */
	private void forget () {
		// Shift buffer contents to the beginning.
		System.arraycopy( buffer, index, buffer, 0, top - index );

		// Shift also adjoint information.
		System.arraycopy( position, index, position, 0, top - index );
		System.arraycopy( line, index, line, 0, top - index );
		System.arraycopy( column, index, column, 0, top - index );

		// Adjust the fill level and current index.
		top -= index;
		index = 0;
	}

	/**
	 * Incorporate newly read characters into the buffer.
	 * These characters still contain control characters,
	 * which must be deleted. The error information arrays
	 * position, line, and column must be updated accordingly.
	 *
	 * The assumption on entry is that the raw data is stored in
	 * buffer[] beginning at buffer[top] upwards.
	 *
	 * At the end, the <CODE>EDIFACTReadBuffer</CODE> should be
	 * left in a consistent state. The field top then has been
	 * increased to cover also the newly incorporated data.
	 *
	 * @param rawTop specifies the index after the last raw data
	 * within buffer[].
	 */
	private void incorporate ( final int rawTop ) {
		int i;

		// Delete control characters.
		for ( i = top; i < rawTop; ++i ) {

			// Remember any pending CR.
			// Interpret lonely characters '\r' as fully
			// valid new line indicators.
			// If there is a '\n', then the pending '\r' was
			// part of a "\r\n" sequence and thus
			// does not count.
			if ( pendingCR && buffer[i] != '\n' ) {
				++topLine;
				topColumn = 0;
				pendingCR = false;
			}

			// If noncontrol
			if ( !Character.isISOControl( buffer[i] ) ) {
				buffer[top] = buffer[i];
				position[top] = topPosition++;
				line[top] = topLine;
				column[top] = topColumn++;
				++top;
			// if LF
			} else if ( buffer[i] == '\n' ) {
				++topPosition;
				++topLine;
				topColumn = 0;
				// if there was a pending '\r', then it was
				// part of a "\r\n" sequence.
				pendingCR = false;
			// if CR
			} else if ( buffer[i] == '\r' ) {
				++topPosition;
				++topColumn;
				pendingCR = true;
			// if other control character
			} else {
				if ( "\t".indexOf( buffer[i] ) < 0 ) {
					componentHelper.logger.log(
						Level.WARNING,
						componentHelper.logName,
						"buffer.incorporate()",
						"Unusual control character Unicode 0x"
						+ Integer.toHexString( buffer[i] )
						+ " encountered in input data, line "
						+ this.getLineNumber() + ", column "
						+ this.getColumn() + "."
					);
				}
				++topPosition;
				// The column means the position in the
				// line, not a "print column".
				// So it is incremented.
				++topColumn;
			}
		}
	}
}
