// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.edifact;

import java.lang.Object;
import java.lang.String;
import java.lang.Boolean;
import java.lang.Character;
import java.io.Reader;
import java.io.StringReader;
import java.io.IOException;

import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;

import Int.ecb.gesmescb.util.ComponentHelper;

/* ***************************** Class ****************************** */

/**
 * The class <CODE>EDIFACTReadBufferTest</CODE> implements a few tests for the
 * {@link EDIFACTReadBuffer}.
 *
 * @author
 * <A HREF="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</A>
 *
 * @version @(#)$Id: EDIFACTReadBufferTest.java,v 1.1 2017/01/19 16:14:05 trainot Exp $
 */
public class EDIFACTReadBufferTest extends TestCase {

	/* ======================= Constants ======================== */

	/**
	 * Character constants to reduce number of objects needed.
	 */
	private static final Character
		APOSTROPH = new Character( '\'' ),
		PLUS = new Character( '+' ),
		COLON = new Character( ':' );

	/**
	 * Boolean constants to reduce number of objects needed.
	 */
	private static final Boolean
		FALSE = Boolean.FALSE,
		TRUE = Boolean.TRUE;

	/**
	 * The input to the {@link EDIFACTReadBuffer} under Test.
	 * This will be read into it by a {@link java.io.StringReader}.
	 */
	private static final String testInput
		= "sdf'sdgf'er'\n"
		+ "AUX+23:33+29+au:de:fr+12.33+23:11+'"
		+ "TUX+2+2:2+2:2:2+arabica+helvetica:bold:sans'\n"
		+ "TERR+tulip:large+foramen+pilotus:excelsior:"
		+ "stupefactus:nagomonascar'ATT'ATT+3'\n"
		+ "PRX:2+33:23+'\n"
		+ "ATT+23?:44+?'He!?''---LOW---+Language:"
		+ "C?+?++Application:"
		+ "SUN?:JAVA:Version 1?':Release 0'\n"
		+ "++DEG?'?:?+:Hi'\n"
		+ "+++::+:++::+++:+:+:::::+:++:++:+:''\n"
		+ "Jull?\n"
		+ "'nag\r\no: Doc??Ho\r\nl\rli\n"
		+ "da?\n"
		+ "y'\n"
		+ "Test\rExit?";

	/**
	 * The objects into which
	 * {@link EDIFACTReadBuffer#readRawComponent}
	 * and {@link EDIFACTReadBuffer#readRawChar}
	 * should translate
	 * {@link #testInput} when called in the correct way.
	 */
	private static final Object[] testRawOutput = {
		"sdf", APOSTROPH, "sdgf", APOSTROPH, "er", APOSTROPH,
		"AUX", PLUS, "23", COLON, "33", PLUS, "29",
		PLUS, "au", COLON, "de", COLON, "fr", PLUS, "12.33",
		PLUS, "23", COLON, "11", PLUS, "", APOSTROPH,
		"TUX", PLUS, "2", PLUS, "2", COLON, "2",
		PLUS, "2", COLON, "2", COLON, "2",
		PLUS, "arabica", PLUS, "helvetica",
		COLON, "bold", COLON, "sans", APOSTROPH,
		"TERR", PLUS, "tulip", COLON, "large", PLUS, "foramen",
		PLUS, "pilotus", COLON, "excelsior", COLON,
		"stupefactus", COLON, "nagomonascar", APOSTROPH, "ATT",
		APOSTROPH, "ATT", PLUS, "3", APOSTROPH,
		"PRX", COLON, "2", PLUS, "33", COLON, "23",
		PLUS, "", APOSTROPH,
		"ATT", PLUS, "23?:44", PLUS, "?'He!?'",
		APOSTROPH, "---LOW---", PLUS, "Language", COLON,
		"C?+?+", PLUS, "Application", COLON,
		"SUN?:JAVA", COLON, "Version 1?'",
		COLON, "Release 0", APOSTROPH,
		"", PLUS, "", PLUS, "DEG?'?:?+", COLON, "Hi", APOSTROPH,
		"", PLUS, "", PLUS, "", PLUS, "", COLON, "", COLON, "",
		PLUS, "", COLON, "", PLUS, "", PLUS, "", COLON, "", COLON, "",
		PLUS, "", PLUS, "", PLUS, "", COLON, "", PLUS, "", COLON, "",
		PLUS, "", COLON, "", COLON, "", COLON, "", COLON, "",
		COLON, "", PLUS, "", COLON, "", PLUS, "", PLUS, "",
		COLON, "", PLUS, "", PLUS, "", COLON, "",
		PLUS, "", COLON, "", APOSTROPH, "", APOSTROPH,
		"Jull?'nago", COLON, " Doc??Hollida?y", APOSTROPH,
		"TestExit?"
	};

	/**
	 * The objects into which
	 * {@link EDIFACTReadBuffer#readEscapedComponent}
	 * and {@link EDIFACTReadBuffer#readRawChar}
	 * should translate
	 * {@link #testInput} when called in the correct way.
	 * The difference to {@link #testRawOutput} is that the
	 * '?' escape characters are removed unless they are
	 * escaped themselves.
	 */
	private static final Object[] testEscapedOutput = {
		"sdf", APOSTROPH, "sdgf", APOSTROPH, "er", APOSTROPH,
		"AUX", PLUS, "23", COLON, "33", PLUS, "29",
		PLUS, "au", COLON, "de", COLON, "fr", PLUS, "12.33",
		PLUS, "23", COLON, "11", PLUS, "", APOSTROPH,
		"TUX", PLUS, "2", PLUS, "2", COLON, "2",
		PLUS, "2", COLON, "2", COLON, "2",
		PLUS, "arabica", PLUS, "helvetica",
		COLON, "bold", COLON, "sans", APOSTROPH,
		"TERR", PLUS, "tulip", COLON, "large", PLUS, "foramen",
		PLUS, "pilotus", COLON, "excelsior", COLON,
		"stupefactus", COLON, "nagomonascar", APOSTROPH, "ATT",
		APOSTROPH, "ATT", PLUS, "3", APOSTROPH,
		"PRX", COLON, "2", PLUS, "33", COLON, "23",
		PLUS, "", APOSTROPH,
		"ATT", PLUS, "23:44", PLUS, "'He!'",
		APOSTROPH, "---LOW---", PLUS, "Language", COLON,
		"C++", PLUS, "Application", COLON,
		"SUN:JAVA", COLON, "Version 1'",
		COLON, "Release 0", APOSTROPH,
		"", PLUS, "", PLUS, "DEG':+", COLON, "Hi", APOSTROPH,
		"", PLUS, "", PLUS, "", PLUS, "", COLON, "", COLON, "",
		PLUS, "", COLON, "", PLUS, "", PLUS, "", COLON, "", COLON, "",
		PLUS, "", PLUS, "", PLUS, "", COLON, "", PLUS, "", COLON, "",
		PLUS, "", COLON, "", COLON, "", COLON, "", COLON, "",
		COLON, "", PLUS, "", COLON, "", PLUS, "", PLUS, "",
		COLON, "", PLUS, "", PLUS, "", COLON, "",
		PLUS, "", COLON, "", APOSTROPH, "", APOSTROPH,
		"Jull'nago", COLON, " Doc?Holliday", APOSTROPH,
		"TestExit"
	};

	/**
	 * The objects to which
	 * {@link EDIFACTReadBuffer#verifyRawString}
	 * should should compare {@link #testInput}.
	 */
	private static final Object[] testVerifyData = {
		"", TRUE,
		"s", TRUE, "g", FALSE, "de", FALSE,
		"df", TRUE,
		"'", TRUE,
		"", TRUE,
		"sdgf'er'AUX+23.", FALSE,
		"sdgf'er'AUX+23:33+29+au:de:fr+12.33+23:11+'", TRUE,
		"TUX+2+2:2+2:2:2+arabica+helvetica:bold:sans'TERR", TRUE,
		"\n", FALSE,
		"+tulip:large+foramen+pilotus:excelsior:", TRUE,
		"stupefactus:nagomonascar'ATT'ATT+3'\n", FALSE,
		"stupefactus:nagomonascar'ATT'ATT+3'", TRUE,
		"PRX:2+33:23+'", TRUE,
		"\n", FALSE,
		"\nATT", FALSE,
		"ATT+23?:44+?'He!?''---LOW---+Language:", TRUE,
		"C?+?++Application:SUN?:JAVA:Version 1?':Release 0'"
		+ "++DEG?'?:?+:Hi'+++::+:++::+++:+:+:::::+:++:++:+:''", TRUE,
		"Ju", TRUE,
		"ll'", FALSE,
		"ll?", TRUE,
		"'nago: ", TRUE,
		"Doc??Hollida?y'TestE", TRUE,
		"X", FALSE,
		"xit", TRUE,
		"", TRUE,
		"??", FALSE,
		"?", TRUE,
		"", TRUE,
		"?", FALSE
	};

	/* ========================= Fields ========================= */

	/** Drain for logging output. */
	private final ComponentHelper componentHelper = new ComponentHelper( "", this );

	/* ====================== Constructors ====================== */

	public EDIFACTReadBufferTest( String name ) {
		super( name );
	}

	/* ===================== Public Methods ===================== */

	/**
	 * Assembles and returns a test suite for all
	 * the test methods of this test case.
	 *
	 * @return a non-null test suite
	 */
	public static Test suite () {
		TestSuite suite = new TestSuite( EDIFACTReadBufferTest.class );
		return suite;
	}

	/**
	 * @param arg the array of argument {@link String}s.
	 */
	public static void main ( String[] arg ) {
		String[] testCaseName = {
			EDIFACTReadBufferTest.class.getName()
		};
		junit.textui.TestRunner.main( testCaseName );
	}

	/* ---------------------- Test Methods ---------------------- */

	/**
	 * Test {@link EDIFACTReadBuffer#readRawChar}.
	 *
	 * @throws IOException if the StringReader has a problem.
	 */
	public void testReadRawChar ()
		throws IOException
	{
		final Reader input = new StringReader( testInput );

		final EDIFACTReadBuffer buffer = new EDIFACTReadBuffer(
			input, 0, componentHelper
		);

		for ( int i = 0; i < testInput.length(); ++i ) {
			// Skip control characters in reference input.
			while (
				Character.isISOControl( testInput.charAt( i ) )
			) ++i;

			char c = testInput.charAt( i );

			assertTrue(
				"dataAvailable is true",
				buffer.dataAvailable()
			);
			assertTrue(
				"readRawChar = '" + c + "'",
				buffer.readRawChar() == c
			);
		}
		// Nothing left in the buffer.
		assertTrue(
			"Last readRawChar = -1",
			buffer.readRawChar() == -1
		);
		assertTrue(
			"datavailable is false",
			!buffer.dataAvailable()
		);

		// Also nothing in its input.
		assertTrue(
			"No input left.",
			input.read() == -1
		);
	}

	/**
	 * Test {@link EDIFACTReadBuffer#readRawComponent}.
	 *
	 * @throws IOException if the StringReader has a problem.
	 */
	public void testReadRawComponent ()
		throws IOException
	{
		doTestReadComponent( false );
	}

	/**
	 * Test {@link EDIFACTReadBuffer#readEscapedComponent}.
	 *
	 * @throws IOException if the StringReader has a problem.
	 */
	public void testReadEscapedComponent ()
		throws IOException
	{
		doTestReadComponent( true );
	}

	/**
	 * Test {@link EDIFACTReadBuffer#readRawComponent}
	 * or {@link EDIFACTReadBuffer#readEscapedComponent}.
	 *
	 * @param escaped
	 * if true then test {@link EDIFACTReadBuffer#readEscapedComponent}
	 * else test {@link EDIFACTReadBuffer#readRawComponent}.
	 *
	 * @throws IOException if the StringReader has a problem.
	 */
	private void doTestReadComponent (
		final boolean escaped
	)
		throws IOException
	{
		final Reader input = new StringReader( testInput );
		final Object[] referenceOutput
		= escaped ? testEscapedOutput : testRawOutput;
		final String methodName
		= escaped ? "readEscapedComponent" : "readRawComponent";

		final EDIFACTReadBuffer buffer = new EDIFACTReadBuffer(
			input, 0, componentHelper
		);

		for ( int i = 0; i < referenceOutput.length; ) {
			// Get the raw String first.
			assertTrue(
				methodName + " equals \""
				+ referenceOutput[i] + "\"",
				(
					escaped
					? buffer.readEscapedComponent()
					: buffer.readRawComponent()
				).equals( referenceOutput[i] )
			);
			++i;
			assertTrue(
				"dataAvailable is "
				+ new Boolean( i < referenceOutput.length ),
				buffer.dataAvailable()
				== i < referenceOutput.length
			);

			// Eventually leave the for-loop
			if ( i >= referenceOutput.length ) break;

			// Get the raw Character second.
			assertTrue(
				"readRawChar equals \'"
				+ referenceOutput[i] + "\'",
				buffer.readRawChar() == (
					(Character)referenceOutput[i]
				).charValue()
			);
			++i;
			assertTrue(
				"dataAvailable is "
				+ new Boolean( i < referenceOutput.length ),
				buffer.dataAvailable()
				== i < referenceOutput.length
			);
		}

		// Nothing in the input.
		assertTrue(
			"No input left.",
			input.read() == -1
		);
	}

	/**
	 * Test {@link EDIFACTReadBuffer#verifyRawChar}.
	 *
	 * @throws IOException if the StringReader has a problem.
	 */
	public void testVerifyRawChar ()
		throws IOException
	{
		final Reader input = new StringReader( testInput );

		final EDIFACTReadBuffer buffer = new EDIFACTReadBuffer(
			input, 0, componentHelper
		);

		for ( int i = 0; i < testInput.length(); ++i ) {
			char c = testInput.charAt( i );
			assertTrue(
				"dataAvailable is true",
				buffer.dataAvailable()
			);
			// Use controls as "false" input.
			if ( Character.isISOControl( c ) ) {
				assertTrue(
					"verifyRawChar(#"
					+ (int)c
					+ ") must be false",
					!buffer.verifyRawChar( c )
				);
			} else {
				assertTrue(
					"verifyRawChar('"
					+ c
					+ "') must be true",
					buffer.verifyRawChar( c )
				);
			}
		}
		// Nothing left in the buffer.
		assertTrue(
			"Last readRawChar = -1",
			buffer.readRawChar() == -1
		);
		assertTrue(
			"datavailable is false",
			!buffer.dataAvailable()
		);

		// Also nothing in its input.
		assertTrue(
			"No input left.",
			input.read() == -1
		);
	}

	/**
	 * Test {@link EDIFACTReadBuffer#verifyRawString}.
	 *
	 * @throws IOException if the StringReader has a problem.
	 */
	public void testVerifyRawString ()
		throws IOException
	{
		final Reader input = new StringReader( testInput );

		final EDIFACTReadBuffer buffer = new EDIFACTReadBuffer(
			input, 0, componentHelper
		);

		for ( int i = 0; i < testVerifyData.length; i += 2 ) {
			String string = (String)testVerifyData[i];
			Boolean result = (Boolean)testVerifyData[i+1];

			// Verify the raw String.
			assertTrue(
				"verifyRawString(\""
				+ string + "\") == "
				+ result.toString(),
				buffer.verifyRawString( string )
				== result.booleanValue()
			);
		}

		// No data left.
		assertTrue(
			"dataAvailable is false",
			!buffer.dataAvailable()
		);

		// Nothing in the input.
		assertTrue(
			"No input left.",
			input.read() == -1
		);
	}
}
