// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.edifact;

import java.lang.String;
import java.lang.Exception;
import java.lang.NullPointerException;
import java.lang.IllegalArgumentException;

import Int.ecb.gesmescb.error.ErrorWatcher;	// For severity codes.

/* ***************************** Class ****************************** */

/**
 * Serves to communicate error conditions within
 * the {@link EDIFACTReader}.
 *
 * @author
 * <A HREF="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</A>
 *
 * @version @(#)$Id: EDIFACTReadException.java,v 1.1 2017/01/19 16:14:05 trainot Exp $
 */
class EDIFACTReadException extends Exception {

	/* ========================= Fields ========================= */

	/**
	 *  serialVersionUID generated by eclipse.
	 */
	private static final long serialVersionUID = -2644088502256260046L;

	/**
	 * The position within the input data to which
	 * the {@link EDIFACTReadException} refers.
	 *
	 * @serial
	 */
	private final long position;

	/**
	 * The line number to which the {@link EDIFACTReadException} refers.
	 *
	 * @serial
	 */
	private final long line;

	/**
	 * The column within the present line
	 * to which the {@link EDIFACTReadException} refers.
	 *
	 * @serial
	 */
	private final long column;

	/**
	 * A piece of offending text where the Exception refers to.
	 *
	 * @serial
	 */
	private final String text;

	/**
	 * The severity code of the error.
	 * The codes are specified in {@link ErrorWatcher}.
	 *
	 * @serial
	 */
	private final int severity;

	/* ====================== Constructors ====================== */

	/**
	 * Construct an <CODE>EDIFACTReadException</CODE>
	 * with a detailed description of the error encountered.
	 *
	 * @param description contains a description of the error.
	 * @param position the position of the error in the input data.
	 * @param line the line number of the error in the input data.
	 * @param column the column of the error in the given line.
	 * @param text sample text from the offending input data.
	 * @param severity of the {@link Int.ecb.gesmescb.event.ErrorEvent}
	 * to be generated
	 * as result of this <CODE>ReaderException</CODE>.
	 * This should be one of
	 * {@link ErrorWatcher#WARNING}, {@link ErrorWatcher#ERROR},
	 * or {@link ErrorWatcher#FAILURE}.
	 *
	 * @throws NullPointerException if description or text are null
	 * references.
	 * @throws IllegalArgumentException if the severity code is invalid.
	 */
	public EDIFACTReadException (
		final String description,
		final long position,
		final long line,
		final long column,
		final String text,
		final int severity
	)
		throws NullPointerException, IllegalArgumentException
	{
		super( description );
		if ( null == description ) {
			throw new NullPointerException(
				"Attempt to construct an EDIFACTReadException"
				+ " with null description."
			);
		}
		if ( null == text ) {
			throw new NullPointerException(
				"Attempt to construct an EDIFACTReadException"
				+ " with null text sample."
			);
		}
		if (
			severity < ErrorWatcher.WARNING
			||
			ErrorWatcher.FAILURE < severity
		) {
			throw new IllegalArgumentException(
				"Severity code is neither of"
				+ " WARNING, ERROR, or FAILURE."
			);
		}
		this.position = position;
		this.line = line;
		this.column = column;
		this.text = text;
		this.severity = severity;
	}

	/* ===================== Public Methods ===================== */

	/**
	 * Get the position where
	 * the <CODE>EDIFACTReadException</CODE> originated.
	 */
	public long getPosition () {
		return position;
	}

	/**
	 * Get the line number where
	 * the <CODE>EDIFACTReadException</CODE> originated.
	 */
	public long getLine () {
		return line;
	}

	/**
	 * Get the column number in the line where
	 * the <CODE>EDIFACTReadException</CODE>
	 * originated.
	 */
	public long getColumn () {
		return column;
	}

	/**
	 * Get the text sample from where
	 * the <CODE>EDIFACTReadException</CODE>
	 * originated.
	 */
	public String getText () {
		return text;
	}

	/**
	 * Get the severity of
	 * the <CODE>EDIFACTReadException</CODE>.
	 */
	public int getSeverity () {
		return severity;
	}

	/**
	 * Returns a <CODE>String</CODE> representation
	 * of the <CODE>EDIFACTReadException</CODE>.
	 *
	 * @return the <CODE>EDIFACTReadException</CODE>
	 * in human readable form.
	 */
	public String toString () {
		return super.toString()
			+ ": position = "
			+ position
			+ ", line = "
			+ line
			+ ", column = "
			+ column
			+ ", text = \""
			+ text
			+ ", severity = \""
			+ (String)ErrorWatcher.SEVERITY_TRANSLATOR.get( severity )
			+'"';
	}
}
