// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, 2003, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.edifact;

import java.lang.String;
import java.lang.Integer;
import java.lang.RuntimeException;

import java.io.FileInputStream;
import java.io.Reader;
import java.io.InputStreamReader;
//import java.io.BufferedReader;	// imported only for javadoc
import java.io.IOException;

import java.util.Collection;
import java.util.List;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Date;

import Int.ecb.logging.Level;

import Int.ecb.gesmescb.InterchangeBegin;
import Int.ecb.gesmescb.InterchangeEnd;
import Int.ecb.gesmescb.ContactChannel;
import Int.ecb.gesmescb.Contact;
import Int.ecb.gesmescb.MessageBegin;
import Int.ecb.gesmescb.MessageEnd;
import Int.ecb.gesmescb.CodeList;
import Int.ecb.gesmescb.StatisticalConcept;
import Int.ecb.gesmescb.KeyFamily;
import Int.ecb.gesmescb.TimeSeries;
import Int.ecb.gesmescb.Attribute;
import Int.ecb.gesmescb.KeyFamilyComponent;
import Int.ecb.gesmescb.Observation;

import Int.ecb.gesmescb.metabase.Metabase;
import Int.ecb.gesmescb.metabase.MaintenanceAgency;

import Int.ecb.gesmescb.event.ErrorEvent;
import Int.ecb.gesmescb.event.GESMESCBEventSourceAdapter;
import Int.ecb.gesmescb.event.GESMESCBReader;

import Int.ecb.gesmescb.error.ErrorWatcher;	// For severity codes.

import Int.ecb.gesmescb.util.FastStringBuffer;
import Int.ecb.gesmescb.util.ComponentHelper;

import Int.ecb.gesmescb.edifact.EDIFACTReadBuffer;
import Int.ecb.gesmescb.edifact.EDIFACTReadException;

/* ***************************** Class ****************************** */

/**
 * <P>Reads GESMES/CB in EDIFACT syntax
 * and dispatches {@link Int.ecb.gesmescb.GESMESCBObject}s representing the
 * data read.</P>
 *
 * <P>The EDIFACTReader is not internally synchronized.
 * External synchronisation is necessary if multiple threads want to
 * use the same EDIFACTReader.</P>
 *
 * <center><table border="3" cellpadding="3">
 * <caption>
 * <a name="prefs">Available {@link Int.ecb.prefs preference} settings</a>
 * </caption>
 * <tr>
 * <th>preference name</th>
 * <th>type</th>
 * <th>meaning</th>
 * <th>default</th>
 * </tr>
 * <tr><td><code>source</code></td><td>string</td>
 *	<td>source filename</td></tr>
 * <tr><td><code>acceptCrap</code></td><td>boolean</td>
 *	<td>whether to issue {@link ErrorEvent#WARNING}s
 * in place of {@link ErrorEvent#ERROR}s on invalid
 * input - use only if you know what you are doing
 * since the {@link Int.ecb.gesmescb.semanticchecker.SemanticChecker}
 * is not able to detect some syntax-related errors</td>
 *	<td><code>false</code></td></tr>
 * </table></center>
 *
 * @author
 * <A HREF="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</A>
 *
 * @version @(#)$Id: EDIFACTReader.java,v 1.4 2018/09/04 09:22:00 kortest Exp $
 */
public class EDIFACTReader
	extends GESMESCBEventSourceAdapter
	implements GESMESCBReader, Constants
{
	/* ========================= Fields ========================= */

	/**
	 * The ComponentHelper used.
	 */
	private ComponentHelper componentHelper; //TODO make final

	/**
	 * The filename of the file read from if the EDIFACTReader has been
	 * constructed from a filename.
	 * Otherwise it is null.
	 */ 
	private String filename;

	/**
	 * The underlying Reader where the EDIFACTReader
	 * takes its input from.
	 */
	private final Reader reader;
	
	private final InputStreamReader inputStreamReader;
	

	/**
	 * The EDIFACTReadBuffer which will read the input.
	 */
	private final EDIFACTReadBuffer buffer;

	/**
	 * The present GESMES group number parsed.
	 * This is used to spot errors in the input file.
	 */
	private int presentGroupNumber;

	/**
	 * The present EDIFACT segment tag being parsed.
	 * This is used together with {@presentGroupNumber}
	 * to spot errors in the input file.
	 */
	private String presentSegmentTag;

	/**
	 * Counter used to check that the number of
	 * messages contained in an interchange
	 * matches the number shown in the final UNZ segment.
	 */
	private int messageCounter;

	/**
	 * Counter used to check that the number of
	 * segments contained in a message
	 * matches the number shown in the final UNT segment.
	 */
	private int segmentCounter;

	/**
	 * The name of the maintenance agency for the metadata.
	 * This variable is updated according to each message header
	 * read.
	 */
	private String maintenanceAgencyId;

	/**
	 * The present key family.
	 * It is needed to parse the time series keys
	 * in order to cope with the rotten GESMES/TS-EDIFACT syntax,
	 * which separates the period information from the first observation
	 * by a colon!
	 * It is also used to check consistency of incoming attributes' codedness.
	 */
	private KeyFamily keyFamily;

	/** Whether the current message is GESMES/CB, GESMES/TS, or SDMX-EDI. */
	private String messageFormat;

	/**
	 * Whether the current message is a delete message or not.
	 */
	private boolean deleteMessage;

	/**
	 * The preparation date and time of the current Interchange.
	 * To be used as timestamp for Structural Definitions.
	 */
	private Date versionDate;

	/**
	 * Last stored progress info result.
	 * @see #progressInfo
	 */
	private int lastMBread;

	/**
	 * Whether to generate {@link ErrorEvent#WARNING}s in place of
	 * {@link ErrorEvent#ERROR}s on invalid input.
	 *
	 * This is useful if data must be loaded from sources which
	 * are trusted though provenly incapable of generating valid
	 * Gesmes/CB.
	 */
	private final Level exceedUpperLimitLevel;

	/**
	 * If already warned for illegal spaces after segments.
	 */
	private boolean haveWarnedForSpaces;

	/**
	 * Level of the {@link ErrorEvent} to generate exceeding upper limits.
	 *
	 * This is useful if to get rid of the warning when reading e.g. 
	 * configuration files containing code lists with over 9999 elements,
	 * as is the case for financial data.
	 */
	private final boolean acceptCrap;	

	/* ====================== Constructors ====================== */

	/**
	 * Construct an <CODE>EDIFACTReader</CODE> to read from a Reader.
	 * Since an internal buffering scheme is used,
	 * no additional external buffering
	 * (e.g. with a {@link java.io.BufferedReader}) is necessary.
	 *
	 * <p>If the <code>EDIFACTReader</code>
	 * has been constructed
	 * from a filename
	 * using {@link #EDIFACTReader(String,String)},
	 * the input file is closed.</p>
	 *
	 * <p>The {@link Reader} is not closed
	 * if the <code>EDIFACTReader</code> has been
	 * created from a <code>Reader</code>
	 * using {@link #EDIFACTReader(String,Reader)}
	 * because then the <code>Reader</code> is considered
	 * not to be property of the <code>EDIFACTReader</code>.</p>
	 *
	 * @param name used for logging.
	 * @param reader where to read from.
	 */
	public EDIFACTReader ( final String name, final Reader reader ) {
		componentHelper = new ComponentHelper( name, this );

		// Remember the reader.
		this.reader = reader;
		
		// Create the buffer.
		buffer = new EDIFACTReadBuffer( reader, componentHelper );

		// Preset the filename field clear.
		this.filename = null;
		this.inputStreamReader = null;
		
		acceptCrap = "true".equalsIgnoreCase(
			componentHelper.getPreference( "acceptCrap", "false" )
		);
		exceedUpperLimitLevel = Level.parse(
			componentHelper.getPreference( "exceedUpperLimitLevel", "WARNING" )
		);

		init();
	}
	public EDIFACTReader (
			final String name,
			final InputStreamReader inputStreamReader
	) {
		componentHelper = new ComponentHelper( name, this );
		
		this.inputStreamReader = inputStreamReader;
		
		// Create the buffer.
		buffer
		= new EDIFACTReadBuffer( this.inputStreamReader, componentHelper );

		// Preset the filename field clear.
		this.filename = null;
		this.reader = null;

		acceptCrap = "true".equalsIgnoreCase(
			componentHelper.getPreference( "acceptCrap", "false" )
		);
		exceedUpperLimitLevel = Level.parse(
			componentHelper.getPreference( "exceedUpperLimitLevel", "WARNING" )
		);

		init();
	}
		

	/**
	 * Construct an <code>EDIFACTReader</code> to read from a file
	 * specified by the preference <code>source</code>.
	 *
	 * @throws IOException if the file cannot be opened for reading.
	 */
	public EDIFACTReader ()
		throws IOException
	{
		this( "" );
	}

	/**
	 * Construct an <code>EDIFACTReader</code> to read from a file
	 * specified by the preference <code>source<code>.
	 *
	 * @param name used for logging.
	 *
	 * @throws IOException if the file cannot be opened for reading.
	 */
	public EDIFACTReader ( final String name )
		throws IOException
	{
		this( name, (String) null );
	}

	/**
	 * Construct an <CODE>EDIFACTReader</CODE> to read from a file
	 * specified by a given filename.
	 *
	 * @param name used for logging.
	 * @param filename specifies in the local system's manner
	 * where to read from.
	 *
	 * @throws IOException if the file cannot be opened for reading.
	 */
	public EDIFACTReader ( final String name, final String filename )
		throws IOException
	{
		componentHelper = new ComponentHelper( name, this );

		// Remember the filename used.
		// This is also used as a flag
		// the "reader is owned by me and must be closed".
		this.filename = filename != null
			? filename
			: componentHelper.getPreference( "source", null );

		if ( this.filename == null ) {
			componentHelper.logger.log(
				Level.SEVERE, componentHelper.logName, "EDIFACTReader",
				"Cannot read preference \"source\"."
			);
		}

		try {
			this.inputStreamReader = null;
			
			this.reader = new InputStreamReader(
				new FileInputStream( filename ),
				ENCODING
			);
			buffer = new EDIFACTReadBuffer( reader, componentHelper );
		}
		catch ( final IOException ioe ) {
			componentHelper.logger.log(
				Level.SEVERE, componentHelper.logName, "EDIFACTReader",
				"Opening file failed: " + ioe.getMessage()
			);
			throw ioe;
		}

		acceptCrap = "true".equalsIgnoreCase(
			componentHelper.getPreference( "acceptCrap", "false" )
		);
		exceedUpperLimitLevel = Level.parse(
			componentHelper.getPreference( "exceedUpperLimitLevel", "WARNING" )
		);

		init();
	}

	/* ===================== Public Methods ===================== */

	/**
	 * Parse a GESMES/CB interchange in EDIFACT syntax into
	 * {@link Int.ecb.gesmescb.GESMESCBObject}s.
	 * These then are dispatched to all listeners.
	 */
	public void run () {
		try {
			final String origin
			= filename != null ? " from " + filename : "";

			componentHelper.logger.log(
				Level.INFO,
				componentHelper.logName,
				"run()",
				"Starting to parse GESMES/CB-EDIFACT or GESMES/TS-EDIFACT input"
				+ origin
				+ '.'
			);

			// Initialize megabyte counting for progress info.
			lastMBread = 0;

			// Read all interchanges.
			readInterchangeSequence();

			componentHelper.logger.log(
				Level.INFO,
				componentHelper.logName,
				"run()",
				"Ending to parse "
				+ messageFormat
				+ " input"
				+ origin
				+ ". Processed " + buffer.getPosition() + " byte."
			);
		}
		catch ( final EDIFACTReadException ere ) {
			complain( ere.getMessage() );
		}
		catch ( final IOException ioe ) {
			final String desc
			= "I/O problem: " + ioe + ". "
				+ positionString();
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"run()",
				desc,
				ioe
			);
			dispatch(
				new ErrorEvent( desc, ErrorEvent.FAILURE )
			);
		}
	}

	/**
	 * <P>Close the <CODE>EDIFACTReader</CODE>.</P>
	 *
	 * <P>This method should always be called
	 * when the <CODE>EDIFACTReader</CODE>
	 * has served its purpose.</P>
	 */
	public void close () {
		try {
			if ( filename != null && reader != null ) {
				reader.close();
				componentHelper.logger.log(
					Level.INFO,
					componentHelper.logName,
					"close()",
					"Closed "
					+ messageFormat
					+ " input"
					+ (
						filename != null
						? " from " + filename
						: ""
						)
					+ '.'
				);
			}
		}
		catch ( final IOException ioe ) {
			final String desc
			= "I/O error while closing input";
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"close()",
				desc,
				ioe
			);
			// Maybe, someone like a Metabase wants to know
			// about this failure.
			dispatch(
				new ErrorEvent( desc, ErrorEvent.FAILURE )
			);
		}
	}

	/**
	 * Return a string representation of this {@link EDIFACTReader}.
	 * This method is only intended to be used for informal purposes.
	 * Particularly the format of the returned String may vary
	 * between implementations.
	 *
	 * @return a human readable string representation.
	 */
	public String toString () {
		return componentHelper.name + (
			filename != null
			? "(\"" + filename + "\")"
			: "(" + reader.toString() + ")"
		);
	}

	/* ==================== Private Methods ===================== */

	/** Status initialisation. */
	private void init () {
		messageFormat = null;

		// Default setting for group number and TAG
		presentGroupNumber = Integer.MIN_VALUE;
		presentSegmentTag = "???";

		haveWarnedForSpaces = false;
	}

	/**
	 * Try to parse a sequence of GESMES/CB interchanges
	 * and dispatch all events accordingly.
	 *
	 * @throws {@link EDIFACTReadException} in case of erroneous input.
	 * @throws IOException if the underlying Reader does.
	 */
	private void readInterchangeSequence ()
		throws EDIFACTReadException, IOException
	{

		// Absolute top. Outside everything.
		messageCounter = -2;
		presentGroupNumber = -2;
		segmentCounter = -2;

		// First read the UNA segment
		// common to all interchanges
		// UNA
		presentSegmentTag = "UNA";
		expectRawString( "UNA:+.? " );

		endSegment();

		componentHelper.logger.log(
			Level.FINER,
			componentHelper.logName,
			"readInterchangeSequence()",
			"Found UNA."
		);

		// Read all available interchanges.
		int i;
		for ( i = 0; readInterchange(); ++i );

		// Require at least one interchange.
		if ( i < 1 ) {
			throwEDIFACTReadException(
				"No interchange header found."
			);
		} else if ( 1 < i ) {
			// TODO: BB: It seems the FAME writing part
			// has a problem with more than 1 interchange.
			// I think otherwise, it would be ok.
			throwEDIFACTReadException(
				"More than one interchange in the file."
			);
		}

		// Require no remaining text.
		if ( buffer.dataAvailable() ) throwEDIFACTReadException(
			"Unparsed input remaining.",
			ErrorWatcher.ERROR
		);
	}

	/**
	 * Try to parse a GESMES/CB interchange and dispatch all events
	 * accordingly. If no interchange header is found, no characters
	 * are read.
	 *
	 * @return true if an interchange was successfully parsed
	 * and false if no interchange header was found.
	 *
	 * @throws {@link EDIFACTReadException} in case of erroneous input.
	 * @throws IOException if the underlying Reader does.
	 */
	private boolean readInterchange ()
		throws EDIFACTReadException, IOException
	{
		// UNB
		presentSegmentTag = "UNB";
		if ( !buffer.verifyRawString( "UNB+" ) ) return false;

		componentHelper.logger.log(
			Level.FINER,
			componentHelper.logName,
			"readInterchange()",
			"Found UNB."
		);

		// No message yet.
		messageCounter = 0;

		// No group yet. Interchange top.
		presentGroupNumber = -1;

		expectRawString( "UNOC:3+" );

		final String sender = buffer.readEscapedComponent();
		expectRawChar( '+' );

		final String receiver = buffer.readEscapedComponent();
		expectRawChar( '+' );

		final String prepDate = buffer.readEscapedComponent();
		expectRawChar( ':' );

		final String prepTime = buffer.readEscapedComponent();
		expectRawChar( '+' );

		final String iref = buffer.readEscapedComponent();

		expectRawString( "++" );
		messageFormat = buffer.readEscapedComponent();
		if (
			! messageFormat.equals( GESMES_CB )
			&&
			! messageFormat.equals( GESMES_TS )
			&&
			! messageFormat.equals( SDMX_EDI )
		) {
			throwEDIFACTReadException(
				"Unsupported message format identifier \""
				+ messageFormat
				+ "\"."
			);
		}

		final boolean test = buffer.verifyRawString( "++++1" );

		segmentCounter = -1;
		endSegment();

		componentHelper.logger.log(
			Level.INFO,
			componentHelper.logName,
			"readInterchange()",
			"Interchange: " + iref
		);

		componentHelper.logger.log(
			Level.INFO,
			componentHelper.logName,
			"readInterchange()",
			"Sender: " + sender
		);

		componentHelper.logger.log(
			Level.INFO,
			componentHelper.logName,
			"readInterchange()",
			"Receiver: " + receiver
		);

		if ( test ) {
			componentHelper.logger.log(
				Level.INFO,
				componentHelper.logName,
				"readInterchange()",
				"This is a test interchange"
			);
		}

		// Construct the interchange data
		final InterchangeBegin interchangeBegin = new InterchangeBegin(
			iref,
			sender,
			receiver,
			prepDate + ':' + prepTime,
			test
		);

		componentHelper.logger.log(
			Level.FINE,
			componentHelper.logName,
			"readInterchange()",
			"Dispatch",
			interchangeBegin
		);

		// send to the listeners
		dispatch( interchangeBegin );

		// Store extraction Date
		versionDate = interchangeBegin.getPreparationDateTimeAsDate();
		if ( versionDate == null ) {
			final String desc = positionString()
				+ " Invalid preparation date and time \""
				+ interchangeBegin.getPreparationDateTime()
				+ "\".";
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"readInterchange()",
				desc
			);
			dispatch(
				new ErrorEvent(
					desc, ErrorEvent.ERROR
				)
			);
			versionDate = new Date();	// Take a kludge to avoid Exceptions.
		}

		// Parse all GESMES/CB messages
		// TODO: Should the EDIFACTReader check
		// mref == messageCounter? Or the SemanticChecker?
		while ( readMessage() );

		// UNZ
		presentSegmentTag = "UNZ";
		expectRawString( "UNZ+" );

		componentHelper.logger.log(
			Level.FINER,
			componentHelper.logName,
			"readInterchange()",
			"Found UNZ."
		);

		{	// Read and verify the number of messages read.
			final String numberOfMessages = buffer.readRawComponent();
			try {
				int noMe = Integer.parseInt(
					numberOfMessages
				);
				if ( noMe != messageCounter ) {
					throwEDIFACTReadException(
						"The asserted message count "
						+ noMe
						+ " does not match the actual"
						+ " number of messages, which is "
						+ messageCounter
						+ "."
					);
				}
			}
			catch ( final NumberFormatException nfe ) {
				throwEDIFACTReadException(
					"The asserted message count \""
					+ numberOfMessages
					+ "\" is not a parseable integer."
				);
			}
		}
		expectRawChar( '+' );
		expectRawString( iref );
		endSegment();

		// Interchange end event
		final InterchangeEnd interchangeEnd = new InterchangeEnd(
			iref
		);

		componentHelper.logger.log(
			Level.FINE,
			componentHelper.logName,
			"readInterchange()",
			"Dispatch",
			interchangeEnd
		);

		// send end event to listeners
		dispatch( interchangeEnd );

		// Additional check if there was at least one message.
		// The GESMES/CB specification requires this.
		if ( messageCounter == 0 ) {
			throwEDIFACTReadException(
				"Interchange contains no message."
			);
		}
		return true;
	}

	/**
	 * Try to parse a GESMES/CB message and dispatch all events
	 * accordingly. If no message header is found, no characters
	 * are read.
	 *
	 * @return true if a message was successfully parsed
	 * and false if no message header was found.
	 *
	 * @throws {@link EDIFACTReadException} in case of erroneous input.
	 * @throws IOException if the underlying Reader does.
	 */
	private boolean readMessage ()
		throws EDIFACTReadException, IOException
	{
		// The variables from which the Message will be
		// constructed.
		// The maintenance agency name is not here,
		// because it must be stored for further use
		// and hence is a field.
		// Resetting receiver and sender is necessary because
		// these variables are also used as flags.
		String
			mref,
			function,
			receiver = null,
			sender = null,
			identity = "";
		Collection contacts = Collections.EMPTY_LIST;

		// Resetting maintenanceAgencyId is necessary because
		// this variable is also used as a flag.
		maintenanceAgencyId = null;

		// Peek progress.
		progressInfo();

		// UNH
		// Try message header
		// and return false if there is no header.
		presentSegmentTag = "UNH";
		if ( !buffer.verifyRawString( "UNH+" ) ) return false;

		componentHelper.logger.log(
			Level.FINER,
			componentHelper.logName,
			"readMessage()",
			"Found UNH."
		);

		// Advance the message counter
		++messageCounter;

		// No group yet. Message top.
		presentGroupNumber = 0;

		// Reset the segment counter which counts the
		// segments per message.
		segmentCounter = 0;

		mref = buffer.readEscapedComponent();
		expectRawString( "+GESMES:2:1:E6" );

		endSegment();

		// BGM
		presentSegmentTag = "BGM";
		expectRawString( "BGM+" );
		function = buffer.readEscapedComponent();

		endSegment();

		// Group 2
		presentGroupNumber = 2;

		presentSegmentTag = "NAD";

		for ( int i = 1; i <= 3; ++i ) {
			// More precise specification of NAD
			presentSegmentTag = "NAD" + i;
			expectRawString( "NAD+" );

			if ( buffer.verifyRawString( "Z02+" ) ) {
				if ( maintenanceAgencyId != null ) {
					throwEDIFACTReadException(
						"Duplicate NAD+Z02"
					);
				}
				if ( i != 1 ) {
					final String desc
					= "Use of Z02 in the "
						+ i + ordinalEnding( i )
						+ " NAD segment is deprecated."
						+ " Use the first NAD segment instead.";
					componentHelper.logger.log(
						Level.WARNING,
						componentHelper.logName,
						"readMessage()",
						desc
					);
					dispatch(
						new ErrorEvent( desc, ErrorEvent.WARNING )
					);
				}

				maintenanceAgencyId
				= buffer.readEscapedComponent();
				endSegment();
			} else if ( buffer.verifyRawString( "MR+" ) ) {
				if ( receiver != null ) {
					throwEDIFACTReadException(
						"Duplicate NAD+MR"
					);
				}
				if ( i != 2 ) {
					final String desc
					= "Use of MR in the "
						+ i + ordinalEnding( i )
						+ " NAD segment is deprecated."
						+ " Use the second NAD segment instead.";
					componentHelper.logger.log(
						Level.WARNING,
						componentHelper.logName,
						"readMessage()",
						desc
					);
					dispatch(
						new ErrorEvent( desc, ErrorEvent.WARNING )
					);
				}

				receiver = buffer.readEscapedComponent();
				endSegment();
			} else if ( buffer.verifyRawString( "MS+" ) ) {
				if ( sender != null ) {
					throwEDIFACTReadException(
						"Duplicate NAD+MS"
					);
				}
				if ( i != 3 ) {
					final String desc
					= "Use of MS in the "
						+ i + ordinalEnding( i )
						+ " NAD segment is deprecated."
						+ " Use the third NAD segment instead.";
					componentHelper.logger.log(
						Level.WARNING,
						componentHelper.logName,
						"readMessage()",
						desc
					);
					dispatch(
						new ErrorEvent( desc, ErrorEvent.WARNING )
					);
				}

				sender = buffer.readEscapedComponent();
				endSegment();

				// try IDE
				presentSegmentTag = "IDE";
				if ( buffer.verifyRawString( "IDE+" ) ) {
					expectRawString( "10+" );

					// Discard message subject.
					// The Escaped-translation
					// is unnecessary here,
					// but does cost not much
					// and follows the
					// logics of the data.
					identity = buffer.readEscapedComponent();
					endSegment();
				}

				// Parse contact information
				contacts = readContactInfo();
			} else {
				throwEDIFACTReadException(
					"Invalid party type"
				);
			}
		}

		componentHelper.logger.log(
			Level.INFO,
			componentHelper.logName,
			"readInterchange()",
			"Message: " + mref
		);

		componentHelper.logger.log(
			Level.INFO,
			componentHelper.logName,
			"readMessage()",
			"Sender: " + sender
		);

		componentHelper.logger.log(
			Level.INFO,
			componentHelper.logName,
			"readMessage()",
			"Receiver: " + receiver
		);

		componentHelper.logger.log(
			Level.INFO,
			componentHelper.logName,
			"readInterchange()",
			"Subject: " + identity
		);

		// Now decide whether it is a metadata message or not.
		if ( function.equals( MessageBegin.STATISTICAL_DEFINITIONS ) ) {

			// Construct the interchange data
			final MessageBegin messageBegin = new MessageBegin(
				mref,
				sender,
				receiver,
				maintenanceAgencyId,
				function,
				identity,
				contacts
			);

			componentHelper.logger.log(
				Level.FINE,
				componentHelper.logName,
				"readMessage()",
				"Dispatch",
				messageBegin
			);

			// send to the listeners
			dispatch( messageBegin );

			// Store whether it is a delete message.
			deleteMessage = messageBegin.isDeleteAction();

			// Parse code list section
			readCodeLists();

			// Parse statistical concept section
			readStatisticalConcepts();

			// Parse key family structure definition section
			readKeyFamilies();

		} else if (	// Data related message
			function.equals( MessageBegin.STATISTICAL_DATA )
			||
			function.equals( MessageBegin.DATA_REQUEST )
		) {

			// Parse data set administration section.
			// The message event will contain data set information.
			String
				dataSetId,
				status,
				extractionDateTime,
				reportingPeriod,
				reportingPeriodFormat,
				keyFamilyId;

			// Group 13
			presentGroupNumber = 13;

			// DSI
			presentSegmentTag = "DSI";

			expectRawString( "DSI+" );

			componentHelper.logger.log(
				Level.FINER,
				componentHelper.logName,
				"readMessage()",
				"Found DSI."
			);

			dataSetId = buffer.readEscapedComponent();
			endSegment();

			componentHelper.logger.log(
				Level.INFO,
				componentHelper.logName,
				"readMessage()",
				"DataSet: " + dataSetId
			);

			// STS
			presentSegmentTag = "STS";
			expectRawString( "STS+3+" );
			status = buffer.readEscapedComponent();
			endSegment();

			// DTM 1
			presentSegmentTag = "DTM1";
			expectRawString( "DTM+242:" );
			extractionDateTime = buffer.readEscapedComponent();
			expectRawString( ":203" );
			endSegment();

			// DTM 2
			presentSegmentTag = "DTM1";
			if ( buffer.verifyRawString( "DTM+" ) ) {
				expectRawString( "Z02:" );
				reportingPeriod
				= buffer.readEscapedComponent();
				expectRawChar( ':' );
				reportingPeriodFormat
				= buffer.readEscapedComponent();
				endSegment();
			} else {
				reportingPeriod
				= reportingPeriodFormat
				= "";
			}

			// Group 14
			presentGroupNumber = 14;

			// IDE
			presentSegmentTag = "IDE";

			// Though Group 14 is marked as conditional
			// in the majority of branching diagrams in
			// The Guide, the semantics imply that it is
			// in fact mandatory, because the usage
			// rules for Group 13 imply Group 14 to be used.

			expectRawString( "IDE+5+" );
			keyFamilyId = buffer.readEscapedComponent();
			endSegment();

			componentHelper.logger.log(
				Level.INFO,
				componentHelper.logName,
				"readMessage()",
				"KeyFamily: " + keyFamilyId
			);

			// GIS 1
			presentSegmentTag = "GIS1";
			expectRawString( "GIS+AR3" );
			endSegment();

			// GIS 2
			presentSegmentTag = "GIS2";
			expectRawString( "GIS+1:::-" );
			endSegment();

			// Construct the Message and MessageEvent
			// containing data set information.
			// Construct the interchange data
			final MessageBegin messageBegin = new MessageBegin(
				mref,
				sender,
				receiver,
				maintenanceAgencyId,
				function,
				identity,
				contacts,
				dataSetId,
				status,
				extractionDateTime,
				reportingPeriod,
				reportingPeriodFormat,
				keyFamilyId
			);

			componentHelper.logger.log(
				Level.FINE,
				componentHelper.logName,
				"readMessage()",
				"Dispatch",
				messageBegin
			);

			// And let hear who has ears to hear.
			dispatch( messageBegin );

			// Store whether it is a delete message.
			deleteMessage = messageBegin.isDeleteAction();

			// All GESMES/CB implements of Group 17
			// is the ARR sequence.
			// Group 17
			presentGroupNumber = 17;

			// Determine the key length for the current
			// key family.
			// Since the rotten GESMES/CB syntax uses colons
			// to separate the period information
			// from the first observation in the ARR
			// segment, the Metabase must be queried.
			{
				final Metabase metabase
				= Metabase.getMetabase();
				final MaintenanceAgency agency
				= metabase.getMaintenanceAgency(
					maintenanceAgencyId
				);
				if ( agency == null ) {
					throwEDIFACTReadException(
						"The maintenance agency \""
						+ maintenanceAgencyId
						+ "\" is unknown"
						+ " to the metabase."
					);
				}
				keyFamily = agency.getKeyFamily( keyFamilyId );
				if ( keyFamily == null ) {
					throwEDIFACTReadException(
						"The key family \""
						+ keyFamilyId
						+ "\" is unknown to"
						+ " the maintenance agency \""
						+ maintenanceAgencyId
						+ "\"."
					);
				}
			}

			componentHelper.logger.log(
				Level.FINER,
				componentHelper.logName,
				"readMessage()",
				"The " + maintenanceAgencyId
				+ "'s keyfamily " + keyFamilyId
				+ " has a key length " + keyFamily.getKeyDimensionCount()
				+ "."
			);

			// ARRs
			for (
				int i = 0;
				readTimeSeries();
			) {
				// EDIFACT specifies maximum number of ARR segments.
				// The EDICACTReader implements this a a soft limit,
				// i.e. generates a warning only on breach.
				// Check for EDIFACT limits and increment i.
				final int MAX_ARR_LIMIT = 999000;
				if ( i < MAX_ARR_LIMIT ) {
					++i;
				} else if ( i == MAX_ARR_LIMIT ) {
					final String desc
					= "GESMES/CB in EDIFACT syntax permits a maximum of "
						+ MAX_ARR_LIMIT + " ARR segments in Group 17."
						+ " This limit has been broken. "
						+ positionString()
						+ ".";
					componentHelper.logger.log(
						exceedUpperLimitLevel,
						componentHelper.logName,
						"endSegment()",
						desc
					);
					dispatch(
						exceedUpperLimitLevel.intValue() >= Level.WARNING.intValue()
						? new ErrorEvent( desc, ErrorEvent.WARNING )
						: new ErrorEvent( desc, ErrorEvent.INFO)
					);
					++i;	// To generate warning only once.
				}
			}

			// Parse the attribute section.
			readAttributeSection();
		} else {
			throwEDIFACTReadException(
				"The message function \""
				+ function
				+ "\" is not supported."
			);
		}


		// UNT
		presentSegmentTag = "UNT";
		expectRawString( "UNT+" );

		componentHelper.logger.log(
			Level.FINER,
			componentHelper.logName,
			"readMessage()",
			"Found UNT."
		);

		{	// Read and verify the number of segments read.
			final String numberOfSegments = buffer.readRawComponent();
			try {
				int noSe = Integer.parseInt(
					numberOfSegments
				);
				// Compare asserted with actual
				// number of segments.
				// UNT counts, therefore +1.
				if ( noSe != segmentCounter+1 ) {
					throwEDIFACTReadException(
						"The asserted segment count "
						+ noSe
						+ " does not match the actual"
						+ " number of segments, which is "
						+ ( segmentCounter+1 )
						+ "."
					);
				}
			}
			catch ( final NumberFormatException nfe ) {
				throwEDIFACTReadException(
					"The asserted segment count \""
					+ numberOfSegments
					+ "\" is not a parseable integer."
				);
			}
		}
		expectRawChar( '+' );
		expectRawString( mref );
		endSegment();

		// Message end event
		final MessageEnd messageEnd = new MessageEnd(
			mref
		);

		componentHelper.logger.log(
			Level.FINE,
			componentHelper.logName,
			"readMessage()",
			"Dispatch",
			messageEnd
		);

		// send end event to listeners
		dispatch( messageEnd );

		// Signalize that there really was a message parsed
		return true;
	}

	/**
	 * Try to parse and check the conditional GESMES/CB group 3,
	 * which contains contact information.
	 *
	 * @throws {@link EDIFACTReadException} in case of erroneous input.
	 * @throws IOException if the underlying Reader does.
	 */
	private Collection readContactInfo ()
		throws EDIFACTReadException, IOException
	{
		final Collection contacts = new ArrayList();

		// Group 3
		presentGroupNumber = 3;

		// try 3 CTA
		presentSegmentTag = "CTA";
		for (
			int i = 1;
			i <= 3
			&&
			buffer.verifyRawString( "CTA+" );
			++i
		) {
			// More precise specification of CTA now available
			presentSegmentTag = "CTA" + i;

			final String function = buffer.readEscapedComponent();
			expectRawChar( '+' );
			final String identity = buffer.readEscapedComponent();
			expectRawChar( ':' );
			final String name = buffer.readEscapedComponent();
			endSegment();

			final Collection channels = new ArrayList();

			// COM
			presentSegmentTag = "COM";
			for (
				int j = 1;
				j <= 5
				&&
				buffer.verifyRawString( "COM+" );
				++j
			) {
				// More precise specification of COM
				presentSegmentTag = "COM" + j;

				// Discard communication number
				final String number = buffer.readEscapedComponent();
				expectRawChar( ':' );
				final String type = buffer.readEscapedComponent();
				endSegment();

				channels.add( new ContactChannel( type, number ) );
			}

			contacts.add( new Contact( function, identity, name, channels ) );
		}

		return contacts;
	}

	/**
	 * Parse the GESMES/CB codelist section
	 * and dispatch all events accordingly.
	 *
	 * @throws {@link EDIFACTReadException} in case of erroneous input.
	 * @throws IOException if the underlying Reader does.
	 */
	private void readCodeLists ()
		throws EDIFACTReadException, IOException
	{
		// From this, the CodeList will be constructed.
		String codeListId, codeListName;
		List codeValues, codeDescriptions;

		// Temporarily store the CodeList and CodeListEvent.
		CodeList codeList;

		// Group 4
		presentGroupNumber = 4;

		// VLI
		presentSegmentTag = "VLI";

		for (
			int i = 1;
			i <= 9999
			&&
			buffer.verifyRawString( "VLI+" );
			++i
		) {
			componentHelper.logger.log(
				Level.FINER,
				componentHelper.logName,
				"readCodeLists()",
				"Found VLI."
			);

			// Peek progress.
			progressInfo();

			// Group 4
			// Must be set inside the loop, too,
			// because there is also group 7.
			presentGroupNumber = 4;

			// VLI
			presentSegmentTag = "VLI" + i;

			codeListId = buffer.readEscapedComponent();
			expectRawString( "+++" );
			codeListName = buffer.readEscapedComponent();
			endSegment();

			// Construct Lists
			// for codes and their descriptions
			codeValues = new ArrayList( 256 );
			codeDescriptions = new ArrayList( 256 );

			// Group 7
			presentGroupNumber = 7;

			// CDV
			// EDIFACT specifies maximum number of CDV segments.
			// The EDICACTReader implements this a a soft limit,
			// i.e. generates a warning only on breach.
			// (Lisbeth 20080606 states this will be Gesmes standard
			// pretty soon.)
			final int MAX_CDV_LIMIT = 9999;

			presentSegmentTag = "CDV";
			for (
				int j = 1;
				buffer.verifyRawString( "CDV+" );
				++j
			) {
				if ( j == ( MAX_CDV_LIMIT + 1 ) ) {
					final String desc
					= "GESMES/CB in EDIFACT syntax permits a maximum of "
						+ MAX_CDV_LIMIT + " CDV segments in Group 7."
						+ " This limit has been broken. Should this "
						+ "WARNING complain about a structural file "
						+ "(see further in this WARNING), please "
						+ "ignore. Position:" + positionString() + ".";
					componentHelper.logger.log(
						exceedUpperLimitLevel,
						componentHelper.logName,
						"endSegment()",
						desc
					);
					dispatch(
						exceedUpperLimitLevel.intValue() >= Level.WARNING.intValue()
						? new ErrorEvent( desc, ErrorEvent.WARNING )
						: new ErrorEvent( desc, ErrorEvent.INFO)					);
				}

				// More precise specification of CDV
				presentSegmentTag = "CDV" + j;

				// Read codelist entry.
				codeValues.add(
					buffer.readEscapedComponent()
				);
				endSegment();

				codeDescriptions.add ( readFTX( 5 ) );
			}

			codeList = new CodeList(
				codeListId,
				codeListName,
				"",	// No description in EDIFACT
				versionDate,
				codeValues,
				codeDescriptions
			);

			componentHelper.logger.log(
				Level.FINE,
				componentHelper.logName,
				"readCodeLists()",
				"Dispatch",
				codeList
			);

			// dispatch to listeners.
			dispatch( codeList );
		}
	}

	/**
	 * Parse the GESMES/CB statistical concept definition section
	 * and dispatch all events accordingly.
	 *
	 * @throws {@link EDIFACTReadException} in case of erroneous input.
	 * @throws IOException if the underlying Reader does.
	 */
	private void readStatisticalConcepts ()
		throws EDIFACTReadException, IOException
	{
		// From this, the StatisticalConcept will be constructed.
		String
			statisticalConceptId,
			statisticalConceptName;

		// Temporarily store the StatisticalConcept
		// and StatisticalConceptEvent.
		StatisticalConcept statisticalConcept;

		// Group 9
		presentGroupNumber = 9;

		// STC
		presentSegmentTag = "STC";

		for (
			int i = 1;
			i <= 9999
			&&
			buffer.verifyRawString( "STC+" );
			++i
		) {
			componentHelper.logger.log(
				Level.FINER,
				componentHelper.logName,
				"readStatisticalConcepts()",
				"Found STC."
			);

			// Peek progress.
			progressInfo();

			// STC
			presentSegmentTag = "STC" + i;

			statisticalConceptId = buffer.readEscapedComponent();
			endSegment();

			// Idea for future:
			// Admit more 70 character components than one
			// for a longer name.
			statisticalConceptName = readFTX( 1 );

			// Construct the StatisticalConceptEvent.
			statisticalConcept = new StatisticalConcept(
				statisticalConceptId,
				statisticalConceptName,
				"",	// No description supported by EDIFACT
				versionDate
			);

			componentHelper.logger.log(
				Level.FINE,
				componentHelper.logName,
				"readStatisticalConcepts()",
				"Dispatch",
				statisticalConcept
			);

			// And dispatch it.
			dispatch( statisticalConcept );
		}
	}

	/**
	 * Parse the GESMES/CB key family structure definition section
	 * and dispatch all events accordingly.
	 *
	 * @throws {@link EDIFACTReadException} in case of erroneous input.
	 * @throws IOException if the underlying Reader does.
	 */
	private void readKeyFamilies ()
		throws EDIFACTReadException, IOException
	{
		// From this, the KeyFamily will be constructed.
		String keyFamilyId, keyFamilyName;

		// To be put into the KeyFamily.
		List keyFamilyComponents;

		// From these values, the keyFamilyComponent
		// will be constructed.
		String
			conceptId,
			conceptType,
			position,
			format,
			usage,
			attachmentLevel,
			codeListId;

		// To be put into the List keyFamilyComponents;
		KeyFamilyComponent keyFamilyComponent;

		// Temporarily store the CodeList and CodeListEvent.
		KeyFamily keyFamily;

		// Group 10
		presentGroupNumber = 10;

		// ASI
		presentSegmentTag = "ASI";

		for (
			int i = 1;
			i <= 9999
			&&
			buffer.verifyRawString( "ASI+" );
			++i
		) {
			componentHelper.logger.log(
				Level.FINER,
				componentHelper.logName,
				"readKeyFamilies()",
				"Found ASI."
			);

			// Peek progress.
			progressInfo();

			// Group 10
			// Must be set inside the loop, too,
			// because there are other groups in the loop.
			presentGroupNumber = 10;

			// ASI
			presentSegmentTag = "ASI" + i;

			keyFamilyId = buffer.readEscapedComponent();
			endSegment();

			// Idea for future: Admit more than one
			// 70 character components than one
			// for a longer name.
			keyFamilyName = readFTX( 1 );

			// Construct the list for the components.
			keyFamilyComponents = new ArrayList( 16 );

			// Group 11
			presentGroupNumber = 11;

			// SCD
			presentSegmentTag = "SCD";
			for (
				int j = 1;
				j <= 999	// 999 now.
				&&
				buffer.verifyRawString( "SCD+" );
				++j
			) {
				// Group 11
				// Must be set inside the loop, too,
				// because there is group 12
				// in the loop.
				presentGroupNumber = 11;

				// More precise specification of SCD
				presentSegmentTag = "SCD" + j;

				conceptType
				= buffer.readRawComponent();
				expectRawChar( '+' );

				conceptId = buffer.readEscapedComponent();

				// Attributes may lack the position parameter.
				if ( buffer.verifyRawString( "++++:" ) ) {
					position = buffer.readRawComponent();
				} else {
					position = "";
				}
				endSegment();

				// ATT 1
				presentSegmentTag = "ATT1";
				expectRawString( "ATT+3+5+:::" );

				format = buffer.readRawComponent();
				endSegment();

				// Try ATT2 and ATT3
				if ( buffer.verifyRawString( "ATT+" ) ) {
					// ATT 2
					presentSegmentTag = "ATT2";
					expectRawString(
						"3+35+"
					);
					usage = buffer.readRawComponent();
					expectRawString(
						":USS"
					);

					endSegment();

					// ATT 3
					presentSegmentTag = "ATT3";
					expectRawString(
						"ATT+3+32+"
					);
					attachmentLevel
					= buffer.readRawComponent();
					expectRawString(
						":ALV"
					);

					endSegment();
				} else {
					// No usage status given.
					usage = "";
					// No attachment level given.
					attachmentLevel = "";
				}

				// Group 12
				presentGroupNumber = 12;

				// Try IDE
				presentSegmentTag = "IDE";
				if ( buffer.verifyRawString( "IDE+" ) ) {
					expectRawString( "1+" );
					codeListId
					= buffer.readEscapedComponent();
					endSegment();
				} else {
					codeListId = "";
				}

				// Construct the dimension
				keyFamilyComponent
				= new KeyFamilyComponent(
					conceptId,
					conceptType,
					position,
					format,
					usage,
					attachmentLevel,
					codeListId,
					messageFormat.equals( GESMES_TS )
					||
					messageFormat.equals( SDMX_EDI )
				);

				// And add to the List
				keyFamilyComponents.add( keyFamilyComponent );
			}

			// Construct the KeyFamily.
			keyFamily = new KeyFamily(
				keyFamilyId,
				keyFamilyName,
				"",	// no description in EDIFACT
				versionDate,
				keyFamilyComponents
			);

			componentHelper.logger.log(
				Level.FINE,
				componentHelper.logName,
				"readKeyFamilies()",
				"Dispatch",
				keyFamily
			);

			// And let the listers hear.
			dispatch( keyFamily );
		}
	}

	/**
	 * Try to parse a Group 17 ARR segment and generate an according
	 * TimeSeriesEvent.
	 *
	 * @return true if an ARR segment was successfully parsed
	 * and false if no ARR segment header was found.
	 *
	 * @throws {@link EDIFACTReadException} in case of erroneous input.
	 * @throws IOException if the underlying Reader does.
	 */
	private boolean readTimeSeries ()
		throws EDIFACTReadException, IOException
	{
		// The variables from which the TimeSeries will be
		// constructed.
		String
			timeSeriesKey,
			timeSeriesPeriod,
			timeSeriesPeriodFormat;
		List observations;

		// Variables to store the TimeSeries and TimeSeriesEvent.
		TimeSeries timeSeries;

		// ARR
		// Try ARR header
		// and return false if there is no header.
		presentSegmentTag = "ARR";
		if ( !buffer.verifyRawString( "ARR+" ) ) return false;

		componentHelper.logger.log(
			Level.FINER,
			componentHelper.logName,
			"readTimeSeries()",
			"Found ARR."
		);

		// Peek progress.
		progressInfo();

		expectRawChar( '+' );

		// Get the series key.
		timeSeriesKey = readKey( keyFamily.getKeyDimensionCount() );

		// If a colon separates, then period info and
		// possibly data follow.
		// Else, the segment must end, which is used only
		// for time series and sibling group deletions.
		if ( buffer.verifyRawChar( ':' ) ) {

			// Get period and period format
			timeSeriesPeriod = buffer.readEscapedComponent();
			expectRawChar( ':' );
			timeSeriesPeriodFormat = buffer.readEscapedComponent();

			// Get the observations, if any.
			if ( buffer.verifyRawChar( ':' ) ) {
				observations = new ArrayList( 1024 );
				for (
					int i = 0;
					i < 9999
					&&
					(
						i == 0	// colon was read before
						||
						buffer.verifyRawChar( '+' )
					);
					++i
				) {
					// Variable to temporarily store an Observation
					// before it is inserted
					// in the observationsList.
					Observation observation;

					// The variables from which
					// the Observations will be
					// constructed.
					final String
						obsValue,
						obsStatus,
						obsConf,
						preBreakValue;

					obsValue = buffer.readEscapedComponent();

					// Watch out for empty observations
					// in ++ constructs like
					// ARR+...:+obs+++++++++++obs'.
					// "If the observation has a :, it
					// is deemed nonempty.
					if ( buffer.verifyRawChar( ':' ) ) {
						obsStatus = buffer.readEscapedComponent();

						if ( buffer.verifyRawChar( ':' ) ) {
							obsConf = buffer.readEscapedComponent();

							if ( buffer.verifyRawChar( ':' ) ) {
								preBreakValue
								= buffer.readEscapedComponent();
								if ( "".equals( preBreakValue ) ) {
									throwEDIFACTReadException(
										"Too many colons or violation of EDIFACT truncation rules"
										+ " where the observation"
										+ " pre-break value is expected."
									);
								}
							} else {
								preBreakValue = "";
								if ( "".equals( obsConf ) ) {
									throwEDIFACTReadException(
										"Violation of EDIFACT truncation rules"
										+ " where the observation"
										+ " confidentiality is expected."
									);
								}
							}
						} else {
							obsConf = preBreakValue = "";
							if ( "".equals( obsStatus ) ) {
								throwEDIFACTReadException(
									"Violation of EDIFACT truncation rules"
									+ " where the observation status"
									+ " is expected."
								);
							}
						}

						if ( "".equals( obsValue ) ) {
							throwEDIFACTReadException(
								"Violation of EDIFACT truncation rules"
								+ " where the observation value"
								+ " is expected."
							);
						}

						observation = new Observation(
							obsValue,
							obsStatus,
							obsConf,
							preBreakValue
						);
					} else {
						// The observation is empty.
						observation = Observation.EMPTY_OBSERVATION;

						// But make sure it really is,
						// since obsValue has already been read.
						if ( ! "".equals ( obsValue ) ) {
							throwEDIFACTReadException(
								"Nonempty observation \""
								+ obsValue
								+ "\" without observation status."
							);
						}
					}

					// Add the constructed observation.
					observations.add( observation );
				}
			} else {	// No observations. Useful for delete.
				// Do a little harmless check.
				if ( !deleteMessage ) {
					throwEDIFACTReadException(
						"No observations found."
					);
				}
				if ( "".equals( timeSeriesPeriodFormat ) ) {
					throwEDIFACTReadException(
						"Violation of EDIFACT truncation rules"
						+ " where the period format is expected."
					);
				}
				observations = Collections.EMPTY_LIST;
			}
		} else {	// For a TS or SG delete action.
			// Do a little harmless check.
			if ( !deleteMessage ) {
				throwEDIFACTReadException(
					"No period information found."
				);
			}
			timeSeriesPeriod
			= timeSeriesPeriodFormat = "";

			// No observations
			observations = Collections.EMPTY_LIST;
		}
		endSegment();

		// Construct TimeSeries and TimeSeriesEvent
		timeSeries = new TimeSeries(
			timeSeriesKey,
			timeSeriesPeriod,
			timeSeriesPeriodFormat,
			observations
		);

		componentHelper.logger.log(
			Level.FINE,
			componentHelper.logName,
			"readTimeSeries()",
			"Dispatch",
			timeSeries
		);

		// Let hear who has an ear.
		dispatch( timeSeries );

		// ARR segment successfully parsed.
		return true;
	}

	/**
	 * Try to parse the attribute section starting with FNS group 20
	 * and generate Attributes accordingly.
	 *
	 * @return true if the attribute section was successfully parsed
	 * and false if no FNS segment header was found.
	 *
	 * @throws {@link EDIFACTReadException} in case of erroneous input.
	 * @throws IOException if the underlying Reader does.
	 */
	private boolean readAttributeSection ()
		throws EDIFACTReadException, IOException
	{
		// The attachment scope
		// (misleadingly in The Guide also called "attachment level")
		// read from the REL segment.
		String attributeScope;

		// Attribute section
		presentGroupNumber = 20;

		// FNS
		presentSegmentTag = "FNS";

		if ( !buffer.verifyRawString( "FNS+" ) ) return false;

		componentHelper.logger.log(
			Level.FINER,
			componentHelper.logName,
			"readAttributeSection()",
			"Found FNS."
		);

		// Discard set-identifier text.
		buffer.readEscapedComponent();
		expectRawString( ":10" );

		endSegment();

		// Group 21
		presentGroupNumber = 21;

		// RELs
		presentSegmentTag = "REL";
		boolean foundREL = false;
		for (
			int i = 1;
			buffer.verifyRawString( "REL+" );
		) {
			foundREL = true;
			componentHelper.logger.log(
				Level.FINER,
				componentHelper.logName,
				"readAttributeSection()",
				"Found REL."
			);

			// Must be set also in the loop due to other groups
			presentGroupNumber = 21;

			// More specific as to REL
			presentSegmentTag = "REL" + i;

			// Check for EDIFACT limits and increment i.
			{
				final int MAX_REL_LIMIT = 9999;
				if ( i <= MAX_REL_LIMIT ) {
					++i;
				} else if ( i == MAX_REL_LIMIT+1 ) {
					final String desc
					= "GESMES/CB in EDIFACT syntax permits a maximum of "
						+ MAX_REL_LIMIT + " REL segments in Group 21."
						+ " This limit has been broken. "
						+ positionString()
						+ ".";
					componentHelper.logger.log(
						Level.WARNING,
						componentHelper.logName,
						"endSegment()",
						desc
					);
					dispatch(
						new ErrorEvent( desc, ErrorEvent.WARNING )
					);
					++i;	// To generate warning only once.
				}
			}

			expectRawString( "Z01+" );

			// Get attachment scope.
			attributeScope = buffer.readEscapedComponent();
			endSegment();

			// Group 22
			presentGroupNumber = 22;

			for (
				int j = 1;
				readAttributes( attributeScope );
			) {
				// Check for EDIFACT limits and increment j.
				final int MAX_ARR_LIMIT = 9999;
				if ( j <= MAX_ARR_LIMIT ) {
					++j;
				} else if ( j == MAX_ARR_LIMIT+1 ) {
					final String desc
					= "GESMES/CB in EDIFACT syntax permits a maximum of "
						+ MAX_ARR_LIMIT + " ARR segments in Group 22."
						+ " This limit has been broken. "
						+ positionString()
						+ ".";
					componentHelper.logger.log(
						Level.WARNING,
						componentHelper.logName,
						"endSegment()",
						desc
					);
					dispatch(
						new ErrorEvent( desc, ErrorEvent.WARNING )
					);
					++j;	// To generate warning only once.
				}
			}
		}

		if ( !foundREL ) {
			throwEDIFACTReadException(
				"An FNS segment is present, but no attributes follow."
			);
		}

		// An attribute section beginning with FNS was found.
		return true;
	}

	/**
	 * Try to parse all attributes for one object
	 * and generate Attributes accordingly.
	 * The object is the whole data set, a series or sibling group,
	 * or an single observation, according to the attachment level.
	 *
	 * @param attributeScope the attachment scope
	 * as specified by the preceding REL segment.
	 *
	 * @return true if an ARR segment header was successfully parsed
	 * and false if no ARR segment header was found.
	 *
	 * @throws {@link EDIFACTReadException} in case of erroneous input.
	 * @throws IOException if the underlying Reader does.
	 */
	private boolean readAttributes (
		final String attributeScope
	)
		throws EDIFACTReadException, IOException
	{
		// The variables from which the Attribute will be
		// constructed.
		String
			attributeName,
			attributeKey,
			attributePeriod,
			attributePeriodFormat,
			attributeCoded,
			attributeValue;

		// The Attribute and its event.
		Attribute attribute;

		// To get the key
		String lastDimensionPos;
		int lastDimensionPosInt;

		// No ARR header found ?
		if ( !buffer.verifyRawString( "ARR+" ) ) return false;

		componentHelper.logger.log(
			Level.FINER,
			componentHelper.logName,
			"readAttributes()",
			"Found ARR."
		);

		// Peek progress.
		progressInfo();

		// ARRs
		presentSegmentTag = "ARR";

		lastDimensionPos = buffer.readEscapedComponent();
		try {
			lastDimensionPosInt = Integer.parseInt(
				lastDimensionPos
			);
		}
		catch ( final NumberFormatException nfe ) {
			throwEDIFACTReadException(
				"Last dimension position \""
				+ lastDimensionPos
				+ "\" is not a parseable integer."
			);
			// Just to get rid of the warning about
			// uninitialized lastDimensionPosInt.
			throw new RuntimeException(
				"Hey, this cannot happen since"
				+ " I have already thrown up something."
			);
		}

		if ( lastDimensionPosInt > 0 ) {
			if ( DATA_SET_SCOPE.equals( attributeScope ) ) {
				complain(
					"Data set scope attributes require an empty key."
				);
			}
		} else {
			if ( SERIES_OR_SIBLING_SCOPE.equals( attributeScope ) ) {
				complain(
					"Series or sibling group scope attributes"
					+ " require a nonempty key."
				);
			}
			if ( OBSERVATION_SCOPE.equals( attributeScope ) ) {
				complain(
					"Observation scope attributes"
					+ " require a key and period specification."
				);
			}
		}

		// Compute the number of actual dimensions in the
		// key based on the scope.
		if ( attributeScope.equals( OBSERVATION_SCOPE ) ) {
			// The actual key is the two time specs shorter.
			lastDimensionPosInt -= 2;
		}

		// Check before key parsing
		if ( lastDimensionPosInt < 0 ) {
			throwEDIFACTReadException(
				"The number "
				+ lastDimensionPosInt
				+ " of dimensions in the key"
				+ " is negative."
			);
		}

		// Read the dimension values from the ARR segment
		if ( lastDimensionPosInt > 0 ) {
			expectRawChar( '+' );
			attributeKey = readKey( lastDimensionPosInt );
		} else {
			attributeKey = "";
		}

		// Read time specification
		// for observation scope attributes.
		if ( attributeScope.equals( OBSERVATION_SCOPE ) ) {
			if ( lastDimensionPosInt > 0 ) {
				expectRawChar( ':' );
			}
			attributePeriod = buffer.readEscapedComponent();
			expectRawChar( ':' );
			attributePeriodFormat = buffer.readEscapedComponent();
		} else {
			attributePeriod = attributePeriodFormat = "";
		}

		// Conclude ARR segment.
		endSegment();

		// Group 23
		presentGroupNumber = 23;

		// IDE
		presentSegmentTag = "IDE";

		for (
			int i = 1;
			buffer.verifyRawString( "IDE+" );
		) {
			// also set in loop.
			presentGroupNumber = 23;
			presentSegmentTag = "IDE" + i;

			// Check EDIFACT limits and increment i.
			{
				final int MAX_IDE_LIMIT = 9999;
				if ( i <= MAX_IDE_LIMIT ) {
					++i;
				} else if ( i == MAX_IDE_LIMIT+1 ) {
					final String desc
					= "GESMES/CB in EDIFACT syntax permits a maximum of "
						+ MAX_IDE_LIMIT + " IDE segments in Group 23."
						+ " This limit has been broken. "
						+ positionString()
						+ ".";
					componentHelper.logger.log(
						Level.WARNING,
						componentHelper.logName,
						"endSegment()",
						desc
					);
					dispatch(
						new ErrorEvent( desc, ErrorEvent.WARNING )
					);
					++i;	// To generate warning only once.
				}
			}

			// Is the attribute coded or uncoded?
			attributeCoded = buffer.readRawComponent();
			expectRawChar( '+' );

			// Attribute name.
			attributeName = buffer.readEscapedComponent();
			endSegment();

			if ( !deleteMessage ) {
				final KeyFamilyComponent component = keyFamily.getComponent(
					attributeName
				);

				// Finally read the attribute value
				if ( attributeCoded.equals( CODED ) ) {
					if ( component != null && !component.isCoded() ) {
						complain(
							"Wrongly representing value"
							+ " for uncoded attribute \""
							+ attributeName
							+ "\" as if it were for a coded attribute."
						);
					}

					// Group 24
					presentGroupNumber = 24;

					// CDV
					presentSegmentTag = "CDV";

					expectRawString( "CDV+" );
					attributeValue
					= buffer.readEscapedComponent();

					endSegment();
				} else if ( attributeCoded.equals( UNCODED ) ) {
					if ( component != null && component.isCoded() ) {
						complain(
							"Wrongly representing value"
							+ " for coded attribute \""
							+ attributeName
							+ "\" as if it were for an uncoded attribute."
						);
					}
					// FTXs
					// StringBuffer to concatenate the FTX
					// segments' contents.
					FastStringBuffer stringBuffer
					= new FastStringBuffer( 350 );
					int j;

					// One FTS segment's
					// contents at a time.
					String string;

					for (
						j = 0;
						null != (
							string = tryFTX( 5 )
						);
					) {
						// Check EDIFACT limits and increment j.
						{
							final int MAX_FTX_LIMIT = 20;
							if ( j < MAX_FTX_LIMIT ) {
								++j;
							} else if ( j == MAX_FTX_LIMIT ) {
								final String desc
								= "GESMES/CB in EDIFACT syntax permits"
									+ " a maximum of "
									+ MAX_FTX_LIMIT
									+ " FTX segments in Group 23."
									+ " This limit has been broken. "
									+ positionString()
									+ ".";
								componentHelper.logger.log(
									Level.WARNING,
									componentHelper.logName,
									"endSegment()",
									desc
								);
								dispatch(
									new ErrorEvent( desc, ErrorEvent.WARNING )
								);
								++j;	// To generate warning only once.
							}
						}
						stringBuffer.append( string );
					}

					if ( j <= 0 ) {
						throwEDIFACTReadException(
							"No FTX segment found"
							+ " containing the"
							+ " uncoded attribute value."
						);
					}

					attributeValue
					= stringBuffer.toString();
				} else {
					throwEDIFACTReadException(
						"Attribute \"" + attributeName
						+ "\" for key \"" + attributeKey
						+ "\" has invalid type \"" + attributeCoded
						+ "\" which neither signifies coded nor uncoded."
					);
					// Just to get rid of the warning about
					// uninitialized attributeValue.
					throw new RuntimeException(
						"Hey, this cannot happen since"
						+ " I have already thrown up something."
					);
				}
			} else {
				// No attribute value in a delete message.
				attributeValue = "";
			}

			// Construct the Attribute and Event
			attribute = new Attribute(
				attributeKey,
				attributePeriod,
				attributePeriodFormat,
				attributeName,
				attributeValue
			);

			componentHelper.logger.log(
				Level.FINE,
				componentHelper.logName,
				"readAttributes()",
				"Dispatch",
				attribute
			);

			// Listen who has ears to listen!
			dispatch( attribute );
		}

		// There was an ATT segment
		// plus depending segments found
		return true;
	}

	/**
	 * <P>Parse a GESMES/CB time series key.</P>
	 *
	 * Parsing ends when either <CODE>length</CODE>
	 * dimension values have been read
	 * or when a separator different from a colon
	 * is found. The latter condition allows to correctly parse
	 * end-wildcarded sibling group keys.</P>
	 * TODO: Keep leading/generate trailing dots for sibling
	 * keys?
	 * Like "M.AT.CPI.." and "..CPI.CH.Z", "..CPI..", "....", etc.
	 *
	 * @param length how many dimensions to parse.
	 *
	 * @return the {@link TimeSeries#KEY_SEPARATOR}-separated key.
	 *
	 * @throws {@link EDIFACTReadException} in case of erroneous input.
	 * @throws IOException if the underlying Reader does.
	 */
	private String readKey ( final int length )
		throws EDIFACTReadException, IOException
	{
		final FastStringBuffer stringBuffer
		= new FastStringBuffer( 64 );
		int i;
		for (
			i = 0;
			i < length
			&&
			(
				i == 0	// '+' separator was read before
				||
				buffer.verifyRawChar( ':' )
			);
			++i
		) {
			final String dimensionValue = buffer.readEscapedComponent();
			if (
				!acceptCrap
				&&
				dimensionValue.indexOf( TimeSeries.KEY_SEPARATOR ) >= 0
			) {
				final String desc = positionString()
					+ " Dimension value \"" + dimensionValue
					+ "\" contains a \"" + TimeSeries.KEY_SEPARATOR + "\".";
				componentHelper.logger.log(
					Level.SEVERE,
					componentHelper.logName,
					"readKey()",
					desc
				);
				dispatch(
					new ErrorEvent(
						desc, ErrorEvent.ERROR
					)
				);
			}
			if ( i > 0 ) {
				stringBuffer.append(
					TimeSeries.KEY_SEPARATOR
				);
			}
			stringBuffer.append( dimensionValue );
		}
		// Append leading dots for end-wildcarded
		// sibling group keys
		for ( ; i < length; ++i ) {
			if ( i > 0 ) {
				stringBuffer.append(
					TimeSeries.KEY_SEPARATOR
				);
			}
		}
		return stringBuffer.toString();
	}

	/**
	 * <P>Try to parse a GESMES/CB FTX segment to a String but return
	 * null if no FTX tag is found.</P>
	 *
	 * <P>Often, FTX segments follow the pattern
	 * <CODE>FTX+ACM+++</CODE>
	 * <I>part<SUB>1</SUB></I><CODE>:</CODE>
	 * <I>part<SUB>2</SUB></I><CODE>:</CODE>
	 * <I>part<SUB>3</SUB></I><CODE>:</CODE>
	 * <I>part<SUB>4</SUB></I><CODE>:</CODE>
	 * <I>part<SUB>5</SUB></I><CODE>'</CODE>.
	 * The result then is the concatenated string
	 * <I>part<SUB>1</SUB>
	 * part<SUB>2</SUB>
	 * part<SUB>3</SUB>
	 * part<SUB>4</SUB>
	 * part<SUB>5</SUB></I>.
	 * In order to allow for different numbers of components
	 * than 5, the number is specified in the parameter.</P>
	 *
	 * @param componentCount how many 70 character
	 * component elements should be admitted.
	 *
	 * @return the contents of the FTX segment or null if no
	 * FTX segment tag is found.
	 *
	 * @throws {@link EDIFACTReadException} in case of erroneous input.
	 * @throws IOException if the underlying Reader does.
	 */
	private String tryFTX ( final int componentCount )
		throws EDIFACTReadException, IOException
	{
		String text;

		// FTX
		presentSegmentTag = "FTX";

		if ( !buffer.verifyRawString( "FTX+" ) ) return null;

		expectRawString( "ACM" );

		// Follow the truncation rules.
		if ( buffer.verifyRawString( "+++" ) ) {
			// Used for long multi-component-text.
			FastStringBuffer stringBuffer
			= new FastStringBuffer( 350 );

			for ( int i = 1; i <= componentCount; ++i ) {
				final String component = buffer.readEscapedComponent();
				if ( component.length() > 70 ) {
					throwEDIFACTReadException(
						"FTX component longer than 70 characters: \""
						+ component + "\"."
					);
				}
				stringBuffer.append( component );
				if ( !buffer.verifyRawChar( ':' ) ) break;
			}
			text = stringBuffer.toString();
		} else {
			text = "";
		}

		endSegment();

		return text;
	}

	/**
	 * Parse a GESMES/CB FTX segment to a String and throw an
	 * EDIFACTReadException if there is none.
	 *
	 * @param componentCount how many 70 character
	 * component elements should be admitted.
	 *
	 * @return the contents of the FTX segment.
	 *
	 * @throws {@link EDIFACTReadException} in case of erroneous input.
	 * @throws IOException if the underlying Reader does.
	 *
	 * @see #tryFTX
	 */
	private String readFTX ( final int componentCount )
		throws EDIFACTReadException, IOException
	{
		final String string = tryFTX ( componentCount );
		if ( string == null ) {
			throwEDIFACTReadException(
				"FTX segment expected but not found."
			);
		}

		return string;
	}

	/**
	 * <P>Verify the character at the current position.
	 * If it equals the given argument then read it,
	 * else throw an {@link EDIFACTReadException}.</P>
	 *
	 * <P>No characters are read if the string does not match.
	 * Escape sequences are not translated.</P>
	 *
	 * <P>This method is intended to ease syntax checking when particular
	 * characters are expected, for instance syntax characters.</P>
	 *
	 * @param c Character to compare with the character on the current
	 * position.
	 *
	 * @throws {@link EDIFACTReadException}
	 * if the character on the current position
	 * does not match.
	 * @throws IOException if the underlying Reader does so.
	 */
	private void expectRawChar ( final char c )
		throws IOException, EDIFACTReadException
	{
		if ( !buffer.verifyRawChar( c ) ) {
			throwEDIFACTReadException(
				"Verify '" + c + "' failed."
			);
		}
	}

	/**
	 * <P>Verify the characters from the current position onwards.
	 * If they equal the given argument then read them,
	 * else throw an {@link EDIFACTReadException}.</P>
	 *
	 * <P>No characters are read if the string does not match.
	 * Escape sequences are not translated.</P>
	 *
	 * <P>This method is intended to ease syntax checking when particular
	 * strings are expected.</P>
	 *
	 * @param string String to compare with the characters
	 * starting at the current position.
	 *
	 * @throws {@link EDIFACTReadException}
	 * if the string on the current position
	 * does not match.
	 * @throws IOException if the underlying Reader does so.
	 */
	private void expectRawString ( final String string )
		throws IOException, EDIFACTReadException
	{
		if ( !buffer.verifyRawString( string ) ) {
			throwEDIFACTReadException(
				"Verify \"" + string + "\" failed."
			);
		}
	}

	// TODO: Replace this with error-reporting as in the catch-clause
	// but so as to allow continuation of process.
	/**
	 * Throw an {@link EDIFACTReadException} with a detailed description
	 * of the error encountered using data stored in
	 * the underlying {@link EDIFACTReadBuffer}.
	 * This function does not return normally
	 * but throw an {@link EDIFACTReadException} and unwind the stack.
	 * It is a convenience in order to easily construct
	 * an {@link EDIFACTReadException}.
	 *
	 * @param description contains a description of the error.
	 * @param severity for the <CODE>EDIFACTReadException</CODE>.
	 *
	 * @throws {@link EDIFACTReadException} as its desired result.
	 * @throws NullPointerException if the description is a null
	 * reference or if {@link #buffer} returns a null sample text.
	 * @throws IllegalArgumentException if the severity code is invalid.
	 *
	 * @see EDIFACTReadException#severity
	 */
	private void throwEDIFACTReadException (
		final String description,
		final int severity
	)
		throws EDIFACTReadException,
			NullPointerException, IllegalArgumentException
	{
		throw new EDIFACTReadException(
			// The +1 is necessary, because the human eye wants
			// counting to start at 1.
			description,
			buffer.getPosition() + 1,
			buffer.getLineNumber() + 1,
			buffer.getColumn() + 1,
			buffer.getTextSample(),
			severity
		);
	}

	/**
	 * Equivalent to {@link throwEDIFACTReadException(String,int)}
	 * with <CODE>severity</CODE>={@link ErrorWatcher#ERROR}.
	 *
	 * @param description contains a description of the error.
	 *
	 * @throws {@link EDIFACTReadException} as its desired result.
	 * @throws NullPointerException if the description is a null
	 * reference or if {@link #buffer} returns a null sample text.
	 */
	private void throwEDIFACTReadException (
		final String description
	)
		throws EDIFACTReadException, NullPointerException
	{
		throwEDIFACTReadException(
			description, ErrorWatcher.ERROR
		);
	}

	/** Complain to whoever listens or logs. */
	private void complain ( final String description ) {
		complain( description, ErrorWatcher.ERROR );
	}

	/** Complain to whoever listens or logs. */
	private void complain ( final String description, final int severity ) {
		final String desc
		= "Parsing problem: " + description + " "
		+ positionString();

		componentHelper.logger.log(
			severity >= ErrorWatcher.ERROR
				? Level.SEVERE
				: Level.WARNING,
			componentHelper.logName,
			"complain()",
			desc
		);
		dispatch(
			new ErrorEvent( desc, severity )
		);
	}

	/**
	 * <P>Describe the present position in the input stream.</P>
	 *
	 * <P>The returned String contains information about
	 * the present position within the input stream
	 * in a human readable format.
	 * This description is used to be enclosed in
	 * {@link ErrorEvent} objects.</P>
	 *
	 * @return human readable description of the present parsing position.
	 */
	private String positionString () {
		return
			"In " + messageCounter + ordinalEnding( messageCounter )
			+ " message, Group " + presentGroupNumber
			+ ", " + presentSegmentTag + "-segment. "
			+ ( filename != null ? "File " + filename + ", c" : "C" )
			// The following +1's are necessary
			// because the human eye wants counting to start at 1.
			+ "haracter " + ( buffer.getPosition() + 1 )
			+ ", line " + ( buffer.getLineNumber() + 1 )
			+ ", pseudocolumn " + ( buffer.getColumn() + 1 )
			+ ", next unparsed text \"" + buffer.getTextSample() + "\".";
	}

	/**
	 * <P>Get an english ordinal ending for a number.</P>
	 *
	 * <P>This is something like "1st", "2nd", "3rd", "4th", ... "23rd".
	 */
	private static String ordinalEnding ( int i ) {
		if ( i < 0 ) i = -i;
		i %= 100;
		if ( 11 <= i && i <= 13 ) return "th";
		switch ( i % 10 ) {
			case 1: return "st";
			case 2: return "nd";
			case 3: return "rd";
			default: return "th";
		}
	}

	/**
	 * Eventually log progress info for every MB of data read.
	 */
	private void progressInfo () {
		final int mb = (int)( buffer.getPosition() >> 20 );
		if ( mb > lastMBread ) {
			lastMBread = mb;

			componentHelper.logger.log(
				Level.INFO,
				componentHelper.logName,
				"progressInfo()",
				"Processed "
				+ mb
				+ " megabyte."
			);
		}
	}

	/**
	 * Finish a segment.
	 *
	 * @throws {@link EDIFACTReadException} if the terminator is missing.
	 * @throws IOException if the underlying Reader does so.
	 */
	private void endSegment ()
		throws IOException, EDIFACTReadException
	{
		expectRawChar( '\'' );

		// Skip spaces (illegal in GESMES/CB)
		if ( buffer.verifyRawChar( ' ' ) ) {
			if ( !haveWarnedForSpaces ) {
				final String desc
				= "Superfluous space character(s) after segment. "
					+ positionString()
					+ " (This warning is issued only once.)";
				componentHelper.logger.log(
					Level.WARNING,
					componentHelper.logName,
					"endSegment()",
					desc
				);
				dispatch(
					new ErrorEvent( desc, ErrorEvent.WARNING )
				);
			}
			haveWarnedForSpaces = true;
			while ( buffer.verifyRawChar( ' ' ) );
		}

		++segmentCounter;
	}
}
