// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.edifact;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.io.Reader;
import java.io.StringReader;
import java.io.Writer;
import java.util.Calendar;
import java.util.Date;
import java.util.Iterator;

import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;
import Int.ecb.gesmescb.Attribute;
import Int.ecb.gesmescb.CodeList;
import Int.ecb.gesmescb.Contact;
import Int.ecb.gesmescb.ContactChannel;
import Int.ecb.gesmescb.InterchangeBegin;
import Int.ecb.gesmescb.InterchangeEnd;
import Int.ecb.gesmescb.KeyFamily;
import Int.ecb.gesmescb.KeyFamilyComponent;
import Int.ecb.gesmescb.MessageBegin;
import Int.ecb.gesmescb.MessageEnd;
import Int.ecb.gesmescb.Observation;
import Int.ecb.gesmescb.StatisticalConcept;
import Int.ecb.gesmescb.TimeSeries;
import Int.ecb.gesmescb.event.ErrorEvent;
import Int.ecb.gesmescb.event.TestEventListener;
import Int.ecb.gesmescb.event.TestEventSource;
import Int.ecb.gesmescb.metabase.Metabase;

/* ***************************** Class ****************************** */

/**
 * Implements a few tests for the {@link EDIFACTReader}.
 *
 * @author
 * <A HREF="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</A>
 *
 * @version @(#)$Id: EDIFACTReaderTest.java,v 1.3 2018/09/04 09:58:42 kortest Exp $
 */
public class EDIFACTReaderTest extends TestCase implements Constants {

	/* ======================= Constants ======================== */

	/**
	 * The input to the {@link EDIFACTReader} under Test.
	 * This will be read into it by a {@link java.io.StringReader}.
	 * Its contents are meaningless but must correspond to
	 * {@link #testOutputCB}.
	 */
	private static final String testInputCB
		= "UNA:+.? '"
		+ "UNB+UNOC:3+4F0+ZZZ+001201:1410+IREF000215++GESMES/CB++++1'"
		+ "UNH+MREF000001+GESMES:2:1:E6'"
		+ "BGM+73'"
		+ "NAD+Z02+ECB'"
		+ "NAD+MR+ZZZ'"
		+ "NAD+MS+4F0'"
		+ "IDE+10+Test metadata'"
		+ "VLI+CL_OBJ+++Test codes'"
		+ "CDV+EPS'FTX+ACM'"
		+ "CDV+HOX'FTX+ACM+++Hoax'"
		+ "CDV+SOX'FTX+ACM+++Smelly: ObjeX'"
		+ "CDV+OBI'FTX+ACM+++German: drink'"
		+ "VLI+CL_CONF+++Further test codes'"
		+ "CDV+N'FTX+ACM+++Not classified'"
		+ "CDV+C'FTX+ACM+++Confidential'"
		+ "CDV+S'FTX+ACM+++Secret'"
		+ "CDV+T'FTX+ACM+++Top secret'"
		+ "CDV+U'FTX+ACM+++So: secret: that: unknown'"
		+ "STC+SECRECY'FTX+ACM'"
		+ "STC+INSANITY'FTX+ACM+++The degree of'"
		+ "ASI+SIS_FOOB'"
		+ "FTX+ACM+++Secret Intelligence Service?: Foobar stuff'"
		+ "SCD+4+WHO++++:1'ATT+3+5+:::AN3'IDE+1+CL_OBJ'"
		+ "SCD+4+SECURITY++++:2'ATT+3+5+:::AN1'IDE+1+CL_CONF'"
		+ "SCD+1+TIME_PERIOD++++:3'ATT+3+5+:::AN..35'"
		+ "SCD+1+TIME_FORMAT++++:4'ATT+3+5+:::AN3'"
		+ "SCD+3+OBS_VALUE++++:5'ATT+3+5+:::AN..15'"
		+ "SCD+3+OBS_STATUS++++:6'ATT+3+5+:::AN1'ATT+3+35+2:USS'ATT+3+32+5:ALV'IDE+1+CL_OBS_STATUS'"
		+ "SCD+3+OBS_CONF++++:7'ATT+3+5+:::AN1'ATT+3+35+1:USS'ATT+3+32+5:ALV'IDE+1+CL_OBS_CONF'"
		+ "SCD+3+OBS_PRE_BREAK++++:8'ATT+3+5+:::AN..15'ATT+3+35+1:USS'ATT+3+32+5:ALV'"
		+ "SCD+Z09+TITLE'ATT+3+5+:::AN..70'ATT+3+35+1:USS'ATT+3+32+1:ALV'"
		+ "SCD+Z09+UNIT'ATT+3+5+:::AN..12'ATT+3+35+2:USS'ATT+3+32+9:ALV'IDE+1+CL_UNIT'"
		+ "SCD+Z09+ORGAN'ATT+3+5+:::AN..70'ATT+3+35+2:USS'ATT+3+32+4:ALV'IDE+1+CL_UNIT'"
		+ "SCD+Z09+COMMENT'ATT+3+5+:::AN..1750'ATT+3+35+1:USS'ATT+3+32+5:ALV'"
		+ "UNT+77+MREF000001'"
		+ "UNH+MREF000002+GESMES:2:1:E6'"
		+ "BGM+74'"
		+ "NAD+Z02+ECB'"
		+ "NAD+MR+BIS'"
		+ "NAD+MS+4F0'"
		+ "IDE+10+Test data'"
		+ "CTA+CP+Secret Intelligence:Ms Aliia Mendez'"
		+ "CTA+CF+Public Accusor:Ms Hedwig Radovan'"
		+ "COM+hedwig.rado@radical.org:XF'"
		+ "COM+hedvig.radovan@securad.com:XF'"
		+ "CTA+CE+Criminal Action:Ms Victor Hugo'"
		+ "DSI+FOO_BAROON'"
		+ "STS+3+7'DTM+242:200008091923:203'"
		+ "IDE+5+SIS_FOOB'GIS+AR3'GIS+1:::-'"
		+ "ARR++HOX:S:2001:602:13.6:S:C'"
		+ "ARR++SOX:S:20102020:702:13.6:S:N+13.5:A+1:A:C:13+1.2:C'"
		+ "ARR++SOX:N:20102020:702:13.6:S:N+++++++13.5:A+1:A:C:13+1.2:C'"
		+ "FNS+Test Attributes:10'"
		+ "REL+Z01+1'"
		+ "ARR+0'"
		+ "IDE+Z10+STUS'CDV+RUHR'"
		+ "IDE+Z11+SLUS'FTX+ACM+++Kuli:muli:luli'"
		+ "REL+Z01+4'"
		+ "ARR+2+NOX:C'"
		+ "IDE+Z10+XCUS'CDV+XANA'"
		+ "ARR+2+:C'"
		+ "IDE+Z11+APPETITE'"
		+ "FTX+ACM+++This: is: an: attribute text: which'"
		+ "FTX+ACM+++ is: composed: of: multiple: FTX'"
		+ "FTX+ACM+++ segments.'"
		+ "ARR+2+NOX'"
		+ "IDE+Z10+BANANA'CDV+SWEET'"
		+ "REL+Z01+5'"
		+ "ARR+4+SOX:C:2020:602'"
		+ "IDE+Z11+COMMENT'"
		+ "FTX+ACM+++Text:0'"
		+ "FTX+ACM+++1'"
		+ "FTX+ACM+++2'"
		+ "FTX+ACM+++3'"
		+ "FTX+ACM+++4'"
		+ "UNT+48+MREF000002'"
		+ "UNH+MREF000003+GESMES:2:1:E6'"
		+ "BGM+74'"
		+ "NAD+Z02+ECB'"
		+ "NAD+MR+BIS'"
		+ "NAD+MS+4F0'"
		+ "IDE+10+Test data delete'"
		+ "CTA+CC+Secret Intelligence:Mr Orson Leigh'"
		+ "COM+00496913440:TE'"
		+ "COM+00496913446000:FX'"
		+ "COM+orson.leigh@radical.org:EM'"
		+ "CTA+CF+Private Investigation:Ms Ellison Khan'"
		+ "COM+ellison.khan.cube.radical.org:EM'"
		+ "CTA+CE+Criminal Action:Ms Victor Hugo'"
		+ "COM+action@radical-cuboid.org:EM'"
		+ "DSI+FOO_BALLOON'"
		+ "STS+3+6'DTM+242:200008091923:203'DTM+Z02:2000:602'"
		+ "IDE+5+SIS_FOOB'GIS+AR3'GIS+1:::-'"
		+ "ARR++HOX:S:2001:602'"
		+ "ARR++SOX:S'"
		+ "ARR++:S'"
		+ "ARR++SOX'"
		+ "FNS+Delete Attributes:10'"
		+ "REL+Z01+1'"
		+ "ARR+0'"
		+ "IDE+Z10+SMUS'"
		+ "IDE+Z11+SCUS'"
		+ "REL+Z01+4'"
		+ "ARR+2+NOX:S'"
		+ "IDE+Z11+XCAS'"
		+ "ARR+2+:N'"
		+ "IDE+Z10+VANITAS'"
		+ "ARR+2+NOX'"
		+ "IDE+Z10+FLUX'"
		+ "REL+Z01+5'"
		+ "ARR+4+SOX:C:2020:602'"
		+ "IDE+Z11+COMMENT'"
		+ "UNT+41+MREF000003'"
		+ "UNZ+3+IREF000215'";

	/**
	 * Version date of the read Key Family (=extraction date of Message).
	 */
	private static final Date versionDate;
	static {
		final Calendar calendar = Calendar.getInstance();
		calendar.clear();
		calendar.set( 2000, Calendar.DECEMBER, 1, 14, 10 );
		versionDate = calendar.getTime();
	}

	/** The Gesmes/CB key family needed for some data tests. */
	private static final KeyFamily SIS_FOOB = new KeyFamily(
		"SIS_FOOB", "Secret Intelligence Service: Foobar stuff", "",
		versionDate,
		new KeyFamilyComponent[] {
			new KeyFamilyComponent(
				"WHO", KeyFamilyComponent.KEY_STRUCTURE_DIMENSION,
				"1", "AN3", "", "", "CL_OBJ"
			),
			new KeyFamilyComponent(
				"SECURITY", KeyFamilyComponent.KEY_STRUCTURE_DIMENSION,
				"2", "AN1", "", "", "CL_CONF"
			),
			new KeyFamilyComponent(
				"TIME_PERIOD", KeyFamilyComponent.TIME_DIMENSION,
				"3", "AN..35", "", "", ""
			),
			new KeyFamilyComponent(
				"TIME_FORMAT", KeyFamilyComponent.TIME_DIMENSION,
				"4", "AN3", "", "", ""
			),
			new KeyFamilyComponent(
				"OBS_VALUE", KeyFamilyComponent.ARRAY_CELL,
				"5", "AN..15", "", "", ""
			),
			new KeyFamilyComponent(
				"OBS_STATUS", KeyFamilyComponent.ARRAY_CELL,
				"6", "AN1", "2", "5", "CL_OBS_STATUS"
			),
			new KeyFamilyComponent(
				"OBS_CONF", KeyFamilyComponent.ARRAY_CELL,
				"7", "AN1", "1", "5", "CL_OBS_CONF"
			),
			new KeyFamilyComponent(
				"OBS_PRE_BREAK", KeyFamilyComponent.ARRAY_CELL,
				"8", "AN..15", "1", "5", ""
			),
			new KeyFamilyComponent(
				"TITLE", KeyFamilyComponent.ATTRIBUTE,
				"", "AN..70", KeyFamilyComponent.CONDITIONAL,
				KeyFamilyComponent.ATTACHED_TO_DATA_SET,
				""
			),
			new KeyFamilyComponent(
				"UNIT", KeyFamilyComponent.ATTRIBUTE,
				"", "AN..12", KeyFamilyComponent.MANDATORY,
				KeyFamilyComponent.ATTACHED_TO_SIBLING_GROUP,
				"CL_UNIT"
			),
			new KeyFamilyComponent(
				"ORGAN", KeyFamilyComponent.ATTRIBUTE,
				"", "AN..70", KeyFamilyComponent.MANDATORY,
				KeyFamilyComponent.ATTACHED_TO_TIME_SERIES,
				"CL_UNIT"
			),
			new KeyFamilyComponent(
				"COMMENT", KeyFamilyComponent.ATTRIBUTE,
				"", "AN..1750", KeyFamilyComponent.CONDITIONAL,
				KeyFamilyComponent.ATTACHED_TO_OBSERVATION,
				""
			)
		}
	);

	/**
	 * The objects into which
	 * the {@link EDIFACTReader} should translate
	 * {@link #testInputCB} when called in the correct way.
	 */
	private static final Object[] testOutputCB = {
		new InterchangeBegin(
			"IREF000215", "4F0", "ZZZ", "001201:1410", true
		),
		new MessageBegin(
			"MREF000001", "4F0", "ZZZ", "ECB", "73", "Test metadata",
			new Contact[]{}
		),
		new CodeList(
			"CL_OBJ", "Test codes", "", versionDate,
			new String[] { "EPS", "HOX", "SOX", "OBI" },
			new String[] { "", "Hoax", "Smelly ObjeX", "German drink" }
		),
		new CodeList(
			"CL_CONF", "Further test codes", "", versionDate,
			new String[] { "N", "C", "S", "T", "U" },
			new String[] {
				"Not classified", "Confidential",
				"Secret", "Top secret", "So secret that unknown"
			}
		),
		new StatisticalConcept(
			"SECRECY", "", "", versionDate
		),
		new StatisticalConcept(
			"INSANITY", "The degree of", "", versionDate
		),
		SIS_FOOB,
		new MessageEnd(
			"MREF000001"
		),
		new MessageBegin(
			"MREF000002", "4F0", "BIS", "ECB", "74", "Test data",
			new Contact[]{
				new Contact(
					Contact.DATA_PROCESSING,
					"Secret Intelligence",
					"Ms Aliia Mendez",
					new ContactChannel[]{}
				),
				new Contact(
					Contact.HEAD_OF_INFORMATION_PRODUCTION,
					"Public Accusor",
					"Ms Hedwig Radovan",
					new ContactChannel[]{
						new ContactChannel(
							ContactChannel.X400,
							"hedwig.rado@radical.org"
						),
						new ContactChannel(
							ContactChannel.X400,
							"hedvig.radovan@securad.com"
						)
					}
				),
				new Contact(
					Contact.HEAD_OF_DATA_PROCESSING,
					"Criminal Action",
					"Ms Victor Hugo",
					new ContactChannel[]{}
				)
			},
			"FOO_BAROON",
			MessageBegin.UPDATE_OR_REPLACE_ACTION,
			"200008091923", "", "", "SIS_FOOB"
		),
		new TimeSeries(
			"HOX.S", "2001", "602",
			new Observation[] {
				new Observation( "13.6", "S", "C", "" )
			}
		),
		new TimeSeries(
			"SOX.S", "20102020", "702",
			new Observation[] {
				new Observation( "13.6", "S", "N", "" ),
				new Observation( "13.5", "A", "", "" ),
				new Observation( "1", "A", "C", "13" ),
				new Observation( "1.2", "C", "", "" )
			}
		),
		new TimeSeries(
			"SOX.N", "20102020", "702",
			new Observation[] {
				new Observation( "13.6", "S", "N", "" ),
				new Observation( "", "", "", "" ),
				new Observation( "", "", "", "" ),
				new Observation( "", "", "", "" ),
				new Observation( "", "", "", "" ),
				new Observation( "", "", "", "" ),
				new Observation( "", "", "", "" ),
				new Observation( "13.5", "A", "", "" ),
				new Observation( "1", "A", "C", "13" ),
				new Observation( "1.2", "C", "", "" )
			}
		),
		new Attribute(
			"", "", "", "STUS", "RUHR"
		),
		new Attribute(
			"", "", "", "SLUS", "Kulimuliluli"
		),
		new Attribute(
			"NOX.C", "", "", "XCUS", "XANA"
		),
		new Attribute(
			".C", "", "", "APPETITE",
			"This is an attribute text"
			+" which is composed of multiple FTX segments."
		),
		new Attribute(
			"NOX.", "", "", "BANANA", "SWEET"
		),
		new Attribute(
			"SOX.C", "2020", "602", "COMMENT", "Text01234"
		),
		new MessageEnd(
			"MREF000002"
		),
		new MessageBegin(
			"MREF000003", "4F0", "BIS", "ECB", "74", "Test data delete",
			new Contact[]{
				new Contact(
					Contact.INFORMATION_PRODUCTION,
					"Secret Intelligence",
					"Mr Orson Leigh",
					new ContactChannel[]{
						new ContactChannel(
							ContactChannel.TELEPHONE,
							"00496913440"
						),
						new ContactChannel(
							ContactChannel.FAX,
							"00496913446000"
						),
						new ContactChannel(
							ContactChannel.EMAIL,
							"orson.leigh@radical.org"
						)
					}
				),
				new Contact(
					Contact.HEAD_OF_INFORMATION_PRODUCTION,
					"Private Investigation",
					"Ms Ellison Khan",
					new ContactChannel[]{
						new ContactChannel(
							ContactChannel.EMAIL,
							"ellison.khan.cube.radical.org"
						)
					}
				),
				new Contact(
					Contact.HEAD_OF_DATA_PROCESSING,
					"Criminal Action",
					"Ms Victor Hugo",
					new ContactChannel[]{
						new ContactChannel(
							ContactChannel.EMAIL,
							"action@radical-cuboid.org"
						)
					}
				)
			},
			"FOO_BALLOON",
			MessageBegin.DELETE_ACTION,
			"200008091923", "2000", "602", "SIS_FOOB"
		),
		new TimeSeries(
			"HOX.S", "2001", "602", new Observation[] {}
		),
		new TimeSeries(
			"SOX.S", "", "", new Observation[] {}
		),
		new TimeSeries(
			".S", "", "", new Observation[] {}
		),
		new TimeSeries(
			"SOX.", "", "", new Observation[] {}
		),
		new Attribute(
			"", "", "", "SMUS", ""
		),
		new Attribute(
			"", "", "", "SCUS", ""
		),
		new Attribute(
			"NOX.S", "", "", "XCAS", ""
		),
		new Attribute(
			".N", "", "", "VANITAS", ""
		),
		new Attribute(
			"NOX.", "", "", "FLUX", ""
		),
		new Attribute(
			"SOX.C", "2020", "602", "COMMENT", ""
		),
		new MessageEnd(
			"MREF000003"
		),
		new InterchangeEnd(
			"IREF000215"
		)
	};

	/**
	 * The input to the {@link EDIFACTReader} under Test.
	 * This will be read into it by a {@link java.io.StringReader}.
	 * Its contents are meaningless but must correspond to
	 * {@link #testOutputTS}.
	 */
	private static final String testInputTS
		= "UNA:+.? '"
		+ "UNB+UNOC:3+4F0+ZZZ+001201:1410+IREF000216++GESMES/TS'"
		+ "UNH+MREF000001+GESMES:2:1:E6'"
		+ "BGM+73'"
		+ "NAD+Z02+ECB'"
		+ "NAD+MR+ZZZ'"
		+ "NAD+MS+4F0'"
		+ "IDE+10+Test Gesmes/TS metadata'"
		+ "ASI+SIS_FOTS'"
		+ "FTX+ACM+++Planet Ignoramus Servus?: Boofar ftuss'"
		+ "SCD+1+TIME_PERIOD++++:3'ATT+3+5+:::AN..16'"
		+ "SCD+4+ITEM++++:1'ATT+3+5+:::AN3'IDE+1+CL_ITEM'"
		+ "SCD+1+TIME_FORMAT++++:4'ATT+3+5+:::AN3'"
		+ "SCD+3+OBS_VALUE++++:5'ATT+3+5+:::AN..15'"
		+ "SCD+13+FREQUA++++:2'ATT+3+5+:::AN1'IDE+1+CL_FREQ'"
		+ "SCD+3+OBS_STATUS++++:6'ATT+3+5+:::AN1'ATT+3+35+2:USS'ATT+3+32+5:ALV'IDE+1+CL_OBS_STATUS'"
		+ "UNT+26+MREF000001'"
		+ "UNZ+1+IREF000216'";

	/**
	 * The objects into which
	 * the {@link EDIFACTReader} should translate
	 * {@link #testInputTS} when called in the correct way.
	 */
	private static final Object[] testOutputTS = {
		new InterchangeBegin(
			"IREF000216", "4F0", "ZZZ", "001201:1410", false
		),
		new MessageBegin(
			"MREF000001", "4F0", "ZZZ", "ECB", "73", "Test Gesmes/TS metadata",
			new Contact[]{}
		),
		new KeyFamily(
			"SIS_FOTS", "Planet Ignoramus Servus: Boofar ftuss", "",
			versionDate,
			new KeyFamilyComponent[] {
				new KeyFamilyComponent(
					"TIME_PERIOD", KeyFamilyComponent.TIME_DIMENSION,
					"3", "AN..16", "", "", "", true
				),
				new KeyFamilyComponent(
					"ITEM", KeyFamilyComponent.KEY_STRUCTURE_DIMENSION,
					"1", "AN3", "", "", "CL_ITEM", true
				),
				new KeyFamilyComponent(
					"TIME_FORMAT", KeyFamilyComponent.TIME_DIMENSION,
					"4", "AN3", "", "", "", true
				),
				new KeyFamilyComponent(
					"OBS_VALUE", KeyFamilyComponent.ARRAY_CELL,
					"5", "AN..15", "", "", "", true
				),
				new KeyFamilyComponent(
					"FREQUA", KeyFamilyComponent.FREQUENCY_DIMENSION,
					"2", "AN1", "", "", "CL_FREQ", true
				),
				new KeyFamilyComponent(
					"OBS_STATUS", KeyFamilyComponent.ARRAY_CELL,
					"6", "AN1", "2", "5", "CL_OBS_STATUS", true
				)
			}
		),
		new MessageEnd(
			"MREF000001"
		),
		new InterchangeEnd(
			"IREF000216"
		)
	};

	/**
	 * Erroneous input to the {@link EDIFACTReader} under Test.
	 * This will be read into it by a {@link java.io.StringReader}.
	 * Its contents are meaningless but must correspond to
	 * {@link #testScopeMismatchOutput}.
	 */
	private static final String testScopeMismatchInput
		= "UNA:+.? '\n"
		+ "UNB+UNOC:3+4F0+ZZZ+001201:1410+IREF000217++GESMES/TS'\n"
		+ "UNH+MREF000002+GESMES:2:1:E6'\n"
		+ "BGM+74'\n"
		+ "NAD+Z02+ECB'\n"
		+ "NAD+MR+BIS'\n"
		+ "NAD+MS+4F0'\n"
		+ "DSI+FOO_BAROON'\n"
		+ "STS+3+7'DTM+242:200008091923:203'\n"
		+ "IDE+5+SIS_FOOB'GIS+AR3'GIS+1:::-'\n"
		+ "FNS+Test Attributes:10'\n"
		+ "REL+Z01+1'\n"
		+ "ARR+2+NOX:C'IDE+Z11+TITLE'FTX+ACM+++Text:X:Y:Z'\n"
		+ "REL+Z01+4'\n"
		+ "ARR+4+NOX:C:2020:602'IDE+Z10+ORGAN'CDV+XANA'\n"
		+ "ARR+0'IDE+Z10+UNIT'CDV+SWEET'\n"
		+ "REL+Z01+5'\n"
		+ "ARR+0'IDE+Z11+COMMENT'FTX+ACM+++:TextAB::C'\n"
		+ "UNT+28+MREF000002'\n"
		+ "UNZ+1+IREF000217'\n";

	/**
	 * The objects into which
	 * the {@link EDIFACTReader} should translate
	 * {@link #testScopeMismatchInput} when called in the correct way.
	 */
	private static final Object[] testScopeMismatchOutput = {
		new InterchangeBegin(
			"IREF000217", "4F0", "ZZZ", "001201:1410", false
		),
		new MessageBegin(
			"MREF000002", "4F0", "BIS", "ECB",
			MessageBegin.STATISTICAL_DATA,
			"", new Contact[]{},
			"FOO_BAROON", MessageBegin.UPDATE_OR_REPLACE_ACTION,
			"200008091923", "", "", "SIS_FOOB"
		),
		new ErrorEvent(
			"Parsing problem:"
			+ " Data set scope attributes require an empty key."
			+ " In 1st message, Group 22, ARR-segment."
			+ " Character 264, line 13, pseudocolumn 6,"
			+ " next unparsed text"
			+ " \"+NOX:C'IDE+Z11+TITLE'FTX+ACM+++Text:X:Y:Z'"
			+ "REL+Z01+4'ARR+4+NOX:C:...\".",
			ErrorEvent.ERROR
		),
		new Attribute(
			"NOX.C", "", "", "TITLE", "TextXYZ"
		),
		new Attribute(
			"NOX.C.2020.602", "", "", "ORGAN", "XANA"
		),
		new ErrorEvent(
			"Parsing problem:"
			+ " Series or sibling group scope attributes"
			+ " require a nonempty key."
			+ " In 1st message, Group 24, ARR-segment."
			+ " Character 368, line 16, pseudocolumn 6,"
			+ " next unparsed text"
			+ " \"'IDE+Z10+UNIT'CDV+SWEET'"
			+ "REL+Z01+5'ARR+0'IDE+Z11+COMMENT'FTX+ACM+...\".",
			ErrorEvent.ERROR
		),
		new Attribute(
			"", "", "", "UNIT", "SWEET"
		),
		new ErrorEvent(
			"Parsing problem:"
			+ " Observation scope attributes require"
			+ " a key and period specification."
			+ " In 1st message, Group 22, ARR-segment."
			+ " Character 409, line 18, pseudocolumn 6,"
			+ " next unparsed text"
			+ " \"'IDE+Z11+COMMENT'FTX+ACM+++:TextAB::C'"
			+ "UNT+28+MREF000002'UNZ+1+IR...\".",
			ErrorEvent.ERROR
		),
		new ErrorEvent(
			"Parsing problem:"
			+ " The number -2 of dimensions in the key is negative."
			+ " In 1st message, Group 22, ARR-segment."
			+ " Character 409, line 18, pseudocolumn 6,"
			+ " next unparsed text"
			+ " \"'IDE+Z11+COMMENT'FTX+ACM+++:TextAB::C'"
			+ "UNT+28+MREF000002'UNZ+1+IR...\".",
			ErrorEvent.ERROR
		),
	};

	/**
	 * Erroneous input to the {@link EDIFACTReader} under Test.
	 * This will be read into it by a {@link java.io.StringReader}.
	 * Its contents are meaningless but must correspond to
	 * {@link #testCodedMismatchOutput}.
	 */
	private static final String testCodedMismatchInput
		= "UNA:+.? '"
		+ "UNB+UNOC:3+4F0+ZZZ+001201:1410+IREF000217++GESMES/TS'"
		+ "UNH+MREF000002+GESMES:2:1:E6'"
		+ "BGM+74'"
		+ "NAD+Z02+ECB'"
		+ "NAD+MR+BIS'"
		+ "NAD+MS+4F0'"
		+ "DSI+FOO_BAROON'"
		+ "STS+3+7'DTM+242:200008091923:203'"
		+ "IDE+5+SIS_FOOB'GIS+AR3'GIS+1:::-'"
		+ "FNS+Test Attributes:10'"
		+ "REL+Z01+1'"
		+ "ARR+0'IDE+Z10+TITLE'CDV+TextXYZ'"
		+ "REL+Z01+4'"
		+ "ARR+2+NOX:C'IDE+Z11+ORGAN'FTX+ACM+++XANA'"
		+ "UNT+21+MREF000002'"
		+ "UNZ+1+IREF000217'";

	/**
	 * The objects into which
	 * the {@link EDIFACTReader} should translate
	 * {@link #testCodedMismatchInput} when called in the correct way.
	 */
	private static final Object[] testCodedMismatchOutput = {
		new InterchangeBegin(
			"IREF000217", "4F0", "ZZZ", "001201:1410", false
		),
		new MessageBegin(
			"MREF000002", "4F0", "BIS", "ECB",
			MessageBegin.STATISTICAL_DATA,
			"", new Contact[]{},
			"FOO_BAROON", MessageBegin.UPDATE_OR_REPLACE_ACTION,
			"200008091923", "", "", "SIS_FOOB"
		),
		new ErrorEvent(
			"Parsing problem:"
			+ " Wrongly representing value for uncoded attribute \"TITLE\""
			+ " as if it were for a coded attribute."
			+ " In 1st message, Group 23, IDE1-segment."
			+ " Character 267, line 1, pseudocolumn 267,"
			+ " next unparsed text"
			+ " \"CDV+TextXYZ'"
			+ "REL+Z01+4'ARR+2+NOX:C'IDE+Z11+ORGAN'FTX+ACM+++XANA'U...\".",
			ErrorEvent.ERROR
		),
		new Attribute(
			"", "", "", "TITLE", "TextXYZ"
		),
		new ErrorEvent(
			"Parsing problem:"
			+ " Wrongly representing value for coded attribute \"ORGAN\""
			+ " as if it were for an uncoded attribute."
			+ " In 1st message, Group 23, IDE1-segment."
			+ " Character 315, line 1, pseudocolumn 315,"
			+ " next unparsed text"
			+ " \"FTX+ACM+++XANA'UNT+21+MREF000002'UNZ+1+IREF000217'\".",
			ErrorEvent.ERROR
		),
		new Attribute(
			"NOX.C", "", "", "ORGAN", "XANA"
		),
		new MessageEnd(
			"MREF000002"
		),
		new InterchangeEnd(
			"IREF000217"
		)
	};

	/**
	 * Erroneous input to the {@link EDIFACTReader} under Test.
	 * This will be read into it by a {@link java.io.StringReader}.
	 * Its contents are meaningless but must correspond to
	 * {@link #testTruncationViolationOutput}.
	 */
	private static final String testTruncationViolationInput
		= "UNA:+.? '"
		+ "UNB+UNOC:3+4F0+ZZZ+001201:1410+IREF000218++GESMES/TS'"
		+ "UNH+MREF000002+GESMES:2:1:E6'"
		+ "BGM+74'"
		+ "NAD+Z02+ECB'"
		+ "NAD+MR+BIS'"
		+ "NAD+MS+4F0'"
		+ "IDE+10+Badly truncated EDIFACT'"
		+ "DSI+FOO_BAROON'"
		+ "STS+3+7'DTM+242:200008091923:203'DTM+Z02:199822200543:716'"
		+ "IDE+5+SIS_FOOB'GIS+AR3'GIS+1:::-'"
		+ "ARR++HOX:S:2001:602:13.6:S:C:'"
		+ "ARR++SOX:S:20102020:702:13.6:S:N+13.5:A+1:A:C:13+'"
		+ "ARR++SOX:N:20102020:702:13.6:S:N+++++++13.5:A+1:A:C:13+:'"
		+ "UNT+17+MREF000002'"
		+ "UNH+MREF000003+GESMES:2:1:E6'"
		+ "BGM+74'"
		+ "NAD+Z02+ECB'"
		+ "NAD+MR+BIS'"
		+ "NAD+MS+4F0'"
		+ "IDE+10+Invalidly truncated delete'"
		+ "CTA+CC+Corporate Confidentiality:Lady Spiff O?'Henderson'"
		+ "CTA+CP+Corporate Privatisation:Ms Julija Chobgornika'"
		+ "CTA+CE+Criminal Event Management:Sir Arthur Hammerclang'"
		+ "DSI+FOO_BALLOON'"
		+ "STS+3+6'DTM+242:200008091923:203'"
		+ "IDE+5+SIS_FOOB'GIS+AR3'GIS+1:::-'"
		+ "ARR++HOX:S:2001:602::'"
		+ "ARR++SOX:R::::'"
		+ "ARR++SOX:S::::+'"
		+ "ARR++SOX:T::::+:'"
		+ "UNT+20+MREF000003'"
		+ "UNZ+2+IREF000218'";

	/**
	 * The objects into which
	 * the {@link EDIFACTReader} should translate
	 * {@link #testTruncationViolationInput} when called in the correct way.
	 */
	private static final Object[] testTruncationViolationOutput = {
		new InterchangeBegin(
			"IREF000218", "4F0", "ZZZ", "001201:1410", false
		),
		new MessageBegin(
			"MREF000002", "4F0", "BIS", "ECB",
			MessageBegin.STATISTICAL_DATA,
			"Badly truncated EDIFACT", new Contact[]{},
			"FOO_BAROON", MessageBegin.UPDATE_OR_REPLACE_ACTION,
			"200008091923", "199822200543", "716",
			"SIS_FOOB"
		),
		new ErrorEvent(
			"Parsing problem:"
			+ " Too many colons or violation of EDIFACT truncation rules"
			+ " where the observation pre-break value is expected."
			+ " In 1st message, Group 17, ARR-segment."
			+ " Character 299, line 1, pseudocolumn 299,"
			+ " next unparsed text"
			+ " \"'ARR++SOX:S:20102020:702:13.6:S:N+13.5:A+1:A:C:13+'"
			+ "ARR++SOX:N:20...\".",
			ErrorEvent.ERROR
		)
	};

	/* ========================= Fields ========================= */

	/**
	 * The file used for the location of test data.
	 */
	private File testFile;

	/* ====================== Constructors ====================== */

	public EDIFACTReaderTest( String name ) {
		super( name );
	}

	/* ================== Protected Methods ===================== */

	/**
	 * Sets up the test fixture.
	 */
	protected void setUp () {
		final TestEventSource reader = new TestEventSource();
		final Metabase metabase = Metabase.getMetabase();
		reader.addListener( metabase, true );
		reader.dispatch( SIS_FOOB );
		try {
			testFile = File.createTempFile( "edifact_testinput.", null );
			testFile.deleteOnExit();

			Reader input = new StringReader( testInputCB );
			Writer data = new BufferedWriter(
				new OutputStreamWriter(
					new FileOutputStream( testFile ),
					"ISO8859-1"
				)
			);
			int c, i = 0;
			while ( ( c = input.read() ) != -1 ) {
				data.write( c );
				if ( ++i % 72 == 0 ) data.write( '\n' );
			}
			data.close();
			input.close();
		}
		catch ( IOException ioe ) {
			fail(
				"Write testFile failed: " + ioe
			);
		}
	}

	/**
	 * Tears down the test fixture.
	 */
	protected void tearDown () {
		testFile.delete();
	}

	/* ===================== Public Methods ===================== */

	/**
	 * Assembles and returns a test suite for all
	 * the test methods of this test case.
	 *
	 * @return a non-null test suite
	 */
	public static Test suite () {
		TestSuite suite = new TestSuite( EDIFACTReaderTest.class );
		return suite;
	}

	/**
	 * @param arg the array of argument {@link String}s
	 */
	public static void main ( String[] arg ) {
		String[] testCaseName = {
			EDIFACTReaderTest.class.getName()
		};
		junit.textui.TestRunner.main( testCaseName );
	}

	/* ---------------------- Test Methods ---------------------- */

	/**
	 * Test {@link EDIFACTReader#run}.
	 *
	 * @throws IOException if the StringReader has a problem.
	 */
	public void testRunFromString ()
		throws IOException
	{
		Reader input = new StringReader( testInputCB );
		EDIFACTReader reader
		= new EDIFACTReader( "Zampano", input );

		// Run the test.
		doTestRun( reader, testOutputCB );

		// Close and check.
		assertTrue( "All input read", -1 == input.read() );
		reader.close();

		// The input is not considered owned by the
		// EDIFACTReader.
		try {
			input.read();
		}
		catch ( IOException ioe ) {
			fail(
				"The input stream was closed but should not be!"
			);
		}
		input.close();
	}

	/**
	 * Test {@link EDIFACTReader#run}.
	 *
	 * @throws IOException if the StringReader has a problem.
	 */
	public void testRunFromFile ()
		throws IOException
	{
		EDIFACTReader reader
		= new EDIFACTReader( "Sargngele", testFile.getPath() );

		// Run the test.
		doTestRun( reader, testOutputCB );

		// Close the reader.
		reader.close();
	}
	
	/**
	 * Test {@link EDIFACTReader#run}.
	 *
	 * @throws IOException if the StringReader has a problem.
	 */
	public void testRunFromStream ()
		throws IOException
	{
		InputStreamReader inputStreamReader = new InputStreamReader(
				new FileInputStream( testFile.getPath() ),
				ENCODING
			);
		
		EDIFACTReader reader
		= new EDIFACTReader( "Sargngele", inputStreamReader );
		

		// Run the test.
		doTestRun( reader, testOutputCB );

		// Close the reader.
		reader.close();
	}

	/**
	 * Test {@link EDIFACTReader#run}.
	 *
	 * @throws IOException if the StringReader has a problem.
	 */
	public void testGesmesTS ()
		throws IOException
	{
		final Reader input = new StringReader( testInputTS );
		final EDIFACTReader reader = new EDIFACTReader( "ZampaTS", input );

		// Run the test.
		doTestRun( reader, testOutputTS );

		// Close and check.
		assertTrue( "All input read", -1 == input.read() );
		reader.close();

		// The input is not considered owned by the
		// EDIFACTReader.
		try {
			input.read();
		}
		catch ( final IOException ioe ) {
			fail(
				"The input stream was closed but should not be!"
			);
		}
		input.close();
	}

	/**
	 * Test {@link EDIFACTReader#run}.
	 *
	 * @throws IOException if the StringReader has a problem.
	 */
	public void testSdmxEdi ()
		throws IOException
	{
		final int formatPosition = testInputTS.indexOf( Constants.GESMES_TS );
		final Reader input = new StringReader(
			testInputTS.substring( 0, formatPosition )
			+ Constants.SDMX_EDI
			+ testInputTS.substring(
				formatPosition + Constants.GESMES_TS.length()
			)
		);
		final EDIFACTReader reader = new EDIFACTReader( "ZampaTS", input );

		// Run the test.
		doTestRun( reader, testOutputTS );

		// Close and check.
		assertTrue( "All input read", -1 == input.read() );
		reader.close();

		// The input is not considered owned by the
		// EDIFACTReader.
		try {
			input.read();
		}
		catch ( final IOException ioe ) {
			fail(
				"The input stream was closed but should not be!"
			);
		}
		input.close();
	}

	/**
	 * Test warning and if whitespace remains after
	 * {@link EDIFACTReader#run}.
	 *
	 * @throws IOException if the StringReader has a problem.
	 */
	public void testRemainingSpaces ()
		throws IOException
	{
		final String spacyInput = testInputCB + " ";
		final Object[] spacyOutput = new Object[testOutputCB.length + 1];
		System.arraycopy(
			testOutputCB, 0, spacyOutput, 0, testOutputCB.length
		);
		spacyOutput[testOutputCB.length] = spacyOutput[testOutputCB.length - 1];
		spacyOutput[testOutputCB.length - 1] = new ErrorEvent(
			"Superfluous space character(s) after segment."
			+ " In 3rd message, Group 23, UNZ-segment."
			+ " Character 3037, line 1, pseudocolumn 3037,"
			+ " next unparsed text \"\"."
			+ " (This warning is issued only once.)",
			ErrorEvent.WARNING
		);
		final Reader input = new StringReader( spacyInput );
		final EDIFACTReader reader = new EDIFACTReader(
			"ZampaSpace", input
		);
		doTestRun( reader, spacyOutput );
		assertTrue( "All input read", -1 == input.read() );
		reader.close();
		input.close();
	}

	/**
	 * Test error and if non-whitespace text remains after
	 * {@link EDIFACTReader#run}.
	 *
	 * @throws IOException if the StringReader has a problem.
	 */
	public void testRemainingText ()
		throws IOException
	{
		final String textyInput = testInputCB + "TEX";
		final Object[] textyOutput = new Object[testOutputCB.length + 1];
		System.arraycopy(
			testOutputCB, 0, textyOutput, 0, testOutputCB.length
		);
		textyOutput[testOutputCB.length] = new ErrorEvent(
			"Parsing problem: Unparsed input remaining."
			+ " In 3rd message, Group 23, UNB-segment."
			+ " Character 3036, line 1, pseudocolumn 3036,"
			+ " next unparsed text \"TEX\".",
			ErrorEvent.ERROR
		);
		{
			final Reader input = new StringReader( textyInput );
			final EDIFACTReader reader = new EDIFACTReader(
				"ZampaText", input
			);
			doTestRun( reader, textyOutput );
			assertTrue( "All input read", -1 == input.read() );
			reader.close();
			input.close();
		}
	}

	/**
	 * Tests {@link EDIFACTReader#run} to
	 * refuse multiple interchanges per file.
	 *
	 * @throws IOException if the StringReader has a problem.
	 */
	public void testMultipleInterchanges ()
		throws IOException
	{
		final Reader input = new StringReader(
			testInputCB
			+ testInputCB.substring( testInputCB.indexOf( "UNB" ) )
		);
		final EDIFACTReader reader
		= new EDIFACTReader( "Zampano", input );
		final Object[] twiceOutputCB = new Object[
			testOutputCB.length + testOutputCB.length + 1
		];
		System.arraycopy(
			testOutputCB, 0,
			twiceOutputCB, 0, testOutputCB.length
		);
		System.arraycopy(
			testOutputCB, 0,
			twiceOutputCB, testOutputCB.length, testOutputCB.length
		);
		twiceOutputCB[2*testOutputCB.length] = new ErrorEvent(
			"Parsing problem: More than one interchange in the file."
			+ " In 3rd message, Group 23, UNB-segment."
			+ " Character 6062, line 1, pseudocolumn 6062,"
			+ " next unparsed text \"\".",
			ErrorEvent.ERROR
		);

		// Run the test.
		doTestRun( reader, twiceOutputCB );

		// Close and check.
		assertTrue( "All input read", -1 == input.read() );
		reader.close();

		// The input is not considered owned by the
		// EDIFACTReader.
		try {
			input.read();
		}
		catch ( final IOException ioe ) {
			fail(
				"The input stream was closed but should not be!"
			);
		}
		input.close();
	}

	/**
	 * Test error if attachment scope does not match attachment level.
	 *
	 * @throws IOException if the StringReader has a problem.
	 */
	public void testScopeMismatch ()
		throws IOException
	{
		final Reader input = new StringReader( testScopeMismatchInput );
		final EDIFACTReader reader = new EDIFACTReader( "ZampaTS", input );

		// Run the test.
		doTestRun( reader, testScopeMismatchOutput );

		// Close and check.
		assertTrue( "All input read", -1 == input.read() );
		reader.close();

		// The input is not considered owned by the
		// EDIFACTReader.
		try {
			input.read();
		}
		catch ( final IOException ioe ) {
			fail(
				"The input stream was closed but should not be!"
			);
		}
		input.close();
	}

	/**
	 * Test error if attribute codedness does not match its definition.
	 *
	 * @throws IOException if the StringReader has a problem.
	 */
	public void testCodedMismatch ()
		throws IOException
	{
		final Reader input = new StringReader( testCodedMismatchInput );
		final EDIFACTReader reader = new EDIFACTReader( "ZampaTS", input );

		// Run the test.
		doTestRun( reader, testCodedMismatchOutput );

		// Close and check.
		assertTrue( "All input read", -1 == input.read() );
		reader.close();

		// The input is not considered owned by the
		// EDIFACTReader.
		try {
			input.read();
		}
		catch ( final IOException ioe ) {
			fail(
				"The input stream was closed but should not be!"
			);
		}
		input.close();
	}

	/**
	 * Test error if truncation rules are violated.
	 *
	 * @throws IOException if the StringReader has a problem.
	 */
	public void testTruncationViolation ()
		throws IOException
	{
		final Reader input = new StringReader( testTruncationViolationInput );
		final EDIFACTReader reader = new EDIFACTReader( "ZampaTS", input );

		// Run the test.
		doTestRun( reader, testTruncationViolationOutput );

		// Close and check.
		assertTrue( "All input read", -1 == input.read() );
		reader.close();

		// The input is not considered owned by the
		// EDIFACTReader.
		try {
			input.read();
		}
		catch ( final IOException ioe ) {
			fail(
				"The input stream was closed but should not be!"
			);
		}
		input.close();
	}

	/* ==================== Private Methods ===================== */

	protected static void printHex ( final String string ) {
		/* protected instead of private to make eclipse moan less.*/
		final int length = string.length();
		for ( int i = 0; i < length; ) {
			System.out.print( Integer.toHexString( string.charAt( i++ ) ) );
			System.out.print( i < length ? ' ' : '\n' );
		}
	}

	/**
	 * Test {@link EDIFACTReader#run}.
	 *
	 * @throws IOException if the StringReader has a problem.
	 */
	private void doTestRun (
		final EDIFACTReader reader,
		final Object[] testOutput
	)
		throws IOException
	{
		TestEventListener listener
		= new TestEventListener();
		Metabase metabase = Metabase.getMetabase();
		reader.addListener( metabase, false );
		reader.addListener( listener, true );

		// Do the run.
		reader.run();

		// System.out.println( metabase );

		// Compare the event contents.
		{
			final Iterator iterator = listener.iterator();
			int i = 0;
			while ( iterator.hasNext() && i < testOutput.length ) {
				final Object expected = testOutput[i++];
				final Object received = iterator.next();
				if ( !expected.equals( received ) ) {
					System.out.println( "expected: " + expected );
					System.out.println( "received: " + received );
				}
				assertEquals(
					"Output",
					expected,
					received
				);
			}
			assertTrue(
				"Not too many events",
				!iterator.hasNext()
			);
			assertEquals(
				"Not too few events",
				testOutput.length, i
			);
		}
	}
}
