// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, 2003 European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.edifact;

import java.io.FileOutputStream;
import java.io.Writer;
import java.io.OutputStreamWriter;
import java.io.IOException;

import java.util.Iterator;

import Int.ecb.logging.Level;

import Int.ecb.gesmescb.*;

import Int.ecb.gesmescb.metabase.Metabase;
import Int.ecb.gesmescb.metabase.MaintenanceAgency;

import Int.ecb.gesmescb.event.StructureListener;
import Int.ecb.gesmescb.event.DataListener;
import Int.ecb.gesmescb.event.GESMESCBWriter;
import Int.ecb.gesmescb.event.GESMESCBEventListenerAdapter;

import Int.ecb.gesmescb.util.Named;
import Int.ecb.gesmescb.util.ComponentHelper;

/* ***************************** Class ****************************** */

/**
 * <P>Translates {@link Int.ecb.gesmescb.GESMESCBObject}s
 * into EDIFACT text.
 * This text then is written to an output {@link java.io.Writer}.</P>
 *
 * <CENTER><TABLE BORDER=3 CELLPADDING=3>
 * <CAPTION>
 * <A NAME="prefs">Available {@link Int.ecb.prefs preference} settings</A>
 * </CAPTION>
 * <TR>
 * <TH>preference name</TH>
 * <TH>preference type</TH>
 * <TH>preference value</TH>
 * </TR>
 * <TR><TD>destination</TD><TD>string</TD><TD>output file path</TD></TR>
 * <tr><td>format</td><td>string</td>
 * <td>EDIFACT interchange format</td><td><code>GESMES/TS</code></td></tr>
 * <TR><TD>maxColumn</TD><TD>integer</TD>
 * <TD>break lines after column #, 0 for none (default)</TD></TR>
 * </TABLE></CENTER>
 *
 * <P>The <CODE>EDIFACTUnlimitedWriter</CODE> is not internally synchronized.
 * External synchronisation is necessary if multiple threads want to
 * use the same <CODE>EDIFACTUnlimitedWriter</CODE>.</P>
 *
 * <P>The <CODE>EDIFACTUnlimitedWriter</CODE> relies upon that all GESMES/CB events
 * to be written do not contain <CODE>null</CODE> references
 * or Strings which contain
 * control characters. If the first assumption fails, a NullPointerException
 * may ensue. If the latter assumption fails, truncation rules may not be
 * correctly observed.</P>
 *
 * @author
 * <A HREF="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</A>
 *
 * @version @(#)$Id: EDIFACTUnlimitedWriter.java,v 1.1 2017/01/19 16:14:05 trainot Exp $
 */
class EDIFACTUnlimitedWriter extends GESMESCBEventListenerAdapter
	implements Named, StructureListener, DataListener, GESMESCBWriter, Constants
{
	/* ========================= Fields ========================= */

	/**
	 * Takes care of logging and preferences.
	 */
	protected final ComponentHelper componentHelper;

	/**
	 * The filename of the file to write to
	 * if the <CODE>EDIFACTUnlimitedWriter</CODE> has been
	 * constructed from a filename.
	 * Otherwise it is null.
	 */
	private final String filename;

	/**
	 * The underlying Writer where the <CODE>EDIFACTUnlimitedWriter</CODE>
	 * writes its output to.
	 */
	private final Writer writer;

	/**
	 * The <CODE>EDIFACTWriteBuffer</CODE> which helps to write segments
	 * with properly escaped characters.
	 */
	private final EDIFACTWriteBuffer buffer;

	/** Format specifier for UNB segment. */
	private final String interchangeFormat;

	/**
	 * In order to write a nice UNZ segment at interchange end.
	 * Count all MessageBegin events.
	 */
	private int messageCounter;

	/**
	 * In order to write a nice UNT segment at message end
	 * Count all Segments within a message
	 */
	private int segmentCounter;

	/**
	 * Whether we presently are in an update/replace message.
	 */
	private boolean inURMessage;

	/**
	 * The current key family.
	 * @see #handle( MessageBegin )
	 * @see #handle( Attribute )
	 */
	private KeyFamily currentKeyFamily;

	/**
	 * A flag is necessary to indicate that no FNS segment
	 * has been written within the current dataset yet.
	 * This is necessary to get only one FNS segment
	 * starting the attribute section.
	 */
	boolean haveNotWrittenFNS;

	/**
	 * A field to store the attachment scope of the previous
	 * attribute's attachment scope.
	 * It is initialized with some impossible
	 * value for an attachment scope.
	 * This is used to write REL segments only if really needed.
	 */
	String previousAttachmentScope;

	/**
	 * A field to store the key of the previous attribute.
	 * It is initialized with some impossible
	 * value for a key.
	 * This is used to write ARR segments only if really needed.
	 */
	String previousKey;

	/**
	 * Together with {@link previousKey} used for ARR segment saving.
	 */
	String previousPeriod, previousPeriodFormat;

	/* ====================== Constructors ====================== */

	/**
	 * Construct an <CODE>EDIFACTUnlimitedWriter</CODE>
	 * with a given writing destination.
	 * Since an internal buffering scheme is used,
	 * no additional external buffering
	 * (e.g. with a {@link java.io.BufferedWriter}) is necessary.
	 *
	 * @param name used for logging and preferences.
	 * @param writer where to write to.
	 */
	public EDIFACTUnlimitedWriter (
		final String name, final Writer writer
	) throws IOException {
		this( name, (Object) writer );
	}

	/**
	 * Construct an <CODE>EDIFACTUnlimitedWriter</CODE>
	 * with given writing destination file.
	 *
	 * @param name used for logging and preferences.
	 * @param filename specifies in the local system's manner
	 * where to write to.
	 *
	 * @throws IOException if the file cannot be opened for writing.
	 */
	public EDIFACTUnlimitedWriter ( final String name, final String filename )
		throws IOException
	{
		this( name, (Object) filename );
	}

	/**
	 * Construct an <CODE>EDIFACTUnlimitedWriter</CODE>
	 * with writing destination file according to preferences.
	 *
	 * @param name used for logging and preferences.
	 *
	 * @throws IOException if the file cannot be opened for writing.
	 * @throws NullPointerException if the preference variable "destination"
	 * is undefined.
	 */
	public EDIFACTUnlimitedWriter ( final String name )
		throws IOException
	{
		this( name, (Object) null );
	}

	/**
	 * Construct an <CODE>EDIFACTUnlimitedWriter</CODE>
	 * with a given writing destination in the form of an
	 * <code>Object</code>.
	 *
	 * @param name used for logging and preferences.
	 * @param destination where to write to.
	 */
	private EDIFACTUnlimitedWriter (
		final String name, final Object destination
	) throws IOException {
		componentHelper = new ComponentHelper( name, this );
		if ( destination == null || destination instanceof String ) {
			filename = destination != null
				? (String) destination
				: componentHelper.getPreference( "destination", null );
			if ( filename == null ) {
				final String desc
				= "No preference variable \"destination\" found"
					+ " to determine the output file.";
				componentHelper.logger.log(
					Level.SEVERE,
					componentHelper.logName,
					"EDIFACTUnlimitedWriter()",
					desc
				);
				throw new NullPointerException( desc );
			}
			writer = new OutputStreamWriter(
				new FileOutputStream( this.filename ),
				ENCODING
			);
		} else {
			filename = null;
			writer = (Writer) destination;
		}
		buffer = new EDIFACTWriteBuffer ( writer, componentHelper );
		interchangeFormat = componentHelper.getPreference(
			"format", GESMES_TS
		);
		if (
			!GESMES_CB.equals( interchangeFormat )
			&&
			!GESMES_TS.equals( interchangeFormat )
		) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"EDIFACTUnlimitedWriter()",
				"Unknown interchange format \""
				+ interchangeFormat
				+ "\" in configuration file."
			);
		}
		messageCounter = segmentCounter = 0;
		currentKeyFamily = null;
		inURMessage = true;
		haveNotWrittenFNS = true;
		previousAttachmentScope = "<first>";
	}

	/* ===================== Public Methods ===================== */

	public String getName () {
		return componentHelper.name;
	}

	/**
	 * <P>Close the <CODE>EDIFACTUnlimitedWriter</CODE>.</P>
	 *
	 * <P>This method should always be called
	 * when the <CODE>EDIFACTUnlimitedWriter</CODE>
	 * has served its purpose.</P>
	 *
	 * <P>The internal buffer is flushed
	 * and if the <CODE>EDIFACTUnlimitedWriter</CODE>
	 * has been constructed
	 * from a filename
	 * using {@link #EDIFACTUnlimitedWriter(String,String)},
	 * the output file is closed.</P>
	 *
	 * <P>The {@link Writer} is not closed
	 * if the <CODE>EDIFACTUnlimitedWriter</CODE> has been
	 * created from a <CODE>Writer</CODE>
	 * using {@link #EDIFACTUnlimitedWriter(String,Writer)}
	 * because then the <CODE>Writer</CODE> is considered
	 * not to be property of the <CODE>EDIFACTUnlimitedWriter</CODE>.</P>
	 */
	public void close () {
		// Remove the sources.
		setActive( false );

		try {
			// Flush the buffer.
			buffer.close();

			if ( filename != null && writer != null ) {
				writer.close();
				componentHelper.logger.log(
					Level.INFO,
					componentHelper.logName,
					"close()",
					"Closed " + interchangeFormat
					+ "-EDIFACT output"
					+ (
						filename != null
						? " to " + filename
						: ""
						)
					+ '.'
				);
			}
		}
		catch ( final IOException ioe ) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"close()",
				"Output problem.",
				ioe
			);
		}
	}

	/* ---------------------- Interchange ----------------------- */

	public void handle ( final InterchangeBegin interchangeBegin ) {
		try {
			componentHelper.logger.log(
				Level.INFO,
				componentHelper.logName,
				"handle( InterchangeBegin )",
				"Begin writing interchange "
				+ interchangeBegin.getIref() + "."
			);

			// Initialize message counting
			messageCounter = 0;

			// UNA
			buffer.writeUntruncated( "UNA:+.? '" );
			buffer.newLine();
			++segmentCounter;

			// UNB
			buffer.writeRaw( "UNB+UNOC:3+" );
			buffer.writeEscaped( interchangeBegin.getSender() );
			buffer.writeRaw( '+' );
			buffer.writeEscaped( interchangeBegin.getReceiver() );
			// Do not to use writeEscaped,
			// because date and time come together.
			buffer.writeRaw( '+' );
			buffer.writeRaw( interchangeBegin.getPreparationDateTime() );
			buffer.writeRaw( '+' );
			buffer.writeEscaped( interchangeBegin.getIref() );
			buffer.writeRaw( "++" );
			buffer.writeEscaped( interchangeFormat );
			if ( interchangeBegin.isTest() ) {
				buffer.writeRaw( "++++1" );
			}
			buffer.writeRaw( '\'' );
			buffer.newLine();
			++segmentCounter;
		}
		catch ( final IOException ioe ) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"handle( InterchangeBegin )",
				"Output problem.",
				ioe
			);
			close();
		}
	}

	public void handle ( final InterchangeEnd interchangeEnd ) {
		try {
			// UNZ
			buffer.writeRaw( "UNZ+" + messageCounter );
			buffer.writeRaw( '+' );
			buffer.writeEscaped( interchangeEnd.getIref() );
			buffer.writeRaw( '\'' );
			buffer.newLine();
			++segmentCounter;

			componentHelper.logger.log(
				Level.INFO,
				componentHelper.logName,
				"handle( InterchangeEnd )",
				"Ended writing interchange "
				+ interchangeEnd.getIref() + "."
			);
		}
		catch ( final IOException ioe ) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"handle( InterchangeEnd )",
				"Output problem.",
				ioe
			);
			close();
		}
	}

	/* ------------------------ Message ------------------------- */

	public void handle ( final MessageBegin messageBegin ) {
		try {
			componentHelper.logger.log(
				Level.FINE,
				componentHelper.logName,
				"handle( MessageBegin )",
				"Begin writing "
				+ ( messageBegin.isStatisticalData()
					? "data-" : "" )
				+ ( messageBegin.isStatisticalDefinitions()
					? "structure-" : "" )
				+ "message " + messageBegin.getMref() + "."
			);

			// Count the message
			++messageCounter;

			// Initialize segment counting
			segmentCounter = 0;

			// Set the flag to indicate
			// that no FNS segment has been
			// written for this message yet.
			// Only relevant for messages
			// which contain attributes.
			haveNotWrittenFNS = true;

			// UNH
			buffer.writeRaw( "UNH+" );
			buffer.writeEscaped( messageBegin.getMref() );
			buffer.writeRaw( "+GESMES:2:1:E6'" );
			buffer.newLine();
			++segmentCounter;

			// BGM
			buffer.writeRaw( "BGM" );
			buffer.writeRaw( '+' );
			buffer.writeEscaped( messageBegin.getFunction() );
			buffer.writeRaw( '\'' );
			buffer.newLine();
			++segmentCounter;

			// NAD 1
			buffer.writeRaw( "NAD+Z02" );
			buffer.writeRaw( '+' );
			buffer.writeEscaped( messageBegin.getMaintenanceAgencyId() );
			buffer.writeRaw( '\'' );
			buffer.newLine();
			++segmentCounter;

			// NAD 2
			buffer.writeRaw( "NAD+MR" );
			buffer.writeRaw( '+' );
			buffer.writeEscaped( messageBegin.getReceiver() );
			buffer.writeRaw( '\'' );
			buffer.newLine();
			++segmentCounter;

			// NAD 3
			buffer.writeRaw( "NAD+MS" );
			buffer.writeRaw( '+' );
			buffer.writeEscaped( messageBegin.getSender() );
			buffer.writeRaw( '\'' );
			buffer.newLine();
			++segmentCounter;

			// IDE+10
			if ( ! "".equals( messageBegin.getIdentity() ) ) {
				buffer.writeRaw( "IDE+10+" );
				buffer.writeEscaped( messageBegin.getIdentity() );
				buffer.writeRaw( '\'' );
				buffer.newLine();
				++segmentCounter;
			}

			// CTA
			for (
				final Iterator iterator = messageBegin.contactIterator();
				iterator.hasNext();
			) {
				final Contact contact = (Contact) iterator.next();

				buffer.writeRaw( "CTA" );
				buffer.writeRaw( '+' );
				buffer.writeEscaped( contact.getFunction() );
				buffer.writeRaw( '+' );
				buffer.writeEscaped( contact.getIdentity() );
				buffer.writeRaw( ':' );
				buffer.writeEscaped( contact.getName() );
				buffer.writeRaw( '\'' );
				buffer.newLine();
				++segmentCounter;

				for (
					final Iterator channels = contact.channelIterator();
					channels.hasNext();
				) {
					final ContactChannel channel
					= (ContactChannel) channels.next();

					buffer.writeRaw( "COM" );
					buffer.writeRaw( '+' );
					buffer.writeEscaped( channel.getNumber() );
					buffer.writeRaw( ':' );
					if ( channel.isEmailAddress() ) {
						buffer.writeRaw( "EM" );
					} else if ( channel.isTelephoneNumber() ) {
						buffer.writeRaw( "TE" );
					} else if ( channel.isFaxNumber() ) {
						buffer.writeRaw( "FX" );
					} else if ( channel.isX400Address() ) {
						buffer.writeRaw( "XF" );
					} else {
						componentHelper.logger.log(
							Level.SEVERE,
							componentHelper.logName,
							"handle( MessageBegin )",
							"Unknown contact channel type \""
							+ channel.getType() + "\"."
						);
						buffer.writeEscaped( channel.getType() );
					}

					buffer.writeRaw( '\'' );
					buffer.newLine();
					++segmentCounter;
				}
			}

			// Data set info?
			if ( !messageBegin.isStatisticalDefinitions() ) {

				// DSI
				buffer.writeRaw( "DSI+" );
				buffer.writeEscaped( messageBegin.getDataSetId() );
				buffer.writeRaw( '\'' );
				buffer.newLine();
				++segmentCounter;

				// STS
				buffer.writeRaw( "STS+3+" );
				buffer.writeEscaped( messageBegin.getAction() );
				buffer.writeRaw( '\'' );
				buffer.newLine();
				++segmentCounter;

				// DTM 1
				buffer.writeRaw( "DTM+242:" );
				buffer.writeEscaped( messageBegin.getExtractionDateTime() );
				buffer.writeRaw( ":203'" );
				buffer.newLine();
				++segmentCounter;

				// DTM 2
				if (
					! "".equals( messageBegin.getReportingPeriod() )
					||
					! "".equals( messageBegin.getReportingPeriodFormat() )
				) {
					buffer.writeRaw( "DTM+Z02:" );
					buffer.writeEscaped( messageBegin.getReportingPeriod() );
					buffer.writeRaw( ':' );
					buffer.writeEscaped( messageBegin.getReportingPeriodFormat() );
					buffer.writeRaw( '\'' );
					buffer.newLine();
					++segmentCounter;
				}

				// IDE
				buffer.writeRaw( "IDE+5+" );
				buffer.writeEscaped( messageBegin.getKeyFamilyId() );
				buffer.writeRaw( '\'' );
				buffer.newLine();
				++segmentCounter;

				// GIS 1
				buffer.writeRaw( "GIS+AR3'" );
				buffer.newLine();
				++segmentCounter;

				// GIS 2
				buffer.writeRaw( "GIS+1:::-'" );
				buffer.newLine();
				++segmentCounter;

				// prepare key length
				final String maintenanceAgencyId
				= messageBegin.getMaintenanceAgencyId();
				final Metabase metabase = Metabase.getMetabase();
				final MaintenanceAgency maintenanceAgency
				= metabase.getMaintenanceAgency( maintenanceAgencyId );
				if ( maintenanceAgency != null ) {
					final String keyFamilyId = messageBegin.getKeyFamilyId();
					currentKeyFamily = maintenanceAgency.getKeyFamily(
						keyFamilyId
					);
					if ( currentKeyFamily == null ) {
						componentHelper.logger.log(
							Level.SEVERE,
							componentHelper.logName,
							"handle( MessageBegin )",
							"Maintenance agengy \"" + maintenanceAgencyId
							+ "\" does not know key family \""
							+ keyFamilyId + "\"."
						);
						close();
					}
				} else {
					componentHelper.logger.log(
						Level.SEVERE,
						componentHelper.logName,
						"handle( MessageBegin )",
						"Unknown maintenance agengy \"" + maintenanceAgencyId
						+ "\"."
					);
					close();
				}
			}

			inURMessage = messageBegin.isUpdateOrReplaceAction();
		}
		catch ( final IOException ioe ) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"handle( MessageBegin )",
				"Output problem.",
				ioe
			);
			close();
		}
	}

	public void handle ( final MessageEnd messageEnd ) {
		try {
			// UNT
			// Since segmentCounting inludes UNT,
			// but the segment is not yet complete, +1 is necessary
			buffer.writeRaw( "UNT+" );
			buffer.writeRaw( segmentCounter + 1 );
			buffer.writeRaw( '+' );
			buffer.writeEscaped( messageEnd.getMref() );
			buffer.writeRaw( '\'' );
			buffer.newLine();
			++segmentCounter;

			componentHelper.logger.log(
				Level.FINE,
				componentHelper.logName,
				"handle( MessageEnd )",
				"End writing message "
				+ messageEnd.getMref() + "."
			);
		}
		catch ( final IOException ioe ) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"handle( MessageEnd )",
				"Output problem.",
				ioe
			);
			close();
		}
	}

	/* ----------------------- Code List ------------------------ */

	public void handle ( final CodeList codeList ) {
		try {
			// VLI
			buffer.writeRaw( "VLI+" );
			buffer.writeEscaped( codeList.getId() );
			buffer.writeRaw( "+++" );
			buffer.writeEscaped( codeList.getName() );
			buffer.writeRaw( '\'' );
			buffer.newLine();
			++segmentCounter;

			for (
				final Iterator
					values = codeList.codeValueIterator(),
					descriptions = codeList.codeDescriptionIterator();
				values.hasNext() && descriptions.hasNext();
			) {
				final String
					value = (String)values.next(),
					description = (String)descriptions.next();

				// CDV
				buffer.writeRaw( "CDV" );
				buffer.writeRaw( '+' );
				buffer.writeEscaped( value );
				buffer.writeRaw( '\'' );
				++segmentCounter;

				// FTX
				writeFTXSegments( description );
			}
		}
		catch ( final IOException ioe ) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"handle( CodeList )",
				"Output problem.",
				ioe
			);
			close();
		}
	}

	/* ------------------ Statistical Concept ------------------- */

	public void handle ( final StatisticalConcept statisticalConcept ) {
		try {
			// STC
			buffer.writeRaw( "STC+" );
			buffer.writeEscaped( statisticalConcept.getId() );
			buffer.writeRaw( '\'' );
			++segmentCounter;

			// FTX
			writeFTXSegments( statisticalConcept.getName() );
		}
		catch ( final IOException ioe ) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"handle( StatisticalConcept )",
				"Output problem.",
				ioe
			);
			close();
		}
	}

	/* ----------------------- Key Family ----------------------- */

	public void handle ( final KeyFamily keyFamily ) {
		try {
			// ASI
			buffer.writeRaw( "ASI+" );
			buffer.writeEscaped( keyFamily.getId() );
			buffer.writeRaw( '\'' );
			++segmentCounter;

			// FTX
			writeFTXSegments( keyFamily.getName() );

			// Group 11: SCD, ATT
			for (
				final Iterator iterator
				= keyFamily.componentIterator();
				iterator.hasNext();
			) {
				final KeyFamilyComponent component
				= (KeyFamilyComponent)iterator.next();

				// SCD
				buffer.writeRaw( "SCD+" );
				if (
					component.isFrequencyDimension()
					&&
					GESMES_TS.equalsIgnoreCase( interchangeFormat )
				) {
					buffer.writeRaw( "13" );
				} else if ( component.isTimeDimension() ) {
					buffer.writeRaw( "1" );
				} else if ( component.isKeyStructureDimension() ) {
					buffer.writeRaw( "4" );
				} else if ( component.isArrayCell() ) {
					buffer.writeRaw( "3" );
				} else if ( component.isAttribute() ) {
					buffer.writeRaw( "Z09" );
				}
				buffer.writeRaw( '+' );
				buffer.writeEscaped( component.getConceptId() );
				buffer.writeRaw( "++++:" );
				buffer.writeEscaped( component.getPosition() );
				buffer.writeRaw( '\'' );
				++segmentCounter;

				// ATT 1
				buffer.writeRaw( "ATT+3+5" );
				buffer.writeRaw( "+:::" );
				buffer.writeEscaped( component.getFormat() );
				buffer.writeRaw( '\'' );
				++segmentCounter;

				if ( !"".equals( component.getUsageStatus() ) ) {
					// ATT 2
					buffer.writeRaw( "ATT+3+35+" );
					if ( component.isMandatory() ) {
						buffer.writeRaw( '2' );
					} else {
						buffer.writeRaw( '1' );
					}
					buffer.writeRaw( ":USS'" );
					++segmentCounter;
				}

				if ( !"".equals( component.getAttachmentLevel() ) ) {
					// ATT 3
					buffer.writeRaw( "ATT+3+32+" );
					if ( component.isAttachedToDataSet() ) {
						buffer.writeRaw( "1" );
					} else if ( component.isAttachedToSiblingGroup() ) {
						buffer.writeRaw( "9" );
					} else if ( component.isAttachedToTimeSeries() ) {
						buffer.writeRaw( "4" );
					} else if ( component.isAttachedToObservation() ) {
						buffer.writeRaw( "5" );
					}
					buffer.writeRaw( ":ALV'" );
					++segmentCounter;
				}

				// IDE
				if ( !"".equals( component.getCodeListId() ) ) {
					buffer.writeRaw( "IDE+1+" );
					buffer.writeEscaped(
						component.getCodeListId()
					);
					buffer.writeRaw( '\'' );
					++segmentCounter;
				}

				buffer.newLine();
			}
		}
		catch ( final IOException ioe ) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"handle( KeyFamily )",
				"Output problem.",
				ioe
			);
			close();
		}
	}

	/* ---------------------- Time Series ----------------------- */

	/**
	 * <P>Process a TimeSeries.</P>
	 *
	 * <p>Being an unlimited writer, no splits are done when the
	 * series length exceeds 9999 observations.</p>
	 *
	 * @see EDIFACTWriter
	 */
	public void handle ( final TimeSeries timeSeries ) {
		try {
			// The position of the last written observation within
			// observationList.
			// int position = 0;	// TODO: split long ARR segments!

			// ARR
			buffer.writeRaw( "ARR++" );
			writeKey( timeSeries.getKey() );
			buffer.writeRaw( ':' );

			// Write the series' time specification
			buffer.writeRaw( timeSeries.getPeriod() );
			buffer.writeRaw( ':' );
			buffer.writeRaw( timeSeries.getPeriodFormat() );

			// Write the data contents.
			// A note on performance optimisation:
			// The way, conditional information fields are handled below
			// performs well under the assumption that observations
			// usually carry only a value and a status.
			// If there are lots of observations with also
			// confidentiality or prebreak value information,
			// some comparisons with the empty string
			// and related method calls are performed twice or more
			// often. In this case it may be better to compute the
			// boolean expressions once and store them in a local variable.
			//
			// It may be considered to extract all observation
			// fields into local variables to minimize the number of
			// method calls. A smart compiler may do equivalent
			// optimisation on its own.
			//
			// At least try this when performance-tuning.
			boolean firstObservation = true;
			for (
				final Iterator iterator
				= timeSeries.observationIterator();
				iterator.hasNext();
				firstObservation = false
			) {
				final Observation observation
				= (Observation)iterator.next();

				// The : or + should be written for every observation
				// regardless of if there are actually
				// any nonvoid observations left.
				// At least the specification says so.
				buffer.writeRaw( firstObservation ? ':' : '+' );

				// Catch absolutely void observations
				if ( observation.isEmpty() ) continue;

				// Nonvoid observations must carry both,
				// a value and a status.
				buffer.writeEscaped( observation.getValue() );
				buffer.writeRaw( ':' );
				buffer.writeEscaped( observation.getStatus() );

				// Catch observations without additional
				// confidentiality or prebreak value information
				if (
					"".equals( observation.getConfidentiality() )
					&&
					"".equals( observation.getPreBreakValue() )
				) continue;

				// Write the confidentiality information
				// or simply the colon, if this is empty.
				buffer.writeRaw( ':' );
				buffer.writeEscaped( observation.getConfidentiality() );

				// Catch observations without
				// prebreak value information
				if (
					"".equals( observation.getPreBreakValue() )
				) continue;

				// Write the prebreak value
				buffer.writeRaw( ':' );
				buffer.writeEscaped( observation.getPreBreakValue() );
			}

			buffer.writeRaw( '\'' );
			buffer.newLine();
			++segmentCounter;
		}
		catch ( final IOException ioe ) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"handle( TimeSeries )",
				"Output problem.",
				ioe
			);
			close();
		}
	}

	/* ----------------------- Attribute ------------------------ */

	/**
	 * <P>Process an Attribute.</P>
	 *
	 * <p>Being an unlimited writer, no splits are done when some
	 * segment counts exceeds their maximum number of iterations.</p>
	 *
	 * @see EDIFACTWriter
	 */
	public void handle ( final Attribute attribute ) {
		try {
			final KeyFamilyComponent component
			= currentKeyFamily.getComponent( attribute.getName() );
			final String scope;
			final int dimCount;
			if ( component.isAttachedToDataSet() ) {
				scope = DATA_SET_SCOPE;
				dimCount = 0;
			} else if (
				component.isAttachedToSiblingGroup()
				||
				component.isAttachedToTimeSeries()
			) {
				scope = SERIES_OR_SIBLING_SCOPE;
				dimCount = currentKeyFamily.getKeyDimensionCount();
			} else if ( component.isAttachedToObservation() ) {
				scope = OBSERVATION_SCOPE;
				dimCount = currentKeyFamily.getKeyDimensionCount()
					+ currentKeyFamily.getTimeDimensionCount();
			} else {
				componentHelper.logger.log(
					Level.SEVERE,
					componentHelper.logName,
					"handle( Attribute )",
					"Invalid attachment level \""
					+ component.getAttachmentLevel()
					+ "\" for attribute \"" + attribute.getName()
					+ "\".",
					attribute
				);
				scope = "";
				dimCount = 0;
			}

			// Only the first attribute event per data set
			// implies an FNS segment to be written.
			// FNS
			if ( haveNotWrittenFNS ) {
				buffer.writeRaw( "FNS+Attributes:10'" );
				buffer.newLine();
				++segmentCounter;

				// The FNS segment now has been written
				haveNotWrittenFNS = false;

				// A REL segment will be needed for the next attribute
				previousAttachmentScope = "<first>";
			}

			// A REL segment is written only if it is the first one
			// or the previous one had a different attachment scope.
			// The below test is also true for the first attribute,
			// because previousAttachmentScope has been initialized
			// with a value impossible for the attachment scope.
			//
			// REL
			if ( ! previousAttachmentScope.equals( scope ) ) {
				buffer.writeRaw( "REL+Z01+" );
				buffer.writeEscaped( scope );
				buffer.writeRaw( '\'' );
				buffer.newLine();
				++segmentCounter;

				// Now the attachment scope has been set.
				previousAttachmentScope = scope;

				// Need ARR segment
				previousKey = "<first>";
				previousPeriod = previousPeriodFormat = "";
			}

			if ( ! (
				previousKey.equals( attribute.getKey() )
				&&
				previousPeriod.equals( attribute.getPeriod() )
				&&
				previousPeriodFormat.equals( attribute.getPeriodFormat() )
			) ) {
				// ARR
				buffer.writeRaw( "ARR+" );
				buffer.writeRaw( dimCount );
				buffer.writeRaw( '+' );
				writeKey( attribute.getKey() );
				buffer.writeRaw( ':' );
				buffer.writeEscaped( attribute.getPeriod() );
				buffer.writeRaw( ':' );
				buffer.writeEscaped( attribute.getPeriodFormat() );
				buffer.writeRaw( '\'' );
				buffer.newLine();
				++segmentCounter;

				// Need ARR segment
				previousKey = attribute.getKey();
				previousPeriod = attribute.getPeriod();
				previousPeriodFormat = attribute.getPeriodFormat();
			}

			// IDE + ( CDV | FTX )
			buffer.writeRaw( "IDE+" );
			if ( component.isCoded() ) {
				// IDE
				buffer.writeRaw( CODED );
				buffer.writeRaw( '+' );
				buffer.writeEscaped( attribute.getName() );
				buffer.writeRaw( '\'' );
				++segmentCounter;

				if ( inURMessage ) {
					// CDV
					buffer.writeRaw( "CDV+" );
					buffer.writeEscaped( attribute.getValue() );
					buffer.writeRaw( '\'' );
					++segmentCounter;
				}
			} else {
				// IDE
				buffer.writeRaw( UNCODED );
				buffer.writeRaw( '+' );
				buffer.writeEscaped( attribute.getName() );
				buffer.writeRaw( '\'' );
				++segmentCounter;

				if ( inURMessage ) {
					// FTX segments to hold the value
					writeFTXSegments( attribute.getValue() );
				}
			}

			buffer.newLine();
		}
		catch ( final IOException ioe ) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"handle( Attribute )",
				"Output problem.",
				ioe
			);
			close();
		}
	}

	/* ==================== Private Methods ===================== */

	/**
	 * Find first character after start in string not equal c.
	 *
	 * @param string the string where to search.
	 * @param index the starting position from where to search.
	 * It must hold 0 &lt;= index &lt;= string.length().
	 * Otherwise the behaviour is undefined.
	 * @param c the character the occurrences of which must be skipped.
	 *
	 * @return the position of the first character
	 * in string beginning from index which is not equal to c.
	 * If there are no other characters than c from position index
	 * until the end of string the length of string will be returned.
	 */
	private static int findNot (
		final String string, int index, final char c
	) {
		final int stringLen = string.length();

		while (
			index < stringLen
			&&
			c == string.charAt( index )
		) ++index;

		return index;
	}

	/**
	 * <P>Write a String into FTX segments.</P>
	 *
	 * <P>The String must be written with
	 * properly escaped syntax characters.
	 * Further it must be split into components of no more than
	 * 70 characters size (including escape characters).
	 * Since each FTX segment can hold no more than five such
	 * components, it may be necessary to write multiple FTX
	 * segments.</P>
	 *
	 * @param string the String to be in the FTX segment.
	 * This argument must not contain control characters.
	 *
	 * @return the length of string if the whole string
	 * was written or the position after the last one written.
	 *
	 */
	private void writeFTXSegments ( final String string )
		throws IOException
	{
		final int stringLen = string.length();

		// The number of space characters outstanding to write.
		int pendingSpaces = 0;
		int ftxSegment = 0;
		int positionInSegment = 0;

		for ( int i = 0; i < stringLen; ++i ) {
			final char c = string.charAt( i );
			final boolean escapeChar = ":+'?".indexOf( c ) >= 0;

			if ( c == ' ' ) {
				++pendingSpaces;
			} else {
				while ( pendingSpaces + positionInSegment >= 350 ) {
					if ( pendingSpaces >= 350 ) {
						// Cannot build valid FTX segments
						// to represent 350 spaces.
						// Therefore I deliberately cheat here.
						pendingSpaces = 0;
						componentHelper.logger.log(
							Level.SEVERE,
							componentHelper.logName,
							"writeFTXSegments()",
							"EDIFACT does not allow to write"
							+ " 350 consecutive spaces in \"" + string + "\"."
						);
					}
					if ( positionInSegment > 0 ) {
						buffer.writeRaw( '\'' );
						positionInSegment = 0;
						++segmentCounter;
					}
				}
				SPCWRITE:
				while ( pendingSpaces > 0 ) {
					switch ( positionInSegment ) {
						case 0:
							buffer.writeRaw( "FTX+ACM+++" );
							if ( ++ftxSegment > 20 ) {
								componentHelper.logger.log(
									Level.SEVERE,
									componentHelper.logName,
									"writeFTXSegments()",
									"Need more FTX segments than allowed"
									+ " to write \"" + string + "\"."
								);
							}
							break;
						case 70:
						case 140:
						case 210:
						case 280:
							buffer.writeRaw( ':' ); break;
						case 350:
							buffer.writeRaw( '\'' );
							positionInSegment = 0;
							continue SPCWRITE;
					}
					buffer.writeRaw( ' ' );
					--pendingSpaces;
					++positionInSegment;
				}
				if ( escapeChar ) {
					if ( ( positionInSegment + 1 ) % 70 == 0 ) {
						// force a premature component break
						// by faking to use up one position.
						++positionInSegment;
					}
					switch ( positionInSegment ) {
						case 0:
							buffer.writeRaw( "FTX+ACM+++" );
							if ( ++ftxSegment > 20 ) {
								componentHelper.logger.log(
									Level.SEVERE,
									componentHelper.logName,
									"writeFTXSegments()",
									"Need more FTX segments than allowed"
									+ " to write \"" + string + "\"."
								);
							}
							break;
						case 70:
						case 140:
						case 210:
						case 280:
							buffer.writeRaw( ':' ); break;
						case 350:
							buffer.writeRaw( "'FTX+ACM+++" );
							positionInSegment = 0;
							if ( ++ftxSegment > 20 ) {
								componentHelper.logger.log(
									Level.SEVERE,
									componentHelper.logName,
									"writeFTXSegments()",
									"Need more FTX segments than allowed"
									+ " to write \"" + string + "\"."
									);
							}
							++segmentCounter;
					}
					buffer.writeRaw( '?' );
					++positionInSegment;
				}
				switch ( positionInSegment ) {
					case 0:
						buffer.writeRaw( "FTX+ACM+++" );
						if ( ++ftxSegment > 20 ) {
							componentHelper.logger.log(
								Level.SEVERE,
								componentHelper.logName,
								"writeFTXSegments()",
								"Need more FTX segments than allowed"
								+ " to write \"" + string + "\"."
							);
						}
						break;
					case 70:
					case 140:
					case 210:
					case 280:
						buffer.writeRaw( ':' ); break;
					case 350:
						buffer.writeRaw( "'FTX+ACM+++" );
						positionInSegment = 0;
						if ( ++ftxSegment > 20 ) {
							componentHelper.logger.log(
								Level.SEVERE,
								componentHelper.logName,
								"writeFTXSegments()",
								"Need more FTX segments than allowed"
								+ " to write \"" + string + "\"."
							);
						}
						++segmentCounter;
				}
				buffer.writeRaw( c );
				++positionInSegment;
			}
		}

		if ( ftxSegment > 0 ) {
			// complete the segment
			buffer.writeRaw( '\'' );
		} else {
			buffer.writeRaw( "FTX+ACM'" );
		}
		buffer.newLine();
		++segmentCounter;
	}

	/**
	 * Write a {@link TimeSeries} or {@link Attribute} key.
	 * The key separator dot must be replaced by a colon.
	 *
	 * @throws IOException in case of output problems.
	 *
	 * @see Int.ecb.gesmescb.TimeSeries#KEY_SEPARATOR
	 * @see Int.ecb.gesmescb.Attribute#KEY_SEPARATOR
	 */
	private void writeKey ( final String key ) throws IOException {
		final int length = key.length();
		for ( int i = 0; i < length; ++i ) {
			final char c = key.charAt( i );
			if ( c == Attribute.KEY_SEPARATOR ) {
				buffer.writeRaw( ':' );
			} else {
				buffer.writeEscaped( c );
			}
		}
	}
}
