// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.edifact;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.LineNumberReader;
import java.io.Reader;
import java.io.StringReader;
import java.util.Calendar;
import java.util.Date;

import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;
import Int.ecb.gesmescb.Attribute;
import Int.ecb.gesmescb.CodeList;
import Int.ecb.gesmescb.Contact;
import Int.ecb.gesmescb.ContactChannel;
import Int.ecb.gesmescb.InterchangeBegin;
import Int.ecb.gesmescb.InterchangeEnd;
import Int.ecb.gesmescb.KeyFamily;
import Int.ecb.gesmescb.KeyFamilyComponent;
import Int.ecb.gesmescb.MessageBegin;
import Int.ecb.gesmescb.MessageEnd;
import Int.ecb.gesmescb.Observation;
import Int.ecb.gesmescb.StatisticalConcept;
import Int.ecb.gesmescb.TimeSeries;
import Int.ecb.gesmescb.event.TestEventSource;
import Int.ecb.gesmescb.metabase.Metabase;

/* ***************************** Class ****************************** */

/**
 * Tests the {@link EDIFACTUnlimitedWriter}.
 *
 * @author
 * <A HREF="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</A>
 *
 * @version @(#)$Id: EDIFACTUnlimitedWriterTest.java,v 1.1 2017/01/19 16:14:05 trainot Exp $
 */
public class EDIFACTUnlimitedWriterTest extends TestCase {

	/* ======================= Constants ======================== */

	protected static final Date versionDate;
	static {
		final Calendar calendar = Calendar.getInstance();
		calendar.clear();
		versionDate = calendar.getTime();
	}

	protected static final InterchangeBegin
		INTERCHANGE_BEGIN = new InterchangeBegin(
			"IREF000215", "4F0", "ZZZ", "001201:1410", true
		);

	protected static final String
		INTERCHANGE_BEGIN_OUT = "UNA:+.? '\n"
			+ "UNB+UNOC:3+4F0+ZZZ+001201:1410+IREF000215++GESMES/TS++++1'\n";

	protected static final MessageBegin
		MESSAGE_BEGIN_1 = new MessageBegin(
			"MREF000001", "4F0", "ZZZ", "ECB",
			MessageBegin.STATISTICAL_DEFINITIONS,
			"Test metadata",
			new Contact[] {
				new Contact(
					Contact.DATA_PROCESSING,
					"SIS",
					"Obelix",
					new ContactChannel[]{
						new ContactChannel(
							ContactChannel.EMAIL,
							"obelix@village.fr"
						),
						new ContactChannel(
							ContactChannel.TELEPHONE,
							"003"
						)
					}
				),
				new Contact(
					Contact.HEAD_OF_DATA_PROCESSING,
					"SIS",
					"Majestix",
					new ContactChannel[]{
						new ContactChannel(
							ContactChannel.FAX,
							"0002"
						)
					}
				)
			}
		);

	protected static final String
		MESSAGE_BEGIN_1_OUT = "UNH+MREF000001+GESMES:2:1:E6'\n"
			+ "BGM+73'\n"
			+ "NAD+Z02+ECB'\n"
			+ "NAD+MR+ZZZ'\n"
			+ "NAD+MS+4F0'\n"
			+ "IDE+10+Test metadata'\n"
			+ "CTA+CP+SIS:Obelix'\n"
			+ "COM+obelix@village.fr:EM'\n"
			+ "COM+003:TE'\n"
			+ "CTA+CE+SIS:Majestix'\n"
			+ "COM+0002:FX'\n";

	protected static final CodeList
		CL_OBJ = new CodeList(
			"CL_OBJ", "Test codes", "", versionDate,
			new String[] { "EPS", "HOX", "SOX", "OBI", "NOX" },
			new String[] { "", "Hoax", "Smelly ObjeX", "German drink", "Latin night" }
		),
		CL_CONF = new CodeList(
			"CL_CONF", "Further test codes", "", versionDate,
			new String[] { "N", "C", "S", "T", "U" },
			new String[] {
				"Not classified", "Confidential",
				"Secret", "Top secret", "So secret that unknown"
			}
		);

	protected static final String
		CL_OBJ_OUT = "VLI+CL_OBJ+++Test codes'\n"
			+ "CDV+EPS'FTX+ACM'\n"
			+ "CDV+HOX'FTX+ACM+++Hoax'\n"
			+ "CDV+SOX'FTX+ACM+++Smelly ObjeX'\n"
			+ "CDV+OBI'FTX+ACM+++German drink'\n"
			+ "CDV+NOX'FTX+ACM+++Latin night'\n",
		CL_CONF_OUT = "VLI+CL_CONF+++Further test codes'\n"
			+ "CDV+N'FTX+ACM+++Not classified'\n"
			+ "CDV+C'FTX+ACM+++Confidential'\n"
			+ "CDV+S'FTX+ACM+++Secret'\n"
			+ "CDV+T'FTX+ACM+++Top secret'\n"
			+ "CDV+U'FTX+ACM+++So secret that unknown'\n";

	protected static final String
		SECRECY_OUT = "STC+SECRECY'FTX+ACM'\n",
		INSANITY_OUT = "STC+INSANITY'FTX+ACM+++The degree of'\n";

	protected static final StatisticalConcept
		SECRECY = new StatisticalConcept(
			"SECRECY", "", "", versionDate
		),
		INSANITY = new StatisticalConcept(
			"INSANITY", "The degree of", "", versionDate
		);

	protected static final KeyFamily
		SIS_FOOB = new KeyFamily(
			"SIS_FOOB", "Secret Intelligence Service: Foobar stuff", "",
			versionDate,
			new KeyFamilyComponent[] {
				new KeyFamilyComponent(
					"WHO",
					KeyFamilyComponent.KEY_STRUCTURE_DIMENSION,
					"1", "AN3", "", "", "CL_OBJ"
				),
				new KeyFamilyComponent(
					"SECURITY", KeyFamilyComponent.KEY_STRUCTURE_DIMENSION,
					"2", "AN1", "", "", "CL_CONF"
				),
				new KeyFamilyComponent(
					"TIME_PERIOD", KeyFamilyComponent.TIME_DIMENSION,
					"3", "AN..35", "", "", ""
				),
				new KeyFamilyComponent(
					"TIME_FORMAT", KeyFamilyComponent.TIME_DIMENSION,
					"4", "AN3", "", "", ""
				),
				new KeyFamilyComponent(
					"OBS_VALUE", KeyFamilyComponent.ARRAY_CELL,
					"5", "AN..15", "", "", ""
				),
				new KeyFamilyComponent(
					"OBS_STATUS", KeyFamilyComponent.ARRAY_CELL,
					"6", "AN1",
					KeyFamilyComponent.MANDATORY,
					KeyFamilyComponent.ATTACHED_TO_OBSERVATION,
					"CL_OBS_STATUS"
				),
				new KeyFamilyComponent(
					"OBS_CONF", KeyFamilyComponent.ARRAY_CELL,
					"7", "AN1",
					KeyFamilyComponent.CONDITIONAL,
					KeyFamilyComponent.ATTACHED_TO_OBSERVATION,
					"CL_OBS_CONF"
				),
				new KeyFamilyComponent(
					"OBS_PRE_BREAK", KeyFamilyComponent.ARRAY_CELL,
					"8", "AN..15",
					KeyFamilyComponent.CONDITIONAL,
					KeyFamilyComponent.ATTACHED_TO_OBSERVATION,
					""
				),
				new KeyFamilyComponent(
					"STUS", KeyFamilyComponent.ATTRIBUTE, "", "AN..70",
					KeyFamilyComponent.CONDITIONAL,
					KeyFamilyComponent.ATTACHED_TO_DATA_SET,
					"CL_UNIT"
				),
				new KeyFamilyComponent(
					"SLUS", KeyFamilyComponent.ATTRIBUTE, "", "AN..12",
					KeyFamilyComponent.CONDITIONAL,
					KeyFamilyComponent.ATTACHED_TO_DATA_SET,
					""
				),
				new KeyFamilyComponent(
					"XCUS", "Z09", "", "AN..70",
					KeyFamilyComponent.MANDATORY,
					KeyFamilyComponent.ATTACHED_TO_TIME_SERIES,
					"CL_UNIT"
				),
				new KeyFamilyComponent(
					"APPETITE", KeyFamilyComponent.ATTRIBUTE, "", "AN..1050",
					KeyFamilyComponent.MANDATORY,
					KeyFamilyComponent.ATTACHED_TO_SIBLING_GROUP,
					""
				),
				new KeyFamilyComponent(
					"MARIA", KeyFamilyComponent.ATTRIBUTE, "", "AN..1750",
					KeyFamilyComponent.CONDITIONAL,
					KeyFamilyComponent.ATTACHED_TO_SIBLING_GROUP,
					""
				),
				new KeyFamilyComponent(
					"BANANA", KeyFamilyComponent.ATTRIBUTE, "", "AN..170",
					KeyFamilyComponent.CONDITIONAL,
					KeyFamilyComponent.ATTACHED_TO_SIBLING_GROUP,
					"CL_UNIT"
				),
				new KeyFamilyComponent(
					"COMMENT", KeyFamilyComponent.ATTRIBUTE, "", "AN..100",
					KeyFamilyComponent.CONDITIONAL,
					KeyFamilyComponent.ATTACHED_TO_OBSERVATION,
					""
				)
			}
		);

	protected static final String
		SIS_FOOB_OUT = "ASI+SIS_FOOB'FTX+ACM+++Secret Intelligence Service?: Foobar stuff'\n"
			+ "SCD+13+WHO++++:1'ATT+3+5+:::AN3'IDE+1+CL_OBJ'\n"
			+ "SCD+4+SECURITY++++:2'ATT+3+5+:::AN1'IDE+1+CL_CONF'\n"
			+ "SCD+1+TIME_PERIOD++++:3'ATT+3+5+:::AN..35'\n"
			+ "SCD+1+TIME_FORMAT++++:4'ATT+3+5+:::AN3'\n"
			+ "SCD+3+OBS_VALUE++++:5'ATT+3+5+:::AN..15'\n"
			+ "SCD+3+OBS_STATUS++++:6'ATT+3+5+:::AN1'ATT+3+35+2:USS'ATT+3+32+5:ALV'IDE+1+CL_OBS\n_STATUS'\n"
			+ "SCD+3+OBS_CONF++++:7'ATT+3+5+:::AN1'ATT+3+35+1:USS'ATT+3+32+5:ALV'IDE+1+CL_OBS_C\nONF'\n"
			+ "SCD+3+OBS_PRE_BREAK++++:8'ATT+3+5+:::AN..15'ATT+3+35+1:USS'ATT+3+32+5:ALV'\n"
			+ "SCD+Z09+COMMENT'ATT+3+5+:::AN..100'ATT+3+35+1:USS'ATT+3+32+5:ALV'\n"
			+ "SCD+Z09+XCUS'ATT+3+5+:::AN..70'ATT+3+35+2:USS'ATT+3+32+4:ALV'IDE+1+CL_UNIT'\n"
			+ "SCD+Z09+APPETITE'ATT+3+5+:::AN..1050'ATT+3+35+2:USS'ATT+3+32+9:ALV'\n"
			+ "SCD+Z09+BANANA'ATT+3+5+:::AN..170'ATT+3+35+1:USS'ATT+3+32+9:ALV'IDE+1+CL_UNIT'\n"
			+ "SCD+Z09+MARIA'ATT+3+5+:::AN..1750'ATT+3+35+1:USS'ATT+3+32+9:ALV'\n"
			+ "SCD+Z09+SLUS'ATT+3+5+:::AN..12'ATT+3+35+1:USS'ATT+3+32+1:ALV'\n"
			+ "SCD+Z09+STUS'ATT+3+5+:::AN..70'ATT+3+35+1:USS'ATT+3+32+1:ALV'IDE+1+CL_UNIT'\n";

	protected static final MessageEnd
		MESSAGE_END_1 = new MessageEnd(
			"MREF000001"
		);

	protected static final String
		MESSAGE_END_1_OUT = "UNT+97+MREF000001'\n";

	protected static final MessageBegin
		MESSAGE_BEGIN_2 = new MessageBegin(
			"MREF000002", "4F0", "BIS", "ECB",
			MessageBegin.STATISTICAL_DATA,
			"Test data",
			new Contact[] {
				new Contact(
					Contact.INFORMATION_PRODUCTION,
					"SIS",
					"Troubadix",
					new ContactChannel[] {}
				)
			},
			"FOO_BAROON", MessageBegin.UPDATE_OR_REPLACE_ACTION,
			"200008091923", "", "", "SIS_FOOB"
		);

	protected final static String
		MESSAGE_BEGIN_2_OUT = "UNH+MREF000002+GESMES:2:1:E6'\n"
			+ "BGM+74'\n"
			+ "NAD+Z02+ECB'\n"
			+ "NAD+MR+BIS'\n"
			+ "NAD+MS+4F0'\n"
			+ "IDE+10+Test data'\n"
			+ "CTA+CC+SIS:Troubadix'\n"
			+ "DSI+FOO_BAROON'\n"
			+ "STS+3+7'\n"
			+ "DTM+242:200008091923:203'\n"
			+ "IDE+5+SIS_FOOB'\n"
			+ "GIS+AR3'\n"
			+ "GIS+1:::-'\n";

	protected static final TimeSeries
		HOX_S = new TimeSeries(
			"HOX.S", "2001", "602",
			new Observation[] {
				new Observation( "13.6", "S", "C", "" )
			}
		),
		SOX_S = new TimeSeries(
			"SOX.S", "20102013", "702",
			new Observation[] {
				new Observation( "13.6", "S", "N", "" ),
				new Observation( "13.5", "A", "", "" ),
				new Observation( "1", "A", "C", "13" ),
				new Observation( "1.2", "C", "", "" )
			}
		),
		SOX_N = new TimeSeries(
			"SOX.N", "20102019", "702",
			new Observation[] {
				new Observation( "13.6", "S", "N", "" ),
				new Observation( "", "", "", "" ),
				new Observation( "", "", "", "" ),
				new Observation( "", "", "", "" ),
				new Observation( "", "", "", "" ),
				new Observation( "", "", "", "" ),
				new Observation( "", "", "", "" ),
				new Observation( "13.5", "A", "", "" ),
				new Observation( "1", "A", "C", "13" ),
				new Observation( "1.2", "C", "", "" )
			}
		);

	protected static final String
		HOX_S_OUT = "ARR++HOX:S:2001:602:13.6:S:C'\n",
		SOX_S_OUT = "ARR++SOX:S:20102013:702:13.6:S:N+13.5:A+1:A:C:13+1.2:C'\n",
		SOX_N_OUT = "ARR++SOX:N:20102019:702:13.6:S:N+++++++13.5:A+1:A:C:13+1.2:C'\n";

	protected static final Attribute
		STUS = new Attribute(
			"", "", "", "STUS", "RUHR"
		),
		SLUS = new Attribute(
			"", "", "", "SLUS", "Kulimuliluli"
		),
		XCUS = new Attribute(
			"NOX.C", "", "", "XCUS", "XANA"
		),
		APPETITE = new Attribute(
			".C", "", "", "APPETITE",
			"A123456789012345678901234567890123456789"
			+ "0123456789012345678901234 spc "
			+ "B123456789012345678901234567890123456789"
			+ "01234567890123456789 spcs5" + spcs( 4 )
			+ "C123456789012345678901234567890123456789"
			+ "012345678901234567890123456789"
			+ "D123456789012345678901234567890123456789"
			+ "012345678901234567890123456789"
			+ "E123456789012345678901234567890123456789"
			+ "01234567890123456789 spcs5" + spcs( 4 )
			+ "F123456789012345678901234567890123456789"
			+ "012345678901234567890123456789"
			+ "G123456789012345678901234567890123456789"
			+ "012345678901234567 spcs3" + spcs( 5 ) + "9"
			+ "H123456789012345678901234567890123456789"
			+ "012345678901234567890123456789"
			+ "I123456789012345678901234567890123456789"
			+ "012345678901234567890123456789"
			+ "J123456789012345678901234567890123456789"
			+ "012345678901234567 s3" + spcs( 8 ) + "9"
		),
		MARIA = new Attribute(
			".C", "", "", "MARIA",
			"A123456789012345678901234567890123456789"
			+ "01234567890123456789012345678:"
			+ "B123456789012345678901234567890123456789"
			+ "01234567890123456789012345 +++"
			+ "C"
			+ spcs( 69 )
			+ "D123456789012345678901234567890123456789"
			+ "0123456789012345678901234567 +"
			+ "E123456789012345678901234567890123456789"
			+ "01234567890123456789" + spcs( 3 ) + "''''''?"
		),
		BANANA = new Attribute(
			"NOX.", "", "", "BANANA", "SWEET"
		),
		COMMENT = new Attribute(
			"SOX.C", "2020", "602", "COMMENT", "Text01234"
		);

	protected static final String
		ATT_1_OUT = "FNS+Attributes:10'\n"
			+ "REL+Z01+1'\n"
			+ "ARR+0'\n"
			+ "IDE+Z10+STUS'CDV+RUHR'\n"
			+ "IDE+Z11+SLUS'FTX+ACM+++Kulimuliluli'\n"
			+ "REL+Z01+4'\n"
			+ "ARR+2+NOX:C'\n"
			+ "IDE+Z10+XCUS'CDV+XANA'\n"
			+ "ARR+2+:C'\n"
			+ "IDE+Z11+APPETITE'FTX+ACM+++"
			+ "A123456789012345678901234567890123456789"
			+ "0123456789012345678901234 spc :"
			+ "B123456789012345678901234567890123456789"
			+ "01234567890123456789 spcs5" + spcs( 4 ) + ":"
			+ "C123456789012345678901234567890123456789"
			+ "012345678901234567890123456789:"
			+ "D123456789012345678901234567890123456789"
			+ "012345678901234567890123456789:"
			+ "E123456789012345678901234567890123456789"
			+ "01234567890123456789 spcs5'FTX+ACM+++" + spcs( 4 )
			+ "F123456789012345678901234567890123456789"
			+ "01234567890123456789012345:6789"
			+ "G123456789012345678901234567890123456789"
			+ "012345678901234567 spcs3" + spcs( 2 ) + ":" + spcs( 3 ) + "9"
			+ "H123456789012345678901234567890123456789"
			+ "01234567890123456789012345:6789"
			+ "I123456789012345678901234567890123456789"
			+ "01234567890123456789012345:6789"
			+ "J123456789012345678901234567890123456789"
			+ "012345678901234567 s3'FTX+ACM+++" + spcs( 8 ) + "9'\n"
			+ "IDE+Z11+MARIA'FTX+ACM+++"
			+ "A123456789012345678901234567890123456789"
			+ "01234567890123456789012345678:?:"
			+ "B123456789012345678901234567890123456789"
			+ "01234567890123456789012345 :?+?+?+"
			+ "C" + spcs( 63 )
			+ ":" + spcs( 6 )
			+ "D123456789012345678901234567890123456789"
			+ "012345678901234567890123:4567 ?+"
			+ "E123456789012345678901234567890123456789"
			+ "01234567890123456789'FTX+ACM+++" + spcs( 3 ) + "?'?'?'?'?'?'??'"
			+ "ARR+2+NOX'\n"
			+ "IDE+Z10+BANANA'CDV+SWEET'\n"
			+ "REL+Z01+5'\n"
			+ "ARR+4+SOX:C:2020:602'\n"
			+ "IDE+Z11+COMMENT'\n"
			+ "FTX+ACM+++Text01234'\n";

	protected static final MessageEnd
		MESSAGE_END_2 = new MessageEnd(
			"MREF000002"
		);

	protected static final String
		MESSAGE_END_2_OUT = "UNT+43+MREF000002'\n";

	protected static final MessageBegin
		MESSAGE_BEGIN_3 = new MessageBegin(
			"MREF000003", "4F0", "BIS", "ECB",
			MessageBegin.STATISTICAL_DATA,
			"Test data delete",
			new Contact[]{},
			"FOO_BALLOON", MessageBegin.DELETE_ACTION,
			"200008091923", "2000", "602", "SIS_FOOB"
		);

	protected static final String
		MESSAGE_BEGIN_3_OUT = "UNH+MREF000003+GESMES:2:1:E6'\n"
			+ "BGM+74'\n"
			+ "NAD+Z02+ECB'\n"
			+ "NAD+MR+BIS'\n"
			+ "NAD+MS+4F0'\n"
			+ "IDE+10+Test data delete'\n"
			+ "DSI+FOO_BALLOON'\n"
			+ "STS+3+6'\n"
			+ "DTM+242:200008091923:203'\n"
			+ "DTM+Z02:2000:602'\n"
			+ "IDE+5+SIS_FOOB'\n"
			+ "GIS+AR3'\n"
			+ "GIS+1:::-'\n";

	protected static final TimeSeries
		HOX_S_2 = new TimeSeries(
			"HOX.S", "2001", "602", new Observation[] {}
		),
		SOX_S_2 = new TimeSeries(
			"SOX.S", "", "", new Observation[] {}
		),
		S_2 = new TimeSeries(
			".S", "", "", new Observation[] {}
		),
		SOX_2 = new TimeSeries(
			"SOX.", "", "", new Observation[] {}
		);

	protected static final String
		HOX_S_2_OUT = "ARR++HOX:S:2001:602'\n",
		SOX_S_2_OUT = "ARR++SOX:S'\n",
		S_2_OUT = "ARR++:S'\n",
		SOX_2_OUT = "ARR++SOX'\n";

	protected static final Attribute
		STUS_2 = new Attribute(
			"", "", "", "STUS", ""
		),
		SLUS_2 = new Attribute(
			"", "", "", "SLUS", ""
		),
		XCUS_2 = new Attribute(
			"NOX.S", "", "", "XCUS", ""
		),
		APPETITE_2 = new Attribute(
			".N", "", "", "APPETITE", ""
		),
		MARIA_2 = new Attribute(
			"NOX.", "", "", "MARIA", ""
		),
		COMMENT_2 = new Attribute(
			"SOX.C", "2020", "602", "COMMENT", ""
		);

	protected static final String
		ATT_2_OUT = "FNS+Attributes:10'\n"
			+ "REL+Z01+1'\n"
			+ "ARR+0'\n"
			+ "IDE+Z10+STUS'\n"
			+ "IDE+Z11+SLUS'\n"
			+ "REL+Z01+4'\n"
			+ "ARR+2+NOX:S'\n"
			+ "IDE+Z10+XCUS'\n"
			+ "ARR+2+:N'\n"
			+ "IDE+Z11+APPETITE'\n"
			+ "ARR+2+NOX'\n"
			+ "IDE+Z11+MARIA'\n"
			+ "REL+Z01+5'\n"
			+ "ARR+4+SOX:C:2020:602'\n"
			+ "IDE+Z11+COMMENT'\n";

	protected static final MessageEnd
		MESSAGE_END_3 = new MessageEnd(
			"MREF000003"
		);

	protected static final String
		MESSAGE_END_3_OUT = "UNT+33+MREF000003'\n";

	protected static final InterchangeEnd
		INTERCHANGE_END = new InterchangeEnd(
			"IREF000215"
		);

	protected static final String
		INTERCHANGE_END_OUT = "UNZ+3+IREF000215'\n";

	/**
	 * The input to the {@link EDIFACTUnlimitedWriter} under Test.
	 * Its contents are meaningless but must correspond to
	 * {@link #testOutput}.
	 */
	protected static final Object[] testInput = {
		INTERCHANGE_BEGIN,
		MESSAGE_BEGIN_1,
		CL_OBJ,
		CL_CONF,
		SECRECY,
		INSANITY,
		SIS_FOOB,
		MESSAGE_END_1,
		MESSAGE_BEGIN_2,
		HOX_S,
		SOX_S,
		SOX_N,
		STUS,
		SLUS,
		XCUS,
		APPETITE,
		MARIA,
		BANANA,
		COMMENT,
		MESSAGE_END_2,
		MESSAGE_BEGIN_3,
		HOX_S_2,
		SOX_S_2,
		S_2,
		SOX_2,
		STUS_2,
		SLUS_2,
		XCUS_2,
		APPETITE_2,
		MARIA_2,
		COMMENT_2,
		MESSAGE_END_3,
		INTERCHANGE_END
	};

	/**
	 * The text into which
	 * the {@link EDIFACTUnlimitedWriter} should translate
	 * {@link #testInput} when called in the correct way.
	 */
	protected static final String testOutput
		= INTERCHANGE_BEGIN_OUT
			+ MESSAGE_BEGIN_1_OUT
			+ CL_OBJ_OUT
			+ CL_CONF_OUT
			+ SECRECY_OUT
			+ INSANITY_OUT
			+ SIS_FOOB_OUT
			+ MESSAGE_END_1_OUT
			+ MESSAGE_BEGIN_2_OUT
			+ HOX_S_OUT
			+ SOX_S_OUT
			+ SOX_N_OUT
			+ ATT_1_OUT
			+ MESSAGE_END_2_OUT
			+ MESSAGE_BEGIN_3_OUT
			+ HOX_S_2_OUT
			+ SOX_S_2_OUT
			+ S_2_OUT
			+ SOX_2_OUT
			+ ATT_2_OUT
			+ MESSAGE_END_3_OUT
			+ INTERCHANGE_END_OUT;

	/* ========================= Fields ========================= */

	/**
	 * The file used for the location of test data.
	 */
	private File testFile;

	/* ====================== Constructors ====================== */

	public EDIFACTUnlimitedWriterTest( final String name ) {
		super( name );
	}

	/* ================== Protected Methods ===================== */

	/**
	 * Sets up the test fixture.
	 */
	protected void setUp () {
		try {
			testFile = File.createTempFile( "edifact_testoutput.", null );
			testFile.deleteOnExit();
		}
		catch ( IOException ioe ) {
			fail(
				"Write testFile failed: " + ioe
			);
		}

		final TestEventSource source = new TestEventSource();
		source.addListener( Metabase.getMetabase(), true );
		source.dispatch( MESSAGE_BEGIN_1 );
		source.dispatch( SIS_FOOB );
		source.dispatch( MESSAGE_END_1 );
	}

	/**
	 * Tears down the test fixture.
	 */
	protected void tearDown () throws IOException {
		final String writerClassName = createdEDIFACTWriterClassName( "Zampano " );
		System.getProperties().remove( writerClassName + ".format" );
		testFile.delete();
	}

	/** Factors EDIFACTUnlimitedWriters. */
	protected EDIFACTUnlimitedWriter createEDIFACTWriter (
		final String name, final String filename
	) throws IOException {
		return new EDIFACTUnlimitedWriter( name, filename );
	}

	/**
	 * Returns the name of the <code>EDIFACTUnlimitedWriter</code>
	 * created by {@link createEDIFACTWriter}.
	 */
	protected String createdEDIFACTWriterClassName (
		final String name
	) throws IOException {
		final EDIFACTUnlimitedWriter writer = createEDIFACTWriter(
			name, testFile.getPath()
		);
		writer.close();
		return writer.getClass().getName() + "." + name;
	}

	/* ===================== Public Methods ===================== */

	/**
	 * Assembles and returns a test suite for all
	 * the test methods of this test case.
	 *
	 * @return a non-null test suite
	 */
	public static Test suite () {
		return new TestSuite( EDIFACTUnlimitedWriterTest.class );
	}

	/**
	 * @param arg the array of argument {@link String}s.
	 */
	public static void main ( final String[] arg ) {
		final String[] testCaseName = {
			EDIFACTUnlimitedWriterTest.class.getName()
		};
		junit.textui.TestRunner.main( testCaseName );
	}

	/* ---------------------- Test Methods ---------------------- */

	/**
	 * Test {@link EDIFACTUnlimitedWriter}.
	 *
	 * @throws IOException if opening the EDIFACTUnlimitedWriter yields a problem.
	 */
	public void test () throws IOException {
		outputTest( testInput, testOutput );
	}

	/**
	 * Test Gesmes/CB capability.
	 */
	public void testUnlimitedCB () throws IOException {
		final Object[] input = new Object[] {
			INTERCHANGE_BEGIN,
			MESSAGE_BEGIN_1,
			SIS_FOOB,
			MESSAGE_END_1,
			INTERCHANGE_END
		};
		final StringBuffer kfOut = new StringBuffer( SIS_FOOB_OUT );
		kfOut.replace(
			SIS_FOOB_OUT.indexOf( "SCD+13" ),
			SIS_FOOB_OUT.indexOf( "+WHO" ),
			"SCD+4"
		);
		final String output = "UNA:+.? '\n"
			+ "UNB+UNOC:3+4F0+ZZZ+001201:1410+IREF000215++GESMES/CB++++1'\n"
			+ MESSAGE_BEGIN_1_OUT
			+ kfOut.toString()
			+ "UNT+71+MREF000001'\n"
			+ "UNZ+1+IREF000215'\n";
		final String writerClassName = createdEDIFACTWriterClassName( "Zampano" );
		System.setProperty( writerClassName + ".format", "GESMES/CB" );
		
		outputTest( input, output );
		
		System.getProperties().remove(writerClassName + ".format");		
	}

	/* =================== Protected Methods ==================== */

	/* ---------------------- Auxiliaries ----------------------- */

	protected void outputTest (
		final Object[] input, final String expectedOutput
	) throws IOException {
		final TestEventSource source = new TestEventSource();
		final Metabase metabase = Metabase.getMetabase();
		final EDIFACTUnlimitedWriter writer = createEDIFACTWriter(
			"Zampano", testFile.getPath()
		);

		source.addListener( metabase, false );
		source.addListener( writer, true );

		// Run the test.
		source.dispatch( input );

		// Close.
		writer.close();

		final LineNumberReader
			actual = new LineNumberReader(
				new InputStreamReader(
					new FileInputStream( testFile.getPath() ),
					"ISO8859-1"
				)
			),
			expected = new LineNumberReader(
				new StringReader( expectedOutput )
			);

		try {
			String a, e, line;
			do {
				a = readSegment( actual );
				e = readSegment( expected );
				line = "Line " + expected.getLineNumber();

				if ( e != null && a == null ) {
					System.out.println( "Received null but expected: " + e );
				}
				if ( e == null && a != null ) {
					System.out.println( "Expected null but received: " + a );
				}
				assertEquals( "Both same time zero" + line, a == null, e == null );

				if ( e == null ) break;

				if ( !a.equals( e ) ) {
					System.out.println( "expected: " + e );
					System.out.println( "received: " + a );
				}
				assertEquals( line, e, a );
			} while ( e != null );
		}
		finally {
			// Close also in case of failed assertions.
			actual.close();
			expected.close();
		}
	}

	/** Reads a junk of text until an apostropus. */
	protected static String readSegment ( final Reader reader )
		throws IOException
	{
		StringBuffer buffer = null;
		int c;
		boolean escaped= false;
		OUTER: while ( ( c = reader.read() ) >= 0 ) {
			if ( Character.isISOControl( (char)c ) ) continue;
			if ( buffer == null ) {
				buffer = new StringBuffer();
			}
			buffer.append( (char)c );
			switch ( c ) {
				case '?': escaped = !escaped; break;
				case '\'': if ( !escaped ) break OUTER;
				default: escaped = false;
			}
		}
		return buffer != null ? buffer.toString() : null;
	}

	protected static String spcs ( final int length ) {
		return stringMult( " ", length );
	}

	protected static String stringMult (
		final String string, final int multiplicity
	) {
		final StringBuffer buffer = new StringBuffer(
			string.length() * multiplicity
		);
		for ( int i = 0; i < multiplicity; ++i ) {
			buffer.append( string );
		}
		return buffer.toString();
	}
}
