// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.edifact;

import java.lang.String;
import java.lang.NullPointerException;
import java.io.Writer;
import java.io.IOException;
// import java.io.BufferedWriter;	// imported only for javadoc

import Int.ecb.logging.Level;

import Int.ecb.gesmescb.util.ComponentHelper;

/**
 * <P>The EDIFACTWriteBuffer helps the {@link EDIFACTWriter} to compose
 * EDIFACT segments and write them to a {@link java.io.Writer}.
 * Its main task is to help produce properly escaped output.
 * The <CODE>EDIFACTWriteBuffer</CODE> implements its own buffering scheme
 * such that no external {@link java.io.BufferedWriter} is needed.</P>
 *
 * <P>Choosing a reasonably large buffer reduces the effect
 * of i/o overhead, if writing is done directly to a {@link java.io.FileWriter},
 * not a {@link java.io.BufferedWriter}.</P>
 *
 * <P>The EDIFACT truncation rules for :, +, and ' are automatically
 * enforced. Also newline characters receive a special treatment.
 * Particularly, they are translated to the platform's newline sequence
 * and successive newlines are replaced by a single one.</P>
 *
 * <P>It is not permitted to write other control characters than
 * newlines \n and \r. Everything else will most likely mess
 * up the truncation rule compliance.</P>
 *
 * <P>The <CODE>EDIFACTWriteBuffer</CODE> is not internally synchronized.
 * External synchronisation is necessary if multiple threads want to
 * use the same <CODE>EDIFACTWriteBuffer</CODE>.</P>
 *
 * @author
 * <A HREF="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</A>
 *
 * @version @(#)$Id: EDIFACTWriteBuffer.java,v 1.1 2017/01/19 16:14:05 trainot Exp $
 */
class EDIFACTWriteBuffer {

	/* ======================= Constants ======================== */

	/**
	 * A sensible default buffer capacity.
	 */
	private static final int DEFAULT_CAPACITY = 2048;

	/**
	 * The string used for line termination in order to split the
	 * output nicely into lines.
	 * It is used in the method {@link #newLine}.
	 */
	private static final String LINE_BREAK
	= System.getProperty( "line.separator", "\n" );

	/** The length of {@link #LINE_BREAK}. */
	private static final int LINE_BREAK_LENGTH = LINE_BREAK.length();

	/* ========================= Fields ========================= */

	/**
	 * Where to write the EDIFACT output to.
	 */
	private final Writer writer;

	/** Pluses written but still waiting. Could be truncated. */
	private int pendingPlus;

	/** Colons written but still waiting. Could be truncated. */
	private int pendingColon;

	/** A last character ? has led to escaped status. */
	private boolean escaped;

	/** LineBreak written but still waiting until after possible truncation. */
	private boolean pendingLineBreak;

	/**
	 * A buffer to be used for writing.
	 * This is less expensive than using a BufferedWriter since
	 * no synchronisation overhead is incurred.
	 */
	private final char[] buffer;

	/**
	 * A second buffer to handle insertion of additional line
	 * breaks.
	 */
	private final char[] secondaryBuffer;

	/**
	 * The number of valid characters in the buffer.
	 * These are then buffer[0], ..., buffer[top-1].
	 */
	private int top;

	/**
	 * The maximum number of columns per line for automatic line breaking.
	 * If this number is <= 0, then no such line breaking will take place.
	 */
	private final int maxColumn;

	/**
	 * Column counter for automatic line breaking.
	 */
	private int columnCount;

	/* ====================== Constructors ====================== */

	/**
	 * Construct an <CODE>EDIFACTWriteBuffer</CODE>
	 * with specified minimum initial capacity
	 * from a {@link java.io.Writer}.
	 *
	 * @param writer where to write to.
	 * @param capacity the minimal initial buffer capacity to be used.
	 * If necessary, the <CODE>EDIFACTWriteBuffer</CODE>
	 * may transparently increase
	 * its used internal buffer capacity during operation.
	 * Nevertheless, a reasonably large initial buffer capacity should be
	 * chosen in order to avoid loss of performance.
	 * @param componentHelper of the client EDIFACTWriter.
	 *
	 * @throws NullPointerException if <CODE>writer</CODE> is null.
	 */
	public EDIFACTWriteBuffer (
		final Writer writer,
		final int capacity,
		final ComponentHelper componentHelper
	)
		throws NullPointerException
	{
		if ( ( this.writer = writer ) == null ) {
			throw new NullPointerException(
				"Construction of a "
				+ getClass().getName()
				+ " from a null pointer is inhibited."
			);
		}

		// Initialize the limit and counter for automatic line breaking.
		columnCount = 0;
		{
			final String maxCol
			= componentHelper.getPreference( "maxColumn", "0" );
			int maxC = 0;	// temporary
			try {
				maxC = Integer.parseInt( maxCol );
			}
			catch ( final NumberFormatException e ) {
				componentHelper.logger.log(
					Level.SEVERE,
					componentHelper.logName,
					"EDIFACTWriteBuffer()",
					"The EDIFACTWriter preference parameter maxColumn "
					+ "should be an integer number but is \"" + maxCol + "\"."
				);
			}
			maxColumn = maxC;
		}

		buffer = new char[ capacity > 0 ? capacity : DEFAULT_CAPACITY ];

		secondaryBuffer = maxColumn > 0
			? new char[
				buffer.length
				+ ( buffer.length/maxColumn + 1 ) * LINE_BREAK_LENGTH
			] : null;

		// Initialize the buffer
		top = 0;

		// Initialize status for truncation rules.
		pendingPlus = pendingColon = 0;
		pendingLineBreak = escaped = false;
	}

	/**
	 * Construct an <CODE>EDIFACTWriteBuffer</CODE>
	 * with default initial capacity
	 * from a {@link java.io.Writer}.
	 *
	 * @param writer where to write to.
	 * @param componentHelper of the client EDIFACTWriter.
	 *
	 * @throws NullPointerException if <CODE>writer</CODE> is null.
	 */
	public EDIFACTWriteBuffer (
		final Writer writer, final ComponentHelper componentHelper
	)
		throws NullPointerException
	{
		this( writer, DEFAULT_CAPACITY, componentHelper );
	}

	/* ===================== Public Methods ===================== */

	/**
	 * Write a String with all necessary characters escaped.
	 *
	 * @param string is the String to write.
	 *
	 * @throws IOException if the underlying Writer does so.
	 */
	public void writeEscaped ( final String string )
		throws IOException
	{
		writeEscaped( string, 0, string.length() );
	}

	/**
	 * Write part of a String with all necessary characters escaped.
	 *
	 * @param string is the String a substring of which is to be written.
	 * @param start is the first position of the substring within string.
	 * This parameter must be in the range of 0 to string.length().
	 * @param string is the position after the last of the substring.
	 * This parameter must be in the range of start to string.length().
	 *
	 * @throws IOException if the underlying Writer does so.
	 *
	 * @deprecated Will be dropped due to lack of use.
	 */
	public void writeEscaped (
		final String string, final int start, final int end
	)
		throws IOException
	{
		for ( int i = start; i < end; ++i ) {
			final char c = string.charAt( i );
			writeEscaped( c );
		}
	}

	/**
	 * Write a char escaped if necessary.
	 *
	 * @param c is the character to write.
	 *
	 * @throws IOException if the underlying Writer does so.
	 */
	public void writeEscaped ( final char c )
		throws IOException
	{
		switch ( c ) {
			case ':':
			case '+':
			case '?':
			case '\'':
				writeRaw ( '?' );
			default:
				writeRaw ( c );
		}
	}

	/**
	 * Write a line break.
	 *
	 * @throws IOException if the underlying Writer does so.
	 */
	public void newLine ()
		throws IOException
	{
		writeRaw( '\n' );	// will be translated via pendingLineBreak.
	}

	/**
	 * Write a String as it is. Particularly do not escape.
	 *
	 * @param string is the String to write.
	 *
	 * @throws IOException if the underlying Writer does so.
	 */
	public void writeRaw ( final String string )
		throws IOException
	{
		final int stringLen = string.length();

		for ( int i = 0; i < stringLen; ++i ) {
			writeRaw( string.charAt( i ) );
		}
	}

	/**
	 * Write an integer.
	 * This is faster than transforming the integer to a string and
	 * then writing the string since no characters must be escaped
	 * in numbers.
	 *
	 * @param n is the integer to write.
	 *
	 * @throws IOException if the underlying Writer does so.
	 */
	public void writeRaw ( final int n )
		throws IOException
	{
		writeRaw( String.valueOf( n ) );
	}

	/**
	 * Write a char as it is. Particularly do not escape.
	 * Truncation rules for : and + are still enforced.
	 * Truncation rules for leading 0 and trailing SPC are semantically
	 * determined by the type numeric/alphanumeric and not implemented here.
	 *
	 * @param c is the character to write.
	 *
	 * @throws IOException if the underlying Writer does so.
	 */
	public void writeRaw ( final char c )
		throws IOException
	{
		if ( escaped ) {
			writeUntruncated( c );
			escaped = false;
		} else {
			switch ( c ) {
				case '\n':
				case '\r': pendingLineBreak = true; break;
				case ':': ++pendingColon; break;
				case '+': pendingColon = 0; ++pendingPlus; break;
				case '\'': pendingColon = pendingPlus = 0;
					writeUntruncated( c ); break;
				case '?': escaped = true;
				default: writeUntruncated( c );
			}
		}
	}

	/**
	 * <P>Write a char as it is even without respect to truncation rules.</P>
	 *
	 * <P>Except for internal use in the <CODE>EDIFACTWriteBuffer</CODE>,
	 * this method should only be used to write the <CODE>UNA</CODE>-segment,
	 * which would be crippled by truncation.</P>
	 *
	 * @param string to write.
	 *
	 * @throws IOException if the underlying Writer does so.
	 */
	public void writeUntruncated ( final String string )
		throws IOException
	{
		final int stringLen = string.length();

		for ( int i = 0; i < stringLen; ++i ) {
			writeUntruncated( string.charAt( i ) );
		}
	}

	/**
	 * <P>Close down.
	 * This method should always be called at the end of writing.
	 * Internal buffers are flushed.
	 * However, the underlying Writer is neither flushed nor closed. </P>
	 *
	 * @throws IOException if the underlying Writer does so.
	 */
	public void close ()
		throws IOException
	{
		flush();
		writer.write( LINE_BREAK );
	}

	/* =================== Protected Methods ==================== */

	/**
	 * <P>Flush the internal buffer.</P>
	 *
	 * @throws IOException if the underlying Writer does so.
	 */
	protected void flush ()
		throws IOException
	{
		if ( maxColumn > 0 ) {
			int secondaryTop = 0;
			for ( int i = 0; i < top; ++i ) {
				final char cc = buffer[i];
				switch ( cc ) {
					case '\n':
					case '\r':
						columnCount = 0;
						break;
					default:
						if ( columnCount >= maxColumn ) {
							for ( int j = 0; j < LINE_BREAK_LENGTH; ++j ) {
								secondaryBuffer[secondaryTop++]
								= LINE_BREAK.charAt( j );
							}
							columnCount = 0;
						}
						++columnCount;
				}
				secondaryBuffer[secondaryTop++] = cc;
			}
			writer.write( secondaryBuffer, 0, secondaryTop );
		} else {
			writer.write( buffer, 0, top );
		}
		top = 0;
	}

	/* ==================== Private Methods ===================== */

	/** Writes first pending : and + and finally the argument. */
	private void writeUntruncated ( final char c )
		throws IOException
	{
		while ( pendingPlus > 0 ) {
			writeToBuf( '+' );
			--pendingPlus;
		}
		while ( pendingColon > 0 ) {
			writeToBuf( ':' );
			--pendingColon;
		}
		if ( pendingLineBreak ) {
			for ( int i = 0; i < LINE_BREAK_LENGTH; ++i ) {
				writeToBuf( LINE_BREAK.charAt( i ) );
			}
			pendingLineBreak = false;
		}
		writeToBuf( c );
	}

	/** Encapsulates writing to buffer. */
	private void writeToBuf ( final char c )
		throws IOException
	{
		if ( top >= buffer.length ) flush();
		buffer[top++] = c;
	}
}
