// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.edifact;

import java.lang.String;
import java.io.Writer;
import java.io.StringWriter;
import java.io.IOException;

import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;

import Int.ecb.gesmescb.edifact.EDIFACTWriteBuffer;
import Int.ecb.gesmescb.util.ComponentHelper;

/* ***************************** Class ****************************** */

/**
 * Performs component test of the {@link EDIFACTWriteBuffer}.
 *
 * @author
 * <A HREF="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</A>
 *
 * @version @(#)$Id: EDIFACTWriteBufferTest.java,v 1.1 2017/01/19 16:14:05 trainot Exp $
 */
public class EDIFACTWriteBufferTest extends TestCase {

	/* ======================= Constants ======================== */

	private static final String NL
	= System.getProperty( "line.separator", "\n" );

	/* ====================== Constructors ====================== */

	public EDIFACTWriteBufferTest( String name ) {
		super( name );
	}

	/* ================== Protected Methods ===================== */

	/**
	 * Sets up the test fixture.
	 */
	protected void setUp () {
		// Set preference for EDIFACTWriteBufferTest, not EDIFACTWriteBuffer
		// because the latter is constructed to use preferences from the
		// former.
		System.setProperty(
			EDIFACTWriteBufferTest.class.getName() + ".maxColumn",
			"80"
		);
	}

	/* ===================== Public Methods ===================== */

	/**
	 * Assembles and returns a test suite for all
	 * the test methods of this test case.
	 *
	 * @return a non-null test suite
	 */
	public static Test suite () {
		final TestSuite suite = new TestSuite( EDIFACTWriteBufferTest.class );
		return suite;
	}

	/**
	 * @param arg the array of argument {@link String}s
	 */
	public static void main ( String[] arg ) {
		String[] testCaseName = {
			EDIFACTWriteBufferTest.class.getName()
		};
		junit.textui.TestRunner.main( testCaseName );
	}

	/* ---------------------- Test Methods ---------------------- */

	/**
	 * Test basically correct operation.
	 *
	 * @throws IOException if the StringReader has a problem.
	 */
	public void testBasicWrite ()
		throws IOException
	{
		final StringWriter writer = new StringWriter();
		final EDIFACTWriteBuffer buffer = createEDIFACTWriteBuffer( writer );

		buffer.writeRaw( 'a' );
		buffer.writeRaw( "bc" );
		buffer.writeRaw( -12 );
		buffer.writeEscaped( 'a' );
		buffer.writeEscaped( "bc" );
		buffer.writeEscaped( ':' );
		buffer.writeEscaped( '+' );
		buffer.writeEscaped( '\'' );
		buffer.writeEscaped( '?' );
		buffer.writeEscaped( ":+'?" );
		buffer.newLine();

		buffer.close();
		writer.close();

		assertEquals(
			"abc-12abc?:?+?'???:?+?'??" + NL,
			writer.toString()
		);
	}

	/**
	 * Test on truncation rules.
	 *
	 * @throws IOException if the StringReader has a problem.
	 */
	public void testTruncationRules ()
		throws IOException
	{
		final StringWriter writer = new StringWriter();
		final EDIFACTWriteBuffer buffer = createEDIFACTWriteBuffer( writer );

		buffer.writeRaw( ":+'a+b:+:c::+d+++'" );
		buffer.writeRaw( "e?:+f?+'" );
		buffer.writeRaw( "g:?++?+'" );
		buffer.writeRaw( "h:??++??+'++?'" );
		buffer.writeRaw( "i:?:+?++'" );

		buffer.close();
		writer.close();

		assertEquals(
			"'a+b+:c+d'"
			+ "e?:+f?+'"
			+ "g:?++?+'"
			+ "h:??++??'++?'"
			+ "i:?:+?+'" + NL,
			writer.toString()
		);
	}

	/**
	 * Test truncation rules in presence of newline.
	 *
	 * @throws IOException if the StringReader has a problem.
	 */
	public void testNewlineTruncationRules ()
		throws IOException
	{
		final StringWriter writer = new StringWriter();
		final EDIFACTWriteBuffer buffer = createEDIFACTWriteBuffer( writer );

		buffer.writeRaw( "a" + NL + ":" + NL + "+" + NL + "'" );

		buffer.close();
		writer.close();

		assertEquals(
			"a" + NL + "'" + NL,
			writer.toString()
		);
	}

	/**
	 * Test truncation rules in presence of newline.
	 *
	 * @throws IOException if the StringReader has a problem.
	 */
	public void testLineBreaking ()
		throws IOException
	{
		final StringWriter writer = new StringWriter();
		final EDIFACTWriteBuffer buffer = createEDIFACTWriteBuffer( writer );

		buffer.writeRaw(
			"aaaa\n"
			+ "bbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbb"
			+ "bbbbbbbbbbbbbbbbbbbbbbbbbbbbbbcccccccccc"
		);

		buffer.close();
		writer.close();

		assertEquals(
			"aaaa" + NL
			+ "bbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbb"
			+ "bbbbbbbbbbbbbbbbbbbbbbbbbbbbbb" + NL
			+ "cccccccccc" + NL,
			writer.toString()
		);
	}

	/* ==================== Private Methods ===================== */

	private EDIFACTWriteBuffer createEDIFACTWriteBuffer (
		final Writer writer
	) {
		return new EDIFACTWriteBuffer(
			writer, new ComponentHelper( "", this )
		);
	}
}
