// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2002, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.edifact;

import java.util.List;
import java.util.ArrayList;
import java.util.Iterator;

import java.io.File;
import java.io.IOException;

import Int.ecb.logging.Level;

import Int.ecb.gesmescb.time.RegularInterval;
import Int.ecb.gesmescb.time.GregorianRegularInterval;

import Int.ecb.gesmescb.event.SerialWriter;
import Int.ecb.gesmescb.event.SerialReader;

import Int.ecb.gesmescb.*;

/* ***************************** Class ****************************** */

/**
 * <p>Translates {@link Int.ecb.gesmescb.GESMESCBObject}s
 * into EDIFACT text with awkward EDIFACT multiplicity count
 * restrictions observed.</p>
 *
 * <center><table border="3" cellpadding="3">
 * <caption>
 * <a name="prefs">Available {@link Int.ecb.prefs preference} settings</a>
 * </caption>
 * <tr>
 * <th>preference name</th>
 * <th>type</th>
 * <th>meaning</th>
 * <th>default</th>
 * </tr>
 * <tr><td><code>destination</code></td><td>string</td>
 *	<td>output file path</td></tr>
 * <tr><td>format</td><td>string</td>
 * <td>EDIFACT interchange format</td><td><code>GESMES/TS</code></td></tr>
 * <tr><td>maxColumn</td><td>integer</td>
 * <td>break lines after column #, 0 for none (default)</td></tr>
 * <tr><td><code>postponeAttributes</code></td><td>boolean</td>
 *	<td>Whether to allow occurrence of {@link Attribute}s not only after
 * all {@link TimeSeries} in a message; in this case, <code>Attribute</code>s
 * must be temporarily stored and written only at message end.</td>
 *	<td><code>true</code></td></tr>
 * </table></center>
 *
 * @see Int.ecb.gesmescb.edifact.EDIFACTUnlimitedWriter
 *
 * @author
 * <A HREF="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</A>
 *
 * @version @(#)$Id: EDIFACTWriter.java,v 1.1 2017/01/19 16:14:05 trainot Exp $
 */
public class EDIFACTWriter extends EDIFACTUnlimitedWriter {

	/* ========================= Fields ========================= */

	/** Whether to temporarily store Attributes. */
	private final boolean postponeAttributes;

	/** Temporarily stores Attributes, if postponeAttributes is true. */
	private final File attributeTempFile;

	/** Temporarily stores Attributes, if postponeAttributes is true. */
	private SerialWriter attributeTempWriter;

	/**
	 * Attribute segment counter.
	 */
	private int arrCounter, ideCounter;

	/* ====================== Constructors ====================== */

	/**
	 * Construct a <CODE>EDIFACTWriter</CODE>
	 * with given writing destination file.
	 *
	 * @param name used for logging and preferences.
	 * @param filename specifies in the local system's manner
	 * where to write to.
	 *
	 * @throws IOException if the file cannot be opened for writing.
	 */
	public EDIFACTWriter ( final String name, final String filename )
		throws IOException
	{
		super( name, filename );
		postponeAttributes = !"false".equalsIgnoreCase(
			componentHelper.getPreference( "postponeAttributes", "true" )
		);
		if ( postponeAttributes ) {
			attributeTempFile = File.createTempFile(
				"attributes", ".gesmes.serial"
			);
			attributeTempFile.deleteOnExit();
		} else {
			attributeTempFile = null;
		}
		arrCounter = ideCounter = 0;
	}

	/**
	 * Construct a <CODE>EDIFACTWriter</CODE>
	 * with writing destination file according to preferences.
	 *
	 * @param name used for logging and preferences.
	 *
	 * @throws IOException if the file cannot be opened for writing.
	 * @throws NullPointerException if the preference variable "destination"
	 * is undefined.
	 */
	public EDIFACTWriter ( final String name )
		throws IOException
	{
		this( name, null );
	}

	/* ===================== Public Methods ===================== */

	public void close () {
		if ( attributeTempFile != null ) {
			attributeTempFile.delete();
		}
		super.close();
	}

	/* ------------------------ Message ------------------------- */

	public void handle ( final MessageBegin messageBegin ) {
		if ( postponeAttributes ) {
			try {
				attributeTempWriter = new SerialWriter(
					"attributeTempWriter",
					attributeTempFile.getPath()
				);
			}
			catch ( final IOException ioe ) {
				componentHelper.logger.log(
					Level.SEVERE,
					componentHelper.logName,
					"handle(MessageBegin)",
					"Output problem when opening temporary"
					+ " file to write attributes.",
					ioe
				);
			}
		}
		super.handle( messageBegin );
	}

	public void handle ( final MessageEnd messageEnd ) {
		if ( postponeAttributes ) {
			attributeTempWriter.close();
			try {
				final SerialReader attributeTempReader = new SerialReader(
					"attributeTempReader",
					attributeTempFile.getPath()
				);
				while ( attributeTempReader.hasNext() ) {
					final Attribute attribute
					= (Attribute) attributeTempReader.next();
					writeAttribute( attribute );
				}
				attributeTempReader.close();
			} catch ( final IOException ioe ) {
				componentHelper.logger.log(
					Level.SEVERE,
					componentHelper.logName,
					"handle(MessageEnd)",
					"Output problem when opening temporary"
					+ " file to read attributes.",
					ioe
				);
			}
		}
		super.handle( messageEnd );
	}

	/* ---------------------- Time Series ----------------------- */

	/**
	 * <p>Write a <code>TimeSeries</code>, split if necessary.</p>
	 *
	 * <p>The particular problem with <code>TimeSeries</code> is that they
	 * must to be split into more than one ARR segment when the
	 * size exceeds 9999.
	 * This also implies to compute starting and ending dates for
	 * the segments created.</p>
	 *
	 * <p>The current implementation ignores the limit of 999000
	 * <code>ARR</code> segments in the main <code>ARR</code> section
	 * of a message.</p>
	 */
	public void handle ( final TimeSeries timeSeries ) {
		final Iterator observations = timeSeries.observationIterator();
		if ( observations.hasNext() ) {
			final int MAX_OBS = 9999;
			final RegularInterval interval = timeSeries.getInterval();
			for (
				final Iterator intervals = interval.iterator();
				observations.hasNext() && intervals.hasNext();
			) {
				final List observationList = new ArrayList( MAX_OBS );
				RegularInterval first = null;
				int emptyObservations = 0;
				while (
					observations.hasNext()
					&&
					intervals.hasNext()
					&&
					observationList.size() + emptyObservations < MAX_OBS
				) {
					final Observation observation
					= (Observation) observations.next();
					final RegularInterval period
					= (RegularInterval) intervals.next();
					if ( observation.isEmpty() ) {
						if ( first != null ) {
							++emptyObservations;
						}
					} else {
						if ( first == null ) first = period;
						while ( emptyObservations > 0 ) {
							observationList.add(
								Observation.EMPTY_OBSERVATION
							);
							--emptyObservations;
						}
						observationList.add( observation );
					}
				}
				if ( first != null ) {
					final RegularInterval intervalSection
					= new GregorianRegularInterval(
						first.getFrequency(),
						first.getBegin(),
						observationList.size()
					);
					final TimeSeries seriesSection = new TimeSeries(
						timeSeries.getKey(),
						intervalSection,
						observationList
					);
					super.handle( seriesSection );
				}
			}
		} else {
			super.handle( timeSeries );
		}
	}

	/* ----------------------- Attribute ------------------------ */

	/**
	 * <p>Write an <code>Attribute</code>.</p>
	 *
	 * <p>Attributes can re-use the <code>ARR</code> segments
	 * for the object key and also the <code>REL</code> segment for
	 * the attachment scope.
	 * However, re-use must not exceed 9999 times.
	 * After that number, the <code>ARR</code> or <code>REL</code>
	 * segment must be issued again.</p>
	 *
	 * <p>The current implementation ignores the limit of 9999
	 * <code>REL</code> segments in the <code>FNS</code> section
	 * of a message.</p>
	 */
	public void handle ( final Attribute attribute ) {
		if ( postponeAttributes ) {
			attributeTempWriter.handle( attribute );
		} else {
			writeAttribute( attribute );
		}
	}

	/* ==================== Private Methods ===================== */

	private void writeAttribute ( final Attribute attribute ) {
		final int MAX_ARR = 9999;
		final int MAX_IDE = 9999;
		final String entryAttachmentScope = previousAttachmentScope;
		final String entryKey = previousKey;
		final String entryPeriod = previousPeriod;
		final String entryPeriodFormat = previousPeriodFormat;

		if ( haveNotWrittenFNS ) {
			arrCounter = ideCounter = 0;
		}

		super.handle( attribute );

		if ( entryAttachmentScope.equals( previousAttachmentScope ) ) {
			if (
				entryKey.equals( previousKey )
				&&
				entryPeriod.equals( previousPeriod )
				&&
				entryPeriodFormat.equals( previousPeriodFormat )
			) {
				if ( ++ideCounter >= MAX_IDE ) {
					previousKey = "<new>";	// some impossible value
				}
			} else {
				if ( ++arrCounter >= MAX_ARR ) {
					previousAttachmentScope = "<new>";	// some impossible value
				}
				ideCounter = 1;
			}
		} else {
			arrCounter = ideCounter = 1;
		}
	}
}
