// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2002, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.edifact;

import java.util.*;
import java.text.*;
import java.io.*;

import junit.framework.Test;
import junit.framework.TestSuite;

import Int.ecb.gesmescb.*;

/* ***************************** Class ****************************** */

/**
 * Tests the {@link EDIFACTWriter}.
 *
 * @author
 * <A HREF="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</A>
 *
 * @version @(#)$Id: EDIFACTWriterTest.java,v 1.1 2017/01/19 16:14:05 trainot Exp $
 */
public class EDIFACTWriterTest extends EDIFACTUnlimitedWriterTest {

	/* ====================== Constructors ====================== */

	public EDIFACTWriterTest( final String name ) {
		super( name );
	}

	/* ================== Protected Methods ===================== */

	protected EDIFACTUnlimitedWriter createEDIFACTWriter (
		final String name, final String filename
	) throws IOException {
		return new EDIFACTWriter( name, filename );
	}
	/* ===================== Public Methods ===================== */

	/**
	 * Assembles and returns a test suite for all
	 * the test methods of this test case.
	 *
	 * @return a non-null test suite
	 */
	public static Test suite () {
		return new TestSuite( EDIFACTWriterTest.class );
	}

	/**
	 * @param arg the array of argument {@link String}s.
	 */
	public static void main ( final String[] arg ) {
		final String[] testCaseName = {
			EDIFACTWriterTest.class.getName()
		};
		junit.textui.TestRunner.main( testCaseName );
	}

	/* ---------------------- Test Methods ---------------------- */

	public void testSplitSeries () throws Exception {
		final DateFormat format = new SimpleDateFormat( "yyyyMMdd" );
		final int MILLIS_PER_DAY = 24 * 60 * 60 * 1000;

		final Date d1 = format.parse( "19000101" );
		final Date d2 = format.parse( "20000101" );
		final int daysBetween
		= (int) ( ( d2.getTime() - d1.getTime() ) / MILLIS_PER_DAY );
		assertEquals( 36524, daysBetween );
		final Calendar calendar = Calendar.getInstance();

		final TimeSeries series = new TimeSeries(
			"HOG.SOB", "1900010119991231", "711",
			Collections.nCopies(
				daysBetween,
				new Observation( "13.6", "A", "", "" )
			)
		);
		final Object[] input = {
			INTERCHANGE_BEGIN,
			MESSAGE_BEGIN_2,
			series,
			MESSAGE_END_2,
			INTERCHANGE_END
		};

		final StringBuffer expected = new StringBuffer();
		expected.append(
			INTERCHANGE_BEGIN_OUT
			+ MESSAGE_BEGIN_2_OUT
		);
		int segments = 13;

		calendar.setTime( d1 );
		while ( true ) {
			final Date start = calendar.getTime();
			if ( start.compareTo( d2 ) >= 0 ) break;

			final long begin = start.getTime();
			final long end = d2.getTime();

			int days = (int) ( ( end - begin ) / MILLIS_PER_DAY );
			if ( days > 9999 ) {
				days = 9999;
			}

			expected.append( "ARR++HOG:SOB:" );
			expected.append( format.format( start ) );
			calendar.add( Calendar.DAY_OF_YEAR, days - 1 );
			expected.append( format.format( calendar.getTime() ) );
			expected.append( ":711" );
			calendar.add( Calendar.DAY_OF_YEAR, 1 );

			expected.append( ":13.6:A" );
			expected.append( stringMult( "+13.6:A", days - 1 ) );
			expected.append( "'\n" );

			++segments;
		}

		expected.append( "UNT+" );
		expected.append( ++segments );
		expected.append( "+MREF000002'\n" + "UNZ+1+IREF000215'\n" );

		outputTest( input, expected.toString() );
	}

	public void testSplitSparseSeries () throws Exception {
		final int daysBetween = 36524;

		final List observations = new ArrayList(
			Collections.nCopies(
				daysBetween,
				Observation.EMPTY_OBSERVATION
			)
		);
		final Observation first = new Observation( "13.6", "A", "", "" );
		final Observation middle1 = new Observation( "13.7", "A", "", "" );
		final Observation middle2 = new Observation( "13.8", "A", "", "" );
		final Observation last = new Observation( "13.9", "A", "", "" );
		observations.set( 0, first );
		observations.set( 4, middle1 );
		observations.set( 9999, middle2 );
		observations.set( daysBetween - 1, last );

		final TimeSeries series = new TimeSeries(
			"HOG.SOB", "1900010119991231", "711", observations
		);
		final Object[] input = {
			INTERCHANGE_BEGIN,
			MESSAGE_BEGIN_2,
			series,
			MESSAGE_END_2,
			INTERCHANGE_END
		};

		outputTest(
			input,
			INTERCHANGE_BEGIN_OUT
			+ MESSAGE_BEGIN_2_OUT
			+ "ARR++HOG:SOB:1900010119000105:711:13.6:A++++13.7:A'\n"
			+ "ARR++HOG:SOB:19270519:102:13.8:A'\n"
			+ "ARR++HOG:SOB:19991231:102:13.9:A'\n"
			+ "UNT+17+MREF000002'\n"
			+ "UNZ+1+IREF000215'\n"
		);
	}

	public void testSplitAttributesARR () throws Exception {
		final List objects = new ArrayList(
			Collections.nCopies(
				20004,
				new Attribute(
					"HOG.FOB", "", "", "XCUS", "A"
				)
			)
		);
		objects.set( 0, INTERCHANGE_BEGIN );
		objects.set( 1, MESSAGE_BEGIN_2 );
		objects.set( 20002, MESSAGE_END_2 );
		objects.set( 20003, INTERCHANGE_END );

		outputTest(
			objects.toArray(),
			INTERCHANGE_BEGIN_OUT
			+ MESSAGE_BEGIN_2_OUT
			+ "FNS+Attributes:10'\n"
			+ "REL+Z01+4'\n"
			+ "ARR+2+HOG:FOB'\n"
			+ stringMult(
				"IDE+Z10+XCUS'\n"
				+ "CDV+A'\n",
				9999
			)
			+ "ARR+2+HOG:FOB'\n"
			+ stringMult(
				"IDE+Z10+XCUS'\n"
				+ "CDV+A'\n",
				9999
			)
			+ "ARR+2+HOG:FOB'\n"
			+ "IDE+Z10+XCUS'"
			+ "CDV+A'\n"
			+ "IDE+Z10+XCUS'"
			+ "CDV+A'\n"
			+ "UNT+40019+MREF000002'\n"
			+ "UNZ+1+IREF000215'\n"
		);
	}

	public void testSplitAttributesREL () throws Exception {
		final Attribute[] attributes = {
			new Attribute(
				"HOG.FOB", "", "", "XCUS", "A"
			),
			new Attribute(
				"HOB.FOG", "", "", "XCUS", "B"
			)
		};

		final Object[] objects = new Object[20004];
		objects[0] = INTERCHANGE_BEGIN;
		objects[1] = MESSAGE_BEGIN_2;
		for ( int i = 2; i < 20002; ++i ) {
			objects[i] = attributes[i%2];
		}
		objects[20002] = MESSAGE_END_2;
		objects[20003] = INTERCHANGE_END;

		outputTest(
			objects,
			INTERCHANGE_BEGIN_OUT
			+ MESSAGE_BEGIN_2_OUT
			+ "FNS+Attributes:10'\n"
			+ "REL+Z01+4'\n"
			+ stringMult(
				"ARR+2+HOG:FOB'\n"
				+ "IDE+Z10+XCUS'\n"
				+ "CDV+A'\n"
				+ "ARR+2+HOB:FOG'\n"
				+ "IDE+Z10+XCUS'\n"
				+ "CDV+B'\n",
				9999/2
			)
			+ "ARR+2+HOG:FOB'\n"
			+ "IDE+Z10+XCUS'\n"
			+ "CDV+A'\n"
			+ "REL+Z01+4'\n"
			+ stringMult(
				"ARR+2+HOB:FOG'\n"
				+ "IDE+Z10+XCUS'\n"
				+ "CDV+B'\n"
				+ "ARR+2+HOG:FOB'\n"
				+ "IDE+Z10+XCUS'\n"
				+ "CDV+A'\n",
				9999/2
			)
			+ "ARR+2+HOB:FOG'\n"
			+ "IDE+Z10+XCUS'\n"
			+ "CDV+B'\n"
			+ "REL+Z01+4'\n"
			+ "ARR+2+HOG:FOB'\n"
			+ "IDE+Z10+XCUS'"
			+ "CDV+A'\n"
			+ "ARR+2+HOB:FOG'\n"
			+ "IDE+Z10+XCUS'"
			+ "CDV+B'\n"
			+ "UNT+60018+MREF000002'\n"
			+ "UNZ+1+IREF000215'\n"
		);
	}

	public void testSplitAttributesARRandREL () throws Exception {
		final Attribute[] attributes = {
			new Attribute(
				"HOG.FOB", "1999", "602", "COMMENT", "A"
			),
			new Attribute(
				"HOG.FOB", "2000", "602", "COMMENT", "A"
			)
		};

		final Object[] objects = new Object[20004];
		objects[0] = INTERCHANGE_BEGIN;
		objects[1] = MESSAGE_BEGIN_2;
		for ( int i = 2; i < 10002; ++i ) {
			objects[i] = attributes[0];
		}
		for ( int i = 10002; i < 20002; ++i ) {
			objects[i] = attributes[i%2];
		}
		objects[20002] = MESSAGE_END_2;
		objects[20003] = INTERCHANGE_END;

		outputTest(
			objects,
			INTERCHANGE_BEGIN_OUT
			+ MESSAGE_BEGIN_2_OUT
			+ "FNS+Attributes:10'\n"
			+ "REL+Z01+5'\n"
			+ "ARR+4+HOG:FOB:1999:602'\n"
			+ stringMult(
				"IDE+Z11+COMMENT'\n"
				+ "FTX+ACM+++A'\n",
				9999
			)
			+ "ARR+4+HOG:FOB:1999:602'\n"
			+ "IDE+Z11+COMMENT'\n"
			+ "FTX+ACM+++A'\n"
			+ "IDE+Z11+COMMENT'\n"
			+ "FTX+ACM+++A'\n"
			+ "ARR+4+HOG:FOB:2000:602'\n"
			+ "IDE+Z11+COMMENT'\n"
			+ "FTX+ACM+++A'\n"
			+ stringMult(
				"ARR+4+HOG:FOB:1999:602'\n"
				+ "IDE+Z11+COMMENT'\n"
				+ "FTX+ACM+++A'\n"
				+ "ARR+4+HOG:FOB:2000:602'\n"
				+ "IDE+Z11+COMMENT'\n"
				+ "FTX+ACM+++A'\n",
				9997/2
			)
			+ "REL+Z01+5'\n"
			+ "ARR+4+HOG:FOB:1999:602'\n"
			+ "IDE+Z11+COMMENT'"
			+ "FTX+ACM+++A'\n"
			+ "ARR+4+HOG:FOB:2000:602'\n"
			+ "IDE+Z11+COMMENT'"
			+ "FTX+ACM+++A'\n"
			+ "UNT+50018+MREF000002'\n"
			+ "UNZ+1+IREF000215'\n"
		);
	}

	/**
	 * Test whether empty {@link Observation}s are cut
	 * from the {@link TimeSeries}.
	 *
	 * @throws IOException if opening the EDIFACTUnlimitedWriter yields a problem.
	 */
	public void testDataOrder () throws IOException {
		final Object[] input = new Object[]{
			INTERCHANGE_BEGIN,
			MESSAGE_BEGIN_1,
			CL_OBJ,
			CL_CONF,
			SECRECY,
			INSANITY,
			SIS_FOOB,
			MESSAGE_END_1,
			MESSAGE_BEGIN_2,
			HOX_S,
			STUS,
			SLUS,
			SOX_S,
			XCUS,
			APPETITE,
			MARIA,
			BANANA,
			SOX_N,
			COMMENT,
			MESSAGE_END_2,
			MESSAGE_BEGIN_3,
			STUS_2,
			SLUS_2,
			HOX_S_2,
			SOX_S_2,
			XCUS_2,
			APPETITE_2,
			S_2,
			MARIA_2,
			COMMENT_2,
			SOX_2,
			MESSAGE_END_3,
			INTERCHANGE_END
		};
		outputTest( input, testOutput );
	}


	/**
	 * Test whether empty {@link Observation}s are cut
	 * from the {@link TimeSeries}.
	 *
	 * @throws IOException if opening the EDIFACTUnlimitedWriter yields a problem.
	 */
	public void testTrimEmpty () throws IOException {
		final Object[] input = new Object[]{
			INTERCHANGE_BEGIN,
			MESSAGE_BEGIN_2,
			new TimeSeries(
				"PAU.SAD", "20022006", "702",
				new Observation[]{
					new Observation( "", "", "", "" ),
					new Observation( "42.3", "S", "N", "" ),
					new Observation( "", "", "", "" ),
					new Observation( "-", "M", "C", "" ),
					new Observation( "", "", "", "" )
				}
			),
			MESSAGE_END_2,
			INTERCHANGE_END
		};
		final String output
		= INTERCHANGE_BEGIN_OUT
			+ MESSAGE_BEGIN_2_OUT
			+ "ARR++PAU:SAD:20032005:702:42.3:S:N++-:M:C'\n"
			+ "UNT+15+MREF000002'\n"
			+ "UNZ+1+IREF000215'\n";
		outputTest( input, output );
	}
}
