// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.edifact;

import java.lang.String;
import java.lang.Number;
import java.lang.IllegalArgumentException;

import java.util.Map;
import java.util.HashMap;
import java.util.Collections;
import java.util.Date;

import java.text.Format;
import java.text.FieldPosition;
import java.text.ParsePosition;
import java.text.ParseException;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.text.NumberFormat;
import java.text.DecimalFormat;

import Int.ecb.gesmescb.time.RegularInterval;
import Int.ecb.gesmescb.time.GregorianRegularInterval;

/* ***************************** Class ****************************** */

/**
 * Formats {@link RegularInterval}s.
 *
 * @author
 * <A HREF="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</A>
 *
 * @version @(#)$Id: RegularIntervalFormat.java,v 1.1 2017/01/19 16:14:05 trainot Exp $
 */
public class RegularIntervalFormat extends Format {

	/* ======================= Constants ======================== */

	/**
	 *  serialVersionUID generated by eclipse.
	 */
	private static final long serialVersionUID = 3178097581345189715L;

	/** Field number for the period start specification within the range field.
	 * @see #PERIOD_RANGE_FIELD
	 * @see FieldPosition */
	public static final int START_PERIOD_FIELD = 0x1;

	/** Field number for the period end specification within the range field.
	 * @see #PERIOD_RANGE_FIELD
	 * @see FieldPosition */
	public static final int END_PERIOD_FIELD = 0x2;

	/** Field number for the colon separating range specification
	 * from format code.
	 * @see FieldPosition */
	public static final int COLON_FIELD = 0x4;

	/** Field number for the EDIFACT time format code.
	 * This is the three-letter code following the colon.
	 * @see FieldPosition */
	public static final int PERIOD_FORMAT_FIELD = 0x8;

	/** Field number for the period range start and (if present) end text.
	 * This is everything up to the colon, i.e. start plus end specification.
	 * @see #START_PERIOD_FIELD
	 * @see #END_PERIOD_FIELD
	 * @see FieldPosition */
	public static final int PERIOD_RANGE_FIELD
	= START_PERIOD_FIELD | END_PERIOD_FIELD;

	/** Field number for the whole time period text.
	 * @see FieldPosition */
	public static final int TIME_PERIOD_FIELD
	= PERIOD_RANGE_FIELD | COLON_FIELD | PERIOD_FORMAT_FIELD;

	/* ========================= Fields ========================= */

	private transient Map SIMPLE_FORMATS_BY_FREQUENCY;
	private transient Map RANGE_FORMATS_BY_FREQUENCY;
	private transient Map FORMATS_BY_PERIOD_FORMAT;

	/* ====================== Constructors ====================== */

	/**
	 * Construct a RegularIntervalFormat.
	 */
	public RegularIntervalFormat () {}

	/* ===================== Public Methods ===================== */

	/* --------------------- Format Methods --------------------- */

	public final StringBuffer format (
		final Object object,
		final StringBuffer toAppendTo,
		final FieldPosition position
	) {
		if ( object instanceof RegularInterval ) {
			return format( (RegularInterval) object, toAppendTo, position );
		} else {
			throw new IllegalArgumentException(
				"The object is not a RegularInterval:" + object
			);
		}
	}

	/**
	 * Compute the full EDIFACT representation
	 * of a <CODE>RegularInterval</CODE>.
	 * This consists of its date and format contents
	 * separated by a colon.
	 *
	 * @return <CODE>toAppendTo</CODE>
	 *
	 * @throws IllegalArgumentException if the <CODE>interval</CODE>
	 * carries invalid data.
	 */
	public StringBuffer format (
		final RegularInterval interval,
		final StringBuffer toAppendTo,
		final FieldPosition position
	) {
		final boolean intervalIsRange = interval.getSize() != 1;
		final int size = interval.getSize();
		
		final FixedFormat format = (FixedFormat) (
			intervalIsRange
				? getRangeFormatByFrequency( interval.getFrequency() )
				: getSimpleFormatByFrequency( interval.getFrequency() )
			);
		if ( format == null ) {
			throw new IllegalArgumentException(
				"There is no " + (
					intervalIsRange ? "range" : "single period"
				) + " format for frequency " + interval.getFrequency()
				+ ". Size is: " + size + ". From " + interval.getBegin() 
				+ " to " + interval.getEnd() + "."
			);
		}

		updateFieldPosition( position, 0, toAppendTo.length() );
		format.format( interval, toAppendTo );
		updateFieldPosition( position, 1, toAppendTo.length() );
		if ( format.isRange() ) {
			format.format(
				interval.computeBasePeriod( interval.getSize() - 1 ),
				toAppendTo
			);
		}
		updateFieldPosition( position, 2, toAppendTo.length() );
		toAppendTo.append( ':' );
		updateFieldPosition( position, 3, toAppendTo.length() );
		toAppendTo.append( format.getPeriodFormat() );
		updateFieldPosition( position, 4, toAppendTo.length() );
		return toAppendTo;
	}

	/* --------------------- Parse Methods ---------------------- */

	/**
	 * <P>Parse a <CODE>RegularInterval</CODE> from EDIFACT format into an
	 * <CODE>Object</CODE>.</P>
	 *
	 * <P>The method returns <CODE>null</CODE> if parsing fails.
	 * Note that {@link Format#parseObject(String,ParsePosition)}
	 * does not throw {@link ParseException}s.</P>
	 *
	 * @param text to parse.
	 * @param position tracker for parsing process.
	 *
	 * @return the represented <CODE>RegularInterval</CODE>
	 * as <CODE>Object</CODE>.
	 *
	 * @see #parse(String,ParsePosition)
	 */
	public final Object parseObject(
		final String text,
		final ParsePosition position
	) {
		return parse( text, position );
	}

	/**
	 * Parse a <CODE>RegularInterval</CODE> from EDIFACT format.
	 * A time period in EDIFACT format looks like 199801010101:203.
	 *
	 * @param text to parse.
	 * @param position tracker for parsing process.
	 *
	 * @return the new <CODE>RegularInterval</CODE>
	 * or <CODE>null</CODE> if parsing fails.
	 */
	public RegularInterval parse (
		final String text,
		final ParsePosition position
	) {
		final int startIndex = position.getIndex();
		final int colonIndex = text.indexOf( ':', startIndex );
		final int endIndex = colonIndex + 4;
		final int textLength = text.length();
		if ( endIndex > textLength ) {
			position.setErrorIndex( textLength );
			return null;
		}

		if ( colonIndex < 0 ) {
			final int errorPosition = textLength;
			if ( position.getErrorIndex() < 0 ) position.setErrorIndex(
				errorPosition
			);
			return null;
		}

		final String
			period = text.substring( startIndex, colonIndex ),
			periodFormat = text.substring( colonIndex + 1, endIndex );
		final FixedFormat format
		= (FixedFormat) getFormatByPeriodFormat( periodFormat );

		if ( format == null ) {
			position.setErrorIndex( colonIndex + 1 );
		} else {
			RegularInterval interval = format.parse( period, position );
			if ( interval != null && format.isRange() ) {
				final RegularInterval lastInterval
				= format.parse( period, position );
				if ( lastInterval != null 
					&&
					lastInterval.getEnd().after(interval.getBegin())
				) {
					interval = new GregorianRegularInterval(
						format.getFrequency(),
						interval.getBegin(),
						lastInterval.getEnd()
					);
				} else {
					interval = null;
				}
			}
			if ( interval != null ) {
				if ( position.getIndex() == colonIndex ) {
					position.setIndex( endIndex );
					return interval;
				} else {
					if ( position.getErrorIndex() < 0 ) {
						position.setErrorIndex( position.getIndex() );
					}
				}
			}
		}
		position.setIndex( startIndex );
		return null;
	}

	/**
	 * Parse a <CODE>RegularInterval</CODE> from EDIFACT format.
	 * Parsing starts at position 0.
	 * @see #parse(String,ParsePosition)
	 */
	public final RegularInterval parse(
		final String text
	) throws ParseException {
		final ParsePosition position = new ParsePosition( 0 );
		final RegularInterval interval = parse( text, position );
		if ( position.getErrorIndex() >= 0 ) {
			throw new ParseException(
				"Cannot parse \"" + text + "\" to a time interval.",
				position.getErrorIndex()
			);
		}
		if ( position.getIndex() < text.length() ) {
			throw new ParseException(
				"Unparsed text after EDIFACT time period string\""
				+ text + "\".",
				position.getIndex()
			);
		}
		return interval;
	}

	/**
	 * Parse a <CODE>RegularInterval</CODE> from separated EDIFACT format.
	 * A time period in EDIFACT format consists of
	 * a period like 199801010101 and a time format code like 203.
	 *
	 * @param period to parse.
	 * @param periodFormat code to use for parsing.
	 *
	 * @return the represented <CODE>RegularInterval</CODE>.
	 *
	 * @throws ParseException if parsing fails
	 */
	public RegularInterval parse(
		final String period,
		final String periodFormat
	) throws ParseException {
		return parse( period + ":" + periodFormat );
	}

	/* ==================== Private Methods ===================== */

	/** Compute begin and end index in <CODE>position</CODE>.
	 * The algorithm relies on well-chosen bit patterns for
	 * the field identifier constants like {@link TIME_PERIOD_FIELD}.
	 * There must be one bit for each portion in the output string,
	 * the combinations then choose the whole field.
	 *
	 * @param position to be updated.
	 * @param currentFieldNumber starting from 0 for first portion.
	 * @param index in the <CODE>StringBuffer</CODE>.
	 */
	private void updateFieldPosition (
		final FieldPosition position,
		final int currentFieldNumber,
		final int index
	) {
		if ( position != null ) {
			final int block = position.getField();

			// If first nonzero bit, then set begin
			if (
				( block & ( -1 >>> 31 - currentFieldNumber ) ) != 0
				&& (
					currentFieldNumber == 0
					||
					( block & ( -1 >>> 32 - currentFieldNumber ) ) == 0
				)
			) {
				position.setBeginIndex( index );
			}
			// If after last nonzero bit, then set end
			if (
				( block & ( -1 << currentFieldNumber ) ) == 0
				&& (
					currentFieldNumber != 0
					&&
					( block & ( -1 << currentFieldNumber - 1 ) ) != 0
				)
			) {
				position.setEndIndex( index );
			}
		}
	}

	/* -------------- Transient Support Functions --------------- */

	private void initMaps () {
		final FixedFormat[] formats = {
			// time stamp format;
			new PatternFormat(
				RegularInterval.MINUTELY, false, "203", "yyyyMMddHHmm"
			),
			// range format; parser will evaluated the same pattern twice 
			new PatternFormat(
				RegularInterval.MINUTELY, true, "719", "yyyyMMddHHmm"
			),
			new PatternFormat(
				RegularInterval.DAILY, false, "102", "yyyyMMdd"
			),
			new PatternFormat(
				RegularInterval.DAILY, true, "711", "yyyyMMdd"
			),
			new YearIndexFormat( RegularInterval.WEEKLY, false, "616", 2 ),
			new YearIndexFormat( RegularInterval.WEEKLY, true, "716", 2 ),
			new YearIndexFormat( RegularInterval.MONTHLY, false, "610", 2 ),
			new YearIndexFormat( RegularInterval.MONTHLY, true, "710", 2 ),
			new YearIndexFormat( RegularInterval.QUARTERLY, false, "608", 1 ),
			new YearIndexFormat( RegularInterval.QUARTERLY, true, "708", 1 ),
			new YearIndexFormat( RegularInterval.SEMIANNUAL, false, "604", 1 ),
			new YearIndexFormat( RegularInterval.SEMIANNUAL, true, "704", 1 ),
			new YearIndexFormat( RegularInterval.ANNUAL, false, "602", 0 ),
			new YearIndexFormat( RegularInterval.ANNUAL, true, "702", 0 )
		};
		final Map
			singleByFrequency = new HashMap( formats.length * 2 ),
			rangeByFrequency = new HashMap( formats.length * 2 ),
			byCode = new HashMap( formats.length * 2 );
		for ( int i = 0; i < formats.length; ++i ) {
			final FixedFormat format = formats[i];
			if ( format.isRange() ) {
				rangeByFrequency.put(
					new Integer( format.getFrequency() ), format
				);
			} else {
				singleByFrequency.put(
					new Integer( format.getFrequency() ), format
				);
			}
			byCode.put( format.getPeriodFormat(), format );
		}
		SIMPLE_FORMATS_BY_FREQUENCY = Collections.unmodifiableMap(
			singleByFrequency
		);
		RANGE_FORMATS_BY_FREQUENCY = Collections.unmodifiableMap(
			rangeByFrequency
		);
		FORMATS_BY_PERIOD_FORMAT = Collections.unmodifiableMap( byCode );
	}

	/** Get simple format by frequency. */
	private synchronized FixedFormat getSimpleFormatByFrequency (
		final int frequency
	) {
		if ( SIMPLE_FORMATS_BY_FREQUENCY == null ) {
			initMaps();
		}
		return (FixedFormat) SIMPLE_FORMATS_BY_FREQUENCY.get(
			new Integer( frequency )
		);
	}

	/** Get range format by frequency. */
	private synchronized FixedFormat getRangeFormatByFrequency (
		final int frequency
	) {
		if ( RANGE_FORMATS_BY_FREQUENCY == null ) {
			initMaps();
		}
		return (FixedFormat) RANGE_FORMATS_BY_FREQUENCY.get(
			new Integer( frequency )
		);
	}

	/** Get format by period format code. */
	private synchronized FixedFormat getFormatByPeriodFormat (
		final String periodFormat
	) {
		if ( FORMATS_BY_PERIOD_FORMAT == null ) {
			initMaps();
		}
		return (FixedFormat) FORMATS_BY_PERIOD_FORMAT.get( periodFormat );
	}
}

/* ***************************** Class ****************************** */

/** Encapsulates <CODE>frequency</CODE>-specific parameters
 * and format method.
 */
abstract class FixedFormat {

	/* ========================= Fields ========================= */

	/** @see #getFrequency */
	private final int frequency;

	/** @see #isRange */
	private final boolean isRange;

	/** EDIFACT time time format code.
	 * @see #getPeriodFormat */
	private final String periodFormat;

	/* ====================== Constructors ====================== */

	/** Constructs with the main characteristics. */
	public FixedFormat (
		final int frequency,
		final boolean isRange,
		final String periodFormat
	) {
		this.frequency = frequency;
		this.isRange = isRange;
		this.periodFormat = periodFormat;
	}

	/* ===================== Public Methods ===================== */

	/** The frequency.
	 * @see RegularInterval#getFrequency */
	public int getFrequency () {
		return frequency;
	}

	/** Whether the formatter expects/produces a range represented by
	 * first and last base period, or only a single base period. */
	public boolean isRange () {
		return isRange;
	}

	/**
	 * Get EDIFACT time format code handled by this <CODE>FixedFormat</CODE>.
	 */
	public String getPeriodFormat () {
		return periodFormat;
	}

	/* -------------------- Abstract Methods -------------------- */

	/**
	 * Format the first base period of a <CODE>RegularInterval</CODE>
	 * into a <CODE>StringBuffer</CODE>.
	 * To format a longer range, the operation must
	 * be used twice.
	 *
	 * @see RegularIntervalFormat#format
	 */
	public abstract void format (
		RegularInterval interval,
		StringBuffer toAppendTo
	);

	/**
	 * Parse a base period portion of a String.
	 * To parse a longer range, the operation must
	 * be used twice.
	 * Of course, it does not include the colon and time period code.
	 *
	 * @return the new <CODE>RegularInterval</CODE>
	 * or <CODE>null</CODE> if parsing fails.
	 */
	public abstract RegularInterval parse (
		String text,
		ParsePosition position
	);
}

/* ***************************** Class ****************************** */

/** Converts {@link RegularInterval}s
 * according to {@link SimpleDateFormat} patterns. */
class PatternFormat extends FixedFormat {

	/* ========================= Fields ========================= */

	private final String pattern;

	/* ====================== Constructors ====================== */

	public PatternFormat (
		final int frequency,
		final boolean isRange,
		final String periodFormat,
		final String pattern
	) {
		super( frequency, isRange, periodFormat );
		this.pattern = pattern;
	}

	/* ===================== Public Methods ===================== */

	public void format (
		final RegularInterval interval,
		final StringBuffer toAppendTo
	) {
		createFormat().format(
			interval.getBegin(),
			toAppendTo,
			new FieldPosition( 0 )
		);
	}

	public RegularInterval parse (
		final String text,
		final ParsePosition position
	) {
		final int endIndex = position.getIndex() + pattern.length();
		if ( endIndex > text.length() ) {
			position.setErrorIndex( text.length() );
			return null;
		}

		// Unfortunately, SimpleDateFormat might not stop parsing
		// at pattern end. So need to cut the rest.
		// See Java Bug Parade bug nr. 4227284.
		final String period = text.substring( 0, endIndex );
		final Date begin = createFormat().parse( period, position );
		if ( begin == null ) return null;
		return new GregorianRegularInterval( getFrequency(), begin, 1 );
	}

	/* ==================== Private Methods ===================== */

	/** Create a strict <CODE>DateFormat</CODE> according to the
	 * used <CODE>pattern</CODE>.
	 *
	 * The class <CODE>SimpleDateFormat</CODE> is not thread safe
	 * and therefore a <CODE>DateFormat</CODE> cannot be reused
	 * without external synchronisation.</P>
	 */
	private DateFormat createFormat () {
		final DateFormat format = new SimpleDateFormat( pattern );
		format.setLenient( false );
		return format;
	}
}

/* ***************************** Class ****************************** */

/** Converts {@link RegularInterval}s
 * using year and index therein.
 *
 * @see RegularInterval#getFirstPeriodYear
 * @see RegularInterval#getFirstPeriodIndex
 */
class YearIndexFormat extends FixedFormat {

	/* ========================= Fields ========================= */

	/** Number of digits for the index in year. */
	private final int indexDigits;

	/* ====================== Constructors ====================== */

	public YearIndexFormat (
		final int frequency,
		final boolean isRange,
		final String periodFormat,
		final int indexDigits
	) {
		super( frequency, isRange, periodFormat );
		this.indexDigits = indexDigits;
	}

	/* ===================== Public Methods ===================== */

	public void format (
		final RegularInterval interval,
		final StringBuffer toAppendTo
	) {
		createFormat().format(
			interval.getFirstPeriodYear() * pow10( indexDigits )
			+ ( indexDigits > 0 ? interval.getFirstPeriodIndex() + 1 : 0 ),
			toAppendTo,
			new FieldPosition( 0 )
		);
	}

	public RegularInterval parse (
		final String text,
		final ParsePosition position
	) {
		final int startIndex = position.getIndex();
		final int endIndex = startIndex + 4 + indexDigits;
		if ( endIndex > text.length() ) {
			position.setErrorIndex( text.length() );
			return null;
		}

		// Compare remark above about end of parsing.
		final String period = text.substring( 0, endIndex );

		final Number yearIndexInt
		= (Number) createFormat().parse( period, position );
		if ( yearIndexInt != null ) {
			if ( position.getIndex() == endIndex ) {
				final int yearIndex = yearIndexInt.intValue();
				final int pow = pow10( indexDigits );
				final int year = yearIndex / pow;
				final int index = indexDigits > 0 ? yearIndex % pow - 1 : 0;
				try {
					final RegularInterval interval
					= new GregorianRegularInterval(
						getFrequency(), year, index, 1
					);
					position.setIndex( endIndex );
					return interval;
				}
				catch ( final IllegalArgumentException e ) {
					position.setErrorIndex( endIndex );
				}
			} else {
				position.setErrorIndex( position.getIndex() );
			}
		}
		position.setIndex( startIndex );
		return null;
	}

	/* ==================== Private Methods ===================== */

	/**
	 * Create a <CODE>NumberFormat</CODE> for the year portion.
	 *
	 * @see http://developer.java.sun.com/developer/bugParade/bugs/4101500.html
	 */
	private NumberFormat createFormat () {
		final NumberFormat format = new DecimalFormat( "####" );
		format.setParseIntegerOnly( true );
		format.setMinimumIntegerDigits( 4 + indexDigits );
		format.setMaximumIntegerDigits( 4 + indexDigits );
		return format;
	}

	/**
	 * Compute power of 10.
	 *
	 * @param exponent greater or equal zero.
	 */
	private static int pow10 ( int exponent ) {
		int value = 1;
		while ( exponent-- > 0 ) value *= 10;
		return value;
	}
}
