// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.edifact;

import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.text.ParsePosition;
import java.text.FieldPosition;
import java.text.ParseException;

import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;

import Int.ecb.gesmescb.time.RegularInterval;
import Int.ecb.gesmescb.time.GregorianRegularInterval;

import Int.ecb.gesmescb.edifact.RegularIntervalFormat;

/* ***************************** Class ****************************** */

/**
 * <p>Tests the implementation of {@link RegularIntervalFormat}. </p>
 *
 * @author
 * <A HREF="mailto:Sami.Airo@ECB.int">Sami Airo</A>
 * @author
 * <A HREF="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</A>
 *
 * @version @(#)$Id: RegularIntervalFormatTest.java,v 1.1 2017/01/19 16:14:05 trainot Exp $
 */
public class RegularIntervalFormatTest extends TestCase {

	/* ====================== Constructors ====================== */

	/**
	 * Constructs a RegularIntervalFormatTest with the specified name.
	 *
	 * @param name Test case name
	 */
	public RegularIntervalFormatTest ( final String name ) {
		super( name );
	}

	/* ===================== Public Methods ===================== */

	/**
	 * Assembles and returns a test suite for all
	 * the test methods of this test case.
	 *
	 * @return a non-null test suite
	 */
	public static Test suite () {
		return new TestSuite( RegularIntervalFormatTest.class );
	}

	/**
	 * Runs the test case.
	 */
	public static void main ( final String args[] ) {
		String[] testCaseName = { RegularIntervalFormatTest.class.getName() };
		junit.textui.TestRunner.main( testCaseName );
	}

	/* --------------------- Test Methods --------------------- */

	public void testDirectDate () throws ParseException {
		final DateFormat dateFormat = new SimpleDateFormat(
			"yyyy-MM-dd'T'HH:mm:ss"
		);
		final RegularIntervalFormat format = new RegularIntervalFormat();
		final Object[][] testPairs = {
			{
				new GregorianRegularInterval(
					RegularInterval.WEEKLY,
					dateFormat.parse( "2001-01-01T00:00:00" ),
					52
				),
				"200101200152:716"
			},
			{
				new GregorianRegularInterval(
					RegularInterval.WEEKLY,
					dateFormat.parse( "2001-12-31T00:00:00" ),
					52
				),
				"200201200252:716"
			},
			{
				new GregorianRegularInterval(
					RegularInterval.WEEKLY,
					dateFormat.parse( "2003-12-29T00:00:00" ),
					53
				),
				"200401200453:716"
			},
			{
				new GregorianRegularInterval(
					RegularInterval.WEEKLY,
					dateFormat.parse( "2005-01-03T00:00:00" ),
					52
				),
				"200501200552:716"
			},
			{
				new GregorianRegularInterval(
					RegularInterval.WEEKLY,
					dateFormat.parse( "2008-12-29T00:00:00" ),
					53
				),
				"200901200953:716"
			},
			{
				new GregorianRegularInterval(
					RegularInterval.WEEKLY,
					dateFormat.parse( "2008-12-22T00:00:00" ),
					55
				),
				"200852201001:716"
			}
		};

		for ( int i = 0; i < testPairs.length; ++i ) {
			final RegularInterval interval = (RegularInterval) testPairs[i][0];
			final String text = (String) testPairs[i][1];
			assertEquals(
				"format[" + i + "]",
				text, format.format( interval )
			);
			assertEquals(
				"parse[" + i + "]",
				interval, format.parse( text )
			);
		}
	}

	public void testDirect () throws ParseException {
		final RegularIntervalFormat format = new RegularIntervalFormat();

		final String text = "2000020120000313:711";
		final RegularInterval interval = new GregorianRegularInterval(
			RegularInterval.DAILY, 2000, 31, 29 + 13
		);

		assertEquals( text, format.format( interval ) );
		assertEquals( interval, format.parse( text ) );
	}

	public void testParseSeparated () throws ParseException {
		final RegularIntervalFormat format = new RegularIntervalFormat();
		assertEquals(
			new GregorianRegularInterval( RegularInterval.ANNUAL, 1998, 0, 2 ),
			format.parse( "19981999", "702" )
		);
		try {
			format.parse( "1998199x", "702" );
			fail( "No throw on x" );
		}
		catch ( final ParseException e ) {}
		try {
			format.parse( "19981999", ":702" );
			fail( "No throw on :702" );
		}
		catch ( final ParseException e ) {}
		try {
			format.parse( "19981999", "703" );
			fail( "No throw on 703" );
		}
		catch ( final ParseException e ) {}
		try {
			format.parse( "19981999", "7023" );
			fail( "No throw on 7023" );
		}
		catch ( final ParseException e ) {}
		try {
			format.parse( "199819990", "702" );
			fail( "No throw on 19990" );
		}
		catch ( final ParseException e ) {}
	}

	public void testFieldPosition () {
		final RegularIntervalFormat format = new RegularIntervalFormat();
		final RegularInterval
			interval1 = new GregorianRegularInterval(
				RegularInterval.MONTHLY, 1998, 4, 2
			),
			interval2 = new GregorianRegularInterval(
				RegularInterval.WEEKLY, 1998, 0, 1
			);
		final FieldPosition
			startField = new FieldPosition( RegularIntervalFormat.START_PERIOD_FIELD ),
			endField = new FieldPosition( RegularIntervalFormat.END_PERIOD_FIELD ),
			rangeField = new FieldPosition( RegularIntervalFormat.PERIOD_RANGE_FIELD ),
			colonField = new FieldPosition( RegularIntervalFormat.COLON_FIELD ),
			formatField = new FieldPosition( RegularIntervalFormat.PERIOD_FORMAT_FIELD ),
			allField = new FieldPosition( RegularIntervalFormat.TIME_PERIOD_FIELD );
		final StringBuffer buffer = new StringBuffer();

		format.format( interval1, buffer, startField ); buffer.setLength( 0 );
		format.format( interval1, buffer, endField ); buffer.setLength( 0 );
		format.format( interval1, buffer, rangeField ); buffer.setLength( 0 );
		format.format( interval1, buffer, colonField ); buffer.setLength( 0 );
		format.format( interval1, buffer, formatField ); buffer.setLength( 0 );
		format.format( interval1, buffer, allField ); buffer.setLength( 0 );

		assertEquals( "start1start", 0, startField.getBeginIndex() );
		assertEquals( "start1end", 6, startField.getEndIndex() );
		assertEquals( "end1start", 6, endField.getBeginIndex() );
		assertEquals( "end1end", 12, endField.getEndIndex() );
		assertEquals( "range1start", 0, rangeField.getBeginIndex() );
		assertEquals( "range1end", 12, rangeField.getEndIndex() );
		assertEquals( "colon1start", 12, colonField.getBeginIndex() );
		assertEquals( "colon1end", 13, colonField.getEndIndex() );
		assertEquals( "format1start", 13, formatField.getBeginIndex() );
		assertEquals( "format1end", 16, formatField.getEndIndex() );
		assertEquals( "all1start", 0, allField.getBeginIndex() );
		assertEquals( "all1end", 16, allField.getEndIndex() );

		format.format( interval2, buffer, startField ); buffer.setLength( 0 );
		format.format( interval2, buffer, endField ); buffer.setLength( 0 );
		format.format( interval2, buffer, rangeField ); buffer.setLength( 0 );
		format.format( interval2, buffer, colonField ); buffer.setLength( 0 );
		format.format( interval2, buffer, formatField ); buffer.setLength( 0 );
		format.format( interval2, buffer, allField ); buffer.setLength( 0 );

		assertEquals( "start2start", 0, startField.getBeginIndex() );
		assertEquals( "start2end", 6, startField.getEndIndex() );
		assertEquals( "end2start", 6, endField.getBeginIndex() );
		assertEquals( "end2end", 6, endField.getEndIndex() );
		assertEquals( "range2start", 0, rangeField.getBeginIndex() );
		assertEquals( "range2end", 6, rangeField.getEndIndex() );
		assertEquals( "colon2start", 6, colonField.getBeginIndex() );
		assertEquals( "colon2end", 7, colonField.getEndIndex() );
		assertEquals( "format2start", 7, formatField.getBeginIndex() );
		assertEquals( "format2end", 10, formatField.getEndIndex() );
		assertEquals( "all2start", 0, allField.getBeginIndex() );
		assertEquals( "all2end", 10, allField.getEndIndex() );
	}

	public void testMinutelySingle () {
		validFormatTest( "199801010101:203" );
		validFormatTest( "199801012359:203" );

		invalidFormatTest( "199801010101:719" );
		invalidFormatTest( "199801010161:203" );
		invalidFormatTest( "199813010101:203" );
		invalidFormatTest( "199801330101:203" );
		invalidFormatTest( "199801012601:203" );
		invalidFormatTest( "1998010123:203" );
		invalidFormatTest( "19July310123:102" );
		invalidFormatTest( "1May19310123:102" );
	}

	public void testMinutelyRange () {
		validFormatTest( "199801010101200812312359:719" );
		validFormatTest( "199801012359200801010000:719" );

		invalidFormatTest( "199801010161200801010000:719" );
		invalidFormatTest( "199813010101200801010000:719" );
		invalidFormatTest( "199801330101200801010000:719" );
		invalidFormatTest( "199801012601200801010000:719" );
		invalidFormatTest( "199801010000199801012601:719" );
	}

	public void testDailySingle () {
		validFormatTest( "19980101:102" );

		validFormatTest( "19960329:102" );
		validFormatTest( "19950329:102" );
		validFormatTest( "19730329:102" );
		validFormatTest( "19920629:102" );

		validFormatTest( "19910121:102" );
		validFormatTest( "19910129:102" );
		validFormatTest( "19990113:102" );

		validFormatTest( "19821231:102" );

		invalidFormatTest( "19981301:102" );
		invalidFormatTest( "19981301:102" );
		invalidFormatTest( "19981232:102" );
		invalidFormatTest( "1998123:102" );
		invalidFormatTest( "19July31:102" );
		invalidFormatTest( "1May1931:102" );
	}

	public void testDailyRange () {
		validFormatTest( "1998010119980102:711" );

		validFormatTest( "1991012919990113:711" );
		validFormatTest( "1982123119890420:711" );
		validFormatTest( "1959100119870213:711" );
		validFormatTest( "1987021620010502:711" );
		validFormatTest( "1959100120010502:711" );

		invalidFormatTest( "1998130119980102:711" );
		invalidFormatTest( "1998123419980102:711" );
		invalidFormatTest( "199812349980102:711" );
		invalidFormatTest( "1998120119982302:711" );
		invalidFormatTest( "1998120119981245:711" );
		invalidFormatTest( "199812011998:711" );
		invalidFormatTest( "1998Ja0119980102:711" );
		invalidFormatTest( "-998010119980102:711" );
		invalidFormatTest( "1998010119980102DD:711" );
	}

	public void testWeeklySingle () {
		validFormatTest( "199801:616" );
		validFormatTest( "199853:616" );

		invalidFormatTest( "199953:616" );
		invalidFormatTest( "19995:616" );
		invalidFormatTest( "199957:616" );
		invalidFormatTest( "999576:616" );
		invalidFormatTest( "19995A:616" );
		invalidFormatTest( "199900:616" );
	}

	public void testWeeklyRange () {
		validFormatTest( "199801199802:716" );
		validFormatTest( "199853199901:716" );
		validFormatTest( "197501200511:716" );
		validFormatTest( "200201200601:716" );

		invalidFormatTest( "200101200199:716" );
		invalidFormatTest( "200101200100:716" );
		invalidFormatTest( "2001012001DD:716" );
		invalidFormatTest( "-00101200102:716" );
		invalidFormatTest( "199853199900:716" );
	}

	public void testMonthlySingle () {
		validFormatTest( "199801:610" );

		invalidFormatTest( "200100:610" );
		invalidFormatTest( "200113:610" );
		invalidFormatTest( "200154:610" );
		invalidFormatTest( "2001DD:610" );
	}

	public void testMonthlyRange () {
		validFormatTest( "199801199802:710" );

		invalidFormatTest( "200101200199:710" );
		invalidFormatTest( "200101200100:710" );
		invalidFormatTest( "2001012001DD:710" );
		invalidFormatTest( "-00101200102:710" );
	}

	public void testQuarterlySingle () {
		validFormatTest( "19981:608" );

		validFormatTest( "20001:608" );
		validFormatTest( "19992:608" );
		validFormatTest( "30004:608" );

		invalidFormatTest( "20010:608" );
		invalidFormatTest( "20015:608" );
		invalidFormatTest( "2001D:608" );
		invalidFormatTest( " 11:608" );
		invalidFormatTest( "11:608" );
	}

	public void testQuarterlyRange () {
		validFormatTest( "1998119982:708" );

		invalidFormatTest( "2001120019:708" );
		invalidFormatTest( "2001120010:708" );
		invalidFormatTest( "200112001D:708" );
		invalidFormatTest( "-001120012:708" );
	}

	public void testSemiannualSingle () {
		validFormatTest( "19981:604" );

		invalidFormatTest( "19980:604" );
		invalidFormatTest( "19985:604" );
		invalidFormatTest( "19981A:604" );
	}

	public void testSemiannualRange () {
		validFormatTest( "1998120012:704" );

		invalidFormatTest( "2001120019:704" );
		invalidFormatTest( "2001120010:704" );
		invalidFormatTest( "200112001D:704" );
		invalidFormatTest( "-001120012:704" );
	}

	public void testAnnualSingle () {
		validFormatTest( "1998:602" );

		invalidFormatTest( "199:602" );
		invalidFormatTest( "-199:602" );
		invalidFormatTest( "99999:602" );
		invalidFormatTest( "AAAA:602" );
		invalidFormatTest( "AAAA: " );
		invalidFormatTest( ":602" );
		invalidFormatTest( "19.99:602" );
		invalidFormatTest( "1999:6.2" );
		// This is not invalid, only parsing stops before the "."
		// in line with the java.text.Format specification.
		// invalidFormatTest( "1999:602.4" );
	}

	public void testAnnualRange () {
		validFormatTest( "19982001:702" );

		invalidFormatTest( "1982001:702" );
		invalidFormatTest( "19992000A:702" );
		invalidFormatTest( "12001:702" );
		invalidFormatTest( "-1982001:702" );
	}

	/* ==================== Private Methods ===================== */

	private void validFormatTest ( final String text ) {
		final RegularIntervalFormat format = new RegularIntervalFormat();
		final ParsePosition position = new ParsePosition( 0 );
		final RegularInterval interval = format.parse( text, position );
		assertEquals( "index on end", text.length(), position.getIndex() );
		assertEquals( "no errorIndex", -1, position.getErrorIndex() );
		assertEquals( text, format.format( interval ) );
	}

	private void invalidFormatTest ( final String text ) {
		final RegularIntervalFormat format = new RegularIntervalFormat();
		final ParsePosition position = new ParsePosition( 0 );
		final RegularInterval interval = format.parse( text, position );
		assertSame(
			"Invalid format \"" + text + "\" must yield null.",
			null,
			interval
		);
		assertTrue(
			"Failed to set errorIndex for invalid format \"" + text + "\".",
			position.getErrorIndex() >= 0
		);
		assertEquals(
			"Consumed positions in invalid format \"" + text + "\".",
			0,
			position.getIndex()
		);
	}
}
