// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2002, 2004 European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.error;

import java.io.Writer;
import java.io.FileWriter;
import java.io.FileReader;
import java.io.BufferedReader;

import java.util.Collection;
import java.util.Iterator;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Properties;
import java.util.Enumeration;

import Int.ecb.logging.Level;
import Int.ecb.logging.Handler;

import Int.ecb.gesmescb.*;

import Int.ecb.gesmescb.util.ComponentHelper;
import Int.ecb.gesmescb.event.GESMESCBEventListener;
import Int.ecb.gesmescb.event.StructureListener;
import Int.ecb.gesmescb.event.DataListener;
import Int.ecb.gesmescb.event.GESMESCBEventListenerAdapter;

import Int.ecb.gesmescb.error.ErrorWatcher;

/* ***************************** Class ****************************** */

/**
 * <P>Provides facilities to evaluate the results of a checking/loading run.</P>
 *
 * <CENTER><TABLE BORDER=3 CELLPADDING=3>
 * <CAPTION>
 * <A NAME="prefs">Available {@link Int.ecb.prefs preference} settings</A>
 * </CAPTION>
 * <TR>
 * <TH>preference name</TH>
 * <TH>preference type</TH>
 * <TH>preference value</TH>
 * </TR>
 * <TR>
 * <TD>acknowledgementFile</TD><TD>string</TD>
 * <TD>acknowledgement output file path; default none</TD>
 * </TR>
 * <TR>
 * <TD>headerFile</TD><TD>string</TD><TD>input header file; default none</TD>
 * </TR>
 * <TR>
 * <TD>reportUnexpected</TD><TD>integer</TD>
 * <TD>Maximum number of unexpected objects reported;
 * default 0 - no reporting</TD>
 * </TR>
 * <TR>
 * <TD>mimeCompliant</TD><TD>boolean</TD>
 * <TD>Tells the application to adhere to MIME rules
 * and also forward traditionally
 * unavailable fields from the header file;
 * default false, meaning traditional format</TD>
 * </TR>
 * </TABLE></CENTER>
 *
 * @author
 * <A HREF="mailto:Sami.Airo@ECB.int">Sami Airo</A>
 * @author
 * <A HREF="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</A>
 *
 * @version @(#)$Id: AcknowledgementManager.java,v 1.1 2017/01/19 16:14:05 trainot Exp $
 */
public class AcknowledgementManager {

	/* ======================= Constants ======================== */

	private static final String UNEXPECTED_REPORT_FIELD = "Unexpected objects";

	/* ========================= Fields ========================= */

	private final ComponentHelper componentHelper;

	private final GESMESCBEnvelopeListener envelopeListener;
	private final GESMESCBAnyObjectListener unexpectedListener;
	private final ErrorEventListener errorListener;
	private final ErrorHandler errorHandler;

	/* ====================== Constructors ====================== */

	public AcknowledgementManager () {
		componentHelper = new ComponentHelper( "", this );
		errorListener = new ErrorEventListener();
		int reportUnexpected = 0;
		final String mru = componentHelper.getPreference(
			"reportUnexpected", "0"
		);
		try {
			reportUnexpected = Integer.parseInt( mru );
		}
		catch ( final NumberFormatException nfe ) {
			componentHelper.logger.log(
				Level.WARNING,
				componentHelper.logName,
				"AcknowledgementManager()",
				"Preference reportUnexpected has non-numeric value \""
				+ mru + "\". Using default " + reportUnexpected
				+ " instead.",
				nfe
			);
		}
		envelopeListener = new GESMESCBEnvelopeListener( 2 );
		unexpectedListener = new GESMESCBAnyObjectListener( reportUnexpected );
		errorHandler = new ErrorHandler();
	}

	/* ===================== Public Methods ===================== */

	/** Provide a listener to extract envelope data for the acknowledgement.
	 */
	public GESMESCBEventListener getEnvelopeListener () {
		return envelopeListener;
	}

	/** Provide a listener to watch out for unexpected objects.
	 * It should be connected to the unexpected source
	 * if a {@link Int.ecb.gesmescb.sieve.GESMESCBEventSieve} is used.
	 *
	 * @see Int.ecb.gesmescb.sieve.GESMESCBEventSieve#getRejectedSource()
	 */
	public GESMESCBEventListener getUnexpectedListener () {
		return unexpectedListener;
	}

	/**
	 * Provide a listener to watch out for
	 * {@link Int.ecb.gesmescb.event.ErrorEvent}s.
	 */
	public GESMESCBEventListener getErrorListener () {
		return errorListener;
	}

	/** Provide a <CODE>Handler</CODE> to inspect log messages. */
	public Handler getErrorHandler () {
		return errorHandler;
	}

	public void close () {
		final Properties ackProps = new Properties();

		if (! readHeaderFile( ackProps ) ) return;
		setReturnValues( ackProps );

		final String newReceiver = ackProps.getProperty( "From", null );
		final String newSender = ackProps.getProperty( "To", null );
		if ( newReceiver != null ) {
			ackProps.setProperty( "To", newReceiver );
		} else {
			ackProps.remove( "To" );
		}
		if ( newSender != null ) {
			ackProps.setProperty( "From", newSender );
		} else {
			ackProps.remove( "From" );
		}

		writeAcknowledgementFile( ackProps );
	}

	/** <P>Evaluates the current state to form an exit code.</P>
	 *
	 * <P>The exit code is:</P>
	 * <DL>
	 * <DT>0</DT><DD>if no problems have been detected</DD>
	 * <DT>1</DT><DD>if unexpected objects were sieved out</DD>
	 * <DT>2</DT><DD>if a component, particularly the
	 * {@link Int.ecb.gesmescb.semanticchecker.SemanticChecker},
	 * encountered an {@link ErrorWatcher#ERROR}.</DD>
	 * <DT>other</DT><DD>in case of an internal failure</DD>
	 * </DL>
	 *
	 * <P>The method may be called multiple times.
	 * It does not affect the state of the
	 * <CODE>AcknowledgementManager</CODE>.</P>
	 */
	public int computeExitCode () {
		if ( errorListener.getSeverity() >= ErrorWatcher.FAILURE ) return 99;
		if ( errorListener.getSeverity() == ErrorWatcher.ERROR ) return 2;
		if ( errorHandler.getSeverity() >= ErrorWatcher.FAILURE ) return 98;
		if ( unexpectedListener.receivedGESMESCBObject() ) return 1;
		return	0;
	}

	/* =================== Protected Methods ==================== */

	/**
	 * Read the header file.
	 * Its name is given by the preference <CODE>headerFile</CODE>.
	 * If this preference is not set, no file is read.
	 *
	 * @param properties where to read the name-value-pairs to.
	 *
	 * @return whether the method was completed successfully.
	 */
	protected boolean readHeaderFile ( final Properties properties ) {
		final String headerFilename = componentHelper.getPreference(
			"headerFile", null
		);
		if ( headerFilename == null ) {
			componentHelper.logger.log(
				Level.CONFIG,
				componentHelper.logName,
				"close()",
				"No preference \"headerFile\" set."
			);
			return true;
		}

		try {
			final BufferedReader reader
			= new BufferedReader( new FileReader( headerFilename ) );

			String line;
			while ( ( line = reader.readLine() ) != null ) {
				if ( "".equals( line ) ) continue;
				properties.setProperty(
					getNamePortion( line ), getValuePortion( line )
				);
			}

			reader.close();
		}
		catch ( final Exception e ) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"close()",
				"Reading from header file raised exception",
				e
			);
			return false;
		}

		return true;
	}

	/**
	 * Modify the given properties to reflect possibly encountered problems.
	 *
	 * @param properties wherein the report variables are set.
	 */
	protected void setReturnValues ( final Properties properties ) {
		for (
			final Iterator envelopes = envelopeListener.iterator();
			envelopes.hasNext();
		) {
			final Object object = envelopes.next();
			if ( object instanceof MessageBegin ) {
				final MessageBegin messageBegin = (MessageBegin) object;
				properties.setProperty(
					"Subject", messageBegin.getDataSetId()
				);
				break;
			}
		}

		String procId = "E2";
		String returnCode = "E81";
		String text = "Serious problem with the checker";

		switch ( computeExitCode() ) {
			case 0:
				procId = "F1";
				returnCode = "F1";
				text = "Gesmes message successfully checked and loaded";
				break;
			case 1:
				procId = "F1";
				returnCode = "F2";
				text = "Gesmes message successfully checked and loaded"
					+ " but some unexpected objects found";
				break;
			case 2:
				procId = "E2";
				returnCode = "E21";
				text = "Checker found errors";
				break;
		}

		properties.setProperty( "Proc.Id", procId );
		properties.setProperty( "Rtrn.Code", returnCode );
		properties.setProperty( "Text", text );

		if (
			unexpectedListener.capacity() > 0
			&&
			unexpectedListener.receivedGESMESCBObject()
		) {
			final Iterator iterator = unexpectedListener.iterator();
			final StringBuffer buffer = new StringBuffer();

			Object next = iterator.hasNext() ? iterator.next() : null;
			do {
				final Object object = next;
				next = iterator.hasNext() ? iterator.next() : null;

				final String type;
				final String id;
				if ( object instanceof InterchangeBegin ) {
					if ( next instanceof InterchangeEnd ) {
						type = "Interchange";
						id = ( (InterchangeBegin) object ).getIref();
					} else continue;
				} else if ( object instanceof MessageBegin ) {
					if ( next instanceof MessageEnd ) {
						type = "Message";
						id = ( (MessageBegin) object ).getMref();
					} else continue;
				} else if ( object instanceof CodeList ) {
					type = "CodeList";
					id = ( (CodeList) object ).getId();
				} else if ( object instanceof StatisticalConcept ) {
					type = "StatisticalConcept";
					id = ( (StatisticalConcept) object ).getId();
				} else if ( object instanceof KeyFamily ) {
					type = "KeyFamily";
					id = ( (KeyFamily) object ).getId();
				} else if ( object instanceof TimeSeries ) {
					type = "TimeSeries";
					id = ( (TimeSeries) object ).getKey();
				} else if ( object instanceof Attribute ) {
					final Attribute attribute = (Attribute) object;
					type = "Attribute";
					id = attribute.getName() + "@" + attribute.getKey() + (
						attribute.getPeriod().length() > 0
						||
						attribute.getPeriodFormat().length() > 0
							? "." + attribute.getPeriod()
								+ "." + attribute.getPeriodFormat()
							: ""
					);
				} else continue;
				buffer.append( ' ' );
				buffer.append( type );
				buffer.append( ' ' );
				buffer.append( '"' );
				buffer.append( id );
				buffer.append( '"' );
				if ( next != null ) buffer.append( ',' );
			} while ( next != null );
			properties.setProperty(
				UNEXPECTED_REPORT_FIELD,
				buffer.toString()
			);
		} else {
			properties.remove( UNEXPECTED_REPORT_FIELD );
		}
	}

	/**
	 * Write the acknowledgement file.
	 * Its name is given by the preference <CODE>acknowledgementFile</CODE>.
	 * If this preference is not set, no file is written.
	 * The boolean preference <CODE>mimeCompliant</CODE> states whether
	 * all properties should be written. Otherwise only the traditional ones
	 * are written and some of them traditionally without space after the
	 * colon.
	 *
	 * @param properties where to read the name-value-pairs from.
	 *
	 * @return whether the method was completed successfully.
	 */
	protected boolean writeAcknowledgementFile ( final Properties properties ) {
		final String ackFilename = componentHelper.getPreference(
			"acknowledgementFile", null
		);
		if ( ackFilename == null ) {
			componentHelper.logger.log(
				Level.CONFIG,
				componentHelper.logName,
				"close()",
				"No preference \"acknowledgementFile\" set."
			);
			return true;
		}

		try {
			final boolean mimeCompliant = ! "false".equalsIgnoreCase(
				componentHelper.getPreference( "mimeCompliant", "false" )
			);
			final String[] compatibleFields = {
				"To: ",
				"Orig.ExchangeID: ",
				"Subject: ",
				"IREF: ",
				"MREF: ",
				"Proc.Id:",
				"Rtrn.Code:",
				"Text: "
			};
			final String lineBreak = System.getProperty(
				"line.separator", "\n"
			);
			final String unknown = "unknown";

			final Writer writer = new FileWriter( ackFilename );
			if ( mimeCompliant ) {
				for (
					final Enumeration enumeration = properties.propertyNames();
					enumeration.hasMoreElements();
				) {
					final String field = (String) enumeration.nextElement();
					final String value = properties.getProperty(
						field, unknown
					);
					writer.write( field + ": " + value + lineBreak );
				}
			} else {
				for ( int i = 0; i < compatibleFields.length; ++i ) {
					final String fieldPlusColonSpace = compatibleFields[i];
					final String field = getNamePortion( fieldPlusColonSpace );
					final String value = properties.getProperty(
						field, unknown
					);
					writer.write( fieldPlusColonSpace + value + lineBreak );
				}
				final String unexpectedReport = properties.getProperty(
					UNEXPECTED_REPORT_FIELD, ""
				);
				if ( !"".equals( unexpectedReport ) ) {
					writer.write(
						UNEXPECTED_REPORT_FIELD + ":"
						+ unexpectedReport + lineBreak
					);
				}
			}
			writer.close();
		}
		catch ( final Exception e ) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"close()",
				"Writing acknowledgement file raised exception",
				e
			);
			return false;
		}

		return true;
	}

	/* ==================== Private Methods ===================== */

	/** Extracts from the line everything to the first colon. */
	private static String getNamePortion ( final String line ) {
		if ( line == null ) return null;
		final int index = line.indexOf( ':' );
		if ( index < 0 ) return line;
		return line.substring( 0, index ).trim();
	}

	/** Extracts from the line everything from the first colon. */
	private static String getValuePortion ( final String line ) {
		if ( line == null ) return null;
		final int index = line.indexOf( ':' );
		if ( index < 0 ) return "";
		return line.substring( index+1 ).trim();
	}
}

/* ***************************** Class ****************************** */

/**
 * <p>Watches out for selected {@link GESMESCBObject}s.</p>
 *
 * @author
 * <a href="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</a>
 */
abstract class GESMESCBObjectListener extends GESMESCBEventListenerAdapter {

	/* ========================= Fields ========================= */

	private final Collection receivedGESMESCBObjects;
	private final int limit;
	private boolean receivedGESMESCBObject;

	/* ====================== Constructors ====================== */

	public GESMESCBObjectListener ( final int limit ) {
		receivedGESMESCBObject = false;
		receivedGESMESCBObjects = new ArrayList();
		this.limit = limit;
	}

	/* ===================== Public Methods ===================== */

	public boolean receivedGESMESCBObject () {
		return receivedGESMESCBObject;
	}

	public Iterator iterator () {
		return Collections.unmodifiableCollection(
			receivedGESMESCBObjects
		).iterator();
	}

	public int capacity () {
		return limit;
	}

	/* =================== Protected Methods ==================== */

	protected void remember ( final GESMESCBObject object ) {
		receivedGESMESCBObject = true;
		if ( receivedGESMESCBObjects.size() < limit ) {
			receivedGESMESCBObjects.add( object );
		}
	}
}

/* ***************************** Class ****************************** */

/**
 * <p>Watches out for any {@link GESMESCBObject}s.</p>
 *
 * <p>This class is used to determine
 * whether unexpected objects have been sieved out.</p>
 *
 * @author
 * <a href="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</a>
 */
class GESMESCBAnyObjectListener extends GESMESCBObjectListener
	implements StructureListener, DataListener
{
	/* ====================== Constructors ====================== */

	public GESMESCBAnyObjectListener ( final int limit ) {
		super( limit );
	}

	/* ===================== Public Methods ===================== */

	public void handle ( final InterchangeBegin interchangeBegin ) {
		remember( interchangeBegin );
	}

	public void handle ( final InterchangeEnd interchangeEnd ) {
		remember( interchangeEnd );
	}

	public void handle ( final MessageBegin messageBegin ) {
		remember( messageBegin );
	}

	public void handle ( final MessageEnd messageEnd ) {
		remember( messageEnd );
	}

	public void handle ( final CodeList codeList ) {
		remember( codeList );
	}

	public void handle ( final StatisticalConcept statisticalConcept ) {
		remember( statisticalConcept );
	}

	public void handle ( final KeyFamily keyFamily ) {
		remember( keyFamily );
	}

	public void handle ( final TimeSeries timeSeries ) {
		remember( timeSeries );
	}

	public void handle ( final Attribute attribute ) {
		remember( attribute );
	}
}

/* ***************************** Class ****************************** */

/**
 * <p>Watches out for {@link InterchangeBegin} and {@link MessageBegin}
 * objects.</p>
 *
 * @author
 * <a href="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</a>
 */
class GESMESCBEnvelopeListener extends GESMESCBObjectListener {

	/* ====================== Constructors ====================== */

	public GESMESCBEnvelopeListener ( final int limit ) {
		super( limit );
	}

	/* ===================== Public Methods ===================== */

	public void handle ( final InterchangeBegin interchangeBegin ) {
		remember( interchangeBegin );
	}

	public void handle ( final InterchangeEnd interchangeEnd ) {
		// Ignore
	}

	public void handle ( final MessageBegin messageBegin ) {
		remember( messageBegin );
	}

	public void handle ( final MessageEnd messageEnd ) {
		// Ignore
	}
}
