// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2002, 2004 European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.error;

import java.util.ArrayList;

import java.io.FileWriter;
import java.io.FileReader;
import java.io.BufferedReader;
import java.io.File;

import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;

import Int.ecb.gesmescb.*;

import Int.ecb.gesmescb.event.ErrorEvent;
import Int.ecb.gesmescb.event.TestEventSource;

import Int.ecb.gesmescb.error.AcknowledgementManager;

import Int.ecb.logging.*;

/* ***************************** Class ****************************** */

/**
 * <P>Tests the implementation of the Int.ecb.gesmesdb.sieve.AcknowledgementManager
 * class.</P>
 *
 * @author
 * <A HREF="mailto:Sami.Airo@ECB.int">Sami Airo</A>
 * @author
 * <A HREF="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</A>
 *
 * @version @(#)$Id: AcknowledgementManagerTest.java,v 1.1 2017/01/19 16:14:05 trainot Exp $
 */

public class AcknowledgementManagerTest extends TestCase {

	/* ========================= Fields ========================= */

	private File headerFile;
	private File ackFile;

	/* ====================== Constructors ====================== */

	/**
	 * Constructs a AcknowledgementManagerTest with the specified name.
	 *
	 * @param name Test case name
	 */
	public AcknowledgementManagerTest ( final String name ) {
		super( name );
	}

	/* ================== Protected Methods ===================== */

	/**
	 * Sets up the test fixture.
	 */
	protected void setUp () {
		final String lineBreak = System.getProperty( "line.separator", "\n" );

		System.setProperty( ".level", "ALL" );
		System.setProperty( "handlers", "Int.ecb.logging.ConsoleHandler" );
		System.setProperty(
			"Int.ecb.logging.ConsoleHandler.formatter",
			"Int.ecb.gesmescb.util.NormalLogFormatter"
		);

		try {
			headerFile = File.createTempFile("header.", null);
			FileWriter writer = new FileWriter( headerFile );
			writer.write(
				"EDX: 1.0" + lineBreak
				+ "From: 5B" + lineBreak
				+ "To: EU" + lineBreak
				+ "Orig.ExchangeID: E12033" + lineBreak
				+ "IREF: 587" + lineBreak
				+ "MREF: 1" + lineBreak
				+ "Subject: BIS_MACRO" + lineBreak
			);
			writer.close();
			headerFile.deleteOnExit();
		} catch (Exception e) {
			fail(e.getMessage());
		}

		try {
			ackFile = File.createTempFile("ack.", null);
		} catch (Exception e) {
			fail(e.getMessage());
		}

		final String className = "Int.ecb.gesmescb.error.AcknowledgementManager";
		System.setProperty( className + ".headerFile", headerFile.getAbsolutePath() );
		System.setProperty( className + ".acknowledgementFile", ackFile.getAbsolutePath() );
	}

	/**
	 * Tears down the test fixture.
	 */
	protected void tearDown () {
		headerFile.delete();
		headerFile = null;
		ackFile.delete();
		ackFile = null;
	}

	/* ===================== Public Methods ===================== */

	/**
	 * Assembles and returns a test suite for all
	 * the test methods of this test case.
	 *
	 * @return a non-null test suite.
	 */
	public static Test suite () {
		TestSuite suite = new TestSuite( AcknowledgementManagerTest.class );
		return suite;
	}

	/**
	 * Runs the test case.
	 */
	public static void main ( final String args[] ) {
		final String[] testCaseName = {
			AcknowledgementManagerTest.class.getName()
		};
		junit.textui.TestRunner.main( testCaseName );
	}

	/* --------------------- Test Methods --------------------- */

	public void testNormalInterchange () {
		final AcknowledgementManager manager = new AcknowledgementManager();
		final TestEventSource errorSource = new TestEventSource();
		final Handler handler = manager.getErrorHandler();

		errorSource.addListener( manager.getErrorListener(), true );
		handler.publish( new LogRecord( Level.WARNING, "" ) );
		errorSource.dispatch( new ErrorEvent( "", ErrorEvent.WARNING ) );
		manager.close();

		ArrayList lines = readFile(ackFile);

		assertEquals( "number of lines in ack", 8, lines.size() );
		checkBaseContents( lines );

		/* This is never sent back
		assertEquals( "Proc.Id:E2", lines.get(3) );
		assertEquals( "Rtrn.Code:E2", lines.get(4) );
		assertEquals( "Text: Checker did not find errors", lines.get(5) );
		*/

		assertEquals( "procId", "Proc.Id:F1", lines.get(5) );
		assertEquals( "returnCode", "Rtrn.Code:F1", lines.get(6) );
		assertEquals( "text", "Text: Gesmes message successfully checked and loaded", lines.get(7) );

		assertEquals( "exit code", 0, manager.computeExitCode() );
	}

	public void testErrorInChecking () {
		final AcknowledgementManager manager = new AcknowledgementManager();
		final TestEventSource errorSource = new TestEventSource();
		final Handler handler = manager.getErrorHandler();

		errorSource.addListener( manager.getErrorListener(), true );
		handler.publish( new LogRecord( Level.SEVERE, "" ) );
		errorSource.dispatch( new ErrorEvent( "", ErrorEvent.ERROR ) );
		manager.close();

		ArrayList lines = readFile(ackFile);

		assertEquals( "number of lines in ack", 8, lines.size() );
		checkBaseContents( lines );

		assertEquals( "procId", "Proc.Id:E2", lines.get(5) );
		assertEquals( "returnCode", "Rtrn.Code:E21", lines.get(6) );
		assertEquals( "text", "Text: Checker found errors", lines.get(7) );

		assertEquals( "exit code", 2, manager.computeExitCode() );
	}

	public void testSeriousErrorEvent () {
		final AcknowledgementManager manager = new AcknowledgementManager();
		final TestEventSource errorSource = new TestEventSource();

		errorSource.addListener( manager.getErrorListener(), true );
		errorSource.dispatch( new ErrorEvent( "", ErrorEvent.FAILURE ) );
		manager.close();

		ArrayList lines = readFile(ackFile);

		assertEquals( "number of lines in ack", 8, lines.size() );
		checkBaseContents( lines );

		assertEquals( "procId", "Proc.Id:E2", lines.get(5) );
		assertEquals( "returnCode", "Rtrn.Code:E81", lines.get(6) );
		assertEquals( "text", "Text: Serious problem with the checker", lines.get(7) );

		assertEquals( "exit code", 99, manager.computeExitCode() );
	}

	public void testSeriousError () {
		final AcknowledgementManager manager = new AcknowledgementManager();
		final TestEventSource errorSource = new TestEventSource();
		final Handler handler = manager.getErrorHandler();

		errorSource.addListener( manager.getErrorListener(), true );
		errorSource.dispatch( new ErrorEvent( "", ErrorEvent.WARNING ) );
		handler.publish( new LogRecord( Level.SEVERE, "" ) );
		manager.close();

		ArrayList lines = readFile(ackFile);

		assertEquals( "number of lines in ack", 8, lines.size() );
		checkBaseContents( lines );

		assertEquals( "procId", "Proc.Id:E2", lines.get(5) );
		assertEquals( "returnCode", "Rtrn.Code:E81", lines.get(6) );
		assertEquals( "text", "Text: Serious problem with the checker", lines.get(7) );

		assertEquals( "exit code", 98, manager.computeExitCode() );
	}

	public void testUnexpectedFound () {
		System.setProperty(
			"Int.ecb.gesmescb.error.AcknowledgementManager.reportUnexpected",
			"100"
		);
		AcknowledgementManager manager = new AcknowledgementManager();
		final TestEventSource errorSource = new TestEventSource();
		final TestEventSource unexpectedSource = new TestEventSource();
		final Handler handler = manager.getErrorHandler();

		errorSource.addListener( manager.getErrorListener(), true );
		unexpectedSource.addListener( manager.getUnexpectedListener(), true );
		errorSource.dispatch( new ErrorEvent( "", ErrorEvent.WARNING ) );
		handler.publish( new LogRecord( Level.WARNING, "" ) );
		unexpectedSource.dispatch( new Attribute(
			"KEY", "", "", "NAME", ""
		) );

		manager.close();

		ArrayList lines = readFile(ackFile);

		assertEquals( "number of lines in ack", 9, lines.size() );
		checkBaseContents( lines );

		assertEquals( "procId", "Proc.Id:F1", lines.get( 5 ) );
		assertEquals( "returnCode", "Rtrn.Code:F2", lines.get( 6 ) );
		assertEquals(
			"text",
			"Text: Gesmes message successfully checked and loaded"
			+ " but some unexpected objects found",
			lines.get( 7 )
		);
		assertEquals(
			"unexpected",
			"Unexpected objects: Attribute \"NAME@KEY\"",
			lines.get( 8 )
		);

		assertEquals( "exit code", 1, manager.computeExitCode() );
	}

	public void testErrorsAndUnexpectedFound () {
		System.setProperty(
			"Int.ecb.gesmescb.error.AcknowledgementManager.reportUnexpected",
			"0"
		);
		AcknowledgementManager manager = new AcknowledgementManager();
		final TestEventSource errorSource = new TestEventSource();
		final TestEventSource unexpectedSource = new TestEventSource();

		errorSource.addListener( manager.getErrorListener(), true );
		unexpectedSource.addListener( manager.getUnexpectedListener(), true );
		errorSource.dispatch( new ErrorEvent( "", ErrorEvent.ERROR ) );
		unexpectedSource.dispatch( new Attribute( "", "", "", "", "" ) );

		manager.close();

		ArrayList lines = readFile(ackFile);

		assertEquals( "number of lines in ack", 8, lines.size() );
		checkBaseContents( lines );

		assertEquals( "procId", "Proc.Id:E2", lines.get(5) );
		assertEquals( "returnCode", "Rtrn.Code:E21", lines.get(6) );
		assertEquals( "text", "Text: Checker found errors", lines.get(7) );

		assertEquals( "exit code", 2, manager.computeExitCode() );
	}

	public void testSubject () {
		final AcknowledgementManager manager = new AcknowledgementManager();
		final TestEventSource source = new TestEventSource();
		
		source.addListener( manager.getEnvelopeListener(), true );
		source.dispatch( new InterchangeBegin(
			"IREF000815", "EE2", "5B0", "040214:1845", true
		) );
		source.dispatch( new MessageBegin(
			"MREF000005", "EE2", "5B0", "ECB",
			MessageBegin.STATISTICAL_DATA, "Hokus", new Contact[]{},
			"ECB_HOKUS", MessageBegin.UPDATE_OR_REPLACE_ACTION,
			"200402151845", "", "", "ECB_FICUS"
		) );
		manager.close();

		final ArrayList lines = readFile( ackFile );

		assertEquals( "subject", "Subject: ECB_HOKUS", lines.get( 2 ) );

		assertEquals( "exit code", 0, manager.computeExitCode() );
	}

	/* ==================== Private Methods ===================== */

	private void checkBaseContents ( final ArrayList lines ) {
		assertEquals( "to", "To: 5B", lines.get(0) );
		assertEquals( "origId", "Orig.ExchangeID: E12033", lines.get(1) );
		assertEquals( "subject", "Subject: BIS_MACRO", lines.get(2) );
		assertEquals( "iref", "IREF: 587", lines.get(3) );
		assertEquals( "mref", "MREF: 1", lines.get(4) );
	}

	private ArrayList readFile ( final File file ) {
		final ArrayList lines = new ArrayList();
		try {
			final BufferedReader fileReader = new BufferedReader(new FileReader(file) );
			String line;
			while ( (line = fileReader.readLine()) != null ) {
				lines.add( line );
			}
			fileReader.close();
		} catch (Exception e) {
			lines.clear();
		}
		return lines;
	}
}
