// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.error;

import java.lang.String;
import java.util.List;
import java.util.Arrays;
import java.util.Collections;

/* ************************** Interface *************************** */

/**
 * <P>Reports errors.</P>
 *
 * <P>Classes implementing this interface watch out for particular
 * error conditions. The interface then provides query functions
 * to the outside which can be used to extract stored error reports.</P>
 *
 * <P>A severity code and a human readable descriptive message
 * can be queried.
 * The valid severity codes fulfill the relations
 * <CENTER>
 * {@link #OK} &lt; {@link #WARNING} &lt;
 * {@link #ERROR} &lt; {@link #FAILURE}.
 * </CENTER></P>
 *
 * @author
 * <A HREF="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</A>
 *
 * @version @(#)$Id: ErrorWatcher.java,v 1.1 2017/01/19 16:14:05 trainot Exp $
 */
public interface ErrorWatcher {

	/* ======================= Constants ======================== */

	/**
	 * The severity code for "ok".
	 */
	public static final int OK = 0;

	/**
	 * The severity code for "warning".
	 * This severity may be used in place of {@link #ERROR}
	 * if the problem is not severe enough to affect the integrity
	 * of vital GESMES/TS data but should nevertheless be
	 * reported to the operator.
	 * This code should be used very carefully since the data
	 * is loaded and the warning could be overlooked.
	 * So only use it when with high probability
	 * even overlooking the warning message does not harm.
	 * The recommended policy is to change an <CODE>ERROR</CODE>
	 * into a <CODE>WARNING</CODE> on a case by case basis
	 * and only if there is a sufficient benefit from doing so
	 * and if the potential risks are negligible.
	 */
	public static final int WARNING = 1;

	/**
	 * The severity code for "error".
	 * This severity must be used if the error condition
	 * is so severe that the integrity of vital GESMES/TS data may suffer
	 * if the data were written to its final destination but
	 * if {@link #FAILURE} is not applicable.
	 */
	public static final int ERROR = 2;

	/**
	 * The severity code for "failure".
	 * This severity must be used to report internal
	 * problems of the application such as running out of disk space
	 * or detection of an implementation bug.
	 */
	public static final int FAILURE = 3;

	/**
	 * Translates severity codes
	 * {@link #OK}, {@link #WARNING}, {@link #ERROR}, and
	 * {@link #FAILURE} to {@link String}s.
	 *
	 * This is not an array in order to avoid public modifiable
	 * static data.
	 * The price for security is
	 * that you have to cast the values got to <CODE>String</CODE>.
	 */
	public static final List SEVERITY_TRANSLATOR
	= Collections.unmodifiableList(
		Arrays.asList(
			new String[] { "OK", "WARNING", "ERROR", "FAILURE" }
		)
	);

	/* ======================== Methods ========================= */

	/**
	 * Tells if everything is all right or if problems were encountered.
	 *
	 * @return one of
	 * {@link #OK}, {@link #WARNING}, {@link #ERROR}, and {@link #FAILURE}.
	 */
	int getSeverity ();

	/**
	 * <P>Get a human readable description of the problem
	 * if there was one.</P>
	 *
	 * <P>In case of many problems encountered, the one of highest severity
	 * should be reported and among those, if there are more than one,
	 * the first one.</P>
	 *
	 * <P>These are minimum requirements only.
	 * An <CODE>ErrorWatcher</CODE> may also report more problems than
	 * one. The ordering of reported problems then should be by time.</P>
	 *
	 * @return a message describing the problem.
	 */
	String getMessage ();

	/**
	 * Reset the internal memory.
	 * After <CODE>clear</CODE> has been called,
	 * the <CODE>ErrorWatcher</CODE> should behave
	 * like a new born one. Particularly, {@link #getSeverity}
	 * should return {@link #OK}.
	 */
	void clear ();
}
