// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2007, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.essbase;

import java.util.Date;
import java.util.Iterator;
import java.util.Map;
import java.util.LinkedHashMap;

import java.sql.SQLException;

import Int.ecb.logging.Level;

import Int.ecb.gesmescb.TimeSeries;
import Int.ecb.gesmescb.Observation;

import Int.ecb.gesmescb.event.ConfiguredSourceAdapter;

import Int.ecb.gesmescb.time.RegularInterval;
import Int.ecb.gesmescb.time.GregorianRegularInterval;

import Int.ecb.gesmescb.util.ObservationMap;

import Int.ecb.gesmescb.csv.EaacsLabConfig;

/* ***************************** Class ****************************** */

/**
 * Sends data received from Essbase via custimed defined functions
 * as {@link Int.ecb.gesmescb.GESMESCBObject}s.
 *
 * <center><table border="3" cellpadding="3">
 * <caption>
 * <a name="prefs">Available {@link Int.ecb.prefs preference} settings</a>
 * </caption>
 * <tr>
 * <th>preference name</th>
 * <th>type</th>
 * <th>meaning</th>
 * <th>default</th>
 * <tr>
 *	<td><code>maxCache</td><td>integer</td>
 *	<td>how many time series should be cached internally
 *	to combine observations from random order into few series</td>
 *	<td><code>0</code></td>
 * </tr>
 * <tr>
 *	<td><code>maxGap</td><td>integer</td>
 *	<td>how many successive empty observations are permitted
 *	before the series is split</td>
 *	<td><code>{@link Integer.MAX_VALUE}</code></td>
 * </tr>
 * <tr>
 *	<td><code>normalStatus</td><td>string</td>
 *	<td>observation status code for normal observations</td>
 *	<td><code>A</code></td>
 * </tr>
 * <tr>
 *	<td><code>missingStatus</td><td>string</td>
 *	<td>observation status code for missing (<code>NaN</code>) observations</td>
 *	<td><code>L</code></td>
 * </tr>
 * <tr>
 *	<td><code>confidentiality</td><td>string</td>
 *	<td>observation confidentiality code</td>
 *	<td><code>F</code></td>
 * </tr>
 * </table></center>
 *
 * @see ConfiguredSourceAdapter for configuration
 * of GESMES/TS envelope constructs.
 * @see EaacsLabConfig for configuration of reporting period.
 *
 * @author
 * <a href="mailto:sis.external@ecb.int">Franois Hoeltzli</A>
 * @author
 * <a href="mailto:sis.external@ecb.int">Bernhard Bodenstorfer</A>
 *
 * @version @(#)$Id: FunctionInvokedSource.java,v 1.1 2017/01/19 16:14:05 trainot Exp $
 */
public class FunctionInvokedSource extends ConfiguredSourceAdapter {

	/* ======================= Constants ======================== */

	/* ========================= Fields ========================= */

	private final EaacsLabConfig eaacsLabConfig;
	private final ExhibitEldestMap timeSeriesMap;
	private final String normalStatus, missingStatus, confidentiality;
	private final int maxCache, maxGap;
	private boolean haveOpened;

	/* ====================== Constructors ====================== */

	/**
	 * Construct a <code>FunctionInvokedSource</code>
	 * reading from a database according to preferences.
	 *
	 * @param name used for logging and preferences.
	 *
	 * @throws NullPointerException if the preference variable
	 * <code>source</code> is undefined.
	 * @throws SQLException if the database cannot be opened for reading.
	 */
	public FunctionInvokedSource ( final String name ) {
		super( name );
		componentHelper.logger.log(
			Level.FINE,
			componentHelper.logName,
			"FunctionInvokedSource()",
			"Opening input from function invocations."
		);
		eaacsLabConfig = new EaacsLabConfig( name );
		normalStatus = componentHelper.getPreference( "normalStatus", "A" );
		missingStatus = componentHelper.getPreference( "missingStatus", "L" );
		confidentiality = componentHelper.getPreference(
			"confidentiality", "F"
		);
		{
			int maxCache;
			final String maxCacheString = componentHelper.getPreference(
				"maxCache", "0"
			);
			try {
				maxCache = Integer.parseInt( maxCacheString );
			} catch ( final NumberFormatException nfe ) {
				componentHelper.logger.log(
					Level.SEVERE,
					componentHelper.logName,
					"FunctionInvokedSource()",
					"Invalid configuration maxCache=\""
					+ maxCacheString
					+ "\"; it should be numeric. Using the default."
				);
				maxCache = 0;
			}
			this.maxCache = maxCache;
			componentHelper.logger.log(
				Level.CONFIG,
				componentHelper.logName,
				"FunctionInvokedSource()",
				"Configured maxCache=" + this.maxCache + "."
			);
		}
		{
			int maxGap;
			final String maxGapString = componentHelper.getPreference(
				"maxGap", Integer.toString( Integer.MAX_VALUE )
			);
			try {
				maxGap = Integer.parseInt( maxGapString );
			} catch ( final NumberFormatException nfe ) {
				componentHelper.logger.log(
					Level.SEVERE,
					componentHelper.logName,
					"FunctionInvokedSource()",
					"Invalid configuration maxGap=\""
					+ maxGapString
					+ "\"; it should be numeric. Using the default."
				);
				maxGap = Integer.MAX_VALUE;
			}
			this.maxGap = maxGap;
			componentHelper.logger.log(
				Level.CONFIG,
				componentHelper.logName,
				"FunctionInvokedSource()",
				"Configured maxGap=" + this.maxGap + "."
			);
		}
		timeSeriesMap = new ExhibitEldestMap( 0, 1.5f, false );
		haveOpened = false;
	}

	/* ===================== Public Methods ===================== */

	/**
	 * Extract from the CIS type JDBC database the data as
	 * {@link Int.ecb.gesmescb.GESMESCBObject}s.
	 * These then are dispatched to all listeners.
	 */
	public void handle (
		final String[] coordinates,
		final String year,
		final String period,
		final double value
	) {
		if ( ! haveOpened ) {
			final Date now = new Date();
			componentHelper.logger.log(
				Level.FINER,
				componentHelper.logName,
				"handle()",
				"Starting object stream now: " + now
			);
			dispatchInterchangeBegin( 0, now );
			dispatchMessageBegin( 1, now );
			haveOpened = true;
		}

		final String key;
		{
			final StringBuffer keyBuffer = new StringBuffer();
			for ( int i = 0; i < coordinates.length; ++i ) {
				if ( 0 < i ) keyBuffer.append( '.' );
				keyBuffer.append( coordinates[i] );
			}
			key = keyBuffer.toString();
		}
		componentHelper.logger.log(
			Level.FINEST,
			componentHelper.logName,
			"handle()",
			"Received observation: "
			+ key + "[" + year + "/" + period + "] = " + value + "."
		);
		if ( ! timeSeriesMap.containsKey( key ) ) {
			componentHelper.logger.log(
				Level.FINER,
				componentHelper.logName,
				"handle()",
				"This is a new time series: " + key
			);
			timeSeriesMap.put( key, new ObservationMap() );
		}
		final ObservationMap observationMap
			= (ObservationMap) timeSeriesMap.get( key );
		final int
			frequency,
			index;
		if ( "A".equals( period ) ) {
			frequency = RegularInterval.ANNUAL;
			index = 0;
		} else if ( "Q1".equals( period ) ) {
			frequency = RegularInterval.QUARTERLY;
			index = 0;
		} else if ( "Q2".equals( period ) ) {
			frequency = RegularInterval.QUARTERLY;
			index = 1;
		} else if ( "Q3".equals( period ) ) {
			frequency = RegularInterval.QUARTERLY;
			index = 2;
		} else if ( "Q4".equals( period ) ) {
			frequency = RegularInterval.QUARTERLY;
			index = 3;
		} else {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"handle()",
				"Invalid period: \"" + period + "\"."
			);
			return;
		}

		final int intYear;
		if ( 'Y' == year.charAt( 0 ) ) {
			try {
				intYear = Integer.parseInt( year.substring( 1 ) );
			}
			catch ( final NumberFormatException e ) {
				componentHelper.logger.log(
					Level.SEVERE,
					componentHelper.logName,
					"handle()",
					"Invalid year; not numeric after Y: \"" + year + "\".",
					e
				);
				return;
			}
		} else {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"handle()",
				"Invalid year, should start with Y: \"" + year + "\"."
			);
			return;
		}

		final RegularInterval interval = new GregorianRegularInterval(
			frequency,
			intYear,
			index,
			1
		);
		final boolean missingValue
		= Double.isNaN( value ) || Double.isInfinite( value );
		final Observation observation = new Observation(
			missingValue ? Observation.MISSING_VALUE : Double.toString( value ),
			missingValue ? missingStatus : normalStatus,
			confidentiality,
			""
		);

		observationMap.put( interval, observation );

		if ( maxCache < timeSeriesMap.size() ) {
			final Map.Entry entry = timeSeriesMap.getEldestEntry();
			dispatchAsSeries( entry );
			timeSeriesMap.remove( entry.getKey() );
		}
	}

	/** Close the <code>FunctionInvokedSource</code>. */
	public void close () {
		componentHelper.logger.log(
			Level.FINE,
			componentHelper.logName,
			"close()",
			"Closing input from function invocations."
		);
		if ( ! haveOpened ) {
			final Date now = new Date();
			componentHelper.logger.log(
				Level.FINER,
				componentHelper.logName,
				"handle()",
				"Empty object stream marked: " + now
			);
			dispatchInterchangeBegin( 0, now );
		} else {
			componentHelper.logger.log(
				Level.FINER,
				componentHelper.logName,
				"close()",
				"Need to dispatch " + timeSeriesMap.size() + " series."
			);
			for (
				final Iterator iterator = timeSeriesMap.entrySet().iterator();
				iterator.hasNext();
			) {
				final Map.Entry entry = (Map.Entry) iterator.next();
				dispatchAsSeries( entry );
				iterator.remove();
			}
			dispatchMessageEnd( 1 );
		}
		dispatchInterchangeEnd( 0 );
		componentHelper.logger.log(
			Level.FINER,
			componentHelper.logName,
			"close()",
			"Closed input from function invocations."
		);
	}

	/* =================== Protected Methods ==================== */

	/** Retrieves the reporting interval begin time point from the
	 * {@link EaacsLabConfig}.
	 */
	protected Date getReportingBegin () {
		return eaacsLabConfig.getBegin();
	}

	/** Retrieves the reporting interval end time point from the
	 * {@link EaacsLabConfig}.
	 */
	protected Date getReportingEnd () {
		return eaacsLabConfig.getEnd();
	}
	/* ==================== Private Methods ===================== */

	/** Dispatch all content from {@link ObservationMap} as {@link TimeSeries}.
	 *
	 * @param entry has time series key as key and the observation map as value
	 */
	private void dispatchAsSeries ( final Map.Entry entry ) {
		final String key = (String) entry.getKey();
		final ObservationMap map = (ObservationMap) entry.getValue();
		final RegularInterval interval = map.getInterval();
		boolean hasLogged = false;
		for (
			final Iterator periods = interval.iterator();
			periods.hasNext();
		) {
			final RegularInterval period = (RegularInterval) periods.next();
			if (
				0 > period.getBegin().compareTo( eaacsLabConfig.getBegin() )
				||
				0 <= period.getBegin().compareTo( eaacsLabConfig.getEnd() )
			) {
				if ( ! hasLogged ) {
					componentHelper.logger.log(
						Level.FINER,
						componentHelper.logName,
						"dispatchAsSeries()",
						"Skipping observations of time series " + key
						+ "[" + interval
						+ "] which fall outside of configured interval from "
						+ eaacsLabConfig.getBegin()
						+ " to "
						+ eaacsLabConfig.getEnd()
						+ "."
					);
					hasLogged = true;
				}
				map.remove( period );
			}
		}
		for (
			final Iterator series = map.iterateTimeSeries(
				key, Integer.MAX_VALUE
			);
			series.hasNext();
		) {
			final TimeSeries timeSeries = (TimeSeries) series.next();
			dispatch( timeSeries );
		}
	}
}

/* ***************************** Class ****************************** */
/** Exhibits the best candidate for removal from cache. */
class ExhibitEldestMap extends LinkedHashMap {
	/**
	 *   serialVersionUID generated by eclipse.
	 */
	private static final long serialVersionUID = -3320241327354862595L;
	private Map.Entry eldestEntry;

	public ExhibitEldestMap (
		final int initialCapacity,
		final float loadFactor,
		final boolean access
	) {
		super( initialCapacity, loadFactor, access );
	}

	public Map.Entry getEldestEntry () {
		return eldestEntry;
	}

	protected boolean removeEldestEntry (
		final Map.Entry entry
	) {
		eldestEntry = entry;
		return false;
	}
}
