// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2007, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.essbase;

import java.util.Set;
import java.util.HashSet;
import java.util.Iterator;

import Int.ecb.gesmescb.InterchangeBegin;
import Int.ecb.gesmescb.InterchangeEnd;
import Int.ecb.gesmescb.MessageBegin;
import Int.ecb.gesmescb.MessageEnd;
import Int.ecb.gesmescb.Contact;
import Int.ecb.gesmescb.TimeSeries;
import Int.ecb.gesmescb.Observation;

import Int.ecb.gesmescb.event.TestEventListener;
import Int.ecb.gesmescb.csv.EaacsLabConfig;

import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;

/* ***************************** Class ****************************** */

/**
 * Tests the implementation of the class {@link FunctionInvokedSource}.
 *
 * @author
 * <a href="mailto:sis.external@ecb.int">Bernhard Bodenstorfer</A>
 *
 * @version @(#)$Id: FunctionInvokedSourceTest.java,v 1.1 2017/01/19 16:14:05 trainot Exp $
 */

public class FunctionInvokedSourceTest extends TestCase {

	/* ======================= Constants ======================== */

	private static final String
		objectName = FunctionInvokedSource.class.getName() + ".Hoizbua",
		sharedName = EaacsLabConfig.class.getName() + ".Hoizbua";

	private static final String[][]
		objectProperties = {
			{ "prepared", "2000-12-01T14:10:00" },
			{ "test", "true" },
			{ "maintenanceAgency", "OPS" },
			{ "keyFamily", "OPS_UNFUG" },
			{ "dataSet", "OPS_MIST" },
			{ "extracted", "2008-01-10T23:15:00" },
			{ "normalStatus", "A" },
			{ "missingStatus", "S" },
			{ "confidentiality", "N" },
			{ "maxCache", "100" }
		},
		sharedProperties = {
			{ "begin", "2006-01-01T00:00:00" },
			{ "end", "2010-01-01T00:00:00" }
		};

	/* The output from the {@link FunctionInvokedSource} under Test. */

	private static final InterchangeBegin INTERCHANGE_BEGIN
		= new InterchangeBegin(
			"IREF000000", "ZZZ", "ZZZ", "001201:1410", true
		);

	private static final MessageBegin MESSAGE_BEGIN
		= new MessageBegin(
			"MREF000001", "ZZZ", "ZZZ", "OPS",
			MessageBegin.STATISTICAL_DATA, "",
			new Contact[]{},
			"OPS_MIST", MessageBegin.UPDATE_OR_REPLACE_ACTION,
			"200801102315", "20062009", "702", "OPS_UNFUG"
		);

	private static final MessageEnd MESSAGE_END
		= new MessageEnd( "MREF000001" );

	private static final InterchangeEnd INTERCHANGE_END
		= new InterchangeEnd( "IREF000000" );

	/* ====================== Constructors ====================== */

	/**
	 * Constructs an FunctionInvokedSourceTest with the specified name.
	 *
	 * @param name Test case name
	 */
	public FunctionInvokedSourceTest ( final String name ) {
		super( name );
	}

	/* ================== Protected Methods ===================== */

	protected void setUp () {
		for ( int i = 0; i < sharedProperties.length; ++i ) {
			System.setProperty(
				sharedName + "." + sharedProperties[i][0],
				sharedProperties[i][1]
			);
		}
		for ( int i = 0; i < objectProperties.length; ++i ) {
			System.setProperty(
				objectName + "." + objectProperties[i][0],
				objectProperties[i][1]
			);
		}
	}

	protected void tearDown () {
		for ( int i = 0; i < sharedProperties.length; ++i ) {
			System.getProperties().remove(
				sharedName + "." + sharedProperties[i][0]
			);
		}
		for ( int i = 0; i < objectProperties.length; ++i ) {
			System.getProperties().remove(
				objectName + "." + objectProperties[i][0]
			);
		}
	}

	/* ===================== Public Methods ===================== */

	/**
	 * Assembles and returns a test suite for all
	 * the test methods of this test case.
	 *
	 * @return a non-null test suite
	 */
	public static Test suite () {
		return new TestSuite( FunctionInvokedSourceTest.class );
	}

	/** Run the test case. */
	public static void main ( final String args[] ) {
		final String[] testCaseName = { FunctionInvokedSourceTest.class.getName() };
		junit.textui.TestRunner.main( testCaseName );
	}

	/* ---------------------- Test Methods ---------------------- */

	public void testEmpty () {
		final FunctionInvokedSource source
		= new FunctionInvokedSource( "Hoizbua" );
		final TestEventListener listener = new TestEventListener(
			new Object[]{
				INTERCHANGE_BEGIN,
				INTERCHANGE_END
			}
		);
		source.addListener( listener, true );
		source.close();
		listener.close();
	}

	public void testAnnualSeries () {
		final FunctionInvokedSource source
		= new FunctionInvokedSource( "Hoizbua" );
		final TestEventListener listener = new TestEventListener(
			new Object[]{
				INTERCHANGE_BEGIN,
				MESSAGE_BEGIN,
				new TimeSeries(
					"H.LZ.B", "20062008", "702", new Observation[]{
						new Observation( "0.0", "A", "N", "" ),
						new Observation( "1.0", "A", "N", "" ),
						new Observation( "9.0", "A", "N", "" )
					}
				),
				MESSAGE_END,
				INTERCHANGE_END
			}
		);

		source.addListener( listener, true );

		source.handle( new String[]{ "H", "LZ", "B" }, "Y2006", "A", 10 );
		source.handle( new String[]{ "H", "LZ", "B" }, "Y2007", "A", 1 );
		source.handle( new String[]{ "H", "LZ", "B" }, "Y2008", "A", 9 );
		source.handle( new String[]{ "H", "LZ", "B" }, "Y2006", "A", 0 );

		source.close();
		listener.close();
	}

	public void testQuarterlySeries () {
		final FunctionInvokedSource source
		= new FunctionInvokedSource( "Hoizbua" );
		final TestEventListener listener = new TestEventListener(
			new Object[]{
				INTERCHANGE_BEGIN,
				MESSAGE_BEGIN,
				new TimeSeries(
					"H.LZ.A", "2006220083", "708", new Observation[]{
						new Observation( "1.0", "A", "N", "" ),
						new Observation( "4.0", "A", "N", "" ),
						Observation.EMPTY_OBSERVATION,
						Observation.EMPTY_OBSERVATION,
						new Observation( "8.0", "A", "N", "" ),
						new Observation( "7.0", "A", "N", "" ),
						Observation.EMPTY_OBSERVATION,
						new Observation( "0.0", "A", "N", "" ),
						new Observation(
							Observation.MISSING_VALUE, "S", "N", ""
						),
						new Observation( "9.25", "A", "N", "" )
					}
				),
				MESSAGE_END,
				INTERCHANGE_END
			}
		);

		source.addListener( listener, true );

		source.handle( new String[]{ "H", "LZ", "A" }, "Y2006", "Q3", 19 );
		source.handle(
			new String[]{ "H", "LZ", "A" }, "Y2008", "Q1", Double.NaN
		);
		source.handle( new String[]{ "H", "LZ", "A" }, "Y2006", "Q2", 1 );
		source.handle( new String[]{ "H", "LZ", "A" }, "Y2006", "Q3", 4 );
		source.handle( new String[]{ "H", "LZ", "A" }, "Y2007", "Q2", 8 );
		source.handle( new String[]{ "H", "LZ", "A" }, "Y2007", "Q3", 7 );
		source.handle( new String[]{ "H", "LZ", "A" }, "Y2008", "Q1", 0 );
		source.handle(
			new String[]{ "H", "LZ", "A" }, "Y2008", "Q2", Double.NaN
		);
		source.handle( new String[]{ "H", "LZ", "A" }, "Y2008", "Q3", 9.25 );

		source.close();
		listener.close();
	}

	public void testSeriesMix () {
		final FunctionInvokedSource source
		= new FunctionInvokedSource( "Hoizbua" );
		final TestEventListener listener = new TestEventListener();

		source.addListener( listener, true );

		source.handle( new String[]{ "H", "LZ", "A" }, "Y2006", "A", 0 );
		source.handle( new String[]{ "H", "LZ", "A" }, "Y2007", "A", 1 );
		source.handle( new String[]{ "H", "LZ", "A" }, "Y2008", "A", 2 );
		source.handle( new String[]{ "H", "LZ", "A" }, "Y2009", "A", 3 );
		source.handle(
			new String[]{ "H", "LZ", "B" }, "Y2006", "A", Double.NaN
		);
		source.handle( new String[]{ "H", "LZ", "B" }, "Y2007", "A", 2.125 );
		source.handle( new String[]{ "H", "LZ", "B" }, "Y2008", "A", 1.0625 );
		source.handle(
			new String[]{ "H", "LZ", "B" }, "Y2009", "A", Double.NaN
		);
		source.handle( new String[]{ "H", "LZ", "L" }, "Y2006", "A", 81 );
		source.handle(
			new String[]{ "H", "LZ", "L" }, "Y2007", "A", Double.NaN
		);
		source.handle(
			new String[]{ "H", "LZ", "L" }, "Y2008", "A", Double.NaN
		);
		source.handle( new String[]{ "H", "LZ", "L" }, "Y2009", "A", 82 );

		source.handle( new String[]{ "Q", "LZ", "M" }, "Y2006", "Q3", 0.5 );
		source.handle( new String[]{ "Q", "LZ", "M" }, "Y2006", "Q4", 10 );
		source.handle( new String[]{ "Q", "LZ", "M" }, "Y2007", "Q1", 20 );
		source.handle( new String[]{ "Q", "LZ", "M" }, "Y2007", "Q2", 30 );
		source.handle( new String[]{ "Q", "LZ", "M" }, "Y2007", "Q3", 40 );
		source.handle( new String[]{ "Q", "LZ", "M" }, "Y2007", "Q4", 50 );
		source.handle( new String[]{ "Q", "LZ", "M" }, "Y2008", "Q1", 60 );
		source.handle(
			new String[]{ "Q", "LZ", "M" }, "Y2008", "Q4", Double.NaN
		);
		source.handle(
			new String[]{ "Q", "RZ", "M" }, "Y2008", "Q3", Double.NaN
		);

		source.close();
		listener.close();

		final Set expected = new HashSet();
		expected.add( INTERCHANGE_BEGIN );
		expected.add( MESSAGE_BEGIN );
		expected.add( new TimeSeries(
			"H.LZ.A", "20062009", "702", new Observation[]{
				new Observation( "0.0", "A", "N", "" ),
				new Observation( "1.0", "A", "N", "" ),
				new Observation( "2.0", "A", "N", "" ),
				new Observation( "3.0", "A", "N", "" )
			}
		) );
		expected.add( new TimeSeries(
			"H.LZ.B", "20062009", "702", new Observation[]{
				new Observation(
					Observation.MISSING_VALUE, "S", "N", ""
				),
				new Observation( "2.125", "A", "N", "" ),
				new Observation( "1.0625", "A", "N", "" ),
				new Observation(
					Observation.MISSING_VALUE, "S", "N", ""
				)
			}
		) );
		expected.add( new TimeSeries(
			"H.LZ.L", "20062009", "702", new Observation[]{
				new Observation( "81.0", "A", "N", "" ),
				new Observation(
					Observation.MISSING_VALUE, "S", "N", ""
				),
				new Observation(
					Observation.MISSING_VALUE, "S", "N", ""
				),
				new Observation( "82.0", "A", "N", "" )
			}
		) );
		expected.add( new TimeSeries(
			"Q.LZ.M", "2006320084", "708", new Observation[]{
				new Observation( "0.5", "A", "N", "" ),
				new Observation( "10.0", "A", "N", "" ),
				new Observation( "20.0", "A", "N", "" ),
				new Observation( "30.0", "A", "N", "" ),
				new Observation( "40.0", "A", "N", "" ),
				new Observation( "50.0", "A", "N", "" ),
				new Observation( "60.0", "A", "N", "" ),
				Observation.EMPTY_OBSERVATION,
				Observation.EMPTY_OBSERVATION,
				new Observation(
					Observation.MISSING_VALUE, "S", "N", ""
				)
			}
		) );
		expected.add( new TimeSeries(
			"Q.RZ.M", "20083", "608", new Observation[]{
				new Observation(
					Observation.MISSING_VALUE, "S", "N", ""
				)
			}
		) );
		expected.add( MESSAGE_END );
		expected.add( INTERCHANGE_END );

		for (
			final Iterator iterator = listener.iterator();
			iterator.hasNext();
		) {
			final Object received = iterator.next();
			assertTrue(
				"Received " + received + " was not in " + expected,
				expected.remove( received )
			);
		}

		assertTrue( "Received all expected", expected.isEmpty() );
	}

	public void testZeroCacheLimit () {
		System.setProperty( objectName + ".maxCache", "0" );
		final FunctionInvokedSource source
		= new FunctionInvokedSource( "Hoizbua" );
		final TestEventListener listener = new TestEventListener(
			new Object[]{
				INTERCHANGE_BEGIN,
				MESSAGE_BEGIN,
				new TimeSeries(
					"H.LZ.A", "20071", "608", new Observation[]{
						new Observation( "41.0", "A", "N", "" )
					}
				),
				new TimeSeries(
					"H.LZ.A", "20072", "608", new Observation[]{
						new Observation(
							Observation.MISSING_VALUE, "S", "N", ""
						)
					}
				),
				new TimeSeries(
					"H.LZ.A", "20081", "608", new Observation[]{
						new Observation( "49.5", "A", "N", "" )
					}
				),
				MESSAGE_END,
				INTERCHANGE_END
			}
		);

		source.addListener( listener, true );

		source.handle( new String[]{ "H", "LZ", "A" }, "Y2007", "Q1", 41 );
		source.handle(
			new String[]{ "H", "LZ", "A" }, "Y2007", "Q2", Double.NaN
		);
		source.handle( new String[]{ "H", "LZ", "A" }, "Y2008", "Q1", 49.5 );

		source.close();
		listener.close();
	}

	public void testBeginEnd () {
		final FunctionInvokedSource source
		= new FunctionInvokedSource( "Hoizbua" );
		final TestEventListener listener = new TestEventListener(
			new Object[]{
				INTERCHANGE_BEGIN,
				MESSAGE_BEGIN,
				new TimeSeries(
					"H.LZ.B", "20062009", "702", new Observation[]{
						new Observation( "12.25", "A", "N", "" ),
						new Observation( "128.0", "A", "N", "" ),
						new Observation( "900.0", "A", "N", "" ),
						new Observation(
							Observation.MISSING_VALUE, "S", "N", ""
						)
					}
				),
				MESSAGE_END,
				INTERCHANGE_END
			}
		);

		source.addListener( listener, true );

		source.handle( new String[]{ "H", "LZ", "B" }, "Y2005", "A", 1.5 );
		source.handle( new String[]{ "H", "LZ", "B" }, "Y2006", "A", 12.25 );
		source.handle( new String[]{ "H", "LZ", "B" }, "Y2007", "A", 128 );
		source.handle( new String[]{ "H", "LZ", "B" }, "Y2008", "A", 900 );
		source.handle(
			new String[]{ "H", "LZ", "B" }, "Y2009", "A", Double.NaN
		);
		source.handle( new String[]{ "H", "LZ", "B" }, "Y2010", "A", 19 );
		source.handle( new String[]{ "H", "LZ", "B" }, "Y2011", "A", 0.01 );

		source.close();
		listener.close();
	}
}
