// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2007, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.event;

import java.util.Date;
import java.util.List;
import java.util.ArrayList;

import java.text.SimpleDateFormat;
import java.text.NumberFormat;
import java.text.DecimalFormat;
import java.text.ParseException;

import Int.ecb.logging.Level;

import Int.ecb.gesmescb.InterchangeBegin;
import Int.ecb.gesmescb.InterchangeEnd;
import Int.ecb.gesmescb.MessageBegin;
import Int.ecb.gesmescb.MessageEnd;
import Int.ecb.gesmescb.Contact;
import Int.ecb.gesmescb.ContactChannel;

import Int.ecb.gesmescb.event.GESMESCBEventSourceAdapter;

import Int.ecb.gesmescb.util.Named;
import Int.ecb.gesmescb.util.ComponentHelper;

import Int.ecb.gesmescb.time.RegularInterval;
import Int.ecb.gesmescb.time.GregorianRegularInterval;

import Int.ecb.gesmescb.xml.XMLDateFormat;

/* ***************************** Class ****************************** */

/**
 * Gets interchange and message constant settings from the
 * preferences.
 * This is a useful adapter to read from non-GESMES/TS model data
 * streams which do not themselves provide the information,
 * such as flat files.
 *
 * <center><table border="3" cellpadding="3">
 * <caption>
 * <a name="prefs">Available {@link Int.ecb.prefs preference} settings</a>
 * </caption>
 * <tr>
 * <th>preference name</th>
 * <th>type</th>
 * <th>meaning</th>
 * <th>default</th>
 * </tr>
 * <tr>
 *	<td><code>iref</code></td><td>string</td>
 *	<td>interchange reference number</td>
 *	<td><code>IREF000000</code></td>
 * </tr>
 * <tr>
 *	<td><code>sender</code></td><td>string</td>
 *	<td>sender organisation code</td>
 *	<td><code>ZZZ</code></td>
 * </tr>
 * <tr>
 *	<td><code>receiver</code></td><td>string</td>
 *	<td>receiver organisation code</td>
 *	<td><code>ZZZ</code></td>
 * </tr>
 * <tr>
 *	<td><code>prepared</code></td><td>dateTime</td>
 *	<td>preparation time stamp <code>yyyy-mm-ddThh:mm:ss</code></td>
 *	<td>current time stamp</td>
 * </tr>
 * <tr>
 *	<td><code>test</code></td><td>boolean</td>
 *	<td>whether to flag data as test data</td>
 *	<td><code>false</code></td>
 * </tr>
 * <tr>
 *	<td><code>mref</code></td><td>string</td>
 *	<td>message reference number</td>
 *	<td><code>MREF000001</code></td>
 * </tr>
 * <tr>
 *	<td><code>subject</code></td><td>string</td>
 *	<td>message subject</td>
 *	<td>empty</td>
 * </tr>
 * <tr>
 *	<td><code>contact.<i>i</i>.name</code></td><td>string</td>
 *	<td><i>i</i>-th contact name</td>
 *	<td>none</td>
 * </tr>
 * <tr>
 *	<td><code>contact.<i>i</i>.area</code></td><td>string</td>
 *	<td><i>i</i>-th contact business area code</td>
 *	<td>none</td>
 * </tr>
 * <tr>
 *	<td><code>contact.<i>i</i>.role</code></td>
 *   <td>one of
 *    <code>info_prod</code> (responsible for information production),
 *    <code>head_info_prod</code> (head of information production),
 *    <code>data_proc</code> (responsible for data processing),
 *    <code>head_data_proc</code> (head of data processing)
 *	<td><i>i</i>-th contact role</td>
 *	<td>none</td>
 * </tr>
 * <tr>
 *	<td><code>contact.<i>i</i>.channel.<i>j</i>.type</code></td>
 *  <td>one of <code>telephone</code>, <code>fax</code>,
 *   <code>email</code>, <code>x400</code></td>
 *	<td><i>i</i>-th contact communication channel <i>j</i> type</td>
 *	<td>none</td>
 * </tr>
 * <tr>
 *	<td><code>contact.<i>i</i>.channel.<i>j</i>.address</code></td>
 *  <td>string</td>
 *	<td>none</td>
 * </tr>
 * <tr>
 *	<td><code>maintenanceAgency</code></td><td>string</td>
 *	<td>the key family maintenance agency</td>
 *	<td><code>ECB</code></td>
 * </tr>
 * <tr>
 *	<td><code>keyFamily</code></td><td>string</td>
 *	<td>the identifier of the key family structuring the data</td>
 *	<td><code>ECB_FCT1</code></td>
 * </tr>
 * <tr>
 *	<td><code>dataSet</code></td><td>string</td>
 *	<td>the identifier of the data set</td>
 *	<td><code>ECB_SPF</code></td>
 * </tr>
 * <tr>
 *	<td><code>extracted</code></td><td>dateTime</td>
 *	<td>data extraction time stamp <code>yyyy-mm-ddThh:mm:ss</code></td>
 *	<td>current time stamp</td>
 * </tr>
 * <tr>
 *	<td><code>reportingBegin</code></td><td>dateTime</td>
 *	<td>begin of reporting period <code>yyyy-mm-ddThh:mm:ss</code></td>
 *	<td>none</td>
 * </tr>
 * <tr>
 *	<td><code>reportingEnd</code></td><td>dateTime</td>
 *	<td>end of reporting period <code>yyyy-mm-ddThh:mm:ss</code></td>
 *	<td>none</td>
 * </tr>
 * </table></center>
 *
 * @author
 * <a href="mailto:sis.external@ecb.int">Bernhard Bodenstorfer</a>
 *
 * @version @(#)$Id: ConfiguredSourceAdapter.java,v 1.1 2017/01/19 16:14:04 trainot Exp $
 */
public class ConfiguredSourceAdapter extends GESMESCBEventSourceAdapter
	implements Named
{
	/* ========================= Fields ========================= */

	/** Takes care of logging and preferences. */
	protected final ComponentHelper componentHelper;

	/* ====================== Constructors ====================== */

	/**
	 * Construct a <code>ConfiguredSourceAdapter</code> to read from a file.
	 *
	 * @param name used for logging and preferences.
	 */
	protected ConfiguredSourceAdapter ( final String name )
	{
		componentHelper = new ComponentHelper( name, this );
	}

	/* ===================== Public Methods ===================== */

	public String getName () {
		return componentHelper.name;
	}

	/* =================== Protected Methods ==================== */

	/** Retrieves the interchange reference number.
	 *
	 * @param defaultInterchangeNumber specifies the default
	 * to be chosen if no configuration is found.
	 *
	 * The default implementation reads from the preferences.
	 * Override this method to change the behaviour.
	 */
	protected String getIref ( final int defaultInterchangeNumber ) {
		final NumberFormat irefFormat = new DecimalFormat( "000000" );
		return componentHelper.getPreference(
			"iref",
			"IREF" + irefFormat.format( defaultInterchangeNumber )
		);
	}

	/** Retrieves the sender organisation code.
	 *
	 * The default implementation reads from the preferences.
	 * Override this method to change the behaviour.
	 */
	protected String getSender () {
		return componentHelper.getPreference( "sender", "ZZZ" );
	}

	/** Retrieves the sender contact information.
	 *
	 * The default implementation reads from the preferences.
	 * Override this method to change the behaviour.
	 */
	protected Contact[] getContacts () {
		final List contacts = new ArrayList();
		for ( int i = 1; ; ++i ) {
			final String name = componentHelper.getPreference(
				"contact." + i + ".name", null
			);
			if ( null == name ) break;
			final String area = componentHelper.getPreference(
				"contact." + i + ".area", null
			);
			final String role = componentHelper.getPreference(
				"contact." + i + ".role", null
			);
			final String function;
			if ( "info_prod".equalsIgnoreCase( role ) ) {
				function = Contact.INFORMATION_PRODUCTION;
			} else if ( "data_proc".equalsIgnoreCase( role ) ) {
				function = Contact.DATA_PROCESSING;
			} else if ( "head_info_prod".equalsIgnoreCase( role ) ) {
				function = Contact.HEAD_OF_INFORMATION_PRODUCTION;
			} else if ( "head_data_proc".equalsIgnoreCase( role ) ) {
				function = Contact.HEAD_OF_DATA_PROCESSING;
			} else {
				componentHelper.logger.log(
					Level.WARNING,
					componentHelper.logName,
					"getContacts()",
					"Contact role \"" + role
					+ "\" is not known."
				);
				function = "";
			}

			final List channels = new ArrayList();
			for ( int j = 1; ; ++j ) {
				final String address = componentHelper.getPreference(
					"contact." + i + ".channel." + j + ".address", null
				);
				if ( null == address ) {
					break;
				}

				final String means = componentHelper.getPreference(
					"contact." + i + ".channel." + j + ".type", null
				);
				final String type;
				if ( "telephone".equalsIgnoreCase( means ) ) {
					type = ContactChannel.TELEPHONE;
				} else if ( "fax".equalsIgnoreCase( means ) ) {
					type = ContactChannel.FAX;
				} else if ( "x400".equalsIgnoreCase( means ) ) {
					type = ContactChannel.X400;
				} else if ( "email".equalsIgnoreCase( means ) ) {
					type = ContactChannel.EMAIL;
				} else {
					componentHelper.logger.log(
						Level.WARNING,
						componentHelper.logName,
						"getContacts()",
						"Contact channel type \"" + means
						+ "\" is not known."
					);
					type = "";
				}
				channels.add( new ContactChannel( type, address ) );
			}
			contacts.add( new Contact( function, area, name, channels ) );
		}
		return (Contact[]) contacts.toArray( new Contact[0] );
	}

	/** Retrieves the receiver organisation code.
	 *
	 * The default implementation reads from the preferences.
	 * Override this method to change the behaviour.
	 */
	protected String getReceiver () {
		return componentHelper.getPreference( "receiver", "ZZZ" );
	}

	/** Retrieves the preparation time stamp.
	 *
	 * @param defaultPrepared to be used when nothing is configured.
	 *
	 * The default implementation reads from the preferences.
	 * Override this method to change the behaviour.
	 */
	protected Date getPrepared ( final Date defaultPrepared ) {
		final String preparedString = componentHelper.getPreference(
			"prepared", null
		);
		if ( null != preparedString ) try {
			return new XMLDateFormat().parse( preparedString );
		} catch ( final ParseException pe ) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"getPrepared()",
				"Configured preparation time stamp \""
				+ preparedString
				+ "\" is not in required ISO dateTime format.",
				pe
			);
		}
		return defaultPrepared;
	}

	/** Retrieves whether the interchange should be flagged as test only.
	 *
	 * The default implementation reads from the preferences.
	 * Override this method to change the behaviour.
	 */
	protected boolean getTest () {
		final String testFlag
		= componentHelper.getPreference( "test", "false" );
		if ( "true".equalsIgnoreCase( testFlag ) ) {
			return true;
		} else if ( "false".equalsIgnoreCase( testFlag ) ) {
			return false;
		} else {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"getTest()",
				"Configured test flag \""
				+ testFlag
				+ "\" is neither \"true\" nor \"false\".",
				testFlag
			);
			return true;
		}
	}

	/** Sends an <code>InterchangeBegin</code> according to
	 * what the relevant <code>get</code> functions return.
	 *
	 * @param defaultInterchangeNumber specifies the default reference number.
	 * @param defaultPrepared to be used when no preparation time stamp
	 * is configured.
	 */
	protected void dispatchInterchangeBegin (
		final int defaultInterchangeNumber,
		final Date defaultPrepared
	) {
		final InterchangeBegin interchangeBegin = new InterchangeBegin(
			getIref( defaultInterchangeNumber ),
			getSender(), getReceiver(),
			getPrepared( defaultPrepared ),
			getTest()
		);
		componentHelper.logger.log(
			Level.FINE,
			componentHelper.logName,
			"dispatchInterchangeBegin()",
			"Dispatching InterchangeBegin",
			interchangeBegin
		);
		dispatch( interchangeBegin );
	}

	/** Sends an <code>InterchangeEnd</code> according to
	 * what the relevant <code>get</code> functions return.
	 *
	 * @param defaultInterchangeNumber specifies the default reference number.
	 */
	protected void dispatchInterchangeEnd (
		final int defaultInterchangeNumber
	) {
		final InterchangeEnd interchangeEnd = new InterchangeEnd(
			getIref( defaultInterchangeNumber )
		);
		componentHelper.logger.log(
			Level.FINE,
			componentHelper.logName,
			"dispatchInterchangeEnd()",
			"Dispatching InterchangeEnd",
			interchangeEnd
		);
		dispatch( interchangeEnd );
	}

	/** Retrieves the message reference number.
	 *
	 * @param defaultMessageNumber specifies the default number
	 * to be chosen if no configuration is found.
	 *
	 * The default implementation reads from the preferences.
	 * Override this method to change the behaviour.
	 */
	protected String getMref ( int defaultMessageNumber ) {
		final NumberFormat mrefFormat = new DecimalFormat( "000000" );
		return componentHelper.getPreference(
			"mref",
			"MREF" + mrefFormat.format( defaultMessageNumber )
		);
	}

	/** Retrieves the message subject.
	 *
	 * The default implementation reads from the preferences.
	 * Override this method to change the behaviour.
	 */
	protected String getSubject () {
		return componentHelper.getPreference(
			"subject", ""
		);
	}

	/** Retrieves the key family maintenance agency identifier.
	 *
	 * The default implementation reads from the preferences.
	 * Override this method to change the behaviour.
	 */
	protected String getMaintenanceAgencyId () {
		final String maintenanceAgencyId = componentHelper.getPreference(
			"maintenanceAgency", null
		);
		if ( null == maintenanceAgencyId ) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"getMaintenanceAgencyId()",
				"No maintenance agency identifier configured."
			);
		}
		return null != maintenanceAgencyId
			? maintenanceAgencyId : "";
	}

	/** Retrieves the key family identifier.
	 *
	 * The default implementation reads from the preferences.
	 * Override this method to change the behaviour.
	 */
	protected String getKeyFamilyId () {
		final String keyFamilyId = componentHelper.getPreference(
			"keyFamily", null
		);
		if ( null == keyFamilyId ) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"getKeyFamilyId()",
				"No key family identifier configured."
			);
		}
		return null != keyFamilyId
			? keyFamilyId : "";
	}

	/** Retrieves the data set identifier.
	 *
	 * The default implementation reads from the preferences.
	 * Override this method to change the behaviour.
	 */
	protected String getDataSetId () {
		final String dataSetId = componentHelper.getPreference(
			"dataSet", null
		);
		if ( null == dataSetId ) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"getDataSetId()",
				"No data set identifier configured."
			);
		}
		return null != dataSetId
			? dataSetId : "";
	}

	/** Retrieves the data extraction time stamp.
	 *
	 * @param defaultExtracted to be used when nothing is configured.
	 *
	 * The default implementation reads from the preferences.
	 * Override this method to change the behaviour.
	 */
	protected Date getExtracted ( final Date defaultExtracted ) {
		final String extractedString = componentHelper.getPreference(
			"extracted", null
		);
		if ( null != extractedString ) try {
			return new XMLDateFormat().parse( extractedString );
		} catch ( final ParseException pe ) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"getExtracted()",
				"Configured extraction time stamp \""
				+ extractedString
				+ "\" is not in required ISO dateTime format.",
				pe
			);
		}
		return defaultExtracted;
	}

	/** Retrieves the reporting interval begin time point.
	 *
	 * The default implementation reads from the preferences.
	 * Override this method to change the behaviour.
	 */
	protected Date getReportingBegin () {
		final String reportingBegin = componentHelper.getPreference(
			"reportingBegin", null
		);
		if ( null == reportingBegin ) {
			componentHelper.logger.log(
				Level.FINER,
				componentHelper.logName,
				"getReportingBegin()",
				"No reporting interval begin time configuration found."
			);
			return null;
		}
		try {
			return new SimpleDateFormat(
				"yyyy-MM-dd'T'HH:mm:ss"
			).parse(
				reportingBegin
			);
		} catch ( final ParseException e ) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"getBegin()",
				"Reporting interval begin time configuration \""
				+ reportingBegin
				+ "\" is not in ISO format yyyy-mm-ddThh:mm:ss.",
				e
			);
			return null;
		}
	}

	/** Retrieves the reporting interval end time point.
	 *
	 * The default implementation reads from the preferences.
	 * Override this method to change the behaviour.
	 */
	protected Date getReportingEnd () {
		final String reportingEnd = componentHelper.getPreference(
			"reportingEnd", null
		);
		if ( null == reportingEnd ) {
			componentHelper.logger.log(
				Level.FINER,
				componentHelper.logName,
				"getReportingEnd()",
				"No reporting interval end time configuration found."
			);
			return null;
		}
		try {
			return new SimpleDateFormat(
				"yyyy-MM-dd'T'HH:mm:ss"
			).parse(
				reportingEnd
			);
		} catch ( final ParseException e ) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"getReportingEnd()",
				"Reporting interval end time configuration \""
				+ reportingEnd
				+ "\" is not in ISO format yyyy-mm-ddThh:mm:ss.",
				e
			);
			return null;
		}
	}

	/** Retrieves the reporting interval based on
	 * {@link #getReportingBegin} and {@link #getReportingEnd}.
	 *
	 * The default implementation reads from the preferences.
	 * Override this method to change the behaviour.
	 */
	protected RegularInterval getReportingInterval () {
		final Date
			reportingBegin = getReportingBegin(),
			reportingEnd = getReportingEnd();
		return reportingBegin != null && reportingEnd != null
			? GregorianRegularInterval.searchCover(
				reportingBegin, reportingEnd, true
			) : null;
	}

	/** Sends a <code>MessageBegin</code> according to
	 * what the relevant <code>get</code> functions return.
	 *
	 * @param defaultMessageNumber specifies the default reference number.
	 */
	protected void dispatchMessageBegin (
		final int defaultMessageNumber,
		final Date defaultExtracted
	) {
		final MessageBegin messageBegin = new MessageBegin(
			getMref( defaultMessageNumber ),
			getSender(), getReceiver(),
			getMaintenanceAgencyId(),
			MessageBegin.STATISTICAL_DATA,
			getSubject(),
			getContacts(),
			getDataSetId(),
			MessageBegin.UPDATE_OR_REPLACE_ACTION,
			getExtracted( defaultExtracted ),
			getReportingInterval(),
			getKeyFamilyId()
		);
		componentHelper.logger.log(
			Level.FINE,
			componentHelper.logName,
			"dispatchMessageBegin()",
			"Dispatching MessageBegin",
			messageBegin
		);
		dispatch( messageBegin );
	}

	/** Sends an <code>MessageEnd</code> according to
	 * what the relevant <code>get</code> functions return.
	 *
	 * @param defaultMessageNumber specifies the default reference number.
	 */
	protected void dispatchMessageEnd ( final int defaultMessageNumber ) {
		final MessageEnd messageEnd = new MessageEnd(
			getMref( defaultMessageNumber )
		);
		componentHelper.logger.log(
			Level.FINE,
			componentHelper.logName,
			"dispatchMessageEnd()",
			"Dispatching MessageEnd",
			messageEnd
		);
		dispatch( messageEnd );
	}
}
