// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2007, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.event;

import java.util.Date;

import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.text.ParseException;

import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;

import Int.ecb.gesmescb.InterchangeBegin;
import Int.ecb.gesmescb.InterchangeEnd;
import Int.ecb.gesmescb.MessageBegin;
import Int.ecb.gesmescb.MessageEnd;
import Int.ecb.gesmescb.Contact;
import Int.ecb.gesmescb.ContactChannel;

import Int.ecb.gesmescb.time.RegularInterval;
import Int.ecb.gesmescb.time.GregorianRegularInterval;

import Int.ecb.gesmescb.event.TestEventListener;

/* ***************************** Class ****************************** */

/**
 * Implements a few tests for the {@link ConfiguredSourceAdapter}.
 *
 * @author
 * <a href="mailto:sis.external@ecb.int">Bernhard Bodenstorfer</a>
 *
 * @version @(#)$Id: ConfiguredSourceAdapterTest.java,v 1.1 2017/01/19 16:14:03 trainot Exp $
 */
public class ConfiguredSourceAdapterTest extends TestCase {

	/* ======================= Constants ======================== */

	private static final String
		objectName = ConfiguredSourceAdapter.class.getName() + ".Christina";

	private static final String[][] objectProperties = {
		{ "iref", "IREF070035" },
		{ "sender", "GR7" },
		{ "contact.1.name", "Hugo" },
		{ "contact.1.area", "terrestric" },
		{ "contact.1.role", "info_prod" },
		{ "contact.1.channel.1.type", "telephone" },
		{ "contact.1.channel.1.address", "4339" },
		{ "contact.1.channel.2.type", "fax" },
		{ "contact.1.channel.2.address", "4994" },
		{ "contact.2.name", "Berlingo" },
		{ "contact.2.area", "spatial" },
		{ "contact.2.role", "head_info_prod" },
		{ "contact.2.channel.1.type", "x400" },
		{ "contact.2.channel.1.address", "CN=MOON" },
		{ "contact.3.name", "Varus" },
		{ "contact.3.area", "empire" },
		{ "contact.3.role", "data_proc" },
		{ "contact.3.channel.1.type", "email" },
		{ "contact.3.channel.1.address", "varus@rome.net" },
		{ "contact.4.name", "Lepidus" },
		{ "contact.4.area", "Afrique" },
		{ "contact.4.role", "head_data_proc" },
		{ "receiver", "BBQ" },
		{ "prepared", "1833-01-13T19:34:00" },
		{ "test", "true" },
		{ "mref", "MREF000486" },
		{ "subject", "Configured subject" },
		{ "maintenanceAgency", "GEZ" },
		{ "keyFamily", "GEZ_DARK" },
		{ "dataSet", "GEZ_BLACK" },
		{ "extracted", "1755-12-24T23:00:00" },
		{ "reportingBegin", "1612-01-01T00:00:00" },
		{ "reportingEnd", "1632-01-01T00:00:00" }
	};

	/* ====================== Constructors ====================== */

	public ConfiguredSourceAdapterTest ( final String name ) {
		super( name );
	}

	/* ================== Protected Methods ===================== */

	/** Sets up the test fixture. */
	protected void setUp () {
		for ( int i = 0; i < objectProperties.length; ++i ) {
			System.setProperty(
				objectName + "." + objectProperties[i][0],
				objectProperties[i][1]
			);
		}
	}

	/** Tears down the test fixture. */
	protected void tearDown () {
		for ( int i = 0; i < objectProperties.length; ++i ) {
			System.getProperties().remove(
				objectName + "." + objectProperties[i][0]
			);
		}
	}

	/** Creates an instance to be tested. */
	protected ConfiguredSourceAdapter createInstance( final String name ) {
		return new ConfiguredSourceAdapter( name );
	}

	/* ===================== Public Methods ===================== */

	/**
	 * Assembles and returns a test suite for all
	 * the test methods of this test case.
	 *
	 * @return a non-null test suite
	 */
	public static Test suite () {
		return new TestSuite( ConfiguredSourceAdapterTest.class );
	}

	/**
	 * @param arg the array of argument {@link String}s
	 */
	public static void main ( final String[] arg ) {
		final String[] testCaseName = {
			ConfiguredSourceAdapterTest.class.getName()
		};
		junit.textui.TestRunner.main( testCaseName );
	}

	/* ---------------------- Test Methods ---------------------- */

	/** Test the case that all is configured. */
	public void testAllConfigured () throws ParseException {
		final DateFormat format
		= new SimpleDateFormat( "yyyy-MM-dd'T'HH:mm:ss" );
		final ConfiguredSourceAdapter adapter = createInstance( "Christina" );

		assertEquals( "iref", "IREF070035", adapter.getIref( 0 ) );
		assertEquals( "sender", "GR7", adapter.getSender() );
		assertEquals( "receiver", "BBQ", adapter.getReceiver() );
		assertEquals(
			"prepared",
			format.parse( "1833-01-13T19:34:00" ),
			adapter.getPrepared( new Date( 0 ) )
		);
		assertEquals( "test", true, adapter.getTest() );

		assertEquals( "mref", "MREF000486", adapter.getMref( 1 ) );
		assertEquals( "subject", "Configured subject", adapter.getSubject() );
		assertEquals(
			"maintenance agency",
			"GEZ",
			adapter.getMaintenanceAgencyId()
		);
		assertEquals( "key family", "GEZ_DARK", adapter.getKeyFamilyId() );
		assertEquals( "data set", "GEZ_BLACK", adapter.getDataSetId() );
		assertEquals(
			"extracted",
			format.parse( "1755-12-24T23:00:00" ),
			adapter.getExtracted( new Date( 0 ) )
		);
		assertEquals(
			"reporting period",
			new GregorianRegularInterval(
				RegularInterval.ANNUAL,
				format.parse( "1612-01-01T00:00:00" ),
				20
			),
			adapter.getReportingInterval()
		);

		final TestEventListener listener = new TestEventListener(
			new Object[]{
				new InterchangeBegin(
					"IREF070035",
					"GR7",
					"BBQ",
					new SimpleDateFormat( "yyyy-MM-dd'T'HH:mm:ss" )
						.parse( "1833-01-13T19:34:00" ),
					true
				),
				new MessageBegin(
					"MREF000486", "GR7", "BBQ", "GEZ",
					MessageBegin.STATISTICAL_DATA, "Configured subject",
					new Contact[]{
						new Contact(
							Contact.INFORMATION_PRODUCTION,
							"terrestric", "Hugo",
							new ContactChannel[]{
								new ContactChannel(
									ContactChannel.TELEPHONE,
									"4339"
								),
								new ContactChannel(
									ContactChannel.FAX,
									"4994"
								)
							}
						),
						new Contact(
							Contact.HEAD_OF_INFORMATION_PRODUCTION,
							"spatial", "Berlingo",
							new ContactChannel[]{
								new ContactChannel(
									ContactChannel.X400,
									"CN=MOON"
								)
							}
						),
						new Contact(
							Contact.DATA_PROCESSING,
							"empire", "Varus",
							new ContactChannel[]{
								new ContactChannel(
									ContactChannel.EMAIL,
									"varus@rome.net"
								)
							}
						),
						new Contact(
							Contact.HEAD_OF_DATA_PROCESSING,
							"Afrique", "Lepidus",
							new ContactChannel[0]
						)
					},
					"GEZ_BLACK", MessageBegin.UPDATE_OR_REPLACE_ACTION,
					"175512242300", "16121631", "702", "GEZ_DARK"
				),
				new MessageEnd( "MREF000486" ),
				new InterchangeEnd( "IREF070035" )
			}
		);
		adapter.addListener( listener, true );
		adapter.dispatchInterchangeBegin( 5, null );
		adapter.dispatchMessageBegin( 8, null );
		adapter.dispatchMessageEnd( 9 );
		adapter.dispatchInterchangeEnd( 10 );
		listener.close();
	}

	/** Test the case that nothing is configured. */
	public void testNothingConfigured () throws ParseException {
		final ConfiguredSourceAdapter adapter = createInstance( "Anonymus" );

		assertEquals( "iref", "IREF000016", adapter.getIref( 16 ) );
		assertEquals( "sender", "ZZZ", adapter.getSender() );
		assertEquals( "receiver", "ZZZ", adapter.getReceiver() );
		assertEquals(
			"prepared",
			new Date( 5 ),
			adapter.getPrepared( new Date( 5 ) )
		);
		assertEquals( "test", false, adapter.getTest() );

		assertEquals( "mref", "MREF002000", adapter.getMref( 2000 ) );
		assertEquals( "subject", "", adapter.getSubject() );
		assertEquals(
			"maintenance agency",
			"",
			adapter.getMaintenanceAgencyId()
		);
		assertEquals( "key family", "", adapter.getKeyFamilyId() );
		assertEquals( "data set", "", adapter.getDataSetId() );
		assertEquals(
			"extracted",
			new Date( 9 ),
			adapter.getExtracted( new Date( 9 ) )
		);
		assertEquals(
			"reporting period",
			null,
			adapter.getReportingInterval()
		);

		final TestEventListener listener = new TestEventListener(
			new Object[]{
				new InterchangeBegin(
					"IREF000505", "ZZZ", "ZZZ",
					"", false
				),
				new MessageBegin(
					"MREF000508", "ZZZ", "ZZZ", "",
					MessageBegin.STATISTICAL_DATA, "",
					new Contact[]{},
					"", MessageBegin.UPDATE_OR_REPLACE_ACTION,
					"", "", "", ""
				),
				new MessageEnd( "MREF000509" ),
				new InterchangeEnd( "IREF000510" )
			}
		);
		adapter.addListener( listener, true );
		adapter.dispatchInterchangeBegin( 505, null );
		adapter.dispatchMessageBegin( 508, null );
		adapter.dispatchMessageEnd( 509 );
		adapter.dispatchInterchangeEnd( 510 );
		listener.close();
	}
}
