// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2008, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.event;

import java.util.TreeMap;
import java.util.Iterator;
import java.util.Date;
import java.text.DateFormat;
import java.text.SimpleDateFormat;

import Int.ecb.gesmescb.AbstractCube;
import Int.ecb.gesmescb.TimeSeries;
import Int.ecb.gesmescb.Attribute;
import Int.ecb.gesmescb.InterchangeBegin;
import Int.ecb.gesmescb.InterchangeEnd;
import Int.ecb.gesmescb.MessageBegin;
import Int.ecb.gesmescb.MessageEnd;
import Int.ecb.gesmescb.Observation;

import Int.ecb.gesmescb.util.ComponentHelper;
import Int.ecb.gesmescb.util.ObservationMap;

import Int.ecb.gesmescb.time.RegularInterval;
import Int.ecb.gesmescb.time.GregorianRegularInterval;
import Int.ecb.logging.Level;

/* ***************************** Class ****************************** */

/**
 * <p>Gathers {@link TimeSeries} ( typically, those rejected by 
 * an {@link ObservationSieve} or {@link Attribute}
 * and provide iterators to work on them later on.
 * </p>
 * <p>For simplicity, DataCollector is implemented as a Singleton.
 * </p>
 * @see ObservationSieve
 * @see EventSieve.
 *
 * @author
 * <A HREF="mailto:Francois.Hoeltzli@ECB.int">Francois Hoeltzli</A>
 *
 * @version @(#)$Id: DataCollector.java,v 1.1 2017/01/19 16:14:03 trainot Exp $
 */

public class DataCollector implements DataListener {
	/* ========================= Fields ========================= */

	/** dataCollector is a Singleton. */
	private static DataCollector instance;
	
	/** Takes care of preferences, settings, log. */
	private final ComponentHelper componentHelper;
	
	/** Memorises the {@link TimeSeries} have come in. */
	private final TreeMap timeSeriesMap;

	/** Memorises the {@link Attribute} that have come in. */
	private final TreeMap attributeMap;

	protected static final DateFormat FORMAT = new SimpleDateFormat( "yyyyMMddHHmmss" );

	/* ====================== Constructors ====================== */
	
	private DataCollector ( String name ) {
		componentHelper = new ComponentHelper( name, this );	
		timeSeriesMap = new TreeMap();
		attributeMap = new TreeMap();
	}
	
	private DataCollector () {
		this ( "" );
	}

	/* ===================== Public Methods ===================== */
	
	public static DataCollector getDataCollector() {
		if ( DataCollector.instance == null ) {
			DataCollector.instance = new DataCollector();
		}
		return DataCollector.instance;
	}
	
	public void clear() {
		clearTimeSeries();
		clearAttributes();
	}

	public void clearTimeSeries() {
		if ( DataCollector.instance != null ) {
			timeSeriesMap.clear();
		}
	}
	public void clearAttributes() {
		if ( DataCollector.instance != null ) {
			attributeMap.clear();
		}
	}
	
	/**
	 * Receive and process a {@link TimeSeries}
	 * @return false if no longer listening.
	 */
	public void handle ( TimeSeries timeSeries ) {
		putTimeSeries( timeSeries );	
	}

	/**
	 * Receive and process an {@link Attribute}
	 * @return false if no longer listening.
	 */
	public void handle ( Attribute attribute ) {
		putAttribute( attribute );
	};
	
	/**
	 * Receive and process an {@link Error}
	 * Does nothing in the current implemtation.
	 */
	public void handle ( ErrorEvent errorEvent ) {
		return;
	};

	/**
	 * Receive and process an {@link MessageBegin}
	 * Does nothing in the current implemtation.
	 */
	public void handle ( MessageBegin messageBegin ) {
		return;
	};

	/**
	 * Receive and process an {@link MessageEnd}
	 * Does nothing in the current implemtation.
	 */
	public void handle ( MessageEnd messageEnd ) {
		return;
	};

	/**
	 * Receive and process an {@link InterchangeBegin}
	 * Does nothing in the current implemtation.
	 */
	public void handle ( InterchangeBegin interchangeBegin ) {
		return;
	};

	/**
	 * Receive and process an {@link InterchangeEnd}
	 * Does nothing in the current implemtation.
	 */
	public void handle ( InterchangeEnd interchangeEnd ) {
		return;
	};

	/**
	 * State always that the connection isActive.
	 */
	public boolean isActive () {
		return true;
	};

	/**
	 * Returns an Iterator over all the keys of the {@link TimeSeries}
	 * that have been collected.
	 */
	public Iterator timeSeriesIterator() {
		return timeSeriesMap.keySet().iterator();
	};

	/**
	 * Return an {@link ObservationMap} over all observations for
	 * the time series whose key is passed.
	 */
	public ObservationMap getObservationMap( final String timeSeriesKey ) {
		return timeSeriesMap.containsKey( timeSeriesKey )
			? (ObservationMap) timeSeriesMap.get( timeSeriesKey )
			: null;
	};

	/**
     * Returns an Iterator over all the {@link TimeSeries} contained
	 * in the {@link TimeSeries} and maching the timeSeriesKey.
	 * There can be several timeseries if the gab between some observations
	 * is bigger than Integer.MAX_VALUE
	 */
	public Iterator iterateTimeSeries( final String timeSeriesKey ) {
	    return iterateTimeSeries(
			timeSeriesKey,
			Integer.MAX_VALUE
		);
	};

	/**
     * Returns an Iterator over all the {@link TimeSeries} contained
	 * in the {@link TimeSeries} and maching the timeSeriesKey.
	 * There can be several timeseries if the gab between some observations
	 * is bigger than the specified gap.
	 */
	public Iterator iterateTimeSeries(
		final String timeSeriesKey,
		final int gap
	) {
	    return hasTimeSeries( timeSeriesKey )
	        ? (
	            (ObservationMap) timeSeriesMap.get( timeSeriesKey )
	          ).iterateTimeSeries( timeSeriesKey, gap )
	        : null;
	};

	/**
	 * Return an Iterator over all the {@link Attribute}s keys.
	 */
	public Iterator attributeIterator() {
		return attributeMap.keySet().iterator();
	};
	
	/**
	 * Store a {@link TimeSeries}
	 */
	public void putTimeSeries ( final TimeSeries timeSeries ) {
		ObservationMap observationMap;
		observationMap = timeSeriesMap.containsKey( timeSeries.getKey() )
			? (ObservationMap) timeSeriesMap.get( timeSeries.getKey() )
			: new ObservationMap();
		observationMap.put(timeSeries);
		timeSeriesMap.put(timeSeries.getKey(), observationMap);
		componentHelper.logger.log(
			Level.FINEST, componentHelper.logName, "memoriseObservations",
			"memorising Observations for time series: " + timeSeries
		);
		return;
	};

	/**
	 * Store an {@link Observation} using a daily frequency.
	 */
	public void putObservation (
		final String timeSeriesKey,
		final Date observationDate,
		final Observation observation
	) {
		putObservation (
			timeSeriesKey,
			RegularInterval.DAILY,
			observationDate,
			observation
		);
		return;
	};
	
	/**
	 * Store an {@link Observation}.
	 */
	public void putObservation (
		final String timeSeriesKey,
		int frequency,
		final Date observationDate,
		final Observation observation
	) {
		ObservationMap observationMap;
		observationMap = timeSeriesMap.containsKey( timeSeriesKey )
			? (ObservationMap) timeSeriesMap.get( timeSeriesKey )
			: new ObservationMap();
		RegularInterval regularInterval
		= getOneDateInterval( frequency, observationDate);
		observationMap.put( regularInterval, observation );
		timeSeriesMap.put(timeSeriesKey, observationMap);
		componentHelper.logger.log(
			Level.FINEST, componentHelper.logName, "putObservations",
			"memorising 1 Observation for time series: \""
			+ timeSeriesKey + "\"."
		);
		return;
	};

	/**
	 * Store an {@link Attribute}
	 */
	public void putAttribute ( final Attribute attribute ) {
		String attributeKey = attribute.getKey()
			+ AbstractCube.KEY_SEPARATOR
			+ attribute.getName();
		attributeMap.put( attributeKey, attribute );
		componentHelper.logger.log(
			Level.FINEST, componentHelper.logName, "memoriseAttribute",
			"memorising Attribute: " + attribute
		);
		return;
	};

	/**
	 * Return the first {@link TimeSeries} stored in the TimeSeriesMap
	 * and that correspond to timeSeriesKey.
	 * This series might be truncated should it exhibit a gap bigger
	 * than Integer.MAX_VALUE.
	 */
	public TimeSeries getTimeSeries( final String timeSeriesKey ) {
		if (! hasTimeSeries( timeSeriesKey ) )
			return null;
		ObservationMap observationMap
		= (ObservationMap) timeSeriesMap.get( timeSeriesKey );
		Iterator iterator = observationMap.iterateTimeSeries(
			timeSeriesKey,
			Integer.MAX_VALUE
		);
		return iterator.hasNext()
			? (TimeSeries) iterator.next()
			: null;
	};

	/**
	 * Return an {@link Observation}.
	 */
	public Observation getObservation(
		final String timeSeriesKey,
		final RegularInterval regularInterval
	) {
		/* low level because we lack an efficient hasObservation() */
		if (! timeSeriesMap.containsKey( timeSeriesKey ) )
			return null;
		ObservationMap observationMap
		= (ObservationMap) timeSeriesMap.get( timeSeriesKey );
		Observation observation=null;
		try {
			observation = observationMap.get( regularInterval );
		} catch ( IllegalArgumentException e ) {

			int frequency = regularInterval.getFrequency();
			Date begin = regularInterval.getBegin();
			Date end = regularInterval.getEnd();
			componentHelper.logger.log(
				Level.WARNING, componentHelper.logName, "getObservation",
				"cannot get Observation for "
				+ "series: \"" + timeSeriesKey + "\", "
				+ (regularInterval==null
					? "interval null."
					: "frequency \"" + frequency + "\", "
						+ "begin \"" + begin + "\", "
						+ "end \"" + end + "\"."
					)
			);
		}
		return observation;
	}
	
	/**
	 * Return an {@link Attribute}.
	 */
	public Attribute getAttribute( final String attributeKey ) {
		if (! hasAttribute( attributeKey ) )
			return null;
		return (Attribute) attributeMap.get( attributeKey );
	};

	/**
	 * Tells whether a given {@link TimeSeries} is stored.
	 */
	public boolean hasTimeSeries( final String timeSeriesKey ) {
		return timeSeriesMap.containsKey( timeSeriesKey );
	};

	/**
	 * Tells whether a given {@link Observation} is stored
	 * for a given basePeriod {@link RegularInterval}.
	 */
	public boolean hasObservation( 
		final String timeSeriesKey,
		final RegularInterval interval
	) {
		if (! timeSeriesMap.containsKey( timeSeriesKey ) )
			return false;
		return (
			getObservation( timeSeriesKey, interval )
			!=
			null
		);
	};
	
	/**
	 * Tells whether a given {@link Attribute} is stored.
	 */
	public boolean hasAttribute( final String attributeKey ) {
		return attributeMap.containsKey( attributeKey );
	};

	/* ===================== Private Methods ==================== */

	private RegularInterval getOneDateInterval(
			int frequency,
			Date beginDate
		) {
			RegularInterval regularInterval=null;
			regularInterval = new GregorianRegularInterval(
					frequency, beginDate, 1
			);
			return regularInterval;
		};
		
	
}

