// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, (c) 2003, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.event;

import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.Iterator;
import java.util.Date;
import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;

import Int.ecb.gesmescb.*;
import Int.ecb.gesmescb.event.DataCollector;
import Int.ecb.gesmescb.time.RegularInterval;
import Int.ecb.gesmescb.time.GregorianRegularInterval;

/* **************************** Class ***************************** */

/**
 * @author
 * <A HREF="mailto:francois.hoeltzli@ECB.int">Francois Hoeltzli</A>
 *
 * @version @(#)$Id: DataCollectorTest.java,v 1.1 2017/01/19 16:14:03 trainot Exp $
 */
public class DataCollectorTest extends TestCase
{
	/* ========================= Fields ========================= */
			
	private final TimeSeries arcsineTimeSeries,
		cosineTimeSeries,
		emptyTimeSeries;
	private final Observation o0, o1, o2, o3, o4, o5; 
	protected static final DateFormat FORMAT = new SimpleDateFormat( "yyyyMMddHHmmss" );

	/* ====================== Constructors ====================== */

	/**
	 * Construct a clean and fresh DataCollectorTest.
	 */
	public DataCollectorTest () {
		this( "" );
	}
	
	/**
	 * Constructs a DataCollectorTest with the specified name.
	 *
	 * @param name Test case name
	 */
	public DataCollectorTest ( final String name ) {
		super( name );
		o0 = new Observation( "1", "3", "5", "9" );
		o1 = new Observation( "11", "13", "15", "17" );
		o2 = new Observation( "19", "21", "23", "25" );
		o3 = new Observation( "1", "2", "3", "4" );
		o4 = new Observation( "5", "6", "7", "8" );
		o5 = new Observation( "9", "10", "11", "12" );
		arcsineTimeSeries = new TimeSeries(
			"ARC.SI.NUS", 
			newInterval( "20010901000000", RegularInterval.DAILY, 3),
			new Observation[] {	o0, o1, o2	}
		);
		cosineTimeSeries = new TimeSeries(
			"CO.SI.NUS", "200109200111", "710",
			new Observation[] { o3, o4, o5 }
		);
		emptyTimeSeries = new TimeSeries(
				"EMPTY.TIME.SERIES", "200109200110", "710",
				new Observation[] {}
		);
	}
	
	/* ===================== Public Methods ===================== */
										
	/**
	* Assembles and returns a test suite for all
	* the test methods of this test case.
	*
	* @return a non-null test suite
	*/
	public static Test suite () {
		TestSuite suite = new TestSuite( DataCollectorTest.class );
			return suite;
	}

	/**
	 * Runs the test case.
	 */
 	public static void main ( String args[] ) {
		String[] testCaseName = { DataCollectorTest.class.getName() };
		junit.textui.TestRunner.main( testCaseName );
	}



	public void testEmptyDataCollector() {
		
		DataCollector dataCollector = DataCollector.getDataCollector();
		dataCollector.clear();
		
		Iterator timeSeriesIterator = dataCollector.timeSeriesIterator();
		assertTrue(
			"timeSeriesIterator should not have next",
			!timeSeriesIterator.hasNext()
		);

		Iterator cosineSeriesIterator
		= dataCollector.iterateTimeSeries( cosineTimeSeries.getKey() );
		assertTrue(
			"iterateTimeSeries(cosine) should return null",
			cosineSeriesIterator == null
		);

		Iterator attributeIterator = dataCollector.attributeIterator();
		assertTrue(
			"attributeIterator should not have next",
			!attributeIterator.hasNext()
		);

	}
	
	public void testDataCollectorWEmptySeries() {
		
		DataCollector dataCollector = DataCollector.getDataCollector();
		dataCollector.clear();
		dataCollector.handle( emptyTimeSeries );

		Iterator timeSeriesIterator = dataCollector.timeSeriesIterator();
		assertTrue(
			"there should be at least one series key",
			timeSeriesIterator.hasNext()
		);
		String timeSeriesKey
		= (String) timeSeriesIterator.next();
		
		assertSame( "emptySeries should be there",
			timeSeriesKey,
			emptyTimeSeries.getKey()
		);

		Iterator emptySeriesIterator
		= dataCollector.iterateTimeSeries( timeSeriesKey );
		assertTrue(
			"ObservationMap.iterateTimeSeries should not generate any"
			+ "Time series for keys corresponding to empty series.",
			!emptySeriesIterator.hasNext()
		);
		assertTrue( "should be no other series key in the collector",
			!timeSeriesIterator.hasNext()
		);
	};
	
	public void testDataCollectorWArcsineSeries() {
		
		DataCollector dataCollector = DataCollector.getDataCollector();
		dataCollector.clear();
		dataCollector.handle( arcsineTimeSeries );

		Iterator timeSeriesIterator = dataCollector.timeSeriesIterator();
		assertTrue(
			"there should be at least one series key",
			timeSeriesIterator.hasNext()
		);
		String timeSeriesKey
		= (String) timeSeriesIterator.next();
		
		assertSame( "key for arcsineSeries should be there",
			timeSeriesKey,
			arcsineTimeSeries.getKey()
		);

		Iterator arcsineSeriesIterator
		= dataCollector.iterateTimeSeries( timeSeriesKey );
		assertTrue(
			"we should find one series corresponding to arcsine...",
			arcsineSeriesIterator.hasNext()
		);
		TimeSeries timeSeries = (TimeSeries) arcsineSeriesIterator.next();
		assertTrue( "... and only one series corresponding to arcsine.",
				!arcsineSeriesIterator.hasNext()
		);
		
		assertEquals( "the retrieved time series should be similar to arcsine",
			timeSeries,
			arcsineTimeSeries
		);
		
		assertTrue( "should be no other series key in the collector",
			!timeSeriesIterator.hasNext()
		);
	};

	public void testDataCollectorWArcAndCosineSeries() {
		
		DataCollector dataCollector = DataCollector.getDataCollector();
		dataCollector.clear();
		dataCollector.handle( arcsineTimeSeries );
		dataCollector.handle( cosineTimeSeries );

		Iterator timeSeriesIterator = dataCollector.timeSeriesIterator();
		assertTrue(
			"there should be 2 series key.",
			timeSeriesIterator.hasNext()
		);
		assertTrue(
			"but only 2.",
				!timeSeriesIterator.hasNext()
		);
		/* Arcsine */
		Iterator arcsineSeriesIterator
		= dataCollector.iterateTimeSeries( arcsineTimeSeries.getKey() );
		assertTrue(
			"we should find one series corresponding to arcsine...",
			arcsineSeriesIterator.hasNext()
		);
		TimeSeries timeSeries = (TimeSeries) arcsineSeriesIterator.next();
		assertTrue( "... and only one series corresponding to arcsine.",
				!arcsineSeriesIterator.hasNext()
		);
		assertEquals( "the retrieved time series should be similar to arcsine",
			timeSeries,
			arcsineTimeSeries
		);
		/* Cosine */
		Iterator cosineSeriesIterator
		= dataCollector.iterateTimeSeries( cosineTimeSeries.getKey() );
		assertTrue(
			"we should find one series corresponding to cosine...",
			cosineSeriesIterator.hasNext()
		);
		timeSeries = (TimeSeries) cosineSeriesIterator.next();
		assertTrue( "... and only one series corresponding to cosine.",
				!cosineSeriesIterator.hasNext()
		);
		assertEquals( "the retrieved time series should be similar to cosine",
			timeSeries,
			cosineTimeSeries
		);
		Observation oo2 = dataCollector.getObservation(
			"ARC.SI.NUS",
			newInterval(
				"20010903000000",
				RegularInterval.DAILY,
				1
			)
		);
		assertEquals( "we should have retrieved o2"
				, o2
				, oo2
		);
		Observation oo5 = dataCollector.getObservation(
			"CO.SI.NUS",
			newInterval(
				"20011101000000",
				RegularInterval.MONTHLY,
				1
			)
		);
		assertEquals( "we should have retrieved o5"
				, o5
				, oo5
		);
		assertTrue(
			"we should have o2",
			dataCollector.hasObservation(
				"ARC.SI.NUS",
				newInterval(
					"20010903000000",
					RegularInterval.DAILY,
					1
				)
			)
		);
		assertTrue(
			"we should not have o5 as a daily observation",
			!dataCollector.hasObservation(
				"CO.SI.NUS",
				newInterval(
					"20011101000000",
					RegularInterval.DAILY,
					1
				)
			)
		);
		assertTrue(
			"but not this one",
			!dataCollector.hasObservation(
				"CO.SI.NUS",
				newInterval(
					"20011101000000",
					RegularInterval.DAILY,
					1
				)
			)
		);
		assertTrue(
			"nor that one",
			!dataCollector.hasObservation(
				"CO.SI.NUS",
				newInterval(
					"20081101000000",
					RegularInterval.MONTHLY,
					1
				)
			)
		);
		assertTrue(
			"nor anything for a series we don't have",
			!dataCollector.hasObservation(
				"SI.NU.SITE",
				newInterval(
					"20081101000000",
					RegularInterval.MONTHLY,
					1
				)
			)
		);
		assertTrue(
				"nor anything for a series we don't have at all",
				!dataCollector.hasObservation(
					"SI.NU.SITE",
					newInterval(
						"20081101000000",
						RegularInterval.DAILY,
						1
					)
				)
			);
	};
	
	private GregorianRegularInterval newInterval(
		String dateString, int frequency, int size) {
		try {
			Date date = FORMAT.parse(dateString);
			return new GregorianRegularInterval( frequency, date, size);
		} catch ( Exception e ) {}
		return null;
	}
	
}
