// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.event;

import java.io.Serializable;

import Int.ecb.gesmescb.error.ErrorWatcher;

/* ***************************** Class ****************************** */

/**
 * <p>Transports error information from {@link GESMESCBEventSource}s
 * to {@link GESMESCBEventListener}s.</p>
 *
 * <p>An ErrorEvent carries a code and a descriptive message
 * for the human reader of log files.
 * The valid severity codes fulfill the relations</p>
 *
 * <CENTER>
 * {@link #INFO} &lt; {@link #WARNING} &lt;
 * {@link #ERROR} &lt; {@link #FAILURE}.
 * </CENTER>
 *
 * @author
 * <A HREF="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</A>
 *
 * @version @(#)$Id: ErrorEvent.java,v 1.1 2017/01/19 16:14:04 trainot Exp $
 */

public class ErrorEvent implements Serializable {

	/* ======================= Constants ======================== */

	/**
	 *   serialVersionUID generated by eclipse.
	 */
	private static final long serialVersionUID = -3474299219286217013L;

	/**
	 * <P>The severity code for "info".</P>
	 *
	 * <P>Listeners receiving an ErrorEvent with this severity code should
	 * not disconnect, but continue.
	 * Such events may be useful just to make listeners
	 * log some information, for instance for debugging
	 * purposes.</P>
	 *
	 * <P>The numeric value is the same as {@link ErrorWatcher#OK}.</P>
	 */
	public static final int INFO = ErrorWatcher.OK;

	/**
	 * <P>The severity code for "warning".</P>
	 *
	 * <P>All {@link GESMESCBEventListener} objects handling
	 * such an ErrorEvent may disconnect or otherwise choose to continue.
	 * This severity will typically be used if the warning condition
	 * is not severe enough to affect the integrity
	 * of vital GESMES/TS data but should nevertheless be
	 * reported to the operator.</P>
	 *
	 * <P>The numeric value is the same as {@link ErrorWatcher#WARNING}.</P>
	 */
	public static final int WARNING = ErrorWatcher.WARNING;

	/**
	 * <P>The severity code for "error".</P>
	 *
	 * <P>Most {@link GESMESCBEventListener} objects handling
	 * such an ErrorEvent should disconnect.
	 * There may still be objects like
	 * a {@link Int.ecb.gesmescb.semanticchecker.SemanticChecker} which
	 * may choose to continue nevertheless.
	 * This severity will typically be used if the error condition
	 * is so severe that the integrity of vital GESMES/TS data may suffer
	 * if the data were written to its final destination.</P>
	 *
	 * <P>The numeric value is the same as {@link ErrorWatcher#ERROR}.</P>
	 */
	public static final int ERROR = ErrorWatcher.ERROR;

	/**
	 * <P>The severity code for "failure".</P>
	 *
	 * <P>All {@link GESMESCBEventListener} objects handling
	 * an ErrorEvent with this severity code should disconnect.</P>
	 *
	 * <P>The numeric value is the same as {@link ErrorWatcher#FAILURE}.</P>
	 */
	public static final int FAILURE = ErrorWatcher.FAILURE;

	/* ========================= Fields ========================= */

	/**
	 * The severity of the error.
	 *
	 * @serial
	 */
	private final int severity;

	/**
	 * A human readable description of what has happened.
	 *
	 * @serial
	 */
	private final String message;

	/* ====================== Constructors ====================== */

	/**
	 * Construct a new ErrorEvent with given severity and descriptive
	 * message.
	 *
	 * @param message briefly describes in human readable terms
	 * what is the matter.
	 * @param severity severity code of the ErrorEvent. This
	 * should be one of
	 * {@link #INFO}, {@link #WARNING}, {@link #ERROR},
	 * or {@link #FAILURE}.
	 *
	 * @throws IllegalArgumentException if the severity code is invalid.
	 */
	public ErrorEvent ( String message, int severity ) {
		if (
			INFO <= severity
			&&
			severity <= FAILURE
		) {
			this.severity = severity;
			this.message = message;
		} else {
			throw new IllegalArgumentException(
				"Severity code is neither of"
				+ " INFO, WARNING, ERROR, or FAILURE."
			);
		}
	}

	/* ===================== Public Methods ===================== */

	/** Get the message of the ErrorEvent. */
	public String getMessage () {
		return message;
	}

	/**
	 * Get the severity of the ErrorEvent.
	 *
	 * @return severity code.
	 */
	public int getSeverity () {
		return severity;
	}

	/**
	 * Indicates whether some other Object is "equal to"
	 * this <code>ErrorEvent</code>.
	 *
	 * @param obj the reference object with which to compare
	 *
	 * @return true if this <code>ErrorEvent</code> is equal to the argument;
	 * false otherwise.
	 */
	public boolean equals ( final Object obj ) {
		if ( this == obj ) return true;
		if ( obj instanceof ErrorEvent ) {
			final ErrorEvent other = (ErrorEvent) obj;
			return
				message.equals( other.message )
				&& severity == other.severity;
		}
		return false;
	}

	/**
	 * Returns the hash code value for this <code>ErrorEvent</code>.
	 * If o1.equals(o2) then o1.hashCode()==o2.hashCode()
	 * for any two <code>ErrorEvent</code> o1 and o2,
	 * as required by the general contract of {@link Object#hashCode}.
	 *
	 * @return hash code value for this object.
	 */
	public int hashCode () {
		return message.hashCode() ^ severity;
	}

	/**
	 * Return a human readable description of the reason
	 * why the ErrorEvent was created.
	 *
	 * @return a human readable description of the severity and
	 * causes.
	 */
	public String toString () {
		String string;
		switch ( severity ) {
			case INFO: string = "INFO"; break;
			case WARNING: string = "WARNING"; break;
			case ERROR: string = "ERROR"; break;
			default: string = "FAILURE";
		}
		return getClass().getName() + ": " + string + ": " + message;
	}
}
