// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, (c) 2003, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.event;

import Int.ecb.gesmescb.*;

import Int.ecb.gesmescb.util.Named;
import Int.ecb.gesmescb.util.ComponentHelper;

import Int.ecb.logging.Level;

/* ***************************** Class ****************************** */

/**
 * <P>Provides base class features for event transport classes.</P>
 *
 * <P>These are classes which
 * receive events
 * through their interface {@link GESMESCBEventListener}
 * and possibly {@link StructureListener} and {@link DataListener}
 * and subsequently dispatch these or other events
 * through their interface {@link GESMESCBEventSource}.</P>
 *
 * <P>Subclasses will implement the required <code>handle</code>
 * methods to process incoming events
 * and then use the protected method {@link #dispatch}
 * in order to dispatch the outgoing events.</P>
 *
 * <P>The GESMESCBEventBroker supports optional automatic disconnect
 * if no one listens to it when an event is dispatched.</P>
 *
 * <P>With multiple inheritance, this bastard class
 * would not be necessary.
 * Without, it is handy to implement
 * source and listener behaviour together in one class.</p>
 *
 * @author
 * <A HREF="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</A>
 *
 * @version @(#)$Id: GESMESCBEventBroker.java,v 1.1 2017/01/19 16:14:04 trainot Exp $
 */

public abstract class GESMESCBEventBroker
	extends GESMESCBEventListenerAdapter
	implements Named, StructureListener, DataListener, GESMESCBEventSource
{
	/* ========================= Fields ========================= */

	/**
	 * The Logger object to handle all logging events.
	 * Also from subclasses.
	 */
	protected final ComponentHelper componentHelper;

	/**
	 * Used to dispatch events.
	 */
	final GESMESCBEventSourceHelper sourceHelper;

	/**
	 * Whether automatic disconnect is enabled.
	 */
	private boolean autoDisconnect;

	/* ====================== Constructors ====================== */

	/**
	 * Construct a GESMESCBEventBroker with or without
	 * automatic disconnect enabled.
	 * Automatic disconnect means that the object will
	 * disconnect from its source
	 * when it detects that no important
	 * subsequent listeners are listening.
	 *
	 * @param name used for logging and preferences.
	 * @param autoDisconnect whether automatic disconnect should be enabled.
	 */
	protected GESMESCBEventBroker (
		final String name,
		final boolean autoDisconnect
	) {
		componentHelper = new ComponentHelper( name, this );
		sourceHelper = new GESMESCBEventSourceHelper();
		this.autoDisconnect = autoDisconnect;
	}

	/* ===================== Public Methods ===================== */

	public String getName () {
		return componentHelper.name;
	}

	public void addListener (
		final GESMESCBEventListener listener,
		final boolean isImportant
	) {
		sourceHelper.addListener( listener, isImportant );
	}

	public void removeAllListeners () {
		sourceHelper.removeAllListeners();
	}

	/* ------------------------ Listener ------------------------ */

	/** Forward, unless overridden. */
	public void handle ( final InterchangeBegin interchangeBegin ) {
		dispatch( interchangeBegin );
	}

	/** Forward, unless overridden. */
	public void handle ( final InterchangeEnd interchangeEnd ) {
		dispatch( interchangeEnd );
	}

	/** Forward, unless overridden. */
	public void handle ( final MessageBegin messageBegin ) {
		dispatch( messageBegin );
	}

	/** Forward, unless overridden. */
	public void handle ( final MessageEnd messageEnd ) {
		dispatch( messageEnd );
	}

	/** Forward, unless overridden. */
	public void handle ( final CodeList codeList ) {
		dispatch( codeList );
	}

	/** Forward, unless overridden. */
	public void handle ( final StatisticalConcept statisticalConcept ) {
		dispatch( statisticalConcept );
	}

	/** Forward, unless overridden. */
	public void handle ( final KeyFamily keyFamily ) {
		dispatch( keyFamily );
	}

	/** Forward, unless overridden. */
	public void handle ( final TimeSeries timeSeries ) {
		dispatch( timeSeries );
	}

	/** Forward, unless overridden. */
	public void handle ( final Attribute attribute ) {
		dispatch( attribute );
	}

	/** Forward, unless overridden. */
	public void handle ( final ErrorEvent errorEvent ) {
		dispatch( errorEvent );
	}

	/* =================== Protected Methods ==================== */

	/**
	 * <p>Distribute an <code>InterchangeBegin</code> to all listeners.</p>
	 *
	 * <p>If there is no important listener
	 * and automatic disconnect is enabled,
	 * then self disconnect.</p>
	 *
	 * @return true if at least one {@link GESMESCBEventListener}
	 * was still listening.
	 */
	protected boolean dispatch ( final InterchangeBegin interchangeBegin ) {
		if ( sourceHelper.dispatch( interchangeBegin ) ) {
			return true;
		} else {
			if ( autoDisconnect ) setActive( false );
			return false;
		}
	}

	/**
	 * <p>Distribute an <code>InterchangeEnd</code> to all listeners.</p>
	 *
	 * <p>If there is no important listener
	 * and automatic disconnect is enabled,
	 * then self disconnect.</p>
	 *
	 * @return true if at least one {@link GESMESCBEventListener}
	 * was still listening.
	 */
	protected boolean dispatch ( final InterchangeEnd interchangeEnd ) {
		if ( sourceHelper.dispatch( interchangeEnd ) ) {
			return true;
		} else {
			if ( autoDisconnect ) setActive( false );
			return false;
		}
	}

	/**
	 * <p>Distribute a <code>MessageBegin</code> to all listeners.</p>
	 *
	 * <p>If there is no important listener
	 * and automatic disconnect is enabled,
	 * then self disconnect.</p>
	 *
	 * @return true if at least one {@link GESMESCBEventListener}
	 * was still listening.
	 */
	protected boolean dispatch ( final MessageBegin messageBegin ) {
		if ( sourceHelper.dispatch( messageBegin ) ) {
			return true;
		} else {
			if ( autoDisconnect ) setActive( false );
			return false;
		}
	}

	/**
	 * <p>Distribute a <code>MessageEnd</code> to all listeners.</p>
	 *
	 * <p>If there is no important listener
	 * and automatic disconnect is enabled,
	 * then self disconnect.</p>
	 *
	 * @return true if at least one {@link GESMESCBEventListener}
	 * was still listening.
	 */
	protected boolean dispatch ( final MessageEnd messageEnd ) {
		if ( sourceHelper.dispatch( messageEnd ) ) {
			return true;
		} else {
			if ( autoDisconnect ) setActive( false );
			return false;
		}
	}

	/**
	 * <p>Distribute a <code>CodeList</code> to all listeners.</p>
	 *
	 * <p>If there is no important listener
	 * and automatic disconnect is enabled,
	 * then self disconnect.</p>
	 *
	 * @return true if at least one {@link GESMESCBEventListener}
	 * was still listening.
	 */
	protected boolean dispatch ( final CodeList codeList ) {
		if ( sourceHelper.dispatch( codeList ) ) {
			return true;
		} else {
			if ( autoDisconnect ) setActive( false );
			return false;
		}
	}

	/**
	 * <p>Distribute a <code>StatisticalConcept</code> to all listeners.</p>
	 *
	 * <p>If there is no important listener
	 * and automatic disconnect is enabled,
	 * then self disconnect.</p>
	 *
	 * @return true if at least one {@link GESMESCBEventListener}
	 * was still listening.
	 */
	protected boolean dispatch ( final StatisticalConcept statisticalConcept ) {
		if ( sourceHelper.dispatch( statisticalConcept ) ) {
			return true;
		} else {
			if ( autoDisconnect ) setActive( false );
			return false;
		}
	}

	/**
	 * <p>Distribute a <code>KeyFamily</code> to all listeners.</p>
	 *
	 * <p>If there is no important listener
	 * and automatic disconnect is enabled,
	 * then self disconnect.</p>
	 *
	 * @return true if at least one {@link GESMESCBEventListener}
	 * was still listening.
	 */
	protected boolean dispatch ( final KeyFamily keyFamily ) {
		if ( sourceHelper.dispatch( keyFamily ) ) {
			return true;
		} else {
			if ( autoDisconnect ) setActive( false );
			return false;
		}
	}

	/**
	 * <p>Distribute a <code>TimeSeries</code> to all listeners.</p>
	 *
	 * <p>If there is no important listener
	 * and automatic disconnect is enabled,
	 * then self disconnect.</p>
	 *
	 * @return true if at least one {@link GESMESCBEventListener}
	 * was still listening.
	 */
	protected boolean dispatch ( final TimeSeries timeSeries ) {
		if ( sourceHelper.dispatch( timeSeries ) ) {
			return true;
		} else {
			if ( autoDisconnect ) setActive( false );
			return false;
		}
	}

	/**
	 * <p>Distribute an <code>Attribute</code> to all listeners.</p>
	 *
	 * <p>If there is no important listener
	 * and automatic disconnect is enabled,
	 * then self disconnect.</p>
	 *
	 * @return true if at least one {@link GESMESCBEventListener}
	 * was still listening.
	 */
	protected boolean dispatch ( final Attribute attribute ) {
		if ( sourceHelper.dispatch( attribute ) ) {
			return true;
		} else {
			if ( autoDisconnect ) setActive( false );
			return false;
		}
	}

	/**
	 * <p>Distribute an <code>ErrorEvent</code> to all listeners.</p>
	 *
	 * <p>If there is no important listener
	 * and automatic disconnect is enabled,
	 * then self disconnect.</p>
	 *
	 * @return true if at least one {@link GESMESCBEventListener}
	 * was still listening.
	 */
	protected boolean dispatch ( final ErrorEvent errorEvent ) {
		if ( sourceHelper.dispatch( errorEvent ) ) {
			return true;
		} else {
			if ( autoDisconnect ) setActive( false );
			return false;
		}
	}

	/* ---------------- Useful Complain Utility ----------------- */

	/**
	 * Logs a message and dispatches an {@link ErrorEvent}.
	 * This is internally used as a shortcut.
	 * Note that the method name logged is not always "complain",
	 * but in slight abuse of the logging API, can be
	 * set via a parameter.
	 *
	 * @param description of the problem.
	 * @param methodName the name of the current method, used for logging.
	 * @param object on which the problem was detected,
	 * typically a {@link Int.ecb.gesmescb.GESMESCBObject}.
	 * @param severity of the offence.
	 */
	protected void complain (
		final String description,
		final String methodName,
		final Object object,
		final int severity
	) {
		Level logLevel;
		switch ( severity ) {
			case ErrorEvent.INFO:
				logLevel = Level.FINE;
				break;
			case ErrorEvent.WARNING:
				logLevel = Level.WARNING;
				break;
			default:
				// Since only offences should be
				// reported, everything else would be
				// SEVERE (such as ERROR or FAILURE).
				logLevel = Level.SEVERE;
		}
		componentHelper.logger.log(
			logLevel, componentHelper.logName, methodName, description, object
		);
		dispatch( new ErrorEvent( description, severity ) );
	}
}
