// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, (c) 2003, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.event;

import java.util.Collection;
import java.util.Iterator;
import java.util.ArrayList;

import Int.ecb.gesmescb.*;
import Int.ecb.gesmescb.event.ErrorEvent;
import Int.ecb.gesmescb.util.ComponentHelper;
import Int.ecb.logging.Level;

/* ***************************** Class ****************************** */

/**
 * <P>Eases implementing the interface {@link GESMESCBEventSource}.</P>
 *
 * <P>The protected operation(s) {@link #dispatch} then can
 * be used to distribute information to all
 * registered {@link GESMESCBEventListener} objects.</P>
 *
 * <P>If extending is not possible, classes implementing
 * <CODE>GESMESCBEventSource</CODE> can be created using
 * the class {@link GESMESCBEventSourceHelper}.</P>
 *
 * @author
 * <A HREF="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</A>
 *
 * @version @(#)$Id: GESMESCBEventSourceAdapter.java,v 1.1 2017/01/19 16:14:04 trainot Exp $
 */

public abstract class GESMESCBEventSourceAdapter
	implements GESMESCBEventSource
{
	/* ========================= Fields ========================= */

	/**
	 * The Logger object to handle all logging events.
	 * Also from subclasses.
	 */
	protected final ComponentHelper componentHelper;
	
	/** Lists listener and importance pairs. */
	private final Collection listeners;

	/* ====================== Constructors ====================== */

	/**
	 * Construct a fresh GESMESCBEventSourceAdapter.
	 */
	protected GESMESCBEventSourceAdapter () {
		componentHelper = new ComponentHelper( "GESMESCBEventSourceAdapter", this );
		listeners = new ArrayList();
	}

	/* ===================== Public Methods ===================== */

	/**
	 * Register a listener.
	 *
	 * @param listener to receive events.
	 * @param important whether the listener counts in the
	 * evaluation of interest.
	 */
	public synchronized void addListener (
		final GESMESCBEventListener listener,
		final boolean important
	) {
		listeners.add( new ListenerImportanceMapping( listener, important ) );
	}

	/**
	 * Remove all listeners from the distribution list.
	 */
	public synchronized void removeAllListeners () {
		listeners.clear();
	}

	/* =================== Protected Methods ==================== */

	/**
	 * <p>Distribute an <code>InterchangeBegin</code> to all listeners.</p>
	 *
	 * @return true if at least one {@link GESMESCBEventListener}
	 * was still listening.
	 */
	protected boolean dispatch ( final InterchangeBegin interchangeBegin ) {
		return dispatch( (Object) interchangeBegin );
	}

	/**
	 * <p>Distribute an <code>InterchangeEnd</code> to all listeners.</p>
	 *
	 * @return true if at least one {@link GESMESCBEventListener}
	 * was still listening.
	 */
	protected boolean dispatch ( final InterchangeEnd interchangeEnd ) {
		return dispatch( (Object) interchangeEnd );
	}

	/**
	 * <p>Distribute a <code>MessageBegin</code> to all listeners.</p>
	 *
	 * @return true if at least one {@link GESMESCBEventListener}
	 * was still listening.
	 */
	protected boolean dispatch ( final MessageBegin messageBegin ) {
		return dispatch( (Object) messageBegin );
	}

	/**
	 * <p>Distribute a <code>MessageEnd</code> to all listeners.</p>
	 *
	 * @return true if at least one {@link GESMESCBEventListener}
	 * was still listening.
	 */
	protected boolean dispatch ( final MessageEnd messageEnd ) {
		return dispatch( (Object) messageEnd );
	}

	/**
	 * <p>Distribute a <code>CodeList</code> to all listeners.</p>
	 *
	 * @return true if at least one {@link GESMESCBEventListener}
	 * was still listening.
	 */
	protected boolean dispatch ( final CodeList codeList ) {
		return dispatch( (Object) codeList );
	}

	/**
	 * <p>Distribute a <code>StatisticalConcept</code> to all listeners.</p>
	 *
	 * @return true if at least one {@link GESMESCBEventListener}
	 * was still listening.
	 */
	protected boolean dispatch ( final StatisticalConcept statisticalConcept ) {
		return dispatch( (Object) statisticalConcept );
	}

	/**
	 * <p>Distribute a <code>KeyFamily</code> to all listeners.</p>
	 *
	 * @return true if at least one {@link GESMESCBEventListener}
	 * was still listening.
	 */
	protected boolean dispatch ( final KeyFamily keyFamily ) {
		return dispatch( (Object) keyFamily );
	}

	/**
	 * <p>Distribute a <code>TimeSeries</code> to all listeners.</p>
	 *
	 * @return true if at least one {@link GESMESCBEventListener}
	 * was still listening.
	 */
	protected boolean dispatch ( final TimeSeries timeSeries ) {
		return dispatch( (Object) timeSeries );
	}

	/**
	 * <p>Distribute an <code>Attribute</code> to all listeners.</p>
	 *
	 * @return true if at least one {@link GESMESCBEventListener}
	 * was still listening.
	 */
	protected boolean dispatch ( final Attribute attribute ) {
		return dispatch( (Object) attribute );
	}

	/**
	 * <p>Distribute an <code>ErrorEvent</code> to all listeners.</p>
	 *
	 * @return true if at least one {@link GESMESCBEventListener}
	 * was still listening.
	 */
	protected boolean dispatch ( final ErrorEvent errorEvent ) {
		return dispatch( (Object) errorEvent );
	}

	/* ==================== Package Methods ===================== */

	/**
	 * <p>Distribute an event to all listeners.</p>
	 *
	 * @param event to be distributed.
	 *
	 * @return true if at least one important {@link GESMESCBEventListener}
	 * was listening.
	 */
	boolean dispatch ( final Object event ) {
		boolean wasHeard = false;

		for (
			final Iterator iterator = listeners.iterator();
			iterator.hasNext();
		) {
			final ListenerImportanceMapping pair
			= (ListenerImportanceMapping) iterator.next();

			if ( !pair.listener.isActive() ) {
				componentHelper.logger.log(
						Level.FINEST,
						componentHelper.logName,
						"run",
						"Removing non-active listener: " + pair.listener.toString()
					);
				iterator.remove();
				continue;
			}

			if ( event instanceof TimeSeries ) {
				if ( pair.listener instanceof DataListener ) {
					( (DataListener) pair.listener ).handle(
						(TimeSeries) event
					);
				}
			} else if ( event instanceof Attribute ) {
				if ( pair.listener instanceof DataListener ) {
					( (DataListener) pair.listener ).handle(
						(Attribute) event
					);
				}
			} else if ( event instanceof CodeList ) {
				if ( pair.listener instanceof StructureListener ) {
					( (StructureListener) pair.listener ).handle(
						(CodeList) event
					);
				}
			} else if ( event instanceof StatisticalConcept ) {
				if ( pair.listener instanceof StructureListener ) {
					( (StructureListener) pair.listener ).handle(
						(StatisticalConcept) event
					);
				}
			} else if ( event instanceof KeyFamily ) {
				if ( pair.listener instanceof StructureListener ) {
					( (StructureListener) pair.listener ).handle(
						(KeyFamily) event
					);
				}
			} else if ( event instanceof MessageBegin ) {
				pair.listener.handle( (MessageBegin) event );
			} else if ( event instanceof MessageEnd ) {
				pair.listener.handle( (MessageEnd) event );
			} else if ( event instanceof InterchangeBegin ) {
				pair.listener.handle( (InterchangeBegin) event );
			} else if ( event instanceof InterchangeEnd ) {
				pair.listener.handle( (InterchangeEnd) event );
			} else if ( event instanceof ErrorEvent ) {
				componentHelper.logger.log(
						Level.FINE,
						componentHelper.logName,
						"run",
						"Dispatching ErrorEvent: " + ((ErrorEvent)event).getMessage()
					);
				pair.listener.handle( (ErrorEvent) event );
			} else if ( event == null ) {
				// Should not happen.
				dispatch( new ErrorEvent(
					"Trying to call dispatch(null)",
					ErrorEvent.FAILURE
				) );
			} else {
				// Should not happen.
				componentHelper.logger.log(
						Level.FINEST,
						componentHelper.logName,
						"run",
						"Wrong listener type (ErrorEvent dispatched): " + pair.listener.toString() + ":" + event.getClass().getName()
					);
				dispatch( new ErrorEvent(
					"Encountered dispatch(" + event.getClass().getName() + ")",
					ErrorEvent.FAILURE
				) );
			}
			wasHeard |= pair.important;
		}
		return wasHeard;
	}
}

/* ***************************** Class ****************************** */

/** Pairs up a listener with its importance. */
final class ListenerImportanceMapping {

	/* ========================= Fields ========================= */

	/** The concerned listener. */
	public final GESMESCBEventListener listener;

	/** Whether the listener is important. */
	public final boolean important;

	/* ====================== Constructors ====================== */

	public ListenerImportanceMapping (
		final GESMESCBEventListener listener, final boolean important
	) {
		this.listener = listener;
		this.important = important;
	};

	/* ===================== Public Methods ===================== */

	public boolean equals ( final Object that ) {
		if ( ! ( that instanceof ListenerImportanceMapping ) ) return false;
		return listener.equals( ( (ListenerImportanceMapping) that ).listener );
	}

	public int hashCode () {
		return listener.hashCode();
	}
}
