// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.event;

import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;

import Int.ecb.gesmescb.InterchangeBegin;

/* ***************************** Class ****************************** */

/**
 * Implements a few tests for the {@link GESMESCBEventSourceHelper}.
 *
 * @author
 * <A HREF="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</A>
 *
 * @version @(#)$Id: GESMESCBEventSourceHelperTest.java,v 1.1 2017/01/19 16:14:04 trainot Exp $
 */
public final class GESMESCBEventSourceHelperTest extends TestCase {

	/* ======================= Constants ======================== */

	/**
	 * Test data to be dispatched.
	 */
	private InterchangeBegin[] objects;

	/**
	 * Test listeners who check.
	 */
	TestEventListener[] listener;

	/* ====================== Constructors ====================== */

	public GESMESCBEventSourceHelperTest( final String name ) {
		super( name );
	}

	/* ================== Protected Methods ===================== */

	/**
	 * Sets up the test fixture.
	 */
	protected void setUp () {
		// Build the events.
		objects = new InterchangeBegin[20];
		for ( int i = 0; i < objects.length; ++i ) {
			objects[i] = new InterchangeBegin(
				"IREF" + i,
				"sender" + i,
				"receiver" + i,
				"000501:" + i,
				true
			);
		}

		// Build the listeners.
		listener = new TestEventListener[objects.length];
		for ( int i = 0; i < listener.length; ++i ) {
			listener[i] = new TestEventListener( 1 );
		}
	}

	/**
	 * Tears down the test fixture.
	 */
	protected void tearDown () {
		objects = null;
		listener = null;
	}

	/* ===================== Public Methods ===================== */

	/**
	 * Assembles and returns a test suite for all
	 * the test methods of this test case.
	 *
	 * @return a non-null test suite.
	 */
	public static Test suite () {
		final TestSuite suite = new TestSuite(
			GESMESCBEventSourceHelperTest.class
		);
		return suite;
	}

	/**
	 * @param arg the array of argument {@link String}s
	 */
	public static void main ( String[] arg ) {
		final String[] testCaseName = {
			GESMESCBEventSourceHelperTest.class.getName()
		};
		junit.textui.TestRunner.main( testCaseName );
	}

	/* ---------------------- Test Methods ---------------------- */

	/**
	 * Test {@link GESMESCBEventSourceHelper#addListener},
	 * {@link GESMESCBEventSourceHelper#dispatch}.
	 */
	public void testAddDispatch () {
		GESMESCBEventSourceHelper source = new GESMESCBEventSourceHelper();

		// Incrementally add and dispatch.
		for ( int i = 0; i < listener.length; ++i ) {
			source.addListener( listener[i], true );
			source.dispatch( objects[i] );
			for ( int j = 0; j < listener.length; ++j ) {
				assertSame(
					"Last " + j + " got in adding" + i,
					j<=i ? objects[i] : null,
					listener[j].getLastObject()
				);
			}
		}
	}

	/**
	 * Test {@link GESMESCBEventSourceHelper#removeAllListeners}.
	 */
	public void testRemoveAll () {
		GESMESCBEventSourceHelper source = new GESMESCBEventSourceHelper();

		// Add all.
		for ( int i = 0; i < listener.length; ++i ) {
			source.addListener( listener[i], true );
		}

		// Dispatch and check one event.
		source.dispatch( objects[0] );
		for ( int i = 0; i < listener.length; ++i ) {
			assertSame(
				"Last " + i + " got before removeAllListeners",
				objects[0],
				listener[i].getLastObject()
			);
		}

		// Now remove.
		source.removeAllListeners();

		// Once again dispatch and check that it did not come.
		source.dispatch( objects[1] );
		for ( int i = 0; i < listener.length; ++i ) {
			assertSame(
				"Last " + i + " got after removeAllListeners",
				objects[0],
				listener[i].getLastObject()
			);
		}
	}

	/**
	 * Test {@link GESMESCBEventSourceHelper#dispatch} with unimportant
	 * listeners.
	 */
	public void testUnimportantListening () {
		GESMESCBEventSourceHelper source = new GESMESCBEventSourceHelper();

		assertTrue( "No listener", !source.dispatch( objects[0] ) );
		source.addListener( listener[0], false );
		assertTrue( "No important listener", !source.dispatch( objects[1] ) );
		source.addListener( listener[1], false );
		assertTrue( "Still no important listener", !source.dispatch( objects[2] ) );
		source.addListener( listener[2], true );
		assertTrue( "An important listener", source.dispatch( objects[3] ) );
		source.addListener( listener[3], false );
		assertTrue( "Still an important listener", source.dispatch( objects[4] ) );
		listener[2].setActive( false );
		assertTrue( "No important listener listening", !source.dispatch( objects[4] ) );
		listener[2].setActive( true );
		assertTrue( "Nonlistening was disconnected", !source.dispatch( objects[4] ) );
	}
}
