// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2003, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.event;

import java.util.Iterator;
import java.util.Date;

import java.io.File;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;

import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;

import Int.ecb.gesmescb.*;

/* ***************************** Class ****************************** */

/**
 * <p>Tests the implementation of the
 * {@link SerialWriter} and {@link SerialReader}.
 *
 * @author
 * <a href="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</a>
 *
 * @version @(#)$Id: SerialRWTest.java,v 1.1 2017/01/19 16:14:04 trainot Exp $
 */

public class SerialRWTest extends TestCase {

	/* ========================= Fields ========================= */

	final InterchangeBegin interchangeBegin = new InterchangeBegin(
		"IREF000002", "4F0", "ZZZ", "020101:1444", true
	);

	final MessageBegin messageBegin = new MessageBegin(
		"MREF000004", "4F0", "ZZZ", "ECB",
		MessageBegin.STATISTICAL_DATA,
		"Test serialized message",
		new Contact[]{},
		"ECB_OB1",
		MessageBegin.UPDATE_OR_REPLACE_ACTION,
		"20020101",
		"", "",
		"ECB_OB1"
	);

	final CodeList codeList = new CodeList(
		"CL_PAULA", "Paula's codes", "", new Date(),
		new String[]{}, new String[]{}
	);

	final StatisticalConcept concept = new StatisticalConcept(
		"SC_PAULA", "Paula's concept", "", new Date()
	);

	final KeyFamily keyFamily = new KeyFamily(
		"KF_PAULA", "Paula's family", "", new Date(),
		new KeyFamilyComponent[]{}
	);

	final TimeSeries timeSeries = new TimeSeries(
		"D.PAULA", "20020101", "102", new Observation[]{
			new Observation( "13.6", "P", "F", "" )
		}
	);

	final Attribute attribute = new Attribute(
		"D.PAULA", "20020101", "102", "MALUCITAS", "NNGUF"
	);

	final MessageEnd messageEnd = new MessageEnd( "MREF000004" );

	final InterchangeEnd interchangeEnd = new InterchangeEnd(
		"IREF000002"
	);

	final ErrorEvent error = new ErrorEvent( "Finished", ErrorEvent.INFO );

	/* ====================== Constructors ====================== */

	/**
	 * Constructs a <code>SerialRWTest</code> with the specified name.
	 *
	 * @param name Test case name
	 */
	public SerialRWTest ( final String name ) {
		super( name );
	}

	/* ===================== Public Methods ===================== */

	/**
	 * Assembles and returns a test suite for all
	 * the test methods of this test case.
	 *
	 * @return a non-null test suite
	 */
	public static Test suite () {
		return new TestSuite( SerialRWTest.class );
	}

	/**
	 * Runs the test case.
	 */
	public static void main ( final String args[] ) {
		final String[] testCaseName = { SerialRWTest.class.getName() };
		junit.textui.TestRunner.main( testCaseName );
	}

	/* --------------------- Test Methods --------------------- */

	public void test () throws IOException {
		final File file = File.createTempFile( "serialGesmes", ".tmp" );
		try {
			file.deleteOnExit();

			final SerialWriter serialWriter = new SerialWriter(
				"", file.getPath()
			);
			serialWriter.handle( interchangeBegin );
			serialWriter.handle( messageBegin );
			serialWriter.handle( codeList );
			serialWriter.handle( concept );
			serialWriter.handle( keyFamily );
			serialWriter.handle( timeSeries );
			serialWriter.handle( attribute );
			serialWriter.handle( messageEnd );
			serialWriter.handle( interchangeEnd );
			serialWriter.handle( error );
			serialWriter.close();

			final TestEventListener listener = new TestEventListener( 20 );

			final SerialReader serialReader = new SerialReader(
				"", file.getPath()
			);
			serialReader.addListener( listener, true );
			serialReader.run();
			serialReader.close();

			final Iterator iterator = listener.iterator();
			assertEquals( interchangeBegin, iterator.next() );
			assertEquals( messageBegin, iterator.next() );
			assertEquals( codeList, iterator.next() );
			assertEquals( concept, iterator.next() );
			assertEquals( keyFamily, iterator.next() );
			assertEquals( timeSeries, iterator.next() );
			assertEquals( attribute, iterator.next() );
			assertEquals( messageEnd, iterator.next() );
			assertEquals( interchangeEnd, iterator.next() );
			assertEquals( error, iterator.next() );
			if ( iterator.hasNext() ) {
				fail( "Got another thing: " + iterator.next() );
			}
		}
		finally {
			file.delete();
		}
	}

	public void testStreamConstructor () throws IOException {
		final ByteArrayOutputStream outputStream = new ByteArrayOutputStream();
		final SerialWriter serialWriter = new SerialWriter(
			"", outputStream
		);
		serialWriter.handle( interchangeBegin );
		serialWriter.handle( interchangeEnd );
		serialWriter.close();
		outputStream.close();

		final TestEventListener listener = new TestEventListener( 20 );
		final ByteArrayInputStream inputStream = new ByteArrayInputStream(
			outputStream.toByteArray()
		);
		final SerialReader serialReader = new SerialReader(
			"", inputStream
		);
		serialReader.addListener( listener, true );
		serialReader.run();
		serialReader.close();

		final Iterator iterator = listener.iterator();
		assertEquals( interchangeBegin, iterator.next() );
		assertEquals( interchangeEnd, iterator.next() );
		if ( iterator.hasNext() ) {
			fail( "Got another thing: " + iterator.next() );
		}
	}

	public void testNext () throws IOException {
		final ByteArrayOutputStream outputStream = new ByteArrayOutputStream();
		final SerialWriter serialWriter = new SerialWriter(
			"", outputStream
		);
		serialWriter.handle( messageBegin );
		serialWriter.handle( messageEnd );
		serialWriter.close();
		outputStream.close();

		final ByteArrayInputStream inputStream = new ByteArrayInputStream(
			outputStream.toByteArray()
		);
		final SerialReader serialReader = new SerialReader(
			"", inputStream
		);

		assertTrue( "Some objects 1", serialReader.hasNext() );
		assertEquals( messageBegin, serialReader.next() );
		assertTrue( "Some objects 2", serialReader.hasNext() );
		assertEquals( messageEnd, serialReader.next() );
		if ( serialReader.hasNext() ) {
			fail( "Got another thing: " + serialReader.next() );
		}
		serialReader.close();
	}
}
