// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2003, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.event;

import java.util.NoSuchElementException;

import java.io.Serializable;
import java.io.InputStream;
import java.io.FileInputStream;
import java.io.BufferedInputStream;
import java.io.ObjectInput;
import java.io.ObjectInputStream;
import java.io.IOException;
import java.io.EOFException;

import Int.ecb.logging.Level;

import Int.ecb.gesmescb.event.ErrorEvent;
import Int.ecb.gesmescb.event.GESMESCBReader;
import Int.ecb.gesmescb.event.GESMESCBEventSourceAdapter;

import Int.ecb.gesmescb.util.Named;
import Int.ecb.gesmescb.util.ComponentHelper;

/* ***************************** Class ****************************** */

/**
 * <p>Reads serialized {@link Int.ecb.gesmescb.GESMESCBObject}s.</p>
 *
 * <center><table border="3" cellpadding="3">
 * <caption>
 * <a name="prefs">Available {@link Int.ecb.prefs preference} settings</a>
 * </caption>
 * <tr>
 * <th>preference name</th>
 * <th>type</th>
 * <th>meaning</th>
 * <th>default</th>
 * </tr>
 * <tr><td><code>source</td><td>string</td><td>input file path</td></tr>
 * </table></center>
 *
 * @author
 * <a href="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</a>
 *
 * @version @(#)$Id: SerialReader.java,v 1.1 2017/01/19 16:14:04 trainot Exp $
 */
public final class SerialReader
	extends GESMESCBEventSourceAdapter
	implements GESMESCBReader, Named
{
	/* ========================= Fields ========================= */

	/**
	 * Takes care of logging and preferences.
	 */
	protected final ComponentHelper componentHelper;

	/** Source for serialized input. */
	private final ObjectInput input;

	/**
	 * The name of the file read.
	 * This can be used for log messages.
	 */
	private final String filename;

	/** Next element for {@link #next()}. */
	private Object next;

	/* ====================== Constructors ====================== */

	/**
	 * Construct a <code>SerialReader</code> to read from a file.
	 *
	 * @param name used for logging and preferences.
	 * @param filename where to read from.
	 *
	 * @throws NullPointerException if <code>filename==null</code>
	 * and the preference variable
	 * <code>source</code> is undefined.
	 * @throws IOException if the file cannot be opened for reading.
	 */
	public SerialReader ( final String name, final String filename )
		throws IOException
	{
		componentHelper = new ComponentHelper( name, this );
		this.filename = filename != null
			? filename
			: componentHelper.getPreference( "source", null );
		if ( this.filename == null ) {
			final String desc
			= "No preference variable \"source\" found"
				+ " to determine the input file.";
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"SerialReader()",
				desc
			);
			throw new NullPointerException( desc );
		}
		input = new ObjectInputStream(
			new BufferedInputStream(
				new FileInputStream( this.filename )
			)
		);
		next = null;
	}

	/**
	 * Construct a <code>SerialReader</code>
	 * reading from a file according to preferences.
	 *
	 * @param name used for logging and preferences.
	 *
	 * @throws NullPointerException if the preference variable
	 * <code>source</code> is undefined.
	 * @throws IOException if the file cannot be opened for reading.
	 */
	public SerialReader ( final String name )
		throws IOException
	{
		this( name, (String) null );
	}

	/**
	 * Construct a <code>SerialReader</code> reading from a byte stream.
	 *
	 * @param name used for logging and preferences.
	 *
	 * @throws NullPointerException if the preference variable
	 * <code>destination</code> is undefined.
	 * @throws IOException if the file cannot be opened for writing.
	 */
	public SerialReader ( final String name, final InputStream input )
		throws IOException
	{
		componentHelper = new ComponentHelper( name, this );
		filename = null;
		this.input = new ObjectInputStream(
			input instanceof BufferedInputStream
			? input
			: new BufferedInputStream( input )
		);
	}

	/* ===================== Public Methods ===================== */

	public String getName () {
		return componentHelper.name;
	}

	/**
	 * Deserialise a stream of
	 * {@link Int.ecb.gesmescb.GESMESCBObject}s.
	 * These then are dispatched to all listeners.
	 */
	public void run () {
		componentHelper.logger.log(
			Level.FINE,
			componentHelper.logName,
			"run()",
			"Starting to read."
		);
		while ( hasNext() ) {
			final Object object = next();
			componentHelper.logger.log(
				Level.FINEST,
				componentHelper.logName,
				"run()",
				"Dispatching object.",
				object
			);
			dispatch( object );
		}
		componentHelper.logger.log(
			Level.FINE,
			componentHelper.logName,
			"run()",
			"Finished reading."
		);
	}

	/** Read the next object from the stream.
	 * This method can be used to read the stream one object at
	 * at a time instead of all objects in a bunch.
	 */
	public Serializable next () {
		if ( hasNext() ) {
			final Object object = next;
			next = null;
			return (Serializable) object;
		} else {
			throw new NoSuchElementException(
				"Reached end of object stream."
			);
		}
	}

	/** Return whether there are more objects in the stream.
	 * @see #next()
	 */
	public boolean hasNext () {
		if ( next == null ) {
			try {
				next = input.readObject();
			}
			catch ( final EOFException eof ) {
				// BB: Does anyone know a better handling of EOF for ObjectInput?
				componentHelper.logger.log(
					Level.FINER,
					componentHelper.logName,
					"hasNext()",
					"Reached end of object stream.",
					eof
				);
			}
			catch ( final Exception e ) {
				final String desc = "Problem: " + e;

				componentHelper.logger.log(
					Level.SEVERE,
					getName(),
					"hasNext()",
					desc,
					e
				);
				dispatch(
					new ErrorEvent( desc, ErrorEvent.ERROR )
				);
			}
		}
		return next != null;
	}

	/**
	 * Close the <code>SerialReader</code>.
	 */
	public void close () {
		try {
			if ( filename != null ) {
				input.close();
			}
		}
		catch ( final Exception e ) {
			final String desc = "Problem: " + e.getMessage() + ".";

			componentHelper.logger.log(
				Level.SEVERE,
				getName(),
				"close()",
				desc,
				e
			);
			dispatch(
				new ErrorEvent( desc, ErrorEvent.ERROR )
			);
		}
	}
}
