// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2003, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.event;

import java.io.OutputStream;
import java.io.FileOutputStream;
import java.io.BufferedOutputStream;
import java.io.ObjectOutput;
import java.io.ObjectOutputStream;
import java.io.IOException;

import Int.ecb.logging.Level;

import Int.ecb.gesmescb.*;

import Int.ecb.gesmescb.event.ErrorEvent;
import Int.ecb.gesmescb.event.GESMESCBWriter;
import Int.ecb.gesmescb.event.StructureListener;
import Int.ecb.gesmescb.event.DataListener;
import Int.ecb.gesmescb.event.GESMESCBEventListenerAdapter;

import Int.ecb.gesmescb.util.Named;
import Int.ecb.gesmescb.util.ComponentHelper;

/* ***************************** Class ****************************** */

/**
 * <p>Writes serialized {@link Int.ecb.gesmescb.GESMESCBObject}s.</p>
 *
 * <center><table border="3" cellpadding="3">
 * <caption>
 * <a name="prefs">Available {@link Int.ecb.prefs preference} settings</a>
 * </caption>
 * <tr>
 * <th>preference name</th>
 * <th>type</th>
 * <th>meaning</th>
 * <th>default</th>
 * </tr>
 * <tr><td><code>destination</td><td>string</td><td>output file path</td></tr>
 * </table></center>
 *
 * @author
 * <a href="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</a>
 *
 * @version @(#)$Id: SerialWriter.java,v 1.1 2017/01/19 16:14:03 trainot Exp $
 */
public class SerialWriter extends GESMESCBEventListenerAdapter
	implements Named, StructureListener, DataListener, GESMESCBWriter
{
	/* ========================= Fields ========================= */

	/**
	 * Takes care of logging and preferences.
	 */
	protected final ComponentHelper componentHelper;

	/** Target for serialized output. */
	private final ObjectOutput output;

	/**
	 * The file to write to
	 * if the <code>SerialWriter</code> has been
	 * constructed from a filename.
	 * Otherwise it is null.
	 */
	private final String filename;

	/* ====================== Constructors ====================== */

	/**
	 * Construct a <code>SerialWriter</code> to write to a file.
	 *
	 * @param name used for logging and preferences.
	 * @param filename where to write to.
	 *
	 * @throws NullPointerException if <code>filename==null</code>
	 * and the preference variable
	 * <code>destination</code> is undefined.
	 * @throws IOException if the file cannot be opened for writing.
	 */
	public SerialWriter ( final String name, final String filename )
		throws IOException
	{
		componentHelper = new ComponentHelper( name, this );
		this.filename = filename != null
			? filename
			: componentHelper.getPreference( "destination", null );
		if ( this.filename == null ) {
			final String desc
			= "No preference variable \"destination\" found"
				+ " to determine the output file.";
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"SerialWriter()",
				desc
			);
			throw new NullPointerException( desc );
		}
		output = new ObjectOutputStream(
			new BufferedOutputStream(
				new FileOutputStream( this.filename )
			)
		);
	}

	/**
	 * Construct a <code>SerialWriter</code>
	 * writing to a file according to preferences.
	 *
	 * @param name used for logging and preferences.
	 *
	 * @throws NullPointerException if the preference variable
	 * <code>destination</code> is undefined.
	 * @throws IOException if the file cannot be opened for writing.
	 */
	public SerialWriter ( final String name )
		throws IOException
	{
		this( name, (String) null );
	}

	/**
	 * Construct a <code>SerialWriter</code> writing to a byte stream.
	 *
	 * @param name used for logging and preferences.
	 *
	 * @throws NullPointerException if the preference variable
	 * <code>destination</code> is undefined.
	 * @throws IOException if the file cannot be opened for writing.
	 */
	public SerialWriter ( final String name, final OutputStream output )
		throws IOException
	{
		componentHelper = new ComponentHelper( name, this );
		filename = null;
		this.output = new ObjectOutputStream(
			output instanceof BufferedOutputStream
			? output
			: new BufferedOutputStream( output )
		);
	}

	/* ===================== Public Methods ===================== */

	public String getName () {
		return componentHelper.name;
	}

	/**
	 * Close the <code>SerialWriter</code>.
	 */
	public void close () {
		try {
			if ( filename != null ) {
				output.close();
			} else {
				output.flush();
			}
		}
		catch ( final IOException ioe ) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"close()",
				"Closing failed: " + ioe,
				ioe
			);
		}
		componentHelper.logger.log(
			Level.FINE,
			componentHelper.logName,
			"close()",
			"Closing"
		);
	}

	/* ------------------------ Listener ------------------------ */

	/* ---------------------- Interchange ----------------------- */

	public void handle ( final InterchangeBegin interchangeBegin ) {
		try {
			componentHelper.logger.log(
				Level.FINE,
				componentHelper.logName,
				"handle( InterchangeBegin )",
				"Processing InterchangeBegin",
				interchangeBegin
			);
			output.writeObject( interchangeBegin );
		}
		catch ( final IOException ioe ) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"handle( InterchangeBegin )",
				"Output problem.",
				ioe
			);
			close();
		}
	}

	public void handle ( final InterchangeEnd interchangeEnd ) {
		try {
			componentHelper.logger.log(
				Level.FINE,
				componentHelper.logName,
				"handle( InterchangeEnd )",
				"Processing InterchangeEnd",
				interchangeEnd
			);
			output.writeObject( interchangeEnd );
		}
		catch ( final IOException ioe ) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"handle( InterchangeEnd )",
				"Output problem.",
				ioe
			);
			close();
		}
	}

	/* ------------------------ Message ------------------------- */

	public void handle ( final MessageBegin messageBegin ) {
		try {
			componentHelper.logger.log(
				Level.FINE,
				componentHelper.logName,
				"handle( MessageBegin )",
				"Processing MessageBegin",
				messageBegin
			);
			output.writeObject( messageBegin );
		}
		catch ( final IOException ioe ) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"handle( MessageBegin )",
				"Output problem.",
				ioe
			);
			close();
		}
	}

	public void handle ( final MessageEnd messageEnd ) {
		try {
			componentHelper.logger.log(
				Level.FINE,
				componentHelper.logName,
				"handle( MessageEnd )",
				"Processing MessageEnd",
				messageEnd
			);
			output.writeObject( messageEnd );
		}
		catch ( final IOException ioe ) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"handle( MessageEnd )",
				"Output problem.",
				ioe
			);
			close();
		}
	}

	/* ------------------------ CodeList ------------------------ */

	public void handle ( final CodeList codeList ) {
		try {
			componentHelper.logger.log(
				Level.FINER,
				componentHelper.logName,
				"handle( CodeList )",
				"Writing CodeList",
				codeList
			);
			output.writeObject( codeList );
		}
		catch ( final IOException ioe ) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"handle( CodeList )",
				"Output problem.",
				ioe
			);
			close();
		}
	}

	/* ------------------- StatisticalConcept ------------------- */

	public void handle ( final StatisticalConcept statisticalConcept ) {
		try {
			componentHelper.logger.log(
				Level.FINER,
				componentHelper.logName,
				"handle( StatisticalConcept )",
				"Writing StatisticalConcept",
				statisticalConcept
			);
			output.writeObject( statisticalConcept );
		}
		catch ( final IOException ioe ) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"handle( StatisticalConcept )",
				"Output problem.",
				ioe
			);
			close();
		}
	}

	/* ----------------------- KeyFamily ------------------------ */

	public void handle ( final KeyFamily keyFamily ) {
		try {
			componentHelper.logger.log(
				Level.FINER,
				componentHelper.logName,
				"handle( KeyFamily )",
				"Writing KeyFamily",
				keyFamily
			);
			output.writeObject( keyFamily );
		}
		catch ( final IOException ioe ) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"handle( KeyFamily )",
				"Output problem.",
				ioe
			);
			close();
		}
	}

	/* ---------------------- Time Series ----------------------- */

	public void handle ( final TimeSeries timeSeries ) {
		try {
			componentHelper.logger.log(
				Level.FINER,
				componentHelper.logName,
				"handle( TimeSeries )",
				"Writing TimeSeries",
				timeSeries
			);
			output.writeObject( timeSeries );
		}
		catch ( final IOException ioe ) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"handle( TimeSeries )",
				"Output problem.",
				ioe
			);
			close();
		}
	}

	/* ----------------------- Attribute ------------------------ */

	public void handle ( final Attribute attribute ) {
		try {
			componentHelper.logger.log(
				Level.FINER,
				componentHelper.logName,
				"handle( Attribute )",
				"Writing Attribute",
				attribute
			);
			output.writeObject( attribute );
		}
		catch ( final IOException ioe ) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"handle( Attribute )",
				"Output problem.",
				ioe
			);
			close();
		}
	}

	/* ----------------------- ErrorEvent ----------------------- */

	public void handle ( final ErrorEvent errorEvent ) {
		super.handle( errorEvent );
		try {
			componentHelper.logger.log(
				Level.FINER,
				componentHelper.logName,
				"handle( ErrorEvent )",
				"Writing ErrorEvent",
				errorEvent
			);
			output.writeObject( errorEvent );
		}
		catch ( final IOException ioe ) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"handle( ErrorEvent )",
				"Output problem.",
				ioe
			);
			close();
		}
	}
}
