// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, (c) 2003, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.event;

import java.util.LinkedList;
import java.util.Iterator;
import java.util.Collections;
import java.util.Arrays;

import junit.framework.Assert;

import Int.ecb.gesmescb.*;

/* **************************** Class ***************************** */

/**
 * Listens to an event stream and stores a limited amount of it.
 * This class is intended to be used for testing only.
 *
 * @see TestEventSource
 *
 * @author
 * <A HREF="mailto:Sami.Airo@ECB.int">Sami Airo</A>
 * @author
 * <A HREF="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</A>
 *
 * @version @(#)$Id: TestEventListener.java,v 1.1 2017/01/19 16:14:04 trainot Exp $
 */
public class TestEventListener extends GESMESCBEventListenerAdapter
	implements DataListener, StructureListener, GESMESCBWriter
{
	/* ========================= Fields ========================= */

	/** List of expected events. */
	private final Iterator expectation;

	/**
	 * History of stored events.
	 */
	private final LinkedList events = new LinkedList();

	/**
	 * Maximum number of events stored.
	 */
	private final int capacity;

	/* ====================== Constructors ====================== */

	/**
	 * Construct a clean and fresh TestEventListener.
	 */
	public TestEventListener () {
		capacity = 0;
		expectation = null;
	}

	/**
	 * Construct a clean and fresh TestEventListener with bounded size.
	 *
	 * @param capacity maximum number of events to be stored.
	 * This must be a positive number.
	 */
	public TestEventListener ( final int capacity ) {
		if ( capacity < 0 ) {
			throw new IllegalArgumentException(
				"Constructor does not like negative capacity."
			);
		}
		this.capacity = capacity;
		expectation = null;
	}

	/**
	 * Construct a clean and fresh TestEventListener which expects
	 * a given sequence of events.
	 */
	public TestEventListener( final Object[] expectation ) {
		capacity = 0;
		this.expectation = Arrays.asList( expectation ).iterator();
	}

	/* ===================== Public Methods ===================== */

	public void handle ( final InterchangeBegin InterchangeBegin ) {
		handle( (Object) InterchangeBegin );
	}

	public void handle ( final InterchangeEnd InterchangeEnd ) {
		handle( (Object) InterchangeEnd );
	}

	public void handle ( final MessageBegin MessageBegin ) {
		handle( (Object) MessageBegin );
	}

	public void handle ( final MessageEnd MessageEnd ) {
		handle( (Object) MessageEnd );
	}

	public void handle ( final CodeList CodeList ) {
		handle( (Object) CodeList );
	}

	public void handle ( final StatisticalConcept StatisticalConcept ) {
		handle( (Object) StatisticalConcept );
	}

	public void handle ( final KeyFamily keyFamily ) {
		handle( (Object) keyFamily );
	}

	public void handle ( final TimeSeries TimeSeries ) {
		handle( (Object) TimeSeries );
	}

	public void handle ( final Attribute Attribute ) {
		handle( (Object) Attribute );
	}

	public void handle ( final ErrorEvent ErrorEvent ) {
		handle( (Object) ErrorEvent );
	}

	public void close () {
		if ( expectation != null ) {
			Assert.assertTrue(
				"Too few events received.",
				!expectation.hasNext()
			);
		}
	}

	/**
	 * Returns the number of events stored.
	 */
	public int size () {
		return events.size();
	}

	/**
	 * Returns the last object received.
	 */
	public Object getLastObject () {
		if ( 0 < size() ) {
			return events.getLast();
		} else {
			return null;
		}
	}

	/**
	 * Get an unmodifiable iterator over all stored information.
	 */
	public Iterator iterator () {
		return Collections.unmodifiableList( events ).iterator();
	}

	/**
	 * Clears the history of stored events.
	 */
	public void clearHistory () {
		events.clear();
	}

	/* =================== Protected Methods ==================== */

	/** Verify that the expected event matches the received one.
	 * Override this method to control the degree of equality required.
	 *
	 * @throws junit.framework.AssertionFailedError in case of mismatch.
	 */
	protected void assertEquals ( final Object expected, final Object received ) {
		Assert.assertEquals( "Output", expected, received );
	}

	/* ==================== Private Methods ===================== */

	private void handle ( final Object event ) {
		if ( 0 < capacity && capacity <= size() ) {
			events.removeFirst();
		}
		events.addLast( event );
		if ( expectation != null ) {
			assertEquals( expectation.next(), event );
		}
	}
}
