// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, (c) 2003, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.event;

import java.util.LinkedList;

import Int.ecb.logging.Level;

import Int.ecb.gesmescb.*;

import Int.ecb.gesmescb.util.ComponentHelper;

/* ***************************** Class ****************************** */

/**
 * <p>Collects events to a queue
 * and on its other end, pushes them out
 * using a dedicated {@link Thread}.</P>
 *
 * <P>This class is central to multithreading.
 * It can be plugged in between a source and a listener
 * to separate the threads they are running in from each other.
 * Using <CODE>java -native</CODE>, this can make it possible
 * to use multiple CPUs if available on the system.</P>
 *
 * <P>The intended use of multithreading is as follows:
 * <OL>
 * <LI>Build up the whole source-listener graph using
 * {@link GESMESCBEventSource#addListener}.</LI>
 * <LI>Run {@link GESMESCBReader#run}.</LI>
 * <LI>Call {@link GESMESCBReader#close}.</LI>
 * <LI>For any <CODE>ThreadPump</CODE> used, starting
 * {@link #close}. Proceed downstream, meaning
 * begin with those <CODE>ThreadPump</CODE>s which are next
 * to the <CODE>GESMESCBReader</CODE>.
 * The reason for this is that <CODE>close</CODE> will
 * make the <CODE>ThreadPump</CODE> inactive so that it will
 * not accept any more incoming data
 * and flush the internal queue.
 * So the senders upstream must deliver no more objects
 * while the receivers downstream must still be capable of
 * properly handling all information.</LI>
 * <LI>Call {@link GESMESCBWriter#close} for all
 * {@link GESMESCBWriter}s.</LI>
 * </OL>
 * This proposed sequence takes as an example a
 * tree structured source-listener graph.
 * If there are listeners listening to more
 * than one source, the directed graph should still not
 * contain circles.
 * Then the <B>general rules</B> are:
 * <UL>
 * <LI>On a {@link GESMESCBReader} never call
 * {@link GESMESCBReader#close} before {@link GESMESCBReader#run}
 * has returned.<LI>
 * <LI>On all other objects which must be <CODE>close</CODE>d
 * never call <CODE>close</CODE> before
 * having <CODE>close</CODE>d all upstream objects which need closing.
 * </LI>
 * <LI>On all objects which must be <CODE>close</CODE>d
 * always call <CODE>close</CODE> before program termination.</LI>
 * </UL>
 * </P>
 *
 * <P>The <CODE>ThreadPump</CODE> supports automatic disconnect.
 * If no one listens to it when an event is dispatched,
 * then it will also stop listening.
 * It then also ends the distribution <CODE>Thread</CODE>.</P>
 *
 * @author
 * <A HREF="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</A>
 *
 * @version @(#)$Id: ThreadPump.java,v 1.1 2017/01/19 16:14:03 trainot Exp $
 */

public final class ThreadPump
	extends GESMESCBEventBroker
	implements DataListener, StructureListener, Runnable
{
	/* ========================= Fields ========================= */

	/** The ComponentHelper used for logging. */
	private final ComponentHelper componentHelper;

	/**
	 * The distribution thread.
	 * This dispatches events from the queue.
	 */
	private final Thread thread;

	/**
	 * The event queue.
	 */
	private final LinkedList queue;

	/**
	 * The limit for the queue size.
	 */
	private final int sizeLimit;

	/* ====================== Constructors ====================== */

	/**
	 * Construct a ThreadPump.
	 *
	 * @param sizeLimit limits the growth of the internal queue.
	 * This avoids that the feeding thread would
	 * build up a huge queue in memory
	 * containing all data and thereby stress the system.
	 *
	 * @throws IllegalArgumentException if <CODE>sizeLimit</CODE>
	 * is less than one.
	 */
	public ThreadPump ( final int sizeLimit ) {
		super( "", true );

		// Set up fields for logging.
		componentHelper = new ComponentHelper( getClass().getName(), this );

		queue = new LinkedList();
		if ( 1 > ( this.sizeLimit = sizeLimit ) ) {
			throw new IllegalArgumentException (
				"The ThreadPump.sizeLimit must be at least one."
			);
		}
		thread = new Thread( this, "PumpThread" );
		thread.start();
	}

	/* ===================== Public Methods ===================== */

	/* ------------------------ Listener ------------------------ */

	public void handle ( final InterchangeBegin interchangeBegin ) {
		handle( (Object) interchangeBegin );
	}

	public void handle ( final InterchangeEnd interchangeEnd ) {
		handle( (Object) interchangeEnd );
	}

	public void handle ( final MessageBegin messageBegin ) {
		handle( (Object) messageBegin );
	}

	public void handle ( final MessageEnd messageEnd ) {
		handle( (Object) messageEnd );
	}

	public void handle ( final CodeList codeList ) {
		handle( (Object) codeList );
	}

	public void handle ( final StatisticalConcept statisticalConcept ) {
		handle( (Object) statisticalConcept );
	}

	public void handle ( final KeyFamily keyFamily ) {
		handle( (Object) keyFamily );
	}

	public void handle ( final TimeSeries timeSeries ) {
		handle( (Object) timeSeries );
	}

	public void handle ( final Attribute attribute ) {
		handle( (Object) attribute );
	}

	public void handle ( final ErrorEvent errorEvent ) {
		handle( (Object) errorEvent );
	}

	/* ------------------ Other public methods ------------------ */

	/**
	 * Internally used to implement {@link Runnable}.
	 * Do not call this method.
	 * Its purpose is to take events
	 * from the internal queue and dispatch them to the
	 * registered listeners.
	 *
	 * @throws SecurityException
	 */
	public void run () {
		// If called from the outside in breach of contract.
		if ( Thread.currentThread() != thread ) {
			throw new SecurityException(
				"You tried to call ThreadPump.run()."
			);
		}
		componentHelper.logger.log(
			Level.INFO,
			componentHelper.logName,
			"run()",
			"Started pumping."
		);
		try {
			while ( isActive() || !isEmpty() ) {
				final Object event = getEventFromQueue();
				if ( event == null ) break;	// see close()
				componentHelper.logger.log(
					Level.FINER,
					componentHelper.logName,
					"run()",
					"pump",
					event
				);
				// A bit dirty, package only.
				sourceHelper.dispatch( event );
			}
		}
		catch ( final Throwable t ) {
			final String description
			= "A " + t.getClass().getName()
				+ " was thrown upon the pump.";
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"run()",
				description,
				t
			);
			// Maybe not all will receive this,
			// but at least some will.
			// And the nasty thrower one is among them.
			dispatch( new ErrorEvent(
				description, ErrorEvent.FAILURE
			) );
		}
		componentHelper.logger.log(
			Level.INFO,
			componentHelper.logName,
			"run()",
			"Finished pumping."
		);
	}

	/**
	 * <P>Complete the pumping activities and close down.
	 * This method must be called when the data transfer is
	 * complete so that the <CODE>ThreadPump</CODE>
	 * has a chance to write all remaining events from
	 * the internal queue and afterwards
	 * shut down and free its resources.
	 *
	 * @throws NullPointerException if called twice.
	 */
	public void close () {
		componentHelper.logger.log(
			Level.FINE,
			componentHelper.logName,
			"close()",
			"Close pumping."
		);

		setActive( false );

		componentHelper.logger.log(
			Level.FINE,
			componentHelper.logName,
			"close()",
			"Waiting for pump thread to join."
		);
		try {
			thread.join();
		}
		catch ( final InterruptedException irre ) {
			// Should not happen.
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"close()",
				"Interrupted while waiting for pump thread to join.",
				irre
			);
		}

		componentHelper.logger.log(
			Level.FINE,
			componentHelper.logName,
			"close()",
			"Pump thread has joined."
		);
	}

	/**
	 * More or less forcefully free the resources.
	 */
	public void finalize () throws Throwable {
		// Use the recommended try-finally idiom for finalize().
		try {
			if ( isActive() ) {
				try {
					componentHelper.logger.log(
						Level.SEVERE,
						componentHelper.logName,
						"finalize()",
						"Program implementor forgot"
						+ " to close ThreadPump!"
					);
				}
				finally {
					close();
				}
			}
		}
		finally {
			super.finalize();
		}
	}

	/* =================== Protected Methods ==================== */

	protected void setActive ( final boolean active ) {
		synchronized ( queue ) {
			super.setActive( false );
			queue.notifyAll();
		}
	}

	/* ==================== Private Methods ===================== */

	/**
	 * Add an event to the internal event queue.
	 *
	 * @param event to be handled.
	 */
	private void handle ( final Object event ) {
		if ( event == null ) {
			// Should never happen.
			final String description = "Got a null reference to handle.";
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"handle(Object)",
				description
			);
			// Continuing makes no sense for the listeners.
			dispatch( new ErrorEvent(
				description, ErrorEvent.FAILURE
			) );
			setActive( false );
		} else {
			try {
				putEventToQueue( event );
			}
			catch ( final InterruptedException irre ) {
				// Should never happen.
				final String description = "Interrupted while pumping.";
				componentHelper.logger.log(
					Level.SEVERE,
					componentHelper.logName,
					"handle()",
					description,
					irre
				);
				// Continuing makes no sense for the listeners.
				dispatch( new ErrorEvent(
					description, ErrorEvent.FAILURE
				) );
				setActive( false );
			}
		}
	}

	/**
	 * Put an event to the queue.
	 * This method blocks if the queue hits the size limit.
	 *
	 * @param event to be put on the queue.
	 *
	 * @throws InterruptedException if the internal thread was interrupted.
	 */
	private void putEventToQueue ( final Object event )
		throws InterruptedException
	{
		// Force the queue to be completely emptied
		// after MessageEnd and ErrorEvents
		// so that the Metabase is updated before
		// an EDIFACTReader may query it to parse the
		// next message.
		final boolean needTotalSynchronize
		= event instanceof MessageEnd || event instanceof ErrorEvent;

		// Synchronize on the queue, not on the ThreadPump
		// to hide the lock from the outside.
		synchronized ( queue ) {
			while ( isActive() && queue.size() >= sizeLimit ) {
				queue.wait();
			}
			queue.addLast( event );

			// There is only one one thread reading,
			// namely the private thread.
			// So notifyAll and notify are equivalent.
			// General preference goes for notifyAll.
			queue.notifyAll();

			// If total sync is needed, enforce queue is emptied
			// before accepting further events.
			if ( needTotalSynchronize ) {
				while ( isActive() && !queue.isEmpty() ) {
					queue.wait();
				}
			}
		}
	}

	/**
	 * Get the next event from the queue.
	 * This method blocks until such an element is available.
	 *
	 * @return the next event from the queue.
	 *
	 * @throws InterruptedException if the internal thread was interrupted.
	 */
	private Object getEventFromQueue ()
		throws InterruptedException
	{
		Object event;

		// Synchronize on the queue, not on the object
		// to hide the lock from the outside.
		synchronized ( queue ) {
			while ( isActive() && queue.isEmpty() ) {
				queue.wait();
			}
			event = queue.isEmpty()
				? null
				: queue.removeFirst();
			queue.notifyAll();
		}
		return event;
	}

	/**
	 * Is the queue empty.
	 * Synchronized on the queue to respect the Java Virtual Memory Model.
	 *
	 * @return Whether the internal queue is empty.
	 */
	private boolean isEmpty () {
		synchronized ( queue ) {
			return queue.isEmpty();
		}
	}
}
