// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.fame;

import java.util.Iterator;

import com.fame.timeiq.dates.*;

import Int.ecb.logging.Level;

import Int.ecb.gesmescb.*;
import Int.ecb.gesmescb.fame.FAMEDb;
import Int.ecb.gesmescb.fame.FAMEDeleter;

import Int.ecb.gesmescb.util.ComponentHelper;

/* **************************** Class ***************************** */

/**
 * Deletes time series and attributes from FAME databases.
 *
 * @author
 * <A HREF="mailto:klo@FAME.com">Kam Lo</A>
 * @author
 * <A HREF="mailto:Sami.Airo@ECB.int">Sami Airo</A>
 *
 * @version @(#)$Id: ESDBDeleter.java,v 1.1 2017/01/19 16:14:03 trainot Exp $
 */

public class ESDBDeleter extends FAMEDeleter {

	/* ======================= Constants ======================== */

	/* ========================= Fields ========================= */

	/* ====================== Constructors ====================== */

	public ESDBDeleter(
		final ComponentHelper componentHelper,
		final FAMEDb fameDb,
		final MessageBegin messageBegin
	) {
		super(componentHelper, fameDb, messageBegin);
	}

	/* ===================== Protected Methods ===================== */

	protected boolean deleteAttributeObservation (Attribute attribute , KeyFamilyComponent component) {
		getComponentHelper().logger.log(
			Level.FINER,
			getComponentHelper().logName,
			"deleteAttributeObservation",
			"Deleting observation level attribute value from " + attribute
		);

		IndexValueRange range = getFameRange(attribute.getInterval());
		
		if (range == null) {
			return false;
		}

		return getFAMEDb().deleteTimeSeriesObservations( getFameName(attribute, component), range );
	}

	protected boolean deleteAttribute (Attribute attribute , KeyFamilyComponent component) {
		getComponentHelper().logger.log(
			Level.FINER,
			getComponentHelper().logName,
			"deleteAttribute",
			"Deleting attribute " + attribute
		);

		return getFAMEDb().deleteObject( getFameName(attribute, component) );
	}

	/**
	 * Deletes observations for the specified range.
	 *
	 * @param timeSeries	timeSeries to delete observations from
	 *
	 * @return	True if the observations were deleted succesfully or the time series
	 *			was not in the database; False otherwise.
	 */
	protected boolean deleteObservations (TimeSeries timeSeries) {
		String fameName = getFameName(timeSeries);
		IndexValueRange range = getFameRange(timeSeries.getInterval());

		getComponentHelper().logger.log(
			Level.FINER,
			getComponentHelper().logName,
			"deleteObservations",
			"Deleting observations " + timeSeries
		);

		if (range == null) {
			return false;
		}

		return (		getFAMEDb().deleteTimeSeriesObservations(fameName, range)
					&&	getFAMEDb().deleteTimeSeriesObservations(fameName + KEY_SEPARATOR + "OBS_STATUS", range)
					&& getFAMEDb().deleteTimeSeriesObservations(fameName + KEY_SEPARATOR + "OBS_CONF", range)
					&& getFAMEDb().deleteTimeSeriesObservations(fameName + KEY_SEPARATOR + "BREAK", range) );
	}

	/**
	 * Deletes a time series set from the database.
	 * Deletes all all series and observation level attributes.
	 *
	 * @param timeSeries	timeSeries to delete
	 *
	 * @return True if the time series was deleted succesfully or it was not in the database; False otherwise.
	 */
	protected boolean deleteTimeSeries (TimeSeries timeSeries) {
		getComponentHelper().logger.log(
			Level.FINER,
			getComponentHelper().logName,
			"deleteTimeSeries",
			"Deleting time series " + timeSeries
		);

		// Delete the series itself
		if (!getFAMEDb().deleteObject( getFameName(timeSeries) )) {
			return false;
		}

		// Delete all the attributes
		Iterator it = getKeyFamily().componentIterator();
		while (it.hasNext()) {
			KeyFamilyComponent comp = (KeyFamilyComponent)it.next();

			if ((comp.isAttribute() && (comp.isAttachedToTimeSeries() || comp.isAttachedToObservation()))
					|| comp.isArrayCell()) {
				if ( ! deleteAttribute(
							new Attribute(
									timeSeries.getKey(),
									"",
									"",
									comp.getConceptId(),
									""
							),
							comp
						)
					)
				{
					return false;
				}
			}
		}

		if ( ! deleteAttribute(
					new Attribute(
						timeSeries.getKey(),
						"",
						"",
						"BREAK",
						""
					),
					getKeyFamily().getComponent("OBS_PRE_BREAK")
				)
			)
		{
			return false;
		}

		return true;
	}

	/**
	 * Deletes a full sibling set from the database.
	 * Deletes all times series with series and observation level attributes together
	 * with the sibling level attributes.
	 *
	 * @param timeSeries	timeSeries that contains the sibling key to delete
	 *
	 * @return True if the sibling set was deleted succesfully or it was not in the database; False otherwise.
	 */
	protected boolean deleteSiblingSet (TimeSeries timeSeries) {
		Iterator it = null;

		getComponentHelper().logger.log(
			Level.FINER,
			getComponentHelper().logName,
			"deleteSiblingSet",
			"Deleting sibling set " + timeSeries
		);

		//Delete all time series in the sibling group
		it = getSiblingSet(timeSeries.getKey()).iterator();
		while (it.hasNext()) {
			if (!deleteTimeSeries( (TimeSeries)it.next() )) return false;
		}

		// Delete all sibling level attributes
		it = getKeyFamily().componentIterator();
		while (it.hasNext()) {
			KeyFamilyComponent comp = (KeyFamilyComponent)it.next();
				if (comp.isAttribute() && comp.isAttachedToSiblingGroup()) {
				if ( !deleteAttribute(
							new Attribute(
									timeSeries.getKey(),
									"",
									"",
									comp.getConceptId(),
									""
							),
							comp
						)
					)
				{
					return false;
				}
			}
		}

		return true;
	}

}
