// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.fame;

import java.util.Iterator;

import com.fame.timeiq.*;
import com.fame.timeiq.data.DoubleList;
import com.fame.timeiq.data.StringList;
import com.fame.timeiq.data.ObservationList;
import com.fame.timeiq.dates.*;

import Int.ecb.logging.Level;

import Int.ecb.gesmescb.*;

import Int.ecb.gesmescb.fame.FAMEDb;
import Int.ecb.gesmescb.fame.FAMERevisor;

import Int.ecb.gesmescb.util.ComponentHelper;

/* **************************** Class ***************************** */

/**
 * Writes time series and attribute updates to a FAME database.
 *
 * <P>Configuration:<UL>
 * <LI>MissingValueInFame: Value to be used for reported missing values (NC or ND)
 * <LI>DefaultObservedValue: Default value to be used for Observed attribute
 * (default AVERAGED). See TiqObserved for allowed values.
 * </UL></P>
 *
 * @author
 * <A HREF="mailto:klo@FAME.com">Kam Lo</A>
 * @author
 * <A HREF="mailto:Sami.Airo@ECB.int">Sami Airo</A>
 *
 * @version @(#)$Id: ESDBUpdater.java,v 1.1 2017/01/19 16:14:03 trainot Exp $
 */

public class ESDBUpdater extends FAMERevisor {

	/* ======================= Constants ======================== */

	/* ========================= Fields ========================= */

	/** Missing value indicator in the Fame database */
	private byte missingValueInFame = TiqConstants.STATUS_NC;

	/** Default value to use for observed attribute */
	TiqObserved defaultObservedValue;

	/* ====================== Constructors ====================== */

	public ESDBUpdater(
		final ComponentHelper componentHelper,
		final FAMEDb fameDb,
		final MessageBegin messageBegin
	) {
		super(componentHelper, fameDb, messageBegin);

		String missingValue = getComponentHelper().getPreference("MissingValueInFame", "NC");
		if (missingValue.equals("ND")) {
			missingValueInFame = TiqConstants.STATUS_ND;
		} else {
			if (! missingValue.equals("NC")) {
				getComponentHelper().logger.log(
					Level.WARNING,
					getComponentHelper().logName,
					"ESDBUpdater",
					"Invalid preference MissingValueInFame: " + missingValue
				);
			}
		}

		getComponentHelper().logger.log(
			Level.CONFIG,
			getComponentHelper().logName,
			"ESDBUpdater",
			"MissingValueInFame: " + TiqHelper.statusToString(missingValueInFame)
		);

		String defaultObservedString = getComponentHelper().getPreference("DefaultObservedValue", "AVERAGED");
		defaultObservedValue = TiqObserved.fromString(defaultObservedString);
		if (defaultObservedValue == null) {
				getComponentHelper().logger.log(
					Level.WARNING,
					getComponentHelper().logName,
					"ESDBUpdater",
					"Invalid preference DefaultObservedValue: " + defaultObservedString
				);
			defaultObservedValue = TiqObserved.AVERAGED;
		}

		getComponentHelper().logger.log(
			Level.CONFIG,
			getComponentHelper().logName,
			"ESDBUpdater",
			"DefaultObservedValue: " + defaultObservedValue
		);
	}

	/* ===================== Public Methods ===================== */

	public boolean handle (Attribute attribute) {
		KeyFamilyComponent component = getKeyFamily().getComponent(attribute.getName());
		String fameName = getFameName(attribute, component);

		if (component.isAttachedToObservation()) {
			IndexValueRange range = getFameRange(attribute.getInterval());

			if (range == null) {
				return false;
			}

			// Size of range
			int rangeSize = range.size();

			RegularSeries regSeries;

			byte[] statusValues = new byte[rangeSize];

			if (component.isNumeric()) {
				double[] doubleValues = new double[rangeSize];
				double value = Double.parseDouble(attribute.getValue());

				for (int i = 0; i < rangeSize; i++) {
					doubleValues[i] = value;
					statusValues[i] = TiqConstants.STATUS_OK;
				}

				regSeries = new RegularSeries(range, new DoubleList(doubleValues), statusValues);
				regSeries.setObserved(defaultObservedValue);
			} else {
				String[] stringValues = new String[rangeSize];
				String stringValue = attribute.getValue();

				for (int i = 0; i < rangeSize; i++) {
					stringValues[i] = stringValue;
					statusValues[i] = TiqConstants.STATUS_OK;
				}

				regSeries = new RegularSeries(range, new StringList(stringValues), statusValues);
			}

			return getFAMEDb().createUpdateTimeSeries(fameName, regSeries);
		}

		Scalar attribObj;
		
		
// All attributes are currently stored as strings...
//		if (component.isNumeric()) {
//			attribObj = new Scalar(new Double(attribute.getValue()), TiqConstants.STATUS_OK);
//		} else {
		/**
		 * check if attribute is at dataset level
		 */
		if (attribute.getKey().isEmpty()) {
			attribute.setName(getDataSetId()+"." + attribute.getName());
		}
		attribObj = new Scalar(attribute.getValue());
//		}
		if (!updateObservedValues(attribute, component)) return false;

		return getFAMEDb().overwriteObject(fameName, attribObj);
	}

	/**
	 * Writes updates to or creates a time series data in a database.
	 *
	 * @param timeSeries	 TimeSeries to update/create
	 *
	 * @return boolean indicating success or failure
	 */
	public boolean handle (TimeSeries timeSeries) {
		String fameName = getFameName(timeSeries);
		IndexValueRange range = getFameRange(timeSeries.getInterval());

		if (range == null) {
			return false;
		}

		// Convert vectors to data type arrays which are used by
		// TimeIQ functions.
		int obsCount = timeSeries.length();

		double[] valueArray = new double[obsCount];
		byte[]	 valueStatusArray = new byte[obsCount];

		String[] statusArray = new String[obsCount];
		byte[]	 statusStatusArray = new byte[obsCount];

		String[] confArray = new String[obsCount];
		byte[]	 confStatusArray = new byte[obsCount];

		double[] preBreakArray = new double[obsCount];
		byte[]	 preBreakStatusArray = new byte[obsCount];

		Iterator iterator;
		Int.ecb.gesmescb.Observation obs;

		ObservationList storedValues = null;
		ObservationList storedStatuses = null;
		ObservationList storedConfs = null;
		ObservationList storedPreBreaks = null;
		boolean seriesIsInDatabase = false;

		iterator = timeSeries.observationIterator();
		while (iterator.hasNext()) {
			obs = (Int.ecb.gesmescb.Observation)(iterator.next());

			if (obs.isEmpty()) {
				try {
					storedValues = getObservations(fameName, range);
					seriesIsInDatabase = (storedValues != null);

					if (seriesIsInDatabase) {
						storedStatuses = getObservations(fameName + KEY_SEPARATOR + "OBS_STATUS", range);
						storedConfs = getObservations(fameName + KEY_SEPARATOR + "OBS_CONF", range);
						storedPreBreaks = getObservations(fameName + KEY_SEPARATOR + "BREAK", range);
					}
				} catch (Exception e) {
					getComponentHelper().logger.log(
						Level.SEVERE,
						getComponentHelper().logName,
						"handle",
						"Error in reading previous observations for time series " + fameName,
						e
					);
					return false;
				}

				break;
			}
		}

		String value;
		int i = 0;
		iterator = timeSeries.observationIterator();

		while (iterator.hasNext()) {
			obs = (Int.ecb.gesmescb.Observation)(iterator.next());

			if (obs.isEmpty()) {
				if (seriesIsInDatabase) {
					valueArray[i] = ((com.fame.timeiq.data.Observation)storedValues.get(i)).getDoubleValue();
					valueStatusArray[i] = ((com.fame.timeiq.data.Observation)storedValues.get(i)).getStatus();

					if (storedStatuses != null) {
						statusArray[i] = ((com.fame.timeiq.data.Observation)storedStatuses.get(i)).getStringValue();
						statusStatusArray[i] = ((com.fame.timeiq.data.Observation)storedStatuses.get(i)).getStatus();
					}

					if (storedConfs != null) {
						confArray[i] = ((com.fame.timeiq.data.Observation)storedConfs.get(i)).getStringValue();
						confStatusArray[i] = ((com.fame.timeiq.data.Observation)storedConfs.get(i)).getStatus();
					}

					if (storedPreBreaks != null) {
						preBreakArray[i] = ((com.fame.timeiq.data.Observation)storedPreBreaks.get(i)).getDoubleValue();
						preBreakStatusArray[i] = ((com.fame.timeiq.data.Observation)storedPreBreaks.get(i)).getStatus();
					}
				} else {
					valueStatusArray[i] = TiqConstants.STATUS_ND;
					statusStatusArray[i] = TiqConstants.STATUS_ND;
					confStatusArray[i] = TiqConstants.STATUS_ND;
					preBreakStatusArray[i] = TiqConstants.STATUS_ND;
				}
			} else {
				if (obs.hasMissingValue()) {
					valueStatusArray[i] = missingValueInFame;
				} else {
					valueArray[i] = Double.parseDouble(obs.getValue());
					valueStatusArray[i] = TiqConstants.STATUS_OK;
				}

				// OBS_STATUS is always reported (unless the whole observation is missing)
				statusArray[i] = obs.getStatus();
				statusStatusArray[i] = TiqConstants.STATUS_OK;

				// Check if the OBS_CONF is missing
				if ((confArray[i] = obs.getConfidentiality()).equals("")) {
					// It is missing set it to ND
					confStatusArray[i] = TiqConstants.STATUS_ND;
				} else {
					// Normal value
					confStatusArray[i] = TiqConstants.STATUS_OK;
				}

				// Check if the OBS_PRE_BREAK is missing
				if ((value = obs.getPreBreakValue()).equals("")) {
					// It is missing set it to ND
					preBreakStatusArray[i] = TiqConstants.STATUS_ND;
				} else {
					if ( obs.hasMissingPreBreakValue() ) {
						// It has a reported missing value
						preBreakStatusArray[i] = TiqConstants.STATUS_NC;
					} else {
						// Normal value
						preBreakArray[i] = Double.parseDouble(value);
						preBreakStatusArray[i] = TiqConstants.STATUS_OK;
					}
				}

			}

			i++;
		}

		RegularSeries valueSeries = new RegularSeries(range, valueArray, valueStatusArray);
		valueSeries.setObserved(defaultObservedValue);
		RegularSeries statusSeries = new RegularSeries(range, statusArray, statusStatusArray);
		RegularSeries confSeries = new RegularSeries(range, confArray, confStatusArray);
		RegularSeries breakSeries = new RegularSeries(range, preBreakArray, preBreakStatusArray);
		breakSeries.setObserved(defaultObservedValue);

		return (
				getFAMEDb().createUpdateTimeSeries(fameName, valueSeries)
			&& getFAMEDb().createUpdateTimeSeries(fameName + KEY_SEPARATOR + "OBS_STATUS", statusSeries)
			&& getFAMEDb().deleteTimeSeriesObservations(fameName + KEY_SEPARATOR + "OBS_CONF", range)
			&& getFAMEDb().createUpdateTimeSeries(fameName + KEY_SEPARATOR + "OBS_CONF", confSeries)
			&& getFAMEDb().deleteTimeSeriesObservations(fameName + KEY_SEPARATOR + "BREAK", range)
			&& getFAMEDb().createUpdateTimeSeries(fameName + KEY_SEPARATOR + "BREAK", breakSeries)
					);

	}

	/* ========================= Private methods ========================= */

	private boolean updateObservedValues(Attribute attribute, KeyFamilyComponent component) {
		if (!attribute.getName().equals("COLLECTION")) return true;

		getComponentHelper().logger.log(
				Level.FINEST,
				getComponentHelper().logName,
				"updateObservedValues",
				"COLLECTION attribute: Updating observed values"
		);

		if (component.isAttachedToDataSet()) {
			getComponentHelper().logger.log(
					Level.WARNING,
					getComponentHelper().logName,
					"updateObservedValues",
					"COLLECTION attribute attached to key family: Observed values for series not updated"
			);
			return true;
		}

		if (component.isAttachedToObservation()) {
			getComponentHelper().logger.log(
					Level.WARNING,
					getComponentHelper().logName,
					"updateObservedValues",
					"COLLECTION attribute attached to observation: Observed values for series not updated"
			);
			return true;
		}

		TiqObserved observed = convertCollectionToObserved(attribute.getValue());
		TimeSeries series = new TimeSeries(
											attribute.getKey(),
											"",
											"",
											new Int.ecb.gesmescb.Observation[0]
										);

		if (component.isAttachedToTimeSeries()) {
			if (!getFAMEDb().setObserved( getFameName( series ) , observed)) return false;
			if (!getFAMEDb().setObserved( getFameName( series ) + KEY_SEPARATOR + "BREAK", observed)) return false;
			return true;
		}

		if (component.isAttachedToSiblingGroup()) {
			Iterator it = getSiblingSet( attribute.getKey() ).iterator();

			while (it.hasNext()) {
				TimeSeries siblingSeries = (TimeSeries)it.next();

				if (!getFAMEDb().setObserved(getFameName( siblingSeries ), observed)) return false;
				if (!getFAMEDb().setObserved(getFameName( siblingSeries ) + KEY_SEPARATOR + "BREAK" , observed)) return false;
			}

			return true;
		}

		throw new IllegalArgumentException("Invalid attachment level");
	}

	private TiqObserved convertCollectionToObserved(String collection) {
		if (collection.equals("A")) return TiqObserved.AVERAGED;
		if (collection.equals("B")) return TiqObserved.BEGINNING;
		if (collection.equals("E")) return TiqObserved.END;
		if (collection.equals("H")) return TiqObserved.HIGH;
		if (collection.equals("L")) return TiqObserved.LOW;
		if (collection.equals("S")) return TiqObserved.SUMMED;
		if (collection.equals("Y")) return TiqObserved.AVERAGED; // Should be ANNUALIZED
		if (collection.equals("M")) return TiqObserved.AVERAGED; // Should be FORMULA
		if (collection.equals("U")) return TiqObserved.AVERAGED; // Should be FORMULA
		if (collection.equals("V")) return TiqObserved.AVERAGED; // Should be FORMULA

		getComponentHelper().logger.log(
				Level.WARNING,
				getComponentHelper().logName,
				"convertCollectionToObserved",
				"Unknown collection attribute value " + collection + ". Using default value"
		);

		return defaultObservedValue;
	}

}
