// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.fame;

import java.io.PrintWriter;
import java.io.FileWriter;
import java.io.IOException;
import java.util.Iterator;

import Int.ecb.logging.Level;

import Int.ecb.gesmescb.*;
import Int.ecb.gesmescb.event.*;

import Int.ecb.gesmescb.util.Named;
import Int.ecb.gesmescb.util.ComponentHelper;

/* **************************** Class ***************************** */

/**
 * FAME 4GL writer for metadata.
 *
 * @author
 * <A HREF="mailto:Sami.Airo@ECB.int">Sami Airo</A>
 *
 * @version @(#)$Id: FAME4GLWriter.java,v 1.1 2017/01/19 16:14:03 trainot Exp $
 */

public class FAME4GLWriter
	extends GESMESCBEventListenerAdapter
	implements Named, StructureListener, DataListener, GESMESCBWriter
{
	/* ======================= Constants ======================== */

	private final int FIRST_CASE_INDEX = 1;

	/* ========================= Fields ========================= */

	private ComponentHelper componentHelper;	
	private PrintWriter writer;
	private FileWriter fileWriter;
	private String filename;
	private int statisticalConceptCount;
	private String maintenanceAgencyId;

	/* ====================== Constructors ====================== */

	public FAME4GLWriter () 
		throws IOException
	{
		this( "" );
	}

	public FAME4GLWriter ( final String name ) 
		throws IOException
	{
		componentHelper = new ComponentHelper( name, this );
		open();
	}

	/* ===================== Public Methods ===================== */

	public String getName () {
		return componentHelper.name;
	}

	public void close () {
		componentHelper.logger.log(
			Level.FINE,
			componentHelper.logName,
			"close",
			"Closing."
		);
		
		// Remove the sources.
		setActive( false );

		try {
			// Flush the buffer.
			writer.close();

			if ( filename != null && fileWriter != null ) {
				fileWriter.close();
				componentHelper.logger.log(
					Level.INFO,
					componentHelper.logName,
					"close()",
					"Closed FAME 4GL output"
					+ (
						filename != null
						? " to " + filename
						: ""
						)
					+ '.'
				);
			}
		}
		catch ( IOException ioe ) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"close()",
				"Output problem.",
				ioe
			);
		}
	}


	public String getFAMEString(final String string) {
		StringBuffer buffer = new StringBuffer();
		for ( int i = 0; i < string.length(); i++ ) {
			final char c = string.charAt( i );
			switch ( c ) {
				case '"':
					buffer.append("\"");
				default:
					buffer.append( c );
			}
		}
		return "\"" + buffer.toString() + "\"";
	}


	public void handle ( InterchangeBegin interchangeBegin ) {
		componentHelper.logger.log(
			Level.FINE,
			componentHelper.logName,
			"handle",
			"Interchange begins"
		);		
	}

	public void handle ( InterchangeEnd interchangeEnd ) {
		componentHelper.logger.log(
			Level.FINE,
			componentHelper.logName,
			"handle",
			"Interchange ends"
		);
	}

	public void handle ( MessageBegin messageBegin ) {
		componentHelper.logger.log(
			Level.FINE,
			componentHelper.logName,
			"handle",
			"Message begins"
		);

		maintenanceAgencyId = componentHelper.getPreference("maintenanceAgencyId", messageBegin.getMaintenanceAgencyId());
	
		if (maintenanceAgencyId.equals("")) {
			maintenanceAgencyId = messageBegin.getMaintenanceAgencyId();
		}

		componentHelper.logger.log(
			Level.CONFIG,
			componentHelper.logName,
			"handle",
			"MaintenanceAgencyId set to " + maintenanceAgencyId
		);

		statisticalConceptCount = FIRST_CASE_INDEX;
	}

	public void handle ( MessageEnd messageEnd ) {
		componentHelper.logger.log(
			Level.FINE,
			componentHelper.logName,
			"handle",
			"Message ends"
		);
	}

	public void handle ( StatisticalConcept statisticalConcept ) {
		componentHelper.logger.log(
			Level.FINE,
			componentHelper.logName,
			"handle",
			"Handling statistical concept"
		);

		String caseSeriesName = maintenanceAgencyId + ".CL_CONCEPT";
		
		if (statisticalConceptCount == FIRST_CASE_INDEX) {
			writer.println("SERIES <STRINGLENGTH AUTO> " + caseSeriesName + ".VALUE: STRING INDEXED BY CASE");
			writer.println("SERIES <STRINGLENGTH AUTO> " + caseSeriesName + ".DESCR: STRING INDEXED BY CASE");
		}		
	
		writer.println("SET " + caseSeriesName + ".VALUE[" + statisticalConceptCount + "]=" + getFAMEString(statisticalConcept.getId())
						+ "; SET " + caseSeriesName + ".DESCR[" + statisticalConceptCount + "]=" + getFAMEString(statisticalConcept.getName()));		
	
		statisticalConceptCount++;
	}

	public void handle ( CodeList codeList ) {
		componentHelper.logger.log(
			Level.FINE,
			componentHelper.logName,
			"handle",
			"Handling code list"
		);
		
		String codelistName = maintenanceAgencyId + "." + codeList.getId();
		
		writer.println("SERIES <STRINGLENGTH AUTO> " + codelistName + ".VALUE: STRING INDEXED BY CASE");
		writer.println("SERIES <STRINGLENGTH AUTO> " + codelistName + ".DESCR: STRING INDEXED BY CASE");
		
		Iterator vit = codeList.codeValueIterator();
		Iterator dit = codeList.codeDescriptionIterator();

		int i = 1;		
		while (vit.hasNext()) {
			String value = (String) vit.next();
			String desc = (String) dit.next();

			writer.println("SET " + codelistName + ".VALUE[" + i + "]=" + getFAMEString(value)
								+ "; SET " + codelistName + ".DESCR[" + i + "]=" + getFAMEString(desc));
			i++;
		}
	}

	public void handle ( KeyFamily keyFamily ) {
		componentHelper.logger.log(
			Level.FINE,
			componentHelper.logName,
			"handle",
			"Handling key family"
		);
	}	

	public void handle ( TimeSeries timeSeries ) {
		componentHelper.logger.log(
			Level.FINE,
			componentHelper.logName,
			"handle",
			"Handling time series"
		);
	}	

	public void handle ( Attribute attribute ) {
		componentHelper.logger.log(
			Level.FINE,
			componentHelper.logName,
			"handle",
			"Handling attribute"
		);
	}		

	/* =================== Protected Methods ==================== */

	protected void open()
		throws IOException
	{
		filename = componentHelper.getPreference("destination", "");
		if (filename.equals("")) {
			final String desc = "No preference variable \"destination\" found to determine the output file.";
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"getColumnDefinition",
				desc
			);	
			throw new NullPointerException( desc );
		}
		
		componentHelper.logger.log(
			Level.CONFIG,
			componentHelper.logName,
			"open",
			"Writing to file " + filename
		);
		
		fileWriter = new FileWriter(filename);
		writer = new PrintWriter(fileWriter);
	}
}
