// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.fame;

import java.util.Properties;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.Arrays;
import java.util.HashSet;

import com.fame.timeiq.*;
import com.fame.timeiq.data.*;
import com.fame.timeiq.dates.*;
import com.fame.timeiq.persistence.*;

import Int.ecb.logging.Level;

import Int.ecb.gesmescb.util.ComponentHelper;

/**
 * Provides a facade to access FAME.
 *
 * @author
 * <A HREF="mailto:Sami.Airo@ECB.int">Sami Airo</A>
 *
 * @version @(#)$Id: FAMEDb.java,v 1.1 2017/01/19 16:14:03 trainot Exp $
 */
public class FAMEDb {

	/* ===================== Nested Classes ====================== */

	private static class EnumerationToIterator implements Iterator {

		private final Enumeration enumeration;

		public EnumerationToIterator ( final Enumeration enumeration ) {
			this.enumeration = enumeration;
		}

		public static Iterator iterator ( final Enumeration enumeration ) {
			return new EnumerationToIterator( enumeration );
		}

		public final boolean hasNext() {
			return enumeration.hasMoreElements();
		}

		public final Object next() {
			return enumeration.nextElement();
		}

		public void remove() {
			throw new UnsupportedOperationException("remove not supported");
		}
	}


	/* ======================= Constants ======================== */

	/** Character encoding used in Fame db. */
	private final String CHARACTER_ENCODING = "ISO8859_1";
	
	/** Default buffer size for matchwildcard */
	private final int BUFFER_SIZE = 25000;

	/* ========================= Fields ========================= */

	/** TimeIQ session. */
	private Session session;

	/** TimeIQ connection to databases. */
	private Connection connection;

	/** TimeIQ Current database for storing data. */
	private DataStore dataStore;

	/** True if this database has been opened */
	private boolean isOpen = false;

	/* Takes care of logging and prefs */
	private ComponentHelper componentHelper;
	
	/** Lists the series prefixes that indicates the series
	 * has to be stored as business.
	 * GESMES makes no difference between Business and Daily,
	 * if we want to store things as Business, we need to indicate 
	 * it somehow. This is a way to do it. 
	 */
	private HashSet businessSeriesPrefixes; 
	
	/* ====================== Constructors ====================== */

	public FAMEDb( final String name ) {
		
		componentHelper = new ComponentHelper( name, this );
		
		businessSeriesPrefixes = new HashSet();	
		for ( int i=1 ;; ++i ) {
			String pref	= componentHelper.getPreference(
				"businessSeriesPrefix." + i, null
			);
			if ( null == pref ) break;
			businessSeriesPrefixes.add( pref );
			componentHelper.logger.log(
				Level.CONFIG,
				componentHelper.logName,
				"FAMEDb",
				"Series with name starting with \""
				+ pref + "\" will be stored as business."
			);
		}
		if ( businessSeriesPrefixes.isEmpty() )
			componentHelper.logger.log(
				Level.CONFIG,
				componentHelper.logName,
				"FAMEDb",
				"No Series to be stored with business frequency."
			);
				
		session = Server.getInstance().getSession();

		Properties props = new Properties();
		props.put("technique", TiqTechnique.CONFORM);
		props.put("defaultObserved", "end");
	}

	/* ------------------------ Finalize ------------------------ */

	protected void finalize() throws Throwable {
		final String sourceMethod = "finalize";

		try {
			if (connection != null && connection.hasTransaction()) {
				try {
					componentHelper.logger.log(
						Level.SEVERE,
						componentHelper.logName,
						sourceMethod,
						"Rolling back open transaction."
					);
					connection.rollback();
				} catch (TransactionChkException e) {
					componentHelper.logger.log(
						Level.SEVERE,
						componentHelper.logName,
						sourceMethod,
						"Errors occured in rollback",
						e
					);
				} catch (ConnectionFailedChkException e) {
					componentHelper.logger.log(
						Level.SEVERE,
						componentHelper.logName,
						sourceMethod,
						"Connection to Fame database lost",
						e
					);
				} catch (ObjectAccessChkException e) {
					componentHelper.logger.log(
						Level.SEVERE,
						componentHelper.logName,
						sourceMethod,
						"Unexpected I/O errors in rollback",
						e
					);
				}
			}

			if (dataStore != null) {
				try {
					dataStore.close();
				} catch (Throwable e) {
					componentHelper.logger.log(
						Level.SEVERE,
						componentHelper.logName,
						sourceMethod,
						"Error in closing datastore",
						e
					);
				}
				dataStore = null;
			}

			if (connection != null) {
				try {
					connection.close();
				} catch (Throwable e) {
					componentHelper.logger.log(
						Level.SEVERE,
						componentHelper.logName,
						sourceMethod,
						"Error in closing connection",
						e
					);
				}
				connection = null;
			}

			session = null;
		} finally {
			super.finalize();
		}

	}


	/* ===================== Public Methods ===================== */

	public boolean open(String dataBaseName, Properties dataBaseProperties) {
      return open("", dataBaseName, dataBaseProperties);
   }

	public boolean open(String serverName, String dataBaseName, Properties dataBaseProperties) {
		final String sourceMethod = "open";

		if (isOpen) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				sourceMethod,
				"Already open"
			);
			return false;
		}

      if (serverName.equals("") && dataBaseName.equals("")) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				sourceMethod,
				"Local database name not provided"
			);
         return false;
      }

		Properties connectionProperties = new Properties();
		connectionProperties.put("characterEncoding", CHARACTER_ENCODING);

		componentHelper.logger.log(
			Level.FINE,
			componentHelper.logName,
			sourceMethod,
			"Opening connection to server " + serverName
		);

		try {
			// Open connection to database server
			connection = session.createConnection(serverName, connectionProperties);
		} catch (TiqCheckedException e) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				sourceMethod,
				"Error in opening connection to server.",
				e
			);
			return false;
		} catch (TiqUncheckedException e) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				sourceMethod,
				"Error in opening connection to server.",
				e
			);
			return false;
		}

      if (dataBaseName.equals("")) {
         isOpen = true;
         return true;      
      }

		componentHelper.logger.log(
			Level.FINER,
			componentHelper.logName,
			sourceMethod,
			"Getting datastore: " + dataBaseName,
         dataBaseProperties
		);

		try {
			dataStore = connection.getDataStore(dataBaseName, dataBaseProperties);
		} catch (TiqCheckedException e) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				sourceMethod,
				"Error in opening database " + dataBaseName + ".",
				e
			);
			return false;
		} catch (TiqUncheckedException e) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				sourceMethod,
				"Error in opening database " + dataBaseName + ".",
				e
			);
			return false;
		}

		componentHelper.logger.log(
			Level.FINER,
			componentHelper.logName,
			sourceMethod,
			"Beginning transaction"
		);

		try {
			connection.begin();
		} catch (TiqCheckedException e) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				sourceMethod,
				"Error in beginning transaction.",
				e
			);
			return false;
		} catch (TiqUncheckedException e) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				sourceMethod,
				"Error in beginning transaction.",
				e
			);
			return false;
		}

		isOpen = true;

		return true;
	}


	/**
	 * Close database.
	 */
	public boolean close(boolean commitTransaction) {
		final String sourceMethod = "closeDatabase";
		boolean isOK = true;

		if (!isOpen) return true;

		componentHelper.logger.log(
			Level.FINE,
			componentHelper.logName,
			sourceMethod,
			"Closing connection.",
			dataStore
		);

		if (connection.hasTransaction()) {
			try {
				if (commitTransaction) {
					componentHelper.logger.log(
						Level.FINER,
						componentHelper.logName,
						sourceMethod,
						"Commiting the current transaction."
					);
					connection.commit();
				} else {
					componentHelper.logger.log(
						Level.FINER,
						componentHelper.logName,
						sourceMethod,
						"Rolling back the current transaction."
					);
					connection.rollback();
				}
			} catch (TiqCheckedException e) {
				componentHelper.logger.log(
					Level.SEVERE,
					componentHelper.logName,
					sourceMethod,
					"Error in transaction " + (commitTransaction ? "commit" : "rollback") + ".",
					e
				);
			} catch (TiqUncheckedException e) {
				componentHelper.logger.log(
					Level.SEVERE,
					componentHelper.logName,
					sourceMethod,
					"Error in transaction " + (commitTransaction ? "commit" : "rollback") + ".",
					e
				);
			}
		}

		componentHelper.logger.log(
			Level.FINE,
			componentHelper.logName,
			sourceMethod,
			"Closing database",
			dataStore
		);


      if (dataStore != null) {
   		try {
	   		dataStore.close();
		   } catch (TiqCheckedException e) {
			   componentHelper.logger.log(
   				Level.SEVERE,
	   			componentHelper.logName,
		   		sourceMethod,
			   	"Error in closing database.",
   				e
	   		);
		   } catch (TiqUncheckedException e) {
   			componentHelper.logger.log(
	   			Level.SEVERE,
		   		componentHelper.logName,
			   	sourceMethod,
   				"Error in closing database.",
	   			e
		   	);
   		} finally {
	   		dataStore = null;
		   }
      }

		// Close connection
		connection.close();
		connection = null;

		if (isOK) {
			isOpen = false;
		}

		return isOK;
	}


	public boolean deleteObject(String name) {
		final String sourceMethod = "deleteObject";

		if (!isOpen) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				sourceMethod,
				"Database not open"
			);
			return false;
		}

		componentHelper.logger.log(
			Level.FINEST,
			componentHelper.logName,
			sourceMethod,
			"Deleting Fame object: " + name
		);

		try {
			dataStore.deleteObject(name);
		} catch (NotFoundChkException e) {
			// this is normally OK. Object was not in the database
		} catch (TiqCheckedException e) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				sourceMethod,
				"Error in deleting fame object " + name + ".",
				e
			);
			return false;
		} catch (TiqUncheckedException e) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				sourceMethod,
				"Error in deleting fame object " + name + ".",
				e
			);
			return false;
		}

		return true;
	}


	public boolean deleteTimeSeriesObservations(String name, IndexValueRange range) {
		final String sourceMethod = "deleteTimeSeriesObservations";

		if (!isOpen) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				sourceMethod,
				"Database not open"
			);
			return false;
		}

		componentHelper.logger.log(
			Level.FINEST,
			componentHelper.logName,
			sourceMethod,
			"Deleting observations from Fame series: " + name,
			range
		);

		try {
			dataStore.deleteObservations(name, range);
		} catch (NotFoundChkException e) {
			// OK series was not in DB
		} catch (DateAlignmentChkException e) {
			TiqObject series = getObject( name );
			boolean isBusiness
			= series.isSeries() && shouldBeBusiness( name );
			componentHelper.logger.log(
				isBusiness ? Level.WARNING : Level.SEVERE,
				componentHelper.logName,
				sourceMethod,
				( isBusiness ? "Issue" : "Error" )
				+ " in deleting observations from "
				+ name,
				e
			);
			return isBusiness;
		} catch (TiqCheckedException e) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				sourceMethod,
				"TiqChecked Error in deleting observations from " + name,
				e
			);
			return false;
		} catch (TiqUncheckedException e) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				sourceMethod,
				"TiqUnchecked Error in deleting observations from " + name,
				e
			);
			return false;
		}
		return true;
	}


	public boolean overwriteObject(String name, TiqObject obj) {
		final String sourceMethod = "overwriteObject";

		if (!isOpen) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				sourceMethod,
				"Database not open"
			);
			return false;
		}

		componentHelper.logger.log(
			Level.FINEST,
			componentHelper.logName,
			sourceMethod,
			"Overwriting Fame object: " + name,
			obj
		);

		try {
			dataStore.overwriteObject(
				name,
				mdpObjectAsBusiness(name, obj)
			);
		} catch (TiqCheckedException e) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				sourceMethod,
				"Error in overwriting Fame object (TiqChecked!)" + name,
				e
			);
			return false;
		} catch (TiqUncheckedException e) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				sourceMethod,
				"Error in overwriting Fame object (TiqUnchecked!) " + name,
				e
			);
			return false;
		}
		return true;
	}


	public boolean createObject(String name, TiqObject obj) {
		final String sourceMethod = "createObject";

		if (!isOpen) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				sourceMethod,
				"Database not open"
			);
			return false;
		}

		componentHelper.logger.log(
			Level.FINEST,
			componentHelper.logName,
			sourceMethod,
			"Creating Fame object: " + name,
			obj
		);

		TiqObject storeObject;
		
		if (obj.getTiqClass().equals(TiqClass.REGULAR_SERIES)) {
			storeObject = mdpObjectAsBusiness( name, (TiqObject) obj );
			setDefaultObserved((RegularSeries) storeObject );
		} else
			storeObject = obj;

		try {
			dataStore.createObject(
				name,
				storeObject	
			);
		} catch (TiqCheckedException e) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				sourceMethod,
				"Error in creating Fame object " + name,
				e
			);
			return false;
		} catch (TiqUncheckedException e) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				sourceMethod,
				"Error in creating Fame object " + name,
				e
			);
			return false;
		}

		return true;
	} 


	/**
	 * Writes Fame object to database. Creates Fame object if it does not
	 * already exist, updates values, otherwise.
	 *
	 * @param dataStore Fame database.
	 * @param name Fame name.
	 * @param fameObject Fame series.
	 *
	 * @return Fame name
	 */
	public boolean createUpdateTimeSeries(String name, RegularSeries series) {
		final String sourceMethod = "createUpdateTimeSeries";

		if (!isOpen) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				sourceMethod,
				"Database not open"
			);
			return false;
		}

		componentHelper.logger.log(
			Level.FINEST,
			componentHelper.logName,
			sourceMethod,
			"Creating Fame object: " + name,
			series
		);

		RegularSeries storeSeries
		= (RegularSeries) mdpObjectAsBusiness( name, (TiqObject) series);
		setDefaultObserved(storeSeries);

		try {
			dataStore.createObject( name, storeSeries );
		} catch (ObjectExistsChkException e) {
			componentHelper.logger.log(
				Level.FINEST,
				componentHelper.logName,
				sourceMethod,
				"Fame object exists. Updating Fame object: " + name,
				storeSeries
			);

			try {
				// Fame object already exist, update observation values
				dataStore.setObservations(name, storeSeries.getObservations());
			} catch (TiqCheckedException e2) {
				componentHelper.logger.log(
					Level.SEVERE,
					componentHelper.logName,
					sourceMethod,
					"Error in Fame object update: " + name,
					e
				);
				return false;
			} catch (TiqUncheckedException e2) {
				componentHelper.logger.log(
					Level.SEVERE,
					componentHelper.logName,
					sourceMethod,
					"Error in Fame object update: " + name,
					e
				);
				return false;
			}
		} catch (TiqCheckedException e) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				sourceMethod,
				"Error in creating Fame object " + name,
				e
			);
			return false;
		} catch (TiqUncheckedException e) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				sourceMethod,
				"Error in creating Fame object " + name,
				e
			);
			return false;
		}

		return true;
	}

	/**
	 * Writes Fame object to database. Creates Fame object if it does not
	 * already exist, updates values, otherwise.
	 *
	 * @param dataStore Fame database.
	 * @param name Fame name.
	 * @param fameObject Fame series.
	 *
	 * @return Fame name
	 */
	public boolean updateCreateTimeSeries(String name, RegularSeries series) {
		final String sourceMethod = "createUpdateTimeSeries";

		if (!isOpen) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				sourceMethod,
				"Database not open"
			);
			return false;
		}

		componentHelper.logger.log(
			Level.FINEST,
			componentHelper.logName,
			sourceMethod,
			"Updating Fame object: " + name,
			series
		);

		RegularSeries storeSeries
		= (RegularSeries) mdpObjectAsBusiness( name, (TiqObject) series);
		setDefaultObserved(storeSeries);

		try {
			dataStore.setObservations(name, storeSeries.getObservations());
		} catch (NotFoundChkException e) {
			componentHelper.logger.log(
				Level.FINEST,
				componentHelper.logName,
				sourceMethod,
				"Fame object does not exist. Creating Fame object: " + name,
				storeSeries
			);

			setDefaultObserved(storeSeries);

			try {
				dataStore.createObject(name, storeSeries);
			} catch (TiqCheckedException e2) {
				componentHelper.logger.log(
					Level.SEVERE,
					componentHelper.logName,
					sourceMethod,
					"Error in creating Fame object " + name,
					e
				);
				return false;
			} catch (TiqUncheckedException e2) {
				componentHelper.logger.log(
					Level.SEVERE,
					componentHelper.logName,
					sourceMethod,
					"Error in creating Fame object " + name,
					e
				);
				return false;
			}
		} catch (TiqCheckedException e) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				sourceMethod,
				"Error in Fame object update:" + name,
				e
			);
			return false;
		} catch (TiqUncheckedException e) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				sourceMethod,
				"Error in Fame object update:" + name,
				e
			);
			return false;
		}

		return true;
	}


	public TiqObject getObject(String name) {
		return getObject(
			name,
			TiqConstants.RANGE_OPEN_END,
			TiqConstants.RANGE_OPEN_END
		);
	}


	public TiqObject getObject(
		final String name,
		final long startIndex,
		final long endIndex
	) {
		return getObject(
			name,
			startIndex,
			endIndex,
			true // automatically promotes business series to daily.
		);
	}


	public TiqObject getObject(
		final String name,
		final long startIndex,
		final long endIndex,
		boolean seeBusinessAsDaily /* false for tests to check storage in Fame*/
	) {
		final String sourceMethod = "getObject";

		if (!isOpen) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				sourceMethod,
				"Database not open"
			);
			return null;
		}

		componentHelper.logger.log(
			Level.FINEST,
			componentHelper.logName,
			sourceMethod,
			"Getting Fame object: " + name
		);

		try {
			return seeBusinessAsDaily
				?convertBusinessToDaily(
					dataStore.getTiqObjectCopy(name, startIndex, endIndex)
					)
				:dataStore.getTiqObjectCopy(name, startIndex, endIndex);
					
		} catch (NotFoundChkException e) {
			// this is OK. Object was not in the database
			return null;
		} catch (TiqCheckedException e) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				sourceMethod,
				"Error in getting Fame object: " + name,
				e
			);
			return null;
		} catch (TiqUncheckedException e) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				sourceMethod,
				"Error in getting Fame object: " + name,
				e
			);
			return null;
		}
	}


	public Iterator getDataObjectInfoIterator(String wildCard) {
		return getDataObjectInfoIterator(
			wildCard,
			new ScreeningCriteria(),
			false
		);
	}


	public Iterator getDataObjectInfoIterator(
		String wildCard,
		boolean getAllInfo
	) {
		return getDataObjectInfoIterator(
			wildCard,
			new ScreeningCriteria(),
			getAllInfo
		);
	}


	public Iterator getDataObjectInfoIterator(
		String wildCard,
		ScreeningCriteria screen,
		boolean getAllInfo
	) {
		final String sourceMethod = "getDataObjectInfoIterator";
		
		screen.setAllowAliases(false);

		try {
			return EnumerationToIterator.iterator(
				dataStore.matchWildCard(
					wildCard,
					screen,
					BUFFER_SIZE,
					"",
					getAllInfo
				)
			);
		} catch (TiqCheckedException e) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				sourceMethod,
				"Error in matching wildcard: " + wildCard,
				e
			);
			return null;
		} catch (TiqUncheckedException e) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				sourceMethod,
				"Error in matching wildcard: " + wildCard,
				e
			);
			return null;
		}
	}


	/**
	 * Updates the observed attribute value in a Fame database
	 * already exist, updates values, otherwise.
	 *
	 * @param dataStore Fame database.
	 * @param name Fame name.
	 * @param fameObject Fame series.
	 *
	 * @return Fame name
	 */
	public boolean setObserved(String name, TiqObserved observed) {
		final String sourceMethod = "setObserved";

		if (!isOpen) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				sourceMethod,
				"Not open"
			);
			return false;
		}

		RegularSeries series = null;

		try {
			componentHelper.logger.log(
				Level.FINEST,
				componentHelper.logName,
				sourceMethod,
				"Getting tracked Fame object: " + name
			);
			series = (RegularSeries)dataStore.getTrackedObject(name);
		} catch (NotFoundChkException e) {
			componentHelper.logger.log(
				Level.FINEST,
				componentHelper.logName,
				sourceMethod,
				"Observed not updated. Object not in database: " + name
			);
			// this is OK. Object was not in the database
			return true;
		} catch (TiqCheckedException e) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				sourceMethod,
				"Error in getting Fame object: " + name,
				e
			);
			return false;
		} catch (TiqUncheckedException e) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				sourceMethod,
				"Error in getting Fame object: " + name,
				e
			);
			return false;
		}

		if (series.getObserved().equals(observed)) {
			componentHelper.logger.log(
				Level.FINEST,
				componentHelper.logName,
				sourceMethod,
				name + " has correct observed value. Not updated"
			);
			return true;
		}

		componentHelper.logger.log(
			Level.FINEST,
			componentHelper.logName,
			sourceMethod,
			"Setting observed value of " + name + " to " + observed
		);
		series.setObserved(observed);

		try {
			componentHelper.logger.log(
				Level.FINEST,
				componentHelper.logName,
				sourceMethod,
				"Updating tracked Fame object: " + name
			);
			//series.update();
			dataStore.overwriteObject(series.getName(), series);
		} catch (TiqCheckedException e) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				sourceMethod,
				"Error in updating tracked Fame object: " + name,
				e
			);
			return false;
		} catch (TiqUncheckedException e) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				sourceMethod,
				"Error in updating tracked Fame object: " + name,
				e
			);
			return false;
		}

		return true;
	}

	public String toString() {
		return (dataStore == null ? "" : dataStore.toString());
	}

	public boolean isOpen() {
		return isOpen;
	}

	public TiqObject remEval(String expression) {
		if (!isOpen) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"remEval",
				"Database not open"
			);
			return null;
		}

		componentHelper.logger.log(
			Level.FINEST,
			componentHelper.logName,
			"remEval",
			"Evaluating expression on MCADBS: " + expression
		);

		try {
			return connection.remEval(expression);
		} catch (Exception e) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"remEval",
				"Error in evaluating expression on MCADBS.",
				e
			);
			return null;
		}
	}

	public DataObjectInfo getDataObjectInfo( final String name ) {
		final String sourceMethod = "getDataObjectInfo";

		if (!isOpen) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				sourceMethod,
				"Database not open"
			);
			return null;
		}

		componentHelper.logger.log(
			Level.FINEST,
			componentHelper.logName,
			sourceMethod,
			"Getting Fame object info: " + name
		);

		try {
			return dataStore.getObjectInfo(name);
		} catch (NotFoundChkException e) {
			// this is OK. Object was not in the database
			return null;
		} catch (TiqCheckedException e) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				sourceMethod,
				"Error in getting Fame object: " + name,
				e
			);
			return null;
		} catch (TiqUncheckedException e) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				sourceMethod,
				"Error in getting Fame object: " + name,
				e
			);
			return null;
		}
	}

	protected void setDefaultObserved( final RegularSeries series ) {
		if (series.getTiqType().isNumeric() && series.getObserved().equals(TiqObserved.UNDEFINED)) {
			series.setObserved(TiqObserved.SUMMED);
		}
	}

	public TiqObject convertBusinessToDaily( final TiqObject tiqObject ) {
		if (
			tiqObject.isSeries()
			&&
			tiqObject.getCalendar().equals( RegularCalendar.BUSINESS )
		) {
			// Because TimeIQ 1.3 cannot treat OBS_STATUS just using
			// TiqObject.getTiqObjectCopy( TimeScaling ),
			// the following explicit conversion is necessary.
			final ObservationList obsList = tiqObject.getObservations();
			final Series result = new RegularSeries(
				RegularCalendar.DAILY,
				tiqObject.getTiqType(),
				tiqObject.getObserved(),
				tiqObject.getBasis()
			);
			try {
				result.setObservations( obsList );
			} catch ( final TiqCheckedException e ) {
				// Should not be possible
				componentHelper.logger.log(
					Level.SEVERE,
					componentHelper.logName,
					"convertBusinessToDaily",
					"Internal error 846637: " + e,
					e
				);
			}
			return result;
		}
		return tiqObject;
	}

	public TiqObject convertDailyToBusiness( final TiqObject tiqObject ) {
		if (
			tiqObject.isSeries()
			&&
			tiqObject.getCalendar().equals( RegularCalendar.DAILY )
		) {
			// Because TimeIQ 1.3 cannot treat OBS_STATUS just using
			// TiqObject.getTiqObjectCopy( TimeScaling ),
			// the following explicit conversion is necessary.
			final Series result = new RegularSeries(
				RegularCalendar.BUSINESS,
				tiqObject.getTiqType(),
				tiqObject.getObserved(),
				tiqObject.getBasis()
			);
			final ObservationList obsList = tiqObject.getObservations();
			if (! obsList.isEmpty()) { //hasDate() doesn't consider NC as data!
				final long[] indexes = obsList.getIndexes();
				try {
					for (
						final PrimitiveIterator iterator
						= RegularCalendar.BUSINESS.indexIterator(
							tiqObject.getMinIndex()
						);
						iterator.hasNext();
					) {
						final long index = iterator.nextLong();
						final int i = Arrays.binarySearch( indexes, index );
						if ( i < 0 ) break;
						final com.fame.timeiq.data.Observation obs
						= (com.fame.timeiq.data.Observation) obsList.get( i );
						result.setObservation(
							index, obs.getTiqValue()
						);
					}
				} catch ( final TiqCheckedException e ) {
					// Should not be possible
					componentHelper.logger.log(
						Level.SEVERE,
						componentHelper.logName,
						"convertDailyToBusiness",
						"Internal error 846448: " + e,
						e
					);
				}
			}
			return result;
		}
		return tiqObject;
	}
	

	public boolean setObservation(
		final String name,
		final long index, 
		final double value,
		final byte status
	) throws Exception {
		if (!isOpen) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"setObservation",
				"Database not open"
			);
			return false;
		}

		componentHelper.logger.log(
			Level.FINEST,
			componentHelper.logName,
			"setObservation",
			"Setting observation " + index + " to " + value
		);

//		try {
			dataStore.setObservation(name, index, value, status);
/*		} catch (Exception e) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"setObservation",
				"Error in setting observation value",
				e
			);
			return false;
		}
*/		
		return true;
	}


	private TiqObject mdpObjectAsBusiness( String name, TiqObject tiqObject ) {
		return shouldBeBusiness( name )
			? convertDailyToBusiness( tiqObject )
			: tiqObject;
	}

	private boolean shouldBeBusiness( String name ) {
		if ( null == businessSeriesPrefixes || businessSeriesPrefixes.isEmpty() )
			return false;
		for ( Iterator iterator = businessSeriesPrefixes.iterator()
			; iterator.hasNext()
			;
		) {
			String prefix = (String) iterator.next();
			if ( prefix.equalsIgnoreCase(
				name.substring( 0, prefix.length() ) )
			) { 
				return true;
			}
		}
		return false;
	}
}

