// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.fame;

import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;

import java.io.File;
import java.util.Properties;
import java.util.Random;

import com.fame.timeiq.*;
import com.fame.timeiq.dates.*;
import com.fame.timeiq.data.*;
import com.fame.timeiq.functors.operators.NE;
import com.fame.timeiq.functors.util.Which;

/* ***************************** Class ****************************** */

/**
 * <P>
 * Tests the implementation of FAMEDb class in the Int.ecb.gesmescb.fame
 * package.
 * </P>
 * 
 * @author <A HREF="mailto:Sami.Airo@ECB.int">Sami Airo</A>
 * 
 * @version @(#)$Id: FAMEDbFAME10Test.java,v 1.1 2017/01/19 16:14:03 trainot Exp $
 */

public class FAMEDbFAME10Test extends TestCase {

    /* ========================= Fields ========================= */

    final String filename = "/tmp/FAMEDbTest.db";
    private static String formatDbVersion = "4";

    /* ====================== Constructors ====================== */

    /**
     * Constructs a FAMEDbTest with the specified name.
     * 
     * @param name
     *            Test case name
     */
    public FAMEDbFAME10Test(final String name) {
        super(name);
    }

    /* ================== Protected Methods ===================== */

    /**
     * Sets up the test fixture
     */
    protected void setUp() {
        System.setProperty("Int.ecb.logging.config.class",
                "Int.ecb.gesmescb.util.LoggingPropertiesReader");
        System.setProperty(".level", "ALL");
        System.setProperty("handlers", "Int.ecb.logging.ConsoleHandler");
        System.setProperty("Int.ecb.logging.ConsoleHandler.level", "INFO");
    }

    /**
     * Tears sown the test fixture
     */
    protected void tearDown() {
        try {
            File file = new File(filename);
            file.delete();
        } catch (Exception e) {
            fail(e.getMessage());
        }
    }

    /** List characteristics of a TiqView. */
    protected static String toString(final TiqView object) {
        final StringBuffer buffer = new StringBuffer();

        buffer.append(object.getClass().getName());
        buffer.append('(');

        try {
            buffer.append("class=");
            buffer.append(object.getTiqClass().toString());

            buffer.append(",type=");
            buffer.append(object.getTiqType().toString());

            if (!object.isSeries()) {
                buffer.append(",calendar=");
                buffer.append(object.getTiqValue().toString());
            }

            if (object.isSeries()) {
                buffer.append(",calendar=");
                buffer.append(object.getCalendar().toString());
                buffer.append(",calendar.hasGaps=");
                buffer.append(object.getCalendar().hasGaps());

                buffer.append(",basis=");
                buffer.append(object.getBasis().toString());
                buffer.append(",basis.hasGaps=");
                buffer.append(object.getBasis().hasGaps());

                buffer.append(",observed=");
                buffer.append(object.getObserved());

                buffer.append(",minIndex=");
                buffer.append(object.getMinIndex());
                buffer.append(",maxIndex=");
                buffer.append(object.getMaxIndex());
            }
        } catch (final FunctorConsistencyChkException e) {
            buffer.append(",exception=");
            buffer.append(e.toString());
        }

        buffer.append(",version=");
        buffer.append(object.getVersionNumber());

        buffer.append(')');

        return buffer.toString();
    }

    /* ===================== Public Methods ===================== */

    /**
     * Assembles and returns a test suite for all the test methods of this test
     * case.
     * 
     * @return a non-null test suite
     */
    public static Test suite() {
        TestSuite suite = new TestSuite(FAMEDbFAME10Test.class);
        return suite;
    }

    /**
     * Runs the test case
     */
    public static void main(String args[]) {
        String[] testCaseName = {FAMEDbFAME10Test.class.getName()};
        junit.textui.TestRunner.main(testCaseName);
    }

    /* --------------------- Test Methods --------------------- */

    public void testOpenClose() {
            Properties prop = new Properties();
            prop.put("mode", "CREATE");
            prop.put("dbVersion", formatDbVersion);
            FAMEDb db = new FAMEDb("test1");

            db.open(filename, prop);
            db.close(true);
    }

    public void testCreateDeleteScalar() {
            Properties prop = new Properties();
            prop.put("mode", "CREATE");
            prop.put("dbVersion", formatDbVersion);

            FAMEDb db = new FAMEDb("test2");
            db.open(filename, prop);

            Scalar a = new Scalar("test1");

            db.createObject("test1", a);
            Scalar b = (Scalar) db.getObject("test1");

            assertEquals("name", "TEST1", b.getName());
            assertEquals("value", a.getTiqValue().getStringValue(), b
                    .getTiqValue().getStringValue());
            assertEquals("status", a.getTiqValue().getStatus(), b.getTiqValue()
                    .getStatus());

            db.deleteObject("test1");

            b = (Scalar) db.getObject("test1");

            assertNull(b);

            db.close(true);

    }

    public void testCreateDeleteSeries() {
            Properties prop = new Properties();
            prop.put("mode", "CREATE");
            prop.put("dbVersion", formatDbVersion);

            FAMEDb db = new FAMEDb("test3");
            db.open(filename, prop);

            RegularSeries a = getRegularSeries(TiqFrequency.WEEKLY, 1991, 1, 1,
                    10);

            db.createObject("test1", a);
            RegularSeries b = (RegularSeries) db.getObject("test1");

            assertEquals("name", "TEST1", b.getName());

            int passed = -1;

            try {
                NE ne = new NE(a, b);
                Which which = new Which(ne.getTiqObjectCopy());
                passed = which.getPassed();
            } catch (TiqCheckedException e) {
                fail(e.getMessage());
            }

            assertEquals(0, passed);

            db.deleteObject("test1");

            b = (RegularSeries) db.getObject("test1");

            assertNull(b);

            db.close(true);
    }

    public void testOverWriteScalar() {
            Properties prop = new Properties();
            prop.put("mode", "CREATE");
            prop.put("dbVersion", formatDbVersion);

            FAMEDb db = new FAMEDb("test4");
            db.open(filename, prop);

            Scalar a = new Scalar("test1");

            db.createObject("test1", a);
            Scalar b = (Scalar) db.getObject("test1");

            assertNotNull(b);

            assertEquals("name", "TEST1", b.getName());
            assertEquals("value", a.getTiqValue().getStringValue(), b
                    .getTiqValue().getStringValue());
            assertEquals("status", a.getTiqValue().getStatus(), b.getTiqValue()
                    .getStatus());

            a = new Scalar(new Double(999));
            db.overwriteObject("test1", a);

            b = (Scalar) db.getObject("test1");

            assertEquals("name", "TEST1", b.getName());
            assertEquals("value", a.getTiqValue().getIntValue(), b
                    .getTiqValue().getIntValue());
            assertEquals("status", a.getTiqValue().getStatus(), b.getTiqValue()
                    .getStatus());

            db.close(true);
    }

    public void testDeleteTimeSeriesObservations() {
            Properties prop = new Properties();
            prop.put("mode", "CREATE");
            prop.put("dbVersion", formatDbVersion);

            FAMEDb db = new FAMEDb("test5");
            db.open(filename, prop);

            RegularSeries a = getRegularSeries(TiqFrequency.MONTHLY, 1991, 1,
                    1, 5);

            db.createUpdateTimeSeries("test1", a);

            RegularSeries b = (RegularSeries) db.getObject("test1");
            byte[] statuses = b.getObservations().getStatuses();

            assertEquals(statuses[0], TiqConstants.STATUS_OK);
            assertEquals(statuses[1], TiqConstants.STATUS_OK);
            assertEquals(statuses[2], TiqConstants.STATUS_OK);
            assertEquals(statuses[3], TiqConstants.STATUS_OK);
            assertEquals(statuses[4], TiqConstants.STATUS_OK);

            IndexValueRange r = getIndexValueRange(TiqFrequency.MONTHLY, 1991,
                    2, 1, 2);

            db.deleteTimeSeriesObservations("test1", r);

            b = (RegularSeries) db.getObject("test1");
            statuses = b.getObservations().getStatuses();

            assertEquals(statuses[0], TiqConstants.STATUS_OK);
            assertEquals(statuses[1], TiqConstants.STATUS_ND);
            assertEquals(statuses[2], TiqConstants.STATUS_ND);
            assertEquals(statuses[3], TiqConstants.STATUS_OK);
            assertEquals(statuses[4], TiqConstants.STATUS_OK);

            db.close(true);
    }

    public void testcreateUpdateTimeSeries() {
            Properties prop = new Properties();
            prop.put("mode", "CREATE");
            prop.put("dbVersion", formatDbVersion);

            FAMEDb db = new FAMEDb("test6");
            db.open(filename, prop);

            RegularSeries a = getRegularSeries(TiqFrequency.MONTHLY, 1991, 1,
                    1, 4);

            db.createObject("test1", a);
            db.createObject("test3", a);

            RegularSeries b = getRegularSeries(TiqFrequency.MONTHLY, 1991, 3,
                    1, 4);

            db.createObject("test2", b);

            db.createUpdateTimeSeries("test3", b);

            RegularSeries c = (RegularSeries) db.getObject("test3");

            db.close(true);
    }

    public void testRemEval() {
            TiqObject iref = null;
            FAMEDb db = new FAMEDb("test7");
            Properties prop = new Properties();
            prop.put("mode", "CREATE");
            prop.put("dbVersion", formatDbVersion);
            // #2961     - port service 
            // dissserv  - name service 
            db.open("#2961@stat-prod", "", prop);
            iref = db.remEval("$iref_test_normal()");

            // TODO: System.out.println("Some info about the iref: " +
            // iref.getTiqValue());

            assertEquals("000001", iref.getTiqValue().getStringValue());
            iref = db.remEval("$iref_test_error()");
            assertEquals(null, iref);
            db.close(true);
    }

    public void testConvertBusiness() throws TiqCheckedException {
        final long Mon_2001_01_01 = DateHelper.ymdToIndex(2001, 1, 1);
        final long Tue_2001_01_02 = DateHelper.ymdToIndex(2001, 1, 2);
        final long Wed_2001_01_03 = DateHelper.ymdToIndex(2001, 1, 3);
        final long Thr_2001_01_04 = DateHelper.ymdToIndex(2001, 1, 4);
        final long Fri_2001_01_05 = DateHelper.ymdToIndex(2001, 1, 5);
        final long Sat_2001_01_06 = DateHelper.ymdToIndex(2001, 1, 6);
        final long Sun_2001_01_07 = DateHelper.ymdToIndex(2001, 1, 7);
        final long Mon_2001_01_08 = DateHelper.ymdToIndex(2001, 1, 8);
        final byte OK = TiqConstants.STATUS_OK;
        final byte NC = TiqConstants.STATUS_NC;

        final FAMEDb db = new FAMEDb("test8");

        final Series dailyNumbers = new RegularSeries(RegularCalendar.DAILY,
                TiqType.FLOAT, TiqObserved.END, null);
        final Series businessNumbers = new RegularSeries(
                RegularCalendar.BUSINESS, TiqType.FLOAT, TiqObserved.END, null);
        // System.out.println( toString( businessNumbers ) );
        // System.out.println( toString( db.convertDailyToBusiness( dailyNumbers
        // ) ) );
        // System.out.println( toString( dailyNumbers ) );
        // System.out.println( toString( db.convertBusinessToDaily(
        // businessNumbers ) ) );
        assertEquals("Empty to daily", dailyNumbers, db
                .convertBusinessToDaily(businessNumbers));
        assertEquals("Empty to business", businessNumbers, db
                .convertDailyToBusiness(dailyNumbers));

        // Now testing with only NC in the series
        dailyNumbers.setObservation(Mon_2001_01_01, null, NC);
        businessNumbers.setObservation(Mon_2001_01_01, null, NC);

        assertEquals("NC to daily", dailyNumbers, db
                .convertBusinessToDaily(businessNumbers));
        assertEquals("NC to business", businessNumbers, db
                .convertDailyToBusiness(dailyNumbers));

        dailyNumbers.setObservation(Mon_2001_01_01, 13.6, OK);
        businessNumbers.setObservation(Mon_2001_01_01, 13.6, OK);
        assertEquals("One to daily", dailyNumbers, db
                .convertBusinessToDaily(businessNumbers));
        assertEquals("One to business", businessNumbers, db
                .convertDailyToBusiness(dailyNumbers));

        dailyNumbers.setObservation(Tue_2001_01_02, 14.8, OK);
        businessNumbers.setObservation(Tue_2001_01_02, 14.8, OK);
        dailyNumbers.setObservation(Wed_2001_01_03, 14.8, OK);
        businessNumbers.setObservation(Wed_2001_01_03, 14.8, OK);
        dailyNumbers.setObservation(Thr_2001_01_04, 14.8, OK);
        businessNumbers.setObservation(Thr_2001_01_04, 14.8, OK);
        dailyNumbers.setObservation(Fri_2001_01_05, 14.8, OK);
        businessNumbers.setObservation(Fri_2001_01_05, 14.8, OK);
        assertEquals("Week to daily", dailyNumbers, db
                .convertBusinessToDaily(businessNumbers));
        assertEquals("Week to business", businessNumbers, db
                .convertDailyToBusiness(dailyNumbers));

        dailyNumbers.setObservation(Mon_2001_01_08, 15.4, OK);
        businessNumbers.setObservation(Mon_2001_01_08, 15.4, OK);
        assertEquals("All to daily", dailyNumbers, db
                .convertBusinessToDaily(businessNumbers));
        assertEquals("All to business", businessNumbers, db
                .convertDailyToBusiness(dailyNumbers));
    }

    public void testMDPimplicitlyBusiness() {
            Properties prop = new Properties();
            prop.put("mode", "CREATE");
            System
                    .setProperty(
                            "Int.ecb.gesmescb.fame.FAMEDb.test9.businessSeriesPrefix.1",
                            "FM.B.");
            System
                    .setProperty(
                            "Int.ecb.gesmescb.fame.FAMEDb.test9.businessSeriesPrefix.2",
                            "FX.B.");
            prop.put("dbVersion", formatDbVersion);

            FAMEDb db = new FAMEDb("test9");
            db.open(filename, prop);

            RegularSeries data1d = getRegularSeries(TiqFrequency.DAILY, 2001,
                    1, 1, 7);

            RegularSeries data1b = (RegularSeries) db
                    .convertDailyToBusiness(data1d);

            assertEquals("data1b should be business after conversion",
                    RegularCalendar.BUSINESS, data1b.getCalendar());

            // db.createUpdateTimeSeries("FM.B.test1", data1b);
            db.createObject("FM.B.1", data1b); // B on FM-B-series.
            assertEquals("fm.b.1 should be business", RegularCalendar.BUSINESS,
                    db.getObject("FM.B.1", TiqConstants.RANGE_OPEN_END,
                            TiqConstants.RANGE_OPEN_END, false).getCalendar());
            assertEquals("fm.b.1 should be daily", RegularCalendar.DAILY, db
                    .getObject("FM.B.1").getCalendar());

            db.createObject("FM.B.2", data1d); // D on FM-B-series.
            assertEquals("fm.b.2 should be business", RegularCalendar.BUSINESS,
                    db.getObject("FM.B.2", TiqConstants.RANGE_OPEN_END,
                            TiqConstants.RANGE_OPEN_END, false).getCalendar());
            assertEquals("fm.b.2 should be daily", RegularCalendar.DAILY, db
                    .getObject("FM.B.2").getCalendar());

            db.createObject("FX.B.3", data1d); // D on FX-B-series.
            assertEquals("fx.b.3 should be business", RegularCalendar.BUSINESS,
                    db.getObject("FX.B.3", TiqConstants.RANGE_OPEN_END,
                            TiqConstants.RANGE_OPEN_END, false).getCalendar());
            assertEquals("fx.b.3 should be daily", RegularCalendar.DAILY, db
                    .getObject("FX.B.3").getCalendar());
            db.createObject("XX.B.3", data1d); // D on foreign series.
            assertEquals("xx.b.3 should be daily", RegularCalendar.DAILY, db
                    .getObject("XX.B.3", TiqConstants.RANGE_OPEN_END,
                            TiqConstants.RANGE_OPEN_END, false).getCalendar());
            assertEquals("xx.b.3 should be daily - 2", RegularCalendar.DAILY,
                    db.getObject("XX.B.3").getCalendar());

            db.createObject("FM.D.3", data1d); // D on FX-D-series.
            assertEquals("fm.d.3 should be daily", RegularCalendar.DAILY, db
                    .getObject("FM.D.3", TiqConstants.RANGE_OPEN_END,
                            TiqConstants.RANGE_OPEN_END, false).getCalendar());
            assertEquals("fm.d.3 should be daily - 2", RegularCalendar.DAILY,
                    db.getObject("FM.D.3").getCalendar());

            db.close(true);
    }

    private RegularSeries getRegularSeries(TiqFrequency frequency,
            int startYear, int startMonth, int startDay, int obs) {
        IndexValueRange range = getIndexValueRange(frequency, startYear,
                startMonth, startDay, obs);

        byte[] statusValues = new byte[obs];
        double[] doubleValues = new double[obs];

        Random r = new Random();
        for (int i = 0; i < obs; i++) {
            doubleValues[i] = r.nextDouble();
            statusValues[i] = TiqConstants.STATUS_OK;
        }

        return new RegularSeries(range, new DoubleList(doubleValues),
                statusValues);
    }

    private IndexValueRange getIndexValueRange(TiqFrequency frequency,
            int startYear, int startMonth, int startDay, int obs) {
        SimpleCalendar cal = null;
        if (frequency == TiqFrequency.WEEKLY) {
            cal = new SimpleCalendar(frequency, SimpleCalendar.REF_FRIDAY, 1);
        } else {
            cal = new SimpleCalendar(frequency);
        }
        TiqDate start = new TiqDate(startYear, startMonth, startDay, frequency);
        IndexValueRange range = null;

        try {
            range = new IndexValueRange(cal, start, start);
        } catch (Throwable e) {
            fail(e.getMessage());
        }

        try {
            range.setPeriodsFromStart(obs);
        } catch (Throwable e) {
            fail(e.getMessage());
        }

        return range;
    }
}
