// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.fame;

import Int.ecb.logging.Level;

import Int.ecb.gesmescb.*;
import Int.ecb.gesmescb.fame.FAMEDb;
import Int.ecb.gesmescb.fame.FAMERevisor;

import Int.ecb.gesmescb.util.ComponentHelper;

/* **************************** Class ***************************** */

/**
 * Abstract class to delete objects from Fame databases.
 *
 * @author
 * <A HREF="mailto:klo@FAME.com">Kam Lo</A>
 * @author
 * <A HREF="mailto:Sami.Airo@ECB.int">Sami Airo</A>
 *
 * @version @(#)$Id: FAMEDeleter.java,v 1.1 2017/01/19 16:14:03 trainot Exp $
 */

public class FAMEDeleter extends FAMERevisor {

	/* ======================= Constants ======================== */

	/* ========================= Fields ========================= */

	/* ====================== Constructors ====================== */

	public FAMEDeleter(
		final ComponentHelper componentHelper,
		final FAMEDb fameDb,
		final MessageBegin messageBegin
	) {
		super(componentHelper, fameDb, messageBegin);
	}

	/* ===================== Public Methods ===================== */

	public boolean handle ( TimeSeries timeSeries ) {
		getComponentHelper().logger.log(
			Level.FINE,
			getComponentHelper().logName,
			"handle",
			"Handling TimeSeries"
		);

		if (timeSeries.getKey().length() == 0) {
			return deleteKeyFamily(timeSeries);
		}

		if (timeSeries.getKey().charAt(0) == TimeSeries.KEY_SEPARATOR) {
			return deleteSiblingSet(timeSeries);
		}

		if (timeSeries.getPeriodFormat().equals("")) {
			return deleteTimeSeries(timeSeries);
		}

		return deleteObservations(timeSeries);
	}


	public boolean handle ( Attribute attribute ) {
		getComponentHelper().logger.log(
			Level.FINE,
			getComponentHelper().logName,
			"handle",
			"Handling Attribute"
		);

		KeyFamilyComponent component = getKeyFamily().getComponent(attribute.getName());

		if (component == null) {
			getComponentHelper().logger.log(
				Level.SEVERE,
				getComponentHelper().logName,
				"handle",
				"Attribute definition not found for " + attribute
			);
			return false;
		}

		if (attribute.getPeriodFormat().equals("")) {
			return deleteAttribute( attribute, component );
		}

		return deleteAttributeObservation( attribute, component );
	}

	/* ===================== Protected Methods ===================== */

	/**
	 * Deletes all objects belonging to a key family.
	 *
	 * @param timeSeries timeSeries key that has the data set identifier.
	 *
	 * @return true if the objects were deleted succesfully; otherwise false
	 */
	protected boolean deleteKeyFamily (TimeSeries timeSeries) {
		getComponentHelper().logger.log(
			Level.WARNING,
			getComponentHelper().logName,
			"deleteKeyFamily",
			"Deletion of key family is not implemented"
		);
		return true;
	}

	/**
	 * Deletes all objects belonging to a sibling set.
	 *
	 * @param timeSeries timeSeries key that has the sibling set key.
	 *
	 * @return true if the objects were deleted succesfully; otherwise false
	 */
	protected boolean deleteSiblingSet (TimeSeries timeSeries) {
		getComponentHelper().logger.log(
			Level.WARNING,
			getComponentHelper().logName,
			"deleteSiblingSet",
			"Deletion of a sibling set is not implemented"
		);
		return true;
	}

	/**
	 * Deletes all objects belonging to a time series.
	 *
	 * @param timeSeries timeSeries key.
	 *
	 * @return true if the objects were deleted succesfully; otherwise false
	 */
	protected boolean deleteTimeSeries (TimeSeries timeSeries) {
		getComponentHelper().logger.log(
			Level.WARNING,
			getComponentHelper().logName,
			"deleteSiblingSet",
			"Deletion of a time series is not implemented"
		);
		return true;
	}

	/**
	 * Deletes observations from a time series.
	 *
	 * @param timeSeries timeSeries key.
	 *
	 * @return true if the observations were deleted succesfully; otherwise false
	 */
	protected boolean deleteObservations (TimeSeries timeSeries) {
		getComponentHelper().logger.log(
			Level.WARNING,
			getComponentHelper().logName,
			"deleteObservations",
			"Deletion of a time series observations is not implemented"
		);
		return true;
	}

	/**
	 * Deletes an attribute object from the database.
	 *
	 * @param attribute attribute to be deleted.
	 * @param component key family component for the attribute
	 *
	 * @return true if the attribute was deleted succesfully; otherwise false
	 */
	protected boolean deleteAttribute (Attribute attribute, KeyFamilyComponent component) {
		getComponentHelper().logger.log(
			Level.WARNING,
			getComponentHelper().logName,
			"deleteAttribute",
			"Deletion of attributes is not implemented"
		);
		return true;
	}

	/**
	 * Deletes an observation from a observation level attribute.
	 *
	 * @param attribute attribute where observation should be deleted.
	 * @param component key family component for the attribute
	 *
	 * @return true if the observation was deleted succesfully; otherwise false
	 */
	protected boolean deleteAttributeObservation (Attribute attribute, KeyFamilyComponent component) {
		getComponentHelper().logger.log(
			Level.WARNING,
			getComponentHelper().logName,
			"deleteAttributeObservation",
			"Deletion of observation level attribute values is not implemented"
		);
		return true;
	}

}
