// ECB/SIS Public License, version 1.0, document reference SIS/2001/116 qqq
//
// Copyright (c) 2002, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.fame;

import java.io.IOException;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.Properties;
import java.util.Map;
import java.util.HashMap;
import java.util.Iterator;

import com.fame.timeiq.Scalar;
import com.fame.timeiq.TiqConstants;

import Int.ecb.logging.Level;

/* ***************************** Class ****************************** */

/**
 * <P>Reads date series and scalars from FAME databases
 * and dispatches {@link Int.ecb.gesmescb.GESMESCBObject}s
 * representing the data read.</P>
 *
 * <P>This class is a specialisation of the
 * {@link Int.ecb.gesmescb.fame.FAMEReader} where the ENVELOPE scalars,
 * if available, are handled as preferences.</P>
 *
 * @author <A HREF="mailto:francois.hoetlzl@ecb.europa.eu">Francosi Hoeltzli</A>
 *
 * @version @(#)$Id: FAMEIDBReader.java,v 1.1 2017/01/19 16:14:03 trainot Exp $
 */

public class FAMEIDBReader extends FAMEReader {

	/* ========================= Fields ========================= */

	/**
	 * Store the envelope.
	 */	
	private Map envelopeNames;
	private Map envelopeProperties;
	
	/**
	 * Store the name of the source database.
	 */	
	private String filename;
	
	/* ====================== Constructors ====================== */
	
	/**
	 * Construct a <CODE>FAMEIDBReader</CODE> to read from FAME
	 * databases. The difference with the
	 * {@link Int.ecb.fame.gesmescb.fame.FAMEReader} is purely that
	 * here, an attempt is made to get the preferences out of the
	 * envelope scalars instead of the preference file.
	 *
	 * @param name	component used for logging and preferences.
	 * @param filename  source database name to read data from.
	 */
	public FAMEIDBReader(
		String name,
		String filename
	) throws IOException {
		super( name, filename );
		this.filename = filename; 
		this.envelopeNames = new HashMap();
		this.envelopeProperties = new HashMap();
		getEnvelopeProperties();
	}

	/* ===================== Public Methods ===================== */

	/**
	 * Returns the preference from the envelope.
	 *
	 * @param key  the name of the property
	 * @return the value of the preference.
	 */
	public String getEnvelopeProperty(
		String key
	) {
		return (String) envelopeProperties.get(key);
	}

	/* ==================== Private Methods ===================== */

	/**
	 * <P>Retrieves all preference values available in the envelope
	 * and store them in the maps <CODE>envelopeProperties</CODE>.</P>
	 *
	 */
	private void getEnvelopeProperties() {

		envelopeNames.clear();
		envelopeProperties.clear();
		envelopeNames.put("IREF", "ENVELOPE.IREF");
		envelopeNames.put("sender", "ENVELOPE.SENDER");
		envelopeNames.put("receiver", "ENVELOPE.RECEIVER");
		envelopeNames.put("maintenanceAgency", "ENVELOPE.AGENCY");
		envelopeNames.put("identity", "ENVELOPE.IDENTITY");
		envelopeNames.put("dataSetName", "ENVELOPE.DATASET");
		envelopeNames.put("extractionTime", "ENVELOPE.EXTRACTION_TIME");
		envelopeNames.put("keyFamilyName", "ENVELOPE.KEY_FAMILY");
		envelopeNames.put("reportingPeriod", "ENVELOPE.REPORTING_PERIOD");
		//trainot fix reporting period
		envelopeNames.put("reportingPeriodFormat", "ENVELOPE.REPORTING_PERIOD_FORMAT");
		envelopeNames.put("originalDataset", "ENVELOPE.ORIGINAL_DATASET");

		FAMEDb db = getSourceDb();
		boolean wasOpen = db.isOpen();
		if ( !wasOpen ) {
			Properties prop = new Properties();
			prop.put("mode", "READ_ONLY");
			db.open( filename, prop );
		}
		
		for( Iterator it = envelopeNames.keySet().iterator()
			; it.hasNext()
			;
		) {
			String key = (String) it.next();
			Scalar scalar = (Scalar) db.getObject(
				(String) envelopeNames.get( key )
			);
			 // trainot make reporting period format optional (Eurostat)
			if (key.equals("reportingPeriodFormat")) {
				if (scalar != null) {
					envelopeProperties.put(
							key,
							scalar.getTiqValue().getStringValue()
						);
				} else {
					envelopeProperties.put(
							key, "-");
				}
				
			} else {
				envelopeProperties.put(
					key,
					scalar.getTiqValue().getStringValue()
				);
			}
		}
		if ( !wasOpen ) {
			db.close(false);
		}
	}

	/**
	 * <P>Retrieves a preference or envelope value.</P>
	 *
	 * @param key	key for the preference
	 * @param manditoryFlag indicates whether a preference value is
	 * manditory for the this component.
	 *
	 * @return preference value for the specified key
	 *
	 * @throws NullPointerException if the preference variable
	 * for the key is undefined and is manditory.
	 */
	protected String getPreference ( 
		final String key, 
		final boolean mandatory,
		final String defaultValue
	)
		throws NullPointerException 
	{
		String preference = getEnvelopeProperty( key );
		if ( null == preference
			||
			"".equals( preference )
		) {
			preference = super.getPreference( key, mandatory, defaultValue );
		}
		return preference;
	}

	public String getMandatoryPreference (
		final String key
	) throws NullPointerException
	{
		String preference = getPreference(
			key, true, ""
		);
		return preference;
	}
}

