// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2002, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.fame;

import java.io.IOException;

import java.text.DecimalFormat;
import java.text.SimpleDateFormat;

import java.util.ArrayList;
import java.util.Date;
import java.util.Iterator;
import java.util.Properties;
import java.util.TreeMap;
import java.util.TreeSet;
import java.util.Vector;

import Int.ecb.logging.Level;

import Int.ecb.gesmescb.*;
import Int.ecb.gesmescb.metabase.*;

import Int.ecb.gesmescb.event.ErrorEvent;
import Int.ecb.gesmescb.event.GESMESCBReader;
import Int.ecb.gesmescb.event.GESMESCBEventSourceAdapter;

import Int.ecb.gesmescb.util.ComponentHelper;
import Int.ecb.gesmescb.util.DecimalsTrimmer;
import Int.ecb.gesmescb.util.Named;
import Int.ecb.gesmescb.util.KeyTokenizer;

import com.fame.timeiq.FunctorConsistencyChkException;
import com.fame.timeiq.Scalar;
import com.fame.timeiq.TiqClass;
import com.fame.timeiq.TiqConstants;
import com.fame.timeiq.TiqObject;
import com.fame.timeiq.TiqType;
import com.fame.timeiq.TiqView;
import com.fame.timeiq.RegularSeries;
import com.fame.timeiq.TiqObserved;

import com.fame.timeiq.data.ObservationList;

import com.fame.timeiq.dates.IndexValueRange;
import com.fame.timeiq.dates.RangeTooLargeChkException;
import com.fame.timeiq.dates.RegularCalendar;
import com.fame.timeiq.dates.TiqDate;
import com.fame.timeiq.dates.TiqFrequency;

import com.fame.timeiq.functors.operators.And;
import com.fame.timeiq.functors.operators.EQ;
import com.fame.timeiq.functors.operators.IndexBy;
import com.fame.timeiq.functors.operators.NE;
import com.fame.timeiq.functors.operators.Or;

import com.fame.timeiq.functors.util.Which;

import com.fame.timeiq.persistence.ConnectionFailedChkException;
import com.fame.timeiq.persistence.DataObjectInfo;
import com.fame.timeiq.persistence.ObjectAccessChkException;
import com.fame.timeiq.persistence.ScreeningCriteria;

/* ***************************** Class ****************************** */

/**
 * <P>Reads date series and scalars from FAME databases
 * and dispatches {@link Int.ecb.gesmescb.GESMESCBObject}s
 * representing the data read.</P>
 *
 * <P>Data is read from the source database. An optional target
 * database may be specified to define data already stored at the
 * destination database. In this case, the source and target
 * databases are compared for difference. Only data that differ will
 * be dispatched.</P>
 * <P>Optional update time series key list file defines which time
 * series data sets to check for updates. Optional delete time series
 * key list file to defines which time series data sets to delete.</P>
 *
 * <center><table border="3" cellpadding="3">
 * <caption>
 * <a name="prefs">Available {@link Int.ecb.prefs preference} settings</a>
 * </caption>
 * <tr>
 * <th>preference name</th>
 * <th>type</th>
 * <th>meaning</th>
 * <th>default</th>
 * </tr>
 * <tr><td><code>IREF</code></td><td>string</td>
 *	<td>The IREF number to use in the {@link Int.ecb.gesmescb.InterchangeBegin} 
 *	event. For format see {@link Int.ecb.gesmescb.InterchangeBegin}.
 *	Mandatory.</td></tr>
 * <tr><td><code>sender</code></td><td>string</td>
 *	<td>The sender code for the {@link Int.ecb.gesmescb.InterchangeBegin} and 
 *	{@link Int.ecb.gesmescb.MessageBegin} events.
 *	Mandatory.</td></tr>
 * <tr><td><code>receiver</code></td><td>string</td>
 *	<td>The receiver code for the {@link Int.ecb.gesmescb.InterchangeBegin} and 
 *	{@link Int.ecb.gesmescb.MessageBegin} events.
 *	Mandatory.</td></tr>
 * <tr><td><code>isTest</code></td><td>boolean</td>
 *	<td>Determines if the {@link Int.ecb.gesmescb.InterchangeBegin}
 *	is marked as test.
 *	Optional.</td>
 *	<td><code>true</code></td></tr>
 * <tr><td><code>maintenanceAgency</code></td><td>String</td>
 *	<td>Maintenance agency identifier for the
 *	{@link Int.ecb.gesmescb.MessageBegin} event.
 *	Mandatory.</td></tr>
 * <tr><td><code>identity</code></td><td>String</td>
 *	<td>Identity text for the {@link Int.ecb.gesmescb.MessageBegin} event.
 *	Optional.</td></tr>
 * <tr><td><code>dataSetName</code></td><td>String</td>
 *	<td>Data set name for the {@link Int.ecb.gesmescb.MessageBegin} event.
 *	Mandatory.</td></tr>
 * <tr><td><code>extractionTime</code></td><td>String</td>
 *	<td>Data set extraction time stamp for the {@link Int.ecb.gesmescb.MessageBegin} event.
 *	Optional. In ECB internal use also used as the release time.</td></tr>
 * <tr><td><code>reportingPeriod</code></td><td>String</td>
 *	<td>Reporting period for the {@link Int.ecb.gesmescb.MessageBegin} event.
 *	Optional.</td></tr>
 * <tr><td><code>reportingPeriodFormat</code></td><td>String</td>
 *	<td>Reporting period format for the
 *	{@link Int.ecb.gesmescb.MessageBegin} event.
 *	Optional.</td></tr>
 * <tr><td><code>keyFamilyName</code></td><td>String</td>
 *	<td>Name of the key family for the
 *	{@link Int.ecb.gesmescb.MessageBegin} event.
 *	Mandatory.</td></tr>
 * <tr><td><code>internalKeyFamilyName</code></td><td>String</td>
 *	<td>Internal key family name. This is used to search the source database.
 *	Objects in the source database must have this as the name prefix.
 *	Mandatory.</td></tr>
 * <tr><td><code>updateTimeSeriesKeyFile</code></td><td>String</td>
 *	<td>Name of the file that contains the keys for the time series to read.
 *	Keys must not have the data set prefix.
 *	Optional.</td></tr>
 * <tr><td><code>deleteTimeSeriesKeyFile</code></td><td>String</td>
 *	<td>Name of the file that contains the keys for the time series,
 *	attributes, and sibling groups to delete.
 *	Keys must not have the data set prefix.
 *	Optional.</td></tr>
 * <tr><td><code>sourceDatabase</code></td><td>String</td>
 *	<td>Name of the source Fame database.
 *	Mandatory.</td></tr>
 * <tr><td><code>targetDatabase</code></td><td>String</td>
 *	<td>Name of the target Fame database.
 *	Optional.</td></tr>
 * <tr><td><code>attributeDatabase</code></td><td>String</td>
 *	<td>Name of the Fame database that contains the DECIMALS attributes.
 *	Optional.</td></tr>
 * <tr><td><code>suppressAttributes</code></td><td>boolean</td>
 *	<td>Whether to suppress reporting of attributes in the footnote section.
 *	Optional.</td><td><code>false</code></td></tr>
 * <tr><td><code>mixAttributes</code></td><td>boolean</td>
 *	<td>Whether to dispatch {@link Attribute}s before the respective
 *  {@link TimeSeries}. Optional.</td><td><code>false</code></td></tr>
 * <tr><td><code>allowAliases</code></td><td>boolean</td>
 *	<td>Whether to allow aliases in the topic list and return them in a database query.
 * </td><td><code>true</code></td></tr>
 * <tr><td><code>ignoreDecimalsAttributes</code></td><td>boolean</td>
 *	<td>If true observation values are reported with full precision.
 * This may not be compatible with GESMES/TS rules.</td><td><code>false</code></td></tr>
 * <tr><td><code>originalDataset</code></td><td>String</td>
 *	<td>The original Dataset as read from the properties.
 * </td><td>dataSetName</td></tr>
 * </table></center>
 *
 *
 * @author <A HREF="mailto:klo@fame.com">Kam Lo</A>
 * @author <A HREF="mailto:sami.airo@ecb.int">Sami Airo</A>
 * @author <A HREF="mailto:francois.hoeltzli@ecb.europa.eu">Francois Hoeltzli</A>
 *
 * @version @(#)$Id: FAMEReader.java,v 1.2 2018/03/05 10:19:53 trainot Exp $
 */

public class FAMEReader
	extends GESMESCBEventSourceAdapter
	implements GESMESCBReader, Named
{
	/* ======================= Constants ======================== */

	/**
	 * The separator used in the key.
	 */
	private static final String KEY_SEPARATOR = String.valueOf( Attribute.KEY_SEPARATOR );

    /**
     * Time series observation status dimension name.
     */
    private static final String OBS_STATUS = "OBS_STATUS";

    /**
     * Time series observation confidentiality dimension name.
     */
    private static final String OBS_CONF = "OBS_CONF";

    /**
     * Time series pre-break dimension name.
     */
    private static final String BREAK = "BREAK";

    /**
     * Default decimals value
     */
    private static final int DEFAULT_DECIMALS = 2;

    /**
     * Maximum decimals value
     */
    private static final int MAX_DECIMALS = 16;

	/* ========================= Fields ========================= */

	/**
	 * Takes care of logging and preferences.
	 */
	protected final ComponentHelper componentHelper;

	/**
	 * Key family
	 */
	private KeyFamily keyFamily;

	/**
	 * Source database. Contains data for updating/deleting at the
	 * destination database.
	 */
	private FAMEDb sourceDatabase;

	/**
	 * Target database. Contains data already in existence at the
	 * destination database.
	 */
	private FAMEDb targetDatabase;

	/**
	 * Attribute database. Contains attributes needed by the reader.
	 * Mainly the DECIMALS
	 */
	private FAMEDb attributeDatabase;

	/**
	 * Sender ID, read from preferences.
	 */
	private String sender;

	/**
	 * Receiver ID, read from preferences.
	 */
	private String receiver;

	/**
	 * Maintenance agency, read from preferences.
	 */
	private String maintenanceAgency;

	/**
	 * Message identity, read from preferences.
	 */
	private String identity;

	/**
	 * Dataset identity, read from preferences.
	 */
	private String dataSetName;

	/**
	 * Data set extraction time stamp.
	 */
	private String extractionTime;

	/**
	 * Reporting period, read from preferences.
	 */
	private String reportingPeriod;

	/**
	 * Reporting period format, read from preferences.
	 */
	private String reportingPeriodFormat;

	/**
	 * Key family name, read from preferences.
	 */
	private String keyFamilyName;

	/**
	 * Internal key family name, read from preferences.
	 */
	private String internalKeyFamilyName;

	/** Whether to suppress the footnote section. */
	private boolean suppressAttributes;

	/** Whether to dispatch {@link Attribute}s
	 * before the respective {@link TimeSeries}. */
	private boolean mixAttributes;

	/**
	 * IREF string, read from preferences.
	 */
	private String iref;

	/**
	 * isTest flag, read from preferences.
	 */
	private boolean isTest;

	/**
	 * Current MREF number.
	 */
	private int mrefNumber = 1;

	/**
	 * Message dispatched flag.
	 */
	private boolean messageDispatched = false;

	/**
	 * List of time series observations to delete.
	 */
	private Vector timeSeriesObsDeleteList = new Vector();

	/**
	 * List of attributes to delete.
	 */
	private Vector attributeDeleteList = new Vector();

	/**
	 * List of attribute observations to deletion.
	 */
	private Vector attributeObsDeleteList = new Vector();

	/**
	 * List of attributes already dispatched.
	 */
	private TreeSet dispatchedAttributeKeyList = new TreeSet();

	/**
	 * List of time series keys mapped to sibling keys.
	 */
	private TreeMap siblingKeyMap = new TreeMap();

	/**
	 * if true write values with full precision, read from preferences.
	 */
	private boolean ignoreDecimalsAttributes;

	/**
	 * if true, use HalfUp Rounding mode instead of default Half Even.
	 */
	private boolean useHalfUpRounding;
	/**
	 * if true allow aliases in the topic lists and wildcard results
	 */
	private boolean allowAliases = true;

	/**
	 * originalDSI read from preferences (dataSetName by default)
	 */
	private String originalDataset;

	/* ====================== Constructors ====================== */

	public FAMEReader () 
		throws IOException
	{
		this("", "");
	}

	/**
	 * Construct a <CODE>FAMEReader</CODE> to read from FAME
	 * databases. If source database name not defined, it is read
	 * from preferences. Also reads the optional target database
	 * name from preferences.
	 *
	 * @param name	component used for logging and preferences.
	 * @param filename  source database name to read data from.
	 */
	public FAMEReader ( final String name, final String filename )
		throws IOException
	{
		componentHelper = new ComponentHelper( name, this );

		sourceDatabase = new FAMEDb("source");
		targetDatabase = new FAMEDb("target");
		attributeDatabase = new FAMEDb("attribute");
		
		// We don't allow for this to be set in the source or attribute database.
		ignoreDecimalsAttributes = "true".equalsIgnoreCase(
			componentHelper.getPreference( "ignoreDecimalsAttributes", "false" )
		);
		componentHelper.logger.log(
			Level.CONFIG,
			componentHelper.logName,
			"ignoreDecimalsAttributes set to : " + ignoreDecimalsAttributes + ".", ""
		);

		useHalfUpRounding = "true".equalsIgnoreCase(
			componentHelper.getPreference( "useHalfUpRounding", "false" )
		);
		componentHelper.logger.log(
			Level.CONFIG,
			componentHelper.logName,
			"useHalfUpRounding set to : " + useHalfUpRounding + ".", ""
		);
	
	
	}

	/**
	 * Construct a <CODE>FAMEReader</CODE> to read from FAME
	 * databases. Reads the source and optional target database name
	 * from preferences.
	 *
	 * @param name	used for logging and preferences.
	 *
	 * @throws NullPointerException if the preference variable
	 * <CODE>sourceDatabase</CODE> is undefined.
	 */
	public FAMEReader ( final String name )
		throws IOException
	{
		this( name, null );
	}

	/* ===================== Public Methods ===================== */

	/**
	 * Returns the name of the component.
	 *
	 * @return the component name.
	 */
	public String getName () {
		return componentHelper.name;
	}

	/**
	 * <P>Retrieves data from databases and dispatches them as
	 * {@link Int.ecb.gesmescb.GESMESCBObject}s to all listeners.</P>
	 *
     * <P>Time series keys are read from the optional delete files.
     * If deletions keys have been specified, time series data set
     * deletion events are sent to the listeners.</P>
     *
     * <P>If update keys have been specified, these will be used to
     * check in the source database for updates. Otherwise, an update
     * list will be generated by searching for observation status
     * objects in the source database.</P>
     *
     * <P>If a target database has been defined, the source data and
     * target data are compared for changes. Changes are sent to the
     * listeners. If not target database has been defined, all data
     * in the update list will be sent to the listeners.</P>
     *
     * <P>Finally, time series observation level, whole attributes
     * and attribute observation level deletions are dispatched.</P>
	 */
	public void run () {
		try {
			getFAMEReaderPreferences();
		} catch (Exception e) {
			logAndDispatchSevereError(
				"Cannot read mandatory preferences.",
				"run()",
				e
			);
			return;
		}
		
		if ( !openDatabase(
			"sourceDatabase", 
			true, 
			sourceDatabase) 
		) {
			logAndDispatchSevereError(
				"Cannot not open source database.",
				"run()"
			);
			return;
	   }

		if ( !openDatabase(
			"targetDatabase", 
			false, 
			targetDatabase) 
		) {
			logAndDispatchSevereError(
				"Could not open target database.",
				"run()"
	    	);
			return;
		}

		if ( !openDatabase(
			"attributeDatabase", 
			false, 
			attributeDatabase) 
		) {
			logAndDispatchSevereError(
				"Could not open attribute database.",
				"run()"
	    	);
			return;
		}

		MaintenanceAgency agency
			= Metabase.getMetabase().getMaintenanceAgency( maintenanceAgency );
		if ( agency == null ) {
			logAndDispatchSevereError(
				"Could not get Metabase values for maintenance agency \""
			    + maintenanceAgency + "\".",
				"run()"
			);
			return;
		}

		keyFamily = agency.getKeyFamily( keyFamilyName );
		if ( keyFamily == null ) {
			logAndDispatchSevereError(
	    		"Could not get Metabase values for key family \""
			   + keyFamilyName + "\" maintained by agency \""
			   + maintenanceAgency + "\".",
				"run()"
			);
			return;
	   }

		try {
		   // List of time series keys to update
   		TimeSeriesKeyList updateTimeSeriesKeyList = getUpdateTimeSeriesKeyList();
		   // Quit if error getting time series key list
		   if ( updateTimeSeriesKeyList == null ) {
				return;
		   }

		   // List of time series keys to delete
	   	TimeSeriesKeyList deleteTimeSeriesKeyList = getDeleteTimeSeriesKeyList();
			// Quit if error getting time series key list
	   	if ( deleteTimeSeriesKeyList == null ) {
				return;
			}

		   dispatchInterchangeBegin();
		
			// Dispatch events to delete time series
			dispatchTimeSeriesDeletes( deleteTimeSeriesKeyList );

			if ( messageDispatched ) {
				dispatchMessageEnd();
			}

			// Dispatch time series updates
			if ( mixAttributes ) {
				for (
					final Iterator keyListIterator
						= updateTimeSeriesKeyList.iterator();
					keyListIterator.hasNext();
				) {
					final String timeSeriesKey = ( String ) keyListIterator.next();

					if (
						! suppressAttributes
						&&
						! dispatchAttributeUpdates( timeSeriesKey )
					) {
						break;
					}

					if ( ! dispatchTimeSeriesUpdate( timeSeriesKey ) ) {
						break;
					}
				}
			} else {
				for (
					final Iterator keyListIterator
						= updateTimeSeriesKeyList.iterator();
					keyListIterator.hasNext();
				) {
					final String timeSeriesKey
					= (String) keyListIterator.next();
					if ( ! dispatchTimeSeriesUpdate( timeSeriesKey ) ) {
						break;
					}
				}

				if ( ! suppressAttributes ) {
					for (
						final Iterator keyListIterator
							= updateTimeSeriesKeyList.iterator();
						keyListIterator.hasNext();
					) {
						final String timeSeriesKey
						= (String) keyListIterator.next();
						if ( ! dispatchAttributeUpdates( timeSeriesKey ) ) {
							break;
						}
					}
				}
			}

			if ( messageDispatched ) {
				dispatchMessageEnd();
			}

			// Dispatch deletion message
			// Delete Time Series Observations
			dispatchTimeSeriesObsDeletes();

			if ( !suppressAttributes ) {
				// Delete Attributes
				dispatchAttributeDeletes();

				// Delete Attribute Observations
				dispatchAttributeObsDeletes();
			}

			if ( messageDispatched ) {
				dispatchMessageEnd();
			}

			dispatchInterchangeEnd();
		} catch ( final Exception e ) {
			logAndDispatchSevereError(
				"Unexpected error: " + e,
				"run()",
				e
			);
		}
	}

	/**
	 * <P>Close the <CODE>FAMEReader</CODE>.</P>
	 *
	 * <P>This method should always be called
	 * when the <CODE>FAMEReader</CODE>
	 * has served its purpose.</P>
	 *
	 * <P>Closes all databases opened by the
	 * <CODE>FAMEReader</CODE>.</P>
	 */
	public void close () {
		try {
			// do here your closing.
			// Probably you will flush buffers, generate last events,
			// close opened files and log about closing here.

			// Close source database
			sourceDatabase.close( false );

			// Close target database if it has been opened
			if ( targetDatabase.isOpen() ) {
			    targetDatabase.close( false );
			}

			// Close attribute database if it has been opened
			if ( attributeDatabase.isOpen() ) {
			    attributeDatabase.close( false );
			}
		}
		catch ( final Exception e ) {
			logAndDispatchSevereError(
				"Problem: " + e.getMessage() + ".",
				"close()",
				e
			);
		}
	}

	/* ==================== Private Methods ===================== */

	/**
	 * <P>Retrieves all preference value required by the
	 * <CODE>FAMEReader</CODE>.</P>
	 *
	 * @throws NullPointerException if a manditory preference
	 * variable has not been defined.
	 */
	private void getFAMEReaderPreferences ()
		throws NullPointerException 
	{
		// Read IREF from preferences
		iref = getPreference( "IREF", true, "" );
		// Read sender from preferences
		sender = getPreference( "sender", true, "" );
		// Read receiver from preferences
		receiver = getPreference( "receiver", true, "" );
		// Read isTest from preferences
		isTest = getPreference( "isTest", false, "" ).equalsIgnoreCase( "true" );
		// Read maintenanceAgency from preferences
		maintenanceAgency = getPreference( "maintenanceAgency", true, "" );
		// Read identity from preferences
		identity = getPreference( "identity", false, "" );
		// Read datasetid from preferences
		dataSetName = getPreference( "dataSetName", true, "" );
		
		// Read extractionTime from preferences
		// Construct current date and time string in CCYYMMDDhhmm
		// format for extraction date and time
		SimpleDateFormat formatter = new SimpleDateFormat ( "yyyyMMddHHmm" );
		extractionTime = getPreference( "extractionTime", false, formatter.format( new Date() ) );
		
		// Read reportingPeriod from preferences
		reportingPeriod = getPreference( "reportingPeriod", false, "" );
		// Read reportingPeriodFormat from preferences
		reportingPeriodFormat = getPreference( "reportingPeriodFormat", false, "" );
		// Read keyFamilyName from preferences
		keyFamilyName = getPreference( "keyFamilyName", true, "" );
		// Read internalKeyFamilyName from preferences
		internalKeyFamilyName = getPreference( "internalKeyFamilyName", true, ""	);
		// Whether to suppress attributes
		suppressAttributes = "true".equalsIgnoreCase(getPreference( "suppressAttributes", false, "false" ));
		mixAttributes = "true".equalsIgnoreCase(
			getPreference( "mixAttributes", false, "false" )
		);
		// Whether to allow aliases
		allowAliases = "true".equalsIgnoreCase(getPreference( "allowAliases", false, "true"));
		originalDataset = getPreference( "originalDataset", true, dataSetName );
	}


	/**
	 * <P>Returns the associated sibling key for a
	 * time series key.</P>
	 *
	 * @param timeSeriesKey time series key
	 *
	 * @return sibling key for the time series key
	 */
	protected String getSiblingKey ( final String timeSeriesKey, final int freqDimPos ) {
		// Get matching sibling key
		String siblingKey = ( String ) siblingKeyMap.get( timeSeriesKey );

		// Check if sibling key has not yet been generated
		if ( siblingKey == null ) {
			siblingKey = "";

			// Construct sibling level key
			int count = 1;
			KeyTokenizer keyTokenizer = new KeyTokenizer( timeSeriesKey );
			while ( keyTokenizer.hasNext() ) {
				siblingKey +=  (1 < count ? KEY_SEPARATOR : "");
				if (count == freqDimPos ) {
					keyTokenizer.next();
				} else {
					siblingKey += (String)keyTokenizer.next();
				}
				count++;
			}
		}
		
		return siblingKey;
	}

	/**
	 * <P>Retrieves the frequency dimension part of the time
	 * series key.</P>
	 *
	 * @param timeSeriesKey time series key
	 *
	 * @return frequency dimension
	 */
	protected String getFrequencyDimension ( 
		final String timeSeriesKey,
		final int freqDimPos
	) {
		// First part is the internal key family name
		int count = 0;
		KeyTokenizer keyTokenizer = new KeyTokenizer( timeSeriesKey );
		while ( keyTokenizer.hasNext() ) {
			// Found frequency dimension
			if ( count == freqDimPos ) {
				return (String)keyTokenizer.next();
			}
			keyTokenizer.next();
			count++;
		}
		return null;
	}

	/**
	 * <P>Check the frequency of the FAME time series object
	 * matches with the frequency dimension.</P>
	 *
	 * @param frequencyDimension	Frequency dimension part of the
	 * time series key
	 * @param timeSeries	FAME time series object
	 *
	 * @return frequency matches or not
	 */
	protected boolean validateFrequency (
	    final String frequencyDimension, 
		 final RegularCalendar timeSeriesCalendar
	) {
		if ( frequencyDimension.equals( "A" ) 
			&& timeSeriesCalendar.equals( RegularCalendar.ANNUAL ) ) {
				return true;
		}
		if ( frequencyDimension.equals( "B" ) 
//	   	&& timeSeriesCalendar.equals( RegularCalendar.BUSINESS ) ) {
	   	&& timeSeriesCalendar.equals( RegularCalendar.DAILY ) ) {  // Business series are automatically converted to daily
				return true;
		}
		if ( frequencyDimension.equals( "D" ) 
	   	&& timeSeriesCalendar.equals( RegularCalendar.DAILY ) ) {
				return true;
		}
		if ( frequencyDimension.equals( "H" )
			&& timeSeriesCalendar.getFrequency().equals( TiqFrequency.SEMIANNUAL ) ) {
				return true;
		}
		if ( frequencyDimension.equals( "M" )
			&& timeSeriesCalendar.equals( RegularCalendar.MONTHLY ) ) {
				return true;
		}
		if ( frequencyDimension.equals( "N" )
			&& timeSeriesCalendar.equals( RegularCalendar.MINUTELY ) ) {
				return true;
		}
		if ( frequencyDimension.equals( "Q" )
			&& timeSeriesCalendar.equals( RegularCalendar.QUARTERLY ) ) {
				return true;
		}
		if ( frequencyDimension.equals( "T" )
			&& timeSeriesCalendar.getFrequency().equals( TiqFrequency.TWICEMONTHLY ) ) {
				return true;
		}
		if ( frequencyDimension.equals( "W" )
			&& timeSeriesCalendar.getFrequency().equals( TiqFrequency.WEEKLY ) ) {
				return true;
		}
		return false;
	}
	protected FAMEDb getSourceDb () {
		return sourceDatabase;
	}

	/**
	 * <P>Retrieves list of time series keys to check for
	 * updates.</P>
	 *
	 * <P>If optional update time series key list file has
	 * been specified, retrieves the list from file. Otherwise,
	 * the list is generated by finding all time series
	 * observation status objects in the source database.</P>
	 *
	 * @return List of time series keys to update
	 */
	private TimeSeriesKeyList getUpdateTimeSeriesKeyList () {
		// Check if update key file has been specified in preferences
		// Read update list file from preferences
		final String updateTimeSeriesKeyFile = getPreference( "updateTimeSeriesKeyFile", false, "" );
		if ( updateTimeSeriesKeyFile.equals( "" ) ) {
			// No preference given, get list from database
			return getUpdateTimeSeriesKeyList( sourceDatabase );
		}
		
	   // Read from list from file
	   TimeSeriesKeyList keyList = new TimeSeriesKeyList();
	   try {
			keyList.readFromFile( updateTimeSeriesKeyFile );
		} catch ( IOException e ) {
			logAndDispatchSevereError(
				"Error reading update time series key file "
				+ updateTimeSeriesKeyFile + ".",
				"getUpdateTimeSeriesKeyList()",
				e
			);
			return null;
		}

		// Check there are keys defined in the file
		if ( keyList.isEmpty() ) {
			componentHelper.logger.log(
				Level.WARNING,
				componentHelper.logName,
				"getUpdateTimeSeriesKeyList()",
				"No time series keys found in update file "
				+ updateTimeSeriesKeyFile + "."
			);
		}
	
		return keyList;
	}

	/**
	 * <P>Retrieves list of time series keys from database
	 * to check for updates. Looks for all observation status
	 * objects in the database and removes the status suffix
	 * to get the list of time series keys.</P>
	 *
	 * @param database	FAME database to search for observation
	 * status objects
	 *
	 * @return List of time series keys to update
	 */
	private TimeSeriesKeyList getUpdateTimeSeriesKeyList (
	    final FAMEDb database
	) {
		TimeSeriesKeyList keyList = new TimeSeriesKeyList();
		ScreeningCriteria screen = new ScreeningCriteria();
		screen.setAllowAliases(allowAliases);

		// <Internal Key Family>.*.OBS_STATUS objects in database
		Iterator objectIterator
			= database.getDataObjectInfoIterator(
				internalKeyFamilyName + KEY_SEPARATOR + "*" + KEY_SEPARATOR + OBS_STATUS,
				screen,
				false
			);

		if (! objectIterator.hasNext() ) {
			componentHelper.logger.log(
				Level.WARNING,
				componentHelper.logName,
				"getObjectIterator()",
				"No observation status objects ("
				+ internalKeyFamilyName + KEY_SEPARATOR
				+ "*" + KEY_SEPARATOR + OBS_STATUS
				+ ") found in database \""
				+ database.toString()
				+ "\" to generate time series update list."
			);
			return keyList;
		}

		try {
			// Construct new list of names without internal key
			// family prefix and .OBS_STATUS suffix
			while ( objectIterator.hasNext() ) {
				// Get next time series name
				String timeSeriesName = ( ( DataObjectInfo ) objectIterator.next() ).getName();

				// Ignore names without internal key
				// family name prefix
				if ( timeSeriesName.startsWith( internalKeyFamilyName + KEY_SEPARATOR ) ) {
					// Strip off internal key family name
					timeSeriesName	= timeSeriesName.substring( internalKeyFamilyName.length() + 1	);

					// Strip off ".OBS_STATUS" from end of name
					if ( timeSeriesName.endsWith(	KEY_SEPARATOR + OBS_STATUS ) ) {
						timeSeriesName = timeSeriesName.substring( 
										0,
										timeSeriesName.length() - ( KEY_SEPARATOR + OBS_STATUS ).length()
						);
					}

					keyList.add( timeSeriesName );
				} else {
					componentHelper.logger.log(
						Level.INFO,
						componentHelper.logName,
						"getObjectIterator()",
						"Object " + timeSeriesName
						+ " does not belong to internal key series family "
						+ internalKeyFamilyName + "."
					);
				}
			}
	    } catch ( ConnectionFailedChkException e ) {
		 	logAndDispatchSevereError(
				"Connection to the server failed.",
				"getObjectIterator()",
				e
			);
			return null;
	    } catch ( ObjectAccessChkException e ) {
			logAndDispatchSevereError(
				"Error accessing object.",
				"getObjectIterator()",
				e
			);
			return null;
		}
		return keyList;
	}

	/**
	 * <P>Retrieves list of time series keys to delete from deletion
	 * file.</P>
	 *
	 * @return List of time series keys to delete
	 */
	private TimeSeriesKeyList getDeleteTimeSeriesKeyList () {
		TimeSeriesKeyList keyList = new TimeSeriesKeyList();

		// Read delete list file from preferences
		String deleteTimeSeriesKeyFile = getPreference( "deleteTimeSeriesKeyFile", false, "" );

		// Check if delete key file has been specified in preferences
		if ( deleteTimeSeriesKeyFile.equals( "" ) ) {
			return keyList;
		}
			
		// Read from list from file
		try {
			keyList.readFromFile( deleteTimeSeriesKeyFile );
		} catch ( IOException e ) {
			logAndDispatchSevereError(
				"Error reading delete time series key file " + deleteTimeSeriesKeyFile + ".",
		    	"getDeleteTimeSeriesKeyList()",
				e
			);
			return null;
		}
			
		// Check there are keys defined in the file
		if ( keyList.isEmpty() ) {
			componentHelper.logger.log(
				Level.WARNING,
				componentHelper.logName,
				"getDeleteTimeSeriesKeyList()",
				"No time series keys found in deletion file "
				+ deleteTimeSeriesKeyFile + "."
			);
		}

		return keyList;
	}

	/**
	 * <P>Dispatch interchange begin event.</P>
	 */
	private void dispatchInterchangeBegin () {
		// Create string containing current date
		// and time in YYMMDD:hhmm format
		SimpleDateFormat formatter = new SimpleDateFormat ( "yyMMdd:HHmm" );
		String today = formatter.format( new Date() );

		// Dispatch Interchange Begin
		final InterchangeBegin interchangeBegin
			= new InterchangeBegin( iref, sender, receiver, today, isTest );
		
		componentHelper.logger.log(
			Level.FINE,
			componentHelper.logName,
		   "dispatchInterchangeBegin()",
			"Dispatching Interchange Begin",
			interchangeBegin
	    );

		dispatch( interchangeBegin );
	}


	/**
	 * <P>Dispatch interchange end event.</P>
	 */
	private void dispatchInterchangeEnd () {
		final InterchangeEnd interchangeEnd = new InterchangeEnd( iref );
		componentHelper.logger.log(
			Level.FINE,
			componentHelper.logName,
			"dispatchInterchangeEnd()",
			"Dispatching Interchange End",
			interchangeEnd
		);
		dispatch( interchangeEnd );
	}

	/**
	 * <P>Dispatch message begin event.</P>
	 *
	 * @param action	action to perform
	 */
	private void dispatchMessageBegin ( 
		String action 
	) {
		// Construct MREF tag
		DecimalFormat mrefFormat = new DecimalFormat( "000000" );
		String mref = "MREF" + mrefFormat.format( mrefNumber );

		final MessageBegin messageBegin = new MessageBegin(
			mref, sender, receiver, maintenanceAgency,
			MessageBegin.STATISTICAL_DATA,
			identity,
			new Contact[]{},
			dataSetName,
			action,
			extractionTime,
			reportingPeriod, reportingPeriodFormat,
			keyFamilyName,
			originalDataset
		);
		
		componentHelper.logger.log(
		    Level.FINE,
			componentHelper.logName,
			"dispatchMessage",
			"Dispatching MessageBegin " + mref,
			messageBegin
		);
		
		dispatch( messageBegin );

		messageDispatched = true;
	}

	/**
	 * <P>Dispatch message end event.</P>
	 */
	private void dispatchMessageEnd () {
		// Construct MREF tag
		DecimalFormat mrefFormat = new DecimalFormat( "000000" );
		String mref = "MREF" + mrefFormat.format( mrefNumber );
		mrefNumber ++;

		final MessageEnd messageEnd = new MessageEnd( mref );
		componentHelper.logger.log(
			Level.FINE,
			componentHelper.logName,
		   "run()",
			"Dispatching MessageEnd",
			messageEnd
		);
		
	   dispatch( messageEnd );
		messageDispatched = false;
	}

	/**
	 * <P>Dispatchs time series event for updating for the time
	 * series key.</P>
	 *
	 * @param timeSeriesKey time series key
	 *
	 * @return successful update or not
	 */
	private boolean dispatchTimeSeriesUpdate (
	    String timeSeriesKey
	) {
		componentHelper.logger.log(
			Level.FINEST,
   		componentHelper.logName,
			"dispatchTimeSeriesUpdate()",
			"Dispatching time series: " + timeSeriesKey
		);

		try {
			// Construct time series object name
			String timeSeriesName
				= internalKeyFamilyName + KEY_SEPARATOR + timeSeriesKey;

			// Get data from database
			TiqView sourceValues
				= sourceDatabase.getObject( timeSeriesName );
	
			if ( sourceValues == null ) {
				componentHelper.logger.log(
					Level.FINE,
		   		componentHelper.logName,
					"dispatchTimeSeriesUpdate()",
					"Time series \"" + timeSeriesName
					+ "\" is not in database \""
					+ sourceDatabase.toString() + "\"."
				);

				return true;
			}
		
			TiqView sourceObsStatus
				= sourceDatabase.getObject(
						timeSeriesName + KEY_SEPARATOR + OBS_STATUS
					);

			TiqView sourceObsConf
				= getObsConfSeries(
					sourceDatabase,
					timeSeriesName,
					sourceValues
		  		);

			TiqView sourcePreBreak
				= getPreBreakSeries(
						sourceDatabase,
						timeSeriesName,
						sourceValues
					);

			 // Read time series from database.
			if (
				! validateTimeSeries(
					timeSeriesName, sourceDatabase, sourceValues,
					sourceObsStatus, sourceObsConf, sourcePreBreak, true
				)
			) {
				return false;
			}

			TiqView targetValues = null;
			TiqView targetObsStatus = null;
			TiqView targetObsConf = null;
			TiqView targetPreBreak = null;

			// Get target objects
			if ( targetDatabase.isOpen() ) {
				targetValues
					= targetDatabase.getObject(
						timeSeriesName
					);

				
				if ( targetValues != null ) {
					targetObsStatus
						= targetDatabase.getObject(
							timeSeriesName + KEY_SEPARATOR + OBS_STATUS
						);
	
					targetObsConf = getObsConfSeries(
												targetDatabase,
												timeSeriesName,
												targetValues
											);

					targetPreBreak = getPreBreakSeries(
												targetDatabase,
												timeSeriesName,
												targetValues
											);
				}

				if (
					! validateTimeSeries(
			    		timeSeriesName, targetDatabase, targetValues,
		   	 		targetObsStatus, targetObsConf, targetPreBreak, false
					)
				) {
					return false;
				}

		   	if ( targetValues != null ) {
		      	// Make sure source and target object
					// frequencies match
					if (
						! targetValues.getCalendar().getFrequency().equals(
							sourceValues.getCalendar().getFrequency()
						)
					) {
						logAndDispatchSevereError(
							"Time series \"" + timeSeriesName
								+ "\" in database \"" + sourceDatabase.toString()
								+ "\" has different frequency to the one in \""
								+ targetDatabase.toString() +"\".",
							"dispatchTimeSeriesUpdate()"
						);
						return false;
					}
				}
				
			}

			// Find observations to delete first
			findTimeSeriesObsDelete(
				timeSeriesKey, sourceValues, sourceObsStatus
			);
			
			int decimals = getDecimals( timeSeriesKey, sourceDatabase, attributeDatabase );
			if (decimals < 0 || MAX_DECIMALS < decimals) {
				logAndDispatchSevereError(
					"Invalid decimals value " + decimals 
					+ " (max allowed: " + MAX_DECIMALS + ") for Time series \"" + timeSeriesName + "\""
					+ ". Please check the highest value of the CL_DECIMALS codelist of this dataset.",
					"dispatchTimeSeriesUpdate()"
				);
				return false;			
			} else {
				componentHelper.logger.log(
					Level.FINER,
					componentHelper.logName,
					"dispatchTimeSeriesUpdate()",
					"Time series " + timeSeriesKey + " uses " + decimals + " decimals."
				);
			}

			Which interestingObs = null;
			if ( targetValues == null ) {
				interestingObs = getInterestingObservations( sourceObsStatus );
			} else {
				interestingObs = getInterestingObservations( 
											sourceValues, sourceObsStatus, sourceObsConf, sourcePreBreak, 
											targetValues, targetObsStatus, targetObsConf, targetPreBreak
										);
			}
		
			if (! (interestingObs.getPassed() > 0) ) {
				componentHelper.logger.log(
					Level.FINEST,
			   	componentHelper.logName,
					"dispatchTimeSeries()",
					"Time series does not contain data. Not dispatched."
				);			
				return true;
			}

			IndexValueRange range = getNonMissingTimeSeriesRange( interestingObs, sourceValues );

			componentHelper.logger.log(
				Level.FINEST,
	   		componentHelper.logName,
				"dispatchTimeSeriesUpdate()",
				"Dispatching range: " + range
			);

			dispatchContinousRangeTimeSeries(
				timeSeriesKey,
				range,
				sourceValues,
				sourceObsStatus,
				sourceObsConf,
				sourcePreBreak,
				decimals
			);
		} catch ( FunctorConsistencyChkException e ) {
			logAndDispatchSevereError(
				"Error in dispatching time series \""
					+ timeSeriesKey + "\".",
				"dispatchTimeSeries()",
				e
			);
			
			return false;
		}

		return true;
	}

	/**
	 * <P>Validate the frequencies and types of the four
	 * FAME objects that makes up the time series data.</P>
	 *
	 * @param timeSeriesName	time series name
	 * @param fameDb	FAME database
	 * @param values	FAME object containing time series values
	 * @param obsStatus	FAME object containing time series
	 * observation status
	 * @param obsConf	FAME object containing time series
	 * observation confidentiality
	 * @param preBreak	FAME object containing time series
	 * pre-break values
	 * @param requiredFlag	indicates if all FAME objects must exist
	 * in the database
	 *
	 * @return all objects are valid or not
	 */
	private boolean validateTimeSeries (
	    String timeSeriesName,
	    FAMEDb fameDb,
	    TiqView values,
	    TiqView obsStatus,
	    TiqView obsConf,
	    TiqView preBreak,
	    boolean requiredFlag
	) {
		if (
			! requiredFlag && values == null && obsStatus == null
			&&
			obsConf == null && preBreak == null
		) {
			// All object can be null if time series
			// is not compulsary
			return true;
		}

		String freqDim = getFrequencyDimension( timeSeriesName, keyFamily.getFrequencyDimensionPosition() );
		if ( freqDim == null ) {
			logAndDispatchSevereError(
				"Could not find the frequency dimension to time series \""
				+ timeSeriesName + "\".",
				"validateTimeSeries()"
			);
			return false;
		}

		if ( values == null ) {
			logAndDispatchSevereError(
				"Time series \"" + timeSeriesName
				+ "\" is not in database \""
				+ fameDb.toString() + "\".",
				"validateTimeSeries()"
			);
			return false;
		}

		try {
			if (! 
				validateTiqView( 
					timeSeriesName,
					fameDb.toString(), 
					values, 
					freqDim, 
					true, 
					true 
				) 
			) {
				return false;
			}

			if (! 
				validateTiqView( 
					timeSeriesName + KEY_SEPARATOR + OBS_STATUS,
					fameDb.toString(), 
					obsStatus, 
					freqDim, 
					true, 
					false 
				) 
			) {
				return false;
			}

			if (! 
				validateTiqView( 
					timeSeriesName + KEY_SEPARATOR + OBS_CONF,
					fameDb.toString(), 
					obsConf, 
					freqDim, 
					false, 
					false 
				) 
			) {
				return false;
			}
	    			
			if (! 
				validateTiqView( 
					timeSeriesName + KEY_SEPARATOR + BREAK,
					fameDb.toString(), 
					preBreak, 
					freqDim, 
					false, 
					true 
				) 
			) {
				return false;
			}
		} catch ( FunctorConsistencyChkException e ) {
			logAndDispatchSevereError(
				"Problem: " + e.getMessage() + ".",
				"validateTimeSeries()",
				e
			);
			return false;
		}
		
		return true;
	}

	/**
	 * <P>Finds the observations in a time series that
	 * needs to be deleted.</P>
	 *
	 * @param timeSeriesKey time series key
	 * @param values	FAME object containing time series values
	 * @param obsStatus	FAME object containing time series
	 * observation status
	 *
	 * @throws RangeTooLargeChkException if the data range for
	 * retrieving time series data is too large.
	 * @throws FunctorConsistencyChkException if an error occurs
	 * retrieving data from the FAME objects.
	 */
	private void findTimeSeriesObsDelete(
	    final String timeSeriesKey,
	    final TiqView values,
	    final TiqView obsStatus
	)
		throws RangeTooLargeChkException, FunctorConsistencyChkException
	{
		// Find observations to delete first, OBS_STATUS = NC
		EQ deletion = new EQ( obsStatus,	new Scalar(	obsStatus.getTiqType(), TiqConstants.STATUS_NC ) );
		Which whichNCObs = new Which( deletion );

		// Check for any observation deletions
		if ( whichNCObs.getPassed() == 0 ) {
			return;
		}
		
		// Construct frequency and range where values are NC
		ObservationList ncObsList = whichNCObs.getObservations();
		IndexValueRange range
			= new IndexValueRange(
				values.getCalendar(),
				((com.fame.timeiq.data.Observation)ncObsList.get( 0 )).getDateValue(),
				((com.fame.timeiq.data.Observation)ncObsList.get(
					ncObsList.size() - 1
				)).getDateValue()
			);

		// Test for contiguous range
		if ( whichNCObs.getPassed() == range.size() ) {
			addTimeSeriesObsDelete( timeSeriesKey, range );
		} else {
			// Get array of date values
			TiqDate[] ncDates = ncObsList.getValues().getDateArray();

			// Non-contiguous range of NC
			for ( int i = 0; i < ncDates.length; i++ ) {
				// Work out GESMES/TS period and format for each observation
				range = new IndexValueRange(
					values.getCalendar(), ncDates[ i ], ncDates[ i ]
				);
				addTimeSeriesObsDelete( timeSeriesKey, range );
			}
		}		
    }
	 
	 
	 protected void addTimeSeriesObsDelete ( 
	 	final String timeSeriesKey,
	 	final IndexValueRange range
	 ) {
		// Add time series to observation time series delete list
		timeSeriesObsDeleteList.add(
			new TimeSeries(
				timeSeriesKey, 
				IntervalConverter.indexValueRangeToRegularInterval(range),
				new ArrayList()
			)
		);
	 }
	 

	/**
	 * <P>Creates a GESMES/TS observation object
	 * {@link Int.ecb.gesmescb.Observation}
	 * from the FAME data.</P>
	 *
	 * @param value time series value
	 * @param valueStatus	time series value status
	 * @param status	time series observation status value
	 * @param statusStatus	status of the time series observation
	 * status value
	 * @param confidentiality	time series observation
	 * confidentiality value
	 * @param confientialityStatus	time series observation
	 * confidentiality status value
	 * @param preBreakValue time series pre-break value
	 * @param preBreakValueStatus	time series pre-break status
	 * value
	 *
	 * @return GESMES/TS observation object
	 * {@link Int.ecb.gesmescb.Observation}
	 */
	private Int.ecb.gesmescb.Observation getObservation (
	    final String value, final byte valueStatus,
	    final String status, final byte statusStatus,
	    final String confidentiality, final byte confidentialityStatus,
	    final String preBreakValue, final byte preBreakValueStatus
	) {
		String obsValue = "";
		String obsStatus = "";
		String obsConf = "";
		String obsPreBreak = "";

		if ( statusStatus != TiqConstants.STATUS_OK ) {
			return Int.ecb.gesmescb.Observation.EMPTY_OBSERVATION;
		}

		obsStatus = status;

		if ( valueStatus == TiqConstants.STATUS_OK ) {
			obsValue = value;
		} else if ( valueStatus == TiqConstants.STATUS_NC ) {
			obsValue = Observation.MISSING_VALUE;
		}

		if ( confidentialityStatus == TiqConstants.STATUS_OK ) {
			obsConf = confidentiality;
		}

		if ( preBreakValueStatus == TiqConstants.STATUS_OK ) {
			obsPreBreak = preBreakValue;
		} else if ( preBreakValueStatus == TiqConstants.STATUS_NC ) {
			obsPreBreak = Observation.MISSING_VALUE;
		}

		return new Int.ecb.gesmescb.Observation( obsValue,	obsStatus, obsConf, obsPreBreak );
    }


	/**
	 * <P>Dispatchs time series events for deleting all
	 * time series data sets in the list of time series
	 * keys.</P>
	 *
	 * @param deleteTimeSeriesKeyList	List of time
	 * series keys to delete
	 */
	private void dispatchTimeSeriesDeletes (
		final TimeSeriesKeyList deleteTimeSeriesKeyList
	) {
		Iterator iterator = deleteTimeSeriesKeyList.iterator();
		while ( iterator.hasNext() ) {
			final TimeSeries timeSeries = new TimeSeries( 
				( String )( iterator.next() ), 
				"", 
				"", 
				new ArrayList()
			);
				
			componentHelper.logger.log(
		  		Level.FINER,
		  		componentHelper.logName,
		  		"dispatchTimeSeriesDeletes()",
		  		"Dispatching Delete TimeSeries",
		  		timeSeries
			);

			if (! messageDispatched ) {
				dispatchMessageBegin( MessageBegin.DELETE_ACTION );
			}

			dispatch( timeSeries );
   	}
	}


	/**
	 * <P>Dispatchs time series events for deleting
	 * the list of time series observations
	 * {@link Int.ecb.gesmescb.TimeSeries}.</P>
	 *
	 * @param timeSeriesObsDeleteList	List of time
	 * series observations
	 * {@link Int.ecb.gesmescb.TimeSeries} to delete
	 */
	private void dispatchTimeSeriesObsDeletes () {
		// Delete time series observations
		Iterator iterator = timeSeriesObsDeleteList.iterator();
		while ( iterator.hasNext() ) {
			final TimeSeries timeSeries = ( TimeSeries ) iterator.next();
			componentHelper.logger.log(
				Level.FINER,
				componentHelper.logName,
				"dispatchTimeSeriesObsDeletes()",
				"Dispatching Time Series Observations Delete",
				timeSeries
			);

			if (! messageDispatched ) {
				dispatchMessageBegin( MessageBegin.DELETE_ACTION );
			}
			
			dispatch( timeSeries );
		}
	}


	/**
	 * <P>Retrieves decimals attribute from FAME database
	 * for a time series key.</P>
	 *
	 * @param timeSeriesKey time series key
	 * @param fameDb	FAME database
     *
     * @return decimal attribute value or -1 if not found
	 */
	private int getDecimalsAttribute (
		final String timeSeriesKey, 
		final FAMEDb fameDb
	) {
		if ( ! fameDb.isOpen() ) {
	   	return -1;
		}

		// Get the key family component
		KeyFamilyComponent attributeComponent = keyFamily.getComponent( "DECIMALS" );
		
		// Key family does not contain DECIMALS
		if (attributeComponent == null) {
			return -1;
		}

		// Construct object name
		String objectName = getAttributeName(
			internalKeyFamilyName,
			timeSeriesKey,
			keyFamily.getFrequencyDimensionPosition(),
			attributeComponent
		);

		// Read decimals attribute object from database.
		TiqObject attributeObject = fameDb.getObject( objectName );

		// Return if attribute object
		// does not exist in source database
		if ( attributeObject == null ) {
			return -1;
		}
		
		if (! validateAttribute( attributeObject, attributeComponent ) ) {
			return -1;
		}
		
		if ( attributeObject.getTiqClass() != TiqClass.SCALAR ) {
			// Observation level
			// Too complicated to deal with and decimals attribute
			// should never be observation level
			componentHelper.logger.log(
				Level.WARNING,
				componentHelper.logName,
				"getDecimalsAttribute()",
				"Decimals attribute object \"" + objectName
		    	+ "\" is attached to observation level. This is not supported."
	   	);
			return -1;
		}

		// Check attribute value is OK
		if ( attributeObject.getTiqValue().getStatus() != TiqConstants.STATUS_OK ) {
			return -1;
		}

		return Integer.parseInt(attributeObject.getTiqValue().getStringValue());
	}


	/**
	 * <P>Dispatchs attribute events for updating attributes
	 * for a list of time series keys.</P>
	 *
	 * @param timeSeriesKey identifies series for which relevant attributes should be sent.
	 *
	 * @return successful update or not
	 */
	private boolean dispatchAttributeUpdates (
		final String timeSeriesKey
	) {
		for (
			final Iterator keyFamilyComponentIterator
				= keyFamily.attributeIterator(
					KeyFamilyComponent.ATTACHED_TO_DATA_SET
				);
			keyFamilyComponentIterator.hasNext();
		) {
			final KeyFamilyComponent component
				= ( KeyFamilyComponent ) keyFamilyComponentIterator.next();
			if ( ! dispatchAttributeUpdate( "", component ) ) {
				return false;
			}
		}

		for (
			final Iterator keyFamilyComponentIterator
				= keyFamily.attributeIterator(
					KeyFamilyComponent.ATTACHED_TO_SIBLING_GROUP
				);
			keyFamilyComponentIterator.hasNext();
		) {
			final KeyFamilyComponent component
				= ( KeyFamilyComponent ) keyFamilyComponentIterator.next();
			final String attributeKey = getSiblingKey( timeSeriesKey, keyFamily.getFrequencyDimensionPosition() );
			if ( ! dispatchAttributeUpdate( attributeKey, component ) ) {
				return false;
			}
		}

		for (
			final Iterator keyFamilyComponentIterator
				= keyFamily.attributeIterator(
					KeyFamilyComponent.ATTACHED_TO_TIME_SERIES
				);
			keyFamilyComponentIterator.hasNext();
		) {
			final KeyFamilyComponent component
				= ( KeyFamilyComponent ) keyFamilyComponentIterator.next();
			if ( ! dispatchAttributeUpdate( timeSeriesKey, component ) ) {
				return false;
			}
		}

		for (
			final Iterator keyFamilyComponentIterator
				= keyFamily.attributeIterator(
					KeyFamilyComponent.ATTACHED_TO_OBSERVATION
				);
			keyFamilyComponentIterator.hasNext();
		) {
			final KeyFamilyComponent component
				= ( KeyFamilyComponent ) keyFamilyComponentIterator.next();
			if ( ! dispatchAttributeUpdate( timeSeriesKey, component ) ) {
				return false;
			}
		}

		return true;
    }


	/**
	 * <P>Dispatch attribute event for updating an attribute.</P>
	 *
	 * @param attributeKey attribute name
	 * @param attributeComponent	Metabase values for the attribute
	 *
	 * @return successful update (if necessary) or not
	 */
	private boolean dispatchAttributeUpdate (
	    final String attributeKey, 
		 final KeyFamilyComponent attributeComponent
	) {
		// Dispatch attribute if not already done so
		if ( ! dispatchedAttributeKeyList.add(
			attributeKey + KEY_SEPARATOR + attributeComponent.getConceptId()
		) ) {
			return true;
		}
				
		// Construct name of attribute object
		String objectName = internalKeyFamilyName;
		// Ioana 10.01.2014 - for attributes at dataset level
		if (attributeKey!= null && !attributeKey.isEmpty() 
				&& (! attributeKey.substring(0, 1).equals(KEY_SEPARATOR)) ) {
			objectName += KEY_SEPARATOR;
		}
		objectName += attributeKey + KEY_SEPARATOR + attributeComponent.getConceptId();

		// Read attribute object from database.
		TiqObject attributeObject = sourceDatabase.getObject( objectName );
		// Return if attribute object does not
		// exist in source database
		if ( attributeObject == null ) {
			return true;
		}

		// Get target attribute object from database
		TiqObject targetAttributeObject = null;
		if ( targetDatabase.isOpen() ) {
			targetAttributeObject = targetDatabase.getObject( objectName );
		}

		if (! validateAttribute( attributeObject, attributeComponent ) ) {
			return false;
		}

		if (targetAttributeObject != null && ! validateAttribute( attributeObject, attributeComponent ) ) {
			return false;
		}
	
		// Key family, sibling or time series level
		if ( attributeObject.getTiqClass() == TiqClass.SCALAR ) {
			dispatchScalarAttributeUpdate( 
				attributeKey,
				attributeObject,
				targetAttributeObject,
				attributeComponent
			);
			return true;
		}
			
		return dispatchSeriesAttributeUpdate(
			attributeKey,
			attributeObject,
			targetAttributeObject,
			attributeComponent
		);
	}


	/**
	 * <P>Dispatch attribute events for deleting a list
	 * of {@link Int.ecb.gesmescb.Attribute}s.</P>
	 *
	 * @param attributeDeleteList	List of
	 * {@link Int.ecb.gesmescb.Attribute}s to delete
	 */
	private void dispatchAttributeDeletes () {
		Iterator iterator = attributeDeleteList.iterator();
		while ( iterator.hasNext() ) {
			final Attribute attribute = ( Attribute ) iterator.next();
			componentHelper.logger.log(
				Level.FINEST,
				componentHelper.logName,
				"dispatchAttributeDeletes()",
				"Dispatching Delete Attribute",
				attribute
			);
			
			if (! messageDispatched ) {
				dispatchMessageBegin( MessageBegin.DELETE_ACTION );
			}
			
			dispatch( attribute );
		}
	}


	/**
	 * <P>Dispatchs attribute events for deleting a list
	 * of {@link Int.ecb.gesmescb.Attribute} observations.</P>
	 *
	 * @param attributeObsDeleteList	List of
	 * {@link Int.ecb.gesmescb.Attribute} observations to delete
	 */
   private void dispatchAttributeObsDeletes () {
		Iterator iterator = attributeObsDeleteList.iterator();
		while ( iterator.hasNext() ) {
			final Attribute attribute = ( Attribute )( iterator.next() );
			componentHelper.logger.log(
				Level.FINEST,
				componentHelper.logName,
				"dispatchAttributeObsDeletes()",
				"Dispatching Attribute Observation Delete",
				attribute
			);

			if (! messageDispatched ) {
				dispatchMessageBegin( MessageBegin.DELETE_ACTION );
			}
			
			dispatch( attribute );
		}
	}


	private TiqView getObsConfSeries(
		final FAMEDb database,
		final String timeSeriesName,
		final TiqView observationSeries
	) 
		throws FunctorConsistencyChkException 
	{
		TiqView obsConf
			= database.getObject(
				timeSeriesName + KEY_SEPARATOR + OBS_CONF
			);

		if (obsConf == null) {
			obsConf = new RegularSeries(
									observationSeries.getCalendar(),
									TiqType.STRING,
									TiqObserved.UNDEFINED,
									null
							);
		}

		return obsConf;
	}
	

	private TiqView getPreBreakSeries(
		final FAMEDb database,
		final String timeSeriesName,
		final TiqView observationSeries
	) 
		throws FunctorConsistencyChkException 
	{
		TiqView preBreak
			= database.getObject(
				timeSeriesName + KEY_SEPARATOR + BREAK
			);

		if (preBreak == null) {
			preBreak = new RegularSeries(
								observationSeries.getCalendar(),
								TiqType.DOUBLE,
								observationSeries.getObserved(),
								observationSeries.getBasis()
							);
		}

		return preBreak;
	}	
	

	private void dispatchContinousRangeTimeSeries(
		final String timeSeriesKey,
		final IndexValueRange range,
		final TiqView sourceValues,
		final TiqView sourceObsStatus,
		final TiqView sourceObsConf,
		final TiqView sourcePreBreak,
		final int decimals
	) 
		throws FunctorConsistencyChkException 
	{	
		ObservationList valList = sourceValues.getObservations(
			range.getFirstIndex(), range.getLastIndex()
		);
		String[] valArray = arrayToStringArray(valList.getValues().getDoubleArray(), decimals);
		byte[] valStatArray = valList.getStatuses();

		ObservationList statList
		    = sourceObsStatus.getObservations(
			range.getFirstIndex(), range.getLastIndex()
		);
		String[] statArray = statList.getValues().getStringArray();
		byte[] statStatArray = statList.getStatuses();

		ObservationList confList
			= sourceObsConf.getObservations(
			range.getFirstIndex(), range.getLastIndex()
	    );
		String[] confArray = confList.getValues().getStringArray();
		byte[] confStatArray = confList.getStatuses();

		ObservationList breakList
			= sourcePreBreak.getObservations(
			range.getFirstIndex(), range.getLastIndex()
		);
		String[] breakArray = arrayToStringArray(breakList.getValues().getDoubleArray(), decimals);
		byte[] breakStatArray = breakList.getStatuses();

		ArrayList observations = new ArrayList( range.size() );
		
		for ( int i = 0; i < range.size(); i++ ) {
			observations.add(
				getObservation (
					valArray[i], valStatArray[i],
					statArray[i], statStatArray[i],
					confArray[i], confStatArray[i],
					breakArray[i], breakStatArray[i]
				)
			);
		}

		dispatchTimeseries(
			new TimeSeries(
					timeSeriesKey, 
					IntervalConverter.indexValueRangeToRegularInterval(range), 
					observations
			)
		);
	}
	
	private void dispatchTimeseries(
		TimeSeries timeSeries
	) {
		if ( ! messageDispatched ) {
			dispatchMessageBegin(
				MessageBegin.UPDATE_OR_REPLACE_ACTION
			);
		}

		componentHelper.logger.log(
			Level.FINER,
			componentHelper.logName,
			"dispatchTimeSeries()",
			"Dispatching TimeSeries",
			timeSeries
		);

		dispatch( timeSeries );
	}
	
	
	private IndexValueRange getNonMissingTimeSeriesRange(
		final Which interestingObs,
		final TiqView sourceValues
	) 
		throws FunctorConsistencyChkException 
	{
		ObservationList whichObsList
			= interestingObs.getObservations();
		
		return 
			new IndexValueRange(
					sourceValues.getCalendar(),
					((com.fame.timeiq.data.Observation)whichObsList.get( 0 )).getDateValue(),
					((com.fame.timeiq.data.Observation)whichObsList.get( whichObsList.size() - 1 )).getDateValue()
			);				
	}

	
	private int getDecimals(
		final String timeSeriesKey,
		final FAMEDb sourceDatabase,
		final FAMEDb attributeDatabase
	) {
		int sourceDecimals
			= getDecimalsAttribute( timeSeriesKey, sourceDatabase );

		if ( sourceDecimals == -1 ) {			
			sourceDecimals = getDecimalsAttribute( timeSeriesKey, attributeDatabase );
		}

		if ( sourceDecimals == -1 ) {
			return DEFAULT_DECIMALS;
		}
				
		return sourceDecimals;
	}


	private Which getInterestingObservations(
		final TiqView sourceObsStatus
	) throws FunctorConsistencyChkException {						
		Scalar deletion = new Scalar( sourceObsStatus.getTiqType(), TiqConstants.STATUS_NC );
		Scalar nd = new Scalar( sourceObsStatus.getTiqType(), TiqConstants.STATUS_ND );
		
		NE notDeletion = new NE( sourceObsStatus, deletion );
		NE notND = new NE( sourceObsStatus, nd );
		
		return new Which(	new And( notDeletion, notND ) );
	}


	private Which getInterestingObservations(
		final TiqView sourceValues,
		final TiqView sourceObsStatus,
		final TiqView sourceObsConf,
		final TiqView sourcePreBreak,
		final TiqView targetValues,
		final TiqView targetObsStatus,
		final TiqView targetObsConf,
		final TiqView targetPreBreak
	) 
		throws FunctorConsistencyChkException 
	{	
		Scalar deletion = new Scalar(	sourceObsStatus.getTiqType(), TiqConstants.STATUS_NC );
		Scalar nd = new Scalar( sourceObsStatus.getTiqType(), TiqConstants.STATUS_ND );
		
		NE notDeletion = new NE( sourceObsStatus, deletion );
		NE notND = new NE( sourceObsStatus, nd );

		And notDeletionOrND = new And( notDeletion, notND );

		NE differentValues = new NE( sourceValues, targetValues );
		NE differentObsStatus = new NE( sourceObsStatus, targetObsStatus );
		NE differentObsConf = new NE( sourceObsConf, targetObsConf );
		NE differentPreBreak = new NE( sourcePreBreak, targetPreBreak );
		Or different = new Or( new Or( differentValues, differentObsStatus ), new Or( differentObsConf, differentPreBreak ) );
																				
		return new Which( new And(	notDeletionOrND, different ) );
	}

	protected String getAttributeName(
		final String internalKeyFamilyName,
		final String timeSeriesKey,
		final int frequencyDimensionPosition,
		final KeyFamilyComponent attributeComponent
	) {
		if ( attributeComponent.isAttachedToDataSet() ) {
			return 	internalKeyFamilyName 
						+ KEY_SEPARATOR + attributeComponent.getConceptId();
		}
		
		if ( attributeComponent.isAttachedToSiblingGroup() ) {
			final String siblingKey = getSiblingKey( timeSeriesKey, frequencyDimensionPosition );
			return 	internalKeyFamilyName 
						+ ( 1 < frequencyDimensionPosition ? KEY_SEPARATOR : "" ) + siblingKey 
						+ KEY_SEPARATOR + attributeComponent.getConceptId();
		}
		
		return 	internalKeyFamilyName 
					+ KEY_SEPARATOR + timeSeriesKey 
					+ KEY_SEPARATOR + attributeComponent.getConceptId();
	}


	protected boolean openDatabase (
		final String preferenceName,
		final boolean mandatory,
		final FAMEDb database
	) {
		String filename = getPreference( preferenceName, mandatory, "" );			
		String servername = getPreference( preferenceName + ".server", false, "" );			

		if ( filename.equals( "" ) ) {
			return !mandatory;
		}
		
		componentHelper.logger.log(
			Level.INFO,
			componentHelper.logName,
			"openDatabase()",
			"Opening " + preferenceName + " Fame database in read mode: " + filename
		);

		Properties properties = new Properties();
		properties.put("mode", "READ_ONLY");

		return database.open( servername, filename, properties );
	}


	protected void logAndDispatchSevereError (
		final String msg,
		final String procedureName
	) {
		componentHelper.logger.log(
			Level.SEVERE,
			componentHelper.logName,
			procedureName,
			msg
		);
		
		dispatch(
			new ErrorEvent( msg, ErrorEvent.ERROR )
		);
	}

	protected void logAndDispatchSevereError (
		final String msg,
		final String procedureName,
		final Exception e
	) {
		componentHelper.logger.log(
			Level.SEVERE,
			componentHelper.logName,
			procedureName,
			msg,
			e
		);
		
		dispatch(
			new ErrorEvent( msg, ErrorEvent.ERROR )
		);
	}

	
	protected boolean validateTiqView (
		final String timeSeriesName,
		final String fameDbName,
		final TiqView view,
		final String freqDim,
		final boolean mandatory,
		final boolean numeric
	) 
		throws FunctorConsistencyChkException
	{	
		if ( view == null ) {
			if ( mandatory ) {
				logAndDispatchSevereError(
					"Time series \"" + timeSeriesName
					+ "\" is not in database \""
					+ fameDbName + "\".",
					"validateTiqView()"
				);
				return false;
			} else {
				return true;
			}
		}

		if ( ! view.isRegular() ) {
			logAndDispatchSevereError(
				"Time series \"" + timeSeriesName
	    		+ "\" in database \"" + fameDbName
	    		+ "\" is not a regular series.",
				"validateTiqView()"
			);
			return false;
		}
		
		if ( ! validateFrequency ( freqDim, view.getCalendar() ) ) {
			logAndDispatchSevereError(
				"Time series \"" + timeSeriesName
	    		+ "\" in database \"" + fameDbName
	    		+ "\" has unexpected frequency \""
	    		+ view.getCalendar().toString() + "\".",
				"validateTiqView()"
			);
			return false;
    	}

		if (numeric) {
	    	if ( ! view.getTiqType().isNumeric() ) {
				logAndDispatchSevereError(
					"Time series \"" + timeSeriesName
					+ "\" in database \"" + fameDbName
					+ "\" is not a numeric type.",
					"validateTiqView()"
				);
				return false;
			}
		} else {
			if ( ! view.getTiqType().equals( TiqType.STRING ) ) {
				logAndDispatchSevereError(
					"Time series \"" + timeSeriesName
					+ "\" in database \"" + fameDbName 
					+ "\" is not a string type.",
					"validateTiqView()"
				);
				return false;
			}
		}

		return true;
	}

	protected boolean validateAttribute (
		final TiqObject attributeObject,
		final KeyFamilyComponent attributeComponent
	) {
		if (
			attributeComponent.isAttachedToDataSet()
			||
			attributeComponent.isAttachedToSiblingGroup()
			||
			attributeComponent.isAttachedToTimeSeries()
		) {
			// Check source attribute
			if ( attributeObject.getTiqClass() != TiqClass.SCALAR ) {
				logAndDispatchSevereError(
					"Attribute object \"" + attributeObject.getName()
			    	+ "\" is not SCALAR class to match attachment level \""
			    	+ attributeComponent.getAttachmentLevel() + "\".",
					"dispatchAttribute()"
				);
				return false;
		   }
		// Time series for Observation level
		} else if ( attributeComponent.isAttachedToObservation() ) {
			// Check source attribute
			if ( attributeObject.getTiqClass() != TiqClass.REGULAR_SERIES ) {
				logAndDispatchSevereError(
					"Source attribute object \"" + attributeObject.getName()
					+ "\" is not a Regular Series class to match attachment level \""
					+ attributeComponent.getAttachmentLevel() + "\".",
					"dispatchAttribute()"
				);
				return false;
			}
		// Unknown attachment level
		} else {
			logAndDispatchSevereError(
				"Attribute object \"" + attributeObject.getName()
				+ "\" has unknown attachment level \""
				+ attributeComponent.getAttachmentLevel() + "\".",
				"dispatchAttribute()"
			);
			return false;
		}
		
		return true;
	}

	protected void dispatchScalarAttributeUpdate (
		final String attributeName,
		final TiqObject attributeObject,
		final TiqObject targetAttributeObject,
		final KeyFamilyComponent attributeComponent
	) {
		// Check attribute value is OK
		if (
			attributeObject.getTiqValue().getStatus() == TiqConstants.STATUS_NC
		) {
			// Add to deletion list
			attributeDeleteList.add(
				new Attribute(
					attributeName, "", "", 
					attributeComponent.getConceptId(),
					""
				)
			);
			return;
		}	
		
		/** trainot 04.07 Ignore if attribute is null, empty or contains only spaces !*/
		if (attributeObject.getTiqValue().getStringValue() == null) {
			return;
		} else {
			if (attributeObject.getTiqValue().getStringValue().replaceAll("\\s","").isEmpty()) {
				return;
			}
		}
		
		// Dispatch attribribute if value not in target database
		// or source and database values are different
		if (
			targetAttributeObject == null
			|| (
				targetAttributeObject.getTiqValue().getStatus() != TiqConstants.STATUS_NC
				&&
				! attributeObject.getTiqValue().getStringValue().equals(
					targetAttributeObject.getTiqValue().getStringValue()
				)
			)
		) {
			// Dispatch message begin event if required
			if ( ! messageDispatched ) {
				dispatchMessageBegin(
					MessageBegin.UPDATE_OR_REPLACE_ACTION
				);
			}

			Attribute attribute = new Attribute(
					attributeName, "", "", 
					attributeComponent.getConceptId(),
					attributeObject.getTiqValue().getStringValue()
			);
			
			dispatch( attribute );

			componentHelper.logger.log(
				Level.FINEST,
				componentHelper.logName,
				"dispatchAttribute()",
				"Dispatching Attribute",
				attribute
			);
		}
	}

	protected boolean dispatchSeriesAttributeUpdate (
		final String attributeName,
		final TiqObject attributeObject,
		final TiqObject targetAttributeObject,
		final KeyFamilyComponent attributeComponent
	) {
		NE notMissing = new NE(	
								attributeObject, 
								new Scalar(
									attributeObject.getTiqType(),
									TiqConstants.STATUS_ND
								)
							);
		NE notEqual = new NE( attributeObject, targetAttributeObject );

		Which interestingObs;
		if ( targetAttributeObject == null ) {
			// Ignore source values equal ND
			interestingObs = new Which( notMissing );
		} else {
			// Ignore source values equal ND and source and target values which are the same
			interestingObs = new Which( new And( notMissing, notEqual ) );
		}
		
		// Observation date range
		try {
			if ( interestingObs.getPassed() == 0 ) {
				return true;
			}
			
			TiqDate[] dateIndexes
				= interestingObs.getObservations().getValues().getDateArray();
			IndexBy indexBy = new IndexBy( attributeObject, interestingObs	);
			ObservationList obsList = indexBy.getObservations();
			String[] values = obsList.getValues().getStringArray();
			byte[] statuses = obsList.getStatuses();
			
			for ( int i = 0; i < dateIndexes.length; i++ ) {
				IndexValueRange range = new IndexValueRange(
					attributeObject.getCalendar(),
					dateIndexes[ i ],
					dateIndexes[ i ]
				);

				if (
					statuses[ i ] == TiqConstants.STATUS_ND
					||
					statuses[ i ] == TiqConstants.STATUS_NC
				) {
					values[ i ] = "";
				}
						
				Attribute attribute = new Attribute(
					attributeName, 
					IntervalConverter.indexValueRangeToRegularInterval( range ),
					attributeComponent.getConceptId(),
					values[ i ]
				);

				if (
					statuses[ i ] != TiqConstants.STATUS_NC
				) {
					/** trainot 04.07 Ignore if attribute is null, empty or contains only spaces !*/
					if (values[i] == null) {
						continue;
					} else {
						if (values[i].replaceAll("\\s","").isEmpty()) {
							continue;
						}
					}
					// Dispatch message begin event if required
					if ( ! messageDispatched ) {
						dispatchMessageBegin(
							MessageBegin.UPDATE_OR_REPLACE_ACTION
						);
					}
					
					dispatch( attribute );
					componentHelper.logger.log(
						Level.FINEST,
						componentHelper.logName,
						"dispatchAttribute()",
						"Dispatching Attribute",
						attribute
					);
				} else {
					// Add time series to deletion list
					attributeObsDeleteList.add( attribute );
				}
			}
	  	} catch ( RangeTooLargeChkException e ) {
			logAndDispatchSevereError(
				"Error in IndexValueRange too large.",
				"dispatchAttribute()",
				e
			);
			return false;
    	} catch ( FunctorConsistencyChkException e ) {
			logAndDispatchSevereError(
				"Error in WHICH statement to find NC observations.",
				"dispatchAttribute()",
				e
			);
			return false;
	   }
		
		return true;
	}
	

	
	
	public String[] arrayToStringArray ( 
		final double[] values, 
		final int decimals
	) {
		String[] strings = new String[values.length];

		if (ignoreDecimalsAttributes) {
			for (int i = 0; i < values.length; i++) {		
				strings[i] = Double.toString((values[i]));
			}
			return strings;
		}
		try {
			for (int i = 0; i < values.length; i++) {		
				strings[i] = DecimalsTrimmer.format(
					values[i], decimals,
					useHalfUpRounding
						? DecimalsTrimmer.HALF_UP_MODE
						: DecimalsTrimmer.HALF_EVEN_MODE
				);
			}
		} catch ( Exception e) {
			logAndDispatchSevereError(
				"FAMEReader error: \"" + e.getMessage()	+ "\".", "arrayToStringArray()"
			);
		}
		return strings;
	}


	/**
	 * <P>Retrieves a preference value.</P>
	 *
	 * @param key	key for the preference
	 * @param manditoryFlag indicates whether a preference value is
	 * manditory for the this component.
	 *
	 * @return preference value for the specified key
	 *
	 * @throws NullPointerException if the preference variable
	 * for the key is undefined and is manditory.
	 */
	protected String getPreference ( 
		final String key, 
		final boolean mandatory,
		final String defaultValue
	)
		throws NullPointerException 
	{
		String preference = componentHelper.getPreference( key, defaultValue );

		if ( preference.equals( "" ) ) {
			if ( mandatory ) {
				final String desc
					= "No value for preference variable \""
					+ key + "\" found.";

				componentHelper.logger.log(
					Level.SEVERE,
					componentHelper.logName,
				   "getPreference()",
					desc
				);

				throw new NullPointerException( desc );
			}

			final String desc
					= "No value for preference variable \""
						+ key + "\" found.";

			componentHelper.logger.log(
				Level.FINER,
				componentHelper.logName,
				"getPreference()",
				desc
			);
			
			return preference;
		}

		final String desc = 
			"Found preference variable \"" + key + "\" with value \""
		    + preference + "\".";

		componentHelper.logger.log(
			Level.FINEST,
			componentHelper.logName,
			"getPreference()",
			desc
		);

		return preference;
	}

}
