// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.fame;

import java.util.List;
import java.util.Iterator;
import java.util.ArrayList;

import com.fame.timeiq.*;
import com.fame.timeiq.data.ObservationList;
import com.fame.timeiq.dates.*;

import Int.ecb.logging.Level;

import Int.ecb.gesmescb.*;
import Int.ecb.gesmescb.time.*;
import Int.ecb.gesmescb.metabase.*;
import Int.ecb.gesmescb.fame.FAMEDb;
import Int.ecb.gesmescb.util.ComponentHelper;

/* **************************** Class ***************************** */

/**
 * Abstract base class for Revisors.
 *
 * @author
 * <A HREF="mailto:klo@FAME.com">Kam Lo</A>
 * @author
 * <A HREF="mailto:Sami.Airo@ECB.int">Sami Airo</A>
 *
 * @version @(#)$Id: FAMERevisor.java,v 1.1 2017/01/19 16:14:03 trainot Exp $
 */

abstract class FAMERevisor {

	/* ======================= Constants ======================== */

	/** Separator character used in Fame object names.	 */
	public static char KEY_SEPARATOR = '.';

	/** Name for frequency code list used by the sibling deletion. */
	private static String FREQUENCY_LIST = "CL_FREQ"; //TODO should use a constant from the Metabase

	/* ========================= Fields ========================= */

	/* Takes care of logging and prefs */
	private ComponentHelper componentHelper;

	/** Current FAMEDb for writing data. */
	private final FAMEDb fameDb;

	/** Internal dataset name */
	private String dataSetId;

	/** Current key family*/
	private KeyFamily keyFamily;

	/** CL_FREQUENCY code list */
	private CodeList frequencyCodeList;

	/* ====================== Constructors ====================== */

	public FAMERevisor(
		final ComponentHelper componentHelper,
		final FAMEDb fameDb,
		final MessageBegin messageBegin
	) {
		this.componentHelper = componentHelper;
		this.fameDb = fameDb;
		this.dataSetId = messageBegin.getDataSetId();

		MaintenanceAgency agency = Metabase.getMetabase().getMaintenanceAgency(messageBegin.getMaintenanceAgencyId());

		if (agency == null) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"FAMERevisor",
				"Metabase does not contain agency with ID " + messageBegin.getMaintenanceAgencyId()
			);
		}

		this.frequencyCodeList = agency.getCodeList(FREQUENCY_LIST);
		this.keyFamily = agency.getKeyFamily( messageBegin.getKeyFamilyId() );
	}

	/* ===================== Public Methods ===================== */

	public abstract boolean handle ( TimeSeries timeSeries );

	public abstract boolean handle ( Attribute attribute );

	/**
	 * Returns the a fame range that conforms to the GESMES/TS range.
	 *
	 * @param interval to be converted
	 *
	 * @return Range that covers the GESMES/TS range
	 */
	public final IndexValueRange getFameRange (
		final RegularInterval interval
	) {
		IndexValueRange range = null;

		try {
			range = IntervalConverter.regularIntervalToIndexValueRange(interval);
		} catch (RangeTooLargeChkException e) {
			getComponentHelper().logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"getFameRange",
				"period range too large",
				e
			);
		} catch (DateOverflowChkException e) {
			getComponentHelper().logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"getFameRange",
				"period range too large",
				e
			);
		}

		return range;
	}

	/**
	 * Returns the Fame database name for the specified attribute.
	 *
	 * @param attribute	attribute the name is returned for
	 * @param component	key family component for the attribute
	 *
	 * @return name to use in a Fame database
	 */
	public final String getFameName (final Attribute attribute, final KeyFamilyComponent component) {
		String key = attribute.getKey().replace(Attribute.KEY_SEPARATOR, KEY_SEPARATOR);

		if (component.isAttachedToDataSet())
			return dataSetId + KEY_SEPARATOR + attribute.getName();

		if (component.isAttachedToSiblingGroup())
			return dataSetId + key + KEY_SEPARATOR + attribute.getName();

		return dataSetId + KEY_SEPARATOR + key + KEY_SEPARATOR + attribute.getName();
	}

	/**
	 * Returns the Fame database name for the specified series.
	 *
	 * @param series	series the name is returned for
	 *
	 * @return name to use in a Fame database
	 */
	public final String getFameName (final TimeSeries series) {
		String tmp = series.getKey().replace(TimeSeries.KEY_SEPARATOR, KEY_SEPARATOR);
		return dataSetId + ( tmp.length() > 0 ? KEY_SEPARATOR + tmp : "" );
	}

	/**
	 * Returns a list of observations for the specified series.
	 *
	 * @param name		Series name the observations are returned for
	 * @param range Range for the observations
	 *
	 * @return List of observations or null if the series was not in the database.
	 */
	public final ObservationList getObservations (String name, IndexValueRange range)
		throws TiqCheckedException, TiqUncheckedException
	{
		ObservationList list = null;

		RegularSeries series = (RegularSeries)fameDb.getObject(name);

		if (series != null) {
			list = series.getObservations(range.getFirstIndex(), range.getLastIndex(), null);
		}

		return list;
	}

	/**
	 * Returns a list of TimeSeries in a sibling set.
	 *
	 * @param siblingKey		key for the sibling set
	 *
	 * @return List of time series in the sibling set.
	 */
	public List getSiblingSet ( String siblingKey ) {
		List list = new ArrayList();
		Iterator it = frequencyCodeList.codeValueIterator();

		while (it.hasNext()) {
			list.add( new TimeSeries(
								(String)it.next() + siblingKey,
								"",
								"",
								new Int.ecb.gesmescb.Observation[0]
						 )
			);
		}

		return list;
	}

	public KeyFamily getKeyFamily () {
		return keyFamily;
	}

	public String getDataSetId() {
		return dataSetId;
	}
	
	public ComponentHelper getComponentHelper () {
		return componentHelper;
	}

	public FAMEDb getFAMEDb () {
		return fameDb;
	}

}
