// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.fame;

import java.util.Date;
import java.util.Calendar;
import java.util.GregorianCalendar;

import com.fame.timeiq.dates.*;
import Int.ecb.gesmescb.time.*;

/**
 * IntervalConverter is a helper class that converts TimeIQ IndexValueRanges and RegularIntervals
 *
 * @author <A HREF="mailto:Sami.Airo@ECB.int">Sami Airo</A>
 *
 * @version @(#)$Id: IntervalConverter.java,v 1.1 2017/01/19 16:14:03 trainot Exp $
 */
final class IntervalConverter {

	/* ===================== Constants ===================== */
	
	public static final int FIRST_DAY_OF_WEEK_CORRECTION = 3;

	/* ===================== Public Static Methods ===================== */
	
	public static int tiqFrequencyToRegularIntervalFrequency( final TiqFrequency frequency ) {
		switch ( frequency.getCode() ) {
			case TiqFrequency.ANNUAL_FREQUENCY:
				return RegularInterval.ANNUAL;
			case TiqFrequency.SEMIANNUAL_FREQUENCY:
				return RegularInterval.SEMIANNUAL;
			case TiqFrequency.QUARTERLY_FREQUENCY:
				return RegularInterval.QUARTERLY;
			case TiqFrequency.MONTHLY_FREQUENCY:
				return RegularInterval.MONTHLY;
			case TiqFrequency.WEEKLY_FREQUENCY:
				return RegularInterval.WEEKLY;
			case TiqFrequency.DAILY_FREQUENCY:
				return RegularInterval.DAILY;
			case TiqFrequency.MINUTELY_FREQUENCY:
				return RegularInterval.MINUTELY;
			default:
				throw new IllegalArgumentException(
					"Invalid TiqFrequency for conversion to RegularInterval frequency: " + frequency
				);
		}
	}

	public static TiqFrequency regularIntervalFrequencyToTiqFrequency( final int frequency ) {
		switch ( frequency ) {
			case RegularInterval.ANNUAL:
				return TiqFrequency.ANNUAL;
			case RegularInterval.SEMIANNUAL:
				return TiqFrequency.SEMIANNUAL;
			case RegularInterval.QUARTERLY:
				return TiqFrequency.QUARTERLY;
			case RegularInterval.MONTHLY:
				return TiqFrequency.MONTHLY;
			case RegularInterval.WEEKLY:
				return TiqFrequency.WEEKLY;
			case RegularInterval.DAILY:
				return TiqFrequency.DAILY;
			case RegularInterval.MINUTELY:
				return TiqFrequency.MINUTELY;
			default:
				throw new IllegalArgumentException(
					"Invalid RegularInterval frequency for conversion to TiqFrequency: " + frequency
				);
		}
	}

	public static IndexValueRange regularIntervalToIndexValueRange( 
		final RegularInterval interval 
	) 
		throws RangeTooLargeChkException, DateOverflowChkException
	{
		TiqFrequency frequency = regularIntervalFrequencyToTiqFrequency(interval.getFrequency());
		SimpleCalendar simpleCalendar = new SimpleCalendar(frequency);

		if (frequency.getCode() == TiqFrequency.WEEKLY_FREQUENCY) {
			simpleCalendar = 	new SimpleCalendar(TiqFrequency.WEEKLY, SimpleCalendar.REF_FRIDAY, 1);
		}

		long begin = javaDateToBeginTiqIndex(interval.getBegin(), interval.getFrequency());

		IndexValueRange range = new IndexValueRange(
				simpleCalendar,
				begin,
				begin 
		);
		
		range.setPeriodsFromStart(interval.getSize());
		return range;
	}

	public static RegularInterval indexValueRangeToRegularInterval( 
		final IndexValueRange range 
	) {
		return new GregorianRegularInterval(
			tiqFrequencyToRegularIntervalFrequency(range.getCalendar().getFrequency()),
			beginTiqIndexToJavaDate(range.getFirstIndex(), range.getCalendar().getFrequency()),
			range.size()
		);
	}
	
	public static Date beginTiqIndexToJavaDate( 
		final long index, 
		final TiqFrequency frequency
	) {
		Date date = DateHelper.indexToJavaDate(index, null);
		
		if (frequency.getCode() == TiqFrequency.WEEKLY_FREQUENCY) {
			Calendar calendar = new GregorianCalendar();
			calendar.setTime(date);
			calendar.add(Calendar.DAY_OF_YEAR, FIRST_DAY_OF_WEEK_CORRECTION);
			return calendar.getTime();
		}
		
		return date;
	}

	public static long javaDateToBeginTiqIndex( 
		final Date date, 
		final int frequency
	) {
		if (frequency == RegularInterval.WEEKLY) {
			Calendar calendar = new GregorianCalendar();
			calendar.setTime(date);
			calendar.add(Calendar.DAY_OF_YEAR, -FIRST_DAY_OF_WEEK_CORRECTION);
			return DateHelper.javaDateToIndex(calendar.getTime(), null);
		}
		
		return DateHelper.javaDateToIndex(date, null);
	}
	
}
