// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.fame;

import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;

import com.fame.timeiq.dates.*;
import Int.ecb.gesmescb.time.*;
import Int.ecb.gesmescb.edifact.RegularIntervalFormat;

import java.text.SimpleDateFormat;
import java.util.*;


/* ***************************** Class ****************************** */

/**
 * <P>Tests the implementation of static methods in IntervalConverter class in the
 * Int.ecb.gesmescb.fame package.</P>
 *
 * @author
 * <A HREF="mailto:Sami.Airo@ECB.int">Sami Airo</A>
 *
 * @version @(#)$Id: IntervalConverterTest.java,v 1.1 2017/01/19 16:14:03 trainot Exp $
 */

public class IntervalConverterTest extends TestCase {

	/* ========================= Fields ========================= */

	Calendar calendar;
	RegularIntervalFormat format;

	/* ====================== Constructors ====================== */

	/**
	 * Constructs a IntervalConverterTest with the specified name.
	 *
	 * @param name Test case name
	 */
	public IntervalConverterTest( final String name ) {
		super( name );
	}

	/* ================== Protected Methods ===================== */

	/**
	 * Sets up the test fixture.
	 */
	protected void setUp() {
		calendar = new GregorianCalendar(1970,0,1,0,0,0);
		format = new RegularIntervalFormat();
	}

	/**
	 * Tears sown the test fixture.
	 */
	protected void tearDown() {
		calendar = null;
		format = null;
	}

	/* ===================== Public Methods ===================== */

	/**
	 * Assembles and returns a test suite for all
	 * the test methods of this test case.
	 *
	 * @return a non-null test suite
	 */
	public static Test suite() {
		TestSuite suite = new TestSuite( IntervalConverterTest.class );
		return suite;
	}

	/**
	 * Runs the test case
	 */
	public static void main(String args[]) {
		String[] testCaseName = {
			IntervalConverterTest.class.getName()
		};
		junit.textui.TestRunner.main(testCaseName);
	}

	/* --------------------- Test Methods --------------------- */

	public void testRangeToInterval1 () 
		throws RangeTooLargeChkException
	{
		calendar.set(2000,0,1,0,0,0);
		Date begin = calendar.getTime();
		calendar.set(2000,0,2,0,0,0);
		Date end = calendar.getTime();
		RegularInterval interval = new GregorianRegularInterval(
			RegularInterval.DAILY,
			begin,
			end
		);
		calendar.add(Calendar.SECOND,-1);
		end = calendar.getTime();
		IndexValueRange range = new IndexValueRange(
			new SimpleCalendar(TiqFrequency.DAILY),
			new TiqDate(begin,TiqFrequency.DAILY),
			new TiqDate(end,TiqFrequency.DAILY)
		);
		assertEquals(interval.getSize(), range.size());
		assertEquals(interval.getBegin(), DateHelper.indexToJavaDate(range.getFirstIndex(),null));
		// FAILS assertEquals(interval.getEnd(), DateHelper.indexToJavaDate(range.getLastIndex(),null));
	}

	public void testRangeToInterval2 () 
		throws Exception
	{
		calendar.set(2000,0,1,0,0,0);
		Date begin = calendar.getTime();
		calendar.set(2000,0,2,0,0,0);
		Date end = calendar.getTime();
		RegularInterval interval1 = new GregorianRegularInterval(
			RegularInterval.DAILY,
			begin,
			end
		);
		
		IndexValueRange range1 = IntervalConverter.regularIntervalToIndexValueRange(interval1);
		RegularInterval interval2 = IntervalConverter.indexValueRangeToRegularInterval(range1);
	
		assertEquals(interval1, interval2);
	}
	

	private void validFormatTest(String in, String notused) {
		RegularInterval i = null;
		String out = null;

		try {
		 	i = format.parse(in);
		} catch (Exception e) {
			fail("Failed to parse " + in + ":" + e);
		}
	
		Date begin1 = i.getBegin();
		Date begin2 = null;
		
		try {
			IndexValueRange r = IntervalConverter.regularIntervalToIndexValueRange(i);
			begin2 = IntervalConverter.beginTiqIndexToJavaDate(r.getFirstIndex(), r.getCalendar().getFrequency());
			i = IntervalConverter.indexValueRangeToRegularInterval(r);
		} catch (Exception e) {
			fail("Failed to convert :" + e);
		}

		Date begin3 = i.getBegin();

		try {
		 	out = format.format( i );
		} catch (Exception e) {
			fail("Failed to format " + in + ":" + e );
		}

		assertEquals(in, out);
		assertEquals(begin1, begin2);
		assertEquals(begin1, begin3);
	}

	private void invalidFormatTest(String in) {
		
		try {
			//RegularInterval p = format.parse(in);
			format.parse(in);
			fail("Failed to catch invalid format " + in);
		} catch (Exception e) {
		}
	}


	public void testMinutelySingle() {
		String s = "199801010101:203";
		validFormatTest(s, s);

		s = "199801012359:203";
		validFormatTest(s, s);

		s = "199801012359:203";
		validFormatTest(s, s);

		s = "199812310101:203";
		validFormatTest(s, s);
		
		s = "199812312359:203";
		validFormatTest(s, s);

		invalidFormatTest("199801010101:777");
		invalidFormatTest("199801010161:203");
		invalidFormatTest("199813010101:203");
		invalidFormatTest("199801330101:203");
		invalidFormatTest("199801012601:203");
		invalidFormatTest("1998010126:203");
		invalidFormatTest("199812312400:203");
	}
	
	public void testMinutelyRange() {
		String
		s = "199801010101199812312359:719";
		validFormatTest(s, s);

		s = "199801012359200101010101:719";
		validFormatTest(s, s);

		s = "199801012359200801010000:719";
		validFormatTest(s, s);

		s = "199812310101200801010000:719";
		validFormatTest(s, s);
		
		s = "199812312359200812312359:719";
		validFormatTest(s, s);

		
		invalidFormatTest("199801010101200101010101:777");
		invalidFormatTest("199801010161200101010101:719");
		invalidFormatTest("199813010101200101010101:719");
		invalidFormatTest("199801330101200101010101:719");
		invalidFormatTest("199801012601200101010101:719");
		invalidFormatTest("1998010126:2001010101:719");
		invalidFormatTest("199812312400200101010101:719");
		// secondly
		invalidFormatTest("1998123123000020010101010100:719");
		// reverse order
		invalidFormatTest("200101010101199812312300:719");
	}


	public void testDailySingle() {
		String s = "19980101:102";
		validFormatTest(s, s);

		validFormatTest("19960329:102", "19960329:102");
		validFormatTest("19950329:102", "19950329:102");
		validFormatTest("19730329:102", "19730329:102");
		validFormatTest("19920629:102", "19920629:102");

		validFormatTest("19910121:102", "19910121:102");
		validFormatTest("19910129:102", "19910129:102");
		validFormatTest("19990113:102", "19990113:102");

		validFormatTest("19821231:102", "19821231:102");


		invalidFormatTest("19981301:102");
		invalidFormatTest("19981301:102");
		invalidFormatTest("19981232:102");
		invalidFormatTest("1998123:102");
	}


/*	public void testAddDailyRange() {
		TimePeriod p = null;
		TimePeriod p1 = null;
		TimePeriod p2 = null;
		TimePeriod p3 = null;

		try {
			p = format.parse("1959100119870213:711");
			p1 = format.parse("1987021620010502:711");
			p2 = format.parse("1959100120010502:711");
		} catch (ParseException e) {
			fail("Failed to parse:" + e);
		}

		p3 = p.add(p1);

		assertEquals(format.format(p2), format.format(p3));

	}
*/

	public void testDailyRange() {
		String s = "1998010119980102:711";
		validFormatTest(s, s);

		validFormatTest("1991012919990113:711", "1991012919990113:711");
		validFormatTest("1982123119890420:711", "1982123119890420:711");
		validFormatTest("1959100119870213:711", "1959100119870213:711");
		validFormatTest("1987021620010502:711", "1987021620010502:711");
		validFormatTest("1959100120010502:711", "1959100120010502:711");

		invalidFormatTest("1998130119980102:711");
		invalidFormatTest("1998123419980102:711");
		invalidFormatTest("199812349980102:711");
		invalidFormatTest("1998120119982302:711");
		invalidFormatTest("1998120119981245:711");
		invalidFormatTest("199812011998:711");
		invalidFormatTest("1998Ja0119980102:711");
		invalidFormatTest("-998010119980102:711");
		invalidFormatTest("1998010119980102DD:711");
	}


	public void testWeeklySingle() 
	{
		String s = "199801:616";
		validFormatTest(s, s);

		s = "199853:616";
		validFormatTest(s, s);

		invalidFormatTest("199953:616");
		invalidFormatTest("19995:616");
		invalidFormatTest("199957:616");
		invalidFormatTest("999576:616");
		invalidFormatTest("19995A:616");
	}
	
	public void testWeeklyConversion()
		throws Exception
	{		
		RegularInterval i = format.parse("200323:616");
		Date begin = i.getBegin();
		SimpleDateFormat dateFormat = new SimpleDateFormat("yyyy-MM-dd hh:mm:ss");
		//"Mon Jun 02 00:00:00 CEST 2003"
 		Date date = dateFormat.parse("2003-06-02 00:00:00");
		assertEquals(date, begin);
		IndexValueRange r = IntervalConverter.regularIntervalToIndexValueRange(i);
		Date firstIndex = DateHelper.indexToJavaDate(r.getFirstIndex(), null);
		//"Fri May 30 00:00:00 CEST 2003"
		date = dateFormat.parse("2003-05-30 00:00:00");
		assertEquals(date, firstIndex);
	}
	
	public void testWeeklyReferenceDayIsFriday() 
		throws Exception
	{		
		RegularInterval i = format.parse("199801:616");
		IndexValueRange r = IntervalConverter.regularIntervalToIndexValueRange(i);
		Date startDate = DateHelper.indexToJavaDate(r.getFirstIndex(), TimeZone.getTimeZone("Europe/Berlin"));
		Calendar calendar = new GregorianCalendar(TimeZone.getTimeZone("Europe/Berlin"));
		calendar.setTime(startDate);
		assertEquals(
			Calendar.FRIDAY,
			calendar.get(Calendar.DAY_OF_WEEK)
		);
	}


	public void testWeeklyRange() {
		String s = "199801199802:716";
		validFormatTest(s, s);

		s = "199853199901:716";
		validFormatTest(s, s);

		s = "196501200301:716";
		validFormatTest(s, s);

		s = "197002198405:716";
		validFormatTest(s, s);

		s = "198801198810:716";
		validFormatTest(s, s);

		s = "198753198801:716";
		validFormatTest(s, s);

		invalidFormatTest("200101200199:716");
		invalidFormatTest("200101200100:716");
		invalidFormatTest("2001012001DD:716");
		invalidFormatTest("-00101200102:716");
		invalidFormatTest("198853200001:716");
	}

	public void testMonthlySingle() {
		String s = "199801:610";
		validFormatTest(s, s);

		invalidFormatTest("200100:610");
		invalidFormatTest("200154:610");
		invalidFormatTest("2001DD:610");
	}


	public void testMonthlyRange() {
		String s = "199801199802:710";
		validFormatTest(s, s);

		invalidFormatTest("200101200199:710");
		invalidFormatTest("200101200100:710");
		invalidFormatTest("2001012001DD:710");
		invalidFormatTest("-00101200102:710");
	}

	public void testQuarterlySingle() {
		String s = "19981:608";
		validFormatTest(s, s);

		validFormatTest("20001:608", "20001:608");
		validFormatTest("19992:608", "19992:608");
		validFormatTest("30004:608", "30004:608");

		invalidFormatTest("20010:608");
		invalidFormatTest("20015:608");
		invalidFormatTest("2001D:608");
		invalidFormatTest(" 11:608");
		invalidFormatTest("11:608");
	}


	public void testQuarterlyRange() {
		String s = "1998119982:708";
		validFormatTest(s, s);

		s = "1998119984:708";
		validFormatTest(s, s);

		invalidFormatTest("2001120019:708");
		invalidFormatTest("2001120010:708");
		invalidFormatTest("200112001D:708");
		invalidFormatTest("-001120012:708");
	}

	public void testSemiannualSingle() {
		String s = "19981:604";
		validFormatTest(s, s);

		invalidFormatTest("19980:604");
		invalidFormatTest("19985:604");
		invalidFormatTest("19981A:604");
	}

	public void testSemiannualRange() {
		String s = "1998120012:704";
		validFormatTest(s, s);

		invalidFormatTest("2001120019:704");
		invalidFormatTest("2001120010:704");
		invalidFormatTest("200112001D:704");
		invalidFormatTest("-001120012:704");
	}

	public void testAnnualSingle() {
		String s = "1998:602";
		validFormatTest(s, s);

		invalidFormatTest("199:602");
		invalidFormatTest("-199:602");
		invalidFormatTest("99999:602");
		invalidFormatTest("AAAA:602");
		invalidFormatTest("AAAA: ");
		invalidFormatTest(":602");
		invalidFormatTest("19.99:602");
		invalidFormatTest("1999:602.4");
		invalidFormatTest("1999:6.2");
	}

	public void testAnnualRange() {
		String s = "19982001:702";
		validFormatTest(s, s);

		invalidFormatTest("1982001:702");
		invalidFormatTest("19992000A:702");
		invalidFormatTest("12001:702");
		invalidFormatTest("-1982001:702");
	}
	
	
	public void ftestRangeIndex() 
		throws Exception
	{
		Calendar calendar = new GregorianCalendar(1995,0,1,0,0,0);
		Date begin = calendar.getTime();
		
		RegularInterval interval = new GregorianRegularInterval(
			RegularInterval.MONTHLY,
			begin,
			30
		);

		IndexValueRange range = new IndexValueRange(
			new SimpleCalendar(TiqFrequency.MONTHLY),
			DateHelper.javaDateToIndex(begin,null),
			DateHelper.javaDateToIndex(begin,null)
		);		
		range.setPeriodsFromStart(30);
		
		Iterator bit = interval.iterator();
		while (bit.hasNext()) {
			RegularInterval baseInterval = (RegularInterval)bit.next();
			System.out.println(baseInterval.getBegin());
		}

		for (int i=0;i<range.size();i++) {
			System.out.println(DateHelper.indexToJavaDate(range.getLong(i),null));			
		}
		/*com.sun.java.util.collections.Iterator iit = range.listIterator();
		while (iit.hasNext()) {
			System.out.println(DateHelper.indexToJavaDate(iit.next(),null));
		}	*/			
	}
}
