// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.fame;

import java.io.File;

import com.fame.timeiq.*;

import Int.ecb.logging.Level;

import Int.ecb.gesmescb.*;

/* **************************** Class ***************************** */

/**
 * Writes each message into separate FAME database. Uses SDPEDeleter for
 * deletions.
 *
 * @author
 * <A HREF="mailto:klo@FAME.com">Kam Lo</A>
 * @author
 * <A HREF="mailto:Sami.Airo@ECB.int">Sami Airo</A>
 *
 * @version @(#)$Id: SDPEWriter.java,v 1.1 2017/01/19 16:14:03 trainot Exp $
 */

public class SDPEWriter extends FAMEWriter {

	/* ======================= Constants ======================== */

	/* ========================= Fields ========================= */

	private FAMERevisor revisor;

	private InterchangeBegin interchangeBegin;
	
	private String loggingFile; 

	/* ====================== Constructors ====================== */

	public SDPEWriter () {
		this( "" );
	}

	public SDPEWriter(final String name) {
		super( name );
		loggingFile
		= System.getProperty("Int.ecb.logging.FileHandler.pattern", "");
	}

	/* ===================== Protected Methods ===================== */

	public void handle ( InterchangeBegin interchangeBegin ) {
		this.interchangeBegin = interchangeBegin;
		super.handle( interchangeBegin );
	}

	public void handle ( MessageBegin messageBegin ) {
		revisor = null;

		if (messageBegin.isUpdateOrReplaceAction()) {
			revisor = new ESDBUpdater( getComponentHelper(), getFAMEDb(), messageBegin );
		}

		if (messageBegin.isDeleteAction()) {
			revisor = new SDPEDeleter( getComponentHelper(), getFAMEDb(), messageBegin );
		}

		if (revisor == null) {
			super.handle( messageBegin );
			return;
		}

		String databaseName = getComponentHelper().getPreference("destination", "");

		if (databaseName.equals("")) {
			getComponentHelper().logger.log(
				Level.SEVERE,
				getComponentHelper().logName,
				"handle",
				"Cannot read FAME database path from preference \"destination\"."
			);
			errorDisconnect();
			return;
		}

		getComponentHelper().logger.log(
			Level.CONFIG,
			getComponentHelper().logName,
			"handle",
			"destination: " + databaseName
		);

		openDatabase(
			new File(
				databaseName
				+ "_" + interchangeBegin.getSender()
				+ "_" + interchangeBegin.getReceiver()
				+ "_" + interchangeBegin.getIref()
				+ "_" + messageBegin.getMref()
				+ "_" + messageBegin.getDataSetId()
				+ ".db"
			)
		);

		super.handle( messageBegin );

		if (!getFAMEDb().createObject(
					messageBegin.getDataSetId() + ".IREF",
					new Scalar(interchangeBegin.getIref())
				)
			)
		{
			getComponentHelper().logger.log(
				Level.SEVERE,
				getComponentHelper().logName,
				"handle",
				"Cannot write IREF number to the FAME database."
			);
			errorDisconnect();
			return;
		}

		if (! (
			getFAMEDb().createObject( "ENVELOPE.IREF",new Scalar( interchangeBegin.getIref() ))
			&& getFAMEDb().createObject( "ENVELOPE.MREF",new Scalar( messageBegin.getMref() ))
			&& getFAMEDb().createObject( "ENVELOPE.SENDER",new Scalar( messageBegin.getSender() ))
			&& getFAMEDb().createObject( "ENVELOPE.RECEIVER",new Scalar( messageBegin.getReceiver() )) 
			&& getFAMEDb().createObject( "ENVELOPE.AGENCY",new Scalar( messageBegin.getMaintenanceAgencyId() ))
			&& getFAMEDb().createObject( "ENVELOPE.KEY_FAMILY",new Scalar( messageBegin.getKeyFamilyId() ))
			&& getFAMEDb().createObject( "ENVELOPE.DATASET",new Scalar( messageBegin.getDataSetId() ))
			&& getFAMEDb().createObject( "ENVELOPE.ORIGINAL_DATASET",new Scalar( messageBegin.getOriginalDataSetId() ))
			&& getFAMEDb().createObject( "ENVELOPE.ACTION", new Scalar( messageBegin.getAction() ))
			&& getFAMEDb().createObject( "ENVELOPE.EXTRACTION_TIME", new Scalar( messageBegin.getExtractionDateTime() )) 
			&& getFAMEDb().createObject( "ENVELOPE.REPORTING_PERIOD", new Scalar( messageBegin.getReportingPeriod() ))
			// trainot fix reporting period
			&& getFAMEDb().createObject( "ENVELOPE.REPORTING_PERIOD_FORMAT", new Scalar( messageBegin.getReportingPeriodFormat() ))
			&& getFAMEDb().createObject( "ENVELOPE.IDENTITY", new Scalar( messageBegin.getIdentity() ))
			&& getFAMEDb().createObject( "ENVELOPE.LOGGING_FILE", new Scalar( loggingFile ))
		)) {
			getComponentHelper().logger.log(
				Level.WARNING,
				getComponentHelper().logName,
				"handle",
				"Cannot write message information to the FAME database."
			);
			errorDisconnect();
			return;
		}
	}

	public void handle ( MessageEnd messageEnd ) {
		super.handle( messageEnd );
		closeDatabase(true);
	}

	protected FAMERevisor getRevisor () {
		return revisor;
	}

}
