// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2002, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


package Int.ecb.gesmescb.fame;

import java.io.IOException;

import java.text.DecimalFormat;
import java.text.SimpleDateFormat;

import java.util.*;

import Int.ecb.logging.Level;

import Int.ecb.gesmescb.*;
import Int.ecb.gesmescb.time.*;

import Int.ecb.gesmescb.event.ErrorEvent;
import Int.ecb.gesmescb.event.GESMESCBReader;
import Int.ecb.gesmescb.event.GESMESCBEventSourceAdapter;

import Int.ecb.gesmescb.util.*;

import com.fame.timeiq.*;
import com.fame.timeiq.data.ObservationList;
import com.fame.timeiq.data.TiqValue;
import com.fame.timeiq.dates.*;
import com.fame.timeiq.persistence.*;

/* ***************************** Class ****************************** */

/**
 * <P>Simple reader for a FAME database. Reads all numeric time series with 
 * the following frequencies: annual, semiannual, quarterly, monthly,
 * weekly, and daily. Dispatches {@link Int.ecb.gesmescb.TimeSeries} objects
 * representing the fame series. Dispatches also empty fame series. The time
 * series key is the name of the Fame series. Normal values of the Fame series are
 * converted to strings with OBS_STATUS value A. NC is converted to - (dash) with
 * OBS_STATUS value M. ND and NA are converter to empty observations.
 *
 * <center><table border="3" cellpadding="3">
 * <caption>
 * <a name="prefs">Available {@link Int.ecb.prefs preference} settings</a>
 * </caption>
 * <tr>
 * <th>preference name</th>
 * <th>type</th>
 * <th>meaning</th>
 * <th>default</th>
 * </tr>
 * <tr><td><code>IREF</code></td><td>string</td>
 *	<td>The IREF number to use in the {@link Int.ecb.gesmescb.InterchangeBegin} 
 * event. For format see {@link Int.ecb.gesmescb.InterchangeBegin}. Optional.</td>
 * <td><code>IREF000001</code></td></tr>
 * <tr><td><code>sender</code></td><td>string</td>
 *	<td>The sender code for the {@link Int.ecb.gesmescb.InterchangeBegin} and 
 * {@link Int.ecb.gesmescb.MessageBegin} events. Optional.</td>
 * <td><code>4F0</code></td></tr>
 * <tr><td><code>receiver</code></td><td>string</td>
 *	<td>The receiver code for the {@link Int.ecb.gesmescb.InterchangeBegin} and 
 * {@link Int.ecb.gesmescb.MessageBegin} events. Optional.</td>
 * <td><code>4F0</code></td></tr>
 * <tr><td><code>isTest</code></td><td>boolean</td>
 *	<td>Determines if the {@link Int.ecb.gesmescb.InterchangeBegin} is marked as test. Optional./td>
 *	<td><code>false</code></td></tr>
 * <tr><td><code>maintenanceAgency</code></td><td>String</td>
 *	<td>Maintenance agency identifier for the {@link Int.ecb.gesmescb.MessageBegin} event. Optional./td>
 *	<td><code>ECB</code></td></tr>
 * <tr><td><code>identity</code></td><td>String</td>
 *	<td>Identity text for the {@link Int.ecb.gesmescb.MessageBegin} event. Optional./td></tr>
 * <tr><td><code>dataSetName</code></td><td>String</td>
 *	<td>Data set name for the {@link Int.ecb.gesmescb.MessageBegin} event. Mandatory./td></tr>
 * <tr><td><code>reportingPeriod</code></td><td>String</td>
 *	<td>Reporting period for the {@link Int.ecb.gesmescb.MessageBegin} event. Optional./td></tr>
 * <tr><td><code>reportingPeriodFormat</code></td><td>String</td>
 *	<td>Reporting period format for the {@link Int.ecb.gesmescb.MessageBegin} event. Optional./td></tr>
 * <tr><td><code>keyFamilyName</code></td><td>String</td>
 *	<td>Name of the key family for the {@link Int.ecb.gesmescb.MessageBegin} event. Mandatory./td></tr>
 * </table></center>
 *
 *
 * @author <A HREF="mailto:sami.airo@ecb.int">Sami Airo</A>
 *
 * @version @(#)$Id: SimpleFAMEReader.java,v 1.1 2017/01/19 16:14:03 trainot Exp $
 */

public class SimpleFAMEReader
	extends GESMESCBEventSourceAdapter
	implements GESMESCBReader, Named
{
	/* ======================= Constants ======================== */

	/* ========================= Fields ========================= */

	/**
	 * Takes care of logging and preferences.
	 */
	protected final ComponentHelper componentHelper;

	/**
	 * Source FAME database.
	 */
	private FAMEDb sourceDatabase;

	/**
	 * Sender ID, read from preferences.
	 */
	private String sender;

	/**
	 * Receiver ID, read from preferences.
	 */
	private String receiver;

	/**
	 * Maintenance agency, read from preferences.
	 */
	private String maintenanceAgency;

	/**
	 * Message identity, read from preferences.
	 */
	private String identity;

	/**
	 * Dataset identity, read from preferences.
	 */
	private String dataSetName;

	/**
	 * Reporting period, read from preferences.
	 */
	private String reportingPeriod;

	/**
	 * Reporting period format, read from preferences.
	 */
	private String reportingPeriodFormat;

	/**
	 * Key family name, read from preferences.
	 */
	private String keyFamilyName;

	/**
	 * IREF string, read from preferences.
	 */
	private String iref;

	/**
	 * isTest flag, read from preferences.
	 */
	private boolean isTest;

	/**
	 * Current mref number.
	 */
	private int mrefNumber;

	/**
	 * Message dispatched flag.
	 */
	private boolean messageDispatched = false;

	/* ====================== Constructors ====================== */

	public SimpleFAMEReader () 
		throws IOException
	{
		this("", "");
	}

	/**
	 * Construct a <CODE>MPDFAMEReader</CODE> to read from FAME
	 * databases. If source database name not defined, it is read
	 * from preferences. 
	 *
	 * @param name	component used for logging and preferences.
	 * @param filename  source database name to read data from.
	 */
	public SimpleFAMEReader ( final String name, final String filename )
		throws IOException
	{
		componentHelper = new ComponentHelper( name, this );
		sourceDatabase = new FAMEDb("source");
		mrefNumber = 1;
	}

	/**
	 * Construct a <CODE>FAMEReader</CODE> to read from FAME
	 * databases. Reads the source and optional target database name
	 * from preferences.
	 *
	 * @param name	used for logging and preferences.
	 *
	 * @throws NullPointerException if the preference variable
	 * <CODE>sourceDatabase</CODE> is undefined.
	 */
	public SimpleFAMEReader ( final String name )
		throws IOException
	{
		this( name, null );
	}

	/* ===================== Public Methods ===================== */

	/**
	 * Returns the name of the component.
	 *
	 * @return the component name.
	 */
	public String getName () {
		return componentHelper.name;
	}


	/**
	 * <P>Retrieves time series from the FAME database and dispatches them as
	 * {@link Int.ecb.gesmescb.TimeSeries} to all listeners.</P>
	 */
	public void run () {
		try {
			getPreferences();
		} catch (Exception e) {
			logAndDispatchSevereError(
				"Cannot read mandatory preferences.",
				"run()",
				e
			);
			return;
		}
		
		if ( !openDatabase( "source", true, sourceDatabase) ) {
			logAndDispatchSevereError(
				"Cannot open source database.",
				"run()"
			);
			return;
	   }

		try {
		   dispatchInterchangeBegin();
		
			if (! dispatchAllTimeSeries( ) ) {
				return;
			}

			if ( messageDispatched ) {
				dispatchMessageEnd();
			}
			
			dispatchInterchangeEnd();
		} catch ( final Exception e ) {
			logAndDispatchSevereError(
				"Unexpected error: " + e,
				"run()",
				e
			);
		}
	}

	/**
	 * <P>Close the <CODE>FAMEReader</CODE>.</P>
	 *
	 * <P>This method should always be called
	 * when the <CODE>FAMEReader</CODE>
	 * has served its purpose.</P>
	 *
	 * <P>Closes all databases opened by the
	 * <CODE>FAMEReader</CODE>.</P>
	 */
	public void close () {
		try {
			sourceDatabase.close( false );
		}
		catch ( final Exception e ) {
			logAndDispatchSevereError(
				"Problem: " + e.getMessage() + ".",
				"close()",
				e
			);
		}
	}


	/* ===================== Package Methods ===================== */

	boolean dispatchAllTimeSeries () {
		ScreeningCriteria screen = new ScreeningCriteria();
		
		screen.setAll(false);
		screen.setClass(TiqClass.SERIES, true);
		screen.setType(TiqType.DOUBLE, true);
		screen.setType(TiqType.FLOAT, true);
		screen.setType(TiqType.INTEGER, true);
		screen.setType(TiqType.LONG, true);
		screen.setType(TiqType.SHORT, true);
		screen.setFrequency(TiqFrequency.ANNUAL, true);
		screen.setFrequency(TiqFrequency.SEMIANNUAL, true);
		screen.setFrequency(TiqFrequency.QUARTERLY, true);
		screen.setFrequency(TiqFrequency.MONTHLY, true);
		screen.setFrequency(TiqFrequency.WEEKLY, true);
		screen.setFrequency(TiqFrequency.DAILY, true);
		
		Iterator objectIterator = sourceDatabase.getDataObjectInfoIterator("*", screen, false);

		if (! objectIterator.hasNext() ) {
			componentHelper.logger.log(
				Level.WARNING,
				componentHelper.logName,
				"dispatchTimeSeries()",
				"Source database \"" + sourceDatabase.toString() + "\" doesn't contain time series to process."
			);
			return true;
		}

		try {
			while ( objectIterator.hasNext() ) {
				DataObjectInfo dataObjectInfo = ( DataObjectInfo ) objectIterator.next();
				
				if ( ! dispatchTimeSeries( dataObjectInfo.getName() ) ) {
					return false;
				}
			}
		} catch ( ConnectionFailedChkException e ) {
		 	logAndDispatchSevereError(
				"Connection to the server failed.",
				"getObjectIterator()",
				e
			);
			return false;
		} catch ( ObjectAccessChkException e ) {
			logAndDispatchSevereError(
				"Error accessing object.",
				"getObjectIterator()",
				e
			);
			return false;
		}

		return true;
	}


	boolean dispatchTimeSeries (
		final String timeSeriesKey
	) {
		componentHelper.logger.log(
			Level.FINEST,
   		componentHelper.logName,
			"dispatchTimeSeriesUpdate()",
			"Dispatching time series " + timeSeriesKey
		);

		TiqView sourceValues = null;

		sourceValues = sourceDatabase.getObject( timeSeriesKey );

		if ( sourceValues == null ) {
			logAndDispatchSevereError(
				"Time series \"" + timeSeriesKey
				+ "\" is not in the source database",
				"dispatchTimeSeries()"
			);			
			return false;
		}
		
		if ( ! messageDispatched ) {
			dispatchMessageBegin(
				MessageBegin.UPDATE_OR_REPLACE_ACTION
			);
		}

		TimeSeries timeSeries = null;
			
		try {		
			timeSeries = 
				convertObservationListToTimeSeries( 
					timeSeriesKey, 
					sourceValues.getObservations() 
				);
		} catch ( RangeTooLargeChkException e ) {
			logAndDispatchSevereError(
				"Error in dispatching time series \""
					+ timeSeriesKey + "\".",
				"dispatchTimeSeries()",
				e
			);			
			return false;
		} catch ( FunctorConsistencyChkException e ) {
			logAndDispatchSevereError(
				"Error in dispatching time series \""
					+ timeSeriesKey + "\".",
				"dispatchTimeSeries()",
				e
			);
				
			return false;
		}			
			
		componentHelper.logger.log(
			Level.FINER,
			componentHelper.logName,
			"dispatchTimeSeries()",
			"Dispatching TimeSeries: " + timeSeriesKey,
			timeSeries
		);

		dispatch( timeSeries );
			
		return true;
	}
	

	TimeSeries convertObservationListToTimeSeries (
		final String key,
		final ObservationList tiqObsList
	) throws RangeTooLargeChkException {
		
		if (tiqObsList.size() == 0) {
			return 
				new TimeSeries(
					key,
					getEmptyRegularInterval( tiqObsList.getFrequency() ),
					new Observation[]{}
				);
		}

		List obsList = new ArrayList();
		
		for (int i = 0; i < tiqObsList.size(); ++i) {
			String obsValue = convertTiqValueToString(((com.fame.timeiq.data.Observation)tiqObsList.get(i)).getTiqValue());
			obsList.add( 
				new Int.ecb.gesmescb.Observation( 
					obsValue, 
					convertObsValueToObsStatus(obsValue), 
					"", 
					""
				) 
			);
		}
		
		return 
			new TimeSeries(
				key, 
				IntervalConverter.indexValueRangeToRegularInterval(
					new IndexValueRange(
						tiqObsList.getCalendar(), 
						tiqObsList.getFirstIndex(),
						tiqObsList.getLastIndex()
					)
				),
				obsList
			);
	}
	
	
	RegularInterval getEmptyRegularInterval ( final TiqFrequency tiqFrequency ) {
		final Date ANY_DATE = new Date(0);

		return 
			new GregorianRegularInterval( 
				IntervalConverter.tiqFrequencyToRegularIntervalFrequency( tiqFrequency ),
				ANY_DATE,
				0
			);
	}
	
	
	String convertTiqValueToString (
		final TiqValue tiqValue
	) {
		if (tiqValue.getStatus() == TiqConstants.STATUS_NA) return "";
		if (tiqValue.getStatus() == TiqConstants.STATUS_ND) return "";
		if (tiqValue.getStatus() == TiqConstants.STATUS_NIC) return "";
		if (tiqValue.getStatus() == TiqConstants.STATUS_NC) return Int.ecb.gesmescb.Observation.MISSING_VALUE;
		
		if (tiqValue.getTiqType().equals(TiqType.DOUBLE)) return Double.toString(tiqValue.getDoubleValue());
		if (tiqValue.getTiqType().equals(TiqType.FLOAT)) return Float.toString(tiqValue.getFloatValue());
		if (tiqValue.getTiqType().equals(TiqType.INTEGER)) return Integer.toString(tiqValue.getIntValue());
		if (tiqValue.getTiqType().equals(TiqType.LONG)) return Long.toString(tiqValue.getLongValue());
		if (tiqValue.getTiqType().equals(TiqType.SHORT)) return Short.toString(tiqValue.getShortValue());
		
		return "";
	}
	
	
	String convertObsValueToObsStatus ( final String value ) {
		if ("".equals(value)) return "";
		if (Int.ecb.gesmescb.Observation.MISSING_VALUE.equals(value)) return "M";
		return "A";
	}
	

	/* ==================== Private Methods ===================== */

	/**
	 * <P>Dispatch message begin event.</P>
	 *
	 * @param action	action to perform
	 */
	private void dispatchMessageBegin ( 
		final String action
	) {
		// Construct MREF tag
		DecimalFormat mrefFormat = new DecimalFormat( "000000" );
		String mref = "MREF" + mrefFormat.format( mrefNumber );

		// Construct current date and time string in CCYYMMDDhhmm
		// format for extraction date and time
		SimpleDateFormat formatter = new SimpleDateFormat ( "yyyyMMddHHmm" );
		String extractionDateTime = formatter.format( new Date() );

		final MessageBegin messageBegin = new MessageBegin(
			mref, sender, receiver, maintenanceAgency,
			MessageBegin.STATISTICAL_DATA,
			identity,
			new Contact[]{},
			dataSetName,
			action,
			extractionDateTime,
			reportingPeriod, reportingPeriodFormat,
			keyFamilyName
		);
		
		componentHelper.logger.log(
		    Level.FINE,
			componentHelper.logName,
			"dispatchMessage",
			"Dispatching MessageBegin " + mref,
			messageBegin
		);
		
		dispatch( messageBegin );

		messageDispatched = true;
	}


	protected void logAndDispatchSevereError (
		final String msg,
		final String procedureName
	) {
		componentHelper.logger.log(
			Level.SEVERE,
			componentHelper.logName,
			procedureName,
			msg
		);
		
		dispatch(
			new ErrorEvent( msg, ErrorEvent.ERROR )
		);
	}

	protected void logAndDispatchSevereError (
		final String msg,
		final String procedureName,
		final Exception e
	) {
		componentHelper.logger.log(
			Level.SEVERE,
			componentHelper.logName,
			procedureName,
			msg,
			e
		);
		
		dispatch(
			new ErrorEvent( msg, ErrorEvent.ERROR )
		);
	}

	/**
	 * <P>Retrieves all preference value required by the
	 * <CODE>FAMEReader</CODE>.</P>
	 *
	 * @throws NullPointerException if a manditory preference
	 * variable has not been defined.
	 */
	private void getPreferences ()
		throws NullPointerException 
	{
		// Read IREF from preferences
		iref = getPreference( "IREF", true, "IREF000001" );
		// Read sender from preferences
		sender = getPreference( "sender", true, "4F0" );
		// Read receiver from preferences
		receiver = getPreference( "receiver", true, "4F0" );
		// Read isTest from preferences
		isTest = getPreference( "isTest", false, "false" ).equalsIgnoreCase( "true" );
		// Read maintenanceAgency from preferences
		maintenanceAgency = getPreference( "maintenanceAgency", true, "ECB" );
		// Read identity from preferences
		identity = getPreference( "identity", false, "" );
		// Read datasetid from preferences
		dataSetName = getPreference( "dataSetName", true, "" );
		// Read reportingPeriod from preferences
		reportingPeriod = getPreference( "reportingPeriod", false, "" );
		// Read reportingPeriodFormat from preferences
		reportingPeriodFormat = getPreference( "reportingPeriodFormat", false, "" );
		// Read keyFamilyName from preferences
		keyFamilyName = getPreference( "keyFamilyName", true, "" );
	}


	/**
	 * <P>Retrieves a preference value.</P>
	 *
	 * @param key	key for the preference
	 * @param manditoryFlag indicates whether a preference value is
	 * manditory for the this component.
	 *
	 * @return preference value for the specified key
	 *
	 * @throws NullPointerException if the preference variable
	 * for the key is undefined and is manditory.
	 */
	private String getPreference ( 
		final String key, 
		final boolean mandatory,
		final String defaultValue
	)
		throws NullPointerException 
	{
		String preference = componentHelper.getPreference( key, defaultValue );

		if ( preference.equals( "" ) ) {
			if ( mandatory ) {
				final String desc
					= "No value for preference variable \""
					+ key + "\" found.";

				componentHelper.logger.log(
					Level.SEVERE,
					componentHelper.logName,
				   "getPreference()",
					desc
				);

				throw new NullPointerException( desc );
			}

			final String desc
					= "No value for preference variable \""
						+ key + "\" found.";

			componentHelper.logger.log(
				Level.FINER,
				componentHelper.logName,
				"getPreference()",
				desc
			);
			
			return preference;
		}

		final String desc = 
			"Found preference variable \"" + key + "\" with value \""
		    + preference + "\".";

		componentHelper.logger.log(
			Level.FINEST,
			componentHelper.logName,
			"getPreference()",
			desc
		);

		return preference;
	}


	private boolean openDatabase (
		final String preferenceName,
		final boolean mandatory,
		final FAMEDb database
	) {
		String filename = getPreference( preferenceName, mandatory, "" );			

		if ( filename.equals( "" ) ) {
			return !mandatory;
		}
		
		componentHelper.logger.log(
			Level.INFO,
			componentHelper.logName,
			"openDatabase()",
			"Opening " + preferenceName + " Fame database in read mode: " + filename
		);

		Properties properties = new Properties();
		properties.put("mode", "READ_ONLY");

		return database.open( filename, properties );
	}

	/**
	 * <P>Dispatch interchange begin event.</P>
	 */
	private void dispatchInterchangeBegin () {
		SimpleDateFormat formatter = new SimpleDateFormat ( "yyMMdd:HHmm" );
		String today = formatter.format( new Date() );

		final InterchangeBegin interchangeBegin
			= new InterchangeBegin( iref, sender, receiver, today, isTest );
		
		componentHelper.logger.log(
			Level.FINE,
			componentHelper.logName,
		   "dispatchInterchangeBegin()",
			"Dispatching Interchange Begin",
			interchangeBegin
	    );

		dispatch( interchangeBegin );
	}

	/**
	 * <P>Dispatch message end event.</P>
	 */
	private void dispatchMessageEnd () {
		DecimalFormat mrefFormat = new DecimalFormat( "000000" );
		String mref = "MREF" + mrefFormat.format( mrefNumber );
		mrefNumber++;

		final MessageEnd messageEnd = new MessageEnd( mref );
		componentHelper.logger.log(
			Level.FINE,
			componentHelper.logName,
		   "run()",
			"Dispatching MessageEnd",
			messageEnd
		);
		
	   dispatch( messageEnd );
		messageDispatched = false;
	}

	/**
	 * <P>Dispatch interchange end event.</P>
	 */
	private void dispatchInterchangeEnd () {
		final InterchangeEnd interchangeEnd = new InterchangeEnd( iref );
		componentHelper.logger.log(
			Level.FINE,
			componentHelper.logName,
			"dispatchInterchangeEnd()",
			"Dispatching Interchange End",
			interchangeEnd
		);
		dispatch( interchangeEnd );
	}

}
