// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.fame;

import java.io.File;
import java.util.Properties;
import java.util.Iterator;

import com.fame.timeiq.RegularSeries;
import com.fame.timeiq.data.TiqValue;
import com.fame.timeiq.TiqConstants;
import com.fame.timeiq.dates.RangeTooLargeChkException;
import com.fame.timeiq.dates.DateOverflowChkException;
import com.fame.timeiq.dates.IndexValueRange;

import Int.ecb.logging.Level;

import Int.ecb.gesmescb.*;
import Int.ecb.gesmescb.event.*;

import Int.ecb.gesmescb.util.ComponentHelper;

/* **************************** Class ***************************** */

/**
 * Writer for writing time series into a FAME database. Writes the 
 * OBS_VALUE attributes into a FAME time series with the same name as
 * the time series key. Supports only update and replace messages.
 * Empty time series are stored as empty FAME series. Creates a new FAME database
 * if it doesn't exist. Existing time series in the FAME database are updated.
 * Doesn't support attributes.
 *
 * @author
 * <A HREF="mailto:Sami.Airo@ECB.int">Sami Airo</A>
 *
 * @version @(#)$Id: SimpleFAMEWriter.java,v 1.1 2017/01/19 16:14:03 trainot Exp $
 */

public class SimpleFAMEWriter extends GESMESCBEventListenerAdapter
	implements DataListener, GESMESCBWriter
{
	/* ======================= Constants ======================== */

	/* ========================= Fields ========================= */

	/** Takes care of logging and prefs */
	private ComponentHelper componentHelper;

	/** Current FAMEDb for writing data. */
	private final FAMEDb fameDb;

	private boolean isUpdateOrReplaceAction;
	
	/* ====================== Constructors ====================== */

	public SimpleFAMEWriter () {
		this( "" );
	}

	public SimpleFAMEWriter(final String name) {
		componentHelper = new ComponentHelper( name, this );
		fameDb = new FAMEDb(name);

		String databaseName = componentHelper.getPreference("destination", "");

		if (databaseName.equals("")) {
			final String desc = "Cannot read FAME database path from preference \"destination\".";
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"SimpleFAMEWriter",
				desc
			);
			errorDisconnect();
			throw new NullPointerException( desc );
		}

		componentHelper.logger.log(
			Level.CONFIG,
			componentHelper.logName,
			"SimpleFAMEWriter",
			"destination: " + databaseName
		);

		openDatabase( new File(databaseName) );
	}

	/* ===================== Public Methods ===================== */

	public void close () {
		componentHelper.logger.log(
			Level.FINE,
			componentHelper.logName,
			"close",
			"Closing."
		);

		if (!closeDatabase(true)) {
			errorDisconnect();
		}
	}

	public void handle ( final InterchangeBegin interchangeBegin ) {
		componentHelper.logger.log(
			Level.FINE,
			componentHelper.logName,
			"handle",
			"Interchange begins"
		);
	}

	public void handle ( final InterchangeEnd interchangeEnd ) {
		componentHelper.logger.log(
			Level.FINE,
			componentHelper.logName,
			"handle",
			"Interchange ends"
		);
	}

	public void handle ( final ErrorEvent event ) {
		componentHelper.logger.log(
			Level.FINE,
			componentHelper.logName,
			"handle",
			"Handling ErrorEvent"
		);

		if (ErrorEvent.WARNING < event.getSeverity()) {
			closeDatabase(false);
		}

		super.handle(event);
	}

	public void handle ( final MessageBegin messageBegin ) {
		componentHelper.logger.log(
			Level.FINE,
			componentHelper.logName,
			"handle",
			"Message begins"
		);

		isUpdateOrReplaceAction = messageBegin.isUpdateOrReplaceAction();
	}

	public void handle ( final MessageEnd messageEnd ) {
		componentHelper.logger.log(
			Level.FINE,
			componentHelper.logName,
			"handle",
			"Message ends"
		);
	}

	public void handle ( final TimeSeries timeSeries ) {
		componentHelper.logger.log(
			Level.FINE,
			componentHelper.logName,
			"handle",
			"Handling TimeSeries " + timeSeries.getKey()
		);
		
		if (!isUpdateOrReplaceAction) return;
		
		if (timeSeries.getInterval() == null) {
			componentHelper.logger.log(
				Level.FINE,
				componentHelper.logName,
				"handle",
				"TimeSeries \"" + timeSeries.getKey() + "\" has no interval"
			);
			return;
		}
		
		RegularSeries tiqSeries = convertTimeSeriesToRegularSeries(timeSeries);
		
		if (tiqSeries == null) {
			errorDisconnect();
			return;
		}
		
		fameDb.createUpdateTimeSeries( timeSeries.getKey(), tiqSeries );
	}

	public void handle ( final Attribute attribute ) {
		componentHelper.logger.log(
			Level.FINE,
			componentHelper.logName,
			"handle",
			"Handling AttributeEvent"
		);
	}

	/* ===================== Protected Methods ===================== */

	/**
	 * Open database for loading data.
	 *
	 * @param file FAME database file to open or create
	 *
	 * @return TRUE if database opened; false otherwise
	 */
	protected boolean openDatabase( File file ) {
		Properties properties = new Properties();
		if (file.exists()) {
			componentHelper.logger.log(
				Level.INFO,
				componentHelper.logName,
				"openDatabase",
				"Opening Fame database in shared mode: " + file.getPath()
			);
			properties.put("mode", "READ_WRITE");
		} else {
			componentHelper.logger.log(
				Level.INFO,
				componentHelper.logName,
				"openDatabase",
				"Creating Fame database: " + file.getPath()
			);
			properties.put("mode", "CREATE");
		}

		return fameDb.open(file.getPath(), properties);
	}

	/**
	 * Closes the FAME database.
	 *
	 *
	 * @return TRUE if database closed; false otherwise
	 */
	protected boolean closeDatabase (final boolean commit) {
		if (!fameDb.isOpen()) return true;

		if (commit) {
			componentHelper.logger.log(
				Level.INFO,
				componentHelper.logName,
				"closeDatabase",
				"Closing Fame database: " + fameDb
			);
		} else {
			componentHelper.logger.log(
				Level.WARNING,
				componentHelper.logName,
				"closeDatabase",
				"Trying to restore and close Fame database: " + fameDb
			);
		}

		return fameDb.close(commit);
	}

	/**
	 * Disconnect from source and close databases because of an error.
	 */
	protected void errorDisconnect() {
		componentHelper.logger.log(
			Level.SEVERE,
			componentHelper.logName,
			"errorDisconnect",
			"Closing databases and disconnecting from source"
		);

		closeDatabase(false);
		setActive( false );
	}


	protected RegularSeries convertTimeSeriesToRegularSeries (
		final TimeSeries series
	) {
		double[] valueArray = new double[series.length()];
		byte[]	 valueStatusArray = new byte[series.length()];

		int i = 0;
		Iterator it = series.observationIterator();

		while (it.hasNext()) {
			TiqValue tiqValue = 
				convertStringToTiqValue( 
					((Observation)it.next()).getValue()
				);

			valueArray[i] = tiqValue.getDoubleValue();
			valueStatusArray[i] = tiqValue.getStatus();
			
			i++;
		}
		
		IndexValueRange range = null;
		try {
			range = IntervalConverter.regularIntervalToIndexValueRange( series.getInterval() );
		} catch (DateOverflowChkException e) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"convertTimeSeriesToRegularSeries",
				"Error in converting time range for series: " + series.getKey(),
				e
			);
			return null;
		} catch (RangeTooLargeChkException e) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"convertTimeSeriesToRegularSeries",
				"Error in converting time range for series: " + series.getKey(),
				e
			);
			return null;
		}
		
		return 
			new RegularSeries(
				range,		
				valueArray, 
				valueStatusArray
			);
	}
	

	protected TiqValue convertStringToTiqValue (
		final String value
	) {
		if (value.equals(Observation.MISSING_VALUE)) return new TiqValue( new Double(-0), TiqConstants.STATUS_NC	);
		if (value.equals("")) return new TiqValue( new Double(-0), TiqConstants.STATUS_ND );
		
		return new TiqValue( new Double(value), TiqConstants.STATUS_OK );
	}

}
