// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.metabase;

import java.lang.String;

import java.util.Map;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Collections;

import Int.ecb.gesmescb.KeyFamily;
import Int.ecb.gesmescb.StatisticalConcept;
import Int.ecb.gesmescb.CodeList;

/* ***************************** Class ****************************** */

/**
 * <P>Holds the code lists, statistical concepts, and key families
 * maintained by a maintenance agency.</P>
 *
 * @author
 * <A HREF="mailto:klo@FAME.com">Kam Lo</A>
 * @author
 * <A HREF="mailto:Sami.Airo@ECB.int">Sami Airo</A>
 * @author
 * <A HREF="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</A>
 *
 * @version @(#)$Id: MaintenanceAgency.java,v 1.1 2017/01/19 16:14:05 trainot Exp $
 */
public final class MaintenanceAgency {

	/* ========================= Fields ========================= */

	/**
	 * Maintenance agency identifier.
	 */
	private final String id;

	/**
	 * Mapping of the code list identifiers to code lists.
	 */
	private final Map codeLists;

	/**
	 * Mapping of the concept identifiers to maintained concepts.
	 */
	private final Map concepts;

	/**
	 * Mapping of the key family identifiers to key families.
	 */
	private final Map keyFamilies;

	/* ====================== Constructors ====================== */

	/**
	 * Construct a new maintenance agency with given identifier.
	 *
	 * @param agencyId maintenance agency identifier.
	 */
	public MaintenanceAgency ( final String agencyId ) {
		id = agencyId;
		codeLists = new HashMap();
		concepts = new HashMap();
		keyFamilies = new HashMap();
	}

	/* ===================== Public Methods ===================== */

	/**
	 * Returns the maintenance agency identifier.
	 *
	 * @return maintenance agency identifier.
	 */
	public String getId () {
		return id;
	}

	/**
	 * Returns the code list for a code list identifier.
	 * Returns null if the maintenance agency does not contain
	 * a code list for this name.
	 *
	 * @param codeListId code list identifier.
	 *
	 * @return code list with the specified identifier.
	 */
	public CodeList getCodeList ( final String codeListId ) {
		return (CodeList)codeLists.get( codeListId );
	}

	/**
	 * Returns true if this maintenance agency contains
	 * the specified code list identifier.
	 *
	 * @param codeListId code list identifier whose presence
	 * in this maintenance agency is to be tested.
	 *
	 * @return true if this maintenance agency contains
	 * the specified code list identifier.
	 */
	public boolean containsCodeList ( final String codeListId ) {
		return codeLists.containsKey( codeListId );
	}

	/**
	 * Returns an iterator over the code list in this maintenance agency.
	 *
	 * @return an iterator over the code lists in this maintenance agency.
	 */
	public Iterator codeListIterator () {
		return
			Collections.unmodifiableCollection(
				codeLists.values()
			).iterator();
	}

	/**
	 * Returns the concept for an identifier.
	 * Returns null if the maintenance agency does not contain
	 * a concept for this identifier.
	 *
	 * @param conceptId concept identifier.
	 *
	 * @return key family with the specified identifier.
	 */
	public StatisticalConcept getStatisticalConcept ( final String conceptId ) {
		return (StatisticalConcept)concepts.get( conceptId );
	}

	/**
	 * Returns true if this maintenance agency contains
	 * the specified concept identifier.
	 *
	 * @param conceptId concept identifier whose presence
	 * in this maintenance agency is to be tested.
	 *
	 * @return true if this maintenance agency contains
	 * the specified concept identifier.
	 */
	public boolean containsStatisticalConcept ( final String conceptId ) {
		return concepts.containsKey( conceptId );
	}

	/**
	 * Returns an iterator over the concepts in this maintenance agency.
	 *
	 * @return an iterator over the concepts in this maintenance agency.
	 */
	public Iterator statisticalConceptIterator () {
		return
			Collections.unmodifiableCollection(
				concepts.values()
			).iterator();
	}

	/**
	 * Returns the key family for an identifier.
	 * Returns null if the maintenance agency does not contain
	 * key family for this identifier.
	 *
	 * @param keyFamilyId key family identifier.
	 *
	 * @return key family with the specified identifier.
	 */
	public KeyFamily getKeyFamily ( final String keyFamilyId ) {
		return (KeyFamily)keyFamilies.get( keyFamilyId );
	}

	/**
	 * Returns true if this maintenance agency contains
	 * the specified key family identifier.
	 *
	 * @param keyFamilyId key family identifier whose presence
	 * in this maintenance agency is to be tested.
	 *
	 * @return true if this maintenance agency contains
	 * the specified key family.
	 */
	public boolean containsKeyFamily ( final String keyFamilyId ) {
		return keyFamilies.containsKey( keyFamilyId );
	}

	/**
	 * Returns an iterator over the key families in this maintenance agency.
	 *
	 * @return an iterator over the key families in this maintenance agency.
	 */
	public Iterator keyFamilyIterator () {
		return
			Collections.unmodifiableCollection(
				keyFamilies.values()
			).iterator();
	}

	/**
	 * <P>Returns a string representation
	 * of this <CODE>MaintenanceAgency</CODE>.</P>
	 *
	 * <P>This method is intended to be used only for debugging purposes,
	 * and the format of the returned string may vary between
	 * implementations.</P>
	 *
	 * <P>The returned string may be empty but may not be null.</P>
	 *
	 * @return a string representation
	 * of this <CODE>MaintenanceAgency</CODE>.
	 */
	public String toString () {
		return getClass().getName()
			+ " ( currently valid CodeLists = " + codeLists
			+ ", currently valid StatisticalConcepts = " + concepts
			+ ", currently valid KeyFamilies = " + keyFamilies
			+ " )";
	}

	/* ==================== Package Methods ===================== */

	/**
	 * Associates the specified code list with its code list identifier
	 * in this maintenance agency.
	 * If the maintenance agency previously contained a mapping
	 * for this identifier, the old code list is replaced.
	 *
	 * @param codeList code list to be associated with the its identifier.
	 */
	void putCodeList ( final CodeList codeList ) {
		codeLists.put( codeList.getId(), codeList );
	}

	/**
	 * Associates the specified concept with its identifier
	 * in this maintenance agency.
	 * If the maintenance agency previously contained a mapping
	 * for this concept identifier, the concept is replaced.
	 *
	 * @param concept concept to be associated with the its identifier.
	 */
	void putStatisticalConcept ( final StatisticalConcept concept ) {
		concepts.put( concept.getId(), concept );
	}

	/**
	 * Associates the specified key family with its identifier
	 * in this maintenance agency.
	 * If the maintenance agency previously contained
	 * a mapping for this key family identifier,
	 * the key family is replaced.
	 *
	 * @param keyFamily key family to be associated with the its identifier.
	 */
	void putKeyFamily ( final KeyFamily keyFamily ) {
		keyFamilies.put( keyFamily.getId(), keyFamily );
	}
}
