// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.metabase;

import java.lang.String;
import java.util.Iterator;
import java.util.Date;

import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;

import Int.ecb.gesmescb.KeyFamily;
import Int.ecb.gesmescb.KeyFamilyComponent;
import Int.ecb.gesmescb.CodeList;
import Int.ecb.gesmescb.StatisticalConcept;

import Int.ecb.gesmescb.metabase.MaintenanceAgency;

/* ***************************** Class ****************************** */

/**
 * <P>Tests the implementation of {@link MaintenanceAgency}.</P>
 *
 * @author
 * <A HREF="mailto:Sami.Airo@ECB.int">Sami Airo</A>
 * @author
 * <A HREF="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</A>
 *
 * @version @(#)$Id: MaintenanceAgencyTest.java,v 1.1 2017/01/19 16:14:05 trainot Exp $
 */

public final class MaintenanceAgencyTest extends TestCase {

	/* ========================= Fields ========================= */

	private KeyFamily[] keyFamilies;
	private CodeList[] codeLists;
	private StatisticalConcept[] concepts;

	/* ====================== Constructors ====================== */

	/**
	 * Constructs a MaintenanceAgencyTest with the specified name.
	 *
	 * @param name Test case name.
	 */
	public MaintenanceAgencyTest ( final String name ) {
		super( name );
	}

	/* ================== Protected Methods ===================== */

	/**
	 * Sets up the test fixture.
	 */
	protected void setUp () {
		codeLists = new CodeList[3];
		concepts = new StatisticalConcept[3];
		keyFamilies = new KeyFamily[3];

		final KeyFamilyComponent[] components = new KeyFamilyComponent[3];
		final Date versionDate = new Date();
		final String[] ids = new String[]{ "A", "B" };
		final String[] descs = new String[]{ "AD", "BD" };

		codeLists[0] = new CodeList(
			"CL0", "name", "desc", versionDate, ids, descs
		);
		codeLists[1] = new CodeList(
			"CL1", "name", "desc", versionDate, ids, descs
		);
		codeLists[2] = new CodeList(
			"CL0", "name1", "desc", versionDate, ids, descs
		);

		concepts[0] = new StatisticalConcept(
			"CONCEPT0", "Concept", "1abc", versionDate
		);
		concepts[1] = new StatisticalConcept(
			"CONCEPT1", "Concept", "1abc", versionDate
		);
		concepts[2] = new StatisticalConcept(
			"CONCEPT0", "Concept1", "1abc", versionDate
		);

		components[0] = new KeyFamilyComponent(
			"CONCEPT1",
			KeyFamilyComponent.KEY_STRUCTURE_DIMENSION,
			"1",
			"AN5",
			KeyFamilyComponent.MANDATORY,
			"",
			"CL_TEST"
		);

		components[1] = new KeyFamilyComponent(
			"FREQ",
			KeyFamilyComponent.KEY_STRUCTURE_DIMENSION,
			"2",
			"AN5",
			KeyFamilyComponent.MANDATORY,
			"",
			"CL_TEST"
		);

		components[2] = new KeyFamilyComponent(
			"CONCEPT3",
			KeyFamilyComponent.ATTRIBUTE,
			"3",
			"AN5",
			KeyFamilyComponent.CONDITIONAL,
			"5",
			"CL_TEST"
		);

		keyFamilies[0] = new KeyFamily( "KF0", "name", "desc", versionDate, components );
		keyFamilies[1] = new KeyFamily( "KF1", "name", "desc", versionDate, components );
		keyFamilies[2] = new KeyFamily( "KF0", "name1", "desc", versionDate, components );
	}

	/**
	 * Tears down the test fixture.
	 */
	protected void tearDown () {
		keyFamilies = null;
		codeLists = null;
		concepts = null;
	}

	/* ===================== Public Methods ===================== */

	/**
	 * Assembles and returns a test suite for all
	 * the test methods of this test case.
	 *
	 * @return a non-null test suite
	 */
	public static Test suite () {
		return new TestSuite( MaintenanceAgencyTest.class );
	}

	/**
	 * Runs the test case.
	 */
	public static void main ( final String args[] ) {
		final String[] testCaseName = { MaintenanceAgencyTest.class.getName() };
		junit.textui.TestRunner.main( testCaseName );
	}

	/* --------------------- Test Methods --------------------- */

	/**
	 * Tests the normal fields.
	 */
	public void testFields () {
		MaintenanceAgency a1 = new MaintenanceAgency( "ECB" );

		assertEquals( "ECB", a1.getId() );
	}

	/**
	 * Tests the code list methods.
	 */
	public void testCodeList () {
		MaintenanceAgency a1 = new MaintenanceAgency( "ECB" );

		a1.putCodeList( codeLists[0] );
		a1.putCodeList( codeLists[1] );
		assertEquals(
			"0", codeLists[0], a1.getCodeList( codeLists[0].getId() )
		);
		assertEquals(
			"1", codeLists[1], a1.getCodeList( codeLists[1].getId() )
		);

		a1.putCodeList( codeLists[2] );
		assertEquals(
			"2", codeLists[2], a1.getCodeList( codeLists[0].getId() )
		);

		{
			final Iterator it = a1.codeListIterator();
			int i = 0;
			while ( it.hasNext() ) {
				//CodeList c = (CodeList)
				it.next();
				try {
					it.remove();
					fail( "No UnsupportedOperationException." );
				} catch ( UnsupportedOperationException e ) {
				}
				++i;
			}
			assertEquals( "i", 2, i );
		}

		assertTrue( "8", a1.containsCodeList( codeLists[0].getId() ) );
		assertTrue( "9", a1.containsCodeList( codeLists[1].getId() ) );
	}

	/**
	 * Tests the concept methods.
	 */
	public void testConcept () {
		MaintenanceAgency a1 = new MaintenanceAgency( "ECB" );

		a1.putStatisticalConcept( concepts[0] );
		a1.putStatisticalConcept( concepts[1] );
		assertEquals(
			"0", concepts[0], a1.getStatisticalConcept( concepts[0].getId() )
		);
		assertEquals(
			"1", concepts[1], a1.getStatisticalConcept( concepts[1].getId() )
		);

		a1.putStatisticalConcept( concepts[2] );
		assertEquals(
			"2", concepts[2], a1.getStatisticalConcept( concepts[0].getId() )
		);

		{
			final Iterator it = a1.statisticalConceptIterator();
			int i = 0;
			while ( it.hasNext() ) {
				//StatisticalConcept c = (StatisticalConcept)
				it.next();
				try {
					it.remove();
					fail( "No UnsupportedOperationException." );
				} catch ( UnsupportedOperationException e ) {
				}
				++i;
			}
			assertEquals( "i", 2, i );
		}

		assertTrue( "8", a1.containsStatisticalConcept( concepts[0].getId() ) );
		assertTrue( "9", a1.containsStatisticalConcept( concepts[1].getId() ) );
	}

	/**
	 * Tests the key family methods.
	 */
	public void testKeyFamilies () {
		MaintenanceAgency a1 = new MaintenanceAgency( "ECB" );

		a1.putKeyFamily( keyFamilies[0] );
		a1.putKeyFamily( keyFamilies[1] );
		assertEquals(
			"0", keyFamilies[0], a1.getKeyFamily( keyFamilies[0].getId() )
		);
		assertEquals(
			"1", keyFamilies[1], a1.getKeyFamily( keyFamilies[1].getId() )
		);

		a1.putKeyFamily( keyFamilies[2] );
		assertEquals(
			"2", keyFamilies[2], a1.getKeyFamily( keyFamilies[0].getId() )
		);

		{
			final Iterator it = a1.keyFamilyIterator();
			int i = 0;
			while ( it.hasNext() ) {
				//KeyFamily c = (KeyFamily)
				it.next();
				try {
					it.remove();
					fail( "No UnsupportedOperationException." );
				} catch ( UnsupportedOperationException e ) {
				}
				++i;
			}
			assertEquals( "i", 2, i );
		}

		assertTrue( "8", a1.containsKeyFamily( keyFamilies[0].getId() ) );
		assertTrue( "9", a1.containsKeyFamily( keyFamilies[1].getId() ) );
	}
}
