// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.metabase;

import java.lang.String;

import java.util.Map;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Collections;

import Int.ecb.logging.Logger;
import Int.ecb.logging.Level;

import Int.ecb.gesmescb.InterchangeBegin;
import Int.ecb.gesmescb.InterchangeEnd;
import Int.ecb.gesmescb.MessageBegin;
import Int.ecb.gesmescb.MessageEnd;
import Int.ecb.gesmescb.CodeList;
import Int.ecb.gesmescb.StatisticalConcept;
import Int.ecb.gesmescb.KeyFamily;

import Int.ecb.gesmescb.event.ErrorEvent;
import Int.ecb.gesmescb.event.GESMESCBEventListenerAdapter;
import Int.ecb.gesmescb.event.StructureListener;

import Int.ecb.gesmescb.metabase.MaintenanceAgency;

/* ***************************** Class ****************************** */

/**
 * <P>Holds a set of {@link MaintenanceAgency} references.</P>
 *
 * <P>The <CODE>Metabase</CODE> is not internally synchronized
 * because it should make its results immediately available.
 * However, the Java Virtual Memory Model may make external synchronisation
 * necessary.</P>
 *
 * @see Int.ecb.gesmescb.event.ThreadPump
 *
 * @author
 * <A HREF="mailto:klo@FAME.com">Kam Lo</A>
 * @author
 * <A HREF="mailto:Sami.Airo@ECB.int">Sami Airo</A>
 * @author
 * <A HREF="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</A>
 *
 * @version @(#)$Id: Metabase.java,v 1.1 2017/01/19 16:14:05 trainot Exp $
 */
public final class Metabase extends GESMESCBEventListenerAdapter
	implements StructureListener
{

	/* ======================= Constants ======================== */

	/**
	 * The reference to the MetaBase singleton used.
	 * The method {@link #getMetabase} returns its value.
	 */
	// Since double checked locking is not guaranteed to work in the
	// Java virtual memory model, initialisation at class loading is
	// the best alternative.
	private static final Metabase metabase = new Metabase();

	/** Class name. */
	private final String CLASS_NAME = getClass().getName();

	/* ========================= Fields ========================= */

	/**
	 * Mapping of maintenance agency identifiers to maintenance agencies
	 */
	private Map maintenanceAgencyMap;

	/**
	 * Current maintenance agency.
	 */
	private MaintenanceAgency maintenanceAgencyCur;

	/**
	 * Logger.
	 */
	private final Logger logger;

	/* ====================== Constructors ====================== */

	/**
	 * Construct a new empty Metabase.
	 */
	protected Metabase () {
		logger = Logger.getLogger( CLASS_NAME );
		logger.log(
			Level.FINE, CLASS_NAME, "Metabase", "Metabase constructed."
		);

		clear();
	}

	/* ===================== Public Methods ===================== */

	/**
	 * Potentially initialize the MetaBase singleton
	 * and get a reference to it.
	 *
	 * @return a reference to the MetaBase singleton.
	 */
	public static Metabase getMetabase () {
		return metabase;
	}

	/* --------------------- Query Methods ---------------------- */

	/**
	 * Returns the maintenance agency for a maintenance agency identifier.
	 * Returns <CODE>null</CODE>
	 * if the Metabase contains no maintenance agency for this identifier.
	 *
	 * @param maintenanceAgencyId maintenance agency identifier.
	 *
	 * @return the maintenence agency for a maintenance agency identifier.
	 */
	public MaintenanceAgency getMaintenanceAgency (
		final String maintenanceAgencyId
	) {
		return (MaintenanceAgency)maintenanceAgencyMap.get(
			maintenanceAgencyId
		);
	}

	/**
	 * Returns true if this metabase contains
	 * a maintenance agency with the specified identifier.
	 *
	 * @param maintenanceAgencyId identifies the maintenance agency
	 * whose presence in this metabase is to be tested.
	 *
	 * @return true if this metabase contains
	 * a maintenance agency with the specified identifier.
	 */
	public boolean containsMaintenanceAgency (
		final String maintenanceAgencyId
	) {
		return maintenanceAgencyMap.containsKey(
			maintenanceAgencyId
		);
	}

	/**
	 * Returns an iterator over the maintenance agencies in this metabase.
	 *
	 * @return an iterator over the maintenance agencies in this metabase.
	 */
	public Iterator maintenanceAgencyIterator () {
		return Collections.unmodifiableCollection(
			maintenanceAgencyMap.values()
		).iterator();
	}

	/**
	 * Returns the number of identifier-maintenance agency mappings
	 * in this metabase.
	 *
	 * @return the number of identifier-maintenance agency mappings.
	 */
	public int getMaintenanceAgencyCount () {
		return maintenanceAgencyMap.size();
	}

	/* ------------------------- clear -------------------------- */

	/**
	 * Deletes all information from the metabase.
	 */
	public void clear () {
		maintenanceAgencyMap = new HashMap();
		maintenanceAgencyCur = null;
	}

	/* --------------------- Object Methods --------------------- */

	/**
	 * <P>Returns a string representation of this <CODE>Metabase</CODE>.</P>
	 *
	 * <P>This method is intended to be used only for debugging purposes,
	 * and the format of the returned string may vary between
	 * implementations.</P>
	 *
	 * <P>The returned string may be empty but may not be null.</P>
	 *
	 * @return a string representation of this <CODE>Metabase</CODE>.
	 */
	public String toString () {
		return getClass().getName()
			+ "( current MaintenanceAgencies = " + maintenanceAgencyMap + " )";
	}

	/* ------------------------ Listener ------------------------ */

	public void handle ( final InterchangeBegin interchangeBegin ) {
		logger.log(
			Level.FINE, CLASS_NAME,
			"handle", "InterchangeBegin", interchangeBegin
		);
	}

	public void handle ( final InterchangeEnd interchangeEnd ) {
		logger.log(
			Level.FINE, CLASS_NAME,
			"handle","InterchangeEnd", interchangeEnd
		);
	}

	public void handle ( final MessageBegin messageBegin ) {
		logger.log(
			Level.FINE, CLASS_NAME,
			"handle", "MessageBegin", messageBegin
		);

		final String agencyId = messageBegin.getMaintenanceAgencyId();

		if ( !maintenanceAgencyMap.containsKey( agencyId ) ) {
			maintenanceAgencyMap.put(
				agencyId,
				new MaintenanceAgency( agencyId )
			);
			logger.log(
					Level.FINEST, CLASS_NAME,
					"handle", "MessageBegin - put agencyId:" + agencyId, messageBegin
				);
			
		}

		maintenanceAgencyCur
		= (MaintenanceAgency)maintenanceAgencyMap.get( agencyId );
	}

	public void handle ( final MessageEnd messageEnd ) {
		logger.log(
			Level.FINE, CLASS_NAME,
			"handle", "MessageEnd", messageEnd
		);

		maintenanceAgencyCur = null;
	}

	public void handle ( final CodeList codeList ) {
		logger.log(
			Level.FINE, CLASS_NAME,
			"handle", "CodeList", codeList
		);

		if ( maintenanceAgencyCur != null ) {
			maintenanceAgencyCur.putCodeList( codeList );
		} else {
			logger.log(
				Level.SEVERE, CLASS_NAME,
				"handle", "CodeList, but no current MaintenanceAgency"
			);
		}
	}

	public void handle ( final StatisticalConcept statisticalConcept ) {
		logger.log(
			Level.FINE, CLASS_NAME,
			"handle", "StatisticalConcept", statisticalConcept
		);

		if ( maintenanceAgencyCur != null ) {
			maintenanceAgencyCur.putStatisticalConcept( statisticalConcept );
		} else {
			logger.log(
				Level.SEVERE, CLASS_NAME,
				"handle", "StatisticalConcept, but no current MaintenanceAgency"
			);
		}
	}

	public void handle ( final KeyFamily keyFamily ) {
		logger.log(
			Level.FINE, CLASS_NAME,
			"handle", "KeyFamily", keyFamily
		);

		if ( maintenanceAgencyCur != null ) {
			maintenanceAgencyCur.putKeyFamily( keyFamily );
		} else {
			logger.log(
				Level.SEVERE, CLASS_NAME,
				"handle", "keyFamily, but no current MaintenanceAgency"
			);
		}
	}

	public void handle ( final ErrorEvent errorEvent ) {
		logger.log(
			Level.FINE, CLASS_NAME,
			"handle", "ErrorEvent", errorEvent
		);
	}
}
