// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.semanticchecker;

import Int.ecb.logging.Level;

import Int.ecb.gesmescb.*;
import Int.ecb.gesmescb.event.ErrorEvent;
import Int.ecb.gesmescb.event.GESMESCBEventBroker;

import Int.ecb.gesmescb.metabase.Metabase;
import Int.ecb.gesmescb.metabase.MaintenanceAgency;

/* ***************************** Class ****************************** */

/**
 * <P>Provides basic features for the semantic checker hierarchy.</P>
 *
 * <p>As a <code>GESMESCBEventBroker</code>, it
 * always has auto disconnect disabled, because it should run as long
 * as possible.</P>
 *
 * @author
 * <A HREF="mailto:klo@FAME.com">Kam Lo</A>
 * @author
 * <A HREF="mailto:Sami.Airo@ECB.int">Sami Airo</A>
 * @author
 * <A HREF="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</A>
 *
 * @see SemanticChecker
 *
 * @version @(#)$Id: CheckerBase.java,v 1.1 2017/01/19 16:14:05 trainot Exp $
 */
public abstract class CheckerBase extends GESMESCBEventBroker {

	/* ========================= Fields ========================= */

	/** The maintenance agency id the current message refers to. */
	private String maintenanceAgencyId;

	/** The maintenance agency the current message refers to. */
	private MaintenanceAgency maintenanceAgency;

	/* ====================== Constructors ====================== */

	/**
	 * Construct a new instance.
	 */
	public CheckerBase () {
		super( "", false );
		maintenanceAgencyId = null;
	}

	/* ===================== Public Methods ===================== */

	/**
	 * <P>Marks the end of transmission.</P>
	 *
	 * <P>This method should be called when all events have
	 * been processed and a final statement about errors
	 * found while checking is desired.</P>
	 *
	 * <P>In of a somehow incomplete transmission,
	 * one or more * {@link Int.ecb.gesmescb.event.ErrorEvent} objects
	 * will be dispatched to the listeners.</P>
	 *
	 * <P>The main purpose of this method is to be able to
	 * check a data flow with missing interchange end event,
	 * say. Without this method, the semantic checker would not
	 * be able to detect such errors.</P>
	 *
	 * <P>{@link ErrorEvent}s may be dispatched accordingly.</P>
	 *
	 * <P>A subclass overriding this method should always first
	 * call its superclass' method and then proceed with its own checks.<P>
	 *
	 * @see Int.ecb.gesmescb.event.GESMESCBReader#close
	 * @see Int.ecb.gesmescb.event.GESMESCBWriter#close
	 */
	public void close () {
		componentHelper.logger.log(
			Level.FINE, getName(),
			"close", "Check close."
		);
	}

	/* ------------------------ Listener ------------------------ */

	/**
	 * Check and then forward an <CODE>InterchangeBegin</CODE>.
	 *
	 * @param interchangeBegin to be processed.
	 */
	public final void handle (
		final InterchangeBegin interchangeBegin
	) {
		componentHelper.logger.log(
			Level.FINE, getName(),
			"handle", "Check InterchangeBegin.",
			interchangeBegin
		);

		check( interchangeBegin );

		componentHelper.logger.log(
			Level.FINE, getName(),
			"handle", "InterchangeBegin checked.",
			interchangeBegin
		);

		dispatch( interchangeBegin );
	}

	/**
	 * Check and then forward an <CODE>InterchangeEnd</CODE>.
	 *
	 * @param interchangeEnd to be processed.
	 */
	public final void handle (
		final InterchangeEnd interchangeEnd
	) {
		componentHelper.logger.log(
			Level.FINE, getName(),
			"handle", "Check InterchangeEnd.",
			interchangeEnd
		);

		check( interchangeEnd );

		componentHelper.logger.log(
			Level.FINE, getName(),
			"handle", "InterchangeEnd checked.",
			interchangeEnd
		);

		dispatch( interchangeEnd );
	}

	/**
	 * Check and then forward a <CODE>MessageBegin</CODE>.
	 *
	 * @param messageBegin to be processed.
	 */
	public final void handle (
		final MessageBegin messageBegin
	) {
		componentHelper.logger.log(
			Level.FINE, getName(),
			"handle", "Check MessageBegin.",
			messageBegin
		);

		check( messageBegin );

		componentHelper.logger.log(
			Level.FINE, getName(),
			"handle", "MessageBegin checked.",
			messageBegin
		);

		dispatch( messageBegin );
	}

	/**
	 * Check and then forward a <CODE>MessageEnd</CODE>.
	 *
	 * @param messageEnd to be processed.
	 */
	public final void handle (
		final MessageEnd messageEnd
	) {
		componentHelper.logger.log(
			Level.FINE, getName(),
			"handle", "Check MessageEnd.",
			messageEnd
		);

		check( messageEnd );

		componentHelper.logger.log(
			Level.FINE, getName(),
			"handle", "MessageEnd checked.",
			messageEnd
		);

		dispatch( messageEnd );
	}

	/**
	 * Check and then forward a <CODE>CodeList</CODE>.
	 *
	 * @param codeList to be processed.
	 */
	public final void handle (
		final CodeList codeList
	) {
		componentHelper.logger.log(
			Level.FINE, getName(),
			"handle", "Check CodeList.",
			codeList
		);

		check( codeList );

		componentHelper.logger.log(
			Level.FINE, getName(),
			"handle", "CodeList checked.",
			codeList
		);

		dispatch( codeList );
	}

	/**
	 * Check and then forward a <CODE>StatisticalConcept</CODE>.
	 *
	 * @param statisticalConcept to be processed.
	 */
	public final void handle (
		final StatisticalConcept statisticalConcept
	) {
		componentHelper.logger.log(
			Level.FINE, getName(),
			"handle", "Check StatisticalConcept.",
			statisticalConcept
		);

		check( statisticalConcept );

		componentHelper.logger.log(
			Level.FINE, getName(),
			"handle", "StatisticalConcept checked.",
			statisticalConcept
		);

		dispatch( statisticalConcept );
	}

	/**
	 * Check and then forward a <CODE>KeyFamily</CODE>.
	 *
	 * @param keyFamily to be processed.
	 */
	public final void handle (
		final KeyFamily keyFamily
	) {
		componentHelper.logger.log(
			Level.FINE, getName(),
			"handle", "Check KeyFamily.",
			keyFamily
		);

		check( keyFamily );

		componentHelper.logger.log(
			Level.FINE, getName(),
			"handle", "KeyFamily checked.",
			keyFamily
		);

		dispatch( keyFamily );
	}

	/**
	 * Check and then forward a <CODE>TimeSeries</CODE>.
	 *
	 * @param timeSeries to be processed.
	 */
	public final void handle (
		final TimeSeries timeSeries
	) {
		componentHelper.logger.log(
			Level.FINE, getName(),
			"handle", "CheckerBase: Check TimeSeries.",
			timeSeries
		);

		check( timeSeries );

		componentHelper.logger.log(
			Level.FINE, getName(),
			"handle", "CheckerBase: TimeSeries checked.",
			timeSeries
		);

		dispatch( timeSeries );
	}

	/**
	 * Check and then forward an <CODE>Attribute</CODE>.
	 *
	 * @param attribute to be processed.
	 */
	public final void handle (
		final Attribute attribute
	) {
		componentHelper.logger.log(
			Level.FINE, getName(),
			"handle", "Check Attribute.",
			attribute
		);

		check( attribute );

		componentHelper.logger.log(
			Level.FINE, getName(),
			"handle", "Attribute checked.",
			attribute
		);

		dispatch( attribute );
	}

	/**
	 * Forward an <CODE>ErrorEvent</CODE>.
	 *
	 * @param errorEvent to be processed.
	 */
	public final void handle (
		final ErrorEvent errorEvent
	) {
		componentHelper.logger.log(
			Level.FINE, getName(),
			"handle", "Forward ErrorEvent.",
			errorEvent
		);

		dispatch( errorEvent );
	}

	/* ==================== Package Methods ===================== */

	/** Supply the maintenance agency id for the current message. */
	String getMaintenanceAgencyId () {
		return maintenanceAgencyId;
	}

	/** Supply the maintenance agency for the current message. */
	MaintenanceAgency getMaintenanceAgency () {
		if ( maintenanceAgency == null ) {
			final Metabase metabase = Metabase.getMetabase();
			maintenanceAgency = metabase.getMaintenanceAgency(
				maintenanceAgencyId
			);
		}
		return maintenanceAgency;
	}

	/* -------------------- Checking Events --------------------- */

	/**
	 * <P>Check if an InterchangeBegin is semantically correct.
	 * Generate one or more {@link ErrorEvent}s if not.</P>
	 *
	 * <P>A subclass overriding this method should always first
	 * call its superclass' method and then proceed with its own checks.<P>
	 *
	 * @param interchangeBegin to be checked.
	 */
	void check (
		final InterchangeBegin interchangeBegin
	) { /* empty body to be overridden */ }

	/**
	 * <P>Check if an InterchangeEnd is semantically correct.
	 * Generate one or more {@link ErrorEvent}s if not.</P>
	 *
	 * <P>A subclass overriding this method should always first
	 * call its superclass' method and then proceed with its own checks.<P>
	 *
	 * @param interchangeEnd to be checked.
	 */
	void check (
		final InterchangeEnd interchangeEnd
	) { /* empty body to be overridden */ }

	/**
	 * <P>Check if an MessageBegin is semantically correct.
	 * Generate one or more {@link ErrorEvent}s if not.</P>
	 *
	 * <P>A subclass overriding this method should always first
	 * call its superclass' method and then proceed with its own checks.<P>
	 *
	 * @param messageBegin to be checked.
	 */
	void check (
		final MessageBegin messageBegin
	) {
		maintenanceAgencyId = messageBegin.getMaintenanceAgencyId();
	}


	/**
	 * <P>Check if an MessageEnd is semantically correct.
	 * Generate one or more {@link ErrorEvent}s if not.</P>
	 *
	 * <P>A subclass overriding this method should always first
	 * call its superclass' method and then proceed with its own checks.<P>
	 *
	 * @param messageEnd to be checked.
	 */
	void check (
		final MessageEnd messageEnd
	) { /* empty body to be overridden */ }

	/**
	 * <P>Check if an CodeList is semantically correct.
	 * Generate one or more {@link ErrorEvent}s if not.</P>
	 *
	 * <P>A subclass overriding this method should always first
	 * call its superclass' method and then proceed with its own checks.<P>
	 *
	 * @param codeList to be checked.
	 */
	void check (
		final CodeList codeList
	) { /* empty body to be overridden */ }

	/**
	 * <P>Check if an StatisticalConcept is semantically correct.
	 * Generate one or more {@link ErrorEvent}s if not.</P>
	 *
	 * <P>A subclass overriding this method should always first
	 * call its superclass' method and then proceed with its own checks.<P>
	 *
	 * @param statisticalConcept to be checked.
	 */
	void check (
		final StatisticalConcept statisticalConcept
	) { /* empty body to be overridden */ }

	/**
	 * <P>Check if an KeyFamily is semantically correct.
	 * Generate one or more {@link ErrorEvent}s if not.</P>
	 *
	 * <P>A subclass overriding this method should always first
	 * call its superclass' method and then proceed with its own checks.<P>
	 *
	 * @param keyFamily to be checked.
	 */
	void check (
		final KeyFamily keyFamily
	) { /* empty body to be overridden */ }

	/**
	 * <P>Check if an TimeSeries is semantically correct.
	 * Generate one or more {@link ErrorEvent}s if not.</P>
	 *
	 * <P>A subclass overriding this method should always first
	 * call its superclass' method and then proceed with its own checks.<P>
	 *
	 * @param timeSeries to be checked.
	 */
	void check (
		final TimeSeries timeSeries
	) { /* empty body to be overridden */ }

	/**
	 * <P>Check if an Attribute is semantically correct.
	 * Generate one or more {@link ErrorEvent}s if not.</P>
	 *
	 * <P>A subclass overriding this method should always first
	 * call its superclass' method and then proceed with its own checks.<P>
	 *
	 * @param attribute to be checked.
	 */
	void check (
		final Attribute attribute
	) { /* empty body to be overridden */ }
}
