// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.semanticchecker;

import java.util.*;

import java.io.StringWriter;
import java.io.PrintWriter;

import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;

import Int.ecb.gesmescb.*;
import Int.ecb.gesmescb.metabase.Metabase;

import Int.ecb.gesmescb.event.ErrorEvent;
import Int.ecb.gesmescb.event.TestEventListener;
import Int.ecb.gesmescb.event.TestEventSource;

/* ***************************** Class ****************************** */

/**
 * Provide common features for tests of subclasses of {@link CheckerBase}.
 *
 * @author
 * <a href="mailto:sis.external@ecb.europa.eu">Bernhard Bodenstorfer</a>
 *
 * @version @(#)$Id: CheckerBaseTest.java,v 1.1 2017/01/19 16:14:05 trainot Exp $
 */
public abstract class CheckerBaseTest extends TestCase {

	/* ======================= Constants ======================== */

	/**
	 * To mark asserted failures in the test output data.
	 * @see doTest(boolean,Object[]).
	 */
	static final ErrorEvent FAILURE = new ErrorEvent(
		"Test failure", ErrorEvent.FAILURE
	);

	/**
	 * To mark asserted errors in the test output data.
	 * @see doTest(boolean,Object[]).
	 */
	static final ErrorEvent ERROR = new ErrorEvent(
		"Test error", ErrorEvent.ERROR
	);

	/**
	 * To mark asserted warnings in the test output data.
	 * @see doTest(boolean,Object[]).
	 */
	static final ErrorEvent WARNING = new ErrorEvent(
		"Test warning", ErrorEvent.WARNING
	);

	/* ----------------------- Test Data ------------------------ */

	protected final InterchangeBegin
		INTERCHANGE_BEGIN = new InterchangeBegin(
			"IREF000001", "4F0", "ZZZ", "010329:1430", false
		);

	protected final InterchangeEnd
		INTERCHANGE_END = new InterchangeEnd( "IREF000001" );

	protected final Contact[] contacts = new Contact[] {
		ContactTest.contact
	};

	protected final ContactChannel[] channels = new ContactChannel[] {
		ContactChannelTest.contactChannel
	};

	protected final MessageBegin
		METADATA_MESSAGE_BEGIN = new MessageBegin(
			"MREF000001", "4F0", "ZZZ", "ECB",
			MessageBegin.STATISTICAL_DEFINITIONS,
			"structural definitions", contacts
		),
		DATA_MESSAGE_BEGIN = new MessageBegin(
			"MREF000001", "ZZZ", "4F0", "ECB",
			MessageBegin.STATISTICAL_DATA,
			"operational data", contacts,
			"KF1", MessageBegin.UPDATE_OR_REPLACE_ACTION,
			"200103291423", "19001999", "702", "KF1"
		);

	protected final MessageEnd
		METADATA_MESSAGE_END = new MessageEnd( "MREF000001" ),
		DATA_MESSAGE_END = new MessageEnd( "MREF000001" );

	protected static final Date versionDate;
	static {
		final Calendar calendar = Calendar.getInstance();
		calendar.clear();
		versionDate = calendar.getTime();
	}

	protected static final CodeList
		CL_ORGANISATION = new CodeList(
			"CL_ORGANISATION",
			"Organisations code list",
			"Lists all organisations",
			versionDate,
			new String[] { "4F0", "ZZZ" },
			new String[] { "European Central Bank (ECB)", "Unspecified" }
		),
		CL_AGENCY = new CodeList(
			"CL_AGENCY",
			"Maintenance agencies code list",
			"Lists all maintenance agencies",
			versionDate,
			new String[] { "ECB", "EUROSTAT", "BIS" },
			new String[] {
				"European Central Bank (ECB)",
				"Eurostat",
				"Bank for International Settlements"
			}
		),
		CL_FREQ = new CodeList(
			"CL_FREQ", "frequency codes", "part of setup", versionDate,
			new String[] { "A", "H", "Q", "M", "W", "D" },
			new String[] {
				"Annual", "Semiannual", "Quarterly",
				"Monthly", "Weekly", "Daily"
			}
		),
		CL_DIM1 = new CodeList(
			"CL_DIM1", "test cl 1", "part of setup", versionDate,
			new String[] { "A", "B", "C" },
			new String[] { "Alpha", "Beta", "Gamma" }
		),
		CL_DIM2 = new CodeList(
			"CL_DIM2", "test cl 2", "part of setup", versionDate,
			new String[] { "WO", "DO", "LO" },
			new String[] { "Wodan", "Donar", "Loki" }
		),
		CL_OBS_CONF = new CodeList(
			"CL_OBS_CONF", "test cl obs_conf", "part of setup", versionDate,
			new String[] { "C", "F", "N", "R" },
			new String[] {
				"Confidential", "Free", "Nonpublishable", "Personal"
			}
		),
		CL_OBS_STATUS = new CodeList(
			"CL_OBS_STATUS", "test cl obs_status", "part of setup", versionDate,
			new String[] { "A", "B", "E", "F", "H" },
			new String[] {
				"Normal", "Break", "Estimate", "Forecast", "Holiday"
			}
		),
		CL_ATT1 = new CodeList(
			"CL_ATT1", "test cl 3", "part of setup", versionDate,
			new String[] { "X", "Y", "Z" },
			new String[] { "Xi", "Ypsilon", "Zeta" }
		),
		CL_ATT2 = new CodeList(
			"CL_ATT2", "test cl 4", "part of setup", versionDate,
			new String[] { "JU", "SA", "UR", "PL" },
			new String[] { "Jupiter", "Saturn", "Uranus", "Pluto" }
		);

	protected static final StatisticalConcept
		SC_FREQ = new StatisticalConcept(
			"FREQ", "Frequency", "part of setup", versionDate
		),
		SC_DIM1 = new StatisticalConcept(
			"DIM1", "test dim 1", "part of setup", versionDate
		),
		SC_DIM2 = new StatisticalConcept(
			"DIM2", "test dim 2", "part of setup", versionDate
		),
		SC_TIME_PERIOD = new StatisticalConcept(
			"TIME_PERIOD", "test sc tp", "part of setup", versionDate
		),
		SC_TIME_FORMAT = new StatisticalConcept(
			"TIME_FORMAT", "test sc tf", "part of setup", versionDate
		),
		SC_OBS_VALUE = new StatisticalConcept(
			"OBS_VALUE", "test obs_value", "part of setup", versionDate
		),
		SC_OBS_STATUS = new StatisticalConcept(
			"OBS_STATUS", "test obs_status", "part of setup", versionDate
		),
		SC_OBS_CONF = new StatisticalConcept(
			"OBS_CONF", "test obs_conf", "part of setup", versionDate
		),
		SC_OBS_PRE_BREAK = new StatisticalConcept(
			"OBS_PRE_BREAK", "test obs_pre_break", "part of setup", versionDate
		),
		SC_ATT1 = new StatisticalConcept(
			"ATT1", "test att 1", "part of setup", versionDate
		),
		SC_ATT2 = new StatisticalConcept(
			"ATT2", "test att 2", "part of setup", versionDate
		),
		SC_ATT3 = new StatisticalConcept(
			"ATT3", "test att 3", "part of setup", versionDate
		),
		SC_ATT4 = new StatisticalConcept(
			"ATT4", "test att 4", "part of setup", versionDate
		),
		SC_EXTRA = new StatisticalConcept(
			"EXTRA", "test sc extra", "part of setup", versionDate
		);

	protected static final KeyFamilyComponent
		FREQ = new KeyFamilyComponent(
			"FREQ", KeyFamilyComponent.KEY_STRUCTURE_DIMENSION,
			"1", "AN..1", "", "", "CL_FREQ"
		),
		DIM1 = new KeyFamilyComponent(
			"DIM1", KeyFamilyComponent.KEY_STRUCTURE_DIMENSION,
			"2", "AN1", "", "", "CL_DIM1"
		),
		DIM2 = new KeyFamilyComponent(
			"DIM2", KeyFamilyComponent.KEY_STRUCTURE_DIMENSION,
			"3", "AN2", "", "", "CL_DIM2"
		),
		TIME_PERIOD = new KeyFamilyComponent(
			"TIME_PERIOD", KeyFamilyComponent.TIME_DIMENSION,
			"4", "AN..35", "", "", ""
		),
		TIME_FORMAT = new KeyFamilyComponent(
			"TIME_FORMAT", KeyFamilyComponent.TIME_DIMENSION,
			"5", "AN3", "", "", ""
		),
		OBS_VALUE = new KeyFamilyComponent(
			"OBS_VALUE", KeyFamilyComponent.ARRAY_CELL,
			"6", "AN..15", "", "", ""
		),
		OBS_STATUS = new KeyFamilyComponent(
			"OBS_STATUS", KeyFamilyComponent.ARRAY_CELL,
			"7", "AN1", KeyFamilyComponent.MANDATORY,
			KeyFamilyComponent.ATTACHED_TO_OBSERVATION, "CL_OBS_STATUS"
		),
		OBS_CONF = new KeyFamilyComponent(
			"OBS_CONF", KeyFamilyComponent.ARRAY_CELL,
			"8", "AN1", KeyFamilyComponent.CONDITIONAL,
			KeyFamilyComponent.ATTACHED_TO_OBSERVATION, "CL_OBS_CONF"
		),
		OBS_PRE_BREAK = new KeyFamilyComponent(
			"OBS_PRE_BREAK", KeyFamilyComponent.ARRAY_CELL,
			"9", "AN..15", KeyFamilyComponent.CONDITIONAL,
			KeyFamilyComponent.ATTACHED_TO_OBSERVATION, ""
		),
		ATT1 = new KeyFamilyComponent(
			"ATT1", KeyFamilyComponent.ATTRIBUTE,
			"", "AN..20", KeyFamilyComponent.CONDITIONAL,
			KeyFamilyComponent.ATTACHED_TO_OBSERVATION, "CL_ATT1"
		),
		ATT2 = new KeyFamilyComponent(
			"ATT2", KeyFamilyComponent.ATTRIBUTE,
			"", "A2", KeyFamilyComponent.MANDATORY,
			KeyFamilyComponent.ATTACHED_TO_TIME_SERIES, "CL_ATT2"
		),
		ATT3 = new KeyFamilyComponent(
			"ATT3", KeyFamilyComponent.ATTRIBUTE,
			"", "N..10", KeyFamilyComponent.CONDITIONAL,
			KeyFamilyComponent.ATTACHED_TO_SIBLING_GROUP, ""
		),
		ATT4 = new KeyFamilyComponent(
			"ATT4", KeyFamilyComponent.ATTRIBUTE,
			"", "A..10", KeyFamilyComponent.MANDATORY,
			KeyFamilyComponent.ATTACHED_TO_DATA_SET, ""
		);

	protected static final KeyFamilyComponent[]
		KFC_ARR = new KeyFamilyComponent[] {
			FREQ,
			DIM1,
			DIM2,
			TIME_PERIOD,
			TIME_FORMAT,
			OBS_VALUE,
			OBS_STATUS,
			OBS_CONF,
			OBS_PRE_BREAK,
			ATT1,
			ATT2,
			ATT3,
			ATT4
		};

	protected static final KeyFamily
		KF1 = new KeyFamily(
			"KF1", "Name", "Description", versionDate, KFC_ARR
		);

	protected static final Observation[]
		OBS_ARR = new Observation[] {
			new Observation( "13.6", "A", "F", "" ),
			new Observation( "13.7", "A", "F", "" ),
			new Observation( "13.8", "A", "F", "" )
		};

	protected static final TimeSeries
		TS1 = new TimeSeries( "M.B.LO", "199010199012", "710", OBS_ARR );

	protected static final Attribute
		AT1 = new Attribute(
			"M.B.LO", "199001", "610", "ATT1", "Z"
		),
		AT2 = new Attribute(
			"M.B.LO", "", "", "ATT2", "JU"
		),
		AT3 = new Attribute(
			".B.LO", "", "", "ATT3", "123"
		),
		AT4 = new Attribute(
			"", "", "", "ATT4", "Irrdzdz"
		);

	/* ========================= Fields ========================= */

	/**
	 * Switch for debugging.
	 * This is set in the constructor.
	 * Subclasses may choose to set this in their constructor to a different
	 * value in order to facilitate debugging once a test case fails.
	 */
	protected boolean debug;

	private boolean metabaseInitialized;

	/* ====================== Constructors ====================== */

	/**
	 * Constructs a CheckerBaseTest with the specified name.
	 *
	 * @param name Test case name
	 */
	public CheckerBaseTest ( String name ) {
		super( name );
		debug = false;
		metabaseInitialized = false;
	}

	/* ================== Protected Methods ===================== */

	/**
	 * Sets up the test fixture.
	 */
	protected void setUp () {
		// Preload metabase with default test data
		if ( !metabaseInitialized ) {
			testAll();
			metabaseInitialized = true;
		}
	}

	/**
	 * Produce a copy with different MREF number.
	 *
	 * @param INTERCHANGE_BEGIN to be cloned with different MREF.
	 * @param iref to be used.
	 *
	 * @return the appropriate changed clone.
	 */
	protected static InterchangeBegin getWithRef(
		final InterchangeBegin INTERCHANGE_BEGIN, final String iref
	) {
		return new InterchangeBegin(
			iref,
			INTERCHANGE_BEGIN.getSender(),
			INTERCHANGE_BEGIN.getReceiver(),
			INTERCHANGE_BEGIN.getPreparationDateTime(),
			INTERCHANGE_BEGIN.isTest()
		);
	}

	/**
	 * Produce a copy with different MREF number.
	 * This function serves symmetry.
	 *
	 * @param INTERCHANGE_END to be cloned with different MREF.
	 * @param iref to be used.
	 *
	 * @return the appropriate changed clone.
	 */
	protected static InterchangeEnd getWithRef(
		final InterchangeEnd INTERCHANGE_END, final String iref
	) {
		return new InterchangeEnd( iref );
	}

	/**
	 * Produce a copy with different MREF number.
	 *
	 * @param messageBegin to be cloned with different MREF.
	 * @param mref to be used.
	 *
	 * @return the appropriate changed clone.
	 */
	protected static MessageBegin getWithRef(
		final MessageBegin messageBegin, final String mref
	) {
		final List contacts = new ArrayList();
		for (
			final Iterator iterator = messageBegin.contactIterator();
			iterator.hasNext();
		) {
			contacts.add( iterator.next() );
		}

		return new MessageBegin(
			mref,
			messageBegin.getSender(),
			messageBegin.getReceiver(),
			messageBegin.getMaintenanceAgencyId(),
			messageBegin.getFunction(),
			messageBegin.getIdentity(),
			(Contact[]) contacts.toArray( new Contact[] {} ),
			messageBegin.getDataSetId(),
			messageBegin.getAction(),
			messageBegin.getExtractionDateTime(),
			messageBegin.getReportingPeriod(),
			messageBegin.getReportingPeriodFormat(),
			messageBegin.getKeyFamilyId()
		);
	}

	/**
	 * Produce a copy with different MREF number.
	 * This function serves symmetry.
	 *
	 * @param messageEnd to be cloned with different MREF.
	 * @param mref to be used.
	 *
	 * @return the appropriate changed clone.
	 */
	protected static MessageEnd getWithRef(
		final MessageEnd messageEnd, final String mref
	) {
		return new MessageEnd( mref );
	}

	/**
	 * Factory for CheckerBases.
	 * Should be overridden by derived classes.
	 *
	 * @see doTest(boolean,Object[])
	 */
	protected abstract CheckerBase checkerFactory ();

	/**
	 * <P>Test reaction to a sequence of objects.</P>
	 *
	 * <P>If the checker is not expected to generate errors or warnings,
	 * the parameter <CODE>objects</CODE> simply is an array
	 * of objects, usually {@link Int.ecb.gesmescb.GESMESCBObject}s
	 * with optional entry <CODE>null</CODE> to signal
	 * {@link SemanticChecker#close}.</P>
	 *
	 * <P>If the Checker is expected to generate errors or warnings,
	 * only the first one needs to be marked by inserting
	 * {@link #ERROR} or {@link #WARNING} after the offending object.
	 * Only in the case where one object should raise a warning and
	 * only a later one should raise an error, the first warning and
	 * the first error must be marked after the corresponding
	 * elements.</P>
	 *
	 * @param checker to be tested.
	 * @param checkEach whether error codes are checked
	 * on a per-object basis or otherwise for the total interchange.
	 * Set true means stricter testing.
	 * Set false is intended for tests which contain only
	 * one single offending element, which is, however, so off-putting
	 * that afterwards no particular behaviour is intended to be specified,
	 * e.g. if a message begin precedes an interchange begin.
	 * @param objects to be fed to the checker.
	 */
	public void doTest (
		final CheckerBase checker,
		final boolean checkEach,
		final Object[] objects
	) {
		if ( debug ) {
			System.out.println( BackTraceException.whereAmI() );
		}

		final TestEventListener listener = new TestEventListener( 1000 );
		final TestEventSource source = new TestEventSource();

		source.addListener( checker, true );
		source.addListener( Metabase.getMetabase(), true );
		checker.addListener( listener, true );

		int expectedSeverityTillHere = ErrorEvent.INFO;

		for ( int i = 0; i < objects.length; ++i ) {
			final Object incoming = objects[i];

			int expectedSeverityHere = ErrorEvent.INFO;
			if (
				i+1 < objects.length && (
					objects[i+1] == FAILURE
					||
					objects[i+1] == ERROR
					||
					objects[i+1] == WARNING
				)
			) {
				++i;
				expectedSeverityHere = ( (ErrorEvent)objects[i] ).getSeverity();
			}
			if ( expectedSeverityTillHere < expectedSeverityHere ) {
				expectedSeverityTillHere = expectedSeverityHere;
			}

			if ( debug ) System.out.println( "SEND " + incoming );

			// dispatch
			if ( incoming != null ) {
				source.dispatch( incoming );
			} else {
				checker.close();
			}

			// compare response
			int receivedSeverity = ErrorEvent.INFO;
			for (
				final Iterator response = listener.iterator();
				response.hasNext();
			) {
				final Object utterance = response.next();
				if ( debug ) System.out.println( "REC: " + utterance );
				if ( utterance instanceof ErrorEvent ) {
					final int utteranceSeverity
					= ( (ErrorEvent)utterance ).getSeverity();
					if ( utteranceSeverity > receivedSeverity ) {
						receivedSeverity = utteranceSeverity;
					}
					if ( checkEach ) {
						assertTrue(
							"Must not raise higher than expected severity: "
							+ utterance,
							utteranceSeverity <= expectedSeverityHere
						);
					} else {
						assertTrue(
							"Must not raise more than expected severity: "
							+ utterance,
							utteranceSeverity <= expectedSeverityTillHere
						);
					}
					if ( utteranceSeverity > receivedSeverity ) {
						receivedSeverity = utteranceSeverity;
					}
					assertTrue(
						"Need finally incoming",
						incoming == null || response.hasNext()
					);
				} else {
					assertSame( incoming, utterance );
					assertTrue(
						"No object after incoming",
						!response.hasNext()
					);
				}
			}

			assertTrue(
				"Must raise expected severity",
				receivedSeverity >= expectedSeverityHere
			);

			// Reset listener
			listener.clearHistory();
		}
	}

	/**
	 * <P>Test reaction to a sequence of objects.</P>
	 *
	 * <P>Calls {@link #doTest(CheckerBase,Object[])} with the checker
	 * returned by {@link #checkerFactory}.</P>
	 *
	 * @param checkEach whether error codes are checked
	 * on a per-object basis or otherwise for the total interchange.
	 * @param objects to be fed to the checker.
	 */
	public final void doTest (
		final boolean checkEach,
		final Object[] objects
	) {
		doTest( checkerFactory(), checkEach, objects );
	}

	/**
	 * <P>Test reaction to a sequence of objects.</P>
	 *
	 * <P>Calls {@link #doTest(CheckerBase,Object[])} with the checker
	 * returned by {@link #checkerFactory}.</P>
	 */
	public final void doTest (
		final Object[] objects
	) {
		doTest( true, objects );
	}

	/* ===================== Public Methods ===================== */

	/**
	 * Assembles and returns a test suite for all
	 * the test methods of this test case.
	 *
	 * @return a non-null test suite
	 */
	public static Test suite () {
		TestSuite suite = new TestSuite( CheckerBaseTest.class );
		return suite;
	}

	/**
	 * Runs the test case
	 */
	public static void main ( String args[] ) {
		String[] testCaseName = { CheckerBaseTest.class.getName() };
		junit.textui.TestRunner.main( testCaseName );
	}

	/* ---------------------- Test Methods ---------------------- */

	/**
	 * Tests all data and metadata.
	 * This is also important to preload the Metabase
	 * for successive tests in subclasses.
	 *
	 * @see #setUp
	 */
	public void testAll () {
		final Object objects[] = {
			INTERCHANGE_BEGIN,
			METADATA_MESSAGE_BEGIN,
			CL_ORGANISATION,
			CL_AGENCY,
			CL_FREQ,
			CL_DIM1,
			CL_DIM2,
			CL_OBS_CONF,
			CL_OBS_STATUS,
			CL_ATT1,
			CL_ATT2,
			SC_FREQ,
			SC_DIM1,
			SC_DIM2,
			SC_TIME_PERIOD,
			SC_TIME_FORMAT,
			SC_OBS_VALUE,
			SC_OBS_STATUS,
			SC_OBS_CONF,
			SC_OBS_PRE_BREAK,
			SC_ATT1,
			SC_ATT2,
			SC_ATT3,
			SC_ATT4,
			SC_EXTRA,
			KF1,
			METADATA_MESSAGE_END,
			getWithRef( DATA_MESSAGE_BEGIN, "MREF000002" ),
			TS1,
			AT1,
			AT2,
			AT3,
			AT4,
			getWithRef( DATA_MESSAGE_END, "MREF000002" ),
			INTERCHANGE_END,
			null
		};

		doTest( true, objects );
	}

	/* =================== Protected Methods ==================== */
}

/** Helps identifying the current test method. */
class BackTraceException extends Exception {
	/**
	 *   serialVersionUID generated by eclipse.
	 */
	private static final long serialVersionUID = -4893862506244629570L;

	public BackTraceException () {
		super( "BacktraceException" );
	}

	public String getTestMethodName () {
		{
			final StringWriter stringWriter = new StringWriter();
			{
				final PrintWriter printWriter = new PrintWriter( stringWriter );
				printStackTrace( printWriter );
				printWriter.close();
			}
			{
				final String total = stringWriter.toString();
				final int start = total.indexOf( "Test.test" ) + 5;
				final int end = total.indexOf( '\n', start );
				return total.substring( start, end );
			}
		}
	}

	public static String whereAmI () {
		return new BackTraceException().getTestMethodName();
	}
}
