// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.semanticchecker;

import java.lang.String;
import java.lang.Double;
import java.lang.NumberFormatException;

import java.text.ParseException;

import java.util.Iterator;

import Int.ecb.gesmescb.GESMESCBObject;
import Int.ecb.gesmescb.CodeList;
import Int.ecb.gesmescb.TimeSeries;
import Int.ecb.gesmescb.Attribute;
import Int.ecb.gesmescb.KeyFamilyComponent;
import Int.ecb.gesmescb.Observation;
import Int.ecb.gesmescb.TextFormat;

import Int.ecb.gesmescb.time.RegularInterval;

import Int.ecb.gesmescb.util.KeyTokenizer;

import Int.ecb.gesmescb.error.ErrorWatcher;

import Int.ecb.gesmescb.semanticchecker.MetadataChecker;

/* ***************************** Class ****************************** */

/**
 * <P>Checks contents of
 * {@link Int.ecb.gesmescb.TimeSeries},
 * and {@link Int.ecb.gesmescb.Attribute} objects
 * in the information flow.</P>
 *
 * <center><table border="3" cellpadding="3">
 * <caption>
 * <a name="prefs">Available {@link Int.ecb.prefs preference} settings</a>
 * </caption>
 * <tr>
 * <th>preference name</th>
 * <th>type</th>
 * <th>meaning</th>
 * <th>default</th>
 * </tr>
 * <tr><td><code>strictDimensionValueCodes</code></td><td>boolean</td>
 *	<td>whether to raise
 * {@link ErrorWatcher#ERROR} (<tt>true</tt>)
 * or only {@link ErrorWatcher#WARNING} (<tt>false</tt>)
 * if a dimension value is not in respective
 * {@link Int.ecb.gesmescb.CodeList}</td>
 *	<td><code>true</code></td></tr>
 * <tr><td><code>strictAttributeValueCodes</code></td><td>boolean</td>
 *	<td>whether to raise
 * {@link ErrorWatcher#ERROR} (<tt>true</tt>)
 * or only {@link ErrorWatcher#WARNING} (<tt>false</tt>)
 * if an attribute value is not in respective
 * {@link Int.ecb.gesmescb.CodeList}</td>
 *	<td><code>true</code></td></tr>
 * </table></center>
 *
 * @author
 * <A HREF="mailto:klo@FAME.com">Kam Lo</A>
 * @author
 * <A HREF="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</A>
 *
<<<<<<< DataChecker.java
 * @version @(#)$Id: DataChecker.java,v 1.2 2017/02/15 12:09:41 trainot Exp $
=======
 * @version @(#)$Id: DataChecker.java,v 1.2 2017/02/15 12:08:37 trainot Exp $
>>>>>>> 1.2
 */
class DataChecker extends MetadataChecker {

	/* ======================= Constants ======================== */

	protected static final TextFormat AN7000_VAR_LENGTH_FORMAT
	= TextFormat.create( TextFormat.ANY, 7000, true );

	/* ========================= Fields ========================= */

	private final boolean strictDimensionValueCodes;
	private final boolean strictAttributeValueCodes;

	private TextFormat obsValueFormat;
	private CodeList statusCodeList;
	private CodeList confidentialityCodeList;
	//private TextFormat preBreakValueFormat;

	/* ====================== Constructors ====================== */

	/**
	 * Construct a new instance.
	 */
	public DataChecker () {
		strictDimensionValueCodes = ! "false".equalsIgnoreCase(
			componentHelper.getPreference( "strictDimensionValueCodes", "true" )
		);
		strictAttributeValueCodes = ! "false".equalsIgnoreCase(
			componentHelper.getPreference( "strictAttributeValueCodes", "true" )
		);

		obsValueFormat = null;
		//preBreakValueFormat = null;
		statusCodeList = confidentialityCodeList = null;
	}

	/* ==================== Package Methods ===================== */

	/* ----------------------- TimeSeries ----------------------- */

	/**
	 * Check the contents of a <CODE>TimeSeries</CODE>.
	 */
	void check (
		final TimeSeries timeSeries
	) {
		final String key = timeSeries.getKey();

		KeyCharacteristics keyCharacteristics;

		super.check( timeSeries );

		// Check key and period information.
		{
			final String period = timeSeries.getPeriod();
			final String periodFormat = timeSeries.getPeriodFormat();

			keyCharacteristics = checkKey(
				timeSeries, key, period, periodFormat
			);

			if (
				!inDeleteMessage
				&&
				keyCharacteristics.keyType < KeyCharacteristics.OBSERVATION_KEY
			) {
				complain(
					"The period information is missing. keyType value is:\"" 
				    + keyCharacteristics.keyType 
					+ "\". Period: \"" +period + "\"."
					+ " Period format: \"" 
					+ periodFormat + "\".",
					"check",
					timeSeries,
					ErrorWatcher.ERROR
				);
			}
		}

		// Check number of observations
		{
			final int length = timeSeries.length();

			if ( inDeleteMessage ) {
				if ( length > 0 ) {
					complain(
						"Observations are present in time series \""
						+ key
						+ "\", but there should be none for this kind of"
						+ " delete operation.",
						"check",
						timeSeries,
						ErrorWatcher.ERROR
					);
				}
			} else if ( length != keyCharacteristics.numOfPeriods ) {
				complain(
					"There are " + length + " observations in time series \""
					+ key
					+ "\", but according to the time range, there should be "
					+ keyCharacteristics.numOfPeriods + ".",
					"check",
					timeSeries,
					ErrorWatcher.ERROR
				);
			}
		}

		// Check observations
		{
			final Iterator observations = timeSeries.observationIterator();
			boolean isFirst = true;

			while ( observations.hasNext() ) {
				final Observation observation
				= (Observation)observations.next();
				check( key, observation );

				if ( isFirst ) {
					if ( observation.isEmpty() ) {
						complain(
							"The first observation of the time series \""
							+ key +"\" is empty.",
							"check",
							timeSeries,
							ErrorWatcher.ERROR
						);
					}
					isFirst = false;
				}
				if ( !observations.hasNext() && observation.isEmpty() ) {
					complain(
						"The last observation of the time series \""
						+ key +"\" is empty.",
						"check",
						timeSeries,
						ErrorWatcher.ERROR
					);
				}
			}
		}
	}

	/* ----------------------- Attribute ------------------------ */

	/**
	 * Check the contents of an <CODE>Attribute</CODE>.
	 */
	void check (
		final Attribute attribute
	) {
		final String conceptId = attribute.getName();
		final String key = attribute.getKey();

		KeyFamilyComponent component = null;
		KeyCharacteristics keyCharacteristics;

		super.check( attribute );

		// check conceptId
		{
			if ( keyFamily != null ) {
				component = keyFamily.getComponent( conceptId );
				if ( component == null ) {
					complain(
						"Attribute \"" + conceptId + "\" for key \"" + key
						+ "\" carries a name which is not a"
						+ " recognized statistical concept id"
						+ " in the key family \"" + keyFamily.getId() + "\".",
						"check",
						attribute,
						ErrorWatcher.ERROR
					);
				}
			}
		}

		// Check referred to time series key
		{
			final String period = attribute.getPeriod();
			final String periodFormat = attribute.getPeriodFormat();

			keyCharacteristics = checkKey(
				attribute, key, period, periodFormat
			);

			if ( keyCharacteristics.numOfPeriods > 1 ) {
				complain(
					"The attribute \"" + conceptId + "\" for key \""
					+ key + "\" obviously wants to simultaneously refer to "
					+ keyCharacteristics.numOfPeriods
					+ " observations at the same time.",
					"check",
					attribute,
					ErrorWatcher.ERROR
				);
			}
		}

		// Check concept type
		if ( component != null && !component.isAttribute() ) {
			complain(
				"Attribute \"" + conceptId
				+ "\" for key \"" + key
				+ "\" refers to a concept"
				+ " which is not suitable as an attribute"
				+ " in the footnote section.",
				"check",
				attribute,
				ErrorWatcher.ERROR
			);
		}

		// Check attachment level
		if ( component != null ) {
			String level = null;
			if ( component.isAttachedToObservation() ) {
				if (
					keyCharacteristics.keyType
					!=
					KeyCharacteristics.OBSERVATION_KEY
				) {
					level = "Observation";
				}
			} else if ( component.isAttachedToTimeSeries() ) {
				if (
					keyCharacteristics.keyType
					!=
					KeyCharacteristics.TIME_SERIES_KEY
				) {
					level = "Time Series";
				}
			} else if ( component.isAttachedToSiblingGroup() ) {
				if (
					keyCharacteristics.keyType
					!=
					KeyCharacteristics.SIBLING_GROUP_KEY
				) {
					level = "Sibling Group";
				}
			} else if ( component.isAttachedToDataSet() ) {
				if (
					keyCharacteristics.keyType
					!=
					KeyCharacteristics.DATA_SET_KEY
				) {
					level = "Data Set";
				}
			}
			if ( level != null ) {
				final String periodInfo
				= attribute.getPeriod() + ":" + attribute.getPeriodFormat();
				complain(
					"Attribute \"" + conceptId
					+ "\" has an unsuitable key \"" + key + "\""
					+ (
						":" != periodInfo
						? " with period information \"" + periodInfo + "\""
						: ""
					)
					+ " for " + level + " attachment level.",
					"check",
					attribute,
					ErrorWatcher.ERROR
				);
			}
		}

		// Check the value
		{
			final String value = attribute.getValue();

			if ( inDeleteMessage ) {
				if ( !value.equals( "" ) ) {
					complain(
						"Attribute \"" + conceptId + "\" for key \"" + key
						+ "\" has a nonempty attribute value \"" + value
						+ "\" in delete message.",
						"check",
						attribute,
						ErrorWatcher.ERROR
					);
				}
			} else if ( component != null ) {
				if (
					getMaintenanceAgency() != null
				) {
					final String codeListId = component.getCodeListId();
					if ( !"".equals( codeListId ) ) {
						final CodeList codeList
						= getMaintenanceAgency().getCodeList( codeListId );
						if (
							codeList != null
							&&
							!codeList.containsCodeValue( value )
						) {
							complain(
								"Attribute \"" + conceptId
								+ "\" for key \"" + key
								+ "\" has a value \"" + value
								+ "\" not in the code list \""
								+ codeListId + "\".",
								"check",
								attribute,
								strictAttributeValueCodes
									? ErrorWatcher.ERROR
									: ErrorWatcher.WARNING
							);
						}
					}
				}

				final String format = component.getFormat();
				try {
					final TextFormat valueFormat
					= TextFormat.parse( format );
					if ( !valueFormat.isValid( value ) ) {
						complain(
							"Attribute \"" + conceptId
							+ "\" for key \"" + key
							+ "\" carries the value \"" + value
							+ "\", which is not in text format \"" + format
							+ "\".",
							"check",
							attribute,
							ErrorWatcher.ERROR
						);
					}
				} catch ( final ParseException pe ) {
					// This would in fact mean invalid structural definitions.
					complain(
						"Attribute \"" + conceptId
						+ "\" for key \"" + key
						+ "\" refers to a key family component"
						+ " with invalid text format \"" + format
						+ "\".",
						"check",
						attribute,
						ErrorWatcher.ERROR
					);
				}
			}
		}
	}

	/* ==================== Private Methods ===================== */

	/**
	 * Check the contents of an <CODE>Observation</CODE>
	 * as far as they do not interrelate with the enclosing
	 * {@link TimeSeries}.
	 *
	 * @param timeSeriesKey for logging.
	 * @param observation to check.
	 */
	private void check (
		final String timeSeriesKey,
		final Observation observation
	) {
		if ( !observation.isEmpty() ) {
			// Check value
			{
				final String value = observation.getValue();
				if ( value.equals( "" ) ) {
					complain(
						"In time series \"" + timeSeriesKey
						+ "\": The value of a nonempty observation"
						+ " must not be empty.",
						"check",
						observation,
						ErrorWatcher.ERROR
					);
				} else {
					final TextFormat format = getObsValueFormat();
					check( timeSeriesKey, observation, value, format );
				}
			}

			// Check status
			{
				final String status = observation.getStatus();
				if ( status.equals( "" ) ) {
					complain(
						"In time series \"" + timeSeriesKey
						+ "\": The status of a nonempty observation"
						+ " must not be empty.",
						"check",
						observation,
						ErrorWatcher.ERROR
					);
				} else {
					final CodeList codeList = getObsStatusCodeList();
					check( timeSeriesKey, observation, status, codeList );
				}
			}

			// Check confidentiality
			{
				final String conf = observation.getConfidentiality();
				final CodeList codeList = getObsConfCodeList();
				check( timeSeriesKey, observation, conf, codeList );
			}

			// Check preBreakValue
			{
				final String preBreak = observation.getPreBreakValue();
				final TextFormat format = getPreBreakValueFormat();
				check( timeSeriesKey, observation, preBreak, format );
			}
		}
	}

	/**
	 * Check a number cell of an {@link Observation}.
	 * Missing value indicator is meant to be accepted.
	 *
	 * @param timeSeriesKey for logging.
	 * @param observation checked, for logging.
	 * @param value of the cell, e.g. the observation or pre break value.
	 * @param format associated with the cell.
	 */
	private void check (
		final String timeSeriesKey,
		final Observation observation,
		final String value,
		final TextFormat format
	) {
		if ( format != null && !format.isValid( value ) ) {
			complain(
				"The observation cell \"" + value
				+ "\" is not in format " + format + " for series key " + timeSeriesKey +  ".",
				"check",
				observation,
				ErrorWatcher.ERROR
			);
		}
		if (
			!value.equals( "" )
			&&
			!value.equals( Observation.MISSING_VALUE )
		) {
			try {
				//final double d = 
				Double.parseDouble( value );
			}
			catch ( NumberFormatException nfe ) {
				complain(
					"In time series \"" + timeSeriesKey
					+ "\": The observation cell \"" + value
					+ "\" cannot be parsed to a number " + "for series key " + timeSeriesKey + ".",
					"check",
					observation,
					ErrorWatcher.ERROR
				);
			}
		}
	}

	/**
	 * Check a code cell of an {@link Observation}.
	 *
	 * @param timeSeriesKey for logging.
	 * @param observation checked, for logging.
	 * @param value of the cell, e.g. the observation or pre break value.
	 * @param codeList associated with the cell.
	 */
	private void check (
		final String timeSeriesKey,
		final Observation observation,
		final String value,
		final CodeList codeList
	) {
		if (
			!value.equals( "" )
			&&
			codeList != null
			&&
			!codeList.containsCodeValue( value )
		) {
			complain(
				"In time series \"" + timeSeriesKey
				+ "\": The code value \""
				+ value + "\" is not known to the code list "
				+ codeList.getId() + ".",
				"check",
				observation,
				ErrorWatcher.ERROR
			);
		}
	}

	/**
	 * Check and classify a time series key and period information.
	 * This is done in the context of the current message,
	 * which means that the current key family and message
	 * type (whether delete message) are used.
	 *
	 * @param object to check, used for logging and distinction
	 * {@link TimeSeries} versus {@link Attribute}.
	 * This distinction is necessary because
	 * sibling group keys taken from {@link TimeSeries}
	 * may only appear in delete messages.
	 * @param key to check
	 * @param period to check
	 * @param periodFormat to check
	 *
	 * @return the key characteristics.
	 */
	private KeyCharacteristics checkKey (
		final GESMESCBObject object,
		final String key,
		final String period,
		final String periodFormat
	) {
		final KeyCharacteristics keyCharacteristics = new KeyCharacteristics();
		final boolean inAttribute = object instanceof Attribute;

		if ( key.equals( "" ) ) {
			keyCharacteristics.keyType = KeyCharacteristics.DATA_SET_KEY;

			if ( !inDeleteMessage && !inAttribute ) {
				complain(
					"The conventional key \"" + key
					+ "\" signifying the total data set"
					+ " is only allowed for"
					+ " a deletion or data set attribute.",
					"check",
					object,
					ErrorWatcher.ERROR
				);
			}
	    /** trainot 25.07.2012 - Spaces are not allowed */	
		} else if (key.matches(".*\\s+.*")) {
			complain(
					"The key \"" + key
					+ "\" contains spaces. ",
					"check",
					object,
					ErrorWatcher.ERROR
				);
		} else {
			final Iterator components
			= keyFamily != null ? keyFamily.keyStructureIterator() : null;
			final KeyTokenizer tokenizer = new KeyTokenizer( key );

			keyCharacteristics.keyType = KeyCharacteristics.TIME_SERIES_KEY;
			while (
				tokenizer.hasNext()
				&& ( components == null || components.hasNext() )
			) {
				final String keyField = tokenizer.nextToken();
				final KeyFamilyComponent component
				= components != null
					? (KeyFamilyComponent)components.next() : null;
				if ( keyField.equals( "" ) ) {
					keyCharacteristics.keyType
					= KeyCharacteristics.SIBLING_GROUP_KEY;
					if ( !inDeleteMessage && !inAttribute ) {
						complain(
							"Sibling group key \"" + key
							+ "\" is not allowed outside"
							+ " a deletion or attribute.",
							"check",
							object,
							ErrorWatcher.ERROR
						);
					}
					if (
						component != null
						&&
						!component.isFrequencyDimension()
					) {
						complain(
							"Sibling group key \"" + key
							+ "\" has wildcarded non-frequency component \""
							+ component.getConceptId()
							+ "\".",
							"check",
							object,
							ErrorWatcher.ERROR
						);
					}
				} else if ( component != null ) {
					final String codeListId = component.getCodeListId();
					final CodeList codeList
					= getMaintenanceAgency().getCodeList( codeListId );
					if ( codeList == null ) {
						complain(
							"CodeList \"" + codeListId
							+ "\" is not known to maintenance agency "
							+ getMaintenanceAgencyId()
							+ ". Unable to check key "
							+ key + " field " + keyField + ".",
							"check",
							object,
							ErrorWatcher.ERROR
						);
					} else 
					  {/** trainot 25.07.2012 
					   Check that the length of the key field is less than the maximum allowed length.
					   Maximum allowed length of format AN<integer>..<integer> or AN<integer>.
					   If the maximum allowed length cannot be obtained, just check that the value is in the codelist.  */
						 try {
					       int maximumAllowedLength = Integer.parseInt(component.getFormat().replaceAll("([A-X])+(([\\d])+(\\.)*?)*?([^\\d])+", ""));
						   if (keyField.length() > maximumAllowedLength) {
								  complain(
										"The length " + keyField.length() + " of the key value \"" + keyField
										+ "\" is bigger than the maximum allowed length " + maximumAllowedLength +
										" of the corresponding codelist values for codelist " + codeList.getId() +
										" and format " + component.getFormat(),
										"check",
										object,
										ErrorWatcher.ERROR
									);
						  } else {
							  if ( !codeList.containsCodeValue( keyField )) {
							  	complain(
							  	"CodeList \"" + codeListId
							  	+ "\" does not contain value \"" + keyField
							  	+ "\" in key \"" + key + "\".",
							  	"check",
							  	object,
							  	strictDimensionValueCodes
							  	? ErrorWatcher.ERROR
							  	: ErrorWatcher.WARNING);
						    }
						 }
					   } catch (Exception e) {
						   if ( !codeList.containsCodeValue( keyField )) {
							  	complain(
							  	"CodeList \"" + codeListId
							  	+ "\" does not contain value \"" + keyField
							  	+ "\" in key \"" + key + "\".",
							  	"check",
							  	object,
							  	strictDimensionValueCodes
							  	? ErrorWatcher.ERROR
							  	: ErrorWatcher.WARNING);
						    }
					   }
					}
				}
			}
			if ( components != null && components.hasNext() ) {
				complain(
					"The key " + key + " has too few dimensions.",
					"check",
					object,
					ErrorWatcher.ERROR
				);
			}
			if ( tokenizer.hasNext() ) {
				complain(
					"The key " + key + " has too many dimensions.",
					"check",
					object,
					ErrorWatcher.ERROR
				);
			}
		}

		// check period information
		{
			if (
				keyCharacteristics.keyType < KeyCharacteristics.TIME_SERIES_KEY
			) {
				if ( !period.equals( "" ) ) {
					complain(
						"Period " + period + " is nonempty,"
						+ " but it should be for this kind of"
						+ " time series key.",
						"check",
						object,
						ErrorWatcher.ERROR
					);
				}
				if ( !periodFormat.equals( "" ) ) {
					complain(
						"Period format " + periodFormat + " is nonempty,"
						+ " but it should be for this kind of"
						+ " time series key.",
						"check",
						object,
						ErrorWatcher.ERROR
					);
				}
			} else {
				if (
					!period.equals( "" )
					||
					!periodFormat.equals( "" )
				) {
					keyCharacteristics.keyType
					= KeyCharacteristics.OBSERVATION_KEY;
					try {
						final RegularInterval timePeriod = PERIOD_FORMATTER.parse(
							period, periodFormat
						);
						// if rep-period is available (i.e. could be parsed)
						if (
							messageReportingPeriod != null
							&&
							(
								messageReportingPeriod.getBegin().compareTo(
									timePeriod.getBegin()
								) > 0
								||
								messageReportingPeriod.getEnd().compareTo(
									timePeriod.getEnd()
								) < 0
							)
						) {
							complain(
								"Period " + timePeriod
								+ " does not lie inside"
								+ " the reporting period "
								+ messageReportingPeriod + ".",
								"check",
								object,
								ErrorWatcher.ERROR
							);
						}
						keyCharacteristics.numOfPeriods
						= timePeriod.getSize();
					}
					catch ( ParseException pe ) {
						complain(
							"Period " + period
							+ " with periodFormat " + periodFormat
							+ " cannot be parsed: " + pe,
							"check",
							object,
							ErrorWatcher.ERROR
						);
					}
				}
			}
		}
		return keyCharacteristics;
	}

	/* ----------------- Get specific Metadata ------------------ */

	// TODO: Reset the metadata references between messages!
	// And get codelists according to positions in ARR segment rather
	// than hardcoded.

	/**
	 * Lazily get the observation value format.
	 * If unknown, <CODE>null</CODE> is returned.
	 *
	 * @return the format for observation values.
	 */
	private TextFormat getObsValueFormat() {
		if ( obsValueFormat == null ) {
			obsValueFormat = getCellFormat( "OBS_VALUE" );
		}
		return obsValueFormat;
	}

	/**
	 * Lazily get the observation pre break value format.
	 * If unknown, <CODE>null</CODE> is returned.
	 *
	 * @return the format for observation pre break values.
	 */
	private TextFormat getPreBreakValueFormat() {
		if ( obsValueFormat == null ) {
			obsValueFormat = getCellFormat( "OBS_PRE_BREAK" );
		}
		return obsValueFormat;
	}

	/**
	 * Get the format for a statistical concept in the current key family.
	 *
	 * @param statisticalConceptId specifies the concept.
	 */
	private TextFormat getCellFormat (
		final String statisticalConceptId
	) {
		if ( keyFamily != null ) {
			final KeyFamilyComponent component
			= keyFamily.getComponent( statisticalConceptId );
			if ( component != null ) {
				final String format = component.getFormat();
				try {
					return TextFormat.parse( format );
				}
				catch ( ParseException pe ) {
					complain(
						"Current key family's component "
						+ statisticalConceptId
						+ " has an invalid format " + format
						+ ".",
						"getCellFormat",
						pe,
						ErrorWatcher.ERROR
					);
				}
			}
		}
		return null;
	}

	/**
	 * Lazily get the observation status code list.
	 * If unknown, <CODE>null</CODE> is returned.
	 *
	 * @return the code list for observation status.
	 */
	private CodeList getObsStatusCodeList() {
		if ( statusCodeList == null ) {
			statusCodeList = getCellCodeList( "OBS_STATUS" );
		}
		return statusCodeList;
	}

	/**
	 * Lazily get the observation confidentiality code list.
	 * If unknown, <CODE>null</CODE> is returned.
	 *
	 * @return the code list for observation confidentiality.
	 */
	private CodeList getObsConfCodeList() {
		if ( confidentialityCodeList == null ) {
			confidentialityCodeList = getCellCodeList( "OBS_CONF" );
			if (confidentialityCodeList == null) {
				confidentialityCodeList = getCellCodeList( "CONF_STATUS" );
			}
		}
		return confidentialityCodeList;
	}

	/**
	 * Get the format for a statistical concept in the current key family.
	 *
	 * @param statisticalConceptId specifies the concept.
	 */
	private CodeList getCellCodeList (
		final String statisticalConceptId
	) {
		if ( keyFamily != null ) {
			final KeyFamilyComponent component
			= keyFamily.getComponent( statisticalConceptId );
			if ( component != null ) {
				final String codeListId = component.getCodeListId();
				if ( getMaintenanceAgency() != null ) {
					return getMaintenanceAgency().getCodeList( codeListId );
				}
			}
		}
		return null;
	}
}
