// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.semanticchecker;

import java.lang.Object;
import java.lang.String;

import junit.framework.Test;
import junit.framework.TestSuite;

import Int.ecb.gesmescb.TimeSeries;
import Int.ecb.gesmescb.Attribute;
import Int.ecb.gesmescb.Observation;

import Int.ecb.gesmescb.semanticchecker.DataChecker;

/* ***************************** Class ****************************** */

/**
 * <P>Tests the implementation of the
 * class {@link DataChecker}.</P>
 *
 * @author
 * <A HREF="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</A>
 *
 * @version @(#)$Id: DataCheckerTest.java,v 1.1 2017/01/19 16:14:05 trainot Exp $
 */

public class DataCheckerTest extends MetadataCheckerTest {

	/* ====================== Constructors ====================== */

	/**
	 * Constructs a SemanticCheckerTest with the specified name.
	 *
	 * @param name Test case name
	 */
	public DataCheckerTest ( final String name ) {
		super( name );
	}

	/* ================== Protected Methods ===================== */

	/**
	 * Factory for DataCheckers.
	 * Should be overridden by derived classes.
	 *
	 * @see doTest(boolean,Object[])
	 */
	protected CheckerBase checkerFactory () {
		return new DataChecker ();
	}

	/* ===================== Public Methods ===================== */

	/**
	 * Assembles and returns a test suite for all
	 * the test methods of this test case.
	 *
	 * @return a non-null test suite
	 */
	public static Test suite () {
		return new TestSuite( DataCheckerTest.class );
	}

	/**
	 * Runs the test case
	 */
	public static void main ( final String args[] ) {
		final String[] testCaseName = { DataCheckerTest.class.getName() };
		junit.textui.TestRunner.main( testCaseName );
	}

	/* ---------------------- Test Methods ---------------------- */

	/* ------------------- Time Series Checks ------------------- */

	public void testGoodTimeSeries () {
		doTest( true, new Object[] {
			INTERCHANGE_BEGIN,
			DATA_MESSAGE_BEGIN,
			new TimeSeries(
				"W.C.LO", "199010", "610",
				new Observation[] {
					new Observation(
						"-13.8", "B", "F", "-"
					)
				}
			),
			TS1,
			DATA_MESSAGE_END,
			INTERCHANGE_END,
			null
		} );
	}

	public void testBadTimeSeriesKeys () {
		doTest( true, new Object[] {
			INTERCHANGE_BEGIN,
			DATA_MESSAGE_BEGIN,
			new TimeSeries( ".B.LO", "199010199012", "710", OBS_ARR ),
			ERROR,
			new TimeSeries( "M..LO", "199010199012", "710", OBS_ARR ),
			ERROR,
			new TimeSeries( "M.B", "199010199012", "710", OBS_ARR ),
			ERROR,
			new TimeSeries( "M.Q.LO", "199010199012", "710", OBS_ARR ),
			ERROR,
			DATA_MESSAGE_END,
			INTERCHANGE_END,
			null
		} );
	}

	public void testBadKeysEnabled () {
		System.setProperty(
			DataChecker.class.getName() + ".strictDimensionValueCodes",
			"false"
		);
		doTest( true, new Object[] {
			INTERCHANGE_BEGIN,
			DATA_MESSAGE_BEGIN,
			new TimeSeries( "X.Q.LO", "199010199012", "710", OBS_ARR ),
			WARNING,
			new TimeSeries( "M.F.LO", "199010199012", "710", OBS_ARR ),
			WARNING,
			new TimeSeries( "M.Q.JU", "199010199012", "710", OBS_ARR ),
			WARNING,
			new Attribute(
				"M.Q.LI", "", "", "ATT2", "JU"
			),
			WARNING,
			DATA_MESSAGE_END,
			INTERCHANGE_END,
			null
		} );
		System.setProperty(
			DataChecker.class.getName() + ".strictDimensionValueCodes",
			"true"
		);
	}

	public void testBadTimeSeriesPeriods () {
		doTest( true, new Object[] {
			INTERCHANGE_BEGIN,
			DATA_MESSAGE_BEGIN,
			new TimeSeries( "M.B.LO", "199010199012", "610", OBS_ARR ),
			ERROR,
			new TimeSeries( "M.B.LO", "199010199012", "702", OBS_ARR ),
			ERROR,
			new TimeSeries( "M.B.LO", "199010", "710", OBS_ARR ),
			ERROR,
			new TimeSeries( "M.B.LO", "199010199012", "t10", OBS_ARR ),
			ERROR,
			new TimeSeries( "M.B.LO", "1990101119901211", "710", OBS_ARR ),
			ERROR,
			new TimeSeries( "M.B.LO", "199010199013", "710", OBS_ARR ),
			ERROR,
			new TimeSeries( "M.B.LO", "19901019901x", "710", OBS_ARR ),
			ERROR,
			DATA_MESSAGE_END,
			INTERCHANGE_END,
			null
		} );
	}

	public void testTimeSeriesOutsideReportingPeriod () {
		doTest( true, new Object[] {
			INTERCHANGE_BEGIN,
			DATA_MESSAGE_BEGIN,
			new TimeSeries( "M.B.LO", "1990119903", "708", OBS_ARR ), //good one
			new TimeSeries( "M.B.LO", "189010189012", "710", OBS_ARR ), ERROR,
			new TimeSeries( "M.B.LO", "1999320001", "708", OBS_ARR ), ERROR,
			DATA_MESSAGE_END,
			INTERCHANGE_END,
			null
		} );
	}

	public void testInvalidEmptyObservations () {
		doTest( true, new Object[] {
			INTERCHANGE_BEGIN,
			DATA_MESSAGE_BEGIN,
			new TimeSeries(
				"M.B.LO", "199010199011", "710",
				new Observation[] {
					OBS_ARR[0],
					OBS_ARR[1]
				}
			),	// just to test with a good one first
			new TimeSeries(
				"M.B.LO", "199010199011", "710",
				new Observation[] {
					OBS_ARR[0],
					Observation.EMPTY_OBSERVATION
				}
			), ERROR,
			new TimeSeries(
				"M.B.LO", "199010199011", "710",
				new Observation[] {
					Observation.EMPTY_OBSERVATION,
					OBS_ARR[0]
				}
			), ERROR,
			DATA_MESSAGE_END,
			INTERCHANGE_END,
			null
		} );
	}

	/* -------------------- Attribute Checks -------------------- */

	public void testGoodAttributes () {
		doTest( true, new Object[] {
			INTERCHANGE_BEGIN,
			DATA_MESSAGE_BEGIN,
			new Attribute(
				"M.B.LO", "199010", "610", "ATT1", "X"
			),
			AT2,
			new Attribute(
				".B.LO", "", "", "ATT3", "12"
			),
			new Attribute(
				"", "", "", "ATT4", "Wien"
			),
			DATA_MESSAGE_END,
			INTERCHANGE_END,
			null
		} );
	}

	public void testUnknownAttributeConcept () {
		doTest( true, new Object[] {
			INTERCHANGE_BEGIN,
			DATA_MESSAGE_BEGIN,
			new Attribute(
				"M.B.LO", "", "", "UNKNOWN", "WO"
			),
			ERROR,
			DATA_MESSAGE_END,
			INTERCHANGE_END,
			null
		} );
	}

	public void testNonAttributeConcept () {
		doTest( true, new Object[] {
			INTERCHANGE_BEGIN,
			DATA_MESSAGE_BEGIN,
			new Attribute(
				"M.B.LO", "199501", "610", "OBS_CONF", "F"
			),
			ERROR,
			DATA_MESSAGE_END,
			INTERCHANGE_END,
			null
		} );
	}

	public void testBadAttributeRefKey () {
		doTest( true, new Object[] {
			INTERCHANGE_BEGIN,
			DATA_MESSAGE_BEGIN,
			new Attribute(
				".B.LO", "", "", "ATT2", "JU"
			),
			ERROR,
			new Attribute(
				"M..LO", "", "", "ATT2", "JU"
			),
			ERROR,
			new Attribute(
				"M.B", "", "", "ATT2", "JU"
			),
			ERROR,
			new Attribute(
				"M.Q.LO", "", "", "ATT2", "JU"
			),
			ERROR,
			DATA_MESSAGE_END,
			INTERCHANGE_END,
			null
		} );
	}

	public void testAttributeOutsideReportingPeriod () {
		doTest( true, new Object[] {
			INTERCHANGE_BEGIN,
			DATA_MESSAGE_BEGIN,
			new Attribute(
				"M.B.LO", "189010", "616", "ATT1", "X"
			), ERROR,
			new Attribute(
				"M.B.LO", "20001", "604", "ATT1", "X"
			), ERROR,
			DATA_MESSAGE_END,
			INTERCHANGE_END,
			null
		} );
	}

	public void testBadAttributeValue () {
		doTest( true, new Object[] {
			INTERCHANGE_BEGIN,
			DATA_MESSAGE_BEGIN,
			new Attribute(
				"A.B.LO", "1999", "602", "ATT1", "T"
			),
			ERROR,
			new Attribute(
				".B.LO", "", "", "ATT3", "1c2"
			),
			ERROR,
			DATA_MESSAGE_END,
			INTERCHANGE_END,
			null
		} );
	}
}
