// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.semanticchecker;

import java.lang.String;

import java.util.HashMap;
import java.util.Iterator;

import Int.ecb.gesmescb.*;

import Int.ecb.gesmescb.metabase.Metabase;
import Int.ecb.gesmescb.metabase.MaintenanceAgency;

import Int.ecb.gesmescb.util.KeyTokenizer;

import Int.ecb.gesmescb.error.ErrorWatcher;

/* ***************************** Class ****************************** */

/**
 * <P>Checks for compliance with ECB restriction environment rules.
 * These rules are further constraints on messages which are agreed
 * in the ESCB's data exchange context.</P>
 *
 * @author
 * <A HREF="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</A>
 *
 * @version @(#)$Id: ECBRestrictionsChecker.java,v 1.1 2017/01/19 16:14:05 trainot Exp $
 */
public class ECBRestrictionsChecker extends CheckerBase {

	/* ======================= Constants ======================== */

	/** ECB's id as a Maintenance Agency. */
	private static final String ECB = "ECB";

	/** Organisation code list name. */
	public static final String CODELIST_ORGANISATION = "CL_ORGANISATION";

	/** MaintenanceAgency code list name. */
	public static final String CODELIST_AGENCY = "CL_AGENCY";

	/** {@link Int.ecb.gesmescb.StatisticalConcept}s code list name. */
	public static final String CODELIST_CONCEPT = "CL_CONCEPT";

	protected static final TextFormat Coordinate18_VAR_LENGTH_FORMAT
	= TextFormat.create( TextFormat.COORDINATE, 18, true );

	public static final HashMap timeFormatDescription = new HashMap();
	/* ========================= Fields ========================= */

	/** Interchange sender */
	private String interchangeSender;

	/** Interchange receiver */
	private String interchangeReceiver;

	/** Position of FREQ in the Key. */
	private int freqPosition;

	/** Position of TIME_FORMAT in the Key. */
	private int timeFormatPosition;

	/* ====================== Constructors ====================== */

	/**
	 * Construct a new instance.
	 */
	public ECBRestrictionsChecker () {
		interchangeSender = null;
		interchangeReceiver = null;
		freqPosition = timeFormatPosition = 0;
		timeFormatDescription.put("203", "Minutely");
		timeFormatDescription.put("102", "Daily");
		timeFormatDescription.put("711", "Daily");
		timeFormatDescription.put("616", "Weekly");
		timeFormatDescription.put("716", "Weekly");
		timeFormatDescription.put("610", "Monthly");
		timeFormatDescription.put("710", "Monthly");
		timeFormatDescription.put("608", "Quarterly");
		timeFormatDescription.put("708", "Quarterly");
		timeFormatDescription.put("604", "Half yearly");
		timeFormatDescription.put("704", "Half yearly");
		timeFormatDescription.put("602", "Annual");
		timeFormatDescription.put("702", "Annual");
		
	}

	/* ==================== Package Methods ===================== */

	/* -------------------- InterchangeBegin -------------------- */

	/**
	 * Check the contents of an InterchangeBegin.
	 */
	void check (
		final InterchangeBegin interchangeBegin
	) {
		super.check( interchangeBegin );

		final Metabase metabase = Metabase.getMetabase();
		final MaintenanceAgency ecb = metabase.getMaintenanceAgency( ECB );
		final CodeList cl_agency
		= ecb != null ? ecb.getCodeList( CODELIST_AGENCY ) : null;

		// Check sender and receiver are in ECB's CL_ORGANISATION
		final CodeList ecb_cl_organisation
		= ecb != null ? ecb.getCodeList( CODELIST_ORGANISATION ) : null;
		interchangeSender = interchangeBegin.getSender();
		interchangeReceiver = interchangeBegin.getReceiver();

		if (
			ecb_cl_organisation != null
			&&
			!ecb_cl_organisation.containsCodeValue( interchangeSender )
		) {
			complain(
				"Interchange sender \"" + interchangeSender
				+ "\" is not in " + ECB + "'s "
				+ CODELIST_ORGANISATION + ".",
				"check",
				interchangeBegin,
				ErrorWatcher.ERROR
			);
		}
		if (
			ecb_cl_organisation != null
			&&
			!ecb_cl_organisation.containsCodeValue( interchangeReceiver )
		) {
			complain(
				"Interchange receiver \"" + interchangeReceiver
				+ "\" is not in " + ECB + "'s "
				+ CODELIST_ORGANISATION + ".",
				"check",
				interchangeBegin,
				ErrorWatcher.ERROR
			);
		}
	}

	/* ---------------------- MessageBegin ---------------------- */

	/**
	 * Check the contents of a MessageBegin.
	 */
	void check (
		final MessageBegin messageBegin
	) {
		super.check( messageBegin );

		final String messageSender = messageBegin.getSender();
		final String messageReceiver = messageBegin.getReceiver();

		final Metabase metabase = Metabase.getMetabase();
		final MaintenanceAgency ecb = metabase.getMaintenanceAgency( ECB );

		if (
			interchangeSender != null
			&&
			!interchangeSender.equals( messageSender )
		) {
			complain(
				"Interchange sender \"" + interchangeSender
				+ " and Message sender \"" + messageSender
				+ "\" differ.",
				"check",
				messageBegin,
				ErrorWatcher.OK
			);
		}

		if (
			interchangeReceiver != null
			&&
			!interchangeReceiver.equals( messageReceiver )
		) {
			complain(
				"Interchange receiver \"" + interchangeReceiver
				+ " and Message receiver \"" + messageReceiver
				+ "\" differ.",
				"check",
				messageBegin,
				ErrorWatcher.OK
			);
		}

		if ( getMaintenanceAgency() != null ) {
			final CodeList cl_organisation
			= getMaintenanceAgency().getCodeList( CODELIST_ORGANISATION );
			final CodeList cl_agency
			= getMaintenanceAgency().getCodeList( CODELIST_AGENCY );

			// Check sender and receiver are in M.A's CL_ORGANISATION.
			if ( cl_organisation != null ) {
				if ( !cl_organisation.containsCodeValue( messageSender ) ) {
					complain(
						"Message sender \"" + messageSender
						+ "\" is not in " + getMaintenanceAgencyId() + "'s "
						+ CODELIST_ORGANISATION + ".",
						"check",
						messageBegin,
						ErrorWatcher.ERROR
					);
				}

				if ( !cl_organisation.containsCodeValue( messageReceiver ) ) {
					complain(
						"Message receiver \"" + messageReceiver
						+ "\" is not in " + getMaintenanceAgencyId() + "'s "
						+ CODELIST_ORGANISATION + ".",
						"check",
						messageBegin,
						ErrorWatcher.ERROR
					);
				}
			} else {
				complain(
					"Maintenance agency \"" + getMaintenanceAgencyId()
					+ "\" does not know code list \""
					+ CODELIST_ORGANISATION
					+ "\".",
					"check",
					messageBegin,
					ErrorWatcher.ERROR
				);
			}

			// Check maintenance agency is in M.A's CL_AGENCY.
			if ( cl_agency != null ) {
				if ( !cl_agency.containsCodeValue(
					getMaintenanceAgencyId()
				) ) {
					complain(
						"Maintenance Agency \""
						+ getMaintenanceAgencyId()
						+ "\" is not in its own "
						+ CODELIST_AGENCY + ".",
						"check",
						messageBegin,
						ErrorWatcher.ERROR
					);
				}
			}

			if (
				messageBegin.isStatisticalData()
				||
				messageBegin.isDataRequest()
			) {
				final KeyFamily keyFamily
				= getMaintenanceAgency().getKeyFamily(
					messageBegin.getKeyFamilyId()
				);
				if ( keyFamily != null ) {
					freqPosition = keyFamily.getFrequencyDimensionPosition();
					final Iterator iterator = keyFamily.timeDimensionIterator();
					if ( iterator.hasNext() ) iterator.next();
					if ( iterator.hasNext() ) {
						final KeyFamilyComponent timeFormat
						= (KeyFamilyComponent) iterator.next();
						timeFormatPosition = timeFormat.getPositionNo();
					} else {
						complain(
							"Used Key Family \"" + keyFamily.getId()
							+ "\" has less than two dimensions for time.",
							"check",
							messageBegin,
							ErrorWatcher.ERROR
						);
					}
				}
			}
		}
	}

	/* ------------------------ CodeList ------------------------ */

	/**
	 * Check the contents of a <code>CodeList</code>.
	 */
	void check (
		final CodeList codeList
	) {
		super.check( codeList );

		final String id = codeList.getId();

		// Check descriptions whitespace structure.
		for (
			final Iterator
				values = codeList.codeValueIterator(),
				descriptions = codeList.codeDescriptionIterator();
			values.hasNext() && descriptions.hasNext();
		) {
			final String value = (String) values.next();
			final String description = (String) descriptions.next();

			if ( "".equals( description ) ) {
				complain(
					"CodeList \"" + id
					+ "\" contains value \""
					+ value
					+ " with empty description: \"" + description + "\".",
					"check",
					codeList,
					ErrorWatcher.WARNING
				);
			}

			if ( !description.equals( description.trim() ) ) {
				complain(
					"CodeList \"" + id
					+ "\" contains value \""
					+ value + "\" with whitespace at beginning or end"
					+ " of its description: \"" + description + "\".",
					"check",
					codeList,
					ErrorWatcher.WARNING
				);
			}

			if ( description.indexOf( " " + " " ) >= 0 ) {
				complain(
					"CodeList \"" + id
					+ "\" contains value \""
					+ value + "\" with duplicate whitespace characters"
					+ " in its description: \"" + description + "\".",
					"check",
					codeList,
					ErrorWatcher.WARNING
				);
			}
		}
	}

	/* ------------------- StatisticalConcept ------------------- */

	/**
	 * Check the contents of a StatisticalConcept.
	 */
	void check (
		final StatisticalConcept statisticalConcept
	) {
		super.check( statisticalConcept );

		final String id = statisticalConcept.getId();

		if ( getMaintenanceAgency() != null ) {
			final CodeList cl_concept
			= getMaintenanceAgency().getCodeList( CODELIST_CONCEPT );

			// Check concept id agency is in M.A's CL_CONCEPT.
			if (
				cl_concept != null
				&&
				!cl_concept.containsCodeValue( id )
			) {
				complain(
					"Statistical Concept Id \"" + id
					+ "\" is not in " + getMaintenanceAgencyId() + "'s "
					+ CODELIST_CONCEPT + ".",
					"check",
					statisticalConcept,
					ErrorWatcher.ERROR
				);
			}
		}
	}

	/* ----------------------- Key Family ----------------------- */

	/**
	 * Check the contents of a Key Family.
	 */
	void check (
		final KeyFamily keyFamily
	) {
		super.check( keyFamily );

		final String id = keyFamily.getId();

		if ( getMaintenanceAgency() != null ) {
			for (
				final Iterator iterator = keyFamily.componentIterator();
				iterator.hasNext();
			) {
				final KeyFamilyComponent component
				= (KeyFamilyComponent) iterator.next();
				final String codeListId = component.getCodeListId();
				if (
					(
						component.isKeyStructureDimension()
						||
						component.isTimeDimension()
					)
					&&
					!"".equals( codeListId )
				) {
					final CodeList codeList
					= getMaintenanceAgency().getCodeList( codeListId );
					if ( codeList != null ) {
						for (
							final Iterator codeIterator
							= codeList.codeValueIterator();
							codeIterator.hasNext();
						) {
							final String code = (String) codeIterator.next();
							if (
								"".equals( code )
								||
								!Coordinate18_VAR_LENGTH_FORMAT.isValid( code )
							) {
								complain(
									getMaintenanceAgencyId()
									+ "'s Key Family \"" + id
									+ "\" assigns the Concept \""
									+ component.getConceptId()
									+ "\" with Code List \""
									+ codeListId
									+ "\". But this contains code \"" + code
									+ "\", which is not suitable"
									+ " for a dimension.",
									"check",
									keyFamily,
									ErrorWatcher.ERROR
								);
							}
						}
					}
				}
			}
		}
	}

	/* ----------------------- TimeSeries ----------------------- */

	/**
	 * Check the contents of a TimeSeries.
	 */
	void check (
		final TimeSeries timeSeries
	) {
		super.check( timeSeries );
		final String
			key = timeSeries.getKey(),
			timeFormat = timeSeries.getPeriodFormat();
		final Object timeFormatDescObj = timeFormatDescription.get(timeFormat);
		String timeFormatDescString = timeFormat;
		if (timeFormatDescObj != null) {
			timeFormatDescString = timeFormatDescObj.toString();
		}
		if ( !freqTimeFormatConsistent( key, timeFormat ) ) {
			complain(
				"Frequency code in key \"" + key
				+ "\" does not match time format \"" + timeFormatDescString
				+ "\".",
				"check",
				timeSeries,
				ErrorWatcher.ERROR
			);
		}
		for (
			final Iterator iterator = timeSeries.observationIterator();
			iterator.hasNext();
		) {
			final Observation observation = (Observation) iterator.next();
			check( observation, key );
		}
	}

	private void check (
		final Observation observation,
		final String key
	) {
		final String value = observation.getValue();
		final String status = observation.getStatus();
		final String preBreak = observation.getPreBreakValue();
		if ( observation.isEmpty() ) {
			// empty is always fine
		} else if ( observation.hasMissingValue() ) {
			if ( "BHLMQJSO".indexOf( status ) < 0 ) {
				complain(
					"Time Series \"" + key
					+ "\" has missing observation with status code \""
					+ status + "\" indicating value to be present.",
					"check",
					observation,
					ErrorWatcher.ERROR
				);
			}
		} else {
			if ( "HLMQO".indexOf( status ) >= 0 ) {
				complain(
					"Time Series \"" + key
					+ "\" has value \"" + value
					+ "\" with status code \""
					+ status + "\" indicating missing value.",
					"check",
					observation,
					ErrorWatcher.ERROR
				);
			}
		}
		if ( !"".equals( preBreak ) && !"B".equals( status ) ) {
			complain(
				"Time Series \"" + key
				+ "\" has pre break value \""
				+ preBreak + "\" with status code \""
				+ status + "\" not indicating a break.",
				"check",
				observation,
				ErrorWatcher.ERROR
			);
		}
		if ( "".equals( preBreak ) && "B".equals( status ) ) {
			complain(
				"Time Series \"" + key
				+ "\" has empty pre break value \""
				+ preBreak + "\" with status code \""
				+ status + "\" indicating a break.",
				"check",
				observation,
				ErrorWatcher.WARNING
			);
		}
	}

	/* ----------------------- Attribute ------------------------ */

	/**
	 * Check the contents of an Attribute.
	 */
	void check (
		final Attribute attribute
	) {
		super.check( attribute );
		final String
			key = attribute.getKey(),
			timeFormat = attribute.getPeriodFormat();
		if ( !freqTimeFormatConsistent( key, timeFormat ) ) {
			complain(
				"Frequency code in key \"" + key
				+ "\" does not match time format \"" + timeFormat
				+ "\".",
				"check",
				attribute,
				ErrorWatcher.ERROR
			);
		}
	}

	/* ==================== Private Methods ===================== */

	/** Check whether frequency code fits time format.
	 * This assumes usual requency encoding.
	 *
	 * @return whether the frequency code is consistent with the time format.
	 */
	private boolean freqTimeFormatConsistent (
		final String key, final String timeFormat
	) {
		if ( "".equals( timeFormat ) ) return true;
		final KeyTokenizer tokenizer = new KeyTokenizer( key );
		for ( int i = 1; i < freqPosition && tokenizer.hasNext(); ++i );
		if ( tokenizer.hasNext() ) {
			final String freq = tokenizer.nextToken();
			final char freqC = freq.length() == 1 ? freq.charAt( 0 ) : '?';
			switch ( freqC ) {
				case 'A': return
					"602".equals( timeFormat ) || "702".equals( timeFormat );
				case 'H': return
					"604".equals( timeFormat ) || "704".equals( timeFormat );
				case 'Q': return
					"608".equals( timeFormat ) || "708".equals( timeFormat );
				case 'M': return
					"610".equals( timeFormat ) || "710".equals( timeFormat );
				case 'N': return
					"203".equals( timeFormat ) || "719".equals( timeFormat );
				case 'W': return
					"616".equals( timeFormat ) || "716".equals( timeFormat );
				case 'B':
				case 'D': return
					"102".equals( timeFormat ) || "711".equals( timeFormat );
				case 'E': return true;
				default:
					complain(
						"Frequency code \"" + freq
						+ "\" is not known in key.",
						"check",
						key,
						ErrorWatcher.ERROR
					);
			}
		}
		return false;
	}
}
