// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.semanticchecker;

import java.lang.String;
import java.lang.Integer;
import java.lang.NumberFormatException;

import java.util.Date;
import java.util.Iterator;

import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.text.ParsePosition;
import java.text.ParseException;

import Int.ecb.logging.Level;

import Int.ecb.gesmescb.InterchangeBegin;
import Int.ecb.gesmescb.InterchangeEnd;
import Int.ecb.gesmescb.ContactChannel;
import Int.ecb.gesmescb.Contact;
import Int.ecb.gesmescb.MessageBegin;
import Int.ecb.gesmescb.MessageEnd;
import Int.ecb.gesmescb.KeyFamily;
import Int.ecb.gesmescb.TextFormat;

import Int.ecb.gesmescb.time.Interval;
import Int.ecb.gesmescb.edifact.EDIFACTReader;
import Int.ecb.gesmescb.edifact.RegularIntervalFormat;

import Int.ecb.gesmescb.error.ErrorWatcher;

import Int.ecb.gesmescb.semanticchecker.SequenceChecker;

/* ***************************** Class ****************************** */

/**
 * <P>Checks the contents of
 * {@link Int.ecb.gesmescb.InterchangeBegin},
 * {@link Int.ecb.gesmescb.InterchangeEnd},
 * {@link Int.ecb.gesmescb.MessageBegin},
 * and {@link Int.ecb.gesmescb.MessageEnd} objects
 * in the information flow.</P>
 *
 * @author
 * <A HREF="mailto:klo@FAME.com">Kam Lo</A>
 * @author
 * <A HREF="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</A>
 *
 * @version @(#)$Id: EnvelopeChecker.java,v 1.1 2017/01/19 16:14:05 trainot Exp $
 */
class EnvelopeChecker extends SequenceChecker {

	/* ======================= Constants ======================== */

	/**
	 * Useful for {@link String} to {@link Interval} conversion.
	 */
	protected static final RegularIntervalFormat PERIOD_FORMATTER
	= new RegularIntervalFormat();

	protected static final TextFormat AN35_VAR_LENGTH_FORMAT
	= TextFormat.create( TextFormat.ANY, 35, true );

	protected static final TextFormat AN17_VAR_LENGTH_FORMAT
	= TextFormat.create( TextFormat.ANY, 17, true );

	protected static final TextFormat AN512_VAR_LENGTH_FORMAT
	= TextFormat.create( TextFormat.ANY, 512, true );

	/* ========================= Fields ========================= */

	/**
	 * The last encountered InterchangeBegin.
	 */
	private InterchangeBegin interchangeBegin;

	/**
	 * Interchange preparation dateTime.
	 */
	private Date preparationDateTime;

	/**
	 * The last encountered MessageBegin.
	 */
	private MessageBegin messageBegin;

	/**
	 * Interchange preparation dateTime.
	 */
	private Date extractionDateTime;

	/**
	 * Counts the messages for MREF verification.
	 */
	private int messageCounter;

	/* -------------------- Protected Fields -------------------- */

	/**
	 * The key family the current message refers to.
	 */
	protected KeyFamily keyFamily;

	/**
	 * Whether the current message is update/replace.
	 */
	protected boolean inUpdateMessage;

	/**
	 * Whether the current message is update/replace.
	 */
	protected boolean inDeleteMessage;

	/**
	 * The current message's reporting period.
	 */
	protected Interval messageReportingPeriod;

	/* ====================== Constructors ====================== */

	/**
	 * Construct a new instance.
	 */
	public EnvelopeChecker () {
		messageCounter = 0;
		keyFamily = null;
		inUpdateMessage = inDeleteMessage = false;
	}

	/* ==================== Package Methods ===================== */

	/* -------------------- InterchangeBegin -------------------- */

	/**
	 * Check the contents of an InterchangeBegin.
	 */
	void check (
		final InterchangeBegin interchangeBegin
	) {
		super.check( interchangeBegin );

		// Store the current interchange info.
		this.interchangeBegin = interchangeBegin;

		// Reset message counting.
		messageCounter = 0;

		// Check IREF
		{
			final String prefix = "IREF";
			final String iref = interchangeBegin.getIref();

			if ( !iref.startsWith( prefix ) ) {
				complain(
					"IREF number \"" + iref
					+ "\" misses prefix " + prefix + ".",
					"check",
					interchangeBegin,
					ErrorWatcher.ERROR
				);
			}

			if ( iref.length() != 10 ) {
				complain(
					"IREF number \"" + iref
					+ "\" has a length different from 10 characters.",
					"check",
					interchangeBegin,
					ErrorWatcher.ERROR
				);
			}

			for ( int i = 4; i < iref.length(); ++i ) {
				if ( !Character.isDigit( iref.charAt( i ) ) ) {
					complain(
						"IREF number \"" + iref
						+ "\" contains non-digits in its number section.",
						"check",
						interchangeBegin,
						ErrorWatcher.ERROR
					);
				}
			}
		}

		// Check sender and receiver are in format AN..35
		{
			final String sender = interchangeBegin.getSender();
			final String receiver = interchangeBegin.getReceiver();

			if ( !AN35_VAR_LENGTH_FORMAT.isValid( sender ) ) {
				complain(
					"Interchange sender \"" + sender
					+ "\" is not in correct format.",
					"check",
					interchangeBegin,
					ErrorWatcher.ERROR
				);
			}
			if ( !AN35_VAR_LENGTH_FORMAT.isValid( receiver ) ) {
				complain(
					"Interchange receiver \"" + receiver
					+ "\" is not in correct format.",
					"check",
					interchangeBegin,
					ErrorWatcher.ERROR
				);
			}
		}

		// Check date and time format of date of preparation
		{
			final String parsePattern = "yyMMdd:HHmm";
			SimpleDateFormat formatter
			= new SimpleDateFormat ( parsePattern );
			formatter.setLenient( false );

			final String dateTime = interchangeBegin.getPreparationDateTime();

			if (
				dateTime.length() != parsePattern.length()
			) {
				complain(
					"Interchange preparation dateTime \"" + dateTime
					+ "\" has a length different from "
					+ parsePattern.length() + " characters.",
					"check",
					interchangeBegin,
					ErrorWatcher.ERROR
				);
			}
			preparationDateTime
			= formatter.parse( dateTime, new ParsePosition( 0 ) );
			if ( preparationDateTime == null ) {
				complain(
					"Interchange preparation dateTime \"" + dateTime
					+ "\" cannot be parsed.",
					"check",
					interchangeBegin,
					ErrorWatcher.ERROR
				);
			}
		}
	}

	/* --------------------- InterchangeEnd --------------------- */

	/**
	 * Check the contents of an InterchangeEnd.
	 */
	void check (
		final InterchangeEnd interchangeEnd
	) {
		super.check( interchangeEnd );

		if ( interchangeBegin == null ) {
			complain(
				"InterchangeEnd without InterchangeBegin",
				"check",
				interchangeEnd,
				ErrorWatcher.ERROR
			);
		} else if (
			!interchangeEnd.isEndFor( interchangeBegin )
		) {
			complain(
				"InterchangeEnd does not match InterchangeBegin",
				"check",
				interchangeEnd,
				ErrorWatcher.ERROR
			);
		}
	}

	/* ---------------------- MessageBegin ---------------------- */

	/**
	 * Check the contents of a MessageBegin.
	 */
	void check (
		final MessageBegin messageBegin
	) {
		super.check( messageBegin );

		// Store the current message info.
		this.messageBegin = messageBegin;

		// Increment the validation reference counter.
		++messageCounter;
		
		// Loader is deprecated. Only BatchTextUI should be taken into account
		// detect the reader in order to check the used standard : EDI or SDMX/ML
		final boolean isXMLFile = 
				System.getProperty("Int.ecb.gesmescb.application.BatchTextUI.dataReaderClass", "")
				.matches(".*XMLReader");

		// Check MREF
		{
			final String prefix = "MREF";
			final String mref = messageBegin.getMref();

			if ( !mref.startsWith( prefix ) ) {
				complain(
					"MREF number \"" + mref
					+ "\" misses prefix " + prefix + ".",
					"check",
					messageBegin,
					ErrorWatcher.ERROR
				);
			}

			if ( mref.length() != 10 ) {
				complain(
					"MREF number \"" + mref
					+ "\" has a length different from 10 characters.",
					"check",
					messageBegin,
					ErrorWatcher.ERROR
				);
			}

			for ( int i = prefix.length(); i < mref.length(); ++i ) {
				if ( !Character.isDigit( mref.charAt( i ) ) ) {
					complain(
						"MREF number \"" + mref
						+ "\" contains non-digits in its number section.",
						"check",
						messageBegin,
						ErrorWatcher.ERROR
					);
				}
			}

			// Verify increase by one of mref number.
			try {
				if (
					Integer.parseInt( mref.substring( prefix.length() ) )
					!=
					messageCounter
				) {
					complain(
						"MREF number \"" + mref
						+ "\" has wrong numeric value. It should be "
						+ messageCounter + ".",
						"check",
						messageBegin,
						ErrorWatcher.ERROR
					);
				}
			}
			catch ( NumberFormatException nfe ) {
				// Impossible to come here, but who knows.
				complain(
					"MREF number \"" + mref
					+ "\" cannot be parsed to a numeric value.",
					"check",
					messageBegin,
					ErrorWatcher.ERROR
				);
			}
		}

		// Check maintenance agency and sender/receiver.
		{
			if ( getMaintenanceAgency() == null ) {
				if ( !messageBegin.isStatisticalDefinitions() ) {
					complain(
						"Maintenance agency \""
						+ getMaintenanceAgencyId()
						+ "\" is not known.",
						"check",
						messageBegin,
						ErrorWatcher.ERROR
					);
				} else {
					componentHelper.logger.log(
						Level.FINE,
						getName(),
						"handle",
						"The maintenance agency \""
						+ getMaintenanceAgencyId()
						+ "\" is new.",
						getMaintenanceAgencyId()
					);
				}
			}
		}

		// Check subject
		{
			final String identity = messageBegin.getIdentity();

			if ( !AN35_VAR_LENGTH_FORMAT.isValid( identity ) ) {
				complain(
					"Message identity \"" + identity
					+ "\" is not in correct format "
					+ AN35_VAR_LENGTH_FORMAT + ".",
					"check",
					messageBegin,
					ErrorWatcher.ERROR
				);
			}
		}

		// Check contacts
		for (
			final Iterator iterator = messageBegin.contactIterator();
			iterator.hasNext();
		) {
			final Contact contact = (Contact) iterator.next();

			{
				final String cFunction = contact.getFunction();
				if (
					!Contact.DATA_PROCESSING.equals( cFunction )
					&&
					!Contact.HEAD_OF_DATA_PROCESSING.equals( cFunction )
					&&
					!Contact.INFORMATION_PRODUCTION.equals( cFunction )
					&&
					!Contact.HEAD_OF_INFORMATION_PRODUCTION.equals( cFunction )
				) {
					complain(
						"Contact function \"" + cFunction
						+ "\" is unknown.",
						"check",
						contact,
						ErrorWatcher.ERROR
					);
				}
			}
			{
				final String cId = contact.getIdentity();
				if ( !AN17_VAR_LENGTH_FORMAT.isValid( cId ) ) {
					complain(
						"Contact identity \"" + cId
						+ "\" is not in correct format "
						+ AN17_VAR_LENGTH_FORMAT + ".",
						"check",
						contact,
						ErrorWatcher.ERROR
					);
				}
			}
			{
				final String cName = contact.getName();
				if ( !AN35_VAR_LENGTH_FORMAT.isValid( cName ) ) {
					complain(
						"Contact name \"" + cName
						+ "\" is not in correct format "
						+ AN35_VAR_LENGTH_FORMAT + ".",
						"check",
						contact,
						ErrorWatcher.ERROR
					);
				}
			}

			for (
				final Iterator channels = contact.channelIterator();
				channels.hasNext();
			) {
				final ContactChannel channel = (ContactChannel) channels.next();
				{
					final String type = channel.getType();
					if (
						!ContactChannel.EMAIL.equals( type )
						&&
						!ContactChannel.FAX.equals( type )
						&&
						!ContactChannel.TELEPHONE.equals( type )
						&&
						!ContactChannel.X400.equals( type )
					) {
						complain(
							"ContactChannel type \"" + type
							+ "\" is unknown.",
							"check",
							contact,
							ErrorWatcher.ERROR
						);
					}
				}
				{
					final String number = channel.getNumber();
					if ( !AN512_VAR_LENGTH_FORMAT.isValid( number ) ) {
						complain(
							"ContactChannel \"number\" \"" + number
							+ "\" is not in correct format "
							+ AN512_VAR_LENGTH_FORMAT + ".",
							"check",
							contact,
							ErrorWatcher.ERROR
						);
					}
				}
			}
		}

		// ----------- Check data set related properties -----------
		// This must be done depending on the message function.

		if ( messageBegin.isStatisticalDefinitions() ) {
			// dataSetId must be empty.
			{
				final String dataSetId = messageBegin.getDataSetId();
				if ( !dataSetId.equals( "" ) ) {
					complain(
						"In a statistical definitions message"
						+ " a nonvoid data set identifier \""
						+ dataSetId + "\" is not allowed.",
						"check",
						messageBegin,
						ErrorWatcher.ERROR
					);
				}
			}

			// action must be empty.
			{
				final String action = messageBegin.getAction();
				if ( !action.equals( "" ) ) {
					complain(
						"In a statistical definitions message"
						+ " a nonvoid action \""
						+ action + "\" is not allowed.",
						"check",
						messageBegin,
						ErrorWatcher.ERROR
					);
				}
			}

			// extractionDateTime must be empty.
			{
				final String extractionDateTime
				= messageBegin.getExtractionDateTime();
				if ( !extractionDateTime.equals( "" ) ) {
					complain(
						"In a statistical definitions message"
						+ " a nonvoid extractionDateTime \""
						+ extractionDateTime + "\" is not allowed.",
						"check",
						messageBegin,
						ErrorWatcher.ERROR
					);
				}
			}

			// reportingPeriod must be empty.
			{
				final String reportingPeriod
				= messageBegin.getReportingPeriod();
				if ( !reportingPeriod.equals( "" ) ) {
					complain(
						"In a statistical definitions message"
						+ " a nonvoid reporting period \""
						+ reportingPeriod + "\" is not allowed.",
						"check",
						messageBegin,
						ErrorWatcher.ERROR
					);
				}
			}

			// reportingPeriodFormat must be empty.
			{
				final String reportingPeriodFormat
				= messageBegin.getReportingPeriodFormat();

				if ( !reportingPeriodFormat.equals( "" ) ) {
					complain(
						"In a statistical definitions message"
						+ " a nonvoid reporting period format \""
						+ reportingPeriodFormat + "\" is not allowed.",
						"check",
						messageBegin,
						ErrorWatcher.ERROR
					);
				}
			}

			// keyFamilyId must be empty.
			{
				final String keyFamilyId
				= messageBegin.getKeyFamilyId();

				if ( !keyFamilyId.equals( "" ) ) {
					complain(
						"In a statistical definitions message"
						+ " a nonvoid keyFamilyId \""
						+ keyFamilyId + "\" is not allowed.",
						"check",
						messageBegin,
						ErrorWatcher.ERROR
					);
				}
			}
		} else if ( messageBegin.isStatisticalData() ) {
			// dataSetId must be AN..35
			{
				final String dataSetId = messageBegin.getDataSetId();
				if ( !AN35_VAR_LENGTH_FORMAT.isValid( dataSetId ) ) {
					complain(
						"Invalid dataSetId \""
						+ dataSetId + "\".",
						"check",
						messageBegin,
						ErrorWatcher.ERROR
					);
				}
			}

			// Check action.
			{
				final String action = messageBegin.getAction();

				inUpdateMessage = messageBegin.isUpdateOrReplaceAction();
				inDeleteMessage = messageBegin.isDeleteAction();

				if (
					messageBegin.isTestingAction()
				) {
					complain(
						"Message has testing action \"" + action
						+ "\".",
						"check",
						messageBegin,
						ErrorWatcher.WARNING
					);
				} else if (
					!inUpdateMessage
					&&
					!inDeleteMessage
				) {
					complain(
						"Message action \"" + action
						+ "\" is invalid.",
						"check",
						messageBegin,
						ErrorWatcher.ERROR
					);
				}
			}

			// Check extractionDateTime.
			{
				final String parsePattern = "yyyyMMddHHmm";
				SimpleDateFormat formatter
				= new SimpleDateFormat ( parsePattern );
				formatter.setLenient( false );

				final String dateTime
				= messageBegin.getExtractionDateTime();

				if (!isXMLFile || 
						(isXMLFile && (dateTime != null) && (!dateTime.isEmpty()))) {
					if (dateTime.length() != parsePattern.length()) {
						complain("Message extractionDateTime \"" + dateTime
								+ "\" has a length different from "
								+ parsePattern.length() + " characters.",
								"check", messageBegin, ErrorWatcher.ERROR);
					}
					extractionDateTime = formatter.parse(dateTime,
							new ParsePosition(0));
					if (extractionDateTime == null) {
						complain("Message extractionDateTime \"" + dateTime
								+ "\" cannot be parsed.", "check",
								messageBegin, ErrorWatcher.ERROR);
					} else if (preparationDateTime != null
							&& extractionDateTime
									.compareTo(preparationDateTime) > 0) {
						final DateFormat dateFormat = new SimpleDateFormat(
								"dd MMM yyyy, HH:mm");
						complain("Interchange preparationDateTime "
								+ dateFormat.format(preparationDateTime)
								+ " precedes" + " message extractionDateTime "
								+ dateFormat.format(extractionDateTime) + ".",
								"check", messageBegin, ErrorWatcher.WARNING);
					}
				}
			}

			// Check reportingPeriod and -format.
			{
				final String reportingPeriod
				= messageBegin.getReportingPeriod();
				final String reportingPeriodFormat
				= messageBegin.getReportingPeriodFormat();

				messageReportingPeriod = null;

				if (
					!reportingPeriod.equals( "" )
					||
					!reportingPeriodFormat.equals( "" )
				) {
					try {
						messageReportingPeriod
						= PERIOD_FORMATTER.parse(
							reportingPeriod, reportingPeriodFormat
						);
						// TODO: Check that reportingperiod starts
						// after preparationdateTime?
						// No: Forecasts.

						if (
							messageReportingPeriod.getBegin().compareTo(
								messageReportingPeriod.getEnd()
							) > 0
						) {
							complain(
								"Negative reporting period \""
								+ messageReportingPeriod
								+ "\".",
								"check",
								messageBegin,
								ErrorWatcher.ERROR
							);
						}
					}
					catch ( ParseException pe ) {
						complain(
							"ReportingPeriod \""
							+ reportingPeriod
							+ "\" cannot be parsed: " + pe,
							"check",
							messageBegin,
							ErrorWatcher.ERROR
						);
					}
				}
			}

			// Set keyFamily.
			// keyFamilyId must be AN..35
			{
				final String keyFamilyId = messageBegin.getKeyFamilyId();

				if ( getMaintenanceAgency() != null ) {
					keyFamily = getMaintenanceAgency().getKeyFamily(
						keyFamilyId
					);
					if ( keyFamily == null ) {
						complain(
							"Maintenance agency \""
							+ getMaintenanceAgencyId()
							+ "\" does not know key family \""
							+ keyFamilyId + "\".",
							"check",
							messageBegin,
							ErrorWatcher.ERROR
						);
					}
				} else {
					keyFamily = null;
				}

				if ( !AN35_VAR_LENGTH_FORMAT.isValid( keyFamilyId ) ) {
					complain(
						"Invalid keyFamilyId \""
						+ keyFamilyId + "\".",
						"check",
						messageBegin,
						ErrorWatcher.ERROR
					);
				}
			}
		} else if ( messageBegin.isDataRequest() ) {
			complain(
				"Message function \"request for data\""
				+ " is not implemented.",
				"check",
				messageBegin,
				ErrorWatcher.ERROR
			);
		} else {
			final String function = messageBegin.getFunction();

			complain(
				"Message function \"" + function + "\" is invalid.",
				"check",
				messageBegin,
				ErrorWatcher.ERROR
			);
		}
	}

	/* ----------------------- MessageEnd ----------------------- */

	/**
	 * Check the contents of a MessageEnd.
	 */
	void check (
		final MessageEnd messageEnd
	) {
		super.check( messageEnd );

		if (
			messageBegin != null
			&&
			!messageEnd.isEndFor( messageBegin )
		) {
			complain(
				"MessageEnd carries different info than MessageBegin",
				"check",
				messageEnd,
				ErrorWatcher.ERROR
			);
		}
	}
}
