// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.semanticchecker;

import java.lang.Object;
import java.lang.String;
import java.lang.StringBuffer;

import junit.framework.Test;
import junit.framework.TestSuite;

import Int.ecb.gesmescb.InterchangeBegin;
import Int.ecb.gesmescb.InterchangeEnd;
import Int.ecb.gesmescb.ContactChannel;
import Int.ecb.gesmescb.Contact;
import Int.ecb.gesmescb.MessageBegin;
import Int.ecb.gesmescb.MessageEnd;

import Int.ecb.gesmescb.semanticchecker.EnvelopeChecker;

/* ***************************** Class ****************************** */

/**
 * Tests the implementation of the class {@link EnvelopeChecker}.
 *
 * @author
 * <a href="mailto:sis.external@ecb.europa.eu">Bernhard Bodenstorfer</a>
 *
 * @version @(#)$Id: EnvelopeCheckerTest.java,v 1.1 2017/01/19 16:14:05 trainot Exp $
 */

public class EnvelopeCheckerTest extends SequenceCheckerTest {

	/* ====================== Constructors ====================== */

	/**
	 * Constructs a SemanticCheckerTest with the specified name.
	 *
	 * @param name Test case name
	 */
	public EnvelopeCheckerTest ( String name ) {
		super( name );
	}

	/* ================== Protected Methods ===================== */

	/**
	 * Factory for EnvelopeCheckers.
	 * Should be overridden by derived classes.
	 *
	 * @see doTest(boolean,Object[])
	 */
	protected CheckerBase checkerFactory () {
		return new EnvelopeChecker ();
	}

	/**
	 * Generates a string of the specified length.
	 *
	 * @param length of the string.
	 * @param c is the character the string consists of.
	 */
	protected static String longString( final int length, final char c ) {
		final StringBuffer buffer = new StringBuffer( length );
		for ( int i = 0; i < length; ++i ) buffer.append( c );
		return buffer.toString();
	}

	/**
	 * Generates a string of the specified length.
	 */
	protected static String longString( final int length ) {
		return longString( length, 'B' );
	}

	/* ===================== Public Methods ===================== */

	/**
	 * Assembles and returns a test suite for all
	 * the test methods of this test case.
	 *
	 * @return a non-null test suite
	 */
	public static Test suite () {
		TestSuite suite = new TestSuite( EnvelopeCheckerTest.class );
		return suite;
	}

	/**
	 * Runs the test case
	 */
	public static void main ( String args[] ) {
		String[] testCaseName = { EnvelopeCheckerTest.class.getName() };
		junit.textui.TestRunner.main( testCaseName );
	}

	/* ---------------------- Test Methods ---------------------- */

	/* ------------------- Interchange Checks ------------------- */

	public void testIrev () {
		doTest( true, new Object[] {
			new InterchangeBegin(
				"IREV000001", "4F0", "ZZZ", "970525:0950", false
			), ERROR
		} );
	}

	public void testTooShortIref () {
		doTest( true, new Object[] {
			new InterchangeBegin(
				"IREF00001", "4F0", "ZZZ", "970525:0950", false
			), ERROR
		} );
	}

	public void testNegativeIref () {
		doTest( true, new Object[] {
			new InterchangeBegin(
				"IREF-00001", "4F0", "ZZZ", "970525:0950", false
			), ERROR
		} );
	}

	public void testUnparsableIref () {
		doTest( true, new Object[] {
			new InterchangeBegin(
				"IREF0x0001", "4F0", "ZZZ", "970525:0950", false
			), ERROR
		} );
	}

	public void testTooLongSender () {
		doTest( true, new Object[] {
			new InterchangeBegin(
				"IREF000001", longString( 36 ), "ZZZ", "970525:0950", false
			), ERROR,
			INTERCHANGE_END,
			null
		} );
	}

	public void testTooLongReceiver () {
		doTest( true, new Object[] {
			new InterchangeBegin(
				"IREF000001", "4F0", longString( 36 ), "970525:0950", false
			), ERROR,
			INTERCHANGE_END,
			null
		} );
	}

	public void testUnparsableDate () {
		doTest( true, new Object[] {
			new InterchangeBegin(
				"IREF000001", "4F0", "ZZZ", "97Jul29:0950", false
			), ERROR,
			INTERCHANGE_END,
			null
		} );
	}

	public void testInvalidPrepDate () {
		doTest( true, new Object[] {
			new InterchangeBegin(
				"IREF000001", "4F0", "ZZZ", "970229:0950", false
			), ERROR,
			INTERCHANGE_END,
			null
		} );
	}

	public void testInvalidPrepTime () {
		doTest( true, new Object[] {
			new InterchangeBegin(
				"IREF000001", "4F0", "ZZZ", "970227:0960", false
			), ERROR,
			INTERCHANGE_END,
			null
		} );
	}

	public void testNiceShortInterchange () {
		doTest( true, new Object[] {
			new InterchangeBegin(
				"IREF000001", "4F0", "ZZZ", "970525:0950", false
			),
			new InterchangeEnd( new String( "IREF000001" ) ),
			null
		} );
	}

	public void testBadShortInterchange () {
		doTest( true, new Object[] {
			new InterchangeBegin(
				"IREF000001", "4F0", "ZZZ", "970525:0950", false
			),
			new InterchangeEnd( "IREF000002" ), ERROR,
			null
		} );
	}

	/* --------------------- Message Checks --------------------- */

	public void testMruf () {
		doTest( true, new Object[] {
			INTERCHANGE_BEGIN,
			new MessageBegin(
				"MRUF000001", "4F0", "ZZZ", "ECB",
				MessageBegin.STATISTICAL_DEFINITIONS,
				"subject", contacts
			), ERROR,
			new MessageEnd( "MRUF000001" ),
			INTERCHANGE_END,
			null
		} );
	}

	public void testTooShortMref () {
		doTest( true, new Object[] {
			INTERCHANGE_BEGIN,
			new MessageBegin(
				"MREF00001", "4F0", "ZZZ", "ECB",
				MessageBegin.STATISTICAL_DEFINITIONS,
				"subject", contacts
			), ERROR,
			new MessageEnd( "MREF00001" ),
			INTERCHANGE_END,
			null
		} );
	}

	public void testTooLongMref () {
		doTest( true, new Object[] {
			INTERCHANGE_BEGIN,
			new MessageBegin(
				"MREF0000001", "4F0", "ZZZ", "ECB",
				MessageBegin.STATISTICAL_DEFINITIONS,
				"subject", contacts
			), ERROR,
			new MessageEnd( "MREF0000001" ),
			INTERCHANGE_END,
			null
		} );
	}

	public void testNegativeMref () {
		doTest( true, new Object[] {
			INTERCHANGE_BEGIN,
			new MessageBegin(
				"MREF-00001", "4F0", "ZZZ", "ECB",
				MessageBegin.STATISTICAL_DEFINITIONS,
				"subject", contacts
			), ERROR,
			new MessageEnd( "MREF-00001" ),
			INTERCHANGE_END,
			null
		} );
	}

	public void testUnparseableMref () {
		doTest( true, new Object[] {
			INTERCHANGE_BEGIN,
			new MessageBegin(
				"MREF0x0001", "4F0", "ZZZ", "ECB",
				MessageBegin.STATISTICAL_DEFINITIONS,
				"subject", contacts
			), ERROR,
			new MessageEnd( "MREF0x0001" ),
			INTERCHANGE_END,
			null
		} );
	}

	public void testBadFirstMref () {
		doTest( true, new Object[] {
			INTERCHANGE_BEGIN,
			new MessageBegin(
				"MREF000002", "4F0", "ZZZ", "ECB",
				MessageBegin.STATISTICAL_DEFINITIONS,
				"subject", contacts
			), ERROR,
			new MessageEnd( "MREF000002" ),
			INTERCHANGE_END,
			null
		} );
	}

	public void testBadSecondMref () {
		doTest( true, new Object[] {
			INTERCHANGE_BEGIN,
			METADATA_MESSAGE_BEGIN,
			METADATA_MESSAGE_END,
			new MessageBegin(
				"MREF000003", "4F0", "ZZZ", "ECB",
					MessageBegin.STATISTICAL_DEFINITIONS,
					"subject", contacts
				), ERROR,
				new MessageEnd( "MREF000003" ),
				INTERCHANGE_END,
				null
		} );
	}

	public void testTooLongSubject () {
		doTest( true, new Object[] {
			INTERCHANGE_BEGIN,
			new MessageBegin(
				METADATA_MESSAGE_BEGIN.getMref(),
				METADATA_MESSAGE_BEGIN.getSender(),
				METADATA_MESSAGE_BEGIN.getReceiver(),
				METADATA_MESSAGE_BEGIN.getMaintenanceAgencyId(),
				MessageBegin.STATISTICAL_DEFINITIONS,
				longString( 36 ),
				contacts
			), ERROR,
			new MessageEnd( "MREF000001" ),
			INTERCHANGE_END,
			null
		} );
	}

	public void testUnknownContactFunction () {
		doTest( true, new Object[] {
			INTERCHANGE_BEGIN,
			new MessageBegin(
				METADATA_MESSAGE_BEGIN.getMref(),
				METADATA_MESSAGE_BEGIN.getSender(),
				METADATA_MESSAGE_BEGIN.getReceiver(),
				METADATA_MESSAGE_BEGIN.getMaintenanceAgencyId(),
				MessageBegin.STATISTICAL_DEFINITIONS,
				METADATA_MESSAGE_BEGIN.getIdentity(),
				new Contact[] {
					new Contact (
						"AV",
						"Bop",
						"Sami",
						channels
					)
				}
			), ERROR,
			new MessageEnd( "MREF000001" ),
			INTERCHANGE_END,
			null
		} );
	}

	public void testTooLongContactIdentity () {
		doTest( true, new Object[] {
			INTERCHANGE_BEGIN,
			new MessageBegin(
				METADATA_MESSAGE_BEGIN.getMref(),
				METADATA_MESSAGE_BEGIN.getSender(),
				METADATA_MESSAGE_BEGIN.getReceiver(),
				METADATA_MESSAGE_BEGIN.getMaintenanceAgencyId(),
				MessageBegin.STATISTICAL_DEFINITIONS,
				METADATA_MESSAGE_BEGIN.getIdentity(),
				new Contact[] {
					new Contact (
						Contact.INFORMATION_PRODUCTION,
						longString( 18 ),
						"Sami",
						channels
					)
				}
			), ERROR,
			new MessageEnd( "MREF000001" ),
			INTERCHANGE_END,
			null
		} );
	}

	public void testTooLongContactName () {
		doTest( true, new Object[] {
			INTERCHANGE_BEGIN,
			new MessageBegin(
				METADATA_MESSAGE_BEGIN.getMref(),
				METADATA_MESSAGE_BEGIN.getSender(),
				METADATA_MESSAGE_BEGIN.getReceiver(),
				METADATA_MESSAGE_BEGIN.getMaintenanceAgencyId(),
				MessageBegin.STATISTICAL_DEFINITIONS,
				METADATA_MESSAGE_BEGIN.getIdentity(),
				new Contact[] {
					new Contact (
						Contact.INFORMATION_PRODUCTION,
						"BoP",
						longString( 36 ),
						channels
					)
				}
			), ERROR,
			new MessageEnd( "MREF000001" ),
			INTERCHANGE_END,
			null
		} );
	}

	public void unknownContactChannelType () {
		doTest( true, new Object[] {
			INTERCHANGE_BEGIN,
			new MessageBegin(
				METADATA_MESSAGE_BEGIN.getMref(),
				METADATA_MESSAGE_BEGIN.getSender(),
				METADATA_MESSAGE_BEGIN.getReceiver(),
				METADATA_MESSAGE_BEGIN.getMaintenanceAgencyId(),
				MessageBegin.STATISTICAL_DEFINITIONS,
				METADATA_MESSAGE_BEGIN.getIdentity(),
				new Contact[] {
					new Contact (
						Contact.INFORMATION_PRODUCTION,
						"BoP",
						"Sumo",
						new ContactChannel[] {
							new ContactChannel(
								"AV", "0815"
							)
						}
					)
				}
			), ERROR,
			new MessageEnd( "MREF000001" ),
			INTERCHANGE_END,
			null
		} );
	}

	public void tooLongContactChannelNumber () {
		doTest( true, new Object[] {
			INTERCHANGE_BEGIN,
			new MessageBegin(
				METADATA_MESSAGE_BEGIN.getMref(),
				METADATA_MESSAGE_BEGIN.getSender(),
				METADATA_MESSAGE_BEGIN.getReceiver(),
				METADATA_MESSAGE_BEGIN.getMaintenanceAgencyId(),
				MessageBegin.STATISTICAL_DEFINITIONS,
				METADATA_MESSAGE_BEGIN.getIdentity(),
				new Contact[] {
					new Contact (
						Contact.INFORMATION_PRODUCTION,
						"BoP",
						"Sumo",
						new ContactChannel[] {
							new ContactChannel(
								ContactChannel.TELEPHONE,
								longString( 513 )
							)
						}
					)
				}
			), ERROR,
			new MessageEnd( "MREF000001" ),
			INTERCHANGE_END,
			null
		} );
	}

	public void testUnknownMaintenanceAgency () {
		// Remark: Must use Statistical data message, because
		// a metadata message would ignore a not yet known M.A.
		doTest( true, new Object[] {
			INTERCHANGE_BEGIN,
			new MessageBegin(
				DATA_MESSAGE_BEGIN.getMref(),
				DATA_MESSAGE_BEGIN.getSender(),
				DATA_MESSAGE_BEGIN.getReceiver(),
				"Aardvark",
				DATA_MESSAGE_BEGIN.getFunction(),
				DATA_MESSAGE_BEGIN.getIdentity(),
				contacts,
				DATA_MESSAGE_BEGIN.getDataSetId(),
				DATA_MESSAGE_BEGIN.getAction(),
				DATA_MESSAGE_BEGIN.getExtractionDateTime(),
				DATA_MESSAGE_BEGIN.getReportingPeriod(),
				DATA_MESSAGE_BEGIN.getReportingPeriodFormat(),
				DATA_MESSAGE_BEGIN.getKeyFamilyId()
			), ERROR,
			DATA_MESSAGE_END,
			INTERCHANGE_END,
			null
		} );
	}

	public void testStatDefsWithNonzeroFields () {
		final String[] foo = {
			"ECB_BOP1",
			MessageBegin.UPDATE_OR_REPLACE_ACTION,
			"200012052219",
			"20002010",
			"702",
			"ECB_BOP1"
		};
		for ( int i = 0; i < foo.length; ++i ) {
			doTest( true, new Object[] {
				INTERCHANGE_BEGIN,
				new MessageBegin(
					"MREF000001", "4F0", "ZZZ", "ECB",
					MessageBegin.STATISTICAL_DEFINITIONS,
					"Junk data to shred", contacts,
					i == 0 ? foo[i] : "",
					i == 1 ? foo[i] : "",
					i == 2 ? foo[i] : "",
					i == 3 ? foo[i] : "",
					i == 4 ? foo[i] : "",
					i == 5 ? foo[i] : ""
				), ERROR,
				new MessageEnd( "MREF000001" ),
				INTERCHANGE_END,
				null
			} );
		}
	}

	public void testTooLongDSI () {
		doTest( true, new Object[] {
			INTERCHANGE_BEGIN,
			new MessageBegin(
				DATA_MESSAGE_BEGIN.getMref(),
				DATA_MESSAGE_BEGIN.getSender(),
				DATA_MESSAGE_BEGIN.getReceiver(),
				DATA_MESSAGE_BEGIN.getMaintenanceAgencyId(),
				DATA_MESSAGE_BEGIN.getFunction(),
				DATA_MESSAGE_BEGIN.getIdentity(),
				contacts,
				longString( 36 ),
				DATA_MESSAGE_BEGIN.getAction(),
				DATA_MESSAGE_BEGIN.getExtractionDateTime(),
				DATA_MESSAGE_BEGIN.getReportingPeriod(),
				DATA_MESSAGE_BEGIN.getReportingPeriodFormat(),
				DATA_MESSAGE_BEGIN.getKeyFamilyId()
			), ERROR,
			DATA_MESSAGE_END,
			INTERCHANGE_END,
			null
		} );
	}

	public void testInvalidAction () {
		doTest( true, new Object[] {
			INTERCHANGE_BEGIN,
			new MessageBegin(
				DATA_MESSAGE_BEGIN.getMref(),
				DATA_MESSAGE_BEGIN.getSender(),
				DATA_MESSAGE_BEGIN.getReceiver(),
				DATA_MESSAGE_BEGIN.getMaintenanceAgencyId(),
				DATA_MESSAGE_BEGIN.getFunction(),
				DATA_MESSAGE_BEGIN.getIdentity(),
				contacts,
				DATA_MESSAGE_BEGIN.getDataSetId(),
				"invalide",
				DATA_MESSAGE_BEGIN.getExtractionDateTime(),
				DATA_MESSAGE_BEGIN.getReportingPeriod(),
				DATA_MESSAGE_BEGIN.getReportingPeriodFormat(),
				DATA_MESSAGE_BEGIN.getKeyFamilyId()
			), ERROR,
			DATA_MESSAGE_END,
			INTERCHANGE_END,
			null
		} );
	}

	public void testTestingAction () {
		doTest( true, new Object[] {
			INTERCHANGE_BEGIN,
			new MessageBegin(
				DATA_MESSAGE_BEGIN.getMref(),
				DATA_MESSAGE_BEGIN.getSender(),
				DATA_MESSAGE_BEGIN.getReceiver(),
				DATA_MESSAGE_BEGIN.getMaintenanceAgencyId(),
				DATA_MESSAGE_BEGIN.getFunction(),
				DATA_MESSAGE_BEGIN.getIdentity(),
				contacts,
				DATA_MESSAGE_BEGIN.getDataSetId(),
				MessageBegin.TESTING_ACTION,
				DATA_MESSAGE_BEGIN.getExtractionDateTime(),
				DATA_MESSAGE_BEGIN.getReportingPeriod(),
				DATA_MESSAGE_BEGIN.getReportingPeriodFormat(),
				DATA_MESSAGE_BEGIN.getKeyFamilyId()
			), WARNING,
			DATA_MESSAGE_END,
			INTERCHANGE_END,
			null
		} );
	}

	public void testTooLateExtractionDateTime () {
		doTest( true, new Object[] {
			INTERCHANGE_BEGIN,
			new MessageBegin(
				DATA_MESSAGE_BEGIN.getMref(),
				DATA_MESSAGE_BEGIN.getSender(),
				DATA_MESSAGE_BEGIN.getReceiver(),
				DATA_MESSAGE_BEGIN.getMaintenanceAgencyId(),
				DATA_MESSAGE_BEGIN.getFunction(),
				DATA_MESSAGE_BEGIN.getIdentity(),
				contacts,
				DATA_MESSAGE_BEGIN.getDataSetId(),
				DATA_MESSAGE_BEGIN.getAction(),
				"200212051430",
				DATA_MESSAGE_BEGIN.getReportingPeriod(),
				DATA_MESSAGE_BEGIN.getReportingPeriodFormat(),
				DATA_MESSAGE_BEGIN.getKeyFamilyId()
			), WARNING,
			DATA_MESSAGE_END,
			INTERCHANGE_END,
			null
		} );
	}

	public void testInvalidExtractionDateTime () {
		doTest( true, new Object[] {
			INTERCHANGE_BEGIN,
			new MessageBegin(
				DATA_MESSAGE_BEGIN.getMref(),
				DATA_MESSAGE_BEGIN.getSender(),
				DATA_MESSAGE_BEGIN.getReceiver(),
				DATA_MESSAGE_BEGIN.getMaintenanceAgencyId(),
				DATA_MESSAGE_BEGIN.getFunction(),
				DATA_MESSAGE_BEGIN.getIdentity(),
				contacts,
				DATA_MESSAGE_BEGIN.getDataSetId(),
				DATA_MESSAGE_BEGIN.getAction(),
				"200202291340",
				DATA_MESSAGE_BEGIN.getReportingPeriod(),
				DATA_MESSAGE_BEGIN.getReportingPeriodFormat(),
				DATA_MESSAGE_BEGIN.getKeyFamilyId()
			), ERROR,
			DATA_MESSAGE_END,
			INTERCHANGE_END,
			null
		} );
	}

	public void testInvalidReportingPeriod () {
		doTest( true, new Object[] {
			INTERCHANGE_BEGIN,
			new MessageBegin(
				DATA_MESSAGE_BEGIN.getMref(),
				DATA_MESSAGE_BEGIN.getSender(),
				DATA_MESSAGE_BEGIN.getReceiver(),
				DATA_MESSAGE_BEGIN.getMaintenanceAgencyId(),
				DATA_MESSAGE_BEGIN.getFunction(),
				DATA_MESSAGE_BEGIN.getIdentity(),
				contacts,
				DATA_MESSAGE_BEGIN.getDataSetId(),
				DATA_MESSAGE_BEGIN.getAction(),
				DATA_MESSAGE_BEGIN.getExtractionDateTime(),
				"2001022920010320", "711",
				DATA_MESSAGE_BEGIN.getKeyFamilyId()
			), ERROR,
			DATA_MESSAGE_END,
			INTERCHANGE_END,
			null
		} );
	}

	public void testNegativeReportingPeriod () {
		doTest( true, new Object[] {
			INTERCHANGE_BEGIN,
			new MessageBegin(
				DATA_MESSAGE_BEGIN.getMref(),
				DATA_MESSAGE_BEGIN.getSender(),
				DATA_MESSAGE_BEGIN.getReceiver(),
				DATA_MESSAGE_BEGIN.getMaintenanceAgencyId(),
				DATA_MESSAGE_BEGIN.getFunction(),
				DATA_MESSAGE_BEGIN.getIdentity(),
				contacts,
				DATA_MESSAGE_BEGIN.getDataSetId(),
				DATA_MESSAGE_BEGIN.getAction(),
				DATA_MESSAGE_BEGIN.getExtractionDateTime(),
				"20022000", "702",
				DATA_MESSAGE_BEGIN.getKeyFamilyId()
			), ERROR,
			DATA_MESSAGE_END,
			INTERCHANGE_END,
			null
		} );
	}

	public void testNoReportingPeriod () {
		doTest( true, new Object[] {
			INTERCHANGE_BEGIN,
			new MessageBegin(
				DATA_MESSAGE_BEGIN.getMref(),
				DATA_MESSAGE_BEGIN.getSender(),
				DATA_MESSAGE_BEGIN.getReceiver(),
				DATA_MESSAGE_BEGIN.getMaintenanceAgencyId(),
				DATA_MESSAGE_BEGIN.getFunction(),
				DATA_MESSAGE_BEGIN.getIdentity(),
				contacts,
				DATA_MESSAGE_BEGIN.getDataSetId(),
				DATA_MESSAGE_BEGIN.getAction(),
				DATA_MESSAGE_BEGIN.getExtractionDateTime(),
				"", "",
				DATA_MESSAGE_BEGIN.getKeyFamilyId()
			),
			DATA_MESSAGE_END,
			INTERCHANGE_END,
			null
		} );
	}

	public void testInvalidReportingPeriodFormat () {
		doTest( true, new Object[] {
			INTERCHANGE_BEGIN,
			new MessageBegin(
				DATA_MESSAGE_BEGIN.getMref(),
				DATA_MESSAGE_BEGIN.getSender(),
				DATA_MESSAGE_BEGIN.getReceiver(),
				DATA_MESSAGE_BEGIN.getMaintenanceAgencyId(),
				DATA_MESSAGE_BEGIN.getFunction(),
				DATA_MESSAGE_BEGIN.getIdentity(),
				contacts,
				DATA_MESSAGE_BEGIN.getDataSetId(),
				DATA_MESSAGE_BEGIN.getAction(),
				DATA_MESSAGE_BEGIN.getExtractionDateTime(),
				"200103200105", "711x",
				DATA_MESSAGE_BEGIN.getKeyFamilyId()
			), ERROR,
			DATA_MESSAGE_END,
			INTERCHANGE_END,
			null
		} );
	}

	public void testUnknownReportingPeriodFormat () {
		doTest( true, new Object[] {
			INTERCHANGE_BEGIN,
			new MessageBegin(
				DATA_MESSAGE_BEGIN.getMref(),
				DATA_MESSAGE_BEGIN.getSender(),
				DATA_MESSAGE_BEGIN.getReceiver(),
				DATA_MESSAGE_BEGIN.getMaintenanceAgencyId(),
				DATA_MESSAGE_BEGIN.getFunction(),
				DATA_MESSAGE_BEGIN.getIdentity(),
				contacts,
				DATA_MESSAGE_BEGIN.getDataSetId(),
				DATA_MESSAGE_BEGIN.getAction(),
				DATA_MESSAGE_BEGIN.getExtractionDateTime(),
				"200103200105", "910",
				DATA_MESSAGE_BEGIN.getKeyFamilyId()
			), ERROR,
			DATA_MESSAGE_END,
			INTERCHANGE_END,
			null
		} );
	}

	public void testUnknownKF () {
		doTest( true, new Object[] {
			INTERCHANGE_BEGIN,
			new MessageBegin(
				DATA_MESSAGE_BEGIN.getMref(),
				DATA_MESSAGE_BEGIN.getSender(),
				DATA_MESSAGE_BEGIN.getReceiver(),
				DATA_MESSAGE_BEGIN.getMaintenanceAgencyId(),
				DATA_MESSAGE_BEGIN.getFunction(),
				DATA_MESSAGE_BEGIN.getIdentity(),
				contacts,
				DATA_MESSAGE_BEGIN.getDataSetId(),
				DATA_MESSAGE_BEGIN.getAction(),
				DATA_MESSAGE_BEGIN.getExtractionDateTime(),
				DATA_MESSAGE_BEGIN.getReportingPeriod(),
				DATA_MESSAGE_BEGIN.getReportingPeriodFormat(),
				"SUCCULENTS"
			), ERROR,
			DATA_MESSAGE_END,
			INTERCHANGE_END,
			null
		} );
	}

	public void testRequestFunction () {
		doTest( true, new Object[] {
			INTERCHANGE_BEGIN,
			new MessageBegin(
				METADATA_MESSAGE_BEGIN.getMref(),
				METADATA_MESSAGE_BEGIN.getSender(),
				METADATA_MESSAGE_BEGIN.getReceiver(),
				METADATA_MESSAGE_BEGIN.getMaintenanceAgencyId(),
				MessageBegin.DATA_REQUEST,
				METADATA_MESSAGE_BEGIN.getIdentity(),
				contacts
			), ERROR,
			METADATA_MESSAGE_END,
			INTERCHANGE_END,
			null
		} );
	}

	public void testBadShortMessage () {
		doTest( true, new Object[] {
			INTERCHANGE_BEGIN,
			METADATA_MESSAGE_BEGIN,
			getWithRef( METADATA_MESSAGE_END, "MREF000002" ), ERROR,
			INTERCHANGE_END,
			null
		} );
	}

	public void testMessageNumbering () {
		doTest( true, new Object[] {
			INTERCHANGE_BEGIN,
			METADATA_MESSAGE_BEGIN,
			METADATA_MESSAGE_END,
			METADATA_MESSAGE_BEGIN, ERROR,	// repeated Mref.
			METADATA_MESSAGE_END,
			INTERCHANGE_END,
			null
		} );
	}

	public void testMessageIncNumbering () {
		doTest( true, new Object[] {
			INTERCHANGE_BEGIN,
			METADATA_MESSAGE_BEGIN,
			METADATA_MESSAGE_END,
			getWithRef( METADATA_MESSAGE_BEGIN, "MREF000003" ), ERROR,
			getWithRef( METADATA_MESSAGE_END, "MREF000003" ),
			INTERCHANGE_END,
			null
		} );
	}
}
