// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, 2003, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.semanticchecker;

import java.text.ParseException;

import java.util.Map;
import java.util.HashMap;
import java.util.TreeMap;
import java.util.Iterator;

import Int.ecb.gesmescb.CodeList;
import Int.ecb.gesmescb.StatisticalConcept;
import Int.ecb.gesmescb.KeyFamily;
import Int.ecb.gesmescb.KeyFamilyComponent;
import Int.ecb.gesmescb.TextFormat;

import Int.ecb.gesmescb.error.ErrorWatcher;

import Int.ecb.gesmescb.semanticchecker.EnvelopeChecker;

/* ***************************** Class ****************************** */

/**
 * <P>Checks contents of
 * {@link Int.ecb.gesmescb.CodeList},
 * {@link Int.ecb.gesmescb.StatisticalConcept},
 * and {@link Int.ecb.gesmescb.KeyFamily} objects
 * in the information flow.</P>
 *
 * @author
 * <A HREF="mailto:klo@FAME.com">Kam Lo</A>
 * @author
 * <A HREF="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</A>
 *
 * @version @(#)$Id: MetadataChecker.java,v 1.1 2017/01/19 16:14:05 trainot Exp $
 */
class MetadataChecker extends EnvelopeChecker {

	/* ======================= Constants ======================== */

	protected static final TextFormat AN70_VAR_LENGTH_FORMAT
	= TextFormat.create( TextFormat.ANY, 70, true );
	protected static final TextFormat AN350_VAR_LENGTH_FORMAT
	= TextFormat.create( TextFormat.ANY, 350, true );

	protected static final TextFormat ID18_VAR_LENGTH_FORMAT
	= TextFormat.create( TextFormat.IDENTIFIER, 18, true );
	protected static final TextFormat ID35_VAR_LENGTH_FORMAT
	= TextFormat.create( TextFormat.IDENTIFIER, 35, true );

	protected static final TextFormat Code18_VAR_LENGTH_FORMAT
	= TextFormat.create( TextFormat.ANY, 18, true );
	protected static final TextFormat Code35_VAR_LENGTH_FORMAT
	= TextFormat.create( TextFormat.ANY, 35, true );

	private static final String[] FIXED_CONCEPTS
	= {
		"TIME_PERIOD",
		"TIME_FORMAT",
		"OBS_VALUE",
		"OBS_STATUS",
		"OBS_CONF",
		"OBS_PRE_BREAK"
	};
/** Array of alternative concepts, implemented for BP6. Should have the 
 * same length as FIXED_CONCEPTS. 
 * */
	private static final String[] ALTERNATIVE_CONCEPTS
	= {
		"",
		"",
		"",
		"",
		"CONF_STATUS",
		"PRE_BREAK_VALUE"
	};
	/* ========================= Fields ========================= */

	/* ====================== Constructors ====================== */

	/**
	 * Construct a new instance.
	 */
	public MetadataChecker () {
	}

	/* ==================== Package Methods ===================== */

	/* ------------------------ CodeList ------------------------ */

	/**
	 * Check the contents of a <CODE>CodeList</CODE>.
	 */
	void check (
		final CodeList codeList
	) {
		super.check( codeList );

		final String id = codeList.getId();

		// Check id is AN..18 (see Guide page 15).
		{
			final TextFormat format = ID18_VAR_LENGTH_FORMAT;
			final TextFormat formatAtLeast = ID35_VAR_LENGTH_FORMAT;

			if ( !format.isValid( id ) ) {
				if ( !formatAtLeast.isValid( id ) ) {
					complain(
						"CodeList id \"" + id
						+ "\" is not even in somewhat relaxed format "
						+ formatAtLeast + ".",
						"check",
						codeList,
						ErrorWatcher.ERROR
					);
				} else {
					complain(
						"CodeList id \"" + id
						+ "\" is not in correct mnemonic format "
						+ format + ".",
						"check",
						codeList,
						ErrorWatcher.WARNING
					);
				}
			}
		}

		// Check name is AN..70
		{
			final TextFormat format = AN70_VAR_LENGTH_FORMAT;
			final String name = codeList.getName();

			if ( !format.isValid( name ) ) {
				complain(
					"CodeList \"" + id
					+ "\" has name \"" + name
					+ "\" which is not in format "
					+ format + ".",
					"check",
					codeList,
					ErrorWatcher.ERROR
				);
			}
		}

		// Check description is AN..350
		{
			final TextFormat format = AN350_VAR_LENGTH_FORMAT;
			final String description = codeList.getDescription();

			if ( !format.isValid( description ) ) {
				complain(
					"CodeList id \"" + id
					+ "\" has description \"" + description
					+ "\" which is not in format "
					+ format + ".",
					"check",
					codeList,
					ErrorWatcher.ERROR
				);
			}
		}

		// Check nonempty.
		{
			final int fill = codeList.getCodeValueCount();

			if ( fill <= 0 ) {
				complain(
					"CodeList \"" + id
					+ "\" is empty.",
					"check",
					codeList,
					ErrorWatcher.ERROR
				);
			}
		}

		// Check uniqueness of code values
		// and matching value-description cardinalities.
		// Also check value and description formats.
		{
			final Map map = new HashMap();
			final Iterator values = codeList.codeValueIterator();
			final Iterator descriptions = codeList.codeDescriptionIterator();

			while ( values.hasNext() ) {
				final String value = (String)values.next();
				final String description;
				if ( descriptions.hasNext() ) {
					description = (String)descriptions.next();
				} else {
					complain(
						"CodeList \"" + id
						+ "\" contains value \""
						+ value + "\" without a description.",
						"check",
						codeList,
						ErrorWatcher.FAILURE
					);
					description = "<null>";
				}
				final Object origValue = map.put( value, description );
				if ( origValue != null ) {
					complain(
						"CodeList \"" + id
						+ "\" code \"" + value
						+ "\" with description \"" + origValue
						+ "\" clashes with same code with description \""
						+ description + "\".",
						"check",
						codeList,
						ErrorWatcher.ERROR
					);
				}
				{
					final TextFormat format
					= Code18_VAR_LENGTH_FORMAT;
					final TextFormat formatAtLeast
					= Code35_VAR_LENGTH_FORMAT;

					if ( !format.isValid( value ) ) {
						if ( !formatAtLeast.isValid( value ) ) {
							complain(
								"CodeList \"" + id
								+ "\" contains value \"" + value
								+ "\", which is not even"
								+ " in somewhat relaxed format "
								+ formatAtLeast + ".",
								"check",
								codeList,
								ErrorWatcher.ERROR
							);
						} else {
							complain(
								"CodeList \"" + id
								+ "\" contains value \"" + value
								+ "\", which is not in correct"
								+ " code value format "
								+ format + ".",
								"check",
								codeList,
								ErrorWatcher.WARNING
							);
						}
					}
				}
				{
					final TextFormat format = AN350_VAR_LENGTH_FORMAT;

					if ( !format.isValid( description ) ) {
						complain(
							"CodeList \"" + id
							+ "\" contains description " + description
							+ "\" which is not in format "
							+ format + ".",
							"check",
							codeList,
							ErrorWatcher.ERROR
						);
					}
				}
			}
			while ( descriptions.hasNext() ) {
				final String description = (String)descriptions.next();
				complain(
					"CodeList \"" + id
					+ "\" contains description \""
					+ description + "\" without a value.",
					"check",
					codeList,
					ErrorWatcher.FAILURE
				);
			}
		}
	}

	/* ------------------- StatisticalConcept ------------------- */

	/**
	 * Check the contents of a <CODE>StatisticalConcept</CODE>.
	 */
	void check (
		final StatisticalConcept statisticalConcept
	) {
		super.check( statisticalConcept );

		final String id = statisticalConcept.getId();

		// Check id is AN..18 (see Guide page 15).
		{
			final TextFormat format = ID18_VAR_LENGTH_FORMAT;
			final TextFormat formatAtLeast = ID35_VAR_LENGTH_FORMAT;

			if ( !format.isValid( id ) ) {
				if ( !formatAtLeast.isValid( id ) ) {
					complain(
						"StatisticalConcept id \"" + id
						+ "\" is not even in somewhat relaxed format "
						+ formatAtLeast + ".",
						"check",
						statisticalConcept,
						ErrorWatcher.ERROR
					);
				} else {
					complain(
						"StatisticalConcept id \"" + id
						+ "\" is not in correct mnemonic format "
						+ format + ".",
						"check",
						statisticalConcept,
						ErrorWatcher.WARNING
					);
				}
			}
		}

		// Check name is AN..70
		{
			final String name = statisticalConcept.getName();
			final TextFormat format = AN70_VAR_LENGTH_FORMAT;

			if ( !format.isValid( name ) ) {
				complain(
					"StatisticalConcept \"" + id
					+ "\" has name \"" + name
					+ "\" which is not in format "
					+ format + ".",
					"check",
					statisticalConcept,
					ErrorWatcher.ERROR
				);
			}
		}

		// Check description is AN..350
		{
			final String description = statisticalConcept.getDescription();
			final TextFormat format = AN350_VAR_LENGTH_FORMAT;

			if ( !format.isValid( description ) ) {
				complain(
					"StatisticalConcept \"" + id
					+ "\" has description \"" + description
					+ "\" which is not in format "
					+ format + ".",
					"check",
					statisticalConcept,
					ErrorWatcher.ERROR
				);
			}
		}

		// Check representation.
		{
			final String representation
			= statisticalConcept.getRepresentation();

			if ( !representation.equals( "" ) ) {
				complain(
					"StatisticalConcept \"" + id
					+ "\" has nonempty representation \"" + representation
					+ "\". This is not yet implemented. Ignoring.",
					"check",
					statisticalConcept,
					ErrorWatcher.WARNING
				);
			}
		}
	}

	/* ----------------------- KeyFamily ------------------------ */

	/**
	 * Check the contents of a <CODE>KeyFamily</CODE>.
	 */
	void check (
		final KeyFamily keyFamily
	) {
		super.check( keyFamily );

		final String id = keyFamily.getId();

		// Check id is AN..18 (see Guide page 15).
		{
			final TextFormat format = ID18_VAR_LENGTH_FORMAT;
			final TextFormat formatAtLeast = ID35_VAR_LENGTH_FORMAT;

			if ( !format.isValid( id ) ) {
				if ( !formatAtLeast.isValid( id ) ) {
					complain(
						"KeyFamily id \"" + id
						+ "\" is not even in somewhat relaxed format "
						+ formatAtLeast + ".",
						"check",
						keyFamily,
						ErrorWatcher.ERROR
					);
				} else {
					complain(
						"KeyFamily id \"" + id
						+ "\" is not in correct mnemonic format "
						+ format + ".",
						"check",
						keyFamily,
						ErrorWatcher.WARNING
					);
				}
			}
		}

		// Check name is AN..70
		{
			final TextFormat format = AN70_VAR_LENGTH_FORMAT;
			final String name = keyFamily.getName();

			if ( !format.isValid( name ) ) {
				complain(
					"KeyFamily id \"" + id
					+ "\" has name \"" + name
					+ "\" which is not in format "
					+ format + ".",
					"check",
					keyFamily,
					ErrorWatcher.ERROR
				);
			}
		}

		// Check description is AN..350
		{
			final TextFormat format = AN350_VAR_LENGTH_FORMAT;
			final String description = keyFamily.getDescription();

			if ( !format.isValid( description ) ) {
				complain(
					"KeyFamily id \"" + id
					+ "\" has description \"" + description
					+ "\" which is not in format "
					+ format + ".",
					"check",
					keyFamily,
					ErrorWatcher.ERROR
				);
			}
		}

		// Check KeyFamilyComponents:
		// That all key positions are occupied exactly once,
		// That no statistical concept appears two times.
		{
			final Map mapPositions = new TreeMap();
			final Map mapConcepts = new HashMap();

			for (
				final Iterator components = keyFamily.componentIterator();
				components.hasNext();
			) {
				final KeyFamilyComponent component
				= (KeyFamilyComponent)components.next();

				// First check Component contents:
				check( component, id );

				// Watch out for position clashes.
				{
					final String position = component.getPosition();

					if ( !component.isAttribute() ) {
						try {
							final Integer pos = Integer.valueOf( position );
							final Object oldValue = mapPositions.put(
								pos, component
							);
							if ( oldValue != null ) {
								complain(
									"KeyFamily \"" + id
									+ "\" component \""
									+ component.getConceptId()
									+ "\" clashes with "
									+ oldValue
									+ " on position " + position +".",
									"check",
									keyFamily,
									ErrorWatcher.ERROR
								);
							}
						}
						catch ( NumberFormatException nfe ) {
							complain(
								"KeyFamily \"" + id
								+ "\" component \"" + component.getConceptId()
								+ "\" has invalid position \""
								+ position + "\".",
								"check",
								keyFamily,
								ErrorWatcher.ERROR
							);
						}
					} else if ( !position.equals( "" ) ) {
						complain(
							"KeyFamily \"" + id
							+ "\" component \"" + component.getConceptId()
							+ "\" has nonvoid position " + position
							+ " though it is no key member nor array cell.",
							"check",
							keyFamily,
							ErrorWatcher.ERROR
						);
					}

				}

				// Watch out for statistical concept double occurrences.
				{
					final String conceptId = component.getConceptId();
					final Object oldValue = mapConcepts.put(
						conceptId, component
					);
					if ( oldValue != null ) {
						complain(
							"KeyFamily \"" + id
							+ "\" component \"" + component.getConceptId()
							+ "\" is used more than once.",
							"check",
							keyFamily,
							ErrorWatcher.ERROR
						);
					}
				}
			}

			final int positionRange = mapPositions.size();
			final int keyLength = keyFamily.getKeyDimensionCount();

			for ( int posi = 1; posi <= positionRange; ++posi ) {
				final Integer pos = new Integer( posi );
				final KeyFamilyComponent component
				= (KeyFamilyComponent)mapPositions.get( pos );

				if ( component == null ) {
					complain(
						"KeyFamily \"" + id
						+ "\" misses the concept at position "
						+ pos + " for the ARR segments.",
						"check",
						keyFamily,
						ErrorWatcher.ERROR
					);
				} else if ( posi <= keyLength ) {
					if ( !component.isKeyStructureDimension() ) {
						complain(
							"KeyFamily \"" + id
							+ "\" component \"" + component.getConceptId()
							+ "\" should be a key structure dimension.",
							"check",
							keyFamily,
							ErrorWatcher.ERROR
						);
					}
				} else {
					if ( posi <= keyLength + 2 ) {
						if ( !component.isTimeDimension() ) {
							complain(
								"KeyFamily \"" + id
								+ "\" component " + component
								+ "\" component \"" + component.getConceptId()
								+ "\" should be a time dimension.",
								"check",
								keyFamily,
								ErrorWatcher.ERROR
							);
						}
					} else if ( posi <= positionRange ) {
						if ( !component.isArrayCell() ) {
							complain(
								"KeyFamily \"" + id
								+ "\" component \"" + component.getConceptId()
								+ "\" should be an array cell.",
								"check",
								keyFamily,
								ErrorWatcher.ERROR
							);
						}
					}
					final int tailPosi = posi - keyLength - 1;
					if ( tailPosi >= FIXED_CONCEPTS.length ) {
						complain(
							"KeyFamily \"" + id
							+ "\" component \"" + component.getConceptId()
							+ "\" is an array cell too many.",
							"check",
							keyFamily,
							ErrorWatcher.ERROR
						);
					} else if ( (!(component.getConceptId().equals(
							FIXED_CONCEPTS[tailPosi]))
									&&
						!(component.getConceptId().equals(
							ALTERNATIVE_CONCEPTS[tailPosi]))) 
							        ||
							(component.getConceptId().isEmpty() )
					  )
					 {
						complain(
							"KeyFamily \"" + id
							+ "\" component \"" + component.getConceptId()
							+ "\" should be for concept \""
							+ FIXED_CONCEPTS[tailPosi]
						    + "\""
							+ " or alternatively for concept \""
							+ ALTERNATIVE_CONCEPTS[tailPosi]
							+ "\" (if alternative concept not empty)."
							+ "Key family component should not be empty.",
							"check",
							keyFamily,
							ErrorWatcher.ERROR
						);
					}
				}
				mapPositions.remove( pos );
			}
			for (
				Iterator rest = mapPositions.values().iterator();
				rest.hasNext();
			) {
				final KeyFamilyComponent component
				= (KeyFamilyComponent)rest.next();

				complain(
					"KeyFamily \"" + id
					+ "\" component \"" + component.getConceptId()
					+ "\" is out of position range [1,"
					+ positionRange + "].",
					"check",
					keyFamily,
					ErrorWatcher.ERROR
				);
			}

			final int frequencyPosition
			= keyFamily.getFrequencyDimensionPosition();
			if ( frequencyPosition == 0 ) {
				complain(
					"KeyFamily \"" + id
					+ "\" does not declare a frequency dimension.",
					"check",
					keyFamily,
					ErrorWatcher.ERROR
				);
			} else if ( frequencyPosition != 1 ) {
				complain(
					"KeyFamily \"" + id
					+ "\" declares the frequency dimension at position "
					+ frequencyPosition + ", different from first one.",
					"check",
					keyFamily,
					ErrorWatcher.ERROR
				);
			}

			if ( positionRange < keyLength + 4 ) {
				complain(
					"KeyFamily \"" + id + "\" does not declare"
					+ " required concepts for time and/or array cells.",
					"check",
					keyFamily,
					ErrorWatcher.ERROR
				);
			}
		}
	}

	/* ==================== Private Methods ===================== */

	/**
	 * Check the contents of a <CODE>KeyFamilyComponent</CODE>
	 * as far as they do not interrelate with the enclosing
	 * {@link KeyFamily}.
	 */
	private void check (
		final KeyFamilyComponent component,
		final String keyFamilyId
	) {
		final String conceptId = component.getConceptId();
		final boolean isObsValue
		= component.getConceptId().equals( "OBS_VALUE" );
		// TODO: Do not rely on id.

		// The usage status and attachment level
		// apply only to attributes and cells,
		// unless OBS_VALUE, which is a dirty exception.
		final boolean isDescribable
		= component.isAttribute()
			||
			component.isArrayCell()
			&&
			!isObsValue;

		// Check conceptId on existence and format.
		{
			if ( getMaintenanceAgency() != null ) {
				final StatisticalConcept concept
				= getMaintenanceAgency().getStatisticalConcept( conceptId );
				if ( concept == null ) {
					complain(
						"Problem in key family \"" + keyFamilyId
						+ "\": Maintenance agency \""
						+ getMaintenanceAgencyId()
						+ "\" does not know statistical concept \""
						+ conceptId + "\".",
						"check",
						component,
						ErrorWatcher.ERROR
					);
				}
			}
		}

		// Check on invalid conceptType.
		{
			final String conceptType = component.getConceptType();
			if (
				!conceptType.equals(
					KeyFamilyComponent.TIME_DIMENSION
				) && !conceptType.equals(
					KeyFamilyComponent.ARRAY_CELL
				) && !conceptType.equals(
					KeyFamilyComponent.FREQUENCY_DIMENSION
				) && !conceptType.equals(
					KeyFamilyComponent.KEY_STRUCTURE_DIMENSION
				) && !conceptType.equals(
					KeyFamilyComponent.ATTRIBUTE
				)
			) {
				complain(
					"Problem in key family \"" + keyFamilyId
					+ "\": Statistical concept \"" + conceptId
					+"\" has an invalid concept type \"" + conceptType
					+ "\".",
					"check",
					component,
					ErrorWatcher.ERROR
				);
			}
		}

		// Check position already done in check( KeyFamily ).

		// Check format
		TextFormat dataValueFormat = null;
		{
			// TODO: TextFormat precompute into KeyFamilyComponent.
			final String format = component.getFormat();

			try {
				dataValueFormat = TextFormat.parse( format );
			}
			catch ( ParseException pe ) {
				complain(
					"Problem in key family \"" + keyFamilyId
					+ "\": Statistical concept \"" + conceptId
					+ "\" specifies an invalid text format \"" + format
					+ "\".",
					"check",
					component,
					ErrorWatcher.ERROR
				);
			}
		}

		// Check usage status.
		{
			final String usageStatus = component.getUsageStatus();
			if ( isDescribable ) {
				if (
					!usageStatus.equals(
						KeyFamilyComponent.CONDITIONAL
					) && !usageStatus.equals(
						KeyFamilyComponent.MANDATORY
					)
				) {
					complain(
						"Problem in key family \"" + keyFamilyId
						+ "\": Statistical concept \"" + conceptId
						+ "\" has an invalid usage status \"" + usageStatus
						+ "\".",
						"check",
						component,
						ErrorWatcher.ERROR
					);
				}
			} else {
				if ( !usageStatus.equals( "" ) ) {
					if ( isObsValue ) {
						complain(
							"Problem in key family \"" + keyFamilyId
							+ "\": Statistical concept \"" + conceptId
							+ "\" specifies a usage status \"" + usageStatus
							+ "\".",
							"check",
							component,
							component.isMandatory()
								? ErrorWatcher.WARNING
								: ErrorWatcher.ERROR
						);
					} else {
						complain(
							"Problem in key family \"" + keyFamilyId
							+ "\": Statistical concept \"" + conceptId
							+ "\" specifies a nonempty usage status \""
							+ usageStatus
							+ "\" in non attribute or cell component.",
							"check",
							component,
							ErrorWatcher.ERROR
						);
					}
				}
			}
		}

		// Check attachment level.
		{
			final String attachmentLevel = component.getAttachmentLevel();
			if ( isDescribable ) {
				if (
					!attachmentLevel.equals(
						KeyFamilyComponent.ATTACHED_TO_OBSERVATION
					) && !attachmentLevel.equals(
						KeyFamilyComponent.ATTACHED_TO_TIME_SERIES
					) && !attachmentLevel.equals(
						KeyFamilyComponent.ATTACHED_TO_SIBLING_GROUP
					) && !attachmentLevel.equals(
						KeyFamilyComponent.ATTACHED_TO_DATA_SET
					)
				) {
					complain(
						"Problem in key family \"" + keyFamilyId
						+ "\": Statistical concept \"" + conceptId
						+ "\" has an invalid attachment level \""
						+ attachmentLevel
						+ "\".",
						"check",
						component,
						ErrorWatcher.ERROR
					);
				}
			} else {
				if ( !attachmentLevel.equals( "" ) ) {
					if ( isObsValue ) {
						complain(
							"Problem in key family \"" + keyFamilyId
							+ "\": Statistical concept \"" + conceptId
							+ "\" has a nonempty attachment level \""
							+ attachmentLevel
							+ "\".",
							"check",
							component,
							component.isAttachedToObservation()
								? ErrorWatcher.WARNING
								: ErrorWatcher.ERROR
						);
					} else {
						complain(
							"Problem in key family \"" + keyFamilyId
							+ "\": Statistical concept \"" + conceptId
							+ "\" has a nonempty attachment level \""
							+ attachmentLevel
							+ "\" in non attribute or cell component.",
							"check",
							component,
							ErrorWatcher.ERROR
						);
					}
				}
			}
		}

		// Check codeListId on existence and format
		{
			final String codeListId = component.getCodeListId();

			if ( component.isCoded() ) {
				if ( getMaintenanceAgency() != null ) {
					final CodeList codeList
					= getMaintenanceAgency().getCodeList( codeListId );
					if ( codeList == null ) {
						complain(
							"Problem in key family \"" + keyFamilyId
							+ "\" with statistical concept \"" + conceptId
							+ "\": Maintenance agency \""
							+ getMaintenanceAgencyId()
							+ "\" does not know CodeList \""
							+ codeListId + "\".",
							"check",
							component,
							ErrorWatcher.ERROR
						);
					} else {
						for (
							final Iterator iterator = codeList.codeValueIterator();
							iterator.hasNext();
						) {
							final String value = (String) iterator.next();
							if ( !dataValueFormat.isValid( value ) ) {
								complain(
									"Problem in key family \"" + keyFamilyId
									+ "\" with statistical concept \""
									+ conceptId
									+ "\": Code list \""
									+ codeListId
									+ "\" contains code \""
									+ value
									+ "\" which is not in format \""
									+ dataValueFormat
									+ "\".",
									"check",
									component,
									ErrorWatcher.ERROR
								);
							}
						}
					}
				}

				// Check associated codeListId
				{
					final TextFormat format
					= ID18_VAR_LENGTH_FORMAT;
					final TextFormat formatAtLeast
					= ID35_VAR_LENGTH_FORMAT;

					if ( !format.isValid( codeListId ) ) {
						if ( !formatAtLeast.isValid( codeListId ) ) {
							complain(
								"Problem in key family \"" + keyFamilyId
								+ "\" with statistical concept \"" + conceptId
								+ "\": CodeList id \"" + codeListId
								+ "\" is not even in somewhat relaxed format "
								+ formatAtLeast + ".",
								"check",
								keyFamily,
								ErrorWatcher.ERROR
							);
						} else {
							complain(
								"Problem in key family \"" + keyFamilyId
								+ "\" with statistical concept \"" + conceptId
								+ "\": CodeList id \"" + codeListId
								+ "\" is not in correct mnemonic format "
								+ format + ".",
								"check",
								keyFamily,
								ErrorWatcher.WARNING
							);
						}
					}
				}
			} else {
				if ( component.isKeyStructureDimension() ) {
					complain(
						"Problem in key family \"" + keyFamilyId
						+ "\": Uncoded dimension \"" + conceptId + "\".",
						"check",
						component,
						ErrorWatcher.ERROR
					);
				}
				if ( !codeListId.equals( "" ) ) {
					complain(
						"Problem in key family \"" + keyFamilyId
						+ "\" with statistical concept \"" + conceptId
						+ "\": CodeList \"" + codeListId
						+ "\" associated with uncoded concept.",
						"check",
						component,
						ErrorWatcher.ERROR
					);
				}
			}
		}
	}
}
