// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, 2003, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.semanticchecker;

import java.lang.Object;
import java.lang.String;

import junit.framework.Test;
import junit.framework.TestSuite;

import Int.ecb.gesmescb.CodeList;
import Int.ecb.gesmescb.StatisticalConcept;
import Int.ecb.gesmescb.KeyFamily;
import Int.ecb.gesmescb.KeyFamilyComponent;

import Int.ecb.gesmescb.semanticchecker.MetadataChecker;

/* ***************************** Class ****************************** */

/**
 * <P>Tests the implementation of the
 * class {@link MetadataChecker}.</P>
 *
 * @author
 * <A HREF="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</A>
 *
 * @version @(#)$Id: MetadataCheckerTest.java,v 1.1 2017/01/19 16:14:05 trainot Exp $
 */

public class MetadataCheckerTest extends EnvelopeCheckerTest {

	/* ====================== Constructors ====================== */

	/**
	 * Constructs a SemanticCheckerTest with the specified name.
	 *
	 * @param name Test case name
	 */
	public MetadataCheckerTest ( String name ) {
		super( name );
	}

	/* ================== Protected Methods ===================== */

	/**
	 * Factory for MetadataCheckers.
	 * Should be overridden by derived classes.
	 *
	 * @see doTest(Object[])
	 */
	protected CheckerBase checkerFactory () {
		return new MetadataChecker ();
	}

	/* ===================== Public Methods ===================== */

	/**
	 * Assembles and returns a test suite for all
	 * the test methods of this test case.
	 *
	 * @return a non-null test suite
	 */
	public static Test suite () {
		TestSuite suite = new TestSuite( MetadataCheckerTest.class );
		return suite;
	}

	/**
	 * Runs the test case
	 */
	public static void main ( String args[] ) {
		String[] testCaseName = { MetadataCheckerTest.class.getName() };
		junit.textui.TestRunner.main( testCaseName );
	}

	/* ---------------------- Test Methods ---------------------- */

	/* -------------------- CodeList Checks --------------------- */

	public void testGoodCodeList () {
		doTest( new Object[] {
			INTERCHANGE_BEGIN,
			METADATA_MESSAGE_BEGIN,
			new CodeList(
				"ID", "Name", "Description", versionDate,
				new String[] { "AT" }, new String[] { "Austria" }
			),
			METADATA_MESSAGE_END,
			INTERCHANGE_END,
			null
		} );
	}

	public void testTooLongCodeListId () {
		doTest( new Object[] {
			INTERCHANGE_BEGIN,
			METADATA_MESSAGE_BEGIN,
			new CodeList(
				"THIS_IS_MORE_THAN_18", "Name", "Description", versionDate,
				new String[] { "AT" }, new String[] { "Austria" }
			), WARNING,
			METADATA_MESSAGE_END,
			INTERCHANGE_END,
			null
		} );
	}

	public void testFarTooLongCodeListId () {
		doTest( new Object[] {
			INTERCHANGE_BEGIN,
			METADATA_MESSAGE_BEGIN,
			new CodeList(
				"THIS_IS_MORE_THAN_35_CHARACTERS_LONG",
				"Name", "Description", versionDate,
				new String[] { "AT" }, new String[] { "Austria" }
			), ERROR,
			METADATA_MESSAGE_END,
			INTERCHANGE_END,
			null
		} );
	}

	public void testBadCharsInCodeListId () {
		doTest( new Object[] {
			INTERCHANGE_BEGIN,
			METADATA_MESSAGE_BEGIN,
			new CodeList(
				"Id", "Name", "Description", versionDate,
				new String[] { "AT" }, new String[] { "Austria" }
			), ERROR,
			METADATA_MESSAGE_END,
			INTERCHANGE_END,
			null
		} );
	}

	public void testTooLongCodeListName () {
		doTest( new Object[] {
			INTERCHANGE_BEGIN,
			METADATA_MESSAGE_BEGIN,
			new CodeList(
				"ID",
				EnvelopeCheckerTest.longString( 71 ),
				"Description",
				versionDate,
				new String[] { "AT" }, new String[] { "Austria" }
			), ERROR,
			METADATA_MESSAGE_END,
			INTERCHANGE_END,
			null
		} );
	}

	public void testTooLongCodeListDesc () {
		doTest( new Object[] {
			INTERCHANGE_BEGIN,
			METADATA_MESSAGE_BEGIN,
			new CodeList(
				"ID", "Name", EnvelopeCheckerTest.longString( 351 ),
				versionDate,
				new String[] { "AT" }, new String[] { "Austria" }
			), ERROR,
			METADATA_MESSAGE_END,
			INTERCHANGE_END,
			null
		} );
	}

	public void testEmptyCodeList () {
		doTest( new Object[] {
			INTERCHANGE_BEGIN,
			METADATA_MESSAGE_BEGIN,
			new CodeList(
				"ID", "Name", "Description", versionDate,
				new String[] {}, new String[] {}
			), ERROR,
			METADATA_MESSAGE_END,
			INTERCHANGE_END,
			null
		} );
	}

	public void testTooFewValuesInCodeList () {
		doTest( new Object[] {
			INTERCHANGE_BEGIN,
			METADATA_MESSAGE_BEGIN,
			new CodeList(
				"ID", "Name", "Description", versionDate,
				new String[] { "AT" }, new String[] { "Austria", "France" }
			), FAILURE,
			METADATA_MESSAGE_END,
			INTERCHANGE_END,
			null
		} );
	}

	public void testTooFewDescriptionsInCodeList () {
		doTest( new Object[] {
			INTERCHANGE_BEGIN,
			METADATA_MESSAGE_BEGIN,
			new CodeList(
				"ID", "Name", "Description", versionDate,
				new String[] { "AT", "FR" }, new String[] { "Austria" }
			), FAILURE,
			METADATA_MESSAGE_END,
			INTERCHANGE_END,
			null
		} );
	}

	public void testDuplicateValuesInCodeList () {
		doTest( new Object[] {
			INTERCHANGE_BEGIN,
			METADATA_MESSAGE_BEGIN,
			new CodeList(
				"ID", "Name", "Description", versionDate,
				new String[] { "AT", "AT" },
				new String[] { "Austria", "Autriche" }
			), ERROR,
			METADATA_MESSAGE_END,
			INTERCHANGE_END,
			null
		} );
	}

	/* --------------- StatisticalConcept Checks ---------------- */

	public void testGoodStatisticalConcept () {
		doTest( new Object[] {
			INTERCHANGE_BEGIN,
			METADATA_MESSAGE_BEGIN,
			new StatisticalConcept(
				"ID", "Name", "Description", versionDate
			),
			METADATA_MESSAGE_END,
			INTERCHANGE_END,
			null
		} );
	}

	public void testTooLongStatisticalConceptId () {
		doTest( new Object[] {
			INTERCHANGE_BEGIN,
			METADATA_MESSAGE_BEGIN,
			new StatisticalConcept(
				"THIS_IS_MORE_THAN_18", "Name", "Description", versionDate
			), WARNING,
			METADATA_MESSAGE_END,
			INTERCHANGE_END,
			null
		} );
	}

	public void testFarTooLongStatisticalConceptId () {
		doTest( new Object[] {
			INTERCHANGE_BEGIN,
			METADATA_MESSAGE_BEGIN,
			new StatisticalConcept(
				"THIS_IS_MORE_THAN_35_CHARACTERS_LONG",
				"Name", "Description", versionDate
			), ERROR,
			METADATA_MESSAGE_END,
			INTERCHANGE_END,
			null
		} );
	}

	public void testBadCharsInStatisticalConceptId () {
		doTest( new Object[] {
			INTERCHANGE_BEGIN,
			METADATA_MESSAGE_BEGIN,
			new StatisticalConcept(
				"I.D", "Name", "Description", versionDate
			), ERROR,
			METADATA_MESSAGE_END,
			INTERCHANGE_END,
			null
		} );
	}

	public void testTooLongStatisticalConceptName () {
		doTest( new Object[] {
			INTERCHANGE_BEGIN,
			METADATA_MESSAGE_BEGIN,
			new StatisticalConcept(
				"ID", EnvelopeCheckerTest.longString( 71 ),
				"Description", versionDate
			), ERROR,
			METADATA_MESSAGE_END,
			INTERCHANGE_END,
			null
		} );
	}

	public void testTooLongStatisticalConceptDesc () {
		doTest( new Object[] {
			INTERCHANGE_BEGIN,
			METADATA_MESSAGE_BEGIN,
			new StatisticalConcept(
				"ID", "Name",
				EnvelopeCheckerTest.longString( 351 ),
				versionDate
			), ERROR,
			METADATA_MESSAGE_END,
			INTERCHANGE_END,
			null
		} );
	}

	public void testStatisticalConceptDescUnimplementedRepresentation () {
		doTest( new Object[] {
			INTERCHANGE_BEGIN,
			METADATA_MESSAGE_BEGIN,
			new StatisticalConcept(
				"ID", "Name", "Description", "Representation"
			), WARNING,
			METADATA_MESSAGE_END,
			INTERCHANGE_END,
			null
		} );
	}

	/* -------------------- KeyFamily Checks -------------------- */

	public void testGoodKeyFamily () {
		doTest( new Object[] {
			INTERCHANGE_BEGIN,
			METADATA_MESSAGE_BEGIN,
			new KeyFamily(
				"KF2", "Name", "Description", versionDate, KFC_ARR
			),
			METADATA_MESSAGE_END,
			INTERCHANGE_END,
			null
		} );
	}

	public void testTooLongKeyFamilyId () {
		doTest( new Object[] {
			INTERCHANGE_BEGIN,
			METADATA_MESSAGE_BEGIN,
			new KeyFamily(
				"THIS_IS_MORE_THAN_18", "Name", "Description", versionDate,
				KFC_ARR
			), WARNING,
			METADATA_MESSAGE_END,
			INTERCHANGE_END,
			null
		} );
	}

	public void testFarTooLongKeyFamilyId () {
		doTest( new Object[] {
			INTERCHANGE_BEGIN,
			METADATA_MESSAGE_BEGIN,
			new KeyFamily(
				"THIS_IS_MORE_THAN_35_CHARACTERS_LONG",
				"Name", "Description", versionDate, KFC_ARR
			), ERROR,
			METADATA_MESSAGE_END,
			INTERCHANGE_END,
			null
		} );
	}

	public void testBadCharsInKeyFamilyId () {
		doTest( new Object[] {
			INTERCHANGE_BEGIN,
			METADATA_MESSAGE_BEGIN,
			new KeyFamily(
				"I/D", "Name", "Description", versionDate, KFC_ARR
			), ERROR,
			METADATA_MESSAGE_END,
			INTERCHANGE_END,
			null
		} );
	}

	public void testTooLongKeyFamilyName () {
		doTest( new Object[] {
			INTERCHANGE_BEGIN,
			METADATA_MESSAGE_BEGIN,
			new KeyFamily(
				"ID", EnvelopeCheckerTest.longString( 71 ), "Description",
				versionDate, KFC_ARR
			), ERROR,
			METADATA_MESSAGE_END,
			INTERCHANGE_END,
			null
		} );
	}

	public void testTooLongKeyFamilyDesc () {
		doTest( new Object[] {
			INTERCHANGE_BEGIN,
			METADATA_MESSAGE_BEGIN,
			new KeyFamily(
				"ID", "Name",
				EnvelopeCheckerTest.longString( 351 ),
				versionDate,
				new KeyFamilyComponent[] {
					FREQ
				}
			), ERROR,
			METADATA_MESSAGE_END,
			INTERCHANGE_END,
			null
		} );
	}

	public void testUnknownConcept () {
		final KeyFamilyComponent[] kfcArr = (KeyFamilyComponent[]) KFC_ARR.clone();
		kfcArr[1] = new KeyFamilyComponent(
			"UNKNOWN",
			KeyFamilyComponent.KEY_STRUCTURE_DIMENSION,
			"2", "AN..35", "", "", "CL_DIM1"
		);
		doTest( new Object[] {
			INTERCHANGE_BEGIN,
			METADATA_MESSAGE_BEGIN,
			new KeyFamily(
				"ID", "Name", "Description", versionDate, kfcArr
			), ERROR,
			METADATA_MESSAGE_END,
			INTERCHANGE_END,
			null
		} );
	}

	public void testAlternateConcept () {
		final KeyFamilyComponent[] kfcArr = (KeyFamilyComponent[]) KFC_ARR.clone();
		kfcArr[1] = new KeyFamilyComponent(
			"CONF_STATUS",
			KeyFamilyComponent.KEY_STRUCTURE_DIMENSION,
			"2", "AN..35", "", "", "CL_DIM1"
		);
		doTest( new Object[] {
			INTERCHANGE_BEGIN,
			METADATA_MESSAGE_BEGIN,
			new KeyFamily(
				"ID", "Name", "Description", versionDate, kfcArr
			), ERROR,
			METADATA_MESSAGE_END,
			INTERCHANGE_END,
			null
		} );
	}
	
	public void testUnknownType () {
		final KeyFamilyComponent[] kfcArr = (KeyFamilyComponent[]) KFC_ARR.clone();
		kfcArr[1] = new KeyFamilyComponent(
			"EXTRA",
			"UNKNOWN",
			"2", "AN..35", "", "", "CL_DIM1"
		);
		doTest( new Object[] {
			INTERCHANGE_BEGIN,
			METADATA_MESSAGE_BEGIN,
			new KeyFamily(
				"ID", "Name", "Description", versionDate, kfcArr
			), ERROR,
			METADATA_MESSAGE_END,
			INTERCHANGE_END,
			null
		} );
	}

	public void testMissingFrequency () {
		final KeyFamilyComponent[] kfcArr = (KeyFamilyComponent[]) KFC_ARR.clone();
		kfcArr[0] = new KeyFamilyComponent(
			"FREQ", KeyFamilyComponent.KEY_STRUCTURE_DIMENSION,
			"1", "AN..1", "", "", "CL_FREQ", true
		);
		doTest( new Object[] {
			INTERCHANGE_BEGIN,
			METADATA_MESSAGE_BEGIN,
			new KeyFamily(
				"ID", "Name", "Description", versionDate, kfcArr
			), ERROR,
			METADATA_MESSAGE_END,
			INTERCHANGE_END,
			null
		} );
	}

	public void testSomeConceptsMissing () {
		doTest( new Object[] {
			INTERCHANGE_BEGIN,
			METADATA_MESSAGE_BEGIN,
			new KeyFamily(
				"KF2", "Name", "Description", versionDate,
				new KeyFamilyComponent[]{
					FREQ, DIM1, DIM2, TIME_PERIOD, TIME_FORMAT, OBS_VALUE
				}
			),
			ERROR,
			METADATA_MESSAGE_END,
			INTERCHANGE_END,
			null
		} );
	}

	public void testUncodedDimension () {
		final KeyFamilyComponent[] kfcArr = (KeyFamilyComponent[]) KFC_ARR.clone();
		kfcArr[1] = new KeyFamilyComponent(
			"EXTRA",
			KeyFamilyComponent.KEY_STRUCTURE_DIMENSION,
			"2", "AN..35", "", "", ""
		);
		doTest( new Object[] {
			INTERCHANGE_BEGIN,
			METADATA_MESSAGE_BEGIN,
			new KeyFamily(
				"ID", "Name", "Description", versionDate, kfcArr
			), ERROR,
			METADATA_MESSAGE_END,
			INTERCHANGE_END,
			null
		} );
	}

	public void testPositionCollision () {
		final KeyFamilyComponent[] kfcArr
			= new KeyFamilyComponent[KFC_ARR.length+1];
		System.arraycopy( KFC_ARR, 0, kfcArr, 0, KFC_ARR.length );
		kfcArr[KFC_ARR.length] = new KeyFamilyComponent(
			"EXTRA",
			KeyFamilyComponent.KEY_STRUCTURE_DIMENSION,
			"9", "AN..35", "", "", "CL_OBS_CONF"
		);
		doTest( new Object[] {
			INTERCHANGE_BEGIN,
			METADATA_MESSAGE_BEGIN,
			new KeyFamily(
				"ID", "Name", "Description", versionDate, kfcArr
			), ERROR,
			METADATA_MESSAGE_END,
			INTERCHANGE_END,
			null
		} );
	}

	public void testNoFirstPosition () {
		final KeyFamilyComponent[] kfcArr
			= new KeyFamilyComponent[KFC_ARR.length-1];
		System.arraycopy( KFC_ARR, 1, kfcArr, 0, KFC_ARR.length-1 );
		doTest( new Object[] {
			INTERCHANGE_BEGIN,
			METADATA_MESSAGE_BEGIN,
			new KeyFamily(
				"ID", "Name", "Description", versionDate, kfcArr
			), ERROR,
			METADATA_MESSAGE_END,
			INTERCHANGE_END,
			null
		} );
	}

	public void testMissingPosition () {
		final KeyFamilyComponent[] kfcArr
			= new KeyFamilyComponent[KFC_ARR.length];
		System.arraycopy( KFC_ARR, 0, kfcArr, 0, KFC_ARR.length );
		kfcArr[kfcArr.length-1] = new KeyFamilyComponent(
			"OBS_PRE_BREAK",
			KeyFamilyComponent.KEY_STRUCTURE_DIMENSION,
			"10", "AN..35", "", "", "CL_OBS_CONF"
		);
		doTest( new Object[] {
			INTERCHANGE_BEGIN,
			METADATA_MESSAGE_BEGIN,
			new KeyFamily(
				"ID", "Name", "Description", versionDate, kfcArr
			), ERROR,
			METADATA_MESSAGE_END,
			INTERCHANGE_END,
			null
		} );
	}

	public void testUnknownFormat () {
		final KeyFamilyComponent[] kfcArr = (KeyFamilyComponent[]) KFC_ARR.clone();
		kfcArr[kfcArr.length-1] = new KeyFamilyComponent(
			"EXTRA",
			KeyFamilyComponent.ATTRIBUTE,
			"", "AN..3x", KeyFamilyComponent.MANDATORY,
			KeyFamilyComponent.ATTACHED_TO_SIBLING_GROUP, ""
		);
		doTest( new Object[] {
			INTERCHANGE_BEGIN,
			METADATA_MESSAGE_BEGIN,
			new KeyFamily(
				"ID", "Name", "Description", versionDate, kfcArr
			), ERROR,
			METADATA_MESSAGE_END,
			INTERCHANGE_END,
			null
		} );
	}

	public void testMissingUsageStatus () {
		final KeyFamilyComponent[] kfcArr = (KeyFamilyComponent[]) KFC_ARR.clone();
		kfcArr[kfcArr.length-1] = new KeyFamilyComponent(
			"EXTRA",
			KeyFamilyComponent.ATTRIBUTE,
			"", "AN..35", "UNKNOWN",
			KeyFamilyComponent.ATTACHED_TO_SIBLING_GROUP, ""
		);
		doTest( new Object[] {
			INTERCHANGE_BEGIN,
			METADATA_MESSAGE_BEGIN,
			new KeyFamily(
				"ID", "Name", "Description", versionDate, kfcArr
			), ERROR,
			METADATA_MESSAGE_END,
			INTERCHANGE_END,
			null
		} );
	}

	public void testUnknownAttachmentLevel () {
		final KeyFamilyComponent[] kfcArr = (KeyFamilyComponent[]) KFC_ARR.clone();
		kfcArr[kfcArr.length-1] = new KeyFamilyComponent(
			"EXTRA",
			KeyFamilyComponent.ATTRIBUTE,
			"", "AN..35", KeyFamilyComponent.CONDITIONAL,
			"UNKNOWN", ""
		);
		doTest( new Object[] {
			INTERCHANGE_BEGIN,
			METADATA_MESSAGE_BEGIN,
			new KeyFamily(
				"ID", "Name", "Description", versionDate, kfcArr
			), ERROR,
			METADATA_MESSAGE_END,
			INTERCHANGE_END,
			null
		} );
	}

	public void testUnknownCodelist () {
		final KeyFamilyComponent[] kfcArr = (KeyFamilyComponent[]) KFC_ARR.clone();
		kfcArr[kfcArr.length-1] = new KeyFamilyComponent(
			"EXTRA",
			KeyFamilyComponent.ATTRIBUTE,
			"", "AN..35", KeyFamilyComponent.CONDITIONAL,
			KeyFamilyComponent.ATTACHED_TO_TIME_SERIES, "UNKNOWN"
		);
		doTest( new Object[] {
			INTERCHANGE_BEGIN,
			METADATA_MESSAGE_BEGIN,
			new KeyFamily(
				"ID", "Name", "Description", versionDate, kfcArr
			), ERROR,
			METADATA_MESSAGE_END,
			INTERCHANGE_END,
			null
		} );
	}

	public void testBadlyCodedConcept () {
		final KeyFamilyComponent[] kfcArr = (KeyFamilyComponent[]) KFC_ARR.clone();
		kfcArr[1] = new KeyFamilyComponent(
			"EXTRA",
			KeyFamilyComponent.KEY_STRUCTURE_DIMENSION,
			"2", "AN..5", "", "", "CL_BAD"
		);
		doTest( new Object[] {
			INTERCHANGE_BEGIN,
			METADATA_MESSAGE_BEGIN,
			new CodeList(
				"CL_BAD", "Contradicts text format", "", versionDate,
				new String[]{ "NO5LONG" },
				new String[]{ "Not AN..5" }
			),
			new KeyFamily(
				"ID", "Name", "Description", versionDate, kfcArr
			), ERROR,
			METADATA_MESSAGE_END,
			INTERCHANGE_END,
			null
		} );
	}
}
