// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.semanticchecker;

import java.lang.String;

import java.util.Iterator;

import Int.ecb.gesmescb.*;

import Int.ecb.gesmescb.error.ErrorWatcher;

/* ***************************** Class ****************************** */

/**
 * <P>Checks for compliance with PreBreak good practice.
 * </P>
 *
 * @author
 * <A HREF="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</A>
 *
 * @version @(#)$Id: PreBreakRestrictionsChecker.java,v 1.1 2017/01/19 16:14:05 trainot Exp $
 */
public class PreBreakRestrictionsChecker extends CheckerBase {

	/* ======================= Constants ======================== */

	/** Organisation code list name. */
	public static final String CODELIST_ORGANISATION = "CL_ORGANISATION";

	/** MaintenanceAgency code list name. */
	public static final String CODELIST_AGENCY = "CL_AGENCY";

	/** {@link Int.ecb.gesmescb.StatisticalConcept}s code list name. */
	public static final String CODELIST_CONCEPT = "CL_CONCEPT";

	protected static final TextFormat Coordinate18_VAR_LENGTH_FORMAT
	= TextFormat.create( TextFormat.COORDINATE, 18, true );

	/* ========================= Fields ========================= */

	/* ====================== Constructors ====================== */

	/**
	 * Construct a new instance.
	 */
	public PreBreakRestrictionsChecker () {
	}

	/* ==================== Package Methods ===================== */

	/* ----------------------- TimeSeries ----------------------- */

	/**
	 * Check the contents of a TimeSeries.
	 */
	void check (
		final TimeSeries timeSeries
	) {
		super.check( timeSeries );
		final String key = timeSeries.getKey();

		for (
			final Iterator iterator = timeSeries.observationIterator();
			iterator.hasNext();
		) {
			final Observation observation = (Observation) iterator.next();
			check( observation, key );
		}
	}

	private void check (
		final Observation observation,
		final String key
	) {
		final String status = observation.getStatus();
		final String preBreak = observation.getPreBreakValue();

		if ( !"".equals( preBreak ) && !"B".equals( status ) ) {
			complain(
				"Time Series \"" + key
				+ "\" has pre break value \""
				+ preBreak + "\" with status code \""
				+ status + "\" not indicating a break.",
				"check",
				observation,
				ErrorWatcher.ERROR
			);
		}
		if ( "".equals( preBreak ) && "B".equals( status ) ) {
			complain(
				"Time Series \"" + key
				+ "\" has empty pre break value \""
				+ preBreak + "\" with status code \""
				+ status + "\" indicating a break.",
				"check",
				observation,
				ErrorWatcher.WARNING
			);
		}
	}

	/* ----------------------- Attribute ------------------------ */

	/**
	 * Check the contents of an Attribute.
	 */
	void check (
		final Attribute attribute
	) {
		super.check( attribute );
	}

	/* ==================== Private Methods ===================== */

}
