// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.semanticchecker;

import Int.ecb.gesmescb.InterchangeBegin;
import Int.ecb.gesmescb.InterchangeEnd;
import Int.ecb.gesmescb.MessageBegin;
import Int.ecb.gesmescb.MessageEnd;
import Int.ecb.gesmescb.CodeList;
import Int.ecb.gesmescb.StatisticalConcept;
import Int.ecb.gesmescb.KeyFamily;
import Int.ecb.gesmescb.TimeSeries;
import Int.ecb.gesmescb.Attribute;

import Int.ecb.gesmescb.error.ErrorWatcher;

import Int.ecb.gesmescb.semanticchecker.CheckerBase;

/* ***************************** Class ****************************** */

/**
 * <p>Checks the correct sequence of GESMES/TS information flow.</p>
 *
 * <P>This responsibility concerns all properties of the information
 * flow which are primarily dependencies between successive pieces of
 * information rather than restrictions of the contents of these pieces.
 * For example, checking that
 * {@link CodeList}, {@link StatisticalConcept},
 * and {@link KeyFamily} objects only occur in metadata
 * messages and
 * {@link TimeSeries} and {@link Attribute} objects
 * only in data or delete messages is done here.</P>
 *
 * @author
 * <A HREF="mailto:klo@FAME.com">Kam Lo</A>
 * @author
 * <A HREF="mailto:Sami.Airo@ECB.int">Sami Airo</A>
 * @author
 * <A HREF="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</A>
 *
 * @version @(#)$Id: SequenceChecker.java,v 1.1 2017/01/19 16:14:05 trainot Exp $
 */
class SequenceChecker extends CheckerBase {

	/* ======================= Constants ======================== */

	/** Non event. */
	private static final int NONE = 0;

	/** Interchange begin event. */
	private static final int INTERCHANGE_BEGIN = 1;

	/** Interchange end event. */
	private static final int INTERCHANGE_END = 2;

	/** Message begin event. */
	private static final int MESSAGE_BEGIN = 3;

	/** Message end event. */
	private static final int MESSAGE_END = 4;

	/** Code list event. */
	private static final int CODE_LIST = 10;

	/** Statistical concept event. */
	private static final int STATISTICAL_CONCEPT = 11;

	/** Key family found event. */
	private static final int KEY_FAMILY = 12;

	/** Time series object found event. */
	private static final int TIME_SERIES_OR_ATTRIBUTE = 20;

	/**
	 * For messages of unknown or unsupported type.
	 * This includes request messages.
	 *
	 * @see #messageType
	 */
	private static final int UNSUPPORTED_MESSAGE = 0;

	/**
	 * For metadata messages.
	 * @see #messageType
	 */
	private static final int METADATA_MESSAGE = 1;

	/**
	 * For data messages.
	 * @see #messageType
	 */
	private static final int DATA_MESSAGE = 2;

	/* ========================= Fields ========================= */

	/**
	 * Type of last event checked.
	 */
	private int lastEventType;

	/**
	 * Whether the present message is concerned with metadata or with data.
	 *
	 * @see #UNSUPPORTED_MESSAGE
	 * @see #METADATA_MESSAGE
	 * @see #DATA_MESSAGE
	 */
	private int messageType;

	/* ====================== Constructors ====================== */

	/**
	 * Construct a new instance.
	 */
	public SequenceChecker () {
		super();
		lastEventType = NONE;
		messageType = UNSUPPORTED_MESSAGE;	// No known type so far.
	}

	/* ==================== Public Methods ====================== */

	/* ==================== Package Methods ===================== */

	/**
	 * <P>Checks that the InterchangeBegin is in correct sequence.</P>
	 */
	void check (
		final InterchangeBegin interchangeBegin
	) {
		super.check( interchangeBegin );

		if (
			lastEventType != NONE
			&&
			lastEventType != INTERCHANGE_END
		) {
			complain(
				"InterchangeBegin out of sequence",
				"check",
				interchangeBegin,
				ErrorWatcher.ERROR
			);
		}

		lastEventType = INTERCHANGE_BEGIN;
	}

	/**
	 * <P>Checks that the InterchangeEnd is in correct sequence.</P>
	 */
	void check (
		final InterchangeEnd interchangeEnd
	) {
		super.check( interchangeEnd );

		if (
			lastEventType != INTERCHANGE_BEGIN
			&&
			lastEventType != MESSAGE_END
		) {
			complain(
				"InterchangeEnd out of sequence",
				"check",
				interchangeEnd,
				ErrorWatcher.ERROR
			);
		}

		lastEventType = INTERCHANGE_END;
	}

	/**
	 * <P>Checks that a MessageBegin is
	 * preceded by InterchangeBegin or MessageEnd.</P>
	 */
	void check (
		final MessageBegin messageBegin
	) {
		super.check( messageBegin );

		if (
			lastEventType != INTERCHANGE_BEGIN
			&&
			lastEventType != MESSAGE_END
		) {
			complain(
				"MessageBegin out of sequence",
				"check",
				messageBegin,
				ErrorWatcher.ERROR
			);
		}

		if ( messageBegin.isStatisticalDefinitions() ) {
			messageType = METADATA_MESSAGE;
		} else if ( messageBegin.isStatisticalData() ) {
			messageType = DATA_MESSAGE;
		} else {
			complain(
				"Message BEGIN with unsupported type "
				+ messageBegin.getFunction(),
				"check",
				messageBegin,
				ErrorWatcher.ERROR
			);
			messageType = UNSUPPORTED_MESSAGE;
		}

		lastEventType = MESSAGE_BEGIN;
	}

	/**
	 * <P>Checks that a MessageEnd is preceded by
	 * MessageBegin,
	 * CodeList, StatisticalConcept, KeyFamily,
	 * TimeSeries, or Attribute.</P>
	 */
	void check (
		final MessageEnd messageEnd
	) {
		super.check( messageEnd );

		if (
			lastEventType != MESSAGE_BEGIN
			&& lastEventType != CODE_LIST
			&& lastEventType != STATISTICAL_CONCEPT
			&& lastEventType != KEY_FAMILY
			&& lastEventType != TIME_SERIES_OR_ATTRIBUTE
		) {
			complain(
				"MessageEnd out of sequence.",
				"check",
				messageEnd,
				ErrorWatcher.ERROR
			);
		}

		lastEventType = MESSAGE_END;
		messageType = UNSUPPORTED_MESSAGE;	// no type active.
	}

	/**
	 * <P>Checks if a CodeList is in correct sequence.</P>
	 */
	void check (
		final CodeList codeList
	) {
		super.check( codeList );

		if (
			lastEventType != MESSAGE_BEGIN
			&& lastEventType != CODE_LIST
		) {
			complain(
				"CodeList out of sequence",
				"check",
				codeList,
				ErrorWatcher.ERROR
			);
		}

		if ( messageType != METADATA_MESSAGE ) {
			complain(
				"CodeList outside metadata message",
				"check",
				codeList,
				ErrorWatcher.ERROR
			);
		}

		lastEventType = CODE_LIST;
	}

	/**
	 * <P>Checks if a StatisticalConcept is in correct sequence.</P>
	 */
	void check (
		final StatisticalConcept statisticalConcept
	) {
		super.check( statisticalConcept );

		if (
			lastEventType != MESSAGE_BEGIN
			&& lastEventType != CODE_LIST
			&& lastEventType != STATISTICAL_CONCEPT
		) {
			complain(
				"StatisticalConcept out of sequence",
				"check",
				statisticalConcept,
				ErrorWatcher.ERROR
			);
		}

		if ( messageType != METADATA_MESSAGE ) {
			complain(
				"StatisticalConcept outside metadata message",
				"check",
				statisticalConcept,
				ErrorWatcher.ERROR
			);
		}

		lastEventType = STATISTICAL_CONCEPT;
	}

	/**
	 * <P>Checks if a KeyFamily is in correct sequence.</P>
	 */
	void check (
		final KeyFamily keyFamily
	) {
		super.check( keyFamily );

		if (
			lastEventType != MESSAGE_BEGIN
			&& lastEventType != STATISTICAL_CONCEPT
			&& lastEventType != CODE_LIST
			&& lastEventType != KEY_FAMILY
		) {
			complain(
				"KeyFamily out of sequence.",
				"check",
				keyFamily,
				ErrorWatcher.ERROR
			);
		}

		if ( messageType != METADATA_MESSAGE ) {
			complain(
				"KeyFamily outside metadata message",
				"check",
				keyFamily,
				ErrorWatcher.ERROR
			);
		}

		lastEventType = KEY_FAMILY;
	}

	/**
	 * <P>Checks if a TimeSeries is in correct sequence.</P>
	 */
	void check (
		final TimeSeries timeSeries
	) {
		super.check( timeSeries );

		if (
			lastEventType != MESSAGE_BEGIN
			&&
			lastEventType != TIME_SERIES_OR_ATTRIBUTE
		) {
			complain(
				"TimeSeries out of sequence",
				"check",
				timeSeries,
				ErrorWatcher.ERROR
			);
		}

		if ( messageType != DATA_MESSAGE ) {
			complain(
				"TimeSeries outside data message",
				"check",
				timeSeries,
				ErrorWatcher.ERROR
			);
		}

		lastEventType = TIME_SERIES_OR_ATTRIBUTE;
	}

	/**
	 * <P>Checks if a Attribute is in correct sequence.</P>
	 */
	void check (
		final Attribute attribute
	) {
		super.check( attribute );

		if (
			lastEventType != MESSAGE_BEGIN
			&& lastEventType != TIME_SERIES_OR_ATTRIBUTE
		) {
			complain(
				"Attribute out of sequence",
				"check",
				attribute,
				ErrorWatcher.ERROR
			);
		}

		if ( messageType != DATA_MESSAGE ) {
			complain(
				"Attribute outside data message",
				"check",
				attribute,
				ErrorWatcher.ERROR
			);
		}

		lastEventType = TIME_SERIES_OR_ATTRIBUTE;
	}

	/**
	 * <P>Checks if the end of information flow is allowed here.</P>
	 */
	public void close () {
		super.close();
		if (
			lastEventType != NONE
			&&
			lastEventType != INTERCHANGE_END
		) {
			complain(
				"Unclosed interchange",
				"close",
				null,
				ErrorWatcher.ERROR
			);
		}
	}
}
