// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.semanticchecker;

import java.lang.Object;

import junit.framework.Test;
import junit.framework.TestSuite;

import Int.ecb.gesmescb.semanticchecker.CheckerBase;
import Int.ecb.gesmescb.semanticchecker.SequenceChecker;

/* ***************************** Class ****************************** */

/**
 * Tests the implementation of the class {@link SequenceChecker}.
 *
 * @author
 * <a href="mailto:sis.external@ecb.europa.eu">Sami Airo</a>
 * @author
 * <a href="mailto:sis.external@ecb.europa.eu">Bernhard Bodenstorfer</a>
 *
 * @version @(#)$Id: SequenceCheckerTest.java,v 1.1 2017/01/19 16:14:05 trainot Exp $
 */

public class SequenceCheckerTest extends CheckerBaseTest {

	/* ====================== Constructors ====================== */

	/**
	 * Constructs a SequenceCheckerTest with the specified name.
	 *
	 * @param name Test case name
	 */
	public SequenceCheckerTest ( String name ) {
		super( name );
	}

	/* ================== Protected Methods ===================== */

	/**
	 * Factory for SequenceCheckers.
	 * Should be overridden by derived classes.
	 *
	 * @see #doTest(boolean,Object[])
	 */
	protected CheckerBase checkerFactory () {
		return new SequenceChecker ();
	}

	/* ===================== Public Methods ===================== */

	/**
	 * Assembles and returns a test suite for all
	 * the test methods of this test case.
	 *
	 * @return a non-null test suite
	 */
	public static Test suite () {
		TestSuite suite = new TestSuite( SequenceCheckerTest.class );
		return suite;
	}

	/**
	 * Runs the test case
	 */
	public static void main ( String args[] ) {
		String[] testCaseName = { SequenceCheckerTest.class.getName() };
		junit.textui.TestRunner.main( testCaseName );
	}

	/* ---------------------- Test Methods ---------------------- */

	/**
	 * Tests interchanges.
	 */
	public void testSuccessInterchange () {
		final Object objects[] = {
			// empty interchange
			INTERCHANGE_BEGIN,
			INTERCHANGE_END,
			// interchange with one message
			INTERCHANGE_BEGIN,
			DATA_MESSAGE_BEGIN,
			DATA_MESSAGE_END,
			INTERCHANGE_END,
			// interchange with two messages
			getWithRef( INTERCHANGE_BEGIN, "IREF000002" ),
			METADATA_MESSAGE_BEGIN,
			METADATA_MESSAGE_END,
			getWithRef( DATA_MESSAGE_BEGIN, "MREF000002" ),
			getWithRef( DATA_MESSAGE_END, "MREF000002" ),
			getWithRef( INTERCHANGE_END, "IREF000002" ),
			null
		};

		doTest( true, objects );
	}

	/**
	 * Tests the handling of messages.
	 */
	public void testSuccessMessage () {
		final Object objects[] = {
			// multiple messages
			INTERCHANGE_BEGIN,

			// followed by concept
			METADATA_MESSAGE_BEGIN,
			SC_FREQ,
			METADATA_MESSAGE_END,

			// followed by code list
			getWithRef( METADATA_MESSAGE_BEGIN, "MREF000002" ),
			CL_FREQ,
			getWithRef( METADATA_MESSAGE_END, "MREF000002" ),

			// followed by key family
			getWithRef( METADATA_MESSAGE_BEGIN, "MREF000003" ),
			KF1,
			getWithRef( METADATA_MESSAGE_END, "MREF000003" ),

			// followed by time series
			getWithRef( DATA_MESSAGE_BEGIN, "MREF000004" ),
			TS1,
			getWithRef( DATA_MESSAGE_END, "MREF000004" ),

			// followed by AT2
			getWithRef( DATA_MESSAGE_BEGIN, "MREF000005" ),
			AT2,
			getWithRef( DATA_MESSAGE_END, "MREF000005" ),

			INTERCHANGE_END,
			null
		};

		doTest( true, objects );
	}

	/**
	 * Tests the handling of concepts.
	 */
	public void testSuccessStatisticalConcept () {
		final Object objects[] = {
			INTERCHANGE_BEGIN,
			// multiple concepts
			METADATA_MESSAGE_BEGIN,
			SC_FREQ,
			SC_FREQ,
			METADATA_MESSAGE_END,
			// followed by code list
			getWithRef( METADATA_MESSAGE_BEGIN, "MREF000002" ),
			CL_FREQ,
			SC_FREQ,
			getWithRef( METADATA_MESSAGE_END, "MREF000002" ),
			// followed by key family
			getWithRef( METADATA_MESSAGE_BEGIN, "MREF000003" ),
			SC_FREQ,
			KF1,
			getWithRef( METADATA_MESSAGE_END, "MREF000003" ),
			INTERCHANGE_END,
			null
		};

		doTest( true, objects );
	}

	/**
	 * Tests the handling of code lists.
	 */
	public void testSuccessCodeList () {
		final Object objects[] = {
			INTERCHANGE_BEGIN,
			// multiple code lists
			METADATA_MESSAGE_BEGIN,
			CL_FREQ,
			CL_FREQ,
			METADATA_MESSAGE_END,
			// followed by key family
			getWithRef( METADATA_MESSAGE_BEGIN, "MREF000002" ),
			CL_FREQ,
			KF1,
			getWithRef( METADATA_MESSAGE_END, "MREF000002" ),
			INTERCHANGE_END,
			null
		};

		doTest( true, objects );
	}

	/**
	 * Tests the handling of a normal interchange with definition messages and
	 * various combinations of items.
	 */
	public void testSuccessKeyFamily () {
		final Object objects[] = {
			INTERCHANGE_BEGIN,
			// multiple key families
			METADATA_MESSAGE_BEGIN,
			KF1,
			KF1,
			METADATA_MESSAGE_END,
			INTERCHANGE_END,
			null
		};

		doTest( true, objects );
	}

	/**
	 * Tests the handling of a time series and an AT2.
	 */
	public void testSuccessDataSet () {
		final Object objects[] = {
			INTERCHANGE_BEGIN,
			// Data set followed by time series
			DATA_MESSAGE_BEGIN,
			TS1,
			DATA_MESSAGE_END,
			// Data set followed by an AT2
			getWithRef( DATA_MESSAGE_BEGIN, "MREF000002" ),
			AT2,
			getWithRef( DATA_MESSAGE_END, "MREF000002" ),
			INTERCHANGE_END,
			null
		};

		doTest( true, objects );
	}

	/**
	 * Tests the handling of time series.
	 */
	public void testSuccessTimeseries () {
		final Object objects[] = {
			INTERCHANGE_BEGIN,
			// multiple time series
			DATA_MESSAGE_BEGIN,
			TS1,
			TS1,
			DATA_MESSAGE_END,
			// Time series and AT2
			getWithRef( DATA_MESSAGE_BEGIN, "MREF000002" ),
			TS1,
			AT2,
			getWithRef( DATA_MESSAGE_END, "MREF000002" ),
			INTERCHANGE_END,
			null
		};

		doTest( true, objects );
	}

	/**
	 * Tests the handling of an AT2.
	 */
	public void testSuccessAttribute () {
		final Object objects[] = {
			INTERCHANGE_BEGIN,
			// multiple AT2s
			DATA_MESSAGE_BEGIN,
			AT2,
			AT2,
			DATA_MESSAGE_END,
			INTERCHANGE_END,
			null
		};

		doTest( true, objects );
	}

	/**
	 * Test reaction to an object outside an interchange.
	 */
	public void testEventBeforeInterchange () {
		final Object objects[] = {
			DATA_MESSAGE_BEGIN, ERROR,
			INTERCHANGE_BEGIN,
			INTERCHANGE_END,
			null
		};

		doTest( false, objects );
	}

	/**
	 * Test reaction to an unclosed message.
	 */
	public void testUnclosedMessageInterchange () {
		final Object objects[] = {
			INTERCHANGE_BEGIN,
			DATA_MESSAGE_BEGIN,
			INTERCHANGE_END, ERROR,
			null
		};

		doTest( false, objects );
	}

	/**
	 * Test reaction to a skew opened message.
	 */
	public void testSkewClosedMessageInterchange () {
		final Object objects[] = {
			INTERCHANGE_BEGIN,
			METADATA_MESSAGE_BEGIN,
			getWithRef( DATA_MESSAGE_BEGIN, "MREF000002" ), ERROR,
			INTERCHANGE_END,
			null
		};

		doTest( false, objects );
	}

	/**
	 * Test reaction to a missing interchange end.
	 */
	public void testUnclosedInterchange () {
		final Object objects[] = {
			INTERCHANGE_BEGIN,
			METADATA_MESSAGE_BEGIN,
			METADATA_MESSAGE_END,
			null, ERROR
		};

		doTest( false, objects );
	}

	/**
	 * Test reaction to duplicate messages.
	 */
	public void testDuplicateMessage () {
		final Object objects[] = {
			INTERCHANGE_BEGIN,
			METADATA_MESSAGE_BEGIN,
			getWithRef( METADATA_MESSAGE_BEGIN, "MREF000002" ), ERROR,
			METADATA_MESSAGE_END,
			METADATA_MESSAGE_END,
			INTERCHANGE_END,
			null
		};

		doTest( false, objects );
	}

	/**
	 * Test reaction to a misplaced code list.
	 */
	public void testMisplacedCodeList () {
		final Object objects[] = {
			INTERCHANGE_BEGIN,
			METADATA_MESSAGE_BEGIN,
			METADATA_MESSAGE_END,
			CL_FREQ, ERROR,
			METADATA_MESSAGE_END,
			INTERCHANGE_END,
			null
		};

		doTest( false, objects );
	}

	/**
	 * Test reaction to a misplaced statistical concept.
	 */
	public void testMisplacedStatisticalConcept () {
		final Object objects[] = {
			INTERCHANGE_BEGIN,
			METADATA_MESSAGE_BEGIN,
			METADATA_MESSAGE_END,
			SC_FREQ, ERROR,
			METADATA_MESSAGE_END,
			INTERCHANGE_END,
			null
		};

		doTest( false, objects );
	}

	/**
	 * Test reaction to a misplaced key family.
	 */
	public void testMisplacedKeyFamily () {
		final Object objects[] = {
			INTERCHANGE_BEGIN,
			DATA_MESSAGE_BEGIN,
			DATA_MESSAGE_END,
			KF1, ERROR,
			DATA_MESSAGE_END,
			INTERCHANGE_END,
			null
		};

		doTest( false, objects );
	}

	/**
	 * Test reaction to a misplaced time series.
	 */
	public void testMisplacedTimeSeries () {
		final Object objects[] = {
			INTERCHANGE_BEGIN,
			METADATA_MESSAGE_BEGIN,
			METADATA_MESSAGE_END,
			TS1, ERROR,
			METADATA_MESSAGE_END,
			INTERCHANGE_END,
			null
		};

		doTest( false, objects );
	}

	/**
	 * Test reaction to a misplaced AT2.
	 */
	public void testMisplacedAttribute () {
		final Object objects[] = {
			INTERCHANGE_BEGIN,
			DATA_MESSAGE_BEGIN,
			DATA_MESSAGE_END,
			AT2, ERROR,
			DATA_MESSAGE_END,
			INTERCHANGE_END,
			null
		};

		doTest( false, objects );
	}
}
