// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.sieve;

import java.lang.String;


import Int.ecb.gesmescb.util.ComponentHelper;
import Int.ecb.logging.Level;

import Int.ecb.gesmescb.Observation;
import Int.ecb.gesmescb.MessageBegin;

import Int.ecb.gesmescb.time.RegularInterval;

/* **************************** Class ***************************** */

/**
 * Filters observations on the basis of their confidentiality.
 *
 * @author
 * <a href="mailto:sis.external@ecb.europa.eu?subject=GESMES/TS-Suite">Sami Airo</a>
 * @author
 * <a href="mailto:sis.external@ecb.europa.eu?subject=GESMES/TS-Suite">François Hoeltzli</a>
 *
 * @version @(#)$Id: ConfidentialityFilter.java,v 1.1 2017/01/19 16:14:04 trainot Exp $
 */

public final class ConfidentialityFilter implements ObservationFilter {

	/* ======================= Constants ======================== */

	/**
	 * By default, only observations where obs_conf=CSI_CODE are discarded.
	 */
	public final String CSI_CODE = "R";

	/* ========================= Fields ========================= */

	/**
	 * Takes care of logging and prefs
	 */
	private ComponentHelper componentHelper;

	/**
	 * Logging level for rejections
	 */
	private Level rejectionLevel;

	/**
	 * If set, only the obs_conf values in the String will be accepted. 
	 */
	private String csiAcceptedCodes;

	/**
	 * If set, only those obs_conf values in the String will be rejected. 
	 */
	private String csiRejectedCodes;

	/* ====================== Constructors ====================== */

	public ConfidentialityFilter () {
		this( "" );
	}

	/**
	 * Constructs a ConfidentialityFilter.
	 */
	public ConfidentialityFilter (final String name) {
		componentHelper = new ComponentHelper( name, this );
	
		rejectionLevel = Level.parse(
			componentHelper.getPreference("rejectionLevel", "FINE")
		);
	
		csiAcceptedCodes = componentHelper.getPreference(
			"csiAcceptedCodes", null
		);
		if ( "".equals( csiAcceptedCodes ) ) {
			csiAcceptedCodes = null;
		}
		if ( null != csiAcceptedCodes ) {
			componentHelper.logger.log(
				Level.CONFIG,
				componentHelper.logName,
				"ConfidentialityFilter",
				"Only observations with the following obs_conf codes "
				+ "will be accepted: \"" + csiAcceptedCodes + "\""
			);
		}
		
		csiRejectedCodes = componentHelper.getPreference(
				"csiRejectedCodes", null
		);
		if ( "".equals( csiRejectedCodes ) ) {
			csiRejectedCodes = null;
		}
		if ( null != csiRejectedCodes ) {
			componentHelper.logger.log(
				Level.CONFIG,
				componentHelper.logName,
				"ConfidentialityFilter",
				"All observations but those with the following obs_conf codes "
				+ "will be rejected: \"" + csiRejectedCodes + "\""
			);
		}
		
		if ( null != csiAcceptedCodes && null != csiRejectedCodes ) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"ConfidentialityFilter",
				"Both csiAcceptedCodes and csiRejectedCodes are non empty!"
			);
		}
		
		if ( null == csiAcceptedCodes && null == csiRejectedCodes ) {
			componentHelper.logger.log(
				Level.CONFIG,
				componentHelper.logName,
				"ConfidentialityFilter",
				"Filter default Mode: only "+ CSI_CODE +" are rejected."
			);
		}
	}

	/* ===================== Public Methods ===================== */

	public boolean accept (
		final MessageBegin messageBegin,
		final String seriesKey,
		final Observation observation,
		final RegularInterval interval
	) {
		componentHelper.logger.log(
			Level.FINE, componentHelper.logName, "accept",
			"checking observation for CSI", interval
		);
		final String conf = observation.getConfidentiality();
		final String confidentiality = "".equals( conf ) ? null : conf;

		if ( null == confidentiality ) {
			if ( csiAcceptedCodes != null ) {
				componentHelper.logger.log(
					rejectionLevel,
					componentHelper.logName,
					"accept",
					"Observation rejected: no confidentiality code "
					+ "found for interval " + interval
				);
				return false;
			} else {
				return true;
			}
		}
		
		if ( null != csiAcceptedCodes ) {
			if ( csiAcceptedCodes.indexOf( confidentiality ) >= 0 ) {
				return true;
			} else {
				componentHelper.logger.log(
					rejectionLevel,
					componentHelper.logName,
					"accept",
					"Observation rejected: observation confidentiality "
					+ "not in accepted CSI codes for interval " + interval
				);
			return false;
			}
		}

		if ( null != csiRejectedCodes ) {
			if ( csiRejectedCodes.indexOf( confidentiality ) >= 0 ) {
				componentHelper.logger.log(
					rejectionLevel,
					componentHelper.logName,
					"accept",
					"Observation rejected: observation confidentiality "
					+ "indicates CSI for interval " + interval
				);
				return false;
			} else {
				return true;
			}
		}

		if ( confidentiality.equals( CSI_CODE ) ) {
			componentHelper.logger.log(
				rejectionLevel,
				componentHelper.logName,
				"accept",
				"Observation rejected: observation confidentiality "
				+ "indicates CSI for interval " + interval
			);
			return false;
		}
		return true;
	}
}
