// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.sieve;

import java.util.Collections;
import java.text.SimpleDateFormat;
import java.text.ParseException;

import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;

import Int.ecb.gesmescb.*;
import Int.ecb.gesmescb.time.*;

/* ***************************** Class ****************************** */

/**
 * Tests the implementation of the class {@link ConfidentialityFilter}.
 *
 * @author
 * <a href="mailto:sis.external@ecb.europa.eu?subject=GESMES/TS-Suite">François Hoeltzli</a>
 *
 * @version @(#)$Id: ConfidentialityFilterTest.java,v 1.1 2017/01/19 16:14:04 trainot Exp $
 */

public class ConfidentialityFilterTest extends TestCase {

	/* ======================= Constants ======================== */

	private final String CLASS_NAME = "Int.ecb.gesmescb.sieve.ConfidentialityFilter";

	/* ========================= Fields ========================= */

	private MessageBegin messageBegin;
	private SimpleDateFormat format = new SimpleDateFormat("yyyyMMdd'T'HHmmss");

	/* ====================== Constructors ====================== */

	/**
	 * Constructs a ConfidentialityFilterTest with the specified name.
	 *
	 * @param name Test case name
	 */
	public ConfidentialityFilterTest ( final String name ) {
		super( name );
	}

	/* ================== Protected Methods ===================== */

	/**
	 * Sets up the test fixture.
	 */
	protected void setUp () {
		messageBegin = new MessageBegin(
			"MREF0001", "SE2", "4F0", "ECB",
			MessageBegin.STATISTICAL_DATA,
			"Test message", Collections.EMPTY_SET,
			"ECB_EXR1",
			MessageBegin.UPDATE_OR_REPLACE_ACTION,
			"20010329:1423", "", "",
			"ECB_EXR1"
		);

		System.setProperty( ".level", "ALL" );
		System.setProperty( "handlers", "Int.ecb.logging.ConsoleHandler" );
		System.setProperty(
			"Int.ecb.logging.ConsoleHandler.formatter",
			"Int.ecb.gesmescb.util.NormalLogFormatter"
		);
	}

	/**
	 * Tears down the test fixture.
	 */
	protected void tearDown () {
		messageBegin = null;
	}

	/* ===================== Public Methods ===================== */

	/**
	 * Assembles and returns a test suite for all
	 * the test methods of this test case.
	 *
	 * @return a non-null test suite.
	 */
	public static Test suite () {
		return new TestSuite( ConfidentialityFilterTest.class );
	}

	/**
	 * Runs the test case.
	 */
	public static void main ( final String args[] ) {
		final String[] testCaseName = {
			ConfidentialityFilterTest.class.getName()
		};
		junit.textui.TestRunner.main( testCaseName );
	}

	/* --------------------- Test Methods --------------------- */

	public void testFilterWithAcceptingList () 
		throws ParseException
	{
		// everything that does not have obs_conf in one of the
		// csiAcceptedCode is rejected.
		System.setProperty(CLASS_NAME + ".csiAcceptedCodes", "FR");
	
		ConfidentialityFilter filter = new ConfidentialityFilter();
				
		assertTrue("Empty not confidential should be refused.", 
			! filter.accept(
				messageBegin, 
				"SERIES.KEY", 
				Observation.EMPTY_OBSERVATION, 
				new GregorianRegularInterval(
					RegularInterval.DAILY,
					format.parse("20030319T000000"),
					format.parse("20030320T000000")
				)
			)
		);
		assertTrue("ND Restricted should be accepted", 
			filter.accept(
				messageBegin, 
				"SERIES.KEY", 
				new Observation ("ND", "A", "R", ""),
				new GregorianRegularInterval(
					RegularInterval.DAILY,
					format.parse("20030319T000000"),
					format.parse("20030320T000000")
				)
			)
		);
		assertTrue("value, F should be accepted", 
			filter.accept(
				messageBegin, 
				"SERIES.KEY", 
				new Observation ("10", "", "F", ""),
				new GregorianRegularInterval(
					RegularInterval.DAILY,
					format.parse("20030319T000000"),
					format.parse("20030320T000000")
				)
			)
		);
		assertTrue("value, C should be rejected", 
			! filter.accept(
				messageBegin, 
				"SERIES.KEY", 
				new Observation ("10", "", "C", ""),
				new GregorianRegularInterval(
					RegularInterval.DAILY,
					format.parse("20030319T000000"),
					format.parse("20030320T000000")
				)
			)
		);
		System.getProperties().remove( CLASS_NAME + ".csiAcceptedCodes" );
	}
	
	public void testFilterWithRejectingList () 
		throws ParseException
	{
		// everything but what has have obs_conf in one of the
		// csiRejectedCode is accepted.
		System.setProperty( CLASS_NAME + ".csiRejectedCodes", "FR");
	
		ConfidentialityFilter filter = new ConfidentialityFilter();
				
		assertTrue("ND F should be rejected", 
			! filter.accept(
				messageBegin, 
				"SERIES.KEY", 
				new Observation ("ND", "A", "F", ""),
				new GregorianRegularInterval(
					RegularInterval.DAILY,
					format.parse("20030319T000000"),
					format.parse("20030320T000000")
				)
			)
		);
		assertTrue("Empty not F not R should be accepted", 
			filter.accept(
				messageBegin, 
				"SERIES.KEY", 
				Observation.EMPTY_OBSERVATION, 
				new GregorianRegularInterval(
					RegularInterval.DAILY,
					format.parse("20030319T000000"),
					format.parse("20030320T000000")
				)
			)
		);
		assertTrue("value, C should be accepted", 
			filter.accept(
				messageBegin, 
				"SERIES.KEY", 
				new Observation ("10", "", "C", ""),
				new GregorianRegularInterval(
					RegularInterval.DAILY,
					format.parse("20030319T000000"),
					format.parse("20030320T000000")
				)
			)
		);
		assertTrue("value R should be rejected", 
			! filter.accept(
				messageBegin, 
				"SERIES.KEY", 
				new Observation ("10", "", "R", ""),
				new GregorianRegularInterval(
					RegularInterval.DAILY,
					format.parse("20030319T000000"),
					format.parse("20030320T000000")
				)
			)
		);
		assertTrue("value F should be rejected", 
			! filter.accept(
				messageBegin, 
				"SERIES.KEY", 
				new Observation ("10", "", "F", ""),
				new GregorianRegularInterval(
					RegularInterval.DAILY,
					format.parse("20030319T000000"),
					format.parse("20030320T000000")
				)
			)
		);
		System.getProperties().remove( CLASS_NAME + ".csiRejectedCodes" );
	}
	
	public void testFilterInDefaultMode () 
		throws ParseException
	{
		ConfidentialityFilter filter = new ConfidentialityFilter();
				
		assertTrue("Empty not R should be accepted", 
			filter.accept(
				messageBegin, 
				"SERIES.KEY", 
				Observation.EMPTY_OBSERVATION, 
				new GregorianRegularInterval(
					RegularInterval.DAILY,
					format.parse("20030319T000000"),
					format.parse("20030320T000000")
				)
			)
		);
		assertTrue("ND Restricted should be rejected", 
			! filter.accept(
				messageBegin, 
				"SERIES.KEY", 
				new Observation ("ND", "A", "R", ""),
				new GregorianRegularInterval(
					RegularInterval.DAILY,
					format.parse("20030319T000000"),
					format.parse("20030320T000000")
				)
			)
		);
		assertTrue("value F should be accepted", 
			filter.accept(
				messageBegin, 
				"SERIES.KEY", 
				new Observation ("10", "", "F", ""),
				new GregorianRegularInterval(
					RegularInterval.DAILY,
					format.parse("20030319T000000"),
					format.parse("20030320T000000")
				)
			)
		);
		assertTrue("value, C should be accepted", 
			filter.accept(
				messageBegin, 
				"SERIES.KEY", 
				new Observation ("10", "", "C", ""),
				new GregorianRegularInterval(
					RegularInterval.DAILY,
					format.parse("20030319T000000"),
					format.parse("20030320T000000")
				)
			)
		);
	}
}
