// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.sieve;

import java.text.DecimalFormat;

import java.util.Iterator;
import java.util.List;
import java.util.ArrayList;

import Int.ecb.logging.Level;

import Int.ecb.gesmescb.*;
import Int.ecb.gesmescb.event.GESMESCBEventBroker;

/* **************************** Class ***************************** */

/**
 * Blocks empty messages and interchanges.
 *
 * @author <A HREF="mailto:Sami.Airo@ECB.int">Sami Airo</A>
 *
 * @version @(#)$Id: EmptyMessageBlocker.java,v 1.1 2017/01/19 16:14:04 trainot Exp $
 */

public class EmptyMessageBlocker extends GESMESCBEventBroker {

	/* ======================= Constants ======================== */

	/* ========================= Fields ========================= */

	private int mrefNumber = 0;	
			
	private InterchangeBegin interchangeBegin;

	private MessageBegin messageBegin;
	
	private boolean interchangeBeginDispatched;
	
	private boolean messageBeginDispatched;

	/* ====================== Constructors ====================== */

	public EmptyMessageBlocker () {
		this( "" );
	}

	public EmptyMessageBlocker( final String name ) {
		super( name, true );
	}

	/* ===================== Public Methods ===================== */

	public void handle ( final InterchangeBegin interchangeBegin ) {
		componentHelper.logger.log(
			Level.FINE, componentHelper.logName, "handle",
			"handling InterchangeBegin", interchangeBegin
		);

		this.interchangeBegin = interchangeBegin;
		interchangeBeginDispatched = false;
	}

	public void handle ( final InterchangeEnd interchangeEnd ) {
		componentHelper.logger.log(
			Level.FINE, componentHelper.logName, "handle",
			"handling InterchangeEnd", interchangeEnd
		);

		if (!interchangeBeginDispatched) return;
		dispatch(interchangeEnd);
	}

	public void handle ( final MessageBegin messageBegin ) {
		componentHelper.logger.log(
			Level.FINE, componentHelper.logName, "handle",
			"handling MessageBegin", messageBegin
		);

		this.messageBegin = messageBegin;
		messageBeginDispatched = false;
	}

	public void handle ( final MessageEnd messageEnd ) {
		componentHelper.logger.log(
			Level.FINE, componentHelper.logName, "handle",
			"handling MessageEnd", messageEnd
		);

		if (!messageBeginDispatched) return;
		dispatch(new MessageEnd( getMREF(mrefNumber) ));		
	}

	public void handle ( final KeyFamily keyFamily ) {
		componentHelper.logger.log(
			Level.FINE, componentHelper.logName, "handle",
			"handling KeyFamily", keyFamily
		);

		dispatchMessageBegin();
		dispatch(keyFamily);
	}	

	public void handle ( final StatisticalConcept statisticalConcept ) {
		componentHelper.logger.log(
			Level.FINE, componentHelper.logName, "handle",
			"handling StatisticalConcept", statisticalConcept
		);

		dispatchMessageBegin();
		dispatch(statisticalConcept);
	}	

	public void handle ( final CodeList codeList ) {
		componentHelper.logger.log(
			Level.FINE, componentHelper.logName, "handle",
			"handling CodeList", codeList
		);

		dispatchMessageBegin();
		dispatch(codeList);
	}	
	
	public void handle ( final TimeSeries timeSeries ) {
		componentHelper.logger.log(
			Level.FINE, componentHelper.logName, "handle",
			"handling TimeSeries", timeSeries
		);

		dispatchMessageBegin();
		dispatch(timeSeries);
	}


	public void handle ( final Attribute attribute ) {
		componentHelper.logger.log(
			Level.FINE, componentHelper.logName, "handle",
			"handling Attribute", attribute
		);

		dispatchMessageBegin();
		dispatch(attribute);
	}


	public void dispatchMessageBegin() {
		if (messageBeginDispatched) return;
	
		dispatchInterchangeBegin();
	
		final List contacts = new ArrayList();
		for (
			final Iterator iterator = messageBegin.contactIterator();
			iterator.hasNext();
		) {
			contacts.add( iterator.next() );
		}

		mrefNumber++;	
		
		dispatch(
			new MessageBegin(
				getMREF(mrefNumber),
				messageBegin.getSender(),
				messageBegin.getReceiver(),
				messageBegin.getMaintenanceAgencyId(),
				messageBegin.getFunction(),
				messageBegin.getIdentity(),
				contacts,
				messageBegin.getDataSetId(),
				messageBegin.getAction(),
				messageBegin.getExtractionDateTime(),
				messageBegin.getReportingPeriod(),
				messageBegin.getReportingPeriodFormat(),
				messageBegin.getKeyFamilyId(),
				messageBegin.getOriginalDataSetId()
			)
		);
	
		messageBeginDispatched = true;
	}

	public void dispatchInterchangeBegin() {
		if (interchangeBeginDispatched) return;
		dispatch(interchangeBegin);
		interchangeBeginDispatched = true;
	}
	
	public String getMREF(final int number) {
		DecimalFormat mrefFormat = new DecimalFormat( "000000" );
		return "MREF" + mrefFormat.format( number );		
	}	
	
}
