// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.sieve;

import java.text.DecimalFormat;

import java.util.Iterator;
import java.util.List;
import java.util.ArrayList;

import Int.ecb.logging.Level;

import Int.ecb.gesmescb.*;
import Int.ecb.gesmescb.event.GESMESCBEventBroker;
import Int.ecb.gesmescb.event.TestEventListener;
import Int.ecb.gesmescb.time.GregorianRegularInterval;
import Int.ecb.gesmescb.time.RegularInterval;

import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;

/* **************************** Class ***************************** */

/**
 * Blocks empty messages and interchanges.
 *
 * @author <A HREF="mailto:francois.hoeltzli@ECB.int">Francois Hoeltzli</A>
 *
 * @version @(#)$Id: EmptyMessageBlockerTest.java,v 1.1 2017/01/19 16:14:04 trainot Exp $
 */

public class EmptyMessageBlockerTest extends TestCase {

	/* ======================= Constants ======================== */

	/* ========================= Fields ========================= */

	private final InterchangeBegin ib = new InterchangeBegin(
		"IREF000001", "DE2", "4F0", "000315:1245", false
	);
	private final InterchangeEnd ie = new InterchangeEnd(
		"IREF000001"
	);
	
	private MessageBegin mb1, mb2;

	private static final Object[] deleteMessage1 = {
		"MREF000001",
		"DE2",
		"4F0",
		"ECB",
		MessageBegin.STATISTICAL_DATA,
		"The subject",
		new Contact[] {
			new Contact(
				Contact.DATA_PROCESSING, "ECB-SIS", "Quasimodo",
				new ContactChannel[] {
					new ContactChannel(
						ContactChannel.EMAIL,
						"s-xtermal@ecb.int"
					)
				}
			)
		},
		"DATASET1",
		MessageBegin.DELETE_ACTION,
		"200103201423",
		"1999",
		"602",
		"ECB_BOP1"
	};
	
	private static final Object[] deleteMessage2 = {
		"MREF000001",
		"DE2",
		"4F0",
		"ECB",
		MessageBegin.STATISTICAL_DATA,
		"The subject",
		new Contact[] {
			new Contact(
				Contact.DATA_PROCESSING, "ECB-SIS", "Quasimodo",
				new ContactChannel[] {
					new ContactChannel(
						ContactChannel.EMAIL,
						"s-xtermal@ecb.int"
					)
				}
			)
		},
		"DATASET1",
		MessageBegin.DELETE_ACTION,
		"200103201423",
		"1999",
		"602",
		"ECB_BOP1"
	};

	private MessageEnd me1, me2;

	private final String
		key = "key",
		period = "20010202",
		periodFormat = "102";
	private final Observation[] observations = {
		new Observation( "", "", "", "" ),
		new Observation( "1", "", "", "" ),
		new Observation( "1", "2", "", "" ),
		new Observation( "1", "2", "3", "" ),
		new Observation( "1", "2", "3", "4" )
	};
	private TimeSeries ts1;
	
	/* ====================== Constructors ====================== */

	public EmptyMessageBlockerTest ( String name ) {
		super( name );
	}

	/**
	 * Assembles and returns a test suite for all
	 * the test methods of this test case.
	 *
	 * @return a non-null test suite
	 */
	public static Test suite () {
		TestSuite suite = new TestSuite( EmptyMessageBlockerTest.class );
		return suite;
	}

	/**
	 * Runs the test case
	 */
	public static void main ( String args[] ) {
		String[] testCaseName = { EmptyMessageBlockerTest.class.getName() };
		junit.textui.TestRunner.main( testCaseName );
	}
	/* ================== Protected Methods ===================== */

	/**
	 * Sets up the test fixture
	 */
	protected void setUp () {
		// Create full MessageBegin.
		mb1 = new MessageBegin(
			(String) deleteMessage1[0],
			(String) deleteMessage1[1],
			(String) deleteMessage1[2],
			(String) deleteMessage1[3],
			(String) deleteMessage1[4],
			(String) deleteMessage1[5],
			(Contact[]) deleteMessage1[6],
			(String) deleteMessage1[7],
			(String) deleteMessage1[8],
			(String) deleteMessage1[9],
			(String) deleteMessage1[10],
			(String) deleteMessage1[11],
			(String) deleteMessage1[12]
		);
		mb2 = new MessageBegin(
			(String) deleteMessage2[0],
			(String) deleteMessage2[1],
			(String) deleteMessage2[2],
			(String) deleteMessage2[3],
			(String) deleteMessage2[4],
			(String) deleteMessage2[5],
			(Contact[]) deleteMessage2[6],
			(String) deleteMessage2[7],
			(String) deleteMessage2[8],
			(String) deleteMessage2[9],
			(String) deleteMessage2[10],
			(String) deleteMessage2[11],
			(String) deleteMessage2[12]
		);
		me1 = new MessageEnd( "MREF000001" );	
		me2 = new MessageEnd( "MREF000002" );	
		
		ts1 = new TimeSeries( key, period, periodFormat, observations );
	}
		
	/* ===================== Public Methods ===================== */
	
	/* ---------------------- Test Methods ---------------------- */
	
	public void testNonEmptyMessage () {
		EmptyMessageBlocker blocker = new EmptyMessageBlocker();
		TestEventListener listener = new TestEventListener();
		blocker.addListener(listener, true);
		
		blocker.handle(ib);
		blocker.handle(mb1);
		blocker.handle(ts1);
		blocker.handle(me1);
		
		Iterator it = listener.iterator();
		assertEquals("Where's my InterchangeBegin?", ib, it.next() );
		assertEquals("Where's my MessageBegin?", mb1, (MessageBegin) it.next() );
		assertEquals("Where's my ts1?", ts1, (TimeSeries) it.next() );
		assertEquals("Where's my MessageEnd?", me1, (MessageEnd) it.next() );
		assertTrue("Should be nothing more!", !it.hasNext() );
	}
	
	public void testEmptyMessages () {
		EmptyMessageBlocker blocker = new EmptyMessageBlocker();
		TestEventListener listener = new TestEventListener();
		blocker.addListener(listener, true);
		
		blocker.handle(ib);
		blocker.handle(mb1);
		blocker.handle(me1);
		blocker.handle(mb2);
		blocker.handle(me2);
		blocker.handle(ie);
		
		Iterator it = listener.iterator();
		assertTrue("Should be nothing more!", !it.hasNext() );
	}
	
	public void testEmptyInterchange () {
		EmptyMessageBlocker blocker = new EmptyMessageBlocker();
		TestEventListener listener = new TestEventListener();
		blocker.addListener(listener, true);
		
		blocker.handle(ib);
		blocker.handle(ie);
		
		Iterator it = listener.iterator();
		assertTrue("Should be nothing more!", !it.hasNext() );
	}
}
