// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (C) 2004, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.sieve;

import Int.ecb.logging.Level;

import Int.ecb.gesmescb.GESMESCBObject;
import Int.ecb.gesmescb.MessageBegin;
import Int.ecb.gesmescb.Attribute;

/* **************************** Class ***************************** */

/**
 * <p>Judges empty {@link GESMESCBObject}s according to preference settings.</p>
 *
 * <center><table border="3" cellpadding="3">
 * <caption>
 * <a name="prefs">Available {@link Int.ecb.prefs preference} settings</a>
 * </caption>
 * <tr>
 * <th>preference name</th>
 * <th>type</th>
 * <th>meaning</th>
 * <th>default</th>
 * </tr>
 * <tr>
 * <td><tt>acceptEmptyAttributeUpdates</tt></td>
 * <td>boolean</td>
 * <td>whether empty attributes in update messages are to be expected</td>
 * <td><tt>true</tt></td>
 * </tr>
 * </table></center>
 *
 * @see GESMESCBEventSieve
 *
 * @author
 * <a href="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</A>
 *
 * @version @(#)$Id: EmptyObjectFilter.java,v 1.1 2017/01/19 16:14:04 trainot Exp $
 */
public class EmptyObjectFilter extends GESMESCBObjectFilterAdapter {

	/* ========================= Fields ========================= */

	/** Whether to treat empty attributes in an update message as expected. */
	private final boolean acceptEmptyAttributeUpdates;

	/** Whether the current message is an update message. */
	private boolean inUpdateMessage;

	/* ====================== Constructors ====================== */

	/** Constructs a EmptyObjectFilter with the default name. */
	public EmptyObjectFilter () {
		this( "" );
	}

	/** Constructs a EmptyObjectFilter with the specified name. */
	public EmptyObjectFilter ( final String name ) {
		super( name );
		acceptEmptyAttributeUpdates
		= "true".equalsIgnoreCase( componentHelper.getPreference(
			"acceptEmptyAttributeUpdates", "true"
		) );
	}

	/* ===================== Public Methods ===================== */

	public boolean accept ( final MessageBegin messageBegin ) {
		inUpdateMessage = messageBegin.isStatisticalData()
			&& messageBegin.isUpdateOrReplaceAction();
		return super.accept( messageBegin );
	}

	public boolean accept ( final Attribute attribute ) {
		if (
			!acceptEmptyAttributeUpdates
			&&
			inUpdateMessage
			&&
			"".equals( attribute.getValue() )
		) {
			componentHelper.logger.log(
				Level.WARNING,
				componentHelper.logName,
				"accept",
				"Empty attribute rejected: "
				+ attribute.getKey() + (
					"".equals( attribute.getPeriod() )
					? ":"
					: "["
						+ attribute.getPeriod() + ":"
						+ attribute.getPeriodFormat() + "]:"
				) + attribute.getName(),
				attribute
			);
			return false;
		}

		return super.accept( attribute );
	}
}
