// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, 2003 European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.sieve;

import Int.ecb.logging.Level;

import Int.ecb.gesmescb.*;

import Int.ecb.gesmescb.event.GESMESCBEventListenerAdapter;
import Int.ecb.gesmescb.event.StructureListener;
import Int.ecb.gesmescb.event.DataListener;
import Int.ecb.gesmescb.event.GESMESCBEventSourceHelper;
import Int.ecb.gesmescb.event.GESMESCBEventSource;
import Int.ecb.gesmescb.event.ErrorEvent;

import Int.ecb.gesmescb.sieve.GESMESCBObjectFilter;

import Int.ecb.gesmescb.util.ComponentHelper;

/* **************************** Class ***************************** */

/**
 * <P>Sieves events into accepted and rejected. The accepted events
 * are passed to the acceptedEventSource and the rejected events to
 * the rejectedEventSource.</P>
 *
 * <center><table border="3" cellpadding="3">
 * <caption>
 * <a name="prefs">Available {@link Int.ecb.prefs preference} settings</a>
 * </caption>
 * <tr>
 * <th>preference name</th>
 * <th>type</th>
 * <th>meaning</th>
 * <th>default</th>
 * </tr>
 * <tr>
 * <td><tt>filter</tt></td>
 * <td>string</td>
 * <td>Name of the <code>GESMESCBObjectFilter</code> class to use</td>
 * <td><tt>Int.ecb.gesmescb.sieve.PreferenceConfiguredFilter</tt></td>
 * </tr>
 * </table></center>
 *
 * @author
 * <A HREF="mailto:Sami.Airo@ECB.int">Sami Airo</A>
 * @author
 * <A HREF="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</A>
 *
 * @version @(#)$Id: GESMESCBEventSieve.java,v 1.1 2017/01/19 16:14:04 trainot Exp $
 */

public class GESMESCBEventSieve extends GESMESCBEventListenerAdapter
	implements StructureListener, DataListener
{

	/* ========================= Fields ========================= */

	/**
	 * Takes care of logging and preferences.
	 */
	protected final ComponentHelper componentHelper;

	/**
	 * Source for the accepted events.
	 */
	private final GESMESCBEventSourceHelper acceptedSource;

	/**
	 * Source for the rejected events.
	 */
	private final GESMESCBEventSourceHelper rejectedSource;

	private final GESMESCBObjectFilter filter;

	private InterchangeBegin currentInterchangeBegin;

	private MessageBegin currentMessageBegin;

	private boolean interchangeBeginDispatchedToAccepted = false;

	private boolean interchangeBeginDispatchedToRejected = false;

	private boolean messageBeginDispatchedToAccepted = false;

	private boolean messageBeginDispatchedToRejected = false;

	private boolean messageBeginAccepted = false;

	private boolean interchangeBeginAccepted = false;

	/** For automatic disconnect. */
	private boolean acceptedActive, rejectedActive;

	/* ====================== Constructors ====================== */

	/**
	 * Constructs a GESMESCBEventSieve.
	 */
	public GESMESCBEventSieve () {
		this( "", null );
	}

	/**
	 * Constructs a GESMESCBEventSieve.
	 */
	public GESMESCBEventSieve ( final String name ) {
		this( name, null );
	}

	/**
	 * Constructs a GESMESCBEventSieve.
	 */
	public GESMESCBEventSieve ( final GESMESCBObjectFilter filter ) {
		this( "", filter );
	}

	/**
	 * Constructs a GESMESCBEventSieve.
	 */
	public GESMESCBEventSieve (
		final String name, final GESMESCBObjectFilter filter
	) {
		componentHelper = new ComponentHelper( name, this );
		acceptedSource = new GESMESCBEventSourceHelper();
		rejectedSource = new GESMESCBEventSourceHelper();

		GESMESCBObjectFilter filterTmp = filter;

		if (filterTmp == null) {
			final String filterClassName = componentHelper.getPreference(
				"filter", "Int.ecb.gesmescb.sieve.PreferenceConfiguredFilter"
			);

			Class classFor = null;

			try {
				classFor = Class.forName(filterClassName);
			} catch ( final Exception e ) {
				componentHelper.logger.log(
					Level.SEVERE,
					componentHelper.logName,
					"GESMESCBEventSieve",
					"cannot load class " + filterClassName,
					e
				);
			}

			if ( classFor != null ) {
				try {
					filterTmp = (GESMESCBObjectFilter) classFor.newInstance();

					componentHelper.logger.log(
						Level.CONFIG,
						componentHelper.logName,
						"GESMESCBEventSieve",
						"GESMESCBEventSieve configured using filter class "
						+ filterClassName,
						filterTmp
					);
				} catch ( final Exception e ) {
					componentHelper.logger.log(
						Level.SEVERE,
						componentHelper.logName,
						"GESMESCBEventSieve",
						"cannot create instance of class " + filterClassName,
						e
					);
				}
			}
		}

		this.filter = filterTmp;

		componentHelper.logger.log(
			Level.FINEST,
			componentHelper.logName,
			"GESMESCBEventSieve",
			"New GESMESCBEventSieve constructed",
			this
		);
		acceptedActive = rejectedActive = true;
	}

	/* ===================== Public Methods ===================== */

	/**
	 * The event source for accepted events.
	 *
	 * @return event source for accepted events
	 */
	public final GESMESCBEventSource getAcceptedSource () {
		return acceptedSource;
	}

	/**
	 * The event source for rejected events.
	 *
	 * @return event source for rejected events
	 */
	public final GESMESCBEventSource getRejectedSource () {
		return rejectedSource;
	}

	/* ------------------------ Listener ------------------------ */

	/**
	 * Receive and cache an <CODE>InterchangeBegin</CODE>.
	 *
	 * @param interchangeBegin to be cached.
	 */
	public void handle (
		final InterchangeBegin interchangeBegin
	) {
		interchangeBeginDispatchedToAccepted = false;
		interchangeBeginDispatchedToRejected = false;

		componentHelper.logger.log(
			Level.FINE,
			componentHelper.logName,
			"handle",
			"InterchangeBegin cached.",
			interchangeBegin
		);
		currentInterchangeBegin = interchangeBegin;

		interchangeBeginAccepted = filter.accept( interchangeBegin );
	}

	/*
	 * Receive and process an <CODE>InterchangeEnd</CODE>.
	 * If InterchangeBegin was dispatched then dispatch.
	 *
	 * @param interchangeEnd to be handled.
	 */
	public void handle (
		final InterchangeEnd interchangeEnd
	) {
		if ( interchangeBeginDispatchedToAccepted ) {
			componentHelper.logger.log	(
				Level.FINE,
				componentHelper.logName,
				"handle",
				"InterchangeEnd dispatched to accepted",
				interchangeEnd
			);
			dispatchToAccepted( interchangeEnd );
		}

		if ( interchangeBeginDispatchedToRejected ) {
			componentHelper.logger.log	(
				Level.FINE,
				componentHelper.logName,
				"handle",
				"InterchangeEnd dispatched to rejected",
				interchangeEnd
			);
			dispatchToRejected( interchangeEnd );
		}
	}

	/**
	 * Inspect and cache a <CODE>MessageBegin</CODE>.
	 *
	 * @param messageBegin to be handled.
	 */
	public void handle (
		final MessageBegin messageBegin
	) {
		messageBeginDispatchedToAccepted = false;
		messageBeginDispatchedToRejected = false;

		currentMessageBegin = messageBegin;
		componentHelper.logger.log(
			Level.FINE,
			componentHelper.logName,
			"handle",
			"MessageBegin cached.",
			messageBegin
		);

		messageBeginAccepted = interchangeBeginAccepted && filter.accept( messageBegin );
		/** I. Trainotti 21.09.2011 
		 * It is needed to dispatch the begin events if they were not yet dispatched.
		 * (e.g. for empty GESMES)
		 * */
		if (messageBeginAccepted) {
			dispatchCurrentBeginEventsToAccepted();
		}
	}

	/*
	 * Receive and process a <CODE>MessageEnd</CODE>.
	 * If MessageBegin was dispatched then dispatch.
	 *
	 * @param messageEnd to be handled.
	 */
	public void handle (
		final MessageEnd messageEnd
	) {
		if ( messageBeginDispatchedToAccepted ) {
			componentHelper.logger.log	(
				Level.FINE,
				componentHelper.logName,
				"handle",
				"MessageEnd dispatched to accepted",
				messageEnd
			);
			dispatchToAccepted( messageEnd );
		}

		if ( messageBeginDispatchedToRejected ) {
			componentHelper.logger.log	(
				Level.FINE,
				componentHelper.logName,
				"handle",
				"MessageEnd dispatched to rejected",
				messageEnd
			);
			dispatchToRejected( messageEnd );
		}
	}

	/**
	 * Inspect and process a <CODE>CodeList</CODE>.
	 *
	 * @param codeList to be handled.
	 */
	public void handle (
		final CodeList codeList
	) {
		if ( messageBeginAccepted && filter.accept( codeList ) ) {
			componentHelper.logger.log(
				Level.FINEST,
				componentHelper.logName,
				"handle",
				"CodeList dispatched to accepted.",
				codeList
			);
			dispatchCurrentBeginEventsToAccepted();
			dispatchToAccepted( codeList );
		} else {
			componentHelper.logger.log(
				Level.FINEST,
				componentHelper.logName,
				"handle",
				"CodeList dispatched to rejected.",
				codeList
			);
			dispatchCurrentBeginEventsToRejected();
			dispatchToRejected( codeList );
		}
	}

	/**
	 * Inspect and process a <CODE>StatisticalConcept</CODE>.
	 *
	 * @param statisticalConcept to be handled.
	 */
	public void handle (
		final StatisticalConcept statisticalConcept
	) {
		if ( messageBeginAccepted && filter.accept( statisticalConcept ) ) {
			componentHelper.logger.log(
				Level.FINEST,
				componentHelper.logName,
				"handle",
				"StatisticalConcept dispatched to accepted.",
				statisticalConcept
			);
			dispatchCurrentBeginEventsToAccepted();
			dispatchToAccepted( statisticalConcept );
		} else {
			componentHelper.logger.log(
				Level.FINEST,
				componentHelper.logName,
				"handle",
				"StatisticalConcept dispatched to rejected.",
				statisticalConcept
			);
			dispatchCurrentBeginEventsToRejected();
			dispatchToRejected( statisticalConcept );
		}
	}

	/**
	 * Inspect and process a <CODE>KeyFamily</CODE>.
	 *
	 * @param keyFamily to be handled.
	 */
	public void handle (
		final KeyFamily keyFamily
	) {
		if ( messageBeginAccepted && filter.accept( keyFamily ) ) {
			componentHelper.logger.log(
				Level.FINEST,
				componentHelper.logName,
				"handle",
				"KeyFamily dispatched to accepted.",
				keyFamily
			);
			dispatchCurrentBeginEventsToAccepted();
			dispatchToAccepted( keyFamily );
		} else {
			componentHelper.logger.log(
				Level.FINEST,
				componentHelper.logName,
				"handle",
				"KeyFamily dispatched to rejected.",
				keyFamily
			);
			dispatchCurrentBeginEventsToRejected();
			dispatchToRejected( keyFamily );
		}
	}

	/**
	 * Inspect and process a <CODE>TimeSeries</CODE>.
	 *
	 * @param timeSeries to be handled.
	 */
	public void handle (
		final TimeSeries timeSeries
	) {
		if ( messageBeginAccepted && filter.accept( timeSeries ) ) {
			componentHelper.logger.log(
				Level.FINEST,
				componentHelper.logName,
				"handle",
				"TimeSeries dispatched to accepted.",
				timeSeries
			);
			dispatchCurrentBeginEventsToAccepted();
			dispatchToAccepted( timeSeries );
		} else {
			componentHelper.logger.log(
				Level.FINEST,
				componentHelper.logName,
				"handle",
				"TimeSeries dispatched to rejected.",
				timeSeries
			);
			dispatchCurrentBeginEventsToRejected();
			dispatchToRejected( timeSeries );
		}
	}

	/**
	 * Inspect and process an <CODE>Attribute</CODE>.
	 *
	 * @param attribute to be handled.
	 */
	public void handle (
		final Attribute attribute
	) {
		if ( messageBeginAccepted && filter.accept( attribute ) ) {
			componentHelper.logger.log(
				Level.FINEST,
				componentHelper.logName,
				"handle",
				"Attribute dispatched to accepted.",
				attribute
			);
			dispatchCurrentBeginEventsToAccepted();
			dispatchToAccepted( attribute );
		} else {
			componentHelper.logger.log(
				Level.FINEST,
				componentHelper.logName,
				"handle",
				"Attribute dispatched to rejected.",
				attribute
			);
			dispatchCurrentBeginEventsToRejected();
			dispatchToRejected( attribute );
		}
	}

	/**
	 * <P>Receive and process an <CODE>ErrorEvent</CODE>.</P>
	 *
	 * Unless overridden, simply dispatch to both acceptedSource.
	 * and rejectedSource.
	 *
	 * @param errorEvent to be handled.
	 */
	public void handle (
		final ErrorEvent errorEvent
	) {
		componentHelper.logger.log(
			Level.FINEST,
			componentHelper.logName,
			"handle",
			"ErrorEvent dispatched to accepted and rejected.",
			errorEvent
		);
		dispatchToAccepted( errorEvent );
		dispatchToRejected( errorEvent );
	}

	public boolean isActive () {
		setActive(
			super.isActive()
			&& (
				acceptedActive
				||
				rejectedActive
			)
		);
		return super.isActive();
	}

	/* ==================== Private Methods ===================== */

	/**
	 * Dispatch cached InterchangeBegin and MessageBegin
	 * to acceptedSource if necessary.
	 */
	private void dispatchCurrentBeginEventsToAccepted () {
		if ( !interchangeBeginDispatchedToAccepted ) {
			dispatchToAccepted( currentInterchangeBegin );
			interchangeBeginDispatchedToAccepted = true;
		}

		if ( !messageBeginDispatchedToAccepted ) {
			dispatchToAccepted( currentMessageBegin );
			messageBeginDispatchedToAccepted = true;
		}
	}

	/**
	 * Dispatch cached InterchangeBegin and MessageBegin
	 * to rejectedSource if necessary.
	 */
	private void dispatchCurrentBeginEventsToRejected () {
		if ( !interchangeBeginDispatchedToRejected ) {
			dispatchToRejected( currentInterchangeBegin );
			interchangeBeginDispatchedToRejected = true;
		}

		if ( !messageBeginDispatchedToRejected ) {
			dispatchToRejected( currentMessageBegin );
			messageBeginDispatchedToRejected = true;
		}
	}

	private void dispatchToAccepted (
		final InterchangeBegin interchangeBegin
	) {
		acceptedActive &= acceptedSource.dispatch( interchangeBegin );
	}

	private void dispatchToAccepted (
		final InterchangeEnd interchangeEnd
	) {
		acceptedActive &= acceptedSource.dispatch( interchangeEnd );
	}

	private void dispatchToAccepted (
		final MessageBegin messageBegin
	) {
		acceptedActive &= acceptedSource.dispatch( messageBegin );
	}

	private void dispatchToAccepted (
		final MessageEnd messageEnd
	) {
		acceptedActive &= acceptedSource.dispatch( messageEnd );
	}

	private void dispatchToAccepted (
		final CodeList codeList
	) {
		acceptedActive &= acceptedSource.dispatch( codeList );
	}

	private void dispatchToAccepted (
		final StatisticalConcept statisticalConcept
	) {
		acceptedActive &= acceptedSource.dispatch( statisticalConcept );
	}

	private void dispatchToAccepted (
		final KeyFamily keyFamily
	) {
		acceptedActive &= acceptedSource.dispatch( keyFamily );
	}

	private void dispatchToAccepted (
		final TimeSeries timeSeries
	) {
		acceptedActive &= acceptedSource.dispatch( timeSeries );
	}

	private void dispatchToAccepted (
		final Attribute attirbute
	) {
		acceptedActive &= acceptedSource.dispatch( attirbute );
	}

	private void dispatchToAccepted (
		final ErrorEvent errorEvent
	) {
		acceptedActive &= acceptedSource.dispatch( errorEvent );
	}

	private void dispatchToRejected (
		final InterchangeBegin interchangeBegin
	) {
		rejectedActive &= rejectedSource.dispatch( interchangeBegin );
	}

	private void dispatchToRejected (
		final InterchangeEnd interchangeEnd
	) {
		rejectedActive &= rejectedSource.dispatch( interchangeEnd );
	}

	private void dispatchToRejected (
		final MessageBegin messageBegin
	) {
		rejectedActive &= rejectedSource.dispatch( messageBegin );
	}

	private void dispatchToRejected (
		final MessageEnd messageEnd
	) {
		rejectedActive &= rejectedSource.dispatch( messageEnd );
	}

	private void dispatchToRejected (
		final CodeList codeList
	) {
		rejectedActive &= rejectedSource.dispatch( codeList );
	}

	private void dispatchToRejected (
		final StatisticalConcept statisticalConcept
	) {
		rejectedActive &= rejectedSource.dispatch( statisticalConcept );
	}

	private void dispatchToRejected (
		final KeyFamily keyFamily
	) {
		rejectedActive &= rejectedSource.dispatch( keyFamily );
	}

	private void dispatchToRejected (
		final TimeSeries timeSeries
	) {
		rejectedActive &= rejectedSource.dispatch( timeSeries );
	}

	private void dispatchToRejected (
		final Attribute attirbute
	) {
		rejectedActive &= rejectedSource.dispatch( attirbute );
	}

	private void dispatchToRejected (
		final ErrorEvent errorEvent
	) {
		rejectedActive &= rejectedSource.dispatch( errorEvent );
	}
}
