// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.sieve;

import java.util.Date;
import java.util.Calendar;

import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;

import Int.ecb.gesmescb.*;

import Int.ecb.gesmescb.event.*;

import Int.ecb.gesmescb.sieve.GESMESCBEventSieve;
import Int.ecb.gesmescb.sieve.TestObjectFilter;

/* ***************************** Class ****************************** */

/**
 * <P>Tests the implementation of {@link GESMESCBEventSieve}.</P>
 *
 * @author
 * <A HREF="mailto:Sami.Airo@ECB.int">Sami Airo</A>
 *
 * @version @(#)$Id: GESMESCBEventSieveTest.java,v 1.1 2017/01/19 16:14:04 trainot Exp $
 */

public class GESMESCBEventSieveTest extends TestCase {

	/* ======================= Constants ======================== */

	/**
	 * Version date of the read Key Family (=extraction date of Message).
	 */
	private static final Date VERSION_DATE;
	static {
		final Calendar calendar = Calendar.getInstance();
		calendar.clear();
		calendar.set( 2000, Calendar.DECEMBER, 1, 14, 10 );
		VERSION_DATE = calendar.getTime();
	}

	/* ========================= Fields ========================= */

	private final InterchangeBegin interchangeBegin = new InterchangeBegin(
		"IREF0001",
		"4F0",
		"4F0",
		"000315124533",
		false
	);

	private final MessageBegin messageBegin = new MessageBegin(
		"MREF0001",
		"4F0",
		"4F0",
		"ECB",
		MessageBegin.STATISTICAL_DATA,
		"De subject",
		new Contact[]{}
	);

	private final CodeList codeList = new CodeList(
		"CL_TEST",
		"Test Codelist",
		"For testing sieve",
		VERSION_DATE,
		new String[]{},
		new String[]{}
	);

	private final StatisticalConcept statisticalConcept
	= new StatisticalConcept(
		"SC_TEST",
		"Test StatisticalConcept",
		"For testing sieve",
		VERSION_DATE
	);

	private final KeyFamily keyFamily = new KeyFamily(
		"KF_TEST",
		"Test KeyFamily",
		"For testing sieve",
		VERSION_DATE,
		new KeyFamilyComponent[]{}
	);

	private final TimeSeries timeSeries = new TimeSeries(
		"A.B.C",
		"19991998",
		"702",
		new Observation[]{}
	);

	private final Attribute attribute = new Attribute(
		"A.B.C", "", "", "TEST_ATT", "0"
	);

	private final MessageEnd messageEnd = new MessageEnd(
		"MREF0001"
	);

	private final InterchangeEnd interchangeEnd = new InterchangeEnd(
		"IREF0001"
	);

	private final ErrorEvent warning = new ErrorEvent("Test warning", ErrorEvent.WARNING);

	private final ErrorEvent info = new ErrorEvent("Test info", ErrorEvent.INFO);

	/* ====================== Constructors ====================== */

	/**
	 * Constructs a GESMESCBEventSieveTest with the specified name.
	 *
	 * @param name Test case name
	 */
	public GESMESCBEventSieveTest ( final String name ) {
		super( name );
	}

	/* ===================== Public Methods ===================== */

	/**
	 * Assembles and returns a test suite for all
	 * the test methods of this test case.
	 *
	 * @return a non-null test suite.
	 */
	public static Test suite () {
		return new TestSuite( GESMESCBEventSieveTest.class );
	}

	/**
	 * Runs the test case.
	 */
	public static void main ( final String args[] ) {
		String[] testCaseName = {
			GESMESCBEventSieveTest.class.getName()
		};
		junit.textui.TestRunner.main( testCaseName );
	}

	/* --------------------- Test Methods --------------------- */

	/**
	 * Tests the handling of a good interchange.
	 */
	public void testAllAccepted () {
		final TestObjectFilter filter = new TestObjectFilter();
		final GESMESCBEventSieve sieve = new GESMESCBEventSieve( filter );
		final TestEventSource source = new TestEventSource();
		final TestEventListener accepted = new TestEventListener( 10 );
		final TestEventListener rejected = new TestEventListener( 1 );

		source.addListener( sieve, true );
		sieve.getAcceptedSource().addListener( accepted, true );
		sieve.getRejectedSource().addListener( rejected, true );

		filter.setDefaultResponse( true );

		source.dispatch( interchangeBegin );
		assertNull( accepted.getLastObject() );

		source.dispatch( messageBegin );
		assertSame(messageBegin, accepted.getLastObject() );

		source.dispatch( codeList );
		assertSame( codeList, accepted.getLastObject() );

		source.dispatch( statisticalConcept );
		assertSame( statisticalConcept, accepted.getLastObject() );

		source.dispatch( keyFamily );
		assertSame( keyFamily, accepted.getLastObject() );

		source.dispatch( timeSeries );
		assertSame( timeSeries, accepted.getLastObject() );

		source.dispatch( attribute );
		assertSame( attribute, accepted.getLastObject() );

		source.dispatch( messageEnd );
		assertSame( messageEnd, accepted.getLastObject() );

		source.dispatch( interchangeEnd );
		assertSame( interchangeEnd, accepted.getLastObject() );

		assertEquals( 9, accepted.size() );
		assertEquals( 0, rejected.size() );
	}

	/**
	 * Tests the handling of a completely unexpected surprise interchange.
	 */
	public void testAllRejected () {
		final TestObjectFilter filter = new TestObjectFilter();
		final GESMESCBEventSieve sieve = new GESMESCBEventSieve( filter );
		final TestEventSource source = new TestEventSource();
		final TestEventListener accepted = new TestEventListener( 1 );
		final TestEventListener rejected = new TestEventListener( 10 );

		source.addListener( sieve, true );
		sieve.getAcceptedSource().addListener( accepted, true );
		sieve.getRejectedSource().addListener( rejected, true );

		filter.setDefaultResponse( false );

		source.dispatch( interchangeBegin );
		assertNull( rejected.getLastObject() );

		source.dispatch( messageBegin );
		assertNull( rejected.getLastObject() );

		source.dispatch( codeList );
		assertSame( codeList, rejected.getLastObject() );

		source.dispatch( statisticalConcept );
		assertSame( statisticalConcept, rejected.getLastObject() );

		source.dispatch( keyFamily );
		assertSame( keyFamily, rejected.getLastObject() );

		source.dispatch( timeSeries );
		assertSame( timeSeries, rejected.getLastObject() );

		source.dispatch( attribute );
		assertSame( attribute, rejected.getLastObject() );

		source.dispatch( messageEnd );
		assertSame( messageEnd, rejected.getLastObject() );

		source.dispatch( interchangeEnd );
		assertSame( interchangeEnd, rejected.getLastObject() );

		assertEquals( 0, accepted.size() );
		assertEquals( 9, rejected.size() );
	}

	/**
	 * Tests the handling of an interchange with an unexpected message.
	 */
	public void testMessageRejected () {
		final TestObjectFilter filter = new TestObjectFilter();
		final GESMESCBEventSieve sieve = new GESMESCBEventSieve( filter );
		final TestEventSource source = new TestEventSource();
		final TestEventListener accepted = new TestEventListener( 10 );
		final TestEventListener rejected = new TestEventListener( 10 );

		source.addListener( sieve, true );
		sieve.getAcceptedSource().addListener( accepted, true );
		sieve.getRejectedSource().addListener( rejected, true );

		filter.setDefaultResponse( true );
		source.dispatch( interchangeBegin );
		assertNull( accepted.getLastObject() );
		assertNull( rejected.getLastObject() );

		filter.setDefaultResponse( false );
		source.dispatch( messageBegin );
		assertNull( accepted.getLastObject() );
		assertNull( rejected.getLastObject() );

		filter.setDefaultResponse( true );
		source.dispatch( codeList );
		assertSame( codeList, rejected.getLastObject() );
		source.dispatch( timeSeries );
		assertSame( timeSeries, rejected.getLastObject() );
		assertNull( accepted.getLastObject() );

		source.dispatch( messageEnd );
		assertSame( messageEnd, rejected.getLastObject() );

		source.dispatch( interchangeEnd );
		assertSame( interchangeEnd, rejected.getLastObject() );

		assertEquals( 0, accepted.size() );
		assertEquals( 6, rejected.size() );
	}

	/**
	 * Tests the handling of an interchange with a message
	 * with unexpected content.
	 */
	public void testContentRejected () {
		final TestObjectFilter filter = new TestObjectFilter();
		final GESMESCBEventSieve sieve = new GESMESCBEventSieve( filter );
		final TestEventSource source = new TestEventSource();
		final TestEventListener accepted = new TestEventListener( 10 );
		final TestEventListener rejected = new TestEventListener( 10 );

		source.addListener( sieve, true );
		sieve.getAcceptedSource().addListener( accepted, true );
		sieve.getRejectedSource().addListener( rejected, true );

		filter.setDefaultResponse( true );
		source.dispatch( interchangeBegin );
		assertNull(accepted.getLastObject());
		assertNull(rejected.getLastObject());
		source.dispatch( messageBegin );
		assertSame(messageBegin, accepted.getLastObject() );
		assertNull( rejected.getLastObject() );

		filter.setDefaultResponse( false );
		source.dispatch( codeList );
		assertSame( codeList, rejected.getLastObject() );
		source.dispatch( timeSeries );
		assertSame( timeSeries, rejected.getLastObject() );
		assertSame(messageBegin, accepted.getLastObject() );

		filter.setDefaultResponse( true );
		source.dispatch( messageEnd );
		assertSame( messageEnd, rejected.getLastObject() );

		source.dispatch( interchangeEnd );
		assertSame( interchangeEnd, rejected.getLastObject() );

		assertEquals( 4, accepted.size() );
		assertEquals( 6, rejected.size() );
	}

	/**
	 * Tests the handling of an interchange with
	 * a good message, an unexpected message, and one with mixed content.
	 */
	public void testMixed () {
		final TestObjectFilter filter = new TestObjectFilter();
		final GESMESCBEventSieve sieve = new GESMESCBEventSieve( filter );
		final TestEventSource source = new TestEventSource();
		final TestEventListener accepted = new TestEventListener( 10 );
		final TestEventListener rejected = new TestEventListener( 10 );

		source.addListener( sieve, true );
		sieve.getAcceptedSource().addListener( accepted, true );
		sieve.getRejectedSource().addListener( rejected, true );

		filter.setDefaultResponse( true );
		source.dispatch( interchangeBegin );
		source.dispatch( messageBegin );
		source.dispatch( codeList );
		source.dispatch( messageEnd );

		assertSame( messageEnd, accepted.getLastObject() );
		assertNull( rejected.getLastObject() );
		assertEquals( 4, accepted.size() );
		assertEquals( 0, rejected.size() );
		accepted.clearHistory();

		filter.setDefaultResponse( false );
		source.dispatch( messageBegin );
		filter.setDefaultResponse( true );
		source.dispatch( timeSeries );
		source.dispatch( attribute );
		source.dispatch( messageEnd );

		assertSame( messageEnd, rejected.getLastObject() );
		assertEquals( 0, accepted.size() );
		assertEquals( 5, rejected.size() );
		rejected.clearHistory();

		filter.setDefaultResponse( true );
		source.dispatch( messageBegin );
		source.dispatch( timeSeries );
		filter.setDefaultResponse( false );
		source.dispatch( attribute );
		filter.setDefaultResponse( true );

		assertSame( timeSeries, accepted.getLastObject() );
		assertSame( attribute, rejected.getLastObject() );

		filter.setDefaultResponse( true );
		source.dispatch( messageEnd );

		assertSame( messageEnd, accepted.getLastObject() );
		assertSame( messageEnd, rejected.getLastObject() );

		source.dispatch( interchangeEnd );
		assertSame( interchangeEnd, accepted.getLastObject() );
		assertSame( interchangeEnd, rejected.getLastObject() );

		assertEquals( 4, accepted.size() );
		assertEquals( 4, rejected.size() );
	}

	public void testErrorWarning () {
		final TestObjectFilter filter = new TestObjectFilter();
		final GESMESCBEventSieve sieve = new GESMESCBEventSieve( filter );
		final TestEventSource source = new TestEventSource();
		final TestEventListener accepted = new TestEventListener( 20 );
		final TestEventListener rejected = new TestEventListener( 20 );

		source.addListener( sieve, true );
		sieve.getAcceptedSource().addListener( accepted, true );
		sieve.getRejectedSource().addListener( rejected, true );

		filter.setDefaultResponse( true );

		source.dispatch( interchangeBegin );
		assertNull( accepted.getLastObject() );

		source.dispatch( messageBegin );
		assertSame(messageBegin, accepted.getLastObject() );

		source.dispatch( info );
		assertSame( info, accepted.getLastObject() );

		source.dispatch( codeList );
		assertSame( codeList, accepted.getLastObject() );

		source.dispatch( warning );
		assertSame( warning, accepted.getLastObject() );

		source.dispatch( statisticalConcept );
		assertSame( statisticalConcept, accepted.getLastObject() );

		source.dispatch( keyFamily );
		assertSame( keyFamily, accepted.getLastObject() );

		source.dispatch( timeSeries );
		assertSame( timeSeries, accepted.getLastObject() );

		source.dispatch( attribute );
		assertSame( attribute, accepted.getLastObject() );

		source.dispatch( messageEnd );
		assertSame( messageEnd, accepted.getLastObject() );

		source.dispatch( interchangeEnd );
		assertSame( interchangeEnd, accepted.getLastObject() );

		assertEquals( 11, accepted.size() );
		assertEquals( 2, rejected.size() );
	}

	/**
	 * Tests the automatic disconnect when nobody important is listening.
	 */
	public void testAutoDisconnect () {
		final TestObjectFilter filter = new TestObjectFilter();
		final GESMESCBEventSieve sieve = new GESMESCBEventSieve( filter );
		final TestEventSource source = new TestEventSource();
		final TestEventListener accepted = new TestEventListener( 10 );
		final TestEventListener rejected = new TestEventListener( 10 );

		source.addListener( sieve, true );
		sieve.getAcceptedSource().addListener( accepted, true );
		sieve.getRejectedSource().addListener( rejected, true );
		assertTrue( "Initially active", sieve.isActive() );

		filter.setDefaultResponse( true );
		source.dispatch( interchangeBegin );
		source.dispatch( messageBegin );
		source.dispatch( codeList );
		assertTrue( "Active after accept", sieve.isActive() );

		filter.setDefaultResponse( false );
		source.dispatch( statisticalConcept );
		assertTrue( "Active after reject", sieve.isActive() );

		sieve.getAcceptedSource().removeAllListeners();
		filter.setDefaultResponse( true );
		source.dispatch( keyFamily );
		filter.setDefaultResponse( false );
		source.dispatch( timeSeries );
		assertTrue( "Active after unheard accept", sieve.isActive() );

		sieve.getRejectedSource().removeAllListeners();
		filter.setDefaultResponse( true );
		source.dispatch( attribute );
		filter.setDefaultResponse( false );
		source.dispatch( messageEnd );
		assertTrue( "Inactive after unheard", !sieve.isActive() );

		source.dispatch( interchangeEnd );
		assertTrue( "Finally inactive", !sieve.isActive() );

		assertSame( codeList, accepted.getLastObject() );
		assertSame( timeSeries, rejected.getLastObject() );
	}
}
