// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.sieve;

import java.lang.String;
import java.util.Date;
import java.text.SimpleDateFormat;
import java.text.ParseException;

import Int.ecb.gesmescb.util.ComponentHelper;
import Int.ecb.logging.Level;

import Int.ecb.gesmescb.Observation;
import Int.ecb.gesmescb.MessageBegin;

import Int.ecb.gesmescb.time.RegularInterval;

/* **************************** Class ***************************** */

/**
 * <P>Filters observations based on the defined interval.</P>
 *
 * @author
 * <A HREF="mailto:Sami.Airo@ECB.int">Sami Airo</A>
 *
 * @version @(#)$Id: IntervalObservationFilter.java,v 1.1 2017/01/19 16:14:04 trainot Exp $
 */

public final class IntervalObservationFilter implements ObservationFilter {

	/* ======================= Constants ======================== */
	
	public final String ISO_DATE_FORMAT = "yyyyMMdd'T'HHmmss";

	/* ========================= Fields ========================= */

	/** 
	 * Takes care of logging and prefs 
	 */
	private ComponentHelper componentHelper;

	/**
	 * First allowed date
	 */
	private Date firstAllowed = new Date(Long.MIN_VALUE);

	/**
	 * Last allowed date
	 */
	private Date lastAllowed = new Date(Long.MAX_VALUE);

	/**
	 * Format for dates
	 */
	private SimpleDateFormat format = new SimpleDateFormat(ISO_DATE_FORMAT);

	/**
	 * Logging level for rejections
	 */
	private Level rejectionLevel;

	/* ====================== Constructors ====================== */

	public IntervalObservationFilter () {
		this( "" );
	}

	/**
	 * Constructs a IntervalObservationFilter.
	 */
	public IntervalObservationFilter (final String name) {
		componentHelper = new ComponentHelper( name, this );
				
		try {
			String beginDate = componentHelper.getPreference("beginDate", "");		
			if (!beginDate.equals("")) {		
				firstAllowed = format.parse(beginDate);

				componentHelper.logger.log(
					Level.CONFIG, 
					componentHelper.logName, 
					"IntervalObservationFilter",
					"First allowed date set to " + format.format(firstAllowed)
				);
			}
		} catch (ParseException e) {
			componentHelper.logger.log(
				Level.SEVERE, 
				componentHelper.logName, 
				"IntervalObservationFilter",
				"Error in parsing beginDate: " + e
			);		
		}

		try {		
			String endDate = componentHelper.getPreference("endDate", "");
			if (!endDate.equals("")) {
				lastAllowed = format.parse(endDate);
	
				componentHelper.logger.log(
					Level.CONFIG, 
					componentHelper.logName, 
					"IntervalObservationFilter",
					"Last allowed date set to " + format.format(lastAllowed)
				);
			}
		} catch (ParseException e) {
			componentHelper.logger.log(
				Level.SEVERE, 
				componentHelper.logName, 
				"IntervalObservationFilter",
				"Error in parsing endDate: " + e
			);		
		}
		
		if (lastAllowed.before(firstAllowed)) {
			componentHelper.logger.log(
				Level.SEVERE, 
				componentHelper.logName, 
				"IntervalObservationFilter",
				"Error in setting interval: end date is before begin date"
			);					
		}

		rejectionLevel = Level.parse( componentHelper.getPreference("rejectionLevel", "FINE") );
	}

	/* ===================== Public Methods ===================== */

	public boolean accept ( 
		final MessageBegin messageBegin,
		final String seriesKey,
		final Observation observation,  
		final RegularInterval interval
	) {
		componentHelper.logger.log(
			Level.FINE, componentHelper.logName, "accept",
			"checking observation interval", interval
		);
				
		if ( interval.getBegin().before(firstAllowed) ) {
			componentHelper.logger.log(
				rejectionLevel, 
				componentHelper.logName, 
				"accept",
				"Observation rejected: interval begin date " 
				+ format.format(interval.getBegin())
				+ " before the allowed first date " + format.format(firstAllowed)
			);
			return false;
		}

		if ( interval.getEnd().after(lastAllowed) ) {
			componentHelper.logger.log(
				rejectionLevel, 
				componentHelper.logName, 
				"accept",
				"Observation rejected: interval end date "
				+ format.format(interval.getEnd())
				+ " after the allowed last date " + format.format(lastAllowed)
			);
			return false;
		}
		
		return true;
	}

}
