// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (C) 2004, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.sieve;

import java.util.TreeMap;
import java.util.Iterator;

import Int.ecb.logging.Level;

import Int.ecb.gesmescb.InterchangeBegin;
import Int.ecb.gesmescb.MessageBegin;
import Int.ecb.gesmescb.CodeList;
import Int.ecb.gesmescb.StatisticalConcept;
import Int.ecb.gesmescb.KeyFamily;
import Int.ecb.gesmescb.TimeSeries;
import Int.ecb.gesmescb.Attribute;

import Int.ecb.gesmescb.time.RegularInterval;
import Int.ecb.gesmescb.util.ComponentHelper;
import Int.ecb.gesmescb.event.DataCollector;


/* **************************** Class ***************************** */

/**
 * <p>Filters {@link Attribute}s. All other entities are accepted.</p>
 * <p>Reject the {@link Attribute}s that are provided
 * in the properties.</p>
 * <p>If a {@link DataCollector} is provided, typically linked to
 * the rejection pipe of a {@link ObservationSieveWithReject},
 * the <code>ObsLevelAttributeFilter</code> rejects the 
 * {@link Observation}-level {@link Attribute}s
 *  for which an {@link Observation} has been collected. Objective
 *  is to discard the Attributes that correspond to the discarded
 *  Observations.
 * </p>
 * @see ObservationSieveWithReject
 *
 * @author
 * <a href="mailto:Francois.Hoeltzli@ECB.int">Francois Hoeltzli</A>
 *
 * @version @(#)$Id: ObsLevelAttributeFilter.java,v 1.1 2017/01/19 16:14:04 trainot Exp $
 */
public class ObsLevelAttributeFilter
	implements GESMESCBObjectFilter
{

	/* ========================= Fields ========================= */

	/** Takes care of logging and preferences. */
	protected final ComponentHelper componentHelper;

	/** Attributes names that are rejected whatever the series and level */
	private final TreeMap rejectedNamesMap;
	
	/** DataCollector where discarded observations are stored. */
	private final DataCollector dataCollector;
	
	/** DataCollector where discarded observations are stored. */
	private final boolean useDataCollector;
	
	/** 
	 * Whether to wait for {@link MessageEnd} before filtering 
	 * the {@link Attribute}s. Otherwise, if the attributes
	 * come before their {@link Observation} they might
	 * get incorrectly accepted.
	 * Note: This behaviour is not implemented yet. TODO!
	 * Default: False. */
	private final boolean waitForMessageEnd;
	
	
	/* ====================== Constructors ====================== */

	/** Constructs a ObsLevelAttributeFilter with the specified name. */
	public ObsLevelAttributeFilter ( final String name ) {
		
		componentHelper = new ComponentHelper( name, this );
		
		rejectedNamesMap = new TreeMap();
		String ignoredAttributes;
		int i=0;
		for(;;) {
			ignoredAttributes = componentHelper.getPreference(
				"ignoredAttributes." + (++i)
				, null
			);
			if ( ignoredAttributes == null )
				break;
			rejectedNamesMap.put( ignoredAttributes, null);
		}
			
		waitForMessageEnd = componentHelper.getPreference(
			"waitForMessageEnd", "false"
		).equalsIgnoreCase( "true" );
		
		useDataCollector = componentHelper.getPreference(
			"useDataCollector", "true"
		).equalsIgnoreCase( "true" );
			
		dataCollector = DataCollector.getDataCollector();
	}
	
	/** Constructs a ObsLevelAttributeFilter. */
	public ObsLevelAttributeFilter () {
		this( "" );
	}
	/* ===================== Public Methods ===================== */

	public boolean accept ( final InterchangeBegin interchangeBegin ) {
		componentHelper.logger.log(
			Level.FINE,
			componentHelper.logName,
			"accept",
			"InterchangeBegin accepted.",
			interchangeBegin
		);
		return true;
	}

	public boolean accept ( final MessageBegin messageBegin ) {
		componentHelper.logger.log(
			Level.FINE,
			componentHelper.logName,
			"accept",
			"MessageBegin accepted.",
			messageBegin
		);
		return true;
	}

	public boolean accept ( final CodeList codeList ) {
		componentHelper.logger.log(
			Level.FINE,
			componentHelper.logName,
			"accept",
			"CodeList accepted.",
			codeList
		);
		return true;
	}

	public boolean accept ( final StatisticalConcept statisticalConcept ) {
		componentHelper.logger.log(
			Level.FINE,
			componentHelper.logName,
			"accept",
			"StatisticalConcept accepted.",
			statisticalConcept
		);
		return true;
	}

	public boolean accept ( final KeyFamily keyFamily ) {
		componentHelper.logger.log(
			Level.FINE,
			componentHelper.logName,
			"accept",
			"KeyFamily accepted.",
			keyFamily
		);
		return true;
	}

	public boolean accept ( final TimeSeries timeSeries ) {
		componentHelper.logger.log(
			Level.FINEST,
			componentHelper.logName,
			"accept",
			"TimeSeries accepted.",
			timeSeries
		);
		return true;
	}

	public boolean accept ( final Attribute attribute ) {
		if ( waitForMessageEnd ) return false; //TODO:accept(MessageEnd){ ...
		/* TODO
		 * check first whether the attribute is in the list of attributes
		 * that are rejected anyway */
		if ( rejectedNamesMap.containsKey( attribute.getName() )) 
			return false;
		if (! useDataCollector)
			return true;
		/* then take care of the special case of the obs-level attributes,
		 * which we might want to filter out when their corresponding
		 * observation is available in the dataCollector.
		 * Note if the interval of the attribute spans several observations:
		 * one of the observation accepted (not in collector)=> the attribute
		 * is accepted with its full interval.*/
		RegularInterval interval = attribute.getInterval();
		if ( interval == null )
			return true;
		String seriesKey = attribute.getKey();
		if (! dataCollector.hasTimeSeries( seriesKey ) )
			return true;
		boolean foundInDataCollector;
		for ( Iterator iterator = interval.iterator()
			; iterator.hasNext()
			;
		) {
			RegularInterval basePeriod = (RegularInterval) iterator.next();
			foundInDataCollector = dataCollector.hasObservation(
				seriesKey,
				basePeriod
			);
			if (! foundInDataCollector) {
				componentHelper.logger.log(
					Level.FINEST,
					componentHelper.logName,
					"accept",
					"Attribute accepted.",
					attribute
				);
				return true;
			}		
		}
		componentHelper.logger.log(
			Level.FINEST, componentHelper.logName, "accept",
			"Attribute rejected.", attribute
		);
		return false;
	}
}
