// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (C) 2004, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.sieve;

import java.util.Collections;
import java.util.Date;
import java.util.Calendar;

import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;

import Int.ecb.gesmescb.InterchangeBegin;
import Int.ecb.gesmescb.MessageBegin;
import Int.ecb.gesmescb.CodeList;
import Int.ecb.gesmescb.Observation;
import Int.ecb.gesmescb.StatisticalConcept;
import Int.ecb.gesmescb.KeyFamily;
import Int.ecb.gesmescb.TimeSeries;
import Int.ecb.gesmescb.Attribute;
import Int.ecb.gesmescb.event.DataCollector;

/* ***************************** Class ****************************** */

/**
 * Provides test data and facilities for {@link ObsLevelAttributeFilter}.
 *
 * @author
 * <a href="mailto:Francois.Hoeltzli@ECB.int">Francois Hoeltzli</A>
 *
 * @version @(#)$Id: ObsLevelAttributeFilterTest.java,v 1.1 2017/01/19 16:14:04 trainot Exp $
 */
public class ObsLevelAttributeFilterTest extends TestCase {

	/* ======================= Constants ======================== */

	/** Version date of the read Key Family (=extraction date of Message). */
	protected static final Date VERSION_DATE;
	static {
		final Calendar calendar = Calendar.getInstance();
		calendar.clear();
		calendar.set( 2000, Calendar.DECEMBER, 1, 14, 10 );
		VERSION_DATE = calendar.getTime();
	}

	protected static final InterchangeBegin INTERCHANGE_BEGIN
	= new InterchangeBegin(
		"IREF000001", "AT2", "4F0", "040804:1143", true
	);

	protected static final MessageBegin STRUCTURE_MESSAGE_BEGIN
	= new MessageBegin(
		"MREF000001", "4F0", "ZZZ", "ECB",
		MessageBegin.STATISTICAL_DEFINITIONS,
		"Test Structure Message", Collections.EMPTY_SET
	);

	protected static final MessageBegin DATA_MESSAGE_BEGIN = new MessageBegin(
		"MREF000001", "AT2", "4F0", "ECB",
		MessageBegin.STATISTICAL_DATA,
		"Test Data Message", Collections.EMPTY_SET,
		"ECB_OB1",
		MessageBegin.UPDATE_OR_REPLACE_ACTION,
		"20040804:1112", "20002003", "702",
		"ECB_OB1"
	);

	protected static final MessageBegin DELETE_MESSAGE_BEGIN
	= new MessageBegin(
		"MREF000001", "4F0", "AT2", "ECB",
		MessageBegin.STATISTICAL_DATA,
		"Test Delete Message", Collections.EMPTY_SET,
		"ECB_OB1",
		MessageBegin.DELETE_ACTION,
		"20040804:1112", "20002003", "702",
		"ECB_OB1"
	);

	protected final CodeList CODE_LIST = new CodeList(
		"CL_TEST", "Test CodeList", "", VERSION_DATE,
		Collections.EMPTY_LIST, Collections.EMPTY_LIST
	);

	protected static final StatisticalConcept STATISTICAL_CONCEPT
	= new StatisticalConcept(
		"SC_TEST", "Test Concept", "", VERSION_DATE
	);

	protected static final KeyFamily KEY_FAMILY = new KeyFamily(
		"KF_TEST", "Test KeyFamily", "", VERSION_DATE, Collections.EMPTY_LIST
	);

	protected static final TimeSeries TIME_SERIES = new TimeSeries(
		"BLACK.SOX", "2003", "602",
		new Observation[]
		{
			new Observation ("ND", "A", "R", "")
		}
	);

	protected static final Attribute ATTRIBUTE = new Attribute(
		"TS.TEST", "", "", "ATT", "13.6"
	);

	protected static final Attribute OBS_LEVEL_ATTRIBUTE = new Attribute(
		"BLACK.SOX", "2003", "602", "OBS_COM", "rosa."		
	);
	protected static final Attribute OBS_LEVEL_ATTRIBUTE_2 = new Attribute(
			"BLACK.PANTHER", "2003", "602", "OBS_COM", "rosa (voc.)"		
		);
	protected static final Attribute OBS_LEVEL_ATTRIBUTE_3 = new Attribute(
			"BLACK.SOX", "2013", "602", "OBS_COM", "rosam."		
		);
	protected static final Attribute SER_LEVEL_ATTRIBUTE = new Attribute(
		"BLACK.SOX", "", "", "TITLE", "Series level."	
	);
	
	
	/* ====================== Constructors ====================== */

	/**
	 * Constructs a ObsLevelAttributeFilterTest with the specified name.
	 *
	 * @param name Test case name
	 */
	public ObsLevelAttributeFilterTest ( final String name ) {
		super( name );
	}

	/**
	 * Assembles and returns a test suite for all
	 * the test methods of this test case.
	 *
	 * @return a non-null test suite.
	 */
	public static Test suite () {
		return new TestSuite( ObsLevelAttributeFilterTest.class );
	}

	/**
	 * Runs the test case.
	 */
	public static void main ( final String args[] ) {
		String[] testCaseName = {
			ObsLevelAttributeFilterTest.class.getName()
		};
		junit.textui.TestRunner.main( testCaseName );
	}

	/* ===================== Public Methods ===================== */

	/* --------------------- Test Methods --------------------- */
	
	public void testScenarioWORejection () {
		DataCollector dataCollector = DataCollector.getDataCollector();
		dataCollector.clear();
		ObsLevelAttributeFilter filter=new ObsLevelAttributeFilter ();
		
		assertTrue(
			"Interchange SimpleScenario"
			, filter.accept( INTERCHANGE_BEGIN )
		);
		assertTrue(
			"Message ATT",
			filter.accept( DATA_MESSAGE_BEGIN )
		);
		assertTrue(
			"Obslevel Attribute should be accepted (dataCollector emtpy)",
			filter.accept( OBS_LEVEL_ATTRIBUTE )
		);
		assertTrue(
			"Series level Attribute should be accepted",
			filter.accept( SER_LEVEL_ATTRIBUTE )
		);
		assertTrue(
			"Series should be accepted",
			filter.accept( TIME_SERIES )
		);
	}

	public void testScenarioWithRejection () {
		DataCollector dataCollector = DataCollector.getDataCollector();
		dataCollector.clear();
		dataCollector.handle( TIME_SERIES );
		ObsLevelAttributeFilter filter=new ObsLevelAttributeFilter ();
		
		assertTrue(
			"Interchange SimpleScenario"
			, filter.accept( INTERCHANGE_BEGIN )
		);
		assertTrue(
			"Message ATT",
			filter.accept( DATA_MESSAGE_BEGIN )
		);
		assertTrue(
			"Obslevel Attribute should be rejected)",
			!filter.accept( OBS_LEVEL_ATTRIBUTE )
		);
		assertTrue(
			"Series level Attribute should be accepted",
			filter.accept( SER_LEVEL_ATTRIBUTE )
		);
		assertTrue(
			"Series should be accepted",
			filter.accept( TIME_SERIES )
		);
		assertTrue(
			"Obs level Attribute for another series should be accepted",
			filter.accept( OBS_LEVEL_ATTRIBUTE_2 )
		);
		assertTrue(
			"Obs level Attribute for another period should be accepted",
			filter.accept( OBS_LEVEL_ATTRIBUTE_3 )
		);			
	}
}
