// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.sieve;

import java.util.Iterator;
import java.util.List;
import java.util.ArrayList;
import java.util.TreeMap;
import java.util.SortedMap;
import java.lang.reflect.Constructor;

import Int.ecb.logging.Level;

import Int.ecb.gesmescb.*;
import Int.ecb.gesmescb.time.*;
import Int.ecb.gesmescb.event.GESMESCBEventBroker;

/* **************************** Class ***************************** */

/**
 * <P>Sieves observations into accepted and rejected. The accepted 
 * observations are passed further.</P>
 *
 * @author
 * <A HREF="mailto:Sami.Airo@ECB.int">Sami Airo</A>
 *
 * @version @(#)$Id: ObservationSieve.java,v 1.1 2017/01/19 16:14:04 trainot Exp $
 */

public class ObservationSieve extends GESMESCBEventBroker {

	/* ======================= Constants ======================== */
	
	/* ========================= Fields ========================= */

	/** 
	 * List of filters to use for sieving 
	 */
	protected List filters;

	/** 
	 * MessageBegin for the current message 
	 */
	protected MessageBegin messageBegin;

	/* ====================== Constructors ====================== */

	public ObservationSieve () {
		this( "" );
	}

	/**
	 * Constructs a ObservationSieve.
	 */
	public ObservationSieve ( final String name ) {
		super( name, true );
		
		filters = new ArrayList();

		int count = 1;
		while (1==1) {
			final String observationFilterClass = componentHelper.getPreference(
				"observationFilterClass." + count, null
			);
	
			if (observationFilterClass == null) {
				break;
			}
	
			final ObservationFilter filter = getFilter(observationFilterClass);
			
			if (filter == null) 
				throw new NullPointerException("Error in adding filter for class " + observationFilterClass);
			
			filters.add(filter);
						
			count++;
		}

		if (filters.isEmpty()) {
			componentHelper.logger.log(
				Level.INFO, componentHelper.logName, "ObservationSieve",
				"ObservationSieve contructed with no filters"
			);
		}
	}

	/* ===================== Public Methods ===================== */

	/**
	 * Inspect and process a <CODE>MessageBegin</CODE>.
	 *
	 * @param messageBegin to be handled.
	 */
	public void handle ( final MessageBegin messageBegin ) {
		this.messageBegin = messageBegin;
		dispatch(messageBegin);
	}
	
	/**
	 * Inspect and process a <CODE>TimeSeries</CODE>.
	 *
	 * @param timeSeries to be handled.
	 */
	public void handle ( final TimeSeries timeSeries ) {
		componentHelper.logger.log(
			Level.FINE, componentHelper.logName, "handle",
			"handling TimeSeries", timeSeries
		);

		SortedMap observationMap = new TreeMap();
		RegularInterval ri = timeSeries.getInterval();
		Iterator iit = ri != null
			? ri.iterator()
			: new NullIterator();
		Iterator oit = timeSeries.observationIterator();
		
		while (iit.hasNext()) {
			RegularInterval interval = (RegularInterval)iit.next();
			Observation observation = (Observation)oit.next();

			if (observation.equals(Observation.EMPTY_OBSERVATION)) continue;

			Iterator fit = filters.iterator();
			boolean accepted = true;
			while (fit.hasNext()) {
				ObservationFilter filter = (ObservationFilter) fit.next();
				accepted = (filter.accept(messageBegin, timeSeries.getKey(), observation, interval) ? accepted : false);
			}

			if (!accepted) {
				componentHelper.logger.log(
					Level.FINEST, componentHelper.logName, "handle",
					"Observation rejected: " + observation + " at " + interval
				);
				continue;
			}

			observationMap.put(interval, observation);			
		}
		
		if (observationMap.isEmpty()) {
			componentHelper.logger.log(
				Level.FINEST, componentHelper.logName, "handle",
				"All time series observation rejected: " + timeSeries
			);
			return;
		}

		RegularInterval first = (RegularInterval)observationMap.firstKey();
		RegularInterval last = (RegularInterval)observationMap.lastKey();

		RegularInterval interval = new GregorianRegularInterval(
			first.getFrequency(),
			first.getBegin(),
			last.getEnd()
		);

		Iterator rit = interval.iterator();
		ArrayList observationList = new ArrayList();
		while (rit.hasNext()) {			
			RegularInterval intervalBase = (RegularInterval)rit.next();
			Observation observation = (Observation)observationMap.get(intervalBase);
			if (observation == null) observation = Observation.EMPTY_OBSERVATION;
			observationList.add(observation);
		}		
	
		dispatch( new TimeSeries( timeSeries.getKey(), interval, observationList) );
	}
		
	private ObservationFilter getFilter(
		final String filterClass
	) {
		componentHelper.logger.log(
			Level.CONFIG,
			componentHelper.logName,
			"getFilter",
			"Using " + filterClass + " for filtering."
		);

		ObservationFilter configuredFilter = null;
		try {
			final Class valueClass = Class.forName( filterClass );
			final Constructor constructor = valueClass.getConstructor(
				new Class[]{ String.class }
			);
			configuredFilter = (ObservationFilter)constructor.newInstance(
				new Object[]{ "" }
			);
		}
		catch ( final Exception e ) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"getFilter",
				"Cannot construct filter from class \"" + filterClass + "\".",
				e
			);
		}
		
		return configuredFilter;
	}

}

/* *********************** Auxiliary Class ************************ */

class NullIterator implements Iterator {
	public NullIterator() {
	}
	public boolean hasNext(){ 
		return false; 
	}
	public Object next(){
		return null;
	}
	public void remove(){};
}