// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.sieve;

import java.util.Iterator;
import java.util.ArrayList;
import java.util.TreeMap;
import java.util.SortedMap;
import java.util.NoSuchElementException;

import Int.ecb.logging.Level;

import Int.ecb.gesmescb.*;
import Int.ecb.gesmescb.time.*;
import Int.ecb.gesmescb.event.GESMESCBEventSourceHelper;
import Int.ecb.gesmescb.event.GESMESCBEventListener;

/* **************************** Class ***************************** */

/**
 * <P>Sieves observations into accepted and rejected.
 * The accepted observations are passed further.
 * The rejected  observations are passed as time series to a rejection 
 * pipe.</P>
 * <table>
 * <caption>
 * <a name="prefs">Available {@link Int.ecb.prefs preference} settings</a>
 * </caption>
 * <tr>
 * <th>preference name</th>
 * <th>type</th>
 * <th>meaning</th>
 * <th>default</th>
 * </tr>
 * <tr>
 * <td><tt>rejectionPipe</tt></td>
 * <td>{@link GESMESCBEventSourceHelper}</td>
 * <td>where rejected {@link Observation}s are sent to</td>
 * <td><tt>-</tt></td>
 * </tr>
 * <tr>
 * <td><tt>atomiseObsLevelAttributes</tt></td>
 * <td>boolean</td>
 * <td>Whether to split {@link Attribute}s spanning over a range 
 * of Observations into atomic attributes, one per Observation.</td>
 * <td><tt>false</tt></td>
 * </tr>
 * </table>
 * @author
 * <A HREF="mailto:Francois.Hoeltzli@ECB.int">Francois Hoeltzli</A>
 *
 * @version @(#)$Id: ObservationSieveWithReject.java,v 1.1 2017/01/19 16:14:04 trainot Exp $
 */

public final class ObservationSieveWithReject extends ObservationSieve
{

	/* ======================= Constants ======================== */

	private final static int ACCEPTED = 1;
	private final static int REJECTED = 2;
	
	/* ========================= Fields ========================= */
	private final GESMESCBEventSourceHelper rejectionPipe;
	private final boolean atomiseObsLevelAttributes;
	private String currentAction;
	/* ====================== Constructors ====================== */
	public ObservationSieveWithReject() {
		this( "" );
	}

	public ObservationSieveWithReject( String name ) {
		super( name );
		atomiseObsLevelAttributes = componentHelper.getPreference(
			"atomiseObsLevelAttributes", "false"
		).equalsIgnoreCase( "true" );
		rejectionPipe = new GESMESCBEventSourceHelper();
	}

	/* ===================== Public Methods ===================== */

	/**
	 *  Reads the MessageBegin to set the currentAction
	 *  
	 *  @param MessageBegin to be handled.
	 */
	public void handle ( final MessageBegin messageBegin ) {
		currentAction = messageBegin.getAction();
		super.handle( messageBegin );
	}
	
	/**
	 *  Reads the MessageEnd and reset currentAction
	 *  
	 *  @param MessageEnd to be handled.
	 */
	public void handle ( final MessageEnd messageEnd ) {
		currentAction = null;
		super.handle( messageEnd );
	}
	
	/**
	 * Inspect and process a <CODE>TimeSeries</CODE>.
	 *
	 * @param timeSeries to be handled.
	 */
	public void handle ( final TimeSeries timeSeries ) {
		if ( messageBegin.isDeleteAction() ) {
			componentHelper.logger.log(
				Level.FINE, componentHelper.logName, "handle",
				"Ignoring (delete) TimeSeries while filtering", timeSeries
			);
			dispatch ( timeSeries );
			return;
		}
		componentHelper.logger.log(
			Level.FINE, componentHelper.logName, "handle",
			"handling TimeSeries", timeSeries
		);

		SortedMap acceptedObservationMap = new TreeMap();
		SortedMap rejectedObservationMap = new TreeMap();
		RegularInterval ri = timeSeries.getInterval();
		Iterator iit = ri != null
			? ri.iterator()
			: new NullIterator();
		Iterator oit = timeSeries.observationIterator();
		
		while (iit.hasNext()) {
			RegularInterval interval = (RegularInterval)iit.next();
			Observation observation = (Observation)oit.next();

			if (observation.equals(Observation.EMPTY_OBSERVATION)) continue;

			Iterator fit = filters.iterator();
			boolean accepted = true;
			while (fit.hasNext()) {
				ObservationFilter filter = (ObservationFilter) fit.next();
				accepted = filter.accept
				(
					messageBegin,
					timeSeries.getKey(),
					observation,
					interval
				)
				? accepted
				: false;
			}

			if (!accepted) {
				componentHelper.logger.log(
					Level.FINEST, componentHelper.logName, "handle",
					"Observation rejected: " + observation + " at " + interval
				);
				rejectedObservationMap.put(interval, observation);			
				continue;
			}
			acceptedObservationMap.put(interval, observation);			
		}
		
		String key = timeSeries.getKey();
		if (acceptedObservationMap.isEmpty()) {
			componentHelper.logger.log(
				Level.FINEST, componentHelper.logName, "handle",
				"All time series observation rejected: " + timeSeries
			);
		} else 
			dispatch( ACCEPTED, key, acceptedObservationMap );
		
		if (rejectedObservationMap.isEmpty()) {
			componentHelper.logger.log(
				Level.FINEST, componentHelper.logName, "handle",
				"No time series observation rejected: " + timeSeries
			);
		} else 
			dispatch( REJECTED, key, rejectedObservationMap );
	}

	/**
	 * Inspect and process an <CODE>Attribute</CODE>.
	 * If the attribute is Obs-level and refers to multiple obs
	 * we need to split it.
	 * @param attribute to be handled.
	 */
	public void handle ( final Attribute attribute ) {
		RegularInterval interval = attribute.getInterval();
		if ( null == interval || !atomiseObsLevelAttributes ) {
			super.handle(attribute);
		} else {
			componentHelper.logger.log(
				Level.WARNING, componentHelper.logName, "handle",
					"handling and atomising Attribute", attribute
			);
			for ( Iterator iterator = interval.iterator()
				; iterator.hasNext()
				;
			) {
				RegularInterval atomicInterval;
				atomicInterval = (RegularInterval) iterator.next();
				
				Attribute atomicAttribute
				= new Attribute (
					attribute.getKey(),
					atomicInterval,
					attribute.getName(),
					attribute.getValue()
				);				
				super.handle( atomicAttribute );				
			}
		}
	}
	
	
	/**
	 * allow to add some listener to the rejection pipe.
	 */
	public void addRejectionListener(
		GESMESCBEventListener listener,
		boolean isImportant
	) {
		rejectionPipe.addListener( listener, isImportant );
	};

	private void dispatch(
		int destination,
		String key,
		SortedMap observationMap
	) {
	
		RegularInterval intervalBegin;
		RegularInterval intervalEnd;
		RegularInterval interval;
		
		try {
			intervalBegin = (RegularInterval) observationMap.firstKey();
			intervalEnd = (RegularInterval) observationMap.lastKey();
		} catch ( NoSuchElementException e ) {
			/* should never happen.*/
			componentHelper.logger.log(
				Level.SEVERE, componentHelper.logName, "dispatch",
				"Trying to dispatch series \"" + key + "\""
				+ "to " + ( destination==1?" accepted":" refused" )
				+ "with an empty observationMap."
			);
			return;
		}

		interval = new GregorianRegularInterval(
			intervalBegin.getFrequency(),
			intervalBegin.getBegin(),
			intervalEnd.getEnd()
		);

		Iterator it = interval.iterator();
		ArrayList observationList = new ArrayList();
		while (it.hasNext()) {			
			RegularInterval intervalBase = (RegularInterval)it.next();
			Observation observation
			= (Observation) observationMap.get(intervalBase);
			if (observation == null)
				observation = Observation.EMPTY_OBSERVATION;
			observationList.add(observation);
		}		
		
		TimeSeries timeSeries
		= new TimeSeries( key, interval, observationList );
		
		switch( destination ) {
			case ACCEPTED:
				this.dispatch( timeSeries );
				break;
			case REJECTED:
				rejectionPipe.dispatch( timeSeries );
				break;
			default:
				componentHelper.logger.log(
					Level.SEVERE,
					componentHelper.logName,
					"observationSieveWithReject",
					"Dispatched called with destination \"" + destination + "\"."
				);
				break;
		}
		
	};
		
};


/* *********************** Auxiliary Class ************************ */
