// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001-2004, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.sieve;

import java.io.File;
import java.io.FileReader;
import java.io.BufferedReader;
import java.io.IOException;

import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;

import Int.ecb.logging.Level;
import Int.ecb.gesmescb.GESMESCBObject;
import Int.ecb.gesmescb.MessageBegin;
import Int.ecb.gesmescb.TimeSeries;
import Int.ecb.gesmescb.Attribute;
import Int.ecb.gesmescb.metabase.MaintenanceAgency;
import Int.ecb.gesmescb.metabase.Metabase;

/* **************************** Class ***************************** */

/**
 * <P>Judges {@link GESMESCBObject}s according to preference settings.</P>
 *
 * <p>The lists of expected data flows are told to the
 * <code>PreferenceConfiguredFilter</code>
 * using {@link Int.ecb.prefs.Preferences}.</p>
 *
 * <p>The acceptance rules are the following:
 * <ul>
 * <li>A data flow is accepted if a preference of the form
 * <code>sender.receiver.maintenanceAgency.keyFamily.dataSet</code>
 * is set which matches the information in the {@link MessageBegin}.
 * A wildcard character "<tt>*</tt>" is allowed in place of
 * any of the five components.</li>
 * <li>The value of the preference describing the flow can point to
 * a file containing a set of time series keys, one in each line.
 * Then data are accepted for those keys.
 * Alternatively, the value can be a single character "<tt>*</tt>",
 * in which case data for all keys are accepted.</li>
 * <li>If a message matches more than one data flow preference
 * (in the presence of wildcards), then the union of all key sets is used.</li>
 * <LI>Structural definitions have also data flows:
 * <CODE>sender.receiver.maintenanceAgency</CODE>.</LI>
 * </UL>
 * </P>
 *
 * <p>Example: Preference setting <tt>EE2.4F0.ECB.*.*=eekeys.txt</tt>
 * allows data for all keys in <tt>eekeys.txt</tt> from
 * sender <tt>EE2</tt> to <tt>4F0</tt> for all <tt>ECB</tt>'s key families,
 * regardless of the data set.</p>
 *
 * <center><table border="3" cellpadding="3">
 * <caption>
 * <a name="prefs">Available {@link Int.ecb.prefs preference} settings</a>
 * </caption>
 * <tr>
 * <th>preference name</th>
 * <th>type</th>
 * <th>meaning</th>
 * <th>default</th>
 * </tr>
 * <tr>
 * <td><i>sender</i>.<i>receiver</i>.<i>maintenanceAgency</i>.<i>keyFamily</i>.<i>dataSet</i></td>
 * <td>string</td><td>key list path or "*"</td>
 * </tr>
 * <tr><td><i>sender</i>.<i>receiver</i>.<i>maintenanceAgency</i></td>
 * <td>string</td><td>anything to make structural definitions accepted, "*" is suggested</td></tr>
 * </table></center>
 *
 * @see GESMESCBEventSieve
 *
 * @author
 * <A HREF="mailto:Sami.Airo@ECB.int">Sami Airo</A>
 * <a href="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</A>
 *
 * @version @(#)$Id: PreferenceConfiguredFilter.java,v 1.1 2017/01/19 16:14:04 trainot Exp $
 */
public class PreferenceConfiguredFilter extends GESMESCBObjectFilterAdapter {

	/* ========================= Fields ========================= */

	/**
	 * Set of expected topic and sibling keys for the current
	 * sender and key family.
	 */
	private Collection expectedKeys;

	/* ====================== Constructors ====================== */

	/**
	 * Constructs a PreferenceConfiguredFilter with the default name.
	 */
	public PreferenceConfiguredFilter () {
		this( "" );
	}

	/**
	 * Constructs a PreferenceConfiguredFilter with the specified name.
	 *
	 * @param name		name for this PreferenceConfiguredFilter
	 */
	public PreferenceConfiguredFilter ( final String name ) {
		super( name );
	}

	/* ===================== Public Methods ===================== */

	public boolean accept ( final MessageBegin messageBegin ) {
		if ( messageBegin.isStatisticalDefinitions() ) {
			final String flowName =
				"StructuralDefinitions" + "."
				+ messageBegin.getSender() + "."
				+ messageBegin.getReceiver() + "."
				+ messageBegin.getMaintenanceAgencyId();
			final String flowValue
			= componentHelper.getPreference( flowName, null );

			if ( flowValue == null ) {
				componentHelper.logger.log(
					Level.WARNING,
					componentHelper.logName,
					"accept",
					"MessageBegin rejected: sender "
					+ messageBegin.getSender()
					+ " is not expected to send structural definitions"
					+ " for maintenance agency "
					+ messageBegin.getMaintenanceAgencyId()
					+ " to receiver " + messageBegin.getReceiver(),
					messageBegin
				);
				return false;
			}

			if ( !"*".equals( flowValue ) ) {
				componentHelper.logger.log(
					Level.SEVERE,
					componentHelper.logName,
					"accept",
					"Error in configuration:"
					+ " Structural Definitions flow \"" + flowName
					+ "\" has acceptance setting \"" + flowValue
					+ "\" other than \"*\".",
					messageBegin
				);
				return true;
			}
		}

		if ( messageBegin.isStatisticalData() ) {
			final Collection filenames = new HashSet();
			for ( int i = 0; i < 32; ++i ) {
				final String flowName = "Data" + "."
				+ ( ( i & 16 ) != 0 ? "*" : messageBegin.getSender() ) + "."
				+ ( ( i & 8 ) != 0 ? "*" : messageBegin.getReceiver() ) + "."
				+ ( ( i & 4 ) != 0 ? "*" : messageBegin.getMaintenanceAgencyId() ) + "."
				+ ( ( i & 2 ) != 0 ? "*" : messageBegin.getKeyFamilyId() ) + "."
				+ ( ( i & 1 ) != 0 ? "*" : messageBegin.getDataSetId() );
				final String filename = componentHelper.getPreference(
					flowName,
					null
				);
				if ( filename != null ) {
					filenames.add( filename );
				} else if ( "*".equals( filename ) ) {
					componentHelper.logger.log(
						Level.CONFIG,
						componentHelper.logName,
						"accept",
						"All keys are accepted for the data flow " + flowName + ".",
						filename
					);
				}
			}
			
			if ( filenames.size() == 0 ) {
				Metabase metabase = Metabase.getMetabase();
				Level level=componentHelper.logger.getLevel();
				if (! metabase.containsMaintenanceAgency(
					messageBegin.getMaintenanceAgencyId()
					))
				{
					componentHelper.logger.setLevel(Level.INFO);
					componentHelper.logger.log(
						Level.WARNING,
						componentHelper.logName,
						"acceptGESMESCBObject",
						"MessageBegin refers to unknown agency:"
						+ "\"" + messageBegin.getMaintenanceAgencyId() + "\" "
						+ "for sender \"" + messageBegin.getSender() + "\", "
						+ "receiver \"" + messageBegin.getReceiver() + "\", "
						+ "key family \"" + messageBegin.getKeyFamilyId() + "\", "
						+ "and dataset \"" + messageBegin.getDataSetId() + "\".",
						messageBegin
					);
				} else {
					for( Iterator it = metabase.maintenanceAgencyIterator()
						; it.hasNext();
					) {
						MaintenanceAgency agency
						= (MaintenanceAgency) it.next();
						if(! agency.containsKeyFamily(
							messageBegin.getKeyFamilyId()
						)){
							componentHelper.logger.log(
								Level.WARNING,
								componentHelper.logName,
								"acceptGESMESCBObject",
								"MessageBegin refers to unknown key \""
								+ messageBegin.getKeyFamilyId() + "\" for "
								+ "mainteance agency "
								+ "\"" + messageBegin.getMaintenanceAgencyId() + "\", "
								+ "sender \"" + messageBegin.getSender() + "\", "
								+ "receiver \"" + messageBegin.getReceiver() + "\", "
								+ "and dataset \"" + messageBegin.getDataSetId() + "\".",
								messageBegin
							);		
						} else break;		
					}
				}
				componentHelper.logger.log(
					Level.WARNING,
					componentHelper.logName,
					"acceptGESMESCBObject",
					"MessageBegin rejected: sender "
					+ "\"" + messageBegin.getSender() + "\" "
					+ "is not expected to send data to receiver "
					+ "\""+ messageBegin.getReceiver() + "\" "
					+ "for \"" + messageBegin.getMaintenanceAgencyId() + "\"'s "
					+ "key family \"" + messageBegin.getKeyFamilyId() + "\" "
					+ "and dataset \"" + messageBegin.getDataSetId() + "\".",
					messageBegin
				);
				componentHelper.logger.setLevel(level);
				
				return false;
			} else if ( filenames.contains( "*" ) ) {
				expectedKeys = null;
			} else {
				expectedKeys = new HashSet();
				for ( final Iterator iterator = filenames.iterator(); iterator.hasNext(); ) {
					final File file = new File( (String) iterator.next() );
					try {
						final BufferedReader br = new BufferedReader(
							new FileReader( file )
						);
						String line;
						while ( ( line = br.readLine() ) != null ) {
							expectedKeys.add( line );
						}
						br.close();
					} catch ( final IOException e ) {
						componentHelper.logger.log(
							Level.SEVERE,
							componentHelper.logName,
							"accept",
							"Error in reading keys for flow from file "
							+ file.getAbsolutePath(),
							e
						);
					}
				}
			}
		}

		return super.accept( messageBegin );
	}

	public boolean accept ( final TimeSeries timeSeries ) {
		if ( expectedKeys != null ) {
			if ( !expectedKeys.contains( timeSeries.getKey() ) ) {
				componentHelper.logger.log(
					Level.WARNING,
					componentHelper.logName,
					"accept",
					"TimeSeries rejected: " + timeSeries.getKey()
					+ " is not expected in the current data flow",
					timeSeries
				);
				return false;
			}
		}

		return super.accept( timeSeries );
	}

	public boolean accept ( final Attribute attribute ) {
		if ( expectedKeys != null ) {
			// Ioana 29.11 - Accept attribute with empty key as dataset level attribute
			// according to the SDMX-EDI standard
			if ( !attribute.getKey().isEmpty() && !expectedKeys.contains( attribute.getKey() ) ) {
				componentHelper.logger.log(
					Level.WARNING,
					componentHelper.logName,
					"accept",
					"Attribute rejected: \"" + attribute.getName()
					+ "\" is not expected for key \""
					+ attribute.getKey()
					+ "\" in the current data flow",
					attribute
				);
				return false;
			}
		}

		return super.accept( attribute );
	}
}
