// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001-2004, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.sieve;

import java.util.Date;
import java.util.Properties;
import java.util.Collections;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;

import junit.framework.Test;
import junit.framework.TestSuite;

import Int.ecb.gesmescb.Contact;
import Int.ecb.gesmescb.InterchangeBegin;
import Int.ecb.gesmescb.InterchangeEnd;
import Int.ecb.gesmescb.KeyFamily;
import Int.ecb.gesmescb.KeyFamilyComponent;
import Int.ecb.gesmescb.MessageBegin;
import Int.ecb.gesmescb.MessageEnd;
import Int.ecb.gesmescb.TimeSeries;
import Int.ecb.gesmescb.Attribute;
import Int.ecb.gesmescb.metabase.Metabase;
import Int.ecb.logging.ConsoleHandler;
import Int.ecb.logging.Handler;
import Int.ecb.logging.LogManager;

/* ***************************** Class ****************************** */

/**
 * Tests the implementation of the class {@link PreferenceConfiguredFilter}.
 *
 * @author
 * <A HREF="mailto:Sami.Airo@ECB.int">Sami Airo</A>
 * <a href="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</A>
 *
 * @version @(#)$Id: PreferenceConfiguredFilterTest.java,v 1.1 2017/01/19 16:14:04 trainot Exp $
 */

public class PreferenceConfiguredFilterTest
	extends GESMESCBObjectFilterAdapterTest
{
	/* ========================= Fields ========================= */

	private File tempFile;
	private Metabase metabase;
	
	/* ====================== Constructors ====================== */

	/**
	 * Constructs a PreferenceConfiguredFilterTest with the specified name.
	 *
	 * @param name Test case name
	 */
	public PreferenceConfiguredFilterTest ( final String name ) {
		super( name );
	}

	/* ================== Protected Methods ===================== */

	/**
	 * Sets up the test fixture.
	 */
	protected void setUp () throws IOException {
		System.setProperty( ".Level", "FINE" );
		System.setProperty( "handlers", "Int.ecb.logging.ConsoleHandler" );
		System.setProperty(
			"Int.ecb.logging.ConsoleHandler.formatter",
			"Int.ecb.gesmescb.util.NormalLogFormatter"
		);

		tempFile = File.createTempFile( "expected", null );

		final FileWriter fileWriter = new FileWriter( tempFile );
		fileWriter.write( "ABC" );
		fileWriter.close();

		tempFile.deleteOnExit();

		final String className = PreferenceConfiguredFilter.class.getName();
		System.setProperty(
			className + ".StructuralDefinitions.4F0.ZZZ.ECB", "*"
		);
		System.setProperty(
			className + ".Data.AT2.4F0.ECB.ECB_OB1.ECB_OB1",
			tempFile.getAbsolutePath()
		);
		System.setProperty(
			className + ".Data.RO5.4F1.ECB.*.*",
			tempFile.getAbsolutePath()
		);
		System.setProperty(
			className + ".Data.CH5.4F1.ECB.*.*",
			"*"
		);
		
		metabase = createMetabase();
	
	}

	/**
	 * Tears down the test fixture.
	 */
	protected void tearDown () {
		final String className = PreferenceConfiguredFilter.class.getName();
		final Properties properties = System.getProperties();
		properties.remove( className + ".StructuralDefinitions.4F0.ZZZ.ECB" );
		properties.remove( className + ".Data.AT2.4F0.ECB.ECB_OB1.ECB_OB1" );
		properties.remove( className + ".Data.RO5.4F1.ECB.*.*" );
		properties.remove( className + ".Data.CH5.4F1.ECB.*.*" );
		tempFile.delete();
		tempFile = null;
		metabase.clear();
	}

	/* ===================== Public Methods ===================== */

	/**
	 * Assembles and returns a test suite for all
	 * the test methods of this test case.
	 *
	 * @return a non-null test suite.
	 */
	public static Test suite () {
		return new TestSuite( PreferenceConfiguredFilterTest.class );
	}

	/**
	 * Runs the test case.
	 */
	public static void main ( final String args[] ) {
		final String[] testCaseName = {
			PreferenceConfiguredFilterTest.class.getName()
		};
		junit.textui.TestRunner.main( testCaseName );
	}

	/* --------------------- Test Methods --------------------- */

	public void testAcceptMessageBegin () {
		
		final GESMESCBObjectFilter filter = new PreferenceConfiguredFilter();

		assertTrue( "Data Message OK", filter.accept( DATA_MESSAGE_BEGIN ) );

		final MessageBegin dataMessageBegin = new MessageBegin(
			"MREF0001", "FI2", "4F0", "ECB",
			MessageBegin.STATISTICAL_DATA,
			"Test message", Collections.EMPTY_SET,
			"ECB_BOP1",
			MessageBegin.UPDATE_OR_REPLACE_ACTION,
			"20010329:1423", "", "",
			"ECB_BOP1"
		);

		assertTrue( "Data Message NOK", !filter.accept( dataMessageBegin ) );

		assertTrue( "Structure Message OK", filter.accept( STRUCTURE_MESSAGE_BEGIN ) );

		final MessageBegin structureMessageBegin = new MessageBegin(
			"MREF0001", "GR2", "4F0", "BIS",
			MessageBegin.STATISTICAL_DEFINITIONS,
			"Test message", Collections.EMPTY_SET
		);

		assertTrue( "Structure Message NOK", ! filter.accept( structureMessageBegin ) );
	}

	public void testAcceptMessageBeginWildcarding () {
		final GESMESCBObjectFilter filter = new PreferenceConfiguredFilter();

		final MessageBegin BOPMessageBegin = new MessageBegin(
			"MREF0001", "RO2", "4F1", "ECB",
			MessageBegin.STATISTICAL_DATA,
			"Test message", Collections.EMPTY_SET,
			"ECB_BOP2",
			MessageBegin.UPDATE_OR_REPLACE_ACTION,
			"20010329:1423", "", "",
			"ECB_BOPT"
		);
		assertTrue( "BOP Message NOK", !filter.accept( BOPMessageBegin ) );

		final MessageBegin RO5MessageBegin = new MessageBegin(
			"MREF0001", "RO5", "4F1", "ECB",
			MessageBegin.STATISTICAL_DATA,
			"Test message", Collections.EMPTY_SET,
			"ECB_BOP2",
			MessageBegin.UPDATE_OR_REPLACE_ACTION,
			"20010329:1423", "", "",
			"ECB_BOPT"
		);
		assertTrue( "RO5 Message OK", filter.accept( RO5MessageBegin ) );

		final MessageBegin CH5MessageBegin = new MessageBegin(
			"MREF0001", "CH5", "4F1", "ECB",
			MessageBegin.STATISTICAL_DATA,
			"Test message", Collections.EMPTY_SET,
			"ECB_BOP2",
			MessageBegin.UPDATE_OR_REPLACE_ACTION,
			"20010329:1423", "", "",
			"ECB_BOPT"
		);
		assertTrue( "CH5 Message OK", filter.accept( CH5MessageBegin ) );
	}

	public void testAcceptStructure () {
		final GESMESCBObjectFilter filter = new PreferenceConfiguredFilter();
		acceptCodeList( filter );
		acceptStatisticalConcept( filter );
		acceptKeyFamily( filter );
	}

	public void testAcceptTimeSeries () {
		final GESMESCBObjectFilter filter = new PreferenceConfiguredFilter();

		assertTrue( "MessageBegin OK", filter.accept( DATA_MESSAGE_BEGIN ) );
		assertTrue(
			"TimeSeries NOK",
			!filter.accept( new TimeSeries(
				"ABCD", "", "", Collections.EMPTY_LIST
			) )
		);

		assertTrue( "MessageBegin OK", filter.accept( DATA_MESSAGE_BEGIN ) );
		assertTrue(
			"TimeSeries OK",
			filter.accept( new TimeSeries(
				"ABC", "", "", Collections.EMPTY_LIST
			) )
		);
	}

	public void testAcceptAttribute () {
		final GESMESCBObjectFilter filter = new PreferenceConfiguredFilter();

		assertTrue( "MessageBegin OK", filter.accept( DATA_MESSAGE_BEGIN ) );
		assertTrue(
			"Attribute NOK",
			!filter.accept( new Attribute(
				"ABCD", "", "", "", "P"
			) )
		);

		assertTrue( "MessageBegin OK", filter.accept( DATA_MESSAGE_BEGIN ) );
		assertTrue(
			"Attribute OK",
			filter.accept( new Attribute(
				"ABC", "", "", "", "P"
			) )
		);
	}

	private Metabase createMetabase() {
		Metabase metabase = Metabase.getMetabase();
		InterchangeBegin interchangeBegin = new InterchangeBegin(
			"IREF0001", "DE2", "4F0", "000315124533", false
		);
		InterchangeEnd interchangeEnd = new InterchangeEnd(
			"IREF0001"
		);
		MessageBegin messageBegin = new MessageBegin(
			"MREF0001", "4F0", "4F0", "ECB",
			MessageBegin.STATISTICAL_DEFINITIONS,
			"subject: definitions", new Contact[] {}
		);
		MessageEnd messageEnd = new MessageEnd(
			"MREF0001"
		);
		MessageBegin messageBeginBIS = new MessageBegin(
			"MREF0002", "5B0", "4F0", "ECB",
			MessageBegin.STATISTICAL_DEFINITIONS,
			"subject: definitions", new Contact[] {}
		);
		MessageEnd messageEndBIS = new MessageEnd(
			"MREF0002"
		);
		KeyFamily keyFamily = new KeyFamily(
			"ECB_BOP1", "name", "desc", new Date(), new KeyFamilyComponent[0]
		);
		metabase.handle( interchangeBegin );
		metabase.handle( messageBegin );
		metabase.handle( keyFamily );
		metabase.handle( messageEnd );
		metabase.handle( messageBeginBIS );
		metabase.handle( messageEndBIS );
		metabase.handle( interchangeEnd );
		
		return metabase;
	}
	
	public void testRefuseInvalidDatasets () {
					
		final GESMESCBObjectFilter filter = new PreferenceConfiguredFilter();
		// TODO: get the console initialised properly in the setup.
		LogManager logmaster = LogManager.getLogManager();
		Handler console = new ConsoleHandler();
		logmaster.addGlobalHandler(console);
		
		assertTrue( "Data Message OK", filter.accept( DATA_MESSAGE_BEGIN ) );

		final MessageBegin messageBegin1 = new MessageBegin(
				"MREF0001", "RO5", "4F1", "ECB",
				MessageBegin.STATISTICAL_DATA,
				"Test message", Collections.EMPTY_SET,
				"ECB_BOP2",
				MessageBegin.UPDATE_OR_REPLACE_ACTION,
				"20010329:1423", "", "",
				"ECB_BOPT"
			);

		assertTrue( "Data Message OK-BOPT", filter.accept( messageBegin1 ) );

		final MessageBegin messageBegin2 = new MessageBegin(
				"MREF0001", "FI2", "4F0", "ECB",
				MessageBegin.STATISTICAL_DATA,
				"Test message", Collections.EMPTY_SET,
				"ECB_UNKNW",
				MessageBegin.UPDATE_OR_REPLACE_ACTION,
				"20010329:1423", "", "",
				"ECB_UNKNW"
			);

		assertTrue( "Data Message NOK", !filter.accept( messageBegin2 ) );

		final MessageBegin messageBegin3 = new MessageBegin(
				"MREF0001", "FI2", "4F0", "BIS",
				MessageBegin.STATISTICAL_DATA,
				"Test message", Collections.EMPTY_SET,
				"ECB_BOP1",
				MessageBegin.UPDATE_OR_REPLACE_ACTION,
				"20010329:1423", "", "",
				"ECB_BOP1"
			);

		assertTrue( "Data Message NOK", !filter.accept( messageBegin3 ) );
		
		logmaster.removeGlobalHandler(console);
	}

}
