// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (C) 2006 European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.sql;

import java.util.Date;
import java.util.List;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.Map;
import java.util.SortedMap;
import java.util.HashMap;
import java.util.TreeMap;

import java.text.NumberFormat;
import java.text.DecimalFormat;
import java.text.SimpleDateFormat;

import java.sql.Connection;
import java.sql.Statement;
import java.sql.ResultSet;
import java.sql.SQLException;

import Int.ecb.logging.Level;

import Int.ecb.gesmescb.InterchangeBegin;
import Int.ecb.gesmescb.MessageBegin;
import Int.ecb.gesmescb.Contact;
import Int.ecb.gesmescb.TimeSeries;
import Int.ecb.gesmescb.Observation;
import Int.ecb.gesmescb.Attribute;
import Int.ecb.gesmescb.MessageEnd;
import Int.ecb.gesmescb.InterchangeEnd;

import Int.ecb.gesmescb.event.ErrorEvent;
import Int.ecb.gesmescb.event.GESMESCBEventSourceAdapter;
import Int.ecb.gesmescb.event.GESMESCBReader;

import Int.ecb.gesmescb.time.RegularInterval;
import Int.ecb.gesmescb.time.GregorianRegularInterval;

import Int.ecb.gesmescb.util.ComponentHelper;
import Int.ecb.gesmescb.util.Named;
import Int.ecb.gesmescb.util.Strings;

/* ***************************** Class ****************************** */

/**
 * <p>Reads from the Currency Information System (CIS) database
 * as {@link Int.ecb.gesmescb.GESMESCBObject}s.</p>
 *
 * <center><table border="3" cellpadding="3">
 * <caption>
 * <a name="prefs">Available {@link Int.ecb.prefs preference} settings</a>
 * </caption>
 * <tr>
 * <th>preference name</th>
 * <th>type</th>
 * <th>meaning</th>
 * <th>default</th>
 * </tr>
 * <tr>
 *	<td><code>iref</td><td>string</td>
 *	<td>interchange reference number</td>
 *	<td><code>IREF000000</code></td>
 * </tr>
 * <tr>
 *	<td><code>sender</td><td>string</td>
 *	<td>sender organisation code</td>
 *	<td><code>ZZZ</code></td>
 * </tr>
 * <tr>
 *	<td><code>receiver</td><td>string</td>
 *	<td>receiver organisation code</td>
 *	<td><code>ZZZ</code></td>
 * </tr>
 * <tr>
 *	<td><code>subject</td><td>string</td>
 *	<td>message title</td>
 * </tr>
 * <tr>
 *	<td><code>test</td><td>boolean</td>
 *	<td>whether to flag data as test data</td>
 *	<td><code>false</code></td>
 * </tr>
 * <tr>
 *	<td><code>maintenanceAgency</td><td>string</td>
 *	<td>the key family maintenance agency</td>
 *	<td><code>ECB</code></td>
 * </tr>
 * <tr>
 *	<td><code>keyFamily</td><td>string</td>
 *	<td>the identifier of the key family structuring the data</td>
 *	<td><code>ECB_BKN1</code></td>
 * </tr>
 * <tr>
 *	<td><code>dataSet</td><td>string</td>
 *	<td>the identifier of the data set</td>
 *	<td><code>ECB_BKN1</code></td>
 * </tr>
 * <tr>
 *	<td><code>source</td><td>string</td>
 *	<td>URL of the database from which to extract the data</td>
 * </tr>
 * <tr>
 *	<td><code>user</td><td>string</td>
 *	<td>the user id for accessing the database</td>
 * </tr>
 * <tr>
 *	<td><code>password</td><td>string</td>
 *	<td>the password for accessing the database</td>
 * </tr>
 * </table></center>
 *
 * @author
 * <a href="mailto:sis.external@ecb.int">Bernhard Bodenstorfer</A>
 *
 * @version @(#)$Id: CISReader.java,v 1.1 2017/01/19 16:14:02 trainot Exp $
 */
public class CISReader
	extends GESMESCBEventSourceAdapter
	implements GESMESCBReader, Named
{
	/* ======================= Constants ======================== */

	private static final Map
		BKN_CATEGORY_MAP = new HashMap();

	static {
		BKN_CATEGORY_MAP.put( "Issued banknotes",
			new CurrencyCategory( "IS10", "B", "F" ) );
		BKN_CATEGORY_MAP.put( "Returned banknotes",
			new CurrencyCategory( "RE10", "B", "F" ) );
		BKN_CATEGORY_MAP.put( "Sorted banknotes",
			new CurrencyCategory( "SR10", "B", "F" ) );
		BKN_CATEGORY_MAP.put( "Banknotes sorted to unfit",
			new CurrencyCategory( "SR11", "B", "F" ) );
		BKN_CATEGORY_MAP.put( "Issued coins",
			new CurrencyCategory( "IS10", "C", "F" ) );
		BKN_CATEGORY_MAP.put( "Returned coins",
			new CurrencyCategory( "RE10", "C", "F" ) );
		BKN_CATEGORY_MAP.put( "Collector coins",
			new CurrencyCategory( "CL10", "C", "S" ) );
		BKN_CATEGORY_MAP.put( "Collector " + " coins",	// CIS db contains this double space!
				BKN_CATEGORY_MAP.get( "Collector coins" ) );
		BKN_CATEGORY_MAP.put( "Net circulation of euro coins",
			new CurrencyCategory( "NC10", "C", "S" ) );
		BKN_CATEGORY_MAP.put( "Number of branches",
			new NonCurrencyCategory( "A030", "Z", "S" ) );
		BKN_CATEGORY_MAP.put( "Number of branches of commercial banks",
			new NonCurrencyCategory( "A020", "Z", "S" ) );
		BKN_CATEGORY_MAP.put( "Number of ATM",
			new NonCurrencyCategory( "A010", "Z", "S" ) );
		BKN_CATEGORY_MAP.put( "Created banknotes",
			SuppressedCategory.SUPPRESSED_CATEGORY );
		BKN_CATEGORY_MAP.put( "Logistical stocks of new banknotes",
			SuppressedCategory.SUPPRESSED_CATEGORY );
		BKN_CATEGORY_MAP.put( "Logistical stocks of reissuable banknotes",
			SuppressedCategory.SUPPRESSED_CATEGORY );
		BKN_CATEGORY_MAP.put( "Stocks of unfit banknotes",
			SuppressedCategory.SUPPRESSED_CATEGORY );
		BKN_CATEGORY_MAP.put( "Stocks of banknotes to be sorted",
			SuppressedCategory.SUPPRESSED_CATEGORY );
		BKN_CATEGORY_MAP.put( "Banknotes held to order",
			SuppressedCategory.SUPPRESSED_CATEGORY );
		BKN_CATEGORY_MAP.put( "Banknotes in transit",
			SuppressedCategory.SUPPRESSED_CATEGORY );
		BKN_CATEGORY_MAP.put( "Eurosystem strategic reserve of new banknotes",
			SuppressedCategory.SUPPRESSED_CATEGORY );
		BKN_CATEGORY_MAP.put( "Eurosystem strategic reserve of reissuable banknotes",
			SuppressedCategory.SUPPRESSED_CATEGORY );
		BKN_CATEGORY_MAP.put( "Destroyed banknotes",
			SuppressedCategory.SUPPRESSED_CATEGORY );
		CurrencyCategory netCirculationOfEuroBanknotesCategory
			= new CurrencyCategory( "NC10", "B", "S" ) ;
		netCirculationOfEuroBanknotesCategory.isExtractable = false;
		BKN_CATEGORY_MAP.put( "Net circulation of euro banknotes",
			netCirculationOfEuroBanknotesCategory );
	}

	private static final String QUERY;
	private static final String QUERY_NC10_B;

	static {
		final StringBuffer buffer = new StringBuffer();
		boolean isFirst = true;
		for (
			final Iterator iterator = BKN_CATEGORY_MAP.entrySet().iterator();
			iterator.hasNext();
		) {
			final Map.Entry entry = (Map.Entry) iterator.next();
			final String key = (String) entry.getKey();
			final BknCategory category = (BknCategory) entry.getValue();
			if ( category.isExtractable() ) {
				if ( isFirst ) {
					isFirst = false;
				} else {
					buffer.append( ", " );
				}
				buffer.append( '\'' );
				buffer.append( key );
				buffer.append( '\'' );
			}
		}

		QUERY = "SELECT"
			+ " cis.v_i_cb.country_iso_code,"
			+ " cis.v_i_category.name,"
			+ " cis.v_i_denomination.amount,"
			+ " cis.v_i_period_type.description,"
			+ " cis.v_i_fact.start_date,"
			+ " cis.v_i_fact.value"
			+ " FROM cis.v_i_cb, cis.v_i_category, cis.v_i_denomination, cis.v_i_fact, cis.v_i_period_type"
			+ " WHERE cis.v_i_cb.cb_id=cis.v_i_fact.cb_from"
			+ " AND cis.v_i_category.category_id=cis.v_i_fact.category_id"
			+ " AND cis.v_i_denomination.denomination_id=cis.v_i_fact.denomination_id"
			+ " AND cis.v_i_period_type.period_type_id=cis.v_i_fact.period_type_id"
			+ " AND cis.v_i_category.name IN ( " + buffer.toString() + " )"
			+ " ORDER BY cis.v_i_cb.cb_id ASC,"
			+ " cis.v_i_category.category_id ASC,"
			+ " cis.v_i_denomination.denomination_id ASC,"
			+ " start_date ASC";

		QUERY_NC10_B = ""
			+ " SELECT cb.country_iso_code"
			+ "      , 'Net circulation of euro banknotes' as name"
			+ "      , denomination.amount"
			+ "      , period_type.description"
			+ "      , start_date, value"
			+ "   FROM ("
			+ "        SELECT cb_from"
			+ "             , 50 as category_id"
			+ "             , start_date"
			+ "             , denomination_id"
			+ "             , sum(value) as value"
			+ "             , period_type_id"
			+ "          FROM ("
			+ "               SELECT cb_from"
			+ "                    , category_id"
			+ "                    , start_date"
			+ "                    , denomination_id"
			+ "                    , sum(value) as value"
			+ "                    , period_type_id"
			+ "                 FROM cis.v_i_fact"
			+ "                WHERE category_id = 1"
			+ "                  AND denomination_id IN (1,2,3,4,5,6,7)"
			+ "                GROUP BY cb_from, category_id, start_date"
			+ "                    , denomination_id, period_type_id"
			+ "                UNION"
			+ "               SELECT cb_from"
			+ "                    , category_id"
			+ "                    , start_date"
			+ "                    , denomination_id"
			+ "                    , -sum(value) as value"
			+ "                    , period_type_id"
			+ "                 FROM cis.v_i_fact"
			+ "                WHERE category_id IN (2,3,4,5,6,7,8,9,14)"
			+ "                  AND denomination_id IN (1,2,3,4,5,6,7)"
			+ "                GROUP BY cb_from, category_id, start_date"
			+ "                    , denomination_id, period_type_id"
			+ "             ) producedAndRemovedBanknotes"
			+ "         GROUP BY cb_from"
			+ "             , start_date"
			+ "             , denomination_id"
			+ "             , period_type_id"
			+ "      ) facts"
			+ "      , cis.v_i_denomination denomination"
			+ "      , cis.v_i_period_type period_type"
			+ "      , cis.v_i_cb cb"
			+ "  WHERE facts.denomination_id in (1,2,3,4,5,6,7)"
			+ "    AND facts.cb_from in (1000)"
			+ "    AND facts.denomination_id=denomination.denomination_id"
			+ "    AND facts.period_type_id=period_type.period_type_id"
			+ "    AND facts.cb_from=cb.cb_id"
			+ " ORDER BY cb.country_iso_code ASC,"
			+ " denomination.amount DESC,"
			+ " start_date ASC"
			+ "";
	}

	/* ========================= Fields ========================= */

	/** Takes care of logging and preferences. */
	protected final ComponentHelper componentHelper;

	/** Takes care of database connectivity. */
	private final SQLConnector connector;

	/* ====================== Constructors ====================== */

	/**
	 * Construct a <code>CISReader</code> to read from a database.
	 *
	 * @param name used for logging and preferences.
	 * @param database URL of the database where to read from.
	 *
	 * @throws NullPointerException if <code>database==null</code>
	 * and the preference variable
	 * <code>source</code> is undefined.
	 * @throws SQLException if the database cannot be opened for reading.
	 */
	public CISReader ( final String name, final String database )
		throws SQLException
	{
		componentHelper = new ComponentHelper( name, this );
		connector = new SQLConnector( componentHelper, database, false );
	}

	/**
	 * Construct a <code>CISReader</code>
	 * reading from a database according to preferences.
	 *
	 * @param name used for logging and preferences.
	 *
	 * @throws NullPointerException if the preference variable
	 * <code>source</code> is undefined.
	 * @throws SQLException if the database cannot be opened for reading.
	 */
	public CISReader ( final String name )
		throws SQLException
	{
		this( name, null );
	}

	/* ===================== Public Methods ===================== */

	public String getName () {
		return componentHelper.name;
	}

	/**
	 * Extract from the CIS type JDBC database the data as
	 * {@link Int.ecb.gesmescb.GESMESCBObject}s.
	 * These then are dispatched to all listeners.
	 */
	public void run () {
		try {
			componentHelper.logger.log(
				Level.INFO,
				componentHelper.logName,
				"run()",
				"Starting to extract CIS data from \""
				+ connector.getDatabaseURL()
				+ "\"."
			);

			final Date now = new Date();
			final String iref
			= componentHelper.getPreference( "iref", "IREF000000" );
			final String sender
			= componentHelper.getPreference( "sender", "ZZZ" );
			final String receiver
			= componentHelper.getPreference( "receiver", "ZZZ" );
			final String prepared
			= new SimpleDateFormat( "yyMMdd:HHmm" ).format( now );
			final boolean test
			= ! "false".equalsIgnoreCase(
				componentHelper.getPreference( "test", "false" )
			);
			final InterchangeBegin interchangeBegin = new InterchangeBegin(
				iref, sender, receiver, prepared, test
			);
			componentHelper.logger.log(
				Level.FINE,
				componentHelper.logName,
				"run()",
				"Dispatching InterchangeBegin",
				interchangeBegin
			);
			dispatch( interchangeBegin );

			final String maintenanceAgencyId
			= componentHelper.getPreference( "maintenanceAgency", "ECB" );
			final String keyFamilyId
			= componentHelper.getPreference( "keyFamily", "ECB_BKN1" );
			final String subject
			= componentHelper.getPreference( "subject", "" );
			final String dataSetId
			= componentHelper.getPreference( "dataSet", "ECB_BKN1" );
			final String extracted
			= new SimpleDateFormat( "yyyyMMddHHmm" ).format( now );
			final MessageBegin messageBegin = new MessageBegin(
				"MREF000001", sender, receiver, maintenanceAgencyId,
				MessageBegin.STATISTICAL_DATA,
				subject,
				new Contact[]{},
				dataSetId,
				MessageBegin.UPDATE_OR_REPLACE_ACTION,
				extracted,
				"", "",
				keyFamilyId
			);
			componentHelper.logger.log(
				Level.FINE,
				componentHelper.logName,
				"run()",
				"Dispatching MessageBegin",
				messageBegin
			);
			dispatch( messageBegin );

			componentHelper.logger.log(
				Level.FINER,
				componentHelper.logName,
				"run()",
				"Get connection to \""
				+ connector.getDatabaseURL()
				+ "\"."
			);
			final Connection connection = connector.getConnection();			
			runQuery( connection, QUERY );
			runQuery( connection, QUERY_NC10_B );

			final MessageEnd messageEnd = new MessageEnd( "MREF000001" );
			componentHelper.logger.log(
				Level.FINE,
				componentHelper.logName,
				"run()",
				"Dispatching MessageEnd",
				messageEnd
			);
			dispatch( new MessageEnd( "MREF000001" ) );

			final InterchangeEnd interchangeEnd = new InterchangeEnd(
				iref
			);
			componentHelper.logger.log(
				Level.FINE,
				componentHelper.logName,
				"run()",
				"Dispatching InterchangeEnd",
				interchangeEnd
			);
			dispatch( interchangeEnd );

			componentHelper.logger.log(
				Level.INFO,
				componentHelper.logName,
				"run()",
				"Ending to extract CIS data from \""
				+ connector.getDatabaseURL()
				+ "\"."
			);
		}
		catch ( final Exception e ) {
			final String desc = "Problem: " + Strings.stackTrace( e );

			componentHelper.logger.log(
				Level.SEVERE,
				getName(),
				"run()",
				desc,
				e
			);
			dispatch(
				new ErrorEvent( desc, ErrorEvent.ERROR )
			);
		}
	}

	/** Close the <code>CISReader</code>. */
	public void close () {
		connector.close();
	}

	/* ==================== Private Methods ===================== */

	private void runQuery (
		final Connection connection,
		final String query
	) throws Exception {
		try {		
			final Statement statement = connection.createStatement();
			componentHelper.logger.log(
				Level.FINER,
				componentHelper.logName,
				"runQuery()",
				"Execute query \"" + query + "\"."
			);
			final ResultSet resultSet
			= statement.executeQuery( query );
			String oldGroupKey = "", oldSeriesKey = "";
			int oldFrequency = Integer.MIN_VALUE;
			final SortedMap observationsMap = new TreeMap();
			while ( resultSet.next() ) {
				final String
					country = resultSet.getString( 1 ),
					category = resultSet.getString( 2 ),
					periodType = resultSet.getString( 4 ),
					value = resultSet.getString( 6 );
				final double amount = resultSet.getDouble( 3 );
				final Date startDate = resultSet.getTimestamp( 5 );
				final BknCategory bknCategory = bknCategoryMap( category );
				if ( null == bknCategory ) {
					final String desc
					= "Skipping data concerning unknown category \""
						+ category + "\".";
					componentHelper.logger.log(
						Level.WARNING,
						getName(),
						"runQuery()",
						desc,
						resultSet
					);
					dispatch(
						new ErrorEvent( desc, ErrorEvent.WARNING )
					);
					continue;
				}
				final String
					freq = freqMap( periodType ),
					refArea = refAreaMap( country ),
					bknItem = bknCategory.getBknItem(),
					bknType = bknCategory.getBknType(),
					bknDenom = bknCategory.getBknDenom( amount ),
					bknSeries = bknCategory.getBknSeries(),
					dataTypeBkn = bknCategory.getDataTypeBkn(),
					seriesDenom = bknCategory.getSeriesDenom( amount );
				final String groupKey = TimeSeries.KEY_SEPARATOR
					+ refArea + TimeSeries.KEY_SEPARATOR
					+ bknItem + TimeSeries.KEY_SEPARATOR
					+ bknType + TimeSeries.KEY_SEPARATOR
					+ bknDenom + TimeSeries.KEY_SEPARATOR
					+ bknSeries + TimeSeries.KEY_SEPARATOR
					+ dataTypeBkn + TimeSeries.KEY_SEPARATOR
					+ seriesDenom;
				final String seriesKey = freq + groupKey;
				// Finish old, cached series
				if ( ! seriesKey.equals( oldSeriesKey ) ) {
					dispatchSeries(
						oldSeriesKey, oldFrequency, observationsMap
					);
					observationsMap.clear();
					oldFrequency = frequencyMap( periodType );
				}
				observationsMap.put( startDate, value );
				// Start new group
				if ( ! groupKey.equals( oldGroupKey ) ) {
					final String
						unit = bknCategory.getUnit( amount ),
						unitMult = "0",
						decimals = bknCategory.getDecimals( amount );
					dispatch( new Attribute(
						groupKey, "", "", "UNIT", unit
					) );
					dispatch( new Attribute(
						groupKey, "", "", "UNIT_MULT", unitMult
					) );
					dispatch( new Attribute(
						groupKey, "", "", "DECIMALS", decimals
					) );
					oldGroupKey = groupKey;
				}
				// Start new series
				if ( ! seriesKey.equals( oldSeriesKey ) ) {
					final String collection = "E";
					dispatch( new Attribute(
						seriesKey, "", "", "COLLECTION", collection
					) );
					oldSeriesKey = seriesKey;
				}
			}
			statement.close();
			dispatchSeries( oldSeriesKey, oldFrequency, observationsMap );
			resultSet.close();
		} catch ( Exception e ) {
			throw e;
		}
	}

	private String freqMap ( final String periodType ) {
		if ( "MONTH".equals( periodType ) ) {
			return "M";
		} else {
			componentHelper.logger.log(
				Level.SEVERE,
				getName(),
				"freqMap()",
				"Unknown period type \"" + periodType + "\"."
			);
			return "?";
		}
	}

	private String refAreaMap ( final String country ) {
		if ( "EU".equals( country ) ) {
			return "4F";
		}
		return country;
	}

	private BknCategory bknCategoryMap ( final String category ) {
		final BknCategory bknCategory = (BknCategory) BKN_CATEGORY_MAP.get( category );
		if ( null == bknCategory ) {
			componentHelper.logger.log(
				Level.SEVERE,
				getName(),
				"bknCategoryMap()",
				"Unknown category \"" + category + "\"."
			);
			return null;
		}
		return bknCategory;
	}

	private int frequencyMap ( final String periodType ) {
		if ( "MONTH".equals( periodType ) ) {
			return RegularInterval.MONTHLY;
		}
		componentHelper.logger.log(
			Level.SEVERE,
			getName(),
			"frequencyMap()",
			"Unknown period type \"" + periodType + "\"."
		);
		return Integer.MIN_VALUE;
	}

	private void dispatchSeries (
		final String seriesKey,
		final int frequency,
		final SortedMap observationsMap
	) {
		if ( observationsMap.isEmpty() ) return;

		final Date begin = (Date) observationsMap.firstKey();
		final Date end = (Date) observationsMap.lastKey();
		final RegularInterval interval = new GregorianRegularInterval(
			frequency, begin,
			new GregorianRegularInterval( frequency, begin, end ).getSize() + 1
		);

		final Iterator observations = observationsMap.entrySet().iterator();
		Map.Entry obsPair = (Map.Entry) observations.next();

		final List observationsList = new ArrayList( observationsMap.size() );

		for (
			final Iterator periods = interval.iterator();
			periods.hasNext();
		) {
			final RegularInterval period = (RegularInterval) periods.next();
			final int comparison = period.getBegin().compareTo( (Date) obsPair.getKey() );
			if ( comparison < 0 ) {
				observationsList.add( Observation.EMPTY_OBSERVATION );
			} else if ( comparison == 0 ) {
				observationsList.add( new Observation( (String) obsPair.getValue(), "A", "F", "" ) );
				obsPair = observations.hasNext() ? (Map.Entry) observations.next() : null;
			} else {
				componentHelper.logger.log(
					Level.SEVERE,
					getName(),
					"dispatchSeries()",
					"Database contains misaligned or multiple observation at "
					+ obsPair.getKey() + " for time slot " + period + "."
				);
			}
		}

		if ( obsPair != null ) {
			componentHelper.logger.log(
				Level.SEVERE,
				getName(),
				"dispatchSeries()",
				"Implementation error at time slot assignment of "
				+ obsPair.getKey() + "."
			);
		}

		dispatch( new TimeSeries(
			seriesKey, interval, observationsList
		) );
	}
}

/* *************************** Interface **************************** */

/** Holds the information corresponding to a category in the CIS database. */
interface BknCategory {
	/** Whether to extract the corresponding data. */
	boolean isExtractable ();
	String getBknItem ();
	String getBknType ();
	String getBknDenom ( double amount );
	String getBknSeries ();
	String getDataTypeBkn ();
	String getSeriesDenom ( double amount );
	String getUnit ( double amount );
	String getDecimals ( double amount );
}

/* ***************************** Class ****************************** */

abstract class AbstractBknCategory implements BknCategory {
	final String bknItem, bknType, dataTypeBkn;

	public AbstractBknCategory (
		final String bknItem,
		final String bknType,
		final String dataTypeBkn
	) {
		this.bknItem = bknItem;
		this.bknType = bknType;
		this.dataTypeBkn = dataTypeBkn;
	}

	public boolean isExtractable () { return true; }
	public String getBknItem () { return bknItem; }
	public String getBknType () { return bknType; }
	public String getDataTypeBkn () { return dataTypeBkn; }
}

/* ***************************** Class ****************************** */

/** Translates amount information into bkn denomination codes for banknotes and coins. */
class CurrencyCategory extends AbstractBknCategory {
	private static final NumberFormat AMOUNT_FORMAT = new DecimalFormat( "0.0E0");
	protected final ComponentHelper componentHelper;
	public boolean isExtractable;

	public CurrencyCategory (
		final String bknItem,
		final String bknType,
		final String dataTypeBkn
	) {
		super( bknItem, bknType, dataTypeBkn );
		isExtractable = true;
		componentHelper = new ComponentHelper( this.getClass().getName(), this );
	}

	public String getBknDenom( final double amount ) {
		if ( 0 == amount ) return "ALLD";
		final String amountString;
		synchronized ( AMOUNT_FORMAT ) {
			amountString = AMOUNT_FORMAT.format( amount );
		}
		if (
				'.' != amountString.charAt( 1 )
				||
				'E' != amountString.charAt( 3 )
		) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.name,
				"bknDenomMap()",
				"Unknown denomination amount \"" + amount + "\"."
			);
		}
		final StringBuffer denomination = new StringBuffer();
		denomination.append( amountString.charAt( 0 ) );
		denomination.append( amountString.charAt( 2 ) );
		if ( '-' == amountString.charAt( 4 ) ) {
			denomination.append( 'N' );
			denomination.append( amountString.substring( 5 ) );
		} else {
			denomination.append( 'P' );
			denomination.append( amountString.substring( 4 ) );
		}
		return denomination.toString();
	}

	public String getBknSeries () {
		return "01";
	}

	public String getSeriesDenom ( final double amount ) {
		return 0 == amount ? "E" : "Q";
	}

	public String getUnit ( final double amount ) {
		return 0 == amount ? "EUR" : "PURE_NUMB";
	}

	public String getDecimals ( final double amount ) {
		return 0 == amount ? "2" : "0";
	}

	public boolean isExtractable () {
		return isExtractable;
	}
}

/* ***************************** Class ****************************** */

/** Translates amount information into bkn denomination codes for non-currency items (e.g. ATMs). */
class NonCurrencyCategory extends AbstractBknCategory {

	public NonCurrencyCategory (
		final String bknItem,
		final String bknType,
		final String dataTypeBkn
	) {
		super( bknItem, bknType, dataTypeBkn );
	}

	public String getBknDenom ( final double amount ) {
		return "ZZZZ";
	}

	public String getBknSeries () {
		return "ZZ";
	}

	public String getSeriesDenom ( final double amount ) {
		return "Q";
	}

	public String getUnit ( final double amount ) {
		return "PURE_NUMB";
	}

	public String getDecimals ( final double amount ) {
		return "0";
	}
}

/* ***************************** Class ****************************** */

class SuppressedCategory implements BknCategory {
	public static final BknCategory SUPPRESSED_CATEGORY
		= new SuppressedCategory();

	private SuppressedCategory () {}

	public boolean isExtractable () { return false; }

	public String getBknItem () { return null; }
	public String getBknType () { return null; }
	public String getBknDenom ( double amount ) { return null; }
	public String getBknSeries () { return null; }
	public String getDataTypeBkn () { return null; }
	public String getSeriesDenom ( double amount ) { return null; }
	public String getUnit ( double amount ) { return null; }
	public String getDecimals ( double amount ) { return null; }
}
