// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (C) 2006, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.sql;

import java.util.Properties;

import java.sql.Connection;
import java.sql.SQLException;

import Int.ecb.gesmescb.InterchangeBegin;
import Int.ecb.gesmescb.MessageBegin;
import Int.ecb.gesmescb.Contact;
import Int.ecb.gesmescb.TimeSeries;
import Int.ecb.gesmescb.Observation;
import Int.ecb.gesmescb.Attribute;
import Int.ecb.gesmescb.MessageEnd;
import Int.ecb.gesmescb.InterchangeEnd;

import Int.ecb.gesmescb.event.TestEventListener;
import Int.ecb.gesmescb.time.Interval;

import junit.framework.Assert;
import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;

/* ***************************** Class ****************************** */

/**
 * <P>Tests the implementation of the class {@link CISReader}.</P>
 *
 * @author
 * <a href="mailto:sis.external@ecb.int">Bernhard Bodenstorfer</A>
 *
 * @version @(#)$Id: CISReaderTest.java,v 1.1 2017/01/19 16:14:02 trainot Exp $
 */

public class CISReaderTest extends TestCase {

	/* ======================= Constants ======================== */

	/** The input to the {@link CISReader} under test. */
	private static final String[]
		testPreparation = {
			"CREATE SCHEMA cis",
			"SET SCHEMA cis",
			"CREATE TABLE v_i_category ("
			+ " category_id NUMERIC(9),"
			+ " name VARCHAR(100)"
			+ " )",
			"CREATE TABLE v_i_cb ("
			+ " cb_id NUMERIC(9),"
			+ " country_iso_code CHAR(2),"
			+ " name VARCHAR(40)"
			+ " )",
			"CREATE TABLE v_i_denomination ("
			+ " denomination_id NUMERIC(9),"
			+ " description VARCHAR(20),"
			+ " amount NUMERIC(9,2)"
			+ " )",
			"CREATE TABLE v_i_period ("
			+ " period_type_id NUMERIC(9),"
			+ " start_date DATE"
			+ " )",
			"CREATE TABLE v_i_period_type ("
			+ " period_type_id NUMERIC(9),"
			+ " description VARCHAR(20)"
			+ " )",
			"CREATE TABLE v_i_fact ("
			+ " cb_from NUMERIC(9),"
			+ " category_id NUMERIC(9),"
			+ " start_date DATE,"
			+ " denomination_id NUMERIC(9),"
			+ " value NUMERIC(21),"
			+ " period_type_id NUMERIC(9)"
			+ " )",
			"INSERT INTO v_i_category VALUES ( 1, 'Created banknotes' )",
			"INSERT INTO v_i_category VALUES ( 2, 'Logistical stocks of new banknotes' )",
			"INSERT INTO v_i_category VALUES ( 3, 'Logistical stocks of reissuable banknotes' )",
			"INSERT INTO v_i_category VALUES ( 4, 'Stocks of unfit banknotes' )",
			"INSERT INTO v_i_category VALUES ( 5, 'Stocks of banknotes to be sorted' )",
			"INSERT INTO v_i_category VALUES ( 6, 'Banknotes held to order' )",
			"INSERT INTO v_i_category VALUES ( 7, 'Banknotes in transit' )",
			"INSERT INTO v_i_category VALUES ( 8, 'Eurosystem strategic reserve of new banknotes' )",
			"INSERT INTO v_i_category VALUES ( 9, 'Eurosystem strategic reserve of reissuable banknotes' )",
			"INSERT INTO v_i_category VALUES ( 10, 'Issued banknotes' )",
			"INSERT INTO v_i_category VALUES ( 11, 'Returned banknotes' )",
			"INSERT INTO v_i_category VALUES ( 12, 'Sorted banknotes' )",
			"INSERT INTO v_i_category VALUES ( 13, 'Banknotes sorted to unfit' )",
			"INSERT INTO v_i_category VALUES ( 14, 'Destroyed banknotes' )",
			"INSERT INTO v_i_category VALUES ( 32, 'Issued coins' )",
			"INSERT INTO v_i_category VALUES ( 33, 'Returned coins' )",
			"INSERT INTO v_i_category VALUES ( 35, 'Collector " + " coins' )",
			"INSERT INTO v_i_category VALUES ( 50, 'Net circulation of euro coins' )",
			"INSERT INTO v_i_category VALUES ( 51, 'Number of branches' )",
			"INSERT INTO v_i_category VALUES ( 55, 'Number of branches of commercial banks' )",
			"INSERT INTO v_i_category VALUES ( 56, 'Number of ATM' )",
			"INSERT INTO v_i_cb VALUES ( 1, 'EU', 'European Central Bank' )",
			"INSERT INTO v_i_cb VALUES ( 2, 'AT', 'Austria' )",
			"INSERT INTO v_i_cb VALUES ( 3, 'LU', 'Luxembourg' )",
			"INSERT INTO v_i_cb VALUES ( 4, 'FR', 'France' )",
			"INSERT INTO v_i_cb VALUES ( 5, 'DE', 'Germany' )",
			"INSERT INTO v_i_cb VALUES ( 6, 'IT', 'Italy' )",
			"INSERT INTO v_i_cb VALUES ( 7, 'ES', 'Spain' )",
			"INSERT INTO v_i_cb VALUES ( 8, 'NL', 'Netherlands' )",
			"INSERT INTO v_i_cb VALUES ( 9, 'BE', 'Belgium' )",
			"INSERT INTO v_i_cb VALUES ( 10, 'IE', 'Ireland' )",
			"INSERT INTO v_i_cb VALUES ( 11, 'FI', 'Finland' )",
			"INSERT INTO v_i_cb VALUES ( 12, 'PT', 'Portugal' )",
			"INSERT INTO v_i_cb VALUES ( 13, 'GR', 'Greece' )",
			"INSERT INTO v_i_cb VALUES ( 14, 'CZ', 'Czech Republic' )",
			"INSERT INTO v_i_cb VALUES ( 15, 'EE', 'Estonia' )",
			"INSERT INTO v_i_cb VALUES ( 16, 'CY', 'Cyprus' )",
			"INSERT INTO v_i_cb VALUES ( 17, 'LV', 'Latvia' )",
			"INSERT INTO v_i_cb VALUES ( 18, 'LT', 'Lithuania' )",
			"INSERT INTO v_i_cb VALUES ( 19, 'HU', 'Hungary' )",
			"INSERT INTO v_i_cb VALUES ( 20, 'MT', 'Malta' )",
			"INSERT INTO v_i_cb VALUES ( 21, 'PL', 'Poland' )",
			"INSERT INTO v_i_cb VALUES ( 22, 'SI', 'Slovenia' )",
			"INSERT INTO v_i_cb VALUES ( 23, 'SK', 'Slovakia' )",
			"INSERT INTO v_i_cb VALUES ( 1000, 'U2', 'Euro area' )",
			"INSERT INTO v_i_denomination VALUES ( 0, 'NULL denomination', 0 )",
			"INSERT INTO v_i_denomination VALUES ( 1, 'EUR 500', 500 )",
			"INSERT INTO v_i_denomination VALUES ( 2, 'EUR 200', 200 )",
			"INSERT INTO v_i_denomination VALUES ( 3, 'EUR 100', 100 )",
			"INSERT INTO v_i_denomination VALUES ( 4, 'EUR 50', 50 )",
			"INSERT INTO v_i_denomination VALUES ( 5, 'EUR 20', 20 )",
			"INSERT INTO v_i_denomination VALUES ( 6, 'EUR 10', 10 )",
			"INSERT INTO v_i_denomination VALUES ( 7, 'EUR 5', 5 )",
			"INSERT INTO v_i_denomination VALUES ( 8, 'EUR 2', 2 )",
			"INSERT INTO v_i_denomination VALUES ( 9, 'EUR 1', 1 )",
			"INSERT INTO v_i_denomination VALUES ( 10, 'EUR 0.50', 0.5 )",
			"INSERT INTO v_i_denomination VALUES ( 11, 'EUR 0.20', 0.2 )",
			"INSERT INTO v_i_denomination VALUES ( 12, 'EUR 0.10', 0.1 )",
			"INSERT INTO v_i_denomination VALUES ( 13, 'EUR 0.05', 0.05 )",
			"INSERT INTO v_i_denomination VALUES ( 14, 'EUR 0.02', 0.02 )",
			"INSERT INTO v_i_denomination VALUES ( 15, 'EUR 0.01', 0.01 )",
			"INSERT INTO v_i_period_type VALUES ( 1, 'MONTH' )",
			"INSERT INTO v_i_period VALUES ( 1, DATE( '2002-01-01' ) )",
			"INSERT INTO v_i_period VALUES ( 1, DATE( '2002-01-02' ) )",
			"INSERT INTO v_i_period VALUES ( 1, DATE( '2002-01-03' ) )",
			"INSERT INTO v_i_period VALUES ( 1, DATE( '2002-01-04' ) )",
			"INSERT INTO v_i_period VALUES ( 1, DATE( '2002-01-05' ) )",
			"INSERT INTO v_i_period VALUES ( 1, DATE( '2002-01-06' ) )",
			"INSERT INTO v_i_period VALUES ( 1, DATE( '2002-01-07' ) )",
			"INSERT INTO v_i_period VALUES ( 1, DATE( '2002-01-08' ) )",
			"INSERT INTO v_i_period VALUES ( 1, DATE( '2002-01-09' ) )",
			"INSERT INTO v_i_period VALUES ( 1, DATE( '2002-01-10' ) )",
			"INSERT INTO v_i_period VALUES ( 1, DATE( '2002-01-11' ) )",
			"INSERT INTO v_i_period VALUES ( 1, DATE( '2002-01-12' ) )",
			"INSERT INTO v_i_period VALUES ( 1, DATE( '2002-01-13' ) )",
			"INSERT INTO v_i_period VALUES ( 1, DATE( '2002-01-14' ) )",
			"INSERT INTO v_i_period VALUES ( 1, DATE( '2002-01-15' ) )",
			"INSERT INTO v_i_fact VALUES ( 1, 10, DATE( '2002-01-01' ), 9, 1235, 1 )",
			"INSERT INTO v_i_fact VALUES ( 1, 10, DATE( '2002-02-01' ), 9, 1236, 1 )",
			"INSERT INTO v_i_fact VALUES ( 1, 10, DATE( '2002-04-01' ), 9, 1238, 1 )",
			"INSERT INTO v_i_fact VALUES ( 2, 11, DATE( '2002-01-01' ), 8, 1358, 1 )",
			"INSERT INTO v_i_fact VALUES ( 3, 12, DATE( '2002-02-01' ), 1, 3121, 1 )",
			"INSERT INTO v_i_fact VALUES ( 4, 13, DATE( '2002-03-01' ), 2, 4142, 1 )",
			"INSERT INTO v_i_fact VALUES ( 5, 32, DATE( '2002-04-01' ), 3, 5323, 1 )",
			"INSERT INTO v_i_fact VALUES ( 6, 33, DATE( '2002-05-01' ), 4, 6334, 1 )",
			"INSERT INTO v_i_fact VALUES ( 7, 35, DATE( '2002-06-01' ), 0, 19, 1 )",
			"INSERT INTO v_i_fact VALUES ( 7, 35, DATE( '2002-06-01' ), 5, 7355, 1 )",
			"INSERT INTO v_i_fact VALUES ( 8, 50, DATE( '2002-07-01' ), 6, 8506, 1 )",
			"INSERT INTO v_i_fact VALUES ( 8, 50, DATE( '2002-07-01' ), 7, 8507, 1 )",
			"INSERT INTO v_i_fact VALUES ( 9, 51, DATE( '2002-08-01' ), 0, 9517, 1 )",
			"INSERT INTO v_i_fact VALUES ( 10, 55, DATE( '2002-09-01' ), 0, 10558, 1 )",
			"INSERT INTO v_i_fact VALUES ( 11, 56, DATE( '2002-10-01' ), 0, 11569, 1 )",
			"INSERT INTO v_i_fact VALUES ( 12, 50, DATE( '2002-11-01' ), 10, 125010, 1 )",
			"INSERT INTO v_i_fact VALUES ( 13, 50, DATE( '2002-12-01' ), 11, 135011, 1 )",
			"INSERT INTO v_i_fact VALUES ( 1, 50, DATE( '2003-01-01' ), 12, 150120301, 1 )",
			"INSERT INTO v_i_fact VALUES ( 1, 50, DATE( '2003-02-01' ), 13, 150130302, 1 )",
			"INSERT INTO v_i_fact VALUES ( 1, 50, DATE( '2003-03-01' ), 14, 150140303, 1 )",
			"INSERT INTO v_i_fact VALUES ( 1, 50, DATE( '2003-04-01' ), 15, 150150304, 1 )",
			"INSERT INTO v_i_fact VALUES ( 1000, 1, DATE( '2002-01-01' ), 1, 1411, 1 )",
			"INSERT INTO v_i_fact VALUES ( 1000, 2, DATE( '2002-01-01' ), 2, 1522, 1 )",
			"INSERT INTO v_i_fact VALUES ( 1000, 3, DATE( '2002-01-01' ), 3, 1633, 1 )",
			"INSERT INTO v_i_fact VALUES ( 1000, 4, DATE( '2002-01-01' ), 4, 1744, 1 )",
			"INSERT INTO v_i_fact VALUES ( 1000, 5, DATE( '2002-01-01' ), 5, 1855, 1 )",
			"INSERT INTO v_i_fact VALUES ( 1000, 6, DATE( '2002-01-01' ), 6, 1966, 1 )",
			"INSERT INTO v_i_fact VALUES ( 1000, 7, DATE( '2002-01-01' ), 7, 2077, 1 )",
			"INSERT INTO v_i_fact VALUES ( 21, 8, DATE( '2002-01-01' ), 8, 2188, 1 )",
			"INSERT INTO v_i_fact VALUES ( 22, 9, DATE( '2002-01-01' ), 9, 2299, 1 )",
			"INSERT INTO v_i_fact VALUES ( 23, 14, DATE( '2002-01-01' ), 10, 231410, 1 )",
			"INSERT INTO v_i_fact VALUES ( 1000, 14, DATE( '2002-01-01' ), 11, 10001411, 1 )"
		},
		testPostparation = {
			"SET SCHEMA cis",
			"DROP TABLE v_i_category",
			"DROP TABLE v_i_cb",
			"DROP TABLE v_i_denomination",
			"DROP TABLE v_i_period",
			"DROP TABLE v_i_period_type",
			"DROP TABLE v_i_fact",
			"DROP SCHEMA cis RESTRICT"
		};

	/** The output from the {@link CISReader} under Test. */
	private static final Object[] testOutput = {
		new InterchangeBegin(
			"IREF000035", "4F1", "4F0", "001201:1410", true
		),
		new MessageBegin(
			"MREF000001", "4F1", "4F0", "ECB",
			MessageBegin.STATISTICAL_DATA, "Test message",
			new Contact[]{},
			"ECB_BKN1", MessageBegin.UPDATE_OR_REPLACE_ACTION,
			"200203140015", "", "", "ECB_BKN1"
		),
		new Attribute( ".4F.IS10.B.10P0.01.F.Q", "", "", "UNIT", "PURE_NUMB" ),
		new Attribute( ".4F.IS10.B.10P0.01.F.Q", "", "", "UNIT_MULT", "0" ),
		new Attribute( ".4F.IS10.B.10P0.01.F.Q", "", "", "DECIMALS", "0" ),
		new Attribute( "M.4F.IS10.B.10P0.01.F.Q", "", "", "COLLECTION", "E" ),
		new TimeSeries(
			"M.4F.IS10.B.10P0.01.F.Q", "200201200204", "710",
			new Observation[] {
				new Observation( "1235", "A", "F", "" ),
				new Observation( "1236", "A", "F", "" ),
				Observation.EMPTY_OBSERVATION,
				new Observation( "1238", "A", "F", "" )
			}
		),
		new Attribute( ".4F.NC10.C.10N1.01.S.Q", "", "", "UNIT", "PURE_NUMB" ),
		new Attribute( ".4F.NC10.C.10N1.01.S.Q", "", "", "UNIT_MULT", "0" ),
		new Attribute( ".4F.NC10.C.10N1.01.S.Q", "", "", "DECIMALS", "0" ),
		new Attribute( "M.4F.NC10.C.10N1.01.S.Q", "", "", "COLLECTION", "E" ),
		new TimeSeries(
			"M.4F.NC10.C.10N1.01.S.Q", "200301", "610",
			new Observation[] {
				new Observation( "150120301", "A", "F", "" )
			}
		),
		new Attribute( ".4F.NC10.C.50N2.01.S.Q", "", "", "UNIT", "PURE_NUMB" ),
		new Attribute( ".4F.NC10.C.50N2.01.S.Q", "", "", "UNIT_MULT", "0" ),
		new Attribute( ".4F.NC10.C.50N2.01.S.Q", "", "", "DECIMALS", "0" ),
		new Attribute( "M.4F.NC10.C.50N2.01.S.Q", "", "", "COLLECTION", "E" ),
		new TimeSeries(
			"M.4F.NC10.C.50N2.01.S.Q", "200302", "610",
			new Observation[] {
				new Observation( "150130302", "A", "F", "" )
			}
		),
		new Attribute( ".4F.NC10.C.20N2.01.S.Q", "", "", "UNIT", "PURE_NUMB" ),
		new Attribute( ".4F.NC10.C.20N2.01.S.Q", "", "", "UNIT_MULT", "0" ),
		new Attribute( ".4F.NC10.C.20N2.01.S.Q", "", "", "DECIMALS", "0" ),
		new Attribute( "M.4F.NC10.C.20N2.01.S.Q", "", "", "COLLECTION", "E" ),
		new TimeSeries(
			"M.4F.NC10.C.20N2.01.S.Q", "200303", "610",
			new Observation[] {
				new Observation( "150140303", "A", "F", "" )
			}
		),
		new Attribute( ".4F.NC10.C.10N2.01.S.Q", "", "", "UNIT", "PURE_NUMB" ),
		new Attribute( ".4F.NC10.C.10N2.01.S.Q", "", "", "UNIT_MULT", "0" ),
		new Attribute( ".4F.NC10.C.10N2.01.S.Q", "", "", "DECIMALS", "0" ),
		new Attribute( "M.4F.NC10.C.10N2.01.S.Q", "", "", "COLLECTION", "E" ),
		new TimeSeries(
			"M.4F.NC10.C.10N2.01.S.Q", "200304", "610",
			new Observation[] {
				new Observation( "150150304", "A", "F", "" )
			}
		),
		new Attribute( ".AT.RE10.B.20P0.01.F.Q", "", "", "UNIT", "PURE_NUMB" ),
		new Attribute( ".AT.RE10.B.20P0.01.F.Q", "", "", "UNIT_MULT", "0" ),
		new Attribute( ".AT.RE10.B.20P0.01.F.Q", "", "", "DECIMALS", "0" ),
		new Attribute( "M.AT.RE10.B.20P0.01.F.Q", "", "", "COLLECTION", "E" ),
		new TimeSeries(
			"M.AT.RE10.B.20P0.01.F.Q", "200201", "610",
			new Observation[] {
				new Observation( "1358", "A", "F", "" )
			}
		),
		new Attribute( ".LU.SR10.B.50P2.01.F.Q", "", "", "UNIT", "PURE_NUMB" ),
		new Attribute( ".LU.SR10.B.50P2.01.F.Q", "", "", "UNIT_MULT", "0" ),
		new Attribute( ".LU.SR10.B.50P2.01.F.Q", "", "", "DECIMALS", "0" ),
		new Attribute( "M.LU.SR10.B.50P2.01.F.Q", "", "", "COLLECTION", "E" ),
		new TimeSeries(
			"M.LU.SR10.B.50P2.01.F.Q", "200202", "610",
			new Observation[] {
				new Observation( "3121", "A", "F", "" )
			}
		),
		new Attribute( ".FR.SR11.B.20P2.01.F.Q", "", "", "UNIT", "PURE_NUMB" ),
		new Attribute( ".FR.SR11.B.20P2.01.F.Q", "", "", "UNIT_MULT", "0" ),
		new Attribute( ".FR.SR11.B.20P2.01.F.Q", "", "", "DECIMALS", "0" ),
		new Attribute( "M.FR.SR11.B.20P2.01.F.Q", "", "", "COLLECTION", "E" ),
		new TimeSeries(
			"M.FR.SR11.B.20P2.01.F.Q", "200203", "610",
			new Observation[] {
				new Observation( "4142", "A", "F", "" )
			}
		),
		new Attribute( ".DE.IS10.C.10P2.01.F.Q", "", "", "UNIT", "PURE_NUMB" ),
		new Attribute( ".DE.IS10.C.10P2.01.F.Q", "", "", "UNIT_MULT", "0" ),
		new Attribute( ".DE.IS10.C.10P2.01.F.Q", "", "", "DECIMALS", "0" ),
		new Attribute( "M.DE.IS10.C.10P2.01.F.Q", "", "", "COLLECTION", "E" ),
		new TimeSeries(
			"M.DE.IS10.C.10P2.01.F.Q", "200204", "610",
			new Observation[] {
				new Observation( "5323", "A", "F", "" )
			}
		),
		new Attribute( ".IT.RE10.C.50P1.01.F.Q", "", "", "UNIT", "PURE_NUMB" ),
		new Attribute( ".IT.RE10.C.50P1.01.F.Q", "", "", "UNIT_MULT", "0" ),
		new Attribute( ".IT.RE10.C.50P1.01.F.Q", "", "", "DECIMALS", "0" ),
		new Attribute( "M.IT.RE10.C.50P1.01.F.Q", "", "", "COLLECTION", "E" ),
		new TimeSeries(
			"M.IT.RE10.C.50P1.01.F.Q", "200205", "610",
			new Observation[] {
				new Observation( "6334", "A", "F", "" )
			}
		),
		new Attribute( ".ES.CL10.C.ALLD.01.S.E", "", "", "UNIT", "EUR" ),
		new Attribute( ".ES.CL10.C.ALLD.01.S.E", "", "", "UNIT_MULT", "0" ),
		new Attribute( ".ES.CL10.C.ALLD.01.S.E", "", "", "DECIMALS", "2" ),
		new Attribute( "M.ES.CL10.C.ALLD.01.S.E", "", "", "COLLECTION", "E" ),
		new TimeSeries(
			"M.ES.CL10.C.ALLD.01.S.E", "200206", "610",
			new Observation[] {
				new Observation( "19", "A", "F", "" )
			}
		),
		new Attribute( ".ES.CL10.C.20P1.01.S.Q", "", "", "UNIT", "PURE_NUMB" ),
		new Attribute( ".ES.CL10.C.20P1.01.S.Q", "", "", "UNIT_MULT", "0" ),
		new Attribute( ".ES.CL10.C.20P1.01.S.Q", "", "", "DECIMALS", "0" ),
		new Attribute( "M.ES.CL10.C.20P1.01.S.Q", "", "", "COLLECTION", "E" ),
		new TimeSeries(
			"M.ES.CL10.C.20P1.01.S.Q", "200206", "610",
			new Observation[] {
				new Observation( "7355", "A", "F", "" )
			}
		),
		new Attribute( ".NL.NC10.C.10P1.01.S.Q", "", "", "UNIT", "PURE_NUMB" ),
		new Attribute( ".NL.NC10.C.10P1.01.S.Q", "", "", "UNIT_MULT", "0" ),
		new Attribute( ".NL.NC10.C.10P1.01.S.Q", "", "", "DECIMALS", "0" ),
		new Attribute( "M.NL.NC10.C.10P1.01.S.Q", "", "", "COLLECTION", "E" ),
		new TimeSeries(
			"M.NL.NC10.C.10P1.01.S.Q", "200207", "610",
			new Observation[] {
				new Observation( "8506", "A", "F", "" )
			}
		),
		new Attribute( ".NL.NC10.C.50P0.01.S.Q", "", "", "UNIT", "PURE_NUMB" ),
		new Attribute( ".NL.NC10.C.50P0.01.S.Q", "", "", "UNIT_MULT", "0" ),
		new Attribute( ".NL.NC10.C.50P0.01.S.Q", "", "", "DECIMALS", "0" ),
		new Attribute( "M.NL.NC10.C.50P0.01.S.Q", "", "", "COLLECTION", "E" ),
		new TimeSeries(
			"M.NL.NC10.C.50P0.01.S.Q", "200207", "610",
			new Observation[] {
				new Observation( "8507", "A", "F", "" )
			}
		),
		new Attribute( ".BE.A030.Z.ZZZZ.ZZ.S.Q", "", "", "UNIT", "PURE_NUMB" ),
		new Attribute( ".BE.A030.Z.ZZZZ.ZZ.S.Q", "", "", "UNIT_MULT", "0" ),
		new Attribute( ".BE.A030.Z.ZZZZ.ZZ.S.Q", "", "", "DECIMALS", "0" ),
		new Attribute( "M.BE.A030.Z.ZZZZ.ZZ.S.Q", "", "", "COLLECTION", "E" ),
		new TimeSeries(
			"M.BE.A030.Z.ZZZZ.ZZ.S.Q", "200208", "610",
			new Observation[] {
				new Observation( "9517", "A", "F", "" )
			}
		),
		new Attribute( ".IE.A020.Z.ZZZZ.ZZ.S.Q", "", "", "UNIT", "PURE_NUMB" ),
		new Attribute( ".IE.A020.Z.ZZZZ.ZZ.S.Q", "", "", "UNIT_MULT", "0" ),
		new Attribute( ".IE.A020.Z.ZZZZ.ZZ.S.Q", "", "", "DECIMALS", "0" ),
		new Attribute( "M.IE.A020.Z.ZZZZ.ZZ.S.Q", "", "", "COLLECTION", "E" ),
		new TimeSeries(
			"M.IE.A020.Z.ZZZZ.ZZ.S.Q", "200209", "610",
			new Observation[] {
				new Observation( "10558", "A", "F", "" )
			}
		),
		new Attribute( ".FI.A010.Z.ZZZZ.ZZ.S.Q", "", "", "UNIT", "PURE_NUMB" ),
		new Attribute( ".FI.A010.Z.ZZZZ.ZZ.S.Q", "", "", "UNIT_MULT", "0" ),
		new Attribute( ".FI.A010.Z.ZZZZ.ZZ.S.Q", "", "", "DECIMALS", "0" ),
		new Attribute( "M.FI.A010.Z.ZZZZ.ZZ.S.Q", "", "", "COLLECTION", "E" ),
		new TimeSeries(
			"M.FI.A010.Z.ZZZZ.ZZ.S.Q", "200210", "610",
			new Observation[] {
				new Observation( "11569", "A", "F", "" )
			}
		),
		new Attribute( ".PT.NC10.C.50N1.01.S.Q", "", "", "UNIT", "PURE_NUMB" ),
		new Attribute( ".PT.NC10.C.50N1.01.S.Q", "", "", "UNIT_MULT", "0" ),
		new Attribute( ".PT.NC10.C.50N1.01.S.Q", "", "", "DECIMALS", "0" ),
		new Attribute( "M.PT.NC10.C.50N1.01.S.Q", "", "", "COLLECTION", "E" ),
		new TimeSeries(
			"M.PT.NC10.C.50N1.01.S.Q", "200211", "610",
			new Observation[] {
				new Observation( "125010", "A", "F", "" )
			}
		),
		new Attribute( ".GR.NC10.C.20N1.01.S.Q", "", "", "UNIT", "PURE_NUMB" ),
		new Attribute( ".GR.NC10.C.20N1.01.S.Q", "", "", "UNIT_MULT", "0" ),
		new Attribute( ".GR.NC10.C.20N1.01.S.Q", "", "", "DECIMALS", "0" ),
		new Attribute( "M.GR.NC10.C.20N1.01.S.Q", "", "", "COLLECTION", "E" ),
		new TimeSeries(
			"M.GR.NC10.C.20N1.01.S.Q", "200212", "610",
			new Observation[] {
				new Observation( "135011", "A", "F", "" )
			}
		),
		new Attribute( ".U2.NC10.B.50P2.01.S.Q", "", "", "UNIT", "PURE_NUMB" ),
		new Attribute( ".U2.NC10.B.50P2.01.S.Q", "", "", "UNIT_MULT", "0" ),
		new Attribute( ".U2.NC10.B.50P2.01.S.Q", "", "", "DECIMALS", "0" ),
		new Attribute( "M.U2.NC10.B.50P2.01.S.Q", "", "", "COLLECTION", "E" ),
		new TimeSeries(
			"M.U2.NC10.B.50P2.01.S.Q", "200201", "610",
			new Observation[] {
				new Observation( "1411", "A", "F", "" )
			}
		),
		new Attribute( ".U2.NC10.B.20P2.01.S.Q", "", "", "UNIT", "PURE_NUMB" ),
		new Attribute( ".U2.NC10.B.20P2.01.S.Q", "", "", "UNIT_MULT", "0" ),
		new Attribute( ".U2.NC10.B.20P2.01.S.Q", "", "", "DECIMALS", "0" ),
		new Attribute( "M.U2.NC10.B.20P2.01.S.Q", "", "", "COLLECTION", "E" ),
		new TimeSeries(
			"M.U2.NC10.B.20P2.01.S.Q", "200201", "610",
			new Observation[] {
				new Observation( "-1522", "A", "F", "" )
			}
		),
		new Attribute( ".U2.NC10.B.10P2.01.S.Q", "", "", "UNIT", "PURE_NUMB" ),
		new Attribute( ".U2.NC10.B.10P2.01.S.Q", "", "", "UNIT_MULT", "0" ),
		new Attribute( ".U2.NC10.B.10P2.01.S.Q", "", "", "DECIMALS", "0" ),
		new Attribute( "M.U2.NC10.B.10P2.01.S.Q", "", "", "COLLECTION", "E" ),
		new TimeSeries(
			"M.U2.NC10.B.10P2.01.S.Q", "200201", "610",
			new Observation[] {
				new Observation( "-1633", "A", "F", "" )
			}
		),
		new Attribute( ".U2.NC10.B.50P1.01.S.Q", "", "", "UNIT", "PURE_NUMB" ),
		new Attribute( ".U2.NC10.B.50P1.01.S.Q", "", "", "UNIT_MULT", "0" ),
		new Attribute( ".U2.NC10.B.50P1.01.S.Q", "", "", "DECIMALS", "0" ),
		new Attribute( "M.U2.NC10.B.50P1.01.S.Q", "", "", "COLLECTION", "E" ),
		new TimeSeries(
			"M.U2.NC10.B.50P1.01.S.Q", "200201", "610",
			new Observation[] {
				new Observation( "-1744", "A", "F", "" )
			}
		),
		new Attribute( ".U2.NC10.B.20P1.01.S.Q", "", "", "UNIT", "PURE_NUMB" ),
		new Attribute( ".U2.NC10.B.20P1.01.S.Q", "", "", "UNIT_MULT", "0" ),
		new Attribute( ".U2.NC10.B.20P1.01.S.Q", "", "", "DECIMALS", "0" ),
		new Attribute( "M.U2.NC10.B.20P1.01.S.Q", "", "", "COLLECTION", "E" ),
		new TimeSeries(
			"M.U2.NC10.B.20P1.01.S.Q", "200201", "610",
			new Observation[] {
				new Observation( "-1855", "A", "F", "" )
			}
		),
		new Attribute( ".U2.NC10.B.10P1.01.S.Q", "", "", "UNIT", "PURE_NUMB" ),
		new Attribute( ".U2.NC10.B.10P1.01.S.Q", "", "", "UNIT_MULT", "0" ),
		new Attribute( ".U2.NC10.B.10P1.01.S.Q", "", "", "DECIMALS", "0" ),
		new Attribute( "M.U2.NC10.B.10P1.01.S.Q", "", "", "COLLECTION", "E" ),
		new TimeSeries(
			"M.U2.NC10.B.10P1.01.S.Q", "200201", "610",
			new Observation[] {
				new Observation( "-1966", "A", "F", "" )
			}
		),
		new Attribute( ".U2.NC10.B.50P0.01.S.Q", "", "", "UNIT", "PURE_NUMB" ),
		new Attribute( ".U2.NC10.B.50P0.01.S.Q", "", "", "UNIT_MULT", "0" ),
		new Attribute( ".U2.NC10.B.50P0.01.S.Q", "", "", "DECIMALS", "0" ),
		new Attribute( "M.U2.NC10.B.50P0.01.S.Q", "", "", "COLLECTION", "E" ),
		new TimeSeries(
			"M.U2.NC10.B.50P0.01.S.Q", "200201", "610",
			new Observation[] {
				new Observation( "-2077", "A", "F", "" )
			}
		),
		new MessageEnd( "MREF000001" ),
		new InterchangeEnd( "IREF000035" )
	};

	/* ========================= Fields ========================= */

	private final SQLTestHelper testHelper;

	/* ====================== Constructors ====================== */

	/**
	 * Constructs an CISReaderTest with the specified name.
	 *
	 * @param name Test case name
	 */
	public CISReaderTest ( final String name ) {
		super( name );
		testHelper = new SQLTestHelper( "cisreader" );
	}

	/* ================== Protected Methods ===================== */

	protected void setUp () {
		testHelper.setUp();

		try {
			final Connection connection = testHelper.createConnection();
			testHelper.executeStatements( connection, testPreparation, false );
			connection.close();
		}
		catch ( final SQLException sqle ) {
			fail( "Preparing test DB failed: " + sqle );
		}

		System.setProperty(
			CISReader.class.getName() + ".Romana.source",
			SQLTestHelper.DB_URL_BASE + ";create=false;readonly=true"
		);
		System.setProperty(
			CISReader.class.getName() + ".Romana.user",
			SQLTestHelper.USER
		);
		System.setProperty(
			CISReader.class.getName() + ".Romana.password",
			SQLTestHelper.PASSWORD
		);
		System.setProperty(
			CISReader.class.getName() + ".Romana.iref",
			"IREF000035"
		);
		System.setProperty(
			CISReader.class.getName() + ".Romana.sender",
			"4F1"
		);
		System.setProperty(
			CISReader.class.getName() + ".Romana.receiver",
			"4F0"
		);
		System.setProperty(
			CISReader.class.getName() + ".Romana.subject",
			"Test message"
		);
		System.setProperty(
			CISReader.class.getName() + ".Romana.test",
			"true"
		);
	}

	protected void tearDown () {
		final Properties properties = System.getProperties();
		properties.remove(
			CISReader.class.getName() + ".Romana.source"
		);
		properties.remove(
			CISReader.class.getName() + ".Romana.user"
		);
		properties.remove(
			CISReader.class.getName() + ".Romana.password"
		);
		properties.remove(
			CISReader.class.getName() + ".Romana.iref"
		);
		properties.remove(
			CISReader.class.getName() + ".Romana.sender"
		);
		properties.remove(
			CISReader.class.getName() + ".Romana.receiver"
		);
		properties.remove(
			CISReader.class.getName() + ".Romana.subject"
		);
		properties.remove(
			CISReader.class.getName() + ".Romana.test"
		);

		testHelper.executeStatements( testPostparation, false );
		testHelper.tearDown();
	}

	/* ===================== Public Methods ===================== */

	/**
	 * Assembles and returns a test suite for all
	 * the test methods of this test case.
	 *
	 * @return a non-null test suite
	 */
	public static Test suite () {
		return new TestSuite( CISReaderTest.class );
	}

	/** Run the test case. */
	public static void main ( final String args[] ) {
		final String[] testCaseName = { CISReaderTest.class.getName() };
		junit.textui.TestRunner.main( testCaseName );
	}

	/* ---------------------- Test Methods ---------------------- */

	public void testExtractDB ()
		throws SQLException
	{
		final CISReader reader = new CISReader( "Romana" );
		final TestEventListener listener = new DateRelaxedTestEventListener( testOutput );
		reader.addListener( listener, true );
		reader.run();
		reader.close();
		listener.close();
	}
}

/* ***************************** Class ****************************** */

class DateRelaxedTestEventListener extends TestEventListener {
	public DateRelaxedTestEventListener ( final Object[] expectation ) {
		super( expectation );
	}
	protected void assertEquals ( final Object expected, final Object received ) {
		if ( expected instanceof InterchangeBegin ) {
			final InterchangeBegin
				expectedI = (InterchangeBegin) expected,
				receivedI = (InterchangeBegin) received;
			Assert.assertEquals( "iref", expectedI.getIref(), receivedI.getIref() );
			Assert.assertEquals( "Interchange sender", expectedI.getSender(), receivedI.getSender() );
			Assert.assertEquals( "Interchange receiver", expectedI.getReceiver(), receivedI.getReceiver() );
			Assert.assertEquals( "testFlag", expectedI.isTest(), receivedI.isTest() );
			// ignore preparation timestamp
		} else if ( expected instanceof MessageBegin ) {
			final MessageBegin
				expectedM = (MessageBegin) expected,
				receivedM = (MessageBegin) received;
			Assert.assertEquals( "mref", expectedM.getMref(), receivedM.getMref() );
			Assert.assertEquals( "Message sender", expectedM.getSender(), receivedM.getSender() );
			Assert.assertEquals( "Message receiver", expectedM.getReceiver(), receivedM.getReceiver() );
			Assert.assertEquals( "Message identity", expectedM.getIdentity(), receivedM.getIdentity() );
			Assert.assertEquals( "Message function", expectedM.getFunction(), receivedM.getFunction() );
			Assert.assertEquals( "maintenanceAgency", expectedM.getMaintenanceAgencyId(), receivedM.getMaintenanceAgencyId() );
			Assert.assertEquals( "keyFamily", expectedM.getKeyFamilyId(), receivedM.getKeyFamilyId() );
			Assert.assertEquals( "dataSet", expectedM.getDataSetId(), receivedM.getDataSetId() );
			Assert.assertEquals( "Message action", expectedM.getAction(), receivedM.getAction() );
			final Interval
				expectedReporting = expectedM.getReportingPeriodAsInterval(),
				receivedReporting = receivedM.getReportingPeriodAsInterval();
			if ( expectedReporting != receivedReporting ) {
				Assert.assertEquals( "reportingBegin", expectedReporting.getBegin(), receivedReporting.getBegin() );
				Assert.assertEquals( "reportingEnd", expectedReporting.getEnd(), receivedReporting.getEnd() );
			}
		} else {
			Assert.assertEquals( "Output", expected, received );
		}
	}
}
