// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (C) 2006, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.sql;

import java.text.ParseException;

import java.util.Date;
import java.util.List;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.Map;
import java.util.HashMap;

import java.sql.Types;
import java.sql.PreparedStatement;

import Int.ecb.gesmescb.KeyFamily;
import Int.ecb.gesmescb.KeyFamilyComponent;
import Int.ecb.gesmescb.Observation;
import Int.ecb.gesmescb.TimeSeries;
import Int.ecb.gesmescb.util.KeyTokenizer;
import Int.ecb.gesmescb.util.ComponentHelper;

import Int.ecb.logging.Level;

/* ***************************** Class ****************************** */

/**
 * Helps preparing records to be written to the ECB Statistical Data
 * Warehouse (SDW) database.
 *
 * A parameter array of {@link Object}s is manipulated instead of directly
 * a {@link java.sql.PreparedStatement} in order to make unit testing easier.
 * Nevertheless, the array is an ordered 1:1 representation of the parameters
 * to be set in the <code>PreparedStatement</code>.
 *
 * @author
 * <a href="mailto:sis.external@ecb.int">Bernhard Bodenstorfer</A>
 *
 * @author
 * <a href="mailto:sis.external@ecb.int">Xavier Sosnowska</A>
 *
 * @version @(#)$Id: SDWLayout.java,v 1.1 2017/01/19 16:14:02 trainot Exp $
 */
public class SDWLayout {

	/* ======================= Constants ======================== */

	/** Represents a NULL of {@link Types#CHAR} type. */
	public static final Object
		NULL_STRING = new NullRepresentation( "String NULL" );

	/** Represents a NULL of {@link Types#NUMERIC} type. */
	public static final Object
		NULL_NUMERIC = new NullRepresentation( "Numeric NULL" );

	/** Represents a NULL of {@link Types#DOUBLE} type. */
	public static final Object
		NULL_DOUBLE = new NullRepresentation( "Double NULL" );

	/** Represents a NULL of {@link Types#DATE} type. */
	public static final Object
		NULL_DATE = new NullRepresentation( "Date NULL" );

	/** "Not a Number" to represent {@link Observation#MISSING_VALUE}. */
	private static final Double NaN = new Double( Double.NaN );

	/* ========================= Fields ========================= */

	/** Takes care of logging and preferences. */
	private final ComponentHelper componentHelper;

	/** The relevant key family describing the table layouts. */
	private final KeyFamily keyFamily;

	/** The dataset identifier. */
	private final String dsi;

	/** The name of the database schema. */
	private final String schema;

	/** ..._TMP table SQL prepared statement */
	private final String obsTmpStatement, seriesTmpStatement, delTmpStatement;

	/** ..._TMP table column count */
	private final int obsTmpWidth, seriesTmpWidth, delTmpWidth;

	/** ..._TMP table column positions per concept function */
	private final int
		obsTmpCoordinates[], seriesTmpCoordinates[],
		obsTmpKey, seriesTmpKey, siblingTmpKey, delTmpKey,
		obsTmpKeyImpact[], seriesTmpKeyImpact[], delTmpKeyImpact[],
		obsTmpSeqId, seriesTmpSeqId, delTmpSeqId,
		obsTmpDt, seriesTmpDt, delTmpDt,
		obsTmpDate, delTmpDate,
		obsTmpCells[], delTmpIsSibling,
		obsTmpNullImpact[], seriesTmpNullImpact[];

	/** ..._TMP table column positions per concept name */
	private final Map obsTmpConcepts, seriesTmpConcepts;

	/* ====================== Constructors ====================== */

	/** Construct a <code>SDWLayout</code> for a key family. */
	public SDWLayout (
		final KeyFamily keyFamily,
		final String dsi,
		final String schema
	) {
		componentHelper = new ComponentHelper( keyFamily.getId(), this );

		this.keyFamily = keyFamily;
		this.dsi = dsi;
		this.schema = schema;

		obsTmpConcepts = new HashMap();
		seriesTmpConcepts = new HashMap();

		final KeyFamilyComponent
			frequency = keyFamily.getFrequency(),
			time = keyFamily.getTimePeriod();
		final boolean
			hasFrequency = null != frequency,
			hasTime = null != time;

		final StringBuffer
			obsTmpStatement = new StringBuffer(),
			obsTmpParameters = new StringBuffer(),
			seriesTmpStatement = new StringBuffer(),
			seriesTmpParameters = new StringBuffer(),
			delTmpStatement = new StringBuffer(),
			delTmpParameters = new StringBuffer();

		obsTmpStatement.append( "INSERT INTO " );
		seriesTmpStatement.append( "INSERT INTO " );
		delTmpStatement.append( "INSERT INTO " );
		if ( null != schema ) {
			obsTmpStatement.append( schema );
			seriesTmpStatement.append( schema );
			delTmpStatement.append( schema );
			obsTmpStatement.append( '.' );
			seriesTmpStatement.append( '.' );
			delTmpStatement.append( '.' );
		}
		obsTmpStatement.append( dsi );
		seriesTmpStatement.append( dsi );
		delTmpStatement.append( dsi );
		obsTmpStatement.append( "_OBS_TMP (" );
		seriesTmpStatement.append( "_SERIES_TMP (" );
		delTmpStatement.append( "_DEL_TMP (" );

		int
			obsTmpIndex = 0,
			seriesTmpIndex = 0,
			delTmpIndex = 0;

		if ( hasFrequency ) {
			obsTmpStatement.append( "FREQUENCY," );
			obsTmpParameters.append( "?," );
			obsTmpConcepts.put(
				frequency.getConceptId(),
				new Integer( obsTmpIndex++ )
			);
			seriesTmpStatement.append( "FREQUENCY," );
			seriesTmpParameters.append( "?," );
			seriesTmpConcepts.put(
				frequency.getConceptId(),
				new Integer( seriesTmpIndex++ )
			);
		}
		if ( hasTime ) {
			obsTmpStatement.append( "OBS_DATE," );
			obsTmpParameters.append( "?," );
			obsTmpDate = obsTmpIndex;
			obsTmpConcepts.put(
				time.getConceptId(),
				new Integer( obsTmpIndex++ )
			);
			delTmpStatement.append( "OBS_DATE," );
			delTmpParameters.append( "?," );
			delTmpDate = delTmpIndex++;
		} else {
			obsTmpDate = delTmpDate = Integer.MIN_VALUE;
		}

		obsTmpCoordinates = new int[ keyFamily.getKeyDimensionCount() ];
		seriesTmpCoordinates = new int[ keyFamily.getKeyDimensionCount() ];
		{
			int i = 0;
			for (
				final Iterator dimensions = keyFamily.keyStructureIterator();
				dimensions.hasNext();
				++i
			) {
				final KeyFamilyComponent dimension
				= (KeyFamilyComponent) dimensions.next();
				if ( dimension == frequency ) {
					obsTmpCoordinates[i] = 0;
					seriesTmpCoordinates[i] = 0;
				} else {
					obsTmpStatement.append( dimension.getConceptId() );
					obsTmpStatement.append( ',' );
					obsTmpParameters.append( "?," );
					obsTmpCoordinates[i] = obsTmpIndex;
					obsTmpConcepts.put(
						dimension.getConceptId(),
						new Integer( obsTmpIndex++ )
					);
					seriesTmpStatement.append( dimension.getConceptId() );
					seriesTmpStatement.append( ',' );
					seriesTmpParameters.append( "?," );
					seriesTmpCoordinates[i] = seriesTmpIndex;
					seriesTmpConcepts.put(
						dimension.getConceptId(),
						new Integer( seriesTmpIndex++ )
					);
				}
			}
		}

		obsTmpStatement.append( "SERIES_KEY" );
		obsTmpParameters.append( "?" );
		obsTmpKey = obsTmpIndex++;
		delTmpStatement.append( "\"KEY\"" );
		delTmpParameters.append( "?" );
		delTmpKey = delTmpIndex++;

		obsTmpCells = new int[ keyFamily.getArrayCellCount() ];
		final List
			obsTmpNullImpact = new ArrayList(),
			seriesTmpNullImpact = new ArrayList();
		{
			int i = 0;
			for (
				final Iterator cells = keyFamily.arrayCellIterator();
				cells.hasNext();
				++i
			) {
				final KeyFamilyComponent cell
				= (KeyFamilyComponent) cells.next();
				obsTmpStatement.append( ',' );
				obsTmpStatement.append( cell.getConceptId() );
				obsTmpParameters.append( ",?" );
				obsTmpCells[i] = obsTmpIndex;
				obsTmpNullImpact.add(
					new Integer( obsTmpIndex )
				);
				obsTmpConcepts.put(
					cell.getConceptId(),
					new Integer( obsTmpIndex++ )
				);
			}
		}

		for (
			final Iterator attributes = keyFamily.attributeIterator();
			attributes.hasNext();
		) {
			final KeyFamilyComponent attribute
			= (KeyFamilyComponent) attributes.next();
			if ( attribute.isAttachedToObservation() ) {
				obsTmpStatement.append( ',' );
				obsTmpStatement.append( attribute.getConceptId() );
				obsTmpParameters.append( ",?" );
				obsTmpNullImpact.add(
					new Integer( obsTmpIndex )
				);
				obsTmpConcepts.put(
					attribute.getConceptId(),
					new Integer( obsTmpIndex++ )
				);
			} else if (
				attribute.isAttachedToTimeSeries()
				||
				attribute.isAttachedToSiblingGroup()
			) {
				seriesTmpStatement.append( attribute.getConceptId() );
				seriesTmpStatement.append( ',' );
				seriesTmpParameters.append( "?," );
				seriesTmpNullImpact.add(
					new Integer( seriesTmpIndex )
				);
				seriesTmpConcepts.put(
					attribute.getConceptId(),
					new Integer( seriesTmpIndex++ )
				);
			} else {
				componentHelper.logger.log(
					Level.SEVERE,
					componentHelper.logName,
					"SDWLayout()",
					"Attribute " + attribute.getConceptId()
					+ " is at "
					+ (
						attribute.isAttachedToDataSet()
						? "dataSet"
						: "\"" + attribute.getAttachmentLevel() + "\""
					)
					+ " level, which is not supported by SDW."
				);
			}
		}

		obsTmpStatement.append( ",SEQ_ID,DT" );
		obsTmpParameters.append( ",?,?" );
		obsTmpSeqId = obsTmpIndex++;
		obsTmpDt = obsTmpIndex++;

		seriesTmpStatement.append( "SEQ_ID,DT" );
		seriesTmpParameters.append( "?,?" );
		seriesTmpSeqId = seriesTmpIndex++;
		seriesTmpDt = seriesTmpIndex++;

		seriesTmpStatement.append( ",SERIES_KEY,SIBLING_KEY" );
		seriesTmpParameters.append( ",?,?" );
		seriesTmpKey = seriesTmpIndex++;
		siblingTmpKey = seriesTmpIndex++;

		delTmpStatement.append( ",ISSIBLING" );
		delTmpParameters.append( ",?" );
		delTmpIsSibling = delTmpIndex++;

		delTmpStatement.append( ",SEQ_ID,DT" );
		delTmpParameters.append( ",?,?" );
		delTmpSeqId = delTmpIndex++;
		delTmpDt = delTmpIndex++;

		obsTmpStatement.append( ") VALUES (" );
		seriesTmpStatement.append( ") VALUES (" );
		delTmpStatement.append( ") VALUES (" );
		obsTmpStatement.append( obsTmpParameters );
		seriesTmpStatement.append( seriesTmpParameters );
		delTmpStatement.append( delTmpParameters );
		obsTmpStatement.append( ")" );
		seriesTmpStatement.append( ")" );
		delTmpStatement.append( ")" );

		this.obsTmpStatement = obsTmpStatement.toString();
		this.seriesTmpStatement = seriesTmpStatement.toString();
		this.delTmpStatement = delTmpStatement.toString();

		obsTmpKeyImpact = new int[ obsTmpCoordinates.length + 1 ];
		for ( int i = 0; i < obsTmpCoordinates.length; ++i ) {
			obsTmpKeyImpact[i] = obsTmpCoordinates[i];
		}
		obsTmpKeyImpact[obsTmpCoordinates.length] = obsTmpKey;

		seriesTmpKeyImpact = new int[ seriesTmpCoordinates.length + 2 ];
		for ( int i = 0; i < seriesTmpCoordinates.length; ++i ) {
			seriesTmpKeyImpact[i] = seriesTmpCoordinates[i];
		}
		seriesTmpKeyImpact[seriesTmpCoordinates.length] = seriesTmpKey;
		seriesTmpKeyImpact[seriesTmpCoordinates.length + 1] = siblingTmpKey;

		delTmpKeyImpact = new int[]{ delTmpKey, delTmpIsSibling };

		this.obsTmpNullImpact = new int[ obsTmpNullImpact.size() ];
		{
			final Iterator iterator = obsTmpNullImpact.iterator();
			for ( int i = 0; i < this.obsTmpNullImpact.length; ++i ) {
				this.obsTmpNullImpact[i]
				= ( (Integer) iterator.next() ).intValue();
			}
		}

		this.seriesTmpNullImpact = new int[ seriesTmpNullImpact.size() ];
		{
			final Iterator iterator = seriesTmpNullImpact.iterator();
			for ( int i = 0; i < this.seriesTmpNullImpact.length; ++i ) {
				this.seriesTmpNullImpact[i]
				= ( (Integer) iterator.next() ).intValue();
			}
		}

		obsTmpWidth = obsTmpIndex;
		seriesTmpWidth = seriesTmpIndex;
		delTmpWidth = delTmpIndex;
	}

	/* ===================== Public Methods ===================== */

	/** Returns the (qualified) table identifier of the ..._OBS_TMP table. */
	public String obsTmpName () {
		return
			( null != schema ? schema + "." : "" )
			+ dsi
			+ "_OBS_TMP";
	}

	/** Returns SQL code of the {@link PreparedStatement}
	 * for the ..._OBS_TMP table. */
	public String obsTmpStatement () {
		return obsTmpStatement;
	}

	/** Returns the number of columns of the ..._OBS_TMP table. */
	public int obsTmpWidth () {
		return obsTmpWidth;
	}

	/** Writes the extraction timestamp into a parameters buffer
	 * for the ..._OBS_TMP table.
	 *
	 * @param parameters represents the fields in a {@link PreparedStatement}
	 * @param extracted determines data version
	 * @return the indices of the updated <code>parameters</code>
	 */
	public int[] obsTmpExtracted (
		final Object[] parameters,
		final Date extracted
	) throws ParseException {
		parameters[obsTmpDt] = extracted;
		return new int[]{ obsTmpDt };
	}

	/** Writes the sequential update counter into a parameters buffer
	 * for the ..._OBS_TMP table.
	 *
	 * @param parameters represents the fields in a {@link PreparedStatement}
	 * @param seqId linear number of update event
	 * @return the indices of the updated <code>parameters</code>
	 */
	public int[] obsTmpSequence (
		final Object[] parameters,
		final int seqId
	) throws ParseException {
		parameters[obsTmpSeqId] = new Integer( seqId );
		return new int[]{ obsTmpSeqId };
	}

	/** Writes a time series key into a parameters buffer
	 * for the ..._OBS_TMP table.
	 *
	 * @param parameters represents the fields in a {@link PreparedStatement}
	 * @param key of the time series
	 * @return the indices of the updated <code>parameters</code>
	 */
	public int[] obsTmpKey (
		final Object[] parameters,
		final String key
	) throws ParseException {
		final KeyTokenizer coordinates = new KeyTokenizer( key );
		for (
			int i = 0;
			i < obsTmpCoordinates.length;
			++i
		) {
			final String coordinate = coordinates.nextToken();
			parameters[obsTmpCoordinates[i]] = coordinate;
		}
		parameters[obsTmpKey] = dsi + TimeSeries.KEY_SEPARATOR + key;
		return (int[]) obsTmpKeyImpact.clone();
	}

	/** Writes an observation time into a parameters buffer for the ..._OBS_TMP table.
	 *
	 * @param parameters represents the fields in a {@link PreparedStatement}
	 * @param date of the observation
	 * @return the indices of the updated <code>parameters</code>
	 */
	public int[] obsTmpDate (
		final Object[] parameters,
		final Date date
	) throws ParseException {
		parameters[obsTmpDate] = date;
		return new int[]{ obsTmpDate };
	}

	/** Resets array cell and attribute fields
	 * in a parameters buffer for the ..._OBS_TMP table.
	 *
	 * @param parameters represents the fields in a {@link PreparedStatement}
	 * @return the indices of the updated <code>parameters</code>
	 */
	public int[] obsTmpNull (
		final Object[] parameters
	) throws ParseException {
		int i = 0;
		parameters[obsTmpCells[i++]] = NULL_DOUBLE;
		parameters[obsTmpCells[i++]] = NULL_STRING;
		parameters[obsTmpCells[i++]] = NULL_STRING;
		parameters[obsTmpCells[i++]] = NULL_DOUBLE;
		while ( i < obsTmpNullImpact.length ) {
			parameters[obsTmpNullImpact[i++]] = NULL_STRING;
		}
		return (int[]) obsTmpNullImpact.clone();
	}

	/** Writes array cells into a parameters buffer for the ..._OBS_TMP table.
	 *
	 * @param parameters represents the fields in a {@link PreparedStatement}
	 * @param observation representing the array cell data
	 * @return the indices of the updated <code>parameters</code>
	 */
	public int[] obsTmpCells (
		final Object[] parameters,
		final Observation observation
	) throws ParseException {
		final String
			value = observation.getValue(),
			status = observation.getStatus(),
			confidentiality = observation.getConfidentiality(),
			preBreakValue = observation.getPreBreakValue();
		int i = 0;
		parameters[obsTmpCells[i++]]
		= "".equals( value ) ? NULL_DOUBLE
			: Observation.MISSING_VALUE.equals( value )	? NaN
				: Double.valueOf( value );
		parameters[obsTmpCells[i++]] = "".equals( status )
			? NULL_STRING : status;
		parameters[obsTmpCells[i++]] = "".equals( confidentiality )
			? NULL_STRING : confidentiality;
		parameters[obsTmpCells[i++]]
		= "".equals( preBreakValue ) ? NULL_DOUBLE
			: Observation.MISSING_VALUE.equals( preBreakValue )	? NaN
				: Double.valueOf( preBreakValue );
		return (int[]) obsTmpCells.clone();
	}

	/** Writes an attribute into a parameters buffer for the ..._OBS_TMP table.
	 *
	 * @param parameters represents the fields in a {@link PreparedStatement}
	 * @param name of the concept underlying the attribute
	 * @param value of the attribute
	 * @return the indices of the updated <code>parameters</code>
	 */
	public int[] obsTmpAttribute (
		final Object[] parameters,
		final String name,
		final String value
	) throws ParseException {
		final Integer indexInt = (Integer) obsTmpConcepts.get( name );
		if ( null == indexInt ) {
			throw new NullPointerException(
				"Unknown observation level attribute \"" + name + "\""
			);
		}
		final int index = indexInt.intValue();
		parameters[index] = value;
		return new int[]{ index };
	}

	/** Returns the (qualified) table identifier of the ..._SERIES_TMP table. */
	public String seriesTmpName () {
		return
			( null != schema ? schema + "." : "" )
			+ dsi
			+ "_SERIES_TMP";
	}

	/** Returns SQL code of the {@link PreparedStatement}
	 * for the ..._SERIES_TMP table. */
	public String seriesTmpStatement () {
		return seriesTmpStatement;
	}

	/** Returns the number of columns of the ..._SERIES_TMP table. */
	public int seriesTmpWidth () {
		return seriesTmpWidth;
	}

	/** Writes the extraction timestamp into a parameters buffer
	 * for the ..._SERIES_TMP table.
	 *
	 * @param parameters represents the fields in a {@link PreparedStatement}
	 * @param extracted determines data version
	 * @return the indices of the updated <code>parameters</code>
	 */
	public int[] seriesTmpExtracted (
		final Object[] parameters,
		final Date extracted
	) throws ParseException {
		parameters[seriesTmpDt] = extracted;
		return new int[]{ seriesTmpDt };
	}

	/** Writes the sequential update counter into a parameters buffer
	 * for the ..._SERIES_TMP table.
	 *
	 * @param parameters represents the fields in a {@link PreparedStatement}
	 * @param seqId linear number of update event
	 * @return the indices of the updated <code>parameters</code>
	 */
	public int[] seriesTmpSequence (
		final Object[] parameters,
		final int seqId
	) throws ParseException {
		parameters[seriesTmpSeqId] = new Integer( seqId );
		return new int[]{ seriesTmpSeqId };
	}

	/** Writes a time series or sibling group key into a parameters buffer
	 * for the ..._SERIES_TMP table.
	 *
	 * @param parameters represents the fields in a {@link PreparedStatement}
	 * @param key of the time series or sibling group
	 * @return the indices of the updated <code>parameters</code>
	 */
	public int[] seriesTmpKey (
		final Object[] parameters,
		final String key
	) throws ParseException {
		final KeyTokenizer coordinates = new KeyTokenizer( key );
		for (
			int i = 0;
			i < seriesTmpCoordinates.length;
			++i
		) {
			final String coordinate = coordinates.nextToken();
			if (
				0 == seriesTmpCoordinates[i]
				&&
				"".equals( coordinate )
			) {
				parameters[seriesTmpCoordinates[i]] = NULL_STRING;
			} else {
				parameters[seriesTmpCoordinates[i]] = coordinate;
			}
		}
		parameters[seriesTmpKey] = dsi + TimeSeries.KEY_SEPARATOR + key;
		parameters[siblingTmpKey] = dsi
			+ TimeSeries.KEY_SEPARATOR + eraseFrequencyFromKey( key );
		return (int[]) seriesTmpKeyImpact.clone();
	}

	/** Resets attribute fields
	 * in a parameters buffer for the ..._SERIES_TMP table.
	 *
	 * @param parameters represents the fields in a {@link PreparedStatement}
	 * @return the indices of the updated <code>parameters</code>
	 */
	public int[] seriesTmpNull (
		final Object[] parameters
	) throws ParseException {
		for ( int i = 0; i < seriesTmpNullImpact.length; ++i ) {
			parameters[seriesTmpNullImpact[i]] = NULL_STRING;
		}
		return (int[]) seriesTmpNullImpact.clone();
	}

	/** Writes an attribute into a parameters buffer
	 * for the ..._SERIES_TMP table.
	 *
	 * @param parameters represents the fields in a {@link PreparedStatement}
	 * @param name of the concept underlying the attribute
	 * @param value of the attribute
	 * @return the indices of the updated <code>parameters</code>
	 */
	public int[] seriesTmpAttribute (
		final Object[] parameters,
		final String name,
		final String value
	) throws ParseException {
		final Integer indexInt = (Integer) seriesTmpConcepts.get( name );
		if ( null == indexInt ) {
			throw new NullPointerException(
				"Unknown observation level attribute \"" + name + "\""
			);
		}
		final int index = indexInt.intValue();
		parameters[index] = value;
		return new int[]{ index };
	}

	/** Returns the (qualified) table identifier of the ..._DEL_TMP table. */
	public String delTmpName () {
		return
			( null != schema ? schema + "." : "" )
			+ dsi
			+ "_DEL_TMP";
	}

	/** Returns SQL code of the {@link PreparedStatement}
	 * for the ..._DEL_TMP table. */
	public String delTmpStatement () {
		return delTmpStatement;
	}

	/** Returns the number of columns of the ..._DEL_TMP table. */
	public int delTmpWidth () {
		return delTmpWidth;
	}

	/** Writes the extraction timestamp into a parameters buffer
	 * for the ..._DEL_TMP table.
	 *
	 * @param parameters represents the fields in a {@link PreparedStatement}
	 * @param extracted determines data version
	 * @return the indices of the updated <code>parameters</code>
	 */
	public int[] delTmpExtracted (
		final Object[] parameters,
		final Date extracted
	) throws ParseException {
		parameters[delTmpDt] = extracted;
		return new int[]{ delTmpDt };
	}

	/** Writes the sequential update counter into a parameters buffer
	 * for the ..._DEL_TMP table.
	 *
	 * @param parameters represents the fields in a {@link PreparedStatement}
	 * @param seqId linear number of update event
	 * @return the indices of the updated <code>parameters</code>
	 */
	public int[] delTmpSequence (
		final Object[] parameters,
		final int seqId
	) throws ParseException {
		parameters[delTmpSeqId] = new Integer( seqId );
		return new int[]{ delTmpSeqId };
	}

	/** Writes a time series or sibling group key into a parameters buffer
	 * for the ..._DEL_TMP table.
	 *
	 * @param parameters represents the fields in a {@link PreparedStatement}
	 * @param key of the time series or sibling group
	 * @return the indices of the updated <code>parameters</code>
	 */
	public int[] delTmpKey (
		final Object[] parameters,
		final String key
	) throws ParseException {
		parameters[delTmpKey] = dsi + TimeSeries.KEY_SEPARATOR + key;
		parameters[delTmpIsSibling] = "".equals(
			keyFamily.computeFrequencyValue( key )
		) ? Boolean.TRUE : Boolean.FALSE;
		return (int[]) delTmpKeyImpact.clone();
	}

	/** Resets date field
	 * in a parameters buffer for the ..._DEL_TMP table.
	 *
	 * @param parameters represents the fields in a {@link PreparedStatement}
	 * @return the indices of the updated <code>parameters</code>
	 */
	public int[] delTmpNull (
		final Object[] parameters
	) throws ParseException {
		parameters[delTmpDate] = NULL_DATE;
		return new int[]{ delTmpDate };
	}

	/** Writes an observation time
	 * into a parameters buffer for the ..._DEL_TMP table.
	 *
	 * @param parameters represents the fields in a {@link PreparedStatement}
	 * @param date of the observation
	 * @return the indices of the updated <code>parameters</code>
	 */
	public int[] delTmpDate (
		final Object[] parameters,
		final Date date
	) throws ParseException {
		parameters[delTmpDate] = date;
		return new int[]{ delTmpDate };
	}

	public String toString () {
		return getClass().getName()
			+ "(" + schema + "," + keyFamily.getId() + "," + dsi + ")";
	}

	/* ==================== Private Methods ===================== */

	private String eraseFrequencyFromKey ( final String key ) {
		final int frequencyPosition = keyFamily.getFrequencyDimensionPosition();
		int start = 0;
		for ( int i = 1; i < frequencyPosition; ++i ) {
			start = key.indexOf( TimeSeries.KEY_SEPARATOR, start ) + 1;
		}
		final int end = key.indexOf( TimeSeries.KEY_SEPARATOR, start );
		return key.substring( 0, start )
			+ ( 0 <= end ? key.substring( end ) : "" );
	}
}

/* ***************************** Class ****************************** */

/** Meant as placeholder for a typed SQL NULL value. */
class NullRepresentation {

	private final String string;

	public NullRepresentation ( final String string ) {
		this.string = string;
	}

	public String toString () {
		return getClass().getName() + "(" + string + ")";
	}
}
