// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (C) 2006, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.sql;

import java.text.DateFormat;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.Date;

import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;

import Int.ecb.gesmescb.KeyFamily;
import Int.ecb.gesmescb.KeyFamilyComponent;
import Int.ecb.gesmescb.Observation;

/* ***************************** Class ****************************** */

/**
 * <p>Determines SDW table names and layouts for a {@link KeyFamily}.</p>
 * <p>The Statistical Data Warehouse (SDW) uses relational databases.</p>
 *
 * @author
 * <A HREF="mailto:sis.external@ecb.int">Bernhard Bodenstorfer</A>
 *
 * @version @(#)$Id: SDWLayoutTest.java,v 1.1 2017/01/19 16:14:02 trainot Exp $
 */

public class SDWLayoutTest extends TestCase {

	/* ======================= Constants ======================== */

	/** To test dates. */
	private static final DateFormat DATE_FORMAT
	= new SimpleDateFormat( "yyyy-MM-dd'T'HH:mm:ss" );

	private static final KeyFamilyComponent
		FREQ = new KeyFamilyComponent(
			"FREQ",
			KeyFamilyComponent.FREQUENCY_DIMENSION,
			"1",
			"AN5",
			KeyFamilyComponent.MANDATORY,
			"",
			"CL_TEST",
			true
		),
		DIM2 = new KeyFamilyComponent(
			"DIM2",
			KeyFamilyComponent.KEY_STRUCTURE_DIMENSION,
			"2",
			"AN5",
			KeyFamilyComponent.MANDATORY,
			"",
			"CL_TEST",
			true
		),
		DIM3 = new KeyFamilyComponent(
			"DIM3",
			KeyFamilyComponent.KEY_STRUCTURE_DIMENSION,
			"3",
			"AN5",
			KeyFamilyComponent.MANDATORY,
			"",
			"CL_TEST",
			true
		),
		TIME_PERIOD = new KeyFamilyComponent(
			"TIME_PERIOD",
			KeyFamilyComponent.TIME_DIMENSION,
			"4",
			"AN5",
			KeyFamilyComponent.MANDATORY,
			"",
			"CL_TEST"
		),
		TIME_FORMAT = new KeyFamilyComponent(
			"TIME_FORMAT",
			KeyFamilyComponent.TIME_DIMENSION,
			"5",
			"N3",
			KeyFamilyComponent.MANDATORY,
			"",
			"CL_TEST"
		),
		OBS_VALUE = new KeyFamilyComponent(
			"OBS_VALUE",
			KeyFamilyComponent.ARRAY_CELL,
			"6",
			"AN..15",
			KeyFamilyComponent.MANDATORY,
			"",
			""
		),
		OBS_STATUS = new KeyFamilyComponent(
			"OBS_STATUS",
			KeyFamilyComponent.ARRAY_CELL,
			"7",
			"AN..15",
			KeyFamilyComponent.MANDATORY,
			"",
			"CL_TEST"
		),
		OBS_CONF = new KeyFamilyComponent(
			"OBS_CONF",
			KeyFamilyComponent.ARRAY_CELL,
			"8",
			"AN..15",
			KeyFamilyComponent.MANDATORY,
			"",
			"CL_TEST"
		),
		OBS_PRE_BREAK = new KeyFamilyComponent(
			"OBS_PRE_BREAK",
			KeyFamilyComponent.ARRAY_CELL,
			"9",
			"AN..15",
			KeyFamilyComponent.MANDATORY,
			"",
			""
		),
		OBS_COM = new KeyFamilyComponent(
			"OBS_COM",
			KeyFamilyComponent.ATTRIBUTE,
			"",
			"AN5",
			KeyFamilyComponent.CONDITIONAL,
			KeyFamilyComponent.ATTACHED_TO_OBSERVATION,
			""
		),
		OBS_ATT2 = new KeyFamilyComponent(
			"OBS_ATT2",
			KeyFamilyComponent.ATTRIBUTE,
			"",
			"AN..5",
			KeyFamilyComponent.CONDITIONAL,
			KeyFamilyComponent.ATTACHED_TO_OBSERVATION,
			"CL_TEST"
		),
		SER_ATT1 = new KeyFamilyComponent(
			"SER_ATT1",
			KeyFamilyComponent.ATTRIBUTE,
			"",
			"A5",
			KeyFamilyComponent.MANDATORY,
			KeyFamilyComponent.ATTACHED_TO_TIME_SERIES,
			"CL_TEST"
		),
		SER_ATT2 = new KeyFamilyComponent(
			"SER_ATT2",
			KeyFamilyComponent.ATTRIBUTE,
			"",
			"A..5",
			KeyFamilyComponent.MANDATORY,
			KeyFamilyComponent.ATTACHED_TO_TIME_SERIES,
			""
		),
		SG_ATT1 = new KeyFamilyComponent(
			"SG_ATT1",
			KeyFamilyComponent.ATTRIBUTE,
			"",
			"N25",
			KeyFamilyComponent.CONDITIONAL,
			KeyFamilyComponent.ATTACHED_TO_SIBLING_GROUP,
			"CL_TEST"
		),
		SG_ATT2 = new KeyFamilyComponent(
			"SG_ATT2",
			KeyFamilyComponent.ATTRIBUTE,
			"",
			"N..25",
			KeyFamilyComponent.CONDITIONAL,
			KeyFamilyComponent.ATTACHED_TO_SIBLING_GROUP,
			""
		),
		DS_ATT1 = new KeyFamilyComponent(
			"DS_ATT1",
			KeyFamilyComponent.ATTRIBUTE,
			"",
			"N20",
			KeyFamilyComponent.CONDITIONAL,
			KeyFamilyComponent.ATTACHED_TO_DATA_SET,
			"CL_TEST"
		),
		DS_ATT2 = new KeyFamilyComponent(
			"DS_ATT2",
			KeyFamilyComponent.ATTRIBUTE,
			"",
			"N..20",
			KeyFamilyComponent.CONDITIONAL,
			KeyFamilyComponent.ATTACHED_TO_DATA_SET,
			""
		);

	private static final KeyFamily KF1 = new KeyFamily(
		"KF1", "name", "desc",
		new Date( 0 ),
		new KeyFamilyComponent[]{
			FREQ, DIM2, DIM3,
			TIME_PERIOD, TIME_FORMAT,
			OBS_VALUE, OBS_STATUS, OBS_CONF, OBS_PRE_BREAK,
			OBS_COM, OBS_ATT2,
			SER_ATT1, SER_ATT2,
			SG_ATT1, SG_ATT2,
			DS_ATT1, DS_ATT2
		}
	);

	/* ========================= Fields ========================= */

	private final Integer testInteger;
	private final Date testDate;

	/* ====================== Constructors ====================== */

	/**
	 * Constructs a SDWLayoutTest with the specified name.
	 *
	 * @param name Test case name
	 */
	public SDWLayoutTest ( final String name ) throws ParseException {
		super( name );
		testInteger = new Integer( -139872 );
		testDate = DATE_FORMAT.parse( "2001-01-01T00:00:00" );
	}

	/* ===================== Public Methods ===================== */

	/**
	 * Assembles and returns a test suite for all
	 * the test methods of this test case.
	 *
	 * @return a non-null test suite
	 */
	public static Test suite () {
		final TestSuite suite = new TestSuite( SDWLayoutTest.class );
		return suite;
	}

	/** Runs the test case. */
	public static void main ( final String args[] ) {
		final String[] testCaseName = { SDWLayoutTest.class.getName() };
		junit.textui.TestRunner.main( testCaseName );
	}

	/* --------------------- Test Methods --------------------- */

	public void testObsTmp () throws ParseException {
		final SDWLayout helper = new SDWLayout( KF1, "DS1", "SCHEMA" );
		assertEquals(
			"Table name",
			"SCHEMA.DS1_OBS_TMP",
			helper.obsTmpName()
		);
		assertEquals(
			"Prepare statement",
			"INSERT INTO SCHEMA.DS1_OBS_TMP"
			+ " (FREQUENCY,OBS_DATE,DIM2,DIM3,SERIES_KEY,OBS_VALUE,OBS_STATUS,OBS_CONF,OBS_PRE_BREAK,OBS_ATT2,OBS_COM,SEQ_ID,DT)"
			+ " VALUES (?,?,?,?,?,?,?,?,?,?,?,?,?)",
			helper.obsTmpStatement()
		);
		assertEquals(
			"Prepare statement without schema",
			"INSERT INTO DS1_OBS_TMP"
			+ " (FREQUENCY,OBS_DATE,DIM2,DIM3,SERIES_KEY,OBS_VALUE,OBS_STATUS,OBS_CONF,OBS_PRE_BREAK,OBS_ATT2,OBS_COM,SEQ_ID,DT)"
			+ " VALUES (?,?,?,?,?,?,?,?,?,?,?,?,?)",
			new SDWLayout( KF1, "DS1", null ).obsTmpStatement()
		);

		final int width = helper.obsTmpWidth();
		assertEquals( "Column count", 13, width );

		{
			final Object[] parameters = new Object[width];
			assertEquals(
				"Extracted index",
				new int[]{ 12 },
				helper.obsTmpExtracted( parameters, testDate )
			);
			assertEquals(
				"Extracted parameters",
				new Object[]{
					null, null, null, null,
					null,
					null, null, null, null,
					null, null,
					null, testDate
				},
				parameters
			);
		}
		{
			final Object[] parameters = new Object[width];
			assertEquals(
				"Sequence number index",
				new int[]{ 11 },
				helper.obsTmpSequence( parameters, testInteger.intValue() )
			);
			assertEquals(
				"Sequence number parameters",
				new Object[]{
					null, null, null, null,
					null,
					null, null, null, null,
					null, null,
					testInteger, null
				},
				parameters
			);
		}
		{
			final Object[] parameters = new Object[width];
			assertEquals(
				"Null index",
				new int[]{ 5, 6, 7, 8, 9, 10 },
				helper.obsTmpNull( parameters )
			);
			assertEquals(
				"Null parameters",
				new Object[]{
					null, null, null, null,
					null,
					SDWLayout.NULL_DOUBLE,
					SDWLayout.NULL_STRING,
					SDWLayout.NULL_STRING,
					SDWLayout.NULL_DOUBLE,
					SDWLayout.NULL_STRING,
					SDWLayout.NULL_STRING,
					null, null
				},
				parameters
			);
		}
		{
			final Object[] parameters = new Object[width];
			assertEquals(
				"Key index",
				new int[]{ 0, 2, 3, 4 },
				helper.obsTmpKey( parameters, "M.DE.100" )
			);
			assertEquals(
				"Key parameters",
				new Object[]{
					"M", null, "DE", "100",
					"DS1.M.DE.100",
					null, null, null, null,
					null, null,
					null, null
				},
				parameters
			);
		}
		{
			final Object[] parameters = new Object[width];
			assertEquals(
				"Strange key index",
				new int[]{ 0, 2, 3, 4 },
				helper.obsTmpKey( parameters, ".." )
			);
			assertEquals(
				"Strange key parameters",
				new Object[]{
					"", null, "", "",
					"DS1...",
					null, null, null, null,
					null, null,
					null, null
				},
				parameters
			);
		}
		{
			final Object[] parameters = new Object[width];
			assertEquals(
				"Date index",
				new int[]{ 1 },
				helper.obsTmpDate( parameters, testDate )
			);
			assertEquals(
				"Date parameters",
				new Object[]{
					null, testDate, null, null,
					null,
					null, null, null, null,
					null, null,
					null, null
				},
				parameters
			);
		}
		{
			final Object[] parameters = new Object[width];
			assertEquals(
				"Full observation index",
				new int[]{ 5, 6, 7, 8 },
				helper.obsTmpCells(
					parameters,
					new Observation( "1.4", "A", "C", "1.9" )
				)
			);
			assertEquals(
				"Full observation parameters",
				new Object[]{
					null, null, null, null,
					null,
					new Double( 1.4 ), "A", "C", new Double( 1.9 ),
					null, null,
					null, null
				},
				parameters
			);
		}
		{
			final Object[] parameters = new Object[width];
			assertEquals(
				"Empty observation index",
				new int[]{ 5, 6, 7, 8 },
				helper.obsTmpCells( parameters, Observation.EMPTY_OBSERVATION )
			);
			assertEquals(
				"Empty observation parameters",
				new Object[]{
					null, null, null, null,
					null,
					SDWLayout.NULL_DOUBLE,
					SDWLayout.NULL_STRING,
					SDWLayout.NULL_STRING,
					SDWLayout.NULL_DOUBLE,
					null, null,
					null, null
				},
				parameters
			);
		}
		{
			final Object[] parameters = new Object[width];
			assertEquals(
				"Missing observation index",
				new int[]{ 5, 6, 7, 8 },
				helper.obsTmpCells(
					parameters,
					new Observation(
						Observation.MISSING_VALUE,
						"M", "F",
						Observation.MISSING_VALUE
					)
				)
			);
			assertEquals(
				"Missing observation parameters",
				new Object[]{
					null, null, null, null,
					null,
					new Double( Double.NaN ),
					"M", "F",
					new Double( Double.NaN ),
					null, null,
					null, null
				},
				parameters
			);
		}
		{
			final Object[] parameters = new Object[width];
			assertEquals(
				"Attribute index",
				new int[]{ 9 },
				helper.obsTmpAttribute(
					parameters,
					"OBS_ATT2",
					"Blah"
				)
			);
			assertEquals(
				"Missing observation parameters",
				new Object[]{
					null, null, null, null,
					null,
					null, null, null, null,
					"Blah", null,
					null, null
				},
				parameters
			);
		}
	}

	public void testSeriesTmp () throws ParseException {
		final SDWLayout helper = new SDWLayout( KF1, "DS1", "SCHEMA" );
		assertEquals(
			"Table name",
			"SCHEMA.DS1_SERIES_TMP",
			helper.seriesTmpName()
		);
		assertEquals(
			"Prepare statement",
			"INSERT INTO SCHEMA.DS1_SERIES_TMP"
			+ " (FREQUENCY,DIM2,DIM3,SER_ATT1,SER_ATT2,SG_ATT1,SG_ATT2,SEQ_ID,DT,SERIES_KEY,SIBLING_KEY)"
			+ " VALUES (?,?,?,?,?,?,?,?,?,?,?)",
			helper.seriesTmpStatement()
		);

		final int width = helper.seriesTmpWidth();
		assertEquals( "Column count", 11, width );

		{
			final Object[] parameters = new Object[width];
			assertEquals(
				"Extracted index",
				new int[]{ 8 },
				helper.seriesTmpExtracted( parameters, testDate )
			);
			assertEquals(
				"Extracted parameters",
				new Object[]{
					null, null, null,
					null, null, null, null,
					null, testDate,
					null, null
				},
				parameters
			);
		}
		{
			final Object[] parameters = new Object[width];
			assertEquals(
				"Extracted index",
				new int[]{ 7 },
				helper.seriesTmpSequence( parameters, testInteger.intValue() )
			);
			assertEquals(
				"Extracted parameters",
				new Object[]{
					null, null, null,
					null, null, null, null,
					testInteger, null,
					null, null
				},
				parameters
			);
		}
		{
			final Object[] parameters = new Object[width];
			assertEquals(
				"Null index",
				new int[]{ 3, 4, 5, 6 },
				helper.seriesTmpNull( parameters )
			);
			assertEquals(
				"Null parameters",
				new Object[]{
					null, null, null,
					SDWLayout.NULL_STRING,
					SDWLayout.NULL_STRING,
					SDWLayout.NULL_STRING,
					SDWLayout.NULL_STRING,
					null, null,
					null, null
				},
				parameters
			);
		}
		{
			final Object[] parameters = new Object[width];
			assertEquals(
				"Key index",
				new int[]{ 0, 1, 2, 9, 10 },
				helper.seriesTmpKey( parameters, "M.DE.100" )
			);
			assertEquals(
				"Key parameters",
				new Object[]{
					"M", "DE", "100",
					null, null, null, null,
					null, null,
					"DS1.M.DE.100", "DS1..DE.100"
				},
				parameters
			);
		}
		{
			final Object[] parameters = new Object[width];
			assertEquals(
				"Sibling key index",
				new int[]{ 0, 1, 2, 9, 10 },
				helper.seriesTmpKey( parameters, ".." )
			);
			assertEquals(
				"Sibling key parameters",
				new Object[]{
					SDWLayout.NULL_STRING, "", "",
					null, null, null, null,
					null, null,
					"DS1...", "DS1..."
				},
				parameters
			);
		}
		{
			final Object[] parameters = new Object[width];
			assertEquals(
				"Series attribute index",
				new int[]{ 3 },
				helper.seriesTmpAttribute(
					parameters,
					"SER_ATT1",
					"Foo"
				)
			);
			assertEquals(
				"Sibling attribute index",
				new int[]{ 5 },
				helper.seriesTmpAttribute(
					parameters,
					"SG_ATT1",
					"Bar"
				)
			);
			assertEquals(
				"Missing observation parameters",
				new Object[]{
					null, null, null,
					"Foo", null, "Bar", null,
					null, null,
					null, null
				},
				parameters
			);
		}
	}

	public void testDelTmp () throws ParseException {
		final SDWLayout helper = new SDWLayout( KF1, "DS1", "SCHEMA" );
		assertEquals(
			"Table name",
			"SCHEMA.DS1_DEL_TMP",
			helper.delTmpName()
		);
		assertEquals(
			"Prepare statement",
			"INSERT INTO SCHEMA.DS1_DEL_TMP"
			+ " (OBS_DATE,\"KEY\",ISSIBLING,SEQ_ID,DT)"
			+ " VALUES (?,?,?,?,?)",
			helper.delTmpStatement()
		);

		final int width = helper.delTmpWidth();
		assertEquals( "Column count", 5, width );

		{
			final Object[] parameters = new Object[width];
			assertEquals(
				"Extracted index",
				new int[]{ 4 },
				helper.delTmpExtracted( parameters, testDate )
			);
			assertEquals(
				"Extracted parameters",
				new Object[]{
					null,
					null, null,
					null, testDate
				},
				parameters
			);
		}
		{
			final Object[] parameters = new Object[width];
			assertEquals(
				"Sequence number index",
				new int[]{ 3 },
				helper.delTmpSequence( parameters, testInteger.intValue() )
			);
			assertEquals(
				"Sequence number parameters",
				new Object[]{
					null,
					null, null,
					testInteger, null
				},
				parameters
			);
		}
		{
			final Object[] parameters = new Object[width];
			assertEquals(
				"Null index",
				new int[]{ 0 },
				helper.delTmpNull( parameters )
			);
			assertEquals(
				"Null parameters",
				new Object[]{
					SDWLayout.NULL_DATE,
					null, null,
					null, null
				},
				parameters
			);
		}
		{
			final Object[] parameters = new Object[width];
			assertEquals(
				"Key index",
				new int[]{ 1, 2 },
				helper.delTmpKey( parameters, "M.DE.100" )
			);
			assertEquals(
				"Key parameters",
				new Object[]{
					null,
					"DS1.M.DE.100", Boolean.FALSE,
					null, null
				},
				parameters
			);
		}
		{
			final Object[] parameters = new Object[width];
			assertEquals(
				"Sibling key index",
				new int[]{ 1, 2 },
				helper.delTmpKey( parameters, ".." )
			);
			assertEquals(
				"Sibling key parameters",
				new Object[]{
					null,
					"DS1...", Boolean.TRUE,
					null, null
				},
				parameters
			);
		}
		{
			final Object[] parameters = new Object[width];
			assertEquals(
				"Date index",
				new int[]{ 0 },
				helper.delTmpDate( parameters, testDate )
			);
			assertEquals(
				"Date parameters",
				new Object[]{
					testDate,
					null, null,
					null, null
				},
				parameters
			);
		}
	}

	/* ==================== Private Methods ===================== */

	private void assertEquals(
		final String message,
		final int[] expected,
		final int[] actual
	) {
		assertEquals( message + " size", expected.length, actual.length );
		for ( int i = 0; i < expected.length; ++i ) {
			assertEquals(
				message + "[" + i + "]",
				expected[i],
				actual[i]
			);
		}
	}

	private void assertEquals(
		final String message,
		final Object[] expected,
		final Object[] actual
	) {
		assertEquals( message + " size", expected.length, actual.length );
		for ( int i = 0; i < expected.length; ++i ) {
			assertEquals(
				message + "[" + i + "]",
				expected[i],
				actual[i]
			);
		}
	}
}
