// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (C) 2007, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.sql;

import java.sql.Connection;
import java.sql.Statement;
import java.sql.CallableStatement;
import java.sql.SQLException;
import java.sql.Types;

import Int.ecb.gesmescb.util.ComponentHelper;

/* ***************************** Class ****************************** */

/**
 * Provides locking to avoid concurrent updates of a single data set in the SDW.
 *
 * @author
 * <a href="mailto:sis.external@ecb.int">Bernhard Bodenstorfer</A>
 *
 * @version @(#)$Id: SDWLocker.java,v 1.1 2017/01/19 16:14:02 trainot Exp $
 */
public class SDWLocker {

	/* ======================= Constants ======================== */

	/** The schema for stored procedures and functions. */
	private static final String SCHEMA = "datamgmt";

	/* ========================= Fields ========================= */

	/** Takes care of logging and preferences. */
	protected final ComponentHelper componentHelper;

	/** The connection to be used to establish and release the lock. */
	private final Connection connection;

	/** For synchronisation table column entry. */
	private final String schema, dataset, name;

	/* ====================== Constructors ====================== */

	/**
	 * Construct a <code>SDWLocker</code> to lock database updates.
	 * The database has to follow the SDW layout.
	 *
	 * @param connection to the database.
	 * @param schema for which the lock should be placed.
	 * @param dataset for which the lock should be placed.
	 * @param name to refer to the lock.
	 */
	public SDWLocker (
		final Connection connection,
		final String schema,
		final String dataset,
		final String name
	) throws SQLException {
		this.componentHelper = new ComponentHelper( name, this );
		this.connection = connection;
		this.schema = schema;
		this.dataset = dataset;
		this.name = name;

		executeAtomically(
			new ProcedureStrategy(
				SCHEMA + ".register_update('"
				+ schema + "','"
				+ dataset + "','"
				+ name
				+ "')"
			)
		);
	}

	/* ===================== Public Methods ===================== */

	/** Request a lock.
	 * The call returns immediately even if no lock could be obtained.
	 * @return whether a lock could be obtained.
	 */
	public boolean request () throws SQLException {
		FunctionStrategy functionStrategy
		= new FunctionStrategy(
			SCHEMA + ".request_update('"
			+ schema + "','"
			+ dataset + "','"
			+ name
			+ "')"
		);
		executeAtomically( functionStrategy );
		return "1".equals ( functionStrategy.getResult() );
	}

	/** Release a previously obtained lock. */
	public void release () throws SQLException {
		executeAtomically(
			new ProcedureStrategy(
				SCHEMA + ".unregister_update('"
				+ schema + "','"
				+ dataset + "','"
				+ name
				+ "')"
			)
		);
	}

	/* ==================== Private Methods ===================== */

	/** Lock synchronisation table, execute, unlock.
	 * The call waits until it can obtain a lock on the table.
	 */
	public void executeAtomically (
		final ConnectionStrategy strategy
	) throws SQLException {
		final boolean autoCommit = connection.getAutoCommit();
		try {
			connection.setAutoCommit( false );
			try {
				strategy.executeOn( connection );
				connection.commit();
			}
			catch ( final SQLException e ) {
				connection.rollback();
				throw e;
			}
		}
		finally {
			connection.setAutoCommit( autoCommit );
		}
	}
}

/* ***************************** Class ****************************** */

/** Runs code against a given JDBC {@link Connection}. */
interface ConnectionStrategy {
	void executeOn ( Connection connection ) throws SQLException;
}

class ProcedureStrategy implements ConnectionStrategy {
	private final String sqlString;

	public ProcedureStrategy ( final String sqlString ) {
		this.sqlString = sqlString;
	}

	public void executeOn ( final Connection connection ) throws SQLException {
		final Statement statement = connection.createStatement();
		statement.executeUpdate( "{ CALL " + sqlString + " }" );
		statement.close();
	}
}

class FunctionStrategy implements ConnectionStrategy {
	private final String sqlString;
    private String result;

	public FunctionStrategy ( final String sqlString ) {
		this.sqlString = sqlString;
	}

	public void executeOn ( final Connection connection ) throws SQLException {
		final CallableStatement callableStatement
		= connection.prepareCall(
			"{ ? = CALL " + sqlString + " }"
		);
		callableStatement.registerOutParameter( 1, Types.VARCHAR );
		callableStatement.executeUpdate();
	result = callableStatement.getString( 1 );
	callableStatement.close();
	}

	public String getResult () {
		return result;
	}
}
