// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (C) 2007, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.sql;

import java.util.List;
import java.util.ArrayList;

import java.sql.Connection;
import java.sql.SQLException;

import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;

/* ***************************** Class ****************************** */

/**
 * Tests the implementation of the class {@link SDWLocker}.
 *
 * @author
 * <a href="mailto:sis.external@ecb.int">Bernhard Bodenstorfer</a>
 *
 * @version @(#)$Id: SDWLockerTest.java,v 1.1 2017/01/19 16:14:02 trainot Exp $
 */

public class SDWLockerTest extends TestCase {

	/* ======================= Constants ======================== */

	/** Schema ID to be used in the database. */
	private static final String SCHEMA = "DATAMGMT";

	/** Creates tables to perform the test. */
	private static final String[]
		testPreparation = {
			"CREATE SCHEMA " + SCHEMA,
			"SET SCHEMA " + SCHEMA,
			"CREATE PROCEDURE register_update ("
			+ " IN iSchema VARCHAR(8),"
			+ " IN iDataset VARCHAR(8),"
			+ " IN iProcessId VARCHAR(120)"
			+ " )"
			+ " PARAMETER STYLE JAVA"
			+ " LANGUAGE JAVA"
			+ " EXTERNAL NAME '"
			+ SDWLockerTest.class.getName()
			+ ".registerUpdate'",
			"CREATE FUNCTION request_update ("
			+ " iSchema VARCHAR(8),"
			+ " iDataset VARCHAR(8),"
			+ " iProcessId VARCHAR(120)"
			+ " ) RETURNS VARCHAR(3) "
			+ " PARAMETER STYLE JAVA"
			+ " LANGUAGE JAVA"
			+ " EXTERNAL NAME '"
			+ SDWLockerTest.class.getName()
			+ ".requestUpdate'",
			"CREATE PROCEDURE unregister_update ("
			+ " IN iSchema VARCHAR(8),"
			+ " IN iDataset VARCHAR(8),"
			+ " IN iProcessId VARCHAR(120)"
			+ " )"
			+ " PARAMETER STYLE JAVA"
			+ " LANGUAGE JAVA"
			+ " EXTERNAL NAME '"
			+ SDWLockerTest.class.getName()
			+ ".unregisterUpdate'"
		},
		testPostparation = {
			"SET SCHEMA " + SCHEMA,
			"DROP PROCEDURE register_update",
			"DROP FUNCTION request_update",
			"DROP PROCEDURE unregister_update",
			"DROP SCHEMA " + SCHEMA + " RESTRICT"
		};

	/* ========================= Fields ========================= */

	private final SQLTestHelper testHelper;

	private static final List registerUpdateCalls = new ArrayList();
	private static final List requestUpdateCalls = new ArrayList();
	private static final List unregisterUpdateCalls = new ArrayList();
	private static String requestUpdateResult;

	/* ====================== Constructors ====================== */

	/**
	 * Constructs an SDWLockerTest with the specified name.
	 *
	 * @param name Test case name
	 */
	public SDWLockerTest ( final String name ) {
		super( name );
		testHelper = new SQLTestHelper( "sdwLocker" );
	}

	/* ================== Protected Methods ===================== */

	protected void setUp () throws SQLException {
		testHelper.setUp();
		testHelper.executeStatements( testPreparation, false );
	}

	protected void tearDown () throws SQLException {
		testHelper.executeStatements( testPostparation, false );
		testHelper.tearDown();
	}

	/* ===================== Public Methods ===================== */

	/**
	 * Assembles and returns a test suite for all
	 * the test methods of this test case.
	 *
	 * @return a non-null test suite
	 */
	public static Test suite () {
		return new TestSuite( SDWLockerTest.class );
	}

	/** Run the test case. */
	public static void main ( final String args[] ) {
		final String[] testCaseName = { SDWLockerTest.class.getName() };
		junit.textui.TestRunner.main( testCaseName );
	}

	/* ---------------- Auxiliary methods ----------------- */

	public static void registerUpdate (
		final String iSchema,
		final String iDataset,
		final String iProcessId
	) {
		registerUpdateCalls.add(
			new String[]{ iSchema, iDataset, iProcessId }
		);
	}

	public static String requestUpdate (
		final String iSchema,
		final String iDataset,
		final String iProcessId
	) {
		requestUpdateCalls.add(
			new String[]{ iSchema, iDataset, iProcessId }
		);
		return requestUpdateResult;
	}

	public static void unregisterUpdate (
		final String iSchema,
		final String iDataset,
		final String iProcessId
	) {
		unregisterUpdateCalls.add(
			new String[]{ iSchema, iDataset, iProcessId }
		);
	}

	public static void assertEquals (
		final String message,
		final String[] expected,
		final String[] actual
	) {
		assertEquals( message + ": length", expected.length, actual.length );
		for ( int i = 0; i < expected.length; ++i ) {
			assertEquals( message + "[" + i + "]", expected[i], actual[i] );
		}
	}

	/* ---------------------- Test Methods ---------------------- */

	/** Scenario with no competition at all. */
	public void testCycle () throws SQLException {
		final Connection connection = testHelper.createConnection();

		registerUpdateCalls.clear();
		requestUpdateCalls.clear();
		unregisterUpdateCalls.clear();

		final SDWLocker locker = new SDWLocker(
			connection, "SDW_WEB", "BLS", "Magda"
		);
		assertEquals(
			"Parameters to registerUpdate",
			new String[]{ "SDW_WEB", "BLS", "Magda" },
			(String[]) registerUpdateCalls.get( 0 )
		);

		assertEquals(
			"No call to requestUpdate",
			0, requestUpdateCalls.size()
		);
		requestUpdateResult = "553";
		assertTrue( "Cannot get lock", !locker.request() );
		assertEquals(
			"Parameters to requestUpdate",
			new String[]{ "SDW_WEB", "BLS", "Magda" },
			(String[]) requestUpdateCalls.get( 0 )
		);
		assertEquals(
			"One call to requestUpdate",
			1, requestUpdateCalls.size()
		);
		requestUpdateResult = "1";
		assertTrue( "Can get lock", locker.request() );
		assertEquals(
			"Parameters to requestUpdate",
			new String[]{ "SDW_WEB", "BLS", "Magda" },
			(String[]) requestUpdateCalls.get( 0 )
		);

		assertEquals(
			"No call to unregisterUpdate",
			0, unregisterUpdateCalls.size()
		);
		locker.release();
		assertEquals(
			"Parameters to unregisterUpdate",
			new String[]{ "SDW_WEB", "BLS", "Magda" },
			(String[]) requestUpdateCalls.get( 0 )
		);

		assertEquals(
			"One call to registerUpdate",
			1, registerUpdateCalls.size()
		);
		assertEquals(
			"Two calls to requestUpdate",
			2, requestUpdateCalls.size()
		);
		assertEquals(
			"One call to unregisterUpdate",
			1, unregisterUpdateCalls.size()
		);

		connection.close();
	}
}
