// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (C) 2007, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.sql;

import java.util.Date;
import java.util.Map;
import java.util.TreeMap;
import java.util.Iterator;

import java.text.DateFormat;
import java.text.SimpleDateFormat;

import java.sql.Connection;
import java.sql.Statement;
import java.sql.PreparedStatement;
import java.sql.CallableStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.SQLWarning;
import java.sql.Types;

import Int.ecb.logging.Level;

import Int.ecb.gesmescb.InterchangeBegin;
import Int.ecb.gesmescb.MessageBegin;
import Int.ecb.gesmescb.KeyFamily;
import Int.ecb.gesmescb.KeyFamilyComponent;
import Int.ecb.gesmescb.TimeSeries;
import Int.ecb.gesmescb.Observation;
import Int.ecb.gesmescb.Attribute;
import Int.ecb.gesmescb.MessageEnd;
import Int.ecb.gesmescb.InterchangeEnd;

import Int.ecb.gesmescb.metabase.Metabase;
import Int.ecb.gesmescb.metabase.MaintenanceAgency;

import Int.ecb.gesmescb.event.DataListener;
import Int.ecb.gesmescb.event.GESMESCBWriter;
import Int.ecb.gesmescb.event.GESMESCBEventListenerAdapter;

import Int.ecb.gesmescb.time.RegularInterval;

import Int.ecb.gesmescb.util.ComponentHelper;
import Int.ecb.gesmescb.util.Named;

import Int.ecb.gesmescb.error.ErrorHandler;

import Int.ecb.logging.LogManager;

/* ***************************** Class ****************************** */

/**
 * <p>
 * Writes {@link Int.ecb.gesmescb.GESMESCBObject}s to the ECB Statistical Data
 * Warehouse (SDW) database.
 * </p>
 *
 * <center><table border="3" cellpadding="3"> <caption> <a
 * name="prefs">Available {@link Int.ecb.prefs preference} settings</a>
 * </caption>
 * <tr>
 * <th>preference name</th>
 * <th>type</th>
 * <th>meaning</th>
 * <th>default</th>
 * <tr>
 *	<td><code>driver</td><td>string</td>
 *	<td>Java class name of the JDBC driver to be used</td>
 * </tr>
 * <tr>
 *  <td><code>destination</td><td>string</td>
 *	<td>URL of the database to which to write the data</td>
 * </tr>
 * <tr>
 *	<td><code>user</td><td>string</td>
 *	<td>user id for accessing the database</td>
 * </tr>
 * <tr>
 *	<td><code>password</td><td>string</td>
 *	<td>password for accessing the database</td>
 * </tr>
 * <tr>
 *	<td><code>schema</td><td>string</td>
 *	<td>schema owner of the relevant database tables</td>
 * </tr>
 * <tr>
 *	<td><code>dataVersion</td><td>string</td>
 *	<td>schema owner of the relevant database tables</td>
 *	<td><code>NONE</code></td>
 * </tr>
 * <tr>
 *	<td><code>refreshLevel</td><td>string</td>
 *	<td>one of:
 *  <code>value</code>, <code>series</code>, <code>dataset</code>
 *  </td>
 *	<td><code>value</code></td>
 * </tr>
 * <tr>
 *	<td><code>processId</td><td>string</td>
 *	<td>to be used to name the lock to guard against
 *	concurrent writing by SDWWritery;
 *	it must be unique among all running writing activities to SDW;
 *	best concatenate: hostname, pid, SDWWriter name,
 *	data source name (e.g. filename)</td>
 * </tr>
 * <tr>
 *	<td><code>maxLockAttempts</td><td>integer</td>
 *	<td>bounds the number of attempts to request access
 *	to the _TMP tables; 0 means unbounded</td>
 *	<td><code>0</code></td>
 * </tr>
 * <tr>
 *	<td><code>lockWaitingTime</td><td>decimal</td>
 *	<td>the time between attempts to request access
 *	to the _TMP tables in seconds</td>
 *	<td><code>60.0</code></td>
 * </tr>
 * <tr>
 *	<td><code>wipeTmp</td><td>string</td>
 *	<td>whether the _TMP load tables should be cleared before loading;
 *  otherwise checked and required to be empty beforehand</td>
 *	<td><code>false</code></td>
 * </tr>
 * <tr>
 *	<td><code>queryTimeout</td><td>integer</td>
 *	<td>How long to wait for the stored procedure
 *	<code>sp_integrate_changes</code></td>
 *	<td>JDBC implementation default</td>
 * </tr>
 * </table></center>
 *
 * @author
 * <a href="mailto:sis.external@ecb.int">Bernhard Bodenstorfer</A>
 *
 * @version @(#)$Id: SDWWriter.java,v 1.1 2017/01/19 16:14:02 trainot Exp $
 */
public class SDWWriter extends GESMESCBEventListenerAdapter implements Named,
		DataListener, GESMESCBWriter {

	/* ======================= Constants ======================== */

	/** Special value mark attributes for deletion. */
	private static final String
		MARK_DELETE = "~",
		ISO_TIME_FORMAT = "yyyy-MM-dd'T'HH:mm:ss";

	/* ========================= Fields ========================= */

	/** Takes care of logging and preferences. */
	protected final ComponentHelper componentHelper;

	/** Data target schema (SQL database schema). */
	private final String schema;

	/** Data source version (SDPE channel). */
	private final String dataVersion;

	/** Data refresh level for clustered data refreshment. */
	private final int refreshLevel;

	/** Lock identifier. */
	private final String processId;

	/** Maximal number of attempts to get a lock granted. */
	private final int maxLockAttempts;

	/** Time in seconds the writer waits to get a lock. */
	private final float lockWaitingTime;

	/** Whether _TMP table content should be removed, if present. */
	private final boolean wipeTmp;

	/** Waiting timeout for stored procedure in seconds. 0 means session default. */
	private final int queryTimeout;

	/** Takes care of database connectivity. */
	private final SQLConnector connector;

	/** Whether interchange is to be ignored. */
	private boolean isTestInterchange;

	/** Watches the log event bus. */
	private final ErrorHandler errorHandler;

	/** Whether message action is delete. */
	private boolean isDeleteMessage;

	/** Current key family. */
	private KeyFamily keyFamily;

	/** Stores info about all DSIs referenced in the interchange. */
	private final Map dataSetInfos;

	/** Message extraction timestamp; used for embargo. */
	private Date extracted;

	/** Sequential linear number of the update/delete table entry. */
	private int seqId;

	/** Helps with SDW table layout. */
	private SDWLayout layout;

	/** Prepared statements to optimise database access. */
	private PreparedStatement
		obsTmpStatement, seriesTmpStatement, delTmpStatement;

	/** Callable statement to start data integration procedure. */
	private final CallableStatement integrateChangesStatement;

	/* ====================== Constructors ====================== */

	/**
	 * Construct a <code>SDWWriter</code> to write to a database. The database
	 * has to follow the SDW layout.
	 *
	 * @param name
	 *	      used for logging and preferences.
	 * @param database
	 *	      URL of the database where to write to.
	 *
	 * @throws NullPointerException
	 *	       if <code>database==null</code> and the preference variable
	 *	       <code>destination</code> is undefined.
	 * @throws SQLException
	 *	       if the database cannot be opened for writing.
	 */
	public SDWWriter ( final String name, final String database )
		throws SQLException
	{
		componentHelper = new ComponentHelper( name, this );
		schema = componentHelper.getPreference( "schema", null );
		if ( null == schema ) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"SDWWriter()",
				"Target database schema not given in configuration."
				+ " Set the configuration preference \"schema\"."
			);
		}
		dataVersion = componentHelper.getPreference( "dataVersion", "NONE" );
		final String refreshSetting = componentHelper.getPreference(
			"refreshLevel", "value"
		);
		if ( "value".equalsIgnoreCase( refreshSetting ) ) {
			refreshLevel = 0;
		} else if ( "series".equalsIgnoreCase( refreshSetting ) ) {
			refreshLevel = 2;
		} else if ( "dataset".equalsIgnoreCase( refreshSetting ) ) {
			refreshLevel = 1;
		} else {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"SDWWriter()",
				"Unknown refresh level \"" + refreshSetting
				+ "\" should be one of: \"value\", \"series\", \"dataset\"."
			);
			refreshLevel = 0;
		}

		wipeTmp = "true".equalsIgnoreCase(
			componentHelper.getPreference( "wipeTmp", "false" )
		);

		final String queryTimeoutString
		= componentHelper.getPreference( "queryTimeout", null );
		if ( null == queryTimeoutString ) {
			queryTimeout = 0;
		} else {
			int queryTimeoutInt = 0;
			try {
				queryTimeoutInt = Integer.parseInt( queryTimeoutString );
				if ( queryTimeoutInt < 0 ) {
					componentHelper.logger.log(
						Level.SEVERE,
						componentHelper.logName,
						"SDWWriter()",
						"Nonpositive queryTimeout value " + queryTimeoutString
						+ " in configuration."
					);
				}
			}
			catch ( final NumberFormatException e ) {
				componentHelper.logger.log(
					Level.SEVERE,
					componentHelper.logName,
					"SDWWriter()",
					"Invalid queryTimeout value " + queryTimeoutString
					+ " in configuration: " + e,
					e
				);
			}
			queryTimeout = queryTimeoutInt;
		}

		processId
		= componentHelper.getPreference( "processId", null );
		if ( null == processId ) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"SDWWriter()",
				"No preference set for processId."
			);
		}

		final String maxLockAttemptsString
		= componentHelper.getPreference( "maxLockAttempts", "0" );
		{
			int maxLockAttemptsInt = 0;
			try {
				maxLockAttemptsInt = Integer.parseInt(
					maxLockAttemptsString
				);
				if ( maxLockAttemptsInt < 0 ) {
					componentHelper.logger.log(
						Level.SEVERE,
						componentHelper.logName,
						"SDWWriter()",
						"Negative maxLockAttempts value "
						+ maxLockAttemptsString
						+ " in configuration."
					);
				}
			}
			catch ( final NumberFormatException e ) {
				componentHelper.logger.log(
					Level.SEVERE,
					componentHelper.logName,
					"SDWWriter()",
					"Invalid maxLockAttempts value " + maxLockAttemptsString
					+ " in configuration: " + e,
					e
				);
			}
			maxLockAttempts = maxLockAttemptsInt;
		}

		final String lockWaitingTimeString
		= componentHelper.getPreference( "lockWaitingTime", "60.0" );
		{
			float lockWaitingTimeFloat = 0;
			try {
				lockWaitingTimeFloat = Float.parseFloat(
					lockWaitingTimeString
				);
				if ( lockWaitingTimeFloat < 0.001 ) {
					componentHelper.logger.log(
						Level.SEVERE,
						componentHelper.logName,
						"SDWWriter()",
						"Too small lockWaitingTime value "
						+ lockWaitingTimeString
						+ " in configuration."
					);
				}
			}
			catch ( final NumberFormatException e ) {
				componentHelper.logger.log(
					Level.SEVERE,
					componentHelper.logName,
					"SDWWriter()",
					"Invalid lockWaitingTime value " + lockWaitingTimeString
					+ " in configuration: " + e,
					e
				);
			}
			lockWaitingTime = lockWaitingTimeFloat;
		}

		dataSetInfos = new TreeMap();
		connector = new SQLConnector( componentHelper, database, true );
		final Connection connection = connector.getConnection();
		integrateChangesStatement = connection.prepareCall(
			"{CALL datamgmt.sp_integrate_changes(?,?,?,?,?,?)}"
		);
		integrateChangesStatement.registerOutParameter(5,Types.VARCHAR);
		integrateChangesStatement.setQueryTimeout( queryTimeout );

		LogManager.getLogManager().addGlobalHandler(
			errorHandler = new ErrorHandler()
		);

		seqId = 0;
	}

	/**
	 * Construct a <code>SDWWriter</code> writing to a database according to
	 * preferences.
	 *
	 * @param name
	 *	      used for logging and preferences.
	 *
	 * @throws NullPointerException
	 *	       if the preference variable <code>destination</code> is
	 *	       undefined.
	 * @throws SQLException
	 *	       if the database cannot be opened for writing.
	 */
	public SDWWriter ( final String name ) throws SQLException {
		this( name, null );
	}

	/* ===================== Public Methods ===================== */

	public String getName () {
		return componentHelper.name;
	}

	/* ------------------------ Listener ------------------------ */

	/* ---------------------- Interchange ----------------------- */

	public void handle ( final InterchangeBegin interchangeBegin ) {
		componentHelper.logger.log(
			Level.INFO,
			componentHelper.logName,
			"handle( InterchangeBegin )",
			"Begin writing interchange " + interchangeBegin.getIref() + "."
		);
		isTestInterchange = interchangeBegin.isTest();
	}

	public void handle ( final InterchangeEnd interchangeEnd ) {
		componentHelper.logger.log(
			Level.INFO,
			componentHelper.logName,
			"handle( InterchangeEnd )",
			"Ended writing interchange " + interchangeEnd.getIref() + "."
		);
	}

	/* ------------------------ Message ------------------------- */

	public void handle ( final MessageBegin messageBegin ) {
		if ( !messageBegin.isStatisticalData() ) {
			componentHelper.logger.log(
				Level.INFO,
				componentHelper.logName,
				"handle( MessageBegin )",
				"Ignoring non-data message " + messageBegin.getMref() + "."
			);
		} else {
			componentHelper.logger.log(
				Level.INFO,
				componentHelper.logName,
				"handle( MessageBegin )",
				"Begin writing data Message " + messageBegin.getMref() + "."
			);

			final Metabase metabase = Metabase.getMetabase();
			final MaintenanceAgency maintenanceAgency
			= metabase.getMaintenanceAgency(
				messageBegin.getMaintenanceAgencyId()
			);
			keyFamily = maintenanceAgency.getKeyFamily(
				messageBegin.getKeyFamilyId()
			);
			final String dsi = messageBegin.getDataSetId();
			if ( null == extracted ) {
				extracted = messageBegin.getExtractionDateTimeAsDate();
			}
			isDeleteMessage = messageBegin.isDeleteAction();

			try {
				final Connection connection = connector.getConnection();
				if ( ! dataSetInfos.containsKey( dsi ) ) {
					layout = new SDWLayout( keyFamily, dsi, schema );
					final SDWLocker locker = new SDWLocker(
						connection,
						schema,
						dsi,
						processId
					);

					boolean lockGranted = false;
					try {
						for (
							int i = 0;
							(
								0 == maxLockAttempts
								||
								i < maxLockAttempts
							)
							&&
							! ( lockGranted = locker.request() );
							++i
						) {
							Thread.sleep( (int) ( 1000 * lockWaitingTime ) );
						}
					} catch ( final InterruptedException e ) {
						componentHelper.logger.log(
							Level.SEVERE,
							componentHelper.logName,
							"handle(MessageBegin)",
							"Received Thread interrupt request while waiting for lock: "
							+ e,
							e
						);
					}
					if ( lockGranted ) {
						final Statement statement = connection.createStatement();
						statement.setQueryTimeout( queryTimeout );
						final String[] table = {
							layout.obsTmpName(),
							layout.seriesTmpName(),
							layout.delTmpName()
						};
						for ( int i = 0; i < table.length; ++i ) {
							final ResultSet inspectResult = statement.executeQuery(
								"SELECT COUNT(*) FROM " + table[i]
							);
							if ( ! inspectResult.next() ) {
								componentHelper.logger.log(
									Level.SEVERE,
									componentHelper.logName,
									"handle( MessageBegin )",
									"Inspecting table " + table[i] + " failed."
								);
							}
							final int rowCount = inspectResult.getInt( 1 );
							if ( 0 != rowCount ) {
								componentHelper.logger.log(
									wipeTmp ? Level.WARNING : Level.SEVERE,
									componentHelper.logName,
									"handle( MessageBegin )",
									"Dirty table " + table[i] + " contains "
									+ rowCount
									+ " rows when it should be expected to be empty."
								);
								if ( wipeTmp ) {
									componentHelper.logger.log(
										Level.WARNING,
										componentHelper.logName,
										"handle( MessageBegin )",
										"Wiping dirty table " + table[i] + "."
									);
									statement.executeUpdate(
										"DELETE FROM " + table[i]
									);
								} else {
									layout = null;	// Prevent table access
								}
							}
							inspectResult.close();
						}

						statement.close();
					} else {
						componentHelper.logger.log(
							Level.SEVERE,
							componentHelper.logName,
							"handle( MessageBegin )",
							"Could not get a lock for schema \""
							+ schema + "\", dataset \"" + dsi + "\"."
						);
						layout = null;
					}

					dataSetInfos.put( dsi, new DataSetInfo( layout, locker ) );
				} else {
					layout = (
						(DataSetInfo) dataSetInfos.get( dsi )
					).getLayout();
				}
				if ( null != layout ) {
					obsTmpStatement = connection.prepareStatement(
						layout.obsTmpStatement()
					);
					obsTmpStatement.setQueryTimeout( queryTimeout );
					seriesTmpStatement = connection.prepareStatement(
						layout.seriesTmpStatement()
					);
					seriesTmpStatement.setQueryTimeout( queryTimeout );
					delTmpStatement = connection.prepareStatement(
						layout.delTmpStatement()
					);
					delTmpStatement.setQueryTimeout( queryTimeout );
				} else {
					// Prevent any insertion into tables for this dataset
					obsTmpStatement = seriesTmpStatement = delTmpStatement
					= null;
				}
			}
			catch ( final SQLException e ) {
				componentHelper.logger.log(
					Level.SEVERE,
					componentHelper.logName,
					"handle( MessageBegin )",
					"Inspecting tables or preparing statements failed: " + e,
					e
				);
			}
		}
	}

	public void handle ( final MessageEnd messageEnd ) {
		componentHelper.logger.log(
			Level.FINER,
			componentHelper.logName,
			"handle( MessageEnd )",
			"End message " + messageEnd.getMref() + "."
		);
		try {
			if ( null != obsTmpStatement ) obsTmpStatement.close();
			if ( null != seriesTmpStatement ) seriesTmpStatement.close();
			if ( null != delTmpStatement ) delTmpStatement.close();
		}
		catch ( final SQLException e ) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"handle( MessageEnd )",
				"Closing statements failed: " + e,
				e
			);
		}
	}

	/* ---------------------- Time Series ----------------------- */

	public void handle ( final TimeSeries timeSeries ) {
		if ( null == layout ) {
			componentHelper.logger.log(
				Level.FINE,
				componentHelper.logName,
				"handle( TimeSeries )",
				"Ignoring series " + timeSeries
				+ " related to dirty table.",
				timeSeries
			);
			return;
		}

		final String key = timeSeries.getKey();
		final RegularInterval range = timeSeries.getInterval();
		try {
			if ( isDeleteMessage ) {
				final Object[] parameters = new Object[ layout.delTmpWidth() ];
				setParameters(
					delTmpStatement,
					layout.delTmpExtracted( parameters, extracted ),
					parameters
				);
				setParameters(
					delTmpStatement,
					layout.delTmpSequence( parameters, ++seqId ),
					parameters
				);
				setParameters(
					delTmpStatement,
					layout.delTmpKey( parameters, key ),
					parameters
				);
				if ( null == range ) {
					setParameters(
						delTmpStatement,
						layout.delTmpNull( parameters ),
						parameters
					);
					final int resultCode = delTmpStatement.executeUpdate();
					if ( 1 != resultCode ) {
						componentHelper.logger.log(
							Level.SEVERE,
							componentHelper.logName,
							"handle( TimeSeries )",
							"Delete time series "
							+ key + " failed."
						);
					}
				} else {
					for (
						final Iterator periods = range.iterator();
						periods.hasNext();
					) {
						final RegularInterval period
						= (RegularInterval) periods.next();
						setParameters(
							delTmpStatement,
							layout.delTmpDate( parameters, period.getBegin() ),
							parameters
						);
						delTmpStatement.addBatch();
					}
					final int[] resultCodes = delTmpStatement.executeBatch();
					for ( int i = 0; i < resultCodes.length; ++i ) {
						if ( Statement.EXECUTE_FAILED == resultCodes[i] ) {
							componentHelper.logger.log(
								Level.SEVERE,
								componentHelper.logName,
								"handle( TimeSeries )",
								"Delete " + key + " observation index "
								+ i + " failed."
							);
						}
					}
				}
			} else {
				final Object[] parameters = new Object[ layout.obsTmpWidth() ];
				setParameters(
					obsTmpStatement,
					layout.obsTmpExtracted( parameters, extracted ),
					parameters
				);
				setParameters(
					obsTmpStatement,
					layout.obsTmpSequence( parameters, ++seqId ),
					parameters
				);
				setParameters(
					obsTmpStatement,
					layout.obsTmpKey( parameters, key ),
					parameters
				);
				setParameters(
					obsTmpStatement,
					layout.obsTmpNull( parameters ),
					parameters
				);
				for (
					final Iterator
						periods = range.iterator(),
						observations = timeSeries.observationIterator();
					periods.hasNext() && observations.hasNext();
				) {
					final RegularInterval period
					= (RegularInterval) periods.next();
					final Observation observation
					= (Observation) observations.next();
					if ( observation.isEmpty() ) continue;
					setParameters(
						obsTmpStatement,
						layout.obsTmpDate( parameters, period.getBegin() ),
						parameters
					);
					setParameters(
						obsTmpStatement,
						layout.obsTmpCells( parameters, observation ),
						parameters
					);
					obsTmpStatement.addBatch();
				}
				final int[] resultCodes = obsTmpStatement.executeBatch();
				for ( int i = 0; i < resultCodes.length; ++i ) {
					if ( Statement.EXECUTE_FAILED == resultCodes[i] ) {
						componentHelper.logger.log(
							Level.SEVERE,
							componentHelper.logName,
							"handle( TimeSeries )",
							"Update " + key + " observation index "
							+ i + " failed."
						);
					}
				}
			}
		}
		catch ( final Exception e ) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"handle( TimeSeries )",
				( isDeleteMessage ? "Delete" : "Update" )
				+ " " + key
				+ ( null == range ? "" : "[" + range + "]" )
				+ " failed: " + e
			);
		}
	}

	/* ----------------------- Attribute ------------------------ */

	public void handle ( final Attribute attribute ) {
		if ( null == layout ) {
			componentHelper.logger.log(
				Level.FINE,
				componentHelper.logName,
				"handle( Attribute )",
				"Ignoring attribute " + attribute
				+ " related to dirty table.",
				attribute
			);
			return;
		}

		final String key = attribute.getKey();
		final RegularInterval range = attribute.getInterval();
		final String name = attribute.getName();
		final String value = isDeleteMessage
			? MARK_DELETE : attribute.getValue();
		final KeyFamilyComponent component = keyFamily.getComponent( name );
		try {
			if ( component.isAttachedToObservation() ) {
				final Object[] parameters = new Object[ layout.obsTmpWidth() ];
				setParameters(
					obsTmpStatement,
					layout.obsTmpExtracted( parameters, extracted ),
					parameters
				);
				setParameters(
					obsTmpStatement,
					layout.obsTmpSequence( parameters, ++seqId ),
					parameters
				);
				setParameters(
					obsTmpStatement,
					layout.obsTmpKey( parameters, key ),
					parameters
				);
				setParameters(
					obsTmpStatement,
					layout.obsTmpNull( parameters ),
					parameters
				);
				for (
					final Iterator periods = range.iterator();
					periods.hasNext();
				) {
					final RegularInterval period = (RegularInterval) periods.next();
					setParameters(
						obsTmpStatement,
						layout.obsTmpDate( parameters, period.getBegin() ),
						parameters
					);
					setParameters(
						obsTmpStatement,
						layout.obsTmpAttribute( parameters, name, value ),
						parameters
					);
					obsTmpStatement.addBatch();
				}
				final int[] resultCodes = obsTmpStatement.executeBatch();
				for ( int i = 0; i < resultCodes.length; ++i ) {
					if ( Statement.EXECUTE_FAILED == resultCodes[i] ) {
						componentHelper.logger.log(
							Level.SEVERE,
							componentHelper.logName,
							"handle( Attribute )",
							"Update " + key + "[" + range + "]."
							+ name + " attribute "
							+ i + " failed."
						);
					}
				}
			} else if (
				component.isAttachedToTimeSeries()
				||
				component.isAttachedToSiblingGroup()
			) {
				final Object[] parameters
				= new Object[ layout.seriesTmpWidth() ];
				setParameters(
					seriesTmpStatement,
					layout.seriesTmpExtracted( parameters, extracted ),
					parameters
				);
				setParameters(
					seriesTmpStatement,
					layout.seriesTmpSequence( parameters, ++seqId ),
					parameters
				);
				setParameters(
					seriesTmpStatement,
					layout.seriesTmpKey( parameters, key ),
					parameters
				);
				setParameters(
					seriesTmpStatement,
					layout.seriesTmpNull( parameters ),
					parameters
				);
				setParameters(
					seriesTmpStatement,
					layout.seriesTmpAttribute( parameters, name, value ),
					parameters
				);
				final int resultCode = seriesTmpStatement.executeUpdate();
				if ( 1 != resultCode ) {
					componentHelper.logger.log(
						Level.SEVERE,
						componentHelper.logName,
						"handle( Attribute )",
						"Update attribute "
						+ key + "." + name + " failed."
					);
				}
			}
		}
		catch ( final Exception e ) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"handle( Attribute )",
				"Update " + key
				+ ( null == range ? "" : "[" + range + "]" )
				+ " failed: " + e
			);
		}
	}

	/* ------------------------- Close -------------------------- */

	/** Close the <code>SDWWriter</code>. */
	public void close () {
		try {
			final Statement clearStatement
			= connector.getConnection().createStatement();
			clearStatement.setQueryTimeout( queryTimeout );
			final DateFormat format = new SimpleDateFormat(
				ISO_TIME_FORMAT
			);
			final boolean commit = ! isTestInterchange
				&&
				isActive()
				&&
				errorHandler.getSeverity() <= ErrorHandler.WARNING
			;

			int i = 0;
			for (
				final Iterator iterator = dataSetInfos.entrySet().iterator();
				iterator.hasNext();
				++i
			) {
				final Map.Entry entry = (Map.Entry) iterator.next();
				final String dsi = (String) entry.getKey();
				final DataSetInfo info = (DataSetInfo) entry.getValue();
				final SDWLayout layout = info.getLayout();
				final SDWLocker locker = info.getLocker();
				if ( null == layout ) {
					componentHelper.logger.log(
						Level.FINE,
						componentHelper.logName,
						"close()",
						"Skipping " + dsi + " related to dirty tables."
					);
				} else if ( commit ) {
					final String releaseDate = format.format( extracted );
					integrateChangesStatement.setString( 1, schema );
					integrateChangesStatement.setString( 2, dsi );
					integrateChangesStatement.setInt( 3, refreshLevel );
					integrateChangesStatement.setString( 4, dataVersion );
					integrateChangesStatement.setString( 5, releaseDate );
					integrateChangesStatement.setInt( 6, i );
					componentHelper.logger.log(
						Level.FINE,
						componentHelper.logName,
						"close()",
						"Call stored procedure to integrate " + dsi + " data."
					);
					componentHelper.logger.log(
						Level.FINER,
						componentHelper.logName,
						"close()",
						"sp_integrate_changes( "
						+ schema
						+ ", "
						+ dsi
						+ ", "
						+ refreshLevel
						+ ", "
						+ dataVersion
						+ ", "
						+ releaseDate
						+ ", "
						+ i
						+ " )"
					);
					integrateChangesStatement.execute();
					componentHelper.logger.log(
						Level.FINE,
						componentHelper.logName,
						"close()",
						"Consolidated " + dsi + " release time: "
						+ integrateChangesStatement.getString( 5 )
					);
					logWarnings( "close()", integrateChangesStatement );
				} else {
					clearStatement.addBatch(
						"DELETE FROM " + layout.obsTmpName()
					);
					clearStatement.addBatch(
						"DELETE FROM " + layout.seriesTmpName()
					);
					clearStatement.addBatch(
						"DELETE FROM " + layout.delTmpName()
					);
					componentHelper.logger.log(
						Level.FINE,
						componentHelper.logName,
						"close()",
						"No integration to happen, removing " + dsi
						+ " data from loading area."
					);
					clearStatement.executeBatch();
				}
				locker.release();
				iterator.remove();
			}
			clearStatement.close();
			integrateChangesStatement.close();
		}
		catch ( final SQLException e ) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"close()",
				"Callable statement failed: " + e,
				e
			);
			e.printStackTrace();
		}
		connector.close();
	}

	/* ==================== Private Methods ===================== */

	/** Use indices and parameters as delivered by {@link SDWLayout}. */
	private void setParameters (
		final PreparedStatement statement,
		final int[] indices,
		final Object[] parameters
	) throws SQLException {
		for ( int i = 0; i < indices.length; ++i ) {
			final int
				index = indices[i],
				index1 = index + 1;
			if ( parameters[index] instanceof String ) {
				statement.setString( index1, (String) parameters[index] );
			} else if ( parameters[index] instanceof Integer ) {
				statement.setInt(
					index1,
					( (Integer) parameters[index] ).intValue()
				);
			} else if ( parameters[index] instanceof Double ) {
				statement.setDouble(
					index1,
					( (Double) parameters[index] ).doubleValue()
				);
			} else if ( parameters[index] instanceof Date ) {
				statement.setTimestamp(
					index1,
					new java.sql.Timestamp(
						( (Date) parameters[index] ).getTime()
					)
				);
			} else if ( parameters[index] instanceof Boolean ) {
				statement.setBoolean(
					index1,
					( (Boolean) parameters[index] ).booleanValue()
				);
			} else if ( parameters[index] == SDWLayout.NULL_STRING ) {
				statement.setNull( index1, Types.CHAR );
			} else if ( parameters[index] == SDWLayout.NULL_NUMERIC ) {
				statement.setNull( index1, Types.NUMERIC );
			} else if ( parameters[index] == SDWLayout.NULL_DOUBLE ) {
				statement.setNull( index1, Types.DOUBLE );
			} else if ( parameters[index] == SDWLayout.NULL_DATE ) {
				statement.setNull( index1, Types.DATE );
			} else {
				throw new IllegalArgumentException(
					null == parameters[index]
					? "Null parameter encountered at index " + i
					: "Unrecognised parameter class " + parameters[index].getClass()
				);
			}
		}
	}

	private void logWarnings (
		final String methodName,
		final Statement statement

	) {
		try {
			for (
				SQLWarning warning
				= statement.getWarnings();
				null != warning;
				warning = warning.getNextWarning()
			) {
				componentHelper.logger.log(
					Level.SEVERE,
					componentHelper.logName,
					methodName,
					warning.toString(),
					warning
				);
			}
		}
		catch ( final SQLException e ) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				methodName,
				"Failed to list SQL warnings: " + e,
				e
			);
		}
	}
}

/* ***************************** Class ****************************** */

class DataSetInfo {
	private final SDWLayout layout;
	private final SDWLocker locker;

	public DataSetInfo ( final SDWLayout layout, final SDWLocker locker) {
		this.layout = layout;
		this.locker = locker;
	}
	public SDWLayout getLayout () {
		return layout;
	}
	public SDWLocker getLocker () {
		return locker;
	}
}
