// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (C) 2007, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.sql;

import java.util.Date;
import java.util.Properties;
import java.util.List;
import java.util.ArrayList;
import java.util.Collections;

import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.text.ParseException;

import java.sql.Connection;
import java.sql.SQLException;

import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;

import org.apache.derby.jdbc.EmbeddedDriver;

import Int.ecb.gesmescb.InterchangeBegin;
import Int.ecb.gesmescb.MessageBegin;
import Int.ecb.gesmescb.Contact;
import Int.ecb.gesmescb.KeyFamily;
import Int.ecb.gesmescb.KeyFamilyComponent;
import Int.ecb.gesmescb.TimeSeries;
import Int.ecb.gesmescb.Observation;
import Int.ecb.gesmescb.Attribute;
import Int.ecb.gesmescb.MessageEnd;
import Int.ecb.gesmescb.InterchangeEnd;

import Int.ecb.gesmescb.metabase.Metabase;

import Int.ecb.gesmescb.event.TestEventSource;
import Int.ecb.gesmescb.event.ErrorEvent;

import Int.ecb.logging.Level;
import Int.ecb.logging.Logger;

/* ***************************** Class ****************************** */

/**
 * Tests the implementation of the class {@link SDWWriter}.
 *
 * @author
 * <a href="mailto:sis.external@ecb.int">Franois Hoeltzli</a>
 * @author
 * <a href="mailto:sis.external@ecb.int">Bernhard Bodenstorfer</a>
 *
 * @version @(#)$Id: SDWWriterTest.java,v 1.1 2017/01/19 16:14:02 trainot Exp $
 */

public class SDWWriterTest extends TestCase {

	/* ======================= Constants ======================== */

	/** To test dates. */
	private static final DateFormat DATE_FORMAT
	= new SimpleDateFormat( "yyyy-MM-dd'T'HH:mm:ss" );

	/** Schema ID to be used in the database. */
	private static final String SCHEMA = "SDW_WEB";

	/** Creates tables to perform the test. */
	private static final String[]
		testPreparation = {
			"CREATE SCHEMA " + SCHEMA,
			"SET SCHEMA " + SCHEMA,
			"CREATE TABLE ds1_obs_tmp ("
			+ " frequency VARCHAR(1) NOT NULL,"
			+ " obs_date TIMESTAMP NOT NULL,"
			+ " ref_area VARCHAR(2) NOT NULL,"
			+ " item VARCHAR(3) NOT NULL,"
			+ " series_key VARCHAR(51) NOT NULL,"
			+ " obs_value FLOAT,"
			+ " obs_status VARCHAR(1),"
			+ " obs_conf VARCHAR(1),"
			+ " obs_pre_break FLOAT,"
			+ " obs_att VARCHAR(64),"
			+ " obs_com VARCHAR(1050),"
			+ " seq_id NUMERIC,"
			+ " dt TIMESTAMP"
			+ " )",
			"CREATE TABLE ds1_series_tmp ("
			+ " frequency VARCHAR(1),"
			+ " ref_area VARCHAR(2) NOT NULL,"
			+ " item VARCHAR(3) NOT NULL,"
			+ " availability VARCHAR(1),"
			+ " compilation VARCHAR(1),"
			+ " title VARCHAR(100),"
			+ " unit VARCHAR(10),"
			+ " seq_id NUMERIC,"
			+ " dt TIMESTAMP,"
			+ " series_key VARCHAR(51) NOT NULL,"
			+ " sibling_key VARCHAR(51) NOT NULL"
			+ " )",
			"CREATE TABLE ds1_del_tmp ("
			+ " obs_date TIMESTAMP,"
			+ " \"KEY\" VARCHAR(51) NOT NULL,"
			+ " issibling VARCHAR(1),"
			+ " seq_id NUMERIC,"
			+ " dt TIMESTAMP"
			+ " )",
			"CREATE SCHEMA datamgmt",
			"SET SCHEMA datamgmt",
			"CREATE PROCEDURE register_update ("
			+ " IN iSchema VARCHAR(8),"
			+ " IN iDataset VARCHAR(8),"
			+ " IN iProcessId VARCHAR(120)"
			+ " )"
			+ " PARAMETER STYLE JAVA"
			+ " LANGUAGE JAVA"
			+ " EXTERNAL NAME '"
			+ SDWWriterTest.class.getName()
			+ ".registerUpdate'",
			"CREATE FUNCTION request_update ("
			+ " iSchema VARCHAR(8),"
			+ " iDataset VARCHAR(8),"
			+ " iProcessId VARCHAR(120)"
			+ " ) RETURNS VARCHAR(3) "
			+ " PARAMETER STYLE JAVA"
			+ " LANGUAGE JAVA"
			+ " EXTERNAL NAME '"
			+ SDWWriterTest.class.getName()
			+ ".requestUpdate'",
			"CREATE PROCEDURE unregister_update ("
			+ " IN iSchema VARCHAR(8),"
			+ " IN iDataset VARCHAR(8),"
			+ " IN iProcessId VARCHAR(120)"
			+ " )"
			+ " PARAMETER STYLE JAVA"
			+ " LANGUAGE JAVA"
			+ " EXTERNAL NAME '"
			+ SDWWriterTest.class.getName()
			+ ".unregisterUpdate'",
			"CREATE PROCEDURE sp_integrate_changes ("
			+ " IN iSchema VARCHAR(8),"
			+ " IN iDataset VARCHAR(8),"
			+ " IN iFullRefresh INTEGER,"
			+ " IN iVersion VARCHAR(8),"
			+ " INOUT iReleaseDate VARCHAR(19),"
			+ " IN iMsg INTEGER"
			+ " )"
			+ " MODIFIES SQL DATA"
			+ " PARAMETER STYLE JAVA"
			+ " LANGUAGE JAVA"
			+ " EXTERNAL NAME '"
			+ SDWWriterTest.class.getName()
			+ ".spIntegrateChanges'"
		},
		testPostparation = {
			"SET SCHEMA " + SCHEMA,
			"DELETE FROM ds1_obs_tmp",
			"DROP TABLE ds1_obs_tmp",
			"DELETE FROM ds1_series_tmp",
			"DROP TABLE ds1_series_tmp",
			"DELETE FROM ds1_del_tmp",
			"DROP TABLE ds1_del_tmp",
			"DROP SCHEMA " + SCHEMA + " RESTRICT",
			"SET SCHEMA datamgmt",
			"DROP PROCEDURE register_update",
			"DROP FUNCTION request_update",
			"DROP PROCEDURE unregister_update",
			"DROP PROCEDURE sp_integrate_changes",
			"DROP SCHEMA datamgmt RESTRICT"
		},
		testDirt = {
			"SET SCHEMA " + SCHEMA,
			"INSERT into ds1_obs_tmp VALUES ("
			+ " 'M', TIMESTAMP( '2005-12-24 00:00:00' ), 'DE', '500',"
			+ " 'DS1.M.DE.500',"
			+ " 13.6, 'B', 'F', NULL, 'Bluups', 'Whoops',"
			+ " 1306, TIMESTAMP( '2003-09-27 00:00:00' )"
			+ " )",
			"INSERT into ds1_series_tmp VALUES ("
			+ " 'M', 'DE', '500',"
			+ " 'A', 'E', 'Hurkk', 'KILO',"
			+ " 1307, TIMESTAMP( '2003-09-27 00:00:00' ),"
			+ " 'DS1.M.DE.500', 'DS1..DE.500'"
			+ " )",
			"INSERT into ds1_del_tmp VALUES ("
			+ " TIMESTAMP( '2005-12-24 00:00:00' ), 'DS1.M.DE.600', 'N',"
			+ " 1308, TIMESTAMP( '2003-09-27 00:00:00' )"
			+ " )"
		};

	/** The test input for the {@link SDWWriter}. */
	private static final InterchangeBegin
		INTERCHANGE_BEGIN = new InterchangeBegin(
			"IREF000035", "4F1", "4F0", "061205:1610", false
		),
		INTERCHANGE_BEGIN_TEST = new InterchangeBegin(
			"IREF000035", "4F1", "4F0", "061205:1620", true
		);

	private static final MessageBegin
		MESSAGE_BEGIN_STRUCTURE = new MessageBegin(
			"MREF000001", "4F1", "4F0", "ECB",
			MessageBegin.STATISTICAL_DEFINITIONS,
			"Test structure message",
			new Contact[]{}
		),
		MESSAGE_BEGIN_UPDATE = new MessageBegin(
			"MREF000001", "4F1", "4F0", "ECB",
			MessageBegin.STATISTICAL_DATA,
			"Test update message",
			new Contact[]{},
			"DS1", MessageBegin.UPDATE_OR_REPLACE_ACTION,
			"200612061559", "", "", "KF1"
		),
		MESSAGE_BEGIN_DELETE = new MessageBegin(
			"MREF000001", "4F2", "4F0", "ECB",
			MessageBegin.STATISTICAL_DATA,
			"Test delete message",
			new Contact[]{},
			"DS1", MessageBegin.DELETE_ACTION,
			"200612061557", "", "", "KF1"
		);

	private static final KeyFamily
		KF1 = new KeyFamily(
			"KF1", "Test structure", "", new Date( 0 ),
			new KeyFamilyComponent[]{
				new KeyFamilyComponent(
					"FREQ",
					KeyFamilyComponent.FREQUENCY_DIMENSION,
					"1", "A1",
					KeyFamilyComponent.MANDATORY,
					"", "CL_FREQ", true
				),
				new KeyFamilyComponent(
					"REF_AREA",
					KeyFamilyComponent.KEY_STRUCTURE_DIMENSION,
					"2", "AN2",
					KeyFamilyComponent.MANDATORY,
					"", "CL_AREA_EE", true
				),
				new KeyFamilyComponent(
					"ITEM",
					KeyFamilyComponent.KEY_STRUCTURE_DIMENSION,
					"3", "N3",
					KeyFamilyComponent.MANDATORY,
					"", "CL_ITEM", true
				),
				new KeyFamilyComponent(
					"TIME_PERIOD",
					KeyFamilyComponent.TIME_DIMENSION,
					"4", "AN..16",
					KeyFamilyComponent.MANDATORY,
					"", "", true
				),
				new KeyFamilyComponent(
					"TIME_FORMAT",
					KeyFamilyComponent.TIME_DIMENSION,
					"5", "N3",
					KeyFamilyComponent.MANDATORY,
					"", "", true
				),
				new KeyFamilyComponent(
					"OBS_VALUE",
					KeyFamilyComponent.ARRAY_CELL,
					"6", "AN..15",
					KeyFamilyComponent.MANDATORY,
					"", "", true
				),
				new KeyFamilyComponent(
					"OBS_STATUS",
					KeyFamilyComponent.ARRAY_CELL,
					"7", "A1",
					KeyFamilyComponent.MANDATORY,
					KeyFamilyComponent.ATTACHED_TO_OBSERVATION,
					"CL_OBS_STATUS", true
				),
				new KeyFamilyComponent(
					"OBS_CONF",
					KeyFamilyComponent.ARRAY_CELL,
					"8", "A1",
					KeyFamilyComponent.CONDITIONAL,
					KeyFamilyComponent.ATTACHED_TO_OBSERVATION,
					"CL_OBS_CONF", true
				),
				new KeyFamilyComponent(
					"OBS_PRE_BREAK",
					KeyFamilyComponent.ARRAY_CELL,
					"9", "AN..15",
					KeyFamilyComponent.CONDITIONAL,
					KeyFamilyComponent.ATTACHED_TO_OBSERVATION,
					"", true
				),
				new KeyFamilyComponent(
					"OBS_COM",
					KeyFamilyComponent.ATTRIBUTE,
					"", "AN..1024",
					KeyFamilyComponent.CONDITIONAL,
					KeyFamilyComponent.ATTACHED_TO_OBSERVATION,
					"", true
				),
				new KeyFamilyComponent(
					"OBS_ATT",
					KeyFamilyComponent.ATTRIBUTE,
					"", "A..64",
					KeyFamilyComponent.CONDITIONAL,
					KeyFamilyComponent.ATTACHED_TO_OBSERVATION,
					"", true
				),
				new KeyFamilyComponent(
					"AVAILABILITY",
					KeyFamilyComponent.ATTRIBUTE,
					"", "A1",
					KeyFamilyComponent.CONDITIONAL,
					KeyFamilyComponent.ATTACHED_TO_TIME_SERIES,
					"CL_AVAILABILITY", true
				),
				new KeyFamilyComponent(
					"COMPILATION",
					KeyFamilyComponent.ATTRIBUTE,
					"", "A1",
					KeyFamilyComponent.CONDITIONAL,
					KeyFamilyComponent.ATTACHED_TO_TIME_SERIES,
					"CL_COMPILATION", true
				),
				new KeyFamilyComponent(
					"TITLE",
					KeyFamilyComponent.ATTRIBUTE,
					"", "AN..100",
					KeyFamilyComponent.CONDITIONAL,
					KeyFamilyComponent.ATTACHED_TO_SIBLING_GROUP,
					"", true
				),
				new KeyFamilyComponent(
					"UNIT",
					KeyFamilyComponent.ATTRIBUTE,
					"", "A3",
					KeyFamilyComponent.CONDITIONAL,
					KeyFamilyComponent.ATTACHED_TO_SIBLING_GROUP,
					"", true
				)
			}
		);

	private static final TimeSeries
		SERIES_1_EMPTY = new TimeSeries(
			"M.HU.100", "", "", new Observation[]{}
		),
		SERIES_1_FULL = new TimeSeries(
			"M.HU.100", "200201200204", "710", new Observation[]{
				new Observation( "1235", "B", "F", "18.4" ),
				new Observation( "1236", "A", "F", "" ),
				Observation.EMPTY_OBSERVATION,
				new Observation( "1238", "E", "F", "" )
			}
		),
		SERIES_2_EMPTY = new TimeSeries(
			"M.HU.200", "200201200203", "710", new Observation[]{}
		),
		SERIES_2_FULL = new TimeSeries(
			"M.HU.200", "200202", "610", new Observation[]{
				new Observation( "835", "B", "F", "818.4" ),
			}
		);

	private static final Attribute
		GROUP_1_ATT_EMPTY = new Attribute( ".HU.100", "", "", "UNIT", "" ),
		GROUP_1_ATT_1 = new Attribute( ".HU.100", "", "", "UNIT", "KILO" ),
		GROUP_1_ATT_2 = new Attribute( ".HU.100", "", "", "TITLE", "HuMp" ),
		SERIES_1_ATT_EMPTY = new Attribute( "M.HU.100", "", "", "COMPILATION", "" ),
		SERIES_1_ATT_1 = new Attribute( "M.HU.100", "", "", "COMPILATION", "E" ),
		SERIES_1_ATT_2 = new Attribute( "M.HU.100", "", "", "AVAILABILITY", "A" ),
		OBS_1_ATT_EMPTY = new Attribute( "M.HU.100", "200201", "610", "OBS_COM", "" ),
		OBS_1_ATT_1 = new Attribute( "M.HU.100", "200201", "610", "OBS_COM", "Terrible break" ),
		OBS_1_ATT_2 = new Attribute( "M.HU.100", "200201", "610", "OBS_ATT", "Unreasonable attribute" ),
		OBS_2_ATT_1 = new Attribute( "M.HU.100", "200202", "610", "OBS_COM", "No break" ),
		OBS_3_ATT_1 = new Attribute( "M.HU.200", "200201", "610", "OBS_COM", "Normal" );

	private static final MessageEnd
		MESSAGE_END = new MessageEnd( "MREF000001" );

	private static final InterchangeEnd
		INTERCHANGE_END = new InterchangeEnd( "IREF000035" );

	private static final ErrorEvent
		WARNING_EVENT = new ErrorEvent( "Warning", ErrorEvent.WARNING ),
		ERROR_EVENT = new ErrorEvent( "Severe", ErrorEvent.ERROR );

	private static final Object[]
		testUpdateInput = {
			INTERCHANGE_BEGIN,
			MESSAGE_BEGIN_UPDATE,
			SERIES_1_FULL,
			SERIES_2_FULL,
			GROUP_1_ATT_1,
			GROUP_1_ATT_2,
			SERIES_1_ATT_1,
			SERIES_1_ATT_2,
			OBS_1_ATT_1,
			OBS_1_ATT_2,
			OBS_2_ATT_1,
			OBS_3_ATT_1,
			MESSAGE_END,
			INTERCHANGE_END
		},
		testDeleteInput = {
			INTERCHANGE_BEGIN,
			MESSAGE_BEGIN_DELETE,
			SERIES_1_EMPTY,
			SERIES_2_EMPTY,
			GROUP_1_ATT_EMPTY,
			SERIES_1_ATT_EMPTY,
			OBS_1_ATT_EMPTY,
			MESSAGE_END,
			INTERCHANGE_END
		},
		testTestInput = {
			INTERCHANGE_BEGIN_TEST,
			MESSAGE_BEGIN_UPDATE,
			SERIES_1_FULL,
			SERIES_2_FULL,
			GROUP_1_ATT_1,
			GROUP_1_ATT_2,
			SERIES_1_ATT_1,
			SERIES_1_ATT_2,
			OBS_1_ATT_1,
			OBS_1_ATT_2,
			OBS_2_ATT_1,
			OBS_3_ATT_1,
			MESSAGE_END,
			MESSAGE_BEGIN_DELETE,
			SERIES_1_EMPTY,
			SERIES_2_EMPTY,
			GROUP_1_ATT_EMPTY,
			SERIES_1_ATT_EMPTY,
			OBS_1_ATT_EMPTY,
			MESSAGE_END,
			INTERCHANGE_END
		};

	/* ========================= Fields ========================= */

	private final Logger logger;

	// As more than one thread uses the fields. See Java Memory Model.
	private volatile static int countRegisterCalls, countUnregisterCalls;
	private volatile static boolean grantLock;

	// As more than one thread uses the List. See Java Memory Model.
	private static final List spIntegrateChangesCalls
		= Collections.synchronizedList( new ArrayList() );

	private final SQLTestHelper testHelper;

	/* ====================== Constructors ====================== */

	/**
	 * Constructs an SDWWriterTest with the specified name.
	 *
	 * @param name Test case name
	 */
	public SDWWriterTest ( final String name ) {
		super( name );
		testHelper = new SQLTestHelper( "sdwwriter" );
		logger = Logger.getLogger( this.getClass().getName() );
	}

	/* ================== Protected Methods ===================== */

	protected void setUp () throws SQLException {
		// Int.ecb.logging.LogManager.getLogManager().addGlobalHandler( new Int.ecb.logging.ConsoleHandler() );

		testHelper.setUp();

		testHelper.executeStatements( testPreparation, false );

		System.setProperty(
			SDWWriter.class.getName() + ".Pete.driver",
			EmbeddedDriver.class.getName()
		);
		System.setProperty(
			SDWWriter.class.getName() + ".Pete.destination",
			SQLTestHelper.DB_URL_BASE + ";create=false;readonly=false"
		);
		System.setProperty(
			SDWWriter.class.getName() + ".Pete.user",
			SQLTestHelper.USER
		);
		System.setProperty(
			SDWWriter.class.getName() + ".Pete.password",
			SQLTestHelper.PASSWORD
		);
		System.setProperty(
			SDWWriter.class.getName() + ".Pete.schema", SCHEMA
		);
		System.setProperty(
			SDWWriter.class.getName() + ".Pete.processId",
			"hostname pid name filename"
		);
		System.setProperty(
			SDWWriter.class.getName() + ".Pete.lockWaitingTime",
			"1"
		);

		final Metabase metabase = Metabase.getMetabase();
		metabase.handle( INTERCHANGE_BEGIN );
		metabase.handle( MESSAGE_BEGIN_STRUCTURE );
		metabase.handle( KF1 );

		countRegisterCalls = countUnregisterCalls = 0;
		grantLock = true;
		spIntegrateChangesCalls.clear();
	}

	protected void tearDown () throws SQLException {
		final Properties properties = System.getProperties();
		properties.remove(
			SDWWriter.class.getName() + ".Pete.driver"
		);
		properties.remove(
			SDWWriter.class.getName() + ".Pete.destination"
		);
		properties.remove(
			SDWWriter.class.getName() + ".Pete.user"
		);
		properties.remove(
			SDWWriter.class.getName() + ".Pete.password"
		);
		properties.remove(
			SDWWriter.class.getName() + ".Pete.schema"
		);
		properties.remove(
			SDWWriter.class.getName() + ".Pete.processId"
		);
		properties.remove(
			SDWWriter.class.getName() + ".Pete.lockWaitingTime"
		);

		testHelper.executeStatements( testPostparation, false );

		testHelper.tearDown();

		countRegisterCalls = countUnregisterCalls = 0;
		spIntegrateChangesCalls.clear();
	}

	/* ===================== Public Methods ===================== */

	/**
	 * Assembles and returns a test suite for all
	 * the test methods of this test case.
	 *
	 * @return a non-null test suite
	 */
	public static Test suite () {
		return new TestSuite( SDWWriterTest.class );
	}

	/** Run the test case. */
	public static void main ( final String args[] ) {
		final String[] testCaseName = { SDWWriterTest.class.getName() };
		junit.textui.TestRunner.main( testCaseName );
	}

	/* ---------------------- Test Methods ---------------------- */

	public void testUpdate ()
		throws SQLException, ParseException
	{
		final TestEventSource source = new TestEventSource();
		final SDWWriter writer = new SDWWriter( "Pete" );
		source.addListener( writer, true );
		source.dispatch( WARNING_EVENT );
		source.dispatch( testUpdateInput );
		logger.log(
			Level.WARNING,
			this.getClass().getName(),
			"testUpdate()",
			"test warning"
		);
		writer.close();

		final Date
			extracted = DATE_FORMAT.parse( "2006-12-06T15:59:00" ),
			Jan2002 = DATE_FORMAT.parse( "2002-01-01T00:00:00" ),
			Feb2002 = DATE_FORMAT.parse( "2002-02-01T00:00:00" ),
			Apr2002 = DATE_FORMAT.parse( "2002-04-01T00:00:00" );

		final Connection connection = testHelper.createConnection();
		testHelper.verifyQueryResult(
			connection,
			"SELECT * FROM " + SCHEMA + ".ds1_obs_tmp"
			+ " WHERE obs_com IS NULL AND obs_att IS NULL"
			+ " ORDER BY item ASC,"
			+ " obs_date ASC",
			new Object[][]{
				{
					"M", Jan2002, "HU", "100", "DS1.M.HU.100",
					new Double( 1235 ), "B", "F", new Double( 18.4 ),
					null, null,
					new Integer( 1 ), extracted
				},
				{
					"M", Feb2002, "HU", "100", "DS1.M.HU.100",
					new Double( 1236 ), "A", "F", null,
					null, null,
					new Integer( 1 ), extracted
				},
				{
					"M", Apr2002, "HU", "100", "DS1.M.HU.100",
					new Double( 1238 ), "E", "F", null,
					null, null,
					new Integer( 1 ), extracted
				},
				{
					"M", Feb2002, "HU", "200", "DS1.M.HU.200",
					new Double( 835 ), "B", "F", new Double( 818.4 ),
					null, null,
					new Integer( 2 ), extracted
				}
			}
		);
		testHelper.verifyQueryResult(
			connection,
			"SELECT * FROM " + SCHEMA + ".ds1_obs_tmp"
			+ " WHERE obs_com IS NOT NULL"
			+ " ORDER BY item ASC,"
			+ " obs_date ASC",
			new Object[][]{
				{
					"M", Jan2002, "HU", "100", "DS1.M.HU.100",
					null, null, null, null,
					null, "Terrible break",
					new Integer( 7 ), extracted
				},
				{
					"M", Feb2002, "HU", "100", "DS1.M.HU.100",
					null, null, null, null,
					null, "No break",
					new Integer( 9 ), extracted
				},
				{
					"M", Jan2002, "HU", "200", "DS1.M.HU.200",
					null, null, null, null,
					null, "Normal",
					new Integer( 10 ), extracted
				}
			}
		);
		testHelper.verifyQueryResult(
			connection,
			"SELECT * FROM " + SCHEMA + ".ds1_obs_tmp"
			+ " WHERE obs_att IS NOT NULL"
			+ " ORDER BY item ASC,"
			+ " obs_date ASC",
			new Object[][]{
				{
					"M", Jan2002, "HU", "100", "DS1.M.HU.100",
					null, null, null, null,
					"Unreasonable attribute", null,
					new Integer( 8 ), extracted
				}
			}
		);
		testHelper.verifyQueryResult(
			connection,
			"SELECT * FROM " + SCHEMA + ".ds1_series_tmp"
			+ " WHERE availability IS NOT NULL",
			new Object[][]{
				{
					"M", "HU", "100",
					"A", null, null, null,
					new Integer( 6 ), extracted,
					"DS1.M.HU.100", "DS1..HU.100"
				}
			}
		);
		testHelper.verifyQueryResult(
			connection,
			"SELECT * FROM " + SCHEMA + ".ds1_series_tmp"
			+ " WHERE compilation IS NOT NULL",
			new Object[][]{
				{
					"M", "HU", "100",
					null, "E", null, null,
					new Integer( 5 ), extracted,
					"DS1.M.HU.100", "DS1..HU.100"
				}
			}
		);
		testHelper.verifyQueryResult(
			connection,
			"SELECT * FROM " + SCHEMA + ".ds1_series_tmp"
			+ " WHERE title IS NOT NULL",
			new Object[][]{
				{
					null, "HU", "100",
					null, null, "HuMp", null,
					new Integer( 4 ), extracted,
					"DS1..HU.100", "DS1..HU.100"
				}
			}
		);
		testHelper.verifyQueryResult(
			connection,
			"SELECT * FROM " + SCHEMA + ".ds1_series_tmp"
			+ " WHERE unit IS NOT NULL",
			new Object[][]{
				{
					null, "HU", "100",
					null, null, null, "KILO",
					new Integer( 3 ), extracted,
					"DS1..HU.100", "DS1..HU.100"
				}
			}
		);
		testHelper.verifyQueryResult(
			connection,
			"SELECT * FROM " + SCHEMA + ".ds1_del_tmp",
			new Object[][]{}
		);
		connection.close();

		final Object[][] changeLog = (Object[][]) spIntegrateChangesCalls.toArray(
			new Object[0][0]
		);
		assertEquals( "Changelog size", 1, changeLog.length );
		assertEquals( "Changelog schema", SCHEMA, changeLog[0][0] );
		assertEquals( "Changelog dataset", "DS1", changeLog[0][1] );
		assertEquals( "Changelog full", new Integer( 0 ), changeLog[0][2] );
		assertEquals( "Changelog version", "NONE", changeLog[0][3] );
		assertEquals(
			"Changelog release",
			new String[]{ DATE_FORMAT.format( extracted ) },
			(String[]) changeLog[0][4]
		);
		assertEquals( "Changelog full", new Integer( 0 ), changeLog[0][5] );
	}

	public void testDelete ()
		throws SQLException, ParseException
	{
		final TestEventSource source = new TestEventSource();
		final SDWWriter writer;
		{
			System.setProperty(
				SDWWriter.class.getName() + ".Pete.dataVersion", "MERGE"
			);
			System.setProperty(
				SDWWriter.class.getName() + ".Pete.refreshLevel", "dataset"
			);
			writer = new SDWWriter( "Pete" );
			final Properties properties = System.getProperties();
			properties.remove(
				SDWWriter.class.getName() + ".Pete.dataVersion"
			);
			properties.remove(
				SDWWriter.class.getName() + ".Pete.refreshLevel"
			);
		}
		source.addListener( writer, true );
		source.dispatch( WARNING_EVENT );
		logger.log(
			Level.WARNING,
			this.getClass().getName(),
			"testDelete()",
			"test warning"
		);
		source.dispatch( testDeleteInput );
		writer.close();

		final Date
			extracted = DATE_FORMAT.parse( "2006-12-06T15:57:00" ),
			Jan2002 = DATE_FORMAT.parse( "2002-01-01T00:00:00" ),
			Feb2002 = DATE_FORMAT.parse( "2002-02-01T00:00:00" ),
			Mar2002 = DATE_FORMAT.parse( "2002-03-01T00:00:00" );

		final Connection connection = testHelper.createConnection();
		testHelper.verifyQueryResult(
			connection,
			"SELECT * FROM " + SCHEMA + ".ds1_del_tmp"
			+ " WHERE obs_date IS NULL",
			new Object[][]{
				{
					null, "DS1.M.HU.100", Boolean.FALSE,
					new Integer( 1 ), extracted
				}
			}
		);
		testHelper.verifyQueryResult(
			connection,
			"SELECT * FROM " + SCHEMA + ".ds1_del_tmp"
			+ " WHERE obs_date IS NOT NULL"
			+ " ORDER BY obs_date ASC",
			new Object[][]{
				{
					Jan2002, "DS1.M.HU.200", Boolean.FALSE,
					new Integer( 2 ), extracted
				},
				{
					Feb2002, "DS1.M.HU.200", Boolean.FALSE,
					new Integer( 2 ), extracted
				},
				{
					Mar2002, "DS1.M.HU.200", Boolean.FALSE,
					new Integer( 2 ), extracted
				}
			}
		);
		testHelper.verifyQueryResult(
			connection,
			"SELECT * FROM " + SCHEMA + ".ds1_obs_tmp",
			new Object[][]{
				{
					"M", Jan2002, "HU", "100", "DS1.M.HU.100",
					null, null, null, null,
					null, "~",
					new Integer( 5 ), extracted
				}
			}
		);
		testHelper.verifyQueryResult(
			connection,
			"SELECT * FROM " + SCHEMA + ".ds1_series_tmp"
			+ " ORDER BY series_key ASC",
			new Object[][]{
				{
					null, "HU", "100",
					null, null, null, "~",
					new Integer( 3 ), extracted,
					"DS1..HU.100", "DS1..HU.100"
				},
				{
					"M", "HU", "100",
					null, "~", null, null,
					new Integer( 4 ), extracted,
					"DS1.M.HU.100", "DS1..HU.100"
				}
			}
		);
		connection.close();

		final Object[][] changeLog = (Object[][]) spIntegrateChangesCalls.toArray(
			new Object[0][0]
		);
		assertEquals( "Changelog size", 1, changeLog.length );
		assertEquals( "Changelog schema", SCHEMA, changeLog[0][0] );
		assertEquals( "Changelog dataset", "DS1", changeLog[0][1] );
		assertEquals( "Changelog full", new Integer( 1 ), changeLog[0][2] );
		assertEquals( "Changelog version", "MERGE", changeLog[0][3] );
		assertEquals(
			"Changelog release",
			new String[]{ DATE_FORMAT.format( extracted ) },
			(String[]) changeLog[0][4]
		);
		assertEquals( "Changelog full", new Integer( 0 ), changeLog[0][5] );
	}

	public void testTestInterchangeIgnored ()
		throws SQLException, ParseException
	{
		final TestEventSource source = new TestEventSource();
		final SDWWriter writer = new SDWWriter( "Pete" );
		source.addListener( writer, true );
		source.dispatch( testTestInput );
		writer.close();

		final Connection connection = testHelper.createConnection();

		testHelper.verifyQueryResult(
			connection,
			"SELECT * FROM " + SCHEMA + ".ds1_obs_tmp",
			new Object[][]{}
		);

		testHelper.verifyQueryResult(
			connection,
			"SELECT * FROM " + SCHEMA + ".ds1_series_tmp",
			new Object[][]{}
		);

		testHelper.verifyQueryResult(
			connection,
			"SELECT * FROM " + SCHEMA + ".ds1_del_tmp",
			new Object[][]{}
		);

		connection.close();

		final Object[][] changeLog = (Object[][]) spIntegrateChangesCalls.toArray(
			new Object[0][0]
		);
		assertEquals( "Changelog size", 0, changeLog.length );
	}

	public void testErrorEventInterchangeIgnored ()
		throws SQLException, ParseException
	{
		final TestEventSource source = new TestEventSource();
		final SDWWriter writer = new SDWWriter( "Pete" );
		source.addListener( writer, true );
		source.dispatch( testUpdateInput );
		source.dispatch( testDeleteInput );
		source.dispatch( ERROR_EVENT );
		writer.close();

		final Connection connection = testHelper.createConnection();

		testHelper.verifyQueryResult(
			connection,
			"SELECT * FROM " + SCHEMA + ".ds1_obs_tmp",
			new Object[][]{}
		);

		testHelper.verifyQueryResult(
			connection,
			"SELECT * FROM " + SCHEMA + ".ds1_series_tmp",
			new Object[][]{}
		);

		testHelper.verifyQueryResult(
			connection,
			"SELECT * FROM " + SCHEMA + ".ds1_del_tmp",
			new Object[][]{}
		);

		connection.close();

		final Object[][] changeLog = (Object[][]) spIntegrateChangesCalls.toArray(
			new Object[0][0]
		);
		assertEquals( "Changelog size", 0, changeLog.length );
	}

	public void testSevereLogEventInterchangeIgnored ()
		throws SQLException, ParseException
	{
		final TestEventSource source = new TestEventSource();
		final SDWWriter writer = new SDWWriter( "Pete" );
		source.addListener( writer, true );
		source.dispatch( testUpdateInput );
		source.dispatch( testDeleteInput );
		logger.log(
			Level.SEVERE,
			this.getClass().getName(),
			"testSevereLogEventInterchangeIgnored()",
			"test severe"
		);
		writer.close();

		final Connection connection = testHelper.createConnection();

		testHelper.verifyQueryResult(
			connection,
			"SELECT * FROM " + SCHEMA + ".ds1_obs_tmp",
			new Object[][]{}
		);

		testHelper.verifyQueryResult(
			connection,
			"SELECT * FROM " + SCHEMA + ".ds1_series_tmp",
			new Object[][]{}
		);

		testHelper.verifyQueryResult(
			connection,
			"SELECT * FROM " + SCHEMA + ".ds1_del_tmp",
			new Object[][]{}
		);

		connection.close();

		final Object[][] changeLog = (Object[][]) spIntegrateChangesCalls.toArray(
			new Object[0][0]
		);
		assertEquals( "Changelog size", 0, changeLog.length );
	}

	public void testDirtyTableSevere ()
		throws SQLException, ParseException
	{
		testHelper.executeStatements( testDirt, false );

		final TestEventSource source = new TestEventSource();
		final SDWWriter writer = new SDWWriter( "Pete" );
		source.addListener( writer, true );
		source.dispatch( testUpdateInput );
		source.dispatch( testDeleteInput );
		writer.close();

		final Date
			extracted = DATE_FORMAT.parse( "2003-09-27T00:00:00" ),
			christmas2005 = DATE_FORMAT.parse( "2005-12-24T00:00:00" );

		final Connection connection = testHelper.createConnection();
		testHelper.verifyQueryResult(
			connection,
			"SELECT * FROM " + SCHEMA + ".ds1_obs_tmp",
			new Object[][]{
				{
					"M", christmas2005, "DE", "500", "DS1.M.DE.500",
					new Double( 13.6 ), "B", "F", null,
					"Bluups", "Whoops",
					new Integer( 1306 ), extracted
				}
			}
		);
		testHelper.verifyQueryResult(
			connection,
			"SELECT * FROM " + SCHEMA + ".ds1_series_tmp",
			new Object[][]{
				{
					"M", "DE", "500",
					"A", "E", "Hurkk", "KILO",
					new Integer( 1307 ), extracted,
					"DS1.M.DE.500", "DS1..DE.500"
				}
			}
		);
		testHelper.verifyQueryResult(
			connection,
			"SELECT * FROM " + SCHEMA + ".ds1_del_tmp",
			new Object[][]{
				{
					christmas2005, "DS1.M.DE.600", "N",
					new Integer( 1308 ), extracted
				}
			}
		);
		connection.close();

		final Object[][] changeLog = (Object[][]) spIntegrateChangesCalls.toArray(
			new Object[0][0]
		);
		assertEquals( "Changelog size", 0, changeLog.length );
	}

	public void testDirtyTableWarning ()
		throws SQLException, ParseException
	{
		testHelper.executeStatements( testDirt, false );

		final TestEventSource source = new TestEventSource();
		final SDWWriter writer;
		{
			System.setProperty(
				SDWWriter.class.getName() + ".Pete.wipeTmp", "true"
			);
			writer = new SDWWriter( "Pete" );
			final Properties properties = System.getProperties();
			properties.remove( SDWWriter.class.getName() + ".Pete.wipeTmp" );
		}
		source.addListener( writer, true );
		source.dispatch( testUpdateInput );
		source.dispatch( testDeleteInput );
		writer.close();

		final Object[][] changeLog = (Object[][]) spIntegrateChangesCalls.toArray(
			new Object[0][0]
		);
		assertEquals( "Changelog size", 1, changeLog.length );
		assertEquals( "Changelog schema", SCHEMA, changeLog[0][0] );
		assertEquals( "Changelog dataset", "DS1", changeLog[0][1] );
		assertEquals( "Changelog full", new Integer( 0 ), changeLog[0][2] );
		assertEquals( "Changelog version", "NONE", changeLog[0][3] );
		assertEquals(
			"Changelog release",
			new String[]{ "2006-12-06T15:59:00" },
			(String[]) changeLog[0][4]
		);
		assertEquals( "Changelog full", new Integer( 0 ), changeLog[0][5] );
	}

	public void testLocking ()
		throws SQLException, ParseException, InterruptedException
	{
		final Thread thread = new Thread() {
			public void run () {
				try {
					final TestEventSource source = new TestEventSource();
					final SDWWriter writer = new SDWWriter( "Pete" );
					source.addListener( writer, true );
					source.dispatch( WARNING_EVENT );
					logger.log(
						Level.WARNING,
						this.getClass().getName(),
						"testLocking()",
						"test warning"
					);
					source.dispatch( testUpdateInput );
					writer.close();
				} catch ( final SQLException e ) {
					fail( "Caught SQLException: " + e );
				}
			}
		};

		grantLock = false;

		thread.start();

		Thread.sleep( 2000 );
		assertEquals(
			"Register should already have been called now",
			1, countRegisterCalls
		);
		assertEquals(
			"Unregister should not have been called yet",
			0, countUnregisterCalls
		);

		final Connection connection = testHelper.createConnection();
		try {
			testHelper.verifyQueryResult(
				connection,
				"SELECT * FROM " + SCHEMA + ".ds1_obs_tmp",
				new Object[][]{}
			);
			testHelper.verifyQueryResult(
				connection,
				"SELECT * FROM " + SCHEMA + ".ds1_series_tmp",
				new Object[][]{}
			);
			testHelper.verifyQueryResult(
				connection,
				"SELECT * FROM " + SCHEMA + ".ds1_del_tmp",
				new Object[][]{}
			);
			assertEquals(
				"No integration should have happened yet",
				0,
				spIntegrateChangesCalls.size()
			);

			grantLock = true;
			thread.join();

			assertEquals(
				"Register should have been called once",
				1, countRegisterCalls
			);
			assertEquals(
				"Unregister should have been called once",
				1, countUnregisterCalls
			);
			assertEquals(
				"Integration should have happened by now",
				1,
				spIntegrateChangesCalls.size()
			);
		} finally {
			connection.close();
		}
	}

	/* ---------------- Auxiliary methods ----------------- */

	public static void registerUpdate (
		final String iSchema,
		final String iDataset,
		final String iProcessId
	) {
		assertEquals(
			"register for correct schema",
			"SDW_WEB",
			iSchema
		);
		assertEquals(
			"register for correct data set",
			"DS1",
			iDataset
		);
		assertEquals(
			"register for correct process id",
			iProcessId,
			"hostname pid name filename"
		);
		++countRegisterCalls;
	}

	public static String requestUpdate (
		final String iSchema,
		final String iDataset,
		final String iProcessId
	) {
		assertEquals(
			"register for correct schema",
			"SDW_WEB",
			iSchema
		);
		assertEquals(
			"register for correct data set",
			"DS1",
			iDataset
		);
		assertEquals(
			"register for correct process id",
			iProcessId,
			"hostname pid name filename"
		);
		return grantLock ? "1" : "43";
	}

	public static void unregisterUpdate (
		final String iSchema,
		final String iDataset,
		final String iProcessId
	) {
		assertEquals(
			"register for correct schema",
			"SDW_WEB",
			iSchema
		);
		assertEquals(
			"register for correct data set",
			"DS1",
			iDataset
		);
		assertEquals(
			"register for correct process id",
			iProcessId,
			"hostname pid name filename"
		);
		++countUnregisterCalls;
	}

	public static void spIntegrateChanges (
		final String iSchema,
		final String iDataset,
		final int iFullRefresh,
		final String iVersion,
		final String[] iReleaseDate,
		final int iMsg
	) {
		spIntegrateChangesCalls.add(
			new Object[]{
				iSchema, iDataset, new Integer( iFullRefresh ),
				iVersion, iReleaseDate, new Integer( iMsg )
			}
		);
	}

	public static void assertEquals (
		final String message,
		final String[] expected,
		final String[] actual
	) {
		assertEquals( message + ": length", expected.length, actual.length );
		for ( int i = 0; i < expected.length; ++i ) {
			assertEquals( message + "[" + i + "]", expected[i], actual[i] );
		}
	}
}
