// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (C) 2006 European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.sql;

import java.util.Properties;

import java.sql.DriverManager;
import java.sql.Connection;
import java.sql.SQLException;

import Int.ecb.logging.Level;

import Int.ecb.gesmescb.util.ComponentHelper;

/* ***************************** Class ****************************** */

/**
 * Connects to a SQL database according to preferences.
 * The client preferences are meant to be used
 * by passing the client's {@link ComponentHelper}
 * to the constructor.
 *
 * <center><table border="3" cellpadding="3">
 * <caption>
 * <a name="prefs">Available {@link Int.ecb.prefs preference} settings</a>
 * </caption>
 * <tr>
 * <th>preference name</th>
 * <th>type</th>
 * <th>meaning</th>
 * <th>default</th>
 * </tr>
 * <tr>
 *	<td><code>driver</td><td>string</td>
 *	<td>Java class name of the JDBC driver to be used</td>
 * </tr>
 * <tr>
 *	<td><code>source</td><td>string</td>
 *	<td>URL of the database from which to select data, if applicable</td>
 * </tr>
 * <tr>
 *	<td><code>destination</td><td>string</td>
 *	<td>URL of the database to which to write to, if applicable</td>
 * </tr>
 * <tr>
 *	<td><code>user</td><td>string</td>
 *	<td>the user id for accessing the database</td>
 * </tr>
 * <tr>
 *	<td><code>password</td><td>string</td>
 *	<td>the password for accessing the database</td>
 * </tr>
 * </table></center>
 *
 * @author
 * <a href="mailto:sis.external@ecb.int">Bernhard Bodenstorfer</A>
 *
 * @version @(#)$Id: SQLConnector.java,v 1.1 2017/01/19 16:14:02 trainot Exp $
 */
public class SQLConnector {

	/* ========================= Fields ========================= */

	/** Takes care of logging and preferences. */
	protected final ComponentHelper componentHelper;

	/** The URL of the database read from. */
	private final String database;

	/** The JDBC connection to read from. */
	private final Connection connection;

	/* ====================== Constructors ====================== */

	/**
	 * Construct a <code>SQLConnector</code> and connect it to a database.
	 *
	 * @param componentHelper to access preferences of the client class.
	 * @param database URL of the database where to connect to.
	 * @param readWrite whether to allow writing to the database, otherwise read only.
	 *
	 * @throws NullPointerException if <code>database==null</code>
	 * and the preference variable
	 * <code>source</code> or <code>destination</code>
	 * (depending on <code>readWrite</code>) is undefined.
	 * @throws SQLException if the database cannot be opened for reading.
	 */
	public SQLConnector (
		final ComponentHelper componentHelper,
		final String database,
		boolean readWrite
	)
		throws SQLException
	{
		this.componentHelper = componentHelper;

		final String driver = componentHelper.getPreference( "driver", null );
		try {
			Class.forName( driver );
		}
		catch ( final Exception e ) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"SQLConnector",
				"Cannot construct JDBC driver from class \""
				+ driver + "\".",
				e
			);
		}

		final String propertyName = readWrite ? "destination" : "source";
		this.database = database != null
			? database
			: componentHelper.getPreference( propertyName, null );
		if ( this.database == null ) {
			final String desc
			= "No preference variable \"" + propertyName + "\" found"
				+ " to determine the database.";
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"SQLConnector()",
				desc
			);
			throw new NullPointerException( desc );
		}

		final Properties properties = new Properties();
		final String user = componentHelper.getPreference( "user", null );
		if ( user != null ) properties.put( "user", user );
		final String password = componentHelper.getPreference( "password", null );
		if ( password != null ) properties.put( "password", password );
		connection = DriverManager.getConnection(
			this.database, properties
		);
		connection.setReadOnly( ! readWrite );
		connection.setAutoCommit( true );
	}

	/* ===================== Public Methods ===================== */

	/** Get the URL used to open the database. */
	public String getDatabaseURL () {
		return database;
	}

	/** Get the connection. */
	public Connection getConnection () {
		return connection;
	}

	/** Close the <code>SQLConnector</code>. */
	public void close () {
		try {
			componentHelper.logger.log(
				Level.FINE,
				componentHelper.logName,
				"close()",
				"Closing SQL connection to \""
				+ database
				+ "\"."
			);
			connection.close();
		}
		catch ( final Exception e ) {
			final String desc = "Problem: " + e.getMessage() + ".";

			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.name,
				"close()",
				desc,
				e
			);
		}
	}
}
