// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (C) 2006, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.sql;

import java.util.Properties;

import java.sql.Connection;
import java.sql.Statement;
import java.sql.ResultSet;
import java.sql.SQLException;

import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;

import org.apache.derby.jdbc.EmbeddedDriver;

import Int.ecb.gesmescb.util.ComponentHelper;

/* ***************************** Class ****************************** */

/**
 * Helps testing the {@link SQLConnector} class.
 * To run it, the Apache Derby SQL database JAR is required.
 *
 * @author
 * <a href="mailto:sis.external@ecb.int">Bernhard Bodenstorfer</a>
 *
 * @version @(#)$Id: SQLConnectorTest.java,v 1.1 2017/01/19 16:14:02 trainot Exp $
 */

public class SQLConnectorTest extends TestCase {

	/* ======================= Constants ======================== */

	/** The database setup for {@link SQLReader} under test. */
	private static final String
		testInput[] = {
			"CREATE TABLE interpretation ("
			+ " variable VARCHAR(100),"
			+ " assigned NUMERIC(9,2)"
			+ " )",
			"INSERT INTO interpretation VALUES ( 'phi', 1.53 )",
			"INSERT INTO interpretation VALUES ( 'psi', 2.71 )"
		},
		testInput1 = "INSERT INTO interpretation VALUES ( 'tau', 1.53 )",
		testInput2 = "CREATE TABLE listing ("
			+ " item VARCHAR(10)"
			+ " )",
		testInput3 = "INSERT INTO listing VALUES ( 'miau' )";

	/* ========================= Fields ========================= */

	private final SQLTestHelper testHelper;
	private ComponentHelper componentHelper;

	/* ====================== Constructors ====================== */

	/**
	 * Constructs an SQLConnectorTest with the specified name.
	 *
	 * @param name Test case name
	 */
	public SQLConnectorTest ( final String name ) {
		super( name );
		testHelper = new SQLTestHelper( "sqlconnector" );
	}

	/* ================== Protected Methods ===================== */

	protected void setUp () {
		testHelper.setUp();

		System.setProperty(
			SQLConnectorTest.class.getName() + ".Linus.driver",
			EmbeddedDriver.class.getName()
		);
		System.setProperty(
			SQLConnectorTest.class.getName() + ".Linus.source",
			SQLTestHelper.DB_URL_BASE + ";create=false;readonly=true"
		);
		System.setProperty(
			SQLConnectorTest.class.getName() + ".Linus.destination",
			SQLTestHelper.DB_URL_BASE + ";create=false;readonly=false"
		);
		System.setProperty(
			SQLConnectorTest.class.getName() + ".Linus.user",
			SQLTestHelper.USER
		);
		System.setProperty(
			SQLConnectorTest.class.getName() + ".Linus.password",
			SQLTestHelper.PASSWORD
		);

		componentHelper = new ComponentHelper( "Linus", this );
	}

	protected void tearDown () {
		componentHelper = null;

		final Properties properties = System.getProperties();
		properties.remove(
			SQLConnectorTest.class.getName() + ".Linus.driver"
		);
		properties.remove(
			SQLConnectorTest.class.getName() + ".Linus.source"
		);
		properties.remove(
			SQLConnectorTest.class.getName() + ".Linus.destination"
		);
		properties.remove(
			SQLConnectorTest.class.getName() + ".Linus.user"
		);
		properties.remove(
			SQLConnectorTest.class.getName() + ".Linus.password"
		);

		testHelper.tearDown();
	}

	/* ===================== Public Methods ===================== */

	/**
	 * Assembles and returns a test suite for all
	 * the test methods of this test case.
	 *
	 * @return a non-null test suite
	 */
	public static Test suite () {
		return new TestSuite( SQLConnectorTest.class );
	}

	/** Run the test case. */
	public static void main ( final String args[] ) {
		final String[] testCaseName = { SQLConnectorTest.class.getName() };
		junit.textui.TestRunner.main( testCaseName );
	}

	/* ---------------------- Test Methods ---------------------- */

	public void testConnectRead ()
		throws SQLException
	{
		{
			final Connection connection = testHelper.createConnection();
			testHelper.executeStatements( connection, testInput, false );
			connection.close();
		}

		final SQLConnector connector
		= new SQLConnector( componentHelper, null, false );
		assertEquals(
			"Database URL",
			SQLTestHelper.DB_URL_BASE + ";create=false;readonly=true",
			connector.getDatabaseURL()
		);

		final Statement statement = connector.getConnection().createStatement();
		final ResultSet resultSet = statement.executeQuery(
			"SELECT assigned FROM interpretation WHERE variable='phi'"
		);

		assertTrue ( "Need one match", resultSet.next() );
		assertEquals( "Value match", 1.53, resultSet.getDouble( 1 ), 0 );
		assertTrue ( "No more match", ! resultSet.next() );

		try {
			statement.executeQuery( testInput1 );
			fail( "Expecting read only violation" );
		} catch ( final SQLException se ) {
		}

		connector.close();
	}

	public void testConnectWrite ()
		throws SQLException
	{
		{
			final SQLConnector connector
			= new SQLConnector( componentHelper, null, true );
			assertEquals(
				"Database URL",
				SQLTestHelper.DB_URL_BASE + ";create=false;readonly=false",
				connector.getDatabaseURL()
			);

			final Statement statement
			= connector.getConnection().createStatement();
			statement.execute( testInput2 );
			statement.execute( testInput3 );

			final ResultSet resultSet = statement.executeQuery(
				"SELECT item FROM listing WHERE item='miau'"
			);

			assertTrue ( "Need the new match", resultSet.next() );
			assertEquals( "New item match", "miau", resultSet.getString( 1 ) );
			assertTrue ( "No more new match", ! resultSet.next() );

			connector.close();
		}

		{
			final Connection connection = testHelper.createConnection();
			final Statement statement = connection.createStatement();
			final ResultSet resultSet = statement.executeQuery(
				"SELECT * FROM listing"
			);

			assertTrue ( "Need the match", resultSet.next() );
			assertEquals( "Item match", "miau", resultSet.getString( 1 ) );
			assertTrue ( "No more match", ! resultSet.next() );

			connection.close();
		}
	}
}
