// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.sql;

import java.util.Iterator;
import java.io.IOException;
import java.util.Date;
import java.text.DateFormat;
import java.text.SimpleDateFormat;

import Int.ecb.logging.Level;

import Int.ecb.gesmescb.*;
import Int.ecb.gesmescb.time.*;
import Int.ecb.gesmescb.metabase.*;

import Int.ecb.gesmescb.util.ComponentHelper;
import Int.ecb.gesmescb.util.KeyTokenizer;

/* **************************** Class ***************************** */

/**
 * SQL writer.
 *
 * @author
 * <A HREF="mailto:Sami.Airo@ECB.int">Sami Airo</A>
 *
 * @version @(#)$Id: SQLDataManipulationWriter.java,v 1.1 2017/01/19 16:14:02 trainot Exp $
 */

public class SQLDataManipulationWriter 
	extends SQLWriter
{

	/* ======================= Constants ======================== */

	final String MISSING_VALUE = "NULL";

	/* ========================= Fields ========================= */

	private MessageBegin messageBegin;

	/* ====================== Constructors ====================== */

	public SQLDataManipulationWriter ()
		throws IOException
	{
		this( "" );
	}

	public SQLDataManipulationWriter (final String name)
		throws IOException
	{
		componentHelper = new ComponentHelper( name, this );
		setSQLSyntax(componentHelper.getPreference("SQLSyntax", "SQL92"));
		open();
	}

	/* ===================== Public Methods ===================== */

	public void handle ( final MessageBegin messageBegin ) {
		componentHelper.logger.log(
			Level.FINE,
			componentHelper.logName,
			"handle",
			"Message begins"
		);

		if (currentSQLSyntax == SQL_SYNTAX_ORACLE8) {
			writer.println("SET DEFINE OFF" + getStatementTerminator());
			writer.println("ALTER SESSION SET NLS_DATE_FORMAT = " + getSQLString("YYYY-MM-dd hh24:mi:ss") + getStatementTerminator());
		}

		if (messageBegin.isStatisticalData()) {			
			writer.println("INSERT INTO " + messageBegin.getKeyFamilyId() + "_DATA_SETS (DATA_SET)"
								+ " VALUES(" + getSQLString(messageBegin.getDataSetId()) + ")" + getStatementTerminator());
		}

		this.messageBegin = messageBegin;
	}

	public void handle ( final CodeList codeList ) {
		componentHelper.logger.log(
			Level.FINE,
			componentHelper.logName,
			"handle",
			"Handling code list"
		);
		
		Iterator vit = codeList.codeValueIterator();
		Iterator dit = codeList.codeDescriptionIterator();
		
		while (vit.hasNext()) {
			String value = (String) vit.next();
			String desc = (String) dit.next();

			writer.println("INSERT INTO " + codeList.getId() + "(VALUE, DESCRIPTION)"
									+ " VALUES(" + getSQLString(value) + "," + getSQLString(desc) + ")" + getStatementTerminator());
		}
	}


	public void handle ( final StatisticalConcept concept ) {
		componentHelper.logger.log(
			Level.FINE,
			componentHelper.logName,
			"handle",
			"Handling statistical concept"
		);		
		
		writer.println("INSERT INTO STATISTICAL_CONCEPTS(ID, NAME)"
							+ " VALUES(" + getSQLString(concept.getId()) + "," + getSQLString(concept.getName()) + ")" + getStatementTerminator());
	}

	
	public void handle ( final TimeSeries timeSeries ) {
		componentHelper.logger.log(
			Level.FINE,
			componentHelper.logName,
			"handle",
			"Handling time series"
		);
		
		KeyFamily keyFamily = Metabase.getMetabase().getMaintenanceAgency(messageBegin.getMaintenanceAgencyId()).getKeyFamily(messageBegin.getKeyFamilyId());
		KeyTokenizer kit = new KeyTokenizer(timeSeries.getKey());
		Iterator it = keyFamily.keyStructureIterator();
		
		StringBuffer seriesColumns = new StringBuffer("SERIES_KEY,SIBLING_KEY");
		StringBuffer siblingColumns = new StringBuffer("DATA_SET,SIBLING_KEY");
		StringBuffer seriesValues = new StringBuffer();
		StringBuffer siblingValues = new StringBuffer();
		
		seriesValues.append(getSQLString(messageBegin.getDataSetId() + "." + timeSeries.getKey()));
		seriesValues.append("," + getSQLString(messageBegin.getDataSetId() + "." + timeSeries.getKey().substring(1)));
		siblingValues.append(getSQLString(messageBegin.getDataSetId()));
		siblingValues.append("," + getSQLString(messageBegin.getDataSetId() + "." + timeSeries.getKey().substring(1)));
		
		while (kit.hasNext()) {
			String dimension = kit.nextToken();
			KeyFamilyComponent component = (KeyFamilyComponent)it.next();
			if (component.getConceptId().equals("FREQ")) {
				seriesColumns.append( "," + component.getConceptId().toUpperCase() );
				seriesValues.append( "," + getSQLString(observationStringToSQLString(dimension)) );
				continue;
			}
			siblingColumns.append( "," + component.getConceptId().toUpperCase() );
			siblingValues.append( "," + getSQLString(observationStringToSQLString(dimension)) );
		}

		writer.println(
			  "INSERT INTO " + messageBegin.getKeyFamilyId() + "_SIBLING_GROUPS"
			+ " (" + siblingColumns.toString().toUpperCase() + ") VALUES(" + siblingValues.toString() + ")"
		 	+ getStatementTerminator()
		);
		
		writer.println(
			  "INSERT INTO " + messageBegin.getKeyFamilyId() + "_TIME_SERIES"
			+ " (" + seriesColumns.toString().toUpperCase() + ") VALUES(" + seriesValues.toString() + ")"		
			+ getStatementTerminator()
		);
		
		final RegularInterval totalInterval = timeSeries.getInterval();
		if ( totalInterval != null ) {
			final Iterator
				intervals = timeSeries.getInterval().iterator(),
				observations = timeSeries.observationIterator();

			while ( intervals.hasNext() && observations.hasNext() ) {
				final RegularInterval interval
				= (RegularInterval) intervals.next();
				final Observation observation
				= (Observation) observations.next();

				if ( observation.isEmpty() ) continue;

				String value = observationValueToSQLString( observation.getValue() );
				String status = observationStringToSQLString( observation.getStatus() );
				String confidentiality = observationStringToSQLString( observation.getConfidentiality() );
				String preBreakValue = observationValueToSQLString( observation.getPreBreakValue() );
				
				writer.println("INSERT INTO " 
								+ messageBegin.getKeyFamilyId() + "_OBSERVATIONS (SERIES_KEY, PERIOD, OBS_VALUE, OBS_STATUS, OBS_CONF, OBS_PRE_BREAK)"
								+ " VALUES(" + getSQLString(messageBegin.getDataSetId() + "." + timeSeries.getKey()) + "," 
								+ intervalToSQLDate( interval ) + "," 
								+ value + "," 
								+ getSQLString(status) + "," 
								+ getSQLString(confidentiality) + "," 
								+ preBreakValue + ")" + getStatementTerminator());
			}
		}		
	}

	public void handle ( final Attribute attribute ) {
		componentHelper.logger.log(
			Level.FINE,
			componentHelper.logName,
			"handle",
			"Handling attribute"
		);		

		KeyFamily keyFamily = Metabase.getMetabase().getMaintenanceAgency(messageBegin.getMaintenanceAgencyId()).getKeyFamily(messageBegin.getKeyFamilyId());
		KeyFamilyComponent component = keyFamily.getComponent(attribute.getName());		
		
		StringBuffer table = new StringBuffer( messageBegin.getKeyFamilyId() );		
		StringBuffer where = new StringBuffer(" WHERE");
		
		if (component.isAttachedToDataSet()) {
			table.append("_DATA_SETS");			
			where.append( " DATA_SET = " + getSQLString(messageBegin.getDataSetId()) );
		} else if (component.isAttachedToSiblingGroup()) {
			table.append("_SIBLING_GROUPS");
			where.append( " SIBLING_KEY = " + getSQLString(messageBegin.getDataSetId() + "." + attribute.getKey()) );
		} else if (component.isAttachedToTimeSeries()) {
			table.append("_TIME_SERIES");
			where.append( " SERIES_KEY = " + getSQLString(messageBegin.getDataSetId() + "." + attribute.getKey()) );
		} else {
			table.append("_DATA");
			where.append( " SERIES_KEY = " + getSQLString(messageBegin.getDataSetId() + "." + attribute.getKey()) );
			where.append(" AND PERIOD = TIMESTAMP " + getSQLString( intervalToSQLDate( attribute.getInterval() ) ) );
		}
					
		writer.println( "UPDATE " + table.toString()
					+ " SET " + attribute.getName().toUpperCase() + " = " + getSQLString(attribute.getValue())
					+ where.toString() + getStatementTerminator() );		
	}

	public String observationValueToSQLString ( final String value ) {
		if (value.equals(Observation.MISSING_VALUE)) return MISSING_VALUE;
		if (value.equals("")) return "NULL";
		return value;
	}

	public String observationStringToSQLString ( final String value ) {
		if (value.equals("")) return "NULL";
		return value;
	}
	
	public String intervalToSQLDate ( final RegularInterval interval ) {
		Date begin = interval.getBegin();
		DateFormat format = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss");
		if (currentSQLSyntax == SQL_SYNTAX_ORACLE8) {
			return getSQLString(format.format(begin));
		}
		return "TIMESTAMP " + getSQLString(format.format(begin));
	}
}
