// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (C) 2006, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.sql;

import java.util.Properties;
import java.util.Date;

import java.io.File;
import java.io.IOException;

import java.sql.DriverManager;
import java.sql.Driver;
import java.sql.Connection;
import java.sql.Statement;
import java.sql.ResultSet;
import java.sql.SQLException;

import org.apache.derby.jdbc.EmbeddedDriver;

import junit.framework.Assert;

/* ***************************** Class ****************************** */

/**
 * Helps testing classes accessing SQL databases.
 * To use it, the Apache Derby SQL database JAR is required.
 *
 * @author
 * <a href="mailto:sis.external@ecb.int">Bernhard Bodenstorfer</a>
 *
 * @version @(#)$Id: SQLTestHelper.java,v 1.1 2017/01/19 16:14:02 trainot Exp $
 */

class SQLTestHelper extends Assert {

	/* ======================= Constants ======================== */

	protected static final String
		DB_URL_BASE = "jdbc:derby:testDB",
		USER = "test",
		PASSWORD = "testword";

	/* ========================= Fields ========================= */

	public final String name;
	private final Properties properties;
	private File testDB;
	private Driver driver;

	/* ====================== Constructors ====================== */

	/**
	 * Constructs an SQLTestHelper with the specified name.
	 *
	 * @param name Test case name
	 */
	public SQLTestHelper ( final String name ) {
		this.name = name;
		properties = new Properties();
		properties.put( "user", USER );
		properties.put( "password", PASSWORD );
		testDB = null;
		driver = null;
	}

	/* ===================== Public Methods ===================== */

	/** Prepare the settings for the test database before a test. */
	protected void setUp () {
		try {
			testDB = File.createTempFile( name + "_test.", ".dbd" );
			assertTrue( "Delete local DB placeholder", testDB.delete() );
			assertTrue( "Create local DB directory", testDB.mkdir() );
			System.setProperty( "derby.system.home", testDB.getPath() );
			driver = new EmbeddedDriver();
		}
		catch ( final IOException ioe ) {
			fail( "Creating test area failed: " + ioe );
		}
	}

	/** Create a connection to the test database during a test. */
	public Connection createConnection () throws SQLException {
		final Connection connection = DriverManager.getConnection(
			DB_URL_BASE + ";create=true",
			properties
		);
		connection.setAutoCommit( true );
		return connection;
	}

	/** Query on the connection and compare results. */
	public void verifyQueryResult (
		final Connection connection,
		final String query,
		final Object[][] expected
	) {
		try {
			final Statement statement = connection.createStatement();
			final ResultSet resultSet = statement.executeQuery( query );
			try {
				for ( int i = 0; i < expected.length; ++i ) {
					assertTrue( "Row[" + i + "] expected", resultSet.next() );
					for ( int j = 0; j < expected[i].length; ++j ) {
						if ( expected[i][j] == null ) {
							assertNull(
								"Null compare row=" + i + " column=" + j
								+ " value=" + resultSet.getObject( j + 1 ),
								resultSet.getObject( j + 1 )
							);
						} else if ( expected[i][j] instanceof String ) {
							assertEquals(
								"String compare row=" + i + " column=" + j,
								(String) expected[i][j],
								resultSet.getString( j + 1 )
							);
						} else if ( expected[i][j] instanceof Integer ) {
							assertEquals(
								"Integer compare row=" + i + " column=" + j,
								( (Integer) expected[i][j] ).intValue(),
								resultSet.getInt( j + 1 )
							);
						} else if ( expected[i][j] instanceof Double ) {
							assertEquals(
								"Double compare row=" + i + " column=" + j,
								( (Double) expected[i][j] ).doubleValue(),
								resultSet.getDouble( j + 1 ),
								0
							);
						} else if ( expected[i][j] instanceof Date ) {
							assertEquals(
								"Date compare row=" + i + " column=" + j,
								(Date) expected[i][j],
								resultSet.getTimestamp( j + 1 )
							);
						} else if ( expected[i][j] instanceof Boolean ) {
							assertEquals(
								"Boolean compare row=" + i + " column=" + j,
								( (Boolean) expected[i][j] ).booleanValue(),
								resultSet.getBoolean( j + 1 )
							);
						} else {
							fail(
								"Unimplemented datatype "
								+ resultSet.getObject( j + 1 ).getClass()
								+ " row=" + i + " column=" + j
							);
						}
					}
					try {
						resultSet.getString( expected[i].length + 1 );
						fail( "Too many columns reported" );
					}
					catch ( final SQLException se ) {}
				}
				assertTrue( "No more rows expected", !resultSet.next() );
			}
			finally {
				resultSet.close();
				statement.close();
			}
		}
		catch ( final SQLException sqle ) {
			fail(
				"Querying SQL failed at query: \""
				+ query + "\": " + sqle
			);
		}
	}

	/** Query on a transient connection and compare results. */
	public void verifyQueryResult (
		final String query,
		final Object[][] expected
	) {
		try {
			final Connection connection = createConnection();
			try {
				verifyQueryResult( connection, query, expected );
			}
			finally {
				connection.close();
			}
		}
		catch ( final SQLException sqle ) {
			fail(
				"Creating or closing SQL connection failed: " + sqle
			);
		}
	}

	/** Execute a number of statements on the connection. */
	public void executeStatements (
		final Connection connection,
		final String[] statements,
		final boolean ignoreExceptions
	) {
		try {
			final Statement statement = connection.createStatement();
			try {
				for ( int i = 0; i < statements.length; ++i ) {
					try {
						statement.execute( statements[i] );
					}
					catch ( final SQLException sqle ) {
						if ( ignoreExceptions ) {
							System.err.println(
								"Warning: SQL statement[" + i + "]: \""
								+ statements[i] + "\": " + sqle
							);
						} else {
							fail(
								"Executing SQL failed at statement[" + i + "]: \""
								+ statements[i] + "\": " + sqle
							);
						}
					}
				}
			}
			finally {
				statement.close();
			}
		}
		catch ( final SQLException sqle ) {
			if ( ignoreExceptions ) {
				System.err.println(
					"Warning: SQL statements: " + sqle
				);
			} else {
				fail( "Executing SQL failed: " + sqle );
			}
		}
	}

	/** Execute a number of statements on a transient connection. */
	public void executeStatements (
		final String[] statements,
		final boolean ignoreExceptions
	) {
		try {
			final Connection connection = createConnection();
			try {
				executeStatements( connection, statements, ignoreExceptions );
			}
			finally {
				connection.close();
			}
		}
		catch ( final SQLException sqle ) {
			fail(
				"Creating or using or closing SQL connection failed: " + sqle
			);
		}
	}

	public void tearDown () {
		try {
			DriverManager.deregisterDriver( driver );

			final Properties properties = System.getProperties();
			properties.remove( "derby.system.home" );

			assertTrue( "Delete local DB directory", recursiveDelete( testDB ) );
		}
		catch ( final SQLException sqle ) {
			fail( "Removing test db failed: " + sqle );
		}
	}

	/* ==================== Private Methods ===================== */

	private static boolean recursiveDelete ( final File file ) {
		boolean success = true;
		if ( file.isDirectory() ) {
			final File[] children = file.listFiles();
			for ( int i = 0; i < children.length; ++i ) {
				success &= recursiveDelete( children[i] );
			}
		}
		return success &= file.delete();
	}
}
