// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.sql;

import java.io.PrintWriter;
import java.io.FileWriter;
import java.io.IOException;

import Int.ecb.logging.Level;

import Int.ecb.gesmescb.*;
import Int.ecb.gesmescb.event.*;

import Int.ecb.gesmescb.util.Named;
import Int.ecb.gesmescb.util.ComponentHelper;

/* **************************** Class ***************************** */

/**
 * SQL writer for metadata.
 *
 * @author
 * <A HREF="mailto:Sami.Airo@ECB.int">Sami Airo</A>
 *
 * @version @(#)$Id: SQLWriter.java,v 1.1 2017/01/19 16:14:02 trainot Exp $
 */

public abstract class SQLWriter extends GESMESCBEventListenerAdapter
	implements Named, StructureListener, DataListener, GESMESCBWriter
{

	/* ======================= Constants ======================== */

	public static final int SQL_SYNTAX_SQL92 = 1;
	public static final int SQL_SYNTAX_ORACLE8 = 2;

	/* ========================= Fields ========================= */

	/* Takes care of logging and prefs */
	protected ComponentHelper componentHelper;

	protected PrintWriter writer;
	private FileWriter fileWriter;
	private String filename;
	protected int currentSQLSyntax = SQL_SYNTAX_SQL92;

	/* ====================== Constructors ====================== */

	/* ===================== Public Methods ===================== */

	public String getName () {
		return componentHelper.name;
	}

	public void close () {
		componentHelper.logger.log(
			Level.FINE,
			componentHelper.logName,
			"close",
			"Closing."
		);

		// Remove the sources.
		setActive( false );

		try {
			// Flush the buffer.
			writer.close();

			if ( filename != null && fileWriter != null ) {
				fileWriter.close();
				componentHelper.logger.log(
					Level.INFO,
					componentHelper.logName,
					"close()",
					"Closed SQL output"
					+ (
						filename != null
						? " to " + filename
						: ""
						)
					+ '.'
				);
			}
		}
		catch ( IOException ioe ) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"close()",
				"Output problem.",
				ioe
			);
		}
	}


	public String getSQLString(final String string) {
		if (string.equals("NULL")) return string;

		StringBuffer buffer = new StringBuffer("'");
		final int stringLen = string.length();

		for ( int i = 0; i < stringLen; ++i ) {
			switch ( string.charAt( i ) ) {
			case '\'':
				buffer.append('\'');
			default:
				buffer.append( string.charAt( i ) );
			}
		}

		buffer.append('\'');

		return buffer.toString();
	}

	public String getStatementTerminator () {
		if ( currentSQLSyntax == SQL_SYNTAX_ORACLE8 ) return ";";
		return "";
	}

	public void handle ( final InterchangeBegin interchangeBegin ) {
		componentHelper.logger.log(
			Level.FINE,
			componentHelper.logName,
			"handle",
			"Interchange begins"
		);
	}

	public void handle ( final InterchangeEnd interchangeEnd ) {
		componentHelper.logger.log(
			Level.FINE,
			componentHelper.logName,
			"handle",
			"Interchange ends"
		);
	}

	public void handle ( final MessageBegin messageBegin ) {
		componentHelper.logger.log(
			Level.FINE,
			componentHelper.logName,
			"handle",
			"Message begins"
		);
	}

	public void handle ( final MessageEnd messageEnd ) {
		componentHelper.logger.log(
			Level.FINE,
			componentHelper.logName,
			"handle",
			"Message ends"
		);
	}

	public void handle ( final StatisticalConcept statisticalConcept ) {
		componentHelper.logger.log(
			Level.FINE,
			componentHelper.logName,
			"handle",
			"Handling statistical concept"
		);
	}

	public void handle ( final CodeList codeList ) {
		componentHelper.logger.log(
			Level.FINE,
			componentHelper.logName,
			"handle",
			"Handling code list"
		);
	}

	public void handle ( final KeyFamily keyFamily ) {
		componentHelper.logger.log(
			Level.FINE,
			componentHelper.logName,
			"handle",
			"Handling key family"
		);
	}

	public void handle ( final TimeSeries timeSeries ) {
		componentHelper.logger.log(
			Level.FINE,
			componentHelper.logName,
			"handle",
			"Handling time series"
		);
	}

	public void handle ( final Attribute attribute ) {
		componentHelper.logger.log(
			Level.FINE,
			componentHelper.logName,
			"handle",
			"Handling attribute"
		);
	}

	protected void setSQLSyntax( final String syntax ) {
		componentHelper.logger.log(
			Level.CONFIG,
			componentHelper.logName,
			"setSQLSyntax",
			"Using SQL syntax: " + syntax
		);

		if (syntax.equals("SQL92")) {
			currentSQLSyntax = SQL_SYNTAX_SQL92;
			return;
		}

		if (syntax.equals("ORACLE8")) {
			currentSQLSyntax = SQL_SYNTAX_ORACLE8;
			return;
		}

		throw new IllegalArgumentException("Invalid SQL syntax option: " + syntax);
	}

	/* =================== Protected Methods ==================== */

	protected void open ()
		throws IOException
	{
		filename = componentHelper.getPreference( "destination", "" );
		if ( filename.equals( "" ) ) {
			final String desc = "No preference variable \"destination\" found to determine the output file.";
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"getColumnDefinition",
				desc
			);
			throw new NullPointerException( desc );
		}
		fileWriter = new FileWriter( filename );
		writer = new PrintWriter( fileWriter );
	}
}
