// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.time;

import java.util.Calendar;
import java.util.Date;
import java.util.Iterator;
import java.util.NoSuchElementException;

import Int.ecb.gesmescb.util.PackageProperties;
import Int.ecb.logging.Level;
import Int.ecb.logging.Logger;
import Int.ecb.prefs.Preferences;

/* ***************************** Class ****************************** */

/**
 * <P>Provides a sceletal implementation of {@link RegularInterval}.</P>
 *
 * <P>For peculiar date calculations, a {@link Calendar} is used,
 * which must be supplied by the subclass via {@link #getCalendar}.
 * To create such a <CODE>Calendar</CODE> and also for other exercises,
 * implementations of <CODE>RegularInterval</CODE>
 * can take advantage of a few static helper methods.</P>
 *
 * @author
 * <A HREF="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</A>
 *
 * @version @(#)$Id: AbstractRegularInterval.java,v 1.1 2017/01/19 16:14:03 trainot Exp $
 */
public abstract class AbstractRegularInterval extends AbstractInterval
	implements RegularInterval
{
	/* ======================= Constants ======================== */

	/** <P>Minutes per day.</P>
	 *
	 * <P>Note that due to change of the earth's rotation speed
	 * the number of seconds per day may vary.
	 * Not so the number of minutes per day, because
	 * additional seconds are always added to the last minute of the year.
	 * So the length of a minute may vary, not the multiplicity.</P>
	 */
	private static final int MINUTES_PER_DAY = 24 * 60;

	/** Whether to use a lenient calendar.
	 * Use the "useLenient" option only if you know what you are doing.
	 * If set, various invalid time formats will not be recognised.
	 * @deprecated USE_LENIENT is not used in this class anymore.
         * Use {@link CalendarHelper#USE_LENIENT} instead
	 */
	public static final boolean USE_LENIENT;

	/* =================== Static Initializer =================== */

	static {
		final String packageName = PackageProperties.getName(AbstractRegularInterval.class.getName());
		final Logger logger = Logger.getLogger( packageName );
		final Preferences
			userPrefs = Preferences.userRoot().node( packageName ),
			systemPrefs = Preferences.systemRoot().node( packageName );
		final String useLenient = systemPrefs.get(
			"useLenient",
			userPrefs.get( "useLenient", "false" )
		);
		USE_LENIENT = "true".equalsIgnoreCase( useLenient );
		logger.log(
			USE_LENIENT ? Level.CONFIG : Level.FINE,
			"AbstractRegularInterval",
			"static initialiser",
			"Using lenient calendars: " + USE_LENIENT
		);
	}

	/* ====================== Constructors ====================== */

	/** Construct one. */
	protected AbstractRegularInterval () {
		super();
	}

	/* ===================== Public Methods ===================== */

	/**
	 * <P>Get the base period frequency.</P>
	 *
	 * <P>Subclasses are expected to implement this method.</P>
	 */
	public abstract int getFrequency ();

	/**
	 * <P>Get the begin point of this {@link RegularInterval}.</P>
	 *
	 * <P>Subclasses of <CODE>AbstractRegularInterval</CODE>
	 * are expected to implement this method
	 * because on it many other methods will be based.</P>
	 */
	public abstract Date getBegin ();

	/**
	 * <P>Get the size measured in base periods.</P>
	 *
	 * <P>Subclasses are expected to implement this method.</P>
	 */
	public abstract int getSize ();

	public Date getEnd () {
		final Calendar calendar = getCalendar();
		advanceCalendar( calendar, getFrequency(), getSize() );
		return calendar.getTime();
	}

	public int getFirstPeriodYear () {
		return getPeriodYear( getCalendar(), getFrequency() );
	}

	public int getFirstPeriodIndex () {
		return getPeriodIndex( getCalendar(), getFrequency() );
	}

	public RegularInterval computeBasePeriod ( final int index ) {
		final Calendar calendar = getCalendar();
		advanceCalendar( calendar, getFrequency(), index );
		return createRegularInterval(
			getFrequency(), calendar.getTime(), 1
		);
	}

	public int computeIndex ( final RegularInterval interval ) {
		final BasePeriodSearch basePeriodSearch
		= new BasePeriodSearch( this );
		return basePeriodSearch.findIndex( interval );
	}

	public Iterator iterator () {
		return new Iterator () {
			final Calendar calendar = getCalendar();
			int index = 0;
			public boolean hasNext () {
				return index < getSize();
			}
			public void remove () {
				throw new UnsupportedOperationException(
					"Cannot remove base periods from a RegularInterval."
				);
			}
			public Object next () {
				if ( hasNext() ) {
					final RegularInterval ri = createRegularInterval(
						getFrequency(), calendar.getTime(), 1
					);
					advanceCalendar( calendar, getFrequency(), 1 );
					++index;
					return ri;
				} else {
					throw new NoSuchElementException(
						"No further base period in RegularInterval."
					);
				}
			}
		};
	}

	public boolean isAligned () {
			final Date begin = getBegin();
			final Calendar calendar = createISOCalendar();
			calendar.setTime( begin );
			projectCalendar( calendar, getFrequency() );
			return begin.equals( calendar.getTime() );
	}

	/**
	 * Tests whether this object basically equals that.
	 *
	 * @param that object which will be compared.
	 *
	 * @return whether <CODE>that</CODE> is a {@link RegularInterval}
	 * and has the same frequency, begin, and size.
	 */
	public boolean equals ( final Object that ) {
		if ( that instanceof RegularInterval ) {
			final RegularInterval other = (RegularInterval) that;
			return
				getFrequency() == other.getFrequency()
				&&
				getSize() == other.getSize()
				&&
				getBegin().equals( other.getBegin() );
		} else return false;
	}

	/**
	 * Computes a hash code compatible with {@link #equals(java.lang.Object)}.
	 */
	public int hashCode () {
		return
			1271 * getFrequency()
			^
			getBegin().hashCode()
			^
			1343 * getSize();
	}

	/** Gives a human readable representation. */
	public String toString () {
		String word;
		switch ( getFrequency() ) {
			case RegularInterval.ANNUAL: word = "year"; break;
			case RegularInterval.SEMIANNUAL: word = "halfyear"; break;
			case RegularInterval.QUARTERLY: word = "quarter"; break;
			case RegularInterval.MONTHLY: word = "month"; break;
			case RegularInterval.WEEKLY: word = "week"; break;
			case RegularInterval.DAILY: word = "day"; break;
			case RegularInterval.MINUTELY: word = "minute"; break;
			default: word = "period";
		}
		return
			getClass().getName() + "("
			+ getSize() + " " + word + ( getSize() == 1 ? "" : "s" )
			+ " starting at " + getBegin() + ")";
	}

	/* =================== Protected Methods ==================== */

	/**
	 * <P>Get an ISO <CODE>Calendar</CODE> which is set to
	 * the begin point of this {@link RegularInterval}.</P>
	 */
	protected Calendar getCalendar () {
		final Calendar calendar = createISOCalendar();
		calendar.setTime( getBegin() );
		return calendar;
	}

	/**
	 * <P>Create a <CODE>RegularInterval</CODE>.</P>
	 *
	 * <P>This factory method is called in functions which require
	 * the creation of new <CODE>RegularInterval</CODE>s
	 * by the <CODE>AbstractRegularInterval</CODE>.
	 * Unless overridden, it creates a {@link GregorianRegularInterval}.
	 * Subclasses may override it to use other implementations,
	 * e.g. for better performance.
	 */
	protected RegularInterval createRegularInterval (
		final int frequency, final Date begin, final int size
	) {
		return new GregorianRegularInterval(
			frequency, begin, size
		);
	}

	/* ---------------- Protected Static Methods ---------------- */

	/**
	 * <P>Create a <CODE>Calendar</CODE> with proper
	 * settings.</P>
	 *
	 * <P>This means ISO weeks are respected.
	 * The returned <CODE>Calendar</CODE> is not lenient by default.
	 * Also it is {@link Calendar#clear}ed so that it
	 * points to the epoch rather than the date of its creation.
	 * The method <CODE>clear</CODE> seems to set to
	 * the epoch of the {@link java.util.TimeZone}
	 * that <CODE>Calendar</CODE> was created in.</P>
	 *
	 * @return a new, ISO conformant <CODE>Calendar</CODE>.
	 */
	public static Calendar createISOCalendar () {
		return CalendarHelper.createISOCalendar();
	}

	/** Get the year a base period lies in.
	 * The base period is characterised by a <CODE>Calendar</CODE>
	 * which must contain the base period's begin point.
	 * Since weeks are not aligned with years,
	 * the begin point may lie in a different year than
	 * the year returned for weekly frequencies.
	 *
	 * <P>Note that the behaviour of this operation is sensitive to
	 * the week settings. Best use only calendars created using
	 * {@link #createISOCalendar} to obtain predictable results.</P>
	 *
	 * @param calendar specifying the base period's begin point.
	 * @param frequency of the base period.
	 *
	 * @see #getPeriodIndex
	 */
	protected static int getPeriodYear (
		Calendar calendar, final int frequency
	) {
		if ( frequency == WEEKLY ) {
			calendar = (Calendar) calendar.clone();
			calendar.add( Calendar.DAY_OF_WEEK, 3 );
		}
		return calendar.get( Calendar.YEAR );
	}

	/** Get the index of a base period within the year it lies in.
	 * The base period is characterised by a <CODE>Calendar</CODE>
	 * which must contain the base period's begin point.
	 * Indexing starts with 0.
	 *
	 * @param calendar specifying the base period's begin point.
	 * @param frequency of the base period.
	 *
	 * @see #getPeriodYear
	 */
	protected static int getPeriodIndex (
		final Calendar calendar, final int frequency
	) {
		switch ( frequency ) {
			case ANNUAL:
				return 0;
			case SEMIANNUAL:
				return calendar.get( Calendar.MONTH ) < Calendar.JULY ? 0 : 1;
			case QUARTERLY:
				final int month = calendar.get( Calendar.MONTH );
				if ( month < Calendar.JULY ) {
					return month < Calendar.APRIL ? 0 : 1;
				} else {
					return month < Calendar.OCTOBER ? 2 : 3;
				}
			case MONTHLY:
				return calendar.get( Calendar.MONTH );
			case WEEKLY:
				return calendar.get( Calendar.WEEK_OF_YEAR ) - 1;
			case DAILY:
				return calendar.get( Calendar.DAY_OF_YEAR ) - 1;
			case MINUTELY:
				return
					( calendar.get( Calendar.DAY_OF_YEAR ) - 1 )
						* MINUTES_PER_DAY
					+ calendar.get( Calendar.HOUR_OF_DAY ) * 60
					+ calendar.get( Calendar.MINUTE );
			default:
				throw new IllegalArgumentException( "Unknown frequency code" );
		}
	}

	/** <P>Set a <CODE>Calendar</CODE> to the begin point of a base period.
	 * The base period is given by its frequency,
	 * the year it lies in and by its index in this year.</P>
	 *
	 * <P>Note that the behaviour of this operation is sensitive to
	 * the week settings. Best use only calendars created using
	 * {@link #createISOCalendar} to obtain predictable results.</P>
	 *
	 * <P>Note that by use of {@link Calendar#setLenient}
	 * the checking behaviour can be influenced. Unfortunately
	 * the specification does not say which kind of {@link Exception}
	 * would be used.</P>
	 *
	 * @param calendar to be set.
	 * @param frequency determines the length of the base period.
	 * @param year the base period should lie in.
	 * @param index of the baseperiod within <CODE>year</CODE>.
	 * Indexing starts with 0.
	 *
	 * @see #getPeriodYear
	 * @see #getPeriodIndex
	 */
	protected static void setCalendar (
		final Calendar calendar,
		final int frequency,
		final int year,
		final int index
	) {
		calendar.clear();
		calendar.set( Calendar.YEAR, year );
		switch ( frequency ) {
			case ANNUAL:
				if ( calendar.isLenient() ) {
					calendar.add( Calendar.YEAR, index );
				} else {
					if ( index != 0 ) {
						throw new IllegalArgumentException(
							"There is only one year (indexed 0) in a year"
						);
					}
				}
				break;
			case SEMIANNUAL:
				calendar.set( Calendar.MONTH, 6 * index );
				break;
			case QUARTERLY:
				calendar.set( Calendar.MONTH, 3 * index );
				break;
			case MONTHLY:
				calendar.set( Calendar.MONTH, index );
				break;
			case WEEKLY:
				// TODO: put non-lenient checking like this into a factory.
				if (
					! calendar.isLenient()
					&& (
						index + 1 < calendar.getActualMinimum(
							Calendar.WEEK_OF_YEAR
						)
						||
						index + 1 > calendar.getActualMaximum(
							Calendar.WEEK_OF_YEAR
						)
					)
				) {
					throw new IllegalArgumentException(
						"Week index " + index + " out of range."
					);
				}
				calendar.set( Calendar.WEEK_OF_YEAR, index + 1 );
				// Unfortunately, GregorianCalendar needs the day of week
				// also set here. It does not automatically use the first
				// day of week.
				calendar.set( Calendar.DAY_OF_WEEK, Calendar.MONDAY );
				sanitiseCalendar( calendar );
				break;
			case DAILY:
				calendar.set( Calendar.DAY_OF_YEAR, index + 1 );
				break;
			case MINUTELY:
				final int dayIndex = index / MINUTES_PER_DAY;
				final int hourOffset = ( index - dayIndex * MINUTES_PER_DAY ) / 60;
				final int minuteOffset
				= index - dayIndex * MINUTES_PER_DAY - hourOffset * 60;
				calendar.set( Calendar.DAY_OF_YEAR, dayIndex + 1 );
				calendar.set( Calendar.HOUR_OF_DAY, hourOffset );
				calendar.set( Calendar.MINUTE, minuteOffset );
				break;
			default:
				throw new IllegalArgumentException( "Unknown frequency code" );
		}
	}

	/** <P>Advance a <CODE>Calendar</CODE> a number of base periods
	 * and ensure that the result is at the beginning of a base period.
	 * The <CODE>Calendar</CODE>
	 * must contain the original base period's begin point.</P>
	 *
	 * @param calendar to be set.
	 * @param frequency determines the length of the base period.
	 * @param number of baseperiods to add. This may be negative, too.
	 */
	protected static void advanceCalendar (
		final Calendar calendar,
		final int frequency,
		final int number
	) {
		CalendarHelper.advanceCalendar(calendar, frequency, number);
	}

	/** <P>Computes the least number of frequency-determined increments
	 * so that the resulting time is after the given <CODE>Date</CODE>.</P>
	 *
	 * <P>The handed-over <CODE>Calendar</CODE> may be changed in an
	 * unspecified way.</P>
	 *
	 * @param calendar to be set.
	 * @param frequency determines the length of the base period.
	 * @param date to be indexed.
	 */
	protected static int computeIndex (
		final int frequency,
		final Calendar calendar,
		final Date date
	) {
		final CalendarSearch calendarSearch
		= new CalendarSearch( frequency, calendar );
		return calendarSearch.findIndex( date );
	}

	/** <P>Set a <CODE>Calendar</CODE> to the base period's begin point.</P>
	 *
	 * <P>This operation acts idempotent on <CODE>Calendar</CODE>s
	 * which means twice execution yields the same one time.
	 * The operation therefore can be called a projection.</P>
	 *
	 * <P>Note that the behaviour of this operation is sensitive to
	 * the week settings. Best use only calendars created using
	 * {@link #createISOCalendar} to obtain predictable results.</P>
	 * 
	 * <P>This method makes no exception for periods with begin point 
	 * 01:00:00 AM (e.g. switch to daylight saving), and sets the period's
	 * begin point at 01:00:00 AM.</P> 
	 *
	 * @param calendar to be set.
	 * @param frequency determines the base period grid.
	 */
	public static void projectCalendar (
		final Calendar calendar,
		final int frequency
	) {
		final int year = calendar.get( Calendar.YEAR );
		final int month = calendar.get( Calendar.MONTH );
		final int week = calendar.get( Calendar.WEEK_OF_YEAR );
		final int day = calendar.get( Calendar.DAY_OF_MONTH );
		final int hour = calendar.get( Calendar.HOUR_OF_DAY );
		final int minute = calendar.get( Calendar.MINUTE );

		calendar.clear();

		switch ( frequency ) {
			case MINUTELY:
				calendar.set( Calendar.YEAR, year );
				calendar.set( Calendar.MONTH, month );
				calendar.set( Calendar.DAY_OF_MONTH, day );
				calendar.set( Calendar.HOUR_OF_DAY, hour );
				calendar.set( Calendar.MINUTE, minute );
				break;
			case DAILY:
				calendar.set( Calendar.YEAR, year );
				calendar.set( Calendar.MONTH, month );
				calendar.set( Calendar.DAY_OF_MONTH, day );
				break;
			case WEEKLY:
				calendar.set( Calendar.YEAR, year );
				calendar.set( Calendar.WEEK_OF_YEAR, week );
				// Unfortunately, GregorianCalendar needs the day of week
				// also set here. It does not automatically use the first
				// day of week.
				calendar.set( Calendar.DAY_OF_WEEK, Calendar.MONDAY );
				sanitiseCalendar( calendar );
				break;
			case MONTHLY:
				calendar.set( Calendar.YEAR, year );
				calendar.set( Calendar.MONTH, month );
				break;
			case QUARTERLY:
				calendar.set( Calendar.YEAR, year );
				calendar.set( Calendar.MONTH, ( month / 3 ) * 3 );
				break;
			case SEMIANNUAL:
				calendar.set( Calendar.YEAR, year );
				calendar.set( Calendar.MONTH, ( month / 6 ) * 6 );
				break;
			case ANNUAL:
				calendar.set( Calendar.YEAR, year );
				break;
			default:
				throw new IllegalArgumentException( "Unknown frequency code" );	
		}
	}

	/** <P>Set a <CODE>Calendar</CODE> to the beginning of a day.</P>
	 *
	 * <P>This operation acts idempotent on <CODE>Calendar</CODE>s
	 * which means twice execution yields the same one time.
	 * The operation therefore can be called a projection.</P>
	 *
	 * <P>This method is useful to clean the result from daylight saving switch
	 * artefacts when using base daily or longer period.
	 * Days get dayProjected on their beginning point, be it 01:00:00 AM (e.g.
	 * switch to daylight saving), or 00:00:00 AM.</P> 
	 *
	 * @param calendar to be set.
	 * @param frequency determines the base period grid.
	 */
	public static void dayProjectCalendar (
		final Calendar calendar
	) {
		CalendarHelper.dayProjectCalendar(calendar);
	}
	
	/* ==================== Private Methods ===================== */

	/** Work around a Java 1.5 misbehaviour.
	 * Java 1.5 non-lenient calendars' getTime() detects
	 * an illegal year when the first week starts outside the year.
	 */
	private static void sanitiseCalendar ( final Calendar calendar ) {
	    CalendarHelper.sanitiseCalendar(calendar);
	}
}

/* ***************************** Class ****************************** */

/**
 * Performs a double binary search in indexed ordered structures.
 * The type of structure is subclass-specific.
 */
abstract class BinarySearch {

	/* =================== Protected Methods ==================== */

	/**
	 * Compare the <CODE>index</CODE>-th element with the <CODE>target</CODE>.
	 * The return code is positive, zero
	 *
	 * @return a negative integer, zero, or a positive integer as the
	 * <CODE>index</CODE>-th object is less than, equal to, or greater
	 * than the <CODE>target</CODE>.
	 *
	 * @throws ClassCastException if target is of the wrong type.
	 */
	protected abstract int compareTo ( int index, Object target );

	/* ===================== Public Methods ===================== */

	/**
	 * Search for the <CODE>target</CODE>.
	 *
	 * The returned <CODE>index</CODE> can be any <CODE>int</CODE> value.
	 * Subclasses may, however, only use a subset of all possible indexes.
	 *
	 * @return the <CODE>index</CODE> of the <CODE>target</CODE>.
	 *
	 * @throws ClassCastException if target is of the wrong type.
	 * @throws IllegalArgumentException if the index is out of the
	 * <CODE>int</CODE> range.
	 * @throws NoSuchElementException if the target is not found.
	 */
	public int findIndex ( final Object target ) {

		final boolean nonnegativeIndex;
		{
			final int cmpv = compareTo( 0, target );
			if ( cmpv == 0 ) return 0;
			nonnegativeIndex = cmpv <= 0;
			// System.out.println( "nonnegativeIndex = " + nonnegativeIndex );
		}

		int bit, guess;
		if ( nonnegativeIndex ) {
			for ( bit = 0; bit < 32; ++bit ) {
				final int index = ( 1 << bit ) - 1;
				final int cmpv = compareTo( index, target );
				if ( cmpv > 0 ) break;
				if ( cmpv == 0 ) return index;
				// System.out.println( "bit = " + bit );
			}
			if ( bit >= 32 ) {
				throw new IllegalArgumentException(
					"The resulting index would be too large for int."
				);
			}
			guess = bit > 0 ? 1 << --bit : 0;	// because 1 << -1 != 0 !!
		} else {
			for ( bit = 0; bit < 32; ++bit ) {
				final int index = -1 << bit;
				final int cmpv = compareTo( index, target );
				if ( cmpv < 0 ) break;
				if ( cmpv == 0 ) return index;
				// System.out.println( "bit = " + bit );
			}
			if ( bit >= 32 ) {
				throw new IllegalArgumentException(
					"The resulting index would be too small for int."
				);
			}
			guess = -1 << bit--;
		}
		while ( bit > 0 ) {
			// System.out.println( "guess = " + guess + "\tbit = " + bit );
			final int next = guess + ( 1 << --bit );
			final int cmpv = compareTo( next, target );
			if ( cmpv <= 0 ) guess = next;
			if ( cmpv == 0 ) break;
		}
		// System.out.println( "final guess = " + guess );
		if ( compareTo( guess, target ) != 0 ) {
			throw new NoSuchElementException(
				"Target not found in indexed structure."
			);
		}
		return guess;
	}
}

/* ***************************** Class ****************************** */

/**
 * <P>Searches in a {@link Calendar}-based grid
 * the base period which a {@link Date} lies in.</P>
 *
 * <P>The <CODE>index</CODE> returned by {@link #findIndex} equals
 * the least number of increments to the <CODE>Calendar</CODE> until it points
 * to a <CODE>Date</CODE> later than the given one.
 * This number can be negative.</P>
 *
 * <P>The method {@link #compareTo} uses an intelligent caching strategy
 * adapted to potentially optimize {@link BinarySearch#findIndex}.</P>
 *
 * <P>This class is used to implement
 * {@link RegularInterval#computeIndex(Calendar,int,Date)}.</P>
 */
final class CalendarSearch extends BinarySearch {

	/* ========================= Fields ========================= */

	/** The origin of the grid. */
	private final Calendar calendar;

	/** The frequency of the grid. */
	private final int frequency;

	/* -------------------- Transient Fields -------------------- */

	/** Cache for last approximations. */
	private transient Calendar lastOnZeroSide;

	/** Index of cached last approximations. */
	private int lastOnZeroSideIndex;

	/* ====================== Constructors ====================== */

	public CalendarSearch (
		final int frequency, final Calendar calendar
	) {
		this.calendar = calendar;
		this.frequency = frequency;
	}

	/* ===================== Public Methods ===================== */

	public int findIndex ( final Object target ) {
		lastOnZeroSide = null;
		lastOnZeroSideIndex = 0;
		return super.findIndex( target );
	}

	/* =================== Protected Methods ==================== */

	protected int compareTo ( final int index, final Object target ) {
		return compareTo ( index, (Date) target );
	}

	/* ==================== Private Methods ===================== */

	private int compareTo ( final int index, final Date target ) {

		final Calendar guess;

		if ( lastOnZeroSide == null ) {
			lastOnZeroSide = (Calendar) calendar.clone();
			lastOnZeroSideIndex = 0;
		}

		if ( index == lastOnZeroSideIndex ) {
			guess = lastOnZeroSide;
		} else {
			guess = (Calendar) lastOnZeroSide.clone();
			AbstractRegularInterval.advanceCalendar(
				guess, frequency, index - lastOnZeroSideIndex
			);
		}

		final int cmpv = guess.getTime().compareTo( target );

		// Cache base period only if it is not farther away from 0.
		// This may improve reliability, since the path is always in
		// the range from 0 to the resulting index.
		if (
			index >= 0 && cmpv <= 0
			||
			index <= 0 && cmpv >= 0
		) {
			lastOnZeroSide = guess;
			lastOnZeroSideIndex = index;
		}

		return cmpv;
	}
}

/* ***************************** Class ****************************** */

/**
 * <P>Searches a base period with respect to a {@link RegularInterval}.</P>
 *
 * <P>The resulting index can be outside the range from
 * <CODE>0</CODE> to <CODE>{@link RegularInterval#getSize}-1</CODE>.</P>
 *
 * <P>The method {@link #compareTo} uses an intelligent caching strategy
 * adapted to potentially optimize {@link BinarySearch#findIndex}.</P>
 *
 * <P>This class is used to implement
 * {@link AbstractRegularInterval#computeIndex}.</P>
 */
final class BasePeriodSearch extends BinarySearch {

	/* ========================= Fields ========================= */

	/** The interval to search in. */
	private final RegularInterval interval;

	/* -------------------- Transient Fields -------------------- */

	/** Cache for last approximations. */
	private transient RegularInterval lastOnZeroSide;

	/** Index of cached last approximations. */
	private transient int lastOnZeroSideIndex;

	/* ====================== Constructors ====================== */

	/** Set the interval to search in. */
	public BasePeriodSearch ( final RegularInterval interval ) {
		this.interval = interval;
	}

	/* ===================== Public Methods ===================== */

	public int findIndex ( final Object target ) {
		lastOnZeroSide = null;
		lastOnZeroSideIndex = 0;
		return super.findIndex( target );
	}

	/* =================== Protected Methods ==================== */

	protected int compareTo ( final int index, final Object target ) {

		final RegularInterval guess;

		// If possible, base on cached base period.
		if ( lastOnZeroSide != null ) {
			if ( index == lastOnZeroSideIndex ) {
				guess = lastOnZeroSide;
			} else {
				guess = lastOnZeroSide.computeBasePeriod(
					index - lastOnZeroSideIndex
				);
			}
		} else {
			guess = interval.computeBasePeriod( index );
		}

		final int cmpv = guess.compareTo( target );

		// Cache base period only if it is not farther away from 0.
		// This may improve reliability, since the path is always in
		// the range from 0 to the resulting index.
		if (
			index >= 0 && cmpv <= 0
			||
			index <= 0 && cmpv >= 0
		) {
			lastOnZeroSide = guess;
			lastOnZeroSideIndex = index;
		}

		return cmpv;
	}
}
