// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.time;

import java.util.Calendar;
import java.util.Date;
import java.util.GregorianCalendar;
import java.util.TimeZone;

import junit.framework.Test;
import junit.framework.TestSuite;

/* ***************************** Class ****************************** */

/**
 * Tests the class {@AbstractRegularInterval}.
 *
 * @author
 * <A HREF="mailto:Sami.Airo@ECB.int">Sami Airo</A>
 * @author
 * <A HREF="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</A>
 *
 * @version @(#)$Id: AbstractRegularIntervalTest.java,v 1.1 2017/01/19 16:14:03 trainot Exp $
 */

public class AbstractRegularIntervalTest extends AbstractIntervalTest {

	/* ====================== Constructors ====================== */

	/**
	 * Constructs a TimePeriodTest with the specified name.
	 *
	 * @param name Test case name
	 */
	public AbstractRegularIntervalTest ( final String name ) {
		super( name );
	}

	/* ===================== Public Methods ===================== */

	/**
	 * Assembles and returns a test suite for all
	 * the test methods of this test case.
	 *
	 * @return a non-null test suite
	 */
	public static Test suite () {
		TestSuite suite = new TestSuite( AbstractRegularIntervalTest.class );
		return suite;
	}

	/**
	 * Runs the test case.
	 */
	public static void main ( final String args[] ) {
		String[] testCaseName = { AbstractRegularIntervalTest.class.getName() };
		junit.textui.TestRunner.main( testCaseName );
	}

	/* ---------------------- Test Methods ---------------------- */

	/**
	 * Tests the alleged ISO Calendar.
	 */
	public void testCreateISOCalendar () {
		final Calendar calendar = AbstractRegularInterval.createISOCalendar();

		assertTrue( "Calendar must not be lenient", !calendar.isLenient() );
		assertEquals(
			"Week must start on Monday",
			Calendar.MONDAY, calendar.getFirstDayOfWeek()
		);
		assertEquals(
			"Minimal days in first week in respective year must be 4",
			4, calendar.getMinimalDaysInFirstWeek()
		);
		assertEquals( new Date( 0L ), new Date( 0L ) );

		{
			final Calendar c = Calendar.getInstance();
			c.clear();
			assertEquals(
				"Alleged new calendar must be cleared.",
				c.getTime(), calendar.getTime()
			);
		}
	}

	public void testGets () {
		final Calendar calendar = Calendar.getInstance();
		calendar.setLenient( false );
		calendar.clear();

		calendar.set( Calendar.YEAR, 2001 );
		assertEquals(
			"Year in annual",
			2001,
			AbstractRegularInterval.getPeriodYear( calendar, RegularInterval.ANNUAL )
		);
		assertEquals(
			"Index in annual",
			0,
			AbstractRegularInterval.getPeriodIndex( calendar, RegularInterval.ANNUAL )
		);

		calendar.set( Calendar.MONTH, Calendar.JULY );
		assertEquals(
			"Year in semiannual",
			2001,
			AbstractRegularInterval.getPeriodYear( calendar, RegularInterval.SEMIANNUAL )
		);
		assertEquals(
			"Index in semiannual",
			1,
			AbstractRegularInterval.getPeriodIndex( calendar, RegularInterval.SEMIANNUAL )
		);
		assertEquals(
			"Year in quarterly",
			2001,
			AbstractRegularInterval.getPeriodYear( calendar, RegularInterval.QUARTERLY )
		);
		assertEquals(
			"Index in quarterly",
			2,
			AbstractRegularInterval.getPeriodIndex( calendar, RegularInterval.QUARTERLY )
		);
		assertEquals(
			"Year in monthly",
			2001,
			AbstractRegularInterval.getPeriodYear( calendar, RegularInterval.MONTHLY )
		);
		assertEquals(
			"Index in monthly",
			6,
			AbstractRegularInterval.getPeriodIndex( calendar, RegularInterval.MONTHLY )
		);

		calendar.set( Calendar.DAY_OF_YEAR, 136 );
		assertEquals(
			"Year in daily",
			2001,
			AbstractRegularInterval.getPeriodYear( calendar, RegularInterval.DAILY )
		);
		assertEquals(
			"Index in daily",
			135,
			AbstractRegularInterval.getPeriodIndex( calendar, RegularInterval.DAILY )
		);
	}

	public void testGets2 () {
		final Calendar calendar = Calendar.getInstance();
		calendar.setLenient( false );
		calendar.clear();

		calendar.set( 1984, Calendar.NOVEMBER, 17 );
		assertEquals(
			"Year in daily",
			1984,
			AbstractRegularInterval.getPeriodYear( calendar, RegularInterval.DAILY )
		);
		assertEquals(
			"Index in daily",
			31+29+31+30 + 31+30+31+31 + 30+31+16,
			AbstractRegularInterval.getPeriodIndex( calendar, RegularInterval.DAILY )
		);

		calendar.set( 1984, Calendar.FEBRUARY, 29 );
		assertEquals(
			"Year in daily",
			1984,
			AbstractRegularInterval.getPeriodYear( calendar, RegularInterval.DAILY )
		);
		assertEquals(
			"Index in daily",
			31+28,
			AbstractRegularInterval.getPeriodIndex( calendar, RegularInterval.DAILY )
		);

		calendar.set( 2000, Calendar.JANUARY, 1, 0, 0 );
		assertEquals(
			"Year in minutely",
			2000,
			AbstractRegularInterval.getPeriodYear( calendar, RegularInterval.MINUTELY )
		);
		assertEquals(
			"Index in minutely",
			0,
			AbstractRegularInterval.getPeriodIndex( calendar, RegularInterval.MINUTELY )
		);

		calendar.add( Calendar.MINUTE, 100000 );
		assertEquals(
			"Year in minutely",
			2000,
			AbstractRegularInterval.getPeriodYear( calendar, RegularInterval.MINUTELY )
		);
		assertEquals(
			"Index in minutely",
			100000,
			AbstractRegularInterval.getPeriodIndex( calendar, RegularInterval.MINUTELY )
		);
	}

	public void testGetsWeekly () {
		final Calendar calendar = AbstractRegularInterval.createISOCalendar();

		calendar.set( 2001, Calendar.JANUARY, 1 );
		assertEquals(
			"Year in weekly",
			2001,
			AbstractRegularInterval.getPeriodYear( calendar, RegularInterval.WEEKLY )
		);
		assertEquals(
			"Index in weekly",
			0,
			AbstractRegularInterval.getPeriodIndex( calendar, RegularInterval.WEEKLY )
		);

		calendar.set( 2001, Calendar.NOVEMBER, 26 );
		assertEquals(
			"Year in weekly",
			2001,
			AbstractRegularInterval.getPeriodYear( calendar, RegularInterval.WEEKLY )
		);
		assertEquals(
			"Index in weekly",
			47,
			AbstractRegularInterval.getPeriodIndex( calendar, RegularInterval.WEEKLY )
		);

		calendar.set( 2001, Calendar.DECEMBER, 24 );
		assertEquals(
			"Year in weekly",
			2001,
			AbstractRegularInterval.getPeriodYear( calendar, RegularInterval.WEEKLY )
		);
		assertEquals(
			"Index in weekly",
			51,
			AbstractRegularInterval.getPeriodIndex( calendar, RegularInterval.WEEKLY )
		);

		calendar.set( 2001, Calendar.DECEMBER, 31 );
		assertEquals(
			"Year in weekly",
			2002,
			AbstractRegularInterval.getPeriodYear( calendar, RegularInterval.WEEKLY )
		);
		assertEquals(
			"Index in weekly",
			0,
			AbstractRegularInterval.getPeriodIndex( calendar, RegularInterval.WEEKLY )
		);
	}

	public void testSet () {
		final Calendar calendar = AbstractRegularInterval.createISOCalendar();
		final Calendar reference = AbstractRegularInterval.createISOCalendar();

		AbstractRegularInterval.setCalendar(
			calendar, RegularInterval.ANNUAL, 1967, 0
		);
		reference.set( 1967, Calendar.JANUARY, 1 );
		assertEquals(
			"Set annual", reference.getTime(), calendar.getTime()
		);

		AbstractRegularInterval.setCalendar(
			calendar, RegularInterval.SEMIANNUAL, 1967, 1
		);
		reference.set( 1967, Calendar.JULY, 1 );
		assertEquals(
			"Set semiannual", reference.getTime(), calendar.getTime()
		);

		AbstractRegularInterval.setCalendar(
			calendar, RegularInterval.QUARTERLY, 1967, 3
		);
		reference.set( 1967, Calendar.OCTOBER, 1 );
		assertEquals(
			"Set quarterly", reference.getTime(), calendar.getTime()
		);

		AbstractRegularInterval.setCalendar(
			calendar, RegularInterval.MONTHLY, 1967, 10
		);
		reference.set( 1967, Calendar.NOVEMBER, 1 );
		assertEquals(
			"Set monthly", reference.getTime(), calendar.getTime()
		);

		AbstractRegularInterval.setCalendar(
			calendar, RegularInterval.WEEKLY, 1967, 6
		);
		reference.set( 1967, Calendar.FEBRUARY, 13 ); // Mon, 13 Feb 1967
		assertEquals(
			"Set weekly", reference.getTime(), calendar.getTime()
		);

		AbstractRegularInterval.setCalendar(
			calendar, RegularInterval.WEEKLY, 2003, 0
		);
		reference.set( 2002, Calendar.DECEMBER, 30 );
		assertEquals(
			"Set weekly2", reference.getTime(), calendar.getTime()
		);

		AbstractRegularInterval.setCalendar(
			calendar, RegularInterval.DAILY, 1967, 13
		);
		reference.set( 1967, Calendar.JANUARY, 14 );
		assertEquals(
			"Set daily", reference.getTime(), calendar.getTime()
		);

		AbstractRegularInterval.setCalendar(
			calendar, RegularInterval.MINUTELY, 1967, 1000
		);
		reference.set( 1967, Calendar.JANUARY, 1 );
		reference.add( Calendar.MINUTE, 1000 );
		assertEquals(
			"Set minutely", reference.getTime(), calendar.getTime()
		);
	}

	public void testAdvance () {
		final Calendar calendar = AbstractRegularInterval.createISOCalendar();
		final Calendar reference = AbstractRegularInterval.createISOCalendar();

		calendar.set( 1980, Calendar.JANUARY, 1 );
		AbstractRegularInterval.advanceCalendar(
			calendar, RegularInterval.ANNUAL, 10
		);
		reference.set( 1990, Calendar.JANUARY, 1 );
		assertEquals(
			"Advance annual", reference.getTime(), calendar.getTime()
		);

		calendar.set( 1980, Calendar.JULY, 1 );
		AbstractRegularInterval.advanceCalendar(
			calendar, RegularInterval.SEMIANNUAL, 10
		);
		reference.set( 1985, Calendar.JULY, 1 );
		assertEquals(
			"Advance semiannual", reference.getTime(), calendar.getTime()
		);

		calendar.set( 1980, Calendar.APRIL, 1 );
		AbstractRegularInterval.advanceCalendar(
			calendar, RegularInterval.QUARTERLY, 10
		);
		reference.set( 1982, Calendar.OCTOBER, 1 );
		assertEquals(
			"Advance quarterly", reference.getTime(), calendar.getTime()
		);

		calendar.set( 1980, Calendar.JANUARY, 1 );
		AbstractRegularInterval.advanceCalendar(
			calendar, RegularInterval.MONTHLY, -100
		);
		reference.set( 1971, Calendar.SEPTEMBER, 1 );
		assertEquals(
			"Advance monthly", reference.getTime(), calendar.getTime()
		);

		calendar.set( 2000, Calendar.DECEMBER, 4 );
		AbstractRegularInterval.advanceCalendar(
			calendar, RegularInterval.WEEKLY, 20
		);
		reference.set( 2001, Calendar.APRIL, 23 );
		assertEquals(
			"Advance weekly", reference.getTime(), calendar.getTime()
		);

		calendar.set( 2000, Calendar.JANUARY, 1 );
		AbstractRegularInterval.advanceCalendar(
			calendar, RegularInterval.DAILY, -1
		);
		reference.set( 1999, Calendar.DECEMBER, 31 );
		assertEquals(
			"Advance daily", reference.getTime(), calendar.getTime()
		);

		calendar.set( 2000, Calendar.JANUARY, 1 );
		AbstractRegularInterval.advanceCalendar(
			calendar, RegularInterval.MINUTELY, 20000
		);
		reference.set( 2000, Calendar.JANUARY, 1 );
		reference.add( Calendar.MINUTE, 20000 );
		assertEquals(
			"Advance minutely", reference.getTime(), calendar.getTime()
		);
	}

	public void testProject () {
		final Calendar original = AbstractRegularInterval.createISOCalendar();
		Calendar calendar;
		final Calendar reference = AbstractRegularInterval.createISOCalendar();

		original.set( 1981, Calendar.NOVEMBER, 17, 13, 25, 20 );
		original.set( Calendar.MILLISECOND, 239 );

		calendar = (Calendar) original.clone();
		reference.set( 1981, Calendar.NOVEMBER, 17, 13, 25, 00 );
		AbstractRegularInterval.projectCalendar(
			calendar, RegularInterval.MINUTELY
		);
		assertEquals(
			"Project minutely", reference.getTime(), calendar.getTime()
		);

		calendar = (Calendar) original.clone();
		reference.set( 1981, Calendar.NOVEMBER, 17, 0, 0, 0 );
		AbstractRegularInterval.projectCalendar(
			calendar, RegularInterval.DAILY
		);
		assertEquals(
			"Project daily", reference.getTime(), calendar.getTime()
		);

		calendar = (Calendar) original.clone();
		reference.set( Calendar.DAY_OF_MONTH, 1 );
		AbstractRegularInterval.projectCalendar(
			calendar, RegularInterval.MONTHLY
		);
		assertEquals(
			"Project monthly", reference.getTime(), calendar.getTime()
		);

		calendar = (Calendar) original.clone();
		reference.set( Calendar.MONTH, Calendar.OCTOBER );
		AbstractRegularInterval.projectCalendar(
			calendar, RegularInterval.QUARTERLY
		);
		assertEquals(
			"Project quarterly", reference.getTime(), calendar.getTime()
		);

		calendar = (Calendar) original.clone();
		reference.set( Calendar.MONTH, Calendar.JULY );
		AbstractRegularInterval.projectCalendar(
			calendar, RegularInterval.SEMIANNUAL
		);
		assertEquals(
			"Project semiannual", reference.getTime(), calendar.getTime()
		);

		calendar = (Calendar) original.clone();
		reference.set( Calendar.MONTH, Calendar.JANUARY );
		AbstractRegularInterval.projectCalendar(
			calendar, RegularInterval.ANNUAL
		);
		assertEquals(
			"Project annual", reference.getTime(), calendar.getTime()
		);

		for ( int day = 1; day <= 14; ++day ) {
			calendar = (Calendar) original.clone();
			calendar.set( 2001, Calendar.JANUARY, day );
			// System.out.println( "1: " + calendar.getTime() );
			reference.set( 2001, Calendar.JANUARY, day <= 7 ? 1 : 8 );
			// System.out.println( "2: " + reference.getTime() );
			AbstractRegularInterval.projectCalendar(
				calendar, RegularInterval.WEEKLY
			);
			// System.out.println( "3: " + calendar.getTime() );
			assertEquals(
				"Project weekly1[" + day + "]",
				reference.getTime(), calendar.getTime()
			);
		}

		for ( int day = 1; day <= 7; ++day ) {
			calendar = (Calendar) original.clone();
			calendar.set( 2004, Calendar.JANUARY, day );
			if ( day <= 4 ) {
				reference.set( 2003, Calendar.DECEMBER, 29 );
			} else {
				reference.set( 2004, Calendar.JANUARY, 5 );
			}
			AbstractRegularInterval.projectCalendar(
				calendar, RegularInterval.WEEKLY
			);
			assertEquals(
				"Project weekly2[" + day + "]",
				reference.getTime(), calendar.getTime()
			);
		}
	}
	
	private static Calendar createRomanISOCalendar () {
		/* force timezone Rome. */
		final Calendar calendar = new GregorianCalendar(
			TimeZone.getTimeZone("Europe/Rome")
		);
		calendar.setFirstDayOfWeek( Calendar.MONDAY );
		calendar.setMinimalDaysInFirstWeek( 4 );
		calendar.setLenient( true );
		calendar.clear();
		return calendar;
	}
	public void testDailightSavingsBug() {
		final int monthly = RegularInterval.MONTHLY;
		Calendar juneRef, julyRef, calendar;
		juneRef = createRomanISOCalendar();
		julyRef = createRomanISOCalendar();
		juneRef.set( 1975, Calendar.JUNE, 1); /* that day starts at 01:00:00 AM*/
		julyRef.set( 1975, Calendar.JULY, 1); /* day starts as usual */ 
		calendar = createRomanISOCalendar();
		calendar.set( 1975, Calendar.MAY, 1 );
		AbstractRegularInterval.advanceCalendar(calendar, monthly, 1);
		assertEquals(
				"June is not june!", juneRef.getTime(), calendar.getTime()
		);
		AbstractRegularInterval.advanceCalendar(calendar, monthly, 1);
		assertEquals(
				"July is not july!", julyRef.getTime(), calendar.getTime()
		);
	}
	public void testWeek() {
		/* 19981224 is thursday, week52 21--27 Dec.
		 * 19981231 is thursday, week53 28 Dec.--3 Jan.
		 * 19990107 is thursday, week01 4--10 Jan.
		 * 20030603 is thursday, week23 2--8 Jun.
		 * */
		Calendar calendar = AbstractRegularInterval.createISOCalendar();
		calendar.setFirstDayOfWeek( Calendar.MONDAY );
		calendar.setMinimalDaysInFirstWeek( 4 );
		Calendar beginCalendar = (Calendar) calendar.clone();
		Calendar endCalendar = (Calendar) calendar.clone();
		beginCalendar.set( 1998, Calendar.DECEMBER, 21 );
		//endCalendar.set( 1999, Calendar.JANUARY, 4 );
		endCalendar.set( 2003, Calendar.JUNE, 9 );
		final int weekly = RegularInterval.WEEKLY;
		RegularInterval ri = new GregorianRegularInterval(
			weekly,
			beginCalendar.getTime(),
			endCalendar.getTime()
		);
		assertEquals( "Wrong interval length",
			ri.getSize(), 233 );
		
	}
}
