// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2004, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.time;

import java.util.Calendar;
import java.util.Date;
import java.util.GregorianCalendar;

import Int.ecb.gesmescb.util.PackageProperties;
import Int.ecb.logging.Level;
import Int.ecb.logging.Logger;
import Int.ecb.prefs.Preferences;

public class CalendarHelper {

    /* ======================= Constants ======================== */
    
    /** 
     * Whether to use a lenient calendar.
     * Use the "useLenient" option only if you know what you are doing.
     * If set, various invalid time formats will not be recognised.
     */
    private static final boolean USE_LENIENT;
    
    /* =================== Static Initializer =================== */

    static {
            final String packageName = PackageProperties.getName(AbstractRegularInterval.class.getName());
            final Logger logger = Logger.getLogger( packageName );
            final Preferences
                    userPrefs = Preferences.userRoot().node( packageName ),
                    systemPrefs = Preferences.systemRoot().node( packageName );
            final String useLenient = systemPrefs.get(
                    "useLenient",
                    userPrefs.get( "useLenient", "false" )
            );
            USE_LENIENT = "true".equalsIgnoreCase( useLenient );
            logger.log(
                    USE_LENIENT ? Level.CONFIG : Level.FINE,
                    "AbstractRegularInterval",
                    "static initialiser",
                    "Using lenient calendars: " + USE_LENIENT
            );
    }
    
    /* ====================== Constructors ====================== */
    
    private CalendarHelper() {
        super();
    }
    
    /* ===================== Public Methods ===================== */
    
    /**
     * <P>Create a <CODE>Calendar</CODE> with proper
     * settings.</P>
     *
     * <P>This means ISO weeks are respected.
     * The returned <CODE>Calendar</CODE> is not lenient by default.
     * Also it is {@link Calendar#clear}ed so that it
     * points to the epoch rather than the date of its creation.
     * The method <CODE>clear</CODE> seems to set to
     * the epoch of the {@link java.util.TimeZone}
     * that <CODE>Calendar</CODE> was created in.</P>
     *
     * @return a new, ISO conformant <CODE>Calendar</CODE>.
     */
    public static Calendar createISOCalendar () {
        final Calendar calendar = new GregorianCalendar();
        calendar.setFirstDayOfWeek( Calendar.MONDAY );
        calendar.setMinimalDaysInFirstWeek( 4 );
        calendar.setLenient( USE_LENIENT );
        calendar.clear();
        return calendar;
    }
    
    /** 
     * Work around a Java 1.5 misbehaviour.
     * Java 1.5 non-lenient calendars' getTime() detects
     * an illegal year when the first week starts outside the year.
     */
    public static void sanitiseCalendar ( final Calendar calendar ) {
        if ( ! calendar.isLenient() ) {
            calendar.setLenient( true );
            calendar.setTime( new Date( calendar.getTime().getTime() ) );
            calendar.setLenient( false );
        }
    }
    
    /** 
     * <P>Advance a <CODE>Calendar</CODE> a number of base periods
     * and ensure that the result is at the beginning of a base period.
     * The <CODE>Calendar</CODE>
     * must contain the original base period's begin point.</P>
     *
     * @param calendar to be set.
     * @param frequency determines the length of the base period.
     * @param number of baseperiods to add. This may be negative, too.
     */
    public static void advanceCalendar (
            final Calendar calendar,
            final int frequency,
            final int number
    ) {
        switch ( frequency ) {
            case RegularInterval.ANNUAL:
                calendar.add( Calendar.YEAR, number );
                // Daylight saving switch in Europe/Rome makes june 1st 1975 
                // start at 01:00:00AM instead of 00:00:00AM.
                // Adding a year to that Date yields a misaligned calendar.
                // dayProjectCalendar() avoids such issues. fh.
                dayProjectCalendar( calendar );
                break;
            case RegularInterval.SEMIANNUAL:
                calendar.add( Calendar.MONTH, 6 * number );
                dayProjectCalendar( calendar );
                break;
            case RegularInterval.QUARTERLY:
                calendar.add( Calendar.MONTH, 3 * number );
                dayProjectCalendar( calendar );
                break;
            case RegularInterval.MONTHLY:
                calendar.add( Calendar.MONTH, number );
                dayProjectCalendar( calendar );
                break;
            case RegularInterval.WEEKLY:
                calendar.add( Calendar.WEEK_OF_YEAR, number );
                dayProjectCalendar( calendar );
                break;
            case RegularInterval.DAILY:
                calendar.add( Calendar.DAY_OF_YEAR, number );
                dayProjectCalendar( calendar );
                break;
            case RegularInterval.MINUTELY:
                calendar.add( Calendar.MINUTE, number );
                break;
            default:
                throw new IllegalArgumentException( "Unknown frequency code" );
        }
    }
    
    /** 
     * <P>Set a <CODE>Calendar</CODE> to the beginning of a day.</P>
     *
     * <P>This operation acts idempotent on <CODE>Calendar</CODE>s
     * which means twice execution yields the same one time.
     * The operation therefore can be called a projection.</P>
     *
     * <P>This method is useful to clean the result from daylight saving switch
     * artefacts when using base daily or longer period.
     * Days get dayProjected on their beginning point, be it 01:00:00 AM (e.g.
     * switch to daylight saving), or 00:00:00 AM.</P> 
     *
     * @param calendar to be set.
     * @param frequency determines the base period grid.
     */
    public static void dayProjectCalendar (
        final Calendar calendar
    ) {
        calendar.set( Calendar.HOUR_OF_DAY, 0 );
        calendar.set( Calendar.MINUTE, 0 );
        calendar.set( Calendar.SECOND , 0 );
        calendar.set( Calendar.MILLISECOND, 0 );
        /* should the period day start at 1:00:00AM, sanitise takes care. */
        sanitiseCalendar( calendar );
    }
}
