// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.time;


import java.lang.NullPointerException;

import java.util.NoSuchElementException;
import java.util.Date;
import java.util.Calendar;
import java.util.GregorianCalendar;

import Int.ecb.gesmescb.time.RegularInterval;
import Int.ecb.gesmescb.time.AbstractRegularInterval;

/* ***************************** Class ****************************** */

/**
 * <P>Implements {@link RegularInterval} using a {@link GregorianCalendar}.</P>
 *
 * <P>Note that badly aligned interval begin points have not been tested.
 * They are not guaranteed to work. To avoid difficulties you might want to
 * use {@link #projectCalendar}.</P>
 *
 * @author
 * <A HREF="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</A>
 *
 * @version @(#)$Id: GregorianRegularInterval.java,v 1.1 2017/01/19 16:14:03 trainot Exp $
 */
public class GregorianRegularInterval extends AbstractRegularInterval {

	/* ========================= Fields ========================= */

	private final int frequency;
	private final Date begin;
	private final int size;

	/* ====================== Constructors ====================== */

	/** Construct one from begin date on size base period lengths.
	 *
	 * @throws NullPointerException
	 * if <CODE>begin</CODE> is <CODE>null</CODE>.
	 */
	public GregorianRegularInterval (
		final int frequency,
		final Date begin,
		final int size
	) {
		this.frequency = frequency;
		if ( ( this.begin = begin ) == null ) {
			throw new NullPointerException(
				"Trying to construct with null begin date."
			);
		}
		this.size = size;
	}

	/** <P>Construct one from begin and end date.</P>
	 *
	 * <P>Precondition: The <CODE>end</CODE> must be on the grid
	 * determined by the <CODE>frequency</CODE> and <CODE>begin</CODE>.</P>
	 *
	 * <P>The <CODE>end</CODE> date should be
	 * the first point in time after the interval
	 * such as it would be returned by {@link #getEnd()}.</P>
	 *
	 * @throws NullPointerException
	 * if <CODE>begin</CODE> or <CODE>end</CODE> is <CODE>null</CODE>.
	 */
	public GregorianRegularInterval (
		final int frequency,
		final Date begin,
		final Date end
	) throws NullPointerException {
		final Calendar calendar = createISOCalendar();
		calendar.setTime( begin );

		this.frequency = frequency;
		if ( ( this.begin = begin ) == null ) {
			throw new NullPointerException(
				"Trying to construct with null begin date."
			);
		}
		if ( end == null ) {
			throw new NullPointerException(
				"Trying to construct with null end date."
			);
		}
		this.size = computeIndex(
			frequency,
			calendar,
			end
		);
	}

	/** Construct one from frequency, year of first period,
	 * index of first period in that year, and size.
	 *
	 * @throws IllegalArgumentException if <CODE>firstPeriodIndex</CODE>
	 * is out of range.
	 */
	public GregorianRegularInterval (
		final int frequency,
		final int firstPeriodYear,
		final int firstPeriodIndex,
		final int size
	) throws IllegalArgumentException {
		final Calendar calendar = createISOCalendar();
		setCalendar( calendar, frequency, firstPeriodYear, firstPeriodIndex );

		this.frequency = frequency;
		this.begin = calendar.getTime();
		this.size = size;
	}

	/**
	 * Construct one from another <CODE>RegularInterval</CODE>.
	 * This is useful to implement bridges to other implementations
	 * of <CODE>RegularInterval</CODE>.
	 */
	public GregorianRegularInterval ( final RegularInterval interval ) {
		this(
			interval.getFrequency(), interval.getBegin(), interval.getSize()
		);
	}

	/* ===================== Public Methods ===================== */

	public int getFrequency () {
		return frequency;
	}

	public Date getBegin () {
		return begin;
	}

	public int getSize () {
		return size;
	}

	/**
	 * Tests whether this object equals that.
	 *
	 * @param that object which will be compared.
	 *
	 * @return whether <CODE>that</CODE> is a
	 * <CODE>GregorianRegularInterval</CODE>
	 * and has the same frequency, begin, and size.
	 */
	public boolean equals ( final Object that ) {
		if ( that.getClass().equals( GregorianRegularInterval.class ) ) {
			return super.equals( that );
		} else return false;
	}

	/** Creates a <code>GregorianRegularInterval</code> with minimum frequency
	 * for given begin and end point.
	 *
	 * @param requireAligned
	 * whether to require alignment with frequency base period.
	 *
	 * @return <code>null</code> if begin and end cannot be matched.
	 */
	public static RegularInterval searchCover (
		final Date begin, final Date end, final boolean requireAligned
	) {
		final int[] frequencies = new int[]{
			RegularInterval.ANNUAL,
			RegularInterval.SEMIANNUAL,
			RegularInterval.QUARTERLY,
			RegularInterval.MONTHLY,
			RegularInterval.WEEKLY,
			RegularInterval.DAILY,
			RegularInterval.MINUTELY
		};
		for ( int i = 0; i < frequencies.length; ++i ) {
			try {
				final RegularInterval interval = new GregorianRegularInterval(
					frequencies[i], begin, end
				);
				if (
					begin.equals( interval.getBegin() )
					&&
					end.equals( interval.getEnd() )
					&&
					( ! requireAligned || interval.isAligned() )
				) {
					return interval;
				}
			} catch ( final NoSuchElementException e ) {
				// try with different frequency.
			}
		}
		return null;
	}
}
