// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.time;

import java.util.Iterator;
import java.util.NoSuchElementException;
import java.util.Date;
import java.text.ParseException;

import junit.framework.Test;
import junit.framework.TestSuite;

import Int.ecb.gesmescb.time.GregorianRegularInterval;

/* ***************************** Class ****************************** */

/**
 * <P>Tests the implementation of the class
 * {@link GregorianRegularInterval}.</P>
 *
 * @author
 * <A HREF="mailto:Sami.Airo@ECB.int">Sami Airo</A>
 * @author
 * <A HREF="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</A>
 *
 * @version @(#)$Id: GregorianRegularIntervalTest.java,v 1.1 2017/01/19 16:14:03 trainot Exp $
 */

public class GregorianRegularIntervalTest extends AbstractRegularIntervalTest {

	/* ====================== Constructors ====================== */

	/**
	 * Constructs a GregorianRegularIntervalTest with the specified name.
	 *
	 * @param name Test case name
	 */
	public GregorianRegularIntervalTest ( final String name ) {
		super( name );
	}

	/* ===================== Public Methods ===================== */

	/**
	 * Assembles and returns a test suite for all
	 * the test methods of this test case.
	 *
	 * @return a non-null test suite
	 */
	public static Test suite () {
		return new TestSuite( GregorianRegularIntervalTest.class );
	}

	/**
	 * Runs the test case
	 */
	public static void main ( final String args[] ) {
		final String[] testCaseName = {
			GregorianRegularIntervalTest.class.getName()
		};
		junit.textui.TestRunner.main( testCaseName );
	}

	/* --------------------- Test Methods --------------------- */

	/**
	 * Tests the constructors.
	 */
	public void testConstructors () throws ParseException {
		{
			final RegularInterval
				interval0 = new GregorianRegularInterval(
					RegularInterval.MONTHLY,
					FORMAT.parse( "20010101000000" ),
					2
				),
				interval1 = new GregorianRegularInterval(
					RegularInterval.MONTHLY,
					FORMAT.parse( "20010101000000" ),
					FORMAT.parse( "20010301000000" )
				),
				interval2 = new GregorianRegularInterval(
					RegularInterval.MONTHLY,
					2001,
					0,
					2
				),
				interval3 = new GregorianRegularInterval(
					interval1
				);
			assertEquals( "M0=1", interval0, interval1 );
			assertEquals( "M0=2", interval0, interval2 );
			assertEquals( "M0=3", interval0, interval3 );
			assertEquals( "M1=2", interval1, interval2 );
			assertEquals( "M1=3", interval1, interval3 );
			assertEquals( "M2=3", interval1, interval3 );
		}
		// Same with WEEKLY.
		{
			final RegularInterval
				interval1 = new GregorianRegularInterval(
					RegularInterval.WEEKLY,
					FORMAT.parse( "20011231000000" ),
					6
				),
				interval2 = new GregorianRegularInterval(
					RegularInterval.WEEKLY,
					2002,
					0,
					6
				),
				interval3 = new GregorianRegularInterval(
					interval1
				);
			assertEquals( "W1=2", interval1, interval2 );
			assertEquals( "W1=3", interval1, interval3 );
			assertEquals( "W2=3", interval1, interval3 );
		}
	}

	/** Test equals, hashCode, and toString. */
	public void testBasic () {
		final RegularInterval[] intervals = new RegularInterval[]{
			new GregorianRegularInterval(
				RegularInterval.SEMIANNUAL, 2004, 1, 3
			),
			new GregorianRegularInterval(
				RegularInterval.SEMIANNUAL, 2004, 1, 3
			),
			new GregorianRegularInterval(
				RegularInterval.QUARTERLY, 2004, 1, 3
			),
			new GregorianRegularInterval(
				RegularInterval.SEMIANNUAL, 2005, 1, 3
			),
			new GregorianRegularInterval(
				RegularInterval.SEMIANNUAL, 2004, 0, 3
			),
			new GregorianRegularInterval(
				RegularInterval.SEMIANNUAL, 2004, 1, 5
			)
		};
		assertEquals( "0==1", intervals[0], intervals[1] );
		assertEquals(
			"0.hash==1.hash",
			intervals[0].hashCode(),
			intervals[1].hashCode()
		);
		assertEquals(
			"0.string==1.string",
			intervals[0].toString(),
			intervals[1].toString()
		);
		for ( int i = 2; i < intervals.length; ++i ) {
			assertTrue( "0!=" + i, ! intervals[0].equals( intervals[i] ) );
			assertTrue(
				"0.hash!=" + i + ".hash",
				intervals[0].hashCode() != intervals[i].hashCode()
			);
			assertTrue(
				"0.string!=" + i + ".string",
				! intervals[0].toString().equals( intervals[i].toString() )
			);
		}
	}

	/**
	 * Tests the genuine fields.
	 */
	public void testFields () throws ParseException {
		final Date begin = FORMAT.parse( "20010101000000" );
		final RegularInterval interval = new GregorianRegularInterval(
			RegularInterval.MONTHLY, begin, 2
		);
		assertSame( "begin", begin, interval.getBegin() );
		assertEquals(
			"frequency", RegularInterval.MONTHLY, interval.getFrequency()
		);
		assertEquals( "size", 2, interval.getSize() );
	}

	/**
	 * Tests
	 * {GregorianRegularCalendar#getEnd},
	 * {GregorianRegularCalendar#getFirstPeriodYear},
	 * {GregorianRegularCalendar#getFirstPeriodIndex}.
	 */
	public void testSimpleOperations () throws ParseException {
		{
			final Date begin = FORMAT.parse( "20010501000000" );
			final Date end = FORMAT.parse( "20010801000000" );
			final RegularInterval interval = new GregorianRegularInterval(
				RegularInterval.MONTHLY, begin, 3
			);
			assertEquals( "monthly end", end, interval.getEnd() );
			assertEquals(
				"monthly firstperiodyear",
				2001,
				interval.getFirstPeriodYear()
			);
			assertEquals(
				"monthly firstperiodindex",
				4,	// 5th month
				interval.getFirstPeriodIndex()
			);
		}
		{
			final Date begin = FORMAT.parse( "20010528000000" );
			final Date end = FORMAT.parse( "20010618000000" );
			final RegularInterval interval = new GregorianRegularInterval(
				RegularInterval.WEEKLY, begin, 3
			);
			assertEquals( "weekly end", end, interval.getEnd() );
			assertEquals(
				"weekly firstperiodyear",
				2001,
				interval.getFirstPeriodYear()
			);
			assertEquals(
				"weekly firstperiodindex",
				21,	// 22nd week
				interval.getFirstPeriodIndex()
			);
		}
	}

	/**
	 * Tests {GregorianRegularCalendar#computeBasePeriod}.
	 */
	public void testComputeBasePeriod () throws ParseException {
		final RegularInterval interval = new GregorianRegularInterval(
			RegularInterval.WEEKLY,
			FORMAT.parse( "20010101000000" ),
			10
		);
		//final RegularInterval
		//	intervalMinus10 = 
			new GregorianRegularInterval(
				RegularInterval.WEEKLY,
				FORMAT.parse( "20001023000000" ),
				1
			);
			final RegularInterval intervalMinus1
			= new GregorianRegularInterval(
				RegularInterval.WEEKLY,
				FORMAT.parse( "20001225000000" ),
				1
			),
			interval0 = new GregorianRegularInterval(
				RegularInterval.WEEKLY,
				FORMAT.parse( "20010101000000" ),
				1
			),
			interval1 = new GregorianRegularInterval(
				RegularInterval.WEEKLY,
				FORMAT.parse( "20010108000000" ),
				1
			),
			interval52 = new GregorianRegularInterval(
				RegularInterval.WEEKLY,
				FORMAT.parse( "20011231000000" ),
				1
			);

		assertEquals(
			"-1",
			intervalMinus1,
			interval.computeBasePeriod( -1 )
		);
		assertEquals(
			"0",
			interval0,
			interval.computeBasePeriod( 0 )
		);
		assertEquals(
			"1",
			interval1,
			interval.computeBasePeriod( 1 )
		);
		assertEquals(
			"52",
			interval52,
			interval.computeBasePeriod( 52 )
		);
	}

	/**
	 * Tests
	 * {GregorianRegularCalendar#computeBasePeriod},
	 * {GregorianRegularCalendar#computeIndex}.
	 */
	public void testBPOperations () throws ParseException {
		final RegularInterval interval = new GregorianRegularInterval(
			RegularInterval.MONTHLY,
			FORMAT.parse( "20010101000000" ),
			10
		);
		final RegularInterval other = new GregorianRegularInterval(
			RegularInterval.MONTHLY,
			FORMAT.parse( "20010501000000" ),
			1
		);
		assertEquals(
			"computeIndex",
			4,
			interval.computeIndex( other )
		);
	}

	/**
	 * Tests
	 * {GregorianRegularCalendar#iterator},
	 * {GregorianRegularCalendar#computeBasePeriod},
	 * {GregorianRegularCalendar#computeIndex},
	 */
	public void testComplexOperations () throws ParseException {
		{
			final Date begin = FORMAT.parse( "20010501000000" );
			final RegularInterval interval = new GregorianRegularInterval(
				RegularInterval.MONTHLY, begin, 3
			);
			final RegularInterval[] basePeriods = {
				new GregorianRegularInterval(
					RegularInterval.MONTHLY,
					FORMAT.parse( "20010501000000" ),
					1
				),
				new GregorianRegularInterval(
					RegularInterval.MONTHLY,
					FORMAT.parse( "20010601000000" ),
					1
				),
				new GregorianRegularInterval(
					RegularInterval.MONTHLY,
					FORMAT.parse( "20010701000000" ),
					1
				)
			};
			final Iterator iterator = interval.iterator();
			for ( int i = 0; i < basePeriods.length; ++i ) {
				assertEquals( "iterator", basePeriods[i], iterator.next() );
				try {
					iterator.remove();
					fail( "Iterator allowed remove()." );
				}
				catch ( UnsupportedOperationException e ) {
				}
				assertEquals(
					"computeBasePeriod",
					basePeriods[i],
					interval.computeBasePeriod( i )
				);
				assertEquals(
					"computeIndex",
					i,
					interval.computeIndex( basePeriods[i] )
				);
			}

			assertTrue( "hasNext false", !iterator.hasNext() );
			try {
				iterator.next();
				fail( "Empty iterator allowed next()." );
			}
			catch ( NoSuchElementException e ) {
			}
		}
		{
			final Date begin = FORMAT.parse( "19470401000000" );
			final int size = 59*12 - 1;	// 194704200602:710
			final RegularInterval interval = new GregorianRegularInterval(
				RegularInterval.MONTHLY, begin, size
			);
			final Iterator iterator = interval.iterator();
			for ( int i = 0; i < size; ++i ) {
				final RegularInterval basePeriod = (RegularInterval) iterator.next();
				try {
					iterator.remove();
					fail( "Iterator allowed remove()." );
				}
				catch ( UnsupportedOperationException e ) {
				}
				assertEquals(
					"computeBasePeriod",
					basePeriod,
					interval.computeBasePeriod( i )
				);
				assertEquals(
					"computeIndex",
					i,
					interval.computeIndex( basePeriod )
				);
			}

			assertTrue( "hasNext false", !iterator.hasNext() );
			try {
				iterator.next();
				fail( "Empty iterator allowed next()." );
			}
			catch ( NoSuchElementException e ) {
			}
		}
		{
			final Date begin = FORMAT.parse( "20031222000000" );
			final RegularInterval interval = new GregorianRegularInterval(
				RegularInterval.WEEKLY, begin, 4
			);
			final RegularInterval[] basePeriods = {
				new GregorianRegularInterval(
					RegularInterval.WEEKLY,
					FORMAT.parse( "20031222000000" ),
					1
				),
				new GregorianRegularInterval(
					RegularInterval.WEEKLY,
					FORMAT.parse( "20031229000000" ),
					1
				),
				new GregorianRegularInterval(
					RegularInterval.WEEKLY,
					FORMAT.parse( "20040105000000" ),
					1
				),
				new GregorianRegularInterval(
					RegularInterval.WEEKLY,
					FORMAT.parse( "20040112000000" ),
					1
				)
			};
			final Iterator iterator = interval.iterator();
			for ( int i = 0; i < basePeriods.length; ++i ) {
				assertEquals( "iterator", basePeriods[i], iterator.next() );
				try {
					iterator.remove();
					fail( "Iterator allowed remove()." );
				}
				catch ( UnsupportedOperationException e ) {
				}
				assertEquals(
					"compute",
					basePeriods[i],
					interval.computeBasePeriod( i )
				);
				assertEquals(
					"computeIndex",
					i,
					interval.computeIndex( basePeriods[i] )
				);
			}

			assertTrue( "hasNext false", !iterator.hasNext() );
			try {
				iterator.next();
				fail( "Empty iterator allowed next()." );
			}
			catch ( NoSuchElementException e ) {
			}
		}
	}

	/** Additional tests for {@link RegularInterval#WEEKLY} frequency.
	 * Good since this involves the most complicated computation and
	 * also is most probably used by other classes
	 * based on {@link AbstractRegularInterval}. */
	public void testWeekly() throws ParseException {
		RegularInterval interval;

		interval = new GregorianRegularInterval(
			GregorianRegularInterval.WEEKLY,
			FORMAT.parse( "20031229000000" ),
			3
		);
		assertEquals(
			"(following) year", 2004, interval.getFirstPeriodYear()
		);
		assertEquals(
			"week of (following) year", 0, interval.getFirstPeriodIndex()
		);

		interval = new GregorianRegularInterval(
			GregorianRegularInterval.WEEKLY,
			FORMAT.parse( "20041227000000" ),
			0
		);
		assertEquals(
			"year", 2004, interval.getFirstPeriodYear()
		);
		assertEquals(
			"week of year", 52, interval.getFirstPeriodIndex()
		);

		interval = interval.computeBasePeriod( 1 );	// next
		assertEquals(
			"following year", 2005, interval.getFirstPeriodYear()
		);
		assertEquals(
			"following week", 0, interval.getFirstPeriodIndex()
		);

		interval = new GregorianRegularInterval(
			GregorianRegularInterval.WEEKLY,
			FORMAT.parse( "19471229000000" ),
			FORMAT.parse( "19490103000000" )
		);
		assertEquals(
			new GregorianRegularInterval(
				GregorianRegularInterval.WEEKLY,
				1948, 0, 53
			),
			interval
		);
		assertEquals(
			"year", 1948, interval.getFirstPeriodYear()
		);
		assertEquals(
			"week of year", 0, interval.getFirstPeriodIndex()
		);
		interval = interval.computeBasePeriod( 52 );	// last
		assertEquals(
			"year", 1948, interval.getFirstPeriodYear()
		);
		assertEquals(
			"week of year", 52, interval.getFirstPeriodIndex()
		);

		interval = new GregorianRegularInterval(
			GregorianRegularInterval.WEEKLY,
			FORMAT.parse( "19480105000000" ),
			FORMAT.parse( "19481227000000" )
		);
		assertEquals(
			new GregorianRegularInterval(
				GregorianRegularInterval.WEEKLY,
				1948, 1, 51
			),
			interval
		);
		assertEquals(
			"year", 1948, interval.getFirstPeriodYear()
		);
		assertEquals(
			"week of year", 1, interval.getFirstPeriodIndex()
		);
		interval = interval.computeBasePeriod( 50 );	// last
		assertEquals(
			"year", 1948, interval.getFirstPeriodYear()
		);
		assertEquals(
			"week of year", 51, interval.getFirstPeriodIndex()
		);

		interval = new GregorianRegularInterval(
			GregorianRegularInterval.WEEKLY,
			FORMAT.parse( "19471222000000" ),
			FORMAT.parse( "19490110000000" )
		);
		assertEquals(
			new GregorianRegularInterval(
				GregorianRegularInterval.WEEKLY,
				1947, 51, 55
			),
			interval
		);
		assertEquals(
			"year", 1947, interval.getFirstPeriodYear()
		);
		assertEquals(
			"week of year", 51, interval.getFirstPeriodIndex()
		);
		interval = interval.computeBasePeriod( 54 );	// last
		assertEquals(
			"year", 1949, interval.getFirstPeriodYear()
		);
		assertEquals(
			"week of year", 0, interval.getFirstPeriodIndex()
		);

		interval = new GregorianRegularInterval(
			GregorianRegularInterval.WEEKLY,
			FORMAT.parse( "19990104000000" ),
			FORMAT.parse( "20000103000000" )
		);
		assertEquals( "1999", 52, interval.getSize() );
	}

	public void testIsAligned () throws ParseException {
		assertTrue(
			"Annual aligned",
			new GregorianRegularInterval(
				GregorianRegularInterval.ANNUAL,
				FORMAT.parse( "20540101000000" ),
				3
			).isAligned()
		);
		assertTrue(
			"Annual misaligned",
			! new GregorianRegularInterval(
				GregorianRegularInterval.ANNUAL,
				FORMAT.parse( "20970701000000" ),
				3
			).isAligned()
		);
		assertTrue(
			"Seminnual aligned",
			new GregorianRegularInterval(
				GregorianRegularInterval.SEMIANNUAL,
				FORMAT.parse( "25300701000000" ),
				392
			).isAligned()
		);
		assertTrue(
			"Seminnual misaligned",
			! new GregorianRegularInterval(
				GregorianRegularInterval.SEMIANNUAL,
				FORMAT.parse( "38711001000000" ),
				1
			).isAligned()
		);
		assertTrue(
			"Quarterly aligned",
			new GregorianRegularInterval(
				GregorianRegularInterval.QUARTERLY,
				FORMAT.parse( "27250401000000" ),
				100
			).isAligned()
		);
		assertTrue(
			"Quarterly misaligned",
			! new GregorianRegularInterval(
				GregorianRegularInterval.QUARTERLY,
				FORMAT.parse( "27250901000000" ),
				82
			).isAligned()
		);
		assertTrue(
			"Monthly aligned",
			new GregorianRegularInterval(
				GregorianRegularInterval.MONTHLY,
				FORMAT.parse( "20241101000000" ),
				34
			).isAligned()
		);
		assertTrue(
			"Monthly misaligned",
			! new GregorianRegularInterval(
				GregorianRegularInterval.MONTHLY,
				FORMAT.parse( "20010108000000" ),
				43
			).isAligned()
		);
		assertTrue(
			"Weekly aligned",
			new GregorianRegularInterval(
				GregorianRegularInterval.WEEKLY,
				FORMAT.parse( "20010115000000" ),
				30
			).isAligned()
		);
		assertTrue(
			"Weekly misaligned",
			! new GregorianRegularInterval(
				GregorianRegularInterval.WEEKLY,
				FORMAT.parse( "20010111000000" ),
				729
			).isAligned()
		);
		assertTrue(
			"Daily aligned",
			new GregorianRegularInterval(
				GregorianRegularInterval.DAILY,
				FORMAT.parse( "26351231000000" ),
				3
			).isAligned()
		);
		assertTrue(
			"Daily misaligned",
			! new GregorianRegularInterval(
				GregorianRegularInterval.DAILY,
				FORMAT.parse( "26351231010000" ),
				3
			).isAligned()
		);
		assertTrue(
			"Minutely aligned",
			new GregorianRegularInterval(
				GregorianRegularInterval.MINUTELY,
				FORMAT.parse( "36251130235600" ),
				3
			).isAligned()
		);
		assertTrue(
			"Minutely misaligned",
			! new GregorianRegularInterval(
				GregorianRegularInterval.MINUTELY,
				FORMAT.parse( "36251130235614" ),
				3
			).isAligned()
		);
	}

	public void testCover () throws ParseException {
		assertEquals(
			"Annual",
			new GregorianRegularInterval(
				GregorianRegularInterval.ANNUAL,
				FORMAT.parse( "20030101000000" ),
				3
			),
			GregorianRegularInterval.searchCover(
				FORMAT.parse( "20030101000000" ),
				FORMAT.parse( "20060101000000" ),
				true
			)
		);
		assertEquals(
			"Annual misaligned",
			new GregorianRegularInterval(
				GregorianRegularInterval.ANNUAL,
				FORMAT.parse( "20030401000000" ),
				3
			),
			GregorianRegularInterval.searchCover(
				FORMAT.parse( "20030401000000" ),
				FORMAT.parse( "20060401000000" ),
				false
			)
		);
		assertEquals(
			"Semiannual 1",
			new GregorianRegularInterval(
				GregorianRegularInterval.SEMIANNUAL,
				FORMAT.parse( "20000101000000" ),
				3
			),
			GregorianRegularInterval.searchCover(
				FORMAT.parse( "20000101000000" ),
				FORMAT.parse( "20010701000000" ),
				true
			)
		);
		assertEquals(
			"Semiannual 2",
			new GregorianRegularInterval(
				GregorianRegularInterval.SEMIANNUAL,
				FORMAT.parse( "20010701000000" ),
				5
			),
			GregorianRegularInterval.searchCover(
				FORMAT.parse( "20010701000000" ),
				FORMAT.parse( "20040101000000" ),
				true
			)
		);
		assertEquals(
			"Semiannual misaligned",
			new GregorianRegularInterval(
				GregorianRegularInterval.SEMIANNUAL,
				FORMAT.parse( "20011001000000" ),
				5
			),
			GregorianRegularInterval.searchCover(
				FORMAT.parse( "20011001000000" ),
				FORMAT.parse( "20040401000000" ),
				false
			)
		);
		assertEquals(
			"Quarterly",
			new GregorianRegularInterval(
				GregorianRegularInterval.QUARTERLY,
				FORMAT.parse( "20500401000000" ),
				4
			),
			GregorianRegularInterval.searchCover(
				FORMAT.parse( "20500401000000" ),
				FORMAT.parse( "20510401000000" ),
				true
			)
		);
		assertEquals(
			"Quarterly misaligned",
			new GregorianRegularInterval(
				GregorianRegularInterval.QUARTERLY,
				FORMAT.parse( "20500201000000" ),
				5
			),
			GregorianRegularInterval.searchCover(
				FORMAT.parse( "20500201000000" ),
				FORMAT.parse( "20510501000000" ),
				false
			)
		);
		assertEquals(
			"Monthly",
			new GregorianRegularInterval(
				GregorianRegularInterval.MONTHLY,
				FORMAT.parse( "184502010000000" ),
				7
			),
			GregorianRegularInterval.searchCover(
				FORMAT.parse( "184502010000000" ),
				FORMAT.parse( "184509010000000" ),
				true
			)
		);
		assertEquals(
			"Monthly misaligned",
			new GregorianRegularInterval(
				GregorianRegularInterval.MONTHLY,
				FORMAT.parse( "284502030000000" ),
				7
			),
			GregorianRegularInterval.searchCover(
				FORMAT.parse( "284502030000000" ),
				FORMAT.parse( "284509030000000" ),
				false
			)
		);
		assertEquals(
			"Weekly",
			new GregorianRegularInterval(
				GregorianRegularInterval.WEEKLY,
				FORMAT.parse( "20010101000000" ),
				4
			),
			GregorianRegularInterval.searchCover(
				FORMAT.parse( "20010101000000" ),
				FORMAT.parse( "20010129000000" ),
				true
			)
		);
		assertEquals(
			"Weekly misaligned",
			new GregorianRegularInterval(
				GregorianRegularInterval.WEEKLY,
				FORMAT.parse( "20010103000000" ),
				1
			),
			GregorianRegularInterval.searchCover(
				FORMAT.parse( "20010103000000" ),
				FORMAT.parse( "20010110000000" ),
				false
			)
		);
		assertEquals(
			"Daily",
			new GregorianRegularInterval(
				GregorianRegularInterval.DAILY,
				FORMAT.parse( "17331231000000" ),
				1
			),
			GregorianRegularInterval.searchCover(
				FORMAT.parse( "17331231000000" ),
				FORMAT.parse( "17340101000000" ),
				true
			)
		);
		// Daily misaligned stays at minute level...
		// TODO: Debug!
		//assertEquals(
		//	"Daily misaligned",
		//	new GregorianRegularInterval(
		//		GregorianRegularInterval.DAILY,
		//		FORMAT.parse( "20331031030000" ),
		//		1
		//	),
		//	GregorianRegularInterval.searchCover(
		//		FORMAT.parse( "20331031030000" ),
		//		FORMAT.parse( "20331101030000" ),
		//		false
		//	)
		//);
		assertEquals(
			"Minutely",
			new GregorianRegularInterval(
				GregorianRegularInterval.MINUTELY,
				FORMAT.parse( "19711214235900" ),
				53
			),
			GregorianRegularInterval.searchCover(
				FORMAT.parse( "19711214235900" ),
				FORMAT.parse( "19711215005200" ),
				true
			)
		);
		assertEquals(
			"Minutely misaligned",
			new GregorianRegularInterval(
				GregorianRegularInterval.MINUTELY,
				FORMAT.parse( "19711214235913" ),
				53
			),
			GregorianRegularInterval.searchCover(
				FORMAT.parse( "19711214235913" ),
				FORMAT.parse( "19711215005213" ),
				false
			)
		);
	}
}
