// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.time;

import java.lang.Comparable;
import java.lang.ClassCastException;

import java.util.Date;

/* *************************** Interface **************************** */

/**
 * <P>Represents a time interval.</P>
 *
 * <P>Intervals are half closed.
 * This means, the time points in the interval are all points which
 * are at or after the begin and before the end of the interval.
 * For this the idiom is used:
 * "From the begin point to before the end point".
 * Note that this implies that the interval is empty if
 * begin and end points coincide.</P>
 *
 * <P><IMG SRC=doc-files/interval.png ALT="Half open interval semantics"></P>
 *
 * <P>In most uses of the framework, the begin point precedes the end point.
 * Such <CODE>Interval</CODE>s are called positive.
 * An <CODE>Interval</CODE> is called negative,
 * when its end point precedes its begin point.
 * Such <CODE>Interval</CODE>s can be illustrated using red ink
 * with the understanding that black and red ink mutually neutralize
 * each other.</P>
 *
 * <P><IMG SRC=doc-files/negatint.png ALT="Negative interval"></P>
 *
 * <P>The same semantics can also be visualized using a function
 * on all points of time which jumps up by one at the begin point, inclusive,
 * and down by one on the end point, inclusive.
 * If the points coincide, these jumps neutralize each other.</P>
 *
 * <P><IMG SRC=doc-files/jumps.png ALT="Step function semantics"></P>
 *
 * <P>The notion of negative <CODE>Interval</CODE>s has been introduced
 * to keep the framework as general as possible.
 * However, care has been taken to give implementations a certain
 * amount of freedom not to support negative intervals.
 * That is, the contract of <CODE>Interval</CODE>
 * does not imply that negative <CODE>Interval</CODE>s
 * could be generated from nonnegative ones by any of
 * <CODE>Interval</CODE>'s methods.
 * Nevertheless, methods which take an <CODE>Interval</CODE> as argument
 * must treat it according to the method's contract even if it is
 * negative.</P>
 *
 * @author
 * <A HREF="mailto:Sami.Airo@ECB.int">Sami Airo</A>
 * @author
 * <A HREF="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</A>
 *
 * @version @(#)$Id: Interval.java,v 1.1 2017/01/19 16:14:03 trainot Exp $
 */
public interface Interval extends Comparable {

	/** Get the <CODE>Interval</CODE>'s begin point.
	 * @return the begin time. */
	Date getBegin ();

	/** Get the <CODE>Interval</CODE>'s end point.
	 * @return the end time. */
	Date getEnd ();

	/**
	 * <P>Decide whether a point in time is inside.</P>
	 *
	 * <P>To be inside an <CODE>Interval</CODE> means to be
	 * later than or equal to the begin point and earlier than the
	 * end point.</P>
	 *
	 * <P>By definition, empty and negative <CODE>Interval</CODE>s
	 * do not contain any point.</P>
	 *
	 * @param date to check.
	 *
	 * @return whether <CODE>date</CODE> is in the interval.
	 */
	boolean contains ( Date date );

	/**
	 * <P>Lexicographically compare two <CODE>Interval</CODE>s.</P>
	 *
	 * <P>The ordering should be lexicographic which means:
	 * Let <I>i</I> and <I>j</I> be <CODE>Interval</CODE>s.
	 * Then <I>i</I><TT>.compareTo(</TT><I>j</I><TT>)</TT> should return
	 * <TABLE>
	 * <TR><TH>negative</TH>
	 * <TD>if <I>i</I> begins before <I>j</I>,
	 * or <I>i</I> begins at the same time as <I>j</I>
	 * but ends before <I>j</I> does so.</TD></TR>
	 * <TR><TH>zero</TH>
	 * <TD>if <I>i</I> begins and ends at the same times as <I>j</I>
	 * does.</TD></TR>
	 * <TR><TH>positive</TH>
	 * <TD>if <I>i</I> begins after <I>j</I>,
	 * or <I>i</I> begins at the same time as <I>j</I>
	 * but ends after <I>j</I> does so.</TD></TR>
	 * </TABLE></P>
	 *
	 * @param o the <CODE>Interval</CODE> to compare to.
	 *
	 * @throws ClassCastException if <CODE>o</CODE> is
	 * not an <CODE>Interval</CODE>.
	 */
	int compareTo ( Object o ) throws ClassCastException;
}
