// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.time;

import java.util.Iterator;

import Int.ecb.gesmescb.time.Interval;

/* *************************** Interface **************************** */

/**
 * <P>Represents a time interval which is well aligned to a time grid
 * according to some frequency. This means the begin and end points lie
 * on this grid. E.g. on month begin points for monthly frequency.</P>
 *
 * <P>The grid cells are the base periods. These are determined by
 * the frequency. For instance, frequency can be daily, then the base
 * periods are days. For frequencies like monthly, base periods are
 * of varying length measured in absolute time.</P>
 *
 * <P><IMG SRC=doc-files/regular.png ALT="Decomposition into base periods"></P>
 *
 * <P>Like usual for {@link Interval}s, a base period always ranges
 * from its begin to before its end. Therefore e.g. the monthly
 * <CODE>RegularInterval</CODE> containing April 2000 to November 2001 begins
 * on 1 April 2000, 0:00:00.000 and extends to before 1 December 2001,
 * 0:00:00.000.</P>
 *
 * <P>A <CODE>RegularInterval</CODE> can be decomposed into its base periods.
 * These then are its subintervals of size one base period each.</P>
 *
 * @author
 * <A HREF="mailto:Sami.Airo@ECB.int">Sami Airo</A>
 * @author
 * <A HREF="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</A>
 *
 * @version @(#)$Id: RegularInterval.java,v 1.1 2017/01/19 16:14:03 trainot Exp $
 */
public interface RegularInterval extends Interval {

	/* ======================= Constants ======================== */

	/**
	 * Minutely frequency.
	 *
	 * @see #getFrequency
	 */
	public static final int MINUTELY = 203;

	/**
	 * Daily frequency.
	 *
	 * @see #getFrequency
	 */
	public static final int DAILY = 102;

	/**
	 * Weekly frequency.
	 *
	 * @see #getFrequency
	 */
	public static final int WEEKLY = 616;

	/**
	 * Monthly frequency.
	 *
	 * @see #getFrequency
	 */
	public static final int MONTHLY = 610;

	/**
	 * Quarterly frequency.
	 *
	 * @see #getFrequency
	 */
	public static final int QUARTERLY = 608;

	/**
	 * Semi-annual frequency.
	 *
	 * @see #getFrequency
	 */
	public static final int SEMIANNUAL = 604;

	/**
	 * Annual frequency.
	 *
	 * @see #getFrequency
	 */
	public static final int ANNUAL = 602;

	/* ========================= Fields ========================= */

	/**
	 * <P>Get the year which the first subinterval lies in.</P>
	 *
	 * <P>For {@link #WEEKLY} frequency the ISO 8601 standard
	 * says that a week begins on Monday and ends with Sunday.
	 * The first week in a year is the first week which
	 * has its Thursday in that year i.e. which has the majority
	 * of its days there.
	 * In consequence, a weekly <CODE>RegularInterval</CODE> may have its
	 * begin point up to three days before the year that
	 * <CODE>getFirstPeriodYear</CODE> returns.</P>
	 *
	 * @see #getFirstPeriodIndex
	 */
	int getFirstPeriodYear ();

	/**
	 * <P>Get the index of the period which the first subinterval
	 * lies in with respect to the year it lies in.</P>
	 *
	 * <P>Period indexing begins at 0 and ends one before the last period
	 * in the year. E.g. for a {@link #MONTHLY} <CODE>RegularInterval</CODE>
	 * beginning with January,
	 * <CODE>getFirstPeriodIndex</CODE> would return 0.</P>
	 *
	 * @see #getFirstPeriodYear
	 */
	int getFirstPeriodIndex ();

	/**
	 * <P>Get the base frequency.
	 * This determines the base period lengths.</P>
	 *
	 * <P>The value must be one of
	 * {@link #MINUTELY},
	 * {@link #DAILY},
	 * {@link #WEEKLY},
	 * {@link #MONTHLY},
	 * {@link #QUARTERLY},
	 * {@link #SEMIANNUAL},
	 * or {@link #ANNUAL}.</P>
	 */
	int getFrequency ();

	/**
	 * <P>Get the size measured in base periods.</P>
	 *
	 * <P>E.g. a {@link #MONTHLY} <CODE>RegularInterval</CODE> consisting of
	 * November 1998 and December 1998 would have size 2.</P>
	 *
	 * <P>The size of base periods is always equal to 1.</P>
	 *
	 * <P>For negative <CODE>RegularInterval</CODE>s,
	 * the size is negative.</P>
	 *
	 * @return the number of contained base periods.
	 */
	int getSize ();

	/**
	 * <P>Get a base period aligned according to the begin time.</P>
	 *
	 * <P>The <CODE>index</CODE>es
	 * in the range from 0 to {@link #getSize}-1
	 * address all contained base periods.
	 * For other <CODE>index</CODE>es, base periods
	 * outside the <CODE>RegularInterval</CODE> are returned.</P>
	 *
	 * <P>The <CODE>index</CODE> 0 corresponds to the base period
	 * which has the same begin time as the <CODE>RegularInterval</CODE>
	 * itself.
	 * The <CODE>index</CODE> {@link #getSize}-1 corresponds to the base period
	 * which has the same end time as the <CODE>RegularInterval</CODE>
	 * itself. The other base periods are indexed in the according
	 * linear order.</P>
	 *
	 * <P><IMG SRC=doc-files/baseperiods.png ALT="Indexing base periods"></P>
	 *
	 * <P>Note that the end time of the <CODE>RegularInterval</CODE>
	 * does not influence the base periods returned by this function.
	 * It is only meaningful when interpreting where they lie
	 * with respect to the <CODE>RegularInterval</CODE>.
	 * Particularly, no negative <CODE>RegularInterval</CODE>s
	 * are returned, but always base periods which by definition
	 * are positive.</P>
	 *
	 * <P>This implies that negative intervals are not treated
	 * specially.</P>
	 *
	 * <P><IMG SRC=doc-files/negativebp.png ALT="Indexing base periods in the negative case"></P>
	 *
	 * @param index of the desired base period.
	 *
	 * @return the desired base period.
	 * This has always {@link #getSize} equal to 1.
	 *
	 * @see #computeIndex
	 */
	RegularInterval computeBasePeriod ( int index );

	/**
	 * <P>Get the index of a base period.</P>
	 *
	 * <P>Precondition: The parameter <CODE>interval</CODE>
	 * has the same frequency and its size is 1.</P>
	 *
	 * <P>The <CODE>index</CODE>es for base periods within the
	 * <CODE>RegularInterval</CODE> are
	 * in the range from 0 to {@link #getSize}-1.
	 * Earlier base periods have negative <CODE>index</CODE>es
	 * and later ones have <CODE>index</CODE>es from {@link #getSize}
	 * onwards.</P>
	 *
	 * <P>Note that the end time of the <CODE>RegularInterval</CODE>
	 * does not influence the base periods returned by this function.</P>
	 *
	 * <P><IMG SRC=doc-files/negativebp.png ALT="Indexing base periods in the negative case"></P>
	 *
	 * @param interval to be indexed.
	 *
	 * @return the desired base period.
	 * This has always {@link #getSize} equal to 1.
	 *
	 * @see #computeBasePeriod
	 */
	int computeIndex ( RegularInterval interval );

	/**
	 * <P>Iterate through all contained base periods.</P>
	 *
	 * <P>The returned <CODE>Iterator</CODE> iterates through
	 * the sequence of base periods from the begin to the end.</P>
	 *
	 * <P>For positive <CODE>RegularInterval</CODE>s, iteration
	 * proceeds in the same ordering as {@link #computeBasePeriod}
	 * indexes the base periods.
	 * The first being the base period with <CODE>index</CODE> equal 0,
	 * the last being the base period with <CODE>index</CODE> equal
	 * {@link #getSize}-1.</P>
	 *
	 * <P>For nonpositive, i.e. empty or negative
	 * <CODE>RegularInterval</CODE>s, no base periods are returned.
	 * This treatment of negative <CODE>RegularInterval</CODE>s
	 * may seem asymmetric.
	 * The rationale behind it is that these do not contain base periods
	 * but inverted base periods. An implementation is free to
	 * add an iteration facility to return those. But the method
	 * <CODE>iterator</CODE> must return only true base periods.</P>
	 *
	 * @return an <CODE>Iterator</CODE> which always returns
	 * <CODE>RegularInterval</CODE>s with {@link #getSize} equal to 1.
	 */
	Iterator iterator ();

	/** Returns whether the interval is aligned according to its frequency.
	 *
	 * That is, for an annual interval, that it starts with a new year,
	 * 1 January 00:00:00.
	 */
	boolean isAligned ();
}
