// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2007, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.time;

import java.util.Date;
import java.util.Calendar;
import java.util.Map;
import java.util.HashMap;

/* ***************************** Class ****************************** */

/**
 * Creates a set of regular intervals to span itself.
 * This object is meant to be immutable.
 *
 * @author
 * <a href="mailto:sis.external@ecb.int">Bernhard Bodenstorfer</a>
 *
 * @version @(#)$Id: RegularisableInterval.java,v 1.1 2017/01/19 16:14:03 trainot Exp $
 */
public class RegularisableInterval extends AbstractInterval {

	/* ======================= Constants ======================== */

	private final int[] FREQUENCIES = new int[]{
		RegularInterval.ANNUAL,
		RegularInterval.SEMIANNUAL,
		RegularInterval.QUARTERLY,
		RegularInterval.MONTHLY,
		RegularInterval.WEEKLY,
		RegularInterval.DAILY,
		RegularInterval.MINUTELY
	};

	/* ========================= Fields ========================= */

	/** The configured begin date. */
	private final Date begin;

	/** The configured end date. */
	private final Date end;

	/** The covering intervals by frequency. */
	private final Map intervalByFrequency;

	/** The covering intervals by size. */
	private final Map intervalBySize;

	/* ====================== Constructors ====================== */

	/** Construct a <code>RegularisableInterval</code>.
	 *
	 * @param begin starting point of the interval.
	 * @param end end point of the interval.
	 */
	public RegularisableInterval ( final Date begin, final Date end ) {

		this.begin = begin;
		this.end = end;

		intervalByFrequency = new HashMap( 2 * FREQUENCIES.length );
		intervalBySize = new HashMap( 2 * FREQUENCIES.length );

		final Calendar calendar
		= AbstractRegularInterval.createISOCalendar();
		for ( int i = 0; i < FREQUENCIES.length; ++i ) {
			calendar.setTime( begin );
			AbstractRegularInterval.projectCalendar(
				calendar, FREQUENCIES[i]
			);
			if ( ! begin.equals( calendar.getTime() ) ) {
				continue;
			}
			calendar.setTime( end );
			AbstractRegularInterval.projectCalendar(
				calendar, FREQUENCIES[i]
			);
			if ( ! end.equals( calendar.getTime() ) ) {
				continue;
			}
			final RegularInterval interval
			= new GregorianRegularInterval( FREQUENCIES[i], begin, end );
			if (
				begin.equals( interval.getBegin() )
				&&
				end.equals( interval.getEnd() )
				&&
				interval.isAligned()
			) {
				intervalByFrequency.put(
					new Integer( interval.getFrequency() ),
					interval
				);
				intervalBySize.put(
					new Integer( interval.getSize() ),
					interval
				);
			}
		}
	}

	/* ===================== Public Methods ===================== */

	/** Retrieve the configured time interval begin. */
	public Date getBegin () {
		return begin;
	}

	/** Retrieve the configured time interval begin. */
	public Date getEnd () {
		return end;
	}

	/** Get an interval from begin to end with given frequency.
	 *
	 * @return <code>null</code> if no suitable has been found.
	 */
	public RegularInterval getIntervalByFrequency ( final int frequency ) {
		final RegularInterval interval
			= (RegularInterval) intervalByFrequency.get(
				new Integer( frequency )
			);
		return interval;
	}

	/** Get an interval from begin to end with given size.
	 *
	 * @return <code>null</code> if no suitable has been found.
	 */
	public RegularInterval getIntervalBySize ( final int size ) {
		final RegularInterval interval
			= (RegularInterval) intervalBySize.get(
				new Integer( size )
			);
		return interval;
	}

	/** Returns whether the other is a RegularisableInterval and
	 * whether their {@link Interval} properties match.
	 */
	public boolean equals ( final Object other ) {
		if ( other instanceof RegularisableInterval ) {
			final Interval that = (Interval) other;
			return AbstractInterval.equals( this, that );
		} else return false;
	}
}
