// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2007, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.time;

import java.util.Date;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.text.ParseException;

import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;

/* ***************************** Class ****************************** */

/**
 * Implements a few tests for the {@link RegularisableInterval}.
 *
 * @author
 * <a href="mailto:sis.external@ecb.int">Bernhard Bodenstorfer</a>
 *
 * @version @(#)$Id: RegularisableIntervalTest.java,v 1.1 2017/01/19 16:14:03 trainot Exp $
 */
public class RegularisableIntervalTest extends TestCase {

	/* ======================= Constants ======================== */

	private static final DateFormat ISO_FORMAT
		= new SimpleDateFormat( "yyyy-MM-dd'T'HH:mm:ss" );

	/* ====================== Constructors ====================== */

	public RegularisableIntervalTest ( final String name ) {
		super( name );
	}

	/* ===================== Public Methods ===================== */

	/**
	 * Assembles and returns a test suite for all
	 * the test methods of this test case.
	 *
	 * @return a non-null test suite
	 */
	public static Test suite () {
		return new TestSuite( RegularisableIntervalTest.class );
	}

	/**
	 * @param arg the array of argument {@link String}s
	 */
	public static void main ( final String[] arg ) {
		final String[] testCaseName = {
			RegularisableIntervalTest.class.getName()
		};
		junit.textui.TestRunner.main( testCaseName );
	}

	/* ---------------------- Test Methods ---------------------- */

	public void testProperties () throws ParseException {
		final Date begin = ISO_FORMAT.parse( "2001-01-01T00:00:00" );
		final Date end = ISO_FORMAT.parse( "2007-01-01T00:00:00" );
		final RegularisableInterval interval
			= new RegularisableInterval( begin, end );

		assertEquals(
			"begin",
			begin,
			interval.getBegin()
		);
		assertEquals(
			"end",
			end,
			interval.getEnd()
		);

		assertEquals(
			"equals",
			new RegularisableInterval( begin, end ),
			interval
		);

		assertEquals(
			"hashCode",
			new RegularisableInterval( begin, end ).hashCode(),
			interval.hashCode()
		);

		assertEquals(
			"toString",
			new RegularisableInterval( begin, end ).toString(),
			interval.toString()
		);
	}

	/** Test
	 * {@link RegularisableInterval#getIntervalByFrequency( int )}
	 * and {@link RegularisableInterval#getIntervalBySize( int )}.
	 */
	public void testAlignedIntervalMaps () throws ParseException {
		final Date
			begin = ISO_FORMAT.parse( "2001-01-01T00:00:00" ),
			end = ISO_FORMAT.parse( "2007-01-01T00:00:00" );

		final RegularisableInterval interval
			= new RegularisableInterval( begin, end );

		assertEquals(
			"Annual frequency",
			new GregorianRegularInterval( RegularInterval.ANNUAL, begin, 6 ),
			interval.getIntervalByFrequency( RegularInterval.ANNUAL )
		);
		assertEquals(
			"Semiannual frequency",
			new GregorianRegularInterval(
				RegularInterval.SEMIANNUAL, begin, 12
			),
			interval.getIntervalByFrequency( RegularInterval.SEMIANNUAL )
		);
		assertEquals(
			"Quarterly frequency",
			new GregorianRegularInterval(
				RegularInterval.QUARTERLY, begin, 24
			),
			interval.getIntervalByFrequency( RegularInterval.QUARTERLY )
		);
		assertEquals(
			"Monthly frequency",
			new GregorianRegularInterval(
				RegularInterval.MONTHLY, begin, 72
			),
			interval.getIntervalByFrequency( RegularInterval.MONTHLY )
		);
		assertEquals(
			"Weekly frequency",
			new GregorianRegularInterval( RegularInterval.WEEKLY, begin, 313 ),
			interval.getIntervalByFrequency( RegularInterval.WEEKLY )
		);
		assertEquals(
			"Daily frequency",
			new GregorianRegularInterval( RegularInterval.DAILY, begin, 2191 ),
			interval.getIntervalByFrequency( RegularInterval.DAILY )
		);

		assertNull(
			"Wrong size",
			interval.getIntervalBySize( 1 )
		);
		assertEquals(
			"Annual size",
			new GregorianRegularInterval( RegularInterval.ANNUAL, begin, 6 ),
			interval.getIntervalBySize( 6 )
		);
		assertEquals(
			"Semiannual size",
			new GregorianRegularInterval(
				RegularInterval.SEMIANNUAL, begin, 12
			),
			interval.getIntervalBySize( 12 )
		);
		assertEquals(
			"Quarterly size",
			new GregorianRegularInterval(
				RegularInterval.QUARTERLY, begin, 24
			),
			interval.getIntervalBySize( 24 )
		);
		assertEquals(
			"Monthly size",
			new GregorianRegularInterval(
				RegularInterval.MONTHLY, begin, 72
			),
			interval.getIntervalBySize( 72 )
		);
		assertEquals(
			"Weekly size",
			new GregorianRegularInterval( RegularInterval.WEEKLY, begin, 313 ),
			interval.getIntervalBySize( 313 )
		);
		assertEquals(
			"Daily size",
			new GregorianRegularInterval( RegularInterval.DAILY, begin, 2191 ),
			interval.getIntervalBySize( 2191 )
		);
	}

	/** Test
	 * {@link RegularisableInterval#getIntervalByFrequency( int )}
	 * and {@link RegularisableInterval#getIntervalBySize( int )}.
	 */
	public void testMisAlignedIntervalMaps () throws ParseException {
		final Date
			begin = ISO_FORMAT.parse( "2001-01-01T00:00:01" ),
			end = ISO_FORMAT.parse( "2007-01-01T00:00:01" );

		final RegularisableInterval interval
			= new RegularisableInterval( begin, end );

		assertNull(
			"Annual frequency",
			interval.getIntervalByFrequency( RegularInterval.ANNUAL )
		);
		assertNull(
			"Semiannual frequency",
			interval.getIntervalByFrequency( RegularInterval.SEMIANNUAL )
		);
		assertNull(
			"Quarterly frequency",
			interval.getIntervalByFrequency( RegularInterval.QUARTERLY )
		);
		assertNull(
			"Monthly frequency",
			interval.getIntervalByFrequency( RegularInterval.MONTHLY )
		);
		assertNull(
			"Weekly frequency",
			interval.getIntervalByFrequency( RegularInterval.WEEKLY )
		);
		assertNull(
			"Daily frequency",
			interval.getIntervalByFrequency( RegularInterval.DAILY )
		);

		assertNull(
			"Wrong size",
			interval.getIntervalBySize( 1 )
		);
		assertNull(
			"Annual size",
			interval.getIntervalBySize( 6 )
		);
		assertNull(
			"Semiannual size",
			interval.getIntervalBySize( 12 )
		);
		assertNull(
			"Quarterly size",
			interval.getIntervalBySize( 24 )
		);
		assertNull(
			"Monthly size",
			interval.getIntervalBySize( 72 )
		);
		assertNull(
			"Weekly size",
			interval.getIntervalBySize( 313 )
		);
		assertNull(
			"Daily size",
			interval.getIntervalBySize( 2191 )
		);
	}
}
