// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2009, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.time;

import java.util.regex.Matcher;
import java.util.regex.Pattern;

import Int.ecb.gesmescb.util.ComponentHelper;
import Int.ecb.logging.Level;

/* ***************************** Class ****************************** */

/**
 * Provide some utility to rewrite TimeStamp Strings.
 *
 * @author
 * <A HREF="mailto:Francoid.Hoeltzli@ecb.europa.eu">Francois Hoeltzli</A>
 * @author
 * <A HREF="mailto:Xavier.Sosnovsky@ecb.europa.eu">Xavier Sosnovsky</A>
 *
 * @version @(#)$Id: TimeZoneUtil.java,v 1.4 2018/03/08 10:48:10 kortest Exp $
 */

public class TimeZoneUtil {
	
	/* ======================= Constants ======================== */
	
	public final static String RFC822 = "RFC822";
	public final static String ISO8601 = "ISO8601";
	
	/* ========================= Fields ========================= */
	
	/** Takes care of logging and preferences. */
	private static ComponentHelper componentHelper;	
	
	private static TimeZoneUtil instance;
	private static Pattern fullPattern;
	private static Pattern fullPatternWithMS;
	private static Pattern rfc822Pattern;
	private static Pattern militaryPattern;
	private static Pattern lousyPattern;
	private static Character A;
	private static Character J;
	private static Character M;
	private static Character Y;
	private static Character Z;
	
	/* ===================== Public Methods ===================== */	
	
	public static TimeZoneUtil getInstance() {
		if (null == instance) {
			instance = new TimeZoneUtil();
			componentHelper = new ComponentHelper( "", instance);
			fullPattern = Pattern.compile("^(\\d{4}-\\d{2}-\\d{2}T\\d{2}:\\d{2}:\\d{2})(.*)$");
			fullPatternWithMS = Pattern.compile("^(\\d{4}-\\d{2}-\\d{2}T\\d{2}:\\d{2}:\\d{2}\\.\\d{3})(.+)$");
			rfc822Pattern = Pattern.compile("^([+-]\\d{2}):?(\\d{2})$");
			militaryPattern = Pattern.compile("^[A-IK-Z]$"); // J is skipped.
			lousyPattern = Pattern.compile("^GMT|UTC|CET|EET|MSK$");
			A = new Character('A');
			J = new Character('J');
			M = new Character('M');
			Y = new Character('Y');
			Z = new Character('Z');
		}
		return instance;
	};

	/**
	 * This function takes some input and tries to reformat it
	 * RFC822 compliant timestamp (Java can work natively with this
	 * format).
	 */
	public String normalize( final String inputDateString ) {
		return normalize( inputDateString, RFC822);
	}

	public String normalize(
		final String inputDateString,
		final String format
	) {
		String separator = "";
		if ( RFC822.equals( format ) ) {
			separator = "";
		} else if ( ISO8601.equals( format) ) {
			separator = ":";
		} else {
			componentHelper.logger.log(
				Level.SEVERE, 
				componentHelper.logName, 
				"normalize",
				"Unknown Format \""
					+ format
					+ "\". Only RFC822 and ISO8601 are allowed.\n"
			);
		}
		Matcher fullMatcher = fullPattern.matcher( inputDateString );
		String returnString = inputDateString;
		Matcher fullWithMSMatcher = fullPatternWithMS.matcher( inputDateString );
		if ( fullMatcher.matches() ) {
			String dateString;
			String timeZoneString;
			if ( fullWithMSMatcher.matches() ){
				dateString = fullWithMSMatcher.group(1);
				// Get rid of the milliseconds:
				dateString = dateString.substring(0, dateString.lastIndexOf("."));
				timeZoneString = fullWithMSMatcher.group(2);
			} else {
				dateString = fullMatcher.group(1);
				timeZoneString = fullMatcher.group(2);
			}
			// provide reasonable default time zone if none
			if ( "".equals( timeZoneString ) ) {
				returnString = dateString + "+00" + separator + "00";
			} else {
				Matcher rfc822Matcher
				= rfc822Pattern.matcher( timeZoneString );
				Matcher militaryMatcher
				= militaryPattern.matcher( timeZoneString );
				Matcher lousyMatcher
				= lousyPattern.matcher( timeZoneString);
				if ( rfc822Matcher.matches() ) {
					/* hopefully usual case, hence first */ 
					 returnString = dateString
					 	+ rfc822Matcher.group(1)
					 	+ separator 
						+ rfc822Matcher.group(2);
				} else if ( militaryMatcher.matches() ) {
					timeZoneString
					= demilitarize(
						new Character( timeZoneString.charAt( 0 ) )
						, separator  
					);
					returnString = dateString + timeZoneString;
				} else if ( lousyMatcher.matches() ) {
					/* we try with a few well known timezones 
					 *  GMT|UTC|CET|EET|MSK
					 */
					if (
						"GMT".equals( timeZoneString )
						||
						"UTC".equals( timeZoneString )
					) {
						returnString = dateString + "+00" + separator + "00";
					} else if ("CET".equals(timeZoneString)) {		
						returnString = dateString + "+01" + separator + "00";
					} else if ("EET".equals(timeZoneString)) {		
						returnString = dateString + "+02" + separator + "00";
					} else if ("MSK".equals(timeZoneString)) {		
						returnString = dateString + "+03" + separator + "00";
					} else { 
						returnString = null;
						componentHelper.logger.log(
							Level.SEVERE, 
							componentHelper.logName, 
							"normalize",
							"TimeZoneString \""
								+ timeZoneString + "\" accepted by "
								+ "ill defined lousyPattern."
						);
					}
				} else { 
					returnString = null;
					componentHelper.logger.log(
						Level.SEVERE, 
						componentHelper.logName, 
						"normalize",
						"Unknown TimeZoneString \""+ timeZoneString + "\""
					);
				}
			}
		} else {
			returnString = null;
			componentHelper.logger.log(
				Level.SEVERE, 
				componentHelper.logName, 
				"normalize",
				"Unreadable inputDateString \""+ inputDateString + "\""
			);
		}
		return returnString;
	}
	
	/* ==================== Private Methods ===================== */	
	
	private String demilitarize( Character zone, String separator ) {
		if ( zone.charValue() < J.charValue() ) {
			return "+"
				+ fourDigitize(
					zone.charValue() - A.charValue() + 1,
					separator
				);
		} else if ( zone.charValue() <= M.charValue() ) {
			return "+"
				+ fourDigitize( zone.charValue() - A.charValue(), separator );
		} else if ( zone.charValue() <= Y.charValue() ) {
			return "-"
				+ fourDigitize( Z.charValue() - zone.charValue(), separator );
		} else if ( zone.charValue() == Z.charValue() ) {
			return "+00" + separator + "00";
		} else {
			return null;
		}
	};
	
	private String fourDigitize( int n, String separator ) {
		/* +8 should be represented as +0800 or +08:00,
		 * -11 as -1100 or -11:00.
		 */
		String mask = "00";
 		String ds = Integer.toString(n); 
 		return mask.substring( 0 , mask.length() - ds.length() )
			+ ds
			+ separator
			+ "00";
	}
}
