// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.util;

import Int.ecb.gesmescb.util.Named;
import Int.ecb.logging.Logger;
import Int.ecb.prefs.Preferences;

/* ***************************** Class ****************************** */

/**
 * Provides commonly used facilities for all GESMES/TS processing components.
 * Typical components are
 * {@link Int.ecb.gesmescb.event.GESMESCBEventSource}s
 * or {@link Int.ecb.gesmescb.event.GESMESCBEventListener}s.
 * Essentially everything that is important enough in that
 * it can be addressed in the configuration or sends messages to log
 * can be viewed as a component.
 * Covered features are naming, logging, and access to preferences.
 *
 * @author
 * <A HREF="mailto:Sami.Airo@ECB.int">Sami Airo</A>
 * @author
 * <A HREF="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</A>
 *
 * @version @(#)$Id: ComponentHelper.java,v 1.4 2007/07/20 12:55:33 michelb Exp $
 */

public class ComponentHelper {

	/* ========================= Fields ========================= */

	/** The component's name. Also useful to implement {@link Named}. */
	public final String name;

	/** The component's package and class name. Used for logging. */
	private final String className;

	/** The component's class name extended by the name.
	 * Used for accessing preferences.
	 */
	private final String pureClassAndInstanceName;

	/** The <CODE>Logger</CODE> to be used by the component.
	 * This field is public for maximum performance.
	 */
	public final Logger logger;

	/** The name to be used by the component for logging.
	 * This field is public for maximum performance.
	 */
	public final String logName;

	/** The component's system preferences. */
	private final Preferences systemPrefs;

	/** The component's user preferences. */
	private final Preferences userPrefs;

	/* ====================== Constructors ====================== */

	/** Prepare a Componenthelper for a <CODE>component</CODE>.
	 * @param name for the <CODE>component</CODE>
	 * @param component to be helped.
	 */
	public ComponentHelper ( final String name, final Object component ) {
		final String nameExtension
		= ( name != null && ! "".equals( name ) ) ? "." + name : "";
	
		this.name = name;
		className = component.getClass().getName();
				
		pureClassAndInstanceName
		= getPureClassName( component ) + nameExtension;
		logName = className + nameExtension;
		
		logger = Logger.getLogger( className + nameExtension );
		systemPrefs = Preferences.systemNodeForPackage( component );
		userPrefs = Preferences.userNodeForPackage( component );
	}

	/* ===================== Public Methods ===================== */

	/**
	 * Returns a preference value.
	 * Looks first to the system preferences and then to the user preferences.
	 * User preference overrides the system preference value.
	 *
	 * @param key	key for the preference
	 * @param def	default value to use
	 *
	 * @return preference value for the specified key or the default value
	 */
	public String getPreference ( final String key, final String def ) {
		final String lookup = pureClassAndInstanceName + "." + key;
		String value = systemPrefs.get( lookup, userPrefs.get( lookup, def ) );
		return value;
	}

	/* ==================== Private Methods ===================== */

	/**
	 * Get the class name without package name.
	 *
	 * @param obj to get the class name of.
	 *
	 * @return the class name without package name.
	 */
	private static String getPureClassName ( final Object obj ) {
		
		if ( obj == null ) {
			return null;
		}

		final String className = obj.getClass().getName();

		final String packageName = PackageProperties.getName(className);
		
		return className.substring( packageName.length() + 1 );
	}
}
