// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (C) 2004, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.util;

import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;

import Int.ecb.gesmescb.KeyFamily;
import Int.ecb.gesmescb.KeyFamilyComponent;
import Int.ecb.gesmescb.TimeSeries;
import Int.ecb.gesmescb.Attribute;

import Int.ecb.gesmescb.util.FilterIterator.Condition;

/* ***************************** Class ****************************** */

/**
 * Filters {@link Int.ecb.gesmescb.AbstractCube} objects of given cube class.
 *
 * @author
 * <A HREF="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</A>
 *
 * @version @(#)$Id: CubeClassCondition.java,v 1.1 2017/01/19 16:14:04 trainot Exp $
 */

public class CubeClassCondition implements Condition {

	/* ======================= Constants ======================== */

	/** Accept data set level {@link Int.ecb.gesmescb.AbstractCube}s. */
	public static final int DATA_SET = 1;

	/** Accept sibling group level {@link Int.ecb.gesmescb.AbstractCube}s. */
	public static final int SIBLING_GROUP = 2;

	/** Accept time series level {@link Int.ecb.gesmescb.AbstractCube}s. */
	public static final int TIME_SERIES = 4;

	/** Accept observation level {@link Int.ecb.gesmescb.AbstractCube}s. */
	public static final int OBSERVATION = 8;

	/* ========================= Fields ========================= */

	/** Which {@link Int.ecb.gesmescb.AbstractCube} levels are eligible. */
	private final int cubeClasses;

	/** The names of all concepts with the given attachment level. */
	private final Collection conceptIds;

	/* ====================== Constructors ====================== */

	/**
	 * Construct a new <code>CubeClassCondition</code>
	 * to filter all {@link Attribute}s with matching attachment level.
	 *
	 * @param keyFamily corresponding to the data set.
	 * @param cubeClasses selected by a binary <code>OR</code> pattern
	 * based on {@link #OBSERVATION}, {@link #TIME_SERIES},
	 * {@link #SIBLING_GROUP}, and {@link #DATA_SET}
	 */
	public CubeClassCondition (
		final KeyFamily keyFamily,
		final int cubeClasses
	) {
		if (
			cubeClasses < 0
			||
			15 < cubeClasses
		) {
			throw new IllegalArgumentException(
				"Invalid cube class selection pattern " + cubeClasses
				+ " to construct a CubeClassCondition."
			);
		}

		this.cubeClasses = cubeClasses;

		conceptIds = new HashSet();
		for (
			final Iterator iterator = keyFamily.componentIterator();
			iterator.hasNext();
		) {
			final KeyFamilyComponent component
			= (KeyFamilyComponent) iterator.next();
			if (
				( cubeClasses & DATA_SET ) != 0
				&&
				component.isAttachedToDataSet()
				||
				( cubeClasses & SIBLING_GROUP ) != 0
				&&
				component.isAttachedToSiblingGroup()
				||
				( cubeClasses & TIME_SERIES ) != 0
				&&
				component.isAttachedToTimeSeries()
				||
				( cubeClasses & OBSERVATION ) != 0
				&&
				component.isAttachedToObservation()
			) {
				conceptIds.add( component.getConceptId() );
			}
		}
	}

	/* ===================== Public Methods ===================== */

	/**
	 * Returns whether the <code>object</code> is an
	 * {@link Int.ecb.gesmescb.AbstractCube} with matching <code>key</code>.
	 */
	public boolean accept ( final Object object ) {
		if ( object instanceof TimeSeries ) {
			return ( cubeClasses & TIME_SERIES ) != 0;
		}
		if ( object instanceof Attribute ) {
			return conceptIds.contains( ( (Attribute) object ).getName() );
		}
		return false;
	}

	public String toString () {
		return super.toString() + "( \"" + conceptIds + "\" )";
	}
}
