// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2007, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.util;

import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.ArrayList;
import java.util.Map;

import java.text.DecimalFormat;

import Int.ecb.logging.Level;

import Int.ecb.gesmescb.*;

import Int.ecb.gesmescb.event.GESMESCBEventBroker;

/* **************************** Class ***************************** */

/**
 * <P>Trims the decimals of every {@link Observation} according
 * to preferences.</P>
 *
 * @author
 * <A HREF="mailto:francois.hoeltzli@ECB.int">Fran&ccedil;ois Hoeltzli</A>
 *
 * @version @(#)$Id: DecimalsTrimmer.java,v 1.1 2017/01/19 16:14:04 trainot Exp $
 */

public final class DecimalsTrimmer extends GESMESCBEventBroker {

	/* ======================= Constants ======================== */

	public static final int HALF_EVEN_MODE = 0;
	public static final int HALF_UP_MODE = 1;
	
	private static final Map formatterMap = new HashMap(); 

	/* ========================= Fields ========================= */

	/**
	 * Number of decimals to be dispatched. 
	 * Negative value means no decimals will be dispatched.
	 */
	private int dispatchedDecimals;
	
	private int MAX_DISPATCHED_DECIMALS = 15;
	private DecimalFormat dispatchFormat;

	/* ====================== Constructors ====================== */

	public DecimalsTrimmer () {
		this( "" );
	}

	/**
	 * Constructs a DecimalsTrimmer.
	 */
	public DecimalsTrimmer ( final String name ) {
		super( name, true );

		final String decs
		= componentHelper.getPreference("dispatchedDecimals", null);
		if (decs == null) { 
			componentHelper.logger.log(
				Level.CONFIG, componentHelper.logName, "DecimalsTrimmer",
				"dispatchedDecimals NOT set. Trimmer will do nothing."
			);
			return;
		}

		try {
			dispatchedDecimals = Integer.parseInt(decs);
			componentHelper.logger.log(
				Level.CONFIG, componentHelper.logName, "DecimalsTrimmer",
				"dispatchedDecimals set to " + dispatchedDecimals 
			);
		} catch (NumberFormatException e) {
			componentHelper.logger.log(
				Level.SEVERE, componentHelper.logName, "DecimalsTrimmer",
				"Invalid preference value for dispatchedDecimals: " + decs
			);
			throw new NullPointerException(e.getMessage());
		}
		if ( dispatchedDecimals > MAX_DISPATCHED_DECIMALS ) {
			componentHelper.logger.log(
				Level.WARNING, componentHelper.logName, "DecimalsTrimmer",
				"Large dispatchedDecimals reduced to MAX_DISPATCHED_DECIMALS :"
				+ " " + Integer.toString( MAX_DISPATCHED_DECIMALS ) + "."
			);
			dispatchedDecimals = MAX_DISPATCHED_DECIMALS;	
		}
		if ( 0 <= dispatchedDecimals ) {
			StringBuffer pattern = new StringBuffer( "0" );
			if ( 0 < dispatchedDecimals ) {
				pattern.append( "." );
				for ( int i = 0; i < dispatchedDecimals; i++ ) {
					pattern.append( "#" );
				}
			}
			dispatchFormat = new DecimalFormat( pattern.toString() );
		} else {
			componentHelper.logger.log(
				Level.WARNING, componentHelper.logName, "DecimalsTrimmer",
				"negative value for dispatchedDecimals will be ignored." 
			);
		}
	}

	/* ===================== Public Methods ===================== */

	/**
	 * Dispatches the observations with up to dispatchedDecimals decimals.
	 *
	 * @param timeSeries coming in.
	 */
	public void handle ( final TimeSeries timeSeries ) {
		
		componentHelper.logger.log(
			Level.FINE, componentHelper.logName, "handle",
			"handling TimeSeries", timeSeries
		);
		
		if ( dispatchFormat == null ) {
			dispatch( timeSeries );
			return;
		}
		
		final List observationList = new ArrayList();
		for( final Iterator it = timeSeries.observationIterator()
			; it.hasNext()
			;
			) {
			final Observation observation = (Observation) it.next();
			if ( Observation.EMPTY_OBSERVATION.equals( observation ) ) {
				observationList.add( observation );
			} else {
				final String value = observation.getValue();
				observationList.add(
					new Observation(
						Observation.MISSING_VALUE.equals( value )
							? Observation.MISSING_VALUE
							: format_with_round_half_up (
									value
							),
						observation.getStatus(),
						observation.getConfidentiality(),
						observation.getPreBreakValue()
					)
				);
			}
		}
		final TimeSeries decimalTrimmedTS = new TimeSeries (
			timeSeries.getKey(),
			timeSeries.getInterval(),
			observationList
		);
		dispatch( decimalTrimmedTS );
	}

	/* ===================== Static Methods ====================== */

	public static String format( final double value, final int decimals, final int mode ) throws Exception {
		
		if ( decimals < 0) {
			throw new Exception(
					"DecimalsTrimmer Exception in format: "
					+ "Can't format with negative number (\""
					+ decimals + "\")of decimals.\n" );
		}
		
		DecimalFormat formatter = getFormatterFromPool(decimals);
		String formatted_value	= formatter.format( value );
		
		switch(mode) {
			case HALF_EVEN_MODE:
				break;
			case HALF_UP_MODE:
				String valueAsString = String.valueOf( value );
				int valueLength = valueAsString.length();
				if ( "5".equals( valueAsString.substring( valueLength-1 ) ) ) {
					if ( valueLength - valueAsString.indexOf(".") - 2 == decimals ) {
						formatted_value = formatter.format (
							Double.valueOf( valueAsString.substring( 0, valueLength-1 ) + "6" )
						);
					}
				}
				break;
			default:
				throw new Exception(
						"DecimalsTrimmer Exception in format: "
						+ "Unknown formatting decimals mode(\""
						+ mode + "\").\n"
				);
		}
		return formatted_value;
	}
	
	/**
	 * Get formatters with different decimals from a pool of formatters. Only one formatter 
	 * is created for each decimals, and it is reused for later requests with the same decimals.
	 * @param decimals
	 * @return
	 */
	private static DecimalFormat getFormatterFromPool(final int decimals) {
		DecimalFormat formatter = (DecimalFormat )formatterMap.get(String.valueOf(decimals));
		if (null == formatter) {
			StringBuffer pattern = new StringBuffer( "0" );
			if ( 0 < decimals ) {
				pattern.append( "." );
				for ( int i = 0; i < decimals; i++ ) {
					pattern.append( "#" );
				}
			}
			formatter = new DecimalFormat( pattern.toString() );
			formatterMap.put(String.valueOf(decimals), formatter);
		}
		return formatter;
	}
	
	/* ==================== Private Methods ===================== */
	
	private String format_with_round_half_up( final String value ) {
	    if (value != null && !value.isEmpty() ) {
		String formatted_value
		= dispatchFormat.format( Double.valueOf ( value ) );
		int valueLength = value.length();
		if ( "5".equals( value.substring( valueLength-1 ) ) ) {
			if ( valueLength - value.indexOf(".") - 2 == dispatchedDecimals ) {
				formatted_value = dispatchFormat.format (
					Double.valueOf( value.substring( 0, valueLength-1 ) + "6" )
				);
			}
		}
		return formatted_value;
	    } else {
	    	return "";
	    }
	    	
	}
}
