// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.util;

import java.lang.String;

/* ***************************** Class ****************************** */

/**
 * A fast and handy unsynchronized replacement of the
 * class {@link java.lang.StringBuffer} handcrafted for the purposes of
 * the {@link Int.ecb.gesmescb.edifact.EDIFACTReadBuffer}
 * and {@link Int.ecb.gesmescb.edifact.EDIFACTReader}.
 *
 * @author
 * <A HREF="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</A>
 *
 * @version @(#)$Id: FastStringBuffer.java,v 1.1 2017/01/19 16:14:04 trainot Exp $
 */

public class FastStringBuffer {

	/* ======================= Constants ======================== */

	/**
	 * The initial size of the internal buffer[].
	 * This should be a power of 2 and not less than 1.
	 */
	private static final int DEFAULT_INITIAL_CAPACITY = 256;

	/* ========================= Fields ========================= */

	/**
	 * The private data storage.
	 */
	private char buffer[];

	/**
	 * The number of valid characters in buffer.
	 */
	private int top;

	/* ====================== Constructors ====================== */

	/**
	 * Construct a FastStringBuffer with default initial capacity.
	 */
	public FastStringBuffer () {
		this ( DEFAULT_INITIAL_CAPACITY );
	}

	/**
	 * Construct a FastStringBuffer with proposed minimum initial capacity.
	 */
	public FastStringBuffer ( int capacity ) {
		if ( capacity < 0 ) {
			capacity = DEFAULT_INITIAL_CAPACITY;
		}
		buffer = new char[ capacity ];
		top = 0;
	}

	/* ===================== Public Methods ===================== */

	/**
	 * Clear all data contents.
	 */
	public void clear () {
		top = 0;
	}

	/**
	 * Append a character to the end of the contents.
	 *
	 * @param c contains the character to append.
	 */
	public void append ( char c ) {

		// Assure that there is enough space in buffer[].
		resize( top + 1 );

		// And append.
		buffer[top++] = c;
	}

	/**
	 * Append a part of a string to the end of the contents.
	 * Unlike with StringBuffer, no attempt is made to read string
	 * beyond its length.
	 *
	 * @param string contains the characters to append.
	 * @param offset the position of the first character to append.
	 * @param len the number of characters to append.
	 */
	public void append ( String string, int offset, int len ) {
		final int stringLen = string.length();
		int appendTop = offset + len;

		// Avoid throwing out-of-bounds exceptions.
		if ( appendTop > stringLen ) {
			appendTop = stringLen;
		}

		// Assure that there is enough space in buffer[].
		resize( top + appendTop - offset );

		// Copy string contents.
		while ( offset < appendTop ) {
			buffer[top++] = string.charAt( offset++ );
		}
	}

	/**
	 * Append a fimal part of a string to the end of the contents.
	 * All characters from position offset onwards are appended.
	 *
	 * @param string contains the characters to append.
	 * @param offset the position of the first character to append.
	 */
	public void append ( String string, int offset ) {
		append( string, offset, string.length() );
	}

	/**
	 * Append a string to the end of the contents.
	 *
	 * @param string contains the characters to append.
	 */
	public void append ( String string ) {
		append( string, 0, string.length() );
	}

	/**
	 * Append a part of a char array to the end of the contents.
	 * Unlike with StringBuffer, no attempt is made to read the array
	 * beyond its bounds.
	 *
	 * @param array contains the characters to append.
	 * @param offset the position in array of the first character
	 * to append.
	 * @param len the number of characters to append.
	 */
	public void append ( char[] array, int offset, int len ) {
		final int arrayLen = array.length;
		int appendTop = offset + len;

		// Avoid throwing out-of-bounds exceptions.
		if ( appendTop > arrayLen ) {
			appendTop = arrayLen;
		}

		// Assure that there is enough space in buffer[].
		resize( top + appendTop - offset );

		int appendLen = appendTop - offset;

		// If there is data to append.
		if ( appendLen > 0 ) {
			// Copy array contents.
			System.arraycopy(
				array, offset, buffer, top, appendLen
			);
			// And adjust top accordingly.
			top += appendLen;
		}

	}

	/**
	 * Get the least nonnegative power of 2
	 * which is greater or equal to n.
	 * The assumption is that the type int can hold such a bound.
	 *
	 * @param n specifies the value to be bounded.
	 *
	 * @return the least nonnegative power of 2 which bounds n.
	 *
	 * @throws IllegalArgumentException if n is so large that
	 * int cannot hold the result.
	 */
	public static int powerBound ( final int n )
		throws IllegalArgumentException
	{
		int i, k;

		// Get a sort of logarithm.
		for ( k = n-1, i = 0 ; k > 0; k >>= 1 ) {
			++i;
		}

		// And now the sought bound.
		// Note that bitwise shift ANDs the second argument.
		int bound = 1 << i;

		// Throw an exception in the unlikely case that
		// int is not large enough to really hold the required bound.
		// In the foreseen use this can only appear
		// for senselessly large array sizes.
		if ( bound < n ) {
			throw new IllegalArgumentException(
				"The number " + n + "is too large to "
				+ "be bounded by an int power of two."
			);
		}

		return bound;
	}

	/**
	 * Transform the present contents into a String.
	 *
	 * @return the present contents.
	 */
	public String toString () {
		return new String( buffer, 0, top );
	}

	/* ==================== Private Methods ===================== */

	/**
	 * Resize buffer[] to at least the given size.
	 * If size is less or equal than the present size,
	 * no action is taken. Particularly no downsizing will occur.
	 *
	 * @param size the proposed buffer size.
	 */
	private void resize ( int size ) {
		if ( size > buffer.length ) {
			// Allocate, fill, and install new buffer[] array.
			char[] newBuffer = new char[ powerBound( size ) ];
			System.arraycopy( buffer, 0, newBuffer, 0, top );
			buffer = newBuffer;
		}
	}
}
