// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (C) 2004, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.util;

import java.util.Iterator;
import java.util.NoSuchElementException;

/* ***************************** Class ****************************** */

/**
 * <p>Filters objects that an {@link Iterator} retrieves.</p>
 *
 * @author
 * <A HREF="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</A>
 *
 * @version @(#)$Id: FilterIterator.java,v 1.1 2017/01/19 16:14:04 trainot Exp $
 */
public final class FilterIterator implements Iterator {

	/* ======================= Interfaces ======================= */

	public interface Condition {
		boolean accept ( Object object );
	}

	/* ========================= Fields ========================= */

	private final Condition condition;
	private final Iterator iterator;

	private boolean hasNext;	// Needed to safely iterate null values.
	private Object next;

	/* ====================== Constructors ====================== */

	/** Create a new instance.
	 * @param iterator to filter objects from.
	 * @param condition to use for filtering.
	 */
	public FilterIterator (
		final Condition condition,
		final Iterator iterator
	) {
		this.condition = condition;
		this.iterator = iterator;
		searchNext();
	}

	/* ===================== Public Methods ===================== */

	public boolean hasNext () {
		return hasNext;
	}

	public Object next () {
		if ( hasNext() ) {
			final Object returnValue = next;
			searchNext();
			return returnValue;
		} else {
			throw new NoSuchElementException( "No more filter results." );
		}
	}

	/** Unimplemented.
	 * In fact, it cannot be done fully in line with the contract.
	 * The call sequence next(), hasNext(), remove() is allowed,
	 * but impossible to implement.
	 * The usual sequence hasNext(), next(), remove() would be unproblematic.
	 */
	public void remove () {
		throw new UnsupportedOperationException(
			"Sorry, this is unimplementable."
		);
	}

	/* ==================== Private Methods ===================== */

	private void searchNext () {
		hasNext = false;
		while ( iterator.hasNext() ) {
			final Object object = iterator.next();
			if ( condition.accept( object ) ) {
				next = object;
				hasNext = true;
				break;
			}
		}
	}
}
