// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2003, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.util;

import java.io.Reader;
import java.io.IOException;

import java.awt.Frame;
import java.awt.Window;
import java.awt.Dialog;
import java.awt.Container;
import java.awt.TextArea;
import java.awt.Button;
import java.awt.BorderLayout;
import java.awt.GridLayout;

import java.awt.event.WindowListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;

/* ***************************** Class ****************************** */

/**
 * Provides a pop-up dialog containing read-only text.
 *
 * @author
 * <a href="mailto:Paula.Rocha@ECB.int">Paula Rocha</a>
 * @author
 * <a href="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</a>
 *
 * @version @(#)$Id: InfoDialog.java,v 1.1 2017/01/19 16:14:04 trainot Exp $
 */
public class InfoDialog extends Dialog {

	/**
	 *   serialVersionUID generated by eclipse.
	 */
	private static final long serialVersionUID = -4732387267729719052L;
	
	/** Stores the chosen option.
	 * @serial Index number of the chosen option.
	 */
	int chosen;

	/* ====================== Constructors ====================== */

	public InfoDialog ( final Frame owner, final String title, final String text ) {
		this( owner, title, text, new String[]{ "Ok" }, false );
	}

	public InfoDialog (
		final Frame owner,
		final String title,
		final String text,
		final String[] choice,
		final boolean modal
	) {
		super( owner, title, modal );
		chosen = -1;
		setLayout( new BorderLayout() );
		final TextArea textArea = new TextArea(
			text, height( text ), width( text ), TextArea.SCROLLBARS_VERTICAL_ONLY
		);
		textArea.setEditable( false );
		add( textArea, BorderLayout.CENTER );
		final Container buttonList = new Container();
		buttonList.setLayout( new GridLayout() );
		add( buttonList, BorderLayout.SOUTH );
		for ( int i = 0; i < choice.length; ++i ) {
			final Button button = new Button( choice[i] );
			buttonList.add( button );
			final ActionListener actionListener = new ButtonListener( this, i );
			button.addActionListener( actionListener );
		}
		final WindowListener windowCloseListener = new WindowCloseListener( this );
		addWindowListener( windowCloseListener );
	}

	public InfoDialog (
		final Frame owner, final String title, final Reader reader
	) throws IOException {
		this( owner, title, InfoDialog.readToString( reader ) );
	}

	/* ===================== Public Methods ===================== */

	public int getChosen () {
		return chosen;
	}

	/* ==================== Private Methods ===================== */

	private static String readToString ( final Reader reader ) throws IOException {
		final StringBuffer buffer = new StringBuffer();
		int c;
		while ( ( c = reader.read() ) != -1 ) {
			buffer.append( (char) c );
		}
		return buffer.toString();
	}

	/** Heuristic to set text area width. */
	private static int width ( final String text ) {
		final int length = text.length();
		if ( length >= 80 ) {
			return 80;
		} else {
			return length;
		}
	}

	/** Heuristic to set text area height. */
	private static int height ( final String text ) {
		final int length = text.length();
		int lines = 1;
		for ( int i = 0; i < length; ++i ) {
			if ( text.charAt( i ) == '\n' ) {
				++lines;
			}
		}
		lines += length/80;
		if ( lines >= 25 ) {
			return 25;
		} else {
			return lines;
		}
	}
}

/* ***************************** Class ****************************** */

class WindowCloseListener extends WindowAdapter implements ActionListener {
	private final Window window;

	public WindowCloseListener ( final Window window ) {
		this.window = window;
	}

	public void windowClosing ( final WindowEvent e ) {
		window.hide();
	}

	public void actionPerformed ( final ActionEvent event ) {
		window.hide();
	}
}

/* ***************************** Class ****************************** */

class ButtonListener extends WindowCloseListener {
	private final InfoDialog infoDialog;
	private final int choice;

	public ButtonListener ( final InfoDialog infoDialog, final int choice ) {
		super( infoDialog );
		this.infoDialog = infoDialog;
		this.choice = choice;
	}

	public void actionPerformed ( final ActionEvent event ) {
		super.actionPerformed( event );
		infoDialog.chosen = choice;
	}
}
