// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2007, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.util;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;

import Int.ecb.logging.Level;

import Int.ecb.gesmescb.*;

import Int.ecb.gesmescb.event.GESMESCBEventBroker;

/* **************************** Class ***************************** */

/**
 * <P>Overwrite the {@link InterchangeBegin} according 
 * to preferences.</P>
 *
 * @author
 * <A HREF="mailto:francois.hoeltzli@ECB.int">Fran&ccedil;ois Hoeltzli</A>
 *
 * @version @(#)$Id: InterchangeBeginOverwriter.java,v 1.1 2017/01/19 16:14:04 trainot Exp $
 */

public final class InterchangeBeginOverwriter extends GESMESCBEventBroker {

	/* ======================= Constants ======================== */

	/* ========================= Fields ========================= */

	final String iref;
	final String sender;
	final String receiver;
	final String preparation;
	final String isTest;
	
	/* ====================== Constructors ====================== */

	public InterchangeBeginOverwriter () {
		this( "" );
	}

	/**
	 * Constructs a InterchangeBeginOverwriter.
	 */
	public InterchangeBeginOverwriter ( final String name ) {

		super( name, true );

		iref = componentHelper.getPreference("iref", null);
		sender = componentHelper.getPreference("sender", null);
		receiver = componentHelper.getPreference("receiver", null);
		preparation = componentHelper.getPreference("preparation", null);
		isTest = componentHelper.getPreference("isTest", null);
		
		componentHelper.logger.log(
			Level.CONFIG, componentHelper.logName,
			"InterchangeBeginOverwriter",
			iref == null
				? "iref not reconfigured."
				: "iref reconfigured to " + iref
		);
		componentHelper.logger.log(
			Level.CONFIG, componentHelper.logName,
			"InterchangeBeginOverwriter",
			sender == null
				? "sender not reconfigured."
				: "sender reconfigured to " + sender
		);
		componentHelper.logger.log(
			Level.CONFIG, componentHelper.logName,
			"InterchangeBeginOverwriter",
			receiver == null
				? "receiver not reconfigured."
				: "receiver reconfigured to " + receiver
		);	
		componentHelper.logger.log(
			Level.CONFIG, componentHelper.logName,
			"InterchangeBeginOverwriter",
			preparation == null
				? "preparation not reconfigured."
				: "preparation reconfigured to " + preparation
		);	
		componentHelper.logger.log(
			Level.CONFIG, componentHelper.logName,
			"InterchangeBeginOverwriter",
			isTest == null
				? "isTest not reconfigured."
				: "isTest reconfigured to " + isTest
		);			
	}

	/* ===================== Public Methods ===================== */

	/**
	 * Dispatches the interchangeBegin as reconfigured.
	 *
	 * @param interchangeBegin coming in.
	 */
	public void handle ( final InterchangeBegin interchangeBegin ) {
		
		componentHelper.logger.log(
			Level.FINE, componentHelper.logName, "handle",
			"handling InterchangeBegin", interchangeBegin
		);
		
		dispatch(
			iref == null 
			&& sender == null 
			&& receiver == null 
			&& preparation == null 
			&& isTest == null
			? interchangeBegin
			: new InterchangeBegin(
				iref == null ? interchangeBegin.getIref() : iref,
				sender == null ? interchangeBegin.getSender() : sender,
				receiver == null ? interchangeBegin.getReceiver() : receiver,
				preparation == null
					? interchangeBegin.getPreparationDateTime()
					: preparation,
				isTest == null
					? interchangeBegin.isTest()
					: "true".equals(isTest)
			)
		);
	}

	/**
	 * Dispatches the interchangeEnd as reconfigured.
	 *
	 * @param interchangeEnd coming in.
	 */
	public void handle ( final InterchangeEnd interchangeEnd ) {
		
		componentHelper.logger.log(
			Level.FINE, componentHelper.logName, "handle",
			"handling InterchangeEnd", interchangeEnd
		);
		dispatch(
			iref == null
				? interchangeEnd
				: new InterchangeEnd ( iref )
		);
	}
	
	/**
	 * Dispatches the messageBegin, replacing the preparation Timestamp.
	 *
	 * @param messageBegin coming in.
	 */
	public void handle ( final MessageBegin messageBegin ) {
		
		componentHelper.logger.log(
			Level.FINE, componentHelper.logName, "handle",
			"handling MessageBegin", messageBegin
		);
		
		if (preparation == null ) {
			dispatch( messageBegin );
		} else {
			Collection contacts = new ArrayList();
			Iterator contactsIterator = messageBegin.contactIterator();
			while( contactsIterator.hasNext() ) {
				contacts.add(contactsIterator.next());
			}
			/* string for preparation is YYMMDD:HHMM in InterchangeBegin
			 * and CCYYMMDDHHMM in MessageBegin. TODO: Fix that!
			 */
			String prepString = "20" + preparation.replaceFirst(":", "" );
			
			dispatch(
				new MessageBegin (
					messageBegin.getMref(),
					messageBegin.getSender(),
					messageBegin.getReceiver(),
					messageBegin.getMaintenanceAgencyId(),
					messageBegin.getFunction(),
					messageBegin.getIdentity(),
					contacts,
					messageBegin.getDataSetId(),
					messageBegin.getAction(),
					prepString,
					messageBegin.getReportingPeriod(),
					messageBegin.getReportingPeriodFormat(),
					messageBegin.getKeyFamilyId(),
					messageBegin.getOriginalDataSetId()
				)
			);
		}
	}

	
	/* ==================== Private Methods ===================== */
}
