// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2007, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.util;


import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;

import Int.ecb.gesmescb.*;

import Int.ecb.gesmescb.event.TestEventListener;


/* ***************************** Class ****************************** */

/**
 * <P>Tests the {@link InterchangeBeginOverwriter}.</P>
 *
 * @author
 * <A HREF="mailto:francois.hoeltzli@ECB.int">Fran&ccedil;ois Hoeltzli</A>
 *
 * @version @(#)$Id: InterchangeBeginOverwriterTest.java,v 1.1 2017/01/19 16:14:04 trainot Exp $
 */

public class InterchangeBeginOverwriterTest extends TestCase {

	/* ======================= Constants ======================== */


	/* ========================= Fields ========================= */

	private final InterchangeBegin ib1 = new InterchangeBegin(
		"IREF0001", "DE2", "4F0", "000315:1245", false
	);
	private final InterchangeEnd ie1 = new InterchangeEnd( "IREF0001" );
	
	private final InterchangeBegin ib2 = new InterchangeBegin(
		/* The timestamp is rotten. Should be YYMMdd:HHmm.
		 * And by the way, this should-be format needs to be corrected some day
		 * since it leads to a Y2K error (in this case there is a year error
		 * from 2027 onwards looming ahead, e.g. in GesmesTSXMLWriter.java )
		 */
		"IREF9999", "4D0", "U2", "271231:2359", true
	);
	private final InterchangeEnd ie2 = new InterchangeEnd( "IREF9999" );
	
	// Strings for Message Begin.
	private static final Object[] mbStringsOrg = {
		"MREF0001",
		"DE2",
		"4F0",
		"ECB",
		MessageBegin.STATISTICAL_DATA,
		"The subject",
		new Contact[] {
			new Contact(
				Contact.DATA_PROCESSING, "ECB-SIS", "Quasimodo",
				new ContactChannel[] {
					new ContactChannel(
						ContactChannel.EMAIL,
						"s-xtermal@ecb.int"
					)
				}
			)
		},
		"DATASET1",
		MessageBegin.DELETE_ACTION,
		"200103201423",
		"1999",
		"602",
		"ECB_BOP1"
	};
	
	private static final Object[] mbStringsExp = {
		"MREF0001",
		"DE2",
		"4F0",
		"ECB",
		MessageBegin.STATISTICAL_DATA,
		"The subject",
		new Contact[] {
			new Contact(
				Contact.DATA_PROCESSING, "ECB-SIS", "Quasimodo",
				new ContactChannel[] {
					new ContactChannel(
						ContactChannel.EMAIL,
						"s-xtermal@ecb.int"
					)
				}
			)
		},
		"DATASET1",
		MessageBegin.DELETE_ACTION,
		"202712312359",
		"1999",
		"602",
		"ECB_BOP1"
	};
	
	private final MessageEnd me = new MessageEnd( "MREF0001" );
	private MessageBegin mb1, mb2;
	
	/* ===================== Protected Methods ===================== */

	protected void setUp() {
		mb1 = new MessageBegin(
			(String) mbStringsOrg[0],
			(String) mbStringsOrg[1],
			(String) mbStringsOrg[2],
			(String) mbStringsOrg[3],
			(String) mbStringsOrg[4],
			(String) mbStringsOrg[5],
			(Contact[]) mbStringsOrg[6],
			(String) mbStringsOrg[7],
			(String) mbStringsOrg[8],
			(String) mbStringsOrg[9],
			(String) mbStringsOrg[10],
			(String) mbStringsOrg[11],
			(String) mbStringsOrg[12]
		);
		mb2 = new MessageBegin(
			(String) mbStringsExp[0],
			(String) mbStringsExp[1],
			(String) mbStringsExp[2],
			(String) mbStringsExp[3],
			(String) mbStringsExp[4],
			(String) mbStringsExp[5],
			(Contact[]) mbStringsExp[6],
			(String) mbStringsExp[7],
			(String) mbStringsExp[8],
			(String) mbStringsExp[9],
			(String) mbStringsExp[10],
			(String) mbStringsExp[11],
			(String) mbStringsExp[12]
		);
	}
	protected void tearDown() {
		System.getProperties().remove(
			"Int.ecb.gesmescb.util.InterchangeBeginOverwriter.iref"
		);
		System.getProperties().remove(
			"Int.ecb.gesmescb.util.InterchangeBeginOverwriter.sender"
		);
		System.getProperties().remove(
			"Int.ecb.gesmescb.util.InterchangeBeginOverwriter.receiver"
		);			
		System.getProperties().remove(
			"Int.ecb.gesmescb.util.InterchangeBeginOverwriter.preparation"
		);
		System.getProperties().remove(
			"Int.ecb.gesmescb.util.InterchangeBeginOverwriter.isTest"
		);	
	}

	/* ====================== Constructors ====================== */

	/**
	 * Constructs a InterchangeBeginOverwriterTest with the specified name.
	 *
	 * @param name Test case name
	 */
	public InterchangeBeginOverwriterTest ( final String name ) {
		super( name );
	}

	/* ===================== Public Methods ===================== */

	/**
	 * Assembles and returns a test suite for all
	 * the test methods of this test case.
	 *
	 * @return a non-null test suite.
	 */
	public static Test suite () {
		return new TestSuite( InterchangeBeginOverwriterTest.class );
	}

	/**
	 * Runs the test case.
	 */
	public static void main ( final String args[] ) {
		String[] testCaseName = {
			InterchangeBeginOverwriterTest.class.getName()
		};
		junit.textui.TestRunner.main( testCaseName );
//		junit.swingui.TestRunner.main( testCaseName );
	}

	/* --------------------- Test Methods --------------------- */

	/**
	 * Test the InterchangeBeginOverwriter 
	 */
	public void testOverwriteNothing () {
		final InterchangeBeginOverwriter writer = new InterchangeBeginOverwriter();
		final TestEventListener listener = new TestEventListener( 10 );
		writer.addListener( listener, true );
		
		writer.handle( ib1 );
		assertSame("interchangeBegin1", ib1, listener.getLastObject() );
		writer.handle( ie1 );
		assertSame("interchangeEnd", ie1, listener.getLastObject() );		
		assertEquals("size", 2, listener.size() );		
	}

	public void testOverwriteAsInProperties () {
		System.setProperty(
			"Int.ecb.gesmescb.util.InterchangeBeginOverwriter.iref",
			"IREF9999"
		);
		System.setProperty(
			"Int.ecb.gesmescb.util.InterchangeBeginOverwriter.sender",
			"4D0"
		);
		System.setProperty(
			"Int.ecb.gesmescb.util.InterchangeBeginOverwriter.receiver",
			"U2"
		);			
		System.setProperty(
			"Int.ecb.gesmescb.util.InterchangeBeginOverwriter.preparation",
			"271231:2359"
		);
		System.setProperty(
			"Int.ecb.gesmescb.util.InterchangeBeginOverwriter.isTest",
			"true"
		);	
		
		final InterchangeBeginOverwriter writer = new InterchangeBeginOverwriter();
		final TestEventListener listener = new TestEventListener( 10 );
		writer.addListener( listener, true );
		
		writer.handle( ib1 );
		assertEquals("interchangeBegin2", ib2, listener.getLastObject() );
		writer.handle( ie1 );
		assertEquals("interchangeEnd", ie2, listener.getLastObject() );		
		assertEquals("size", 2, listener.size() );		
	}	

	public void testOverwriteAsInPropertiesWMessageOverwrite () {
		System.setProperty(
			"Int.ecb.gesmescb.util.InterchangeBeginOverwriter.iref",
			"IREF9999"
		);
		System.setProperty(
			"Int.ecb.gesmescb.util.InterchangeBeginOverwriter.sender",
			"4D0"
		);
		System.setProperty(
			"Int.ecb.gesmescb.util.InterchangeBeginOverwriter.receiver",
			"U2"
		);			
		System.setProperty(
			"Int.ecb.gesmescb.util.InterchangeBeginOverwriter.preparation",
			"271231:2359"
		);
		System.setProperty(
			"Int.ecb.gesmescb.util.InterchangeBeginOverwriter.isTest",
			"true"
		);	
		
		final InterchangeBeginOverwriter writer = new InterchangeBeginOverwriter();
		final TestEventListener listener = new TestEventListener( 10 );
		writer.addListener( listener, true );
		
		writer.handle( ib1 );
		assertEquals("interchangeBegin2", ib2, listener.getLastObject() );
		writer.handle( mb1 );
		assertEquals("messageBegin2", mb2, listener.getLastObject() );
		writer.handle( me );
		assertSame("messageEnd", me, listener.getLastObject() );
		writer.handle( ie1 );
		assertEquals("interchangeEnd", ie2, listener.getLastObject() );		
		assertEquals("size", 4, listener.size() );		
	}	
}
