// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.util;

import java.lang.String;
import java.lang.NullPointerException;
import java.lang.UnsupportedOperationException;
import java.util.Iterator;
import java.util.NoSuchElementException;

import Int.ecb.gesmescb.TimeSeries;

/* ***************************** Class ****************************** */

/**
 * <P>Tokenizes a <CODE>String</CODE>.
 * Particularly useful for keys of {@link TimeSeries}.</P>
 *
 * <P>The main difference from the {@link java.util.StringTokenizer} is
 * that also empty tokens are reported. E.g. the key
 * ".B..E.F" becomes {"","B","","E","F"} rather than {"B","E","F"}.</P>
 *
 * <P>Also, it implements {@link Iterator}
 * in place of {java.util.Enumeration}. </P>
 *
 * @author
 * <A HREF="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</A>
 *
 * @version @(#)$Id: KeyTokenizer.java,v 1.1 2017/01/19 16:14:04 trainot Exp $
 */

public final class KeyTokenizer implements Iterator {

	/* ======================= Constants ======================== */

	/**
	 * The initial size of the internal buffer[].
	 * This should be a power of 2 and not less than 1.
	 */
	private static final char DEFAULT_SEPARATOR = TimeSeries.KEY_SEPARATOR;

	/* ========================= Fields ========================= */

	/**
	 * The private data storage.
	 */
	private final String string;

	/**
	 * The current position.
	 * Points to the start of the next token.
	 * If no more tokens available,
	 * then it equals <CODE>{@link #string}.length()+1</CODE>.
	 */
	private int position;

	/* ====================== Constructors ====================== */

	/**
	 * Construct a KeyTokenizer.
	 *
	 * @param key to tokenize.

	 * @throws NullPointerException if <CODE>key</CODE> is <CODE>null</CODE>.
	 */
	public KeyTokenizer ( final String key ) {
		if ( ( this.string = key ) == null ) {
			throw new NullPointerException(
				"KeyTokenizer refuses to tokenize null."
			);
		}
		position = 0;
	}

	/* ===================== Public Methods ===================== */

	public boolean hasNext () {
		return position <= string.length();
	}

	public String nextToken () {
		if ( hasNext() ) {
			final int next = string.indexOf( DEFAULT_SEPARATOR, position );
			String retval;
			if ( next >= 0 ) {
				retval = string.substring( position, next );
				position = next+1;
			} else {
				retval = string.substring( position );
				position = string.length() + 1;
			}
			return retval;
		} else {
			throw new NoSuchElementException(
				"String \"" + string + "\" has no more tokens."
			);
		}
	}

	public Object next () {
		return nextToken();
	}

	public void remove () {
		throw new UnsupportedOperationException(
			"Java Strings cannot be modified."
		);
	}
}
