// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.util;

import java.io.*;
import java.util.Properties;

import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;

import Int.ecb.logging.LogManager;
import Int.ecb.logging.Logger;
import Int.ecb.logging.LogRecord;
import Int.ecb.logging.Level;
import Int.ecb.logging.TestHandler;

/* ***************************** Class ****************************** */

/**
 * <P>Tests the implementation of loggingPropertiesReader class in
 * the Int.ecb.gesmescb.application package.</P>
 *
 * @author
 * <A HREF="mailto:Sami.Airo@ECB.int">Sami Airo</A>
 *
 * @version @(#)$Id: LoggingPropertiesReaderTest.java,v 1.1 2017/01/19 16:14:04 trainot Exp $
 */

public class LoggingPropertiesReaderTest extends TestCase {

	/* ========================= Fields ========================= */

	private LogManager _manager;
	private Logger _logger;
	private TestHandler _handler;
	private File propertiesFile;
	private File testLogFile;

	/* ====================== Constructors ====================== */

	/**
	 * Constructs a LoggerTest with the specified name.
	 *
	 * @param name Test case name
	 */
	public LoggingPropertiesReaderTest ( final String name ) {
		super( name );
	}

	/* ================== Protected Methods ===================== */

	/**
	 * Sets up the test fixture.
	 */
	protected void setUp () throws IOException {
		propertiesFile = File.createTempFile( "logging", "test.properties" );
		propertiesFile.deleteOnExit();
		final Writer writer = new FileWriter( propertiesFile );
		writer.write(
			".level = ALL\n"
			+ "\n"
			+ "handlers = Int.ecb.logging.FileHandler Int.ecb.logging.ConsoleHandler\n"
			+ "\n"
			+ "Int.ecb.logging.FileHandler.level = INFO\n"
			+ "Int.ecb.logging.FileHandler.pattern = testlog.log\n"
			+ "# Int.ecb.logging.FileHandler.filter = \n"
			+ "# Int.ecb.logging.FileHandler.formatter = \n"
			+ "Int.ecb.logging.ConsoleHandler.level = INFO\n"
			+ "# Int.ecb.logging.FileHandler.filter = \n"
			+ "# Int.ecb.logging.FileHandler.formatter = \n"
			+ "# logging levels for specific logger instances\n"
			+ "test1.level = ALL\n"
			+ "Int.ecb.gesmescb.application.TestProperty = Finland\n"
		);
		writer.close();

		testLogFile = File.createTempFile( "test", "log" );
		testLogFile.deleteOnExit();

		System.setProperty(
			"Int.ecb.logging.config.class",
			"Int.ecb.gesmescb.util.LoggingPropertiesReader"
		);
		System.setProperty(
			"Int.ecb.logging.config.file",
			propertiesFile.getPath()
		);
		System.setProperty(
			"Int.ecb.logging.FileHandler.pattern",
			testLogFile.getPath()
		);
		_manager = LogManager.getLogManager();
		_manager.readConfiguration();
		_logger = Logger.getLogger("Test");
		_handler = new TestHandler();
	}

	/**
	 * Tears down the test fixture.
	 */
	protected void tearDown () throws IOException {
		final Properties sysProps = System.getProperties();
		sysProps.remove( "Int.ecb.logging.config.class" );
		sysProps.remove( "Int.ecb.logging.config.file" );
		sysProps.remove( "Int.ecb.logging.FileHandler.pattern" );
		System.setProperties( sysProps );

		propertiesFile.delete();
		testLogFile.delete();

		final InputStream emptyStream = new ByteArrayInputStream( new byte[0] );
		_manager.readConfiguration( emptyStream );
		emptyStream.close();

		_manager = null;
		_logger = null;
		_handler = null;
	}

	/* ===================== Public Methods ===================== */

	/**
	 * Assembles and returns a test suite for all
	 * the test methods of this test case.
	 *
	 * @return a non-null test suite
	 */
	public static Test suite () {
		return new TestSuite( LoggingPropertiesReaderTest.class );
	}

	/**
	 * Runs the test case
	 */
	public static void main ( final String args[] ) {
		final String[] testCaseName = {
			LoggingPropertiesReaderTest.class.getName()
		};
		junit.textui.TestRunner.main( testCaseName );
	}

	/* --------------------- Test Methods --------------------- */

	/**
	 * Tests logging
	 */
	public void testLogging () {
		LogRecord r = new LogRecord( Level.INFO, "1. Testing" );

		_logger.addHandler(_handler);

		_logger.log(r);
		assertSame(r, _handler.getLastRecord());
		_handler.close();
	}

	/**
	 * Tests logging methods
	 */
	public void testProperties () {
		assertEquals(
			"Finland",
			_manager.getProperty( "Int.ecb.gesmescb.application.TestProperty" )
		);
	}

	/**
	 * Tests isLoggable method
	 */
	public void testIsLoggable () {
		_logger.setLevel( Level.INFO );
		assertTrue( _logger.isLoggable( Level.SEVERE ) );
		assertTrue( !_logger.isLoggable( Level.FINE ) );
	}
}
