// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2007, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.util;

import java.util.Date;
import java.util.List;
import java.util.ArrayList;
import java.util.Map;
import java.util.SortedMap;
import java.util.TreeMap;
import java.util.Iterator;
import java.util.NoSuchElementException;
import java.util.Collections;

import Int.ecb.gesmescb.Observation;
import Int.ecb.gesmescb.TimeSeries;

import Int.ecb.gesmescb.time.RegularInterval;
import Int.ecb.gesmescb.time.GregorianRegularInterval;

/* ***************************** Class ****************************** */

/**
 * Manages a time-ordered map of {@link Observation}s.
 * This is useful to convert loose observations into time series.
 *
 * @author
 * <a href="mailto:sis.external@ecb.int">Bernhard Bodenstorfer</A>
 *
 * @version @(#)$Id: ObservationMap.java,v 1.1 2017/01/19 16:14:04 trainot Exp $
 */
public class ObservationMap {

	/* ========================= Fields ========================= */

	/** Maps starting points of base periods to observations. */
	private final SortedMap observations;

	/** Unspecified if observations is empty; else the frequency
	 * of the base periods to which the observations relate.
	 */
	private int frequency;

	/** Caches the spanned interval. */
	private transient RegularInterval interval;

	/* ====================== Constructors ====================== */

	/** Construct an empty <code>ObservationMap</code>. */
	public ObservationMap () {
		observations = new TreeMap();
	}

	/* ===================== Public Methods ===================== */

	/** Insert all observations of the time series. */
	public void put ( final TimeSeries timeSeries ) {
		final RegularInterval interval = timeSeries.getInterval();
		for (
			final Iterator
				periods = interval.iterator(),
				observations = timeSeries.observationIterator();
			periods.hasNext() && observations.hasNext();
		) {
			final RegularInterval period = (RegularInterval) periods.next();
			final Observation observation = (Observation) observations.next();
			put( period, observation );
		}
	}

	/** Insert an observation for the specified base period. */
	public void put (
		final RegularInterval interval,
		final Observation observation
	) {
		if ( isEmpty() ) {
			frequency = interval.getFrequency();
		} else {
			if ( frequency != interval.getFrequency() ) {
				throw new IllegalArgumentException( "Cannot mix frequencies." );
			}
		}
		if ( 1 != interval.getSize() ) {
			throw new IllegalArgumentException(
				"Expecting exactly one single base period, but received "
				+ interval.getSize() + "."
			);
		}
		if ( ! observation.isEmpty() ) {
			final Date begin = interval.getBegin();
			observations.put( begin, observation );
			this.interval = isEmpty() ? interval : null;
		}
	}

	/** Retrieve an observation for the specified base period. */
	public Observation get ( final RegularInterval interval ) {
		if ( isEmpty() ) {
			return null;
		} else {
			if ( frequency != interval.getFrequency() ) {
				throw new IllegalArgumentException( "Frequency mismatch" );
			}
		}
		if ( 1 != interval.getSize() ) {
			throw new IllegalArgumentException(
				"Expecting exactly one single base period, but received "
				+ interval.getSize() + "."
			);
		}
		final Date begin = interval.getBegin();
		final Observation observation = (Observation) observations.get( begin );
		return observation;
	}

	/** Remove an observation for the specified base period. */
	public Observation remove ( final RegularInterval interval ) {
		if ( isEmpty() ) {
			return null;
		} else {
			if ( frequency != interval.getFrequency() ) {
				throw new IllegalArgumentException( "Frequency mismatch" );
			}
		}
		if ( 1 != interval.getSize() ) {
			throw new IllegalArgumentException(
				"Expecting exactly one single base period, but received "
				+ interval.getSize() + "."
			);
		}
		final Date begin = interval.getBegin();
		final Observation observation = (Observation) observations.remove(
			begin
		);
		return observation;
	}

	/** Return the interval spanned from the first to the last observation. */
	public RegularInterval getInterval () {
		if ( null == interval ) {
			if ( isEmpty() ) {
				interval = null;
			} else {
				final Date begin = (Date) observations.firstKey();
				final Date end = (Date) observations.lastKey();
				interval = new GregorianRegularInterval(
					frequency, begin,
					new GregorianRegularInterval(
						frequency, begin, end
					).getSize() + 1
				);
			}
		}
		return interval;
	}

	/** Retrieve the stored observations in chunks as time series.
	 *
	 * This avoids memory stress in the case of very sparse series,
	 * when packing into a single series can take disproportionately
	 * much memory.
	 *
	 * @param key for the time series
	 * @param maxGap how many successive empty observations are permitted
	 * before the series will be split.
	 */
	public Iterator iterateTimeSeries (
		final String key, final int maxGap
	) {
		final RegularInterval interval = getInterval();
		if ( null == interval ) return Collections.EMPTY_SET.iterator();

		return new TimeSeriesChunkIterator (
			key, frequency,
			observations.entrySet().iterator(),
			maxGap
		);
	}

	/** Return whether there are observations stored. */
	public boolean isEmpty () {
		return observations.isEmpty();
	}

	/** Return the number of (non-empty) observations stored. */
	public int size () {
		return observations.size();
	}

	/** Forget all observations. */
	public void clear () {
		observations.clear();
		interval = null;
	}
}

/* ***************************** Class ****************************** */

/** Iterates through the observation map entries and creates time series. */
class TimeSeriesChunkIterator implements Iterator {
	private final String key;
	private final int frequency;
	private final Iterator pairs;

	/** Stores the date of the first not-yet chunked observation. */
	private Date date;

	/** Stores the first not-yet chunked observation. */
	private Observation observation;

	/** How many successive empty observations are permitted
	 * before the series will be split. */
	private final int maxGap;

	/**
	 * @param key for the time series
	 * @param pairs iterates over {@link Map.Entry} objects containing
	 * a {@link Date} as key and a {@link Observation} as value.
	 */
	public TimeSeriesChunkIterator (
		final String key,
		final int frequency,
		final Iterator pairs,
		final int maxGap
	) {
		this.key = key;
		this.frequency = frequency;
		this.pairs = pairs;
		this.maxGap = maxGap;
	}

	public boolean hasNext () {
		return pairs.hasNext();
	}

	/** Returns the next {@link TimeSeries}. */
	public Object next () {
		Date
			firstDate = null,
			lastDate = null;
		final List observationsList = new ArrayList();
		while ( null != date || pairs.hasNext() ) {
			if ( null == date ) {
				final Map.Entry obsPair = (Map.Entry) pairs.next();
				date = (Date) obsPair.getKey();
				observation = (Observation) obsPair.getValue();
			}
			if ( null == firstDate ) {
				firstDate = date;
			}
			if ( null != lastDate ) {
				final RegularInterval sinceLast = new GregorianRegularInterval(
					frequency,
					lastDate, date
				);
				final int gap = sinceLast.getSize() - 1;
				if ( gap <= maxGap ) {
					for ( int i = 0; i < gap; ++i ) {
						observationsList.add( Observation.EMPTY_OBSERVATION );
					}
				} else {
					break;
				}
			}
			observationsList.add( observation );
			lastDate = date;
			date = null;
			observation = null;
		}
		if ( null == firstDate || null == lastDate ) {
			throw new NoSuchElementException(
				"No more observations in the ObservationMap."
			);
		}
		final RegularInterval
			oneTooSmallInterval = new GregorianRegularInterval(
				frequency, firstDate, lastDate
			),
			interval = new GregorianRegularInterval(
				frequency, firstDate,
				oneTooSmallInterval.getSize() + 1
			);
		final TimeSeries timeSeries = new TimeSeries(
			key, interval, observationsList
		);
		return timeSeries;
	}

	public void remove () {
		throw new UnsupportedOperationException(
			"Cannot remove chunks from an ObservationMap."
		);
	}
}
