// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2007, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.util;

import java.util.Iterator;
import java.util.NoSuchElementException;

import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.text.ParsePosition;

import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;

import Int.ecb.gesmescb.Observation;
import Int.ecb.gesmescb.TimeSeries;

import Int.ecb.gesmescb.time.RegularInterval;
import Int.ecb.gesmescb.time.GregorianRegularInterval;

/* ***************************** Class ****************************** */

/**
 * Implements a few tests for the {@link ObservationMap}.
 *
 * @author
 * <a href="mailto:sis.external@ecb.int">Bernhard Bodenstorfer</a>
 *
 * @version @(#)$Id: ObservationMapTest.java,v 1.1 2017/01/19 16:14:04 trainot Exp $
 */
public class ObservationMapTest extends TestCase {

	/* ======================= Constants ======================== */

	private static final DateFormat FORMAT
	= new SimpleDateFormat( "yyy-MM-dd" );

	private static final RegularInterval
		A_2001 = new GregorianRegularInterval(
			RegularInterval.ANNUAL,
			FORMAT.parse( "2001-01-01", new ParsePosition( 0 ) ),
			1
		),
		A_2005 = new GregorianRegularInterval(
			RegularInterval.ANNUAL,
			FORMAT.parse( "2005-01-01", new ParsePosition( 0 ) ),
			1
		),
		A_2006 = new GregorianRegularInterval(
			RegularInterval.ANNUAL,
			FORMAT.parse( "2006-01-01", new ParsePosition( 0 ) ),
			1
		),
		A_2007 = new GregorianRegularInterval(
			RegularInterval.ANNUAL,
			FORMAT.parse( "2007-01-01", new ParsePosition( 0 ) ),
			1
		),
		A_2008 = new GregorianRegularInterval(
			RegularInterval.ANNUAL,
			FORMAT.parse( "2008-01-01", new ParsePosition( 0 ) ),
			1
		),
		A_2009 = new GregorianRegularInterval(
			RegularInterval.ANNUAL,
			FORMAT.parse( "2009-01-01", new ParsePosition( 0 ) ),
			1
		),
		A_2001_2009 = new GregorianRegularInterval(
			RegularInterval.ANNUAL,
			FORMAT.parse( "2001-01-01", new ParsePosition( 0 ) ),
			9
		),
		A_2005_2006 = new GregorianRegularInterval(
			RegularInterval.ANNUAL,
			FORMAT.parse( "2005-01-01", new ParsePosition( 0 ) ),
			2
		),
		A_2005_2009 = new GregorianRegularInterval(
			RegularInterval.ANNUAL,
			FORMAT.parse( "2005-01-01", new ParsePosition( 0 ) ),
			5
		),
		A_2008_2009 = new GregorianRegularInterval(
			RegularInterval.ANNUAL,
			FORMAT.parse( "2008-01-01", new ParsePosition( 0 ) ),
			2
		),
		Q1_2001 = new GregorianRegularInterval(
			RegularInterval.QUARTERLY,
			FORMAT.parse( "2001-01-01", new ParsePosition( 0 ) ),
			1
		),
		Q4_2001 = new GregorianRegularInterval(
			RegularInterval.QUARTERLY,
			FORMAT.parse( "2001-10-01", new ParsePosition( 0 ) ),
			1
		),
		Q1_Q4_2001 = new GregorianRegularInterval(
			RegularInterval.QUARTERLY,
			FORMAT.parse( "2001-01-01", new ParsePosition( 0 ) ),
			4
		);

	private static final Observation
		OBS_1 = new Observation( "1.6", "A", "N", "" ),
		OBS_2 = new Observation( "2.6", "A", "N", "5.3" ),
		OBS_3 = new Observation( "3.6", "F", "C", "-" ),
		OBS_4 = new Observation( "-", "E", "", "" ),
		OBS_5 = new Observation( "5.6", "A", "", "-5" );

	/* ====================== Constructors ====================== */

	public ObservationMapTest ( final String name ) {
		super( name );
	}

	/* ===================== Public Methods ===================== */

	/**
	 * Assembles and returns a test suite for all
	 * the test methods of this test case.
	 *
	 * @return a non-null test suite
	 */
	public static Test suite () {
		return new TestSuite( ObservationMapTest.class );
	}

	/**
	 * @param arg the array of argument {@link String}s
	 */
	public static void main ( final String[] arg ) {
		final String[] testCaseName = {
			ObservationMapTest.class.getName()
		};
		junit.textui.TestRunner.main( testCaseName );
	}

	/* ---------------------- Test Methods ---------------------- */

	public void testPutEffects () {
		final ObservationMap map = new ObservationMap();

		assertTrue( "Initially empty", map.isEmpty() );
		assertEquals( "Initial size 0", 0, map.size() );
		assertNull( "Null interval spanned", map.getInterval() );
		{
			final Iterator iterator = map.iterateTimeSeries( "W.UR.GS", 0 );
			assertTrue(
				"Initially no series",
				! iterator.hasNext()
			);
			try {
				iterator.next();
				fail( "Really no series" );
			} catch ( final NoSuchElementException e ) {}
		}

		map.clear();

		assertTrue( "Clear has worked on empty, too.", map.isEmpty() );

		map.put( A_2008, Observation.EMPTY_OBSERVATION );

		assertTrue( "Empty is empty", map.isEmpty() );
		assertEquals( "Empty is size 0", 0, map.size() );
		assertNull( "Empty is null interval spanned", map.getInterval() );
		{
			final Iterator iterator = map.iterateTimeSeries( "W.UR.GS", 0 );
			assertTrue(
				"Empty is no series",
				! iterator.hasNext()
			);
			try {
				iterator.next();
				fail( "Empty is really no series" );
			} catch ( final NoSuchElementException e ) {}
		}

		try {
			map.put( A_2001_2009, OBS_5 );
			fail( "Multiple observations." );
		} catch ( final IllegalArgumentException e ) {}

		assertTrue( "Multi-observations must not be added.", map.isEmpty() );

		map.put( A_2006, OBS_1 );

		assertTrue( "No longer empty", ! map.isEmpty() );
		assertEquals( "New size 1", 1, map.size() );
		assertEquals( "2006 spanned", A_2006, map.getInterval() );
		{
			final Iterator iterator = map.iterateTimeSeries( "W.UR.GS", 0 );
			assertTrue( "One series", iterator.hasNext() );
			assertEquals(
				"One series is",
				new TimeSeries(
					"W.UR.GS", A_2006, new Observation[]{ OBS_1 }
				),
				iterator.next()
			);
			assertTrue(
				"No second series",
				! iterator.hasNext()
			);
			try {
				iterator.next();
				fail( "Really no second series" );
			} catch ( final NoSuchElementException e ) {}
		}

		map.clear();

		assertTrue( "Again empty", map.isEmpty() );
		assertEquals( "Again size 0", 0, map.size() );
		assertNull( "Again null interval spanned", map.getInterval() );
		assertTrue(
			"Again no series",
			! map.iterateTimeSeries( "W.UR.GS", 0 ).hasNext()
		);

		map.put( Q1_2001, OBS_2 );

		try {
			map.put( A_2008, OBS_3 );
			fail( "Mixing frequencies." );
		} catch ( final IllegalArgumentException e ) {}
		assertEquals( "Still size 1", 1, map.size() );

		map.put( Q4_2001, OBS_4 );

		assertTrue( "No longer empty", ! map.isEmpty() );
		assertEquals( "New size 2", 2, map.size() );
		assertEquals( "2001 spanned", Q1_Q4_2001, map.getInterval() );
		{
			final Iterator iterator = map.iterateTimeSeries( "F.U2.PS", 0 );
			assertTrue( "Two series", iterator.hasNext() );
			assertEquals(
				"First series is",
				new TimeSeries(
					"F.U2.PS", Q1_2001, new Observation[]{ OBS_2 }
				),
				iterator.next()
			);
			assertEquals(
				"Second series is",
				new TimeSeries(
					"F.U2.PS", Q4_2001, new Observation[]{ OBS_4 }
				),
				iterator.next()
			);
			assertTrue(
				"No third series",
				! iterator.hasNext()
			);
			try {
				iterator.next();
				fail( "Really no third series" );
			} catch ( final NoSuchElementException e ) {}
		}

		map.clear();

		assertTrue( "Yet again empty", map.isEmpty() );
		assertEquals( "Yet again size 0", 0, map.size() );
		assertNull( "Yet again null interval spanned", map.getInterval() );
		assertTrue(
			"Yet again no series",
			! map.iterateTimeSeries( "W.UR.GS", 0 ).hasNext()
		);
	}

	public void testSplitSeries () {
		final ObservationMap map = new ObservationMap();

		map.put( A_2001, OBS_1 );
		map.put( A_2005, OBS_2 );
		map.put( A_2006, OBS_3 );
		map.put( A_2008, OBS_4 );
		map.put( A_2009, OBS_5 );

		for ( int i = 0; i < 1; ++i ) {
			final Iterator iterator = map.iterateTimeSeries( "B.RM.PF", i );
			assertTrue( "Gap limit " + i + ": 3 series", iterator.hasNext() );
			assertEquals( "Gap limit " + i + ": 1 of 3",
				new TimeSeries(
					"B.RM.PF", A_2001, new Observation[]{ OBS_1 }
				),
				iterator.next()
			);
			assertEquals( "Gap limit " + i + ": 2 of 3",
				new TimeSeries(
					"B.RM.PF", A_2005_2006, new Observation[]{ OBS_2, OBS_3 }
				),
				iterator.next()
			);
			assertEquals( "Gap limit " + i + ": 3 of 3",
				new TimeSeries(
					"B.RM.PF", A_2008_2009, new Observation[]{ OBS_4, OBS_5 }
				),
				iterator.next()
			);
			assertTrue(
				"Gap limit " + i + ": No more than 3",
				! iterator.hasNext()
			);
		}

		for ( int i = 1; i < 3; ++i ) {
			final Iterator iterator = map.iterateTimeSeries( "B.RM.PF", i );
			assertTrue( "Gap limit " + i + ": 2 series", iterator.hasNext() );
			assertEquals( "Gap limit " + i + ": 1 of 2",
				new TimeSeries(
					"B.RM.PF", A_2001, new Observation[]{ OBS_1 }
				),
				iterator.next()
			);
			assertEquals( "Gap limit " + i + ": 2 of 2",
				new TimeSeries(
					"B.RM.PF", A_2005_2009, new Observation[]{
						OBS_2, OBS_3,
						Observation.EMPTY_OBSERVATION,
						OBS_4, OBS_5
					}
				),
				iterator.next()
			);
			assertTrue(
				"Gap limit " + i + ": No more than 2",
				! iterator.hasNext()
			);
		}

		for ( int i = 3; i < 10; ++i ) {
			final Iterator iterator = map.iterateTimeSeries( "B.RM.PF", i );
			assertTrue( "Gap limit " + i + ": 1 series", iterator.hasNext() );
			assertEquals( "Gap limit " + i + ": 1 of 1",
				new TimeSeries(
					"B.RM.PF", A_2001_2009, new Observation[]{
						OBS_1,
						Observation.EMPTY_OBSERVATION,
						Observation.EMPTY_OBSERVATION,
						Observation.EMPTY_OBSERVATION,
						OBS_2, OBS_3,
						Observation.EMPTY_OBSERVATION,
						OBS_4, OBS_5
					}
				),
				iterator.next()
			);
			assertTrue(
				"Gap limit " + i + ": No more than 1",
				! iterator.hasNext()
			);
		}

		map.clear();

		assertTrue( "Yet again empty", map.isEmpty() );
		assertEquals( "Yet again size 0", 0, map.size() );
		assertNull( "Yet again null interval spanned", map.getInterval() );
		assertTrue(
			"Yet again no series",
			! map.iterateTimeSeries( "W.UR.GS", 0 ).hasNext()
		);
	}

	public void testPutSeriesAndGet () {
		final TimeSeries series = new TimeSeries(
			"FR.K.SL", "20072014", "702", new Observation[]{
				new Observation( Observation.MISSING_VALUE, "S", "", "" ),
				Observation.EMPTY_OBSERVATION,
				new Observation( "81.9", "A", "F", Observation.MISSING_VALUE ),
				new Observation( "1.7", "A", "", "" ),
				new Observation( "1.7", "A", "", "4.4" ),
				Observation.EMPTY_OBSERVATION,
				Observation.EMPTY_OBSERVATION,
				new Observation( Observation.MISSING_VALUE, "S", "", "" )
			}
		);
		final ObservationMap map = new ObservationMap();
		map.put( series );
		final Iterator iterator = map.iterateTimeSeries( "FR.K.SL", 10 );
		assertEquals( "Exact copy", series, iterator.next() );
		assertTrue( "No more", ! iterator.hasNext() );

		assertNull( "Nothing before", map.get( A_2006 ) );
		assertEquals(
			"First",
			new Observation( Observation.MISSING_VALUE, "S", "", "" ),
			map.get( A_2007 )
		);
		assertEquals(
			"First",
			new Observation( Observation.MISSING_VALUE, "S", "", "" ),
			map.get( A_2007 )
		);
		assertNull(
			"Gap",
			map.get( A_2008 )
		);
		assertEquals(
			"Middle",
			new Observation( "81.9", "A", "F", Observation.MISSING_VALUE ),
			map.remove( A_2009 )
		);
		assertNull( "Middle removed", map.get( A_2009 ) );
		assertNull( "Middle removed", map.remove( A_2009 ) );
	}
}
