// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2007, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.util;

import java.util.ArrayList;
import java.util.Iterator;
import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;

import Int.ecb.logging.Level;

import Int.ecb.gesmescb.*;

import Int.ecb.gesmescb.event.GESMESCBEventBroker;

/* **************************** Class ***************************** */

/**
 * <P>Changes the key of the {@link TimeSeries} and {@link Attribute}s  
 * according to an ordered set of rules configured in the preferences.</P>
 *
 * @author
 * <A HREF="mailto:francois.hoeltzli@ECB.int">Fran&ccedil;ois Hoeltzli</A>
 *
 * @version @(#)$Id: RuleBasedMapper.java,v 1.1 2017/01/19 16:14:04 trainot Exp $
 */
public class RuleBasedMapper extends GESMESCBEventBroker {

	/* ========================= Fields ========================= */
	
	/** List of SimpleRuleBasedMappers to go through to obtain a full mapping. */
	private final ArrayList mappers;
	
	/** Whether to dispatch the unmapped name of the series along with
	 * the mapped series. */
	private boolean dispatchOriginalName;
	
	/** Name of the attribute which would store the original name. */ 
	private String originalNameAttribute; 

	/** Prefix the original name, eg with the input DSI*/
	private String originalNamePrefix; 

	/** Whether to dispatch when TimeSeries and Attributes when the mapping
	 * did not remap all the dimensions. */
	private boolean dispatchWhenIncompleteMapping;
	
	/** How to build the dispatch key from the mapped key so that the dimensions
	 * appear in the order expected in the key family.
	 */
	private final String reorderingKey;
	
	/* ====================== Constructors ====================== */
	
	/**
	 * Construct a new <code>RuleBasedMapper</code> without parameters.
	 */
	public RuleBasedMapper () {
		this("");
	}

	/**
	 * Construct a new <code>RuleBasedMapper</code>.
	 *
	 * @param name name of mapper.
	 */
	public RuleBasedMapper ( final String name ) {
		
		super( name, true );
		mappers = new ArrayList();
	
		String filename = componentHelper.getPreference("filename", null);		
		if (filename != null) {
			getMappersFromFile(filename);
		} else {
			componentHelper.logger.log(
    			Level.WARNING, componentHelper.logName, "RuleBasedMapper",
    			"No filename specified in the preferences.\n"
    		);		    		
		}
		
		/* Read from properties whether to dispach a (e.g.) DOM_SER_IDS
		 * attribute (containing the original key) along the mapped series.*/
		originalNameAttribute = componentHelper.getPreference(
			"originalNameAttribute", ""
		);		
		originalNamePrefix = componentHelper.getPreference(
			"originalNamePrefix", ""
		);		
		if ( ! "".equals( originalNameAttribute ) ) {
			dispatchOriginalName = true;
			componentHelper.logger.log(
    			Level.CONFIG, componentHelper.logName, "RuleBasedMapper",
    			"Original series name (prefixed with \"" + originalNamePrefix
				+ "\") will be stored in attribute \""
    			+ originalNameAttribute + "\".\n"
    		);		    		
		} else {
			dispatchOriginalName = false;
		}
		
		dispatchWhenIncompleteMapping = componentHelper
			.getPreference("dispatchWhenIncompleteMapping", "true")
			.equalsIgnoreCase("true");
		
		reorderingKey = componentHelper.getPreference("reordering", null); 
		componentHelper.logger.log(
			Level.CONFIG, componentHelper.logName, "handle",
			(null == reorderingKey)
				? "No reordering parameter found. "
				: "Reordering key is \"" +  reorderingKey + "\"",
			reorderingKey
		);
	}
	
	/* ==================== Private Methods ===================== */
	
	/**
	 * Read the rule file passed as parameter.
	 * 
	 * @param filename The file to read the rules from.
	 */
	private void getMappersFromFile(final String filename) {
		try {
		    BufferedReader in = new BufferedReader(new FileReader(filename));
		    String line;
		    SimpleRuleBasedMapper mapper = null;
		    while ((line = in.readLine()) != null) {
		    	mapper = handleRuleLine(line, mapper, filename);
		    }
		    in.close();
		} catch (IOException e) {
			componentHelper.logger.log(
    			Level.SEVERE, componentHelper.logName, "RuleBasedMapper",
    			"Could not open filename \"" + filename + "\".\n"
    		);		    		
		}
	}

	/**
	 * This function interprets the text controlling the creation of the 
	 * partial mappers constituting the RuleBasedMapper.
	 * 
	 * @param line: text to be interpreted (add a rule/add a mapper etc.)
	 * @param mapper: SimpleRuleBasedMapper currently used
	 * @param filename: name of the rule file. For error report
	 * @return a SimpleRuleBasedMapper
	 */
	private SimpleRuleBasedMapper handleRuleLine(
		final String line,
		SimpleRuleBasedMapper mapper,
		final String filename
	) {
		if (0 >= line.length())
			return mapper;

    	switch ( line.charAt(0) ) {
    	case '{':
    		mapper = new SimpleRuleBasedMapper();
    		break ;
    	case '}':
    		if( mapper != null ) {
    			mappers.add(mapper);
    			mapper = null;
    		} else {
    			componentHelper.logger.log(
    				Level.SEVERE, componentHelper.logName, "RuleBasedMapper",
    				"Unmatched closing parenthesis in rule file \"" + filename
    				+ "\" in line \"" + line + "\"\n"
    			);
    		}
    		break ;
		case '#': //ignore comments
			break;
    	default:
    		if ( mapper != null ) {
    			mapper.addRule( line );
    		} else if ( "swap ".equals( line.substring(0,5) )) {
    			mapper = new SimpleRuleBasedMapper();
    			mapper.addRule( line );
    			mappers.add(mapper);    
    			mapper = null;
    		} else {
    			componentHelper.logger.log(
	    			Level.SEVERE, componentHelper.logName, "RuleBasedMapper",
	    			"loose rule found, where is the opening parenthesis? "
	    			+"in rule file \"" + filename
	    			+ "\" in line \"" + line + "\"\n"
	    		);		    		
    		}
    		break;
    	}
    	return mapper;
    }

	/* ===================== Public Methods ===================== */

	/*
	 * Dispatch a TimeSeries with a remapped key, and according
	 * to <code>originalSeriesName</code>, dispatch an attribute
	 * as well containing the original name.
	 * 
	 * @params timeSeries the TimeSeries to be mapped.
	 */
	public void handle ( final TimeSeries timeSeries ) {	
		componentHelper.logger.log(
			Level.FINE, componentHelper.logName, "handle",
			"handling timeSeries", timeSeries
		);
		String key = timeSeries.getKey();
		ArrayList status = SimpleRuleBasedMapper.createStatus( key, SimpleRuleBasedMapper.ORIGINAL ); 
		String mappedKey = remap( key, status );
		if ( reorderingKey != null )
			mappedKey = reorder( mappedKey );
		
		boolean fullyRemapped = SimpleRuleBasedMapper.fullyRemapped( status );
		
		boolean partiallyRemapped = SimpleRuleBasedMapper.partiallyRemapped( status );
		
		if ( ! fullyRemapped ) {
			componentHelper.logger.log(
				Level.WARNING, componentHelper.logName, "handle",
				"could not fully remap timeSeries " + key, timeSeries
			);
		}
		if ( fullyRemapped || (dispatchWhenIncompleteMapping && partiallyRemapped)) {
			TimeSeries mappedSeries = new TimeSeries(
				mappedKey,
				timeSeries.getPeriod(),
				timeSeries.getPeriodFormat(),
				timeSeries.observationIterator()
			);
			dispatch( mappedSeries );
			if ( dispatchOriginalName ) {
				Attribute originalSeriesName = new Attribute	(
					mappedKey,
					"",
					"",
					originalNameAttribute,
					"".equals( originalNamePrefix )
						? key
						: originalNamePrefix + timeSeries.KEY_SEPARATOR + key
				);	
				dispatch( originalSeriesName );
			}
		} else {
			componentHelper.logger.log(
				Level.WARNING, componentHelper.logName, "handle",
				"Time Series \"" + key + "\" ignored.", timeSeries
			);
		}
	}

	/*
	 * Dispatch an attribute with a remapped key.
	 * 
	 * @params attribute the Attribute to be mapped.
	 */
	public void handle( final Attribute attribute ) {	
		componentHelper.logger.log(
			Level.FINE, componentHelper.logName, "handle",
			"handling attribute", attribute
		);
		String key = attribute.getKey();
		ArrayList status = SimpleRuleBasedMapper.createStatus( key, SimpleRuleBasedMapper.ORIGINAL ); 
		String mappedKey = remap( key, status );
		if ( reorderingKey != null )
			mappedKey = reorder( mappedKey );

		boolean fullyRemapped = SimpleRuleBasedMapper.fullyRemapped( status );
		
		if ( ! fullyRemapped ) {
			componentHelper.logger.log(
				Level.WARNING, componentHelper.logName, "handle",
				"could not fully remap attribute " + key, attribute
			);
		}
		if ( fullyRemapped || dispatchWhenIncompleteMapping ) {
			Attribute mappedAttribute = new Attribute(
				mappedKey,
				attribute.getPeriod(),
				attribute.getPeriodFormat(),
				attribute.getName(),
				attribute.getValue()
			);
			dispatch( mappedAttribute );
		} else {
			componentHelper.logger.log(
				Level.WARNING, componentHelper.logName, "handle",
				"Attribute \"" + key + "\" ignored.", attribute
			);
		}
	}
		
	
	/**
	 * Do some string manipulation on a key according to the 
	 * SimpleRuleBasedMappers contained in field mappers.
	 * Compute the status as well (Which dimensions have been changed).
	 * 
	 * @param key Textual representation of the key to be remapped
	 * @param status Indicate for each dimension whether it has been replaced 
	 * @return the remapped key; and as a side effect, modify status.
	 */
	public String remap( final String key, ArrayList status ) {
		String mappedKey = key;
		Iterator itMappers = mappers.iterator();
		while ( itMappers.hasNext() ) {
			SimpleRuleBasedMapper mapper
			= (SimpleRuleBasedMapper) itMappers.next();
			mappedKey = mapper.remap( mappedKey, status);
		}
		return mappedKey;
	}	

	/**
	 * Do reorder the dimensions of the key according to the 
	 * reordering property.
	 * The interpretation of the property (eg: 3.4.2.1) converts the original
	 * key a.b.c.d to c.d.b.a (the third element of the property indicates
	 * where the third element of the reordered key comes from).
	 * 
	 * Note that the status array is not affected by the reordering!
	 * 
	 * @param key Textual representation of the key to be reordered
	 * @return the reordered key.
	 * 
	 * @deprecated Reordering should be done in the SimpleRuleBasedMapper.java
	 */
	public String reorder( final String key ) {

		if ( null == reorderingKey ) 
			return key;
		
		String reorderedKey=""; 
		String[] elements = key.split( "[" + TimeSeries.KEY_SEPARATOR + "]");
		KeyTokenizer swapTokenizer = new KeyTokenizer( reorderingKey );
		
		for( ; swapTokenizer.hasNext(); ) {
			int index = Integer.parseInt(swapTokenizer.nextToken());
			reorderedKey += elements[ index - 1 ]; // java uses zero based arrays
			if ( swapTokenizer.hasNext() ) {
				reorderedKey += TimeSeries.KEY_SEPARATOR;
			}			
		}
		return reorderedKey;
	}	

	
	/**
	 * This function wraps private handleRuleLine for testing purposes.
	 * 
	 * @param line: text to be interpreted (add a rule/add a mapper etc.)
	 * @param mapper: SimpleRuleBasedMapper currently used
	 * @param filename: name of the rule file. For error report
	 * @return a SimpleRuleBasedMapper
	 */
	public SimpleRuleBasedMapper addLine(
		final String line,
		SimpleRuleBasedMapper mapper,
		final String filename
	) {
		return handleRuleLine( line, mapper, filename);
	}

}
	

