// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2007, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.util;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.Stack;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.text.StringCharacterIterator;

import Int.ecb.logging.Level;

import Int.ecb.gesmescb.*;

import Int.ecb.gesmescb.event.GESMESCBEventBroker;

/* **************************** Class ***************************** */

/**
 * <P>Changes the key of the {@link TimeSeries} and {@link Attribute}s  
 * according to an ordered set of rules configured in the preferences.</P>
 *
 * @author
 * <A HREF="mailto:francois.hoeltzli@ECB.int">Fran&ccedil;ois Hoeltzli</A>
 *
 * @version @(#)$Id: SimpleRuleBasedMapper.java,v 1.1 2017/01/19 16:14:04 trainot Exp $
 */

public final class SimpleRuleBasedMapper extends GESMESCBEventBroker {

	/* ======================= Constants ======================== */
	public static final Integer ORIGINAL = new Integer(0);
	public static final Integer REPLACED = new Integer(1);
	
	/* ========================= Fields ========================= */

	// Whether to dispatch or not the things that are not completely mapped
	// When some not fully mappable objects comes and allowPartialMap is not set,
	// a WARNING is issued.
	private final boolean allowPartialMap;

	// Exit the mapper as soon as one rule fired. 
	private final boolean fastExit;

	// Replacement rules are stored in two parallel lists,
	// same index refers to the same rule
	private final ArrayList replacementPatterns;  // compiled regexp of the rule
	private final ArrayList replacementReplacers; // right hand side of the rule
	
	// swapRule not final as we can assign via the addRule
	private String swapRule;
	
	/* ====================== Constructors ====================== */

	public SimpleRuleBasedMapper () {
		this( "" );
	}

	/**
	 * Constructs a SimpleRuleBasedMapper.
	 */
	public SimpleRuleBasedMapper ( final String name ) {
		super( name, true );
		allowPartialMap = componentHelper.getPreference(
			"allowPartialMap", "false"
		).equalsIgnoreCase( "true" );
		fastExit = componentHelper.getPreference(
			"fastExit", "false"
		).equalsIgnoreCase( "true" );
		replacementPatterns = new ArrayList();
		replacementReplacers = new ArrayList();
		swapRule = componentHelper.getPreference("swapRule", null);
		if ( null != swapRule ) {
			componentHelper.logger.log(
				Level.CONFIG, componentHelper.logName,
				"SimpleRuleBasedMapper",
				"swapRule is " + swapRule
			);
		}
		int i=0; 
		String rule;
		do {
			rule = componentHelper.getPreference("rule." + i, null);
			if ( null != rule ) {
				addRule(rule);				
				componentHelper.logger.log(
					Level.CONFIG, componentHelper.logName,
					"SimpleRuleBasedMapper",
					"rule " + i + " is " + rule
				);
			}
			++i;
		} while ( rule != null );
	}

	/* ===================== Public Methods ===================== */

	/**
	 * Dispatches the interchangeBegin as reconfigured.
	 *
	 * @param interchangeBegin coming in.
	 */
	public void handle ( final TimeSeries timeSeries ) {	
		componentHelper.logger.log(
			Level.FINE, componentHelper.logName, "handle",
			"handling timeSeries", timeSeries
		);
		String key = timeSeries.getKey();
		ArrayList status = createStatus( key, ORIGINAL ); 
		String mappedKey = remap( key, status );
		if ( allowPartialMap || fullyRemapped( status ) ) {
			TimeSeries mappedSeries = new TimeSeries(
				mappedKey,
				timeSeries.getPeriod(),
				timeSeries.getPeriodFormat(),
				timeSeries.observationIterator()
			);
			dispatch( mappedSeries );
		} else {
			componentHelper.logger.log(
				Level.WARNING, componentHelper.logName, "handle",
				"could not fully remap timeSeries " + key, timeSeries
			);
		}
	}

	public void handle( final Attribute attribute ) {	
		componentHelper.logger.log(
			Level.FINE, componentHelper.logName, "handle",
			"handling attribute", attribute
		);
		String key = attribute.getKey();
		ArrayList status = createStatus( key, ORIGINAL ); 
		String mappedKey = remap( key, status );
		if ( allowPartialMap || fullyRemapped( status ) ) {
			Attribute mappedAttribute = new Attribute(
				mappedKey,
				attribute.getPeriod(),
				attribute.getPeriodFormat(),
				attribute.getName(),
				attribute.getValue()
			);
			dispatch( mappedAttribute );
		} else {
			componentHelper.logger.log(
				Level.WARNING, componentHelper.logName, "handle",
				"could not fully remap attribute " + key, attribute
			);
		}
	}
	
	/**
	 * Returns the remapping of the key passed in
	 * and modifies the status accordingly.
	 * 
	 * Status is an Array giving for each dimension either ORIGINAL or REPLACED,
	 * depending on the origin of the current value of the dimension considered.
	 *
	 * @param key: The key to be remapped.
	 * @param status: The array containing the status for each dimension.
	 * 
	 */
	public String remap( final String key, final ArrayList status ) {
		String mappedKey = key;
		Iterator itPatterns = replacementPatterns.iterator();
		Iterator itReplacers = replacementReplacers.iterator();
		if ( swapRule != null ) {
			mappedKey = swap( key, status );
		}
		while ( itPatterns.hasNext() ) {
			final Pattern pattern = (Pattern) itPatterns.next();
			final String replacer = (String) itReplacers.next();
			final Matcher matcher = pattern.matcher(mappedKey);
			if ( matcher.matches() ) {
				mappedKey
				= applyRuleToString( matcher, replacer, mappedKey, status );
				if ( fastExit ) {
					return mappedKey;
				}
			}
		}
				
		return mappedKey;
	}

	/**
	 * Add a rule to the SimpleRuleBasedMapper. Adding a rule means adding a
	 * pattern to match for the rule to apply and adding a replacer String
	 * to replacementPatterns and replacementReplacers respectively.
	 * 
	 * @param rule: A string containing the rule. A rule has on the left hand side
	 * a pattern to be matched. The pattern falls into 3 parts: the left context,
	 * left from the first opening parenthesis, the replacee, between the first pair
	 * of parenthesis, and the right context, on right from the first closing
	 * parenthesis. A rule has also a right hand side, right from the first '->'
	 * digramme, specifying the replacer (what to replace the replacee with).
	 * 
	 */
	public void addRule ( String rule ) {
		if ( '#' == rule.charAt(0) ) return;
		if ( "swap ".equals(rule.substring(0, 5)) ) {
			swapRule = rule.substring(5);
			componentHelper.logger.log(
				Level.FINE, componentHelper.logName, "addRule",
				"Reordering Rule \"" + swapRule + "\" will be used"
			);
			return;
		}
		// rules look like:
		// leftContextPattern(replacee)rightContextPattern->replacement
		// or .... ->()  indicating replacement to self (toggle status to 1)
		/* what's left from the first parenthesis is only context and won't change*/
		Pattern rulePattern = Pattern.compile("([^()>-]*)\\(([^()>-]*)\\)([^()>-]*)->(.*)");
		Matcher ruleMatcher = rulePattern.matcher(rule);
		if (! ruleMatcher.matches() ) {
			componentHelper.logger.log(
				Level.SEVERE, componentHelper.logName, "addRule",
				"Failed to parse rule \"" + rule + "\". Should be of the form A(B)C->D."
				+ " Rule will be ignored."
			);
		} else {
			/* what's before the replacee is context*/
			String left_context = makePattern( ruleMatcher.group(1) );
			/* replacee is the thing between the first set of brackets*/
			String replacee = makePattern( ruleMatcher.group(2) );
			/* what's after the replacee and before the arrow is context*/
			String right_context = makePattern( ruleMatcher.group(3) );
			/* what's after the arrow is replacementString.
			 * Note that we replace to Strings, not to Patterns!*/
			String replacementString = ruleMatcher.group(4);
			Pattern pattern = Pattern.compile(
				'^' + left_context + replacee + right_context + '$'
			);				
			replacementPatterns.add(pattern);
			replacementReplacers.add(replacementString);
			componentHelper.logger.log(
				Level.FINE, componentHelper.logName, "addRule",
				"Rule \"" + rule + "\" parsed successfully as "
				+ left_context + ", " + replacee + ", " + right_context + ", "
				+ replacementString
			);
		}
	}
	
	public static ArrayList createStatus( String key, Integer value ) {
		ArrayList status = new ArrayList();
		if ( 0 == key.length() ) return status ;
		status.add( value );
		for( StringCharacterIterator it=new StringCharacterIterator(key)
			; StringCharacterIterator.DONE != it.current()
			; it.next()
		) {
			if ( TimeSeries.KEY_SEPARATOR == it.current() )
				status.add( value );
		}
		return status;
	}
	
	public static boolean fullyRemapped( final ArrayList status ) {
		for( Iterator it = status.iterator(); it.hasNext(); ) {
			if( ORIGINAL == (Integer) it.next() ) {
				return false;
			}
		}
		return true;
	}
	
	public static boolean partiallyRemapped( final ArrayList status ) {
		for( Iterator it = status.iterator(); it.hasNext(); ) {
			if( ORIGINAL != (Integer) it.next() ) {
				return true;
			}
		}
		return false;
	}
	
	/* ==================== Private Methods ===================== */

	private String applyRuleToString(
		final Matcher matcher,
		final String replacer,
		final String key,
		final ArrayList status
	){
		String mapped_key = key;
		String leftContextMatch = matcher.group(1);
		String replaceeMatch = matcher.group(2);
		String rightContextMatch = matcher.group(3);
		// special rule A(B)C->() that sets status for B to REPLACED
		String replacement = "()".equals( replacer )
			? replaceeMatch
			: replacer;
		mapped_key = leftContextMatch + replacement + rightContextMatch;
		computeStatus(
			status,
			leftContextMatch,
			//replaceeMatch,
			rightContextMatch,
			replacement
		);
		return mapped_key;
	}
	
	private String makePattern( final String str){
		// stars in the rules set mean "any characters in [A-Z0-9_]"
		String pattern = str.replaceAll( "[*]", "[\\\\w]*");
		// dots in the rules set mean "one dot character and nothing else"
		pattern = pattern.replaceAll("[.]", "[.]");
		// group for later retrieval of the substring matching this pattern
		return "(" + pattern + ")"; 
	}
		
	public static void concatStatus( /* made public for testing */
		final String leftString,
		final ArrayList leftStatus, // the contents get modified intentionally
		final String rightString,
		final ArrayList rightStatus // here as well but we should not care
	) {
		if ( 0 == leftString.length() ) {
			if ( (0 != leftStatus.size()) && (0 != rightStatus.size()) ) {
				if ( TimeSeries.KEY_SEPARATOR == rightString.charAt(0) )
					rightStatus.set( 0, SimpleRuleBasedMapper.REPLACED );
			}
			leftStatus.clear();
			leftStatus.addAll(rightStatus);
			return;
		}
		if ( 0 == rightString.length() ) {
			return;
		}
		if ( TimeSeries.KEY_SEPARATOR
			== leftString.charAt( leftString.length() - 1 )
		) {
			leftStatus.remove( leftStatus.size() - 1 ); 
			if ( TimeSeries.KEY_SEPARATOR
				== rightString.charAt( 0 ) ) {
				rightStatus.remove( 0 );
				leftStatus.add( SimpleRuleBasedMapper.REPLACED ); 
			}
		} else {
			rightStatus.remove(0);
			if ( TimeSeries.KEY_SEPARATOR
				!= rightString.charAt(0) ) {
				leftStatus.remove( leftStatus.size() - 1 );
				leftStatus.add( SimpleRuleBasedMapper.REPLACED ); 
			}
		}
		leftStatus.addAll( rightStatus );
		return;
	}

	public static void computeStatus( /* made public for testing */
		final ArrayList status, // gets modified intentionally
		final String leftContext,
		//final String replacee,
		final String rightContext,
		final String replacement
	) {
		//	store the status from the left context
		ArrayList leftStatus = new ArrayList();
		int numberOfDots = leftContext.replaceAll(
			"[^" + TimeSeries.KEY_SEPARATOR + "]", ""
		).length();
		if ( 0 == leftContext.length() ) 
			--numberOfDots; // empty leftStatus in that case.
		for( int i = 0; i <= numberOfDots ; i++ )
			leftStatus.add( status.get( i ) );
		//	store the status from the right context
		ArrayList rightStatus = new ArrayList();
		numberOfDots = 1 + rightContext.replaceAll(
			"[^" + TimeSeries.KEY_SEPARATOR + "]", ""
		).length();
		for( int i = numberOfDots ; i >= 1 ; i-- )
			rightStatus.add( status.get( status.size() - i ) );
		//	store the replacement status
		ArrayList replacementStatus = new ArrayList();
		numberOfDots = replacement.replaceAll(
			"[^" + TimeSeries.KEY_SEPARATOR + "]", ""
		).length();
		for( int i=0; i <= numberOfDots; i++ ) 
			replacementStatus.add( SimpleRuleBasedMapper.REPLACED );
		//	append the right part
		status.clear();
		status.addAll(leftStatus);
		SimpleRuleBasedMapper.concatStatus(
			leftContext, status,
			replacement, replacementStatus
		);
		SimpleRuleBasedMapper.concatStatus(
			leftContext + replacement, status, rightContext, rightStatus
		);
	}
	
	private String swap(final String key, ArrayList status) {
	/**
	 * Do reorder the dimensions of the key according to the 
	 * reordering property.
	 * The interpretation of the property (eg: 3.4.2.1) converts the original
	 * key a.b.c.d to c.d.b.a (the third element of the property indicates
	 * where the third element of the reordered key comes from).
	 * 
	 * Note that the status array is affected by the reordering!
	 * 
	 * @param key Textual representation of the key to be reordered
	 * @return the reordered key.
	 */
		if ( null == swapRule ) 
			return key;
			
		String reorderedKey=""; 
		ArrayList newStatus=new ArrayList();
		String[] elements = key.split( "[" + TimeSeries.KEY_SEPARATOR + "]");
		String[] swaps = swapRule.split( "," );
		for( int i=0; i<swaps.length; i++ ) {
			int index = Integer.parseInt( swaps[i] );
			reorderedKey += elements[ index - 1 ]; // java uses zero based arrays
			newStatus.add( status.get( index - 1 ) );
			if ( i+1 < swaps.length ) {
				reorderedKey += TimeSeries.KEY_SEPARATOR;
			}			
		}
		if (! newStatus.isEmpty()) 
			status = newStatus;
		return reorderedKey;	
	}
}
