// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2003, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.util;

import java.awt.TextArea;
import java.awt.Color;

import Int.ecb.logging.LogRecord;
import Int.ecb.logging.Level;
import Int.ecb.logging.Formatter;
import Int.ecb.logging.Handler;

/**
 * {@link java.awt} based logging Handler.
 *
 * {@link LogRecord}s are published to a given {@link java.awt.TextArea}.
 *
 * @author
 * <a href="mailto:Paula.Rocha@ECB.int">Paula Rocha</a>
 * @author
 * <a href="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</a>
 *
 * @version @(#)$Id: TextAreaHandler.java,v 1.1 2017/01/19 16:14:04 trainot Exp $
 */
public class TextAreaHandler extends Handler {

	private final TextArea textArea;
	private boolean isFirstRecord;

	/**
	 * Create a TextAreaHandler with a given text area and a given formatter.
	 */
	public TextAreaHandler ( final TextArea textArea, final Formatter formatter ) {
		super( formatter );
		this.textArea = textArea;
		isFirstRecord = true;
	}

	/**
	 * Format and publish a log record on the text area.
	 * The TextAreaHandler first checks if the given
	 * LogRecord has at least the required log level.
	 * If so, it calls its Formatter to format the record and then
	 * writes the result to the text area. If this is the first
	 * LogRecord to be written, the formatter's "head" string
	 * is written to the stream before the LogRecord is written.
	 */
	public void publish ( final LogRecord record ) {
		if ( isLoggable( record ) ) {
			if ( isFirstRecord ) {
				textArea.append( getFormatter().getHead( this ) );
				isFirstRecord = false;
			}
			textArea.append( getFormatter().format( record ) );
			final Color presentColor = textArea.getBackground();
			if ( record.getLevel().intValue() < Level.WARNING.intValue() ) {
				if (
					!presentColor.equals( Color.red )
					&&
					!presentColor.equals( Color.yellow )
				) {
					textArea.setBackground( Color.green );
				}
			} else if ( record.getLevel().intValue() < Level.SEVERE.intValue() ) {
				if ( !presentColor.equals( Color.red ) ) {
					textArea.setBackground( Color.yellow );
				}
			} else {
				textArea.setBackground( Color.red );
			}
		}
	}

	/**
	 * Finishes the log output with the formatter's "tail" string.
	 */
	public void close () {
		if ( !isFirstRecord ) {
			textArea.append( getFormatter().getTail( this ) );
		}
	}
}
