// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.util;

import java.util.Iterator;
import java.util.TreeMap;
import java.util.Map;
import java.util.SortedMap;
import java.util.ArrayList;
import java.util.Vector;

import Int.ecb.logging.Level;

import Int.ecb.gesmescb.*;
import Int.ecb.gesmescb.time.*;

import Int.ecb.gesmescb.event.GESMESCBEventBroker;

/* **************************** Class ***************************** */

/**
 * <P>Joins successive {@link TimeSeries} with the same key.</P>
 *
 * @author
 * <A HREF="mailto:Sami.Airo@ECB.int">Sami Airo</A>
 *
 * @version @(#)$Id: TimeSeriesJoiner.java,v 1.1 2017/01/19 16:14:04 trainot Exp $
 */

public final class TimeSeriesJoiner extends GESMESCBEventBroker {

	/* ======================= Constants ======================== */

	/* ========================= Fields ========================= */

	/**
	 * TimeSeries waiting to be joined
	 */
	// Cache must preserve the reception sequence, not invert it! fh. 
	private Vector cachedTimeSeries = new Vector();

	/**
	 * Whether the current one is a delete message.
	 */
	private boolean inUpdateMessage = false;

	private int maxGapBetweenObservations = 5;

	/* ====================== Constructors ====================== */

	public TimeSeriesJoiner () {
		this( "" );
	}

	/**
	 * Constructs a TimeSeriesJoiner.
	 */
	public TimeSeriesJoiner ( final String name ) {
		super( name, true );

		String gaps = componentHelper.getPreference("maxGapBetweenObservations", null);
		if (gaps == null) return;

		try {
			maxGapBetweenObservations = Integer.parseInt(gaps);
			componentHelper.logger.log(
				Level.CONFIG, componentHelper.logName, "TimeSeriesJoiner",
				"maxGapBetweenObservations set to " + maxGapBetweenObservations
			);
		} catch (NumberFormatException e) {
			componentHelper.logger.log(
				Level.SEVERE, componentHelper.logName, "TimeSeriesJoiner",
				"Invalid preference value for maxGapBetweenObservations: " + gaps
			);
			throw new NullPointerException(e.getMessage());
		}
	}

	/* ===================== Public Methods ===================== */

	/**
	 * Finds out whether the current message is a delete message.
	 *
	 * @param messageBegin coming in.
	 */
	public void handle ( final MessageBegin messageBegin ) {
		componentHelper.logger.log(
			Level.FINE, componentHelper.logName, "handle",
			"handling MessageBegin", messageBegin
		);
		
		inUpdateMessage = messageBegin.isUpdateOrReplaceAction();
		if ( inUpdateMessage ) {
			componentHelper.logger.log(
				Level.FINE, componentHelper.logName, "handle",
				"The message is a update/replace message", messageBegin
			);
		}

		dispatchCache();
		dispatch( messageBegin );
	}

	/**
	 * Tries to cache and join the series.
	 *
	 * @param timeSeries coming in.
	 */
	public void handle ( final TimeSeries timeSeries ) {
		componentHelper.logger.log(
			Level.FINE, componentHelper.logName, "handle",
			"handling TimeSeries", timeSeries
		);

		if ( !inUpdateMessage ) {	// TODO: refine deletion handling.
			componentHelper.logger.log(
				Level.FINER, componentHelper.logName, "handle",
				"Simply forwarding TimeSeries in other than update/replace messages"
			);
			dispatchCache();
			dispatch( timeSeries );
			return;
		}
		
		if ( cachedTimeSeries.isEmpty() ) {
			componentHelper.logger.log(
				Level.FINER, componentHelper.logName, "handle",
				"Setting initial TimeSeries"
			);
			cachedTimeSeries.addElement( timeSeries );
			return;
		}
		
		if ( ((TimeSeries)cachedTimeSeries.lastElement()).getKey().equals( timeSeries.getKey() ) ) {
			componentHelper.logger.log(
				Level.FINER, componentHelper.logName, "handle",
				"Joining TimeSeries"
			);
			cachedTimeSeries.addElement( timeSeries );
			return;
		}
		
		componentHelper.logger.log(
			Level.FINER, componentHelper.logName, "handle",
			"New TimeSeries key dispatching TimeSeries"
		);
		dispatchCache();
		cachedTimeSeries.addElement( timeSeries );
	}


	/**
	 * Dispatches the time series from cache if necessary.
	 *
	 * @param attribute coming in.
	 */
	public void handle ( final Attribute attribute ) {
		componentHelper.logger.log(
			Level.FINEST, componentHelper.logName, "handle",
			"handling Attribute", attribute
		);
		dispatchCache();
		dispatch( attribute );
	}


	/**
	 * Dispatches the time series from cache if necessary.
	 *
	 * @param messageEnd coming in.
	 */
	public void handle ( final MessageEnd messageEnd ) {
		componentHelper.logger.log(
			Level.FINE, componentHelper.logName, "handle",
			"handling MessageEnd", messageEnd
		);
		dispatchCache();
		dispatch( messageEnd );
	}

	/* =================== Protected Methods ==================== */

	/**
	 * Flushes the series cache and then dispatches the incoming information.
	 * This is called for all except incoming {@link TimeSeries}.
	 */
	protected void dispatchCache () {
		if (cachedTimeSeries.isEmpty()) return;
		
		if (cachedTimeSeries.size() == 1 && maxGapBetweenObservations < 0) {
			dispatch( (TimeSeries)cachedTimeSeries.remove(0) );
			return;
		}
		
		TimeSeries topSeries = (TimeSeries)cachedTimeSeries.firstElement();
		SortedMap observationMap = getObservationMap( cachedTimeSeries );

		if (observationMap.isEmpty()) {
			dispatch( topSeries );
			return;
		}
		
		if (maxGapBetweenObservations < 0) {
			dispatchTimeSeries(topSeries.getKey(), observationMap);
		} else if (maxGapBetweenObservations == 0) {
			dispatchSingleObservationPerTimeSeries(topSeries.getKey(), observationMap);
		} else {
			dispatchTimeSeries(topSeries.getKey(), observationMap, maxGapBetweenObservations);
		}		
	}

	protected SortedMap getObservationMap( final Vector timeSeriesVector ) {		
		SortedMap observationMap = new TreeMap();
		while (!timeSeriesVector.isEmpty()) {
			TimeSeries timeSeries = (TimeSeries)timeSeriesVector.remove(0);
			
			if (timeSeries.getInterval() == null) continue;
			
			Iterator oit = timeSeries.observationIterator();		
			Iterator rit = timeSeries.getInterval().iterator();
		
			while (rit.hasNext()) {
				RegularInterval interval = (RegularInterval)rit.next();
				Observation observation = (Observation)oit.next();
				if (observation.equals(Observation.EMPTY_OBSERVATION)) continue;
				observationMap.put(interval, observation);			
			}
		}
		
		return observationMap;
	}
	
	protected void dispatchTimeSeries (
		final String key,
		final SortedMap observationMap
	) {
		RegularInterval first = (RegularInterval)observationMap.firstKey();
		RegularInterval last = (RegularInterval)observationMap.lastKey();

		RegularInterval interval = new GregorianRegularInterval(
			first.getFrequency(),
			first.getBegin(),
			last.getEnd()
		);

		Iterator rit = interval.iterator();
		ArrayList observationList = new ArrayList();
		while (rit.hasNext()) {			
			RegularInterval intervalBase = (RegularInterval)rit.next();
			Observation observation = (Observation)observationMap.get(intervalBase);
			if (observation == null) observation = Observation.EMPTY_OBSERVATION;
			observationList.add(observation);
		}		
	
		dispatch( new TimeSeries( key, interval, observationList) );
	}

	protected void dispatchSingleObservationPerTimeSeries (
		final String key,
		final SortedMap observationMap
	) {		
		Iterator sit = observationMap.entrySet().iterator();
		while (sit.hasNext()) {
			Map.Entry entry = (Map.Entry)sit.next();
			dispatch( 
				new TimeSeries( 
					key, 
					(RegularInterval)entry.getKey(), 
					new Observation[] {(Observation)entry.getValue()}
				)
			);
		}		
	}

	protected void	dispatchTimeSeries(
		final String key, 
		final SortedMap observationMap, 
		final int maxGapBetweenObservations
	) {
		RegularInterval first = (RegularInterval)observationMap.firstKey();
		RegularInterval previous = first;
		Iterator kit = observationMap.keySet().iterator();
		while (kit.hasNext()) {
			RegularInterval current = (RegularInterval)kit.next();
			if (maxGapBetweenObservations < getIntervalDistance(previous, current)) {
				dispatchTimeSeries(key, observationMap.subMap(first, current));
				first = current;
			}
			previous = current;
		}
		dispatchTimeSeries(key, observationMap.tailMap(first));
	}


	protected int getIntervalDistance (
		final RegularInterval first,
		final RegularInterval last
	) {
		RegularInterval interval = new GregorianRegularInterval(
			first.getFrequency(), 
			first.getBegin(), 
			last.getBegin()
		);
		return interval.getSize();
	}			
}
