// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.util;

import java.util.Date;

import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;

import Int.ecb.gesmescb.*;
import Int.ecb.gesmescb.time.*;

import Int.ecb.gesmescb.event.TestEventListener;


/* ***************************** Class ****************************** */

/**
 * <P>Tests the implementation of {@link GESMESCBEventSieve}.</P>
 *
 * @author
 * <A HREF="mailto:Sami.Airo@ECB.int">Sami Airo</A>
 *
 * @version @(#)$Id: TimeSeriesJoinerTest.java,v 1.1 2017/01/19 16:14:04 trainot Exp $
 */

public class TimeSeriesJoinerTest extends TestCase {

	/* ======================= Constants ======================== */


	/* ========================= Fields ========================= */


	private final MessageBegin messageBegin = new MessageBegin(
		"MREF0001",
		"4F0",
		"4F0",
		"ECB",
		MessageBegin.STATISTICAL_DATA,
		"De subject",
		new Contact[]{},
		"DSI",
		MessageBegin.UPDATE_OR_REPLACE_ACTION,
		"",
		"",
		"",
		"KEY_FAMILY"
	);

	private final Attribute attribute = new Attribute(
		"A.B.C", "", "", "TEST_ATT", "0"
	);

	private final MessageEnd messageEnd = new MessageEnd(
		"MREF0001"
	);

	private final TimeSeries series_ABC_1999 = new TimeSeries(
		"A.B.C",
		"1999",
		"602",
		new Observation[]{
			new Observation("1999","A","","")
		}
	);

	private final TimeSeries series_ABC_2000 = new TimeSeries(
		"A.B.C",
		"2000",
		"602",
		new Observation[]{
			new Observation("2000","A","","")
		}
	);

	private final TimeSeries series_ABC_2002 = new TimeSeries(
		"A.B.C",
		"2002",
		"602",
		new Observation[]{
			new Observation("2002","A","","")
		}
	);

	private final TimeSeries series_ABC_19992000 = new TimeSeries(
		"A.B.C",
		"19992000",
		"702",
		new Observation[]{
			new Observation("1999","A","",""),
			new Observation("2000","A","","")
		}
	);

	private final TimeSeries series_ABC_19992002 = new TimeSeries(
		"A.B.C",
		"19992002",
		"702",
		new Observation[]{
			new Observation("1999","A","",""),
			new Observation("2000","A","",""),
			Observation.EMPTY_OBSERVATION,
			new Observation("2002","A","",""),
		}
	);

	private final TimeSeries series_ABC_19961998 = new TimeSeries(
		"A.B.C",
		"19961998",
		"702",
		new Observation[]{
			Observation.EMPTY_OBSERVATION,
			new Observation("1997","A","",""),
			new Observation("1998","A","","")
		}
	);

	private final TimeSeries series_ABC_19972000 = new TimeSeries(
		"A.B.C",
		"19972000",
		"702",
		new Observation[]{
			new Observation("1997","A","",""),
			new Observation("1998","A","",""),
			Observation.EMPTY_OBSERVATION,
			new Observation("2000","A","","")
		}
	);

	private final TimeSeries series_ABC_19972000_2 = new TimeSeries(
		"A.B.C",
		"19972000",
		"702",
		new Observation[]{
			new Observation("1997","A","",""),
			new Observation("1998","A","",""),
			new Observation("1999","A","",""),
			new Observation("2000","A","","")
		}
	);

	private final TimeSeries series_ABC_EMPTY = new TimeSeries(
		"A.B.C",
		new GregorianRegularInterval(
			RegularInterval.ANNUAL,
			new Date(0),
			0
		),
		new Observation[]{}
	);

	private final TimeSeries series_ABC_NO_INTERVAL = new TimeSeries(
		"A.B.C",
		"",
		"",
		new Observation[]{}
	);

	private final TimeSeries series_ACC_2000 = new TimeSeries(
		"A.C.C",
		"2000",
		"602",
		new Observation[]{
			new Observation("2000","A","","")
		}
	);

	private final TimeSeries series_QBC_19991 = new TimeSeries(
		"Q.B.C",
		"19991",
		"608",
		new Observation[]{
			new Observation("19991","A","","")
		}
	);

	private final TimeSeries series_QBC_19992 = new TimeSeries(
		"Q.B.C",
		"19992",
		"608",
		new Observation[]{
			new Observation("19992","A","","")
		}
	);

	private final TimeSeries series_QBC_1999119992 = new TimeSeries(
		"Q.B.C",
		"1999119992",
		"708",
		new Observation[]{
			new Observation("19991","A","",""),
			new Observation("19992","A","","")
		}
	);

	private final TimeSeries series_QBC_20001 = new TimeSeries(
		"Q.B.C",
		"20001",
		"608",
		new Observation[]{
			new Observation("20001","A","","")
		}
	);

	private final TimeSeries series_QBC_1999120001 = new TimeSeries(
		"Q.B.C",
		"1999120001",
		"708",
		new Observation[]{
			new Observation("19991","A","",""),
			new Observation("19992","A","",""),
			Observation.EMPTY_OBSERVATION,
			Observation.EMPTY_OBSERVATION,
			new Observation("20001","A","","")
		}
	);

	private final TimeSeries series_MBC_199901 = new TimeSeries(
		"M.B.C",
		"199901",
		"610",
		new Observation[]{
			new Observation("199901","A","","")
		}
	);

	private final TimeSeries series_MBC_199902 = new TimeSeries(
		"M.B.C",
		"199902",
		"610",
		new Observation[]{
			new Observation("199902","A","","")
		}
	);

	private final TimeSeries series_MBC_199901199902 = new TimeSeries(
		"M.B.C",
		"199901199902",
		"710",
		new Observation[]{
			new Observation("199901","A","",""),
			new Observation("199902","A","","")
		}
	);

	private final TimeSeries series_MBC_199905 = new TimeSeries(
		"M.B.C",
		"199905",
		"610",
		new Observation[]{
			new Observation("199905","A","","")
		}
	);

	private final TimeSeries series_MBC_199901199905 = new TimeSeries(
		"M.B.C",
		"199901199905",
		"710",
		new Observation[]{
			new Observation("199901","A","",""),
			new Observation("199902","A","",""),
			Observation.EMPTY_OBSERVATION,
			Observation.EMPTY_OBSERVATION,
			new Observation("199905","A","","")
		}
	);

	private final TimeSeries series_WBC_199901 = new TimeSeries(
		"W.B.C",
		"199901",
		"616",
		new Observation[]{
			new Observation("199901","A","","")
		}
	);

	private final TimeSeries series_WBC_199902 = new TimeSeries(
		"W.B.C",
		"199902",
		"616",
		new Observation[]{
			new Observation("199902","A","","")
		}
	);

	private final TimeSeries series_WBC_199901199902 = new TimeSeries(
		"W.B.C",
		"199901199902",
		"716",
		new Observation[]{
			new Observation("199901","A","",""),
			new Observation("199902","A","","")
		}
	);

	private final TimeSeries series_WBC_199905 = new TimeSeries(
		"W.B.C",
		"199905",
		"616",
		new Observation[]{
			new Observation("199905","A","","")
		}
	);

	private final TimeSeries series_WBC_199901199905 = new TimeSeries(
		"W.B.C",
		"199901199905",
		"716",
		new Observation[]{
			new Observation("199901","A","",""),
			new Observation("199902","A","",""),
			Observation.EMPTY_OBSERVATION,
			Observation.EMPTY_OBSERVATION,
			new Observation("199905","A","","")
		}
	);

	private final TimeSeries series_DBC_19990101 = new TimeSeries(
		"D.B.C",
		"19990101",
		"102",
		new Observation[]{
			new Observation("19990101","A","","")
		}
	);

	private final TimeSeries series_DBC_19990102 = new TimeSeries(
		"D.B.C",
		"19990102",
		"102",
		new Observation[]{
			new Observation("19990102","A","","")
		}
	);

	private final TimeSeries series_DBC_1999010119990102 = new TimeSeries(
		"D.B.C",
		"1999010119990102",
		"711",
		new Observation[]{
			new Observation("19990101","A","",""),
			new Observation("19990102","A","","")
		}
	);

	private final TimeSeries series_DBC_19990105 = new TimeSeries(
		"D.B.C",
		"19990105",
		"102",
		new Observation[]{
			new Observation("19990105","A","","")
		}
	);

	private final TimeSeries series_DBC_1999010119990105 = new TimeSeries(
		"D.B.C",
		"1999010119990105",
		"711",
		new Observation[]{
			new Observation("19990101","A","",""),
			new Observation("19990102","A","",""),
			Observation.EMPTY_OBSERVATION,
			Observation.EMPTY_OBSERVATION,
			new Observation("19990105","A","","")
		}
	);

	private final TimeSeries series_DBC_1998010119980112 = new TimeSeries(
		"D.B.C",
		"1998010119980112",
		"711",
		new Observation[]{
			Observation.EMPTY_OBSERVATION,
			new Observation("19980102","A","",""),
			new Observation("19980103","A","",""),
			Observation.EMPTY_OBSERVATION,
			Observation.EMPTY_OBSERVATION,
			Observation.EMPTY_OBSERVATION,
			new Observation("19980107","A","",""),
			Observation.EMPTY_OBSERVATION,
			Observation.EMPTY_OBSERVATION,
			Observation.EMPTY_OBSERVATION,
			new Observation("19980111","A","",""),
			Observation.EMPTY_OBSERVATION,
		}
	);

	private final TimeSeries series_DBC_19980111 = new TimeSeries(
		"D.B.C",
		"19980111",
		"102",
		new Observation[]{
			new Observation("19980111","A","","")
		}
	);

	private final TimeSeries series_LastDuplicateWins1 = new TimeSeries (
		"H.DK.XX.803.Z.DKK.Z0Z.E",
		"20072",
		"604",
		new Observation[]{
			new Observation("1","A","","")
		}
	);
	
	private final TimeSeries series_LastDuplicateWins2 = new TimeSeries (
		"H.DK.XX.803.Z.DKK.Z0Z.E",
		"20072",
		"604",
		new Observation[]{
			new Observation("2","A","","")
		}
	);
	
	private final TimeSeries series_LastDuplicateWinsMV = new TimeSeries (
		"H.DK.XX.803.Z.DKK.Z0Z.E",
		"20072",
		"604",
		new Observation[]{
			new Observation( Observation.MISSING_VALUE,"M","","")
		}
	);
	
	/* ===================== Protected Methods ===================== */

	protected void setUp() {
		System.setProperty("Int.ecb.gesmescb.TimeSeriesJoiner.maxGapBetweenObservations", "5");
	}

	/* ====================== Constructors ====================== */

	/**
	 * Constructs a GESMESCBEventSieveTest with the specified name.
	 *
	 * @param name Test case name
	 */
	public TimeSeriesJoinerTest ( final String name ) {
		super( name );
	}

	/* ===================== Public Methods ===================== */

	/**
	 * Assembles and returns a test suite for all
	 * the test methods of this test case.
	 *
	 * @return a non-null test suite.
	 */
	public static Test suite () {
		return new TestSuite( TimeSeriesJoinerTest.class );
	}

	/**
	 * Runs the test case.
	 */
	public static void main ( final String args[] ) {
		String[] testCaseName = {
			TimeSeriesJoinerTest.class.getName()
		};
		junit.textui.TestRunner.main( testCaseName );
//		junit.swingui.TestRunner.main( testCaseName );
	}

	/* --------------------- Test Methods --------------------- */

	/**
	 * Test join of two annual series in normal order
	 */
	public void testJoinAnnual1 () {
		final TimeSeriesJoiner joiner = new TimeSeriesJoiner();
		final TestEventListener listener = new TestEventListener( 10 );

		joiner.addListener( listener, true );
		joiner.handle(messageBegin);
		assertSame("messageBegin", messageBegin, listener.getLastObject() );
		joiner.handle(series_ABC_1999);
		assertSame("series_ABC_1999", messageBegin, listener.getLastObject() );
		joiner.handle(series_ABC_2000);
		assertSame("series_ABC_2000", messageBegin, listener.getLastObject() );
		joiner.handle(series_ACC_2000);
		assertEquals("series_ABC_19992000", series_ABC_19992000, listener.getLastObject() );
		joiner.handle(messageEnd);
		assertSame("messageEnd", messageEnd, listener.getLastObject() );		
		assertEquals("size", 4, listener.size() );		
	}

	/**
	 * Test join of two annual series in reverse order
	 */
	public void testJoinAnnual2 () {
		final TimeSeriesJoiner joiner = new TimeSeriesJoiner();
		final TestEventListener listener = new TestEventListener( 10 );

		joiner.addListener( listener, true );
		joiner.handle(messageBegin);
		assertSame("messageBegin", messageBegin, listener.getLastObject() );
		joiner.handle(series_ABC_2000);
		assertSame("series_ABC_2000", messageBegin, listener.getLastObject() );
		joiner.handle(series_ABC_1999);
		assertSame("series_ABC_1999", messageBegin, listener.getLastObject() );
		joiner.handle(series_ACC_2000);
		assertEquals("series_ABC_19992000", series_ABC_19992000, listener.getLastObject() );
		joiner.handle(messageEnd);
		assertSame("messageEnd", messageEnd, listener.getLastObject() );		
		assertEquals("size", 4, listener.size() );		
	}

	/**
	 * Test join of three annual series in normal order with gap in between
	 */
	public void testJoinAnnual3 () {
		final TimeSeriesJoiner joiner = new TimeSeriesJoiner();
		final TestEventListener listener = new TestEventListener( 10 );

		joiner.addListener( listener, true );
		joiner.handle(messageBegin);
		assertSame("messageBegin", messageBegin, listener.getLastObject() );
		joiner.handle(series_ABC_1999);
		assertSame("series_ABC_1999", messageBegin, listener.getLastObject() );
		joiner.handle(series_ABC_2000);
		assertSame("series_ABC_2000", messageBegin, listener.getLastObject() );
		joiner.handle(series_ABC_2002);
		assertSame("series_ABC_2002", messageBegin, listener.getLastObject() );
		joiner.handle(series_ACC_2000);
		assertEquals("series_ABC_19992002", series_ABC_19992002, listener.getLastObject() );
		assertEquals("size", 2, listener.size() );		
		joiner.handle(messageEnd);
		assertSame("messageEnd", messageEnd, listener.getLastObject() );		
		assertEquals("size", 4, listener.size() );		
	}

	/**
	 * Test join of two annual series followed by an attribute
	 */
	public void testJoinAnnual4 () {
		final TimeSeriesJoiner joiner = new TimeSeriesJoiner();
		final TestEventListener listener = new TestEventListener( 10 );

		joiner.addListener( listener, true );
		joiner.handle(messageBegin);
		assertSame("messageBegin", messageBegin, listener.getLastObject() );
		joiner.handle(series_ABC_1999);
		assertSame("series_ABC_1999", messageBegin, listener.getLastObject() );
		joiner.handle(series_ABC_2000);
		assertSame("series_ABC_2000", messageBegin, listener.getLastObject() );
		joiner.handle(attribute);
		assertEquals("attribute", attribute, listener.getLastObject() );
		assertEquals("size", 3, listener.size() );		
		joiner.handle(messageEnd);
		assertSame("messageEnd", messageEnd, listener.getLastObject() );		
	}

	/**
	 * Test join of two annual series with gap in start and between
	 */
	public void testJoinAnnual5 () {
		final TimeSeriesJoiner joiner = new TimeSeriesJoiner();
		final TestEventListener listener = new TestEventListener( 10 );

		joiner.addListener( listener, true );
		joiner.handle(messageBegin);
		assertSame("messageBegin", messageBegin, listener.getLastObject() );
		joiner.handle(series_ABC_19961998);
		assertSame("series_ABC_19961998", messageBegin, listener.getLastObject() );
		joiner.handle(series_ABC_2000);
		assertSame("series_ABC_2000", messageBegin, listener.getLastObject() );
		joiner.handle(series_ACC_2000);
		assertEquals("series_ABC_19972000", series_ABC_19972000, listener.getLastObject() );
		assertEquals("size", 2, listener.size() );		
		joiner.handle(messageEnd);
		assertSame("messageEnd", messageEnd, listener.getLastObject() );		
		assertEquals("size", 4, listener.size() );		
	}

	/**
	 * Test join of single annual empty series
	 */
	public void testJoinAnnualEmpty1 () {
		final TimeSeriesJoiner joiner = new TimeSeriesJoiner();
		final TestEventListener listener = new TestEventListener( 10 );

		joiner.addListener( listener, true );
		joiner.handle(messageBegin);
		assertSame("messageBegin", messageBegin, listener.getLastObject() );
		joiner.handle(series_ABC_EMPTY);
		assertSame("series_ABC_EMPTY 1", messageBegin, listener.getLastObject() );
		joiner.handle(series_ACC_2000);
		assertEquals("series_ABC_EMPTY", series_ABC_EMPTY, listener.getLastObject() );
		assertEquals("size", 2, listener.size() );		
		joiner.handle(messageEnd);
		assertSame("messageEnd", messageEnd, listener.getLastObject() );		
		assertEquals("size", 4, listener.size() );		
	}

	/**
	 * Test join of two annual empty series
	 */
	public void testJoinAnnualEmpty2 () {
		final TimeSeriesJoiner joiner = new TimeSeriesJoiner();
		final TestEventListener listener = new TestEventListener( 10 );

		joiner.addListener( listener, true );
		joiner.handle(messageBegin);
		assertSame("messageBegin", messageBegin, listener.getLastObject() );
		joiner.handle(series_ABC_EMPTY);
		assertSame("series_ABC_EMPTY 1", messageBegin, listener.getLastObject() );
		joiner.handle(series_ABC_EMPTY);
		assertSame("series_ABC_EMPTY 2", messageBegin, listener.getLastObject() );
		joiner.handle(series_ACC_2000);
		assertEquals("series_ABC_EMPTY", series_ABC_EMPTY, listener.getLastObject() );
		assertEquals("size", 2, listener.size() );		
		joiner.handle(messageEnd);
		assertSame("messageEnd", messageEnd, listener.getLastObject() );		
		assertEquals("size", 4, listener.size() );		
	}

	/**
	 * Test join of annual empty and nonempty series
	 */
	public void testJoinAnnualEmpty3 () {
		final TimeSeriesJoiner joiner = new TimeSeriesJoiner();
		final TestEventListener listener = new TestEventListener( 10 );

		joiner.addListener( listener, true );
		joiner.handle(messageBegin);
		assertSame("messageBegin", messageBegin, listener.getLastObject() );
		joiner.handle(series_ABC_EMPTY);
		assertSame("series_ABC_EMPTY", messageBegin, listener.getLastObject() );
		joiner.handle(series_ABC_19992000);
		assertSame("series_ABC_19992000", messageBegin, listener.getLastObject() );
		joiner.handle(series_ACC_2000);
		assertEquals("series_ACC_2000", series_ABC_19992000, listener.getLastObject() );
		assertEquals("size", 2, listener.size() );		
		joiner.handle(messageEnd);
		assertSame("messageEnd", messageEnd, listener.getLastObject() );		
		assertEquals("size", 4, listener.size() );		
	}

	/**
	 * Test join of single annual empty series with no interval
	 */
	public void testJoinAnnualEmpty5 () {
		final TimeSeriesJoiner joiner = new TimeSeriesJoiner();
		final TestEventListener listener = new TestEventListener( 10 );

		joiner.addListener( listener, true );
		joiner.handle(messageBegin);
		assertSame("messageBegin", messageBegin, listener.getLastObject() );
		joiner.handle(series_ABC_NO_INTERVAL);
		assertSame("series_ABC_NO_INTERVAL", messageBegin, listener.getLastObject() );
		joiner.handle(series_ACC_2000);
		assertEquals("series_ABC_2000", series_ABC_NO_INTERVAL, listener.getLastObject() );
		assertEquals("size", 2, listener.size() );		
		joiner.handle(messageEnd);
		assertSame("messageEnd", messageEnd, listener.getLastObject() );		
		assertEquals("size", 4, listener.size() );		
	}

	/**
	 * Test join of two quarterly series in normal order
	 */
	public void testJoinQuarterly1 () {
		final TimeSeriesJoiner joiner = new TimeSeriesJoiner();
		final TestEventListener listener = new TestEventListener( 10 );

		joiner.addListener( listener, true );
		joiner.handle(messageBegin);
		assertSame("messageBegin", messageBegin, listener.getLastObject() );
		joiner.handle(series_QBC_19991);
		assertSame("series_QBC_19991", messageBegin, listener.getLastObject() );
		joiner.handle(series_QBC_19992);
		assertSame("series_QBC_19992", messageBegin, listener.getLastObject() );
		joiner.handle(series_ACC_2000);
		assertEquals("series_QBC_199911992", series_QBC_1999119992, listener.getLastObject() );
		joiner.handle(messageEnd);
		assertSame("messageEnd", messageEnd, listener.getLastObject() );		
		assertEquals("size", 4, listener.size() );		
	}

	/**
	 * Test join of three quarterly series with gaps
	 */
	public void testJoinQuarterly2 () {
		final TimeSeriesJoiner joiner = new TimeSeriesJoiner();
		final TestEventListener listener = new TestEventListener( 10 );

		joiner.addListener( listener, true );
		joiner.handle(messageBegin);
		assertSame("messageBegin", messageBegin, listener.getLastObject() );
		joiner.handle(series_QBC_19992);
		assertSame("series_QBC_19992", messageBegin, listener.getLastObject() );
		joiner.handle(series_QBC_1999119992);
		assertSame("series_QBC_1999119992", messageBegin, listener.getLastObject() );
		joiner.handle(series_QBC_20001);
		assertSame("series_QBC_20001", messageBegin, listener.getLastObject() );
		joiner.handle(series_ACC_2000);
		assertEquals("series_QBC_1999120001", series_QBC_1999120001, listener.getLastObject() );
		joiner.handle(messageEnd);
		assertSame("messageEnd", messageEnd, listener.getLastObject() );		
		assertEquals("size", 4, listener.size() );		
	}

	/**
	 * Test join of two monthly series in normal order
	 */
	public void testJoinMonthly1 () {
		final TimeSeriesJoiner joiner = new TimeSeriesJoiner();
		final TestEventListener listener = new TestEventListener( 10 );

		joiner.addListener( listener, true );
		joiner.handle(messageBegin);
		assertSame("messageBegin", messageBegin, listener.getLastObject() );
		joiner.handle(series_MBC_199901);
		assertSame("series_MBC_199901", messageBegin, listener.getLastObject() );
		joiner.handle(series_MBC_199902);
		assertSame("series_MBC_199902", messageBegin, listener.getLastObject() );
		joiner.handle(series_ACC_2000);
		assertEquals("series_MBC_199901199902", series_MBC_199901199902, listener.getLastObject() );
		joiner.handle(messageEnd);
		assertSame("messageEnd", messageEnd, listener.getLastObject() );		
		assertEquals("size", 4, listener.size() );		
	}

	/**
	 * Test join of three monthly series with gaps
	 */
	public void testJoinMonthly2 () {
		final TimeSeriesJoiner joiner = new TimeSeriesJoiner();
		final TestEventListener listener = new TestEventListener( 10 );

		joiner.addListener( listener, true );
		joiner.handle(messageBegin);
		assertSame("messageBegin", messageBegin, listener.getLastObject() );
		joiner.handle(series_MBC_199902);
		assertSame("series_MBC_199902", messageBegin, listener.getLastObject() );
		joiner.handle(series_MBC_199901199902);
		assertSame("series_MBC_199901199902", messageBegin, listener.getLastObject() );
		joiner.handle(series_MBC_199905);
		assertSame("series_MBC_199905", messageBegin, listener.getLastObject() );
		joiner.handle(series_ACC_2000);
		assertEquals("series_MBC_199901199905", series_MBC_199901199905, listener.getLastObject() );
		joiner.handle(messageEnd);
		assertSame("messageEnd", messageEnd, listener.getLastObject() );		
		assertEquals("size", 4, listener.size() );		
	}

	/**
	 * Test join of two weekly series in normal order
	 */
	public void testJoinWeekly1 () {
		final TimeSeriesJoiner joiner = new TimeSeriesJoiner();
		final TestEventListener listener = new TestEventListener( 10 );

		joiner.addListener( listener, true );
		joiner.handle(messageBegin);
		assertSame("messageBegin", messageBegin, listener.getLastObject() );
		joiner.handle(series_WBC_199901);
		assertSame("series_WBC_199901", messageBegin, listener.getLastObject() );
		joiner.handle(series_WBC_199902);
		assertSame("series_WBC_199902", messageBegin, listener.getLastObject() );
		joiner.handle(series_ACC_2000);
		assertEquals("series_WBC_199901199902", series_WBC_199901199902, listener.getLastObject() );
		joiner.handle(messageEnd);
		assertSame("messageEnd", messageEnd, listener.getLastObject() );		
		assertEquals("size", 4, listener.size() );		
	}

	/**
	 * Test join of three weekly series with gaps
	 */
	public void testJoinWeekly2 () {
		final TimeSeriesJoiner joiner = new TimeSeriesJoiner();
		final TestEventListener listener = new TestEventListener( 10 );

		joiner.addListener( listener, true );
		joiner.handle(messageBegin);
		assertSame("messageBegin", messageBegin, listener.getLastObject() );
		joiner.handle(series_WBC_199902);
		assertSame("series_WBC_199902", messageBegin, listener.getLastObject() );
		joiner.handle(series_WBC_199901199902);
		assertSame("series_WBC_199901199902", messageBegin, listener.getLastObject() );
		joiner.handle(series_WBC_199905);
		assertSame("series_WBC_199905", messageBegin, listener.getLastObject() );
		joiner.handle(series_ACC_2000);
		assertEquals("series_WBC_199901199905", series_WBC_199901199905, listener.getLastObject() );
		joiner.handle(messageEnd);
		assertSame("messageEnd", messageEnd, listener.getLastObject() );		
		assertEquals("size", 4, listener.size() );		
	}

	/**
	 * Test join of two daily series in normal order
	 */
	public void testJoinDaily1 () {
		final TimeSeriesJoiner joiner = new TimeSeriesJoiner();
		final TestEventListener listener = new TestEventListener( 10 );

		joiner.addListener( listener, true );
		joiner.handle(messageBegin);
		assertSame("messageBegin", messageBegin, listener.getLastObject() );
		joiner.handle(series_DBC_19990101);
		assertSame("series_DBC_19990101", messageBegin, listener.getLastObject() );
		joiner.handle(series_DBC_19990102);
		assertSame("series_DBC_19990102", messageBegin, listener.getLastObject() );
		joiner.handle(series_ACC_2000);
		assertEquals("series_DBC_1999010119990102", series_DBC_1999010119990102, listener.getLastObject() );
		joiner.handle(messageEnd);
		assertSame("messageEnd", messageEnd, listener.getLastObject() );		
		assertEquals("size", 4, listener.size() );		
	}

	/**
	 * Test join of three daily series with gaps
	 */
	public void testJoinDaily2 () {
		final TimeSeriesJoiner joiner = new TimeSeriesJoiner();
		final TestEventListener listener = new TestEventListener( 10 );

		joiner.addListener( listener, true );
		joiner.handle(messageBegin);
		assertSame("messageBegin", messageBegin, listener.getLastObject() );
		joiner.handle(series_DBC_19990102);
		assertSame("series_DBC_19990102", messageBegin, listener.getLastObject() );
		joiner.handle(series_DBC_1999010119990102);
		assertSame("series_DBC_1999010119990102", messageBegin, listener.getLastObject() );
		joiner.handle(series_DBC_19990105);
		assertSame("series_DBC_19990105", messageBegin, listener.getLastObject() );
		joiner.handle(series_ACC_2000);
		assertEquals("series_DBC_1999010119990105", series_DBC_1999010119990105, listener.getLastObject() );
		joiner.handle(messageEnd);
		assertSame("messageEnd", messageEnd, listener.getLastObject() );		
		assertEquals("size", 4, listener.size() );		
	}

	/**
	 * Test join of three daily series with gaps with maxGap 0
	 */
	public void testJoinWithGap0 () {
		System.setProperty("Int.ecb.gesmescb.TimeSeriesJoiner.maxGapBetweenObservations", "0");
		final TimeSeriesJoiner joiner = new TimeSeriesJoiner();
		final TestEventListener listener = new TestEventListener( 10 );

		joiner.addListener( listener, true );
		joiner.handle(messageBegin);
		assertSame("messageBegin", messageBegin, listener.getLastObject() );
		joiner.handle(series_DBC_19990102);
		assertSame("series_DBC_19990102", messageBegin, listener.getLastObject() );
		joiner.handle(series_DBC_1999010119990102);
		assertSame("series_DBC_1999010119990102", messageBegin, listener.getLastObject() );
		joiner.handle(series_DBC_19990105);
		assertSame("series_DBC_19990105", messageBegin, listener.getLastObject() );
		joiner.handle(series_ACC_2000);
		assertEquals("series_DBC_19990105_2", series_DBC_19990105, listener.getLastObject() );
		joiner.handle(messageEnd);
		assertSame("messageEnd", messageEnd, listener.getLastObject() );		
		assertEquals("size", 6, listener.size() );		
	}

	/**
	 * Test join of a daily series with gaps with maxGap 2
	 */
	public void testJoinWithGap2 () {
		System.setProperty("Int.ecb.gesmescb.TimeSeriesJoiner.maxGapBetweenObservations", "2");
		final TimeSeriesJoiner joiner = new TimeSeriesJoiner();
		final TestEventListener listener = new TestEventListener( 10 );

		joiner.addListener( listener, true );
		joiner.handle(messageBegin);
		assertSame("messageBegin", messageBegin, listener.getLastObject() );
		joiner.handle(series_DBC_1998010119980112);
		assertSame("series_DBC_1998010119980112", messageBegin, listener.getLastObject() );
		joiner.handle(series_ACC_2000);
		assertEquals("series_DBC_19980111", series_DBC_19980111, listener.getLastObject() );
		joiner.handle(messageEnd);
		assertSame("messageEnd", messageEnd, listener.getLastObject() );		
		assertEquals("size", 6, listener.size() );		
	}

	/**
	 * Test that empty doesn't overlay value
	 */
	public void testJoinEmpty () {
		final TimeSeriesJoiner joiner = new TimeSeriesJoiner();
		final TestEventListener listener = new TestEventListener( 10 );

		joiner.addListener( listener, true );
		joiner.handle(messageBegin);
		assertSame("messageBegin", messageBegin, listener.getLastObject() );
		joiner.handle(series_ABC_19992000);
		assertSame("series_ABC_19992000", messageBegin, listener.getLastObject() );
		joiner.handle(series_ABC_19972000);
		assertSame("series_ABC_19972000", messageBegin, listener.getLastObject() );
		joiner.handle(series_ACC_2000);
		assertEquals("series_ABC_19972000_2", series_ABC_19972000_2, listener.getLastObject() );
		joiner.handle(messageEnd);
		assertSame("messageEnd", messageEnd, listener.getLastObject() );		
		assertEquals("size", 4, listener.size() );		
	}

	/**
	 * Test that empty doesn't overlay value, also when empty comes first
	 */
	public void testJoinEmpty2 () {
		final TimeSeriesJoiner joiner = new TimeSeriesJoiner();
		final TestEventListener listener = new TestEventListener( 10 );

		joiner.addListener( listener, true );
		joiner.handle(messageBegin);
		assertSame("messageBegin", messageBegin, listener.getLastObject() );
		joiner.handle(series_ABC_19972000);
		assertSame("series_ABC_19972000", messageBegin, listener.getLastObject() );
		joiner.handle(series_ABC_19992000);
		assertSame("series_ABC_19992000", messageBegin, listener.getLastObject() );
		joiner.handle(series_ACC_2000);
		assertEquals("series_ABC_19972000_2", series_ABC_19972000_2, listener.getLastObject() );
		joiner.handle(messageEnd);
		assertSame("messageEnd", messageEnd, listener.getLastObject() );		
		assertEquals("size", 4, listener.size() );		
	}
	/**
	 * Test that the last (non empty) duplicate wins, without Missing value
	 */
	public void testLastDupeWins () {
		final TimeSeriesJoiner joiner = new TimeSeriesJoiner();
		final TestEventListener listener = new TestEventListener( 10 );

		joiner.addListener( listener, true );
		joiner.handle(messageBegin);
		assertSame("messageBegin", messageBegin, listener.getLastObject() );
		joiner.handle(series_LastDuplicateWins1);
		assertSame("series_LastDuplicateWins1", messageBegin, listener.getLastObject() );
		joiner.handle(series_LastDuplicateWins2);
		assertSame("series_LastDuplicateWins2", messageBegin, listener.getLastObject() );
		joiner.handle(series_ACC_2000);
		assertEquals("series_LastDuplicateWins2", series_LastDuplicateWins2, listener.getLastObject() );
		joiner.handle(messageEnd);
		assertSame("messageEnd", messageEnd, listener.getLastObject() );		
		assertEquals("size", 4, listener.size() );		
	}
	/**
	 * Test that the last (non empty) duplicate wins, with Missing as last 
	 */
	public void testLastDupeWinsWithMissingAsLast () {
		final TimeSeriesJoiner joiner = new TimeSeriesJoiner();
		final TestEventListener listener = new TestEventListener( 10 );

		joiner.addListener( listener, true );
		joiner.handle(messageBegin);
		assertSame("messageBegin", messageBegin, listener.getLastObject() );
		joiner.handle(series_LastDuplicateWins1);
		assertSame("series_LastDuplicateWins1", messageBegin, listener.getLastObject() );
		joiner.handle(series_LastDuplicateWinsMV);
		assertSame("series_LastDuplicateWinsMV", messageBegin, listener.getLastObject() );
		joiner.handle(series_ACC_2000);
		assertEquals("series_LastDuplicateWinsMV", series_LastDuplicateWinsMV, listener.getLastObject() );
		joiner.handle(messageEnd);
		assertSame("messageEnd", messageEnd, listener.getLastObject() );		
		assertEquals("size", 4, listener.size() );		
	}
	/**
	 * Test that the last (non empty) duplicate wins, with a Missing Value
	 */
	public void testLastDupeWinsWithMissingFirst () {
		final TimeSeriesJoiner joiner = new TimeSeriesJoiner();
		final TestEventListener listener = new TestEventListener( 10 );

		joiner.addListener( listener, true );
		joiner.handle(messageBegin);
		assertSame("messageBegin", messageBegin, listener.getLastObject() );
		joiner.handle(series_LastDuplicateWinsMV);
		assertSame("series_LastDuplicateWinsMV", messageBegin, listener.getLastObject() );
		joiner.handle(series_LastDuplicateWins2);
		assertSame("series_LastDuplicateWins2", messageBegin, listener.getLastObject() );
		joiner.handle(series_ACC_2000);
		assertEquals("series_LastDuplicateWins2", series_LastDuplicateWins2, listener.getLastObject() );
		joiner.handle(messageEnd);
		assertSame("messageEnd", messageEnd, listener.getLastObject() );		
		assertEquals("size", 4, listener.size() );		
	}
}
