// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2003, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.xml;

import java.util.Calendar;
import java.util.Date;
import java.util.Iterator;

import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;
import Int.ecb.gesmescb.Contact;
import Int.ecb.gesmescb.KeyFamily;
import Int.ecb.gesmescb.KeyFamilyComponent;
import Int.ecb.gesmescb.MessageBegin;
import Int.ecb.gesmescb.MessageEnd;
import Int.ecb.gesmescb.error.ErrorHandler;
import Int.ecb.gesmescb.error.ErrorWatcher;
import Int.ecb.gesmescb.metabase.Metabase;
import Int.ecb.logging.LogManager;

/* ***************************** Class ****************************** */

/**
 * Implements a few tests for the {@link KeyFamilyURIMapping}.
 *
 * @author
 * <a href="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</a>
 *
 * @author
 * <a href="mailto:Sami.Airo@ECB.int">Sami Airo</a>
 *
 * @version @(#)$Id: KeyFamilyURIMappingTest.java,v 1.6 2009/05/13 09:43:28 sosnows Exp $
 */
public class KeyFamilyURIMappingTest extends TestCase {

	/* ======================= Constants ======================== */

	private static final Date VERSION_DATE;
	static {
		final Calendar calendar = Calendar.getInstance();
		calendar.clear();
		calendar.set( 2003, Calendar.NOVEMBER, 12, 14, 34 );
		VERSION_DATE = calendar.getTime();
	}

	private static final KeyFamily
		KF = new KeyFamily(
			"KF", "", "", VERSION_DATE, new KeyFamilyComponent[]{}
		),
		KF1 = new KeyFamily(
			"KF1", "", "", VERSION_DATE, new KeyFamilyComponent[]{}
		),
		BUHDY_KF = new KeyFamily(
			"BUHDY_KF", "", "", VERSION_DATE, new KeyFamilyComponent[]{}
		),
		BUHDY_KF2 = new KeyFamily(
			"BUHDY_KF2", "", "", VERSION_DATE, new KeyFamilyComponent[]{}
		),
		SHARED = new KeyFamily(
			"SHARED", "", "", VERSION_DATE, new KeyFamilyComponent[]{}
		),
		BUHDY_SECRET = new KeyFamily(
			"BUHDY_SECRET", "", "", VERSION_DATE, new KeyFamilyComponent[]{}
		);

	/* ========================= Fields ========================= */

	private ErrorHandler errorHandler;

	/* ====================== Constructors ====================== */

	public KeyFamilyURIMappingTest ( final String name ) {
		super( name );
	}

	/* ================== Protected Methods ===================== */

	/** Sets up the test fixture. */
	protected void setUp () {
		errorHandler = new ErrorHandler();
		LogManager.getLogManager().addGlobalHandler( errorHandler );
		System.setProperty(
			KeyFamilyURIMapping.class.getName() + ".BUHDY.*.namespace",
			"http://www.buhdy.org/star-namespace/kfidx()"
		);
		System.setProperty(
			KeyFamilyURIMapping.class.getName() + ".BUHDY.BUHDY_SECRET.namespace",
			"https://www.buhdy.org/secret-namespace/3.5/sec"
		);
		System.setProperty(
			KeyFamilyURIMapping.class.getName()
			+ ".BUHDY.*.dtd",
			"http://www.buhdy.org/date()/star.dtd"
		);
		System.setProperty(
			KeyFamilyURIMapping.class.getName()
			+ ".BUHDY.BUHDY_SECRET.dtd",
			"http://www.buhdy.org/kfid().dtd"
		);
		System.setProperty(
			KeyFamilyURIMapping.class.getName()
			+ ".BUHDY.*.schema",
			"http://www.buhdy.org/kfidv()/star.xsd"
		);
		System.setProperty(
			KeyFamilyURIMapping.class.getName()
			+ ".BUHDY.BUHDY_SECRET.schema",
			"http://www.buhdy.org/geheim.xsd"
		);
		System.setProperty(
			KeyFamilyURIMapping.class.getName()
			+ ".BUHDY.*.stylesheet",
			"http://www.buhdy.org/kfidx()/date().xsl"
		);
		System.setProperty(
			KeyFamilyURIMapping.class.getName()
			+ ".BUHDY.BUHDY_SECRET.stylesheet",
			"http://www.buhdy.org/date()/kfidx()/kfidx().xsd"
		);
		System.setProperty(
			KeyFamilyURIMapping.class.getName() + ".BUHDY.SHARED.namespace",
			"http://www.bodi.org/shared-vocabulary/3.4/kf2"
		);
		System.setProperty(
			KeyFamilyURIMapping.class.getName() + ".BUDDY.SHARED.namespace",
			"http://www.bodi.org/shared-vocabulary/3.4/kf2"
		);
		final Metabase metabase = Metabase.getMetabase();
		metabase.handle( new MessageBegin(
			"DUMMY", "", "", "BUHDY",
			MessageBegin.STATISTICAL_DEFINITIONS,
			"", new Contact[]{}
		) );
		metabase.handle( KF );
		metabase.handle( KF1 );
		metabase.handle( BUHDY_SECRET );
		metabase.handle( SHARED );
		metabase.handle( new MessageEnd( "DUMMY" ) );
		metabase.handle( new MessageBegin(
			"DUMMY", "", "", "BUDDY",
			MessageBegin.STATISTICAL_DEFINITIONS,
			"", new Contact[]{}
		) );
		metabase.handle( SHARED );
		metabase.handle( new MessageEnd( "DUMMY" ) );
	}

	/** Tears down the test fixture. */
	protected void tearDown () {
		System.getProperties().remove(
			KeyFamilyURIMapping.class.getName() + ".BUHDY.*.namespace"
		);
		System.getProperties().remove(
			KeyFamilyURIMapping.class.getName() + ".BUHDY.BUHDY_SECRET.namespace"
		);
		System.getProperties().remove(
			KeyFamilyURIMapping.class.getName() + ".BUHDY.*.dtd"
		);
		System.getProperties().remove(
			KeyFamilyURIMapping.class.getName() + ".BUHDY.BUHDY_SECRET.dtd"
		);
		System.getProperties().remove(
			KeyFamilyURIMapping.class.getName() + ".BUHDY.*.schema"
		);
		System.getProperties().remove(
			KeyFamilyURIMapping.class.getName() + ".BUHDY.BUHDY_SECRET.schema"
		);
		System.getProperties().remove(
			KeyFamilyURIMapping.class.getName() + ".BUHDY.*.stylesheet"
		);
		System.getProperties().remove(
			KeyFamilyURIMapping.class.getName() + ".BUHDY.BUHDY_SECRET.stylesheet"
		);
		System.getProperties().remove(
			KeyFamilyURIMapping.class.getName() + ".BUHDY.SHARED.namespace"
		);
		System.getProperties().remove(
			KeyFamilyURIMapping.class.getName() + ".BUDDY.SHARED.namespace"
		);
	}

	/* ===================== Public Methods ===================== */

	/**
	 * Assembles and returns a test suite for all
	 * the test methods of this test case.
	 *
	 * @return a non-null test suite
	 */
	public static Test suite () {
		final TestSuite suite = new TestSuite( KeyFamilyURIMappingTest.class );
		return suite;
	}

	/**
	 * @param arg the array of argument {@link String}s
	 */
	public static void main ( final String[] arg ) {
		final String[] testCaseName = {
			KeyFamilyURIMappingTest.class.getName()
		};
		junit.textui.TestRunner.main( testCaseName );
	}

	/* ---------------------- Test Methods ---------------------- */

	public void testXmlNS () {
		final KeyFamilyURIMapping helper = KeyFamilyURIMapping.getInstance();
		assertEquals(
			"setup: " + errorHandler.getMessage(),
			ErrorWatcher.OK,
			errorHandler.getSeverity()
		);
		assertEquals(
			"no strip",
			"http://www.buhdy.org/star-namespace/kf",
			helper.xmlns( "BUHDY", KF )
		);
		assertEquals(
			"number strip",
			"http://www.buhdy.org/star-namespace/kf",
			helper.xmlns( "BUHDY", KF1 )
		);
		assertEquals(
			"agency strip",
			"http://www.buhdy.org/star-namespace/kf",
			helper.xmlns( "BUHDY", BUHDY_KF )
		);
		assertEquals(
			"agency and number strip",
			"http://www.buhdy.org/star-namespace/kf",
			helper.xmlns( "BUHDY", BUHDY_KF2 )
		);
		assertEquals(
			"exact override",
			"https://www.buhdy.org/secret-namespace/3.5/sec",
			helper.xmlns( "BUHDY", BUHDY_SECRET )
		);
		assertEquals(
			"no problem yet: " + errorHandler.getMessage(),
			ErrorWatcher.OK,
			errorHandler.getSeverity()
		);
		assertNull(
			"unknown",
			helper.xmlns( "UNKNOWN", KF )
		);
		assertTrue(
			"now this was an error: " + errorHandler.getMessage(),
			ErrorWatcher.ERROR <= errorHandler.getSeverity()
		);
	}

	public void testDTD () {
		final KeyFamilyURIMapping helper = KeyFamilyURIMapping.getInstance();
		assertEquals(
			"insert date",
			"http://www.buhdy.org/2003-11-12/star.dtd",
			helper.dtd( "BUHDY", KF )
		);
		assertEquals(
			"insert kfid",
			"http://www.buhdy.org/BUHDY_SECRET.dtd",
			helper.dtd( "BUHDY", BUHDY_SECRET )
		);
		assertNull(
			"unknown",
			helper.dtd( "UNKNOWN", KF )
		);
		assertEquals(
			"still fine: " + errorHandler.getMessage(),
			ErrorWatcher.OK,
			errorHandler.getSeverity()
		);
	}

	public void testSchemaLocation () {
		final KeyFamilyURIMapping helper = KeyFamilyURIMapping.getInstance();
		assertEquals(
			"generic KF",
			"http://www.buhdy.org/kf/star.xsd",
			helper.schemaLocation( "BUHDY", KF )
		);
		assertEquals(
			"generic KF1",
			"http://www.buhdy.org/kf/1/star.xsd",
			helper.schemaLocation( "BUHDY", KF1 )
		);
		assertEquals(
			"generic KF2",
			"http://www.buhdy.org/kf/2/star.xsd",
			helper.schemaLocation( "BUHDY", BUHDY_KF2 )
		);
		assertEquals(
			"specified",
			"http://www.buhdy.org/geheim.xsd",
			helper.schemaLocation( "BUHDY", BUHDY_SECRET )
		);
		assertNull(
			"unknown",
			helper.schemaLocation( "UNKNOWN", KF )
		);
		assertEquals(
			"still fine: " + errorHandler.getMessage(),
			ErrorWatcher.OK,
			errorHandler.getSeverity()
		);
	}

	public void testStylesheet () {
		final KeyFamilyURIMapping helper = KeyFamilyURIMapping.getInstance();
		assertEquals(
			"kfidx date",
			"http://www.buhdy.org/kf/2003-11-12.xsl",
			helper.stylesheet( "BUHDY", KF1 )
		);
		assertEquals(
			"multi kfidx",
			"http://www.buhdy.org/2003-11-12/secret/secret.xsd",
			helper.stylesheet( "BUHDY", BUHDY_SECRET )
		);
		assertNull(
			"unknown",
			helper.stylesheet( "UNKNOWN", KF )
		);
		assertEquals(
			"still fine: " + errorHandler.getMessage(),
			ErrorWatcher.OK,
			errorHandler.getSeverity()
		);
	}

	public void testDeprecatedUriProperty () {
		System.setProperty(
			KeyFamilyURIMapping.class.getName() + ".uri.BUHDY",
			"http://www.bodi.org/public-vocabulary"
		);
		System.setProperty(
			KeyFamilyURIMapping.class.getName() + ".uri.BUHDY.BUHDY_SECRET",
			"https://www.bodi.org/secret-vocabulary/3.5/sec"
		);

		final KeyFamilyURIMapping helper = KeyFamilyURIMapping.getInstance();
		assertEquals(
			"no strip",
			"http://www.bodi.org/public-vocabulary/2003-11-12/kf",
			helper.xmlns( "BUHDY", KF )
		);
		assertEquals(
			"number strip",
			"http://www.bodi.org/public-vocabulary/2003-11-12/kf",
			helper.xmlns( "BUHDY", KF1 )
		);
		assertEquals(
			"agency strip",
			"http://www.bodi.org/public-vocabulary/2003-11-12/kf",
			helper.xmlns( "BUHDY", BUHDY_KF )
		);
		assertEquals(
			"agency and number strip",
			"http://www.bodi.org/public-vocabulary/2003-11-12/kf",
			helper.xmlns( "BUHDY", BUHDY_KF2 )
		);
		assertEquals(
			"exact override",
			"https://www.bodi.org/secret-vocabulary/3.5/sec",
			helper.xmlns( "BUHDY", BUHDY_SECRET )
		);
		assertEquals(
			"still fine: " + errorHandler.getMessage(),
			ErrorWatcher.WARNING,
			errorHandler.getSeverity()
		);

		System.getProperties().remove(
			KeyFamilyURIMapping.class.getName() + ".uri.BUHDY"
		);
		System.getProperties().remove(
			KeyFamilyURIMapping.class.getName() + ".uri.BUHDY.BUHDY_SECRET"
		);
	}

	public void testFailingLookup () {
		final KeyFamilyURIMapping helper = KeyFamilyURIMapping.getInstance();
		final Iterator iterator
		= helper.lookup(
			"http://www.bodi.org/vocabulary/2003-11-12/secret"
		);
		assertTrue( "no match", !iterator.hasNext() );
	}

	public void testUniqueLookup () {
		final KeyFamilyURIMapping helper = KeyFamilyURIMapping.getInstance();
		final Iterator iterator
		= helper.lookup(
			"https://www.buhdy.org/secret-namespace/3.5/sec"
		);
		assertTrue( "exact one", iterator.hasNext() );
		assertEquals(
			"exact match",
			new ConcreteLookupResult(
				"BUHDY",
				BUHDY_SECRET
			),
			iterator.next()
		);
		assertTrue( "no more", !iterator.hasNext() );
	}

	public void testAmbiguousLookup () {
		final KeyFamilyURIMapping helper = KeyFamilyURIMapping.getInstance();
		final Iterator iterator
		= helper.lookup(
			"http://www.buhdy.org/star-namespace/kf"
		);
		final KeyFamilyURIMapping.LookupResult
			result1 = new ConcreteLookupResult(
				"BUHDY",
				KF
			),
			result2 =
			new ConcreteLookupResult(
				"BUHDY",
				KF1
			);
		assertTrue( "first", iterator.hasNext() );
		final KeyFamilyURIMapping.LookupResult first
		= (KeyFamilyURIMapping.LookupResult) iterator.next();
		assertTrue(
			"first match",
			first.equals( result1 ) || first.equals( result2 )
		);
		assertTrue( "second", iterator.hasNext() );
		final KeyFamilyURIMapping.LookupResult second
		= (KeyFamilyURIMapping.LookupResult) iterator.next();
		assertTrue(
			"second match",
			second.equals( result1 ) || second.equals( result2 )
		);
		assertTrue(
			"not same",
			! first.equals( second )
		);
		assertTrue( "no more", !iterator.hasNext() );
	}

	public void testAmbiguousAccrossMAsLookup () {
		final KeyFamilyURIMapping helper = KeyFamilyURIMapping.getInstance();
		final Iterator iterator
		= helper.lookup(
			"http://www.bodi.org/shared-vocabulary/3.4/kf2"
		);
		final KeyFamilyURIMapping.LookupResult
			result1 = new ConcreteLookupResult(
				"BUDDY",
				SHARED
			),
			result2 = new ConcreteLookupResult(
				"BUHDY",
				SHARED
			);
		assertTrue( "first", iterator.hasNext() );
		final KeyFamilyURIMapping.LookupResult first
		= (KeyFamilyURIMapping.LookupResult) iterator.next();
		assertTrue(
			"first match",
			first.equals( result1 ) || first.equals( result2 )
		);
		assertTrue( "second", iterator.hasNext() );
		final KeyFamilyURIMapping.LookupResult second
		= (KeyFamilyURIMapping.LookupResult) iterator.next();
		assertTrue(
			"second match",
			second.equals( result1 ) || second.equals( result2 )
		);
		assertTrue(
			"not same",
			! first.equals( second )
		);
		assertTrue( "no more", !iterator.hasNext() );
	}
	
	public void testSDMXRegistryConventions() {
	    System.setProperty(
            KeyFamilyURIMapping.class.getName() + ".useSDMXNamingConventions",
            "true"
        );
	    final KeyFamilyURIMapping helper = KeyFamilyURIMapping.getInstance();
	    assertEquals(
            "Should use SDMX Registry naming conventions",
            "urn:sdmx:org.sdmx.infomodel.keyfamily.KeyFamily=BUHDY:KF",
            helper.xmlns( "BUHDY", KF )
        );
	    
	    System.getProperties().remove(
            KeyFamilyURIMapping.class.getName() + ".useSDMXNamingConventions"
        );
	}
}
