// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (C) 2005-2006, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.xml;

import java.io.IOException;
import java.io.OutputStream;
import java.util.Comparator;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import Int.ecb.gesmescb.Attribute;
import Int.ecb.gesmescb.CodeList;
import Int.ecb.gesmescb.Contact;
import Int.ecb.gesmescb.ContactChannel;
import Int.ecb.gesmescb.InterchangeBegin;
import Int.ecb.gesmescb.InterchangeEnd;
import Int.ecb.gesmescb.KeyFamily;
import Int.ecb.gesmescb.KeyFamilyComponent;
import Int.ecb.gesmescb.MessageBegin;
import Int.ecb.gesmescb.MessageEnd;
import Int.ecb.gesmescb.Observation;
import Int.ecb.gesmescb.StatisticalConcept;
import Int.ecb.gesmescb.TextFormat;
import Int.ecb.gesmescb.TimeSeries;
import Int.ecb.gesmescb.event.DataListener;
import Int.ecb.gesmescb.event.StructureListener;
import Int.ecb.gesmescb.metabase.MaintenanceAgency;
import Int.ecb.gesmescb.metabase.Metabase;
import Int.ecb.gesmescb.time.Interval;
import Int.ecb.gesmescb.time.RegularInterval;
import Int.ecb.gesmescb.time.TimeZoneUtil;
import Int.ecb.gesmescb.util.KeyTokenizer;
import Int.ecb.logging.Level;
import Int.ecb.gesmescb.util.ComponentHelper;

/* ***************************** Class ****************************** */

/**
 * <p>
 * Translates {@link Int.ecb.gesmescb.GESMESCBObject}s into
 * compact/StructureSpecific SDMX-ML text. This text then is written to an
 * output {@link java.io.Writer} or to a file according to preference settings.
 * </p>
 * 
 * <center>
 * <table border="3" cellpadding="3">
 * <caption> <a name="prefs">Available {@link Int.ecb.prefs preference}
 * settings</a> </caption>
 * <tr>
 * <th>preference name</th>
 * <th>type</th>
 * <th>meaning</th>
 * <th>default</th>
 * </tr>
 * <tr>
 * <td><code>destination</td>
 * <td>string</td>
 * <td>output file path</td>
 * </tr>
 * <tr>
 * <td><code>encoding</td>
 * <td>string</td>
 * <td>character encoding</td>
 * <td>platform default, which may not be understood by XML parsers; better use
 * UTF-8 or UTF-16</td>
 * </tr>
 * <tr>
 * <td><code>indent</td>
 * <td>string or integer</td>
 * <td>if integer, then added indentation for enclosed elements; if "
 * <tt>tab</tt>", then use 1 tab character to indent</td>
 * <td><tt>tab</tt></td>
 * </tr>
 * <tr>
 * <td><code>sdmxVersion</code></td>
 * <td>string</td>
 * <td>SDMX-ML syntax version</td>
 * <td><code>1.0</code></td>
 * </tr>
 * <tr>
 * <td><code>message.schema</code></td>
 * <td>string</td>
 * <td>URI to the SDMX-ML message namespace schema</td>
 * </tr>
 * <tr>
 * <td><code>common.schema</code></td>
 * <td>string</td>
 * <td>URI to the SDMX-ML common namespace schema</td>
 * </tr>
 * <tr>
 * <td><code>structure.schema</code></td>
 * <td>string</td>
 * <td>URI to the SDMX-ML structure namespace schema</td>
 * </tr>
 * <tr>
 * <td><code>defaultObsConf</code></td>
 * <td>string</td>
 * <td>What to write when OBS_CONF is empty. When this option is not set, the
 * XML attribute will be suppressed.</td>
 * <td>not set</td>
 * </tr>
 * <tr>
 * <td><code>defaultObsPreBreak</code></td>
 * <td>string</td>
 * <td>What to write when OBS_PRE_BREAK is empty. When this option is not set,
 * the XML attribute will be suppressed.</td>
 * <td>not set</td>
 * </tr>
 * </table>
 * </center>
 * 
 * <p>
 * The <code>SDMXMLWriter</code> is not internally synchronized. External
 * synchronisation is necessary if multiple threads want to use the same
 * <code>SDMXMLWriter</code>.
 * </p>
 * 
 * @author <a href="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard
 *         Bodenstorfer</a>
 * @author <a href="mailto:Paula.Rocha@ECB.int">Paula Rocha</a>
 * 
 * @version @(#)$Id: SDMXMLWriter.java,v 1.29 2014/10/01 12:45:25 trainot Exp $
 * 
 * @see KeyFamilyURIMapping
 */
public class SDMXMLWriter extends XMLWriter implements StructureListener,
		DataListener {
	/* ======================= Constants ======================== */

	/** Status code for documentBegin not issued yet. */
	private static final int OUTSIDE_ROOT_TAG = -1;

	/** Status code for message content: no child of document element */
	private static final int NONE = 0;

	/** Status code: open element CodeLists */
	private static final int CL = 1;

	/** Status code: open element Concepts */
	private static final int SC = 2;

	/** Status code: open element KeyFamilies */
	private static final int KF = 3;

	/** Status code: open tag DataSet */
	private static final int DS = 4;

	/** Status code: open element, but closed tag DataSet (only inside loop) */
	private static final int DSC = 5;

	/** Status code: open tag Group */
	private static final int DS_SG = 6;

	/** Status code: open tag Series */
	private static final int DS_TS = 8;

	/** Status code: open element, but closed tag Series (only inside loop) */
	private static final int DS_TSC = 9;

	/** Status code: open tag Obs */
	private static final int DS_TS_OBS = 10;

	/** Status code: added series attribute, nothing to do */
	private static final int DS_ADD_SER_ATTR = 11;

	/* namespace constants */

	/** Namespace message */
	private static final String MESSAGE_NS = "message";

	/** Namespace common */
	private static final String COMMON_NS = "common";

	/** Default entry point for schema locations */
	private String defaultEntryPoint;

	/** SDMX/ML version */
	private String sdmxVersion = "";

	/* ========================= Fields ========================= */

	/**
	 * Current XML dateTime format. Changes its pattern according to frequency.
	 */
	private final XMLDateFormat xmlDateFormat;

	/** Links key families to their namespace URIs. */
	private final KeyFamilyURIMapping uriMapping;

	/** Controls SDMX syntax version-specific features. */
	private final SDMXVersionFeatures features;

	/**
	 * What to write when OBS_CONF is empty. When this option is not set, the
	 * attribute will be suppressed.
	 */
	private final String defaultObsConf;

	/**
	 * What to write when OBS_PRE_BREAK is empty. When this option is not set,
	 * the attribute will be suppressed.
	 */
	private final String defaultObsPreBreak;

	/** The current interchange. */
	private InterchangeBegin currentInterchangeBegin;

	private String documentElement;

	/** The current maintenance agency id. */
	private String currentMaintenanceAgencyId;

	/** The current <code>KeyFamily</code>. */
	private KeyFamily keyFamily;

	private String keyFamilyNamespace;
	private String keyFamilySchema;

	/** Extra settings for sdmx20c */
	private String sdmxDataStructureNamespace;
	private String sdmxDataStructureSchemaLocation;

	/**
	 * Comparator for sorting the Concept identifiers in the current
	 * {@link #keyFamily}.
	 */
	private Comparator conceptIdComparator;

	/** Whether message action is delete. */
	private boolean isDeleteMessage;

	/** Status for cube render automaton. */
	private int status;

	/** The current sibling group key. Only valid if SG is in status. */
	private String sgKey;

	/** The current sibling group key. Only valid if TS is in status. */
	private String tsKey;

	/** The current observation "key". Only valid if OBS is in status. */
	private RegularInterval obsInterval;

	/**
	 * Whether the time format has been written for a Series. Only valid if TS
	 * is in status.
	 */
	private boolean timeFormatWritten;

	/** Time zone util sanitizing timestamps */
	private final TimeZoneUtil tsUtil;

	/** namespace shorthands to prepend to elements in v21 */
	private String prefix;

	/** temporarily save the old prefix */
	private String oldPrefix;

	/** keep current action for handle(TimeSeries) **/
	private String currentMessageBeginAction;

	/** keep current StructureRef for handle(TimeSeries) **/
	private String currentStructureRef;

	/** whether to allow multiple messages in one interchange */
	private boolean allowMultipleMessages;

	/** map of key positions key -> position in output stream */
	private Map keyPositionsMap = new HashMap();

	/** extra spaces for attributes at series level */
	private int extraSpacesCount = 0;

	/* ====================== Constructors ====================== */

	public SDMXMLWriter(final String name,
			final RandomFileOutputStream outputStream) throws IOException {
		super(name, outputStream);
		String v = initialize();
		features = "1.0".equals(v) ? (SDMXVersionFeatures) new Version1Features()
				: "2.0".equals(v) ? (SDMXVersionFeatures) new Version2Features()
						: "2.1".equals(v) ? (SDMXVersionFeatures) new Version21Features()
								: (SDMXVersionFeatures) new Version1Features();
		defaultObsConf = componentHelper.getPreference("defaultObsConf", null);
		defaultObsPreBreak = componentHelper.getPreference(
				"defaultObsPreBreak", null);
		defaultEntryPoint = componentHelper.getPreference("defaultEntryPoint",
				"https://sreg.escb.eu/ws/rest/schema/datastructure/");
		xmlDateFormat = new XMLDateFormat();
		if ("2.1".equals(v)) {
			xmlDateFormat.setVersion("sdmx21");
		} else {
			xmlDateFormat.setVersion("sdmx20c");
		}
		sdmxVersion = v;
		uriMapping = KeyFamilyURIMapping.getInstance();
		tsUtil = TimeZoneUtil.getInstance();

	}

	/**
	 * Construct an <code>SDMXMLWriter</code> to write to a file.
	 * 
	 * @param name
	 *            used for logging and preferences.
	 * @param filename
	 *            where to write to.
	 * 
	 * @throws NullPointerException
	 *             if <code>filename==null</code> and the preference variable
	 *             <code>destination</code> is undefined.
	 * @throws IOException
	 *             if the file cannot be opened for writing.
	 */
	public SDMXMLWriter(final String name, final String filename)
			throws IOException {
		super(name, filename);
		String v = initialize();
		features = "1.0".equals(v) ? (SDMXVersionFeatures) new Version1Features()
				: "2.0".equals(v) ? (SDMXVersionFeatures) new Version2Features()
						: "2.1".equals(v) ? (SDMXVersionFeatures) new Version21Features()
								: (SDMXVersionFeatures) new Version1Features();
		defaultObsConf = componentHelper.getPreference("defaultObsConf", null);
		defaultObsPreBreak = componentHelper.getPreference(
				"defaultObsPreBreak", null);
		defaultEntryPoint = componentHelper.getPreference("defaultEntryPoint",
				"https://sreg.escb.eu/ws/rest/schema/datastructure/");
		xmlDateFormat = new XMLDateFormat();
		if ("2.1".equals(v)) {
			xmlDateFormat.setVersion("sdmx21");
		} else {
			xmlDateFormat.setVersion("sdmx20c");
		}
		sdmxVersion = v;
		uriMapping = KeyFamilyURIMapping.getInstance();
		tsUtil = TimeZoneUtil.getInstance();

	}

	/**
	 * Construct an <code>SDMXMLWriter</code> writing to a file according to
	 * preferences.
	 * 
	 * @param name
	 *            used for logging and preferences.
	 * 
	 * @throws NullPointerException
	 *             if the preference variable <code>destination</code> is
	 *             undefined.
	 * @throws IOException
	 *             if the file cannot be opened for writing.
	 */
	public SDMXMLWriter(final String name) throws IOException {

		this(name, (String) null);

	}

	private String initialize() {

		final String versionString = componentHelper.getPreference(
				"sdmxVersion", "1.0");
		boolean isV1 = "1.0".equals(versionString);
		boolean isV2 = "2.0".equals(versionString);
		boolean isV21 = "2.1".equals(versionString);
		if (!isV1 && !isV2 && !isV21) {
			componentHelper.logger.log(Level.SEVERE, componentHelper.logName,
					"SDMXMLWriter()", "Unknown SDMX syntax version \""
							+ versionString + "\" configured.");
			isV1 = true;
		} else {
			componentHelper.logger.log(Level.CONFIG, componentHelper.logName,
					"SDMXMLWriter()",
					"Instantiating SDMXMLWriter version is \"" + versionString
							+ "\".");
		}
		currentInterchangeBegin = null;
		documentElement = null;
		currentMaintenanceAgencyId = null;
		keyFamily = null;
		conceptIdComparator = null;
		isDeleteMessage = false;
		keyFamilyNamespace = keyFamilySchema = null;
		status = OUTSIDE_ROOT_TAG;
		tsKey = sgKey = null;
		timeFormatWritten = false;
		allowMultipleMessages = "true".equalsIgnoreCase(componentHelper
				.getPreference("allowMultipleMessages", "false"));
		defaultEntryPoint = componentHelper.getPreference("defaultEntryPoint",
				"https://sreg.escb.eu/ws/rest/schema/datastructure/");
		writeXMLHeader();
		return isV1 ? "1.0" : isV2 ? "2.0" : isV21 ? "2.1" : "1.0";
	}

	private void addExtraSpacesForAttributes(String key) {
		try {
			// add extra space for attributes
			writer.write(new String(new char[extraSpacesCount]).replace('\0',
					' '));
			writer.flush();
		} catch (IOException e) {
			componentHelper.logger.log(Level.SEVERE, componentHelper.logName,
					"addExtraSpacesForAttributes",
					"Could not add extra spaces to file");
		}
	}

	/* ===================== Public Methods ===================== */

	/* ------------------------ Listener ------------------------ */

	/* ---------------------- Interchange ----------------------- */

	public void handle(final InterchangeBegin interchangeBegin) {
		componentHelper.logger.log(Level.INFO, componentHelper.logName,
				"handle( InterchangeBegin )", "Begin writing interchange "
						+ interchangeBegin.getIref() + ".");

		// Store for use in handle( MessageBegin )
		currentInterchangeBegin = interchangeBegin;

	}

	public void handle(final InterchangeEnd interchangeEnd) {
		try {
			// changeToPrefix(MESSAGE_NS);
			closeElement(documentElement);
			close();
		} catch (final IOException ioe) {
			componentHelper.logger.log(Level.SEVERE, componentHelper.logName,
					"handle( InterchangeEnd )", "Output problem.", ioe);
			close();
		}
		componentHelper.logger.log(Level.INFO, componentHelper.logName,
				"handle( InterchangeEnd )", "Ended writing interchange "
						+ interchangeEnd.getIref() + ".");
	}

	/* ------------------------ Message ------------------------- */

	public void handle(final MessageBegin messageBegin) {

		if (documentElement != null) {

			if (features instanceof Version21Features && !allowMultipleMessages) {
				componentHelper.logger.log(Level.SEVERE,
						componentHelper.logName, "handle( MessageBegin )",
						"SDMXMLWriter cannot translate interchanges with"
								+ " more than one message.", messageBegin);

				throw new RuntimeException(
						"Multiple message not allowed by default, please set "
								+ "SDMXMLWriter.allowMultipleMessages to true if needed.");
			}
		}

		componentHelper.logger.log(Level.FINE, componentHelper.logName,
				"handle( MessageBegin )", "Begin writing Message "
						+ messageBegin.getMref() + ".");

		currentMaintenanceAgencyId = messageBegin.getMaintenanceAgencyId();

		isDeleteMessage = messageBegin.isDeleteAction();
		if (!"".equals(messageBegin.getKeyFamilyId())) {
			final Metabase metabase = Metabase.getMetabase();
			final MaintenanceAgency maintenanceAgency = metabase
					.getMaintenanceAgency(currentMaintenanceAgencyId);
			if (maintenanceAgency != null) {
				final String keyFamilyId = messageBegin.getKeyFamilyId();
				keyFamily = maintenanceAgency.getKeyFamily(keyFamilyId);
				if (keyFamily != null) {
					conceptIdComparator = new ConceptIdComparator(keyFamily);
					/** initialise count of extra spaces */
					Iterator it = keyFamily.componentIterator();
					while (it.hasNext()) {
						KeyFamilyComponent currentComponent = (KeyFamilyComponent) it
								.next();
						if (currentComponent.isAttachedToTimeSeries()) {
							extraSpacesCount += currentComponent.getConceptId()
									.length() + 2; // +1 for equal sign, +1 for
													// extra space
							extraSpacesCount += currentComponent
									.getTextFormat().getReferenceLength() + 2; // +2
																				// for
																				// quotation
																				// marks
						}
					}
					componentHelper.logger.log(Level.FINE,
							componentHelper.logName, "handle( MessageBegin )",
							"Extra spaces for attributes:" + extraSpacesCount);
					/** /** Finished initialising count of extra spaces */
				} else {
					componentHelper.logger.log(Level.SEVERE,
							componentHelper.logName, "handle( MessageBegin )",
							"Maintenance agengy " + currentMaintenanceAgencyId
									+ " does not know key family \""
									+ keyFamilyId + "\".");
					close();
				}
			} else {
				componentHelper.logger.log(Level.SEVERE,
						componentHelper.logName, "handle( MessageBegin )",
						"Unknown maintenance agengy \""
								+ currentMaintenanceAgencyId + "\".");
				close();
			}
			keyFamilyNamespace = uriMapping.xmlns(currentMaintenanceAgencyId,
					keyFamily);
			keyFamilySchema = uriMapping.schemaLocation(
					currentMaintenanceAgencyId, keyFamily);
		} else {
			keyFamily = null;
			conceptIdComparator = null;
			keyFamilyNamespace = keyFamilySchema = null;
		}

		if (features instanceof Version2Features) {
			/**
			 * This namespace definition is conform with ESCB registry namespace
			 * !!
			 */
			sdmxDataStructureNamespace = "urn:sdmx:org.sdmx.infomodel.datastructure.DataStructure="
					+ currentMaintenanceAgencyId
					+ ":"
					+ messageBegin.getKeyFamilyId() + "(1.0):compact";
			sdmxDataStructureSchemaLocation = defaultEntryPoint
					+ currentMaintenanceAgencyId + "/"
					+ messageBegin.getKeyFamilyId() + "/1.0/?version=2.0";
		}

		if (features instanceof Version21Features) {
			/**
			 * This namespace definition is conform with ESCB registry namespace
			 * !!
			 */
			sdmxDataStructureNamespace = "urn:sdmx:org.sdmx.infomodel.datastructure.DataStructure="
					+ currentMaintenanceAgencyId
					+ ":"
					+ messageBegin.getKeyFamilyId()
					+ "(1.0):ObsLevelDim:TIME_PERIOD";
			sdmxDataStructureSchemaLocation = defaultEntryPoint
					+ currentMaintenanceAgencyId + "/"
					+ messageBegin.getKeyFamilyId() + "/1.0/?version=2.1";
		}

		currentMessageBeginAction = getActionFromMessageBegin(messageBegin);

		if (status == OUTSIDE_ROOT_TAG) {

			try {
				writeDocumentElement(messageBegin);
			} catch (final IOException ioe) {
				componentHelper.logger.log(Level.SEVERE,
						componentHelper.logName, "handle( MessageBegin )",
						"Output problem on documentElement.", ioe);
				close();
			}
			try {
				writeDocumentHeader(messageBegin);
			} catch (final IOException ioe) {
				componentHelper.logger.log(Level.SEVERE,
						componentHelper.logName, "handle( MessageBegin )",
						"Output problem on Header.", ioe);
				close();
			}
		}
	}

	public void handle(final MessageEnd messageEnd) {
		try {
			closeAllContentElements();
			componentHelper.logger.log(Level.FINE, componentHelper.logName,
					"handle( MessageEnd )",
					"End writing message " + messageEnd.getMref() + ".");

			/** remove all extra spaces from Series elements */
			removeExtraSpacesFromSeriesTagsAndCleanAnomalies();
		} catch (final IOException ioe) {
			componentHelper.logger.log(Level.SEVERE, componentHelper.logName,
					"handle( MessageEnd )", "Output problem.", ioe);
			close();
		}
	}

	/* ----------------------- Code List ------------------------ */

	public void handle(final CodeList codeList) {
		final String id = codeList.getId();
		try {
			if (status != CL) {
				closeAllContentElements();
				openElement("message:CodeLists");
				writeAttribute("xmlns", features.getXmlnsSdmxStructure());
				status = CL;
			}

			openElement("CodeList");
			writeAttribute(features.getAgencyRefAttributeName(),
					currentMaintenanceAgencyId);
			writeAttribute("id", id);

			final String name = codeList.getName();
			if (!"".equals(name)) {
				writeElement("Name", name);
			}

			final String description = codeList.getDescription();
			if (!"".equals(description)) {
				writeElement("Description", description);
			}

			for (final Iterator values = codeList.codeValueIterator(), descriptions = codeList
					.codeDescriptionIterator(); values.hasNext()
					&& descriptions.hasNext();) {
				final String value = (String) values.next();
				openElement("Code");
				writeAttribute("value", value);

				final String desc = (String) descriptions.next();
				if (!"".equals(desc)) {
					writeElement("Description", desc);
				}

				closeElement("Code");
			}

			closeElement("CodeList");
		} catch (final IOException ioe) {
			componentHelper.logger.log(Level.SEVERE, componentHelper.logName,
					"handle( CodeList )", "Output problem.", ioe);
			close();
		}
	}

	/* ------------------ Statistical Concept ------------------- */

	public void handle(final StatisticalConcept statisticalConcept) {
		final String id = statisticalConcept.getId();
		try {
			if (status != SC) {
				closeAllContentElements();
				openElement("message:Concepts");
				writeAttribute("xmlns", features.getXmlnsSdmxStructure());
				status = SC;
			}

			openElement("Concept");
			writeAttribute(features.getAgencyRefAttributeName(),
					currentMaintenanceAgencyId);
			writeAttribute("id", id);

			final String name = statisticalConcept.getName();
			if (!"".equals(name)) {
				writeElement("Name", name);
			}

			final String description = statisticalConcept.getDescription();
			if (!"".equals(description)) {
				writeElement("Description", description);
			}

			closeElement("Concept");
		} catch (final IOException ioe) {
			componentHelper.logger.log(Level.SEVERE, componentHelper.logName,
					"handle( StatisticalConcept )", "Output problem.", ioe);
			close();
		}
	}

	/* ----------------------- Key Family ----------------------- */

	public void handle(final KeyFamily keyFamily) {
		this.keyFamily = keyFamily;
		keyFamilyNamespace = uriMapping.xmlns(currentMaintenanceAgencyId,
				keyFamily);
		// unnecessary to set conceptIdSeparator here.

		final String id = keyFamily.getId();
		try {
			if (status != KF) {
				closeAllContentElements();
				openElement("message:KeyFamilies");
				writeAttribute("xmlns", features.getXmlnsSdmxStructure());
				status = KF;
			}

			openElement("KeyFamily");
			writeAttribute(features.getAgencyRefAttributeName(),
					currentMaintenanceAgencyId);
			writeAttribute("id", id);
			if (keyFamilyNamespace != null) {
				writeAttribute("uri", keyFamilyNamespace);
			}

			final String name = keyFamily.getName();
			if (!"".equals(name)) {
				writeElement("Name", name);
			}

			final String description = keyFamily.getDescription();
			if (!"".equals(description)) {
				writeElement("Description", description);
			}

			openElement("Components");

			for (final Iterator iterator = keyFamily.keyStructureIterator(); iterator
					.hasNext();) {
				final KeyFamilyComponent component = (KeyFamilyComponent) iterator
						.next();
				write(component);
			}
			write(keyFamily.getTimePeriod());

			if (keyFamily.attributeIterator(
					KeyFamilyComponent.ATTACHED_TO_SIBLING_GROUP).hasNext()) {
				openElement("Group");

				writeAttribute("id", "Group");
				for (final Iterator iterator = keyFamily.keyStructureIterator(); iterator
						.hasNext();) {
					final KeyFamilyComponent component = (KeyFamilyComponent) iterator
							.next();
					if (!component.isFrequencyDimension()) {
						writeElement("DimensionRef", component.getConceptId());
					}
				}

				closeElement("Group");
			}

			write(keyFamily.getObsValue());
			write(keyFamily.getTimeFormat());
			for (final Iterator iterator = keyFamily.arrayCellIterator(); iterator
					.hasNext();) {
				final KeyFamilyComponent component = (KeyFamilyComponent) iterator
						.next();
				if (component != keyFamily.getObsValue()) {
					write(component);
				}
			}
			for (final Iterator iterator = keyFamily.attributeIterator(); iterator
					.hasNext();) {
				final KeyFamilyComponent component = (KeyFamilyComponent) iterator
						.next();
				write(component);
			}

			closeElement("Components");

			closeElement("KeyFamily");
		} catch (final IOException ioe) {
			componentHelper.logger.log(Level.SEVERE, componentHelper.logName,
					"handle( KeyFamily )", "Output problem.", ioe);
			close();
		}
	}

	/* ---------------------- Time Series ----------------------- */

	public void handle(final TimeSeries timeSeries) {
		final String key = timeSeries.getKey();
		final RegularInterval range = timeSeries.getInterval();
		final RegularInterval period = range != null ? range
				.computeBasePeriod(0) : null;
		final int obsCount = range != null ? range.getSize() : 0;
		if (!isDeleteMessage && obsCount == 0)
			return;
		try {
			// This loop rattles through a couple of state changes.
			do {
				switch (status) {
				case NONE:
					openElementDataSet();
					status = DS;
					componentHelper.logger.log(Level.FINE,
							componentHelper.logName, "handle( TimeSeries )",
							"status NONE for key: \"" + key + "\".");
					break;
				case DS:
				case DSC:
					openElement("Series", conceptIdComparator);
					writeKey(key);

					if (range != null && !isDeleteMessage) {
						writeTimeFormat(range);
						timeFormatWritten = true;
					} else {
						timeFormatWritten = false;
					}
					tsKey = key;

					status = DS_TS;
					componentHelper.logger.log(Level.FINE,
							componentHelper.logName, "handle( TimeSeries )",
							"status DS, DSC for key: \"" + key + "\".");

					buffer.writeAttributeCache();
					writer.flush();

					// remember current position
					long currentPosition = outputStream.getFilePointer();
					// write extra spaces for attributes after the time format
					keyPositionsMap.put(key, new Long(currentPosition));
					componentHelper.logger.log(Level.FINE,
							componentHelper.logName, "handle( TimeSeries )",
							"Added key: \"" + key + "\" at position: "
									+ outputStream.getFilePointer());

					addExtraSpacesForAttributes(key);
					break;
				case DS_SG:
					closeElement("Group");
					status = DSC;
					componentHelper.logger.log(Level.FINE,
							componentHelper.logName, "handle( TimeSeries )",
							"status DS_SG, close group for key: \"" + key
									+ "\".");

					break;
				case DS_TS:
					if (key.equals(tsKey) && range != null && !isDeleteMessage
							&& !timeFormatWritten) {
						writeTimeFormat(range);
						timeFormatWritten = true;

					}
					if (obsCount != 0) {
						status = DS_TSC;
					}
					componentHelper.logger.log(Level.FINE,
							componentHelper.logName, "handle( TimeSeries )",
							"status DS_TS for key: \"" + key + "\".");
					break;
				case DS_TSC:
					if (!key.equals(tsKey) || range != null
							&& !timeFormatWritten && !isDeleteMessage) {
						closeElement("Series");
						status = DSC;
					} else {
						if (obsCount > 0) {
							openElement("Obs", conceptIdComparator);
							writeTimePeriod(period);
							obsInterval = period;
							status = DS_TS_OBS;
						}
					}
					componentHelper.logger.log(Level.FINE,
							componentHelper.logName, "handle( TimeSeries )",
							"status DS_TSC for key: \"" + key + "\".");

					break;
				case DS_TS_OBS:
					if (!key.equals(tsKey)) {

						closeElement("Obs");
						closeElement("Series");
						status = DSC;
					} else if (!period.equals(obsInterval)) {
						closeElement("Obs");
						status = DS_TSC;
					}
					componentHelper.logger.log(Level.FINE,
							componentHelper.logName, "handle( TimeSeries )",
							"status DS_TS_OBS for key: \"" + key + "\".");

					break;
				}
			} while (obsCount == 0 && status != DS_TS || obsCount != 0
					&& status != DS_TS_OBS);

			if (obsCount > 0) {
				for (final Iterator intervals = range.iterator(), observations = timeSeries
						.observationIterator(); intervals.hasNext()
						&& (observations.hasNext() || isDeleteMessage);) {
					final RegularInterval interval = (RegularInterval) intervals
							.next();
					final Observation observation = observations.hasNext() ? (Observation) observations
							.next() : null;

					if (observation != null && observation.isEmpty()
							&& !isDeleteMessage)
						continue;

					if (!interval.equals(obsInterval)) {
						closeElement("Obs");
						openElement("Obs", conceptIdComparator);
						writeTimePeriod(interval);
						obsInterval = interval;
					}
					if (observation != null) {
						writeObservation(observation);
					}
				}
			}
		} catch (final IOException ioe) {
			componentHelper.logger.log(Level.SEVERE, componentHelper.logName,
					"handle( TimeSeries )", "Output problem.", ioe);
			close();
		}
	}

	/* ----------------------- Attribute ------------------------ */

	public void handle(final Attribute attribute) {
		final String key = attribute.getKey();
		final RegularInterval range = attribute.getInterval();
		final String name = attribute.getName();
		final String value = attribute.getValue();
		final KeyFamilyComponent component = keyFamily.getComponent(name);

		if (component == null) {
			componentHelper.logger.log(
					Level.SEVERE,
					componentHelper.logName,
					"handle( Attribute )",
					"No concept \"" + name + "\" in key family \""
							+ keyFamily.getId() + "\".", attribute);
			close();
		}

		try {
			// This loop rattles through a couple of state changes.
			loop: while (!stateMatches(key, range, component)) {
				switch (status) {
				case NONE:
					openElementDataSet();
					status = DS;
					componentHelper.logger.log(Level.FINE,
							componentHelper.logName, "handle( Atttribute )",
							"status NONE for attribute: \"" + name
									+ "\" and key: \"" + key + "\"");

					break;
				case DS:
				case DSC:
					if (component.isAttachedToDataSet()) {
						changeToPrefix(MESSAGE_NS);
						closeElement("DataSet");
						status = NONE;
					} else if (component.isAttachedToObservation()) {
						openElement("Series", conceptIdComparator);
						writeKey(key);

						if (range != null && !isDeleteMessage) {
							writeTimeFormat(range);
							timeFormatWritten = true;
						} else {
							timeFormatWritten = false;
						}
						tsKey = key;
						status = DS_TS;
						componentHelper.logger.log(Level.FINE,
								componentHelper.logName,
								"handle( Atttribute )",
								"status DS, DSC for attribute: \"" + name
										+ "\" and key: \"" + key + "\"");
					} else if (component.isAttachedToTimeSeries()) {

						writer.flush();

						Object positionObj = keyPositionsMap.get(key);

						/** CASE ATTRIBUTE AT SERIES LEVEL WITH ATTACHED SERIES */
						if (positionObj != null) {
							long currentPosition = outputStream
									.getFilePointer();
							long seriesPosition = Long.parseLong(positionObj
									.toString());

							outputStream.getRandomFile().seek(seriesPosition);
							writer.write(" " + name + "=\"" + value + "\""); 
							writer.flush();

							// update the key position in the map
							keyPositionsMap.put(key,
									new Long(outputStream.getFilePointer()));
							// return to current position
							outputStream.getRandomFile().seek(currentPosition);

							// openElement( "Series", conceptIdComparator );
							// writeKey( key );
							/*
							 * if ( range != null && ! isDeleteMessage ) {
							 * writeTimeFormat( range ); timeFormatWritten =
							 * true; } else { timeFormatWritten = false; }
							 */
							tsKey = key;
							// status = DS_TS;
							status = DS_ADD_SER_ATTR;
							componentHelper.logger.log(Level.FINE,
									componentHelper.logName,
									"handle( Atttribute )",
									"status DS, DSC for attribute: \"" + name
											+ "\" and key: \"" + key + "\"");

							break loop;
						} else {
							/** CASE ATTRIBUTE AT SERIES LEVEL WITHOUT ATTACHED SERIES */
							openElement("Series", conceptIdComparator);
							writeKey(key);

							if (range != null && !isDeleteMessage) {
								writeTimeFormat(range);
								timeFormatWritten = true;
							} else {
								timeFormatWritten = false;
							}
							tsKey = key;
							status = DS_TS;
							componentHelper.logger.log(Level.FINE,
									componentHelper.logName,
									"handle( Atttribute )",
									"status DS, DSC for attribute: \"" + name
											+ "\" and key: \"" + key + "\"");
							break;
						} //END PROCESSING ATTRIBUTE ATTACHED AT SERIES LEVEL 
						
					} else if (component.isAttachedToSiblingGroup()) {
						componentHelper.logger.log(Level.FINE,
								componentHelper.logName, "handle( Attribute )",
								"opening Group with key \"" + key + "\".");
						openElement("Group", conceptIdComparator);
						writeKey(key);
						sgKey = key;
						status = DS_SG;
					}
					break;
				case DS_SG:
					closeElement("Group");
					status = DSC;
					componentHelper.logger.log(Level.FINE,
							componentHelper.logName, "handle( Atttribute )",
							"status DS_SG for attribute: \"" + name
									+ "\" and key: \"" + key + "\"");
					break;
				case DS_TS:
				case DS_TSC:
					if (component.isAttachedToObservation()
							&& key.equals(tsKey)) {
						if (!isDeleteMessage && !timeFormatWritten) {
							writeTimeFormat(range);
							timeFormatWritten = true;
						}
						openElement("Obs", conceptIdComparator);
						writeTimePeriod(range);
						obsInterval = range;
						status = DS_TS_OBS;
					} else {
						closeElement("Series");
						status = DSC;
					}
					componentHelper.logger.log(Level.FINE,
							componentHelper.logName, "handle( Atttribute )",
							"status DS_TS, DS_TSC for attribute: \"" + name
									+ "\" and key: \"" + key + "\"");
					break;
				case DS_TS_OBS:
					closeElement("Obs");
					status = DS_TSC;
					componentHelper.logger.log(Level.FINE,
							componentHelper.logName, "handle( Atttribute )",
							"status DS_TS_OBS for attribute: \"" + name
									+ "\" and key: \"" + key + "\"");
					break;
				}
			} // end loop
			if (status != DS_ADD_SER_ATTR) {
				writeAttribute(name, value);
			} else {
				status = DS;
			}
		} catch (final IOException ioe) {
			componentHelper.logger.log(Level.SEVERE, componentHelper.logName,
					"handle( Attribute )", "Output problem.", ioe);
			close();
		}
	}

	/* for unit testing purpose */
	public ComponentHelper getComponentHelper() {
		return componentHelper;
	}

	/* =================== Protected Methods ==================== */

	protected void openElement(String element) throws IOException {
		super.openElement(prefixElement(element));
	}

	protected void closeElement(String element) throws IOException {
		super.closeElement(prefixElement(element));
	}

	protected void writeElement(String element, String value)
			throws IOException {
		super.writeElement(prefixElement(element), value);
	}

	/* ==================== Private Methods ===================== */

	private void writeDocumentElement(final MessageBegin messageBegin)
			throws IOException {

		final String dtd, stylesheet;

		if (messageBegin.isStatisticalDefinitions()) {
			documentElement = "Structure";
			dtd = componentHelper.getPreference("structure.dtd", null);
			stylesheet = componentHelper.getPreference("structure.stylesheet",
					null);
		} else if (messageBegin.isStatisticalData()) {
			documentElement = features.getDataDocumentElement();
			dtd = uriMapping.dtd(currentMaintenanceAgencyId, keyFamily);
			stylesheet = uriMapping.stylesheet(currentMaintenanceAgencyId,
					keyFamily);
		} else {
			componentHelper.logger.log(Level.SEVERE, componentHelper.logName,
					"handle( MessageBegin )", "Unsupported message function \""
							+ messageBegin.getFunction() + "\".");
			dtd = stylesheet = null;
		}

		if (dtd != null) {
			writeRaw("<!DOCTYPE " + documentElement + " SYSTEM \"" + dtd
					+ "\">");
			writeRaw(NEW_LINE);
		}

		if (stylesheet != null) {
			writeRaw("<?xml-stylesheet href=\"" + stylesheet
					+ "\" type=\"text/xsl\"?>");
			writeRaw(NEW_LINE);
		}

		writeRaw(NEW_LINE);
		changeToPrefix(MESSAGE_NS);

		openElement(documentElement);
		status = NONE;

		writeAttribute(features.getXmlnsAttributeName(),
				features.getXmlnsSdmxMessage());

		if (messageBegin.isStatisticalData()) {
			writeRaw(features.getFurtherNamespaces());
			writeRaw(features
					.getXmlnsDsd(currentMaintenanceAgencyId, keyFamily));
		}

		if (messageBegin.isStatisticalDefinitions()) {
			writeAttribute("xmlns:message", features.getXmlnsSdmxMessage());
		}

		final String messageSchema = componentHelper.getPreference(
				"message.schema", null);
		final String commonSchema = componentHelper.getPreference(
				"common.schema", null);
		final String structureSchema = componentHelper.getPreference(
				"structure.schema", null);
		if (messageSchema != null || messageBegin.isStatisticalDefinitions()
				&& structureSchema != null || messageBegin.isStatisticalData()
				&& keyFamilySchema != null
				|| features instanceof Version21Features
				|| features instanceof Version2Features) {
			writeAttribute("xmlns:xsi", Constants.XMLNS_XSI);
		}
		if (messageSchema != null || messageBegin.isStatisticalDefinitions()
				&& structureSchema != null) {
			writeAttribute(
					"xsi:schemaLocation",
					(features.getSchemaLocation(messageSchema, commonSchema,
							currentMaintenanceAgencyId, keyFamily,
							keyFamilySchema) + (messageBegin
							.isStatisticalDefinitions()
							&& structureSchema != null ? (messageSchema != null ? " "
							: "")
							+ features.getXmlnsSdmxStructure()
							+ ' '
							+ structureSchema
							: "")));
		}
	}

	private void writeDocumentHeader(final MessageBegin messageBegin)
			throws IOException {

		openElement("Header");
		writeElement("ID", currentInterchangeBegin.getIref());

		writeElement("Test", currentInterchangeBegin.isTest() ? "true"
				: "false");

		if (!sdmxVersion.equalsIgnoreCase("2.1")) {
			final String subject = messageBegin.getIdentity();
			if (!"".equals(subject)) {
				changeToPrefix(COMMON_NS);
				writeElement("Name", messageBegin.getIdentity());
				restorePrefix();
			}
		}
		xmlDateFormat.applyPattern(XMLDateFormat.TIMESTAMP_PATTERN);

		{
			final Date preparationDate = currentInterchangeBegin
					.getPreparationDateTimeAsDate();
			if (preparationDate != null) {
				final String prepDateTimeXML = tsUtil.normalize(
						xmlDateFormat.format(preparationDate),
						TimeZoneUtil.ISO8601);
				writeElement("Prepared", prepDateTimeXML);
			}
		}

		{
			final String senderId = messageBegin.getSender();
			if (!currentInterchangeBegin.getSender().equals(senderId)) {
				componentHelper.logger.log(
						Level.SEVERE,
						componentHelper.logName,
						"handle( MessageBegin )",
						"Interchange sender \""
								+ currentInterchangeBegin.getSender()
								+ "\" does not match message sender \""
								+ senderId + "\".");
			}

			openElement("Sender");
			writeAttribute("id", senderId);

			for (final Iterator iterator = messageBegin.contactIterator(); iterator
					.hasNext();) {
				final Contact contact = (Contact) iterator.next();

				openElement("Contact");

				changeToPrefix(COMMON_NS);
				writeElement("Name", contact.getName());
				restorePrefix();
				writeElement("Department", contact.getIdentity());

				final String role;
				if (contact.isResponsibleForInformationProduction()) {
					role = "Responsible for Information Production";
				} else if (contact.isHeadOfInformationProduction()) {
					role = "Head of Information Production";
				} else if (contact.isResponsibleForDataProcessing()) {
					role = "Responsible for Data Processing";
				} else if (contact.isHeadOfDataProcessing()) {
					role = "Head of Data Processing";
				} else {
					role = contact.getFunction();
					componentHelper.logger.log(Level.WARNING,
							componentHelper.logName, "handle( MessageBegin )",
							"Contact function \"" + contact.getFunction()
									+ "\" is not known.");
				}
				writeElement("Role", role);

				for (final Iterator channels = contact.channelIterator(); channels
						.hasNext();) {
					final ContactChannel channel = (ContactChannel) channels
							.next();

					final String media;
					if (channel.isEmailAddress()) {
						media = "URI";
					} else if (channel.isFaxNumber()) {
						media = "Fax";
					} else if (channel.isTelephoneNumber()) {
						media = "Telephone";
					} else if (channel.isX400Address()) {
						media = "X400";
					} else {
						media = channel.getType();
						componentHelper.logger.log(Level.WARNING,
								componentHelper.logName,
								"handle( MessageBegin )",
								"Contact channel type \"" + channel.getType()
										+ "\" is not known.");
					}

					writeElement(media, channel.isEmailAddress() ? "mailto:"
							+ channel.getNumber() : channel.getNumber());
				}

				closeElement("Contact");
			}

			closeElement("Sender");
		}

		{
			final String receiverId = messageBegin.getReceiver();
			if (!currentInterchangeBegin.getReceiver().equals(receiverId)) {
				componentHelper.logger.log(
						Level.SEVERE,
						componentHelper.logName,
						"handle( MessageBegin )",
						"Interchange receiver \""
								+ currentInterchangeBegin.getReceiver()
								+ "\" does not match message receiver \""
								+ receiverId + "\".");
			}

			if (!"".equals(receiverId)) {
				openElement("Receiver");
				writeAttribute("id", receiverId);
				closeElement("Receiver");
			}
		}

		if (sdmxVersion.equalsIgnoreCase("2.1")) {
			final String subject = messageBegin.getIdentity();
			if (!"".equals(subject)) {
				changeToPrefix(COMMON_NS);
				writeElement("Name", messageBegin.getIdentity());
				restorePrefix();
			}
		}

		{
			final String keyFamilyRef = messageBegin.getKeyFamilyId();
			if (!"".equals(keyFamilyRef) && messageBegin.isStatisticalData()
					&& !(features instanceof Version21Features)) {
				openElement("KeyFamilyRef");
				writeCharacters(keyFamilyRef);
				closeElement("KeyFamilyRef");
			}

			final String keyFamilyAgency = messageBegin
					.getMaintenanceAgencyId();
			if (!"".equals(keyFamilyAgency) && messageBegin.isStatisticalData()
					&& !(features instanceof Version21Features)) {
				openElement("KeyFamilyAgency");
				writeCharacters(keyFamilyAgency);
				closeElement("KeyFamilyAgency");
			}

			final String dsi = messageBegin.getDataSetId();
			if (!"".equals(dsi) && !(features instanceof Version21Features)) {
				final String dsa = componentHelper.getPreference(
						"dataSetAgency", null);
				if (null != dsa) {
					writeElement("DataSetAgency", dsa);
				}
				writeElement("DataSetID", dsi);
				// only for 2.1... currentStructureRef = dsi;
			}

			if (!"".equals(keyFamilyRef) && !"".equals(keyFamilyAgency)
					&& !"".equals(dsi)
					&& (features instanceof Version21Features)) {
				openElement("Structure");
				writeAttribute("dimensionAtObservation", "TIME_PERIOD");
				writeAttribute("structureID", dsi);
				currentStructureRef = dsi;
				writeAttribute("namespace", features.getDataStructureNamespace(
						keyFamilyAgency, keyFamilyRef));
				changeToPrefix(COMMON_NS);
				openElement("Structure");
				openElement("Ref");
				writeAttribute("agencyID", keyFamilyAgency);
				writeAttribute("id", keyFamilyRef);
				closeElement("Ref");
				closeElement("Structure");
				restorePrefix();
				closeElement("Structure");
			}
		}

		{
			if (!"".equals(currentMessageBeginAction)
					&& features instanceof Version1Features) {
				writeElement("DataSetAction", currentMessageBeginAction);
			}
		}

		{ // misused to indicate the external DSI/dataflowID.
			// (sdmx foresees it as a kind of mref.)
			final String dsi = messageBegin.getDataSetId();
			if (!"".equals(dsi) && (features instanceof Version21Features)) {
				writeElement("DataSetID", dsi);
			}
		}

		{
			final Date extractionDate = messageBegin
					.getExtractionDateTimeAsDate();
			componentHelper.logger.log(Level.INFO, componentHelper.logName,
					"handle( MessageBegin )", "Extraction date \""
							+ extractionDate + "\".", messageBegin);
			if (extractionDate != null) {
				final String extractDateTimeXML = tsUtil.normalize(
						xmlDateFormat.format(extractionDate),
						TimeZoneUtil.ISO8601);
				writeElement("Extracted", extractDateTimeXML);
			}
		}

		{
			final Interval reportingInterval = messageBegin
					.getReportingPeriodAsInterval();

			if (reportingInterval != null) {
				final String periodBegin = tsUtil.normalize(
						xmlDateFormat.format(reportingInterval.getBegin()),
						TimeZoneUtil.ISO8601);
				final String periodEnd = tsUtil.normalize(
						xmlDateFormat.format(reportingInterval.getEnd()),
						TimeZoneUtil.ISO8601);
				writeElement("ReportingBegin", periodBegin);
				writeElement("ReportingEnd", periodEnd);
			}
		}

		closeElement("Header");
	}

	private String getActionFromMessageBegin(MessageBegin messageBegin) {
		final String action;
		if ("".equals(messageBegin.getAction())) {
			action = "";
		} else if (messageBegin.isUpdateOrReplaceAction()) {
			action = features.getUpdateReplaceAction();
		} else if (messageBegin.isDeleteAction()) {
			action = "Delete";
		} else if (messageBegin.isTestingAction()) {
			action = "Test";
			componentHelper.logger.log(Level.WARNING, componentHelper.logName,
					"handle( MessageBegin )", "Using deprecated test action.",
					messageBegin);
		} else {
			action = messageBegin.getAction();
			componentHelper.logger.log(Level.SEVERE, componentHelper.logName,
					"handle( MessageBegin )", "Message action \""
							+ messageBegin.getAction() + "\" is not known.",
					messageBegin);
		}
		return action;
	}

	private void openElementDataSet() throws IOException {
		openElement(prefixElement("DataSet"));

		if (features instanceof Version21Features) {
			writeAttribute("data:action", currentMessageBeginAction);
			writeAttribute("data:structureRef", currentStructureRef);
			writeAttribute("data:dataScope", "DataStructure");
			writeAttribute("xsi:type", "dsd:DataSetType");
			writeAttribute("xsi:schemaLocation", sdmxDataStructureNamespace
					+ " " + sdmxDataStructureSchemaLocation);

		} else if (features instanceof Version2Features) {
			writeAttribute("action", currentMessageBeginAction);

			if (keyFamilyNamespace != null) {
				writeAttribute("xmlns", keyFamilyNamespace);
			} else {
				writeAttribute("xmlns", sdmxDataStructureNamespace);
			}

			if (keyFamilySchema != null) {
				writeAttribute("xsi:schemaLocation", keyFamilyNamespace + " "
						+ keyFamilySchema);
			} else {
				writeAttribute("xsi:schemaLocation", sdmxDataStructureNamespace
						+ " " + sdmxDataStructureSchemaLocation);
			}

		} else {
			writeAttribute("xmlns", keyFamilyNamespace);
			if (keyFamilySchema != null) {
				writeAttribute("xsi:schemaLocation", keyFamilyNamespace + " "
						+ keyFamilySchema);
			}
		}
	}

	private boolean shouldBePrefixed(final String element) {
		if (element.indexOf(':') != -1) {
			return false;
		}
		if (!(features instanceof Version21Features)) {
			return false;
		}
		return "Obs".equalsIgnoreCase(element) ? false : "Series"
				.equalsIgnoreCase(element) ? false : "Group"
				.equalsIgnoreCase(element) ? false : "Ref"
				.equalsIgnoreCase(element) ? false : true;
	}

	private String prefixElement(final String element) {
		return (shouldBePrefixed(element) && (null != prefix)) ? prefix + ":"
				+ element : element;
	}

	private void changeToPrefix(final String newPrefix) {
		oldPrefix = prefix;
		prefix = newPrefix;
	}

	private void restorePrefix() {
		prefix = oldPrefix;
		oldPrefix = null;
	}

	/* ------------------------ Writing ------------------------- */

	/* ------------------------- Header ------------------------- */

	private void writeXMLHeader() {
		componentHelper.logger.log(Level.FINE, componentHelper.logName,
				"writeXMLHeader()", "Writing header.");

		try {
			writeRaw("<!-- SDMX-ML from Gesmes/TS Suite -->");
			writeRaw(NEW_LINE);
			writeRaw("<!-- @(#)$Id: SDMXMLWriter.java,v 1.29 2014/10/01 12:45:25 trainot Exp $ -->");
			writeRaw(NEW_LINE);
		} catch (final IOException ioe) {
			componentHelper.logger.log(Level.SEVERE, componentHelper.logName,
					"writeXMLHeader", "Output problem.", ioe);
			close();
		}
	}

	/* ------------------- KeyFamilyComponent ------------------- */

	/** Write a <code>KeyFamilyComponent</code>. */
	private void write(final KeyFamilyComponent component) throws IOException {
		final String id = component.getConceptId();
		final String element;
		if (component.isKeyStructureDimension()) {
			element = "Dimension";
		} else if (component == keyFamily.getTimePeriod()) {
			element = "TimeDimension";
		} else if (component == keyFamily.getTimeFormat()) {
			element = "Attribute";
		} else if (component == keyFamily.getObsValue()) {
			element = "PrimaryMeasure";
		} else if (component.isArrayCell() || component.isAttribute()) {
			element = "Attribute";
		} else {
			componentHelper.logger.log(Level.SEVERE, componentHelper.logName,
					"handle( KeyFamily )",
					"Key family component \"" + id + "\" has invalid type \""
							+ component.getConceptType() + "\".", keyFamily);
			element = component.getConceptType();
		}
		openElement(element);
		// TODO: avoid attribute names illegal in XML
		writeAttribute(features.getConceptRefAttributeName(), id);

		{
			final String attachmentLevel;
			if (component.isAttachedToDataSet()) {
				attachmentLevel = "DataSet";
			} else if (component.isAttachedToSiblingGroup()) {
				attachmentLevel = "Group";
			} else if (component.isAttachedToTimeSeries()
					|| component == keyFamily.getTimeFormat()) {
				attachmentLevel = "Series";
			} else if (component.isAttachedToObservation()
					|| component.isArrayCell()
					&& component != keyFamily.getObsValue()) {
				attachmentLevel = "Observation";
			} else {
				attachmentLevel = null;
			}
			if (attachmentLevel != null) {
				writeAttribute("attachmentLevel", attachmentLevel);
			}
		}

		// TODO: special case for CL_TIME_FORMAT, if exists
		if (component.isCoded()) {
			writeAttribute("codelist", component.getCodeListId());
		}
		if (component.isKeyStructureDimension()) {
			writeAttribute("isFrequencyDimension",
					component.isFrequencyDimension() ? "true" : "false");
		}

		{
			final String usageStatus;
			if (component.isMandatory()
					|| component == keyFamily.getTimeFormat()) {
				usageStatus = "Mandatory";
			} else if ("".equals(component.getUsageStatus())) {
				usageStatus = null;
			} else {
				usageStatus = "Conditional";
			}
			if (usageStatus != null) {
				writeAttribute("assignmentStatus", usageStatus);
			}
		}

		{
			final String isTimeFormat;
			if (component == keyFamily.getTimeFormat()) {
				isTimeFormat = "true";
			} else if (component.isArrayCell()
					&& component != keyFamily.getObsValue()
					|| component.isAttribute()) {
				isTimeFormat = "false";
			} else {
				isTimeFormat = null;
			}
			if (isTimeFormat != null) {
				writeAttribute("isTimeFormat", isTimeFormat);
			}
		}

		if (!component.isKeyStructureDimension()
				&& component != keyFamily.getTimePeriod()
				&& component != keyFamily.getObsValue()) {
			TextFormat format = component.getTextFormat();
			if (format != null) {
				if (component == keyFamily.getTimeFormat()) {
					format = TextFormat.create(TextFormat.ANY, 3, false);
				}

				openElement("TextFormat");

				final String type;
				switch (format.getCharacterType()) {
				case TextFormat.ANY:
					type = features.getAlphanumericType();
					break;
				case TextFormat.ALPHABETIC:
					type = features.getAlphabeticType();
					break;
				case TextFormat.NUMERIC:
					type = features.getNumericType();
					break;
				default:
					type = "";
					componentHelper.logger.log(Level.SEVERE,
							componentHelper.logName, "handle( KeyFamily )",
							"Key family component \"" + id
									+ "\" has invalid text format \""
									+ component.getFormat() + "\".", keyFamily);
				}
				if (features instanceof Version1Features) {
					writeAttribute(features.getTextTypeAttributeName(),
							format.isVariableLength() ? type : type + "Fixed");

					writeAttribute("length",
							Integer.toString(format.getReferenceLength()));
				} else if (features instanceof Version2Features) {
					writeAttribute(features.getTextTypeAttributeName(), type);
					writeAttribute("minLength", Integer.toString(format
							.isVariableLength() ? 0 : format
							.getReferenceLength()));
					writeAttribute("maxLength",
							Integer.toString(format.getReferenceLength()));
				} else {
					componentHelper.logger.log(Level.SEVERE,
							componentHelper.logName,
							"write(KeyFamilyComponent)",
							"SDMX syntax version not properly implemented.",
							features);
				}

				closeElement("TextFormat");
			}
		}

		if (component.isAttachedToSiblingGroup()) {
			writeElement("AttachmentGroup", "Group");
		}

		closeElement(element);
	}

	/* --------------------- Identification --------------------- */

	private void writeKey(final String key) throws IOException {
		if (key.length() > 0) {
			for (final Iterator dimensions = keyFamily.keyStructureIterator(), values = new KeyTokenizer(
					key); dimensions.hasNext() && values.hasNext();) {
				final KeyFamilyComponent dimension = (KeyFamilyComponent) dimensions
						.next();
				final String value = (String) values.next();
				if ("".equals(value))
					continue;
				writeAttribute(dimension.getConceptId(), value);
			}
		}
	}

	private void writeFrequency(final String key) throws IOException {
		if (key.length() > 0) {
			for (final Iterator dimensions = keyFamily.keyStructureIterator(), values = new KeyTokenizer(
					key); dimensions.hasNext() && values.hasNext();) {
				final KeyFamilyComponent dimension = (KeyFamilyComponent) dimensions
						.next();
				final String value = (String) values.next();
				if ("".equals(value))
					continue;
				if (dimension == keyFamily.getFrequency()) {
					writeAttribute(dimension.getConceptId(), value);
				}
			}
		}
	}

	private void writeTimeFormat(final RegularInterval interval)
			throws IOException {
		final int duration = interval.getFrequency();
		writeAttribute(keyFamily.getTimeFormat().getConceptId(),
				XMLDateFormat.formatDuration(duration));
	}

	private void writeTimePeriod(final RegularInterval interval)
			throws IOException {
		final Date time = interval.getBegin();
		final int duration = interval.getFrequency();
		xmlDateFormat.setFrequency(duration);
		writeAttribute(keyFamily.getTimePeriod().getConceptId(),
				xmlDateFormat.format(time));
	}

	/* ---------------------- Observation ----------------------- */

	private void writeObservation(final Observation observation)
			throws IOException {
		final Iterator arrayCells = keyFamily.arrayCellIterator();
		{
			final String value = observation.getValue();
			if (arrayCells.hasNext()) {
				final KeyFamilyComponent obsValue = (KeyFamilyComponent) arrayCells
						.next();
				writeAttribute(obsValue.getConceptId(),
						observation.hasMissingValue() ? "NaN" : value);
			}
		}
		{
			final String status = observation.getStatus();
			if (arrayCells.hasNext()) {
				final KeyFamilyComponent obsStatus = (KeyFamilyComponent) arrayCells
						.next();
				writeAttribute(obsStatus.getConceptId(), status);
			}
		}
		{
			final String conf = "".equals(observation.getConfidentiality()) ? defaultObsConf
					: observation.getConfidentiality();
			if (arrayCells.hasNext()) {
				final KeyFamilyComponent obsConf = (KeyFamilyComponent) arrayCells
						.next();
				if (null != conf) {
					writeAttribute(obsConf.getConceptId(), conf);
				}
			}
		}
		{
			final String preBreak = "".equals(observation.getPreBreakValue()) ? defaultObsPreBreak
					: observation.getPreBreakValue();
			if (arrayCells.hasNext()) {
				final KeyFamilyComponent obsPreBreak = (KeyFamilyComponent) arrayCells
						.next();
				if (null != preBreak) {
					writeAttribute(obsPreBreak.getConceptId(),
							observation.hasMissingPreBreakValue() ? "NaN"
									: preBreak);
				}
			}
		}
	}

	/* ---------------------- Close cubes ----------------------- */

	/** Close all open Cube elements. */
	public void closeAllContentElements() throws IOException {
		// This loop rattles through a couple of state changes.
		do {
			switch (status) {
			case NONE:
				break;
			case CL:
				closeElement("message:CodeLists");
				status = NONE;
				break;
			case SC:
				closeElement("message:Concepts");
				status = NONE;
				break;
			case KF:
				closeElement("message:KeyFamilies");
				status = NONE;
				break;
			case DS:
				changeToPrefix(MESSAGE_NS);
				closeElement("DataSet");
				status = NONE;
				break;
			case DS_SG:
				closeElement("Group");
				status = DS;
				break;
			case DS_TS:
				closeElement("Series");
				status = DS;
				break;
			case DS_TS_OBS:
				closeElement("Obs");
				status = DS_TS;
				break;
			}
		} while (status != NONE);
	}

	/* ----------------------- Auxiliary ------------------------ */

	/** Determine if the state matches the requested Cube. */
	private boolean stateMatches(final String key,
			final RegularInterval interval, final KeyFamilyComponent component) {
		switch (status) {
		case NONE:
		case DSC:
		case DS_TSC:
			return false;
		case DS:
			return component.isAttachedToDataSet();
		case DS_SG:
			return key.equals(sgKey) && component.isAttachedToSiblingGroup();
		case DS_TS:
			return key.equals(tsKey) && component.isAttachedToTimeSeries();
		case DS_TS_OBS:
			return key.equals(tsKey) && interval != null
					&& interval.equals(obsInterval)
					&& component.isAttachedToObservation();
		case DS_ADD_SER_ATTR:
			return component.isAttachedToTimeSeries();
		}
		throw new RuntimeException("Failure!");
	}

	private boolean keyExtends(final String seriesKey, final String groupKey) {
		for (final Iterator dimensions = keyFamily.keyStructureIterator(), seriesKeyValues = new KeyTokenizer(
				seriesKey), groupKeyValues = new KeyTokenizer(groupKey); dimensions
				.hasNext();) {
			final KeyFamilyComponent dimension = (KeyFamilyComponent) dimensions
					.next();
			final String seriesKeyValue = seriesKeyValues.hasNext() ? (String) seriesKeyValues
					.next() : "";
			final String groupKeyValue = groupKeyValues.hasNext() ? (String) groupKeyValues
					.next() : "";
			if (!groupKeyValue.equals(seriesKeyValue)) {
				if (!dimension.isFrequencyDimension()
						|| !"".equals(groupKeyValue)) {
					return false;
				}
			}
		}
		return true;
	}
}

/* *********************** Interface ************************ */

interface SDMXVersionFeatures {

	/** xmlns or xmlns:message in v21 */
	String getXmlnsAttributeName();

	/** further namespaces for v21 */
	String getFurtherNamespaces();

	/** datastructure namespace for v21 */
	String getDataStructureNamespace(String agency, String dsd);

	/** provide the location (plural for v21) */
	String getSchemaLocation(String messageSchema, String commonSchema,
			String currentMaintenanceAgencyId, KeyFamily keyFamily,
			String keyFamilySchema);

	/** provide the xmlns:dsd for v21 */
	String getXmlnsDsd(String currentMaintenanceAgencyId, KeyFamily keyFamily);

	/** SDMX-ML message namespace URI */
	String getXmlnsSdmxMessage();

	/** SDMX-ML structure namespace URI */
	String getXmlnsSdmxStructure();

	/** XML attribute name to reference the structure maintenance agency. */
	String getAgencyRefAttributeName();

	/** XML attribute name to reference concepts in Key Family Components. */
	String getConceptRefAttributeName();

	/** XML attribute name to set character set constraint. */
	String getTextTypeAttributeName();

	/** Text type attribute code for alpha-numeric coding. */
	String getAlphanumericType();

	/** Text type attribute code for alphabetic coding. */
	String getAlphabeticType();

	/** Text type attribute code for numeric coding. */
	String getNumericType();

	/** Code for update/replace action. */
	String getUpdateReplaceAction();

	/** Code for update/replace action. */
	String getDataDocumentElement();
}

/* ************************* Class ************************** */

class Version1Features implements SDMXVersionFeatures {

	public String getXmlnsAttributeName() {
		return "xmlns";
	}

	public String getFurtherNamespaces() {
		return "";
	}

	public String getDataStructureNamespace(final String agency,
			final String dsd) {
		return "";
	}

	public String getSchemaLocation(final String messageSchema,
			final String commonSchema, final String currentMaintenanceAgencyId,
			final KeyFamily keyFamily, final String keyFamilySchema) {
		return messageSchema != null ? getXmlnsSdmxMessage() + ' '
				+ messageSchema : "";
	}

	public String getXmlnsDsd(final String currentMaintenanceAgencyId,
			final KeyFamily keyFamily) {
		return "";
	}

	public String getXmlnsSdmxMessage() {
		return Constants.XMLNS_SDMX_MESSAGE_V1_0;
	}

	public String getXmlnsSdmxStructure() {
		return Constants.XMLNS_SDMX_STRUCTURE_V1_0;
	}

	public String getAgencyRefAttributeName() {
		return "agency";
	}

	public String getConceptRefAttributeName() {
		return "concept";
	}

	public String getTextTypeAttributeName() {
		return "TextType";
	}

	public String getAlphanumericType() {
		return "AlphaNum";
	}

	public String getAlphabeticType() {
		return "Alpha";
	}

	public String getNumericType() {
		return "Num";
	}

	public String getUpdateReplaceAction() {
		return "Update";
	}

	public String getDataDocumentElement() {
		return "MessageGroup";
	}
}

/* ************************* Class ************************** */

class Version2Features implements SDMXVersionFeatures {

	public String getXmlnsAttributeName() {
		return "xmlns";
	}

	public String getFurtherNamespaces() {
		return "";
	}

	public String getDataStructureNamespace(final String agency,
			final String dsd) {
		return "";
	}

	public String getSchemaLocation(final String messageSchema,
			final String commonSchema, final String currentMaintenanceAgencyId,
			final KeyFamily keyFamily, final String keyFamilySchema) {
		return messageSchema != null ? getXmlnsSdmxMessage() + ' '
				+ messageSchema : "";
	}

	public String getXmlnsDsd(final String currentMaintenanceAgencyId,
			final KeyFamily keyFamily) {
		return "";
	}

	public String getXmlnsSdmxMessage() {
		return Constants.XMLNS_SDMX_MESSAGE_V2_0;
	}

	public String getXmlnsSdmxStructure() {
		return Constants.XMLNS_SDMX_STRUCTURE_V2_0;
	}

	public String getAgencyRefAttributeName() {
		return "agencyID";
	}

	public String getConceptRefAttributeName() {
		return "conceptRef";
	}

	public String getTextTypeAttributeName() {
		return "textType";
	}

	public String getAlphanumericType() {
		return "String";
	}

	public String getAlphabeticType() {
		return "String";
	}

	public String getNumericType() {
		return "BigInteger";
	}

	public String getUpdateReplaceAction() {
		return "Replace";
	}

	public String getDataDocumentElement() {
		return "MessageGroup";
	}
}

/* ************************* Class ************************** */

class Version21Features implements SDMXVersionFeatures {

	public String getXmlnsAttributeName() {
		return "xmlns:message";
	}

	public String getFurtherNamespaces() {
		return " xmlns:common=\"" + Constants.XMLNS_SDMX_COMMON_V2_1 + "\""
				+ " xmlns:data=\""
				+ Constants.XMLNS_SDMX_DATA_STRUCTURESPECIFIC_V2_1 + "\"";
	}

	public String getDataStructureNamespace(final String agency,
			final String dsd) {
		return "urn:sdmx:org.sdmx.infomodel.datastructure.DataStructure="
				+ agency + ":" + dsd + "(1.0):ObsLevelDim:TIME_PERIOD";
	}

	public String getSchemaLocation(final String messageSchema,
			final String commonSchema, final String currentMaintenanceAgencyId,
			final KeyFamily keyFamily, final String keyFamilySchema) {
		String result = "";
		if (messageSchema != null) {
			result = result + getXmlnsSdmxMessage() + ' ' + messageSchema + ' ';
		}
		if (commonSchema != null) {
			result = result + getXmlnsSdmxCommon() + ' ' + commonSchema + ' ';
		}
		result = result
				+ getDataStructureNamespace(currentMaintenanceAgencyId,
						keyFamily.getId()) + ' ' + keyFamilySchema
		// no! + ' '
		;
		return result;
	}

	public String getXmlnsDsd(final String currentMaintenanceAgencyId,
			final KeyFamily keyFamily) {
		return " xmlns:dsd=\""
				+ getDataStructureNamespace(currentMaintenanceAgencyId,
						keyFamily.getId()) + "\"";
	}

	private String getXmlnsSdmxCommon() { // not in the interface
		return Constants.XMLNS_SDMX_COMMON_V2_1;
	}

	public String getXmlnsSdmxMessage() {
		return Constants.XMLNS_SDMX_MESSAGE_V2_1;
	}

	public String getXmlnsSdmxStructure() {
		return Constants.XMLNS_SDMX_STRUCTURE_V2_1;
	}

	public String getAgencyRefAttributeName() {
		return "agencyID";
	}

	public String getConceptRefAttributeName() {
		return "conceptRef";
	}

	public String getTextTypeAttributeName() {
		return "textType";
	}

	public String getAlphanumericType() {
		return "String";
	}

	public String getAlphabeticType() {
		return "String";
	}

	public String getNumericType() {
		return "BigInteger";
	}

	public String getUpdateReplaceAction() {
		return "Replace";
	}

	public String getDataDocumentElement() {
		return "StructureSpecificData";
	}

}
