// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2002, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.xml;

import java.text.ParseException;
import java.text.ParsePosition;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.HashMap;
import java.util.Locale;
import java.util.Map;

import Int.ecb.gesmescb.time.RegularInterval;

/* ***************************** Class ****************************** */

/**
 * <P>Formats {@link Date}s to XML dateTime strings.
 *
 * @author
 * <A HREF="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</A>
 *
 * @version @(#)$Id: XMLDateFormat.java,v 1.19 2014/08/04 12:57:06 trainot Exp $
 */
public class XMLDateFormat {

	/* ======================= Constants ======================== */

	/**
	 *   serialVersionUID generated by eclipse.
	 */
	private static final long serialVersionUID = 4176030302688310660L;
	static final String
		TIMESTAMP_PATTERN = "yyyy-MM-dd'T'HH:mm:ssZ",
		SECONDLY_PATTERN = "yyyy-MM-dd'T'HH:mm:ss",
		MINUTELY_PATTERN = "yyyy-MM-dd'T'HH:mm",
		DAILY_PATTERN = "yyyy-MM-dd",
		WEEKLY_PATTERN = DAILY_PATTERN,
		EXOTIC_WEEKLY_PATTERN = "yyyy-'W'ww",
		MONTHLY_PATTERN = "yyyy-MM",
		QUARTERLY_PATTERN = MONTHLY_PATTERN,
		SEMIANNUAL_PATTERN = MONTHLY_PATTERN,
		ANNUAL_PATTERN = "yyyy",
		
		/**
		 * Internal formats are the formats used in SDMX-EDI
		 * and accepted by the checker
		 */
		MINUTELY_PATTERN_INTERN = "yyyyMMddHHmm",
		DAILY_PATTERN_INTERN = "yyyyMMdd",
		WEEKLY_PATTERN_INTERN = "yyyyww",
		MONTHLY_PATTERN_INTERN = "yyyyMM",
		QUARTERLY_PATTERN_INTERN = MONTHLY_PATTERN_INTERN,
		SEMIANNUAL_PATTERN_INTERN = MONTHLY_PATTERN_INTERN,
		ANNUAL_PATTERN_INTERN = "yyyy";
	
	public static Map xmlDateFormatToEDI = new HashMap();

	static {
		xmlDateFormatToEDI.put("P1Y", "602");
		xmlDateFormatToEDI.put("P6M", "604");
		xmlDateFormatToEDI.put("P3M", "608");
		xmlDateFormatToEDI.put("P1M", "610");
		xmlDateFormatToEDI.put("P7D", "616");
		xmlDateFormatToEDI.put("P1D", "102");
		xmlDateFormatToEDI.put("PT1M", "203");
	}
	/** If set, quarters are formatted as Q1-4, semesters as S1-2. */
	private String version;
	
	private SimpleDateFormat simpleDateFormat;
	private int frequency;
	
	/* ====================== Constructors ====================== */

	/** Construct an instance. */
	public XMLDateFormat () {
		simpleDateFormat = new SimpleDateFormat( MINUTELY_PATTERN, Locale.ENGLISH );
		version=null; 
	}

	/* ===================== Public Methods ===================== */

	/** Sets the version of sdmxml to which we write.
	 *  If no version is given, then sdwxml is assumed, whereby
	 *  quarters get formatted as yyyy-qq and semesters as yyyy-ss.
	 *  Otherwise, quarters get formatted as yyyy-Q[1-4] ans
	 *  semesters as yyyy-S[1-2].
	 *  
	 * @param inputVersion the version to which we write.
	 * 
	 */
	public void setVersion( final String inputVersion ) {
		this.version = inputVersion;		
	}
	
	/** Set the translation pattern for frequency.
	 * @param frequency one of {@link RegularInterval}'s codes.
	 *
	 * {@link RegularInterval#ANNUAL},
	 * {@link RegularInterval#SEMIANNUAL},
	 * {@link RegularInterval#QUARTERLY},
	 * {@link RegularInterval#MONTHLY},
	 * {@link RegularInterval#WEEKLY},
	 * {@link RegularInterval#DAILY},
	 * {@link RegularInterval#MINUTELY}.
	 */
	public void setFrequency ( final int frequency ) {
		this.frequency = frequency;
		switch ( frequency ) {
			case RegularInterval.ANNUAL:
				applyPattern( ANNUAL_PATTERN );
				break;
			case RegularInterval.SEMIANNUAL:
				applyPattern( SEMIANNUAL_PATTERN );
				break;
			case RegularInterval.QUARTERLY:
				applyPattern( QUARTERLY_PATTERN );
				break;
			case RegularInterval.MONTHLY:
				applyPattern( MONTHLY_PATTERN );
				break;
			case RegularInterval.WEEKLY:
				applyPattern( WEEKLY_PATTERN );
				break;
			case RegularInterval.DAILY:
				applyPattern( DAILY_PATTERN );
				break;
			case RegularInterval.MINUTELY:
			default:
				applyPattern( MINUTELY_PATTERN );
				break;
		}
	}

	/** Adapt to the obvious frequency and parse. */
	public Date parse ( final String text, final ParsePosition pos ) throws ParseException {
		final int hyphens = count( text, '-' );
		final int ts = count( text, 'T' );
		final int colons = count( text, ':' );
		String rewordedText = text;
		switch ( hyphens + ts + colons ) {
			case 0: setFrequency( RegularInterval.ANNUAL ); break;
			case 1: 
				final String periodTypeIndicator = text.substring(5,6);
				if ( "W".equals( periodTypeIndicator ) ) {
					setFrequency( RegularInterval.WEEKLY );
					applyPattern( EXOTIC_WEEKLY_PATTERN );
					break;
				} else if ( "Q".equals( periodTypeIndicator ) 
							||
							"B".equals( periodTypeIndicator )
							||
							"S".equals( periodTypeIndicator )
							||
							"H".equals( periodTypeIndicator )
				) {
					final String year = text.substring(0,4);
					final String subPeriod = text.substring(6, text.length());
					setFrequency( RegularInterval.MONTHLY );
					rewordedText
						= subPeriod.equals("1") ? year + "-01"
						: subPeriod.equals("2") ?
							year + ( periodTypeIndicator.equals("Q") ? "-04" : "-07"  )
						: subPeriod.equals("3") ? 
							periodTypeIndicator.equals("Q") ? (year + "-07") : (text + "is a wrong period")  
						: subPeriod.equals("4") ?
							periodTypeIndicator.equals("Q") ? (year + "-10") : (text + "is a wrong period")  
						: text + "is a wrongPeriod"
						;
					break;
				} else {
					setFrequency( RegularInterval.MONTHLY ); break;
				}
			case 2: setFrequency( RegularInterval.DAILY ); break;
			case 4: setFrequency( RegularInterval.MINUTELY ); break;
			case 5: applyPattern( SECONDLY_PATTERN ); break;
			default:
				return null;
		}
		final Date result = simpleDateFormat.parse( rewordedText, pos );
		if ( null == result ) {
				throw new ParseException( "Text unparsable as date: \""
						+ text + "\"", pos.getErrorIndex() );
		}
		return result;
		
	}

	/** Return the duration code corresponding to a frequency.
	 * @param frequency one of {@link RegularInterval}'s codes.
	 *
	 * {@link RegularInterval#ANNUAL},
	 * {@link RegularInterval#SEMIANNUAL},
	 * {@link RegularInterval#QUARTERLY},
	 * {@link RegularInterval#MONTHLY},
	 * {@link RegularInterval#WEEKLY},
	 * {@link RegularInterval#DAILY},
	 * {@link RegularInterval#MINUTELY}.
	 */
	public static String formatDuration ( final int frequency ) {
		switch ( frequency ) {
			case RegularInterval.ANNUAL: return "P1Y";
			case RegularInterval.SEMIANNUAL: return "P6M";
			case RegularInterval.QUARTERLY: return "P3M";
			case RegularInterval.MONTHLY: return "P1M";
			case RegularInterval.WEEKLY: return "P7D";
			case RegularInterval.DAILY: return "P1D";
			case RegularInterval.MINUTELY: return "PT1M";
		}
		throw new IllegalArgumentException(
			"Unknown frequency numeric code " + frequency + "."
		);
	}
/**
 * Obtains internal Gesmes Suite representation of an XML date
 * based on the frequency
 * 
 * @param frequency
 * @param xmlDate
 * @return
 */
	
	
	public String formatDateToInternalFormat(final int frequency,
			final Date date) {
		SimpleDateFormat sf = null;
		switch (frequency) {
		case RegularInterval.ANNUAL:
			 sf = new SimpleDateFormat( ANNUAL_PATTERN_INTERN, Locale.ENGLISH );
			return sf.format(date);
		case RegularInterval.SEMIANNUAL:
			 setFrequency(RegularInterval.SEMIANNUAL);
			 return formatSDMXQuarterlyAndSeminannualToInternalFormat(date);
		case RegularInterval.QUARTERLY:
		 	 setFrequency(RegularInterval.QUARTERLY);
			 return formatSDMXQuarterlyAndSeminannualToInternalFormat(date);
		case RegularInterval.MONTHLY:
			 sf = new SimpleDateFormat( MONTHLY_PATTERN_INTERN, Locale.ENGLISH );
			return sf.format(date);
		case RegularInterval.WEEKLY:
			 sf = new SimpleDateFormat( WEEKLY_PATTERN_INTERN, Locale.ENGLISH );
			return sf.format(date);
		case RegularInterval.DAILY:
			 sf = new SimpleDateFormat( DAILY_PATTERN_INTERN, Locale.ENGLISH );
			return sf.format(date);
		case RegularInterval.MINUTELY:
			 sf = new SimpleDateFormat( MINUTELY_PATTERN_INTERN, Locale.ENGLISH );
			return sf.format(date);
		}
		throw new IllegalArgumentException("Unknown frequency numeric code "
				+ frequency + ".");
	}
	
	/** Return the frequency corresponding to a duration code.
	 * @return one of {@link RegularInterval}'s codes.
	 *
	 * {@link RegularInterval#ANNUAL},
	 * {@link RegularInterval#SEMIANNUAL},
	 * {@link RegularInterval#QUARTERLY},
	 * {@link RegularInterval#MONTHLY},
	 * {@link RegularInterval#WEEKLY},
	 * {@link RegularInterval#DAILY},
	 * {@link RegularInterval#MINUTELY}.
	 */
	public static int parseDuration (
		final String duration
	) throws ParseException {
		if ( duration.length() < 3 ) {
			throw new ParseException(
				"Too short duration code \"" + duration + "\".", 0
			);
		}
		switch ( duration.charAt( 2 ) ) {
			case 'Y': if ( "P1Y".equals( duration ) ) {
					return RegularInterval.ANNUAL;
				}
				break;
			case 'M': if ( "P1M".equals( duration ) ) {
					return RegularInterval.MONTHLY;
				} else if ( "P3M".equals( duration ) ) {
					return RegularInterval.QUARTERLY;
				} else if ( "P6M".equals( duration ) ) {
					return RegularInterval.SEMIANNUAL;
				}
				break;
			case 'D': if ( "P1D".equals( duration ) ) {
					return RegularInterval.DAILY;
				} else if ( "P7D".equals( duration ) ) {
					return RegularInterval.WEEKLY;
				}
				break;
			default: if ( "PT1M".equals( duration ) ) {
				return RegularInterval.MINUTELY;
			}
		}
		throw new ParseException(
			"Cannot map duration code \"" + duration + "\" to known frequency.",
			0
		);
	}
	
	/* ======== Composed methods from SimpleDateFormat ========== */

	public void applyPattern( final String pattern ) {
		simpleDateFormat.applyPattern(pattern);
	};
	
	public Date parse(String preparedString) throws ParseException {
		return this.parse(preparedString, new ParsePosition(0));
	}
	
	public String format( final Date date) {
		final String tempFormatted = simpleDateFormat.format(date);
		if (null == version 
			||
			(frequency != RegularInterval.QUARTERLY && frequency != RegularInterval.SEMIANNUAL)
		) {
			return tempFormatted;
		}
		/* version not null. Quarters and Semesters have been formatted as YYYY-XX */
		String number = tempFormatted.substring(5, 7);
		final String year = tempFormatted.substring(0, 5);
		/* Q: -01 -> -Q1 
		 *    -04 -> -Q2
		 *    -07 -> -Q3
		 *    -10 -> -Q4 
		 * S: -01 -> -S1
		 *    -07 -> -S2 */
		if (RegularInterval.QUARTERLY==frequency) {
			if      ( "01".equals(number) ) { number = "Q1"; } 
			else if ( "04".equals(number) ) { number = "Q2"; }
			else if ( "07".equals(number) ) { number = "Q3"; }
			else if ( "10".equals(number) ) { number = "Q4"; }
			else { 
				throw new IllegalArgumentException(
					"formatting unknown quarter " + date + "."
				);
			}
		} else if (RegularInterval.SEMIANNUAL==frequency) {
			if      ( "01".equals(number) ) 
				{ 
				    if (version.equalsIgnoreCase("sdmx21")){
				    	number = "S1";	
				    } else {
				    	number = "B1";
				    }
				     
				} 
			else if ( "07".equals(number) ) 
			    { 
			    	if (version.equalsIgnoreCase("sdmx21")){
			    		number = "S2";	
			    	} else {
			    		number = "B2";
			    	}
				}
			else { 
				throw new IllegalArgumentException(
					"formatting unknown semester " + date + "."
				);
			}			
		}
		return year+number;
	};

	public String formatSDMXQuarterlyAndSeminannualToInternalFormat( final Date date) {
		final String tempFormatted = simpleDateFormat.format(date);
		
		String number = tempFormatted.substring(5, 7);
		final String year = tempFormatted.substring(0, 4);

		if (RegularInterval.QUARTERLY==frequency) {
			if      ( "01".equals(number) ) { number = "1"; } 
			else if ( "04".equals(number) ) { number = "2"; }
			else if ( "07".equals(number) ) { number = "3"; }
			else if ( "10".equals(number) ) { number = "4"; }
			else { 
				throw new IllegalArgumentException(
					"formatting unknown quarter " + date + "."
				);
			}
		} else if (RegularInterval.SEMIANNUAL==frequency) {
			if      ( "01".equals(number) ) { number = "1"; } 
			else if ( "07".equals(number) ) { number = "2"; }
			else { 
				throw new IllegalArgumentException(
					"formatting unknown semester " + date + "."
				);
			}			
		}
		return year+number;
	};
	
	/* ==================== Private Methods ===================== */

	private static int count ( final String string, final char c ) {
		final int length = string.length();
		int occurrences = 0;
		for ( int i = 0; i < length; ++i ) {
			if ( string.charAt( i ) == c ) ++occurrences;
		}
		return occurrences;
	}


}
