// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2003, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.xml;

import java.io.Writer;
import java.io.IOException;

/* ***************************** Class ****************************** */

/**
 * <p>Helps writing XML elements and attributes.</p>
 *
 * @author
 * <A HREF="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</A>
 *
 * @version @(#)$Id: XMLWriteBuffer.java,v 1.1 2003/10/26 17:07:42 bodenst Exp $
 */
class XMLWriteBuffer {

	/* ======================= Constants ======================== */

	/** Status codes. */
	static final int
		FRESH_LINE = 0,
		TAG_OPEN = 1,
		TAG_CLOSED = 2;

	/* ========================= Fields ========================= */

	/** Writes to the destination. */
	private final Writer writer;

	/** Character to use for indentation. */
	private final char indentChar;

	/** The indentation step width for nested elements. */
	private final int indentStep;

	/** The line break sequence. */
	private final String newLine;

	/** The current indentation depth for nested elements. */
	private int indent;

	/** Status if a tag is open etc. */
	int status;

	/* ====================== Constructors ====================== */

	/**
	 * Construct an <code>XMLWriteBuffer</code> with specified
	 * indentation setting.
	 */
	public XMLWriteBuffer (
		final Writer writer,
		final char indentChar,
		final int indentStep,
		final String newLine
	) {
		this.writer = writer;
		this.indentChar = indentChar;
		this.indentStep = indentStep;
		this.newLine = newLine;
		indent = 0;
		status = FRESH_LINE;
	}

	/* ===================== Public Methods ===================== */

	/** Get the indentation level. */
	public int getIndent () {
		return indent;
	}

	/** Set the indentation level. */
	public void setIndent ( final int indent ) {
		this.indent = indent;
	}

	/** Open an element. */
	public void openElement ( final String name ) throws IOException {
		switch ( status ) {
			case TAG_OPEN:
				writer.write( '>' );
				// fall through.
			case TAG_CLOSED:
				writer.write( newLine );
		}
		indent();
		writer.write( '<' );
		writer.write( name );
		status = TAG_OPEN;
		++indent;
	}

	/** Write attributes to an opening element. */
	public void writeAttribute (
		final String name, final String value
	) throws IOException {
		writer.write( ' ' );
		writer.write( name );
		writer.write( '=' );
		writer.write( '"' );
		if (
			value.indexOf( '&' ) >= 0
			||
			value.indexOf( '<' ) >= 0
			||
			value.indexOf( '>' ) >= 0
			||
			value.indexOf( '"' ) >= 0
		) {
			final int length = value.length();
			for ( int i = 0; i < length; ++i ) {
				final char c = value.charAt( i );
				switch ( c ) {
					case '&': writer.write( "&amp;" ); break;
					case '<': writer.write( "&lt;" ); break;
					case '>': writer.write( "&gt;" ); break;
					case '"': writer.write( "&quot;" ); break;
					default: writer.write( c );
				}
			}
		} else {
			writer.write( value );
		}
		writer.write( '"' );
	}

	/** Write character content into an element. */
	public void writeCharacters ( final String string ) throws IOException {
		if ( "".equals( string ) ) return;
		switch ( status ) {
			case TAG_OPEN:
				writer.write( '>' );
				status = TAG_CLOSED;
		}
		if (
			string.indexOf( '&' ) >= 0
			||
			string.indexOf( '<' ) >= 0
			||
			string.indexOf( '>' ) >= 0
		) {
			final int length = string.length();
			for ( int i = 0; i < length; ++i ) {
				final char c = string.charAt( i );
				switch ( c ) {
					case '&': writer.write( "&amp;" ); break;
					case '<': writer.write( "&lt;" ); break;
					case '>': writer.write( "&gt;" ); break;
					default: writer.write( c );
				}
			}
		} else {
			writer.write( string );
		}
	}

	/** Write attributes to an opening element. */
	public void writeElement (
		final String name, final String value
	) throws IOException {
		openElement( name );
		writeCharacters( value );
		closeElement( name );
	}

	/** Close an element. */
	public void closeElement ( final String name ) throws IOException {
		--indent;
		switch ( status ) {
			case TAG_OPEN:
				writer.write( '/' );
				writer.write( '>' );
				writer.write( newLine );
				break;
			case FRESH_LINE:
				indent();
				// fall through
			case TAG_CLOSED:
				writer.write( '<' );
				writer.write( '/' );
				writer.write( name );
				writer.write( '>' );
				writer.write( newLine );
		}
		status = FRESH_LINE;
	}

	/**
	 * <p>Close the <code>XMLWriteBuffer</code>.</p>
	 *
	 * <p>This method should always be called
	 * when the <code>XMLWriteBuffer</code>
	 * has served its purpose.</p>
	 */
	public void close () throws IOException {
	}

	/* ==================== Private Methods ===================== */

	/** Write indentation if appropriate. */
	private void indent () throws IOException {
		final int actualIndent = indent * indentStep;
		for ( int i = 0; i < actualIndent; ++i ) {
			writer.write( indentChar );
		}
	}
}
