// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2002-2004, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.xml;

import java.io.BufferedOutputStream;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.RandomAccessFile;
import java.io.Writer;
import java.io.BufferedWriter;
import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.nio.channels.FileChannel;
import java.util.Comparator;

import Int.ecb.logging.Level;
import Int.ecb.gesmescb.event.GESMESCBWriter;
import Int.ecb.gesmescb.event.GESMESCBEventListenerAdapter;
import Int.ecb.gesmescb.util.Named;
import Int.ecb.gesmescb.util.ComponentHelper;

/* ***************************** Class ****************************** */

/**
 * <P>Translates {@link Int.ecb.gesmescb.GESMESCBObject}s
 * into XML text.
 * This text then is written to an output {@link java.io.Writer}
 * or to a file according to preference settings.</P>
 *
 * <center><table border="3" cellpadding="3">
 * <caption>
 * <a name="prefs">Available {@link Int.ecb.prefs preference} settings</a>
 * </caption>
 * <tr>
 * <th>preference name</th>
 * <th>type</th>
 * <th>meaning</th>
 * <th>default</th>
 * </tr>
 * <tr><td><code>destination</td><td>string</td><td>output file path</td></tr>
 * <tr><td><code>encoding</td><td>string</td><td>character encoding</td>
 * <td>platform default, which may not be understood by XML parsers;
 * better use UTF-8 or UTF-16</td></tr>
 * <tr><td><code>indent</td><td>string or integer</td>
 * <td>if integer, then added indentation for enclosed elements;
 * if "<tt>tab</tt>", then use 1 tab character
 * to indent</td><td><tt>tab</tt></td></tr>
 * <tr><td><code>gesmes.schemaLocation</code></td>
 * <td>string</td>
 * <td>URI to the Gesmes/TS-XML schema</td></tr>
 * <tr><td><code>edifact.schemaLocation</code></td>
 * <td>string</td>
 * <td>URI to the Gesmes/TS-XML EDIFACT extension schema</td></tr>
 * </table></center>
 *
 * <p>The <code>XMLWriter</code> is not internally synchronized.
 * External synchronisation is necessary if multiple threads want to
 * use the same <code>XMLWriter</code>.</p>
 *
 * @author
 * <a href="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</a>
 * @author
 * <a href="mailto:Paula.Rocha@ECB.int">Paula Rocha</a>
 *
 * @version @(#)$Id: XMLWriter.java,v 1.39 2014/10/01 12:46:52 trainot Exp $
 */
public abstract class XMLWriter
	extends GESMESCBEventListenerAdapter
	implements Named, GESMESCBWriter
{
	/* ======================= Constants ======================== */

	/**
	 * The string used for line termination in order to split the
	 * output nicely into lines.
	 */
	protected static final String NEW_LINE
	= System.getProperty( "line.separator", "\n" );

	/* ========================= Fields ========================= */

	/** Takes care of logging and preferences. */
	protected final ComponentHelper componentHelper;

	/**
	 * The filename of the file to write to
	 * if the <CODE>XMLWriter</CODE> has been
	 * constructed from a filename.
	 * Otherwise it is null.
	 */
	private final String filename;

	protected final RandomFileOutputStream outputStream;
	
	/** Writes to the destination. */
	protected final Writer writer;

	/** Writes the XML text. */
	protected final CheckingXMLWriteBuffer buffer;

	/* ====================== Constructors ====================== */

	public XMLWriter ( final String name, final RandomFileOutputStream outputStream )
	throws IOException
	{
		componentHelper = new ComponentHelper( name, this );

		this.filename = null;
		this.outputStream = outputStream;
		{
			
			XMLOutputStreamWriter target = null;
			final String configuredEncoding = componentHelper.getPreference(
				"encoding", null
			);
			try {
				target = new XMLOutputStreamWriter(
					outputStream, configuredEncoding
				);
			}
			catch ( final UnsupportedEncodingException ue ) {
				componentHelper.logger.log(
					Level.WARNING,
					componentHelper.logName,
					"XMLWriter()",
					"Configured encoding \"" + configuredEncoding
					+ "\" is not supported on this platform."
					+ " Trying platform default instead.",
					ue
				);
				try {
					target = new XMLOutputStreamWriter(
						outputStream
					);
				}
				catch ( final UnsupportedEncodingException ue2 ) {
					componentHelper.logger.log(
						Level.SEVERE,
						componentHelper.logName,
						"XMLWriter()",
						"Default encoding is unavailable.",
						ue2
					);
				}
			}
			if ( configuredEncoding == null ) {
				componentHelper.logger.log(
					Level.CONFIG,
					componentHelper.logName,
					"XMLWriter()",
					"No encoding configured. Using platform default \""
					+ target.getEncoding()
					+ "\", which may not be supported by XML parsers"
					+ " even on the same platform."
					+ " Advice: Best configure preference \"encoding\""
					+ " to equal \"UTF-8\"."
				);
			}
			writer = new BufferedWriter( target );
		}
	
		final String configuredIndent = componentHelper.getPreference(
			"indent", "tab"
		);
		if ( "tab".equalsIgnoreCase( configuredIndent ) ) {
			buffer = new CheckingXMLWriteBuffer(
				writer, '\t', 1, NEW_LINE
			);
		} else {
			int indentStep = 0;
			try {
				indentStep = Integer.parseInt( configuredIndent );
			}
			catch ( final NumberFormatException nfe ) {
				componentHelper.logger.log(
					Level.WARNING,
					componentHelper.logName,
					"XMLWriter()",
					"Preference value \"indent\" should be numeric"
					+ " or \"tab\", but it is \""
					+ configuredIndent + "\". Using no indentation."
				);
			}
			buffer = new CheckingXMLWriteBuffer(
				writer, ' ', indentStep, NEW_LINE
			);
		}
	}
	
	/**
	 * Construct an <CODE>XMLWriter</CODE> to write to a file.
	 *
	 * @param name used for logging and preferences.
	 * @param filename where to write to.
	 *
	 * @throws NullPointerException if <CODE>filename==null</CODE>
	 * and the preference variable
	 * <CODE>destination</CODE> is undefined.
	 * @throws IOException if the file cannot be opened for writing.
	 */
	public XMLWriter ( final String name, final String filename )
		throws IOException
	{
		componentHelper = new ComponentHelper( name, this );

		this.filename = filename != null
			? filename
			: componentHelper.getPreference( "destination", null );
		if ( this.filename == null ) {
			final String desc
			= "No preference variable \"destination\" found"
				+ " to determine the output file.";
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"XMLWriter()",
				desc
			);
			throw new NullPointerException( desc );
		}

		{
			outputStream = new RandomFileOutputStream(
				this.filename
			);
			XMLOutputStreamWriter target = null;
			final String configuredEncoding = componentHelper.getPreference(
				"encoding", null
			);
			try {
				target = new XMLOutputStreamWriter(
					outputStream, configuredEncoding
				);
			}
			catch ( final UnsupportedEncodingException ue ) {
				componentHelper.logger.log(
					Level.WARNING,
					componentHelper.logName,
					"XMLWriter()",
					"Configured encoding \"" + configuredEncoding
					+ "\" is not supported on this platform."
					+ " Trying platform default instead.",
					ue
				);
				try {
					target = new XMLOutputStreamWriter(
						outputStream
					);
				}
				catch ( final UnsupportedEncodingException ue2 ) {
					componentHelper.logger.log(
						Level.SEVERE,
						componentHelper.logName,
						"XMLWriter()",
						"Default encoding is unavailable.",
						ue2
					);
				}
			}
			if ( configuredEncoding == null ) {
				componentHelper.logger.log(
					Level.CONFIG,
					componentHelper.logName,
					"XMLWriter()",
					"No encoding configured. Using platform default \""
					+ target.getEncoding()
					+ "\", which may not be supported by XML parsers"
					+ " even on the same platform."
					+ " Advice: Best configure preference \"encoding\""
					+ " to equal \"UTF-8\"."
				);
			}
			writer = new BufferedWriter( target );
		}

		final String configuredIndent = componentHelper.getPreference(
			"indent", "tab"
		);
		if ( "tab".equalsIgnoreCase( configuredIndent ) ) {
			buffer = new CheckingXMLWriteBuffer(
				writer, '\t', 1, NEW_LINE
			);
		} else {
			int indentStep = 0;
			try {
				indentStep = Integer.parseInt( configuredIndent );
			}
			catch ( final NumberFormatException nfe ) {
				componentHelper.logger.log(
					Level.WARNING,
					componentHelper.logName,
					"XMLWriter()",
					"Preference value \"indent\" should be numeric"
					+ " or \"tab\", but it is \""
					+ configuredIndent + "\". Using no indentation."
				);
			}
			buffer = new CheckingXMLWriteBuffer(
				writer, ' ', indentStep, NEW_LINE
			);
		}
	}

	/**
	 * Construct an <CODE>XMLWriter</CODE>
	 * writing to a file according to preferences.
	 *
	 * @param name used for logging and preferences.
	 *
	 * @throws NullPointerException if the preference variable
	 * <CODE>destination</CODE> is undefined.
	 * @throws IOException if the file cannot be opened for writing.
	 */
	public XMLWriter ( final String name )
		throws IOException
	{
		this( name, (String)null );
	}

	/* ===================== Public Methods ===================== */

	public String getName () {
		return componentHelper.name;
	}

	/**
	 * <P>Close the <CODE>XMLWriter</CODE>.</P>
	 *
	 * <P>This method should always be called
	 * when the <CODE>XMLWriter</CODE>
	 * has served its purpose.</P>
	 */
	public void close () {
		try {
			componentHelper.logger.log(
				Level.FINE,
				componentHelper.logName,
				"close()",
				"Closing"
			);
			setActive( false );

			buffer.close();
			writer.close();
		}
		catch ( final IOException ioe ) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"close()",
				"Closing failed.",
				ioe
			);
		}
	}

	/* =================== Protected Methods ==================== */

	/* --------------------- Basic Writing ---------------------- */

	/**
	 * Directly write characters to the output stream.
	 * Use with care, e.g. to write XML comments.
	 */
	protected void writeRaw ( final String string ) throws IOException {
		writer.write( string );
	}

	/** Open an element with the given name.
	 * @param attributeComparator sorts the attributes by name
	 */
	protected void openElement (
		final String name, final Comparator attributeComparator
	) throws IOException {
		if ( attributeComparator != null ) {
			buffer.openSortedElement( name, attributeComparator );
		} else {
			buffer.openElement( name );
		}
	}

	/** Open an element with the given name. */
	protected void openElement ( final String name ) throws IOException {
		buffer.openElement( name );
	}

	/** Add an attribute with the given name to a previously opened element. */
	protected void writeAttribute (
		final String name, final String value
	) throws IOException {
		buffer.writeAttribute( name, value );
		writer.flush();
	}

	
	/** Write character content. */
	protected void writeCharacters (
		final String name
	) throws IOException {
		buffer.writeCharacters( name );
	}

	/** Write a simple element with given name and character content. */
	protected void writeElement (
		final String name, final String value
	) throws IOException {
		buffer.writeElement( name, value );
	}

	/** Close the element with given name. */
	protected void closeElement ( final String name ) throws IOException {
		buffer.closeElement( name );
	}
	
	protected void removeExtraSpacesFromSeriesTagsAndCleanAnomalies() throws IOException {
		// firstly make sure all content is available
		writer.flush();
		
		File tempFileObject = new File(filename + "~");
		// get a channel to the output stream
		FileChannel sourceChannel = outputStream.getRandomFile().getChannel();
		RandomAccessFile tempFile = new RandomAccessFile(tempFileObject, "rw");
		FileChannel tempChannel = tempFile.getChannel();
		 
		long fileSize = outputStream.getRandomFile().length(); // some info on current status
	    tempFile.setLength(0); // discard temporary content, if any
        sourceChannel.transferTo(0, fileSize,tempChannel); // update temporary content with current status

        //empty the new output (the new output is the original source)
        outputStream.getRandomFile().setLength(0);
     
        //prepare the input
        BufferedReader brTemp = new BufferedReader(new FileReader(tempFileObject));
        
        //read the input, process, write to output
        String currentLine;
        while ( (currentLine=brTemp.readLine())!=null) {
        	if (currentLine.matches("(\\s*)<Series(.+)")) {
                // remove all extra spaces before end of tag but keep indentation
        		writer.write(currentLine.replaceAll("(\\s*)<Series(.*?)(\\s+)>","$1<Series$2>")+ NEW_LINE);
        	} else {
        		writer.write(currentLine + NEW_LINE);
        	}
        }
        
		//sourceChannel.close(); - do not close the source, it's done elsewhere !
		tempChannel.close();
		tempFile.close();
		brTemp.close();
		tempFileObject.delete();

	}
}
