// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2002-2004, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.xml;

import java.io.BufferedReader;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.StringReader;
import java.util.Calendar;
import java.util.Date;

import junit.framework.TestCase;
import Int.ecb.gesmescb.Attribute;
import Int.ecb.gesmescb.CodeList;
import Int.ecb.gesmescb.Contact;
import Int.ecb.gesmescb.ContactChannel;
import Int.ecb.gesmescb.InterchangeBegin;
import Int.ecb.gesmescb.InterchangeEnd;
import Int.ecb.gesmescb.KeyFamily;
import Int.ecb.gesmescb.KeyFamilyComponent;
import Int.ecb.gesmescb.MessageBegin;
import Int.ecb.gesmescb.MessageEnd;
import Int.ecb.gesmescb.Observation;
import Int.ecb.gesmescb.StatisticalConcept;
import Int.ecb.gesmescb.TimeSeries;
import Int.ecb.gesmescb.event.TestEventSource;
import Int.ecb.gesmescb.metabase.Metabase;

/* ***************************** Class ****************************** */

/**
 * Provides basic facilities to test {@link XMLWriter} classes.
 *
 * @author
 * <a href="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</a>
 * @author
 * <a href="mailto:Paula.Rocha@ECB.int">Paula Rocha</a>
 *
 * @version @(#)$Id: XMLWriterTest.java,v 1.16 2014/10/01 12:46:52 trainot Exp $
 */
public abstract class XMLWriterTest extends TestCase {

	/* ======================= Constants ======================== */

	private static final boolean DEBUG = false;

	protected static final String WILDCARD = "!WILDCARD!";

	protected static final String NEW_LINE
	= System.getProperty( "line.separator" );
	
	protected  RandomFileOutputStream outputStream;

	/**
	 * Version date of the Key Family.
	 */
	protected static final Date VERSION_DATE;
	static {
		final Calendar calendar = Calendar.getInstance();
		calendar.clear();
		calendar.set( 2002, Calendar.SEPTEMBER, 11, 14, 10 );
		VERSION_DATE = calendar.getTime();
	}

	protected static final InterchangeBegin INTERCHANGE_BEGIN
	= new InterchangeBegin(
		"IREF000012", "BBB", "ZZZ", "020817:1748", true
	);

	protected static final InterchangeEnd INTERCHANGE_END
	= new InterchangeEnd( "IREF000012" );

	protected static final MessageBegin STRUCTURE_MESSAGE_BEGIN
	= new MessageBegin(
		"MREF000001", "BBB", "ZZZ", "BODI",
		MessageBegin.STATISTICAL_DEFINITIONS,
		"The subject",
		new Contact[]{}
	);

	protected static final MessageBegin MESSAGE_BEGIN
	= new MessageBegin(
		"MREF000001", "BBB", "ZZZ", "BODI",
		MessageBegin.STATISTICAL_DATA,
		"The subject",
		new Contact[]{
			new Contact(
				Contact.INFORMATION_PRODUCTION, "BOP", "BOP Hotline",
				new ContactChannel[]{
					new ContactChannel(
						ContactChannel.TELEPHONE,
						"++34987230763"
					)
				}
			),
			new Contact(
				Contact.HEAD_OF_INFORMATION_PRODUCTION, "BOP", "BOP Head",
				new ContactChannel[]{
					new ContactChannel(
						ContactChannel.FAX,
						"++34987234762"
					)
				}
			),
			new Contact(
				Contact.DATA_PROCESSING, "SIS", "External Systems Hotline",
				new ContactChannel[]{
					new ContactChannel(
						ContactChannel.EMAIL,
						"sis.external@ecb.int"
					)
				}
			),
			new Contact(
				Contact.HEAD_OF_DATA_PROCESSING, "SIS", "SIS Head",
				new ContactChannel[]{
					new ContactChannel(
						ContactChannel.X400,
						"CO=DE; REST=unknown"
					)
				}
			),
		},
		"DS1",
		MessageBegin.UPDATE_OR_REPLACE_ACTION,
		"200011011310",
		"200001200012",
		"710",
		"KFSUMPF"
	);

	protected static final MessageBegin DELETE_MESSAGE_BEGIN
	= new MessageBegin(
		"MREF000001", "BBB", "ZZZ", "BODI",
		MessageBegin.STATISTICAL_DATA,
		"Subject is \"Deletor\"",
		new Contact[]{},
		"DS1",
		MessageBegin.DELETE_ACTION,
		"200011011310",
		"2000120002",
		"708",
		"KFSUMPF"
	);

	protected static final MessageEnd MESSAGE_END
	= new MessageEnd( "MREF000001" );

	protected static final CodeList CODE_LIST
	= new CodeList(
		"CL_SUMPF", "Sumpfige CodeList", "", VERSION_DATE,
		new String[]{ "S", "U", "M", "P", "F" },
		new String[]{ "Sumpf", "Und", "Manche", "Parteien", "Finanzierung" }
	);

	protected static final StatisticalConcept CONCEPT
	= new StatisticalConcept( "SUMPF", "Sumpfiges Concept", "", VERSION_DATE );

	protected static final KeyFamily KEY_FAMILY
	= new KeyFamily(
		"KFSUMPF", "Sample Key Family", "", VERSION_DATE,
		new KeyFamilyComponent[]{
			new KeyFamilyComponent(
				"DS_COMMENT", KeyFamilyComponent.ATTRIBUTE,
				"", "AN..64", KeyFamilyComponent.CONDITIONAL,
				KeyFamilyComponent.ATTACHED_TO_DATA_SET,
				""
			),
			new KeyFamilyComponent(
				"TIME_PERIOD", KeyFamilyComponent.TIME_DIMENSION,
				"3", "N..16", "", "", ""
			),
			new KeyFamilyComponent(
				"TIME_FORMAT", KeyFamilyComponent.TIME_DIMENSION,
				"4", "N3", "", "", "CL_TIME_FORMAT"
			),
			new KeyFamilyComponent(
				"FREQ", KeyFamilyComponent.KEY_STRUCTURE_DIMENSION,
				"1", "A1", "", "", "CL_FREQ"
			),
			new KeyFamilyComponent(
				"TITLE", KeyFamilyComponent.ATTRIBUTE,
				"", "AN..64", KeyFamilyComponent.CONDITIONAL,
				KeyFamilyComponent.ATTACHED_TO_SIBLING_GROUP,
				""
			),
			new KeyFamilyComponent(
				"OBS_PRE_BREAK", KeyFamilyComponent.ARRAY_CELL,
				"8", "AN..15", KeyFamilyComponent.CONDITIONAL, "", ""
			),
			new KeyFamilyComponent(
				"OBS_CONF", KeyFamilyComponent.ARRAY_CELL,
				"7", "A1", KeyFamilyComponent.CONDITIONAL, "",
				"CL_OBS_CONF"
			),
			new KeyFamilyComponent(
				"SUMPF", KeyFamilyComponent.KEY_STRUCTURE_DIMENSION,
				"2", "A1", "", "", "CL_SUMPF"
			),
			new KeyFamilyComponent(
				"OBS_COM", KeyFamilyComponent.ATTRIBUTE,
				"", "AN..1050", KeyFamilyComponent.CONDITIONAL,
				KeyFamilyComponent.ATTACHED_TO_OBSERVATION,
				""
			),
			new KeyFamilyComponent(
				"OBS_STATUS", KeyFamilyComponent.ARRAY_CELL,
				"6", "A1", KeyFamilyComponent.MANDATORY, "",
				"CL_OBS_STATUS"
			),
			new KeyFamilyComponent(
				"OBS_VALUE", KeyFamilyComponent.ARRAY_CELL,
				"5", "AN..16", "", "", ""
			),
			new KeyFamilyComponent(
				"UNIT", KeyFamilyComponent.ATTRIBUTE,
				"", "AN..4", KeyFamilyComponent.MANDATORY,
				KeyFamilyComponent.ATTACHED_TO_TIME_SERIES,
				"CL_UNIT"
			)
		}
	);

	protected static final TimeSeries SERIES1
	= new TimeSeries( "M.P", "200301200302", "710", new Observation[]{
		new Observation( "18.5", "A", "", "" ),
		new Observation( "-", "M", "C", "" ),
	} );

	protected static final TimeSeries SERIES2
	= new TimeSeries( "M.U", "", "", new Observation[]{} );

	protected static final TimeSeries SERIES3
	= new TimeSeries( "Q.P", "2000120002", "708", new Observation[]{} );

	protected final Attribute DS_ATT = new Attribute(
		"", "", "", "DS_COMMENT", "dscom"
	);

	protected final Attribute SG_ATT = new Attribute(
		".P", "", "", "TITLE", "sgcom"
	);

	protected final Attribute TS_ATT = new Attribute(
		"M.P", "", "", "UNIT", "tscom"
	);

	protected final Attribute OBS_ATT = new Attribute(
		"M.P", "200303", "610", "OBS_COM", "obscom"
	);

	/* ========================= Fields ========================= */

	/**
	 * The file used for the location of test data.
	 */
	private File testFile;

	/* ====================== Constructors ====================== */

	public XMLWriterTest( final String name ) {
		super( name );
		
	}

	/* ================== Protected Methods ===================== */

	/** Sets up the test fixture. */
	protected void setUp () {
		try {
			testFile = File.createTempFile( "xml_testoutput.", null );
			outputStream = new  RandomFileOutputStream(testFile.getAbsoluteFile());
			if ( DEBUG ) {
				System.out.println( "Test file: " + testFile );
			} else {
				testFile.deleteOnExit();
			}
		}
		catch ( IOException ioe ) {
			fail(
				"Write testFile failed: " + ioe
			);
		}
		System.setProperty(
			getClassName() + ".destination",
			testFile.getPath()
		);
		System.setProperty(
			getClassName() + ".indent", "tab"
		);
		System.setProperty(
			getClassName() + ".encoding", "UTF-8"
		);
		System.setProperty(
			"Int.ecb.gesmescb.xml.KeyFamilyURIMapping.BODI.KFSUMPF.namespace",
			"http://www.sumpfland.org/vocabulary/2002-09-12/kfsumpf"
		);
		final Metabase metabase = Metabase.getMetabase();
		metabase.handle( INTERCHANGE_BEGIN );
		metabase.handle( STRUCTURE_MESSAGE_BEGIN );
		metabase.handle( KEY_FAMILY );
	}

	/** Tears down the test fixture. */
	protected void tearDown () {
		System.getProperties().remove(
			"Int.ecb.gesmescb.xml.XMLWriter.uri.BODI.KFSUMPF"
		);
		System.getProperties().remove(
			getClassName() + ".encoding"
		);
		System.getProperties().remove(
			getClassName() + ".indent"
		);
		System.getProperties().remove(
			getClassName() + ".destination"
		);
		if ( !DEBUG ) {
			testFile.delete();
		}
	}

	/** Returns the qualified name of the class to be tested. */
	protected abstract String getClassName ();

	/** Returns an instance of the class to be tested.*/
	protected abstract XMLWriter factory () throws IOException;
	
	protected abstract XMLWriter streamFactory () throws IOException;

	protected void writeTempFile ( final Object object ) throws IOException {
		writeTempFile( new Object[]{ object } );
	}

	protected void writeTempFile ( final Object[] objects ) throws IOException {
		final TestEventSource source = new TestEventSource();
		final XMLWriter writer = factory();
	
		source.addListener( writer, true );
		source.dispatch( objects );

		writer.close();
	}
	
	protected void writeStream ( final Object[] objects ) throws IOException {
		final TestEventSource source = new TestEventSource();
		final XMLWriter writer = streamFactory();

		source.addListener( writer, true );
		source.dispatch( objects );

		writer.close();
	}

	
	protected String readTempFile () throws IOException {
		final byte[] buffer = new byte[ (int) testFile.length() ];
		final InputStream in = new FileInputStream( testFile );
		in.read( buffer );
		in.close();
		return new String( buffer );
	}
	
	protected String readStream () throws IOException {
		return new String( this.outputStream.toString() );
	}
	
	protected void assertEqualsLinewise (
		final String expected, final String actual
	) throws Exception {
		final BufferedReader expectedReader = new BufferedReader(
			new StringReader( expected )
		);
		final BufferedReader actualReader = new BufferedReader(
			new StringReader( actual )
		);
		String expectedLine, actualLine;
		int lineNo = 0;
		do {
			++lineNo;
			expectedLine = expectedReader.readLine();
			actualLine = actualReader.readLine();
		} while (
			expectedLine != null
			&&
			actualLine != null
			&&
			equalsWithWildcardStars( expectedLine, actualLine )
		);
		if ( expectedLine == null ) {
			assertNull( "Line " + lineNo, actualLine );
		} else {
			assertEquals( "Line " + lineNo, expectedLine, actualLine );
		}
		expectedReader.close();
		actualReader.close();
	}

	/* ==================== Private Methods ===================== */

	private static boolean equalsWithWildcardStars(
		final String expected, final String actual
	) {
		final int wildBegin = expected.indexOf( WILDCARD );
		if ( wildBegin < 0 ) return expected.equals( actual );

		final String expectedBefore = expected.substring( 0, wildBegin );
		if ( actual.length() < wildBegin ) {
			// fail
			return expected.equals( actual );
		}

		final String actualBefore = actual.substring( 0, wildBegin );
		final String expectedAfter = expected.substring(
			wildBegin + WILDCARD.length()
		);
		final String actualAfter = actual.substring(
			actual.length() - expectedAfter.length()
		);
		return
			expectedBefore.equals( actualBefore )
			&&
			expectedAfter.equals( actualAfter );
	}

	/* ================== Protected Methods ===================== */

	protected static String moreIndent (
		final String text, final int indent
	) throws IOException {
		final BufferedReader reader = new BufferedReader(
			new StringReader( text )
		);
		final StringBuffer buffer = new StringBuffer();

		for ( int i = 0; i < indent; ++i ) buffer.append( '\t' );
		final String indentation = buffer.toString();

		buffer.delete( 0, buffer.length() );
		String line;
		while ( ( line = reader.readLine() ) != null ) {
			buffer.append( indentation );
			buffer.append( line );
			buffer.append( NEW_LINE );
		}
		reader.close();
		return buffer.toString();
	}
}
