// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2004, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.xml;

import java.io.IOException;

import java.util.Collection;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.Map;
import java.util.HashMap;
import java.util.TreeMap;
import java.util.Date;

import Int.ecb.logging.Level;

import Int.ecb.gesmescb.*;
import Int.ecb.gesmescb.time.RegularInterval;
import Int.ecb.gesmescb.metabase.*;

import Int.ecb.gesmescb.event.StructureListener;
import Int.ecb.gesmescb.event.DataListener;

import Int.ecb.gesmescb.util.KeyTokenizer;

/* ***************************** Class ****************************** */

/**
 * <p>Translates {@link Int.ecb.gesmescb.GESMESCBObject}s
 * into SDMX Case Study's XML draft text.
 * This text then is written to an output {@link java.io.Writer}
 * or to a file according to preference settings.</p>
 *
 * <p><blink>Please note</blink>
 * that this class is intended only for demonstration,
 * not for production use.</p>
 *
 * <center><table border="3" cellpadding="3">
 * <caption>
 * <a name="prefs">Available {@link Int.ecb.prefs preference} settings</a>
 * </caption>
 * <tr>
 * <th>preference name</th>
 * <th>type</th>
 * <th>meaning</th>
 * <th>default</th>
 * </tr>
 * <tr><td><code>destination</td><td>string</td><td>output file path</td></tr>
 * <tr><td><code>encoding</td><td>string</td><td>character encoding</td>
 * <td>platform default, which may not be understood by XML parsers;
 * better use UTF-8 or UTF-16</td></tr>
 * <tr><td><code>indent</td><td>string or integer</td>
 * <td>if integer, then added indentation for enclosed elements;
 * if "<tt>tab</tt>", then use 1 tab character
 * to indent</td><td><tt>tab</tt></td></tr>
 * <tr><td><code>schemaLocation</code></td>
 * <td>string</td>
 * <td>URI to the CaseStudy-XML schema</td></tr>
 * </table></center>
 *
 * <p>The <code>CaseStudyXMLWriter</code> is not internally synchronized.
 * External synchronisation is necessary if multiple threads want to
 * use the same <code>CaseStudyXMLWriter</code>.</p>
 *
 * @author
 * <a href="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</a>
 *
 * @version @(#)$Id: CaseStudyXMLWriter.java,v 1.1 2017/01/19 16:14:03 trainot Exp $
 *
 * @see KeyFamilyURIMapping
 */
public class CaseStudyXMLWriter
	extends XMLWriter
	implements StructureListener, DataListener, Constants
{
	/* ========================= Fields ========================= */

	/** Current XML dateTime format.
	 * Changes its pattern according to frequency. */
	private final XMLDateFormat xmlDateFormat;

	/** Links key families to their namespace URIs. */
	private final KeyFamilyURIMapping uriMapping;

	/** The current <code>MessageBegin</code>. */
	private MessageBegin messageBegin;

	/** The current <code>KeyFamily</code>. */
	private KeyFamily keyFamily;

	/** Current {@link KeyFamilyComponent}s. */
	private final Collection
		dataSetAttributes,
		siblingGroupAttributes,
		timeSeriesAttributes,
		observationAttributes;

	/** Temporary structural definitions store. */
	private final Collection codeLists, concepts, keyFamilies;

	/** Temporary concept descriptions map. */
	private final Map conceptNames;

	/**
	 * Temporary data store.
	 * It (at least indirectly) holds all data objects in a message,
	 * so it consumes potentially unlimited amounts of memory.
	 * TODO: Work out a better scheme.
	 */
	private final Map dataSet;

	/* ====================== Constructors ====================== */

	/**
	 * Construct an <code>CaseStudyXMLWriter</code> to write to a file.
	 *
	 * @param name used for logging and preferences.
	 * @param filename where to write to.
	 *
	 * @throws NullPointerException if <code>filename==null</code>
	 * and the preference variable
	 * <code>destination</code> is undefined.
	 * @throws IOException if the file cannot be opened for writing.
	 */
	public CaseStudyXMLWriter ( final String name, final String filename )
		throws IOException
	{
		super( name, filename );

		xmlDateFormat = new XMLDateFormat();
		uriMapping = KeyFamilyURIMapping.getInstance();

		messageBegin = null;
		keyFamily = null;

		dataSetAttributes = new ArrayList();
		siblingGroupAttributes = new ArrayList();
		timeSeriesAttributes = new ArrayList();
		observationAttributes = new ArrayList();

		codeLists = new ArrayList();
		concepts = new ArrayList();
		keyFamilies = new ArrayList();

		conceptNames = new HashMap();
		dataSet = new TreeMap();

		writeXMLHeader();
	}

	/**
	 * Construct an <code>CaseStudyXMLWriter</code>
	 * writing to a file according to preferences.
	 *
	 * @param name used for logging and preferences.
	 *
	 * @throws NullPointerException if the preference variable
	 * <code>destination</code> is undefined.
	 * @throws IOException if the file cannot be opened for writing.
	 */
	public CaseStudyXMLWriter ( final String name )
		throws IOException
	{
		this( name, null );
	}

	/* ===================== Public Methods ===================== */

	/* ------------------------ Listener ------------------------ */

	/* ---------------------- Interchange ----------------------- */

	public void handle ( final InterchangeBegin interchangeBegin ) {
		try {
			componentHelper.logger.log(
				Level.INFO,
				componentHelper.logName,
				"handle( InterchangeBegin )",
				"Begin writing interchange "
				+ interchangeBegin.getIref() + "."
			);

			openElement( "SDMXTimeSeries" );
			writeAttribute( "xmlns", XMLNS_CASESTUDY );
			writeAttribute( "xmlns:xsi", XMLNS_XSI );

			final String schema = componentHelper.getPreference(
				"schemaLocation", null
			);
			if ( schema != null ) {
				writeAttribute(
					"xsi:schemaLocation",
					XMLNS_CASESTUDY + ' ' + schema
				);
			}

			openElement( "Header" );
			writeElement( "Name", interchangeBegin.getIref() );
			writeElement( "Sender", interchangeBegin.getSender() );
			writeElement( "Receiver", interchangeBegin.getReceiver() );
			writeElement(
				"TestIndicator",
				interchangeBegin.isTest() ? "true" : "false"
			);

			final Date preparationDate
			= interchangeBegin.getPreparationDateTimeAsDate();
			if ( preparationDate != null ) {
				xmlDateFormat.applyPattern( XMLDateFormat.SECONDLY_PATTERN );
				final String prepDateTimeXML = xmlDateFormat.format(
					preparationDate
				);
				writeElement( "Prepared", prepDateTimeXML );
			}

			closeElement( "Header" );
		}
		catch ( final IOException ioe ) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"handle( InterchangeBegin )",
				"Output problem.",
				ioe
			);
			close();
		}
	}

	public void handle ( final InterchangeEnd interchangeEnd ) {
		try {
			closeElement( "SDMXTimeSeries" );

			componentHelper.logger.log(
				Level.INFO,
				componentHelper.logName,
				"handle( InterchangeEnd )",
				"Ended writing interchange "
				+ interchangeEnd.getIref() + "."
			);
		}
		catch ( final IOException ioe ) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"handle( InterchangeEnd )",
				"Output problem.",
				ioe
			);
			close();
		}
	}

	/* ------------------------ Message ------------------------- */

	public void handle ( final MessageBegin messageBegin ) {
		this.messageBegin = messageBegin;

		try {
			componentHelper.logger.log(
				Level.FINE,
				componentHelper.logName,
				"handle( MessageBegin )",
				"Begin writing Message " + messageBegin.getMref() + "."
			);

			if ( messageBegin.isStatisticalData() ) {
				final String maintenanceAgencyId
				= messageBegin.getMaintenanceAgencyId();
				final Metabase metabase = Metabase.getMetabase();
				final MaintenanceAgency maintenanceAgency
				= metabase.getMaintenanceAgency( maintenanceAgencyId );
				if ( maintenanceAgency != null ) {
					final String keyFamilyId = messageBegin.getKeyFamilyId();
					keyFamily = maintenanceAgency.getKeyFamily( keyFamilyId );
					initAttributeLists();
					if ( keyFamily != null ) {
						openElement( "StructureDefinition" );
						openElement( "KeyFamilies" );
						openElement( "KeyFamily" );
						writeAttribute( "uniqueID", keyFamilyId );
						writeAttribute( "name", keyFamily.getName() );
						openElement( "ExtKeyRef" );
						writeAttribute(
							"extKeyURI",
							uriMapping.xmlns( maintenanceAgencyId, keyFamily )
						);
						writeCharacters( keyFamily.getName() );
						closeElement( "ExtKeyRef" );
						closeElement( "KeyFamily" );
						closeElement( "KeyFamilies" );
						closeElement( "StructureDefinition" );
					} else {
						componentHelper.logger.log(
							Level.SEVERE,
							componentHelper.logName,
							"handle( MessageBegin )",
							"Maintenance agengy " + maintenanceAgencyId
							+ " does not know key family \""
							+ keyFamilyId + "\"."
						);
						close();
					}
				} else {
					componentHelper.logger.log(
						Level.SEVERE,
						componentHelper.logName,
						"handle( MessageBegin )",
						"Unknown maintenance agengy \"" + maintenanceAgencyId
						+ "\"."
					);
					close();
				}
			} else {
				keyFamily = null;
				initAttributeLists();
			}
		}
		catch ( final IOException ioe ) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"handle( MessageBegin )",
				"Output problem.",
				ioe
			);
			close();
		}
	}

	public void handle ( final MessageEnd messageEnd ) {
		try {
			if ( messageBegin.isStatisticalDefinitions() ) {
				openElement( "StructureDefinition" );
				if ( !keyFamilies.isEmpty() ) {
					openElement( "KeyFamilies" );
					for (
						final Iterator iterator = keyFamilies.iterator();
						iterator.hasNext();
					) {
						final KeyFamily keyFamily = (KeyFamily) iterator.next();
						write( keyFamily );
					}
					closeElement( "KeyFamilies" );
					keyFamilies.clear();
				}
				if ( !codeLists.isEmpty() ) {
					openElement( "CodeLists" );
					for (
						final Iterator iterator = codeLists.iterator();
						iterator.hasNext();
					) {
						final CodeList codeList = (CodeList) iterator.next();
						write( codeList );
					}
					closeElement( "CodeLists" );
					codeLists.clear();
				}
				if ( !concepts.isEmpty() ) {
					openElement( "ConceptList" );
					for (
						final Iterator iterator = concepts.iterator();
						iterator.hasNext();
					) {
						final StatisticalConcept concept = (StatisticalConcept) iterator.next();
						write( concept );
					}
					closeElement( "ConceptList" );
					concepts.clear();
					conceptNames.clear();
				}
				closeElement( "StructureDefinition" );
			} else {
				openElement( "Data" );
				if ( !dataSet.isEmpty() ) {
					openElement( "DataSet" );
					writeAttribute( "name", messageBegin.getDataSetId() );
					writeAttributeList( dataSet, dataSetAttributes );
					writeElement( "KeyRef", messageBegin.getKeyFamilyId() );
					for (
						final Iterator groupKeys = dataSet.keySet().iterator();
						groupKeys.hasNext();
					) {
						final String groupKey = (String) groupKeys.next();
						final Map siblingGroup
						= (Map) dataSet.get( groupKey );
						groupKeys.remove();

						openElement( "SiblingGroup" );
						writeAttribute(
							"variableDimension",
							keyFamily.getFrequency().getConceptId()
						);
						writeAttributeList(
							siblingGroup, siblingGroupAttributes
						);
						for (
							final Iterator seriesKeys
							= siblingGroup.keySet().iterator();
							seriesKeys.hasNext();
						) {
							final String seriesKey = (String) seriesKeys.next();
							final Map timeSeries
							= (Map) siblingGroup.get( seriesKey );
							seriesKeys.remove();

							openElement( "TimeSeries" );
							writeAttributeList(
								timeSeries, timeSeriesAttributes
							);
							writeKey( seriesKey );
							if ( !timeSeries.isEmpty() ) {
								openElement( "Observations" );
								for ( final Iterator periods
									= timeSeries.keySet().iterator();
									periods.hasNext();
								) {
									final String period
									= (String) periods.next();
									final Map observation
									= (Map) timeSeries.get( period );
									periods.remove();

									openElement( "Observation" );
									writeAttributeList(
										observation, observationAttributes
									);

									writeElement( "Time", period );

									final String value
									= (String) observation.remove(
										keyFamily.getObsValue().getConceptId()
									);
									if ( value != null ) {
										writeElement( "Value", value );
									}

									closeElement( "Observation" );
								}

								closeElement( "Observations" );
							}
							closeElement( "TimeSeries" );
						}
						closeElement( "SiblingGroup" );
					}
					closeElement( "DataSet" );
				}
				closeElement( "Data" );
			}

			componentHelper.logger.log(
				Level.FINE,
				componentHelper.logName,
				"handle( MessageBegin )",
				"End writing message "
				+ messageEnd.getMref() + "."
			);
		}
		catch ( final IOException ioe ) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"handle( MessageEnd )",
				"Output problem.",
				ioe
			);
			close();
		}
	}

	/* ----------------------- Code List ------------------------ */

	public void handle ( final CodeList codeList ) {
		codeLists.add( codeList );
	}

	private void write ( final CodeList codeList ) {
		final String id = codeList.getId();
		try {
			openElement( "CodeList" );
			writeAttribute( "uniqueID", id );

			final String name = codeList.getName();
			if ( !"".equals( name ) ) {
				writeAttribute( "name", name );
			}

			final String description = codeList.getDescription();
			if ( !"".equals( description ) ) {
				writeAttribute( "description", description );
			}

			openElement( "Codes" );

			for (
				final Iterator
					values = codeList.codeValueIterator(),
					descriptions = codeList.codeDescriptionIterator();
				values.hasNext() && descriptions.hasNext();
			) {
				final String value = (String) values.next();
				openElement( "Code" );
				writeElement( "Value", value );

				final String desc = (String) descriptions.next();
				if ( !"".equals( desc ) ) {
					writeElement( "Description", desc );
				}

				closeElement( "Code" );
			}

			closeElement( "Codes" );
			closeElement( "CodeList" );
		}
		catch ( final IOException ioe ) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"handle( CodeList )",
				"Output problem.",
				ioe
			);
			close();
		}
	}

	/* ------------------ Statistical Concept ------------------- */

	public void handle ( final StatisticalConcept statisticalConcept ) {
		concepts.add( statisticalConcept );
		conceptNames.put(
			statisticalConcept.getId(),
			statisticalConcept.getName()
		);
	}

	private void write ( final StatisticalConcept statisticalConcept ) {
		final String id = statisticalConcept.getId();
		try {
			openElement( "Concept" );
			writeElement( "Identifier", id );

			final String name = statisticalConcept.getName();
			if ( !"".equals( name ) ) {
				writeElement( "Label", name );
			}

			closeElement( "Concept" );
		}
		catch ( final IOException ioe ) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"handle( StatisticalConcept )",
				"Output problem.",
				ioe
			);
			close();
		}
	}

	/* ----------------------- Key Family ----------------------- */

	public void handle ( final KeyFamily keyFamily ) {
		keyFamilies.add( keyFamily );
	}

	private void write ( final KeyFamily keyFamily ) {
		final String id = keyFamily.getId();
		try {
			openElement( "KeyFamily" );
			writeAttribute( "uniqueID", id );

			final String name = keyFamily.getName();
			if ( !"".equals( name ) ) {
				writeAttribute( "name", name );
			}

			final String description = keyFamily.getDescription();
			if ( !"".equals( description ) ) {
				writeAttribute( "description", description );
			}

			openElement( "Components" );

			for (
				final Iterator iterator = keyFamily.componentIterator();
				iterator.hasNext();
			) {
				final KeyFamilyComponent component
				= (KeyFamilyComponent) iterator.next();
				if (
					component.isKeyStructureDimension()
					||
						component.isArrayCell()
						&&
						keyFamily.getObsValue() != component
					||
					component.isAttribute()
				) {
					write( component, keyFamily );
				}
			}

			closeElement( "Components" );
			closeElement( "KeyFamily" );
		}
		catch ( final IOException ioe ) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"handle( KeyFamily )",
				"Output problem.",
				ioe
			);
			close();
		}
	}

	/* ---------------------- Time Series ----------------------- */

	public void handle ( final TimeSeries timeSeries ) {
		// TODO: Avoid clashes between attributes and group/series
		final RegularInterval range = timeSeries.getInterval();

		if (
			range == null
			||
			range.getSize() == 0
		) return;

		final String key = timeSeries.getKey();
		final String groupKey = groupKey( key );

		if ( !dataSet.containsKey( groupKey ) ) {
			dataSet.put(
				groupKey,
				new TreeMap()
			);
		}
		final Map siblingGroup = (Map) dataSet.get( groupKey );

		if ( !siblingGroup.containsKey( key ) ) {
			siblingGroup.put(
				key,
				new TreeMap()
			);
		}
		final Map series = (Map) siblingGroup.get( key );

		xmlDateFormat.setFrequency( RegularInterval.DAILY );

		for (
			final Iterator
				intervals = range.iterator(),
				observations = timeSeries.observationIterator();
			intervals.hasNext() && observations.hasNext();
		) {
			final RegularInterval interval
			= (RegularInterval) intervals.next();
			final Observation observation
			= (Observation) observations.next();
			final Date time = interval.getBegin();
			final String period = xmlDateFormat.format( time );

			if ( observation.isEmpty() ) continue;

			if ( !series.containsKey( period ) ) {
				series.put(
					period,
					new TreeMap()
				);
			}
			final Map obs = (Map) series.get( period );
			Object oldValue = null,
				oldStatus = null,
				oldConf = null,
				oldPreBreak = null;

			{
				final String value = observation.getValue();
				final KeyFamilyComponent obsValue = keyFamily.getObsValue();
				if ( obsValue != null ) {
					oldValue = obs.put(
						obsValue.getConceptId(),
						observation.hasMissingValue()
						? "NaN" : value
					);
				}
			}

			{
				final String status = observation.getStatus();
				final KeyFamilyComponent obsStatus = keyFamily.getObsStatus();
				if ( obsStatus != null ) {
					oldStatus = obs.put( obsStatus.getConceptId(), status );
				}
			}

			{
				final String conf = observation.getConfidentiality();
				final KeyFamilyComponent obsConf = keyFamily.getObsConf();
				if ( obsConf != null ) {
					oldConf = obs.put( obsConf.getConceptId(), conf );
				}
			}

			{
				final String preBreak = observation.getPreBreakValue();
				final KeyFamilyComponent obsPreBreak = keyFamily.getObsPreBreak();
				if ( obsPreBreak != null ) {
					oldPreBreak = obs.put(
						obsPreBreak.getConceptId(),
						observation.hasMissingPreBreakValue()
						? "NaN" : preBreak
					);
				}
			}

			if ( oldValue != null ) {
				componentHelper.logger.log(
					Level.WARNING,
					componentHelper.logName,
					"handle( TimeSeries )",
					"Time series " + key
					+ " overwrites previous observation "
					+ oldValue + ":"
					+ oldStatus + ":"
					+ oldConf + ":"
					+ oldPreBreak + " with "
					+ observation + " at "
					+ period + ".",
					timeSeries
				);
			}
		}
	}

	/* ----------------------- Attribute ------------------------ */

	public void handle ( final Attribute attribute ) {
		final String name = attribute.getName();
		final String value = attribute.getValue();
		final KeyFamilyComponent component = keyFamily.getComponent(
			name
		);
		if ( component.isAttachedToDataSet() ) {
			dataSet.put( name, value );
		} else {
			final String key = attribute.getKey();
			final String groupKey = groupKey( key );
			if ( !dataSet.containsKey( groupKey ) ) {
				dataSet.put(
					groupKey,
					new TreeMap()
				);
			}
			final Map siblingGroup = (Map) dataSet.get( groupKey );

			if ( component.isAttachedToSiblingGroup() ) {
				siblingGroup.put( name, value );
			} else {
				if ( !siblingGroup.containsKey( key ) ) {
					siblingGroup.put(
						key,
						new TreeMap()
					);
				}
				final Map series = (Map) siblingGroup.get( key );

				if ( component.isAttachedToTimeSeries() ) {
					series.put( name, value );
				} else {
					final RegularInterval interval = attribute.getInterval();
					xmlDateFormat.setFrequency( RegularInterval.DAILY );
					final Date time = interval.getBegin();
					final String period = xmlDateFormat.format( time );

					if ( !series.containsKey( period ) ) {
						series.put(
							period,
							new TreeMap()
						);
					}
					final Map obs = (Map) series.get( period );

					obs.put( name, value );
				}
			}
		}
	}

	/* ==================== Private Methods ===================== */

	private void writeXMLHeader () {
		componentHelper.logger.log(
			Level.FINE,
			componentHelper.logName,
			"writeXMLHeader()",
			"Writing header."
		);

		try {
			writeRaw( "<!-- SDMX Case Study's draft XML from Gesmes/TS Suite -->" );
			writeRaw( NEW_LINE );
			writeRaw( "<!-- @(#)$Id: CaseStudyXMLWriter.java,v 1.1 2017/01/19 16:14:03 trainot Exp $ -->" );
			writeRaw( NEW_LINE );

			final String stylesheetURI = componentHelper.getPreference(
				"stylesheet", null
			);
			if ( stylesheetURI != null ) {
				writeRaw(
					"<?xml-stylesheet href=\""
					+ stylesheetURI
					+ "\" type=\"text/xsl\""
					+ " media=\"screen,html\"?>"
				);
				writeRaw( NEW_LINE );
			}
			writeRaw( NEW_LINE );
		}
		catch ( final IOException ioe ) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"writeXMLHeader",
				"Output problem.",
				ioe
			);
			close();
		}
	}

	/** Write a <code>KeyFamilyComponent</code>. */
	private void write (
		final KeyFamilyComponent component,
		final KeyFamily keyFamily
	) throws IOException {
		final String id = component.getConceptId();
		final String codeListId = component.getCodeListId();
		if (
			component.isKeyStructureDimension()
			||
				component.isTimeDimension()
				&&
				component != keyFamily.getTimeFormat()
		) {
			openElement( "Dimension" );
			writeAttribute( "conceptName", id );
			if ( conceptNames.containsKey( id ) ) {
				writeAttribute(
					"description",
					(String) conceptNames.get( id )
				);
			}
			writeAttribute( "representation", codeListId );
			closeElement( "Dimension" );
		} else {
			openElement( "Attribute" );
			writeAttribute( "conceptName", id );
			if ( conceptNames.containsKey( id ) ) {
				writeAttribute(
					"description",
					(String) conceptNames.get( id )
				);
			}

			final String attachmentLevel;
			if (
				component.isArrayCell()
				||
				component.isAttachedToObservation()
			) {
				attachmentLevel = "Observation";
			} else if ( component.isAttachedToTimeSeries() ) {
				attachmentLevel = "TimeSeries";
			} else if ( component.isAttachedToSiblingGroup() ) {
				attachmentLevel = "SiblingGroup";
			} else if ( component.isAttachedToDataSet() ) {
				attachmentLevel = "DataSet";
			} else {
				attachmentLevel = component.getAttachmentLevel();
			}
			writeAttribute( "assignmentLevel", attachmentLevel );

			final String usage;
			if (
				component.isMandatory()
				||
				component == keyFamily.getObsValue()
				||
				component == keyFamily.getTimeFormat()
				// BB: I model time format as an attribute
				// because time series deletion does not require it stated.
			) {
				usage = "Mandatory";
			} else {
				usage = "Optional";
			}
			writeAttribute( "assignmentStatus", usage );

			if ( !"".equals( codeListId ) ) {
				writeElement( "Coded", codeListId );
			} else {
				final TextFormat format = component.getTextFormat();
				if ( format != null ) {
					if ( TextFormat.NUMERIC == format.getCharacterType() ) {
						openElement( "Numeric" );
						writeAttribute(
							"maxlength",
							Integer.toString( format.getReferenceLength() )
						);
						writeAttribute( "decimals", "0" );
						closeElement( "Numeric" );
					} else {
						openElement( "String" );
						writeAttribute(
							"maxlength",
							Integer.toString( format.getReferenceLength() )
						);
						closeElement( "String" );
					}
				}
			}
			closeElement( "Attribute" );
		}
	}

	private void writeKey ( final String key ) throws IOException {
		openElement( "Key" );
		for (
			final Iterator
				dimensions = keyFamily.keyStructureIterator(),
				values = new KeyTokenizer( key );
			dimensions.hasNext() && values.hasNext();
		) {
			final KeyFamilyComponent dimension
			= (KeyFamilyComponent) dimensions.next();
			final String value = (String) values.next();
			if ( "".equals( value ) ) continue;
			openElement( "KeyValue" );
			writeAttribute( "name", dimension.getConceptId() );
			writeElement( "Value", value );
			closeElement( "KeyValue" );
		}
		closeElement( "Key" );
	}

	private void writeAttributeList (
		final Map attributeMap, final Collection components
	) throws IOException {
		boolean wroteAttributeList = false;
		for (
			final Iterator iterator
			= components.iterator();
			iterator.hasNext();
		) {
			final KeyFamilyComponent component
			= (KeyFamilyComponent) iterator.next();
			final String id = component.getConceptId();
			final String value = (String) attributeMap.remove( id );
			if ( value != null ) {
				if ( !wroteAttributeList ) {
					openElement( "AttributeList" );
					wroteAttributeList = true;
				}
				openElement( "AttVal" );
				writeAttribute( "name", id );
				writeElement( "Value", value );
				closeElement( "AttVal" );
			}
		}
		if ( wroteAttributeList ) {
			closeElement( "AttributeList" );
		}
	}

	/* ----------------------- Auxiliary ------------------------ */

	private String groupKey ( final String seriesKey ) {
		final StringBuffer groupKey = new StringBuffer();
		for (
			final Iterator
				dimensions = keyFamily.keyStructureIterator(),
				dimensionValues = new KeyTokenizer( seriesKey );
			dimensions.hasNext() && dimensionValues.hasNext();
		) {
			final KeyFamilyComponent dimension
			= (KeyFamilyComponent) dimensions.next();
			final String value = (String) dimensionValues.next();
			if ( !dimension.isFrequencyDimension() ) {
				groupKey.append( value );
				groupKey.append( TimeSeries.KEY_SEPARATOR );
			}
		}
		if ( groupKey.length() > 0 ) {
			groupKey.setLength( groupKey.length() - 1 );
		}
		return groupKey.toString();
	}

	/** Initialise the attribute container contents. */
	private void initAttributeLists () {
		dataSetAttributes.clear();
		siblingGroupAttributes.clear();
		timeSeriesAttributes.clear();
		observationAttributes.clear();

		if ( keyFamily == null ) return;

		for (
			final Iterator iterator = keyFamily.componentIterator();
			iterator.hasNext();
		) {
			final KeyFamilyComponent component
			= (KeyFamilyComponent) iterator.next();

			if (
				component.isArrayCell()
				||
				component.isAttribute()
			) {
				if (
					component.isAttachedToObservation()
					||
						component.isArrayCell()
						&&
						component != keyFamily.getObsValue()
				) {
					observationAttributes.add( component );
				} else if (
					component.isAttachedToTimeSeries()
				) {
					timeSeriesAttributes.add( component );
				} else if ( component.isAttachedToSiblingGroup() ) {
					siblingGroupAttributes.add( component );
				} else if ( component.isAttachedToDataSet() ) {
					dataSetAttributes.add( component );
				}
			}
		}
	}
}
