// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2003, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.xml;

import java.io.Writer;
import java.io.IOException;

import java.util.List;
import java.util.ArrayList;
import java.util.Map;
import java.util.SortedMap;
import java.util.HashMap;
import java.util.TreeMap;
import java.util.Comparator;
import java.util.Iterator;

/* ***************************** Class ****************************** */

/**
 * <p>Helps safer and nicer writing of XML elements and attributes.</p>
 *
 * <p>It helps to ensure correct XML syntax and can sort attributes
 * for the human reader. Note that XML infoset foresees no attribute
 * ordering.</p>
 *
 * @author
 * <A HREF="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</A>
 *
 * @version @(#)$Id: CheckingXMLWriteBuffer.java,v 1.1 2017/01/19 16:14:03 trainot Exp $
 */
class CheckingXMLWriteBuffer extends XMLWriteBuffer {

	/* ========================= Fields ========================= */

	private final List elementStack;
	private Map attributeMap;

	/* ====================== Constructors ====================== */

	/**
	 * Construct a <code>CheckingXMLWriteBuffer</code> with specified
	 * indentation setting.
	 */
	public CheckingXMLWriteBuffer (
		final Writer writer,
		final char indentChar,
		final int indentStep,
		final String newLine
	) {
		super( writer, indentChar, indentStep, newLine );
		elementStack = new ArrayList();
		attributeMap = null;
	}

	/* ===================== Public Methods ===================== */

	/**
	 * Opens an element for attribute sorting.
	 * @param attributeComparator will determine the attribute order.
	 */
	public void openSortedElement (
		final String name, final Comparator attributeComparator
	) throws IOException {
		writeAttributeCache();
		super.openElement( name );
		elementStack.add( name );
		checkName( name );
		attributeMap = new TreeMap( attributeComparator );
	}

	public void openElement ( final String name ) throws IOException {
		writeAttributeCache();
		super.openElement( name );
		elementStack.add( name );
		checkName( name );
		attributeMap = new HashMap();
	}

	public void writeAttribute (
		final String name, final String value
	) throws IOException {
		if ( status != TAG_OPEN ) {
			throw new RuntimeException(
				"No tag open to add attribute \"" + name
				+ "\" to."
			);
		}
		checkName( name );
		if ( attributeMap.put( name, value ) != null ) {
			// TODO: Implement a better error response.
			throw new RuntimeException(
				"Duplicate attribute \"" + name
				+ "\"."
			);
		}
		// If no re-order, write immediate
		if ( ! ( attributeMap instanceof SortedMap ) ) {
			super.writeAttribute( name, value );
		}
	}

	public void writeCharacters ( final String string ) throws IOException {
		if ( "".equals( string ) ) return;
		writeAttributeCache();
		super.writeCharacters ( string );
	}

	public void closeElement ( final String name ) throws IOException {

		writeAttributeCache();
		super.closeElement( name );
		if ( elementStack.size() > 0 ) {
			final String opening = (String) elementStack.remove(
				elementStack.size() - 1
			);
			if ( !name.equals( opening ) ) {
				throw new RuntimeException(
					"Trying to close opening tag \"" + opening
					+ "\" with closing tag \"" + name + "\"."
				);
			}
		} else {
			throw new RuntimeException(
				"Trying no tag \"" + name
				+ "\" available to be closed."
			);
		}
		attributeMap = null;
	}

	/**
	 * <p>Close the <code>CheckingXMLWriteBuffer</code>.</p>
	 *
	 * <p>This method should always be called
	 * when the <code>CheckingXMLWriteBuffer</code>
	 * has served its purpose.</p>
	 */
	public void close () throws IOException {
		if ( elementStack.size() > 0 ) {
			final StringBuffer buffer = new StringBuffer(
				"Trying to close output with some elements remaining unclosed: "
			);
			for (
				final Iterator iterator = elementStack.iterator();
				iterator.hasNext();
			) {
				buffer.append( '"' );
				buffer.append( (String) iterator.next() );
				buffer.append( '"' );
				buffer.append( iterator.hasNext() ? ", " : "." );
			}
			throw new RuntimeException(
				buffer.toString()
			);
		}
		super.close();
	}

	/* ==================== Private Methods ===================== */

	private void checkName ( final String name ) {
		if ( "".equals( name ) ) {
			throw new RuntimeException(
				"Empty name."
			);
		}
		{
			final char c = name.charAt( 0 );
			if (
				!Character.isLetter( c )
				&&
				"_:".indexOf( c ) < 0
			) {
				throw new RuntimeException(
					"Illegal first character in name \"" + name + "\"."
				);
			}
		}
		final int length = name.length();
		for ( int i = 1; i < length; ++i ) {
			final char c = name.charAt( i );
			if (
				!Character.isLetterOrDigit( c )
				&&
				".-_:".indexOf( c ) < 0
			) {
				throw new RuntimeException(
					"Dubious character '" + c
					+ "' in name \"" + name
					+ "\" at position " + i + "."
				);
			}
		}
	}

	protected void writeAttributeCache () throws IOException {
		// If re-order, write delayed
		if ( attributeMap instanceof SortedMap ) {
			for (
				final Iterator iterator = attributeMap.entrySet().iterator();
				iterator.hasNext();
			) {
				final Map.Entry entry = (Map.Entry) iterator.next();
				final String name = (String) entry.getKey();
				final String value = (String) entry.getValue();
				super.writeAttribute( name, value );
				iterator.remove();
			}
		}
	}
}
