// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2003, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.xml;

import java.util.Iterator;
import java.util.Date;

import java.text.ParseException;

// used but not importable: import org.xml.sax.XMLReader;
import org.xml.sax.Attributes;

import Int.ecb.logging.Level;

import Int.ecb.gesmescb.Observation;
import Int.ecb.gesmescb.TimeSeries;
import Int.ecb.gesmescb.KeyFamily;
import Int.ecb.gesmescb.KeyFamilyComponent;

import Int.ecb.gesmescb.time.RegularInterval;
import Int.ecb.gesmescb.time.GregorianRegularInterval;

/* ***************************** Class ****************************** */

/**
 * <p>Processes a <code>Cube</code> hierarchy
 * from SAX2 parser output.</p>
 *
 * <center><table border="3" cellpadding="3">
 * <caption>
 * <a name="prefs">Available {@link Int.ecb.prefs preference} settings</a>
 * </caption>
 * <tr>
 * <th>preference name</th>
 * <th>type</th>
 * <th>meaning</th>
 * <th>default</th>
 * </tr>
 * <tr><td><code>fakeTimeForEmptySeries</td><td>boolean</td>
 * <td>whether a fake period start should be supplied to empty series
 * when only period format is in the data</td>
 * <td><code>false</code></td></tr>
 * </table></center>
 *
 * @author
 * <a href="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</a>
 *
 * @author
 * <a href="mailto:Sami.Airo@ECB.int">Sami Airo</a>
 *
 * @version @(#)$Id: CubeReader.java,v 1.1 2017/01/19 16:14:02 trainot Exp $
 */
final class CubeReader extends SAX2Reader {

	/* ========================= Fields ========================= */

	private final XMLDateFormat format;

	private final MapStack mapStack;

	private final boolean fakeTimeForEmptySeries;

	private KeyFamily keyFamily;

	/* ====================== Constructors ====================== */

	public CubeReader (
		final org.xml.sax.XMLReader parser,
		final SAX2Reader parent
	) {
		super( parser, parent );
		format = new XMLDateFormat();
		mapStack = new MapStack();
		fakeTimeForEmptySeries = componentHelper.getPreference(
			"fakeTimeForEmptySeries", "false"
		).equalsIgnoreCase( "true" );

		keyFamily = null;
	}

	/* ===================== Public Methods ===================== */

	/** Set the data structure. */
	public void setKeyFamily ( final KeyFamily keyFamily ) {
		this.keyFamily = keyFamily;
	}

	/* --------------------- ContentHandler --------------------- */

	public void startElement (
		final String namespaceURI,
		final String localName,
		final String qName,
		Attributes atts
	) {
		super.startElement( namespaceURI, localName, qName, atts );
		mapStack.push();
		final int length = atts.getLength();
		for ( int i = 0; i < length; ++i ) {
			mapStack.put( atts.getLocalName( i ), atts.getValue( i ) );
		}

		final String key = distilTimeSeriesKey();

		if ( key != null ) {
			final Observation observation = distilObservation();
			final boolean haveObservations
			= observation != null && ! observation.isEmpty();
			final RegularInterval interval = distilRegularInterval(
				haveObservations ? 1 : 0,
				fakeTimeForEmptySeries && !haveObservations
			);
			if (
				interval != null
				&&
				haveObservations
			) {
				dispatch(
					new TimeSeries(
						key,
						interval,
						new Observation[]{
							observation
						}
					)
				);
			} else {
				dispatch(
					new TimeSeries(
						key,
						interval,
						new Observation[]{}
					)
				);
			}
		}
	}

	public void endElement (
		final String namespaceURI,
		final String localName,
		final String qName
	) {
		super.endElement( namespaceURI, localName, qName );
		mapStack.pop();
		if ( getNestingDepth() == 0 ) {
			finish();
		}
	}

	/* ==================== Package Methods ===================== */

	String distilTimeSeriesKey () {
		final StringBuffer buffer = new StringBuffer();
		for (
			final Iterator iterator = keyFamily.keyStructureIterator();
			iterator.hasNext();
		) {
			final KeyFamilyComponent component
			= (KeyFamilyComponent) iterator.next();
			final String value = mapStack.get( component.getConceptId() );
			if ( value == null ) {
				return null;
			}
			buffer.append( value );
			if ( iterator.hasNext() ) {
				buffer.append( TimeSeries.KEY_SEPARATOR );
			}
		}
		return buffer.toString();
	}

	RegularInterval distilRegularInterval (
		final int size, final boolean fakePeriod
	) {
		final String origPeriod = keyFamily.getTimePeriod() != null
			? mapStack.get(
				keyFamily.getTimePeriod().getConceptId()
			) : null;
		final String period = origPeriod != null
			? origPeriod : ( fakePeriod ? "2001-01-01" : null );
		final String periodFormat = keyFamily.getTimeFormat() != null
			? mapStack.get(
				keyFamily.getTimeFormat().getConceptId()
			) : null;
		if ( period == null || periodFormat == null ) return null;
		try {
			final Date date = format.parse( period );
			final int frequency = XMLDateFormat.parseDuration( periodFormat );
			return new GregorianRegularInterval(
				frequency, date, size
			);
		}
		catch ( final ParseException e ) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"distilRegularInterval()",
				"Cannot parse date \"" + period
				+ "\" and duration \"" + periodFormat + "\" to a base period.",
				e
			);
			return null;
		}
	}

	Observation distilObservation () {
		final String value = keyFamily.getObsValue() != null
			? mapStack.get(
				keyFamily.getObsValue().getConceptId()
			) : null;
		final String status = keyFamily.getObsStatus() != null
			? mapStack.get(
				keyFamily.getObsStatus().getConceptId()
			) : null;
		final String conf = keyFamily.getObsConf() != null
			? mapStack.get(
				keyFamily.getObsConf().getConceptId()
			) : null;
		final String preBreak = keyFamily.getObsPreBreak() != null
			? mapStack.get(
				keyFamily.getObsPreBreak().getConceptId()
			) : null;
		return new Observation(
			value == null ? ""
				: ( "NaN".equals( value ) ? "-" : value ),
			status == null ? "" : status,
			conf == null ? "" : conf,
			preBreak == null ? ""
				: ( "NaN".equals( preBreak ) ? "-" : preBreak )
		);
	}
}
