// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2003, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.xml;

import java.util.Iterator;
import java.util.Date;
import java.util.Calendar;

import org.xml.sax.SAXException;

import junit.framework.Test;
import junit.framework.TestSuite;

import Int.ecb.gesmescb.*;

/* ***************************** Class ****************************** */

/**
 * Implements a few tests for the {@link CubeReader}.
 *
 * @author
 * <a href="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</a>
 *
 * @author
 * <a href="mailto:Sami.Airo@ECB.int">Sami Airo</a>
 *
 * @version @(#)$Id: CubeReaderTest.java,v 1.1 2017/01/19 16:14:02 trainot Exp $
 */
public class CubeReaderTest extends SAX2ReaderTest {

	/* ======================= Constants ======================== */

	private static final Date VERSION_DATE;
	static {
		final Calendar calendar = Calendar.getInstance();
		calendar.clear();
		calendar.set( 2003, Calendar.NOVEMBER, 10, 15, 22 );
		VERSION_DATE = calendar.getTime();
	}

	private static final KeyFamilyComponent
		FREQ = new KeyFamilyComponent(
			"FREQ", KeyFamilyComponent.KEY_STRUCTURE_DIMENSION,
			"1", "AN..1", "", "", ""
		),
		ITEM = new KeyFamilyComponent(
			"ITEM", KeyFamilyComponent.KEY_STRUCTURE_DIMENSION,
			"2", "AN1", "", "", ""
		),
		TIME_PERIOD = new KeyFamilyComponent(
			"TIME_PERIOD", KeyFamilyComponent.TIME_DIMENSION,
			"4", "AN..35", "", "", ""
		),
		TIME_FORMAT = new KeyFamilyComponent(
			"TIME_FORMAT", KeyFamilyComponent.TIME_DIMENSION,
			"5", "AN3", "", "", ""
		),
		OBS_VALUE = new KeyFamilyComponent(
			"OBS_VALUE", KeyFamilyComponent.ARRAY_CELL,
			"6", "AN..15", "", "", ""
		),
		OBS_STATUS = new KeyFamilyComponent(
			"OBS_STATUS", KeyFamilyComponent.ARRAY_CELL,
			"7", "AN1", KeyFamilyComponent.MANDATORY,
			KeyFamilyComponent.ATTACHED_TO_OBSERVATION, ""
		),
		OBS_CONF = new KeyFamilyComponent(
			"OBS_CONF", KeyFamilyComponent.ARRAY_CELL,
			"8", "AN1", KeyFamilyComponent.CONDITIONAL,
			KeyFamilyComponent.ATTACHED_TO_OBSERVATION, ""
		),
		OBS_PRE_BREAK = new KeyFamilyComponent(
			"OBS_PRE_BREAK", KeyFamilyComponent.ARRAY_CELL,
			"9", "AN..15", KeyFamilyComponent.CONDITIONAL,
			KeyFamilyComponent.ATTACHED_TO_OBSERVATION, ""
		),
		OBS_ATT = new KeyFamilyComponent(
			"OBS_ATT", KeyFamilyComponent.ATTRIBUTE,
			"", "AN..20", KeyFamilyComponent.CONDITIONAL,
			KeyFamilyComponent.ATTACHED_TO_OBSERVATION, ""
		),
		TS_ATT = new KeyFamilyComponent(
			"TS_ATT", KeyFamilyComponent.ATTRIBUTE,
			"", "A2", KeyFamilyComponent.MANDATORY,
			KeyFamilyComponent.ATTACHED_TO_TIME_SERIES, ""
		),
		SG_ATT = new KeyFamilyComponent(
			"SG_ATT", KeyFamilyComponent.ATTRIBUTE,
			"", "N..10", KeyFamilyComponent.CONDITIONAL,
			KeyFamilyComponent.ATTACHED_TO_SIBLING_GROUP, ""
		),
		DS_ATT = new KeyFamilyComponent(
			"DS_ATT", KeyFamilyComponent.ATTRIBUTE,
			"", "A..10", KeyFamilyComponent.MANDATORY,
			KeyFamilyComponent.ATTACHED_TO_DATA_SET, ""
		);

	private static final KeyFamily KF = new KeyFamily(
		"KF", "test KF", "", VERSION_DATE,
		new KeyFamilyComponent[]{
			FREQ, ITEM, TIME_PERIOD, TIME_FORMAT,
			OBS_VALUE, OBS_STATUS, OBS_CONF, OBS_PRE_BREAK,
			OBS_ATT, TS_ATT, SG_ATT, DS_ATT
		}
	);

	/* ====================== Constructors ====================== */

	public CubeReaderTest ( final String name ) {
		super( name );
	}

	/* ================== Protected Methods ===================== */

	/** Tears down the test fixture. */
	protected void tearDown () {
		super.tearDown();
		System.getProperties().remove(
			CubeReader.class.getName() + ".fakeTimeForEmptySeries"
		);
	}

	protected SAX2Reader factory (
		final org.xml.sax.XMLReader parser,
		final SAX2Reader parent
	) {
		final CubeReader cubeReader = new CubeReader( parser, parent );
		cubeReader.setKeyFamily( KF );
		return cubeReader;
	}

	protected void openCube ( final String string ) throws SAXException {
		openElement( "Cube", string );
	}

	protected void closeCube () throws SAXException {
		closeElement( "Cube" );
	}

	/* ===================== Public Methods ===================== */

	/**
	 * Assembles and returns a test suite for all
	 * the test methods of this test case.
	 *
	 * @return a non-null test suite
	 */
	public static Test suite () {
		final TestSuite suite = new TestSuite( CubeReaderTest.class );
		return suite;
	}

	/**
	 * @param arg the array of argument {@link String}s
	 */
	public static void main ( final String[] arg ) {
		final String[] testCaseName = {
			CubeReaderTest.class.getName()
		};
		junit.textui.TestRunner.main( testCaseName );
	}

	/* ---------------------- Test Methods ---------------------- */

	public void testEmptyTimeSeries () {
		getReader().startElement(
			"someURI",
			"AnySeries",
			"somePrefix:AnySeries",
			new TestAttributes(
				new String[]{ "FREQ", "ITEM" },
				new String[]{ "M", "602" }
			)
		);
		getReader().endElement(
			"someURI",
			"AnySeries",
			"somePrefix:AnySeries"
		);
		assertEquals( "1 object", 1, listener.size() );
		assertEquals(
			"series",
			new TimeSeries(
				"M.602", "", "", new Observation[]{}
			),
			listener.getLastObject()
		);
	}

	public void testSingleObservation () {
		getReader().startElement(
			"someURI2",
			"AnySeries2",
			"somePrefix:AnySeries2",
			new TestAttributes(
				new String[]{
					"FREQ", "ITEM", "TIME_PERIOD", "TIME_FORMAT",
					"OBS_VALUE", "OBS_STATUS", "OBS_CONF", "OBS_PRE_BREAK"
				},
				new String[]{
					"Q", "ANYITEM", "2002-04", "P3M",
					"123.4", "A", "C", "-118"
				}
			)
		);
		getReader().endElement(
			"someURI2",
			"AnySeries2",
			"somePrefix:AnySeries2"
		);
		assertEquals( "1 object", 1, listener.size() );
		assertEquals(
			"series",
			new TimeSeries(
				"Q.ANYITEM", "20022", "608",
				new Observation[]{
					new Observation( "123.4", "A", "C", "-118" )
				}
			),
			listener.getLastObject()
		);
	}

	public void testNaNObservation () {
		getReader().startElement(
			"someURI2",
			"AnySeries2",
			"somePrefix:AnySeries2",
			new TestAttributes(
				new String[]{
					"FREQ", "ITEM", "TIME_PERIOD", "TIME_FORMAT",
					"OBS_VALUE", "OBS_STATUS", "OBS_CONF", "OBS_PRE_BREAK"
				},
				new String[]{
					"Q", "ANYITEM", "2002-04", "P3M",
					"NaN", "NaN", "NaN", "NaN"
				}
			)
		);
		getReader().endElement(
			"someURI2",
			"AnySeries2",
			"somePrefix:AnySeries2"
		);
		assertEquals( "1 object", 1, listener.size() );
		assertEquals(
			"series",
			new TimeSeries(
				"Q.ANYITEM", "20022", "608",
				new Observation[]{
					new Observation( "-", "NaN", "NaN", "-" )
				}
			),
			listener.getLastObject()
		);
	}

	public void testMultipleObservations () {
		getReader().startElement(
			"someURI3",
			"AnySeries3",
			"somePrefix:AnySeries3",
			new TestAttributes(
				new String[]{ "FREQ", "ITEM", "TIME_FORMAT" },
				new String[]{ "W", "ANYITEM", "P7D" }
			)
		);
		assertEquals(
			"key",
			"W.ANYITEM",
			( (CubeReader) getReader() ).distilTimeSeriesKey()
		);
		getReader().startElement(
			"someURI3",
			"AnyObs3",
			"somePrefix:AnyObs3",
			new TestAttributes(
				new String[]{ "TIME_PERIOD", "OBS_VALUE", "OBS_STATUS", "OBS_CONF", "OBS_PRE_BREAK" },
				new String[]{ "2003-01-06", "234.5", "A", "", "" }
			)
		);
		getReader().endElement(
			"someURI3",
			"AnyObs3",
			"somePrefix:AnyObs3"
		);
		getReader().startElement(
			"someURI3",
			"AnyObs3",
			"somePrefix:AnyObs3",
			new TestAttributes(
				new String[]{ "TIME_PERIOD", "OBS_VALUE", "OBS_STATUS", "OBS_CONF", "OBS_PRE_BREAK" },
				new String[]{ "2003-01-13", "NaN", "M", "", "" }
			)
		);
		getReader().endElement(
			"someURI3",
			"AnyObs3",
			"somePrefix:AnyObs3"
		);
		getReader().endElement(
			"someURI3",
			"AnySeries3",
			"somePrefix:AnySeries3"
		);
		final Iterator iterator = listener.iterator();
		assertTrue( "zerot", iterator.hasNext() );
		assertEquals(
			"series0",
			new TimeSeries(
				"W.ANYITEM", "", "",
				new Observation[]{}
			),
			iterator.next()
		);
		assertTrue( "first", iterator.hasNext() );
		assertEquals(
			"series1",
			new TimeSeries(
				"W.ANYITEM", "200302", "616",
				new Observation[]{
					new Observation( "234.5", "A", "", "" )
				}
			),
			iterator.next()
		);
		assertTrue( "second", iterator.hasNext() );
		assertEquals(
			"series2",
			new TimeSeries(
				"W.ANYITEM", "200303", "616",
				new Observation[]{
					new Observation( "-", "M", "", "" )
				}
			),
			iterator.next()
		);
		assertTrue( "no more", !iterator.hasNext() );
	}

	public void testDataSetInheritance () throws SAXException {
		openElement(
			"DataSet",
			"ITEM=GOODS OBS_STATUS=A OBS_CONF=F OBS_PRE_BREAK=NaN"
		);
		openElement(
			"Series",
			"FREQ=A TIME_FORMAT=P1Y OBS_CONF=C"
		);
		openElement(
			"Obs",
			"TIME_PERIOD=2000 OBS_VALUE=13.6"
		);
		closeElement( "Obs" );
		closeElement( "Series" );
		closeElement( "DataSet" );

		final Iterator iterator = listener.iterator();
		assertTrue( "first", iterator.hasNext() );
		assertEquals(
			"series1",
			new TimeSeries(
				"A.GOODS", "", "",
				new Observation[]{}
			),
			iterator.next()
		);
		assertTrue( "second", iterator.hasNext() );
		assertEquals(
			"series2",
			new TimeSeries(
				"A.GOODS", "2000", "602",
				new Observation[]{
					new Observation( "13.6", "A", "C", "-" )
				}
			),
			iterator.next()
		);
		assertTrue( "no more", !iterator.hasNext() );
	}

	public void testNoFakePeriod () throws SAXException {
		openElement(
			"Series",
			"FREQ=A ITEM=FOODS TIME_FORMAT=P1Y"
		);
		closeElement( "Series" );

		final Iterator iterator = listener.iterator();
		assertTrue( "first", iterator.hasNext() );
		assertEquals(
			"series1",
			new TimeSeries(
				"A.FOODS", "", "", new Observation[]{}
			),
			iterator.next()
		);
		assertTrue( "no more", !iterator.hasNext() );
	}

	public void testFakePeriod () throws SAXException {
		System.setProperty(
			CubeReader.class.getName() + ".fakeTimeForEmptySeries",
			"true"
		);
		setUp();	// create another test arrangement
		openElement(
			"Series",
			"FREQ=A ITEM=FOODS TIME_FORMAT=P1Y"
		);
		closeElement( "Series" );

		final Iterator iterator = listener.iterator();
		assertTrue( "first", iterator.hasNext() );
		assertEquals(
			"series1",
			new TimeSeries(
				"A.FOODS", "20012000", "702", new Observation[]{}
			),
			iterator.next()
		);
		assertTrue( "no more", !iterator.hasNext() );
	}

	public void testSuppressFakePeriodIfObservations () throws SAXException {
		System.setProperty(
			CubeReader.class.getName() + ".fakeTimeForEmptySeries",
			"true"
		);
		setUp();	// create another test arrangement
		openElement(
			"Obs",
			"FREQ=A ITEM=FOODS TIME_FORMAT=P1Y OBS_VALUE=4"
		);
		closeElement( "Obs" );

		final Iterator iterator = listener.iterator();
		assertTrue( "first", iterator.hasNext() );
		assertEquals(
			"series1",
			new TimeSeries(
				"A.FOODS", "", "", new Observation[]{}
			),
			iterator.next()
		);
		assertTrue( "no more", !iterator.hasNext() );
	}
}
