// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2003, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.xml;

import java.util.Iterator;
import java.util.Date;

import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.text.ParseException;

// used but not importable: import org.xml.sax.XMLReader;
import org.xml.sax.Attributes;

import Int.ecb.logging.Level;

import Int.ecb.gesmescb.InterchangeBegin;
import Int.ecb.gesmescb.InterchangeEnd;
import Int.ecb.gesmescb.MessageBegin;
import Int.ecb.gesmescb.MessageEnd;
import Int.ecb.gesmescb.Contact;
import Int.ecb.gesmescb.KeyFamily;

import Int.ecb.gesmescb.time.RegularInterval;
import Int.ecb.gesmescb.time.GregorianRegularInterval;

import Int.ecb.gesmescb.metabase.Metabase;
import Int.ecb.gesmescb.metabase.MaintenanceAgency;

import Int.ecb.gesmescb.event.GESMESCBEventListener;
import Int.ecb.gesmescb.event.ErrorEvent;

import Int.ecb.gesmescb.edifact.RegularIntervalFormat;

/* ***************************** Class ****************************** */

/**
 * <p>Processes an <code>Envelope</code> with contents
 * from SAX2 parser output.</p>
 *
 * <p>In doing so it delegates to other <code>SAX2Reader<code>s.</p>
 *
 * @author
 * <a href="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</a>
 *
 * @author
 * <a href="mailto:Sami.Airo@ECB.int">Sami Airo</a>
 *
 * @version @(#)$Id: EnvelopeReader.java,v 1.1 2017/01/19 16:14:03 trainot Exp $
 */
final class EnvelopeReader extends SAX2Reader {

	/* ======================= Constants ======================== */

	/** Internal principal state codes. */
	private static final int
		START = 0,
		ENVELOPE = 1,
		TEST = 2,
		NAME = 3,
		PREPARED = 4,
		SENDER = 5,
		RECEIVER = 6,
		DATASET_ID = 8,
		ACTION = 9,
		REPORTING_BEGIN = 10,
		REPORTING_END = 11,
		EXTENSION = 32,
		MAINTENANCE_AGENCY_ID = 33,
		KEY_FAMILY_ID = 34;
		//FINISH = 64;

	/** Internal codes for dispatched envelopes level. */
	private static final int
		NONE = 0,
		INTERCHANGE = 1,
		MESSAGE = 2;

	/* ========================= Fields ========================= */

	/** Date format for preparation time stamp. */
	final DateFormat preparationDateFormat;

	/** Date format for extraction time stamp. */
	final DateFormat extractionDateFormat;

	/** Current XML dateTime format.
	 * Changes its pattern according to frequency. */
	private final XMLDateFormat xmlDateFormat;

	/** EDIFACT interval format for reporting period. */
	private final RegularIntervalFormat edifactIntervalFormat;

	/** Delegate syntactic automaton. */
	private final CubeReader cubeReader;

	/* -------------------- Automaton state --------------------- */

	/** The principal parsing state of the automaton. */
	private int status;

	/** What has been dispatched so far in terms of envelopes. */
	private int dispatched;

	/** The currently processed Envelope. */
	private Envelope envelope;

	/** The principal parsing state of the automaton. */
	private final StringBuffer accumulator;

	/** The nesting depth inside an extension element. */
	private int extensionDepth;

	/* ====================== Constructors ====================== */

	public EnvelopeReader (
		final org.xml.sax.XMLReader parser,
		final SAX2Reader parent
	) {
		super( parser, parent );
		preparationDateFormat = new SimpleDateFormat( "yyMMdd:HHmm" );
		extractionDateFormat = new SimpleDateFormat( "yyyyMMddHHmm" );
		xmlDateFormat = new XMLDateFormat();
		edifactIntervalFormat = new RegularIntervalFormat();
		envelope = null;
		accumulator = new StringBuffer();
		dispatched = NONE;
		extensionDepth = 0;
		cubeReader = new CubeReader( parser, this );
		init();
	}

	/* ===================== Public Methods ===================== */

	public void addListener (
		final GESMESCBEventListener listener,
		final boolean isImportant
	) {
		super.addListener( listener, isImportant );
		cubeReader.addListener( listener, isImportant );
	}

	public void removeAllListeners () {
		super.removeAllListeners();
		cubeReader.removeAllListeners();
	}

	public void reset () {
		super.reset();
		init();
	}

	/* --------------------- ContentHandler --------------------- */

	public void startElement (
		final String namespaceURI,
		final String localName,
		final String qName,
		Attributes atts
	) {
		super.startElement( namespaceURI, localName, qName, atts );

		switch ( status ) {
			case START:
				if (
					XMLNS_GESMES.equals( namespaceURI )
					&&
					"Envelope".equals( localName )
				) {
					status = ENVELOPE;
					envelope = new Envelope();
					dispatched = NONE;
					envelope.ref = atts.getValue( "id" );
					if ( envelope.ref == null ) {
						envelope.ref = "";
					}
				}
				break;
			case ENVELOPE:
				if ( XMLNS_GESMES.equals( namespaceURI ) ) {
					if ( "test".equals( localName ) ) {
						envelope.test = true;
						status = TEST;
						break;
					}
					if ( "name".equals( localName ) ) {
						status = NAME;
						accumulator.setLength( 0 );
						break;
					}
					if ( "prepared".equals( localName ) ) {
						status = PREPARED;
						accumulator.setLength( 0 );
						break;
					}
					if ( "Sender".equals( localName ) ) {
						status = SENDER;
						envelope.senderId = atts.getValue( "id" );
						break;
					}
					if ( "Receiver".equals( localName ) ) {
						status = RECEIVER;
						envelope.receiverId = atts.getValue( "id" );
						break;
					}
					if ( "dataSetId".equals( localName ) ) {
						status = DATASET_ID;
						accumulator.setLength( 0 );
						break;
					}
					if ( "action".equals( localName ) ) {
						status = ACTION;
						accumulator.setLength( 0 );
						break;
					}
					if ( "reportingBegin".equals( localName ) ) {
						status = REPORTING_BEGIN;
						accumulator.setLength( 0 );
						break;
					}
					if ( "reportingEnd".equals( localName ) ) {
						status = REPORTING_END;
						accumulator.setLength( 0 );
						break;
					}
					if ( "Extension".equals( localName ) ) {
						status = EXTENSION;
						extensionDepth = getNestingDepth();
						break;
					}
					if ( "Envelope".equals( localName ) ) {
						dispatchEnvelopes();
						envelope = new Envelope( envelope );
						envelope.ref = atts.getValue( "id" );
						if ( envelope.ref == null ) {
							envelope.ref = "";
						}
						break;
					}
				} else {
					final KeyFamily keyFamily;
					if (
						! "".equals( envelope.getMaintenanceAgencyId() )
						&&
						! "".equals( envelope.getKeyFamilyId() )
					) {
						final String
							maintenanceAgencyId
							= envelope.getMaintenanceAgencyId(),
							keyFamilyId = envelope.getKeyFamilyId();
						final Metabase metabase = Metabase.getMetabase();
						final MaintenanceAgency maintenanceAgency
						= metabase.getMaintenanceAgency( maintenanceAgencyId );
						if ( maintenanceAgency != null ) {
							keyFamily
							= maintenanceAgency.getKeyFamily( keyFamilyId );
							if ( keyFamily == null ) {
								final String desc
								= "Maintenance agency \"" + maintenanceAgencyId
									+ "\" does not know key family \""
									+ keyFamilyId
									+ "\".";
								componentHelper.logger.log(
									Level.SEVERE,
									componentHelper.logName,
									"startElement()",
									desc
								);
								dispatch(
									new ErrorEvent(
										desc, ErrorEvent.ERROR
									)
								);
							}
						} else {
							final String desc
							= "Maintenance agency \"" + maintenanceAgencyId
								+ "\" is not known.";
							componentHelper.logger.log(
								Level.SEVERE,
								componentHelper.logName,
								"startElement()",
								desc
							);
							dispatch(
								new ErrorEvent(
									desc, ErrorEvent.ERROR
								)
							);
							keyFamily = null;
						}
					} else {
						final KeyFamilyURIMapping mapping
						= KeyFamilyURIMapping.getInstance();
						final Iterator iterator = mapping.lookup(
							namespaceURI
						);
						if ( iterator.hasNext() ) {
							final KeyFamilyURIMapping.LookupResult result
							= (KeyFamilyURIMapping.LookupResult) iterator.next();
							keyFamily = result.getKeyFamily();
							envelope.keyFamilyId = keyFamily.getId();
							envelope.maintenanceAgencyId
							= result.getMaintenanceAgencyId();
						} else {
							final String desc
							= "Namespace URI \"" + namespaceURI
								+ "\" does not correspond"
								+ " to a known key family.";
							componentHelper.logger.log(
								Level.SEVERE,
								componentHelper.logName,
								"startElement()",
								desc
							);
							dispatch(
								new ErrorEvent(
									desc, ErrorEvent.ERROR
								)
							);
							keyFamily = null;
						}
						if ( iterator.hasNext() ) {
							final String desc
							= "Namespace URI \"" + namespaceURI
								+ "\" corresponds"
								+ " to more than one key family.";
							componentHelper.logger.log(
								Level.SEVERE,
								componentHelper.logName,
								"startElement()",
								desc
							);
							dispatch(
								new ErrorEvent(
									desc, ErrorEvent.ERROR
								)
							);
						}
					}
					dispatchEnvelopes();
					if ( keyFamily != null ) {
						componentHelper.logger.log(
							Level.FINE,
							componentHelper.logName,
							"startElement()",
							"Data is for key family \""
							+ keyFamily.getId()
							+ "\"."
						);
						cubeReader.setKeyFamily( keyFamily );
						delegate(
							cubeReader,
							namespaceURI, localName, qName, atts
						);
					}
				}
				break;
			case EXTENSION:
				if ( XMLNS_EDIFACT.equals( namespaceURI ) ) {
					if ( "maintenanceAgencyId".equals( localName ) ) {
						status = MAINTENANCE_AGENCY_ID;
						accumulator.setLength( 0 );
						break;
					}
					if ( "keyFamilyId".equals( localName ) ) {
						status = KEY_FAMILY_ID;
						accumulator.setLength( 0 );
						break;
					}
				}
				break;
			default:
				complain(
					"Unexpected element: \"" + qName + "\".",
					"startElement()",
					qName,
					ErrorEvent.ERROR
				);
		}
	}

	public void characters (
		final char[] ch,
		final int start,
		final int length
	) {
		super.characters( ch, start, length );

		switch ( status ) {
			case NAME:
			case PREPARED:
			case DATASET_ID:
			case ACTION:
			case REPORTING_BEGIN:
			case REPORTING_END:
			case MAINTENANCE_AGENCY_ID:
			case KEY_FAMILY_ID:
				accumulator.append( ch, start, length );
				break;
			default:
				checkWhitespace( ch, start, length );
		}
	}

	public void endElement (
		final String namespaceURI,
		final String localName,
		final String qName
	) {
		super.endElement( namespaceURI, localName, qName );

		switch ( status ) {
			case ENVELOPE:
				dispatchEnvelopes();
				if ( dispatched == MESSAGE ) {
					dispatch( new MessageEnd(
						envelope.parent != null
							? envelope.getRef() : "MREF000001"
					) );
					if ( envelope.parent != null ) {
						envelope = envelope.parent;
					}
					dispatched = INTERCHANGE;
				}
				if ( dispatched == INTERCHANGE ) {
					dispatch( new InterchangeEnd( envelope.getRef() ) );
					envelope = null;
					dispatched = NONE;
				}
				break;
			case TEST:
				checkNamespace( XMLNS_GESMES, namespaceURI );
				checkLocalName( "test", localName );
				status = ENVELOPE;
				break;
			case NAME:
				checkNamespace( XMLNS_GESMES, namespaceURI );
				checkLocalName( "name", localName );
				envelope.name = accumulator.toString();
				status = ENVELOPE;
				break;
			case PREPARED:
				checkNamespace( XMLNS_GESMES, namespaceURI );
				checkLocalName( "prepared", localName );
				try {
					xmlDateFormat.applyPattern(
						XMLDateFormat.SECONDLY_PATTERN
					);
					envelope.prepared
					= xmlDateFormat.parse( accumulator.toString() );
				} catch ( final ParseException e ) {
					complain(
						"Preparation time stamp \""
						+ accumulator.toString() + "\" badly formatted.",
						"endElement()",
						qName,
						ErrorEvent.ERROR
					);
				}
				status = ENVELOPE;
				break;
			case SENDER:
			case RECEIVER:
				status = ENVELOPE;
				break;
			case DATASET_ID:
				checkNamespace( XMLNS_GESMES, namespaceURI );
				checkLocalName( "dataSetId", localName );
				envelope.dataSetId = accumulator.toString();
				status = ENVELOPE;
				break;
			case ACTION:
				checkNamespace( XMLNS_GESMES, namespaceURI );
				checkLocalName( "action", localName );
				envelope.action = accumulator.toString();
				if (
					"update".equals( envelope.action )
				) {
					envelope.action = MessageBegin.UPDATE_OR_REPLACE_ACTION;
				} else if ( "delete".equals( envelope.action ) ) {
					envelope.action = MessageBegin.DELETE_ACTION;
				} else {
					complain(
						"Unknown message action \"" + envelope.action + "\".",
						"endElement()",
						qName,
						ErrorEvent.ERROR
					);
				}
				status = ENVELOPE;
				break;
			case REPORTING_BEGIN:
				checkNamespace( XMLNS_GESMES, namespaceURI );
				checkLocalName( "reportingBegin", localName );
				try {
					envelope.reportingBegin
					= xmlDateFormat.parse( accumulator.toString() );
				} catch ( final ParseException e ) {
					complain(
						"Reporting period begin time point \""
						+ accumulator.toString() + "\" badly formatted.",
						"endElement()",
						qName,
						ErrorEvent.ERROR
					);
				}
				status = ENVELOPE;
				break;
			case REPORTING_END:
				checkNamespace( XMLNS_GESMES, namespaceURI );
				checkLocalName( "reportingEnd", localName );
				try {
					envelope.reportingEnd
					= xmlDateFormat.parse( accumulator.toString() );
				} catch ( final ParseException e ) {
					complain(
						"Reporting period end time point \""
						+ accumulator.toString() + "\" badly formatted.",
						"endElement()",
						qName,
						ErrorEvent.ERROR
					);
				}
				status = ENVELOPE;
				break;
			case MAINTENANCE_AGENCY_ID:
				checkNamespace( XMLNS_EDIFACT, namespaceURI );
				checkLocalName( "maintenanceAgencyId", localName );
				envelope.maintenanceAgencyId = accumulator.toString();
				status = EXTENSION;
				break;
			case KEY_FAMILY_ID:
				checkNamespace( XMLNS_EDIFACT, namespaceURI );
				checkLocalName( "keyFamilyId", localName );
				envelope.keyFamilyId = accumulator.toString();
				status = EXTENSION;
				break;
			case EXTENSION:
				if (
					getNestingDepth() < extensionDepth
					&&
					XMLNS_GESMES.equals( namespaceURI )
					&&
					"Extension".equals( localName )
				) {
					status = ENVELOPE;
				}
				break;
			default: {
				complain(
					"Unexpected element end: \"" + qName + "\".",
					"endElement()",
					qName,
					ErrorEvent.ERROR
				);
			}
		}
	}

	/* ==================== Private Methods ===================== */

	/** (Re-)initialise the internal state. */
	private void init () {
		status = START;
	}

	private boolean interchangeWaiting () {
		return envelope != null;
	}

	private boolean messageWaiting () {
		return envelope != null && (
			envelope.parent != null
			||
			! "".equals( envelope.getName() )
			||
			! "".equals( envelope.getMaintenanceAgencyId() )
			||
			! "".equals( envelope.getKeyFamilyId() )
			||
			! "".equals( envelope.getDataSetId() )
			||
			! "".equals( envelope.getAction() )
		);
	}

	private void dispatchEnvelopes () {
		if ( dispatched == NONE && interchangeWaiting() ) {
			dispatch(
				new InterchangeBegin(
					envelope.getRef(),
					envelope.getSenderId(),
					envelope.getReceiverId(),
					envelope.getPrepared() != null
						? preparationDateFormat.format(
							envelope.getPrepared()
						) : "",
					envelope.isTest()
				)
			);
			dispatched = INTERCHANGE;
		}
		if ( dispatched == INTERCHANGE && messageWaiting() ) {
			final Date reportingBegin = envelope.getReportingBegin();
			final Date reportingEnd = envelope.getReportingEnd();
			final RegularInterval interval;
			if ( reportingBegin != null && reportingEnd != null ) {
				interval = GregorianRegularInterval.searchCover(
					reportingBegin, reportingEnd, true
				);
			} else {
				interval = null;
			}
			final String reportingPeriod, reportingPeriodFormat;
			if ( interval != null ) {
				final String reportingPaF = edifactIntervalFormat.format(
					interval
				);
				final int separator = reportingPaF.indexOf( ':' );
				reportingPeriod = reportingPaF.substring( 0, separator );
				reportingPeriodFormat = reportingPaF.substring( separator + 1 );
			} else {
				reportingPeriod = reportingPeriodFormat = "";
			}
			dispatch(
				new MessageBegin(
					envelope.parent != null ? envelope.getRef() : "MREF000001",
					envelope.getSenderId(),
					envelope.getReceiverId(),
					envelope.getMaintenanceAgencyId(),
					MessageBegin.STATISTICAL_DATA,
					envelope.getName(),
					new Contact[]{},
					envelope.getDataSetId(),
					"".equals( envelope.getAction() )
						? MessageBegin.UPDATE_OR_REPLACE_ACTION
						: envelope.getAction(),
					envelope.getPrepared() != null
						? extractionDateFormat.format(
							envelope.getPrepared()
						) : "",
					reportingPeriod,
					reportingPeriodFormat,
					envelope.getKeyFamilyId()
				)
			);
			dispatched = MESSAGE;
		}
	}
}

/* ***************************** Class ****************************** */

class Envelope {

	/* ========================= Fields ========================= */

	/** The nesting parent. */
	public final Envelope parent;

	/** IREF or MREF number. */
	public String ref;

	/** Test indicator. */
	public boolean test;

	/** Message subject name. */
	public String name;

	/** Interchange preparation timestamp. */
	public Date prepared;

	/** The sender id. */
	public String senderId;

	/** The receiver id. */
	public String receiverId;

	/** Envelope dataset id. */
	public String dataSetId;

	/** Envelope action. */
	public String action;

	/** Envelope reporting period begin point. */
	public Date reportingBegin;

	/** Envelope reporting period end point. */
	public Date reportingEnd;

	/** maintenance agency id for the message. */
	public String maintenanceAgencyId;

	/** key family id for the message. */
	public String keyFamilyId;

	/* ====================== Constructors ====================== */

	public Envelope () {
		this( null );
	}

	public Envelope ( final Envelope parent ) {
		this.parent = parent;
	}

	/* ===================== Public Methods ===================== */

	/** IREF or MREF number. */
	public String getRef () {
		if ( ref != null ) return ref;
		if ( parent != null ) return parent.getRef();
		return "";
	}

	/** Test indicator. */
	public boolean isTest () {
		if ( test ) return true;
		if ( parent != null ) return parent.isTest();
		return false;
	}

	/** Message subject name. */
	public String getName () {
		if ( name != null ) return name;
		if ( parent != null ) return parent.getName();
		return "";
	}

	/** Interchange preparation timestamp. */
	public Date getPrepared () {
		if ( prepared != null ) return prepared;
		if ( parent != null ) return parent.getPrepared();
		return null;
	}

	/** Dender id. */
	public String getSenderId () {
		if ( senderId != null ) return senderId;
		if ( parent != null ) return parent.getSenderId();
		return "";
	}

	/** Receiver id. */
	public String getReceiverId () {
		if ( receiverId != null ) return receiverId;
		if ( parent != null ) return parent.getReceiverId();
		return "";
	}

	/** Target data set id. */
	public String getDataSetId () {
		if ( dataSetId != null ) return dataSetId;
		if ( parent != null ) return parent.getDataSetId();
		return "";
	}

	/** Transaction command. */
	public String getAction () {
		if ( action != null ) return action;
		if ( parent != null ) return parent.getAction();
		return "";
	}

	/** Reporting period begin point. */
	public Date getReportingBegin () {
		if ( reportingBegin != null ) return reportingBegin;
		if ( parent != null ) return parent.getReportingBegin();
		return null;
	}

	/** Reporting period end point. */
	public Date getReportingEnd () {
		if ( reportingEnd != null ) return reportingEnd;
		if ( parent != null ) return parent.getReportingEnd();
		return null;
	}

	/** Structural definitions maintenance agency id. */
	public String getMaintenanceAgencyId () {
		if ( maintenanceAgencyId != null ) return maintenanceAgencyId;
		if ( parent != null ) return parent.getMaintenanceAgencyId();
		return "";
	}

	/** key family id for the message. */
	public String getKeyFamilyId () {
		if ( keyFamilyId != null ) return keyFamilyId;
		if ( parent != null ) return parent.getKeyFamilyId();
		return "";
	}
}
