// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2003, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.xml;

import java.util.Iterator;

import org.xml.sax.SAXException;

import junit.framework.Test;
import junit.framework.TestSuite;

import Int.ecb.gesmescb.*;
import Int.ecb.gesmescb.metabase.*;

/* ***************************** Class ****************************** */

/**
 * Implements a few tests for the {@link EnvelopeReader}.
 *
 * @author
 * <a href="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</a>
 *
 * @author
 * <a href="mailto:Sami.Airo@ECB.int">Sami Airo</a>
 *
 * @version @(#)$Id: EnvelopeReaderTest.java,v 1.1 2017/01/19 16:14:02 trainot Exp $
 */
public class EnvelopeReaderTest extends SAX2ReaderTest {

	/* ====================== Constructors ====================== */

	public EnvelopeReaderTest ( final String name ) {
		super( name );
	}

	/* ================== Protected Methods ===================== */

	/** Sets up the test fixture. */
	protected void setUp () {
		super.setUp();
		setNamespace( XMLNS_GESMES, null );
		final Metabase metabase = Metabase.getMetabase();
		metabase.handle( new MessageBegin(
			"DUMMY", "", "", "ECB",
			MessageBegin.STATISTICAL_DEFINITIONS,
			"", new Contact[]{}
		) );
		metabase.handle( KF );
		System.setProperty(
			KeyFamilyURIMapping.class.getName() + ".uri.ECB.KF",
			"http://www.ecb.int/vocabulary/2003-11-10/kf"
		);
	}

	/** Tears down the test fixture. */
	protected void tearDown () {
		System.getProperties().remove(
			KeyFamilyURIMapping.class.getName() + ".uri.ECB.KF"
		);
	}

	protected SAX2Reader factory (
		final org.xml.sax.XMLReader parser,
		final SAX2Reader parent
	) {
		final EnvelopeReader envelopeReader
		= new EnvelopeReader( parser, parent );
		return envelopeReader;
	}

	/* ===================== Public Methods ===================== */

	/**
	 * Assembles and returns a test suite for all
	 * the test methods of this test case.
	 *
	 * @return a non-null test suite
	 */
	public static Test suite () {
		final TestSuite suite = new TestSuite( EnvelopeReaderTest.class );
		return suite;
	}

	/**
	 * @param arg the array of argument {@link String}s
	 */
	public static void main ( final String[] arg ) {
		final String[] testCaseName = {
			EnvelopeReaderTest.class.getName()
		};
		junit.textui.TestRunner.main( testCaseName );
	}

	/* ---------------------- Test Methods ---------------------- */

	public void testEmptyInterchange () throws SAXException {
		openElement( "Envelope", "" );
		closeElement( "Envelope" );

		final Iterator iterator = listener.iterator();
		assertTrue( "first", iterator.hasNext() );
		assertEquals(
			"interchange begin",
			new InterchangeBegin(
				"", "", "", "", false
			),
			iterator.next()
		);
		assertTrue( "second", iterator.hasNext() );
		assertEquals(
			"interchange end",
			new InterchangeEnd( "" ),
			iterator.next()
		);
		assertTrue( "no more", !iterator.hasNext() );
	}

	public void testOnlyInterchange () throws SAXException {
		openElement( "Envelope", "id=IREF000483" );
		openElement( "test", "" );
		closeElement( "test" );
		closeElement( "Envelope" );

		final Iterator iterator = listener.iterator();
		assertTrue( "first", iterator.hasNext() );
		assertEquals(
			"interchange begin",
			new InterchangeBegin(
				"IREF000483", "", "", "", true
			),
			iterator.next()
		);
		assertTrue( "second", iterator.hasNext() );
		assertEquals(
			"interchange end",
			new InterchangeEnd( "IREF000483" ),
			iterator.next()
		);
		assertTrue( "no more", !iterator.hasNext() );
	}

	public void testNestedMessage () throws SAXException {
		openElement( "Envelope", "id=IREF000007" );
		writeElement( "prepared", "2003-02-01T01:00:00" );
		openElement( "Envelope", "id=MREF000001" );
		writeElement( "prepared", "2003-02-01T00:00:00" );
		closeElement( "Envelope" );
		closeElement( "Envelope" );

		final Iterator iterator = listener.iterator();
		assertTrue( "first", iterator.hasNext() );
		assertEquals(
			"interchange begin",
			new InterchangeBegin(
				"IREF000007", "", "", "030201:0100", false
			),
			iterator.next()
		);
		assertTrue( "second", iterator.hasNext() );
		assertEquals(
			"message begin",
			new MessageBegin(
				"MREF000001", "", "", "",
				MessageBegin.STATISTICAL_DATA,
				"", new Contact[]{},
				"", MessageBegin.UPDATE_OR_REPLACE_ACTION,
				"200302010000", "", "", ""
			),
			iterator.next()
		);
		assertTrue( "third", iterator.hasNext() );
		assertEquals(
			"message end",
			new MessageEnd( "MREF000001" ),
			iterator.next()
		);
		assertTrue( "fourth", iterator.hasNext() );
		assertEquals(
			"interchange end",
			new InterchangeEnd( "IREF000007" ),
			iterator.next()
		);
		assertTrue( "no more", !iterator.hasNext() );
	}

	public void testInterchangeMessage () throws SAXException {
		openElement( "Envelope", "id=EREF-3" );
		openElement( "name", "" );
		characters( "Sax sox" );
		closeElement( "name" );
		closeElement( "Envelope" );

		final Iterator iterator = listener.iterator();
		assertTrue( "first", iterator.hasNext() );
		assertEquals(
			"interchange begin",
			new InterchangeBegin(
				"EREF-3", "", "", "", false
			),
			iterator.next()
		);
		assertTrue( "second", iterator.hasNext() );
		assertEquals(
			"message begin",
			new MessageBegin(
				"MREF000001", "", "", "",
				MessageBegin.STATISTICAL_DATA,
				"Sax sox", new Contact[]{},
				"", MessageBegin.UPDATE_OR_REPLACE_ACTION,
				"", "", "", ""
			),
			iterator.next()
		);
		assertTrue( "third", iterator.hasNext() );
		assertEquals(
			"message end",
			new MessageEnd( "MREF000001" ),
			iterator.next()
		);
		assertTrue( "fourth", iterator.hasNext() );
		assertEquals(
			"interchange end",
			new InterchangeEnd( "EREF-3" ),
			iterator.next()
		);
		assertTrue( "no more", !iterator.hasNext() );
	}

	public void testMessageProperties () throws SAXException {
		openElement( "Envelope", "id=IREF000001" );
		writeElement( "name", "Some test stuff" );
		writeElement( "prepared", "2003-06-13T23:56:00" );
		openElement( "Sender", "id=DE2" );
		closeElement( "Sender" );
		openElement( "Receiver", "id=EE2" );
		closeElement( "Receiver" );
		writeElement( "dataSetId", "MPD_DATA" );
		writeElement( "action", "delete" );
		writeElement( "reportingBegin", "2003-01-01T00:00:00" );
		writeElement( "reportingEnd", "2003-07-01T00:00:00" );
		openElement( "Extension", "" );
		setNamespace( XMLNS_EDIFACT, null );
		writeElement( "maintenanceAgencyId", "ECB" );
		writeElement( "keyFamilyId", "KF" );
		setNamespace( XMLNS_GESMES, null );
		closeElement( "Extension" );
		closeElement( "Envelope" );

		final Iterator iterator = listener.iterator();
		assertTrue( "first", iterator.hasNext() );
		assertEquals(
			"interchange begin",
			new InterchangeBegin(
				"IREF000001", "DE2", "EE2", "030613:2356", false
			),
			iterator.next()
		);
		assertTrue( "second", iterator.hasNext() );
		assertEquals(
			"message begin",
			new MessageBegin(
				"MREF000001", "DE2", "EE2", "ECB",
				MessageBegin.STATISTICAL_DATA,
				"Some test stuff", new Contact[]{},
				"MPD_DATA", MessageBegin.DELETE_ACTION,
				"200306132356", "20031", "604", "KF"
			),
			iterator.next()
		);
		assertTrue( "third", iterator.hasNext() );
		assertEquals(
			"message end",
			new MessageEnd( "MREF000001" ),
			iterator.next()
		);
		assertTrue( "fourth", iterator.hasNext() );
		assertEquals(
			"interchange end",
			new InterchangeEnd( "IREF000001" ),
			iterator.next()
		);
		assertTrue( "no more", !iterator.hasNext() );
	}

	public void testDataMessage () throws SAXException {
		openElement( "Envelope", "id=IREF000001" );
		openElement( "dataSetId", "" );
		characters( "MPD_DATA" );
		closeElement( "dataSetId" );
		openElement( "Extension", "" );
		setNamespace( XMLNS_EDIFACT, null );
		openElement( "maintenanceAgencyId", "" );
		characters( "ECB" );
		closeElement( "maintenanceAgencyId" );
		openElement( "keyFamilyId", "" );
		characters( "KF" );
		closeElement( "keyFamilyId" );
		setNamespace( XMLNS_GESMES, null );
		closeElement( "Extension" );
		setNamespace(
			"http://www.ecb.int/vocabulary/2003-11-10/irrelevant", null
		);
		openElement(
			"Cube",
			"FREQ=Q ITEM=80 TIME_PERIOD=2003-07"
			+ " TIME_FORMAT=P3M OBS_VALUE=4 OBS_STATUS=A"
		);
		closeElement( "Cube" );
		setNamespace( XMLNS_GESMES, null );
		closeElement( "Envelope" );

		final Iterator iterator = listener.iterator();
		assertTrue( "first", iterator.hasNext() );
		assertEquals(
			"interchange begin",
			new InterchangeBegin(
				"IREF000001", "", "", "", false
			),
			iterator.next()
		);
		assertTrue( "second", iterator.hasNext() );
		assertEquals(
			"message begin",
			new MessageBegin(
				"MREF000001", "", "", "ECB",
				MessageBegin.STATISTICAL_DATA,
				"", new Contact[]{},
				"MPD_DATA", MessageBegin.UPDATE_OR_REPLACE_ACTION,
				"", "", "", "KF"
			),
			iterator.next()
		);
		assertTrue( "third", iterator.hasNext() );
		assertEquals(
			"time series",
			new TimeSeries(
				"Q.80", "20033", "608", new Observation[]{
					new Observation(
						"4", "A", "", ""
					)
				}
			),
			iterator.next()
		);
		assertTrue( "fourth", iterator.hasNext() );
		assertEquals(
			"message end",
			new MessageEnd( "MREF000001" ),
			iterator.next()
		);
		assertTrue( "fifth", iterator.hasNext() );
		assertEquals(
			"interchange end",
			new InterchangeEnd( "IREF000001" ),
			iterator.next()
		);
		assertTrue( "no more", !iterator.hasNext() );
	}

	public void testNoEdifactHintsDataMessage () throws SAXException {
		openElement( "Envelope", "id=IREF000001" );
		openElement( "dataSetId", "" );
		characters( "MPD_DATA" );
		closeElement( "dataSetId" );
		setNamespace(
			"http://www.ecb.int/vocabulary/2003-11-10/kf", null
		);
		openElement(
			"Cube",
			"FREQ=Q ITEM=80 TIME_PERIOD=2003-07"
			+ " TIME_FORMAT=P3M OBS_VALUE=4 OBS_STATUS=A"
		);
		closeElement( "Cube" );
		setNamespace( XMLNS_GESMES, null );
		closeElement( "Envelope" );

		final Iterator iterator = listener.iterator();
		assertTrue( "first", iterator.hasNext() );
		assertEquals(
			"interchange begin",
			new InterchangeBegin(
				"IREF000001", "", "", "", false
			),
			iterator.next()
		);
		assertTrue( "second", iterator.hasNext() );
		assertEquals(
			"message begin",
			new MessageBegin(
				"MREF000001", "", "", "ECB",
				MessageBegin.STATISTICAL_DATA,
				"", new Contact[]{},
				"MPD_DATA", MessageBegin.UPDATE_OR_REPLACE_ACTION,
				"", "", "", "KF"
			),
			iterator.next()
		);
		assertTrue( "third", iterator.hasNext() );
		assertEquals(
			"time series",
			new TimeSeries(
				"Q.80", "20033", "608", new Observation[]{
					new Observation(
						"4", "A", "", ""
					)
				}
			),
			iterator.next()
		);
		assertTrue( "fourth", iterator.hasNext() );
		assertEquals(
			"message end",
			new MessageEnd( "MREF000001" ),
			iterator.next()
		);
		assertTrue( "fifth", iterator.hasNext() );
		assertEquals(
			"interchange end",
			new InterchangeEnd( "IREF000001" ),
			iterator.next()
		);
		assertTrue( "no more", !iterator.hasNext() );
	}
}
