// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2002-2006, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.xml;

import java.io.IOException;

import java.util.Iterator;
import java.util.Comparator;
import java.util.Date;

import Int.ecb.logging.Level;

import Int.ecb.gesmescb.*;
import Int.ecb.gesmescb.time.Interval;
import Int.ecb.gesmescb.time.RegularInterval;
import Int.ecb.gesmescb.metabase.*;

import Int.ecb.gesmescb.event.StructureListener;
import Int.ecb.gesmescb.event.DataListener;

import Int.ecb.gesmescb.util.KeyTokenizer;

/* ***************************** Class ****************************** */

/**
 * <p>Translates {@link Int.ecb.gesmescb.GESMESCBObject}s
 * into Gesmes/TS-XML text.
 * This text then is written to an output {@link java.io.Writer}
 * or to a file according to preference settings.</p>
 *
 * <center><table border="3" cellpadding="3">
 * <caption>
 * <a name="prefs">Available {@link Int.ecb.prefs preference} settings</a>
 * </caption>
 * <tr>
 * <th>preference name</th>
 * <th>type</th>
 * <th>meaning</th>
 * <th>default</th>
 * </tr>
 * <tr><td><code>destination</td><td>string</td><td>output file path</td></tr>
 * <tr><td><code>encoding</td><td>string</td><td>character encoding</td>
 * <td>platform default, which may not be understood by XML parsers;
 * better use UTF-8 or UTF-16</td></tr>
 * <tr><td><code>indent</td><td>string or integer</td>
 * <td>if integer, then added indentation for enclosed elements;
 * if "<tt>tab</tt>", then use 1 tab character
 * to indent</td><td><tt>tab</tt></td></tr>
 * <tr><td><code>gesmes.schemaLocation</code></td>
 * <td>string</td>
 * <td>URI to the Gesmes/TS-XML schema</td></tr>
 * <tr><td><code>edifact.schemaLocation</code></td>
 * <td>string</td>
 * <td>URI to the Gesmes/TS-XML EDIFACT extension schema</td></tr>
 * </table></center>
 *
 * <p>The <code>GesmesTSXMLWriter</code> is not internally synchronized.
 * External synchronisation is necessary if multiple threads want to
 * use the same <code>GesmesTSXMLWriter</code>.</p>
 *
 * @author
 * <a href="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</a>
 * @author
 * <a href="mailto:Paula.Rocha@ECB.int">Paula Rocha</a>
 *
 * @version @(#)$Id: GesmesTSXMLWriter.java,v 1.1 2017/01/19 16:14:02 trainot Exp $
 *
 * @see KeyFamilyURIMapping
 */
public class GesmesTSXMLWriter
	extends XMLWriter
	implements StructureListener, DataListener, Constants
{
	/* ======================= Constants ======================== */

	/** Current Cube nesting code: no Cube */
	private static final int NONE = 0;

	/** Current Cube nesting code: DataSet */
	private static final int DS = 1;

	/** Current Cube nesting code: DataSet Series */
	private static final int DS_TS = 5;

	/** Current Cube nesting code: DataSet Series Obs */
	private static final int DS_TS_OBS = 13;

	/** Current Cube nesting code: DataSet Group */
	private static final int DS_SG = 3;

	/** Current Cube nesting code: DataSet Group Series */
	private static final int DS_SG_TS = 7;

	/** Current Cube nesting code: DataSet Group Series Obs */
	private static final int DS_SG_TS_OBS = 15;

	/** Current Cube nesting code: DataSet DataSetAtt */
	private static final int DS_ATT = 17;

	/** Current Cube nesting code: DataSet Series SeriesAtt */
	private static final int DS_TS_ATT = 21;

	/** Current Cube nesting code: DataSet Group GroupAtt */
	private static final int DS_SG_ATT = 19;

	/** Current Cube nesting code: DataSet Group Series SeriesAtt */
	private static final int DS_SG_TS_ATT = 23;

	/* ========================= Fields ========================= */

	/** Current XML dateTime format.
	 * Changes its pattern according to frequency. */
	private final XMLDateFormat xmlDateFormat;

	/** Links key families to their namespace URIs. */
	private final KeyFamilyURIMapping uriMapping;

	/** The current maintenance agency id. */
	private String currentMaintenanceAgencyId;

	/** The current <code>KeyFamily</code>. */
	private KeyFamily keyFamily;

	/**
	 * Comparator for sorting the Concept identifiers in the
	 * current {@link #keyFamily}.
	 */
	private Comparator conceptIdComparator;

	/** Whether message action is delete. */
	private boolean isDeleteMessage;

	/** The current Cube status. */
	private int cubeStatus;

	/** The current sibling group key. Only valid if SG is in status. */
	private String sgKey;

	/** The current sibling group key. Only valid if TS is in status. */
	private String tsKey;

	/** The current observation "key". Only valid if OBS is in status. */
	private RegularInterval obsInterval;

	/**
	 * Whether the time format has been written for a Series.
	 * Only valid if TS is in status.
	 */
	private boolean timeFormatWritten;

	/* ====================== Constructors ====================== */

	/**
	 * Construct an <code>GesmesTSXMLWriter</code> to write to a file.
	 *
	 * @param name used for logging and preferences.
	 * @param filename where to write to.
	 *
	 * @throws NullPointerException if <code>filename==null</code>
	 * and the preference variable
	 * <code>destination</code> is undefined.
	 * @throws IOException if the file cannot be opened for writing.
	 */
	public GesmesTSXMLWriter ( final String name, final String filename )
		throws IOException
	{
		super( name, filename );

		xmlDateFormat = new XMLDateFormat();
		uriMapping = KeyFamilyURIMapping.getInstance();

		currentMaintenanceAgencyId = null;
		keyFamily = null;
		conceptIdComparator = null;
		isDeleteMessage = false;

		cubeStatus = NONE;
		tsKey = sgKey = null;
		timeFormatWritten = false;

		writeXMLHeader();
	}

	/**
	 * Construct an <code>GesmesTSXMLWriter</code>
	 * writing to a file according to preferences.
	 *
	 * @param name used for logging and preferences.
	 *
	 * @throws NullPointerException if the preference variable
	 * <code>destination</code> is undefined.
	 * @throws IOException if the file cannot be opened for writing.
	 */
	public GesmesTSXMLWriter ( final String name )
		throws IOException
	{
		this( name, null );
	}

	/* ===================== Public Methods ===================== */

	/* ------------------------ Listener ------------------------ */

	/* ---------------------- Interchange ----------------------- */

	public void handle ( final InterchangeBegin interchangeBegin ) {
		try {
			componentHelper.logger.log(
				Level.INFO,
				componentHelper.logName,
				"handle( InterchangeBegin )",
				"Begin writing interchange "
				+ interchangeBegin.getIref() + "."
			);

			openElement( "Envelope" );
			writeAttribute( "id", interchangeBegin.getIref() );
			writeAttribute( "xmlns", XMLNS_GESMES );
			writeAttribute( "xmlns:gesmes", XMLNS_GESMES );
			writeAttribute( "xmlns:edifact", XMLNS_EDIFACT );
			writeAttribute( "xmlns:xsi", XMLNS_XSI );

			final String gesmesSchema = componentHelper.getPreference(
				"gesmes.schemaLocation", null
			);
			final String edifactSchema = componentHelper.getPreference(
				"edifact.schemaLocation", null
			);
			if ( gesmesSchema != null || edifactSchema != null ) {
				writeAttribute(
					"xsi:schemaLocation",
					(
						gesmesSchema != null
							? XMLNS_GESMES + ' ' + gesmesSchema : ""
					) + (
						edifactSchema != null
							? ' ' + XMLNS_EDIFACT + ' ' + edifactSchema : ""
					)
				);
			}

			if ( interchangeBegin.isTest() ) {
				writeElement( "test", "" );
			}

			xmlDateFormat.applyPattern( XMLDateFormat.SECONDLY_PATTERN );

			{
				final Date preparationDate
				= interchangeBegin.getPreparationDateTimeAsDate();
				if ( preparationDate != null ) {
					final String prepDateTimeXML = xmlDateFormat.format(
						preparationDate
					);
					writeElement( "prepared", prepDateTimeXML );
				}
			}

			openElement( "Sender" );
			writeAttribute( "id", interchangeBegin.getSender() );
			closeElement( "Sender" );

			final String receiverId = interchangeBegin.getReceiver();
			if ( ! "".equals( receiverId ) ) {
				openElement( "Receiver" );
				writeAttribute( "id", receiverId );
				closeElement( "Receiver" );
			}
		}
		catch ( final IOException ioe ) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"handle( InterchangeBegin )",
				"Output problem.",
				ioe
			);
			close();
		}
	}

	public void handle ( final InterchangeEnd interchangeEnd ) {
		try {
			closeElement( "Envelope" );

			componentHelper.logger.log(
				Level.INFO,
				componentHelper.logName,
				"handle( InterchangeEnd )",
				"Ended writing interchange "
				+ interchangeEnd.getIref() + "."
			);
		}
		catch ( final IOException ioe ) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"handle( InterchangeEnd )",
				"Output problem.",
				ioe
			);
			close();
		}
	}

	/* ------------------------ Message ------------------------- */

	public void handle ( final MessageBegin messageBegin ) {
		currentMaintenanceAgencyId = messageBegin.getMaintenanceAgencyId();
		isDeleteMessage = messageBegin.isDeleteAction();

		try {
			componentHelper.logger.log(
				Level.FINE,
				componentHelper.logName,
				"handle( MessageBegin )",
				"Begin writing Message " + messageBegin.getMref() + "."
			);

			openElement( "Envelope" );
			writeAttribute( "id", messageBegin.getMref() );

			if ( ! "".equals( messageBegin.getKeyFamilyId() ) ) {
				// Set current keyfamily
				final String maintenanceAgencyId
				= messageBegin.getMaintenanceAgencyId();
				final Metabase metabase = Metabase.getMetabase();
				final MaintenanceAgency maintenanceAgency
				= metabase.getMaintenanceAgency( maintenanceAgencyId );
				if ( maintenanceAgency != null ) {
					final String keyFamilyId = messageBegin.getKeyFamilyId();
					keyFamily = maintenanceAgency.getKeyFamily( keyFamilyId );
					if ( keyFamily != null ) {
						conceptIdComparator
						= new ConceptIdComparator( keyFamily );
					} else {
						componentHelper.logger.log(
							Level.SEVERE,
							componentHelper.logName,
							"handle( MessageBegin )",
							"Maintenance agengy " + maintenanceAgencyId
							+ " does not know key family \""
							+ keyFamilyId + "\"."
						);
						close();
					}
				} else {
					componentHelper.logger.log(
						Level.SEVERE,
						componentHelper.logName,
						"handle( MessageBegin )",
						"Unknown maintenance agengy \"" + maintenanceAgencyId
						+ "\"."
					);
					close();
				}

				final String keyFamilySchema;
				if (uriMapping != null) {
				    keyFamilySchema = uriMapping.schemaLocation(
					maintenanceAgencyId, keyFamily);
				} else {
					keyFamilySchema = componentHelper.getPreference(
							"defaultEntryPoint", "https://sreg.escb.eu/ws/rest/schema/datastructure/");
				}
				
				if ( keyFamilySchema != null ) {
					writeAttribute(
						"xsi:schemaLocation",
						keyFamilySchema + maintenanceAgencyId +
						"/" + messageBegin.getKeyFamilyId()  + "/1.0/?version=2.0"
					);
				}
			} else {
				keyFamily = null;
				conceptIdComparator = null;
			}

			{
				final String subject = messageBegin.getIdentity();
				if ( !"".equals( subject ) ) {
					writeElement( "name", messageBegin.getIdentity() );
				}
			}

			xmlDateFormat.applyPattern( XMLDateFormat.SECONDLY_PATTERN );

			{
				final Date extractionDate
				= messageBegin.getExtractionDateTimeAsDate();
				if ( extractionDate != null ) {
					final String extractDateTimeXML
					= xmlDateFormat.format(
						extractionDate
					);
					writeElement( "prepared", extractDateTimeXML );
				}
			}

			{
				openElement( "Sender" );
				writeAttribute( "id", messageBegin.getSender() );

				for (
					final Iterator iterator = messageBegin.contactIterator();
					iterator.hasNext();
				) {
					final Contact contact = (Contact) iterator.next();

					openElement( "Contact" );
					writeElement( "name", contact.getName() );

					writeElement( "department", contact.getIdentity() );

					final String role;
					if ( contact.isResponsibleForInformationProduction() ) {
						role = "Responsible for Information Production";
					} else if ( contact.isHeadOfInformationProduction() ) {
						role = "Head of Information Production";
					} else if ( contact.isResponsibleForDataProcessing() ) {
						role = "Responsible for Data Processing";
					} else if ( contact.isHeadOfDataProcessing() ) {
						role = "Head of Data Processing";
					} else {
						role = contact.getFunction();
						componentHelper.logger.log(
							Level.WARNING,
							componentHelper.logName,
							"handle( MessageBegin )",
							"Contact function \"" + contact.getFunction()
							+ "\" is not known."
						);
					}
					writeElement( "role", role );

					for (
						final Iterator channels = contact.channelIterator();
						channels.hasNext();
					) {
						final ContactChannel channel
						= (ContactChannel) channels.next();

						final String media;
						if ( channel.isEmailAddress() ) {
							media = "uri";
						} else if ( channel.isFaxNumber() ) {
							media = "fax";
						} else if ( channel.isTelephoneNumber() ) {
							media = "telephone";
						} else if ( channel.isX400Address() ) {
							media = "x400";
						} else {
							media = channel.getType();
							componentHelper.logger.log(
								Level.WARNING,
								componentHelper.logName,
								"handle( MessageBegin )",
								"Contact channel type \"" + channel.getType()
								+ "\" is not known."
							);
						}

						writeElement(
							media,
							channel.isEmailAddress()
							? "mailto:" + channel.getNumber()
							: channel.getNumber()
						);
					}

					closeElement( "Contact" );
				}

				closeElement( "Sender" );
			}

			final String receiverId = messageBegin.getReceiver();
			if ( ! "".equals( receiverId ) ) {
				openElement( "Receiver" );
				writeAttribute( "id", receiverId );
				closeElement( "Receiver" );
			}

			{
				final Interval reportingInterval
				= messageBegin.getReportingPeriodAsInterval();

				if ( reportingInterval != null ) {
					final String periodBegin = xmlDateFormat.format(
						reportingInterval.getBegin()
					);
					final String periodEnd = xmlDateFormat.format(
						reportingInterval.getEnd()
					);

					writeElement( "reportingBegin", periodBegin );
					writeElement( "reportingEnd", periodEnd );
				}
			}

			if ( ! "".equals( messageBegin.getDataSetId() ) ) {
				writeElement( "dataSetId", messageBegin.getDataSetId() );
			}

			final String action;
			if ( "".equals( messageBegin.getAction() ) ) {
				action = "";
			} else if ( messageBegin.isUpdateOrReplaceAction() ) {
				action = "update";
			} else if ( messageBegin.isDeleteAction() ) {
				action = "delete";
			} else if ( messageBegin.isTestingAction() ) {
				action = "test";
				componentHelper.logger.log(
					Level.WARNING,
					componentHelper.logName,
					"handle( MessageBegin )",
					"Using deprecated test action.",
					messageBegin
				);
			} else {
				action = "";
				componentHelper.logger.log(
					Level.SEVERE,
					componentHelper.logName,
					"handle( MessageBegin )",
					"Message action \"" + messageBegin.getAction()
					+ "\" is not known.",
					messageBegin
				);
			}

			if ( ! "".equals( action ) ) {
				writeElement( "action", action );
			}

			openElement( "Extension" );

			writeElement(
				"edifact:maintenanceAgencyId",
				messageBegin.getMaintenanceAgencyId()
			);

			if ( ! "".equals( messageBegin.getKeyFamilyId() ) ) {
				writeElement(
					"edifact:keyFamilyId",
					messageBegin.getKeyFamilyId()
				);
			}

			closeElement( "Extension" );
		}
		catch ( final IOException ioe ) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"handle( MessageBegin )",
				"Output problem.",
				ioe
			);
			close();
		}
	}

	public void handle ( final MessageEnd messageEnd ) {
		try {
			closeAllCubeElements();

			closeElement( "Envelope" );

			componentHelper.logger.log(
				Level.FINE,
				componentHelper.logName,
				"handle( MessageBegin )",
				"End writing message "
				+ messageEnd.getMref() + "."
			);
		}
		catch ( final IOException ioe ) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"handle( MessageEnd )",
				"Output problem.",
				ioe
			);
			close();
		}
	}

	/* ----------------------- Code List ------------------------ */

	public void handle ( final CodeList codeList ) {
		final String id = codeList.getId();
		try {
			openElement( "CodeList" );
			writeAttribute( "id", id );

			final String name = codeList.getName();
			if ( !"".equals( name ) ) {
				writeElement( "name", name );
			}

			final String description = codeList.getDescription();
			if ( !"".equals( description ) ) {
				writeElement( "description", description );
			}

			for (
				final Iterator
					values = codeList.codeValueIterator(),
					descriptions = codeList.codeDescriptionIterator();
				values.hasNext() && descriptions.hasNext();
			) {
				final String value = (String) values.next();
				openElement( "Code" );
				writeAttribute( "value", value );

				final String desc = (String) descriptions.next();
				if ( !"".equals( desc ) ) {
					writeElement( "description", desc );
				}

				closeElement( "Code" );
			}

			closeElement( "CodeList" );
		}
		catch ( final IOException ioe ) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"handle( CodeList )",
				"Output problem.",
				ioe
			);
			close();
		}
	}

	/* ------------------ Statistical Concept ------------------- */

	public void handle ( final StatisticalConcept statisticalConcept ) {
		final String id = statisticalConcept.getId();
		try {
			openElement( "Concept" );
			writeAttribute( "id", id );

			final String name = statisticalConcept.getName();
			if ( !"".equals( name ) ) {
				writeElement( "name", name );
			}

			final String description = statisticalConcept.getDescription();
			if ( !"".equals( description ) ) {
				writeElement( "description", description );
			}

			closeElement( "Concept" );
		}
		catch ( final IOException ioe ) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"handle( StatisticalConcept )",
				"Output problem.",
				ioe
			);
			close();
		}
	}

	/* ----------------------- Key Family ----------------------- */

	public void handle ( final KeyFamily keyFamily ) {
		this.keyFamily = keyFamily;
		// unnecessary to set conceptIdSeparator here.

		final String id = keyFamily.getId();
		try {
			openElement( "KeyFamily" );
			writeAttribute( "id", id );

			final String name = keyFamily.getName();
			if ( !"".equals( name ) ) {
				writeElement( "name", name );
			}

			final String description = keyFamily.getDescription();
			if ( !"".equals( description ) ) {
				writeElement( "description", description );
			}

			writeElement(
				"uri",
				"urn:sdmx:org.sdmx.infomodel.datastructure.DataStructure="+currentMaintenanceAgencyId
				+":" + id  + "(1.0):compact"
			);

			for (
				final Iterator iterator = keyFamily.componentIterator();
				iterator.hasNext();
			) {
				final KeyFamilyComponent component
				= (KeyFamilyComponent) iterator.next();
				write( component );
			}

			final StringBuffer[] dimensions = new StringBuffer[4];
			final StringBuffer[] attributes = new StringBuffer[4];
			for ( int i = 0; i < 4; ++i ) {
				dimensions[i] = new StringBuffer();
				attributes[i] = new StringBuffer();
			}

			for (
				final Iterator iterator = keyFamily.componentIterator();
				iterator.hasNext();
			) {
				final KeyFamilyComponent component
				= (KeyFamilyComponent) iterator.next();
				final String conceptId = component.getConceptId();

				if (
					component.isKeyStructureDimension()
					||
					component.isTimeDimension()
					&&
					component != keyFamily.getTimeFormat()
				) {
					dimensions[0].append( conceptId );
					dimensions[0].append( ' ' );
					if ( !component.isTimeDimension() ) {
						dimensions[1].append( conceptId );
						dimensions[1].append( ' ' );
						if ( component != keyFamily.getFrequency() ) {
							dimensions[2].append( conceptId );
							dimensions[2].append( ' ' );
						}
					}
				} else {
					if (
						component.isAttachedToObservation()
						||
						component.isArrayCell()
					) {
						attributes[0].append( conceptId );
						attributes[0].append( ' ' );
					} else if (
						component.isAttachedToTimeSeries()
						||
						component.isTimeDimension()
						&&
						component == keyFamily.getTimeFormat()
					) {
						attributes[1].append( conceptId );
						attributes[1].append( ' ' );
					} else if ( component.isAttachedToSiblingGroup() ) {
						attributes[2].append( conceptId );
						attributes[2].append( ' ' );
					} else if ( component.isAttachedToDataSet() ) {
						attributes[3].append( conceptId );
						attributes[3].append( ' ' );
					}
				}
			}

			for ( int i = 0; i < 4; ++i ) {
				if ( attributes[i].length() == 0 ) continue;
				openElement( "Level" );
				writeAttribute(
					"dimensions",
					dimensions[i].toString().trim()
				);
				writeAttribute(
					"attributes",
					attributes[i].toString().trim()
				);
				closeElement( "Level" );
			}

			closeElement( "KeyFamily" );
		}
		catch ( final IOException ioe ) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"handle( KeyFamily )",
				"Output problem.",
				ioe
			);
			close();
		}
	}

	/* ---------------------- Time Series ----------------------- */

	public void handle ( final TimeSeries timeSeries ) {
		final String key = timeSeries.getKey();
		final RegularInterval range = timeSeries.getInterval();
		final RegularInterval period = range != null
			? range.computeBasePeriod( 0 )
			: null;
		final int obsCount = range != null ? range.getSize() : 0;
		try {
			// This loop rattles through a couple of state changes.
			do {
				switch ( cubeStatus ) {
					case NONE:
						openElement( "DataSet" );
						writeAttribute(
							"xmlns",
							"urn:sdmx:org.sdmx.infomodel.datastructure.DataStructure="+currentMaintenanceAgencyId
							+":" + keyFamily.getId()  + "(1.0):compact"
						);
						cubeStatus = DS;
						break;
					case DS:
						openElement( "Series", conceptIdComparator );
						writeKey( key );
						if (
							range != null
							&&
							! isDeleteMessage
						) {
							writeTimeFormat( range );
							timeFormatWritten = true;
						} else {
							timeFormatWritten = false;
						}
						tsKey = key;
						cubeStatus = DS_TS;
						break;
					case DS_SG:
						if ( keyExtends( key, sgKey ) ) {
							openElement( "Series", conceptIdComparator );
							writeFrequency( key );
							if (
								range != null
								&&
								! isDeleteMessage
							) {
								writeTimeFormat( range );
								timeFormatWritten = true;
							} else {
								timeFormatWritten = false;
							}
							tsKey = key;
							cubeStatus = DS_SG_TS;
						} else {
							closeElement( "Group" );
							cubeStatus = DS;
						}
						break;
					case DS_TS:
						if ( !key.equals( tsKey ) ) {
							closeElement( "Series" );
							cubeStatus = DS;
						} else if ( range != null ) {
							if (
								! isDeleteMessage
								&&
								! timeFormatWritten
							) {
								writeTimeFormat( range );
								timeFormatWritten = true;
							}
							if ( obsCount > 0 ) {
								openElement( "Obs", conceptIdComparator );
								writeTimePeriod( period );
								obsInterval = period;
								cubeStatus = DS_TS_OBS;
							}
						}
						break;
					case DS_SG_TS:
						if ( !key.equals( tsKey ) ) {
							closeElement( "Series" );
							cubeStatus = DS_SG;
						} else if ( range != null ) {
							if (
								! isDeleteMessage
								&&
								! timeFormatWritten
							) {
								writeTimeFormat( range );
								timeFormatWritten = true;
							}
							if ( obsCount > 0 ) {
								openElement( "Obs", conceptIdComparator );
								writeTimePeriod( period );
								obsInterval = period;
								cubeStatus = DS_SG_TS_OBS;
							}
						}
						break;
					case DS_TS_OBS:
						if (
							!key.equals( tsKey )
						) {
							closeElement( "Obs" );
							closeElement( "Series" );
							cubeStatus = DS;
						} else if (
							!period.equals( obsInterval )
						) {
							closeElement( "Obs" );
							cubeStatus = DS_TS;
						}
						break;
					case DS_SG_TS_OBS:
						if (
							!key.equals( tsKey )
						) {
							closeElement( "Obs" );
							closeElement( "Series" );
							cubeStatus = DS_SG;
						} else if (
							!period.equals( obsInterval )
						) {
							closeElement( "Obs" );
							cubeStatus = DS_SG_TS;
						}
						break;
					case DS_ATT:
						closeElement( "DataSetAtt" );
						cubeStatus = DS;
						break;
					case DS_SG_ATT:
						closeElement( "GroupAtt" );
						cubeStatus = DS_SG;
						break;
					case DS_TS_ATT:
						if ( !key.equals( tsKey ) ) {
							closeElement( "SeriesAtt" );
							closeElement( "Series" );
							cubeStatus = DS;
						} else if ( range != null ) {
							if (
								! isDeleteMessage
								&&
								! timeFormatWritten
							) {
								writeTimeFormat( range );
								timeFormatWritten = true;
							}
							if ( obsCount > 0 ) {
								closeElement( "SeriesAtt" );
								cubeStatus = DS_TS;
							}
						}
						break;
					case DS_SG_TS_ATT:
						if ( !key.equals( tsKey ) ) {
							closeElement( "SeriesAtt" );
							closeElement( "Series" );
							cubeStatus = DS_SG;
						} else if ( range != null ) {
							if (
								! isDeleteMessage
								&&
								! timeFormatWritten
							) {
								writeTimeFormat( range );
								timeFormatWritten = true;
							}
							if ( obsCount > 0 ) {
								closeElement( "SeriesAtt" );
								cubeStatus = DS_SG_TS;
							}
						}
						break;
				}
			} while (
				obsCount == 0
				&&
				cubeStatus != DS_TS
				&&
				cubeStatus != DS_TS_ATT
				&&
				cubeStatus != DS_SG_TS
				&&
				cubeStatus != DS_SG_TS_ATT
				||
				obsCount != 0
				&&
				cubeStatus != DS_TS_OBS
				&&
				cubeStatus != DS_SG_TS_OBS
			);

			if ( obsCount > 0 ) {
				for (
					final Iterator
						intervals = range.iterator(),
						observations = timeSeries.observationIterator();
					intervals.hasNext() && ( observations.hasNext() || isDeleteMessage );
				) {
					final RegularInterval interval
					= (RegularInterval) intervals.next();
					final Observation observation
					= observations.hasNext() ?
						(Observation) observations.next() : null;

					if (
						observation != null
						&&
						observation.isEmpty()
						&&
						!isDeleteMessage
					) continue;

					if ( !interval.equals( obsInterval ) ) {
						closeElement( "Obs" );
						openElement( "Obs", conceptIdComparator );
						writeTimePeriod( interval );
						obsInterval = interval;
					}
					if ( observation != null ) {
						writeObservation( observation );
					}
				}
			}
		}
		catch ( final IOException ioe ) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"handle( TimeSeries )",
				"Output problem.",
				ioe
			);
			close();
		}
	}

	/* ----------------------- Attribute ------------------------ */

	public void handle ( final Attribute attribute ) {
		final String key = attribute.getKey();
		final RegularInterval range = attribute.getInterval();
		final String name = attribute.getName();
		final String value = attribute.getValue();
		final KeyFamilyComponent component = keyFamily.getComponent(
			name
		);

		if ( component == null ) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"handle( Attribute )",
				"No concept \"" + name
				+ "\" in key family \"" + keyFamily.getId() + "\".",
				attribute
			);
			close();
		}

		try {
			// This loop rattles through a couple of state changes.
			STATUS_SEQUENCE: while ( true ) {
				switch ( cubeStatus ) {
					case NONE:
						openElement( "DataSet" );
						writeAttribute(
							"xmlns",
							"urn:sdmx:org.sdmx.infomodel.datastructure.DataStructure="+currentMaintenanceAgencyId
							+":" + keyFamily.getId()  + "(1.0):compact"
						);
						cubeStatus = DS;
						break;
					case DS:
						if (
							component.isAttachedToObservation()
							||
							component.isAttachedToTimeSeries()
						) {
							openElement( "Series", conceptIdComparator );
							writeKey( key );
							if (
								range != null
								&&
								! isDeleteMessage
							) {
								writeTimeFormat( range );
								timeFormatWritten = true;
							} else {
								timeFormatWritten = false;
							}
							tsKey = key;
							cubeStatus = DS_TS;
						} else if ( component.isAttachedToSiblingGroup() ) {
							openElement( "Group", conceptIdComparator );
							writeKey( key );
							sgKey = key;
							cubeStatus = DS_SG;
						} else if ( component.isAttachedToDataSet() ) {
							break STATUS_SEQUENCE;
						}
						break;
					case DS_SG:
						if (
							(
								component.isAttachedToObservation()
								||
								component.isAttachedToTimeSeries()
							)
							&&
							keyExtends( key, sgKey )
						) {
							openElement( "Series", conceptIdComparator );
							writeFrequency( key );
							if (
								range != null
								&&
								! isDeleteMessage
							) {
								writeTimeFormat( range );
								timeFormatWritten = true;
							} else {
								timeFormatWritten = false;
							}
							tsKey = key;
							cubeStatus = DS_SG_TS;
						} else if (
							component.isAttachedToSiblingGroup()
							&&
							key.equals( sgKey )
						) {
							break STATUS_SEQUENCE;
						} else if (
							component.isAttachedToDataSet()
						) {
							closeElement( "Group" );
							openElement( "DataSetAtt", conceptIdComparator );
							cubeStatus = DS_ATT;
						} else {
							closeElement( "Group" );
							cubeStatus = DS;
						}
						break;
					case DS_TS:
						if (
							component.isAttachedToObservation()
							&&
							key.equals( tsKey )
						) {
							if (
								! isDeleteMessage
								&&
								! timeFormatWritten
							) {
								writeTimeFormat( range );
								timeFormatWritten = true;
							}
							openElement( "Obs", conceptIdComparator );
							writeTimePeriod( range );
							obsInterval = range;
							cubeStatus = DS_TS_OBS;
						} else if (
							component.isAttachedToTimeSeries()
							&&
							key.equals( tsKey )
						) {
							break STATUS_SEQUENCE;
						} else if (
							component.isAttachedToDataSet()
						) {
							closeElement( "Series" );
							openElement( "DataSetAtt", conceptIdComparator );
							cubeStatus = DS_ATT;
						} else {
							closeElement( "Series" );
							cubeStatus = DS;
						}
						break;
					case DS_SG_TS:
						if (
							component.isAttachedToObservation()
							&&
							key.equals( tsKey )
						) {
							if (
								! isDeleteMessage
								&&
								! timeFormatWritten
							) {
								writeTimeFormat( range );
								timeFormatWritten = true;
							}
							openElement( "Obs", conceptIdComparator );
							writeTimePeriod( range );
							obsInterval = range;
							cubeStatus = DS_SG_TS_OBS;
						} else if (
							component.isAttachedToTimeSeries()
							&&
							key.equals( tsKey )
						) {
							break STATUS_SEQUENCE;
						} else if (
							component.isAttachedToSiblingGroup()
							&&
							key.equals( sgKey )
						) {
							closeElement( "Series" );
							openElement( "GroupAtt", conceptIdComparator );
							cubeStatus = DS_SG_ATT;
						} else {
							closeElement( "Series" );
							cubeStatus = DS_SG;
						}
						break;
					case DS_TS_OBS:
						if (
							component.isAttachedToObservation()
							&&
							key.equals( tsKey )
							&&
							range.equals( obsInterval )
						) {
							break STATUS_SEQUENCE;
						} else if (
							component.isAttachedToTimeSeries()
							&&
							key.equals( tsKey )
						) {
							closeElement( "Obs" );
							openElement( "SeriesAtt", conceptIdComparator );
							cubeStatus = DS_TS_ATT;
						} else {
							closeElement( "Obs" );
							cubeStatus = DS_TS;
						}
						break;
					case DS_SG_TS_OBS:
						if (
							component.isAttachedToObservation()
							&&
							key.equals( tsKey )
							&&
							range.equals( obsInterval )
						) {
							break STATUS_SEQUENCE;
						} else if (
							component.isAttachedToTimeSeries()
							&&
							key.equals( tsKey )
						) {
							closeElement( "Obs" );
							openElement( "SeriesAtt", conceptIdComparator );
							cubeStatus = DS_SG_TS_ATT;
						} else {
							closeElement( "Obs" );
							cubeStatus = DS_SG_TS;
						}
						break;
					case DS_ATT:
						if ( component.isAttachedToDataSet() ) {
							break STATUS_SEQUENCE;
						} else {
							closeElement( "DataSetAtt" );
							cubeStatus = DS;
						}
						break;
					case DS_SG_ATT:
						if (
							component.isAttachedToSiblingGroup()
							&&
							key.equals( sgKey )
						) {
							break STATUS_SEQUENCE;
						} else {
							closeElement( "GroupAtt" );
							cubeStatus = DS_SG;
						}
						break;
					case DS_TS_ATT:
						if (
							component.isAttachedToObservation()
							&&
							key.equals( tsKey )
						) {
							if (
								! isDeleteMessage
								&&
								! timeFormatWritten
							) {
								writeTimeFormat( range );
								timeFormatWritten = true;
							}
							closeElement( "SeriesAtt" );
							cubeStatus = DS_TS;
						} else if (
							component.isAttachedToTimeSeries()
							&&
							key.equals( tsKey )
						) {
							break STATUS_SEQUENCE;
						} else {
							closeElement( "SeriesAtt" );
							cubeStatus = DS_TS;
						}
						break;
					case DS_SG_TS_ATT:
						if (
							component.isAttachedToObservation()
							&&
							key.equals( tsKey )
						) {
							if (
								! isDeleteMessage
								&&
								! timeFormatWritten
							) {
								writeTimeFormat( range );
								timeFormatWritten = true;
							}
							closeElement( "SeriesAtt" );
							cubeStatus = DS_SG_TS;
						} else if (
							component.isAttachedToTimeSeries()
							&&
							key.equals( tsKey )
						) {
							break STATUS_SEQUENCE;
						} else {
							closeElement( "SeriesAtt" );
							cubeStatus = DS_SG_TS;
						}
						break;
				}
			}
			writeAttribute( name, value );
		}
		catch ( final IOException ioe ) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"handle( Attribute )",
				"Output problem.",
				ioe
			);
			close();
		}
	}

	/* ==================== Private Methods ===================== */

	/* ------------------------ Writing ------------------------- */

	/* ------------------------- Header ------------------------- */

	private void writeXMLHeader () {
		componentHelper.logger.log(
			Level.FINE,
			componentHelper.logName,
			"writeXMLHeader()",
			"Writing header."
		);

		try {
			writeRaw( "<!-- Gesmes/TS-XML from Gesmes/TS Suite -->" );
			writeRaw( NEW_LINE );
			writeRaw( "<!-- @(#)$Id: GesmesTSXMLWriter.java,v 1.1 2017/01/19 16:14:02 trainot Exp $ -->" );
			writeRaw( NEW_LINE );

			final String stylesheetURI = componentHelper.getPreference(
				"stylesheet", null
			);
			if ( stylesheetURI != null ) {
				writeRaw(
					"<?xml-stylesheet href=\""
					+ stylesheetURI
					+ "\" type=\"text/xsl\""
					+ " media=\"screen,html\"?>"
				);
				writeRaw( NEW_LINE );
			}
			writeRaw( NEW_LINE );
		}
		catch ( final IOException ioe ) {
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"writeXMLHeader",
				"Output problem.",
				ioe
			);
			close();
		}
	}

	/* ------------------- KeyFamilyComponent ------------------- */

	/** Write a <code>KeyFamilyComponent</code>. */
	private void write (
		final KeyFamilyComponent component
	) throws IOException {
		final String id = component.getConceptId();
		// TODO: avoid attribute names illegal in XML

		openElement( "Concept" );
		writeAttribute( "ref", id );

		if (
			component.isMandatory()
			||
			component == keyFamily.getObsValue()
			||
			component == keyFamily.getTimeFormat()
			// BB: I model time format as an attribute
			// because time series deletion does not require it stated.
		) {
			writeElement( "usage", "mandatory" );
		} else if ( !"".equals( component.getUsageStatus() ) ) {
			writeElement( "usage", "conditional" );
		}

		{
			final TextFormat format = component.getTextFormat();
			if ( format != null ) {
				final String type;
				switch ( format.getCharacterType() ) {
					case TextFormat.ANY: type = null; break;
					case TextFormat.ALPHABETIC: type = "alphabetic"; break;
					case TextFormat.NUMERIC: type = "numeric"; break;
					default: type = null;
						componentHelper.logger.log(
							Level.SEVERE,
							componentHelper.logName,
							"handle( KeyFamily )",
							"Key family component \"" + id
							+ "\" has invalid text format \""
							+ component.getFormat() + "\".",
							keyFamily
						);
				}

				openElement( "TextFormat" );

				if ( type != null ) {
					writeElement( "cType", type );
				}

				openElement( "length" );
				writeAttribute(
					"exact",
					format.isVariableLength() ? "false" : "true"
				);
				writeCharacters( Integer.toString(
					format.getReferenceLength()
				) );
				closeElement( "length" );

				closeElement( "TextFormat" );
			}
		}

		if ( component.isTimeDimension() ) {
			if ( component != keyFamily.getTimeFormat() ) {
				openElement( "TimePointSet" );
				openElement( "periodicity" );
				openElement( "value-of" );
				writeAttribute(
					"concept",
					keyFamily.getTimeFormat().getConceptId()
				);
				closeElement( "value-of" );
				closeElement( "periodicity" );
				closeElement( "TimePointSet" );
			} else {
				writeElement( "TimeDurationSet", "" );
			}
		}

		if (
			component == keyFamily.getObsValue()
			||
			component == keyFamily.getObsPreBreak()
		) {
			writeElement( "RealNumberSet", "" );
		}

		// TODO: special case for CL_TIME_FORMAT, if exists
		if ( component.isCoded() ) {
			openElement( "CodeList" );
			writeAttribute( "ref", component.getCodeListId() );
			closeElement( "CodeList" );
		}

		final String conceptType;
		if ( component.isFrequencyDimension() ) {
			conceptType = "frequency dimension";
		} else if ( component.isKeyStructureDimension() ) {
			conceptType = "dimension";
		} else if ( component.isTimeDimension() ) {
			if ( component != keyFamily.getTimeFormat() ) {
				conceptType = "time dimension";
			} else {
				conceptType = "duration attribute";
			}
		} else if ( component.isArrayCell() ) {
			conceptType = "array cell attribute";
		} else if ( component.isAttribute() ) {
			conceptType = "attribute";
		} else {
			conceptType = null;
		}

		final String attachmentLevel;
		if (
			component.isAttachedToObservation()
			||
			component.isArrayCell() // to handle OBS_VALUE specialities
		) {
			attachmentLevel = "observation";
		} else if (
			component.isAttachedToTimeSeries()
			||
			component == keyFamily.getTimeFormat()
		) {
			attachmentLevel = "time series";
		} else if ( component.isAttachedToSiblingGroup() ) {
			attachmentLevel = "sibling group";
		} else if ( component.isAttachedToDataSet() ) {
			attachmentLevel = "data set";
		} else {
			attachmentLevel = null;
		}

		final boolean conceptTyped = conceptType != null;
		final boolean attachmentLevelled = attachmentLevel != null;
		if ( conceptTyped || attachmentLevelled ) {
			openElement( "Extension" );
			if ( conceptTyped ) {
				writeElement( "edifact:conceptType", conceptType );
			}
			if ( attachmentLevelled ) {
				writeElement( "edifact:attachmentLevel", attachmentLevel );
			}
			closeElement( "Extension" );
		}

		closeElement( "Concept" );
	}

	/* --------------------- Identification --------------------- */

	private void writeKey ( final String key ) throws IOException {
		if ( key.length() > 0 ) {
			for (
				final Iterator
					dimensions = keyFamily.keyStructureIterator(),
					values = new KeyTokenizer( key );
				dimensions.hasNext() && values.hasNext();
			) {
				final KeyFamilyComponent dimension
				= (KeyFamilyComponent) dimensions.next();
				final String value = (String) values.next();
				if ( "".equals( value ) ) continue;
				writeAttribute( dimension.getConceptId(), value );
			}
		}
	}

	private void writeFrequency ( final String key ) throws IOException {
		if ( key.length() > 0 ) {
			for (
				final Iterator
					dimensions = keyFamily.keyStructureIterator(),
					values = new KeyTokenizer( key );
				dimensions.hasNext() && values.hasNext();
			) {
				final KeyFamilyComponent dimension
				= (KeyFamilyComponent) dimensions.next();
				final String value = (String) values.next();
				if ( "".equals( value ) ) continue;
				if ( dimension == keyFamily.getFrequency() ) {
					writeAttribute( dimension.getConceptId(), value );
				}
			}
		}
	}

	private void writeTimeFormat (
		final RegularInterval interval
	) throws IOException {
		final int duration = interval.getFrequency();
		writeAttribute(
			keyFamily.getTimeFormat().getConceptId(),
			XMLDateFormat.formatDuration( duration )
		);
	}

	private void writeTimePeriod (
		final RegularInterval interval
	) throws IOException {
		final Date time = interval.getBegin();
		final int duration = interval.getFrequency();
		xmlDateFormat.setFrequency( duration );
		writeAttribute(
			keyFamily.getTimePeriod().getConceptId(),
			xmlDateFormat.format( time )
		);
	}

	/* ---------------------- Observation ----------------------- */

	private void writeObservation (
		final Observation observation
	) throws IOException {
		final Iterator
			arrayCells = keyFamily.arrayCellIterator();
		{
			final String value = observation.getValue();
			if ( arrayCells.hasNext() ) {
				final KeyFamilyComponent obsValue
				= (KeyFamilyComponent) arrayCells.next();
				writeAttribute(
					obsValue.getConceptId(),
					observation.hasMissingValue()
					? "NaN" : value
				);
			}
		}
		{
			final String status = observation.getStatus();
			if ( arrayCells.hasNext() ) {
				final KeyFamilyComponent obsStatus
				= (KeyFamilyComponent) arrayCells.next();
				writeAttribute( obsStatus.getConceptId(), status );
			}
		}
		{
			final String conf = observation.getConfidentiality();
			if ( arrayCells.hasNext() ) {
				final KeyFamilyComponent obsConf
				= (KeyFamilyComponent) arrayCells.next();
				writeAttribute( obsConf.getConceptId(), conf );
			}
		}
		{
			final String preBreak = observation.getPreBreakValue();
			if ( arrayCells.hasNext() ) {
				final KeyFamilyComponent obsPreBreak
				= (KeyFamilyComponent) arrayCells.next();
				writeAttribute(
					obsPreBreak.getConceptId(),
					observation.hasMissingPreBreakValue()
					? "NaN" : preBreak
				);
			}
		}
	}

	/* ---------------------- Close cubes ----------------------- */

	/** Close all open Cube elements. */
	public void closeAllCubeElements () throws IOException {
		// This loop rattles through a couple of state changes.
		do {
			switch ( cubeStatus ) {
				case NONE: break;
				case DS:
					closeElement( "DataSet" );
					cubeStatus = NONE;
					break;
				case DS_SG:
					closeElement( "Group" );
					cubeStatus = DS;
					break;
				case DS_TS:
					closeElement( "Series" );
					cubeStatus = DS;
					break;
				case DS_SG_TS:
					closeElement( "Series" );
					cubeStatus = DS_SG;
					break;
				case DS_TS_OBS:
					closeElement( "Obs" );
					cubeStatus = DS_TS;
					break;
				case DS_SG_TS_OBS:
					closeElement( "Obs" );
					cubeStatus = DS_SG_TS;
					break;
				case DS_ATT:
					closeElement( "DataSetAtt" );
					cubeStatus = DS;
					break;
				case DS_SG_ATT:
					closeElement( "GroupAtt" );
					cubeStatus = DS_SG;
					break;
				case DS_TS_ATT:
					closeElement( "SeriesAtt" );
					cubeStatus = DS_TS;
					break;
				case DS_SG_TS_ATT:
					closeElement( "Series" );
					cubeStatus = DS_SG_TS;
					break;
			}
		} while ( cubeStatus != NONE );
	}

	/* ----------------------- Auxiliary ------------------------ */

	/** Determine if the state matches the requested Cube. */
	private boolean stateMatches(
		final String key,
		final RegularInterval interval,
		final String conceptId
	) {
		final KeyFamilyComponent component = keyFamily.getComponent(
			conceptId
		);
		switch ( cubeStatus ) {
			case NONE: return false;
			case DS: return component.isAttachedToDataSet();
			case DS_SG:
				return
					key.equals( sgKey )
					&&
					component.isAttachedToSiblingGroup();
			case DS_TS:
			case DS_SG_TS:
				return
					key.equals( tsKey )
					&&
					component.isAttachedToTimeSeries();
			case DS_TS_OBS:
			case DS_SG_TS_OBS:
				return
					key.equals( tsKey )
					&&
					interval.equals( obsInterval )
					&&
					component.isAttachedToObservation();
		}
		throw new RuntimeException( "Failure!" );
	}

	private boolean keyExtends (
		final String seriesKey, final String groupKey
	) {
		for (
			final Iterator
				dimensions = keyFamily.keyStructureIterator(),
				seriesKeyValues = new KeyTokenizer( seriesKey ),
				groupKeyValues = new KeyTokenizer( groupKey );
			dimensions.hasNext();
		) {
			final KeyFamilyComponent dimension
			= (KeyFamilyComponent) dimensions.next();
			final String seriesKeyValue = seriesKeyValues.hasNext()
				? (String) seriesKeyValues.next() : "";
			final String groupKeyValue = groupKeyValues.hasNext()
				? (String) groupKeyValues.next() : "";
			if ( !groupKeyValue.equals( seriesKeyValue ) ) {
				if (
					!dimension.isFrequencyDimension()
					||
					!"".equals( groupKeyValue )
				) {
					return false;
				}
			}
		}
		return true;
	}
}
