// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2003, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.xml;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.NoSuchElementException;

import Int.ecb.gesmescb.KeyFamily;
import Int.ecb.gesmescb.metabase.MaintenanceAgency;
import Int.ecb.gesmescb.metabase.Metabase;
import Int.ecb.gesmescb.time.RegularInterval;
import Int.ecb.gesmescb.util.ComponentHelper;
import Int.ecb.logging.Level;

/* ***************************** Class ****************************** */

/**
 * Singleton to find the URI for a {@link KeyFamily}.
 *
 * <center><table border="3" cellpadding="3">
 * <caption>
 * <a name="prefs">Available {@link Int.ecb.prefs preference} settings</a>
 * </caption>
 * <tr>
 * <th>preference name</th>
 * <th>type</th>
 * <th>meaning</th>
 * <th>default</th>
 * </tr>
 * <tr><td><code><i>Maintenance Agency id</i>.<i>Key Family id</i>.namespace</code></td>
 * <td>string</td><td>Key Family's namespace URI</td></tr>
 * <tr><td><code><i>Maintenance Agency id</i>.<i>Key Family id</i>.dtd</code></td>
 * <td>string</td><td>Key Family's XML system DTD URI</td></tr>
 * <tr><td><code><i>Maintenance Agency id</i>.<i>Key Family id</i>.schema</code></td>
 * <td>string</td><td>Key Family's XML schema location URI</td></tr>
 * <tr><td><code><i>Maintenance Agency id</i>.<i>Key Family id</i>.stylesheet</code></td>
 * <td>string</td><td>Key Family's XML XSLT stylesheet URI</td></tr>
 * </table></center>
 *
 * <center><table border="3" cellpadding="3">
 * <p>The URI string can contain function calls to the following functions</p>
 * <table>
 * <tr><td><code>date()</code></td><td>Key Family's version date</td></tr>
 * <tr><td><code>kfid()</code></td><td>Full Key Family identifier</td></tr>
 * <tr><td><code>kfidv()</code></td><td>Versioned Key Family identifier</td></tr>
 * <tr><td><code>kfidx()</code></td><td>Plain Key Family identifier</td></tr>
 * </table></center>
 *
 * @author
 * <A HREF="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</A>
 *
 * @version @(#)$Id: KeyFamilyURIMapping.java,v 1.1 2017/01/19 16:14:03 trainot Exp $
 */
public class KeyFamilyURIMapping {

	/* ========================= Nested ========================= */

	/**
	 * Pairs a maintenance agency id and a key family.
	 * This is what corresponds to a namespace URI.
	 */
	public interface LookupResult {
		String getMaintenanceAgencyId ();
		KeyFamily getKeyFamily ();
	}

	/** Implements calculations needed to expand preference values. */
	interface Function {
		String expand ( String maintenanceAgencyId, KeyFamily keyFamily );
	}

	/* ========================= Fields ========================= */

	/** The singleton. */
	private static KeyFamilyURIMapping mapping;

	/** Map to all available {@link Int.ecb.gesmescb.xml.KeyFamilyURIMapping.Function}s. */
	protected static final Map FUNCTIONS = new HashMap();
	static {
		FUNCTIONS.put( "date", new VersionDateFunction() );
		FUNCTIONS.put( "kfid", new KeyFamilyIdFunction() );
		FUNCTIONS.put( "kfidv", new KeyFamilyIdExtractFunction( true ) );
		FUNCTIONS.put( "kfidx", new KeyFamilyIdExtractFunction( false ) );
	}

	/** Takes care of logging and preferences. */
	protected final ComponentHelper componentHelper;

	/* ====================== Constructors ====================== */

	/** Construct a <code>KeyFamilyURIMapping</code>. */
	protected KeyFamilyURIMapping () {
		componentHelper = new ComponentHelper( "", this );
	}

	/* ===================== Public Methods ===================== */

	/** Get the class' singleton. */
	public static synchronized KeyFamilyURIMapping getInstance () {
		if ( mapping == null ) {
			mapping = new KeyFamilyURIMapping();
		}
		return mapping;
	}

	/** Compute the <code>xmlns</code> for a Key Family.
	 * This uses preferences to obtain a clue on the namespace uri. */
	public String xmlns (
		final String maintenanceAgencyId,
		final KeyFamily keyFamily
	) {
		final String maPreferenceDeprecated = "uri." + maintenanceAgencyId;
		final String kfPreferenceDeprecated = maPreferenceDeprecated + "." + keyFamily.getId();

		String kfURI = componentHelper.getPreference(
			kfPreferenceDeprecated, null
		);
		if ( kfURI != null ) {
			componentHelper.logger.log(
				Level.WARNING,
				componentHelper.logName,
				"xmlns()",
				"Deprecated preference \"" + kfPreferenceDeprecated
				+ "\" should be replaced by a namespace preference."
			);
			return kfURI;
		} else {
			kfURI = componentHelper.getPreference(
				maPreferenceDeprecated, null
			);
		}
		if ( kfURI != null ) {
			componentHelper.logger.log(
				Level.WARNING,
				componentHelper.logName,
				"xmlns()",
				"Deprecated preference \"" + maPreferenceDeprecated
				+ "\" should be replaced by a namespace preference."
			);
			while ( kfURI.endsWith( "/" ) || kfURI.endsWith( "\\" ) ) {
				kfURI = kfURI.substring( 0, kfURI.length() - 1 );
			}
			{
				final int lastPortion = kfURI.lastIndexOf( '/' );
				boolean hasVersionDate = false;
				if ( lastPortion >= 0 ) {
					final int length = kfURI.length();
					for ( int i = lastPortion + 1; i < length; ++i ) {
						hasVersionDate = true;
						char c = kfURI.charAt( i );
						if ( ! (
							Character.isDigit( c ) || c == '-' || c == 'T'
						) ) {
							hasVersionDate = false;
							break;
						}
					}
				}
				if ( !hasVersionDate ) {
					kfURI += "/date()";
				}
			}
			kfURI += "/kfidx()";
			return expand( maintenanceAgencyId, keyFamily, kfURI );
		} else {
		    final boolean useSDMXNamingConventions 
		    = componentHelper.getPreference(
		       "useSDMXNamingConventions", "false"
		    ).equalsIgnoreCase( "true" );
		    if (useSDMXNamingConventions) {
		        return "urn:sdmx:org.sdmx.infomodel.keyfamily.KeyFamily=" +
		            maintenanceAgencyId + ":" + keyFamily.getId();
		    } else {
    			return expand(
    				maintenanceAgencyId,
    				keyFamily,
    				getSetting(
    					maintenanceAgencyId, keyFamily, "namespace", Level.SEVERE
    				)
    			);
		    }
		}
	}

	/**
	 * Query the <code>dtd</code> URI for a Key Family
	 * from preferences.
	 */
	public String dtd (
		final String maintenanceAgencyId,
		final KeyFamily keyFamily
	) {
		return expand(
			maintenanceAgencyId,
			keyFamily,
			getSetting(
				maintenanceAgencyId, keyFamily, "dtd", Level.FINE
			)
		);
	}

	/**
	 * Query the <code>schemaLocation</code> URI for a Key Family
	 * from preferences.
	 */
	public String schemaLocation (
		final String maintenanceAgencyId,
		final KeyFamily keyFamily
	) {
		return expand(
			maintenanceAgencyId,
			keyFamily,
			getSetting(
				maintenanceAgencyId, keyFamily, "schema", Level.FINE
			)
		);
	}

	/**
	 * Query the <code>stylesheet</code> URI for a Key Family
	 * from preferences.
	 */
	public String stylesheet (
		final String maintenanceAgencyId,
		final KeyFamily keyFamily
	) {
		return expand(
			maintenanceAgencyId,
			keyFamily,
			getSetting(
				maintenanceAgencyId, keyFamily, "stylesheet", Level.FINE
			)
		);
	}

	/**
	 * Look up the maintenance agency and  key family
	 * for a given <code>xmlns</code>.
	 * All matching combinations from the {@link Metabase}
	 * are reported as {@link LookupResult} pairs.
	 */
	public Iterator lookup (
		final String xmlns
	) {
		return new LookupIterator( this, xmlns );
	}

	/* =================== Private Functions ==================== */

	/**
	 * Get a preference setting via key family or maintenance agency.
	 *
	 * @param ifFails level to be logged if no appropriate
	 * setting can be looked up.
	 */
	private String getSetting (
		final String maintenanceAgencyId,
		final KeyFamily keyFamily,
		final String setting,
		final Level ifFails
	) {
		final String kfPreference
		= maintenanceAgencyId + "." + keyFamily.getId() + "." + setting;
		{
			final String value = componentHelper.getPreference(
				kfPreference, null
			);
			if ( value != null ) {
				componentHelper.logger.log(
					Level.FINE,
					componentHelper.logName,
					"getSetting()",
					"Found preference \"" + kfPreference
					+ "\" value \"" + value + "\"."
				);
				return value;
			}
		}

		final String maPreference
		= maintenanceAgencyId + ".*." + setting;
		{
			final String value = componentHelper.getPreference(
				maPreference, null
			);
			if ( value != null ) {
				componentHelper.logger.log(
					Level.FINE,
					componentHelper.logName,
					"getSetting()",
					"Found preference \"" + maPreference
					+ "\" value \"" + value + "\"."
				);
				return value;
			}
		}

		/*componentHelper.logger.log(
			ifFails,
			componentHelper.logName,
			"getSetting()",
			"Neither preference \"" + kfPreference
			+ "\" nor preference \"" + maPreference + "\" is set."
		);*/
		return null;
	}

	/** Interpret a setting from the preferences according to key family. */
	private String expand (
		final String maintenanceAgencyId,
		final KeyFamily keyFamily,
		final String string
	) {
		if ( string == null ) return null;
		final StringBuffer buffer = new StringBuffer();
		int tokenLength = 0;
		for ( int i = 0; i < string.length(); ++i ) {
			final char c = string.charAt( i );
			if ( Character.isLetter( c ) ) {
				++tokenLength;
			} else {
				final String token = string.substring( i - tokenLength, i );
				tokenLength = 0;
				if ( c == '(' ) {
					if ( string.charAt( ++i ) != ')' ) {
						componentHelper.logger.log(
							Level.SEVERE,
							componentHelper.logName,
							"expand()",
							"Preference setting \"" + string + "\" is invalid at position "
							+ i + "."
						);
					}
					final KeyFamilyURIMapping.Function function
					= (KeyFamilyURIMapping.Function) FUNCTIONS.get( token );
					if ( function != null ) {
						buffer.append(
							function.expand( maintenanceAgencyId, keyFamily )
						);
					} else {
						componentHelper.logger.log(
							Level.SEVERE,
							componentHelper.logName,
							"expand()",
							"Preference setting \"" + string
							+ "\" uses unknown function \""
							+ token + "\" at position " + i + "."
						);
					}
				} else {
					buffer.append( token );
					buffer.append( c );
				}
			}
		}
		final String token = string.substring( string.length() - tokenLength );
		buffer.append( token );
		return buffer.toString();
	}
}

/* ***************************** Class ****************************** */

class VersionDateFunction implements KeyFamilyURIMapping.Function {
	/** XML dateTime format used for version timestamps. */
	protected final XMLDateFormat xmlDateFormat;

	public VersionDateFunction () {
		xmlDateFormat = new XMLDateFormat();
		xmlDateFormat.setFrequency( RegularInterval.DAILY );
	}

	public String expand (
		final String maintenanceAgencyId,
		final KeyFamily keyFamily
	) {
		return xmlDateFormat.format(
			keyFamily.getVersionDate()
		);
	}
}

class KeyFamilyIdFunction implements KeyFamilyURIMapping.Function {
	public KeyFamilyIdFunction () {}

	public String expand (
		final String maintenanceAgencyId,
		final KeyFamily keyFamily
	) {
		return keyFamily.getId();
	}
}

class KeyFamilyIdExtractFunction implements KeyFamilyURIMapping.Function {
	private final boolean versioning;

	public KeyFamilyIdExtractFunction ( final boolean versioning ) {
		this.versioning = versioning;
	}

	public String expand (
		final String maintenanceAgencyId,
		final KeyFamily keyFamily
	) {
		final String kfId = keyFamily.getId().toLowerCase();
		int idStart =
			kfId.startsWith(
				maintenanceAgencyId.toLowerCase()
			)
			? maintenanceAgencyId.length()
			: 0;
		while (
			idStart < kfId.length()
			&&
			0 <= "_-".indexOf( kfId.charAt( idStart ) )
		) {
			++idStart;
		}
		int verStart = kfId.length();
		while (
			0 <= --verStart
			&&
			Character.isDigit( kfId.charAt( verStart ) )
		);
		int idEnd = ++verStart;
		while (
			idStart <= --idEnd
			&&
			0 <= "_-".indexOf( kfId.charAt( idEnd ) )
		);
		final String idx = kfId.substring( idStart, ++idEnd );
		return
			versioning && verStart != kfId.length()
			? idx + "/" + kfId.substring( verStart )
			: idx;
	}
}

/* ***************************** Class ****************************** */

final class LookupIterator implements Iterator {
	final KeyFamilyURIMapping mapping;
	final String xmlns;
	final Iterator maintenanceAgencies;
	MaintenanceAgency maintenanceAgency;
	Iterator keyFamilies;
	KeyFamily keyFamily;

	public LookupIterator (
		final KeyFamilyURIMapping mapping,
		final String xmlns
	) {
		this.mapping = mapping;
		this.xmlns = xmlns;
		final Metabase metabase = Metabase.getMetabase();
		maintenanceAgencies = metabase.maintenanceAgencyIterator();
		maintenanceAgency = null;
		keyFamilies = null;
		keyFamily = null;
	}

	public boolean hasNext () {
		if ( keyFamily != null ) return true;
		LOOKUP:
		while (
			maintenanceAgencies.hasNext()
			||
			keyFamilies != null && keyFamilies.hasNext()
		) {
			if ( keyFamilies == null ) {
				maintenanceAgency = (MaintenanceAgency) maintenanceAgencies.next();
				keyFamilies = maintenanceAgency.keyFamilyIterator();
			}
			while ( keyFamily == null && keyFamilies.hasNext() ) {
				final KeyFamily kf = (KeyFamily) keyFamilies.next();
				final String namespace
				= mapping.xmlns( maintenanceAgency.getId(), kf );
				if ( xmlns.equals( namespace ) ) {
					keyFamily = kf;
					break LOOKUP;
				}
			}
			keyFamilies = null;
		}
		return keyFamily != null;
	}

	public Object next () {
		if ( hasNext() ) {
			KeyFamilyURIMapping.LookupResult result
			= new ConcreteLookupResult(
				maintenanceAgency.getId(),
				keyFamily
			);
			keyFamily = null;
			return result;
		} else {
			throw new NoSuchElementException( "No more lookup results." );
		}
	}

	public void remove () {
		if ( keyFamilies != null ) {
			keyFamilies.remove();
		} else {
			throw new IllegalStateException(
				"Use remove() only after next()!"
			);
		}
	}
}

/* ***************************** Class ****************************** */

final class ConcreteLookupResult implements KeyFamilyURIMapping.LookupResult {
	final String maintenanceAgencyId;
	final KeyFamily keyFamily;

	public ConcreteLookupResult (
		final String maintenanceAgencyId,
		final KeyFamily keyFamily
	) {
		this.maintenanceAgencyId = maintenanceAgencyId;
		this.keyFamily = keyFamily;
	}

	public String getMaintenanceAgencyId () {
		return maintenanceAgencyId;
	}

	public KeyFamily getKeyFamily () {
		return keyFamily;
	}

	public boolean equals ( Object other ) {
		if ( other instanceof KeyFamilyURIMapping.LookupResult ) {
			KeyFamilyURIMapping.LookupResult that
			= (KeyFamilyURIMapping.LookupResult) other;
			return
				getMaintenanceAgencyId().equals(
					that.getMaintenanceAgencyId()
				)
				&&
				getKeyFamily().equals( that.getKeyFamily() );
		}
		return false;
	}
}
