// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2003, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.xml;

// used but not importable: import org.xml.sax.XMLReader;
import org.xml.sax.Attributes;
import org.xml.sax.ContentHandler;
import org.xml.sax.ErrorHandler;
import org.xml.sax.Locator;
import org.xml.sax.SAXParseException;

import Int.ecb.gesmescb.event.ErrorEvent;
import Int.ecb.gesmescb.event.GESMESCBEventSourceAdapter;
import Int.ecb.gesmescb.util.ComponentHelper;
import Int.ecb.logging.Level;

/* ***************************** Class ****************************** */

/**
 * <p>Builds Gesmes/TS objects from SAX2 parser events.</p>
 *
 * <p>The main idea is to build a delegation hierarchy of
 * <code>SAX2Reader</code>s which parallels the XML syntax
 * element hierarchy.</p>
 *
 * @version @(#)$Id: SAX2Reader.java,v 1.1 2017/01/19 16:14:02 trainot Exp $
 */
public abstract class SAX2Reader
	extends GESMESCBEventSourceAdapter
	implements Constants, ContentHandler, ErrorHandler
{

	/* ========================= Fields ========================= */

	/** Takes care of logging. */
	protected final ComponentHelper componentHelper;

	/** The SAX parser in action. */
	private final org.xml.sax.XMLReader parser;

	/** The parent <code>SAX2Reader</code> in a delegation hierarchy. */
	private final SAX2Reader parent;

	/** Difference opened minus closed elements. */
	private int nestingDepth;

	/* ====================== Constructors ====================== */

	/** Construct an instance. */
	public SAX2Reader (
		final org.xml.sax.XMLReader parser,
		final SAX2Reader parent
	) {
		componentHelper = new ComponentHelper( "", this );
		this.parser = parser;
		this.parent = parent;
		nestingDepth = 0;
	}

	/* ===================== Public Methods ===================== */

	/** Reset the internal parsing state. */
	public void reset () {}

	/* --------------------- ContentHandler --------------------- */

	public void startDocument () {
		componentHelper.logger.log(
			Level.FINE,
			componentHelper.logName,
			"startDocument()",
			"XML document starts here."
		);
	}

	public void setDocumentLocator ( final Locator locator ) {
		componentHelper.logger.log(
			Level.FINE,
			componentHelper.logName,
			"setDocumentLocator()",
			"Locator: " + locator
		);
	}

	public void startPrefixMapping (
		final String prefix,
		final String uri
	) {
		componentHelper.logger.log(
			Level.FINER,
			componentHelper.logName,
			"startPrefixMapping()",
			"Prefix: \"" + prefix + "\", uri: \"" + uri + "\"."
		);
	}

	public void startElement (
		final String namespaceURI,
		final String localName,
		final String qName,
		Attributes atts
	) {
		componentHelper.logger.log(
			Level.FINEST,
			componentHelper.logName,
			"startElement()",
			"XML Element: \"" + qName + "\"."
		);
		++nestingDepth;
	}

	public void characters (
		final char[] ch,
		final int start,
		final int length
	) {
		componentHelper.logger.log(
			Level.FINEST,
			componentHelper.logName,
			"characters()",
			"XML pcdata: \"" + new String( ch, start, length ) + "\"."
		);
	}

	public void ignorableWhitespace (
		final char[] ch,
		final int start,
		final int length
	) {
		componentHelper.logger.log(
			Level.FINEST,
			componentHelper.logName,
			"ignorableWhitespace()",
			"XML whitespace: \"" + new String( ch, start, length ) + "\"."
		);
	}

	public void endElement (
		final String namespaceURI,
		final String localName,
		final String qName
	) {
		componentHelper.logger.log(
			Level.FINEST,
			componentHelper.logName,
			"endElement()",
			"XML Element: \"" + qName + "\"."
		);
		--nestingDepth;
	}

	public void processingInstruction (
		final String target,
		final String data
	) {
		componentHelper.logger.log(
			Level.FINEST,
			componentHelper.logName,
			"processingInstruction()",
			"Target: \"" + target + "\", data: \"" + data + "\"."
		);
	}

	public void skippedEntity ( final String entity ) {
		componentHelper.logger.log(
			Level.FINE,
			componentHelper.logName,
			"skippedEntity()",
			"Entity: \"" + entity + "\"."
		);
	}

	public void endPrefixMapping ( final String prefix ) {
		componentHelper.logger.log(
			Level.FINE,
			componentHelper.logName,
			"endPrefixMapping()",
			"Prefix: \"" + prefix + "\"."
		);
	}

	public void endDocument () {
		componentHelper.logger.log(
			Level.FINE,
			componentHelper.logName,
			"endDocument()",
			"XML document ends here."
		);
	}

	/* ---------------------- ErrorHandler ---------------------- */

	public void warning ( final SAXParseException exception ) {
		handleParserError( exception );
	}

	public void error ( final SAXParseException exception ) {
		handleParserError( exception );
	}

	public void fatalError ( final SAXParseException exception ) {
		handleParserError( exception );
	}

	/* =================== Protected Methods ==================== */

	/**
	 * Let delegate process the given element open
	 * and the following SAX2 events.
	 *
	 * The nesting depth will be reduced by one,
	 * because the current element goes to the delegate.
	 */
	protected void delegate (
		final SAX2Reader delegate,
		final String namespaceURI,
		final String localName,
		final String qName,
		Attributes atts
	) {
		--nestingDepth;
		delegate.startElement( namespaceURI, localName, qName, atts );
		delegate( delegate );
	}

	/** Let delegate process the following SAX2 events. */
	protected void delegate ( final SAX2Reader delegate ) {
		parser.setContentHandler( delegate );
		parser.setErrorHandler( delegate );
	}

	/**
	 * Finish activity as a delegate.
	 *
	 * Subsequent SAX2 events to be processed by the parent.
	 */
	protected void finish () {
		parser.setContentHandler( parent );
		parser.setErrorHandler( parent );
	}

	/** Get difference between numbers of opened and closed elements. */
	protected int getNestingDepth () {
		return nestingDepth;
	}

	/* ==================== Package Methods ===================== */

	/**
	 * Check matching element local name.
	 *
	 * @return whether it is o.k.
	 */
	boolean checkLocalName (
		final String expectedElement,
		final String actualElement
	) {
		if ( expectedElement.equals( actualElement ) ) {
			return true;
		} else {
			complain(
				"Expected element \""
				+ expectedElement
				+ "\" but got \""
				+ actualElement + "\".",
				"checkNamespace()",
				null,
				ErrorEvent.ERROR
			);
			return false;
		}
	}

	/**
	 * Check matching namespace URI.
	 *
	 * @return whether it is o.k.
	 */
	boolean checkNamespace (
		final String expectedNamespaceURI,
		final String actualNameSpaceURI
	) {
		if ( expectedNamespaceURI.equals( actualNameSpaceURI ) ) {
			return true;
		} else {
			complain(
				"Expected element in namespace \""
				+ expectedNamespaceURI
				+ "\" but got one in \""
				+ actualNameSpaceURI + "\".",
				"checkNamespace()",
				null,
				ErrorEvent.ERROR
			);
			return false;
		}
	}

	/**
	 * Check that it is only whitespace.
	 *
	 * @return whether it is o.k.
	 */
	boolean checkWhitespace (
		final char[] ch,
		final int start,
		final int length
	) {
		for ( int i = start; i < start + length; ++i ) {
			if ( !Character.isWhitespace( ch[i] ) ) {
				complain(
					"Unexpected non-whitespace character data: \""
					+ new String( ch, start, length ) + "\".",
					"characters()",
					new String( ch, start, length ),
					ErrorEvent.ERROR
				);
				return false;
			}
		}
		return true;
	}

	/** Check that <code>attr</code> is empty. */
	boolean checkNoAttributes ( final Attributes attr ) {
		final int length = attr.getLength();
		if ( length == 0 ) {
			return true;
		} else {
			complain(
				"Unexpected attributes.",
				"checkNoAttributes()",
				null,
				ErrorEvent.ERROR
			);
			return false;
		}
	}

	/**
	 * Logs a message and dispatches an {@link ErrorEvent}.
	 * This is internally used as a shortcut.
	 * Note that the method name logged is not always "complain",
	 * but in slight abuse of the logging API, can be
	 * set via a parameter.
	 *
	 * @param description of the problem.
	 * @param methodName the name of the current method, used for logging.
	 * @param object on which the problem was detected,
	 * typically a {@link Int.ecb.gesmescb.GESMESCBObject}.
	 * @param severity of the offence.
	 */
	void complain (
		final String description,
		final String methodName,
		final Object object,
		final int severity
	) {
		Level logLevel;
		switch ( severity ) {
			case ErrorEvent.INFO:
				logLevel = Level.FINE;
				break;
			case ErrorEvent.WARNING:
				logLevel = Level.WARNING;
				break;
			default:
				// Since only offences should be
				// reported, everything else would be
				// SEVERE (such as ERROR or FAILURE).
				logLevel = Level.SEVERE;
		}
		componentHelper.logger.log(
			logLevel, componentHelper.logName, methodName, description, object
		);
		dispatch( new ErrorEvent( description, severity ) );
	}

	/* ==================== Private Methods ===================== */

	private void handleParserError ( final SAXParseException exception ) {
		complain(
			"XML parsing problem: " + exception.getMessage() + ".",
			"handleParserError()",
			exception,
			ErrorEvent.ERROR
		);
	}
}
