// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2003, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.xml;

import org.xml.sax.Attributes;
import org.xml.sax.SAXException;

import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;

import Int.ecb.gesmescb.event.TestEventListener;

/* ***************************** Class ****************************** */

/**
 * Implements a few tests for the {@link SAX2Reader}.
 *
 * @author
 * <a href="mailto:sis.external@ecb.int">Bernhard Bodenstorfer</a>
 *
 * @version @(#)$Id: SAX2ReaderTest.java,v 1.1 2017/01/19 16:14:02 trainot Exp $
 */
public class SAX2ReaderTest extends TestCase implements TestConstants {

	/* ========================= Fields ========================= */

	/** The fake parser to the tested. */
	protected final TestParser parser;

	/** The parent to the tested {@link #reader}. */
	protected final TestSAX2Reader parent;

	/** The listener to receive test output. */
	protected final TestEventListener listener;

	/* ====================== Constructors ====================== */

	public SAX2ReaderTest ( final String name ) {
		super( name );
		parser = new TestParser();
		parent = new TestSAX2Reader( parser, null );
		listener = new TestEventListener();
	}

	/* ================== Protected Methods ===================== */

	/** Sets up the test fixture. */
	protected void setUp () {
		final SAX2Reader reader = factory( parser, parent );
		reader.addListener( listener, true );
		parser.setContentHandler( reader );
		parser.setErrorHandler( reader );
	}

	/** Tears down the test fixture. */
	protected void tearDown () {
		listener.clearHistory();
	}

	/** Generate the reader to be tested.
	 * Uses {@link #parser} and {@link #parent}.
	 */
	protected SAX2Reader factory (
		final org.xml.sax.XMLReader parser,
		final SAX2Reader parent
	) {
		return new TestSAX2Reader( parser, parent );
	}

	protected SAX2Reader getReader () {
		return (SAX2Reader) parser.getContentHandler();
	}

	
	/* ===================== Public Methods ===================== */

	/**
	 * Assembles and returns a test suite for all
	 * the test methods of this test case.
	 *
	 * @return a non-null test suite
	 */
	public static Test suite () {
		final TestSuite suite = new TestSuite( SAX2ReaderTest.class );
		return suite;
	}

	/**
	 * @param arg the array of argument {@link String}s
	 */
	public static void main ( final String[] arg ) {
		final String[] testCaseName = {
			SAX2ReaderTest.class.getName()
		};
		junit.textui.TestRunner.main( testCaseName );
	}
    
    public void setNamespace ( final String uri, final String prefix ) {
        parser.setNamespace ( uri, prefix );
    }

    public void openElement (
        final String name, final String attributes
    ) throws SAXException {
        parser.openElement( name, attributes );
    }

    public void characters ( final String string ) throws SAXException {
        parser.characters( string );
    }

    public void closeElement ( final String name ) throws SAXException {
        parser.closeElement( name );
    }

    public void writeElement (
        final String name, final String value
    ) throws SAXException {
        parser.writeElement( name, value );
    }

	/* ---------------------- Test Methods ---------------------- */

	public void testCheckLocalName () {
		assertTrue( "Cube=Cube",
			getReader().checkLocalName(
				"Cube", "Cube"
			)
		);
		assertTrue( "Cube!=Kebap",
			! getReader().checkLocalName(
				"Cube", "Kebap"
			)
		);
	}

	public void testCheckNamespace () {
		assertTrue( "bop=bop",
			getReader().checkNamespace(
				"bop", "bop"
			)
		);
		assertTrue( "bop!=bsi",
			! getReader().checkNamespace(
				"bop", "bsi"
			)
		);
	}

	public void testCheckNoAttributes () {
		final Attributes noAttributes = new TestAttributes(
			new String[]{}, new String[]{}
		);
		final Attributes someAttributes = new TestAttributes(
			new String[]{ "OBS_COM" }, new String[]{ "No comment" }
		);
		assertTrue( "bop=bop",
			getReader().checkNoAttributes( noAttributes )
		);
		assertTrue( "bop!=bsi",
			! getReader().checkNoAttributes( someAttributes )
		);
	}

	public void testNestingDepth () throws SAXException {
		assertEquals( "0", 0, getReader().getNestingDepth() );
		openElement( "element1", "id=43.4" );
		assertEquals( "1", 1, getReader().getNestingDepth() );
		openElement( "element2", "id=43.4" );
		assertEquals( "2", 2, getReader().getNestingDepth() );
		closeElement( "element2" );
		assertEquals( "3", 1, getReader().getNestingDepth() );
		closeElement( "element1" );
		assertEquals( "4", 0, getReader().getNestingDepth() );
	}

	public void testDelegation () {
		final SAX2Reader reader = getReader();
		final SAX2Reader child = new TestSAX2Reader( parser, reader );
		reader.delegate( child );
		assertSame( "child", child, getReader() );
		child.finish();
		assertSame( "reader", reader, getReader() );
		reader.finish();
		assertSame( "parent", parent, getReader() );
	}
}

/* ***************************** Class ****************************** */

/** Concrete, testable subclass. */
class TestSAX2Reader extends SAX2Reader {
	public TestSAX2Reader (
		final org.xml.sax.XMLReader parser,
		final SAX2Reader parent
	) {
		super( parser, parent );
	}
}
