// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2003, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.xml;

import java.util.Iterator;
import java.util.Date;
import java.util.Calendar;
import java.util.List;

import org.xml.sax.SAXException;

import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;

import Int.ecb.gesmescb.*;

/* ***************************** Class ****************************** */

/**
 * Implements a few tests for the {@link StructureReader}.
 *
 * @author
 * <a href="mailto:Xavier.Sosnowska@ECB.europa.eu">Xavier Sosnovsky</a>
 *
 * @author
 * <a href="mailto:Francois.Hoeltzli@ECB.europa.eu">Francois Hoeltzli</a>
 *
 * @version @(#)$Id: StructureReaderTest.java,v 1.1 2017/01/19 16:14:02 trainot Exp $
 */
public class StructureReaderTest extends SAX2ReaderTest {

	/* ======================= Constants ======================== */

	private static final Date VERSION_DATE;
	static {
		final Calendar calendar = Calendar.getInstance();
		calendar.clear();
		calendar.set( 2003, Calendar.NOVEMBER, 10, 15, 22 );
		VERSION_DATE = calendar.getTime();
	}

	private static final KeyFamilyComponent
		FREQ = new KeyFamilyComponent(
			"FREQ", KeyFamilyComponent.KEY_STRUCTURE_DIMENSION,
			"1", "AN..1", "", "", ""
		),
		ITEM = new KeyFamilyComponent(
			"ITEM", KeyFamilyComponent.KEY_STRUCTURE_DIMENSION,
			"2", "AN1", "", "", ""
		),
		TIME_PERIOD = new KeyFamilyComponent(
			"TIME_PERIOD", KeyFamilyComponent.TIME_DIMENSION,
			"4", "AN..35", "", "", ""
		),
		TIME_FORMAT = new KeyFamilyComponent(
			"TIME_FORMAT", KeyFamilyComponent.TIME_DIMENSION,
			"5", "AN3", "", "", ""
		),
		OBS_VALUE = new KeyFamilyComponent(
			"OBS_VALUE", KeyFamilyComponent.ARRAY_CELL,
			"6", "AN..15", "", "", ""
		),
		OBS_STATUS = new KeyFamilyComponent(
			"OBS_STATUS", KeyFamilyComponent.ARRAY_CELL,
			"7", "AN1", KeyFamilyComponent.MANDATORY,
			KeyFamilyComponent.ATTACHED_TO_OBSERVATION, ""
		),
		OBS_CONF = new KeyFamilyComponent(
			"OBS_CONF", KeyFamilyComponent.ARRAY_CELL,
			"8", "AN1", KeyFamilyComponent.CONDITIONAL,
			KeyFamilyComponent.ATTACHED_TO_OBSERVATION, ""
		),
		OBS_PRE_BREAK = new KeyFamilyComponent(
			"OBS_PRE_BREAK", KeyFamilyComponent.ARRAY_CELL,
			"9", "AN..15", KeyFamilyComponent.CONDITIONAL,
			KeyFamilyComponent.ATTACHED_TO_OBSERVATION, ""
		),
		OBS_ATT = new KeyFamilyComponent(
			"OBS_ATT", KeyFamilyComponent.ATTRIBUTE,
			"", "AN..20", KeyFamilyComponent.CONDITIONAL,
			KeyFamilyComponent.ATTACHED_TO_OBSERVATION, ""
		),
		TS_ATT = new KeyFamilyComponent(
			"TS_ATT", KeyFamilyComponent.ATTRIBUTE,
			"", "A2", KeyFamilyComponent.MANDATORY,
			KeyFamilyComponent.ATTACHED_TO_TIME_SERIES, ""
		),
		SG_ATT = new KeyFamilyComponent(
			"SG_ATT", KeyFamilyComponent.ATTRIBUTE,
			"", "N..10", KeyFamilyComponent.CONDITIONAL,
			KeyFamilyComponent.ATTACHED_TO_SIBLING_GROUP, ""
		),
		DS_ATT = new KeyFamilyComponent(
			"DS_ATT", KeyFamilyComponent.ATTRIBUTE,
			"", "A..10", KeyFamilyComponent.MANDATORY,
			KeyFamilyComponent.ATTACHED_TO_DATA_SET, ""
		);

	private static final KeyFamily KF = new KeyFamily(
		"KF", "test KF", "", VERSION_DATE,
		new KeyFamilyComponent[]{
			FREQ, ITEM, TIME_PERIOD, TIME_FORMAT,
			OBS_VALUE, OBS_STATUS, OBS_CONF, OBS_PRE_BREAK,
			OBS_ATT, TS_ATT, SG_ATT, DS_ATT
		}
	);

	/* ====================== Constructors ====================== */

	public StructureReaderTest ( final String name ) {
		super( name );
	}

	/* ================== Protected Methods ===================== */

	/** Tears down the test fixture. */
	protected void tearDown () {
		super.tearDown();
		System.getProperties().remove(
			CubeReader.class.getName() + ".fakeTimeForEmptySeries"
		);
	}

	protected SAX2Reader factory (
		final org.xml.sax.XMLReader parser,
		final SAX2Reader parent
	) {
		final CubeReader cubeReader = new CubeReader( parser, parent );
		cubeReader.setKeyFamily( KF );
		return cubeReader;
	}

	protected void openCube ( final String string ) throws SAXException {
		openElement( "Cube", string );
	}

	protected void closeCube () throws SAXException {
		closeElement( "Cube" );
	}

	/* ===================== Public Methods ===================== */

	/**
	 * Assembles and returns a test suite for all
	 * the test methods of this test case.
	 *
	 * @return a non-null test suite
	 */
	public static Test suite () {
		final TestSuite suite = new TestSuite( StructureReaderTest.class );
		return suite;
	}

	/**
	 * @param arg the array of argument {@link String}s
	 */
	public static void main ( final String[] arg ) {
		final String[] testCaseName = {
			StructureReaderTest.class.getName()
		};
		junit.textui.TestRunner.main( testCaseName );
	}

	/* ---------------------- Test Methods ---------------------- */

	
	
	// Following the suggestion in the comments on top of the SAXparser code,
	// we should define a StructureReader as something that delegates to some
	// component readers.
	//
	// a StructureReader needs at least (from the samples):
	//
	// a StructureHeaderReader
	// a CodeListReader (allowing for Name, Code, Description)
	// a keyFamilyReader
	// a conceptReader
	//
	
	// tests for a StructureHeaderReader
/*	<Header>
	<ID>BIS_01</ID>
	<Test>true</Test>
	<Truncated>false</Truncated>
	<Name xml:lang="en">Trans46301</Name>
	<Prepared>2002-03-11T09:30:47-05:00</Prepared>
	<Sender id="BIS">
		<Name xml:lang="en">Bank for International Settlements</Name>
		<Contact>
			<Name xml:lang="en">G.B. Smith</Name>
			<Telephone>+000.000.0000</Telephone>
		</Contact>
	</Sender>
	<Receiver id="ECB">
		<Name xml:lang="en">European Central Bank</Name>
		<Contact>
			<Name xml:lang="en">B.S. Featherstone</Name>
			<Department xml:lang="en">Statistics Division</Department>
			<Telephone>+000.000.0001</Telephone>
		</Contact>
	</Receiver>
	<Extracted>2002-03-11T09:30:47-05:00</Extracted>
</Header>*/


	public void testStructureHeaderReader () {
		getReader().startElement( "someuri", "Header", "Header", new TestAttributes( "" ) );
		
		getReader().startElement( "someuri", "ID", "ID", new TestAttributes( "" ) );		
		getReader().characters( "BIS_01".toCharArray(), 0, 0 ); //TODO: check characters()
		getReader().endElement( "someuri", "ID", "ID" );
		
		getReader().startElement( "someuri", "Test", "Test", new TestAttributes( "" ) );		
		getReader().characters( "true".toCharArray(), 0, 0 ); 
		getReader().endElement( "someuri", "Test", "Test" );
		
		getReader().startElement( "someuri", "Truncated", "Truncated", new TestAttributes( "" ) );		
		getReader().characters( "false".toCharArray(), 0, 0 ); 
		getReader().endElement( "someuri", "Truncated", "Truncated" );
		
		getReader().startElement( "someuri", "Name", "Name", new TestAttributes( "xml:lang=en" ) );		
		getReader().characters( "Trans46301".toCharArray(), 0, 0 ); 
		getReader().endElement( "someuri", "Name", "Name" );	
		
		getReader().startElement( "someuri", "Prepared", "Prepared", new TestAttributes( "" ) );		
		getReader().characters( "2002-03-11T09:30:47-05:00".toCharArray(), 0, 0 ); 
		getReader().endElement( "someuri", "Prepared", "Prepared" );	
		
		getReader().startElement( "someuri", "Sender", "Sender", new TestAttributes( "id=BIS" ) );		
		getReader().startElement( "someuri", "Name", "Name", new TestAttributes( "xml:lang=en" ) );		
		getReader().characters( "Bank for International Settlements".toCharArray(), 0, 0 ); 
		getReader().endElement( "someuri", "Name", "Name" );		
		getReader().startElement( "someuri", "Contact", "Contact", new TestAttributes( "xml:lang=en" ) );		
		getReader().startElement( "someuri", "Name", "Name", new TestAttributes( "xml:lang=en" ) );		
		getReader().characters( "G.B. Smith".toCharArray(), 0, 0 ); 
		getReader().endElement( "someuri", "Name", "Name" );
		getReader().startElement( "someuri", "Telephone", "Telephone", new TestAttributes( "" ) );		
		getReader().characters( "+000.000.0000".toCharArray(), 0, 0 ); 
		getReader().endElement( "someuri", "Telephone", "Telephone" );		
		getReader().endElement( "someuri", "Contact", "Contact" );
		getReader().endElement( "someuri", "Sender", "Sender" );		
		
		getReader().startElement( "someuri", "Receiver", "Receiver", new TestAttributes( "id=ECB" ) );		
		getReader().startElement( "someuri", "Name", "Name", new TestAttributes( "xml:lang=en" ) );		
		getReader().characters( "European Central Bank".toCharArray(), 0, 0 ); 
		getReader().endElement( "someuri", "Name", "Name" );		
		getReader().startElement( "someuri", "Contact", "Contact", new TestAttributes( "xml:lang=en" ) );		
		getReader().startElement( "someuri", "Name", "Name", new TestAttributes( "xml:lang=en" ) );		
		getReader().characters( "B.S. Featherstone".toCharArray(), 0, 0 ); 
		getReader().endElement( "someuri", "Name", "Name" );
		getReader().startElement( "someuri", "Department", "Department", new TestAttributes( "xml:lang=en" ) );		
		getReader().characters( "Statistics Division".toCharArray(), 0, 0 ); 
		getReader().endElement( "someuri", "Department", "Department" );
		getReader().startElement( "someuri", "Telephone", "Telephone", new TestAttributes( "" ) );		
		getReader().characters( "+000.000.0001".toCharArray(), 0, 0 ); 
		getReader().endElement( "someuri", "Telephone", "Telephone" );		
		getReader().endElement( "someuri", "Contact", "Contact" );
		getReader().endElement( "someuri", "Receiver", "Receiver" );
		
		getReader().startElement( "someuri", "Extracted", "Extracted", new TestAttributes( "" ) );		
		getReader().characters( "2002-03-11T09:30:47-05:00".toCharArray(), 0, 0 ); 
		getReader().endElement( "someuri", "Extracted", "Extracted" );
		
		getReader().endElement( "someuri", "Header", "Header" );
		
		assertEquals( "1 object", 1, listener.size() );
		 /* commented out for further implementation : assertEquals( "Header",
				new Header
				listener.getLastObject();
		) */
		
				
	}
	
	/* commented out for further implementation : 
	// tests for a CodeListReader
	// grmbl, there is a <codelists> tag around the whole crap.. TODO!
	public void testEmptyCodeList () {
		getReader().startElement(namespaceURI, localName, qName, atts)
		getReader().startElement(
			"http://www.SDMX.org/resources/SDMXML/schemas/v2_0/structure",
			"TheExampleCodeList",
			"structure:CodeList",
			new TestAttributes(
				new String[]{ "id", "agency", "version", "uri" },
				new String[]{
					"CL_OBS_STATUS",
					"BIS",
					"1.0",
					"http://www.bis.org/structure/codelists/cl_obs_status.xml"
				}
			)
		);
		getReader().endElement(
			"http://www.SDMX.org/resources/SDMXML/schemas/v2_0/structure",
			"TheExampleCodeList",
			"structure:CodeList"
		);
		assertEquals( "1 object", 1, listener.size() );
		assertEquals(
			"CodeList",
			new CodeList (
				"CL_OBS_STATUS",
				null, //String name,
				null, //String description,
				null, //Date versionDate,
				null, //List codeValues,
				null //List codeDescriptions
			),
			listener.getLastObject()
		);
	}
   */
	public void testExampleCodeList () {
/*		<structure:CodeList id="CL_OBS_STATUS" agencyID="BIS" version="1.0" uri="http://www.bis.org/structure/codelists/cl_obs_status.xml">
			<structure:Name xml:lang="en">Observation Status</structure:Name>
			<structure:Code value="A">
				<structure:Description xml:lang="en">Present</structure:Description>
			</structure:Code>
			<structure:Code value="M">
				<structure:Description xml:lang="en">Missing</structure:Description>
			</structure:Code>
		</structure:CodeList>
*/		
		getReader().startElement(
			"http://www.SDMX.org/resources/SDMXML/schemas/v2_0/structure",
			"CodeLists",
			"structure:CodeLists",
			new TestAttributes( "" )
		);
		getReader().startElement(
			"http://www.SDMX.org/resources/SDMXML/schemas/v2_0/structure",
			"CodeList",
			"structure:CodeList",
			new TestAttributes(
				new String[]{ "id", "agencyID", "version", "uri" },
				new String[]{
					"CL_OBS_STATUS",
					"BIS",
					"1.0",
					"http://www.bis.org/structure/codelists/cl_obs_status.xml"
				}
			)
		);
		getReader().startElement(
			"",
			"nameOfTheCodelist",
			"structure:Name",
				new TestAttributes(
					new String[]{ "xml:lang" },
					new String[]{ "en" }
				)
		);
		getReader().characters( "Observation Status".toCharArray(), 0, 18 );
		getReader().endElement(
			"",
			"nameOfTheCodeList",
			"structure:Name"
		);
/*		<structure:Code value="A">
			<structure:Description xml:lang="en">Present</structure:Description>
		</structure:Code>*/
		getReader().startElement(
			"",
			"Code for present",
			"structure:Code",
			new TestAttributes(	new String[]{ "value" }, new String[]{ "A" } )
		);
		getReader().startElement(
			"",
			"Description for present",
			"structure:Description",
			new TestAttributes(
				new String[]{ "xml:lang" }, new String[]{ "en" }
			)
		);
		getReader().characters( "Present".toCharArray(), 0, 7 );			
		getReader().endElement(
			"",
			"Description for present",
			"structure:Description"
		);
		getReader().endElement(
			"",
			"Code for present",
			"structure:Code"
		);
/*	
<structure:Code value="M">
	<structure:Description xml:lang="en">Missing</structure:Description>
</structure:Code>
*/
		getReader().startElement(
			"",
			"Code for missing",
			"structure:Code",
			new TestAttributes(	new String[]{ "value" }, new String[]{ "M" } )
		);
		getReader().startElement(
			"",
			"Description for missing",
			"structure:Description",
			new TestAttributes(
				new String[]{ "xml:lang" }, new String[]{ "en" }
			)
		);
		getReader().characters( "Missing".toCharArray(), 0, 7 );			
		getReader().endElement(
				"",
				"Description for missing",
				"structure:Description"
		);
		getReader().endElement(
				"",
				"Code for missing",
				"structure:Code"
		);	
		// end the codelist
		getReader().endElement(
			"http://www.SDMX.org/resources/SDMXML/schemas/v2_0/structure",
			"ExampleCodeList",
			"structure:CodeList"
		);		
		getReader().endElement(
			"http://www.SDMX.org/resources/SDMXML/schemas/v2_0/structure",
			"CodeLists",
			"structure:CodeLists"
		);
		// done
		assertEquals( "1 object", 1, listener.size() );
		assertEquals(
			"codelist",
			new CodeList(
				"identifier",
				"name",
				"description",
				new Date(), //versionDate,
				new String[]{ "A", "M" },
				new String[]{ "present", "missing" }
			),
			listener.getLastObject()
		);
	}

	public void testNaNObservation () {
		getReader().startElement(
			"someURI2",
			"AnySeries2",
			"somePrefix:AnySeries2",
			new TestAttributes(
				new String[]{
					"FREQ", "ITEM", "TIME_PERIOD", "TIME_FORMAT",
					"OBS_VALUE", "OBS_STATUS", "OBS_CONF", "OBS_PRE_BREAK"
				},
				new String[]{
					"Q", "ANYITEM", "2002-04", "P3M",
					"NaN", "NaN", "NaN", "NaN"
				}
			)
		);
		getReader().endElement(
			"someURI2",
			"AnySeries2",
			"somePrefix:AnySeries2"
		);
		assertEquals( "1 object", 1, listener.size() );
		assertEquals(
			"series",
			new TimeSeries(
				"Q.ANYITEM", "20022", "608",
				new Observation[]{
					new Observation( "-", "NaN", "NaN", "-" )
				}
			),
			listener.getLastObject()
		);
	}

	public void testMultipleObservations () {
		getReader().startElement(
			"someURI3",
			"AnySeries3",
			"somePrefix:AnySeries3",
			new TestAttributes(
				new String[]{ "FREQ", "ITEM", "TIME_FORMAT" },
				new String[]{ "W", "ANYITEM", "P7D" }
			)
		);
		assertEquals(
			"key",
			"W.ANYITEM",
			( (CubeReader) getReader() ).distilTimeSeriesKey()
		);
		getReader().startElement(
			"someURI3",
			"AnyObs3",
			"somePrefix:AnyObs3",
			new TestAttributes(
				new String[]{ "TIME_PERIOD", "OBS_VALUE", "OBS_STATUS", "OBS_CONF", "OBS_PRE_BREAK" },
				new String[]{ "2003-01-06", "234.5", "A", "", "" }
			)
		);
		getReader().endElement(
			"someURI3",
			"AnyObs3",
			"somePrefix:AnyObs3"
		);
		getReader().startElement(
			"someURI3",
			"AnyObs3",
			"somePrefix:AnyObs3",
			new TestAttributes(
				new String[]{ "TIME_PERIOD", "OBS_VALUE", "OBS_STATUS", "OBS_CONF", "OBS_PRE_BREAK" },
				new String[]{ "2003-01-13", "NaN", "M", "", "" }
			)
		);
		getReader().endElement(
			"someURI3",
			"AnyObs3",
			"somePrefix:AnyObs3"
		);
		getReader().endElement(
			"someURI3",
			"AnySeries3",
			"somePrefix:AnySeries3"
		);
		final Iterator iterator = listener.iterator();
		assertTrue( "zerot", iterator.hasNext() );
		assertEquals(
			"series0",
			new TimeSeries(
				"W.ANYITEM", "", "",
				new Observation[]{}
			),
			iterator.next()
		);
		assertTrue( "first", iterator.hasNext() );
		assertEquals(
			"series1",
			new TimeSeries(
				"W.ANYITEM", "200302", "616",
				new Observation[]{
					new Observation( "234.5", "A", "", "" )
				}
			),
			iterator.next()
		);
		assertTrue( "second", iterator.hasNext() );
		assertEquals(
			"series2",
			new TimeSeries(
				"W.ANYITEM", "200303", "616",
				new Observation[]{
					new Observation( "-", "M", "", "" )
				}
			),
			iterator.next()
		);
		assertTrue( "no more", !iterator.hasNext() );
	}

	public void testDataSetInheritance () throws SAXException {
		openElement(
			"DataSet",
			"ITEM=GOODS OBS_STATUS=A OBS_CONF=F OBS_PRE_BREAK=NaN"
		);
		openElement(
			"Series",
			"FREQ=A TIME_FORMAT=P1Y OBS_CONF=C"
		);
		openElement(
			"Obs",
			"TIME_PERIOD=2000 OBS_VALUE=13.6"
		);
		closeElement( "Obs" );
		closeElement( "Series" );
		closeElement( "DataSet" );

		final Iterator iterator = listener.iterator();
		assertTrue( "first", iterator.hasNext() );
		assertEquals(
			"series1",
			new TimeSeries(
				"A.GOODS", "", "",
				new Observation[]{}
			),
			iterator.next()
		);
		assertTrue( "second", iterator.hasNext() );
		assertEquals(
			"series2",
			new TimeSeries(
				"A.GOODS", "2000", "602",
				new Observation[]{
					new Observation( "13.6", "A", "C", "-" )
				}
			),
			iterator.next()
		);
		assertTrue( "no more", !iterator.hasNext() );
	}

	public void testNoFakePeriod () throws SAXException {
		openElement(
			"Series",
			"FREQ=A ITEM=FOODS TIME_FORMAT=P1Y"
		);
		closeElement( "Series" );

		final Iterator iterator = listener.iterator();
		assertTrue( "first", iterator.hasNext() );
		assertEquals(
			"series1",
			new TimeSeries(
				"A.FOODS", "", "", new Observation[]{}
			),
			iterator.next()
		);
		assertTrue( "no more", !iterator.hasNext() );
	}

	public void testFakePeriod () throws SAXException {
		System.setProperty(
			CubeReader.class.getName() + ".fakeTimeForEmptySeries",
			"true"
		);
		setUp();	// create another test arrangement
		openElement(
			"Series",
			"FREQ=A ITEM=FOODS TIME_FORMAT=P1Y"
		);
		closeElement( "Series" );

		final Iterator iterator = listener.iterator();
		assertTrue( "first", iterator.hasNext() );
		assertEquals(
			"series1",
			new TimeSeries(
				"A.FOODS", "20012000", "702", new Observation[]{}
			),
			iterator.next()
		);
		assertTrue( "no more", !iterator.hasNext() );
	}

	public void testSuppressFakePeriodIfObservations () throws SAXException {
		System.setProperty(
			CubeReader.class.getName() + ".fakeTimeForEmptySeries",
			"true"
		);
		setUp();	// create another test arrangement
		openElement(
			"Obs",
			"FREQ=A ITEM=FOODS TIME_FORMAT=P1Y OBS_VALUE=4"
		);
		closeElement( "Obs" );

		final Iterator iterator = listener.iterator();
		assertTrue( "first", iterator.hasNext() );
		assertEquals(
			"series1",
			new TimeSeries(
				"A.FOODS", "", "", new Observation[]{}
			),
			iterator.next()
		);
		assertTrue( "no more", !iterator.hasNext() );
	}
}
