// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2003, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.xml;

import java.util.Date;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.text.ParseException;

import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;

import Int.ecb.gesmescb.time.RegularInterval;

/* ***************************** Class ****************************** */

/**
 * Implements a few tests for the {@link XMLDateFormat}.
 *
 * @author
 * <a href="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</a>
 *
 * @author
 * <a href="mailto:Sami.Airo@ECB.int">Sami Airo</a>
 *
 * @version @(#)$Id: XMLDateFormatTest.java,v 1.1 2017/01/19 16:14:02 trainot Exp $
 */
public class XMLDateFormatTest extends TestCase {

	/* ====================== Constructors ====================== */

	public XMLDateFormatTest( final String name ) {
		super( name );
	}

	/* ================== Protected Methods ===================== */
		
	/* ===================== Public Methods ===================== */

	/**
	 * Assembles and returns a test suite for all
	 * the test methods of this test case.
	 *
	 * @return a non-null test suite
	 */
	public static Test suite () {
		final TestSuite suite = new TestSuite( XMLDateFormatTest.class );
		return suite;
	}

	/**
	 * @param arg the array of argument {@link String}s
	 */
	public static void main ( final String[] arg ) {
		final String[] testCaseName = {
			XMLDateFormatTest.class.getName()
		};
		junit.textui.TestRunner.main( testCaseName );
	}

	/* ---------------------- Test Methods ---------------------- */

	public void testFormat () throws ParseException {
		final XMLDateFormat format = new XMLDateFormat();
		format.setVersion("not the old sdmx, the new one....");
		
		final Date annualDate
		= new SimpleDateFormat( "yyyyMMdd" ).parse( "20020101");
		format.setFrequency( RegularInterval.ANNUAL );
		assertEquals( "annual", "2002", format.format( annualDate ) );
		format.setFrequency( RegularInterval.WEEKLY );
		assertEquals( "weekly", "2002-01-01", format.format( annualDate ) );
		
		final Date otherDate
		= new SimpleDateFormat( "yyyyMMdd" ).parse( "20020701");
		format.setFrequency( RegularInterval.SEMIANNUAL );
		assertEquals( "semiannual", "2002-B2", format.format( otherDate ) );
		format.setFrequency( RegularInterval.QUARTERLY );
		assertEquals( "quarterly", "2002-Q3", format.format( otherDate ) );
		format.setFrequency( RegularInterval.MONTHLY );
		assertEquals( "monthly", "2002-07", format.format( otherDate ) );
		format.setFrequency( RegularInterval.DAILY );
		assertEquals( "daily", "2002-07-01", format.format( otherDate ) );
		format.setFrequency( RegularInterval.MINUTELY );
		assertEquals( "minutely", "2002-07-01T00:00", format.format( otherDate ) );
		format.applyPattern( XMLDateFormat.SECONDLY_PATTERN );
		assertEquals(
			"secondly", "2002-07-01T00:00:00", format.format( otherDate )
		);
	}

	public void testFormatWOldDates () throws ParseException {
		final XMLDateFormat xmlformat = new XMLDateFormat();
		final String preparedString = "1833-01-13T19:34:00";
		final Date oldDate = xmlformat.parse(preparedString);
		//assertEquals(
	//		"olddate",
	//		preparedString,
	//		new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss").format( oldDate )
	//	);
		final DateFormat format
		= new SimpleDateFormat( "yyyy-MM-dd'T'HH:mm:ss" );
		assertEquals ( "Kedel!",
			format.parse( preparedString ),
			oldDate
		);

	}

	public void testParse () throws ParseException {
		final XMLDateFormat format = new XMLDateFormat();
		format.setVersion("sdmx");
		final Date date = new SimpleDateFormat( "yyyyMMdd" ).parse(
			"20020101"
		);
		assertEquals( "annual", date, format.parse( "2002" ) );
		assertEquals( "monthly", date, format.parse( "2002-01" ) );
		assertEquals( "daily", date, format.parse( "2002-01-01" ) );
		assertEquals( "minutely", date, format.parse( "2002-01-01T00:00" ) );
		assertEquals( "secondly", date, format.parse( "2002-01-01T00:00:00" ) );
	}

	public void testFormatDuration () throws ParseException {
		assertEquals(
			"annual",
			"P1Y",
			XMLDateFormat.formatDuration( RegularInterval.ANNUAL )
		);
		assertEquals(
			"semiannual",
			"P6M",
			XMLDateFormat.formatDuration( RegularInterval.SEMIANNUAL )
		);
		assertEquals(
			"quarterly",
			"P3M",
			XMLDateFormat.formatDuration( RegularInterval.QUARTERLY )
		);
		assertEquals(
			"monthly",
			"P1M",
			XMLDateFormat.formatDuration( RegularInterval.MONTHLY )
		);
		assertEquals(
			"weekly",
			"P7D",
			XMLDateFormat.formatDuration( RegularInterval.WEEKLY )
		);
		assertEquals(
			"daily",
			"P1D",
			XMLDateFormat.formatDuration( RegularInterval.DAILY )
		);
		assertEquals(
			"minutely",
			"PT1M",
			XMLDateFormat.formatDuration( RegularInterval.MINUTELY )
		);
	}

	public void testParseDuration () throws ParseException {
		assertEquals(
			"annual",
			RegularInterval.ANNUAL,
			XMLDateFormat.parseDuration( "P1Y" )
		);
		assertEquals(
			"semiannual",
			RegularInterval.SEMIANNUAL,
			XMLDateFormat.parseDuration( "P6M" )
		);
		assertEquals(
			"quarterly",
			RegularInterval.QUARTERLY,
			XMLDateFormat.parseDuration( "P3M" )
		);
		assertEquals(
			"monthly",
			RegularInterval.MONTHLY,
			XMLDateFormat.parseDuration( "P1M" )
		);
		assertEquals(
			"weekly",
			RegularInterval.WEEKLY,
			XMLDateFormat.parseDuration( "P7D" )
		);
		assertEquals(
			"daily",
			RegularInterval.DAILY,
			XMLDateFormat.parseDuration( "P1D" )
		);
		assertEquals(
			"minutely",
			RegularInterval.MINUTELY,
			XMLDateFormat.parseDuration( "PT1M" )
		);
	}

	public void testExoticFormats () throws ParseException {
		
		final XMLDateFormat format = new XMLDateFormat();
		final SimpleDateFormat sdf = new SimpleDateFormat("yyyyMMdd");

		//do we want weeks to start on sundays? or mondays? done here with sundays.

		assertEquals( "weekly", sdf.parse("20121230"), format.parse( "2013-W1" ) );
		assertEquals( "weekly", sdf.parse("20130106"), format.parse( "2013-W2" ) );
		assertEquals( "weekly", sdf.parse("20130106"), format.parse( "2013-W02" ) );
		assertEquals( "weekly", sdf.parse("20131222"), format.parse( "2013-W52" ) );
		assertEquals( "weekly", sdf.parse("20131229"), format.parse( "2013-W53" ) );
		assertEquals( "weekly", sdf.parse("20091227"), format.parse( "2009-W53" ) );
		assertEquals( "q1", sdf.parse("20130101"), format.parse( "2013-Q1" ) );
		assertEquals( "q2", sdf.parse("20130401"), format.parse( "2013-Q2" ) );
		assertEquals( "q3", sdf.parse("20130701"), format.parse( "2013-Q3" ) );
		assertEquals( "q4", sdf.parse("20131001"), format.parse( "2013-Q4" ) );
		assertEquals( "s1", sdf.parse("20130101"), format.parse( "2013-S1" ) );
		assertEquals( "s2", sdf.parse("20130701"), format.parse( "2013-S2" ) );
		assertEquals( "b1", sdf.parse("20130101"), format.parse( "2013-B1" ) );
		assertEquals( "b2", sdf.parse("20130701"), format.parse( "2013-B2" ) );
		try {
			format.parse( "2013-K2" );
			assertEquals( "2013-K2 should throw an exception", 0, 1);
		}
		catch (ParseException p) { assertEquals ( "ok", 1, 1); }
		catch (Exception e) { assertEquals ( "wrong Exception ...", 0, 1); };
		try {
			format.parse( "2013-B9" );
			assertEquals( "2013-B9 should throw an exception", 0, 1);
		}
		catch (ParseException p) { assertEquals ( "ok", 1, 1); }
		catch (Exception e) { assertEquals ( "wrong Exception ...", 0, 1); };
		try {
			format.parse( "2013-S3" );
			assertEquals( "2013-S3 should throw an exception", 0, 1);
		}
		catch (ParseException p) { assertEquals ( "ok", 1, 1); }
			
	}	
}
