// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2003, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.xml;

import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.IOException;
import java.io.UnsupportedEncodingException;

/* ***************************** Class ****************************** */

/**
 * <p>Writes the xml magic line and a character stream to a byte stream.</p>
 * <p>The XML magic line is something like
 * <tt>&lt;?xml version="1.0" encoding="UTF-8"?&gt;</tt>.
 * As the encoding appears in the line, it is linked to the actual
 * character translation.</p>
 *
 * @author
 * <A HREF="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</A>
 *
 * @version @(#)$Id: XMLOutputStreamWriter.java,v 1.1 2017/01/19 16:14:02 trainot Exp $
 */
class XMLOutputStreamWriter
	extends OutputStreamWriter
{
	/* ======================= Constants ======================== */

	/**
	 * The string used for line termination in order to split the
	 * output nicely into lines.
	 */
	private static final String LINE_BREAK
	= System.getProperty( "line.separator", "\n" );

	/* ========================= Fields ========================= */

	/** Character encoding used by {@link #writer}. */
	private final String encoding;

	/** If headline has been written yet. */
	private boolean headLineWritten;

	/* ====================== Constructors ====================== */

	/**
	 * Construct an <code>XMLOutputStreamWriter</code>
	 * to the <code>OutputStream</code> with the default <tt>encoding</tt>.
	 */
	public XMLOutputStreamWriter (
		final OutputStream outputStream
	) throws UnsupportedEncodingException {
		super( outputStream );
		this.encoding = super.getEncoding();
		headLineWritten = false;
	}

	/**
	 * Construct an <code>XMLOutputStreamWriter</code>
	 * to the <code>OutputStream</code> with the given <tt>encoding</tt>.
	 */
	public XMLOutputStreamWriter (
		final OutputStream outputStream, final String encoding
	) throws UnsupportedEncodingException {
		super( outputStream, encoding );
		this.encoding = encoding;
		headLineWritten = false;
	}

	/* ===================== Public Methods ===================== */

	/** Returns encoding as it was set for the constructor. */
	public String getEncoding () {
		return encoding;
	}

	public void write ( String str ) throws IOException {
		headLine();
		super.write( str );
	}

	public void write ( String str, int off, int len ) throws IOException {
		headLine();
		super.write( str, off, len );
	}

	public void write ( char[] cbuf ) throws IOException {
		headLine();
		super.write( cbuf );
	}

	public void write ( char[] cbuf, int off, int len ) throws IOException {
		headLine();
		super.write( cbuf, off, len );
	}

	public void write ( int c ) throws IOException {
		headLine();
		super.write( c );
	}

	/* ==================== Private Methods ===================== */

	private synchronized void headLine () throws IOException {
		if ( !headLineWritten ) {
			headLineWritten = true;
			write( "<?xml version=\"1.0\" encoding=\"" );
			write( encoding );
			write( "\"?>" );
			write( LINE_BREAK );
			write( LINE_BREAK );
		}
	}
}
