// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2003, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.xml;

import java.io.InputStream;
import java.io.FileInputStream;
import java.io.BufferedInputStream;
import java.io.IOException;

import java.lang.reflect.Constructor;

// used but not importable: import org.xml.sax.XMLReader;
import org.xml.sax.helpers.XMLReaderFactory;
import org.xml.sax.InputSource;

import Int.ecb.logging.Level;

import Int.ecb.gesmescb.event.ErrorEvent;
import Int.ecb.gesmescb.event.GESMESCBReader;
import Int.ecb.gesmescb.event.GESMESCBEventSourceAdapter;
import Int.ecb.gesmescb.event.GESMESCBEventListener;

import Int.ecb.gesmescb.util.Named;
import Int.ecb.gesmescb.util.ComponentHelper;

import Int.ecb.gesmescb.xml.SAX2Reader;

/* ***************************** Class ****************************** */

/**
 * <p>Parses Gesmes/TS-XML
 * and dispatches {@link Int.ecb.gesmescb.GESMESCBObject}s representing the
 * data read.</p>
 *
 * <center><table border="3" cellpadding="3">
 * <caption>
 * <a name="prefs">Available {@link Int.ecb.prefs preference} settings</a>
 * </caption>
 * <tr>
 * <th>preference name</th>
 * <th>type</th>
 * <th>meaning</th>
 * <th>default</th>
 * </tr>
 * <tr><td><code>source</td><td>string</td><td>input file path</td></tr>
 * <tr><td><code>parser</td><td>string</td><td>XML SAX2 parser</td>
 * <td><tt>org.apache.xerces.parsers.SAXParser</tt></td></tr>
 * <tr><td><code>documentReader</td><td>string</td>
 * <td>The {@link SAX2Reader} class to process the document element</td>
 * <td><tt>EnvelopeReader</tt></td></tr>
 * </table></center>
 *
 * @author
 * <a href="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</a>
 *
 * @version @(#)$Id: XMLReader.java,v 1.1 2017/01/19 16:14:02 trainot Exp $
 */
public class XMLReader
	extends GESMESCBEventSourceAdapter
	implements GESMESCBReader, Named
{
	/* ======================= Constants ======================== */

    /** Default parser name. */
    private static final String DEFAULT_PARSER_CLASS
	= "org.apache.xerces.parsers.SAXParser";

	/* ========================= Fields ========================= */

	/** Takes care of logging and preferences. */
	protected final ComponentHelper componentHelper;

	/**
	 * The name of the file read.
	 * This can be used for log messages.
	 */
	private final String filename;

	/** The input XML data. */
	private InputStream input;

	/** The SAX2 parser. */
	private org.xml.sax.XMLReader parser;

	private SAX2Reader documentReader;

	/* ====================== Constructors ====================== */

	/**
	 * Construct an <code>XMLReader</code> to read from a file.
	 *
	 * @param name used for logging and preferences.
	 * @param filename where to read from.
	 *
	 * @throws NullPointerException if <code>filename==null</code>
	 * and the preference variable
	 * <code>source</code> is undefined.
	 * @throws IOException if the file cannot be opened for reading.
	 */
	public XMLReader ( final String name, final String filename )
		throws IOException
	{
		componentHelper = new ComponentHelper( name, this );
		this.filename = filename != null
			? filename
			: componentHelper.getPreference( "source", null );
		if ( this.filename == null ) {
			final String desc
			= "No preference variable \"source\" found"
				+ " to determine the input file.";
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"XMLReader()",
				desc
			);
		}
		try {
			input = new BufferedInputStream(
				new FileInputStream( this.filename )
			);
		}
		catch ( final IOException ioe ) {
			final String desc = "Failed to open input file: " + ioe;
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"XMLReader()",
				desc
			);
			input = null;
		}
		try {
			final String parserClass = componentHelper.getPreference(
				"parser", DEFAULT_PARSER_CLASS
			);

			componentHelper.logger.log(
				Level.FINE,
				componentHelper.logName,
				"XMLReader",
				"Trying to use " + parserClass + " to parse XML."
			);

			parser = XMLReaderFactory.createXMLReader( parserClass );

			componentHelper.logger.log(
				Level.CONFIG,
				componentHelper.logName,
				"XMLReader",
				"Using " + parserClass + " to parse XML."
			);

			final String documentReaderClass = componentHelper.getPreference(
				"documentReader", EnvelopeReader.class.getName()
			);

			componentHelper.logger.log(
				Level.FINE,
				componentHelper.logName,
				"XMLReader",
				"Trying to use " + documentReaderClass + " to read XML."
			);

			final Class valueClass = Class.forName( documentReaderClass );
			final Constructor constructor = valueClass.getConstructor(
				new Class[]{ org.xml.sax.XMLReader.class, SAX2Reader.class }
			);
			documentReader = (SAX2Reader) constructor.newInstance(
				new Object[]{ parser, null }
			);

			componentHelper.logger.log(
				Level.CONFIG,
				componentHelper.logName,
				"XMLReader",
				"Using " + documentReaderClass + " to read XML."
			);

			parser.setContentHandler( documentReader );
			parser.setErrorHandler( documentReader );
		}
		catch ( final Exception e ) {
			final String desc = "Failed to instantiate SAX parser: " + e;
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"XMLReader()",
				desc
			);
			parser = null;
		}
	}

	/**
	 * Construct an <code>XMLReader</code>
	 * reading from a file according to preferences.
	 *
	 * @param name used for logging and preferences.
	 *
	 * @throws NullPointerException if the preference variable
	 * <code>source</code> is undefined.
	 * @throws IOException if the file cannot be opened for reading.
	 */
	public XMLReader ( final String name )
		throws IOException
	{
		this( name, null );
	}

	/* ===================== Public Methods ===================== */

	public String getName () {
		return componentHelper.name;
	}

	public void addListener (
		final GESMESCBEventListener listener,
		final boolean isImportant
	) {
		super.addListener( listener, isImportant );
		documentReader.addListener( listener, isImportant );
	}

	public void removeAllListeners () {
		super.removeAllListeners();
		documentReader.removeAllListeners();
	}

	/**
	 * Parse a Gesmes/TS-XML interchange into
	 * {@link Int.ecb.gesmescb.GESMESCBObject}s.
	 * These then are dispatched to all listeners.
	 */
	public void run () {
		try {
			final String origin
			= filename != null ? " from " + filename : "";

			componentHelper.logger.log(
				Level.INFO,
				componentHelper.logName,
				"run()",
				"Starting to parse XML input"
				+ origin + '.'
			);

			parser.parse( new InputSource( input ) );
		}
		catch ( final Exception e ) {
			final String desc = "Problem: " + e.getMessage() + ".";

			componentHelper.logger.log(
				Level.SEVERE,
				getName(),
				"run()",
				desc,
				e
			);
			dispatch(
				new ErrorEvent( desc, ErrorEvent.ERROR )
			);
		}
	}

	/**
	 * Close the <code>XMLReader</code>.
	 */
	public void close () {
		try {
			input.close();

			final String desc = "Closed input from " + filename + ".";
			componentHelper.logger.log(
				Level.INFO,
				getName(),
				"close()",
				desc
			);
			dispatch(
				new ErrorEvent( desc, ErrorEvent.INFO )
			);
		}
		catch ( final Exception e ) {
			final String desc = "Problem: " + e.getMessage() + ".";

			componentHelper.logger.log(
				Level.SEVERE,
				getName(),
				"close()",
				desc,
				e
			);
			dispatch(
				new ErrorEvent( desc, ErrorEvent.ERROR )
			);
		}
	}
}
