// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2003, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.xml;

import java.io.*;

import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;

/* ***************************** Class ****************************** */

/**
 * Implements a few tests for the {@link XMLWriteBuffer}.
 *
 * @author
 * <A HREF="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</A>
 *
 * @version @(#)$Id: XMLWriteBufferTest.java,v 1.1 2017/01/19 16:14:02 trainot Exp $
 */
public class XMLWriteBufferTest extends TestCase {

	/* ======================= Constants ======================== */

	private static final String NEW_LINE
	= System.getProperty( "line.separator" );

	/* ====================== Constructors ====================== */

	public XMLWriteBufferTest( final String name ) {
		super( name );
	}

	/* =================== Protected Methods ==================== */

	protected XMLWriteBuffer factory (
		final Writer writer,
		final char indentChar,
		final int indentStep,
		final String newLine
	) {
		return new XMLWriteBuffer( writer, indentChar, indentStep, newLine );
	}

	/* ===================== Public Methods ===================== */

	/**
	 * Assembles and returns a test suite for all
	 * the test methods of this test case.
	 *
	 * @return a non-null test suite
	 */
	public static Test suite () {
		final TestSuite suite = new TestSuite( XMLWriteBufferTest.class );
		return suite;
	}

	/**
	 * @param arg the array of argument {@link String}s
	 */
	public static void main ( final String[] arg ) {
		final String[] testCaseName = {
			XMLWriteBufferTest.class.getName()
		};
		junit.textui.TestRunner.main( testCaseName );
	}

	/* ---------------------- Test Methods ---------------------- */

	public void testIndent () throws IOException {
		final CharArrayWriter writer = new CharArrayWriter();
		final XMLWriteBuffer buffer = factory(
			writer, '\t', 1, NEW_LINE
		);

		assertEquals( "Initial", 0, buffer.getIndent() );
		buffer.setIndent( 3 );
		assertEquals( "Brute set 3", 3, buffer.getIndent() );
		buffer.setIndent( 1 );
		assertEquals( "Brute set back 1", 1, buffer.getIndent() );
		buffer.openElement( "DataSet" );
		assertEquals( "After DataSet", 2, buffer.getIndent() );
		buffer.writeAttribute( "id", "someID" );
		assertEquals( "In DataSet", 2, buffer.getIndent() );
		buffer.openElement( "Series" );
		assertEquals( "After Series", 3, buffer.getIndent() );
		buffer.closeElement( "Series" );
		assertEquals( "After Series close", 2, buffer.getIndent() );
		buffer.closeElement( "DataSet" );
		assertEquals( "After DataSet close", 1, buffer.getIndent() );
		buffer.close();

		assertEqualsLinewise(
			"\t<DataSet id=\"someID\">" + NEW_LINE
			+ "\t\t<Series/>" + NEW_LINE
			+ "\t</DataSet>" + NEW_LINE,
			writer.toString()
		);
	}

	public void testEscaping () throws IOException {
		final CharArrayWriter writer = new CharArrayWriter();
		final XMLWriteBuffer buffer = factory(
			writer, '\t', 1, NEW_LINE
		);

		buffer.openElement( "DataSet" );
		buffer.writeAttribute( "id", "<>&\"" );
		buffer.writeCharacters( "<>&\"" );
		buffer.closeElement( "DataSet" );
		buffer.close();

		assertEqualsLinewise(
			"<DataSet id=\"&lt;&gt;&amp;&quot;\">&lt;&gt;&amp;\"</DataSet>"
			+ NEW_LINE,
			writer.toString()
		);
	}

	public void testConfig () throws IOException {
		final CharArrayWriter writer = new CharArrayWriter();
		final XMLWriteBuffer buffer = factory(
			writer, 't', 3, "nl"
		);

		buffer.openElement( "DataSet" );
		buffer.writeAttribute( "the", "" );
		buffer.writeCharacters( "2" );
		buffer.openElement( "Series" );
		buffer.closeElement( "Series" );
		buffer.closeElement( "DataSet" );
		buffer.close();

		assertEqualsLinewise(
			"<DataSet the=\"\">2nlttt<Series/>nl</DataSet>nl",
			writer.toString()
		);
	}

	public void testWriteElement () throws IOException {
		final CharArrayWriter writer = new CharArrayWriter();
		final XMLWriteBuffer buffer = factory(
			writer, '\t', 1, NEW_LINE
		);

		buffer.writeElement( "test", "" );
		buffer.writeElement( "prepared", "45" );
		buffer.close();

		assertEqualsLinewise(
			"<test/>" + NEW_LINE
			+ "<prepared>45</prepared>" + NEW_LINE,
			writer.toString()
		);
	}

	/* ==================== Private Methods ===================== */

	void assertEqualsLinewise (
		final String expected, final String actual
	) throws IOException {
		final BufferedReader expectedReader = new BufferedReader(
			new StringReader( expected )
		);
		final BufferedReader actualReader = new BufferedReader(
			new StringReader( actual )
		);

		String expectedLine, actualLine;
		int lineNo = 0;
		do {
			++lineNo;
			expectedLine = expectedReader.readLine();
			actualLine = actualReader.readLine();
			if ( expectedLine != null ) {
				assertEquals( "Line " + lineNo, expectedLine, actualLine );
			} else {
				assertNull( "Line " + lineNo, actualLine );
			}
		} while ( expectedLine != null );

		expectedReader.close();
		actualReader.close();
	}
}
