package Int.ecb.gesmescb.xml.sdmx.v2;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashSet;
import java.util.List;
import java.util.Locale;
import java.util.Set;

import Int.ecb.gesmescb.CodeList;
import Int.ecb.gesmescb.Contact;
import Int.ecb.gesmescb.KeyFamily;
import Int.ecb.gesmescb.KeyFamilyComponent;
import Int.ecb.gesmescb.MessageBegin;
import Int.ecb.gesmescb.StatisticalConcept;
import Int.ecb.gesmescb.event.ErrorEvent;
import Int.ecb.gesmescb.event.GESMESCBEventSourceAdapter;
import Int.ecb.gesmescb.event.GESMESCBReader;
import Int.ecb.gesmescb.util.ComponentHelper;
import Int.ecb.logging.Level;
import eu.ecb.stats.sdmx.api.IDataStructureApi;
import eu.ecb.stats.sdmx.api.ISDMXService;
import eu.ecb.stats.sdmx.api.MetadataQuery;
import eu.ecb.stats.sdmx.model.base.UsageStatus;
import eu.ecb.stats.sdmx.model.base.representation.FacetBasedRepresentation;
import eu.ecb.stats.sdmx.model.base.representation.FacetType;
import eu.ecb.stats.sdmx.model.base.representation.IFacet;
import eu.ecb.stats.sdmx.model.base.representation.IRepresentation;
import eu.ecb.stats.sdmx.model.structure.code.ICode;
import eu.ecb.stats.sdmx.model.structure.code.ICodelist;
import eu.ecb.stats.sdmx.model.structure.concept.ConceptRole;
import eu.ecb.stats.sdmx.model.structure.concept.IConcept;
import eu.ecb.stats.sdmx.model.structure.datastructure.AttachmentLevel;
import eu.ecb.stats.sdmx.model.structure.datastructure.IDataAttribute;
import eu.ecb.stats.sdmx.model.structure.datastructure.IDataStructure;
import eu.ecb.stats.sdmx.model.structure.datastructure.IDimension;
import eu.ecb.stats.sdmx.model.structure.datastructure.IPrimaryMeasure;
import eu.ecb.stats.sdmx.model.structure.datastructure.ITimeDimension;
import eu.ecb.stats.sdmx.service.registry.SdmxRegistry;
import eu.ecb.stats.sdmx.service.sdmxml.SDMXMLService;

/* ***************************** Class ****************************** */

/**
 * This class uses an existing library, written by Xavier Sosnovsky,
 * to read data structure definitions from the ECB registry,
 * translates the output into JavaSuite objects and finally 
 * dispatches them to all listeners.
 * 
 * @author kortest
 *
 */

public class SDMX21StructureReader 
	extends GESMESCBEventSourceAdapter
	implements GESMESCBReader
{	
	/* ========================= Fields ========================= */
	
	private ISDMXService[] serviceInstances;
	/**
	 * The componentHelper that is used, e.g. for logging.
	 */
	private ComponentHelper componentHelper;
	
	/**
	 * The timestamp when the structural definitions were read 
	 * off the registry.
	 */
	private Date versionDate;
	
	/**
	 * SDMXServices acts as a factory that provides access to a specific
	 * data provider, in this case the ECB registry. This is were the data structure
	 * definitions are read off.
	 */
	private String[] registryLocations={
			"https://sdw-ecb-wsrest.ecb.de/service/",
			//"https://registry.sdmx.org/FusionRegistry/ws/rest/"
	};
	
	
	/**
	 * Used when structures are read off of .xml files.
	 */
	private FileInputStream fileInputStream;

	/**
	 * Used when structures are read off of .xml files.
	 */
	private String filename;
	
	/* ====================== Constructors ====================== */
	
	/**
	 * Contructs an SDMX21StructureReader that reads the data structure
	 * definitions from the registry.
	 * 
	 * @param name Used for logging.
	 */
	public SDMX21StructureReader(String name){
		
		this.componentHelper = new ComponentHelper(name, this);
		if (System.getProperty("Int.ecb.prefs.config.file").contains("reception")){
			serviceInstances = new ISDMXService[1];
			serviceInstances[0] = new SdmxRegistry(registryLocations[0]);
			componentHelper.logger.log(
					Level.INFO,
					componentHelper.logName,
					"SDMX21StructureReader(String name)",
					"Reading structures from the ECB registry"
			);
		} else {
			serviceInstances = new ISDMXService[1];
			serviceInstances[0] = new SdmxRegistry(registryLocations[0]);
			componentHelper.logger.log(
					Level.INFO,
					componentHelper.logName,
					"SDMX21StructureReader(String name)",
					"Reading structures from the ESCB registry"
			);
		}

		versionDate = new Date();
	}
	
	/**
	 * Contructs an SDMX21StructureReader that reads data structure
	 * definitions from the a file.
	 * 
	 * @param name Used for logging.
	 * @filename is the filename.
	 */
	public SDMX21StructureReader(String name, String filename) throws FileNotFoundException {
		
		this.componentHelper = new ComponentHelper(name, this);
		componentHelper.logger.log(
				Level.INFO,
				componentHelper.logName,
				"SDMX21StrucutreReader(String name)",
				"Reading SDMX-ML Structures from file: " + filename
		);
		try{
			serviceInstances = new ISDMXService[1];
			this.filename=filename;
			fileInputStream = new FileInputStream(this.filename);
			serviceInstances[0] = new SDMXMLService(fileInputStream);
		} catch (FileNotFoundException fnfe){
			componentHelper.logger.log(
					Level.SEVERE, componentHelper.logName, "SDMX21StructureReader",
					"Opening file failed: " + fnfe.getMessage()
				);
				throw fnfe;
		}
		
		versionDate = new Date();
		
	}
	
	/* ===================== Public Methods ===================== */

	/**
	 * Parse a data structure definiton in SDMX21 
	 * syntax into {@link Int.ecb.gesmescb.GESMESCBObject}s.
	 * These are then dispatched to all listeners.
	 */
	
	public void run () {
		for (ISDMXService sdmxServiceInstance : serviceInstances){
			final IDataStructureApi dsdDao = sdmxServiceInstance.getDataStructureDAO();
			final Set<IDataStructure> dsds = dsdDao.getArtefacts(MetadataQuery.newInstance(null, null, null, null, null));	
			for (IDataStructure dsd : dsds){
				MessageBegin messageBegin = new MessageBegin(
						"MREF000001","","",dsd.getMaintainer().getIdentifier(),"","",new Contact[]{},
						"","","","","",""
				);
				componentHelper.logger.log(
						Level.FINE,
						componentHelper.logName,
						"readMessage()",
						"Dispatch",
						messageBegin
				);
				
				dispatch( messageBegin );
				
				readCodeLists( dsd );
				
				readKeyFamilies( dsd );
				
				readStatisticalConcepts( dsd );
			}
		}
	}
	
	public void close (){
		
		try{
			if (fileInputStream != null && this.filename != null){
				fileInputStream.close();
				componentHelper.logger.log(
						Level.INFO,
						componentHelper.logName,
						"close()",
						"Closed "
						+ " input"
						+ (
							this.filename != null
							? " from " + this.filename
							: ""
							)
						+ '.'
					);
			}
		} catch ( final IOException ioe ) {
			final String desc
			= "I/O error while closing input";
			componentHelper.logger.log(
				Level.SEVERE,
				componentHelper.logName,
				"close()",
				desc,
				ioe
			);
			// Maybe, someone like a Metabase wants to know
			// about this failure.
			dispatch(
				new ErrorEvent( desc, ErrorEvent.FAILURE )
			);
		}
	}
	
	public Date getDate(){
		return versionDate;
	}
	
	/* ==================== Private Methods ===================== */

	/**
	 * Reads the codelists that are used in a specific data
	 * structure definition. 
	 * 
	 * @param dsd The data structure definition.
	 */
	private void readCodeLists ( final IDataStructure dsd ) {
		
		Set<ICodelist> usedLists = new HashSet<>();
		for (final IDimension dim : dsd.getKeyDescriptor()) {
			if (!(dim instanceof ITimeDimension)) {
				final IRepresentation list =
					null != dim.getLocalRepresentation() ? dim.
					getLocalRepresentation() : dim.
					getConceptIdentity().getCoreRepresentation();
				if (list instanceof ICodelist) {
					usedLists.add((ICodelist) list);
	            }
			}
		}
		for (final IDataAttribute att : dsd.getAttributeDescriptor()) {
			final IRepresentation list =
				null != att.getLocalRepresentation() ? att.
				getLocalRepresentation() : att.
				getConceptIdentity().getCoreRepresentation();
			if (list instanceof ICodelist) {
				usedLists.add((ICodelist) list);
			}
		}
		for (final IPrimaryMeasure mea : dsd.getMeasureDescriptor()){
			final IRepresentation list =
				null != mea.getLocalRepresentation() ? mea.
				getLocalRepresentation() : mea.
				getConceptIdentity().getCoreRepresentation();
			if (list instanceof ICodelist) {
				usedLists.add((ICodelist) list);
			}
		}
		
		for (ICodelist cl : usedLists){

			List codeValues = new ArrayList(){}; 
			List descriptions = new ArrayList(){};
			for (ICode c : cl.getChildren()){
				codeValues.add(c.getIdentifier());
				descriptions.add(c.getName().get(Locale.ENGLISH));
			}
			String description = (cl.getDescription().get(Locale.ENGLISH)==null) ? "" : cl.getDescription().get(Locale.ENGLISH);
			CodeList codelist = new CodeList( 
					cl.getIdentifier(),
					cl.getName().get(Locale.ENGLISH),
					description,
					versionDate,
					codeValues,
					descriptions
			);
			
			componentHelper.logger.log(
					Level.FINE,
					componentHelper.logName,
					"readCodeLists()",
					"Dispatch",
					codelist
			); 
			dispatch ( codelist );
		}
	}
	
	/**
	 * Converts the data structure definition into KeyFamily.java
	 * instance. 
	 * 
	 * @param dsd The data structure definition.
	 */
	private void readKeyFamilies ( final IDataStructure dsd ) {
		
		
		// To be put into the KeyFamily.
		ArrayList<KeyFamilyComponent> keyFamilyComponents = new ArrayList<KeyFamilyComponent>();
				
		String 
			conceptId,
			conceptType,
			position,
			format,
			usageStatus,
			attachmentLevel,
			codeListId;
		
		int dimCounter = 1;
		
		for (final IDimension dimension : dsd.getKeyDescriptor()){
			
			conceptId = dimension.getConceptIdentity().getIdentifier();
			if (dimension.getRole() != null && dimension.getRole().equals(ConceptRole.TIME)){
				conceptType = "1";
			} else {
				conceptType = "4";
			}
			position = String.valueOf(dimCounter);
			format = "";
			usageStatus="";
			attachmentLevel = "";
			
			if (dimension.getLocalRepresentation() instanceof ICodelist){
				codeListId = ((ICodelist) dimension.getLocalRepresentation()).
		                   getIdentifier();
			} else {
				codeListId = "";
			}
			
			KeyFamilyComponent keyFamilyComponent = new KeyFamilyComponent(
					conceptId,
					conceptType,
					position,
					format,
					usageStatus,
					attachmentLevel,
					codeListId
			);				
			keyFamilyComponents.add(keyFamilyComponent);
			dimCounter++;
		}

		for (final IDataAttribute attribute : dsd.getAttributeDescriptor()) {
				
			conceptId = attribute.getConceptIdentity().getIdentifier();
			conceptType = "Z09";
			position = "";
			int minLength =0, maxLength =0;
			if ( attribute.getLocalRepresentation() instanceof FacetBasedRepresentation){
				final FacetBasedRepresentation representation = (FacetBasedRepresentation) attribute.getLocalRepresentation();
				for( IFacet facet : representation.getFacets()){
					if ( facet.getType() == FacetType.MIN_LENGTH){
						minLength = facet.getValue();
					} else if( facet.getType() == FacetType.MAX_LENGTH) {
						maxLength = facet.getValue(); 
					}
				}
				if (maxLength == minLength){
					format = "AN"+String.valueOf(maxLength);
				} else {
					format = "AN.."+String.valueOf(maxLength);
				}
			} else {
				format = "";
			}
			
			if (attribute.getUsageStatus().equals(UsageStatus.MANDATORY)){
				usageStatus = "2";
			} else {
				usageStatus = "1";
			} 
			
			AttachmentLevel aL = attribute.getAttachmentLevel();
			
			if ( aL.equals(AttachmentLevel.DATASET)){
				attachmentLevel = "1";
			} else if (aL.equals(AttachmentLevel.SERIES)){
				attachmentLevel = "4";
			} else if (aL.equals(AttachmentLevel.GROUP)){
				attachmentLevel = "9";
			} else if (aL.equals(AttachmentLevel.OBSERVATION)){
				attachmentLevel = "5";
			} else {
				attachmentLevel = "8";
			} 
			codeListId = (attribute.getLocalRepresentation() instanceof FacetBasedRepresentation) ? "" : ((ICodelist)attribute.getLocalRepresentation()).getIdentifier();
			
			KeyFamilyComponent keyFamilyComponent = new KeyFamilyComponent(
				conceptId,
				conceptType,
				position,
				format,
				usageStatus,
				attachmentLevel,
				codeListId
			);
			keyFamilyComponents.add(keyFamilyComponent);
		}
		
		for (final IPrimaryMeasure measure : dsd.getMeasureDescriptor()){
			
			conceptId = measure.getConceptIdentity().getIdentifier();
			conceptType = "3";
			position = "";
			format = "";
			usageStatus = "";
			attachmentLevel = "";
			codeListId = (measure.getLocalRepresentation()==null) ? "":((ICodelist) measure.getLocalRepresentation()).getIdentifier();
			KeyFamilyComponent keyFamilyComponent = new KeyFamilyComponent(
					conceptId,
					conceptType,
					position,
					format,
					usageStatus,
					attachmentLevel,
					codeListId
			);
			keyFamilyComponents.add(keyFamilyComponent);
		}
		
		String description= (dsd.getDescription().get(Locale.ENGLISH)==null)? "" : dsd.getDescription().get(Locale.ENGLISH);
		
		KeyFamily keyFamily = new KeyFamily(
				dsd.getIdentifier(),
				dsd.getDefaultName(),
				description,
				versionDate,
				keyFamilyComponents
		);
		
		componentHelper.logger.log(
				Level.FINE,
				componentHelper.logName,
				"readKeyFamilies()",
				"Dispatch",
				keyFamily
		);
		
		dispatch( keyFamily );
	}
	
	/**
	 * Reads the statistical Concepts that are used in a specific data
	 * structure definition. 
	 * 
	 * @param dsd The data structure definition.
	 */
	private void readStatisticalConcepts ( IDataStructure dsd ){
		
		String description;
		Set<IConcept> usedConcepts = new HashSet<>();
		
		for (final IDimension dim : dsd.getKeyDescriptor()) {
			if (!(dim instanceof ITimeDimension)) {
				usedConcepts.add(dim.getConceptIdentity());
			}
		}
		for (final IDataAttribute att : dsd.getAttributeDescriptor()) {
			usedConcepts.add(att.getConceptIdentity());
		}
		for (final IPrimaryMeasure mea : dsd.getMeasureDescriptor()){
			usedConcepts.add(mea.getConceptIdentity());
		}
		for( final IConcept con : usedConcepts){
			description = (con.getDescription().toString() != "{}") ? "" : con.getDescription().toString();
			StatisticalConcept concept = new StatisticalConcept(
					con.getIdentifier().toString(),
					con.getDefaultName().toString(),
					description,
					versionDate
			);
			
			componentHelper.logger.log(
					Level.FINE,
					componentHelper.logName,
					"readStatisticalConcepts()",
					"Dispatch",
					con
				);
			
			dispatch( concept );
		}
	}
}
