// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2004, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.xml.sdmx.v2;

import Int.ecb.gesmescb.time.RegularInterval;
import Int.ecb.gesmescb.util.ComponentHelper;
import Int.ecb.logging.Level;

/**
 * Guesses the frequency and the appropriate date format out of the supplied 
 * time format. 
 * 
 * @author Francois Hoeltzli
 * @author Xavier Sosnovsky
 */
public class SDMXHelper {
    
    /* ========================= Fields ========================= */

    /** Takes care of logging. */
    protected final ComponentHelper componentHelper;
    
    private int frequency;
    
    private String dateFormat;
    
    /* ====================== Constructors ====================== */

    /**
     * Constructs a new SDMXHelper.
     * 
     * @param timeFormat the time format out of which the frequency and the
     * date format will be guessed.
     */
    public SDMXHelper(final String timeFormat) {
        super();
        componentHelper = new ComponentHelper( "", this );
        initializeFields(timeFormat);
    }
    
    /* ===================== Public Methods ===================== */

    /**
     * @return The frequency guessed out of the time format supplied in the
     * constructor
     */
    public int getFrequency() {
        return frequency;
    }
    
    /**
     * @return The date format guessed out of the time format supplied in the
     * constructor
     */
    public String getDateFormat() {
        return dateFormat;
    }
    
    /* ==================== Private Methods ===================== */
    
    private void initializeFields(final String timeFormat){
        if ("P1D".equals(timeFormat)) {
            dateFormat = "yyyy-MM-dd";
            frequency = RegularInterval.DAILY;
        } else if ("P1M".equals(timeFormat)) {
            dateFormat = "yyyy-MM";
            frequency = RegularInterval.MONTHLY;
        } else if ("P1Y".equals(timeFormat)) {
            dateFormat = "yyyy";
            frequency = RegularInterval.ANNUAL;
        } else if ("P3M".equals(timeFormat)) {
            dateFormat = "yyyy-MM";
            frequency = RegularInterval.QUARTERLY;
        } else if ("P6M".equals(timeFormat)) {
            dateFormat = "yyyy-MM";
            frequency = RegularInterval.SEMIANNUAL;
        } else if ("P7D".equals(timeFormat)) {
            dateFormat = "yyyy-MM-dd";
            frequency = RegularInterval.WEEKLY;
        } else if ("PT1M".equals(timeFormat)) {
            dateFormat = "yyyy-MM-dd'T'HH:mm";
            frequency = RegularInterval.MINUTELY;
        } else if ("".equals(timeFormat)) {
            componentHelper.logger.log(
                    Level.SEVERE,
                    componentHelper.logName,
                    "TimeSeriesReader",
                    "empty TIME_FORMAT in sdmx-ml file."
                );
        } else if (null == timeFormat) {
            componentHelper.logger.log(
                    Level.SEVERE,
                    componentHelper.logName,
                    "TimeSeriesReader",
                    "no TIME_FORMAT found in sdmx-ml file."
                );
        } else {
            componentHelper.logger.log(
                    Level.SEVERE,
                    componentHelper.logName,
                    "TimeSeriesReader",
                    "unknown TIME_FORMAT \"" + timeFormat
                    + "\" found in sdmx-ml file."
                );
        }
    }
}
