package Int.ecb.gesmescb.xml.sdmx.v2;

import Int.ecb.gesmescb.CodeList;
import Int.ecb.gesmescb.Contact;
import Int.ecb.gesmescb.KeyFamily;
import Int.ecb.gesmescb.KeyFamilyComponent;
import Int.ecb.gesmescb.MessageBegin;
import Int.ecb.gesmescb.StatisticalConcept;
import Int.ecb.gesmescb.event.ErrorEvent;
import Int.ecb.gesmescb.event.GESMESCBEventSourceAdapter;
import Int.ecb.gesmescb.event.GESMESCBReader;
import Int.ecb.gesmescb.time.TimeZoneUtil;
import Int.ecb.gesmescb.util.ComponentHelper;
import Int.ecb.logging.Level;

import java.net.InetAddress;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.UnknownHostException;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.nio.file.StandardCopyOption;
import java.text.DateFormat;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Date;
import java.util.List;

import org.w3c.dom.*;
import org.xml.sax.SAXException;

import javax.net.ssl.HttpsURLConnection;
import javax.xml.parsers.*;
import java.io.*;
import java.util.concurrent.TimeUnit;
	
public class SDMXMLStructureReader 	
	extends GESMESCBEventSourceAdapter
	implements GESMESCBReader {
	
	/* ========================= Constants ========================= */
	
	private static final String HEADERTAGNAME = "Header";
	
	private static final String PREPAREDTAGNAME = "Prepared";

	private static final String DATASTRUCTURETAGNAME = "DataStructure";
	
	private static final String DIMENSIONLISTTAGNAME = "DimensionList";
	
	private static final String DIMENSIONTAGNAME = "Dimension";
	
	private static final String ENUMERATIONTAGNAME = "Enumeration";
	
	private static final String ENUMERATIONFORMATTAGNAME = "EnumerationFormat";
	
	private static final String ATTRIBUTETAGNAME = "Attribute";
	
	private static final String AGENCYATTRIBUTENAME = "agencyID";
	
	private static final String POSITIONATTRIBUTENAME = "position";
	
	private static final String IDATTRIBUTENAME = "id";

	private static final String FREQ_DIM_NAME = "FREQ";
	
	private static final String TIME_FORMAT_NAME = "TIME_FORMAT";
	
	private static final String OBS_STATUS_NAME = "OBS_STATUS";
	
	private static final String OBS_CONF_1_NAME = "CONF_STATUS";
	
	private static final String OBS_CONF_2_NAME = "OBS_CONF";

	private static final String OBS_BREAK_1_NAME = "OBS_PRE_BREAK";
	
	private static final String OBS_BREAK_2_NAME = "PRE_BREAK_VALUE";
	
	private static final String DEFAULT_FORMAT = "AN..1998";
	
	private static final String DSD_STORE = "/global/proddatadg/sdpe/tools/config/gesmescb/fusionreg/";
//	private static final String DSD_STORE = "/tmp/";
	
	private List<String> internalFR = Arrays.asList("BIS_MACRO", "ECB_RMP1", "ECB_SUR1","ECB_IBX1", "ECB_ALF1","ECB_AWM1","ECB_BMA1","ECB_BPS1","ECB_COF1","ECB_DTCC","ECB_DTS1","ECB_EA1","ECB_ESAGFS1","ECB_GFS1","ECB_IBX1","ECB_IEA0","ECB_IFS1","ECB_IFS2","ECB_MIS1","ECB_MTS1","ECB_OB1","ECB_OPD1","ECB_ORG1","ECB_PUB1","ECB_QLF1","ECB_RMP1","ECB_S2LOA1","ECB_S2OTH1","ECB_SFT1","ECB_STA1","ECB_STI1","ECB_SUR1","ECB_TMA1");
		
	/* ========================= Fields ========================= */
	
	private InputStream ips;
	
	public String registryURL;
	
	public String registryURLRAW;
	
	private String currentAgencyId;
	
	private Date versionDate;
	
	private DateFormat isoDateParser;
	
	private String lang; 
	
	private ComponentHelper componentHelper;
	
	private Element root; 
	
	private String host;
	
	private String maintenanceAgency;
	
	private String loadAllDSDs;	

	private String allDSDsURL;
	
	/* ========================= Constructors ========================= */
	
	public SDMXMLStructureReader (String name, String filename) throws Exception {
		this.componentHelper = new ComponentHelper(name, this);
		lang = componentHelper.getPreference("language", "en");
		try {
			ips = new BufferedInputStream(new FileInputStream(filename));
		} catch (FileNotFoundException e) {
			String desc = "File not found: " + filename;
			dispatch(new ErrorEvent(desc, ErrorEvent.FAILURE));
			componentHelper.logger.log(Level.SEVERE, componentHelper.logName, "SDMXMLStructureReader", desc + e.getMessage());
			throw e;
		}

			parseIPS(ips, filename);
	}
	
	
	public SDMXMLStructureReader(String name) throws Exception {
		this.componentHelper = new ComponentHelper(name, this);
		lang = componentHelper.getPreference("language", "en");
		String ma = componentHelper.getPreference("MaintenanceAgency", "ECB");
		String kf = componentHelper.getPreference("KeyFamily", "ALL");
		loadAllDSDs = componentHelper.getPreference("loadAllDSDs", "");
		componentHelper.logger.log(Level.INFO, componentHelper.logName, "SDMXMLStructureReader", "Maintenance Agency: " + ma);
		componentHelper.logger.log(Level.INFO, componentHelper.logName, "SDMXMLStructureReader", "Key Family: " + kf);
		
		if (ma.equals("") || ma == null) {
			ma = "ECB";
		}		
		maintenanceAgency = ma;
		registryURL = System.getenv("REGISTRY_WEBSERVICE_URL");
		registryURLRAW = System.getenv("REGISTRY_WEBSERVICE_URL");
		
      
      if (registryURL == null) {
			componentHelper.logger.log(Level.SEVERE, componentHelper.logName, "SDMXMLStructureReader", "Could not load Fusion Registry Webservice URLfrom ENV var: REGISTRY_WEBSERVICE_URL");
			dispatch(new ErrorEvent("Could not load Fusion Registry Webservice URL", ErrorEvent.FAILURE));
			throw new IllegalStateException("Failed to read registryURL environment variable");
      }
		
//		registryURL = "https://PRE_HOST_VAR.HOST_VAR/ws/rest/datastructure/MA_VAR/KF_VAR/latest/?detail=full&references=children";
		
		String acc_or_prod;
		try {
			acc_or_prod = InetAddress.getLocalHost().getHostName().contains(System.getenv("test_hostname")) ? "a-sreg" : "sreg";
		} catch (UnknownHostException ex) {
			componentHelper.logger.log(Level.WARNING, componentHelper.logName, "SDMXMLStructureReader", "Could not determine the hostname, defaulting to production registry. " + ex.getMessage());
			acc_or_prod = "sreg";
		}

		host = (!componentHelper.getPreference("direction", "incoming").equalsIgnoreCase("incoming")) ? "escb.eu" : "ecb.de";
		
		host = ma.equalsIgnoreCase("BIS") ? "ecb.de": host;
		
//		host = internalFR.contains(kf) ? "ecb.de": host;
		
		registryURLRAW = registryURL.replace("PRE_HOST_VAR", acc_or_prod).replace("HOST_VAR", host).replace("datastructure", "codelist");

		allDSDsURL = registryURL.replace("PRE_HOST_VAR", acc_or_prod).replace("HOST_VAR", "ecb.de").replace("MA_VAR", loadAllDSDs).replace("KF_VAR", "ALL");
		
		registryURL = registryURL.replace("PRE_HOST_VAR", acc_or_prod).replace("HOST_VAR", host).replace("MA_VAR", ma).replace("KF_VAR", kf);
		
		componentHelper.logger.log(Level.INFO, componentHelper.logName, "URL constructor", "Registry webservice URL " + registryURL);

		String filenameDSD = DSD_STORE + host + "/" + ma + "_" + kf + ".DSD";

		root = getURL(registryURL, filenameDSD);

	}
	
	private Element getURL(String url, String localReadPath) throws Exception {
		InputStream tmp_ips = null;
		URL u = null;
		HttpsURLConnection huc = null;
		int code = 0;
		try {
			u = new URL(url);
		} catch (MalformedURLException e1) {
			componentHelper.logger.log(Level.INFO, componentHelper.logName, "getURL", "Malformed URL: " + url);
		} 
		try {
			huc = (HttpsURLConnection) u.openConnection(); 
			tmp_ips = huc.getInputStream();
		} catch (IOException e) {
			try {
				code = huc.getResponseCode();
			} catch (IOException e1) {
			}
			componentHelper.logger.log(Level.WARNING, componentHelper.logName, "getURL",  "Could not load SDMX artefact from Fusion Registry! " + Integer.toString(code) + " response. Looking for local SDMX artefact ...");
			File f = new File(localReadPath);
			if (f.exists()){
				componentHelper.logger.log(Level.INFO, componentHelper.logName, "getURL", "Reading SDMX artefact from: " + localReadPath);
				tmp_ips = isFromFilePath(localReadPath);
			}
			else {
				componentHelper.logger.log(Level.WARNING, componentHelper.logName, "getURL", localReadPath + " does not exist! Could not load SDMX artefact from registry or local file system!");
			}
		}
		Element local_el = null;
		
		local_el = parseIPS(tmp_ips, localReadPath);


	    if (tmp_ips != null) {
	        try {
	        	tmp_ips.close();
	        } catch (Exception e) {
	        	}
	    }		

		return local_el;
	}
	
	private InputStream isFromFilePath(String filePath) {
	    File initialFile = new File(filePath);
	    try {
	    	return new FileInputStream(initialFile);
		} catch (IOException e) {
			componentHelper.logger.log(Level.SEVERE, componentHelper.logName, "isFromFilePath", e.getMessage());
		}
		return null;
	}

	private Element parseIPS(InputStream ips_local, String localReadPath) throws Exception {
		if (ips_local == null) {
			componentHelper.logger.log(Level.WARNING, componentHelper.logName, "parseIPS",
					"Received inputStream is null!");
			return null;

		}
		
		Document document = null;		
		try {
			DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
			factory.setNamespaceAware(true);
			document = factory.newDocumentBuilder().parse(ips_local);
		} catch (SAXException e) {
			componentHelper.logger.log(Level.WARNING, componentHelper.logName, "parseIPS", "Failure parsing the XMl from the registry.");
			File f = new File(localReadPath);
			if (f.exists()){
				componentHelper.logger.log(Level.INFO, componentHelper.logName, "parseIPS", "Reading SDMX artefact from: " + localReadPath);
				ips_local = isFromFilePath(localReadPath);
				try {
					DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
					factory.setNamespaceAware(true);
					document = factory.newDocumentBuilder().parse(ips_local);
				} catch (SAXException e1) {
					final String desc = "SAX exception: " + e1 + ".";
					componentHelper.logger.log(Level.SEVERE, componentHelper.logName, "parseIPS",
							"SAX XML Parser exception while parsing XML: " + e1.getMessage(), e1);
					dispatch(new ErrorEvent(desc, ErrorEvent.FAILURE));
					throw e;	
				}
			}			
			else {
				final String desc = "SAX exception: " + e + ".";
				componentHelper.logger.log(Level.SEVERE, componentHelper.logName, "parseIPS",
						"SAX XML Parser exception while parsing XML: " + e.getMessage(), e);
				dispatch(new ErrorEvent(desc, ErrorEvent.FAILURE));
				throw e;	
			}
			
	
		} catch (IOException e) {
			final String desc = "I/O problem: " + e + ".";
			dispatch(new ErrorEvent(desc, ErrorEvent.FAILURE));
			componentHelper.logger.log(Level.SEVERE, componentHelper.logName, "parseIPS",
					"IO exception while parsing XML: " + e.getMessage(), e);
			throw e;
		} catch (ParserConfigurationException e) {
			final String desc = "Parser configuration problem: " + e + ".";
			dispatch(new ErrorEvent(desc, ErrorEvent.FAILURE));
			componentHelper.logger.log(Level.SEVERE, componentHelper.logName, "parseIPS",
					"Parser Configuration exception while parsing XML: " + e.getMessage());
			throw e;
		}
		
		document.getDocumentElement().normalize();
		Element local_root = document.getDocumentElement();
		
		if (local_root == null) {
			componentHelper.logger.log(Level.SEVERE, componentHelper.logName, "parseIPS",
					"No root element found in SDMX file");
			dispatch(new ErrorEvent("No root element found in SDMX file", ErrorEvent.FAILURE));
			throw new Exception();	
		}			
		return local_root;
	}
	
	
	/* ========================= Public Methods ========================= */

	public void run () {

		readHeader( root );
		
		readConcepts( root );
		
		readCodeLists( root );
		
		getUnattachedCodelist();
		
		readKeyFamilies( root );

		if (!loadAllDSDs.isEmpty() && loadAllDSDs != null) {
			componentHelper.logger.log(Level.INFO, componentHelper.logName, "SDMXMLStructureReader", "loadAllDSDs is set to: " + loadAllDSDs );
			getAllDSDs();
		}	
				
	}
	
	private void getAllDSDs() {
		componentHelper.logger.log(Level.INFO, componentHelper.logName, "getAllDSDs", "URL: " + allDSDsURL);
		String filenameAllDSDs = DSD_STORE + host + "/" + loadAllDSDs + "_ALL.DSD";
		Element rootAllStructures = null;
		
		try {
			rootAllStructures = getURL(allDSDsURL, filenameAllDSDs);
		} catch (Exception e) {
			componentHelper.logger.log(Level.SEVERE, componentHelper.logName, "getAllDSDs", "URL: " + allDSDsURL);
			dispatch(new ErrorEvent("Could not parse inputStream!", ErrorEvent.FAILURE));
		}
		
		if (rootAllStructures == null) {
			componentHelper.logger.log(Level.SEVERE, componentHelper.logName, "getAllDSDs", "Could not get SDMX structures!");
			dispatch(new ErrorEvent("Could not get SDMX structures!", ErrorEvent.FAILURE));
		}

		readHeader(rootAllStructures);
		readConcepts(rootAllStructures);
		readCodeLists(rootAllStructures);
		readKeyFamilies(rootAllStructures);	
	}

	
	/* ========================= Private Methods ========================= */

	private void getUnattachedCodelist() {
		String url = "";
		String agencyId = currentAgencyId;
		for (String codelist : Arrays.asList("CL_ORGANISATION")) {
			agencyId = maintenanceAgency.equalsIgnoreCase("ALL") ? "ECB": agencyId;
			Element rootCodelist = null;
			url = registryURLRAW.replace("KF_VAR", codelist).replace("MA_VAR", agencyId);

			componentHelper.logger.log(Level.INFO, componentHelper.logName, "getUnattachedCodelist",
					"get Codelist URL " + url);
			String filenameCL = DSD_STORE + host + "/" + agencyId + "_CL_ORGANISATION.CL";
			try {
				rootCodelist = getURL(url, filenameCL);
			} catch (Exception e) {
			
			}
			
			if (rootCodelist == null && !maintenanceAgency.equals("ECB")) {
				agencyId = "ECB";
				url = registryURLRAW.replace("KF_VAR", codelist).replace("MA_VAR", agencyId);
				componentHelper.logger.log(Level.INFO, componentHelper.logName, "getUnattachedCodelist",
						"Could not get " + codelist + " codelist for agency " + maintenanceAgency + " - will try " + agencyId + " instead.");				
				componentHelper.logger.log(Level.INFO, componentHelper.logName, "getUnattachedCodelist", "get Codelist URL " + url);
				filenameCL = DSD_STORE + host + "/" + agencyId + "_CL_ORGANISATION.CL";
				try {
					rootCodelist = getURL(url, filenameCL);
				} catch (Exception ex) {
					componentHelper.logger.log(Level.SEVERE, componentHelper.logName, "getUnattachedCodelist", "URL: " + url);
					dispatch(new ErrorEvent("Could not parse inputStream!", ErrorEvent.FAILURE));
				}	
			}
		    
			if (rootCodelist == null) {
				componentHelper.logger.log(Level.SEVERE, componentHelper.logName, "getUnattachedCodelist", "Could not get SDMX structures!");
				dispatch(new ErrorEvent("Could not get SDMX structures!", ErrorEvent.FAILURE));
			}
			
			readUnattachedCodeLists(rootCodelist);

		}
	}
	
	private void readUnattachedCodeLists(final Element localroot){
		Element clElement, cElement;
		String id, name= "";
		String agencyId = "";
		NodeList codeNodes;
		ArrayList<String> codes = new ArrayList<String>();
		ArrayList<String> descs = new ArrayList<String>();
		
		NodeList clRoots = localroot.getElementsByTagNameNS("*","Codelist");
		
		if (clRoots == null) {
			componentHelper.logger.log(Level.WARNING, componentHelper.logName, "readUnattachedCodeLists", "unattached Codelist root not found.");
		}
		
		for (int i = 0; i < clRoots.getLength(); i++){
			clElement = (Element) clRoots.item(i);
			id = clElement.getAttribute(IDATTRIBUTENAME);
			agencyId = clRoots.item(0).getAttributes().getNamedItem(AGENCYATTRIBUTENAME).getNodeValue();
			name = getNameFromElement(clElement, lang);
			codeNodes = clElement.getElementsByTagNameNS("*", "Code");
			for (int j = 0; j < codeNodes.getLength(); j++){
				cElement = (Element) codeNodes.item(j);
				codes.add(cElement.getAttribute(IDATTRIBUTENAME));
				descs.add(getNameFromElement(cElement, lang));
			}
			componentHelper.logger.log(Level.FINE, componentHelper.logName, "readUnattachedCodeLists", "Starting to read unattached codelists: " + agencyId);
			componentHelper.logger.log(Level.FINE, componentHelper.logName, "readUnattachedCodeLists", "Dispatching unattached CodeList.");

			dispatch(new MessageBegin("MREF000001", "", "", agencyId, "", "", new Contact[] {}, "", "", "", "", "", ""));
			componentHelper.logger.log(Level.FINE, componentHelper.logName, "readUnattachedCodeLists", "New Agency: " + agencyId);	
			
			dispatch(new CodeList(id, name, "", versionDate, codes, descs));
			componentHelper.logger.log(Level.FINEST, componentHelper.logName, "readUnattachedCodeLists", "Name: " + name);
			componentHelper.logger.log(Level.FINEST, componentHelper.logName, "readUnattachedCodeLists", "Id: " + id);
			codes.clear();
			descs.clear();
		}
	}
		
	private void readHeader(Element root) {
		versionDate = parseDate(((Element) ((Element) root.getElementsByTagNameNS("*", HEADERTAGNAME).item(0))
				.getElementsByTagNameNS("*", PREPAREDTAGNAME).item(0)).getTextContent());

		if (versionDate == null) {
			componentHelper.logger.log(Level.SEVERE, componentHelper.logName, "Header", "versionDate is not available.");
		}
	}
	
	private void readKeyFamilies (final Element root){
		
		String position, 
			conceptId, 
			conceptType,  
			usageStatus, 
			attachmentLevel, 
			codeListId,
			agencyId, 
			id;
		
		componentHelper.logger.log(Level.FINE, componentHelper.logName, "readKeyFamilies",
				"Start reading Key Families.");
		
		String format = "", name = "", description = "";
		ArrayList<KeyFamilyComponent> keyFamilyComponents = new ArrayList<KeyFamilyComponent>();
		int dimNum;
		Element structureRoot, eElement, aLElement;		
		
		NodeList structureRoots = root.getElementsByTagNameNS("*", DATASTRUCTURETAGNAME);
		
		if (structureRoots == null) {
			componentHelper.logger.log(Level.SEVERE, componentHelper.logName, "readKeyFamilies", "Data structure root not found.");
			dispatch(new ErrorEvent("Data structure root not found", ErrorEvent.FAILURE));
		}		
		
		for (int i = 0; i < structureRoots.getLength(); i++) {
			structureRoot = (Element) structureRoots.item(i);
			agencyId = structureRoot.getAttribute(AGENCYATTRIBUTENAME);
			if (!agencyId.equals(currentAgencyId)) {
				componentHelper.logger.log(Level.FINE, componentHelper.logName, "readKeyFamilies",
						"dispatch messagebegin.");
				dispatch(new MessageBegin("MREF000001", "", "", agencyId, "", "", new Contact[] {}, "", "", "", "", "", ""));
				componentHelper.logger.log(Level.FINE, componentHelper.logName, "readKeyFamilies",
						"New Agency: " + agencyId);				
				currentAgencyId = agencyId;
			}
			
			id = structureRoot.getAttribute(IDATTRIBUTENAME);
			
			name = getNameFromElement(structureRoot, lang);

			NodeList dims = ((Element) structureRoot.getElementsByTagNameNS("*", DIMENSIONLISTTAGNAME).item(0))
					.getElementsByTagNameNS("*", DIMENSIONTAGNAME);
			
			if (dims == null) {
				componentHelper.logger.log(Level.SEVERE, componentHelper.logName, "readKeyFamilies",
						"DimensionList root not found.");
				dispatch(new ErrorEvent("DimensionList root not found", ErrorEvent.FAILURE));
			}	
			dimNum = dims.getLength();
			for (int temp = 0; temp < dimNum; temp++) {
				eElement = (Element) dims.item(temp);
				attachmentLevel="";
				usageStatus="";
				position = eElement.getAttribute(POSITIONATTRIBUTENAME);
				conceptId = eElement.getAttribute(IDATTRIBUTENAME);
				conceptType = (conceptId.equals(FREQ_DIM_NAME)) ? 
						KeyFamilyComponent.FREQUENCY_DIMENSION : 
						KeyFamilyComponent.KEY_STRUCTURE_DIMENSION;
				if (eElement.getElementsByTagNameNS("*", ENUMERATIONTAGNAME).item(0) == null){
					codeListId = "";
					format = getFormatFromElement(eElement, false);
				} else {
					codeListId = ((Element)((Element) eElement.getElementsByTagNameNS("*", ENUMERATIONTAGNAME).item(0))
							.getElementsByTagNameNS("*", "Ref").item(0)).getAttribute(IDATTRIBUTENAME);				
					format = getFormatFromElement(eElement, true);
				}
				keyFamilyComponents.add( 
						new KeyFamilyComponent(
							conceptId, conceptType, position, format, usageStatus, attachmentLevel, codeListId
						)
				);
			}

			eElement = (Element) structureRoot.getElementsByTagNameNS("*", "TimeDimension").item(0);
			conceptId = eElement.getAttribute(IDATTRIBUTENAME);
			conceptType = KeyFamilyComponent.TIME_DIMENSION;
			position = String.valueOf(dimNum+1);
			format = "AN..35";
			usageStatus = "";
			attachmentLevel = "";
			codeListId = "";
			keyFamilyComponents.add( 
					new KeyFamilyComponent(
						conceptId, conceptType, position, format, usageStatus, attachmentLevel, codeListId
					)
			);
			
			eElement = (Element)((Element) structureRoot.getElementsByTagNameNS("*", "MeasureList").item(0))
					.getElementsByTagNameNS("*", "PrimaryMeasure").item(0);
			
			if (eElement == null) {
				componentHelper.logger.log(Level.SEVERE, componentHelper.logName, "readKeyFamilies",
						"MeasureList is not available.");
			}	
			
			conceptId = eElement.getAttribute(IDATTRIBUTENAME);
			conceptType = KeyFamilyComponent.ARRAY_CELL;
			position = String.valueOf(dimNum+3);
			format = getFormatFromElement(eElement, false);
			usageStatus = "";
			attachmentLevel = "";
			
			keyFamilyComponents.add( 
					new KeyFamilyComponent(
						conceptId, conceptType, position, format, usageStatus, attachmentLevel, codeListId
					)
			);
			
			NodeList atts = ((Element) structureRoot.getElementsByTagNameNS("*", "AttributeList").item(0))
					.getElementsByTagNameNS("*", ATTRIBUTETAGNAME);
			
			if (atts == null) {
				componentHelper.logger.log(Level.SEVERE, componentHelper.logName, "readKeyFamilies",
						"AttributeList root not found.");
				dispatch(new ErrorEvent("AttributeList root not found.", ErrorEvent.FAILURE));
			}	
			for (int j = 0; j < atts.getLength(); j++){
				eElement = (Element) atts.item(j);
				conceptId = eElement.getAttribute(IDATTRIBUTENAME);
				if (conceptId.equals(OBS_STATUS_NAME)){
					conceptType = KeyFamilyComponent.ARRAY_CELL;
					position = String.valueOf(dimNum+4);
				} else if (conceptId.equals(OBS_CONF_1_NAME) ||
						conceptId.equals(OBS_CONF_2_NAME)){
					conceptType = KeyFamilyComponent.ARRAY_CELL;
					position = String.valueOf(dimNum+5);
				} else if (conceptId.equals(OBS_BREAK_1_NAME) ||
						conceptId.equals(OBS_BREAK_2_NAME)){
					conceptType = KeyFamilyComponent.ARRAY_CELL;
					position = String.valueOf(dimNum+6);
				} else if (conceptId.equals(TIME_FORMAT_NAME)){
					conceptType = "1";
					position = String.valueOf(dimNum+2);
				} else {
					conceptType = "Z09";
					position = "";
				}
				if (eElement.getAttribute("assignmentStatus").equals("Conditional")){
					usageStatus = KeyFamilyComponent.CONDITIONAL;
				} else {
					usageStatus = KeyFamilyComponent.MANDATORY;
				}
				if (eElement.getElementsByTagNameNS("*", ENUMERATIONTAGNAME).item(0) == null){
					codeListId = "";
					format = getFormatFromElement(eElement, false);
				} else {
					codeListId = ((Element)((Element) eElement.getElementsByTagNameNS("*", ENUMERATIONTAGNAME).item(0))
							.getElementsByTagNameNS("*", "Ref").item(0)).getAttribute(IDATTRIBUTENAME);
					format = getFormatFromElement(eElement, true);
				}

				aLElement = (Element) eElement.getElementsByTagNameNS("*", "AttributeRelationship").item(0);
				for (int n = 0; n < aLElement.getChildNodes().getLength(); n++){
				}	
				if (aLElement.getFirstChild().getLocalName().equals("PrimaryMeasure")){
					attachmentLevel = KeyFamilyComponent.ATTACHED_TO_OBSERVATION;
				} else if (aLElement.getChildNodes().getLength() == dimNum){
					attachmentLevel = KeyFamilyComponent.ATTACHED_TO_TIME_SERIES;
				} else if (aLElement.getChildNodes().getLength() > 1 &&
						aLElement.getChildNodes().getLength() < dimNum){
					attachmentLevel = KeyFamilyComponent.ATTACHED_TO_SIBLING_GROUP;
				} else {
					attachmentLevel = KeyFamilyComponent.ATTACHED_TO_DATA_SET;
				}
				keyFamilyComponents.add( 
						new KeyFamilyComponent(
							conceptId, conceptType, position, format, usageStatus, attachmentLevel, codeListId
						)
				);	
			}
			componentHelper.logger.log(Level.FINE, componentHelper.logName, "readKeyFamilies",
					"Dispatch KeyFamily: " + currentAgencyId + " " + id);

			dispatch(
					new KeyFamily(
						id,name,description,versionDate,keyFamilyComponents
					)
			);
			
			keyFamilyComponents.clear();
		}	
	}

	
	private void readCodeLists(final Element root){
		Element clElement, cElement;
		String id,
			agencyId,
			name= "";
		NodeList codeNodes;
		ArrayList<String> codes = new ArrayList<String>();
		ArrayList<String> descs = new ArrayList<String>();
		
		componentHelper.logger.log(Level.FINE, componentHelper.logName, "readCodeLists",
				"Starting to read codelists.");
		
		NodeList clRoots = root.getElementsByTagNameNS("*","Codelist");
		
		if (clRoots == null) {
			//TODO
			componentHelper.logger.log(Level.WARNING, componentHelper.logName, "readCodeLists",
					"Codelist root not found");
			//dispatch(new ErrorEvent("Codelist root not found.", ErrorEvent.FAILURE));
		}
		
		for (int i = 0; i < clRoots.getLength(); i++){
			clElement = (Element) clRoots.item(i);
			id = clElement.getAttribute(IDATTRIBUTENAME);
			agencyId = clElement.getAttribute(AGENCYATTRIBUTENAME); 
			if (!agencyId.equals(currentAgencyId)){
				dispatch(new MessageBegin("MREF000001", "", "", agencyId, "", "", new Contact[] {}, "", "", "", "", "",
						""));
				componentHelper.logger.log(Level.FINE, componentHelper.logName, "readCodeLists",
						"New Agency: " + agencyId);	
				currentAgencyId = agencyId;
			}
			name = getNameFromElement(clElement, lang);
			codeNodes = clElement.getElementsByTagNameNS("*", "Code");
			for (int j = 0; j < codeNodes.getLength(); j++){
				cElement = (Element) codeNodes.item(j);
				codes.add(cElement.getAttribute(IDATTRIBUTENAME));
				descs.add(getNameFromElement(cElement,lang));
			}
			componentHelper.logger.log(Level.FINE, componentHelper.logName, "readCodeLists",
					"Dispatching CodeList " + id);
			dispatch(
				new CodeList(
					id, name, "", versionDate, codes, descs
				)
			);

			codes.clear();
			descs.clear();
		}
		
	}
	
	private void readConcepts(final Element root){
		String	agencyId;
		Element cSElement, cElement;
		NodeList concepts;
		componentHelper.logger.log(Level.FINE, componentHelper.logName, "readConcepts",
				"Starting to read concepts.");
		NodeList csRoots = root.getElementsByTagNameNS("*","ConceptScheme");
		
		if (csRoots == null) {
			componentHelper.logger.log(Level.SEVERE, componentHelper.logName, "readConcepts",
					"Concepts root not found");
			dispatch(new ErrorEvent("Concepts root not found.", ErrorEvent.FAILURE));
		}		
		
		for (int i = 0; i < csRoots.getLength();i++){
			cSElement = (Element) csRoots.item(i); 
			agencyId = cSElement.getAttribute(AGENCYATTRIBUTENAME);
			if (!agencyId.equals(currentAgencyId)){
				dispatch(
					new MessageBegin(
						"MREF000001","","",agencyId,"","",new Contact[]{},
						"","","","","",""
					)
				);
				currentAgencyId = agencyId;
			}
			concepts = cSElement.getElementsByTagNameNS("*", "Concept");
			for (int j = 0; j < concepts.getLength(); j++){
				cElement = (Element) concepts.item(j);
				componentHelper.logger.log(Level.FINE, componentHelper.logName, "readConcepts",
						"Dispatching StatisticalConcept " + cElement.getAttribute(IDATTRIBUTENAME));
				dispatch(
					new StatisticalConcept(
						cElement.getAttribute(IDATTRIBUTENAME),
						getNameFromElement(cElement, lang),
						"", versionDate
					)
				);
			}
		}		
	}
	
    private Date parseDate(final String input) {
        if (null == isoDateParser) {
            isoDateParser = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ssZ");
        }
        Date output = null;
        try {
           output = isoDateParser.parse(
               TimeZoneUtil.getInstance().normalize( input ) 
           );            
        } catch (final ParseException e) {
            componentHelper.logger.log(
                Level.SEVERE,
                componentHelper.logName,
                "endElement",
                "Problem when parsing date: \"" 
                + "\"."
            );
        }
        return output;
    }
    
    private String getNameFromElement(final Element el, final String lang){
    	Element nElement;
    	NodeList names = el.getElementsByTagNameNS("*", "Name");
    	String name = null;
		if (names.getLength() > 0) {
			name = ((Element) names.item(0)).getTextContent();
		}
    	for (int k = 1; k < names.getLength(); k++){
			nElement = (Element) names.item(k);
			if (nElement.getAttribute("xml:lang").equals(lang) 
					&& ((Element)nElement.getParentNode()).equals(el)){
				name = nElement.getTextContent();
				break;
			}
		}
		return name;
    }
    
	private String getFormatFromElement(final Element el, final boolean isCoded) {
		String minLength, maxLength, format = DEFAULT_FORMAT;
		NamedNodeMap nnm;
		if (isCoded) {
			if (el.getElementsByTagNameNS("*", ENUMERATIONFORMATTAGNAME).item(0) == null) {
				return format;
			}
			nnm = ((Element) el.getElementsByTagNameNS("*", ENUMERATIONFORMATTAGNAME).item(0)).getAttributes();
		} else {
			if (el.getElementsByTagNameNS("*", "TextFormat").item(0) == null) {
				return format;
			}
			nnm = ((Element) el.getElementsByTagNameNS("*", "TextFormat").item(0)).getAttributes();
		}

		Node minLengthNode = nnm.getNamedItem("minLength");
		Node maxLengthNode = nnm.getNamedItem("maxLength");
		minLength = minLengthNode != null ? nnm.getNamedItem("minLength").getNodeValue() : "";
		maxLength = maxLengthNode != null ? nnm.getNamedItem("maxLength").getNodeValue() : "";

		if (minLength.equals(maxLength)) {
			format = "AN" + maxLength;
		} else {
			format = "AN.." + maxLength;
		}
		return format;
	}


	@Override
	public void close() {
		// TODO Auto-generated method stub
		
	}

}


