package Int.ecb.gesmescb.xml.sdmx.v2;

import java.io.IOException;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.FactoryConfigurationError;
import javax.xml.parsers.ParserConfigurationException;

import org.w3c.dom.DOMException;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.xml.sax.SAXException;

import Int.ecb.gesmescb.util.ComponentHelper;
import Int.ecb.logging.Level;

/**
 * Parses an SDMX-ML Structure message, containing a data structure definition,
 * and extracts the key family identifier and the agency maintaining the key 
 * family. 
 * 
 * @author Francois Hoeltzli
 * @author Xavier Sosnovsky
 */
public class SimpleStructureProxy {
    
    /* ========================= Fields ========================= */
    
    /** Takes care of logging. */
    protected final ComponentHelper componentHelper;
    
    private Element keyFamily;

    /* ====================== Constructors ====================== */
    
    /**
     * Constructs a new SimpleStructureProxy.
     * 
     * @param keyFamilyURI URL of the SDMX-ML Structure file containing the
     * data structure definition. 
     */
    public SimpleStructureProxy(final String keyFamilyURI) {
        componentHelper = new ComponentHelper( "", this );
        try {
            final DocumentBuilderFactory factory
                = DocumentBuilderFactory.newInstance();
            factory.setNamespaceAware( true );
            final DocumentBuilder domParser = factory.newDocumentBuilder();
            final Document xmlDoc  
                = domParser.parse( keyFamilyURI );
            final String structureNS = 
                "http://www.SDMX.org/resources/SDMXML/schemas/v2_0/structure";
            final String messageNS = 
                "http://www.SDMX.org/resources/SDMXML/schemas/v2_0/message";
            final Element root = xmlDoc.getDocumentElement();
            final Element keyFamilies = 
                (Element) root.getElementsByTagNameNS( messageNS, "KeyFamilies" ).item(0);
            keyFamily = 
                (Element) keyFamilies.getElementsByTagNameNS( structureNS, "KeyFamily" ).item(0);
        } catch ( FactoryConfigurationError fce ) {
            componentHelper.logger.log(
                Level.SEVERE,
                componentHelper.logName,
                this.getClass().getSimpleName(),
                "Cannot get DOM parser! " + fce
            );
        } catch ( ParserConfigurationException pce ) {
            componentHelper.logger.log(
                Level.SEVERE,
                componentHelper.logName,
                this.getClass().getSimpleName(),
                "Cannot configure parser. " + pce
            );
        } catch ( SAXException sxe ) {
            Exception initialException = sxe;
            if ( sxe.getException() != null ) {
                initialException = sxe.getException();
            }
            componentHelper.logger.log(
                Level.SEVERE,
                componentHelper.logName,
                this.getClass().getSimpleName(),
                "SAX Exception: " + initialException
            );
        } catch ( DOMException doe) {
            componentHelper.logger.log(
                Level.SEVERE,
                componentHelper.logName,
                this.getClass().getSimpleName(),
                "DOM Exception: " + doe
            );
        } catch ( IOException ioe ) {
            componentHelper.logger.log(
                Level.SEVERE,
                componentHelper.logName,
                this.getClass().getSimpleName(),
                "Can't find XML structure file. " + ioe
            );
        }
        
    }
    
    /* ===================== Public Methods ===================== */
    
    /**
     * @return The key family identifier (e.g.: ECB_EXR1)
     */
    public String getKeyFamilyRef() {
        return keyFamily.getAttribute("id");
    }
    
    /**
     * @return The agency maintaining the key family (e.g.: ECB) 
     */
    public String getKeyFamilyAgency() {
        return keyFamily.getAttribute("agencyID");
    }
}
