// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2002, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
package Int.ecb.gesmescb.xml.sdmx.v2.compact;

import java.util.Iterator;

import org.xml.sax.Attributes;
import org.xml.sax.XMLReader;

import Int.ecb.gesmescb.Attribute;
import Int.ecb.gesmescb.KeyFamilyComponent;
import Int.ecb.gesmescb.TimeSeries;
import Int.ecb.gesmescb.time.GregorianRegularInterval;
import Int.ecb.gesmescb.xml.MapStack;
import Int.ecb.gesmescb.xml.SAX2Reader;
import Int.ecb.logging.Level;

/**
 * Content handler for sibling groups.  
 * 
 * @author Francois Hoeltzli
 * @author Xavier Sosnovsky
 */
class GroupReader extends KeyReader {
    
    /* ====================== Constructors ====================== */
    /**
     * Constructs a new GroupReader.
     * 
     * @param parser The parser of the XML document
     * @param parent The class responsible for handling content
     */
    public GroupReader(final XMLReader parser, final SAX2Reader parent) {
        super(parser, parent);
    }
    
    /* ===================== Public Methods ===================== */
    
    /**
     * {@inheritDoc} 
     */
    public void startElement (
        final String namespaceURI,
        final String localName,
        final String qName,
        final Attributes atts
    ) {
        super.startElement( namespaceURI, localName, qName, atts );
        if ( "Group".equals( localName ) ) {
            handleGroup(atts);
        } else {
            componentHelper.logger.log(
                    Level.SEVERE,
                    componentHelper.logName,
                    "GroupReader",
                    "Unknown (for SMDX-ML) element : \""
                    + localName + "\""
                );
        }
    }
    
    /**
     * {@inheritDoc} 
     */
    public void endElement (
            final String namespaceURI,
            final String localName,
            final String qName
        ) {
        super.endElement( namespaceURI, localName, qName );
        if ( "Group".equals( localName ) ) {
            finish();
        } else {
            componentHelper.logger.log(
                Level.SEVERE,
                componentHelper.logName,
                "GroupReader",
                "Closing unexpected element : \""
                + localName + "\""
            );
        }
    }
    
    /* ===================== Protected Methods ================== */
    
    /**
     * {@inheritDoc} 
     */
    protected boolean isSiblingGroup() {
        return true;
    }
    
    /**
     * {@inheritDoc} 
     */
	protected String distilKey (final MapStack xmlAttributes) {
    	//
		// SDMXMLWriter knows only sibling groups and expect them
		// to be written with an empty leading frequency dimension.
		//
    	return TimeSeries.KEY_SEPARATOR + super.distilKey(xmlAttributes);
    }
	
	/* ===================== Private Methods ================== */

    private void handleGroup(final Attributes atts) {
        final MapStack xmlSeriesAtts = readXMLAttributes(atts);
        key = distilKey(xmlSeriesAtts);
        
        if ( key != null ) {
            for (
                final Iterator iterator = keyFamily.attributeIterator(
                        KeyFamilyComponent.ATTACHED_TO_SIBLING_GROUP);
                iterator.hasNext();
            ) {
                final KeyFamilyComponent component
                = (KeyFamilyComponent) iterator.next();
                final String value = 
                    xmlSeriesAtts.get( component.getConceptId() );
                if ( value != null ) {
                    dispatch(new Attribute(key, 
                        (GregorianRegularInterval) null, 
                        component.getConceptId(), value));
                }
            }
        }
    }
}
