// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2002, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
package Int.ecb.gesmescb.xml.sdmx.v2.compact;

import java.text.DateFormat;
import java.text.NumberFormat;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.Hashtable;
import java.util.Iterator;

import org.xml.sax.Attributes;
import org.xml.sax.XMLReader;

import Int.ecb.gesmescb.Attribute;
import Int.ecb.gesmescb.Contact;
import Int.ecb.gesmescb.InterchangeBegin;
import Int.ecb.gesmescb.InterchangeEnd;
import Int.ecb.gesmescb.KeyFamily;
import Int.ecb.gesmescb.KeyFamilyComponent;
import Int.ecb.gesmescb.MessageBegin;
import Int.ecb.gesmescb.MessageEnd;
import Int.ecb.gesmescb.event.GESMESCBEventListener;
import Int.ecb.gesmescb.metabase.Metabase;
import Int.ecb.gesmescb.time.GregorianRegularInterval;
import Int.ecb.gesmescb.time.TimeZoneUtil;
import Int.ecb.gesmescb.xml.MapStack;
import Int.ecb.gesmescb.xml.SAX2Reader;
import Int.ecb.gesmescb.xml.sdmx.v2.KeyFamilyDependentReader;
import Int.ecb.logging.Level;

/**
 * Content handler for SDMX-ML Message Header and DataSet.
 * 
 * @author Maria Stahl
 * @author Francois Hoeltzli
 */
class HeaderReader21 extends KeyFamilyDependentReader {
    
    /* ======================= Constants ======================== */
   	
	private final String INVALID_ACTION="InvalidAction";
	
	// should be from Constants, but Constants is in another package.
	private final String DATA_NS_URI
	="http://www.sdmx.org/resources/sdmxml/schemas/v2_1/data/structurespecific";
	
	/* ========================= Fields ========================= */
    
    private String iref;
    private boolean test;
    private String preparationTimeStamp;
    private String sender;
    private String receiver;
    private String name;
    private String keyFamilyRef;
    private String keyFamilyAgency;
    private KeyFamily keyFamily;
    private String dataSetID;
    private String dataSetAction;
    private String extracted;
    private String reportingBegin;
    private String reportingEnd;
    private int mrefCounter;
    private NumberFormat numberFormatter;
    private DateFormat isoDateParser;
    private StringBuffer accumulator;
    private String mref;
    private boolean isSenderOpen;
    private String headerAction;
    
    private final Hashtable refAgencies = new Hashtable();
    private final Hashtable refIdentifiers = new Hashtable();
    private String reference;
    
    private final GroupReader groupReader;
    private final TimeSeriesReader seriesReader;

    private String defaultExtractionDate;
    private boolean defaultToPreparationDate;
	private String currentDatasetAction;
    
    /* ====================== Constructors ====================== */
    
    /**
     * Constructs a new HeaderReader21.
     * 
     * @param parser The parser of the XML document
     * @param parent The class responsible for handling content
     */
	
	/*In computer technology, a parser is a program, usually part of a compiler, that receives input in the form of sequential source program instructions, 
	 * interactive online commands, markup tags, or some other defined interface and breaks them up into parts (for example, the nouns (objects), 
	 * verbs (methods), and their attributes or options) that can then be managed by other programming (for example, other components in a compiler). 
	 * A parser may also check to see that all input has been provided that is necessary.*/
	
	/*HeaderReader21.java assumes that the action is mandatory in the DataSet tag, which it is not according to SDMX documentation:
	 *"This purpose is noted in the action for the data set, which is either inherited from the header of the data message or explicitly stated at the data set level."
	 * However, the action is also NOT mandatory in the header. It just needs to specified somewhere. The HeaderReader should check if it is given in the header and then check if it is overwritten in any of the DataSet tags, if so it should be overwritten for all datasets where it is specified and otherwise inherit it from the Header of the message.
	 */
	
    public HeaderReader21(final XMLReader parser, final SAX2Reader parent) {
        super(parser, parent);
        mrefCounter = 0;
        groupReader = new GroupReader( parser, this );  
        seriesReader = new TimeSeriesReader( parser, this );
        isSenderOpen = false;
		defaultExtractionDate
			= componentHelper.getPreference(
        		"defaultExtractionDate", null 
            );
		if (null != defaultExtractionDate) {
			defaultToPreparationDate
			= "preparationDate".equalsIgnoreCase( defaultExtractionDate );
		}
    }

    /* ===================== Public Methods ===================== */ 

   
/* Public means that the method is visible and can be called from other objects of other types. 
   Other alternatives are private, protected, package and package-private.
   Static means that the method is associated with the class, not a specific instance (object) of that class. 
   This means that you can call a static method without creating an object of the class.
   Void means that the method has no return value. If the method returned an int(eger) you would write int(eger) instead of void.
   The combination of all three of these is most commonly seen on the main method which most tutorials will include */  
 
    /**
     * {@inheritDoc} 
     */
    public void addListener (
            final GESMESCBEventListener listener,
            final boolean isImportant
        ) {
        super.addListener( listener, isImportant );
        groupReader.addListener( listener, isImportant );
        seriesReader.addListener( listener, isImportant );
    }

/* Final is a class that cannot be extended. 
 * This does not mean that all references to objects of the class would act as if they were declared as final */
    
    /**
     * {@inheritDoc} 
     */
    public void removeAllListeners () {
        super.removeAllListeners();
        groupReader.removeAllListeners();
        seriesReader.removeAllListeners();
    }
    
    /**
     * {@inheritDoc} 
     */
    public void startElement (
            final String namespaceURI,
            final String localName,
            final String qName,
            final Attributes atts
        ) {
        /*Here it probably starts with the elements*/
        super.startElement( namespaceURI, localName, qName, atts );
        accumulator = new StringBuffer();
		
        if ( "Sender".equals( localName ) ) {
            isSenderOpen = true;
            sender = atts.getValue( "", "id" );
            if ( null == sender ) {
                componentHelper.logger.log(
                    Level.SEVERE,
                    componentHelper.logName,
                    "startElement",
                    "Found sender element without id attribute."
                );
            } else {
                componentHelper.logger.log(
                    Level.INFO,
                    componentHelper.logName,
                    "startElement",
                    "Sender:" + sender + "."
                );
            }
        } else if ( "Receiver".equals( localName ) ) {
            receiver = atts.getValue( "", "id");
            if ( null == receiver ) {
                componentHelper.logger.log(
                    Level.SEVERE,
                    componentHelper.logName,
                    "startElement",
                    "Found receiver element without id attribute."
                );
            } else {
                componentHelper.logger.log(
                    Level.INFO,
                    componentHelper.logName,
                    "startElement",
                    "Receiver:" + receiver + "."
                );
            }
        } else if ( "Structure".equals( localName )
               && XMLNS_SDMX_MESSAGE_V2_1.equals( namespaceURI ) ) {
            reference = atts.getValue( "", "structureID" );            
        } else if ( "Structure".equals( localName )
               && XMLNS_SDMX_COMMON_V2_1.equals( namespaceURI ) ) {
        } else if ( "Ref".equals( localName ) ) {
            refAgencies.put( reference,
                atts.getValue( "", "agencyID" )
            );
            refIdentifiers.put( reference,
                atts.getValue( "", "id" )
            );


            
        } else if ( "DataSet".equals( localName ) ) {
			//debugDisplayAttributes(atts);
            String currentReference = 
                atts.getValue( XMLNS_SDMX_DATA_STRUCTURESPECIFIC_V2_1,
                    "structureRef" )
            ;       
            componentHelper.logger.log(
                    Level.INFO,
                    componentHelper.logName,
                    "startElement",
                    "structureRef is \"" + currentReference + "\""
                );

            keyFamilyAgency = (String) refAgencies.get(
                 (String) currentReference
            );
            keyFamilyRef = (String) refIdentifiers.get(
                 (String) currentReference
            );
                        
            if ( null == keyFamilyAgency ) {
                componentHelper.logger.log(
                    Level.SEVERE,
                    componentHelper.logName,
                    "startElement",
                    "no KeyFamilyAgency found in the xml header."
                );
            } 
            if ( null == keyFamilyRef ) {
              componentHelper.logger.log(
                    Level.SEVERE,
                    componentHelper.logName,
                    "startElement",
                    "no KeyFamilyRef found in the xml header."
                );
            } 
            if ( null == Metabase.getMetabase()
                .getMaintenanceAgency( keyFamilyAgency )
            ) {
                componentHelper.logger.log(
                    Level.SEVERE,
                    componentHelper.logName,
                    "startElement",
                    "Metabase does not know keyFamilyAgency \""
                    + keyFamilyAgency + "\"."
                );
            } 
            if ( null == Metabase.getMetabase()
                .getMaintenanceAgency( keyFamilyAgency )
                .getKeyFamily( keyFamilyRef )
            ) {
                componentHelper.logger.log(
                    Level.SEVERE,
                    componentHelper.logName,
                    "startElement",
                    "Metabase does not know keyFamily \""
                    + keyFamilyRef + "\"."
                );
            } else {
                keyFamily = Metabase.getMetabase()
                    .getMaintenanceAgency( keyFamilyAgency )
                    .getKeyFamily( keyFamilyRef ); 
                groupReader.setKeyFamily(keyFamily);
                seriesReader.setKeyFamily(keyFamily);
            }
            mrefCounter++; 

            if (null == numberFormatter) {
                numberFormatter = NumberFormat.getInstance();
                numberFormatter.setGroupingUsed(false);
                numberFormatter.setMaximumIntegerDigits(6);
                numberFormatter.setMinimumIntegerDigits(6);
            }
            String dsReportingBegin = reportingBegin;
            String dsReportingEnd = reportingEnd;
            String dsDataProviderID = sender;
            componentHelper.logger.log(
                Level.INFO,
                componentHelper.logName,
                "startElement",
                "Assigning dsDataProviderID with sender \"" + sender + "\"."
            );
			
            if (null != atts.getValue( "", "reportingBeginDate")) {
                dsReportingBegin = 
                    atts.getValue( "", "reportingBeginDate");
            }
            if (null != atts.getValue( "", "reportingEndDate")) {
                dsReportingEnd = 
                    atts.getValue( "", "reportingEndDate");
            }
            if (null != atts.getValue( "", "dataProviderID")) {
                dsDataProviderID = 
                    atts.getValue( "", "dataProviderID");
            }
            final String action = null != atts.getValue( DATA_NS_URI, "action")
				? atts.getValue( DATA_NS_URI, "action")
				: atts.getValue( "", "action")
				;

			final String dsAction =
					null == action ? headerAction
				:	"Append".equals(action)  ? MessageBegin.OBJECT_LEVEL_ACTION
				:	"Update".equals(action)  ? MessageBegin.UPDATE_OR_REPLACE_ACTION
				:	"Replace".equals(action) ? MessageBegin.UPDATE_OR_REPLACE_ACTION
				:	"Delete".equals(action)  ? MessageBegin.DELETE_ACTION
				:	"Information".equals(action) ? MessageBegin.TESTING_ACTION
				:	INVALID_ACTION
				;
			
			if (dsAction == null){
				componentHelper.logger.log(
	                Level.SEVERE,
	                componentHelper.logName,
	                "dataSet Element",
	                "The action is neither defined in the header nor"
	                + " in the dataset element."
	            );
			}
			if ( INVALID_ACTION.equals(dsAction) ){
                componentHelper.logger.log(
                    Level.SEVERE,
                    componentHelper.logName,
					"dataSet Element",
                    "data:action has unknown value \"" + action + "\""
                    + " or dataSet element does not define an action and"
                    + " the action defined in the header is invalid."
                );
			}
            if (null != atts.getValue( "", "datasetID")) {
                dataSetID = atts.getValue( "", "datasetID");
            }
            
            String periodRange  = "";
            String periodFormat = ""; 
            if ((null != dsReportingBegin && null == dsReportingEnd) || 
                (null == dsReportingBegin && null != dsReportingEnd)) {
                componentHelper.logger.log(
                    Level.SEVERE,
                    componentHelper.logName,
                    "startElement",
                    (null == dsReportingBegin)
                    ? "Reporting end without reporting begin."
                    : "Reporting begin missing without reporting end."
                );
            } else if (null != dsReportingBegin && null != dsReportingEnd) {
				try {
					periodRange = (dsReportingBegin.substring(0, 10) + dsReportingEnd
							.substring(0, 10)).replace("-", "");
					periodFormat = "711";
				} catch (Exception e) {
					periodRange = (dsReportingBegin + dsReportingEnd).replace(
							"-", "");
					if (10 == dsReportingBegin.length()) {
						periodFormat = "711";
					} else if (4 == dsReportingBegin.length()) {
						periodFormat = "702";
					} else if (7 == dsReportingBegin.length()) {
						final char periodTypeIndicator = dsReportingBegin
								.charAt(5);
						if ('Q' == periodTypeIndicator) {
							periodFormat = "708";
							periodRange = periodRange.replace("Q", "");
						} else if ('B' == periodTypeIndicator) {
							periodFormat = "704";
							periodRange = periodRange.replace("B", "");
						} else if ('W' == periodTypeIndicator) {
							periodFormat = "716";
							periodRange = periodRange.replace("W", "");
						} else {
							periodFormat = "710";
						}
					} else {
						componentHelper.logger.log(Level.SEVERE,
								componentHelper.logName, "startElement",
								"Could not guess period format for: "
										+ periodRange);
					}
				}

			}
            mref = "MREF" + numberFormatter.format(mrefCounter);
            // set reasonnable defaults
			receiver = (null==receiver) ? "" : receiver; 
			keyFamilyAgency = (null == keyFamilyAgency) ? "" : keyFamilyAgency;
            final Date preparationDate = parseDate( preparationTimeStamp );

			final Date extractionDate = 
			    (null == extracted)
				? (null != defaultExtractionDate)
					? defaultToPreparationDate
						? preparationDate
						: parseDate(defaultExtractionDate)
					: null
				: parseDate(extracted)
				;
            keyFamilyRef = (null == keyFamilyRef) ? "" : keyFamilyRef;
            if ( null == dataSetID || "".equals(dataSetID) ) {
                componentHelper.logger.log(
                        Level.SEVERE,
                        componentHelper.logName,
                        "startElement",
                        "File Header contains no DataSetID element."
                    );
            }
            name = (null == name) ? "" : name;
			
			currentDatasetAction = dsAction;
			
            dispatch( new MessageBegin( mref, dsDataProviderID,
				receiver, 
                keyFamilyAgency, 
                MessageBegin.STATISTICAL_DATA, name, 
                new Contact[] {}, dataSetID, dsAction,
                extractionDate,
                periodRange, periodFormat, 
                keyFamilyRef,
                dataSetID )
			); 
            // Ioana Trainotti 20.01.2014 - Process attributes at dataset level
            final MapStack xmlSeriesAtts = KeyReader.readXMLAttributes( atts );
            for (
                    final Iterator iterator = keyFamily.attributeIterator(
                            KeyFamilyComponent.ATTACHED_TO_DATA_SET);
                    iterator.hasNext();
                ) {
                    final KeyFamilyComponent component
                    = (KeyFamilyComponent) iterator.next();
                    final String value = 
                        xmlSeriesAtts.get( component.getConceptId() );
                    if ( value != null ) {
                        componentHelper.logger.log(
                                Level.FINEST,
                                componentHelper.logName,
                                this.getClass().getSimpleName(),
                                "Dispatching attributes \""
                                + component.getConceptId() 
                                + "\" with value \"" 
                                + value + "\""
                        );
                       
                        dispatch(new Attribute("", // dataset level, no key provided
                        			(GregorianRegularInterval) null, 
                        			component.getConceptId(), value));
                        
                    }
                }
            
        } else if ("Group".equals(localName)) {
            delegate(
                groupReader,
                namespaceURI, localName, qName, atts
            );
        } else if ("Series".equals(localName)) {
			// if only empty obs make the series empty.
			// Otherwise a potential gesmes writer will fully ignore the series. 	
			seriesReader.setEmptySeriesOfEmptyObs(
				MessageBegin.DELETE_ACTION.equals(currentDatasetAction)
			);
            delegate(
                    seriesReader,
                    namespaceURI, localName, qName, atts
                );
        } else if (
        	// handled when their end tag is encountered.
        	"Header".equals(localName)
        	||
        	"ID".equals(localName)
        	||
        	"Test".equals(localName)
        	||
        	"Name".equals(localName)
        	||
        	"Prepared".equals(localName)
        	||
        	"Contact".equals(localName)
        	||
        	"Department".equals(localName)
        	||
        	"URI".equals(localName)
        	||
        	"DataSetAction".equals(localName)
        	||
        	"DataSetAgency".equals(localName)
        	||
        	"DataSetID".equals(localName)
        	||
        	"KeyFamilyAgency".equals(localName)
        	||
        	"KeyFamilyRef".equals(localName)
        	||
        	"Extracted".equals(localName)
        	||
        	"ReportingBegin".equals(localName)
        	||
        	"Truncated".equals(localName)
        	||
            "Source".equals(localName)
            ||
            "Role".equals(localName)
            ||
            "Telephone".equals(localName)
            ||
            "Fax".equals(localName)
            ||
            "X400".equals(localName)
            ||
            "Email".equals(localName)
            ||
            "ReportingEnd".equals(localName)
        ){}
        else {
        	componentHelper.logger.log(
                Level.SEVERE,
                componentHelper.logName,
                "HeaderReader21",
                "StartElement in HeaderReader21:"
                + "Unknown (for SMDX-ML) element : \""
                + localName + "\""
            );
        }
    }
    
    /**
     * {@inheritDoc} 
     */
    public void characters (
        final char[] ch,
        final int start,
        final int length
    ) {
/*Here is the super with the character elements*/
    	super.characters( ch, start, length );
        accumulator.append( ch, start, length );        
    }

    /**
     * {@inheritDoc} 
     */
    public void endElement (
        final String namespaceURI,
        final String localName,
        final String qName
    ) {
        super.endElement( namespaceURI, localName, qName );
        if ( "ID".equals( localName ) ) {
            iref = accumulator.toString();
        } else if ( "Test".equals( localName ) ) { 
            test = "true".equals( accumulator.toString().toLowerCase() );
            if (test == true) {
          	  componentHelper.logger.log(
                        Level.INFO,
                        componentHelper.logName,
                        "HeaderReader",
                        "This is a TEST interchange"
                    );
          }
        } else if ( "Prepared".equals( localName ) ) { 
            preparationTimeStamp = accumulator.toString();
        } else if ( "Header".equals( localName ) ) {
            if (null == preparationTimeStamp) {
                componentHelper.logger.log(
                    Level.SEVERE,
                    componentHelper.logName,
                    "HeaderReader21",
                    "The preparation time stamp cannot be null."
                );
            }
            if (null == iref) {
                componentHelper.logger.log(
                    Level.SEVERE,
                    componentHelper.logName,
                    "HeaderReader21",
                    "The iref cannot be null."
                );
            }
            if (null == sender) {
                componentHelper.logger.log(
                    Level.SEVERE,
                    componentHelper.logName,
                    "HeaderReader21",
                    "The sender cannot be null."
                );
            }
            
            final Date preparationDate = parseDate( preparationTimeStamp );
            dispatch( new InterchangeBegin(
                iref, sender, (null == receiver) ? "" : receiver, 
                preparationDate, test ) 
            );
        } else if (
            "StructureSpecificData".equals( localName )
            ||
            "StructureSpecificTimeSeries".equals( localName ) 
        ) {
            dispatch( new InterchangeEnd( iref ) );
            finish();
        } else if ( "Name".equals( localName ) && !isSenderOpen ) {
            name = accumulator.toString();
        } else if ( "KeyFamilyRef".equals( localName ) ) {
            keyFamilyRef = accumulator.toString();
        } else if ( "KeyFamilyAgency".equals( localName ) ) {
            keyFamilyAgency = accumulator.toString(); 
        } else if ( "DataSetID".equals( localName ) ) {
            dataSetID = accumulator.toString(); 
        } else if ( "Extracted".equals( localName ) ) {
            extracted = accumulator.toString();             
        } else if ( "ReportingBegin".equals( localName ) ) {
            reportingBegin = accumulator.toString(); 
        } else if ( "ReportingEnd".equals( localName ) ) {
            reportingEnd = accumulator.toString(); 
        } else if ( "DataSetAction".equals( localName ) ) {
			// do nothing, this should come from the action attribute in the DataSet Element
        	String action = accumulator.toString();
        	headerAction =
				action == null ? action
        		:	"Append".equals(action)  ? MessageBegin.OBJECT_LEVEL_ACTION
				:	"Update".equals(action)  ? MessageBegin.UPDATE_OR_REPLACE_ACTION
				:	"Replace".equals(action) ? MessageBegin.UPDATE_OR_REPLACE_ACTION
				:	"Delete".equals(action)  ? MessageBegin.DELETE_ACTION
				:	"Information".equals(action) ? MessageBegin.TESTING_ACTION
				:	INVALID_ACTION
				;
        	if (headerAction == INVALID_ACTION){
        		componentHelper.logger.log(
        			Level.WARNING,
                    componentHelper.logName,
     				"HeaderReader21",
                    "DataSetAction in header has unknown value \"" + action + "\"."
                    + " Message will be accepted if correct actions are given in"
                    + " the dataSet element."
                 );
        	}
        	/*headerAction=accumulator.toString();
			componentHelper.logger.log(
				Level.WARNING,
				componentHelper.logName,
				"Element DataSetAction",
				"Action should be specified in the data:action attribute from Element DataSet. Contents of DataSetAction ignored."
			);*/
        } else if ( "DataSet".equals( localName ) ) {
            dispatch(new MessageEnd(mref));
        } else if ("Sender".equals( localName )) {
            isSenderOpen = false;
        } else if (
        	"Contact".equals( localName )
        	||
        	"DataSetAgency".equals( localName )
        	||
        	"Department".equals( localName )
        	||
        	"URI".equals( localName )
        	||
        	"Receiver".equals( localName )
        	||
            "Truncated".equals(localName)
            ||
            "Source".equals(localName)
            ||
            "Role".equals(localName)
            ||
            "Telephone".equals(localName)
            ||
            "Fax".equals(localName)
            ||
            "X400".equals(localName)
            ||
            "Email".equals(localName)
			||
			( "Name".equals(localName) && isSenderOpen )
         ||
         "Structure".equals(localName)
         ||
         "Ref".equals(localName)
   		)
         {}
        else {
            componentHelper.logger.log(
                Level.SEVERE,
                componentHelper.logName,
                "HeaderReader21",
                "EndElement in HeaderReader21:"
                + "Unknown (for SMDX-ML) element : \""
                + localName + "\""
            );
        }
    }
    
    /* ===================== Private Methods ===================== */
    
    private Date parseDate(final String input) {
        if (null == isoDateParser) {
            isoDateParser = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ssZ");
        }
        Date output = null;
        try {
           output = isoDateParser.parse(
               TimeZoneUtil.getInstance().normalize( input ) 
           );            
        } catch (final ParseException e) {
            componentHelper.logger.log(
                Level.SEVERE,
                componentHelper.logName,
                "endElement",
                "Problem when parsing date: \"" + extracted
                + "\"."
            );
        }
        return output;
    }
    
	private void debugDisplayAttributes (final Attributes attributes) {
		int length = attributes.getLength();  
		for (int i=0; i<length; i++) {  
			String name = attributes.getQName(i);  
			System.out.println("Name:" + name);  
			String value = attributes.getValue(i);  
			System.out.println("Value:" + value);  
			//String nsUri = attributes.getURI(i);  
			//System.out.println("NS Uri:" + nsUri);  
			String lName = attributes.getLocalName(i);  
			System.out.println("Local Name:" + lName);  
		}
	}
}
