// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2004, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.gesmescb.xml.sdmx.v2.compact;

import java.text.DateFormat;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.Iterator;

import junit.framework.Test;
import junit.framework.TestSuite;

import org.xml.sax.SAXException;

import Int.ecb.gesmescb.Attribute;
import Int.ecb.gesmescb.Contact;
import Int.ecb.gesmescb.InterchangeBegin;
import Int.ecb.gesmescb.InterchangeEnd;
import Int.ecb.gesmescb.MessageBegin;
import Int.ecb.gesmescb.MessageEnd;
import Int.ecb.gesmescb.Observation;
import Int.ecb.gesmescb.TimeSeries;
import Int.ecb.gesmescb.metabase.Metabase;
import Int.ecb.gesmescb.time.GregorianRegularInterval;
import Int.ecb.gesmescb.time.RegularInterval;
import Int.ecb.gesmescb.time.TimeZoneUtil;
import Int.ecb.gesmescb.xml.KeyFamilyURIMapping;
import Int.ecb.gesmescb.xml.SAX2Reader;
import Int.ecb.gesmescb.xml.SAX2ReaderTest;

/* ***************************** Class ****************************** */

public class StructureSpecificReaderTest extends SAX2ReaderTest {

    private final DateFormat dailyDateFormat;
    
    Date pDate;
    
    /* ====================== Constructors ====================== */
    
    public StructureSpecificReaderTest(final String name) {
        super(name);
        dailyDateFormat = new SimpleDateFormat("yyyy-MM-dd");
    }
    
    /* ================== Protected Methods ===================== */

    /** Sets up the test fixture. */
    protected void setUp () {
        super.setUp();
        //setNamespace( XMLNS_GESMES, null );
        final Metabase metabase = Metabase.getMetabase();
        metabase.handle( new MessageBegin(
            "DUMMY", "", "", "ECB",
            MessageBegin.STATISTICAL_DEFINITIONS,
            "", new Contact[]{}
        ) );
        metabase.handle( KF );
        System.setProperty(
            KeyFamilyURIMapping.class.getName() + ".uri.ECB.KF",
            "http://www.ecb.int/vocabulary/2003-11-10/kf"
        );
        final SimpleDateFormat parser = 
            new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ssZ");
        pDate = null;
        try {
            pDate = parser.parse(
               TimeZoneUtil.getInstance().normalize(
                   "2008-12-03T09:38:52+0100" ) 
           );            
        } catch (final ParseException e) {
            fail("Exception when parsing date");
        }
    }

    /** Tears down the test fixture. */
    protected void tearDown () {
        System.getProperties().remove(
            KeyFamilyURIMapping.class.getName() + ".uri.ECB.KF"
        );
    }

    protected SAX2Reader factory (
        final org.xml.sax.XMLReader parser,
        final SAX2Reader parent
    ) {
        final StructureSpecificReader reader = new StructureSpecificReader(
            parser, parent );
        reader.setKeyFamily( KF );
        return reader;
    }

    /* ===================== Public Methods ===================== */

    /**
     * Assembles and returns a test suite for all
     * the test methods of this test case.
     *
     * @return a non-null test suite
     */
    public static Test suite () {
        return new TestSuite( StructureSpecificReaderTest.class );
    }

    /**
     * @param arg the array of argument {@link String}s
     */
    public static void main ( final String[] arg ) {
        final String[] testCaseName = {
            StructureSpecificReaderTest.class.getName()
        };
        junit.textui.TestRunner.main( testCaseName );
    }

    /* ---------------------- Test Methods ---------------------- */
    
    public void testMinmalHeader () throws SAXException {
        openElement( "StructureSpecificData", "" );
        openElement( "Header", "" );
        openElement("ID", "");
        characters ("IREF000638");
        closeElement("ID");
        openElement("Test", "");
        characters("false");
        closeElement("Test");
        openElement("Name", "");
        characters("Some test");
        closeElement("Name");
        openElement("Prepared", "");
        characters("2008-12-03T09:38:52+01:00");
        closeElement("Prepared");
        openElement("Sender", "id=4F0");
        closeElement("Sender");
        openElement("Receiver", "id=DE2");
        closeElement("Receiver");
        setNamespace(XMLNS_SDMX_MESSAGE_V2_1,"message");
        openElement("Structure", "dimensionAtObservation=TIME_PERIOD structureID=KF namespace=urn:sdmx:org.sdmx.infomodel.datastructure.DataStructure=ECB:KF(1.0):ObsLevelDim:TIME_PERIOD");
        setNamespace( "" ,"somethingElse");
        openElement("Structure",""); 
        openElement("Ref","agencyID=ECB id=KF");
        closeElement("Ref");
        closeElement("Structure");		
        closeElement("Structure");		
        closeElement( "Header" );
        closeElement( "StructureSpecificData" );

        final Iterator iterator = listener.iterator();
        assertTrue( "first", iterator.hasNext() );
        assertEquals(
            "interchange begin",
            new InterchangeBegin(
                "IREF000638", "4F0", "DE2", pDate, false
            ),
            iterator.next()
        );
        assertTrue( "second", iterator.hasNext() );
        assertEquals(
            "interchange end",
            new InterchangeEnd( "IREF000638" ),
            iterator.next()
        );
        assertTrue( "no more", !iterator.hasNext() );
    }    

    public void testMessageWithHeaderAndDataSet () throws SAXException {
        openElement( "StructureSpecificData", "" );
        openElement( "Header", "" );
        openElement("ID", "");
        characters ("IREF000638");
        closeElement("ID");
        openElement("Test", "");
        characters("false");
        closeElement("Test");
        openElement("Name", "");
        characters("Some test");
        closeElement("Name");
        openElement("Prepared", "");
        characters("2008-12-03T09:38:52+01:00");
        closeElement("Prepared");
        openElement("Sender", "id=4F0");
        closeElement("Sender");
        openElement("Receiver", "id=DE2");
        closeElement("Receiver");
        setNamespace(XMLNS_SDMX_MESSAGE_V2_1,"message");
        openElement("Structure", "dimensionAtObservation=TIME_PERIOD structureID=KF namespace=urn:sdmx:org.sdmx.infomodel.datastructure.DataStructure=ECB:KF(1.0):ObsLevelDim:TIME_PERIOD");
        setNamespace( "" ,"somethingElse");
        openElement("Structure",""); 
        openElement("Ref","agencyID=ECB id=KF");
        closeElement("Ref");
        closeElement("Structure");		
        closeElement("Structure");
        openElement("DataSetID","");
        characters("ICP");
        closeElement("DataSetID");        		
        closeElement( "Header" );
        openElement("DataSet", "reportingBeginDate=2007-Q1 reportingEndDate=2007-Q2 structureRef=KF action=Delete");
       /* closeElement("DataSet");
        /*closeElement( "StructureSpecificData" );
        
		final Iterator iterator = listener.iterator();
        assertTrue( "first", iterator.hasNext() );
        assertEquals(
            "interchange begin",
            new InterchangeBegin(
                "IREF000638", "4F0", "DE2", pDate, false
            ),
            iterator.next()
        );
        assertTrue("There should be a MessageBegin", iterator.hasNext());
        final MessageBegin expectedMB = new MessageBegin(
                "MREF000001",
                "4F0",
                "DE2",
                "ECB",
                MessageBegin.STATISTICAL_DATA,
                "Some test",
                new Contact[] {},
                "ICP",
                MessageBegin.DELETE_ACTION,
                "",
                "2007120072",
                "708",
                "KF"
                );
        assertEquals("Did not receive the expected MessageBegin", expectedMB, 
                iterator.next());
        assertTrue("Expecting MessageEnd", iterator.hasNext());
        final MessageEnd expectedME = new MessageEnd("MREF000001");
        assertEquals("Did not receive the expected MessageEnd", expectedME, 
                iterator.next());
        assertEquals(
            "interchange end",
            new InterchangeEnd( "IREF000638" ),
            iterator.next()
        );
        assertTrue( "no more", !iterator.hasNext() );
		*/
    }
    
    public void testMessageWithHeaderAndDataSetAndGroup () throws SAXException {
        openElement( "StructureSpecificData", "" );
        openElement( "Header", "" );
        openElement("ID", "");
        characters ("IREF000638");
        closeElement("ID");
        openElement("Test", "");
        characters("false");
        closeElement("Test");
        openElement("Name", "");
        characters("Some test");
        closeElement("Name");
        openElement("Prepared", "");
        characters("2008-12-03T09:38:52+01:00");
        closeElement("Prepared");
        openElement("Sender", "id=4F0");
        closeElement("Sender");
        openElement("Receiver", "id=DE2");
        closeElement("Receiver");
        setNamespace(XMLNS_SDMX_MESSAGE_V2_1,"message");
        openElement("Structure", "dimensionAtObservation=TIME_PERIOD structureID=KF namespace=urn:sdmx:org.sdmx.infomodel.datastructure.DataStructure=ECB:KF(1.0):ObsLevelDim:TIME_PERIOD");
        setNamespace( "" ,"somethingElse");
        openElement("Structure",""); 
        openElement("Ref","agencyID=ECB id=KF");
        closeElement("Ref");
        closeElement("Structure");		
        closeElement("Structure");		
        closeElement( "Header" );
        openElement("DataSet", "reportingBeginDate=2007-Q1 reportingEndDate=2007-Q2 structureRef=KF dataProviderID=5D0 datasetID=ICP action=Delete");
        openElement("Group", "ITEM=AUD SG_ATT=TEST");
        closeElement("Group");
        closeElement("DataSet");
        closeElement( "StructureSpecificData" );
        
        final Iterator iterator = listener.iterator();
        assertTrue( "first", iterator.hasNext() );
        assertEquals(
            "interchange begin",
            new InterchangeBegin(
                "IREF000638", "4F0", "DE2", pDate, false
            ),
            iterator.next()
        );
        assertTrue("There should be a MessageBegin", iterator.hasNext());
        final MessageBegin expectedMB = new MessageBegin(
                "MREF000001",
                "5D0",
                "DE2",
                "ECB",
                MessageBegin.STATISTICAL_DATA,
                "Some test",
                new Contact[] {},
                "ICP",
                MessageBegin.DELETE_ACTION,
                "",
                "2007120072",
                "708",
                "KF"
                );
        assertEquals("Did not receive the expected MessageBegin", expectedMB, 
                iterator.next());
        assertTrue("There's room for one attribute", iterator.hasNext());
        final Attribute expectedAttribute = new Attribute(".AUD", 
                (GregorianRegularInterval) null, "SG_ATT", "TEST");
        assertEquals("The attributes should be equal", expectedAttribute, 
                iterator.next());        
        assertTrue("Expecting MessageEnd", iterator.hasNext());
        final MessageEnd expectedME = new MessageEnd("MREF000001");
        assertEquals("Did not receive the expected MessageEnd", expectedME, 
                iterator.next());
        assertEquals(
            "interchange end",
            new InterchangeEnd( "IREF000638" ),
            iterator.next()
        );
        assertTrue( "no more", !iterator.hasNext() );
    }

    public void testMessageWithSeries () throws SAXException, ParseException {
        openElement( "StructureSpecificTimeSeries", ""); // variatio delectat
        openElement( "Header", "" );
        openElement("ID", "");
        characters ("IREF000638");
        closeElement("ID");
        openElement("Test", "");
        characters("false");
        closeElement("Test");
        openElement("Name", "");
        characters("Some test");
        closeElement("Name");
        setNamespace(XMLNS_SDMX_MESSAGE_V2_1,"message");
        openElement("Structure", "dimensionAtObservation=TIME_PERIOD structureID=KF namespace=urn:sdmx:org.sdmx.infomodel.datastructure.DataStructure=ECB:KF(1.0):ObsLevelDim:TIME_PERIOD");
        setNamespace( "" ,"somethingElse");
        openElement("Structure",""); 
        openElement("Ref","agencyID=ECB id=KF");
        closeElement("Ref");
        closeElement("Structure");		
        closeElement("Structure");		
        openElement("Prepared", "");
        characters("2008-12-03T09:38:52+01:00");
        closeElement("Prepared");
        openElement("Sender", "id=4F0");
        closeElement("Sender");
        openElement("Receiver", "id=DE2");
        closeElement("Receiver");

        //openElement("KeyFamilyRef", "");
        //characters("KF");
        //closeElement("KeyFamilyRef");
        
        //openElement("KeyFamilyAgency", "");
        //characters("ECB");
        //closeElement("KeyFamilyAgency");
        closeElement( "Header" );
        openElement("DataSet", "reportingBeginDate=2008-01-04 reportingEndDate=2008-01-05 structureRef=KF dataProviderID=5D0 datasetID=ICP action=Delete");
        setNamespace( "", "somethingElse" );
        openElement("Series", "FREQ=D ITEM=AUD TIME_FORMAT=P1D");
        openElement("Obs", "TIME_PERIOD=2008-01-04 OBS_VALUE=1.36 OBS_STATUS=N OBS_CONF=C");
        closeElement("Obs");
        openElement("Obs", "TIME_PERIOD=2008-01-05 OBS_VALUE=1.45 OBS_STATUS=N OBS_CONF=C OBS_PRE_BREAK=2.45");
        closeElement("Obs");
        closeElement("Series");
        closeElement("DataSet");
        closeElement( "StructureSpecificTimeSeries" );
        
        final Iterator iterator = listener.iterator();
        assertTrue( "first", iterator.hasNext() );
        assertEquals(
            "interchange begin",
            new InterchangeBegin(
                "IREF000638", "4F0", "DE2", pDate, false
            ),
            iterator.next()
        );
        assertTrue("There should be a MessageBegin", iterator.hasNext());
        final MessageBegin expectedMB = new MessageBegin(
                "MREF000001",
                "5D0",
                "DE2",
                "ECB",
                MessageBegin.STATISTICAL_DATA,
                "Some test",
                new Contact[] {},
                "ICP",
                MessageBegin.DELETE_ACTION,
                "",
                "2008010420080105",
                "711",
                "KF"
        );
        assertEquals("Did not receive the expected MessageBegin", expectedMB, 
                iterator.next());
        assertTrue("There's room for one series", iterator.hasNext());
        final Observation[] observations = {
                new Observation( "1.36", "N", "C", "" ),
                new Observation( "1.45", "N", "C", "2.45" ),
            };
        final TimeSeries expectedSeries = new TimeSeries("D.AUD",
                new GregorianRegularInterval(
                        RegularInterval.DAILY, 
                        dailyDateFormat.parse("2008-01-04"), 
                        dailyDateFormat.parse("2008-01-06")),
                observations
        );
        assertEquals(
                "Did not recieve the expected Series",
                expectedSeries,
                iterator.next()
        );
        assertTrue("Expecting MessageEnd", iterator.hasNext());
        final MessageEnd expectedME = new MessageEnd("MREF000001");
        assertEquals("Did not receive the expected MessageEnd", expectedME, 
                iterator.next());
        assertEquals(
            "interchange end",
            new InterchangeEnd( "IREF000638" ),
            iterator.next()
        );
        assertTrue( "no more", !iterator.hasNext() );
    }    
    
	public void testInterchangeWith2Messages () throws SAXException, ParseException {
        openElement( "StructureSpecificData", ""); 
        openElement( "Header", "" );
        openElement("ID", "");
        characters ("IREF000638");
        closeElement("ID");
        openElement("Test", "");
        characters("false");
        closeElement("Test");
        openElement("Name", "");
        characters("Some test");
        closeElement("Name");
        setNamespace(XMLNS_SDMX_MESSAGE_V2_1,"message");
        openElement("Structure", "dimensionAtObservation=TIME_PERIOD structureID=KF namespace=urn:sdmx:org.sdmx.infomodel.datastructure.DataStructure=ECB:KF(1.0):ObsLevelDim:TIME_PERIOD");
        setNamespace( "" ,"somethingElse");
        openElement("Structure",""); 
        openElement("Ref","agencyID=ECB id=KF");
        closeElement("Ref");
        closeElement("Structure");		
        closeElement("Structure");		
        openElement("Prepared", "");
        characters("2008-12-03T09:38:52+01:00");
        closeElement("Prepared");
        openElement("Sender", "id=4F0");
        closeElement("Sender");
        openElement("Receiver", "id=DE2");
        closeElement("Receiver");

        openElement("KeyFamilyRef", "");
        characters("KF");
        closeElement("KeyFamilyRef");
        
        openElement("KeyFamilyAgency", "");
        characters("ECB");
        closeElement("KeyFamilyAgency");
        closeElement( "Header" );
        openElement("DataSet", "reportingBeginDate=2008-01-04 reportingEndDate=2008-01-05 structureRef=KF dataProviderID=5D0 datasetID=ICP action=Delete");
        setNamespace( "", "somethingElse" );
        openElement("Series", "FREQ=D ITEM=AUD TIME_FORMAT=P1D");
        openElement("Obs", "TIME_PERIOD=2008-01-04 OBS_VALUE=1.36 OBS_STATUS=N OBS_CONF=C");
        closeElement("Obs");
        openElement("Obs", "TIME_PERIOD=2008-01-05 OBS_VALUE=1.45 OBS_STATUS=N OBS_CONF=C OBS_PRE_BREAK=2.45");
        closeElement("Obs");
        closeElement("Series");
        closeElement("DataSet");
// second dataset
        openElement("DataSet", "reportingBeginDate=2013-12 reportingEndDate=2014-01 structureRef=KF dataProviderID=5D0 datasetID=ICP action=Replace");
        setNamespace( "", "somethingElse" );
        openElement("Series", "FREQ=M ITEM=USD");
        openElement("Obs", "TIME_PERIOD=2013-12 OBS_VALUE=1.99 OBS_STATUS=N OBS_CONF=C");
        closeElement("Obs");
        openElement("Obs", "TIME_PERIOD=2014-01 OBS_VALUE=2.99 OBS_STATUS=N OBS_CONF=C OBS_PRE_BREAK=2.45");
        closeElement("Obs");
        closeElement("Series");
        closeElement("DataSet");

        closeElement( "StructureSpecificData" );
        
        final Iterator iterator = listener.iterator();
		//skip the interchange begin
		iterator.next();
		//skip the messagebegin
		iterator.next();
		//check the first series in the first message
        final Observation[] observations1= {
                new Observation( "1.36", "N", "C", "" ),
                new Observation( "1.45", "N", "C", "2.45" ),
            };
        final TimeSeries expectedSeries1 = new TimeSeries("D.AUD",
                new GregorianRegularInterval(
                        RegularInterval.DAILY, 
                        dailyDateFormat.parse("2008-01-04"), 
                        dailyDateFormat.parse("2008-01-06")),
                observations1
        );
        assertEquals(
                "Did not recieve the first expected Series",
                expectedSeries1,
                iterator.next()
        );
		//skip the messageEnd
		iterator.next();

        assertTrue("There should be a second MessageBegin", iterator.hasNext());
        final MessageBegin expectedMB = new MessageBegin(
                "MREF000002",
                "5D0",
                "DE2",
                "ECB",
                MessageBegin.STATISTICAL_DATA,
                "Some test",
                new Contact[] {},
                "ICP",
                MessageBegin.UPDATE_OR_REPLACE_ACTION,
                "",
                "201312201401",
                "710",
                "KF"
        );
        assertEquals("Did not receive the expected MessageBegin", expectedMB, 
                iterator.next());
        assertTrue("There's room for one series", iterator.hasNext());
        final Observation[] observations = {
                new Observation( "1.99", "N", "C", "" ),
                new Observation( "2.99", "N", "C", "2.45" ),
            };
        final TimeSeries expectedSeries = new TimeSeries("M.USD",
                new GregorianRegularInterval(
                        RegularInterval.MONTHLY, 
                        dailyDateFormat.parse("2013-12-01"), 
                        dailyDateFormat.parse("2014-02-01")),
                observations
        );
        assertEquals(
                "Did not recieve the expected Series",
                expectedSeries,
                iterator.next()
        );
        assertTrue("Expecting MessageEnd", iterator.hasNext());
        final MessageEnd expectedME = new MessageEnd("MREF000002");
        assertEquals("Did not receive the expected MessageEnd", expectedME, 
                iterator.next());
        assertTrue( "still expecting an interchange end", iterator.hasNext() );
        assertEquals(
            "interchange end",
            new InterchangeEnd( "IREF000638" ),
            iterator.next()
        );
        assertTrue( "no more", !iterator.hasNext() );
	}
}
