// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2004, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
package Int.ecb.gesmescb.xml.sdmx.v2.compact;

import java.text.DateFormat;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.SortedMap;
import java.util.TreeMap;
import java.util.Map;

import org.xml.sax.Attributes;
import org.xml.sax.XMLReader;

import Int.ecb.gesmescb.Attribute;
import Int.ecb.gesmescb.KeyFamilyComponent;
import Int.ecb.gesmescb.Observation;
import Int.ecb.gesmescb.TimeSeries;
import Int.ecb.gesmescb.time.CalendarHelper;
import Int.ecb.gesmescb.time.GregorianRegularInterval;
import Int.ecb.gesmescb.time.RegularInterval;
import Int.ecb.gesmescb.xml.XMLDateFormat;
import Int.ecb.gesmescb.xml.MapStack;
import Int.ecb.gesmescb.xml.SAX2Reader;
import Int.ecb.gesmescb.xml.sdmx.v2.SDMXHelper;
import Int.ecb.logging.Level;

/**
 * Content handler for series and observations.  
 * 
 * @author Francois Hoeltzli
 * @author Xavier Sosnovsky
 */
class TimeSeriesReader extends KeyReader {
	
    /* ========================= Fields ========================= */
    
    private TimeSeries series;
    
    private String timeFormat;
    
    private SortedMap observations;
    
    private XMLDateFormat dateFormat;
    
    private int frequency;
    
	// Whether to empty the series if it consists only of empty
	// observations. We might want to do this in the case where
	// we read a dataset elements marked with action delete.
	private boolean emptySeriesOfEmptyObs;

	
	// Variable holding the current status of attributes at series level (SCALAR values in FAME)
	// false if no attributes at series level are present
	// true if attributes at series level are present
	private boolean isSeriesElementWithAttributesAtSeriesLevel = false; 
	
	// number of observations for one <Series> element
	private int obsCount;
	
	// number of completely empty <Obs> elements (invalid)
    private int invalidCompletelyEmptyObsCount;
    
    // number of observations with only attributes at obs level for one <Series> element
    private int obsOnlyAttrAtObsLevelCount;
		
    /* ====================== Constructors ====================== */

    /**
     * Constructs a new TimeSeriesReader.
     * 
     * @param parser The parser of the XML document
     * @param parent The class responsible for handling content
     */
    public TimeSeriesReader(final XMLReader parser, final SAX2Reader parent) {
		super(parser, parent);
        emptySeriesOfEmptyObs = false;
    }
    
	/**
     * Constructs a new TimeSeriesReader.
     * 
     * @param parser The parser of the XML document
     * @param parent The class responsible for handling content
     * @param emptify Whether to suppress observations from series consisting only of empty observations.
     */

    public TimeSeriesReader(final XMLReader parser, final SAX2Reader parent,
		final boolean emptify
	) {
		super(parser, parent);
        emptySeriesOfEmptyObs = emptify;
    }
    
    /* ===================== Public Methods ===================== */
   
   	public void setEmptySeriesOfEmptyObs(final boolean emptify) {
		emptySeriesOfEmptyObs = emptify;	
	}

    /**
     * {@inheritDoc} 
     */
    public void startElement (
        final String namespaceURI,
        final String localName,
        final String qName,
        final Attributes atts
    ) {
    	
        super.startElement( namespaceURI, localName, qName, atts );
		componentHelper.logger.log(
            	Level.FINEST,
            	componentHelper.logName,
            	this.getClass().getSimpleName(),
            	"startElement: " + localName 
        	);
        if ( "Series".equals( localName ) ) {
        	obsCount = 0;
        	invalidCompletelyEmptyObsCount = 0;
        	obsOnlyAttrAtObsLevelCount = 0;
            handleSeries(atts);
        } else if ( "Obs".equals( localName ) ) {
            try {
                handleObservation(atts);
            } catch (final ParseException e) {
                componentHelper.logger.log(
                        Level.SEVERE,
                        componentHelper.logName,
                        this.getClass().getSimpleName(),
                        "Could not parse date for observation: \""
                        + atts
                    );
            }
        } else {
            componentHelper.logger.log(
                    Level.SEVERE,
                    componentHelper.logName,
                    this.getClass().getSimpleName(),
                    "Unknown (for SMDX-ML) element : \""
                    + localName + "\""
                );
        }
    }

    /**
     * {@inheritDoc} 
     */
    public void endElement (
    	final String namespaceURI,
    	final String localName,
    	final String qName
    ) {
    	super.endElement( namespaceURI, localName, qName );
    	componentHelper.logger.log(
            	Level.FINEST,
            	componentHelper.logName,
            	this.getClass().getSimpleName(),
            	"endElement: " + localName 
        	);
    	
        if ("Series".equals( localName )) {
        	timeFormat=""; //reset value for next element
        	if ((obsCount != 0) && 
        			(obsCount != invalidCompletelyEmptyObsCount) && 
        			(obsCount != obsOnlyAttrAtObsLevelCount)) {
        		dispatchSeries();
            	componentHelper.logger.log(
                    	Level.FINEST,
                    	componentHelper.logName,
                    	this.getClass().getSimpleName(),
                    	"endElement:" + localName+ " -  dispatchSeries in scenario 1"
                	);
            } else {
            	if (emptySeriesOfEmptyObs) { //deletion message
            		dispatchSeries();
                	componentHelper.logger.log(
                        	Level.FINEST,
                        	componentHelper.logName,
                        	this.getClass().getSimpleName(),
                        	"endElement:" + localName+ " -  dispatchSeries in scenario 2"
                    	);
            	}
            }
        	
            // test empty series:
        	// no deletions and
        	// either no observation elements (<Obs>) at all
        	// or only empty observation elements for the current <series> element
        	// and no attributes at series level in <Series> element
            if (!emptySeriesOfEmptyObs && // no deletions
                (((obsCount == 0) || (obsCount!=0 && obsCount==invalidCompletelyEmptyObsCount)) && 
            	(!isSeriesElementWithAttributesAtSeriesLevel))) {
            	 componentHelper.logger.log(
                         Level.SEVERE,
                         componentHelper.logName,
                         this.getClass().getSimpleName(),
                         "Could not find any observations for empty series " + 
                         "with no series level (SCALAR) attributes: \""
                         + key + "\". ObsCount: " + obsCount
                     );
            }
            //reset values for next <series> element :
        	isSeriesElementWithAttributesAtSeriesLevel=false;

            finish();
        } else if ("Obs".equals( localName )) {
		} else {
            componentHelper.logger.log(
                Level.SEVERE,
                componentHelper.logName,
                this.getClass().getSimpleName(),
                "Unknown (for SMDX-ML) element : \""
                + localName + "\""
            );
        }
    }
    
    /* ==================== Private Methods ===================== */
        
    private void dispatchSeries() {
		try {
			if (observations.entrySet().isEmpty()) {
				series = new TimeSeries(key, (GregorianRegularInterval) null,
						new Observation[] {});
			} else {
				final Calendar calendar = CalendarHelper.createISOCalendar();
				calendar.setTime((Date) observations.lastKey());
				CalendarHelper.advanceCalendar(calendar, frequency, 1);

				final RegularInterval interval = new GregorianRegularInterval(
						frequency, (Date) observations.firstKey(),
						calendar.getTime());

				if (!emptySeriesOfEmptyObs || someObservationsAreNonEmpty()) {
					series = new TimeSeries(key, interval, observations);
				} else {
					series = new TimeSeries(key, interval, new Observation[] {});
				}
			}
		    dispatch(series);
		}
		catch (Exception e) {
            componentHelper.logger.log(
                    Level.SEVERE,
                    componentHelper.logName,
                    this.getClass().getSimpleName(),
                    "Could not parse date \""
                    + observations.firstKey() 
                    + "\" or \""
                    + observations.lastKey()
                    + "\"." 
			);
		}
	}

	private boolean someObservationsAreNonEmpty() {
		final Iterator iterator = observations.values().iterator();
		boolean nonEmptyObservationFound = false;
		while(iterator.hasNext()) {
			final Observation observation = (Observation) iterator.next();
			nonEmptyObservationFound = !observation.isEmpty();
			if (nonEmptyObservationFound) break;
		}
		return nonEmptyObservationFound;
	}
        
    private Observation distilObservation (final MapStack obsAttributes) {
        final String value = getObsValue(obsAttributes);
        final String status = getObsStatus(obsAttributes);
        final String conf = getObsConf(obsAttributes);
        final String preBreak = getObsPreBreak(obsAttributes);

        return new Observation(
            value == null ? ""
                : ( "NaN".equals( value ) ? "-" : value ),
            status == null ? "" : status,
            conf == null ? "" : conf,
            preBreak == null ? ""
                : ( "NaN".equals( preBreak ) ? "-" : preBreak )
        );
    }
    
    private void distilAndDispatchObservationAttributes (final Observation observation,
    		final MapStack xmlObsAttributes,
    		final Date period, String timeFormat) {
   
    	boolean isObsWithAttrAtObsLevel = false;
    	ArrayList conceptsToIgnoreForObservationLevelAttributes =
    			new ArrayList();
    	conceptsToIgnoreForObservationLevelAttributes.
    	   add(keyFamily.getObsValue().getConceptId());
    	conceptsToIgnoreForObservationLevelAttributes.
    	   add(keyFamily.getObsConf().getConceptId());
    	conceptsToIgnoreForObservationLevelAttributes.
    	   add(keyFamily.getObsPreBreak().getConceptId());
    	conceptsToIgnoreForObservationLevelAttributes.
    	   add(keyFamily.getObsStatus().getConceptId());
    	conceptsToIgnoreForObservationLevelAttributes.
 	       add(keyFamily.getTimePeriod().getConceptId());
    	
    	Object[] xmlObsAttributesArray = xmlObsAttributes.getStack().toArray();
        for (int i=0;i<xmlObsAttributesArray.length;i++) {
			TreeMap oneObservationLine = (TreeMap)xmlObsAttributesArray[i];
			
			Set attributeKeys = oneObservationLine.keySet();
			Iterator attributeKeysIterator = attributeKeys.iterator();
			
			while (attributeKeysIterator.hasNext()) {
				String attributeName = (String) attributeKeysIterator.next();
				String attributeValue = oneObservationLine.get(attributeName)
						.toString();

				if (!conceptsToIgnoreForObservationLevelAttributes
						.contains(attributeName)) {
					/**
					 * Dealing with an explicit attribute at observation level,
					 * which should be immediately dispatched
					 */
					dispatch(new Attribute(key,
							dateFormat.formatDateToInternalFormat(frequency,
									period), XMLDateFormat.xmlDateFormatToEDI
									.get(timeFormat).toString(), attributeName,
							attributeValue));
					isObsWithAttrAtObsLevel = true;
				}
			}
		} 
        obsCount++;
        if (!observation.isEmpty()) {	
        	observations.put( period, observation );
        } else { // empty observation (i.e. no OBS_VALUE, OBS_STATUS and OBS_CONF)
        	if (emptySeriesOfEmptyObs) { // deletions, we need our empty observations !!
        		observations.put( period, observation );
        	}
        	if (!isObsWithAttrAtObsLevel) { // no attributes at obs level
        		invalidCompletelyEmptyObsCount++;
        	} else {
        		obsOnlyAttrAtObsLevelCount++;
        	}
        }
        
        
    } // end distilAndDispatchObservationAttributes
    
    private String handleSeries (final Attributes atts) {
    	// Firstly assume only attributes at obs level for one series with no observations
    	observations = new TreeMap(); //reset observations !!
        final MapStack xmlSeriesAtts = readXMLAttributes( atts );

        final String timeFormatAttribute = getTimeFormat( xmlSeriesAtts );
		final String timeFormat = ( null == timeFormatAttribute )
			? inferTimeFormatFromFrequency( xmlSeriesAtts )
			: timeFormatAttribute
			;
		this.timeFormat = timeFormat;
        final SDMXHelper helper = new SDMXHelper(timeFormat);
        frequency = helper.getFrequency();
        dateFormat = new XMLDateFormat();
        
        key = distilKey(xmlSeriesAtts);
        
        if ( key != null ) {
            for (
                final Iterator iterator = keyFamily.attributeIterator(
                        KeyFamilyComponent.ATTACHED_TO_TIME_SERIES);
                iterator.hasNext();
            ) {
                final KeyFamilyComponent component
                = (KeyFamilyComponent) iterator.next();
                final String value = 
                    xmlSeriesAtts.get( component.getConceptId() );
                if ( value != null ) {
                    componentHelper.logger.log(
                            Level.FINEST,
                            componentHelper.logName,
                            this.getClass().getSimpleName(),
                            "Dispatching attributes \""
                            + component.getConceptId() 
                            + "\" with value \"" 
                            + value + "\""
                    );
                   
                    dispatch(new Attribute(key, 
                    			(GregorianRegularInterval) null, 
                    			component.getConceptId(), value));
                    isSeriesElementWithAttributesAtSeriesLevel = true;
                }
            }
        }
        return timeFormat;
    }
    
    private void handleObservation (final Attributes atts)
			throws ParseException {
    	final MapStack xmlObsAttributes = readXMLAttributes(atts);
        final String period = keyFamily.getTimePeriod() == null
            ? null: xmlObsAttributes.get(
                keyFamily.getTimePeriod().getConceptId()
            );
            
        if (null != period) {
        	Date date;
        	try {
    			date = dateFormat.parse(period);					
        	} catch ( ParseException e ) {
				componentHelper.logger.log(
                	Level.INFO,
                	componentHelper.logName,
                	this.getClass().getSimpleName(),
                	"Unparsable period: \"" 
                	+ period + "\"." 
            	);
				throw (e);
			}	
        	Observation observation;
        	observation = distilObservation(xmlObsAttributes);
			
			distilAndDispatchObservationAttributes(observation, xmlObsAttributes, date, timeFormat);

        } else {
            componentHelper.logger.log(
                Level.SEVERE,
                componentHelper.logName,
                this.getClass().getSimpleName(),
                "Could not find TIME_PERIOD for observation for key \""
                + key + "\"." 
            );
        }
    }
    
    private String getObsValue (final MapStack obsAttributes) {
    	return keyFamily.getObsValue() == null
            ? null: obsAttributes.get(
                keyFamily.getObsValue().getConceptId()
            );
    }
    
    private String getObsStatus (final MapStack obsAttributes){
    	return keyFamily.getObsStatus() == null
    	    ? null : obsAttributes.get(
    	        keyFamily.getObsStatus().getConceptId()
    	    );
    }
    
    private String getObsConf (final MapStack obsAttributes) {
    	return keyFamily.getObsConf() == null
    	    ? null : obsAttributes.get(
	        keyFamily.getObsConf().getConceptId()
    	    );
    }
    
    private String getObsPreBreak (final MapStack obsAttributes) {
    	return keyFamily.getObsPreBreak() == null
	    ? null : obsAttributes.get(
	        keyFamily.getObsPreBreak().getConceptId()
	    );
    }

	private String getTimeFormat (final MapStack xmlSeriesAtts) {

		final KeyFamilyComponent timeFormatAttribute
			= keyFamily.getTimeFormat();
		
		if ( null == timeFormatAttribute ) { 
        	componentHelper.logger.log(
                    Level.INFO,
                    componentHelper.logName,
                    this.getClass().getSimpleName(),
                    "The keyFamily \"" + keyFamily.toString()
					+ "\" does not define any time dimension."
            );
			return null;
		}
		
		final String timeFormat = xmlSeriesAtts.get(
			timeFormatAttribute.getConceptId()
		);

		return timeFormat;	
	}

	private String inferTimeFormatFromFrequency (final MapStack xmlSeriesAtts) {

		final KeyFamilyComponent frequencyDimension
			= keyFamily.getFrequency();
			
		if ( null == frequencyDimension ) {
        	componentHelper.logger.log(
                    Level.INFO,
                    componentHelper.logName,
                    this.getClass().getSimpleName(),
                    "The keyFamily \"" + keyFamily.toString()
					+ "\" does not define any frequency dimension."
            );
			return null;
		}
		
		final String frequencyCode = xmlSeriesAtts.get(
			frequencyDimension.getConceptId()
		);

		if ( null == frequencyCode ) {
			debugAttributes( xmlSeriesAtts );
			return null;
		}

		final int frequency =
				( "A".equals( frequencyCode ) ) ? RegularInterval.ANNUAL
			:	( "H".equals( frequencyCode ) ) ? RegularInterval.SEMIANNUAL
			:	( "S".equals( frequencyCode ) ) ? RegularInterval.SEMIANNUAL
			:	( "Q".equals( frequencyCode ) ) ? RegularInterval.QUARTERLY
			:	( "M".equals( frequencyCode ) ) ? RegularInterval.MONTHLY
			:	( "W".equals( frequencyCode ) ) ? RegularInterval.WEEKLY
			:	( "D".equals( frequencyCode ) ) ? RegularInterval.DAILY
			:	( "N".equals( frequencyCode ) ) ? RegularInterval.MINUTELY
			:	-1
			;
		
    	return XMLDateFormat.formatDuration( frequency );
	}

	private void debugAttributes( final MapStack xmlSeriesAtts ) {
		final Iterator it = xmlSeriesAtts.topEntryIterator();
		for( ; it.hasNext(); ) {
			final Map.Entry e = (Map.Entry) it.next();
			componentHelper.logger.log(
       	    	Level.INFO,
       	    	componentHelper.logName,
       	    	this.getClass().getSimpleName(),
				"Series: " + e.toString()
			);
		}	
	}

}
