// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2002, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
package Int.ecb.gesmescb.xml.sdmx.v2.compact;

import java.text.DateFormat;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.Iterator;
import java.util.Date;
import java.util.ArrayList;

import junit.framework.Test;
import junit.framework.TestSuite;

import org.xml.sax.SAXException;

import Int.ecb.gesmescb.Attribute;
import Int.ecb.gesmescb.Observation;
import Int.ecb.gesmescb.TimeSeries;
import Int.ecb.gesmescb.metabase.Metabase;
import Int.ecb.gesmescb.time.GregorianRegularInterval;
import Int.ecb.gesmescb.time.RegularInterval;
import Int.ecb.gesmescb.xml.SAX2Reader;
import Int.ecb.gesmescb.xml.SAX2ReaderTest;

public class TimeSeriesReaderTest extends SAX2ReaderTest {
    
    private final DateFormat dailyDateFormat;
    private final DateFormat monthlyDateFormat;
    private final DateFormat weeklyDateFormat;
    private final DateFormat annualDateFormat;
    private final DateFormat minutelyDateFormat;
    private final DateFormat quarterlyDateFormat;
    private final DateFormat semiannualDateFormat;
	
    /* ====================== Constructors ====================== */
    
	public TimeSeriesReaderTest(String name) {
		super(name);
		dailyDateFormat = new SimpleDateFormat("yyyy-MM-dd");
		monthlyDateFormat = new SimpleDateFormat("yyyy-MM");
		weeklyDateFormat = new SimpleDateFormat("yyyy-MM-dd");
		annualDateFormat = new SimpleDateFormat("yyyy");
		minutelyDateFormat = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ssZ");
		quarterlyDateFormat = new QuarterlyFormatter("yyyyMM");
		semiannualDateFormat = new SemiannualFormatter("yyyyM");
	}
    
    /* ================== Protected Methods ===================== */

    /** Sets up the test fixture. */
    protected void setUp () {
        super.setUp();
        //setNamespace( XMLNS_GESMES, null );
        final Metabase metabase = Metabase.getMetabase();
        metabase.handle( KF );
        /*System.setProperty(
            KeyFamilyURIMapping.class.getName() + ".uri.ECB.KF",
            "http://www.ecb.int/vocabulary/2003-11-10/kf"
        );*/
    }

    /** Tears down the test fixture. */
    protected void tearDown () {
        /*System.getProperties().remove(
            KeyFamilyURIMapping.class.getName() + ".uri.ECB.KF"
        );*/
    }

    protected SAX2Reader factory (
        final org.xml.sax.XMLReader parser,
        final SAX2Reader parent
    ) {
        final TimeSeriesReader reader
            = new TimeSeriesReader( parser, parent );
        reader.setKeyFamily( KF );
       	reader.setEmptySeriesOfEmptyObs(false);
		return reader;

    }

	/* ===================== Public Methods ===================== */

	/**
	 * Assembles and returns a test suite for all
	 * the test methods of this test case.
	 *
	 * @return a non-null test suite
	 */
	public static Test suite () {
		final TestSuite suite = new TestSuite( TimeSeriesReaderTest.class );
		return suite;
	}

	/**
	 * @param arg the array of argument {@link String}s
	 */
	public static void main ( final String[] arg ) {
		final String[] testCaseName = {
			TimeSeriesReaderTest.class.getName()
		};
		junit.textui.TestRunner.main( testCaseName );
	}

	/* ---------------------- Test Methods ---------------------- */
    
    public void testGetEmptyTimeSeries() throws SAXException
    {
        openElement("Series", "FREQ=D ITEM=AUD TIME_FORMAT=P3M");
        closeElement("Series");
        final Iterator iterator = listener.iterator();
        assertTrue("At least one object", !iterator.hasNext());
    }
	
    public void testGetEmptyTimeSeriesWithAttributes() throws SAXException {
        openElement("Series", "FREQ=D ITEM=AUD TS_ATT=BB TIME_FORMAT=P3M");
        closeElement("Series");
        final Iterator iterator = listener.iterator();
        assertTrue("At least one object", iterator.hasNext());
        final Attribute attribute = (Attribute) iterator.next();
     
        final Attribute expectedAttribute = new Attribute("D.AUD", 
                (GregorianRegularInterval) null, "TS_ATT", "BB");
        final TimeSeries expectedSeries = new TimeSeries(
                "D.AUD",
                (GregorianRegularInterval) null,
                new Observation[]{}
        );
        assertEquals("The attributes should be equal", expectedAttribute, 
                attribute);
    }
    
    public void testGetTimeSeriesWithDailyObservations() throws SAXException, NullPointerException, ParseException
    {
        openElement("Series", "FREQ=D ITEM=AUD TIME_FORMAT=P1D");
        openElement("Obs", "TIME_PERIOD=2008-01-04 OBS_VALUE=1.36 OBS_STATUS=N OBS_CONF=C");
        closeElement("Obs");
        openElement("Obs", "TIME_PERIOD=2008-01-05 OBS_VALUE=1.45 OBS_STATUS=N OBS_CONF=C OBS_PRE_BREAK=2.45");
        closeElement("Obs");
        closeElement("Series");
        final Iterator iterator = listener.iterator();
        assertTrue("At least one object", iterator.hasNext());
        final TimeSeries series = (TimeSeries) iterator.next();
        final Observation[] observations = {
                new Observation( "1.36", "N", "C", "" ),
                new Observation( "1.45", "N", "C", "2.45" ),
            };
        final TimeSeries expectedSeries = new TimeSeries("D.AUD",
                new GregorianRegularInterval(
                        RegularInterval.DAILY, 
                        dailyDateFormat.parse("2008-01-04"), 
                        dailyDateFormat.parse("2008-01-06")),
                observations
        );
        assertEquals("The series should be equal", expectedSeries, series);
        assertTrue("Should be only one", !iterator.hasNext());
    }
    
    public void testGetTimeSeriesWithMonthlyObservations() throws SAXException, NullPointerException, ParseException
    {
        openElement("Series", "FREQ=M ITEM=AUD TIME_FORMAT=P1M");
        openElement("Obs", "TIME_PERIOD=2008-01 OBS_VALUE=1.36 OBS_STATUS=N OBS_CONF=C");
        closeElement("Obs");
        openElement("Obs", "TIME_PERIOD=2008-02 OBS_VALUE=1.45 OBS_STATUS=N OBS_CONF=C OBS_PRE_BREAK=2.45");
        closeElement("Obs");
        closeElement("Series");
        final Iterator iterator = listener.iterator();
        assertTrue("At least one object", iterator.hasNext());
        final TimeSeries series = (TimeSeries) iterator.next();
        final Observation[] observations = {
                new Observation( "1.36", "N", "C", "" ),
                new Observation( "1.45", "N", "C", "2.45" ),
            };
        final TimeSeries expectedSeries = new TimeSeries("M.AUD",
                new GregorianRegularInterval(
                        RegularInterval.MONTHLY, 
                        monthlyDateFormat.parse("2008-01"), 
                        monthlyDateFormat.parse("2008-03")),
                observations
        );
        assertEquals("The series should be equal", expectedSeries, series);
        assertTrue("Should be only one", !iterator.hasNext());
    }
    
    public void testGetTimeSeriesWithObservationsAndHoles() throws SAXException, NullPointerException, ParseException
    {
        openElement("Series", "FREQ=D ITEM=AUD TIME_FORMAT=P1D");
        openElement("Obs", "TIME_PERIOD=2008-01-04 OBS_VALUE=1.36 OBS_STATUS=N OBS_CONF=C");
        closeElement("Obs");
        openElement("Obs", "TIME_PERIOD=2008-01-06 OBS_VALUE=1.45 OBS_STATUS=N OBS_CONF=C OBS_PRE_BREAK=2.45");
        closeElement("Obs");
        closeElement("Series");
        final Iterator iterator = listener.iterator();
        assertTrue("At least one object", iterator.hasNext());
        final TimeSeries series = (TimeSeries) iterator.next();
        final Observation[] observations = {
                new Observation( "1.36", "N", "C", "" ),
                Observation.EMPTY_OBSERVATION,
                new Observation( "1.45", "N", "C", "2.45" ),
            };
        final TimeSeries expectedSeries = new TimeSeries("D.AUD",
                new GregorianRegularInterval(
                        RegularInterval.DAILY, 
                        dailyDateFormat.parse("2008-01-04"), 
                        dailyDateFormat.parse("2008-01-07")),
                observations
        );
        assertEquals("The series should be equal", expectedSeries, series);
        assertTrue("Should be only one", !iterator.hasNext());
    }
	/*public void testGetObservation() throws SAXException {
        openElement( "Obs", "TIME_PERIOD=1998-10 OBS_VALUE=54955 OBS_STATUS=A OBS_CONF=C" );
        closeElement("Obs");
        final Iterator iterator = listener.iterator();
        assertTrue("At least one obs", iterator.hasNext());
        assertEquals("Correct observation", new Observation("54955", "A", "C", ""), iterator.next());
        assertTrue("Should be only one", !iterator.hasNext());
	}*/
    
    public void testEmptyAndMissingObservations()
        throws SAXException, NullPointerException, ParseException
    {
    	  openElement("Series", "FREQ=D ITEM=AUD TIME_FORMAT=P1D");
          openElement("Obs", "TIME_PERIOD=2008-01-04 OBS_VALUE=NaN OBS_STATUS=N OBS_CONF=C");
          closeElement("Obs");
          openElement("Obs", "TIME_PERIOD=2008-01-05");
          closeElement("Obs");
          closeElement("Series");
          final Iterator iterator = listener.iterator();
          assertTrue("At least one object", iterator.hasNext());
          final TimeSeries series = (TimeSeries) iterator.next();
          final Observation[] observations = {
              new Observation( Observation.MISSING_VALUE, "N", "C", "" ),
              Observation.EMPTY_OBSERVATION,
          };
          final TimeSeries expectedSeries = new TimeSeries("D.AUD",
              new GregorianRegularInterval(
                  RegularInterval.DAILY, 
                  dailyDateFormat.parse("2008-01-04"), 
                  dailyDateFormat.parse("2008-01-05")),
              observations
          );
          assertTrue("Should be only one", !iterator.hasNext());
         
    }

	// test series with only empty obs get empty wehn deletion

    public void testDumpEmptyObs()
        throws SAXException, NullPointerException, ParseException
    {

		TimeSeriesReader reader = (TimeSeriesReader) getReader(); 
		reader.setEmptySeriesOfEmptyObs(true);

        openElement("Series", "FREQ=D ITEM=AUD TIME_FORMAT=P1D");
        openElement("Obs", "TIME_PERIOD=2008-01-04");
        closeElement("Obs");
        openElement("Obs", "TIME_PERIOD=2008-01-05");
        closeElement("Obs");
        closeElement("Series");
		
        final Iterator iterator = listener.iterator();
        assertTrue("At least one object", iterator.hasNext());
    }

	// test without time_format

    public void testGetTimeSeriesWithoutTimeFormatMonthly() throws SAXException, NullPointerException, ParseException {
		privateTestWFrequency(
			"M", "2013-11", "2013-12", "2014-01",
			monthlyDateFormat,
			RegularInterval.MONTHLY 
		);
	}
    public void testGetTimeSeriesWithoutTimeFormatAnnual() throws SAXException, NullPointerException, ParseException {
		privateTestWFrequency(
			"A", "2013", "2014", "2015",
			annualDateFormat,
			RegularInterval.ANNUAL 
		);
	}
    public void testGetTimeSeriesWithoutTimeFormatMinutely() throws SAXException, NullPointerException, ParseException {
		privateTestWFrequency(
			"N",
			"2013-01-29T14:21:00+0100",
			"2013-01-29T14:22:00+0100",
			"2013-01-29T14:23:00+0100",
			minutelyDateFormat,
			RegularInterval.MINUTELY 
		);
	}
    public void testGetTimeSeriesWithoutTimeFormatDaily() throws SAXException, NullPointerException, ParseException {
		privateTestWFrequency(
			"D", "2013-01-29", "2013-01-30", "2013-01-31",
			dailyDateFormat,
			RegularInterval.DAILY 
		);
	}
    public void testGetTimeSeriesWithoutTimeFormatWeekly() throws SAXException, NullPointerException, ParseException {
		privateTestWFrequency(
			"W", "2013-01-06", "2013-01-13", "2013-01-20",
			weeklyDateFormat,
			RegularInterval.WEEKLY 
		);
	}
    public void testGetTimeSeriesWithoutTimeFormatSemiannual() throws SAXException, NullPointerException, ParseException {
		privateTestWFrequency(
			"S", "2012-S2", "2013-S1", "2013-S2",
			semiannualDateFormat,
			RegularInterval.SEMIANNUAL 
		);
	}
    public void testGetTimeSeriesWithoutTimeFormatQuarterly() throws SAXException, NullPointerException, ParseException {
		privateTestWFrequency(
			"Q", "2012-Q4", "2013-Q1", "2013-Q2",
			quarterlyDateFormat,
			RegularInterval.QUARTERLY 
		);
	}

	public void testTimeFormatHasPrecedenceOverFreq() throws SAXException, NullPointerException, ParseException {
	
        openElement("Series", "FREQ=M ITEM=AUD TIME_FORMAT=P1Y");

        openElement("Obs", "TIME_PERIOD=2012 OBS_VALUE=1.36 OBS_STATUS=N OBS_CONF=C");
        closeElement("Obs");
        openElement("Obs", "TIME_PERIOD=2013 OBS_VALUE=1.45 OBS_STATUS=N OBS_CONF=C OBS_PRE_BREAK=2.45");
        closeElement("Obs");
        closeElement("Series");
        final Iterator iterator = listener.iterator();
        assertTrue("At least one object", iterator.hasNext());
        final TimeSeries series = (TimeSeries) iterator.next();
        final Observation[] observations = {
                new Observation( "1.36", "N", "C", "" ),
                new Observation( "1.45", "N", "C", "2.45" ),
            };
        final TimeSeries expectedSeries = new TimeSeries("M.AUD",
                new GregorianRegularInterval(
                        RegularInterval.ANNUAL,
                        annualDateFormat.parse("2012"), 
                        annualDateFormat.parse("2014")),
                observations
        );
        assertEquals("The series should be equal", expectedSeries, series);
	}

	private void privateTestWFrequency(
		String freq,
		String first,
		String second,
		String third,
		DateFormat formatter,
		int interval
	) throws SAXException, NullPointerException, ParseException {
	
        openElement("Series", "FREQ="+freq+" ITEM=AUD");

        openElement("Obs", "TIME_PERIOD="+first+" OBS_VALUE=1.36 OBS_STATUS=N OBS_CONF=C");
        closeElement("Obs");
        openElement("Obs", "TIME_PERIOD="+second+" OBS_VALUE=1.45 OBS_STATUS=N OBS_CONF=C OBS_PRE_BREAK=2.45");
        closeElement("Obs");
        closeElement("Series");
        final Iterator iterator = listener.iterator();
        assertTrue("At least one object", iterator.hasNext());
        final TimeSeries series = (TimeSeries) iterator.next();
        final Observation[] observations = {
                new Observation( "1.36", "N", "C", "" ),
                new Observation( "1.45", "N", "C", "2.45" ),
            };
        final TimeSeries expectedSeries = new TimeSeries(freq+".AUD",
                new GregorianRegularInterval(
                        interval,
                        formatter.parse(first), 
                        formatter.parse(third)),
                observations
        );
        assertEquals("The series should be equal", expectedSeries, series);
		assertTrue("Should be only one", !iterator.hasNext());
	}

	class QuarterlyFormatter extends SimpleDateFormat {
		QuarterlyFormatter ( String format ) {
			super(format);
		}
		public Date parse( final String quarter ) throws NullPointerException, java.text.ParseException {
			final String parseableQuarter =
				"2012-Q4".equals(quarter) ? "201210"
			: 	"2013-Q1".equals(quarter) ? "201301"
			: 	"2013-Q2".equals(quarter) ? "201304"
			:   "UtterlyWrongQuarter"
			;
			return super.parse(parseableQuarter);
		}
	}

	class SemiannualFormatter extends SimpleDateFormat {
		SemiannualFormatter ( String format ) {
			super(format);
		}
		public Date parse( final String semester ) throws NullPointerException, java.text.ParseException {
			final String parseableSemester =
				"2012-S2".equals(semester) ? "201207"
			: 	"2013-S1".equals(semester) ? "201301"
			: 	"2013-S2".equals(semester) ? "201307"
			:   "UtterlyWrongSemester"
			;
			return super.parse(parseableSemester);
		}
	}
}
