// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.logging;

import Int.ecb.logging.Level;
import Int.ecb.logging.LogRecord;
import Int.ecb.logging.Filter;
import Int.ecb.logging.Formatter;

/**
 * <P>Takes log messages from a Logger and exports them.</P>
 *
 * <P>It might for example, write them to a console or write them to a file,
 * or send them to a network logging service, or forward them to an OS log,
 * or whatever. A Handler can be disabled by doing a setLevel(Level.OFF) and
 * can be re-enabled by doing a setLevel with an appropri-ate level. Handler
 * classes typically use LogManager properties to set default values for Filter,
 * Formatter, and Level. See the specific documentation for each concrete
 * Handler class.</P>
 *
 * @author
 * <A HREF="mailto:Sami.Airo@ECB.int">Sami Airo</A>
 *
 * @version @(#)$Id: Handler.java,v 1.1 2017/01/19 16:14:05 trainot Exp $
 */
public abstract class Handler {

	private Level level;
	private Filter filter;
	private Formatter formatter;

	/**
	 * Default constructor.
	 * The resulting Handler has a log level of ALL, no formatter,
	 * and no filter.
	 */
	protected Handler() {
		level = Level.ALL;
	}


	/**
	 * Construct a Handler with the given formatter.
	 * @param	formatter	Formatter to be used to format output.
	 */
	protected Handler(Formatter formatter) {
		this();
		this.formatter = formatter;
	}


	/**
	 * Get the log level specifying which messages will be logged by this Handler.
	 * Message levels lower than this level will be discarded.
	 *
	 * @return		the level of messages being logged.
	 */
	public Level getLevel() {
		return level;
	}


	/**
	 * Return the Formatter for this Handler.
	 *
	 * @return		the formatter (may be null).
	 */
	public Formatter getFormatter() {
		return formatter;
	}


	 /**
	 * Close the Handler and free all associated resources.
	 * The close method will perform a "flush" and then close the Handler.
	 * After close has been called this Handler should no longer be used.
	 * Method calls may either be silently ignored or may throw runtime
	 * exceptions.
	 *
	 * @throws	SecurityException	if a security manager exists and if the
	 *																 caller does not have LoggingPermission("control").
	 */
	 public abstract void close();


	/**
	 * Publish a LogRecord.
	 * The logging request was made initially to a Logger object, which initialized the
	 * LogRecord and forwarded it here.
	 *
	 * The Handler is responsible for formatting the message, when and if necessary.
	 * The formatting should include local-ization.
	 *
	 * @param record		description of the log event
	 */
	public abstract void publish(LogRecord record);


	/**
	 * Set the log level specifying which message levels will be logged by this
	 * handler. Message levels lower than this value will be discarded.
	 * The intention is to allow developers to turn on voluminous logging, but to
	 * limit the messages that are sent to certain Handlers.
	 *
	 * @param newLevel	 the new value for the log level
	 */
	public void setLevel(Level newLevel) {
		level = newLevel;
	}


	/**
	 * Check if this Handler would actually log a given LogRecord.
	 * This method checks if the LogRecord has an appropriate level and whether it
	 * satisfies any Filter. It also may make other Handler specific checks that
	 * might prevent a handler from logging the LogRecord.
	 */
	public boolean isLoggable(LogRecord record) {
		if (level.intValue() <= record.getLevel().intValue()) {
			if (filter != null) {
				return filter.isLoggable(record);
			} else {
				return true;
			}
		} else {
			return false;
		}
	}


	/**
	 * Set a filter to control output on this Handler.
	 * For each call of "publish" the Handler will call this Filter (if it is
	 * non-null) to check if the log record should be published or discarded.
	 *
	 * @param newFilter a filter object (may be null)
	 * @throws SecurityException
	 * if a security manager exists and if the
	 * caller does not have LoggingPermission("control").
	 */
	public void setFilter(Filter newFilter) {
		filter = newFilter;
	}

	/**
	 * Set a Formatter.
	 * This formatter will be used to format LogRecords for this Handler.
	 * Some Handlers may not use Formatters,
	 * in which case the formatter will be remembered, but not used.
	 *
	 * @param newFormatter the formatter to use (may not be null)
	 */
	public void setFormatter(Formatter newFormatter) {
		formatter = newFormatter;
	}
}
