// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.logging;

import java.util.HashMap;

/**
 * <P>Defines a set of standard logging levels that can be used to
 * control logging output.</P>
 *
 * <P>The logging Level objects are ordered and are
 * specified by ordered integers. Enabling logging at a given level also enables
 * logging at all higher levels.</P>
 *
 * <P>Clients should normally use the predefined Level constants
 * such as {@link #SEVERE}.</P>
 *
 * @author
 * <A HREF="mailto:Sami.Airo@ECB.int">Sami Airo</A>
 *
 * @version @(#)$Id: Level.java,v 1.1 2017/01/19 16:14:05 trainot Exp $
 */
public class Level { //implements java.io.Serializable { TODO

	private final java.lang.String name;
	private final int value;
	private static HashMap levels = new HashMap(9);

	public static final Level OFF;
	public static final Level SEVERE;
	public static final Level WARNING;
	public static final Level INFO;
	public static final Level CONFIG;
	public static final Level FINE;
	public static final Level FINER;
	public static final Level FINEST;
	public static final Level ALL;


	static {
		OFF = new Level("OFF", 1000);		 levels.put("OFF", OFF);			 levels.put("1000", OFF);

		SEVERE = new Level("SEVERE", 100);
		levels.put("SEVERE", SEVERE);
		levels.put("100", SEVERE);

		WARNING = new Level("WARNING", 90);
		levels.put("WARNING", WARNING);
		levels.put("90", WARNING);

		INFO = new Level("INFO", 80);
		levels.put("INFO", INFO);
		levels.put("80", INFO);

		CONFIG = new Level("CONFIG", 70);
		levels.put("CONFIG", CONFIG);
		levels.put("70", CONFIG);

		FINE = new Level("FINE", 60);
		levels.put("FINE", FINE);
		levels.put("60", FINE);

		FINER = new Level("FINER", 50);
		levels.put("FINER", FINER);
		levels.put("50", FINER);

		FINEST = new Level("FINEST", 40);
		levels.put("FINEST", FINEST);
		levels.put("40", FINEST);

		ALL = new Level("ALL", 0);
		levels.put("ALL", ALL);
		levels.put("0", ALL);
	}


	protected Level(java.lang.String name, int value) {
		this.name = name;
		this.value = value;
	}


	/**
	 * Get the integer value for this level. This integer value can be used for
	 * efficient ordering comparisons between Level objects.
	 *
	 * @return the integer value for this level.
	 */
	public final int intValue() {
		return value;
	}


	/**
	 * Parse a level name string into a Level.
	 * The argument string may consist of either a level name or an integer value.
	 * For example:
	 * <UL><LI>"SEVERE"
	 * <LI>"1000"</UL>
	 * @param name of level to be parsed
	 * @return parsed value
	 * @throws NullPointerException if the name is null
	 * @throws IllegalArgumentException
	 * if the value is neither one of the known names nor an integer.
	 */
	public static Level parse ( final String name ) {
		if (name == null) {
			throw new NullPointerException();
		} else {
			Level level = (Level)levels.get(name);

			if (level == null) {
				throw new IllegalArgumentException();
			} else {
				return level;
			}
		}
	}


	/**
	 * @return the string name of the Level, for example "INFO".
	 */
	public final java.lang.String toString() {
		return name;
	}
}
