// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.logging;

import java.lang.Throwable;
import java.lang.System;
import java.lang.String;

import java.io.InputStream;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.FileNotFoundException;

import java.util.Map;
import java.util.HashMap;
import java.util.Set;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Properties;
import java.util.StringTokenizer;

import Int.ecb.logging.Logger;
import Int.ecb.logging.Handler;
import Int.ecb.logging.LogRecord;

/* ***************************** Class ****************************** */

/**
 * <P>Provides is a singleton to maintain a set of
 * shared state about Loggers and log services.</P>
 *
 * <P>This LogManager object:
 * <UL>
 * <LI>Manages a hierarchical namespace of Logger objects. All named Loggers are
 * stored in this namespace.
 * <LI> Manages a set of logging control properties. These are simple key-value
 * pairs that can be used by Handlers and other logging objects to configure themselves.
 * <LI> Manages a list of "global" logging handlers. By default Loggers will publish
 * all (relevant) LogRecords to all the global handlers.
 * </UL></P>
 *
 * <P>The global LogManager object can be retrieved using
 * {@link #getLogManager}.
 * The LogManager object is created during class initialization
 * and cannot subsequently be changed.</P>
 *
 * @author
 * <A HREF="mailto:Sami.Airo@ECB.int">Sami Airo</A>
 *
 * @version @(#)$Id: LogManager.java,v 1.1 2017/01/19 16:14:05 trainot Exp $
 */
public class LogManager {

	/* ========================= Fields ========================= */

	/* --------------------- Static Fields ---------------------- */

	private static final LogManager manager = new LogManager();

	// And the static initializer:
	static {
		manager.readConfiguration();
	}

	/* -------------------- Instance Fields --------------------- */

	private final String className = getClass().getName();
	private final Map loggers;
	private final Set handlers;
	private final Properties properties;

	/* ====================== Constructors ====================== */

	protected LogManager () {
		handlers = new HashSet();
		loggers = new HashMap();
		properties = new Properties();

		// add the root logger
		//addLogger(Logger.getLogger(""));
	}


	protected void finalize() throws Throwable {
		try {
			Iterator handlerIterator = handlers.iterator();

			while (handlerIterator.hasNext()) {
				((Handler)handlerIterator.next()).close();
			}
		} finally {
			super.finalize();
		}
	}

	/* ===================== Public Methods ===================== */

	/**
	 * Add a global log Handler to receive output from all loggers.
	 *
	 * @param handler output Handler
	 *
	 * @throws SecurityException
	 * if a security manager exists and if the caller does not have
	 * LoggingPermission("control").
	 */
	public void addGlobalHandler ( Handler handler ) throws SecurityException {
		handlers.add(handler);
	}

	/**
	 * Remove a global log Handler to no longer receive output
	 * from all loggers.
	 *
	 * @param handler to be removed
	 */
	public void removeGlobalHandler (
		final Handler handler
	) throws SecurityException {
		handlers.remove( handler );
	}

	/**
	 * Return the global LogManager object.
	 */
	public static LogManager getLogManager () {
		return manager;
	}


	/**
	 * Add a named logger.
	 * This does nothing and returns false if a logger with the
	 * same name is already registered.
	 *
	 * @param l the new logger.
	 * @return true if the argument logger was registered successfully, false
	 * if a logger of that name already exists.
	 * @throws NullPointerException if the logger name is null.
	 */
	public boolean addLogger ( Logger l ) throws NullPointerException {

		if (loggers.containsKey(l.getName())) {
			return false;
		} else {
			loggers.put(l.getName(), l);
			return true;
		}
	}


	/**
	 * Method to find a named logger.
	 *
	 * @param name name of the logger
	 * @return matching logger or null if none is found
	 */
	public Logger getLogger ( String name ) {
		return (Logger)loggers.get(name);
	}


	/**
	 * Get the value of a logging property.
	 * @param	name property name
	 * @return				property value
	 */
	public String getProperty ( String name ) {
		return properties.getProperty(name);
	}


	/**
	 * Get the log level for a given logger name. See the setLevel method
	 * for more details.
	 *
	 * @param name	The logger name
	 *
	 * @return	the log level
	 */
	public Level getLevel ( java.lang.String name ) {
		String level;
		Logger logger = getLogger(name);

		// first check if we have the logger already
		if (logger == null) {
			// Logger is not yet known. Check if we have level specified
			level = properties.getProperty(name + ".level");

			if (level == null) {
				// level is not specified. Check if we are at the root already
				if (0 < name.length()) {
					String parent = "";
					// not at the root yet. Look up the parent name if we are no just below root.
					if (-1 < name.lastIndexOf(".")) {
						parent = name.substring(0, name.lastIndexOf("."));
					}
					level = getLevel(parent).toString();
				} else {
					// we are at the root. Return a default value
					level = "INFO";
				}
			}
		} else {
			level = logger.getLevel().toString();
		}

		return Level.parse(level);
	}


	/**
	 * Get an array of all the global Handlers.
	 *
	 * @return array of global Handlers
	 * @throws SecurityException
	 * if a security manager exists and if the caller
	 * does not have LoggingPermission("control").
	 */
	public Handler[] getGlobalHandlers () throws SecurityException {
		return (Handler[])handlers.toArray(new Handler[0]);
	}


	/**
	 * Set a log level for a given set of loggers.
	 * Subsequently the target loggers
	 * will only log messages whose types are greater than or equal to the given level.
	 * The level value Level.OFF can be used to turn off logging.
	 *
	 * The empty string "" can be used to identify the root.
	 *
	 * @param name The logger name to update.
	 * @param level the new log level, e.g. Level.SEVERE or Level.FINER
	 * @return the previous value of the level
	 * @throws SecurityException
	 * if a security manager exists and if the caller does not have
	 * LoggingPermission("control").
	 */
	public static Level setLevel ( String name, Level level ) throws SecurityException {
		Iterator i = getLogManager().loggers.values().iterator();

		while (i.hasNext()) {
			Logger l = (Logger)i.next();

			if ( name.equals("") || l.getName().equals(name) || l.getName().startsWith(name + ".") ) {
				l.setLevel(level);
			}
		}

		return level; //TODO
	}


	/**
	 * Method to post a log record to all global handlers.
	 * This is used by Loggers when they wish to publish a LogRecord to all the global handlers.
	 */
	public void publish ( LogRecord record ) {
		Iterator handlerIterator = handlers.iterator();

		while (handlerIterator.hasNext()) {
			((Handler)handlerIterator.next()).publish(record);
		}
	}


	/*
	 * Reinitialize the logging properties and reread the logging configuration.
	 * The same rules are used for locating the configuration properties as are
	 * used at startup. So normally the logging properties will be re-read from
	 * the same file that was used at startup.
	 *
	 * All existing global Handlers will be closed and a new set of global
	 * Handlers will be created based on the new properties.
	 *
	 * Any log level definitions in the new configuration file will be applied
	 * using LogManager.setLevel() in the order they are read.
	 *
	 * @throws SecurityException
	 * if a security manager exists and if the
	 * caller does not have LoggingPermission("control").
	 */
	public void readConfiguration () throws SecurityException {
		InputStream ins = null;

		String reader = System.getProperty("Int.ecb.logging.config.class");

		if (reader == null ) {
			String fileName = System.getProperty("Int.ecb.logging.config.file");

			if (fileName == null) {
				fileName = "logging.properties";
			}

			try {
				ins = new FileInputStream(fileName);
			} catch (FileNotFoundException e) {
				log( Level.SEVERE, className, "readConfiguration", "cannot find configuration file", e);
			} catch (IOException e) {
				log( Level.SEVERE, className, "readConfiguration", "cannot read configuration file", e);
			}

			if (ins != null ) {
				readConfiguration( ins );
			}
		} else {
			Class classFor = null;

			try {
				classFor = Class.forName(reader);
			} catch (ClassNotFoundException e) {
				log( Level.SEVERE, className, "readConfiguration", "cannot load class " + reader, e);
			}

			if (classFor != null) {
				try {
					Object tmp = classFor.newInstance();
				} catch (InstantiationException e) {
					log( Level.SEVERE, className, "readConfiguration", "cannot create instance of class " + reader, e);
				} catch (IllegalAccessException e) {
					log( Level.SEVERE, className, "readConfiguration", "cannot create instance of class " + reader, e);
				}
			}
		}

	}


	/** Reinitialize the logging properties and
	 * reread the logging configuration from the given stream,
	 * which should be in java.util.Properties format.
	 * A PropertyChangeEvent will be fired after
	 * the properties are read.
	 *
	 * All existing global Handlers will be closed
	 * and a new set of global Handlers will be created
	 * based on the new prop-erties.
	 * Any log level definitions in the new configuration file will
	 * be applied using LogManager.setLevel() in the order they are read.
	 *
	 * @param ins stream to read properties from
	 *
	 * @throws SecurityException
	 * if a security manager exists
	 * and if the caller does not have LoggingPermission("control").
	 */
	public void readConfiguration ( InputStream ins ) throws SecurityException {

		if (0 < handlers.size()) {
			Iterator handlerIterator = handlers.iterator();

			while (handlerIterator.hasNext()) {
				((Handler)handlerIterator.next()).close();
			}

			handlers.clear();
		}

		try {
			properties.load( ins );
		} catch (IOException e) {
			log( Level.SEVERE, className, "readConfiguration", "cannot process configuration", e);
		}

		String sHandlers = getProperty("handlers");
		if (sHandlers != null) {
			StringTokenizer st = new StringTokenizer(sHandlers);

			while (st.hasMoreTokens()) {
				Class classFor = null;

				try {
					classFor = Class.forName(st.nextToken());
				} catch (ClassNotFoundException e) {
					log( Level.SEVERE, className, "readConfiguration", "cannot process configuration", e);
				}

				try {
					 addGlobalHandler((Handler) classFor.newInstance());
				} catch (InstantiationException e) {
					log( Level.SEVERE, className, "readConfiguration", "cannot process configuration", e);
				} catch (IllegalAccessException e) {
					log( Level.SEVERE, className, "readConfiguration", "cannot process configuration", e);
				}
			}
		}

	}


	public void log ( Level level, String className, String sourceMethod, String msg, Throwable thrown ) {
		LogRecord record = new LogRecord(level, msg);
		record.setSourceClassName(className);
		record.setSourceMethodName(sourceMethod);
		record.setThrown(thrown);
		publish(record);
	}
}
