// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.logging;

import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;

import Int.ecb.logging.LogManager;
import Int.ecb.logging.Logger;
import Int.ecb.logging.LogRecord;
import Int.ecb.logging.Level;
import Int.ecb.logging.Handler;
import Int.ecb.logging.TestHandler;

/* ***************************** Class ****************************** */

/**
 * <P>Tests the implementation of handler classes int the Int.ecb.logging
 * package.</P>
 *
 * @author
 * <A HREF="mailto:Sami.Airo@ECB.int">Sami Airo</A>
 *
 * @version @(#)$Id: LogManagerTest.java,v 1.1 2017/01/19 16:14:05 trainot Exp $
 */

public class LogManagerTest extends TestCase {

	/* ========================= Fields ========================= */

	private LogManager _manager;

	/* ====================== Constructors ====================== */

	/**
	 * Constructs a LogManagerTest with the specified name.
	 *
	 * @param name Test case name
	 */
	public LogManagerTest(String name) {
		super(name);
	}

	/* ================== Protected Methods ===================== */

	/**
	 * Sets up the test fixture
	 */
	protected void setUp() {
//		System.setProperty("Int.ecb.logging.config.file", "../gesmescb/test/logging.test.properties");
		System.setProperty("Int.ecb.logging.config.class", "Int.ecb.gesmescb.application.LoggingPropertiesReader");
		_manager = LogManager.getLogManager();
	}

	/**
	 * Tears sown the test fixture
	 */
	protected void tearDown() {
		_manager = null;
	}

	/* ===================== Public Methods ===================== */

	/**
	 * Assembles and returns a test suite for all
	 * the test methods of this test case.
	 *
	 * @return a non-null test suite
	 */
	public static Test suite() {
		TestSuite suite = new TestSuite(LogManagerTest.class);
		return suite;
	}

	/**
	 * Runs the test case
	 */
	public static void main(String args[]) {
		String[] testCaseName = {LogManagerTest.class.getName()};
					 junit.textui.TestRunner.main(testCaseName);
	}

	/* --------------------- Test Methods --------------------- */

	/**
	 * Tests getting a named logger
	 */
	public void testGetLogger() {
		Logger l1 = Logger.getLogger("Test1");
		Logger l2 = Logger.getLogger("Test1");
		Logger l3 = _manager.getLogger("Test1");

		assertTrue("l1 not null", l1 != null);
		assertTrue("l2 not null", l2 != null);
		assertTrue("l3 not null", l3 != null);
		assertSame(l1, l2);
		assertSame(l1, l3);
	}

	/**
	 * Tests adding and removing a global handler
	 */
	public void testAddGetRemoveGlobalhandler () {
		final TestHandler t = new TestHandler();

		_manager.addGlobalHandler( t );
		{
			final Handler[] ha = _manager.getGlobalHandlers();
			boolean found = false;
			for (int i = 0; i < ha.length; i++) {
				if (ha[i] == t) {
					found = true;
				}
			}
			assertTrue( "global handler registered", found );
		}

		_manager.removeGlobalHandler( t );
		{
			final Handler[] ha = _manager.getGlobalHandlers();
			boolean found = false;
			for (int i = 0; i < ha.length; i++) {
				if (ha[i] == t) {
					found = true;
				}
			}
			assertTrue( "global handler not registered", ! found );
		}
	}

	/**
	 * Tests publishing a record to global handlers
	 */
	public void testPublish() {
		LogRecord r = new LogRecord( Level.SEVERE, "test" );
		TestHandler t = new TestHandler();

		_manager.addGlobalHandler(t);
		_manager.publish(r);

		assertSame(r, t.getLastRecord());
	}


	/**
	 * Tests setting the logging level
	 */
	public void testSetLevel() {
		Logger l1 = Logger.getLogger("Test1");
		Logger l2 = Logger.getLogger("Test1.Test2");

		_manager.setLevel("", Level.FINE);

		assertEquals(Level.FINE, l1.getLevel());
		assertEquals(Level.FINE, l2.getLevel());

		_manager.setLevel("Test1", Level.ALL);

		assertEquals("l1 level equals ALL", Level.ALL, l1.getLevel());
		assertEquals("l2 level equals ALL", Level.ALL, l2.getLevel());
	}

	/**
	 * Tests setting the logging level
	 */
	public void testGetLevel() {
		Logger l1 = Logger.getLogger("Test1");
		Logger l2 = Logger.getLogger("Test1.Test2");

		_manager.setLevel("", Level.FINE);

		assertEquals(Level.FINE, l1.getLevel());
		assertEquals(Level.FINE, l2.getLevel());

		_manager.setLevel("Test1", Level.ALL);

		assertEquals("l1 level equals ALL", Level.ALL, l1.getLevel());
		assertEquals("l2 level equals ALL", Level.ALL, l2.getLevel());

		assertEquals("l1 level equals ALL", Level.ALL, _manager.getLevel("Test1"));
		assertEquals("l2 level equals ALL", Level.ALL, _manager.getLevel("Test1.Test2"));

	}
}
