// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.logging;

import java.io.*;

/**
 * Carries a piece of log information.
 *
 * @author
 * <A HREF="mailto:Sami.Airo@ECB.int">Sami Airo</A>
 *
 * @version @(#)$Id: LogRecord.java,v 1.1 2017/01/19 16:14:05 trainot Exp $
 */
public class LogRecord {
	private static long nextSequenceNumber = 1;

	private Level level;
	private java.lang.String msg;
	private long sequenceNumber;
	private long millis;
	private int threadID;
	private Object parameter;
	private Throwable thrown;
	private java.lang.String sourceClassName;
	private java.lang.String sourceMethodName;
	private java.lang.String thrownBackTrace;


	/**
	 * Construct a LogRecord with the given level and message values.
	 */
	public LogRecord(Level level, java.lang.String msg) {
		this.level = level;
		this.msg = msg;
		sequenceNumber = nextSequenceNumber++;
		millis = System.currentTimeMillis();
		threadID = 1;
		//threadID = Thread.currentThread().getName();
	}


	/**
	 * Get event time in milliseconds since 1970.
	 * @return event time in millis since 1970
	 */
	public long getMillis() {
		return millis;
	}


	/**
	 * Get the name of the class that (allegedly) issued the logging request.
	 */
	public java.lang.String getSourceClassName() {
		return sourceClassName;
	}


	/**
	 * Get the name of the method that (allegedly) issued the logging request.
	 */
	public java.lang.String getSourceMethodName() {
		return sourceMethodName;
	}

	/**
	 * Get the Level.
	 */
	public Level getLevel() {
		return level;
	}

	/**
	 * Get the Message.
	 */
	public java.lang.String getMessage() {
		return msg;
	}

	/**
	 * Get the sequence number.
	 */
	public long getSequenceNumber() {
		return sequenceNumber;
	}


	/**
	 * Get the parameters to the log message.
	 * @return the log message parameters. May be null if there are no parameters.
	 */
	public java.lang.Object getParameter() {
		return parameter;
	}

	/**
	 * Get any throwable associated with the log record.
	 */
	public java.lang.Throwable getThrown() {
		return thrown;
	}


	/**
	 * Get a backtrace for any thrown associated with the log record.
	 * If the event involved an exception, this will be the the backtrace from the
	 * exception object. Otherwise it will return null. The exact backtrace format
	 * is implementation specific and will vary beween VMs.
	 */
	public java.lang.String getThrownBackTrace() {
		return thrownBackTrace;
	}


	/**
	 * Set the parameters to the log message.
	 *
	 * @param parameter	the log message parameters.
	 */
	public void setParameter(java.lang.Object parameter) {
		this.parameter = parameter;
	}


	/**
	 * Set the name of the class that (allegedly) issued the logging request.
	 * @param sourceClassName	the source class name
	 */
	public void setSourceClassName(java.lang.String sourceClassName) {
		this.sourceClassName = sourceClassName;
	}


	/**
	 * Set the name of the method that (allegedly) issued the logging request.
	 * @param sourceMethodName the source method name
	 */
	public void setSourceMethodName(java.lang.String sourceMethodName) {
		this.sourceMethodName = sourceMethodName;
	}


	/**
	 * Set a throwable associateded with the log event.
	 * @param thrown stuff
	 */
	public void setThrown ( final Throwable thrown ) {
		this.thrown = thrown;

		java.io.StringWriter sw = new java.io.StringWriter();
		java.io.PrintWriter pw = new java.io.PrintWriter(sw);

		thrown.printStackTrace(pw);

		thrownBackTrace = sw.toString();
	}

	public String toString() {
		return "<" + getClass().getName() + "["
			+ "level=" + level + ", "
			+ "sourceClassName=" + sourceClassName + ", "
			+ "sourceMethodName=" + sourceMethodName + ", "
			+ "message=" + msg + ", "
			+ "sequenceNumber=" + sequenceNumber + ", "
			+ "millis=" + millis + ", "
			+ "Parameter=" + parameter + ", "
			+ "thrown=" + thrown + ", "
			+ "thrownBackTrace=" + thrownBackTrace
			+ "]>";
	}
}
