// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.logging;

import java.net.URL;
import java.net.URLClassLoader;
import java.util.Set;
import java.util.HashSet;
import java.util.Iterator;
import Int.ecb.logging.Level;
import Int.ecb.logging.LogManager;
import Int.ecb.logging.Handler;
import Int.ecb.logging.LogRecord;

/**
 * <P>
 * Receives logging information and forwards it.
 * </P>
 * 
 * @author <A HREF="mailto:Sami.Airo@ECB.int">Sami Airo</A>
 * 
 * @version @(#)$Id: Logger.java,v 1.2 2018/06/19 16:13:49 kortest Exp $
 */
public class Logger {

	private Level level;
	private java.lang.String name;
	private static final LogManager manager = LogManager.getLogManager();
	private Set handlers;
	private boolean useGlobalHandlers = true;
	private static int MAXIMUM_LOG_LINES = 10000;
	private int logLines = 0;

	private Logger(java.lang.String name) {
		this.name = name;
		handlers = new java.util.HashSet();
	}

	protected void finalize() throws Throwable {
		try {
			Iterator handlerIterator = handlers.iterator();

			while (handlerIterator.hasNext()) {
				((Handler) handlerIterator.next()).close();
			}
		} finally {
			super.finalize();
		}
	}

	public static synchronized Logger getLogger(java.lang.String name) {

		Logger logger = (Logger) manager.getLogger(name);

		if (logger == null) {
			logger = new Logger(name);

			logger.level = manager.getLevel(name);

			manager.addLogger(logger);
		}

		return logger;
	}

	public void log(LogRecord record) {
		if (this.level.intValue() <= record.getLevel().intValue()) {
         logLines++;
			if (logLines <= MAXIMUM_LOG_LINES) {
				if (0 < handlers.size()) {
					Iterator handlerIterator = handlers.iterator();

					while (handlerIterator.hasNext()) {
						((Handler) handlerIterator.next()).publish(record);
					}
				}

				if (useGlobalHandlers) {
					manager.publish(record);
				}
			}
		}
	}

	public void log(Level level, java.lang.String sourceClass,
			java.lang.String sourceMethod, java.lang.String msg) {
		if (this.level.intValue() <= level.intValue()) {
      	logLines++;
			if (logLines <= MAXIMUM_LOG_LINES) {
				LogRecord record = new LogRecord(level, msg);
				record.setSourceClassName(sourceClass);
				record.setSourceMethodName(sourceMethod);
				log(record);
			}
		}
	}

	public void log(Level level, java.lang.String sourceClass,
			java.lang.String sourceMethod, java.lang.String msg,
			java.lang.Object param) {
		if (this.level.intValue() <= level.intValue()) {
         logLines++;
			if (logLines <= MAXIMUM_LOG_LINES) {
				LogRecord record = new LogRecord(level, msg);
				record.setSourceClassName(sourceClass);
				record.setSourceMethodName(sourceMethod);
				record.setParameter(param);
				log(record);
			}
		}
	}

	public void log(Level level, java.lang.String sourceClass,
			java.lang.String sourceMethod, java.lang.String msg,
			java.lang.Throwable thrown) {
		if (this.level.intValue() <= level.intValue()) {
      	logLines++;
			if (logLines <= MAXIMUM_LOG_LINES) {
				LogRecord record = new LogRecord(level, msg);
				record.setSourceClassName(sourceClass);
				record.setSourceMethodName(sourceMethod);
				record.setThrown(thrown);
				log(record);
			}
		}

	}

	public void setLevel(Level newLevel) {
		level = newLevel;
	}

	public Level getLevel() {
		return level;
	}

	public java.lang.String getName() {
		return name;
	}

	/**
	 * Add a log Handler to receive logging messages.
	 * 
	 * @param handler
	 *            a logging Handler
	 */
	public void addHandler(Handler handler) {
		handlers.add(handler);
	}

	/**
	 * Check if a message of the given level would actually be logged by this
	 * logger.
	 * 
	 * @param level
	 *            a message logging level
	 * @return true if the given message level is currently being logged.
	 */
	public boolean isLoggable(Level level) {
		return (this.level.intValue() <= level.intValue());
	}
}
