// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.logging;

import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;

import Int.ecb.logging.LogManager;
import Int.ecb.logging.Logger;
import Int.ecb.logging.LogRecord;
import Int.ecb.logging.Level;
import Int.ecb.logging.Handler;
import Int.ecb.logging.TestHandler;

/* ***************************** Class ****************************** */

/**
 * <P>Tests the implementation of {@link Logger}.</P>
 *
 * @author
 * <A HREF="mailto:Sami.Airo@ECB.int">Sami Airo</A>
 *
 * @version @(#)$Id: LoggerTest.java,v 1.1 2017/01/19 16:14:05 trainot Exp $
 */

public class LoggerTest extends TestCase {

	/* ========================= Fields ========================= */

	private LogManager _manager;
	private Logger _logger;
	private TestHandler _handler;

	/* ====================== Constructors ====================== */

	/**
	 * Constructs a LoggerTest with the specified name.
	 *
	 * @param name Test case name
	 */
	public LoggerTest(String name) {
		super(name);
	}

	/* ================== Protected Methods ===================== */

	/**
	 * Sets up the test fixture
	 */
	protected void setUp() {
		System.setProperty("Int.ecb.logging.config.class", "Int.ecb.gesmescb.application.LoggingPropertiesReader");
		_manager = LogManager.getLogManager();
		_logger = Logger.getLogger("Test");
		_handler = new TestHandler();
	}

	/**
	 * Tears sown the test fixture
	 */
	protected void tearDown() {
		_manager = null;
		_logger = null;
		_handler = null;
	}

	/* ===================== Public Methods ===================== */

	/**
	 * Assembles and returns a test suite for all
	 * the test methods of this test case.
	 *
	 * @return a non-null test suite
	 */
	public static Test suite() {
		TestSuite suite = new TestSuite(LoggerTest.class);
		return suite;
	}

	/**
	 * Runs the test case
	 */
	public static void main(String args[]) {
		String[] testCaseName = {LoggerTest.class.getName()};
		junit.textui.TestRunner.main(testCaseName);
	}

	/* --------------------- Test Methods --------------------- */

	/**
	 * Tests setting the log level
	 */
	public void testSetLevel() {
		LogRecord r = new LogRecord( Level.INFO, "test" );

		_logger.addHandler(_handler);

		_logger.setLevel(Level.ALL);
		_logger.log(r);
		assertSame(r, _handler.getLastRecord());
		_handler.close();

		_logger.setLevel(Level.SEVERE);
		_logger.log(r);
		assertNull(_handler.getLastRecord());
	}

	/**
	 * Tests logging methods
	 */
	public void testLogging() {
		LogRecord r = null;
		Level l = Level.INFO;
		String sc = "sourceClass";
		String sm = "sourceMethod";
		String msg = "msg";
		Throwable t = new Throwable();								;

		_logger.addHandler(_handler);
		_logger.setLevel(Level.ALL);

		_logger.log(l, sc, sm, msg);
		r = _handler.getLastRecord();
		assertEquals(l, r.getLevel());
		assertEquals(sc, r.getSourceClassName());
		assertEquals(sm, r.getSourceMethodName());
		assertEquals(msg, r.getMessage());
		_handler.close();

		_logger.log(l, sc, sm, msg, _manager);
		r = _handler.getLastRecord();
		assertEquals(l, r.getLevel());
		assertEquals(sc, r.getSourceClassName());
		assertEquals(sm, r.getSourceMethodName());
		assertEquals(msg, r.getMessage());
		assertSame("parameter", _manager, r.getParameter());
		_handler.close();

		_logger.log(l, sc, sm, msg, t);
		r = _handler.getLastRecord();
		assertEquals(l, r.getLevel());
		assertEquals(sc, r.getSourceClassName());
		assertEquals(sm, r.getSourceMethodName());
		assertEquals(msg, r.getMessage());
		assertSame("thrown", t, r.getThrown());
	}

	/**
	 * Tests isLoggable method
	 */
	public void testIsLoggable() {
		_logger.setLevel(Level.INFO);
		assertTrue(_logger.isLoggable(Level.SEVERE));
		assertTrue(!_logger.isLoggable(Level.FINE));
	}

}
