// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.logging;

import java.io.*;

import Int.ecb.logging.Handler;
import Int.ecb.logging.LogRecord;
import Int.ecb.logging.SimpleFormatter;

/**
 * Stream based logging Handler.
 *
 * This is primarily intended as a base class or support class to be used in
 * implementing other logging Handlers.
 *
 * LogRecords are published to a given java.io.OutputStream.
 *
 * @author
 * <A HREF="mailto:Sami.Airo@ECB.int">Sami Airo</A>
 *
 * @version @(#)$Id: StreamHandler.java,v 1.1 2017/01/19 16:14:05 trainot Exp $
 */
public abstract class StreamHandler extends Int.ecb.logging.Handler {

	private java.io.PrintWriter pw;
	private boolean isFirstRecord;


	/**
	 * Create a StreamHandler, with no current output stream.
	 */
	public StreamHandler() {
		super( new SimpleFormatter() );
		isFirstRecord = true;
	}


	/**
	 * Create a StreamHandler with a given formatter and output stream.
	 *
	 * @param out the target output stream
	 * @param formatter to be used to format output
	 */
	public StreamHandler (
		final OutputStream out,
		final Formatter formatter
	) {
		super( formatter );
		setOutputStream(out);
		isFirstRecord = true;
	}


	/**
	 * Close the current output stream.
	 * The Formatter's "tail" string is written to the stream before it is closed.
	 * In addition, if the formatter's "head" string has not yet been written to
	 * the stream, it will be written before the "tail" string.
	 *
	 * @throws SecurityException
	 * if a security manager exists and if the caller
	 * does not have LoggingPermission("control").
	 */
	public void close() {
		if (pw != null) {
			pw.print(getFormatter().getTail(this));
			pw.close();
		}
	}


	/**
	 * Flush any buffered messages.
	 */
	public void flush() {
		if (pw != null) pw.flush();
	}


	/**
	 * Format and publish a log record.
	 * The StreamHandler first checks if there is an OutputStream and if the given
	 * LogRecord has at least the required log level. If not it silently returns.
	 * If so, it calls any associated Filter to check if the record should be
	 * published. If so, it calls its Formatter to format the record and then
	 * writes the result to the current output stream. If this is the first
	 * LogRecord to be written to a given OutputStream, the Formatter's "head" string
	 * is written to the stream before the LogRecord is written.
	 *
	 * @param record	description of the log event
	 */
	public void publish(LogRecord record) {
		if ((pw != null) & (isLoggable(record))) {
			if (isFirstRecord) {
				pw.print(getFormatter().getHead(this));
				isFirstRecord = false;
			}
			pw.print(getFormatter().format(record));
			// if handler level is set to ALL show also StackTrace
			if (super.getLevel() == Level.ALL &&
					null != record.getThrownBackTrace()) {
				pw.print(record.getThrownBackTrace() 
						+ System.getProperty( "line.separator", "\n" ));
			}
			pw.flush();
		}
	}


	/**
	 * Change the output stream.
	 * If there is a current output stream then the formatter's tail string is
	 * written and the stream is flushed and closed. Then the output stream is
	 * replaced with the new output stream.
	 *
	 * @param	 out			 New output stream. May not be null.
	 * @throws	SecurityException	 if a security manager exists and if the
	 *					 caller does not have LoggingPermission("control").
	 */
	protected void setOutputStream(java.io.OutputStream out) {
		if (this.pw != null) {
			flush();
			close();
		}

		pw = new java.io.PrintWriter(out);
		isFirstRecord = true;
	}
}
