// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.prefs;

import java.util.Properties;
import java.util.Map;
import java.util.HashMap;
import java.util.StringTokenizer;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;

/* ***************************** Class ****************************** */

/**
 * <P>Implements subset of forthcoming java.util.prefs.Preferences class.</P>

 * <P>All preferences are stored in a single file in Java properties format.
 * The file name is read from the system property Int.ecb.prefs.config.file
 * (Start Java with option -DInt.ecb.prefs.config.file=filename).
 * If this is not defined then file "default.properties" in the current path is used.</P>
 *
 * @author
 * <A HREF="mailto:Sami.Airo@ECB.int">Sami Airo</A>
 * @author
 * <A HREF="mailto:Bernhard.Bodenstorfer@ECB.int">Bernhard Bodenstorfer</A>
 *
 * @version @(#)$Id: Preferences.java,v 1.1 2017/01/19 16:14:05 trainot Exp $
 */

public class Preferences {

	/* ========================= Fields ========================= */

	/* --------------------- Static Fields ---------------------- */

	/**
	 * Global properties.
	 */
	private static final Properties properties = new Properties();

	/**
	 * Mapping of user paths to nodes.
	 */
	private static final Map userPreferencesMap = new HashMap();

	/**
	 * Mapping of system paths to nodes.
	 */
	private static final Map systemPreferencesMap = new HashMap();

	/**
	 * The root node user preferences.
	 */
	private static final Preferences userRoot = new Preferences();

	/**
	 * The root node system preferences.
	 */
	private static final Preferences systemRoot = new Preferences();

	/* -------------------- Instance Fields --------------------- */

	/**
	 * Absolute path name for the node.
	 */
	private String absolutePathName;

	/**
	 * Parent node.
	 */
	private Preferences parent;

	/**
	 * Relative path name for the node.
	 */
	private String relativePathName;

	/**
	 * Package name for the node.
	 */
	private String packageName;

	/**
	 * True if this node is a system node; false otherwise.
	 */
	private boolean isSystem;

	/* ================= Initialization Blocks ================== */

	/* -------------- Static Initialization Blocks -------------- */

	static {
		reload();

		userRoot.absolutePathName = "/";
		userRoot.relativePathName = null;
		userRoot.packageName = "";
		userRoot.parent = null;
		userRoot.isSystem = false;

		systemRoot.absolutePathName = "/";
		systemRoot.relativePathName = null;
		systemRoot.packageName = "";
		systemRoot.parent = null;
		systemRoot.isSystem = true;

		userPreferencesMap.put( userRoot.absolutePathName, userRoot );
		systemPreferencesMap.put( systemRoot.absolutePathName, systemRoot );
	}

	/* ====================== Constructors ====================== */

	/**
	 * Sole constructor.
	 * (For invocation by subclass constructors, typically implicit.)
	 */
	protected Preferences() {
	}

	/* ===================== Public Methods ===================== */

	/**
	 * Returns the preference node from the calling user's preference tree that is associated
	 * (by convention) with the specified object's package.
	 *
	 * @return the user preference node associated with the package in which o's class is declared.
	 */
	public static Preferences userNodeForPackage(java.lang.Object o) {
		String path = (o.getClass().getPackage() == null ? "<unnamed>" : o.getClass().getPackage().getName());
		return userRoot.node(path.replace('.', '/'));
	}

	/**
	 * Returns the preference node from the system preference tree that is associated
	 * (by convention) with the specified object's package.
	 *
	 * @return the system preference node associated with the package in which o's class is declared.
	 */
	public static Preferences systemNodeForPackage(java.lang.Object o) {
		String path = (o.getClass().getPackage() == null ? "<unnamed>" : o.getClass().getPackage().getName());
		return systemRoot.node(path.replace('.', '/'));
	}

	/**
	 * Returns the value associated with the specified key in this preference node.
	 * Returns the specified default if there is no value associated with the key, or the backing store is
	 * inaccessible.
	 *
	 * @param key	whose associated value is to be returned.
	 * @param def	to be returned in the event
	 * that this preference node has no value
	 * associated with key.
	 *
	 * @return the value associated with key,
	 * or def if no value is associated with key, or the
	 * backing store is inaccessible.
	 */
	public java.lang.String get(java.lang.String key, java.lang.String def) {
		String prop = null;

		if (isSystem) {
			prop = System.getProperty(packageName + "." + key);
		} else {
			prop = properties.getProperty(packageName + "." + key);
		}

		if (prop == null) {
			if (parent != null) {
				return parent.get(key, def);
			} else {
				return def;
			}
		} else {
			return prop;
		}

	}

	/**
	 * Returns the parent of this preference node, or null if this is the root.
	 *
	 * @return the parent of this preference node.
	 */
	public Preferences parent() {
		return parent;
	}

	/**
	 * Returns the int value represented by the string associated with the specified key in this
	 * preference node.
	 *
	 * @param	key	key whose associated value is to be returned as an int.
	 * @param	def	the value to be returned in the event that this preference node has no value
	 *					 associated with key or the associated value cannot be interpreted as an int, or the
	 *					 backing store is inaccessible.
	 *
	 * @return the int value represented
	 * by the string associated with key in this preference node,
	 * or def if the associated value does not exist or
	 * cannot be interpreted as an int.
	 */
	public int getInt(java.lang.String key, int def) {

		String val = get(key, null);

		if (val != null) {
			try {
				return Integer.parseInt( val );
			} catch (NumberFormatException e) {
				return def;
			}
		} else {
			return def;
		}

	}

	/**
	 * <P>Returns the named preference node in the same tree as this node, creating it and any of
	 * its ancestors if they do not already exist.Accepts a relative or absolute path name.
	 * Relative path names (which do not begin with the slash character ('/')) are
	 * interpreted relative to this preference node.<P/>
	 *
	 * @param pathName	the path name of the preference node to return.
	 *
	 * @return the specified preference node
	 */
	public Preferences node(String pathName) {

		String path = (pathName.charAt(0) == '/' ? pathName : absolutePathName + pathName);

		Preferences pref = null;
		if (isSystem) {
			pref = (Preferences)systemPreferencesMap.get(path);
		} else {
			pref = (Preferences)userPreferencesMap.get(path);
		}

		if (pref == null) {
			pref = new Preferences();

			pref.absolutePathName = path;
			pref.relativePathName = null;
			pref.packageName = path.replace('/','.').substring(1);
			pref.isSystem = isSystem;

			if (isSystem) {
				systemPreferencesMap.put(pref.absolutePathName, pref);
			} else {
				userPreferencesMap.put(pref.absolutePathName, pref);
			}

			if (pref.absolutePathName.equals("/")) {
				pref.parent = null;
			} else {
				pref.parent = node( pref.absolutePathName.substring(0,
					(pref.absolutePathName.lastIndexOf("/") == 0 ? 1 : pref.absolutePathName.lastIndexOf("/") )) );
			}
		}

		return pref;
	}

	/**
	 * Returns the root preference node for the calling user.
	 *
	 * @return the root preference node for the calling user.
	 */
	public static Preferences userRoot() {
		return userRoot;
	}

	/**
	 * Returns the root preference node for the system.
	 *
	 * @return the root preference node for the system.
	 */
	public static Preferences systemRoot() {
		return systemRoot;
	}

	/** Reload the preferences from the location specified in
	 * system property <code>Int.ecb.prefs.config.file</code>.
	 */
	public static void reload () {
		final String filename = System.getProperty(
			"Int.ecb.prefs.config.file", "default.properties"
		);
		clear();
		load( filename );
	}

	/** Load additional preferences from the location specified. */
	public static void load ( final String filename ) {
		try {
			final FileInputStream fileStream = new FileInputStream( filename );
			try {
				properties.load(fileStream);
			} finally {
				// close on read errors
				fileStream.close();
			}
		} catch ( FileNotFoundException e ) {
			// do nothing if we get an error in reading from file -> use defaults
		} catch ( final IOException e ) {
			// do nothing if we get an error in reading from file -> use defaults
		}
	}

	/** Reset all loaded preferences. */
	public static void clear () {
		properties.clear();
	}
}
