// ECB/SIS Public License, version 1.0, document reference SIS/2001/116
//
// Copyright (c) 2001, European Central Bank. All rights reserved.
//
// Redistribution and use in source and binary forms,
// with or without modification, are permitted
// provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// Neither the name of the European Central Bank
// nor the names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package Int.ecb.prefs;

import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;

import Int.ecb.prefs.Preferences;


/* ***************************** Class ****************************** */

/**
 * <P>Tests the implementation of the Int.ecb.prefs.Preferences class.</P>
 *
 * <P>Depends on the preferences file default.properties to be in the current
 * path.</P>
 *
 * @author
 * <A HREF="mailto:Sami.Airo@ECB.int">Sami Airo</A>
 *
 * @version @(#)$Id: PreferencesTest.java,v 1.1 2017/01/19 16:14:05 trainot Exp $
 */

public class PreferencesTest extends TestCase {

	/* ========================= Fields ========================= */

	private Preferences _user_pref;
	private Preferences _system_pref;

	/* ====================== Constructors ====================== */

	/**
	 * Constructs a PreferencesTes with the specified name.
	 *
	 * @param name Test case name
	 */
	public PreferencesTest(String name) {
		super(name);
	}

	/* ================== Protected Methods ===================== */

	/**
	 * Sets up the test fixture
	 */
	protected void setUp() {
		System.setProperty("Int.ecb.prefs.config.file", "../gesmescb/test/default.test.properties");
		_user_pref = Preferences.userNodeForPackage(this);
		_system_pref = Preferences.systemNodeForPackage(this);
	}

	/**
	 * Tears sown the test fixture
	 */
	protected void tearDown() {
		_user_pref = null;
		_system_pref = null;
	}

	/* ===================== Public Methods ===================== */

	/**
	 * Assembles and returns a test suite for all
	 * the test methods of this test case.
	 *
	 * @return a non-null test suite
	 */
	public static Test suite() {
		TestSuite suite = new TestSuite(PreferencesTest.class);
		return suite;
	}

	/**
	 * Runs the test case
	 */
	public static void main(String args[]) {
		String[] testCaseName = {PreferencesTest.class.getName()};
		junit.textui.TestRunner.main(testCaseName);
	}

	/* --------------------- Test Methods --------------------- */

	/**
	 * Tests getting a defined string property
	 */
	public void testGetDefinedString() {
		String t = _user_pref.get("TestString", "not found");

		assertEquals("TestString", t);
	}

	/**
	 * Tests getting a defined string property
	 */
	public void testGetDefinedStringSystem() {
		String test = "system test string";
		System.setProperty("<unnamed>.SystemTestString", test);
		String t = _system_pref.get("SystemTestString", "not found");

		assertEquals(test, t);
	}

	/**
	 * Tests getting a not defined string property
	 */
	public void testGetUndefinedString() {
		String def = "not found";
		String t = _user_pref.get("UndefinedString", def);

		assertEquals(def, t);
	}

	/**
	 * Tests getting a defined int property
	 */
	public void testGetDefinedInt() {
		int t = _user_pref.getInt("TestInt", 9999);

		assertEquals(100, t);
	}

	/**
	 * Tests getting a undefined int property
	 */
	public void testGetUndefinedInt() {
		int def = 9999;
		int t = _user_pref.getInt("TestInt2", def);

		assertEquals(def, t);
	}

	/**
	 * Tests getting a int property that is not defined as int
	 */
	public void testGetStringAsInt() {
		int def = 9999;
		int t = _user_pref.getInt("TestNotInt", def);

		assertEquals(def, t);

		String s = _user_pref.get("TestNotInt", "not found");
		assertEquals("TestNotInt", s);
	}

	/**
	 * Tests getting and using the root node
	 */
	public void testRootNode() {
		Preferences root = Preferences.userRoot();

		assertNotNull(root);
		assertEquals(100, root.getInt("TestInt", 9999));
		assertEquals(9999, root.getInt("TestInt2", 9999));
		assertEquals("TestString", root.get("TestString", "not found"));
		assertEquals("not found", root.get("TestString2", "not found"));
	}

	/**
	 * Tests getting and using a absolutely named node
	 */
	public void testAbsoluteNode() {
		String nodePath = "/test1/test2";
		Preferences p2 = Preferences.userRoot().node(nodePath);

		assertNotNull(p2);
		assertEquals(100, p2.getInt("TestInt", 9999));
		assertEquals(9999, p2.getInt("TestInt2", 9999));
		assertEquals("TestString", p2.get("TestString", "not found"));
		assertEquals("not found", p2.get("TestString2", "not found"));
	}

	/**
	 * Tests getting and using a relatively named node
	 */
	public void testRelativeNode() {
		String nodePath = "test1/test2";
		Preferences p2 = Preferences.userRoot().node(nodePath);

		assertNotNull(p2);
		assertEquals(100, p2.getInt("TestInt", 9999));
		assertEquals(9999, p2.getInt("TestInt2", 9999));
		assertEquals("TestString", p2.get("TestString", "not found"));
		assertEquals("not found", p2.get("TestString2", "not found"));
	}

	/**
	 * Tests getting a preference set at only at the parent level
	 */
	public void testGetParentString() {
		String nodePath = "/test1/test2";
		Preferences p2 = Preferences.userRoot().node(nodePath);

		assertNotNull(p2);
		assertEquals("RootTestString", p2.get("RootTestString", "not found"));
	}

	/**
	 * Tests getting the parents of a node
	 */
	public void testGetParent() {
		String nodePath = "/test1/test2";
		Preferences p2 = Preferences.userRoot().node(nodePath);

		assertNotNull(p2);
		assertEquals("RootTestString", p2.get("RootTestString", "not found"));
	}
}
